<?php

namespace Console;

use Wildix\DependencyInjection\Container;
use Wildix\InvalidDataException;

abstract class BaseWorker {

    const PARAM_CREATED     = 'created';
    const PARAM_RETRY_COUNT = 'retry_count';

    /**
     * Dependency injection container
     *
     * @var Container
     */
    private $_container;

    /**
     * Expire seconds for task.
     * Default: 1 day
     *
     * @var integer
     */
    protected $_expireSeconds = 86400 * 1;

    /**
     * Limit retry count
     * Default: 3 times
     *
     * @var integer
     */
    protected $_limitRetryCount = 3;

    /**
     * Array of worker params.
     *
     * @var array
     */
    protected $_params = [];

    /**
     * Logger
     *
     * @var \Logger\Log
     */
    protected $_logger;

    /**
     * Main task processing logic
     *
     * @return void
     */
    abstract public function run() : void;

    /**
     * Need return array of required params
     *
     * @return array
     */
    abstract protected function getRequiredParams() : array;

    /**
     * Worker constructor.
     *
     * @param array     $params      Worker params.
     * @param Container $diContainer Dependency injection container.
     *
     * @throws InvalidDataException Invalid item.
     */
    public function __construct(array $params, Container $diContainer) {
        $this->_params    = $params;
        $this->_container = $diContainer;
        $this->_logger    = $diContainer->get(\Logger\Log::class);

        $this->checkTask();
    }

    /**
     * Check valid task for required params, expired date and retry count limit.
     *
     * @throws InvalidDataException Invalid data.
     *
     * @return void
     */
    private function checkTask() : void {
        if (!$this->checkParams()) {
            throw new InvalidDataException('Required param missed.');
        }

        if (isset($this->_params[self::PARAM_CREATED]) && $this->isExpired($this->_params[self::PARAM_CREATED])) {
            throw new InvalidDataException('Expired item.');
        }

        if (isset($this->_params[self::PARAM_RETRY_COUNT]) && $this->isRetryCountLimitReached($this->_params[self::PARAM_RETRY_COUNT])) {
            throw new InvalidDataException('Retry count limit reached.');
        }
    }

    /**
     * Check existing requirement params
     *
     * @return boolean
     */
    private function checkParams() {
        foreach ($this->getRequiredParams() as $paramName) {
            if (!isset($this->_params[$paramName])) {
                return false;
            }
        }
        return true;
    }

    /**
     * Check item expired date
     *
     * @param integer $createdAt Created date.
     *
     * @return boolean
     */
    private function isExpired($createdAt) : bool {
        return time() > $createdAt + $this->_expireSeconds;
    }

    /**
     * Check item retry count limit
     *
     * @param integer $retryCount Retry count.
     *
     * @return boolean
     */
    private function isRetryCountLimitReached($retryCount) {
        return $retryCount >= $this->_limitRetryCount;
    }

    /**
     * Return dependency container
     *
     * @param string $service Service name from container.
     *
     * @return mixed Service
     */
    protected function getServiceFromContainer(string $service) {
        return $this->_container->get($service);
    }
}
