<?php
namespace Console;

use Wildix\DependencyInjection\Container;
use Wildix\InvalidDataException;

if (PHP_SAPI !== 'cli') {
    die('only for cli');
}

require_once '/var/www/api/init.php';

/**
 * Class Runner
 */
class Runner {

    /**
     * Array mapping for queue task type and worker
     *
     * @var array
     */
    private static $_typeWorkerMapper = [
        'deviceSync' => DeviceSync::class,
        'office365PhonebookSync' => Office365PhonebookSync::class,
        'office365PhonebookCleaner' => Office365PhonebookCleaner::class
    ];

    /**
     * Runner constructor.
     *
     * @param array     $params      Console command params.
     * @param Container $diContainer Dependency injection container.
     */
    public function __construct(array $params, Container $diContainer) {
        try {
            $workerClass = self::getWorkerByType($this->getTypeFromParams($params));
            $worker      = new $workerClass($this->prepareParamsForWorker($params), $diContainer);
            $worker->run();
        } catch (InvalidDataException $e) {
            $logger = $diContainer->get(\Logger\Log::class);
            $logger->error($e->getLine(), 'Queue manager: [ ERROR ] ' . $e->getMessage() . ' Params: ' . json_encode($params));
        }
    }

    /**
     * Return task type (first param)
     *
     * @param array $params Array of all params.
     *
     * @return string Type of queue task
     */
    private function getTypeFromParams(array $params) : string {
        return $params[1] ?? '';
    }

    /**
     * Return worker class by task type
     *
     * @param string $type Type of queue task.
     *
     * @return string Worker class name.
     *
     * @throws InvalidDataException Invalid or empty type.
     */
    private static function getWorkerByType(string $type) : string {
        if ($type === '') {
            throw new InvalidDataException('Empty type.');
        }

        if (!isset(self::$_typeWorkerMapper[$type])) {
            throw new InvalidDataException('Unknown task type.');
        }
        return self::$_typeWorkerMapper[$type];
    }

    /**
     * Return params for worker from all of the params
     *
     * @param array $params Array of all params.
     *
     * @return array Array params only for worker.
     */
    private function prepareParamsForWorker(array $params) : array {
        $params = array_slice($params, 2);

        $result = [];

        foreach ($params as $param) {
            $divided             = explode('=', $param);
            $result[$divided[0]] = $divided[1];
        }
        return $result;
    }
}

$diContainer = new Container();

$runner = new Runner($_SERVER['argv'], $diContainer);
