<?php
namespace Controllers;

use Models\PBX;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Slim\Exception\HttpNotFoundException;
use Wildix\CallFiles;
use Wildix\CallweaverManager;
use Wildix\Error;
use Wildix\Exception;
use Wildix\Sms\Exception\SmsViaGsmQueueException;
use Wildix\Sms\Factory\SmsViaGsmDtoFactory;
use Wildix\Sms\SmsViaGsmService;
use Wildix\Validator;

class Astman extends Controller {

    public function postOriginate(Request $request, Response $response, array $args): Response {
        $params = $request->getParsedBody();

        $rc = array("\n", "\r");

        $channel     = str_replace($rc, '', $params['channel'] ?? null);
        $exten       = str_replace($rc, '', $params['exten'] ?? null);
        $context     = str_replace($rc, '', $params['context'] ?? null);
        $priority    = str_replace($rc, '', $params['priority'] ?? null);
        $application = str_replace($rc, '', $params['application'] ?? null);
        $data        = str_replace($rc, '', $params['data'] ?? null);
        $timeout     = str_replace($rc, '', $params['timeout'] ?? null);
        $callerid    = str_replace($rc, '', $params['callerid'] ?? null);
        $variable    = str_replace($rc, '', $params['variable'] ?? null);
        $account     = str_replace($rc, '', $params['account'] ?? null);
        $async       = str_replace($rc, '', $params['async'] ?? null);
        $actionid    = str_replace($rc, '', $params['actionid'] ?? null);

        $async = $async == 'true';

        $result = $this->getCallWeaverManager()->originate(
            $channel,
            $exten,
            $context,
            $priority,
            $application,
            $data,
            $timeout,
            $callerid,
            $variable,
            $account,
            $async,
            $actionid
        );

        if ($result !== true) {
            throw new Exception('Internal server error');
        }

        return $this->success($response);
    }

    public function postPlayback(Request $request, Response $response, array $args): Response {
        $params = $request->getParsedBody();
        $rc     = array("\n", "\r");
        $result = false;
        $data   = str_replace($rc, '', $params['path'] ?? '');
        $async  = str_replace($rc, '', $params['async'] ?? '');

        if ($data) {
            $channel     = 'Local/' . $this->getUser()->getExtension() . '@pbxsound';
            $application = 'Playback';
            $callerid    = '"Playback" <' . substr($data, strrpos($data, '/') + 1) . '>';
            $async       = $async == 'true';

            $result = $this->getCallWeaverManager()->originate(
                $channel,
                null,
                null,
                null,
                $application,
                $data,
                null,
                $callerid,
                null,
                null,
                $async,
                null
            );
        }

        if ($result !== true) {
            throw new Exception('Internal server error');
        }

        return $this->success($response);
    }

    public function postRingtone(Request $request, Response $response, array $args): Response {
        $params = $request->getParsedBody();
        $rc     = array("\n", "\r");
        $result = false;
        $data   = str_replace($rc, '', $params['path'] ?? '');
        $async  = str_replace($rc, '', $params['async'] ?? '');

        if ($data) {
            $channel     = 'Local/ringtone@pbxservices';
            $application = 'Hangup';
            $callerid    = '"Ringtone" <' . substr($data, strrpos($data, '/') + 1) . '>';
            $variable    = 'RINGTONEUSER=' . $this->getUser()->getExtension() . '|RINGTONE=' . $data;
            $async       = $async == 'true';

            $result = $this->getCallWeaverManager()->originate(
                $channel,
                null,
                null,
                null,
                $application,
                null,
                null,
                $callerid,
                $variable,
                null,
                $async,
                null
            );
        }
        if ($result !== true) {
            throw new Exception('Internal server error');
        }

        return $this->success($response);
    }

    public function deleteCall(Request $request, Response $response, array $args): Response {
        $type   = $args['type'] ?? '';
        $callId = $args['id'] ?? '';

        if (strtolower($type) != 'alarmclocks' || !$this->getCallFiles()->removeCallById($type, $callId)) {
            throw new Exception('Internal server error');
        }

        return $this->success($response);
    }

    public function getCall(Request $request, Response $response, array $args): Response {
        $type   = $args['type'] ?? '';
        $callId = $args['id'] ?? '';

        if (strtolower($type) != 'alarmclocks') {
            throw new Exception('Internal server error');
        }

        $call = $this->getCallFiles()->getCallById($type, $callId);
        if (!$call) {
            throw new Exception($type . ' "' . $callId . '" not found', 404);
        }

        return $this->result($call, $response);
    }

    public function postMobilityCall(Request $request, Response $response, array $args): Response {
        $params            = $request->getParsedBody();
        $user              = $this->getUser();
        $number            = trim($params['number'] ?? '');
        $extension         = $user->getExtension();
        $mobilityNumber    = $user->getMobilePhone();
        $context           = $user->getDialplan();
        $userGroupDn       = $user->getGroupDn();
        $preferredLanguage = $user->getLanguage();

        $number = preg_replace('/[^*+#A-Z0-9]/i', '', $number);

        if (empty($extension)) {
            throw new Exception('User extension is empty', 400);
        }
        if (empty($number)) {
            throw new Exception('Destination number is empty', 400);
        }

        if (empty($mobilityNumber)) {
            throw new Exception('Mobile number cannot be empty', 400);
        }

        if (empty($context)) {
            throw new Exception('User dialplan is empty', 404);
        }

        $result = $this->getCallFiles()->makeMobilityCall($extension, $mobilityNumber, $userGroupDn, $number, $context, $preferredLanguage);
        if ($result !== true) {
            throw new Exception('Internal server error');
        }

        return $this->success($response);
    }

    private function externalPrefix($number) {
        // which do not start with + and are longer than 7 digits and no have conference feature code
        if ($number[0] != '+' && strlen($number) > 6) {
            if (!str_starts_with($number, $this->getPbx()->getFeatureAccessCode('CONFERENCE'))) {
                $externalLinePrefix = $this->getPbx()->getExternalLinePrefix();
                if ($externalLinePrefix !== null) {
                    $number = $externalLinePrefix . $number;
                }
            }
        }

        return $number;
    }

    public function postCall(Request $request, Response $response, array $args): Response {
        $params    = $request->getParsedBody();
        $user      = $this->getUser();
        $userPhone = $user->getExtension();
        $number    = trim($params['number'] ?? '');
        $name      = trim($params['name'] ?? '');
        $postpone  = trim($params['postpone'] ?? '');

        $pos = strpos($number, '(0)');
        if ($pos !== false && $pos > 0) {
            $number = str_replace('(0)', '', $number);
        }
        $number = preg_replace('/[^*+#A-Z0-9]/i', '', $number);

        if (empty($number)) {
            throw new Exception('Destination number is empty', 400);
        }

        if (empty($name)) {
            $name = $number;
        }

        $number = $this->externalPrefix($number);

        if (!empty($postpone)) {
            // 14/07/2014 15:35
            $postpone = \DateTime::createFromFormat('d/m/Y H:i', $postpone);
            if ($postpone === false) {
                throw new Exception('Bad format postponed date', 400);
            } else {
                $nowDate = new \DateTime();
                if ($postpone->getTimestamp() <= $nowDate->getTimestamp()) {
                    throw new Exception('Postponed date must be greater than the current', 400);
                }
            }
        } else {
            $postpone = null;
        }

        $result = $this->getCallFiles()->makeCall($userPhone, $number, $name, $number, $postpone);
        if ($result !== true) {
            throw new Exception('Internal server error');
        }

        return $this->success($response);
    }

    public function startCall(Request $request, Response $response, array $args): Response {
        $extension = $args['extension'] ?? '';

        $params = $request->getParsedBody();

        $toNumber = $this->_normalizeDestinationNumber(trim($params['number'] ?? ''));

        if (empty($toNumber)) {
            $this->getLogger()->warning(__LINE__, 'StartCall: Destination number is empty');
            throw new Exception('Destination number is empty', 400);
        }

        $registrations = $this->getSipProxy()->getUsersRegistrations([$extension]);

        if (!count($registrations)) {
            $this->getLogger()->warning(__LINE__, 'StartCall: User do not have any active device');
            throw new Exception(Error::ActiveDeviceNotFound, 400);
        }

        $uaId = $this->getActiveDeviceUaId($extension, $registrations);
        $user = $this->getColleagues()->getUserByExtension($extension);

        try {
            $result = $this->_startUserCallFromDevice($user, $uaId, $toNumber);
        } catch (\Exception $e) {
            $errorMessage = sprintf('Start call error: [%s] %s', $e->getCode(), $e->getMessage());
            throw new Exception($errorMessage);
        }

        return $this->result($result, $response);
    }

    public function endCall(Request $request, Response $response, array $args): Response {
        $channel = $args['channel'] ?? '';

        $result = $this->getCallWeaverManager()->hangup($channel);

        return $this->result($result, $response);
    }

    public function answerCall(Request $request, Response $response, array $args): Response {
        $channel = $args['channel'] ?? '';
        $params  = $request->getParsedBody();

        $extension = (string)$params['extension'] ?? '';

        if ($extension === '') {
            throw new Exception('Extension is empty', 400);
        }

        $registrations = $this->getSipProxy()->getUsersRegistrations([$extension]);
        $uaId          = $this->getActiveDeviceUaId($extension, $registrations);

        if ($uaId === 'any_device') {
            throw new HttpNotFoundException($request, Error::ActiveDeviceNotFound);
        }

        $this->getLogger()->info(__LINE__, sprintf('Answer call. For Extension - %s | For UAID - %s', $extension, $uaId));
        try {
            $result = $this->getCallWeaverManager()->answer($channel, $uaId);
        } catch (\Exception $e) {
            throw new Exception(sprintf('Answer call error. [%s] %s', $e->getCode(), $e->getMessage()));
        }

        return $this->result($result, $response);
    }

    public function postAlarmClock(Request $request, Response $response, array $args): Response {
        $callId = $args['id'] ?? null;
        $user   = $this->getUser();

        $params   = $request->getParsedBody();
        $number   = trim($params['number'] ?? '');
        $postpone = trim($params['postpone'] ?? '');
        $dialplan = trim($params['dialplan'] ?? '');

        if (empty($number)) {
            throw new Exception('Destination number is empty', 400);
        }

        if (!empty($postpone)) {
            // 14/07/2014 15:35
            $postpone = \DateTime::createFromFormat('d/m/Y H:i', $postpone);
            if ($postpone === false) {
                throw new Exception('Bad format postponed date', 400);
            } else {
                // remove the load if many alarms at the exact same moment
                $seconds = rand(0, 30);
                $postpone->setTimestamp($postpone->getTimestamp() + $seconds);

                $nowDate = new \DateTime();
                if ($postpone->getTimestamp() <= $nowDate->getTimestamp()) {
                    throw new Exception('Postponed date must be greater than the current', 400);
                }
            }
        } else {
            $postpone = null;
        }

        if (empty($dialplan)) {
            $dialplan = 'alarmclock';
        }

        $result = $this->getCallFiles()->makeAlarmClock($user, $number, $dialplan, $postpone, $callId);
        if ($result !== true) {
            throw new Exception('Internal server error');
        }

        return $this->result($result, $response);
    }

    public function editAlarmClock(Request $request, Response $response, array $args): Response {
        $callId = $args['id'] ?? null;

        if (!$this->getCallFiles()->removeCallById('AlarmClocks', $callId)) {
            throw new Exception('Internal server error');
        }

        return $this->postAlarmClock($request, $response, $args);
    }

    public function getAlarmClocks(Request $request, Response $response, array $args): Response {
        $result = $this->getCallFiles()->getAlarmClocks();

        return $this->result(
            [
                'total'   => count($result),
                'records' => $result
            ],
            $response
        );
    }

    public function deleteAlarmClock(Request $request, Response $response, array $args): Response {
        $callId = $args['id'] ?? null;

        if (!$this->getCallFiles()->removeCallById('alarmclocks', $callId)) {
            throw new Exception('Internal server error');
        }

        return $this->success($response);
    }

    public function sendSms(Request $request, Response $response, array $args): Response {
        $smsViaGsmService    = $this->getContainer()->get(SmsViaGsmService::class);
        $smsViaGsmDtoFactory = $this->getContainer()->get(SmsViaGsmDtoFactory::class);
        try {
            $smsViaGsmDto = $smsViaGsmDtoFactory->createFromRequest($request);
            $smsViaGsmService->sendSms($smsViaGsmDto);
        } catch (SmsViaGsmQueueException $e) {
            return $this->warning($e->getMessage(), $response, $e->getFailedNumbers());
        } catch (\Throwable $e) {
            throw new Exception($e->getMessage(), (int)$e->getCode() ?: 400);
        }

        return $this->success($response);
    }

    public function getCallStatus(Request $request, Response $response, array $args): Response {
        $channel = $args['channel'] ?? '';
        $result  = $this->getCallWeaverManager()->getCallStatus($channel);

        return $this->result($result, $response);
    }

    public function sendCall(Request $request, Response $response, array $args): Response {
        $params = $request->getQueryParams();
        $user   = $this->getUser();
        $to     = $this->_normalizeDestinationNumber(trim($params['number'] ?? ''));
        $name   = $params['name'] ?? null;
        $uaid   = trim($params['device'] ?? '');

        if (empty($uaid)) {
            $uaid = 'any_device';
        }

        if (empty($to)) {
            throw new Exception('Destination number is empty', 400);
        }

        $result = $this->_startUserCallFromDevice($user, $uaid, $to, $name);

        return $this->result($result, $response);
    }

    public function editTags(Request $request, Response $response, array $args): Response {
        $channel = $args['channel'] ?? '';

        $params = $request->getParsedBody();
        $tags   = trim($params['tags'] ?? '');
        $result = $this->getCallWeaverManager()->setTags($channel, $tags);

        if ($result !== true) {
            throw new Exception('Internal server error');
        }

        return $this->success($response);
    }

    public function getTags(Request $request, Response $response, array $args): Response {
        $channel = $args['channel'] ?? '';

        $result = $this->getCallWeaverManager()->getTags($channel);

        return $this->result($result, $response);
    }

    public function getRecordings(Request $request, Response $response, array $args): Response {
        $channel = $args['channel'] ?? '';
        $records = $this->getCallWeaverManager()->getRecordings($channel);

        return $this->result(['total' => count($records), 'records' => $records], $response);
    }

    public function postTransfer(Request $request, Response $response, array $args): Response {
        $channel = $args['channel'] ?? '';

        $user = $this->getUser();

        $params = $request->getParsedBody();

        $type     = strtolower(trim($params['type'] ?? ''));
        $to       = trim($params['to'] ?? '');
        $context  = trim($params['context'] ?? '');
        $priority = intval(trim($params['priority'] ?? '1'));

        if (empty($type) || !in_array($type, ['redirect', 'bridge'])) {
            $type = 'redirect';
        }

        if (empty($to)) {
            throw new Exception('Destination channel is empty', 400);
        }

        if ($type == 'redirect') {
            if (empty($context)) {
                $context = $user->getDialplan();
            }
        }

        if ($type == 'redirect') {
            $result = $this->getCallWeaverManager()->redirect($channel, '', $to, $context, $priority);
        } else {
            $result = $this->getCallWeaverManager()->bridge($channel, $to);
        }

        return $this->result($result, $response);
    }

    public function sipUpdate(Request $request, Response $response, array $args): Response {
        $channel = $args['channel'] ?? '';
        $params  = $request->getParsedBody();

        $number = trim($params['number'] ?? '');
        $name   = trim($params['name'] ?? '');
        $data   = [];

        if ($number) {
            $data['ConnectedLineNum'] = $number;
        }

        if ($name) {
            $data['ConnectedLineName'] = $name;
        }

        $result = $this->getCallWeaverManager()->sipUpdate($channel, $data);

        return $this->result($result, $response);
    }

    public function playDtmf(Request $request, Response $response, array $args): Response {
        $channel = $args['channel'] ?? '';
        $params  = $request->getParsedBody();

        $digit   = $params['dtmf'] ?? '';
        $bridged = Validator::isTrue($params['sendOnBridgedChannel'] ?? true);

        $validator = new Validator();
        $validator->rule('dtmf', 'digit');
        $validator->rule('required', 'digit');

        $validateErrors = $validator->validateData(['digit' => $digit], ['digit']);
        if ($validateErrors) {
            throw new Exception(Error::getMessage(Error::InvalidDtmfCode), 400);
        }

        $result = $this->getCallWeaverManager()->playDtmf($channel, $digit, $bridged);

        return $this->result($result, $response);
    }

    private function _normalizeDestinationNumber($number) {
        $pos = strpos($number, '(0)');

        if ($pos !== false && $pos > 0) {
            $number = str_replace('(0)', '', $number);
        }

        $number = preg_replace('/[^*+#A-Z0-9]/i', '', $number);

        return $this->externalPrefix($number);
    }

    /**
     * Initiate call from a concrete user device.
     *
     * @param \Wildix\Colleagues\Colleague $user     Device owner.
     * @param string                       $uaId     Device identifier.
     * @param string                       $toNumber Destination number.
     * @param string|null                  $toName   Name of a called user.
     *
     * @return boolean
     * @throws \Exception If failed start call.
     */
    private function _startUserCallFromDevice(\Wildix\Colleagues\Colleague $user, $uaId, $toNumber, $toName = null) {
        $fromExtension = $user->getExtension();
        $this->getLogger()->debug(
            __LINE__,
            sprintf('Start user call. From Extension - %s | From UAID - %s | To Number - %s | To Name - %s', $fromExtension, $uaId, $toNumber, $toName)
        );

        if ($uaId == 'any_device') {
            /*
                Action: Originate.
                Channel: SIP/125200.
                Application: Transfer.
                Data: SIP/125100.
                Async: true.
                CallerID: 125100
            */

            $result = $this->_originateCallInvite($fromExtension, $toNumber, isset($toName) ? $toName : $toNumber);
        } else if ($uaId == 'mobility') {
            /*
                Action: Originate.
                Channel: Local/125300@mobility.
                Exten: 125200.
                Context: users.
                Priority: 1.
                Async: true.
                CallerID: 125300.
                Variable: MOBILITY_OPTS=|0|users|en|gid=946731237;o=com0;dc=wildix.
            */

            $result = $this->_originateMobileCall($user, $toNumber);
        } else {
            /*
                Action: SendRefer.
                Dest: 125200.
                To: 125100.
                UaID: sip:125200@10.100.1.173:33035.
            */

            $result = $this->getCallWeaverManager()->sendRefer($fromExtension, $toNumber, $uaId);
        }

        return $result;
    }

    private function _originateCallInvite($fromNumber, $toNumber, $toName) {
        $channel     = 'SIP/' . $fromNumber;
        $application = 'Transfer';
        $data        = 'SIP/' . $toNumber;
        $async       = true;
        $callerid    = '"' . $toName . '" <' . $toNumber . '>';

        return $this->getCallWeaverManager()->originate(
            $channel,
            null,
            null,
            null,
            $application,
            $data,
            null,
            $callerid,
            null,
            null,
            $async,
            null
        );
    }

    private function _originateMobileCall(\Wildix\Colleagues\Colleague $fromUser, $toNumber) {
        $channel  = 'Local/' . $fromUser->getExtension() . '@mobility';
        $exten    = $toNumber;
        $context  = $fromUser->getDialplan();
        $priority = 1;
        $async    = true;
        $callerid = $fromUser->getExtension();
        $variable = 'MOBILITY_OPTS=' . $fromUser->getMobilePhone() . '||0|' . $context . '|' . $fromUser->getLanguage() . '|' . $fromUser->getGroupDn();

        return $this->getCallWeaverManager()->originate(
            $channel,
            $exten,
            $context,
            $priority,
            null,
            null,
            null,
            $callerid,
            $variable,
            null,
            $async,
            null
        );
    }

    /**
     * Get active device user agent by extension.
     *
     * @param string $extension     Extension.
     * @param array  $registrations User sip registrations.
     *
     * @return string User agent.
     */
    private function getActiveDeviceUaId(string $extension, array $registrations): string {
        try {
            $activeDevice = current($this->getSipProxy()->getUsersActiveDevices([$extension]));
        } catch (\Exception $e) {
            $this->getLogger()->info(__LINE__, 'User do not have any active device. Message: ' . $e->getMessage());
            $activeDevice = [
                'type' => 'any_device',
                'useragent' => 'unknown',
            ];
        }

        if (isset($registrations[$extension]['registrations'])
            && is_array($registrations[$extension]['registrations'])
        ) {
            $userAgentLambda = static function (array $registration) use ($activeDevice) {
                return $registration['useragent'] === $activeDevice['useragent'];
            };
            // web phone
            if (strpos($activeDevice['useragent'], 'WebRTC-') !== false) {
                $userAgentLambda = static function (array $registration) {
                    return (strpos($registration['useragent'], 'WebRTC-') !== false);
                };
            }

            // use registration array as source of truth
            $activeDeviceFromRegistrations = current(array_filter($registrations[$extension]['registrations'], $userAgentLambda));
            if (is_array($activeDeviceFromRegistrations)
                && array_key_exists('contact', $activeDeviceFromRegistrations)
                && array_key_exists('useragent', $activeDeviceFromRegistrations)
            ) {
                $activeDevice = [
                    'extension' => $extension,
                    'type'      => 'device',
                    'contact'   => $activeDeviceFromRegistrations['contact'],
                    'useragent' => $activeDeviceFromRegistrations['useragent'],
                ];
            }
        }

        return $activeDevice['type'] === 'device' ? $activeDevice['contact'] : $activeDevice['type'];
    }

    private function getCallWeaverManager(): CallweaverManager {
        return $this->getContainer()->get(CallweaverManager::class);
    }

    private function getSipProxy() : \Wildix\SipProxy {
        return $this->getContainer()->get(\Wildix\SipProxy::class);
    }

    private function getCallFiles(): CallFiles {
        return $this->getContainer()->get(CallFiles::class);
    }

    private function getColleagues(): \Wildix\Colleagues\Colleagues {
        return $this->getContainer()->get(\Wildix\Colleagues\Colleagues::class);
    }

    private function getPbx(): PBX {
        return $this->getContainer()->get(PBX::class);
    }
}
