<?php
namespace Controllers;

use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\CDRView;
use Wildix\Colleagues\Colleague;
use Wildix\Colleagues\Colleagues;
use Wildix\Colleagues\Groups;
use Wildix\Databases\WMSDB;
use Wildix\PBXes\PBXes;

class CDRViewController extends Controller {

    protected string $licenceAcl = 'cti_cdr_view';

    public function getSettings(Request $request, Response $response): Response {
        $settings = $this->service()->getSettings($this->getUser()->getID());
        unset($settings['id']);
        return $this->result($settings, $response);
    }

    public function setSettings(Request $request, Response $response): Response {
        $params = $request->getParsedBody();
        $data   = [
            'period'               => trim($params['period']) ?? null,
            'sla_period'           => trim($params['sla_period']) ?? null,
            'miss_call_sec_period' => trim($params['miss_call_sec_period']) ?? null,
            'grade_of_accuracy' => (int)($params['grade_of_accuracy'] ?? 2),
        ];

        $this->service()->setSettings($this->getUser()->getID(), $data);

        return $this->success($response);
    }

    public function getExtraProperties(Request $request, Response $response): Response {
        $res = [
            'groups' => [],
            'trunks' => $this->getWmsDb()->getTrunksNamesList(),
            'pbxes'  => []
        ];

        // Put groups
        foreach ($this->serviceGroups()->getGroups() as $group) {
            $users = [];

            foreach ($this->serviceColleagues()->getUsersByGroup($group) as $user) {
                $users[] = $user->getExtension();
            }
            $res['groups'][] = [
                $group->getID(),
                $group->getName(),
                $users,
            ];
        }

        foreach ($this->servicePbxEs()->getPBXes() as $pbx) {
            $res['pbxes'][] = [
                $pbx->getSerial(),
                $pbx->getDomain(),
            ];
        }

        return $this->result($res, $response);
    }

    public function getInfoProperties(Request $request, Response $response): Response {
        try {
            $res = $this->service()->getCDRInfo();
        } catch (\Exception $e) {
            $res = [
                'skipped' => 1,
                'timeout' => 1,
                'exception' => $e->getMessage()
            ];
        }

        return $this->result($res, $response);
    }

    public function getCallsHistory(Request $request, Response $response): Response {
        $data   = $request->getQueryParams();
        $filter = $data['filter'];
        $start  = $data['start'];
        $count  = $data['count'];
        $sort   = $data['sort'];
        $dir    = $data['dir'];

        try {
            if (!empty($filter['id']) && is_array($filter['id'])) {
                $filter['rowid'] = $filter['id'];
            }
            unset($filter['id']);
            if (!empty($filter['date']) && is_array($filter['date'])) {
                $filter['start']         = $filter['date'];
                $filter['start']['from'] = $filter['start']['from'] . ' 00:00:00';
                $filter['start']['to']   = $filter['start']['to'] . ' 23:59:59';
            }
            unset($filter['date']);
            // this is needed becoase call model want to give only for your pbx by default
            // "exclude" will goes here too
            $filter['pbxid'] = '*';

            $result = $this->service()->getCallsHistory(
                $filter,
                $start,
                $count,
                $sort,
                $dir,
                null,
                $this->getUser()->can(Colleague::SEE_CALL_RECORDINGS)
            );
        } catch (\Exception $exc) {
            $result = [
                'events' => [],
                'skipped' => 1,
                'timeout' => 1,
                'exception' => $exc->getMessage()
            ];
        }
        return $this->result($result, $response);
    }

    public function getFilters(Request $request, Response $response): Response {
        $result       = [];
        $data         = $request->getQueryParams();
        $filter       = $data['filter'] ?? [];
        $search       = $data['search'] ?? '';
        $searchFields = $data['searchFields'] ?? '';
        $start        = $data['start'] ?? 0;
        $count        = $data['count'] ?? 100;
        $sort         = $data['sort'];
        $dir          = $data['dir'];
        if ($searchFields) {
            $searchFields = explode(',', $searchFields);
        }
        $total = $this->service()->getCountFilters($filter, $search, $searchFields, $this->getUser());

        if ($total > 0) {
            $result = $this->service()->getFilters($filter, $search, $searchFields, $start, $count, $sort, $dir, $this->getUser());
        }

        return $this->resultList($result, $response, ['total' => $total]);
    }

    public function addFilter(Request $request, Response $response): Response {
        $params = $request->getParsedBody();
        $data   = [
            'name'        => trim($params['name']) ?? '',
            'description' => trim($params['description']) ?? '',
            'shared'      => trim($params['shared']) ?? '',
            'filter'      => trim($params['filter']) ?? ''
        ];

        $result = $this->service()->addFilter($this->getUser()->getID(), $data);

        return $this->result(['id' => $result], $response);
    }

    public function deleteFilter(Request $request, Response $response, $params): Response {
        $query        = $request->getQueryParams();
        $filter       = $query['filter'] ?? [];
        $search       = $query['search'] ?? '';
        $searchFields = $query['searchFields'] ?? '';

        if (!is_array($filter)) {
            $filter = [];
        }
        if (!empty($params['id'])) {
            $filter['id'] = $params['id'];
        }

        if ($searchFields) {
            $searchFields = explode(',', $searchFields);
        }
        $result = $this->service()->deleteFilters($filter, $search, $searchFields, $this->getUser());

        if (is_numeric($result)) {
            $result = ['filters' => $result];
        }

        return $this->result($result, $response);
    }

    private function service(): CDRView {
        return $this->getContainer()->get(CDRView::class);
    }

    private function serviceGroups(): Groups {
        return $this->getContainer()->get(Groups::class);
    }

    private function servicePbxEs(): PBXes {
        return $this->getContainer()->get(PBXes::class);
    }

    private function serviceColleagues(): Colleagues {
        return $this->getContainer()->get(Colleagues::class);
    }

    private function getWmsDb(): WMSDB {
        return $this->getContainer()->get(WMSDB::class);
    }
}
