<?php
namespace Controllers;

use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\CallHistory\Call;
use Wildix\CallHistory\Calls;
use Wildix\Colleagues\Colleague;
use Wildix\Validator;

class CallHistory extends Controller {

    public function getItems(Request $request, Response $response, array $args): Response {
        $params = $request->getQueryParams();

        $user = $this->getUser();

        if (!empty($args['extension'])) {
            if (!$user->isAdministrator()) {
                return $this->error('No access for use history (' . $args['extension'] . ')', $response, 403);
            } else {
                $user = $this->_getColleaguesClass()->getUserByExtension($args['extension']);
                if ($user === null) {
                    return $this->error('User not found', $response, 404);
                }
            }
        }

        $filter = $params['filter'] ?? null;
        if (!is_array($filter)) {
            $filter = [];
        }

        if (array_key_exists('archived', $filter)) {
            $filter['archived'] = Validator::isTrue($filter['archived']);
        }

        $search       = (string)($params['search'] ?? '');
        $searchFields = null;
        if (!empty($params['searchFields'])) {
            $searchFields = explode(',', (string)$params['searchFields']);
        }

        $start  = (int)($params['start'] ?? 0);
        $count  = (int)($params['count'] ?? 100);
        $sort   = (string)($params['sort'] ?? '');
        $dir    = (string)($params['dir'] ?? '');
        $fields = !empty($params['fields']) ? array_map('trim', explode(',', (string)$params['fields'])) : [];

        return $this->_mackCallHistoryResponse(
            $response,
            $filter,
            $count,
            $start,
            $fields,
            $sort,
            $dir,
            $search,
            $searchFields,
            $user
        );
    }

    public function getItem(Request $request, Response $response, array $args): Response {
        $params = $request->getQueryParams();
        $user   = $this->getUser();

        $fields = $params['fields'];
        if ($fields) {
            $fields = explode(',', $fields);
        }

        $gmt  = date('O');
        $call = $this->_getCallsHistory()->getByID($args['id'], $user, $user->can(Colleague::SEE_CALL_RECORDINGS));
        if ($call) {
            $record = array_merge($call->getAsArray($fields), array('timezone' => $gmt));
            if ($this->_needShowAttachments($fields)) {
                $record[Calls::ATTACHMENTS] = $this->_getCallsHistory()->formatAttachments(
                    $record,
                    (string)$this->getUser()->getExtension()
                );
            }

            return $this->result($record, $response);
        } else {
            return $this->error('History call with id ' . $args['id'] . ' is not found', $response, 404);
        }
    }

    public function update(Request $request, Response $response, array $args): Response {
        $getParams  = $request->getQueryParams();
        $postParams = $request->getParsedBody();

        $requestData = $postParams['data'];
        if (empty($requestData) || !is_array($requestData)) {
            return $this->error('Call data cannot be empty', $response, 400);
        }

        $user = $this->getUser();

        if (!empty($args['id'])) {
            $filter[\Wildix\CallHistory\Call::ID] = $args['id'];
        } else {
            $filter = $getParams['filter'];
            if (!is_array($filter)) {
                $filter = [];
            }
        }

        $search       = (string)($getParams['search'] ?? '');
        $searchFields = null;
        if (!empty($getParams['searchFields'])) {
            $searchFields = explode(',', (string)$getParams['searchFields']);
        }

        // allow changing only archived status
        $data = [];
        if (array_key_exists('archived', $requestData) && $user) {
            $call = $this->_getCallsHistory()->getByID($args['id'], $user, $user->can(Colleague::SEE_CALL_RECORDINGS));
            if ($call) {
                $archived = ($requestData['archived'] === 'true' || $requestData['archived'] === true) ? 1 : 0;
                if ($call->getFromNumber() == $user->getExtension()) {
                    $data['from_archived'] = $archived;
                } else {
                    $data['to_archived'] = $archived;
                }
            } else {
                return $this->error('History call with id ' . $args['id'] . ' is not found', $response, 404);
            }
        } else {
            return $this->error('Not found archived status', $response, 400);
        }

        $result = $this->_getCallsHistory()->updateCall($data, $filter, $search, $searchFields, $user);

        if (is_numeric($result)) {
            $result = array('changed' => $result);
        }

        return $this->result($result, $response);
    }

    public function remove(Request $request, Response $response, array $args): Response {
        $params = $request->getQueryParams();
        $user   = $this->getUser();
        $result = false;

        if (!empty($args['id'])) {
            $filter[\Wildix\CallHistory\Call::ID] = $args['id'];
        } else {
            $filter = $params['filter'];
            if (!is_array($filter)) {
                $filter = [];
            }
        }

        $search       = (string)($params['search'] ?? '');
        $searchFields = null;
        if (!empty($params['searchFields'])) {
            $searchFields = explode(',', (string)$params['searchFields']);
        }

        $result = $this->_getCallsHistory()->deleteCall($filter, $search, $searchFields, $user);

        if (is_numeric($result)) {
            $result = ['deleted' => $result];
        }

        return $this->result($result, $response);
    }

    public function getHistory(Request $request, Response $response): Response {
        $params = $request->getQueryParams();

        $filter = [];

        if (!empty($params['filter'][Call::START])) {
            $filter[Call::START] = $params['filter'][Call::START];
        }

        if (!empty($params['filter'][Call::END])) {
            $filter[Call::END] = $params['filter'][Call::END];
        }

        $limit = (int)($params['count'] ?? 100);
        $limit = ($limit <= 100 ? $limit : 100);

        $offset = (int)($params['start'] ?? 0);

        $fields = isset($params['fields']) ? array_map('trim', explode(',', $params['fields'])) : [];

        return $this->_mackCallHistoryResponse($response, $filter, $limit, $offset, $fields);
    }

    public function updateTags(Request $request, Response $response, array $args): Response {
        $params = $request->getParsedBody();
        $rowid  = (int)$args['id'];
        $cdrDb  = $this->_getCdrDb();
        $user   = $this->getUser();

        $call = $this->_getCallsHistory()->getByID($rowid, $user, $user->can(Colleague::SEE_CALL_RECORDINGS));
        if (!$call) {
            return $this->error('History call with id ' . $rowid . ' is not found', $response, 404);
        }

        $allowedTags = [];
        $tagSetting  = $this->_getCdrGeneralSettings()->get('tags');
        if (!empty($tagSetting)) {
            $allowedTags = array_map('trim', explode(',', $tagSetting));
        }

        $tags = trim($params['tags']);
        if ($tags) {
            $tagsArray      = explode(',', $tags);
            $notAllowedTags = !empty($allowedTags) ? array_diff($tagsArray, $allowedTags) : [];

            if ($notAllowedTags) {
                $validator = new \Wildix\Validator();
                $validator->rule('CallHistoryTag', 'tags');
                $validateErrors = $validator->validateData(['tags' => implode(',', $notAllowedTags)], ['tags']);
            }

            if ($notAllowedTags && $validateErrors) {
                return $this->error('Not allowed tags', $response, 400);
            }
        }

        $tags = $cdrDb->realEscapeString('tags:' . $tags);

        if ($cdrDb->isMsSql()) {
            $query = 'UPDATE [' . $cdrDb->getCdrTableName() . '] SET lastdata = ' . $tags . '  WHERE rowid = ' . $rowid . ';';
        } else {
            $query = 'UPDATE `' . $cdrDb->getCdrTableName() . '` SET lastdata = ' . $tags . '  WHERE rowid = ' . $rowid . ';';
        }
        $result = $cdrDb->exec($query);


        if ($result === false) {
            return $this->error('Internal server error', $response);
        }

        return $this->success($response);
    }

    public function readVoicemail(Request $request, Response $response, array $args): Response {
        return $this->actionVoicemail($response, (int)$args['id'], 'read');
    }

    public function removeVoicemail(Request $request, Response $response, array $args): Response {
        return $this->actionVoicemail($response, (int)$args['id'], 'delete');
    }

    private function actionVoicemail(Response $response, $callID, $action): Response {
        $user = $this->getUser();

        $call = $this->_getCallsHistory()->getByID($callID, $user, $user->can(Colleague::SEE_CALL_RECORDINGS));
        if (!$call) {
            return $this->error('History call with id ' . $callID . ' is not found', $response, 404);
        }

        if ($call->getLastapp() != 'VoiceMail') {
            return $this->error('History call with id ' . $callID . ' is not voicemail', $response, 404);
        }

        $start = strrpos($call->getUserfield(), '/') + 1;
        $end   = strrpos($call->getUserfield(), '.');
        $msgID = substr($call->getUserfield(), $start, $end - $start);

        $result = $this->_getLuaService()->manageVoicemail($user->getExtension(), $msgID, $action);
        if ((int)$result['changed'] === 1) {
            return $this->success($response);
        } else {
            if ($result['error'] == 'vmemptylist') {
                $result['error'] = 'Voicemail is not found';
            }
            return $this->error($result['error'], $response, 400);
        }
    }

    private function _mackCallHistoryResponse(
        Response $response,
        array $filter,
        int $limit,
        int $offset,
        array $fields = [],
        ?string $sortBy = null,
        ?string $orderDir = null,
        ?string $search = null,
        ?array $searchFields = null,
        ?Colleague $forUser = null
    ): Response {
        $result = [
            'total' => 0,
            'records' => [],
        ];

        $callsHistory = $this->_getCallsHistory();

        $result['total'] = $callsHistory->getCount($filter, $search, $searchFields, $forUser);

        if ($result['total'] > 0) {
            $result['records'] = $callsHistory->getCalls(
                $filter,
                $search,
                $searchFields,
                $offset,
                $limit,
                $sortBy,
                $orderDir,
                $forUser,
                $this->getUser()->can(Colleague::SEE_CALL_RECORDINGS)
            );

            $gmt   = date('O');
            $count = count($result['records']);
            for ($i = 0; $i < $count; $i++) {
                $record = array_merge($result['records'][$i]->getAsArray($fields), array('timezone' => $gmt));

                if ($this->_needShowAttachments($fields)) {
                    $extension = $forUser !== null ? (string)$forUser->getExtension() : '';

                    $record[Calls::ATTACHMENTS] = $callsHistory->formatAttachments($record, $extension);
                }

                $result['records'][$i] = $record;
            }
        }


        return $this->result($result, $response);
    }

    private function _needShowAttachments($fields) {
        return empty($fields) || in_array(Calls::ATTACHMENTS, $fields);
    }

    private function _getCallsHistory(): Calls {
        return $this->getContainer()->get(Calls::class);
    }

    private function _getCdrDb(): \Wildix\Databases\CDRDB {
        return $this->getContainer()->get(\Wildix\Databases\CDRDB::class);
    }

    private function _getCdrGeneralSettings(): \Wildix\Settings\CDR\General {
        return $this->getContainer()->get(\Wildix\Settings\CDR\General::class);
    }

    private function _getLuaService(): \Services\LuaService {
        return $this->getContainer()->get(\Services\LuaService::class);
    }

    private function _getColleaguesClass(): \Wildix\Colleagues\Colleagues {
        return $this->getContainer()->get(\Wildix\Colleagues\Colleagues::class);
    }
}
