<?php
namespace Controllers;

use Cache\FileCache;
use Cache\FileCacheInterface;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\Colleagues\Colleague;
use Wildix\Colleagues\Group;
use Wildix\Error;
use Wildix\Exception;
use Wildix\InvalidDataException;
use Wildix\UserDataCleaner;
use Wildix\UserDataCleaner\Factories\UserDataFactory;

/**
 * Colleagues Controller
 */
class Colleagues extends Controller {

    private const CACHE_PREFIX = 'colleague';

    /**
     * Time of сache life in seconds.
     * After this time, the cache will be rebuilt.
     */
    private const CACHE_TTL = 600;

    /**
     * Real cache time of life in seconds.
     */
    private const CACHE_REAL_TTL = 86400;

    /**
     * Managed groups list.
     *
     * @var array $_managedGroups
     */
    private $_managedGroups;

    public function getItems(Request $request, Response $response): Response {
        return $this->result($this->_getColleagues($request, $response, false), $response);
    }

    public function getItemsAcl(Request $request, Response $response): Response {
        return $this->result($this->_getColleagues($request, $response, true), $response);
    }

    private function _getColleagues(Request $request, Response $response, bool $checkAcl): array {
        try {
            $emptyResult = [
                'total'   => 0,
                'records' => []
            ];

            $params       = $request->getQueryParams();
            $filter       = $params['filter'] ?? [];
            $search       = $params['search'];
            $searchFields = $params['searchFields'];
            if ($searchFields) {
                $searchFields = explode(',', $searchFields);
            }
            $start  = $params['start'];
            $count  = $params['count'];
            $sort   = $params['sort'];
            $dir    = $params['dir'];
            $fields = $params['fields'];

            if ($fields) {
                $fields = explode(',', $fields);
            }
            $searchStrategy = $params['searchStrategy'] ?? \Wildix\Colleagues\Colleagues::SEARCH_STRATEGY_CONTAIN;
        } catch (\Throwable $exception) {
            throw new Exception(Error::BadRequest, 400);
        }
        if ($checkAcl) {
            $managedGroups = array_map(
                function ($group) {
                    return $group[Group::DN];
                },
                $this->getManegedGroups()
            );

            $filteredGroups = $managedGroups;

            if (!empty($filter[Colleague::GROUP_DN])) {
                $filteredGroups = array_intersect($filter[Colleague::GROUP_DN], $managedGroups);
            }

            if (empty($filteredGroups)) {
                return $emptyResult;
            }

            $filter[Colleague::GROUP_DN] = $filteredGroups;
        }

        $colleagueRepository = $this->getColleagueRepository();

        $result    = false;
        $fromCache = true;
        $cacheKey  = '';
        try {
            $user = $this->getUser();

            $isCachedRequest = (int)$count > 499 && $sort === null &&
                                $dir === null && $fields === null &&
                                $search === null && $searchFields === null &&
                                count($filter) === 0;

            if ($isCachedRequest) {
                $cacheKey = sprintf('%s_%s_%s', $start, $count, $user->getGroupDn());
                $result   = $this->getCache()->get($cacheKey, self::CACHE_PREFIX);
            }

            if ($result === false) {
                $result    = $emptyResult;
                $fromCache = false;

                if ($checkAcl) {
                    $user = null;
                }

                $this->getLogger()->info(__LINE__, sprintf('Fetch colleagues from storage %s', $cacheKey));
                $result['records'] = $colleagueRepository->getUsers($filter, $search, $searchFields, $start, $count, $sort, $dir, $user, $searchStrategy);

                $count = count($result['records']);
                for ($i = 0; $i < $count; $i++) {
                    if ($result['records'][$i]) {
                        $result['records'][$i] = $result['records'][$i]->getAsArray($fields);
                    }
                }

                $result['total'] = $colleagueRepository->getTotalLastQuery();

                if ($isCachedRequest) {
                    $this->getCache()->setRealTtl(self::CACHE_REAL_TTL);
                    $this->getCache()->set($cacheKey, $result, self::CACHE_TTL, self::CACHE_PREFIX);
                }
            }
        } catch (\Exception $e) {
            throw new InvalidDataException($e->getCode());
        }

        $this->getLogger()->info(__LINE__, sprintf('Get colleagues %s. Use cache %d', $cacheKey, (int)$fromCache));

        return $result;
    }

    public function getItem(Request $request, Response $response, array $args): Response {
        return $this->_getColleague($request, $response, $args['id'], false);
    }

    public function getItemAcl(Request $request, Response $response, array $args): Response {
        return $this->_getColleague($request, $response, $args['id'], true);
    }

    private function _getColleague(Request $request, Response $response, $colleagueId, $checkAcl): Response {
        $params = $request->getQueryParams();
        $fields = (string)$params['fields'] ?? '';
        $fields = ($fields !== '') ? explode(',', $fields) : [];

        $colleague = $this->findOrFail($colleagueId);

        if ($checkAcl && $this->_checkPermissionToGroup($colleague->getGroupDn())) {
            throw new Exception(Error::getMessage(\Wildix\Error::AccessError), 403);
        }

        return $this->result($colleague->getAsArray($fields), $response);
    }

    private function _checkPermissionToGroup(string $groupDn): bool {
        $managedGroups = array_column($this->getManegedGroups(), \Wildix\Colleagues\Group::DN);
        return !in_array($groupDn, $managedGroups);
    }

    public function create(Request $request, Response $response): Response {
        $colleague = $this->colleagueFactory()->build($request->getParsedBody());

        // check permission managed group
        if ($this->_checkPermissionToGroup($colleague->getGroupDn())) {
            return $this->error(Error::getMessage(\Wildix\Error::AccessError), $response);
        }

        $this->getColleagueRepository()->addUser($colleague);

        $user = $this->findOrFail($colleague->getID());

        return $this->result($user->getAsArray(), $response);
    }

    public function update(Request $request, Response $response, array $args): Response {
        $data      = $request->getParsedBody();
        $colleague = $this->findOrFail($args['id']);

        $isGroupChanged = (isset($data[Colleague::GROUP_DN]) && $data[Colleague::GROUP_DN] !== $colleague->getGroupDn());

        // check permission to manage of current colleague group
        if ($this->_checkPermissionToGroup($colleague->getGroupDn())) {
            return $this->error(Error::getMessage(\Wildix\Error::AccessError), $response);
        }

        $newGroupDn = $data[Colleague::GROUP_DN] ?? false;
        // check permission to manage of target group
        if ($newGroupDn && $isGroupChanged && $this->_checkPermissionToGroup($newGroupDn)) {
            return $this->error(Error::getMessage(\Wildix\Error::AccessError), $response);
        }

        $this->colleagueFactory()->update($colleague, $data);
        $this->getColleagueRepository()->updateUser($colleague);

        return $this->result($colleague->getAsArray(), $response);
    }

    public function remove(Request $request, Response $response, array $args): Response {
        $params = $request->getQueryParams();

        $colleague = $this->findOrFail($args['id']);

        // check permission to manage of current colleague group
        if ($this->_checkPermissionToGroup($colleague->getGroupDn())) {
            return $this->error(Error::getMessage(\Wildix\Error::AccessError), $response);
        }

        $userDeletingData = UserDataFactory::make(
            array_flip($params['data'] ?? [])
        );

        $this->getContainer()->get(UserDataCleaner::class)->clearUserData($colleague, $userDeletingData);

        if ($userDeletingData->isNeedToDeleteUser()) {
            $this->getColleagueRepository()->removeUser($colleague, $this->getCurrentPbxDn());
        }

        return $this->success($response);
    }

    public function updateColleagueByExtension(Request $request, Response $response, array $args): Response {
        $colleagueRepository = $this->getColleagueRepository();

        $params    = $request->getParsedBody();
        $colleague = $colleagueRepository->getUserByExtension($args['extension'], $this->getUser());
        $result    = null;

        if ($colleague) {
            $result = $colleagueRepository->updateUserProperty($params['data'], $colleague);
            if ($result === true) {
                $result = array('changed' => '1');
            } else {
                // Error
                $result = array('changed' => '0');
            }
        } else {
            return $this->error('Colleague with extension ' . $args['extension'] . ' is not found', $response, 404);
        }

        return $this->result($result, $response);
    }

    /**
     * Load colleague by id or terminate execution with HTTP error 404
     *
     * @param integer $colleagueId Id of looking for colleague.
     *
     * @return Colleague
     * @throws InvalidDataException Error.
     */
    private function findOrFail($colleagueId): Colleague {
        $colleague = $this->getColleagueRepository()->getUserByID($colleagueId, $this->getUser());

        if (!($colleague instanceof Colleague)) {
            throw new InvalidDataException(Error::getMessage(Error::NotFound), [], 404);
        }

        return $colleague;
    }

    /**
     * Build factory for colleague instances
     *
     * @return \Wildix\Colleagues\ColleagueFactory
     */
    private function colleagueFactory(): \Wildix\Colleagues\ColleagueFactory {
        return $this->getContainer()->get(\Wildix\Colleagues\ColleagueFactory::class);
    }

    /**
     * Return LDAP DN of current PBX
     *
     * @return string
     */
    private function getCurrentPbxDn() {
        $currentPBX = \WMS_Ldap::getInstance()->getEntry(
            \WMS_Ldap::PBX,
            '(pbxSerial=' . $this->getPbx()->getSerial() . ')',
            array('pbxSerial', 'dn')
        );

        return isset($currentPBX['dn']) ? $currentPBX['dn'] : '';
    }

    /**
     * Get list of managed groups.
     *
     * @return array
     */
    private function getManegedGroups() {
        if ($this->_managedGroups === null) {
            $this->_managedGroups = $this->getGroupsService()->getManagedGroups([\Wildix\Colleagues\Group::DN]);
        }

        return $this->_managedGroups;
    }

    /**
     * Returns \Wildix\Colleagues\Colleagues object.
     *
     * @return \Wildix\Colleagues\Colleagues
     */
    private function getColleagueRepository(): \Wildix\Colleagues\Colleagues {
        return $this->getContainer()->get(\Wildix\Colleagues\Colleagues::class);
    }

    /**
     * Get Groups service with permissions.
     *
     * @return \Wildix\Colleagues\Groups
     */
    private function getGroupsService(): \Wildix\Colleagues\Groups {
        $groups = $this->getContainer()->get(\Wildix\Colleagues\Groups::class);
        $groups->setPermissions($this->getContainer()->get(\WMS_Permissions::class));

        return $groups;
    }

    private function getCache(): FileCacheInterface {
        return $this->getContainer()->get(FileCache::class);
    }

    private function getPbx(): \Models\PBX {
        return $this->getContainer()->get(\Models\PBX::class);
    }
}
