<?php
namespace Controllers;

use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\Phonebooks\Contact;
use Wildix\Phonebooks\Contacts;
use Wildix\Phonebooks\Phonebooks;

class ContactsController extends Controller {

    public function getItems(Request $request, Response $response, array $args): Response {
        $result = [];
        $params = $request->getQueryParams();
        $filter = $params['filter'] ?? [];

        if (!empty($args['ids'])) {
            $ids                 = array_map('intval', explode(',', $args['ids']));
            $filter[Contact::ID] = array('or' => $ids);
        } else {
            $ids = [];
        }
        if (!empty($args['phonebookId'])) {
            $this->servicePhonebooks()->getByID($args['phonebookId'], $this->getUser());
            $filter[Contact::PHONEBOOK_ID] = $args['phonebookId'];
        }

        // fix for cdrview reconnect action
        if (isset($filter['last_update']['from']) && $filter['last_update']['from'] === '') {
            unset($filter['last_update']);
        }

        $search       = $params['search'];
        $searchFields = $params['searchFields'] ?? false;
        if ($searchFields) {
            $searchFields = explode(',', $searchFields);
        }
        $start  = $params['start'];
        $count  = $params['count'];
        $sort   = $params['sort'];
        $dir    = $params['dir'];
        $fields = $params['fields'] ?? false;
        if ($fields) {
            $fields = explode(',', $fields);
        }
        $withTotal = $params['withTotal'];
        if ($withTotal === 'no' || $withTotal === 0 || $withTotal === false || $withTotal === 'false') {
            $withTotal = false;
        } else {
            $withTotal = true;
        }

        if ($withTotal) {
            $total = $this->service()->getCount($filter, $search, $searchFields, $this->getUser());
        }

        if (!$withTotal || $total > 0) {
            $contacts = $this->service()->getContacts($filter, $search, $searchFields, $start, $count, $sort, $dir, $this->getUser());
            foreach ($contacts as $key => $contact) {
                $result[] = $contact->getAsArray($fields);
            }
        } else {
            if (!empty($args['ids'])) {
                return $this->error('Contact with id ' . $args['ids'] . ' is not found', $response, 404);
            }
        }

        if (count($ids) !== 1) {
            $return = array('records' => $result);
            if ($withTotal) {
                $return['total'] = $total;
            }
        } else {
            $return = $result[0];
        }

        return $this->result($return, $response);
    }

    public function getPhonebookContacts(Request $request, Response $response, array $args): Response {
        return $this->getItems($request, $response, $args);
    }

    public function getPhonebookContact(Request $request, Response $response, array $args): Response {
        return $this->getItems($request, $response, $args);
    }

    public function getContact(Request $request, Response $response, array $args): Response {
        $params = $request->getQueryParams();
        $fields = $params['fields'];
        if ($fields) {
            $fields = explode(',', $fields);
        }

        $contact = $this->service()->getByID($args['contactId'], $this->getUser());

        if (!$contact) {
            return $this->error('Contact with id ' . $args['contactId'] . ' is not found', $response, 404);
        }

        return $this->result($contact->getAsArray($fields), $response);
    }

    public function create(Request $request, Response $response, array $args): Response {
        $params = $request->getParsedBody();
        $data   = $params['data'] ?? [];

        if (!empty($args['phonebookId'])) {
            $this->servicePhonebooks()->getByID($args['phonebookId'], $this->getUser());
            $data[Contact::PHONEBOOK_ID] = $args['phonebookId'];
        }

        $result = $this->service()->addContact($data, $this->getUser());

        return $this->result(['id' => $result], $response);
    }

    public function update(Request $request, Response $response, array $args): Response {
        $postParams = $request->getParsedBody();
        $getParams  = $request->getQueryParams();

        $filter = [];
        if (isset($args['contactId'])) {
            $contact = $this->service()->getByID($args['contactId'], null);
            if (!$contact) {
                return $this->error('Contact with id ' . $args['contactId'] . ' is not found', $response, 404);
            }
            $filter[Contact::ID] = $args['contactId'];
        } else {
            $filter = $getParams['filter'];
            if (!is_array($filter)) {
                $filter = [];
            }
        }

        if (isset($args['phonebookId'])) {
            $this->servicePhonebooks()->getByID($args['phonebookId'], $this->getUser());
            $filter[Contact::PHONEBOOK_ID] = $args['phonebookId'];
        }

        $searchFields = $getParams['searchFields'];
        if ($searchFields) {
            $searchFields = explode(',', $searchFields);
        }

        $result = $this->service()->updateContacts(
            $postParams['data'],
            $filter,
            $getParams['search'],
            $searchFields,
            $this->getUser()
        );

        if (is_numeric($result)) {
            $result = ['changed' => $result];
        }

        return $this->result($result, $response);
    }

    public function updatePhonebookContact(Request $request, Response $response, array $args): Response {
        return $this->update($request, $response, $args);
    }

    public function deleteContacts(Request $request, Response $response, array $args): Response {
        $params = $request->getQueryParams();
        $filter = $params['filter'];
        if (!is_array($filter)) {
            $filter = [];
        }

        if (!empty($args['contactId'])) {
            $filter[Contact::ID] = $args['contactId'];
        }
        if (!empty($args['phonebookId'])) {
            $filter[Contact::PHONEBOOK_ID] = $args['phonebookId'];
        }

        $searchFields = $params['searchFields'];
        if ($searchFields) {
            $searchFields = explode(',', $searchFields);
        }

        $result = $this->service()->deleteContacts($filter, $params['search'], $searchFields, $this->getUser());

        if (is_numeric($result)) {
            $result = ['contacts' => $result];
        }


        return $this->result($result, $response);
    }

    public function remove(Request $request, Response $response, array $args): Response {
        return $this->deleteContacts($request, $response, $args);
    }

    public function removePhonebookContacts(Request $request, Response $response, array $args): Response {
        return $this->deleteContacts($request, $response, $args);
    }

    public function removePhonebookContact(Request $request, Response $response, array $args): Response {
        return $this->deleteContacts($request, $response, $args);
    }

    private function servicePhonebooks(): Phonebooks {
        return $this->getContainer()->get(Phonebooks::class);
    }

    private function service(): Contacts {
        return $this->getContainer()->get(Contacts::class);
    }
}
