<?php

namespace Controllers;

use Api\Support\BaseCollectObject;
use Api\Support\Interfaces\ArrayableInterface;
use Api\Translation\Interfaces\TranslationStrategyInterface;
use DataEncoder\JSON;
use DataEncoder\XML;
use Logger\Log;
use Psr\Container\ContainerInterface;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Slim\Psr7\Factory\StreamFactory;
use User;
use Wildix\Error;
use Wildix\Exception;
use Wildix\License;

abstract class Controller {

    const SUCCESS = 'Success';
    const RECORDS = 'records';

    protected string $licenceAcl = '';

    public function __construct(private ContainerInterface $container) {
        if (!empty($this->licenceAcl) && !$this->getContainer()->get(License::class)->haveLicense($this->licenceAcl)) {
            throw new Exception(Error::AccessError, 403);
        }
    }

    public function getLogger(): Log {
        return $this->getContainer()->get(Log::class);
    }

    /**
     * Instance of dependence injection container.
     *
     * @return ContainerInterface
     */
    protected function getContainer(): ContainerInterface {
        return $this->container;
    }

    /**
     * Get Colleague User.
     *
     * @return \Wildix\Colleagues\Colleague
     */
    protected function getUser() {
        return $this->getContainer()->get(User::class);
    }

    public function error($error, Response $response, $httpCode = 500, $formErrors = null): Response {
        $code   = null;
        $reason = $error;

        if (\Wildix\Error::isErrorCode($error)) {
            $reason = \Wildix\Error::getMessage($error);
            $code   = $error;
        }

        $responseData = array(
            'type' => 'error',
            'reason' => $reason,
            'code' => $code,
            'result' => null,
        );

        if ($formErrors) {
            $responseData['errors'] = $formErrors;
        }

        return $this->response($httpCode, $response, $responseData);
    }

    public function badRequest($errors, Response $response): Response {
        return $this->error(\Wildix\Error::BadRequest, $response, 400, $errors);
    }

    public function warning($reason, Response $response, $result, array $warnings = []): Response {
        $responseData = array(
            'type' => 'warning',
            'reason' => $reason,
            'result' => $result,
        );

        if (!empty($warnings)) {
            $responseData['warnings'] = $warnings;
        }

        return $this->response(200, $response, $responseData);
    }

    public function result($result, Response $response): Response {
        $responseData = [
            'type' => 'result',
            'result' => $result,
        ];
        return $this->response(200, $response, $responseData);
    }

    public function success(Response $response): Response {
        return $this->result(static::SUCCESS, $response);
    }

    /**
     * Return collection result
     *
     * @param array    $list     Collection.
     * @param Response $response Response.
     * @param array    $options  Header options.
     *
     * @return Response
     */
    public function resultList(array $list, Response $response, array $options = []): Response {
        return $this->result(array_merge($options, [self::RECORDS => $list]), $response);
    }

    /**
     * Return collection result
     *
     * @param BaseCollectObject $list     Collection.
     * @param Response          $response Response.
     * @param array             $options  Header options.
     *
     * @return Response
     */
    protected function resultListObject(BaseCollectObject $list, Response $response, array $options = []): Response {
        return $this->resultList($list->toArray(), $response, $options);
    }

    public function resultObject(object $list, Response $response): Response {
        if ($list instanceof ArrayableInterface) {
            $list = $list->toArray();
        }

        return $this->result($list, $response);
    }

    public static function getResponse($data, Response $response, $resultType = 'application/json'): Response {
        $responseType = isset($_REQUEST['responseType']) ? strtolower($_REQUEST['responseType']) : $resultType;

        switch ($responseType) {
            case 'xml':
                $body              = XML::encode($data);
                $contentTypeHeader = 'text/xml; charset=utf-8';
                break;

            default:
                $body              = JSON::encode($data);
                $contentTypeHeader = 'application/json; charset=utf-8';
                break;
        }

        $jsonCallback = $_REQUEST['callback'] ?? '';
        if ($jsonCallback !== '') {
            if ($responseType === 'xml') {
                $body = sprintf("%s('%s');", $jsonCallback, $body);
            } else {
                $body = sprintf('%s(%s);', $jsonCallback, $body);
            }

            $contentTypeHeader = 'application/javascript; charset=utf-8';
        }

        $streamFactory = new StreamFactory();
        $response      = $response->withBody($streamFactory->createStream($body));

        return $response->withHeader('Content-Type', $contentTypeHeader);
    }

    protected function isOldVersion(Request $request) {
        $userAgent             = $request->getServerParams()['HTTP_USER_AGENT'] ?? '';
        $isCompatibleUserAgent = preg_match(
            '/^Collaboration\s\d*\.\d*\.\d*\s\([^;]*;\si(Phone\s)?OS.*$/i',
            htmlspecialchars($userAgent)
        );

        return ($userAgent === 'Android CTIconnect' || $isCompatibleUserAgent);
    }

    protected function translate(string $key): string {
        return $this->getContainer()->get(TranslationStrategyInterface::class)->t($key);
    }

    protected function response($code, Response $response, array $data = []): Response {
        // @TODO decouple?!
        $response = self::getResponse($data, $response);

        return $response->withStatus($code);
    }
}
