<?php

namespace Controllers\Devices;

use Api\Support\BaseCollectObject;
use Cache\FileCache;
use Logger\Log;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\Core\AbstractCrudController;
use Wildix\Core\Container;
use Wildix\Devices\Device;
use Wildix\Devices\Devices;
use Wildix\Devices\Vendors;
use Wildix\Error;
use Wildix\Exception;
use Wildix\InvalidDataException;
use Wildix\License;
use Wildix\Services\Portal;
use Wildix\Services\VideoConference;

class DevicesController extends AbstractCrudController {

    const CACHE_TTL = 30;

    const CACHE_GROUP = 'devices';

    protected $_service = Devices::class;

    public function getItems(Request $request, Response $response): Response {
        $params = $request->getQueryParams();
        $start  = (int)($params['start'] ?? 0);
        $count  = (int)($params['count'] ?? 15);

        $cacheKey        = 'devices_' . json_encode($params);
        $isValidCacheKey = 'devices_is_valid' . $cacheKey;

        $records      = $this->cache()->get($cacheKey, self::CACHE_GROUP);
        $cacheIsValid = $this->cache()->get($isValidCacheKey, self::CACHE_GROUP);
        if ($records === false || $cacheIsValid === false) {
            $records = [];
            $this->cache()->set($isValidCacheKey, date('Y-m-d H:i:s'), self::CACHE_TTL, self::CACHE_GROUP);
            $items = $this->_service()->getItems($start, $count, ['user', 'description', 'settings']);
            foreach ($items as $item) {
                $records[] = $this->_service()->buildItem($item);
            }
            $this->cache()->set($cacheKey, $records, (self::CACHE_TTL * 3), self::CACHE_GROUP);
        }

        return $this->resultList($records, $response);
    }

    public function create(Request $request, Response $response): Response {
        $params = $request->getParsedBody();
        $records = $this->_service()->create(
            $params['macs'] ?? [],
            $request->getUri()->getHost()
        );
        return $this->resultList($records, $response);
    }

    public function synchronization(Request $request, Response $response, array $args): Response {
        $params = $request->getParsedBody();
        $ip = (string)$params['address'] ?? '';
        $this->_service()->synchronizationItem($args['id'], $ip);
        return $this->success($response);
    }

    public function synchronizationItems(Request $request, Response $response): Response {
        $params = $request->getParsedBody();
        $items = (array)$params['items'] ?? [];
        $this->_service()->synchronizationItems($items);
        return $this->success($response);
    }

    public function connect(Request $request, Response $response, array $args): Response {
        $connect = $this->_service()->connect($args['id']);
        return $this->result($connect, $response);
    }

    public function disconnect(Request $request, Response $response, array $args): Response {
        $this->_service()->disconnect($args['id']);
        return $this->success($response);
    }

    public function scanAddress(Request $request, Response $response): Response {
        $params = $request->getParsedBody();
        $res    = [];
        $ranges = [];
        $range  = $params['range'] ?? '';

        $ips = explode(',', $range);

        foreach ($ips as $item) {
            $range    = trim($item, "- , \t\n");
            $ranges[] = $this->_service()->parseIpRange($range);
        }

        foreach ($ranges as $range) {
            $items = $this->_service()->scanIpAddressRange($range['start'], $range['end']);
            if (!empty($items)) {
                array_push($res, ...$items);
            }
        }
        return $this->result(['records' => $res], $response);
    }

    public function verify(Request $request, Response $response): Response {
        $params = $request->getQueryParams();
        $ip   = (string)($params['address'] ?? '');
        $port = (int)($params['port'] ?? 0);
        return $this->result($this->_service()->verifyDevices($ip, $port), $response, false);
    }

    public function login(Request $request, Response $response): Response {
        $params = $request->getParsedBody();
        $mac    = $params['mac'] ?? '';
        $info   = Vendors::getDeviceInfoByMac($mac);

        if ($info['modelLine'] !== 'W-AIR') {
            $this->_service()->update($mac, (array)$params);
        }
        return $this->result('Login Successfu', $response);
    }

    public function logout(Request $request, Response $response): Response {
        $params = $request->getParsedBody() ?? [];
        $mac    = $params['mac'] ?? '';
        $data   = array_merge($params, ['user' => 'unknown']);
        $this->_service()->update($mac, (array)$data);
        return $this->result('Logout Successful', $response);
    }

    public function getWizyconfConfig(Request $request, Response $response): Response {
        $pbx = $this->getContainer()->get(\Models\PBX::class);
        // see device middleware / get device from request attribute
        $device = $request->getAttribute('device');

        $conferenceService = $this->getContainer()->get(VideoConference::class);
        $licenseToken      = $this->getContainer()->get(License::class)->getLimitations('token');

        $assignedUser = $device->getAssignedUsers()[0];

        $assignedUserInfo             = $assignedUser->getAsArray();
        $assignedUserInfo['password'] = $assignedUser->getPassword();

        $deviceAccessTokenLifetime = 60 * 60 * 24 * 30 * 6;
        $userAccessTokenLifetime   = 60 * 60 * 24;

        $deviceAccessToken = $this->_service()->createWizyconfAccessToken(
            $device->getMac(),
            $device->getToken(),
            $deviceAccessTokenLifetime
        );
        $userAccessToken   = $this->_service()->createWizyconfAccessToken(
            $device->getMac(),
            $licenseToken,
            $userAccessTokenLifetime,
            [
                'user' => $assignedUser->getExtension(),
            ]
        );

        $config = [
            'host' => $pbx->getDomain(),
            'securePort' => $pbx->getCustomHttpsPort() ?: 443,
            'refreshToken' => $deviceAccessToken,
            'accessToken' => $userAccessToken,
            'conferenceHost' => $conferenceService->getConfig('host'),
            'user' => $assignedUser->getExtension(),
            'userInfo' => $assignedUserInfo
        ];

        return $this->result($config, $response, false);
    }

    /**
     * Synchronization device with portal.
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     *
     * @return Response
     */
    public function sync(Request $request, Response $response): Response {
        $task = new \BackgroundTask('\Controllers\Devices\DevicesController:progressSyncDevices');
        if ($task->getStatus() == \BackgroundTask::LAUNCHED) {
            $result = 'Background task in progress';
        } else {
            $result = $task->runAsync();
            if ($result !== false) {
                $result = 'success';
            }
        }

        if ($result !== false) {
            return $this->result($result, $response);
        }

        return $this->error('Background task not running', $response);
    }

    public static function progressSyncDevices() {
        $service = Container::get(Devices::class);
        $logger  = Container::get(Log::class);
        if ($service->isDisableSync(Devices::DEVICE_SYNC_PORTAL)) {
            return false;
        }

        $pbxDevices = [];
        $result     = [
                    'updated' => 0,
                    'added' => 0,
                    'removed' => 0
                     ];

        $countDevices = $service->getCount();
        $start        = 0;
        $count        = 500;
        while ($start < $countDevices) {
            $_devices = $service->getDevices(null, null, null, $start, $count);
            foreach ($_devices as $value) {
                $pbxDevices[$value->getMac()] = $value->getAsArray();
            }
            $start += $count;
        }

        if ($countDevices !== count($pbxDevices)) {
            // something wrong: count query different from fetched devices - need interrupt sync
            $logger->warning(
                __LINE__,
                sprintf('Device sync from portal. Count query(%s) is different from fetched devices(%s)', $countDevices, count($pbxDevices))
            );
        }

        $countReceivedDevices = 0;
        $start                = 0;
        $count                = 500;

        $portalService = Container::get(Portal::class);

        do {
            try {
                $portalResult = $portalService->getDevices($start, $count);
                if ($portalResult === false) {
                    $logger->error(
                        __LINE__,
                        sprintf('Device sync from portal. Failed get devices from portal. Result: false')
                    );

                    return array(
                        'reason' => 'Unknown error',
                        'code'   => '1'
                    );
                }
                $portalDevices         = $portalResult['records'];
                $countReceivedDevices += count($portalDevices);
                $totalDevices          = $portalResult['total'];
            } catch (\Exception $e) {
                $logger->error(
                    __LINE__,
                    sprintf('Device sync from portal. Failed get devices from portal. [message]: %s, [code]: %s', $e->getMessage(), $e->getCode())
                );

                return array(
                    'reason' => $e->getMessage(),
                    'code'   => $e->getCode()
                    );
            }

            $logger->info(
                __LINE__,
                sprintf('Device sync from portal. Devices from portal count: %s', is_array($portalDevices) ? count($portalDevices) : 0)
            );

            foreach ($portalDevices as $device) {
                $logger->info(
                    __LINE__,
                    sprintf('Device sync from portal. Processing device from portal: %s', json_encode($device))
                );

                if (empty($device['serial'])) {
                    continue;
                }
                if (!empty($pbxDevices[$device['serial']])) {
                    // device have on this PBX check if from portal update attribute if needed
                    if ($pbxDevices[$device['serial']]['value']['from_wmp'] != 1) {
                        $data = array(
                                'value' => $pbxDevices[$device['serial']]['value']
                        );
                        $data['value']['from_wmp'] = 1;

                        $changed = $service->updateDevice($data, array('id' => $pbxDevices[$device['serial']]['id']));
                        $logger->info(
                            __LINE__,
                            sprintf('Device sync from portal. Update device with mac: "%s", model: "%s"', $device['serial'], $device['model'])
                        );
                        if (is_numeric($changed)) {
                            $result['updated'] += $changed;
                        }
                    }
                    unset($pbxDevices[$device['serial']]);
                } else {
                    $changed = $service->create([$device['serial']]);
                    $logger->info(
                        __LINE__,
                        sprintf('Device sync from portal. Add device with mac: "%s", model: "%s"', $device['serial'], $device['model'])
                    );
                    if (isset($changed[0])) {
                        $result['added'] += 1;
                    }
                }
            }
            $start += $count;
        } while ($countReceivedDevices < $totalDevices);


        if (count($pbxDevices) > 0) {
            // remove
            foreach ($pbxDevices as $serial => $device) {
                $devicePassword          = (string)($device['token'] ?? '');
                $isDeviceShouldBeRemoved = isset($device['value']) && (int)$device['value']['from_wmp'] === 1 && $devicePassword === '';
                $logger->info(
                    __LINE__,
                    sprintf('Device sync from portal. Device with mac "%s" not from Portal. Device details: %s', $device['mac'], json_encode($device))
                );
                // remove if from_wmp(previously received from portal) is 1 and token(password) is empty
                if ($isDeviceShouldBeRemoved && !empty($device['id'])) {
                    try {
                        $deleted = $service->remove($device['mac']);
                        $logger->info(
                            __LINE__,
                            sprintf('Device sync from portal. Device with mac "%s" is removed from DB', $device['mac'])
                        );
                        $result['removed'] += 1;
                    } catch (\Exception $e) {
                        $logger->error(
                            __LINE__,
                            sprintf('Device sync from portal. Can not remove device "%s" from DB: %s', $device['mac'], $e->getMessage())
                        );
                    }
                }
            }
        }

        $logger->info(
            __LINE__,
            sprintf('Device sync from portal. Done. Updated: %d, Added: %d, Removed: %d', $result['updated'], $result['added'], $result['removed'])
        );

        return $result;
    }

    /**
     * Reset a device token by PIN.
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     *
     * @return Response
     * @throws Exception if is invalid PIN (400) or if device not found (404)
     */
    public function resetToken(Request $request, Response $response): Response {
        $params = $request->getParsedBody();
        $pin    = (string)($params['pin'] ?? '');

        if (empty($pin)) {
            throw new Exception(Error::InvalidCode, 400);
        }

        $this->_service()->resetByPin($pin, [Device::TOKEN]);

        return $this->success($response);
    }

    public function result($result, Response $response, $needClearCache = true): Response {
        if ($needClearCache) {
            $this->cache()->clear(self::CACHE_GROUP);
        }

        return parent::result($result, $response);
    }

    private function canDelete(
        $trunkName,
        $device
    ): void {
        $findLinkInDialplan = $this->_service()->findLinkInDialplan($trunkName);
        if ($findLinkInDialplan) {
            $find = new BaseCollectObject($findLinkInDialplan);
            $find = $find->keyBy('rulename');
            throw new InvalidDataException(
                sprintf(
                    'You can\'t remove %s. It\'s linked trunks to dialplan %s',
                    $device->getModel(),
                    $find->keys()->implode(', ')
                ),
                [],
                400
            );
        }
    }

    private function cache(): FileCache {
        return $this->getContainer()->get(FileCache::class);
    }
}
