<?php

namespace Controllers;

use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\System\Dhcp\Leases;
use Wildix\System\Dhcp\ServerManager;

class DhcpServer extends Controller {

    private $_allowedLeasesStatus = [
        Leases::LEASES_STATIC,
        Leases::LEASES_DYNAMIC,
    ];

    public function getSettings(Request $request, Response $response, array $args): Response {
        $config   = $this->_getServerManager()->getActualServerConfig();
        $messages = $this->_getServerManager()->getMessages();

        if (!empty($messages) && (!empty($messages['reason']) || !empty($messages['warnings']))) {
            return $this->warning($messages['reason'], $response, $config, $messages['warnings']);
        } else {
            return $this->result($config, $response);
        }
    }

    public function setSettings(Request $request, Response $response, array $args): Response {
        $settings = $request->getParsedBody();

        if (!isset($settings) || !is_array($settings)) {
            return $this->error('settings are empty', $response);
        }

        $saveResult = $this->_getServerManager()->saveConfig($settings);
        $messages   = $this->_getServerManager()->getMessages();
        $config     = $this->_getServerManager()->getActualServerConfig();

        if (!$saveResult || (!empty($messages) && (!empty($messages['notifications']) || !empty($messages['errors'])))) {
            return $this->warning($messages, $response, $config);
        } else {
            return $this->result($config, $response);
        }
    }

    public function getLeases(Request $request, Response $response, array $args): Response {
        $result = $this->_getServerManager()->getActualLeases();

        return $this->result(
            [
                'records' => $result,
                'total' => count($result),
            ],
            $response
        );
    }

    public function deleteLeases(Request $request, Response $response, array $args): Response {
        $request = $request->getQueryParams();

        if (!isset($request) || !isset($request['leases'])) {
            return $this->error('leases are empty', $response);
        }
        $leases = explode('|', (string)$request['leases']);
        $this->_getServerManager()->removeLeases($leases);

        return $this->result(['records' => $leases], $response);
    }

    public function purgeLeases(Request $request, Response $response, array $args): Response {
        $result = $this->_getServerManager()->purgeLeases();

        return $this->result(['status' => $result], $response);
    }

    public function setLeases(Request $request, Response $response, array $args): Response {
        $request = $request->getParsedBody();

        if (!isset($request) || !is_array($request) || !isset($request['leases']) || !isset($request['status'])) {
            return $this->error('request are empty', $response);
        }
        $leases = explode('|', (string)$request['leases']);

        $status = (string)$request['status'];
        if (!in_array($status, $this->_allowedLeasesStatus)) {
            return $this->error('Not allowed status', $response);
        }
        $this->_getServerManager()->setLeasesStatus($leases, $status);

        return $this->result(['records' => $leases, 'status' => $status], $response);
    }

    /**
     * Get Server manager class
     *
     * @return ServerManager
     */
    private function _getServerManager() {
        return $this->getContainer()->get(ServerManager::class);
    }
}
