<?php
namespace Controllers\Dialplans;

use Controllers\Controller;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\CallweaverManager;

class DbController extends Controller {

    public function getItems(Request $request, Response $response, array $args): Response {
        $originAttribute = $request->getAttribute('originAttribute');
        if ($originAttribute['key']) {
            return $this->getValue($originAttribute['family'], explode('/', $originAttribute['key']), $response);
        }

        return $this->getFamily($originAttribute['family'], $response);
    }

    public function create(Request $request, Response $response, array $args): Response {
        return $this->setValue($request, $response, $args);
    }

    public function update(Request $request, Response $response, array $args): Response {
        return $this->setValue($request, $response, $args);
    }

    public function remove(Request $request, Response $response, array $args): Response {
        $family = $args['family'] ?? '';
        $keys   = $args['key'] ?? '';

        if ($keys) {
            return $this->delValue($family, explode('/', $keys), $response);
        }

        return $this->delFamily($family, $response);
    }

    private function delValue(string $family, array $keys, Response $response): Response {
        if ($this->isReadOnly($family)) {
            return $this->error(sprintf('Failed to update database. Family %s is read only', $family), $response);
        }

        $key = $this->prepareKey($keys);

        $result = $this->getCallweaverManager()->databaseDel($family, $key);

        if ($result === false) {
            return $this->error('Failed to update database', $response);
        }

        return $this->success($response);
    }

    private function delFamily(string $family, Response $response): Response {
        if ($this->isReadOnly($family)) {
            return $this->error(sprintf('Failed to update database. Family %s is read only', $family), $response);
        }

        $result = $this->getCallweaverManager()->databaseDelTree($family);

        if ($result === false) {
            return $this->error('Failed to update database', $response);
        }

        return $this->success($response);
    }

    private function getFamily(string $family, Response $response): Response {
        $result = $this->getCallweaverManager()->databaseGetFamilyTree($family);
        if ($result === false) {
            return $this->error(sprintf('Family %s not found', $family), $response, 404);
        }

        return $this->result(
            [
                'total'   => count($result),
                'records' => $result
            ],
            $response
        );
    }

    private function getValue(string $family, array $keys, Response $response): Response {
        $key = $this->prepareKey($keys);

        $value = $this->getCallweaverManager()->databaseGet($family, $key);

        if ($value === false) {
            return $this->error(sprintf('Key %s not found', $key), $response, 404);
        }

        return $this->result($value, $response);
    }

    private function setValue(Request $request, Response $response, array $args): Response {
        $family = $args['family'] ?? '';
        $keys   = $args['key'] ?? '';
        $params = $request->getParsedBody();

        if ($this->isReadOnly($family)) {
            return $this->error(sprintf('Failed to update database. Family %s is read only', $family), $response);
        }

        $key   = $this->prepareKey(explode('/', $keys));
        $value = trim($params['value'] ?? '');

        $result = $this->getCallweaverManager()->databasePut($family, $key, $value);

        if ($result === false) {
            return $this->error('Failed to update database', $response);
        }

        return $this->success($response);
    }

    private function prepareKey($keys): string {
        $last = array_pop($keys);
        if (!empty($last)) {
            $keys[] = $last;
        }

        return join('/', $keys);
    }

    private function isReadOnly(string $family): bool {
        return in_array(strtolower($family), ['customdevstate', 'sip', 'queue']);
    }

    private function getCallweaverManager(): CallweaverManager {
        return $this->getContainer()->get(CallweaverManager::class);
    }
}
