<?php
namespace Controllers\Dialplans;

use Api\Dialplans\Rules\Imports\Interfaces\ImportServiceInterface;
use Api\Dialplans\Rules\Imports\Request\ImportRequest;
use Controllers\Controller;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Slim\Psr7\UploadedFile;
use Wildix\Dialplan\ApplicationService;
use Wildix\Dialplan\DialplanService;
use Wildix\Dialplan\Rules;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\Validator;

class RulesController extends Controller {

    /**
     * Get dialplans
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     *
     * @return Response
     *
     * @throws \Exception Exception.
     */
    public function getItems(Request $request, Response $response): Response {
        $result = [];
        $data   = $this->getDialplanService()->getDialplans();
        foreach ($data as $item) {
            $result[] = $this->_dialplanFormatter($item);
        }

        return $this->result(['records' => $result], $response);
    }

    /**
     * Get dialplan config.
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     * @param array    $args     Arguments.
     *
     * @return Response
     *
     * @throws \Wildix\Exception Exception.
     */
    public function getItem(Request $request, Response $response, array $args): Response {
        $result = $this->getDialplanService()->getDialplanById($args['id']);
        if ($result === null) {
            throw new \Wildix\Exception(\Wildix\Error::DialplanNotFound, 404);
        }
        $result  = $this->_dialplanFormatter($result);
        $result += $this->getDialplanService()->getAdditionalData($result[\Wildix\Dialplan\DialplanService::PARAM_ID]);

        return $this->result($result, $response);
    }

    /**
     * Format dialplan description
     *
     * @param array $data Dialplan data.
     *
     * @return array
     */
    private function _dialplanFormatter(array $data): array {
        $predefinedDescription = [
            'users',
            'main',
            'classound',
            'kite',
        ];

        $data[DialplanService::PARAM_DESCRIPTION] = $data[DialplanService::PARAM_DESCRIPTION] ?? '';
        if (in_array($data[DialplanService::PARAM_NAME], $predefinedDescription, true)) {
            $data[DialplanService::PARAM_DESCRIPTION] = $this->translate('dialplan.rule.' . $data[DialplanService::PARAM_NAME]);
        }

        return $data;
    }

    /**
     * Import
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     *
     * @return Response
     *
     * @throws InvalidDataException Exception.
     */
    public function import(Request $request, Response $response): Response {
        $validator = new Validator($request->getUploadedFiles());
        $validator->rule('instanceOf', ['importDialplanFile'], UploadedFile::class);
        $validator->rule('required', ['importDialplanFile']);
        if (!$validator->validate()) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $validator->errors());
        }

        $importRequest = ImportRequest::make($request);
        $this->getImportService()->validateBackupFile($importRequest);
        if (!$importRequest->isDialPlanFile()) {
            throw new InvalidDataException('Import failed! Not valid file backup');
        }

        $this->getImportService()->dropBackup();
        if ($this->getImportService()->moveUploadedImportFile($importRequest)) {
            $this->getImportService()->importRules($importRequest);
            $this->getImportService()->updateFromNewTables();
        }

        $result = $this->getImportService()->getAllNameFromBackup();
        if ($result->getRuleNames()->isEmpty()) {
            throw new InvalidDataException('Import failed! Unsuccessful attempt to load rules. No rule names found');
        }
        return $this->result($result->toArray(), $response);
    }

    /**
     * Restore backup
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     *
     * @return Response
     */
    public function restoreBackup(Request $request, Response $response): Response {
        $params = $request->getQueryParams();
        $return = $this->getImportService()->restoreBackup($params['idNameRules'] ?? []);

        return $this->resultList($return, $response);
    }

    /**
     * Create dialplan.
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     *
     * @return Response
     *
     * @throws InvalidDataException Exception.
     * @throws \Wildix\Exception    Exception.
     */
    public function create(Request $request, Response $response): Response {
        $params = $request->getParsedBody();

        $dialplanId = $this->getDialplanService()->createDialplan(
            (string)$params[DialplanService::PARAM_NAME] ?? '',
            (string)$params[DialplanService::PARAM_DESCRIPTION] ?? '',
            (array)$params[DialplanService::PARAM_NUMBERS] ?? [],
            (array)$params[DialplanService::PARAM_INCLUDES] ?? []
        );

        return $this->getItem($request, $response, ['id' => $dialplanId]);
    }

    /**
     * Update dialplan by ID.
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     * @param array    $args     Arguments.
     *
     * @return Response
     *
     * @throws InvalidDataException Exception.
     * @throws \Wildix\Exception    Exception.
     */
    public function update(Request $request, Response $response, array $args): Response {
        $params = $request->getParsedBody();

        $this->getDialplanService()->updateDialplan(
            $args['id'],
            (string)$params[DialplanService::PARAM_DESCRIPTION] ?? '',
            (array)$params[DialplanService::PARAM_NUMBERS] ?? [],
            (array)$params[DialplanService::PARAM_INCLUDES] ?? []
        );

        return $this->getItem($request, $response, ['id' => $args['id']]);
    }

    /**
     * Delete dialplans by IDs.
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     * @param array    $args     Arguments.
     *
     * @return Response
     *
     * @throws \Exception Exception.
     */
    public function remove(Request $request, Response $response, array $args): Response {
        $ids = array_filter(explode(',', $args['ids']));

        $this->getDialplanService()->deleteDialplans($ids);

        return $this->success($response);
    }

    /**
     * Return applications list.
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     * @param array    $args     Arguments.
     *
     * @return Response
     */
    public function getApplications(Request $request, Response $response, array $args): Response {
        $params      = $request->getQueryParams();
        $withConfigs = (bool)$params['withConfigs'] ?? false;
        $data        = $this->getApplicationService()->getApplications($withConfigs);
        return $this->result(['records' => $data], $response);
    }

    /**
     * Export dialplans.
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     *
     * @return Response
     *
     * @throws InvalidDataException Exception.
     */
    public function export(Request $request, Response $response): Response {
        $params = $request->getParsedBody();
        $result = $this->service()->export($params['rulesIds'] ?? 0);

        return $this->result($result, $response);
    }

    public function pingExport(Request $request, Response $response): Response {
        $params    = $request->getQueryParams();
        $validator = new Validator(['exportedFile' => $params['exportedFile']]);
        $validator->rule('required', ['exportedFile']);
        if (!$validator->validate()) {
            throw new InvalidDataException('Validation error', $validator->errors());
        }

        $result = $this->service()->pingExport($params['exportedFile']);

        return $this->result($result, $response);
    }

    /**
     * Download backup dialplan.
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     *
     * @return Response
     *
     * @throws InvalidDataException Exception.
     */
    public function downloadBackup(Request $request, Response $response): Response {
        $exportedFile = $request->getQueryParams()['exportedFile'];

        if (strpos($exportedFile, '../') !== false) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR));
        }

        $backup = Rules::TMP_BACKUP_DIR . $exportedFile;

        if (!file_exists($backup)) {
            return $this->error(\Wildix\Error::FileNotFound, $response);
        }

        ini_set('zlib.output_compression', 'Off');
        return $response->withHeader('Content-Disposition', 'attachment; filename="' . $exportedFile . '"')
            ->withHeader('X-Accel-Redirect', '/inner-storage' . $backup);
    }

    private function getDialplanService(): DialplanService {
        return $this->getContainer()->get(DialplanService::class);
    }

    private function getImportService(): ImportServiceInterface {
        return $this->getContainer()->get(ImportServiceInterface::class);
    }

    private function getApplicationService(): ApplicationService {
        return $this->getContainer()->get(ApplicationService::class);
    }

    private function service(): Rules {
        return $this->getContainer()->get(Rules::class);
    }
}
