<?php
namespace Controllers;

use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Slim\Psr7\Factory\StreamFactory;
use Slim\Psr7\Factory\UploadedFileFactory;
use Slim\Psr7\UploadedFile;
use Wildix\Error;
use Wildix\Exception;
use Wildix\Faxes\Faxes;
use Wildix\InvalidDataException;
use Wildix\Validator;

/**
 * REST API Controller for sending faxes
 */
class FaxController extends Controller {

    protected string $licenceAcl = 'faxes';

    public function sendFax(Request $request, Response $response) {
        $user     = $this->getUser();
        $file     = $request->getUploadedFiles();
        $params   = $request->getParsedBody();
        $fileName = $this->service()->getSafeFilename($params['fileName'] ?? '');
        if (isset($file['fileName'])) {
            $fileName = $this->uploadFile($file['fileName'])['filename'];
        }

        $faxNumbers = $params['faxNumbers'] ?? [];
        $postpone   = $params['postpone'] ? $params['postponeDate'] : null;
        $faxCover   = [
            'faxCover'        => Validator::isTrue($params['faxCover']),
            'to'              => $params['faxCoverTo'],
            'toPhone'         => $params['faxCoverToPhone'],
            'toCompany'       => $params['faxCoverToCompany'],
            'from'            => $params['faxCoverFrom'],
            'fromPhone'       => $params['faxCoverFromPhone'],
            'fromCompany'     => $params['faxCoverFromCompany'],
            'subject'         => $params['faxCoverSubject'],
            'notes'           => $params['faxCoverNotes'],
            'fromFax'         => $params['faxCoverFromFax'],
            'checkboxUrgent'  => $params['faxCoverCheckboxUrgent'],
            'checkboxReview'  => $params['faxCoverCheckboxReview'],
            'checkboxComment' => $params['faxCoverCheckboxComment'],
            'checkboxReply'   => $params['faxCoverCheckboxReply'],
            'checkboxForward' => $params['faxCoverCheckboxForward']
        ];

        $fileExists = $fileName && file_exists(\Wildix\Faxes\Faxes::FAXCOVER_PATH . $fileName);

        $result = $this->service()->sendFax($faxNumbers, $postpone, $faxCover, $fileName, $user, $request);
        if ($result !== true) {
            return $this->warning('Error send fax', $response, ['code' => $result, 'fileExists' => $fileExists]);
        }

        return $this->result(['status' => 'Success', 'fileExists' => $fileExists], $response);
    }

    public function uploadFax(Request $request, Response $response) {
        $file   = $request->getUploadedFiles();
        $params = $request->getParsedBody();
        if (isset($file['qqfile'])) {
            $result = $this->service()->uploadFax($file['qqfile']);
        } else {
            $validator = new Validator($params);
            $validator->rule('required', 'name')->reason('Name file is empty');
            $validator->rule('required', 'size')->reason('Size file is empty');
            if (!$validator->validate()) {
                throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $validator->errors());
            }

            $filename = $this->service()->getSafeFilename($params['name'] ?? '');
            $size     = $params['size'];
            $end      = $params['end'] ?? true;
            $chunks   = $params['chunks'] ?? [];

            $fileExtension = pathinfo($filename, PATHINFO_EXTENSION);

            $data = $this->makeChunk($chunks);

            $result     = $filename;
            $chunksSize = strlen($data);
            $tmpFile    = sys_get_temp_dir() . DIRECTORY_SEPARATOR . md5($filename) . ($fileExtension ? '.' . $fileExtension : '');

            if (file_put_contents($tmpFile, $data, FILE_APPEND) != $chunksSize) {
                throw new Exception('error_disk');
            } else if (Validator::isTrue($end)) {
                $newStream      = new StreamFactory();
                $newUploadFiles = new UploadedFileFactory();
                $stream         = $newStream->createStreamFromFile($tmpFile);
                $file           = $newUploadFiles->createUploadedFile($stream, $size, UPLOAD_ERR_OK, $filename, $fileExtension);

                $result = $this->service()->uploadFax($file);
            }
        }

        return $this->result($result, $response);
    }

    public function getQueue(Request $request, Response $response) {
        $result = $this->service()->getQueueFiles();

        return $this->result($result, $response);
    }

    private function makeChunk(array $chunks): string {
        $data = '';
        foreach ($chunks as $chunk) {
            $data .= base64_decode($chunk);
        }

        return $data;
    }

    private function uploadFile(UploadedFile $file): array {
        return $this->service()->uploadFax($file);
    }

    private function service(): Faxes {
        return $this->getContainer()->get(Faxes::class);
    }
}
