<?php

namespace Controllers;

use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\BatchOperationsException;
use Wildix\Colleagues\Group;
use Wildix\Error;

class Groups extends Controller {
    public function getItems(Request $request, Response $response, array $args): Response {
        $result = [];

        $user   = $this->getUser();
        $groups = $this->serviceGroups()->getGroups($user);

        if ($groups) {
            foreach ($groups as $group) {
                $groupData = $group->getGroupAsArray(Group::getFields());

                $result[] = [
                    'dn' => $groupData[Group::DN],
                    'id' => $groupData[Group::ID],
                    'name' => $groupData[Group::NAME],
                    'wcgrp' => $groupData[Group::WCGRP],
                ];
            }
        }

        return $this->result(
            [
                'total' => count($result),
                'records' => $result,
            ],
            $response
        );
    }

    public function getManagedGroups(Request $request, Response $response, array $args): Response {
        $params = $request->getQueryParams();

        $fields = (string)($params['fields'] ?? '');
        $fields = ($fields !== '') ? explode(',', $fields) : [];

        return $this->result(['records' => $this->serviceGroups()->getManagedGroups($fields)], $response);
    }

    public function create(Request $request, Response $response, array $args): Response {
        $groupData = $this->collectGroupData($request);

        $groupDn = $this->serviceGroups()->saveGroup($groupData);
        $group   = $this->serviceGroups()->getGroupByDn(
            $groupDn,
            [
                Group::DN,
                Group::ID,
                Group::NAME,
                Group::WCGRP,
                Group::OWNER,
                Group::ACL_WG2G,
            ]
        );

        if (!$group instanceof Group) {
            return $this->error(Error::UnknownError, $response);
        }

        return $this->result($this->serviceGroups()->normalizeGroupData($group->getGroupAsArray()), $response);
    }

    public function update(Request $request, Response $response, array $args): Response {
        $id = (int)$args['id'];

        $group = $this->serviceGroups()->getGroupById($id, [Group::DN, Group::NAME]);

        if (!isset($group)) {
            return $this->error(sprintf('Group %d is not found.', $id), $response, 404);
        }

        $groupData = $this->collectGroupData($request);

        $groupData[\Wildix\Colleagues\Groups::FIELD_DN] = $group->getDn();

        if (!$this->getUser()->isRoot()) {
            $groupData[\Wildix\Colleagues\Groups::FIELD_NAME] = $group->getName();
            unset($groupData[\Wildix\Colleagues\Groups::FIELD_WCGRP], $groupData[\Wildix\Colleagues\Groups::FIELD_INHERITS]);
        } else if (!isset($groupData[\Wildix\Colleagues\Groups::FIELD_NAME])) {
            $groupData[\Wildix\Colleagues\Groups::FIELD_NAME] = $group->getName();
        }

        $this->serviceGroups()->saveGroup($groupData);

        return $this->success($response);
    }

    public function remove(Request $request, Response $response, array $args): Response {
        $ids = $args['ids'];

        $errors = [];

        foreach (explode(',', $ids) as $id) {
            $group = $this->serviceGroups()->getGroupById((int)$id, [Group::DN, Group::NAME]);

            if (!isset($group)) {
                $errors[$id] = Error::ACL_GROUP_NOT_FOUND;

                continue;
            }
            try {
                $this->serviceGroups()->removeGroup($group);
            } catch (\Exception $ex) {
                $errors[$id] = $ex->getCode();
            }
        }

        if (count($errors) > 0) {
            throw new BatchOperationsException($errors);
        }

        return $this->success($response);
    }

    public function getRules(Request $request, Response $response, array $args): Response {
        $rules = $this->serviceGroups()->getRules();

        $abilitySetsForUse    = [\Wildix\Colleagues\Groups::RULE_ABILITY_OPTIONS_YES, \Wildix\Colleagues\Groups::RULE_ABILITY_OPTIONS_USE];
        $abilitySetsForManage = [\Wildix\Colleagues\Groups::RULE_ABILITY_OPTIONS_SET];

        $getAccessNameFromTranslation = function (string $prefix, string $ruleKey): string {
            return $this->translate($prefix . '.' . $ruleKey);
        };
        $callClasses                  = get_call_classes_no_assoc();

        foreach ($rules as &$subset) {
            $translationPrefix = null;

            if (in_array($subset[\Wildix\Colleagues\Groups::RULES_SET_ABILITY], $abilitySetsForUse, true)) {
                $translationPrefix = 'acl';
            } else if (in_array($subset[\Wildix\Colleagues\Groups::RULES_SET_ABILITY], $abilitySetsForManage, true)) {
                $translationPrefix = 'features';
            }

            foreach ($subset[\Wildix\Colleagues\Groups::RULES_SET_ACCESS] as &$item) {
                if (isset($translationPrefix)) {
                    $itemName = $getAccessNameFromTranslation($translationPrefix, $item[\Wildix\Colleagues\Groups::RULES_SET_ACCESS_KEY]);
                } else {
                    $itemName = $callClasses[$item[\Wildix\Colleagues\Groups::RULES_SET_ACCESS_KEY]] ?? $item[\Wildix\Colleagues\Groups::RULES_SET_ACCESS_KEY];
                }

                $item['name'] = $itemName;
            }
        }

        unset($subset, $item);

        return $this->result(
            [
                'list' => $rules,
                'default' => $this->serviceGroups()->getDefault(true),
            ],
            $response->withHeader('Cache-Control', 'private, max-age=1800')
        );
    }

    private function serviceGroups(): \Wildix\Colleagues\Groups {
        static $repository;

        if (!isset($repository)) {
            $repository = $this->getContainer()->get(\Wildix\Colleagues\Groups::class);
            $repository->setPermissions($this->getContainer()->get(\WMS_Permissions::class));
        }

        return $repository;
    }

    private function collectGroupData(Request $request) {
        $groupData = $request->getParsedBody();

        if (!empty($groupData[\Wildix\Colleagues\Groups::FIELD_RULES]) && $groupData[\Wildix\Colleagues\Groups::FIELD_RULES] === [0 => '']) {
            // Empty array can't be sent through form data.
            // Value for next data "paramName[]=" will be converted to "array(0 => '')".
            // The value replaced with an empty array.
            $groupData[\Wildix\Colleagues\Groups::FIELD_RULES] = [];
        }

        return $groupData;
    }
}
