<?php
namespace Controllers;

use Models\PBX;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\Colleagues\Colleague;
use Wildix\Colleagues\Group;
use Wildix\Databases\DEVDB;
use Wildix\Dialplan\ExtensionsVariables;
use Wildix\Dialplan\FeatureCodes\FeatureConfigFile;
use Wildix\Dialplan\PagingGroups\PagingGroupsStorage;
use Wildix\Services\GroupChat;
use Wildix\Settings\CDR\General;
use Wildix\SipRtp;

/**
 * IOS Controller
 */
class Ios extends Controller {

    public function getSettings(Request $request, Response $response, array $args): Response {
        $need = $this->_defineRequestedConfigs($request);

        $data = $this->_getConfigRelatedToCurrentUser($need, $request);

        if (isset($need['dialplanSettings'])) {
            $data['dialplanSettings'] = [
                FeatureConfigFile::TRANSFER_DIGIT_TIMEOUT => $this->configFeature()->get(FeatureConfigFile::TRANSFER_DIGIT_TIMEOUT, FeatureConfigFile::GENERAL_SECTION),
                FeatureConfigFile::TRANSFER_INTER_DIGIT_TIMEOUT => $this->configFeature()->get(FeatureConfigFile::TRANSFER_INTER_DIGIT_TIMEOUT, FeatureConfigFile::GENERAL_SECTION),
                ExtensionsVariables::END_OF_DIGIT => $this->serviceExtensionsVariables()->get(ExtensionsVariables::END_OF_DIGIT)
            ];
        }

        if (isset($need['cdrSettings'])) {
            $data['cdrSettings'] = [
                'save_chat_history' => $this->configGeneral()->getBoolean(General::SAVE_CHAT_HISTORY)
            ];
        }

        if (isset($need['pbxInfo'])) {
            $data['pbxInfo'] = $this->getPbxInfo();
        }

        if (isset($need['wideband'])) {
            $data['wideband'] = $this->servicePbx()->isWideband($_SERVER['REMOTE_ADDR']);
        }

        if (isset($need['sip-rtp'])) {
            $data['sip-rtp'] = $this->_getSipRtpSettings();
        }

        if (isset($need['conference'])) {
            $data['conference'] = [
                'groupChatDomain' => $this->serviceGroupChat()->getGroupChatDomain(),
                'smsBotDomain' => $this->serviceGroupChat()->getSmsBotDomain()
            ];
        }

        return $this->result($data, $response);
    }

    private function getPbxInfo() {
        $pbxInfo = [
            'networkAddresses' => [],
            'ports' => [],
            'externalLinePrefix' => $this->servicePbx()->getExternalLinePrefix(),
            'version' => $this->servicePbx()->getVersion(),
            'serial' => $this->servicePbx()->getSerial()
        ];

        $pbxInfo['networkAddresses'] = $this->servicePbx()->getCandidates();

        $pbxInfo['ports'][] = '443';
        $port               = $this->servicePbx()->getCustomHttpsPort();
        if ($port) {
            $pbxInfo['ports'][] = $port;
        }
        $pbxInfo['ports'] = array_values(array_unique($pbxInfo['ports']));

        // failovers
        $failovers = $this->servicePbx()->getFailovers();
        if (count($failovers)) {
            foreach ($failovers as $pbx) {
                $net = [];
                if (isset($pbx['description']) && $pbx['description']['pbxData']) {
                    $net = $pbx['description']['pbxData']['addrSet']['list'];
                }
                $net[] = $pbx['domain'];
                usort($net, ['Models\PBX', 'cmpNetworkAddresses']);

                $ports = ['443'];
                if (isset($pbx['publicPort']) && $pbx['publicPort'] != '443') {
                    $ports[] = $pbx['publicPort'];
                }
                $pbxInfo['failovers'][] = [
                    'networkAddresses' => $net,
                    'ports' => $ports,
                ];
            }
        }

        return $pbxInfo;
    }

    private function getPersonalInfo(Colleague $user) {
        $personalInfo = array_merge($user->getAsArray(), ['sipPassword' => $user->getSipPassword()]);

        if (!empty($_SESSION['withAdminPwd'])) {
            $personalInfo['password'] = $user->getPassword();
        }

        if (isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] == 'on') {
            $personalInfo['xmppPassword'] = $user->getPassword();
        }

        return $personalInfo;
    }

    private function getColleagues(Colleague $user) {
        $resColleagues = [];
        $colleagues    = $this->serviceColleagues()->getUsers(null, null, null, 0, 1500, 'id', 'asc', $user);
        if ($colleagues) {
            foreach ($colleagues as $colleague) {
                $resColleagues[] = $colleague->getAsArray();
            }
        }

        return $resColleagues;
    }

    private function getGroups(Colleague $user) {
        $resGroups = [];
        $groups    = $this->serviceGroups()->getGroups($user);
        if ($groups) {
            foreach ($groups as $group) {
                $groupData   = $group->getGroupAsArray(Group::getFields());
                $resGroups[] = [
                    'dn' => $groupData[Group::DN],
                    'id' => $groupData[Group::ID],
                    'name' => $groupData[Group::NAME],
                    'wcgrp' => $groupData[Group::WCGRP],
                ];
            }
        }

        return $resGroups;
    }

    private function _defineRequestedConfigs(Request $request): array {
        $params = $request->getQueryParams();

        $fields = $params['fields'] ?? [
                'groups',
                'colleagues',
                'personalInfo',
                'pbxInfo',
                'wideband',
                'dialplanSettings',
                'cdrSettings',
                'sip-rtp',
                'personalSettings',
                'pagingGroups',
                'conference',
            ];

        if (!is_array($fields)) {
            $fields = explode(',', $fields);
            $fields = array_map('trim', $fields);
        }

        return array_combine($fields, $fields);
    }

    private function _getConfigRelatedToCurrentUser(array $need, Request $request) {
        $params = $request->getQueryParams();

        $data = [];

        $user = $this->getUser();

        if (isset($need['acl'])) {
            $filter      = $params['filter'] ?? [];
            $data['acl'] = $user->getAclByFilter($filter);
        }

        if (isset($need['groups'])) {
            $data['groups'] = $this->getGroups($user);
        }

        if (isset($need['colleagues'])) {
            $data['colleagues'] = $this->getColleagues($user);
        }

        if (isset($need['personalInfo'])) {
            $data['personalInfo'] = $this->getPersonalInfo($user);
        }

        if (isset($need['personalSettings'])) {
            $data['personalSettings'] = $this->_getPersonalSettings($user, $request);
        }

        if (isset($need['pagingGroups'])) {
            $pagingGroups         = PagingGroupsStorage::getForUser($user->getExtension());
            $data['pagingGroups'] = [
                'total' => count($pagingGroups),
                'records' => $pagingGroups
            ];
        }

        return $data;
    }

    private function _getSipRtpSettings() {
        $jbSettings = [
            'min' => $this->serviceDevDb()->getSystemVariable(DEVDB::SYSTEM_JITTER_MIN, '50'),
            'average' => $this->serviceDevDb()->getSystemVariable(DEVDB::SYSTEM_JITTER_NOM, '100'),
            'max' => $this->serviceDevDb()->getSystemVariable(DEVDB::SYSTEM_JITTER_MAX, '200'),
        ];

        $sipOpts = $this->serviceSipRtp()->getGeneralSip();

        return [
            'jitter' => $jbSettings,
            'DSCP' => [
                'RTP' => $sipOpts['tos'] ?? '0',
                'SIP' => $sipOpts['tos_sip'] ?? '0',
            ],
        ];
    }

    private function _getPersonalSettings(Colleague $user, Request $request) {
        $personalSettingsFields = [
            'features',
            'personal',
            'ringtones',
            'fkeys',
            'sharedvm',
            'cwtone',
        ];

        return $user->getSettings()->groupByFields($personalSettingsFields, $this->isOldVersion($request));
    }

    private function serviceGroupChat(): GroupChat {
        return $this->getContainer()->get(GroupChat::class);
    }

    private function servicePbx(): PBX {
        return $this->getContainer()->get(PBX::class);
    }

    private function serviceColleagues(): \Wildix\Colleagues\Colleagues {
        return $this->getContainer()->get(\Wildix\Colleagues\Colleagues::class);
    }

    private function serviceGroups(): \Wildix\Colleagues\Groups {
        return $this->getContainer()->get(\Wildix\Colleagues\Groups::class);
    }

    private function serviceSipRtp(): SipRtp {
        return $this->getContainer()->get(SipRtp::class);
    }

    private function serviceDevDb(): DEVDB {
        return $this->getContainer()->get(DEVDB::class);
    }

    private function serviceExtensionsVariables(): ExtensionsVariables {
        return $this->getContainer()->get(ExtensionsVariables::class);
    }

    private function configFeature() {
        return $this->getContainer()->get(FeatureConfigFile::class);
    }

    private function configGeneral(): General {
        return $this->getContainer()->get(General::class);
    }
}
