<?php
namespace Controllers;

use InvalidArgumentException;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\Dialplan\ExtensionsVariables;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\LanguageAndRegion;
use Wildix\SipRtp;
use Wildix\System\Ntp as NtpService;
use Wildix\Validator;

class LanguageAndRegionController extends Controller {

    public function getSettings(Request $request, Response $response) {
        $data = array();

        $tmp = $this->serviceSipRtp()->getGeneralSip();
        foreach (array('language', 'tonezone') as $val) {
            if (isset($tmp[$val])) {
                $data[$val] = $tmp[$val];
            }
        }

        $tmp = $this->serviceNtp()->getTimezone();

        if ($tmp) {
            $data['timezone'] = $tmp;
        }

        $data['installedLangs']             = $this->service()->getInstalledLanguages();
        $data['maxInstall']                 = count($data['installedLangs']) + $this->service()->getCountInstallLangsForMemory();
        $data['countrycode']                = $this->serviceExtensionsVariables()->get(ExtensionsVariables::COUNTRY_CODE);
        $data['lists']['tonezones']         = $this->service()->getListTonezones();
        $data['lists']['timezones']         = $this->serviceNtp()->getListTimezones();
        $data['lists']['langs_for_install'] = $this->service()->getListLanguagesForInstall();

        return $this->result($data, $response);
    }

    public function setSettings(Request $request, Response $response) {
        $maxInstall   = $this->service()->getCountInstallLangsForMemory();
        $allowedLangs = $this->service()->getListLanguagesForInstall();
        $params       = $request->getParsedBody();
        $post         = [
            'language'    => $params['language'] ?? 'en',
            'tonezone'    => $params['tonezone'] ?? '',
            'area'        => $params['area'],
            'zone'        => $params['zone'],
            'countrycode' => $params['countrycode'],
            'langs'       => $params['lang']
        ];

        $validator = new Validator($post);

        $sip = $this->serviceSipRtp()->getGeneralSip();
        unset($sip['language'], $sip['tonezone']);
        if (isset($post['language'])) {
            $sip['language'] = $post['language'];
        }

        if (isset($post['tonezone'])) {
            $sip['tonezone'] = $post['tonezone'];
        }

        if ($sip['tonezone'] && !array_key_exists($sip['tonezone'], $this->service()->getListTonezones())) {
            throw new InvalidArgumentException('Not allowed Tone Zone');
        }

        $validator->rule('in', 'language', $allowedLangs)->reason('Not allowed Language');
        $validator->rule('required', 'langs')->reason('Sounds packages is empty');
        $validator->rule('required', 'language')->reason('Not allowed Language');
        $validator->rule(
            function ($field, $value) use ($allowedLangs) {
                $value = $value ?? [];
                return !count(array_diff($value, $allowedLangs));
            },
            'langs',
            'Not allowed sound package'
        );
        $validator->rule(
            function ($field, $value) use ($maxInstall) {
                $value = $value ?? [];
                return count($value) < $maxInstall;
            },
            'langs',
            sprintf('Select max %d sound package', $maxInstall)
        );

        $validator->rule('CountryCode', 'countrycode')->message(Error::CountryCode);
        $validator->rule('required', 'countrycode')->message(Error::FieldRequired);


        if (!$validator->validate()) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $validator->errors());
        }


        if ($this->service()->setLanguages($post['langs']) !== true) {
            throw new InvalidArgumentException('Error install sounds');
        }

        if ($this->serviceSipRtp()->setGeneralSip($sip) !== true) {
            throw new InvalidArgumentException('Not save sip config');
        }

        if (!$this->serviceNtp()->hasTimezone($post['area'], $post['zone']) || !$this->serviceNtp()->setTimezone($post['area'], $post['zone'])) {
            throw new InvalidArgumentException('Not save timezone');
        }

        $this->serviceExtensionsVariables()->set(ExtensionsVariables::COUNTRY_CODE, intval($post['countrycode']));
        if ($this->serviceExtensionsVariables()->save() !== true) {
            throw new InvalidArgumentException('Not save country code');
        }


        exec('callweaver -rx "dialplan reload"');
        \PBX::getInstance()->put_cli_reload('pbx_lua');
        \PBX::getInstance()->put_cli_reload('sip');

        return $this->success($response);
    }

    public function getStatus(Request $request, Response $response) {
        if ($this->service()->getStatus() === true) {
            return $this->result(['processes' => 1], $response);
        } else {
            return $this->success($response);
        }
    }

    private function service(): LanguageAndRegion {
        return $this->getContainer()->get(LanguageAndRegion::class);
    }

    private function serviceNtp(): NtpService {
        return $this->getContainer()->get(NtpService::class);
    }

    private function serviceSipRtp(): SipRtp {
        return $this->getContainer()->get(SipRtp::class);
    }

    private function serviceExtensionsVariables(): ExtensionsVariables {
        return $this->getContainer()->get(ExtensionsVariables::class);
    }
}
