<?php
namespace Controllers;

use Models\Note;
use Models\Notes as Model;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\Phonebooks\Contacts;
use Wildix\Validator;

class Notes extends Controller {

    public function getItems(Request $request, Response $response): Response {
        $params = $request->getQueryParams();
        $filter = [];

        // dateFrom comes with d/m/Y H:i format, but also can be with seconds
        $dateFrom = strtotime(str_replace('/', '-', $params['dateFrom']));

        if ($dateFrom !== false) {
            $filter['dateFrom'] = $dateFrom;
        }

        $notes    = $this->service()->getNotes($filter, 100, 'DESC');
        $arrNotes = [];
        foreach ($notes as $note) {
            $arrNotes[] = $note->getAsArray();
        }

        return $this->result($arrNotes, $response);
    }

    public function remove(Request $request, Response $response, array $args): Response {
        $user = $this->getUser();

        $note = $this->service()->getNoteByID($args['noteId']);

        if (!$note) {
            return $this->error('Note not found', $response, 404);
        }

        if ($note->getOwner() != $user->getExtension()) {
            return $this->error('Forbidden', $response, 403);
        }

        $result = $this->service()->deleteNote($note);


        if ($result) {
            return $this->result(['contactId' => '', 'noteId' => $args['noteId']], $response);
        }

        return $this->error('Note not deleted', $response, 501);
    }

    public function getNotesByContact(Request $request, Response $response, array $args): Response {
        $contact = $this->serviceContacts()->getByID($args['contactId']);
        if (!empty($contact)) {
            $arrNotes = [];
            $notes    = $this->service()->getLastNotesByContactID($contact->getID(), 100, 'DESC');

            foreach ($notes as $note) {
                $arrNotes[] = $note->getAsArray();
            }

            return $this->result(['contact' => $contact->getAsArray(), 'notes' => $arrNotes], $response);
        }

        return $this->result($args['contactId'], $response);
    }

    public function addNoteToContact(Request $request, Response $response, array $args): Response {
        $params   = $request->getParsedBody();
        $user     = $this->getUser();
        $noteBody = trim($params['note']);
        $contact  = $this->serviceContacts()->getByID($args['contactId']);
        if (empty($contact)) {
            return $this->error('Contact not found', $response, 404);
        }
        $validator = new Validator($params, ['note', 'contactId']);
        $validator->rule('required', ['note', 'contactId']);

        if (!$validator->validate()) {
            throw new InvalidDataException(Error::VALIDATION_ERROR, $validator->errors());
        }

        $note = new Note();

        $note->setContactID($args['contactId']);
        $note->setBody($noteBody);
        $note->setTime(time());
        $note->setOwner($user->getExtension());

        if ($this->service()->saveNote($note)) {
            return $this->result(['contactId' => $args['contactId'], 'note' => $note->getAsArray()], $response);
        }

        return $this->error('Note not added', $response, 501);
    }

    public function deleteNoteFromContact(Request $request, Response $response, array $args): Response {
        $user = $this->getUser();
        $note = $this->service()->getNoteByID($args['noteId']);

        if (!$note) {
            return $this->error('Note not found', $response, 404);
        }

        if (! ($note->getContactID() == $args['contactId'] && $note->getOwner() == $user->getExtension())) {
            return $this->error('Forbidden', $response, 403);
        }

        $result = $this->service()->deleteNote($note);


        if ($result) {
            return $this->result($args, $response);
        }

        return $this->error('Note not deleted', $response, 501);
    }

    private function serviceContacts(): Contacts {
        return $this->getContainer()->get(Contacts::class);
    }

    private function service(): Model {
        return $this->getContainer()->get(\Models\Notes::class);
    }
}
