<?php

namespace Controllers\OAuth2;

use Controllers\Controller;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Slim\Psr7\Factory\StreamFactory;
use Slim\Views\PhpRenderer;
use Wildix\Auth\OAuth2\OAuth2Exception;
use Wildix\Auth\OAuth2\OAuth2Service;
use Wildix\Auth\OAuth2\OAuth2ServiceInterface;

class Auth extends Controller {

    /**
     * Get OAuth2 authorization form
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     *
     * @return Response
     */
    public function getAuthForm(Request $request, Response $response) {
        $data = $this->_getAuthorizeDataFromRequest($request);
        try {
            $this->_getService()->handleAuthorizationRequest($data);

            $oath2Mode      = true;
            $oauth2GrantUrl = $this->_getService()->generateGrantUrl($data);

            $stringHtml = include '../authorization/index.php';

            $streamFactory = new StreamFactory;

            return $response->withBody($streamFactory->createStream($stringHtml));
        } catch (OAuth2Exception $e) {
            $info         = $e->getFormattedError();
            $info['code'] = $e->getCode();

            return $this->showInfoWindow($response, $info);
        }
    }

    /**
     * Get OAuth2 grant confirm form
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     *
     * @return Response
     */
    public function getGrantForm(Request $request, Response $response): Response {
        $data = $this->_getAuthorizeDataFromRequest($request);
        try {
            $app  = $this->_getService()->handleAuthorizationGrantRequest($data);
            $vars = [
                'oauth2' => $this->_getService()->getSessionData(),
                'user'   => $this->getUser(),
                'grantMessage'  => sprintf($this->translate('application.oauth2.pleaseGrantAccess'), $app->getName()),
                'appName'  => $app->getName(),
                'title' => $this->translate('oauth2.window.title'),
                'buttonYes'  => $this->translate('buttons.Yes'),
                'buttonNo'  => $this->translate('buttons.No'),
            ];

            $renderer = new PhpRenderer(__DIR__ . '/../../templates/OAuth2');
            return $renderer->render($response, 'Grant.php', $vars);
        } catch (OAuth2Exception $e) {
            $info         = $e->getFormattedError();
            $info['code'] = $e->getCode();

            return $this->showInfoWindow($response, $info);
        }
    }


    /**
     * Confirm grant and redirect page
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     *
     * @return Response|void
     */
    public function grantConfirm(Request $request, Response $response) {
        $data = $this->_getAuthorizeDataFromRequest($request);
        try {
            $confirm = $this->_getService()->handleGrantConfirmRequest($data);
            $this->_getService()->redirectAfterGrant($confirm);
        } catch (OAuth2Exception $e) {
            $info         = $e->getFormattedError();
            $info['code'] = $e->getCode();

            return $this->showInfoWindow($response, $info);
        }
    }

    /**
     * Get OAuth2 tokens
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     *
     * @throws OAuth2Exception Exception.
     * @return Response
     */
    public function getTokens(Request $request, Response $response): Response {
        $response = $response->withoutHeader('Access-Control-Allow-Origin');
        try {
            if ($request->getHeaderLine('X-Requested-With') === 'XMLHttpRequest') {
                throw new OAuth2Exception('Not supported AJAX request', OAuth2Exception::ERROR_INVALID_REQUEST);
            }
            $data   = $this->_getAccessTokenRequest($request);
            $result = $this->_getService()->handleAccessTokenRequest($data);
            $code   = 200;
        } catch (OAuth2Exception $e) {
            $result = $e->getFormattedError();
            $code   = $e->getCode();

            $this->getLogger()->error(__LINE__, json_encode($result));
        }

        // @TODO fix signature
        return $this->response($code, $response, $result);
    }

    private function showInfoWindow(Response $response, array $data): Response {
        $renderer = new PhpRenderer(__DIR__ . '/../../templates/OAuth2');
        return $renderer->render($response, 'Info.php', $data);
    }

    /**
     * Get OAuth data from response
     *
     * Sample https://developers.google.com/identity/protocols/oauth2/javascript-implicit-flow?hl=RU#redirecting
     *
     * @param Request $request Request.
     *
     * @return string[]
     */
    private function _getAuthorizeDataFromRequest(Request $request): array {
        $getParams  = $request->getQueryParams();
        $postParams = $request->getParsedBody();

        return [
            OAuth2Service::PARAM_CLIENT_ID     => (string)($getParams[OAuth2Service::PARAM_CLIENT_ID] ?? ''),
            OAuth2Service::PARAM_RESPONSE_TYPE => (string)($getParams[OAuth2Service::PARAM_RESPONSE_TYPE] ?? ''),
            OAuth2Service::PARAM_REDIRECT_URI  => (string)($getParams[OAuth2Service::PARAM_REDIRECT_URI] ?? ''),
            OAuth2Service::PARAM_SCOPE         => (string)($getParams[OAuth2Service::PARAM_SCOPE] ?? ''),
            OAuth2Service::PARAM_STATE         => (string)($getParams[OAuth2Service::PARAM_STATE] ?? ''),
            OAuth2Service::PARAM_PROMPT        => (string)($getParams[OAuth2Service::PARAM_PROMPT] ?? 'consent'),
            OAuth2Service::PARAM_GRANT_CONFIRM => (string)($postParams[OAuth2Service::PARAM_GRANT_CONFIRM] ?? ''),
        ];
    }

    private function _getAccessTokenRequest(Request $request): array {
        $params = $request->getParsedBody();

        return [
            OAuth2Service::PARAM_CLIENT_ID     => (string)($params[OAuth2Service::PARAM_CLIENT_ID] ?? ''),
            OAuth2Service::PARAM_GRANT_TYPE    => (string)($params[OAuth2Service::PARAM_GRANT_TYPE] ?? ''),
            OAuth2Service::PARAM_REDIRECT_URI  => (string)($params[OAuth2Service::PARAM_REDIRECT_URI] ?? ''),
            OAuth2Service::PARAM_CLIENT_SECRET => (string)($params[OAuth2Service::PARAM_CLIENT_SECRET] ?? ''),
            OAuth2Service::PARAM_CODE          => (string)($params[OAuth2Service::PARAM_CODE] ?? ''),
            OAuth2Service::PARAM_REFRESH_TOKEN => (string)($params[OAuth2Service::PARAM_REFRESH_TOKEN] ?? ''),
        ];
    }

    /**
     * Get Oauth service
     *
     * @return OAuth2ServiceInterface
     */
    private function _getService(): OAuth2ServiceInterface {
        return $this->getContainer()->get(OAuth2ServiceInterface::class);
    }
}
