<?php
namespace Controllers;

use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\Core\Container;
use Wildix\Error;
use Wildix\Exception;
use Wildix\License;
use Wildix\PBXes\PBXes;
use Wildix\Services\Portal;

class PBX extends Controller {

    const TURN_API_URL = 'https://:domain/api/pbx/ice/:serial/:extension';

    public function ping(Request $request, Response $response, array $args) {
        return $this->result('Pong', $response);
    }

    public function pingError(Request $request, Response $response, array $args) {
        // @TODO remove
        return $this->error('Pong error', $response, 403);
    }

    public function pingFatal(Request $request, Response $response, array $args) {
        // @TODO remove
        throw new \Error('something bad happened');
    }

    public function pingWarning(Request $request, Response $response, array $args) {
        // @TODO remove
        return $this->warning('Pong has warnings', $response, 'Pong', ['warning' => 123, 'warning3' => 321]);
    }

    public function pingJsonBody(Request $request, Response $response, array $args) {
        // @TODO remove
        return $this->result($request->getParsedBody(), $response);
    }

    public function getVersion(Request $request, Response $response, array $args) {
        return $this->result($this->service()->getVersion(), $response);
    }

    public function getSecurePort(Request $request, Response $response, array $args) {
        $port = $this->service()->getCustomHttpsPort();

        return $this->result(['port' => $port ?: '443'], $response);
    }

    public function getCandidates(Request $request, Response $response, array $args) {
        $result = array(
                'networkAddresses' => array(),
                'ports'            => array(),
                'failovers'        => array()
        );

        // candidates
        $result['networkAddresses'] = $this->service()->getCandidates();

        // ports
        $result['ports'][] = '443';
        $port              = $this->service()->getCustomHttpsPort();
        if ($port) {
            $result['ports'][] = $port;
        }
        $result['ports'] = array_values(array_unique($result['ports']));

        // failovers
        $failovers = $this->service()->getFailovers();
        if (count($failovers)) {
            foreach ($failovers as $pbx) {
                $net = array();
                if ($pbx['description'] && $pbx['description']['pbxData']) {
                    $net = $pbx['description']['pbxData']['addrSet']['list'];
                }
                $net[] = $pbx['domain'];
                usort($net, array('Models\PBX', 'cmpNetworkAddresses'));

                $ports = array('443');
                if ($pbx['publicPort'] && $pbx['publicPort'] != '443') {
                    $ports[] = $pbx['publicPort'];
                }
                $result['failovers'][] = array(
                        'networkAddresses' => $net,
                        'ports'            => $ports
                );
            }
        }

        return $this->result($result, $response);
    }

    public function getIceCandidates(Request $request, Response $response, array $args) {
        $serial     = $this->service()->getSerial();
        $extension  = $this->getUser()->getExtension();
        $kiteApiUrl = str_replace(':domain', $this->service()->getKiteConf('domain'), self::TURN_API_URL);
        $kiteApiUrl = str_replace(':extension', $extension, str_replace(':serial', $serial, $kiteApiUrl));

        $ch = curl_init($kiteApiUrl);
        curl_setopt($ch, CURLOPT_USERAGENT, 'PBX Api');
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_TIMEOUT, 10);
        curl_setopt($ch, CURLOPT_AUTOREFERER, true);
        $kiteResponse = curl_exec($ch);
        curl_close($ch);
        // Kite returns public ice servers: stun and turn with valid credentials
        $result = array();
        try {
            if (!empty($kiteResponse)) {
                $result = json_decode($kiteResponse, true);
            }
        } catch (\Exception $exc) {
            // Magic ??
        }
        // we add one more stun for lan
        $networkAddresses = $this->service()->getNetworkInterfaces();
        if ($networkAddresses && is_array($networkAddresses['interfaces'])) {
            foreach ($networkAddresses['interfaces'] as $address) {
                $result[] = array(
                    'host'      => $address['ip'],
                    'port'      => null,
                    'transport' => 'udp',
                    'type'      => 'stun'
                );
            }
        }
        return $this->result($result, $response);
    }

    public function getExternalLinePrefix(Request $request, Response $response, array $args) {
        $result = [
            'externalLinePrefix' => $this->service()->getExternalLinePrefix()
        ];
        return $this->result($result, $response);
    }

    public function redirect(Request $request, Response $response, array $args) {
        // Get pbx info from LDap but not for our own pbx
        if ($args['serial'] == $this->service()->getSerial()) {
            $pbx    = $this->service();
            $serial = $pbx->getSerial();
            $domain = $this->servicePbxes()->getPBXBySerial($serial)->getDomain();
        } else {
            $pbx    = $this->servicePbxes()->getPBXBySerial($args['serial']);
            if ($pbx == null) {
                throw new Exception(Error::getMessage(Error::NotFound), 404);
            }

            $domain = $pbx->getDomain();
        }


        try {
            // Get pbx secure port or 443 as fallback
            $port = $pbx->getCustomHttpsPort();
            if (!$port) {
                $port = 443;
            }
        } catch (\Exception $exc) {
            // Magic ??
        }

        $url = 'https://' . $domain . ($port == 443 ? '' : ':' . $port) . '/';
        if (strpos($url, '.wildixin.com') === false) {
            $url = 'https://' . $pbx->getSerial() . '.wildixin.com' . ($port == 443 ? '' : ':' . $port) . '/';
        }

        return $response->withHeader('Location', $url)->withStatus(301);
    }

    public function recheckPortsStatus(Request $request, Response $response, array $args) {
        $result = self::progressUploadPbxPorts(true);
        if (!empty($result['message'])) {
            return $this->error($result['message'], $response, $result['code']);
        }

        return $this->result($result, $response);
    }

    public function getPortsStatus(Request $request, Response $response, array $args) {
        $params = $request->getQueryParams();
        $force  = $params['force'] ?? false;
        if ($force != 'true') {
            // if force true get from portal
            $force = false;
        }

        $result = self::progressGetPortsStatus($force);

        if (!empty($result['message'])) {
            return $this->error($result['message'], $response, $result['code']);
        }

        return $this->result($result, $response);
    }

    public static function progressGetPortsStatus($force = true) {
        try {
            $result = self::serviceLicense()->isCloud() ? self::_getDefaultPorts() : self::_getPortalService()->getPortsStatus($force);
            if ($result) {
                return $result;
            } else {
                return array(
                    'code' => 500,
                    'message' => 'Unknown error'
                );
            }
        } catch (\Exception $e) {
            return array(
                'code' => 500,
                'message' => $e->getMessage()
            );
        }
    }

    public static function progressUploadPbxPorts($force = false) {
        $result = false;
        $pbx     = Container::get(\Models\PBX::class);
        if ($force === false || $force === '' || $force === 'false' || $force === 0 || $force === '0') {
            $force = false;
        } else {
            $force = true;
        }

        try {
            $ports  = $pbx->getPorts();
            $result = self::serviceLicense() ? self::_getDefaultPorts() : self::_getPortalService()->setPorts($ports, $force);
        } catch (\Exception $e) {
            return array(
                'code' => 500,
                'message' => $e->getMessage()
            );
        }
        if ($result) {
            return $result;
        } else {
            return array(
                'code' => 500,
                'message' => 'Unknown error'
            );
        }
    }

    public static function progressUploadPbxInterfaces() {
        $portalResult = false;

        if (self::serviceLicense()->isCloud()) {
            return true;
        }


        $interfaces = array();
        $network    = new \network();
        // phpcs:disable
        if (!empty($network->active_interfaces)) {
            foreach ($network->active_interfaces as $key => $value) {
        // phpcs:enable
                if (strpos($value['address'], '5.15.') === 0) {
                    continue;
                }

                $interfaces[] = array(
                        'type' => $value['type'],
                        'name' => $key,
                        'address' => $value['address'],
                        'netmask' => $value['netmask'],
                        'broadcast' => $value['broadcast'],
                        'network' => $value['network'],
                        'alias' => $value['alias'],

                );
            }
            $portalResult = self::_getPortalService()->setInterfaces($interfaces);
        }


        if ($portalResult) {
            return $portalResult;
        } else {
            return array(
                'code' => 500,
                'message' => 'Unknown error'
            );
        }
    }

    private static function _getPortalService(): Portal {
        return Container::get(Portal::class);
    }

    private static function _getDefaultPorts(): array {
        $pbx               = Container::get(\Models\PBX::class);
        $ports             = self::_preparePortsData($pbx->getPorts());
        $result['total']   = count($ports);
        $result['records'] = $ports;
        return $result;
    }

    private static function _preparePortsData(array $data): array {
        $result    = [];
        $lastcheck = date('Y-m-d H:i:s');
        foreach ($data as $item) {
            $item['status']    = '1';
            $item['lastcheck'] = $lastcheck;
            array_push($result, $item);
        }
        return $result;
    }

    private function service(): \Models\PBX {
        return $this->getContainer()->get(\Models\PBX::class);
    }

    private function servicePbxes(): PBXes {
        return $this->getContainer()->get(PBXes::class);
    }

    private static function serviceLicense(): License {
        return Container::get(License::class);
    }
}
