<?php
namespace Controllers;

use Models\PBX;
use PHPGangsta_GoogleAuthenticator as GoogleAuthenticator;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Session\Session;
use User;
use Wildix\Auth\TwoFactor\TwoFactorAuthenticatorFactory;
use Wildix\CallweaverManager;
use Wildix\Colleagues\Colleague;
use Wildix\Colleagues\Colleagues;
use Wildix\Colleagues\Groups;
use Wildix\Colleagues\Settings;
use Wildix\Dialplan\PagingGroups\PagingGroupsStorage;
use Wildix\Ejabberd;
use Wildix\Error;
use Wildix\Exception;
use Wildix\InvalidDataException;
use Wildix\License;
use Wildix\PBXes\PBXes;
use Wildix\Utils\JWT;
use Wildix\Validator;

/**
 * Personal Controller
 */
class Personal extends Controller {

    public function logout(Request $request, Response $response): Response {
        $_SESSION = [];
        session_write_close();
        if (!empty($_COOKIE['__rc'])) {
            setcookie('__rc', '', -1, '/');
        }
        return $this->result('OK', $response);
    }

    private function autchFail(Response $response, $message, string $login = null): Response {
        $this->_getUser()->authFail($login);
        return $this->error($message, $response, 401);
    }

    public function login(Request $request, Response $response): Response {
        $params = $request->getParsedBody();
        $login  = $params['login'] ?? '';
        if ($this->_getUser()->isForbidden($login)) {
            $messageBlockTime = $login === 'admin' ? '1 hour' : 'one minute';
            return $this->error(sprintf('Blocked for %s because of too many login attempts', $messageBlockTime), $response, 403);
        }
        $type     = $params['type'];
        $code     = $params['code'];
        $remember = $params['remember'] === '1';
        $password = $params['password'];
        $token    = $params['token'];

        $isMsTeams   = filter_var($params['isMsTeams'] ?? false, FILTER_VALIDATE_BOOLEAN);
        $loginStatus = null;

        if (!empty($type)) {
            if ($type == 'next') {
                if (!$this->_getUser()->checkNextStage($code)) {
                    return $this->autchFail($response, 'Invalid code entered. Please check the accuracy of the time settings on your device.', $login);
                }
            } else {
                $loginStatus = $this->_getUser()->login($login, $token, $type, $isMsTeams);
            }
        } else {
            if (empty($login) || empty($password)) {
                return $this->autchFail($response, 'Login or password empty', $login);
            }

            $loginStatus = $this->_getUser()->login($login, $password, null, $isMsTeams);
        }

        if (is_array($loginStatus) && $loginStatus['errorCode'] === 403) {
            $hasError = array_key_exists('errorMessage', $loginStatus);
            return $this->error(Error::AccessError, $response, 403, $hasError ? [$loginStatus['errorMessage']] : []);
        }

        $next = $this->_getUser()->getNextStage();
        if ($next !== false) {
            if (TwoFactorAuthenticatorFactory::hasAuthenticator($next)) {
                $authenticatorFactory = $this->getContainer()->get(TwoFactorAuthenticatorFactory::class);

                $pbx = $this->servicePbxEs()->getPBXByDN($this->_getUser()->getPBXDn());
                $authenticatorFactory->createBy2ndFactor($next, isset($pbx) ? $pbx->getDomain() : '')->sendCodeTo2ndFactor($this->_getUser());
            }

            return $this->result(['next' => $next], $response);
        }

        if ($this->_getUser()->isLogged()) {
            if ($remember) {
                $this->_getUser()->setRememberCookie();
            }
            $response = $response->withHeader(Session::CSRF_TOKEN_HEADER, Session::getCsrfToken());
        } else {
            return $this->autchFail($response, 'Login or password incorrect', $login);
        }

        if (env('CALL_CONTROL_FEATURE', false)) {
            $user        = $this->_getUser();
            $queryParams = $request->getQueryParams();
            $sipModeFlag = $queryParams['sip'] ?? '';
            $this->getLogger()->info(
                __LINE__,
                sprintf(
                    "[features: second collaboration] user '%s' with license '%s' logged in '%s' mode",
                    $user->getExtension(),
                    $user->getLicenseType(),
                    $sipModeFlag === 'no' ? 'call control' : 'standard'
                )
            );
        }

        return $this->result('OK', $response);
    }

    private function cutRoster($roster) {
        // remove
        $countUsers     = count($roster, COUNT_RECURSIVE) - count($roster);
        $maxRosterUsers = $this->serviceLicense()->getMaxRosterUsers();

        // delete if colleagues > limit
        if ($countUsers >= $maxRosterUsers) {
            $whileCounts = $countUsers + 1;
            while ($whileCounts > $maxRosterUsers) {
                $removeUsers           = array();
                $random                = rand(0, count($roster) - 1);
                $tOne                  = array_slice($roster, $random, 1);
                $tOnekey               = key($tOne);
                $random                = rand(0, count($tOne[$tOnekey]) - 1);
                $removeUsers[$tOnekey] = array_slice($tOne[$tOnekey], $random, 1);

                unset($roster[$tOnekey][key($removeUsers[$tOnekey])]);
                $whileCounts--;
            }
        }
        return $roster;
    }

    public function postRoster(Request $request, Response $response): Response {
        $params    = $request->getParsedBody();
        $user      = $this->getUser();
        $extension = trim($params['user']);

        if ($user->getExtension() == $extension) {
            return $this->error('Not add roster himself', $response, 404);
        }

        $addUser = $this->serviceColleagues()->getUserByExtension($extension);

        if (!$addUser || empty($addUser)) {
            return $this->error('User with number ' . $extension . ' is not found', $response, 404);
        }

        $userRoster = $user->getSettings()->getRoster();

        $group = $this->serviceGroups()->getGroupByDn($addUser->getGroupDn())->getName();
        $name  = $addUser->getName() . ' [' . $addUser->getExtension() . ']';

        if (empty($userRoster[$group][$name])) {
            $maxRosterUsers            = $this->serviceLicense()->getMaxRosterUsers();
            $countUsers                = count($userRoster, COUNT_RECURSIVE) - count($userRoster);
            $userRoster                = $this->cutRoster($userRoster);
            $userRoster[$group][$name] = $addUser->getExtension();

            $user->getSettings()->setRoster($userRoster);
            $user->getSettings()->save();

            if ($countUsers > ($maxRosterUsers - License::ROSTER_DIFF_FOR_WARNING)) {
                if ($countUsers < $maxRosterUsers) {
                    return $this->warning('LimitWarning', $response, ($countUsers + 1) . '/' . $maxRosterUsers);
                } else {
                    return $this->warning('LimitError', $response, $maxRosterUsers);
                }
            }
        }

        return $this->result('OK', $response);
    }

    public function deleteRoster(Request $request, Response $response): Response {
        $params    = $request->getParsedBody();
        $extension = trim($params['user']);
        $user      = $this->getUser();
        $roster    = $user->getSettings()->getRoster();

        $dirty = false;
        foreach ($roster as $key => $group) {
            foreach ($group as $name => $val) {
                if (preg_match('/^[^\[]+ \[' . $extension . '\]$/', $name) === 1) {
                    unset($roster[$key][$name]);
                    $dirty = true;
                    break;
                }
            }
        }

        if ($dirty) {
            $roster = $this->fillRosterIfEmpty($roster);
            $user->getSettings()->setRoster($roster);
            $user->getSettings()->save();
        }

        return $this->result('OK', $response);
    }

    public function putRoster(Request $request, Response $response): Response {
        $params          = $request->getParsedBody();
        $user            = $this->getUser();
        $maxRosterUsers  = $this->serviceLicense()->getMaxRosterUsers();
        $limitForWarning = $this->serviceLicense()->getMaxRosterUsers() - License::ROSTER_DIFF_FOR_WARNING;

        $roster     = $this->fillRosterIfEmpty($this->fillRoster($params['users']));
        $countUsers = count($roster, COUNT_RECURSIVE) - count($roster);

        $user->getSettings()->setRoster($roster);
        $user->getSettings()->save();

        if ($countUsers > $limitForWarning) {
            if ($countUsers <= $maxRosterUsers) {
                return $this->warning('LimitWarning', $response, ($countUsers + 1) . '/' . $maxRosterUsers);
            } else {
                return $this->warning('LimitError', $response, $maxRosterUsers);
            }
        }

        return $this->success($response);
    }

    private function fillRoster($users) {
        $roster     = array();
        $colleagues = $this->filterUsers($users);

        if (empty($colleagues)) {
            return $roster;
        }

        $groups = array();
        $user   = $this->getUser();
        $gr     = $this->serviceGroups()->getGroups($user);

        if ($gr) {
            foreach ($gr as $group) {
                $groups[$group->getDn()] = $group->getName();
            }
        }

        foreach ($colleagues as $colleague) {
            if ($colleague && $groups[$colleague->getGroupDn()]) {
                $key = $colleague->getName() . ' [' . $colleague->getExtension() . ']';
                $roster[$groups[$colleague->getGroupDn()]][$key] = $colleague->getExtension();
            }
        }

        return $this->cutRoster($roster);
    }

    private function filterUsers($users) {
        if (empty($users)) {
            return array();
        }

        $filter         = null;
        $user           = $this->getUser();
        $countUsers     = count($users);
        $maxRosterUsers = $this->serviceLicense()->getMaxRosterUsers();

        if ($countUsers > $maxRosterUsers) {
            $users = array_slice($users, 0, $maxRosterUsers);
        }

        foreach ($users as $key => $extension) {
            if ($extension instanceof Colleague) {
                // exclude himself
                if ($extension->getExtension() === $user->getExtension()) {
                    array_splice($users, $key, 1);
                }
            } else if ($extension !== $user->getExtension()) {
                // not himself
                $filter[Colleague::EXTENSION][] = $extension;
            }
        }

        if (isset($filter)) {
            $users = $this->serviceColleagues()->getUsers($filter, null, null, null, null, null, null, $user);
        }

        return $users;
    }

    private function fillRosterIfEmpty($roster) {
        $countRosterUsers = count($roster, COUNT_RECURSIVE) - count($roster);

        if (!empty($roster) && $countRosterUsers > 0) {
            return $roster;
        }

        $user  = $this->getUser();
        $pbxDn = $user->getPBXDn();
        $users = $this->serviceColleagues()->getUsersByDepartmentAndPbx(
            $user->getDepartment(),
            $pbxDn,
            $user,
            License::ROSTER_DEFAULT_SIZE
        );

        if (empty($users) || count($users) === 1) {
            $users = $this->serviceColleagues()->getUsersByGroupAndPbx(
                $this->serviceGroups()->getGroupByDn($user->getGroupDn()),
                $pbxDn,
                License::ROSTER_DEFAULT_SIZE
            );
        }

        return $this->fillRoster($users);
    }

    public function getInfo(Request $request, Response $response): Response {
        $user   = $this->getUser();
        $params = $request->getQueryParams();
        $fields = $params['fields'] ?? [];

        if (!empty($fields)) {
            $fields = explode(',', $fields);
        }

        $data = $user->getAsArray($fields);

        if (in_array(Colleague::PASSWORD, $fields)) {
            // show web password if force requested web password
            $data[Colleague::PASSWORD] = $user->getPassword();
        }

        if (empty($fields) || in_array(Colleague::SIP_PASSWORD, $fields)) {
            // show sip password if $fields empty or force requested sip password
            $data[Colleague::SIP_PASSWORD] = $user->getSipPassword();
        }

        return $this->result($data, $response);
    }

    private function setFeatures($request, $features, $userSettings) {
        $oldVersion = $this->isOldVersion($request);

        if (empty($features)) {
            return true;
        }

        if (!empty($features['internal'])) {
            if ($oldVersion) {
                $get                  = $userSettings->getInternalClassFeatures();
                $get['default']       = $features['internal'];
                $features['internal'] = $get;
            }
            $userSettings->setInternalClassFeatures($features['internal']);
        }
        if (!empty($features['external'])) {
            if ($oldVersion) {
                $get                  = $userSettings->getExternalClassFeatures();
                $get['default']       = $features['external'];
                $features['external'] = $get;
            }
            $userSettings->setExternalClassFeatures($features['external']);
        }
        if (!empty($features['blacklist'])) {
            if ($oldVersion) {
                $get                   = $userSettings->getBlacklistClassFeatures();
                $get['default']        = $features['blacklist'];
                $features['blacklist'] = $get;
            }
            $userSettings->setBlacklistClassFeatures($features['blacklist']);
        }
        if (!empty($features['whitelist'])) {
            if ($oldVersion) {
                $get                   = $userSettings->getWhitelistClassFeatures();
                $get['default']        = $features['whitelist'];
                $features['whitelist'] = $get;
            }
            $userSettings->setWhitelistClassFeatures($features['whitelist']);
        }
        return true;
    }





    public function getSettings(Request $request, Response $response): Response {
        $params = $request->getQueryParams();
        $fields = $params['fields'];

        if (empty($fields)) {
            $fields = [
               'features',
               'personal',
               'ringtones',
               'fkeys',
               'sharedvm',
               'cwtone',
               'contactCenter',
            ];
        } else {
            $fields = explode(',', $fields);
        }

        return $this->result($this->getUserSettings()->groupByFields($fields, $this->isOldVersion($request)), $response);
    }

    public function postSettings(Request $request, Response $response): Response {
        $params   = $request->getParsedBody();
        $user     = $this->getUser();
        $data     = 'fail';
        $settings = $params['settings'];

        if (!empty($settings)) {
            $settings = json_decode($settings, true);
            $personal = $settings['personal'];

            if (!empty($personal)) {
                if ($personal['email'] !== null) {
                    $user->setEmail($personal['email']);
                }

                if ($this->isAllowToSavePhones($personal['mobilePhone'], $user->getOfficePhone())) {
                     $user->setMobilePhone($personal['mobilePhone']);
                }
                $data = 'success';
            }

            $userSettings = $this->getUserSettings();

            if ($this->setFeatures($request, $settings['features'] ?? [], $userSettings)
                && $this->setCallCenter($settings['contactCenter'] ?? [], $userSettings)
                && $userSettings->save()
            ) {
                $data = 'success';
            }
        }

        return $this->result($data, $response);
    }

    private function setCallCenter(array $settings, Settings $userSettings): bool {
        if (!$settings) {
            return true;
        }

        $validator = new Validator($settings, ['enable', 'statusSync', 'callGroups']);

        foreach ($settings['callGroups'] as $key => $callGroup) {
            $validator->rule('StrictInteger', sprintf('callGroups.%d.id', $key));
            $validator->rule('required', sprintf('callGroups.%d.id', $key));
            $validator->rule('boolean', sprintf('callGroups.%d.active', $key));
            $validator->rule('required', sprintf('callGroups.%d.active', $key));
        }

        $validator->rule('array', 'callGroups');
        $validator->rule('boolean', ['enable', 'statusSync']);
        $validator->rule('required', ['enable', 'statusSync', 'callGroups']);

        if (!$validator->validate()) {
            throw new InvalidDataException(Error::VALIDATION_ERROR, $validator->errors());
        }

        $callCenterSettings = [
            'enable' => $settings['enable'] ? '1' : '0',
            'statusSync' => $settings['statusSync'] ? '1' : '0',
            'CALLCENTER_MODE' => $settings['enable'] ? '1' : '0',
            'callGroups' => $settings['callGroups'] ?: ['1' => ['id' => '']],
        ];

        $userSettings->setCallCenter($callCenterSettings);

        return true;
    }

    private function getUserSettings(): Settings {
        return $this->getUser()->getSettings();
    }

    /**
     * Check is allow to save phones
     *
     * @param string $phoneOne Phone number.
     * @param string $phoneTwo Phone number.
     *
     * @return boolean
     */
    private function isAllowToSavePhones($phoneOne = '', $phoneTwo = '') {
        $res = true;
        if (strval($phoneOne) !== '') {
            $res = preg_match('/^[\+]?[0-9]{2,50}\z/', strval($phoneOne))
                && (empty($phoneTwo) || preg_replace('/\D/', '', strval($phoneOne)) !== preg_replace('/\D/', '', strval($phoneTwo)));
        }

        return $res;
    }

    public function postTwoFactor(Request $request, Response $response): Response {
        $params   = $request->getParsedBody();
        $enable   = trim($params['enable']);
        $type     = trim($params['type'] ?? 'totp');
        $app      = trim($params['app']);
        $code     = trim($params['code']);
        $password = trim($params['password']);
        $secret   = trim($params['secret']);

        if (empty($app)) {
            return $this->error(Error::EmptyApp, $response, 400);
        }
        if (empty($code)) {
            return $this->error(Error::EmptyCode, $response, 400);
        }
        if (empty($password)) {
            return $this->error(Error::EmptyPassword, $response, 400);
        }
        if (empty($secret)) {
            return $this->error(Error::EmptySecret, $response, 400);
        }
        if (empty($type)) {
            return $this->error(Error::EmptyType, $response, 400);
        }


        $ga          = $this->getContainer()->get(GoogleAuthenticator::class);
        $checkResult = $ga->verifyCode(
            $secret,
            $code,
            // (6 or 2) is clock tolerance: (6 or 2) * 30sec
            TwoFactorAuthenticatorFactory::hasAuthenticator($app) ? 6 : 2
        );

        if (!$checkResult) {
            return $this->error(\Wildix\Error::InvalidCode, $response, 400);
        }


        $user = $this->getUser();

        if (!$user->isValidPassword($password)) {
            return $this->error(Error::InvalidPassword, $response, 400);
        }

        $settings = $user->getSettings();
        $settings->setTwoFactor(
            [
                'enable' => $enable,
                'type' => $type,
                'app' => $app,
                'secret' => $secret
            ]
        );
        $settings->save();
        $this->_getUser()->setTwoFactorState(true);

        return $this->success($response);
    }

    public function deleteTwoFactor(Request $request, Response $response): Response {
        $user = $this->getUser();

        $settings = $user->getSettings();
        $settings->setTwoFactor(
            [
                'enable' => '0',
                'type' => 'totp',
                'app' => '',
                'secret' => ''
            ]
        );
        $settings->save();
        $this->_getUser()->setTwoFactorState(false);

        return $this->success($response);
    }

    public function putPassword(Request $request, Response $response): Response {
        $params  = $request->getParsedBody();
        $user    = $this->getUser();
        $current = trim($params['current']);
        $new     = trim($params['new']);
        $confirm = trim($params['confirm']);

        if (empty($current)) {
            return $this->error(Error::EmptyCurrentPassword, $response, 400);
        }
        if (empty($new)) {
            return $this->error(Error::EmptyPassword, $response, 400);
        }
        if (empty($confirm)) {
            return $this->error(Error::EmptyConfirmPassword, $response, 400);
        }
        if ($new != $confirm) {
            return $this->error(Error::NotEqualConfirmPassword, $response, 400);
        }

        if (!($user->getPassword() == $current || $user->isEqualPasswords($user->getPassword(), $current))) {
            return $this->error(Error::InvalidPassword, $response, 400);
        }

        $passwordHash = $user->cryptPassword($new);
        $this->serviceColleagues()->updateUserProperty(
            [
                Colleague::PASSWORD => $passwordHash,
                Colleague::USER_PIN => $user->makeUserPinFromPassword($new)
            ],
            $user
        );

        // update user cache
        $user->updateSessionPassword($passwordHash);


        return $this->success($response);
    }

    public function putPresence(Request $request, Response $response): Response {
        $params  = $request->getParsedBody();
        $user    = $this->getUser();
        $status  = trim($params['status'] ?? '');
        $message = trim($params['message'] ?? '');
        $until   = trim($params['until'] ?? '');

        if ($status == 'mur' && $user->getRole() !== 'room') {
            return $this->error('Status "mur" can be installed only for users "Room"', $response, 400);
        }

        $this->eJabberd()->setUserPresence($user->getExtension(), $status, $message, $until);
        return $this->success($response);
    }

    public function putPresenceLocation(Request $request, Response $response): Response {
        $params    = $request->getParsedBody();
        $lat       = trim($params['lat'] ?? '');
        $lng       = trim($params['lng'] ?? '');
        $address   = trim($params['address'] ?? '');
        $extension = $this->getUser()->getExtension();
        if ($extension !== "") {
            $this->eJabberd()->setUserPresenceLocation($extension, $lat, $lng, $address);
        } else {
            return $this->error('Invalid extension', $response, 400);
        }

        return $this->success($response);
    }

    public function getPagingGroups(Request $request, Response $response): Response {
        $result = PagingGroupsStorage::getForUser($this->getUser()->getExtension());

        return $this->resultList($result, $response, ['total' => count($result)]);
    }

    public function getAuthToken(Request $request, Response $response): Response {
        $params  = $request->getQueryParams();
        $service = (string)($params['service'] ?? '');

        $licenseToken = $this->getContainer()->get(License::class)->getLimitations('token');

        $authService = parse_url(env('SERVICE_AUTH_URI'), PHP_URL_HOST);

        if ($service !== $authService) {
            return $this->result(['token' => $this->_getUser()->getJwtAuthToken($licenseToken)], $response);
        }

        $time    = time();
        $payload = [
            'iss'    => $this->modelsPbx()->getDomain(),
            'sub'    => 'PBX',
            'aud'    => $service,
            'iat'    => $time,
            'exp'    => $time + 60,
            'id'     => substr(uniqid() . uniqid() . uniqid(), 0, 32),
            'user'   => array(
                    'name'      => $this->_getUser()->getName(),
                    'extension' => $this->_getUser()->getExtension(),
                    'jid'       => $this->_getUser()->getJid(),
                    'email'     => $this->_getUser()->getEmail(),
                    'remote_id' => $this->_getUser()->getID(),
                    'lang'      => $this->_getUser()->getLanguage()
            ),
            'serial' => $this->modelsPbx()->getSerial(),
            'token'  => $this->_getUser()->getJwtAuthToken($licenseToken)
        ];

        $secret = md5(md5($payload['iss']) . md5($payload['iat'])) . md5($payload['token']);

        return $this->result(['token' => JWT::encode($payload, $secret)], $response);
    }

    public function sendCodeTo2ndFactor(Request $request, Response $response, $args): Response {
        $params        = $request->getParsedBody();
        $requestErrors = [];

        if (!TwoFactorAuthenticatorFactory::hasAuthenticator($args['secondFactor'])) {
            $requestErrors['secondFactor'] = [
                'code' => Error::UnknownError,
                'reason' => Error::getMessage(Error::UnknownError),
            ];
        }

        $colleague = $this->getUser();
        $settings  = $colleague->getSettings();

        $secret = $params['secret'];

        if (strlen($secret) !== 32) {
            $twoFactorSettings = $settings->getTwoFactor();
            $secret            = !empty($twoFactorSettings['secret']) ? $twoFactorSettings['secret'] : null;
        }

        if (!isset($secret)) {
            $requestErrors['secret'] = [
                'code' => Error::UnknownError,
                'reason' => Error::getMessage(Error::UnknownError),
            ];
        }

        if (count($requestErrors) > 0) {
            throw new InvalidDataException('Can\'t send authentication code to an user\'s second factor', $requestErrors);
        }

        $settings->setTwoFactor(
            [
                'enable' => 1,
                'type' => 'totp',
                'app' => $args['secondFactor'],
                'secret' => $secret,
            ]
        );

        $twoFactorAuthenticatorFactory = $this->getContainer()->get(TwoFactorAuthenticatorFactory::class);

        $pbx                    = $this->servicePbxEs()->getPBXByDN($this->_getUser()->getPBXDn());
        $twoFactorAuthenticator = $twoFactorAuthenticatorFactory->createBy2ndFactor((string)$args['secondFactor'], isset($pbx) ? $pbx->getDomain() : '');
        $twoFactorAuthenticator->sendCodeTo2ndFactor($colleague);

        return $this->success($response);
    }

    public function getAcl(Request $request, Response $response): Response {
        $params = $request->getQueryParams();
        $filter = $params['filter'];
        $user   = $this->getUser();
        if (!$filter || !is_array($filter)) {
            return $this->error('filter is required and it must be an array', $response, 400);
        }
        $aclList = $user->getAclByFilter($filter);

        return $this->result($aclList, $response);
    }

    public function addDevice(Request $request, Response $response): Response {
        $params        = $request->getParsedBody();
        $pairingCode   = trim($params['code']);
        $targetChannel = '';
        $user          = $this->getUser();
        $extension     = $user->getExtension();

        if (!$this->serviceCallweaverManager()->isConnected()) {
            throw new Exception(Error::CallweaverNotAvailable, 404);
        }

        $channels = $this->serviceCallweaverManager()->ChannelsStatus('123');

        if (!count($channels)) {
            throw new Exception(Error::EmptyChannelsList, 404);
        }

        $unknownChannels = [];

        foreach ($channels as $item) {
            if (isset($item['ConnectedLineNum']) && $item['ConnectedLineNum'] == '<unknown>') {
                $unknownChannels[$item['Channel']] = $item['Channel'];
            }
        }

        if (count($unknownChannels)) {
            foreach ($unknownChannels as $channel) {
                $result = $this->serviceCallweaverManager()->getVar($channel, 'ASSIGNCODE');

                if ($result['Response'] == 'Success' && $result['Value'] == $pairingCode) {
                    $targetChannel = $channel;
                    break;
                }
            }

            if ($targetChannel) {
                $this->serviceCallweaverManager()->redirect($targetChannel, '', $extension, 'assigntouser', '1');
            } else {
                throw new Exception(Error::ChannelNotFound, 404);
            }
        } else {
            throw new Exception(Error::ChannelNotFound, 404);
        }

        return $this->success($response);
    }

    private function eJabberd(): Ejabberd {
        return $this->getContainer()->get(Ejabberd::class);
    }

    private function serviceColleagues(): Colleagues {
        return $this->getContainer()->get(Colleagues::class);
    }

    private function servicePbxEs(): PBXes {
        return $this->getContainer()->get(PBXes::class);
    }

    private function _getUser(): User {
        return $this->getContainer()->get(User::class);
    }

    private function serviceLicense(): License {
        return $this->getContainer()->get(License::class);
    }

    private function serviceGroups(): Groups {
        return $this->getContainer()->get(Groups::class);
    }

    private function modelsPbx(): PBX {
        return $this->getContainer()->get(PBX::class);
    }

    private function serviceCallweaverManager(): CallweaverManager {
        return $this->getContainer()->get(CallweaverManager::class);
    }
}
