<?php
namespace Controllers;

use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\Smtp\SmtpValidator;
use Wildix\Validator;

class Smtp extends Controller {
    public function getSettings(Request $request, Response $response, array $args): Response {
        return $this->result($this->_formatResponse($this->_getSmtpService()->getMailOptions()), $response);
    }

    public function setSettings(Request $request, Response $response, array $args): Response {
        $data = $request->getParsedBody();

        $data['is_default_settings'] = Validator::isTrue($data['isDefaultSettings'] ?? '') ? Validator::ON : Validator::OFF;

        if ($data['is_default_settings'] === Validator::OFF) {
            $validator = $this->_getSmtpValidator();
            if (!$validator->validateSmptSettings($data)) {
                return $this->badRequest($validator->getErrors(), $response);
            }

            $data['user']     = (string)($data['user'] ?? '');
            $data['password'] = (string)($data['password'] ?? '');
        }
        $data['tls']          = Validator::isTrue($data['tls'] ?? '') ? Validator::ON : Validator::OFF;
        $data['tls_starttls'] = Validator::isTrue($data['tlsStartTls'] ?? '') ? Validator::ON : Validator::OFF;

        $result = $this->_getSmtpService()->setMailOptions($data);

        return $this->result($this->_formatResponse($result), $response);
    }

    public function runTest(Request $request, Response $response, array $args): Response {
        $data = $request->getParsedBody();

        $validator = $this->_getSmtpValidator();
        if (!$validator->validateSmptPingRequest($data)) {
            return $this->badRequest($validator->getErrors(), $response);
        }

        return $this->result($this->_getSmtpService()->runTest($data), $response);
    }

    public function getTestStatus(Request $request, Response $response, array $args): Response {
        return $this->result($this->_getSmtpService()->getTestStatus($args['testId']), $response);
    }

    private function _formatResponse($data) {
        $output = [
            'account' => $data['account'],
            'auth' => $data['auth'],
            'domain' => $data['domain'],
            'from' => $data['from'],
            'host' => $data['host'],
            'port' => (int)$data['port'],
            'user' => $data['user'],
            'password' => $data['password'],
            'timeout' => (int)$data['timeout'],
            'tls' => $data['tls'] === Validator::ON,
            'tlsStartTls' => $data['tls_starttls'] === Validator::ON,
            'isDefaultSettings' => $data['is_default_settings'] === Validator::ON,
        ];

        // not show defaults user and password for cloud PBX
        if ($output['isDefaultSettings']) {
            $output['user']     = '';
            $output['password'] = '';
        }

        return $output;
    }

    /**
     * Is cloud PBX.
     *
     * @return boolean
     */
    private function _isCloudPbx(): bool {
        return $this->getContainer()->get(\Wildix\License::class)->isCloud();
    }

    /**
     * Return SMTP service.
     *
     * @return \Wildix\Smtp
     */
    private function _getSmtpService(): \Wildix\Smtp {
        return $this->getContainer()->get(\Wildix\Smtp::class);
    }

    /**
     * Return Smtp validator.
     *
     * @return SmtpValidator
     */
    private function _getSmtpValidator(): SmtpValidator {
        return $this->getContainer()->get(SmtpValidator::class);
    }
}
