<?php

namespace Controllers;

use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\InvalidDataException;
use Wildix\Sounds\SoundsService;

class SoundsController extends Controller {

    private function _service(): SoundsService {
        return $this->getContainer()->get(SoundsService::class);
    }

    /**
     * Get items
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     * @param array    $args     Params.
     *
     * @return Response
     * @throws InvalidDataException InvalidDataException.
     */
    public function getItems(Request $request, Response $response, array $args = []): Response {
        $args = $request->getAttribute('originAttribute');
        $args = array_filter($args);
        $path = preg_replace('/(\/){2,}/', '$1', implode('/', $args));

        $res = $this->_service()->getItems($path);

        return $this->resultList($res, $response);
    }

    /**
     * Create item
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     * @param array    $args     Params.
     *
     * @return Response
     * @throws \Wildix\InvalidDataException InvalidDataException
     */
    public function create(Request $request, Response $response, array $args): Response {
        $requestParams = $request->getParsedBody();
        $file          = $request->getUploadedFiles()['file'];
        $params        = [
            'text'      => (string)$requestParams['text'] ?? '',
            'voice'     => (string)$requestParams['voice'] ?? 'Microsoft Server Speech Text to Speech Voice (en-US, ZiraPro)',
            'quality'   => (string)$requestParams['quality'] ?? 'SAFT8kHz16BitMono',
            'type'      => $requestParams['type'] ?? 'mp3',
            'volume'    => (integer)$requestParams['volume'] ?? 100,
            'speed'     => (integer)$requestParams['speed'] ?? 0,
            'pitch'     => (integer)$requestParams['pitch'] ?? 0,
            'directory' => (string)$requestParams['directory'] ?? '/',
            'name'      => (string)$requestParams['filename'] ?? 'tts',
            'file'      => $file,
        ];
        $this->_service()->create($params);
        return $this->success($response);
    }

    /**
     * Update item
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     * @param array    $args     Params.
     *
     * @return Response
     * @throws InvalidDataException InvalidDataException.
     */
    public function update(Request $request, Response $response, array $args = []) {
        $requestParams = $request->getParsedBody();

        $path = implode('/', $args);
        $data = [
            'name' => (string)($requestParams['newName'] ?? ''),
            'move' => (string)($requestParams['move'] ?? ''),
        ];
        $this->_service()->update($path, $data);

        return $this->success($response);
    }

    /**
     * Remove item
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     * @param array    $args     Params.
     *
     * @return Response
     * @throws \Wildix\InvalidDataException InvalidDataException.
     */
    public function remove(Request $request, Response $response, array $args = []): Response {
        $args = $request->getAttribute('originAttribute');
        $path = implode('/', $args);
        $this->_service()->remove($path);

        return $this->success($response);
    }

    /**
     * Check item
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     * @param array    $args     Params.
     *
     * @return Response
     * @throws \Wildix\InvalidDataException InvalidDataException.
     */
    public function check(Request $request, Response $response, array $args = []) {
        $args = $request->getAttribute('originAttribute');
        $path = implode('/', $args);
        $this->_service()->checkItem($path);

        return $this->success($response);
    }

    /**
     * Create directory
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     *
     * @return mixed
     * @throws \Wildix\InvalidDataException InvalidDataException
     */
    public function createDirectory(Request $request, Response $response) {
        $path = $request->getParsedBody();

        if (!isset($path['path'])) {
            throw new InvalidDataException('Empty name');
        }

        $this->_service()->createDirectory($path['path']);

        return $this->success($response);
    }

    /**
     * Edit directory
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     * @param array    $args     Params.
     *
     * @return Response
     * @throws \Wildix\InvalidDataException InvalidDataException.
     */
    public function editDirectory(Request $request, Response $response, array $args = []): Response {
        $args          = $request->getAttribute('originAttribute');
        $path          = implode('/', $args);
        $paramsRequest = $request->getParsedBody();
        $data          = [
            'name' => $paramsRequest['newName'] ?? '',
            'move' => $paramsRequest['move'] ?? '',
        ];

        $this->_service()->editDirectory($path, $data);

        return $this->success($response);
    }

    /**
     * Remove directory
     *
     * @param Request  $request  Request.
     * @param Response $response Response.
     * @param array    $args     Params.
     *
     * @return Response
     * @throws InvalidDataException Exception.
     */
    public function removeDirectory(Request $request, Response $response, array $args = []) {
        $args = $request->getAttribute('originAttribute');
        $path = implode('/', $args);
        $this->_service()->removeDirectory($path);

        return $this->success($response);
    }

    /**
     * Get path
     *
     * @param Request           $request Request.
     * @param array|string|null $path    Path.
     *
     * @return string
     */
    private function getPath(Request $request, array|string|null $path): string {
        $path = explode('?', stristr($request->getRequestTarget(), $path));
        array_splice($path, -1);

        return implode('/', $path);
    }
}
