<?php
namespace Controllers;

use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\System\Mount;
use Wildix\System\Storage\Storage;

class Storages extends Controller {

    private function storage(): Storage {
        return $this->getContainer()->get(Storage::class);
    }

    public function getData(Request $request, Response $response, array $args): Response {
        return $this->result($this->storage()->getData(), $response);
    }

    public function initUsb(Request $request, Response $response, array $args): Response {
        $body    = $request->getParsedBody();
        $initUsb = $body['initUsb'] ?? null;

        if (!$initUsb) {
            return $this->error('Missing initUsb param', $response);
        }

        $storage = $this->storage();
        $result  = $storage->initUsb($initUsb);
        $errors  = $storage->getErrors();

        if ($errors && $errors[0]) {
            return $this->error($errors[0], $response);
        }

        return $this->result($result, $response);
    }

    public function mountUsb(Request $request, Response $response, array $args) {
        $body       = $request->getParsedBody();
        $mountUsb   = $body['mountUsb'] ?? null;
        $mountPoint = $body['mountPoint'] ?? null;

        if (!$mountUsb) {
            return $this->error('Missing mountUsb param', $response);
        }
        if (!$mountPoint) {
            return $this->error('Missing mountPoint param', $response);
        }

        $storage = $this->storage();
        $result  = $storage->mountUsb($mountUsb, $mountPoint);
        $errors  = $storage->getErrors();

        if ($errors && $errors[0]) {
            return $this->error($errors[0], $response);
        }

        if ($mountPoint == '/mnt/cdr') {
            \CDR::getInstance()->moveChatHistoryDb(true);
        }

        return $this->result($result, $response);
    }

    public function mountToExistingShare(Request $request, Response $response, array $args) {
        $body        = $request->getParsedBody();
        $shareType   = $body['shareType'] ?? null;
        $sharePath   = $body['sharePath'] ?? null;
        $mountPoint  = $body['mountPoint'] ?? null;
        $shareFolder = $body['shareFolder'] ?? null;

        if (!$shareType) {
            return $this->error('Missing shareType param', $response);
        }
        if (!$sharePath) {
            return $this->error('Missing sharePath param', $response);
        }
        if (!$mountPoint) {
            return $this->error('Missing mountPoint param', $response);
        }
        if (!$shareFolder) {
            return $this->error('Missing shareFolder param', $response);
        }

        $storage = $this->storage();
        $result  = $storage->mountToExistingShare($mountPoint, $shareType, $sharePath, $shareFolder);
        $errors  = $storage->getErrors();

        if ($errors && $errors[0]) {
            return $this->error($errors[0], $response);
        }

        return $this->result($result, $response);
    }

    public function unMountUsb(Request $request, Response $response, array $args) {
        $body       = $request->getParsedBody();
        $mountPoint = $body['mountPoint'] ?? null;

        if (!$mountPoint) {
            return $this->error('Missing mountPoint param', $response);
        }

        $storage = $this->storage();

        if ($mountPoint == '/mnt/cdr') {
            \CDR::getInstance()->moveChatHistoryDb(false);
        }

        $result = $storage->unMount($mountPoint, 'usb');
        $errors = $storage->getErrors();

        if ($errors && $errors[0]) {
            return $this->error($errors[0], $response);
        }

        return $this->result($result, $response);
    }

    public function remountShare(Request $request, Response $response, array $args) {
        $body       = $request->getParsedBody();
        $shareType  = $body['shareType'] ?? null;
        $mountPoint = $body['mountPoint'] ?? null;

        if (!$shareType) {
            return $this->error('Missing shareType param', $response);
        }

        if (!$mountPoint) {
            return $this->error('Missing mountPoint param', $response);
        }

        $storage = $this->storage();
        $result  = $storage->remountShare($mountPoint, $shareType);
        $errors  = $storage->getErrors();

        if ($errors && $errors[0]) {
            return $this->error($errors[0], $response);
        }

        return $this->result($result, $response);
    }

    public function unMountSmb(Request $request, Response $response, array $args) {
        $body       = $request->getParsedBody();
        $mountPoint = $body['mountPoint'] ?? null;

        if (!$mountPoint) {
            return $this->error('Missing mountPoint param', $response);
        }

        $storage = $this->storage();
        $result  = $storage->unMount($mountPoint, 'smb');
        $errors  = $storage->getErrors();

        if ($errors && $errors[0]) {
            return $this->error($errors[0], $response);
        }

        return $this->result($result, $response);
    }

    public function mountSmb(Request $request, Response $response, array $args) {
        $params = ['type' => 'smb'];
        $body   = $request->getParsedBody();

        $mountPoint = trim($body['mountPoint'] ?? null) . '/';
        if (!preg_match('/^[^`;\(\)\[\]\{\}\|]+$/', $mountPoint)) {
            return $this->error(_t('devices.err.CharsNotRequired'), $response);
        }

        $params['servername'] = trim($body['name'] ?? null);
        $params['serverpath'] = trim($body['path'] ?? null);
        $params['domain']     = trim($body['domain'] ?? null);
        $params['username']   = trim($body['username'] ?? null);
        $params['password']   = trim($body['password'] ?? null);
        $params['serverpath'] = trim($params['serverpath'], '\/');
        $params['smbversion'] = $body['smbVersion'] ?? Mount::DEFAULT_SMB_VERSION;

        if ($params['servername'] == '') {
            return $this->error(_t('devices.err.PleaseInputStorage'), $response);
        }

        if ($params['username'] != '') {
            if (!preg_match('/^[^`;\(\)\[\]\{\}\|]+$/', $params['username'])) {
                return $this->error(_t('devices.err.CharsNotRequired'), $response);
            }
        }

        $storage  = $this->storage();
        $validate = $storage->validateShareMount($mountPoint, $params);
        if ($validate === true) {
            $result = $storage->mountShare($mountPoint, $params);
            $errors = $storage->getErrors();

            if ($errors && $errors[0]) {
                return $this->error($errors[0], $response);
            }

            return $this->result($result, $response);
        } else {
            return $this->error($validate, $response);
        }
    }

    public function mountNfs(Request $request, Response $response, array $args) {
        $params = ['type' => 'nfs'];
        $body   = $request->getParsedBody();

        $mountPoint = trim($body['mountPoint'] ?? null) . '/';
        if (!preg_match('/^[^`;\(\)\[\]\{\}\|]+$/', $mountPoint)) {
            return $this->error(_t('devices.err.CharsNotRequired'));
        }

        $params['servername'] = trim($body['name'] ?? null);
        $params['serverpath'] = trim($body['path'] ?? null);
        $params['serverpath'] = trim($params['serverpath'], '\/');

        if ($params['servername'] == '') {
            return $this->error(_t('devices.err.PleaseInputStorage'), $response);
        }

        $storage  = $this->storage();
        $validate = $storage->validateShareMount($mountPoint, $params);

        if ($validate === true) {
            $result = $storage->mountShare($mountPoint, $params);
            $errors = $storage->getErrors();

            if ($errors && $errors[0]) {
                return $this->error($errors[0], $response);
            }

            return $this->result($result, $response);
        } else {
            return $this->error($validate, $response);
        }
    }
}
