<?php
namespace Controllers;

use Models\PBX as PBXModel;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\Colleagues\Colleague;
use Wildix\Colleagues\Colleagues;
use Wildix\Ejabberd;
use Wildix\Error;
use Wildix\Exception;
use Wildix\PBXes\PBX;
use Wildix\PBXes\PBXes;

/**
 * Controller User
 */
class User extends Controller {

    public function sendMessage(Request $request, Response $response, array $args): Response {
        $params = $request->getParsedBody();
        $from   = $this->getUser();

        $extensions = explode(',', $args['extension']);
        $message    = trim($params['message']);
        $type       = trim($params['type'] ?? 'chat');

        foreach ($extensions as $extension) {
            $user = $this->getColleagues()->getUserByExtension($extension);
            if (!$user) {
                throw new Exception('User not found', 404);
            }
            $this->eJabberd()->sendMessage($from->getJid(), $user->getJid(), $type, $message);
        }

        return $this->success($response);
    }

    public function getUserInfo(Request $request, Response $response, array $args): Response {
        $user = $this->getColleagues()->getUserByExtension($args['extension']);
        if (!$user) {
            throw new Exception('User not found', 404);
        }
        $presence = $this->eJabberd()->getUserPresence($user->getExtension());
        $uid      = $user->getID();
        $name     = $user->getName();
        $avatar   = $this->getUserAvatarUrl($user->getExtension());
        $pbx      = $this->getPbxEs()->getPBXByDN($user->getPBXDn());

        $data = [
            'uid' => $uid,
            'name' => $name,
            'avatar' => $avatar,
            'pbx' => $pbx->getSerial(),
            'pbxHostname' => $pbx->getHostname(),
            'presence' => $presence,
        ];

        return $this->result($data, $response);
    }

    public function getPublicUserInfo(Request $request, Response $response, array $args): Response {
        $extension = explode(',', $args['extension']);
        if (count($extension) == 1) {
            $user = $this->getColleagues()->getUserByExtension($extension[0]);
            if (!$user || !$user->can(Colleague::SHARE_STATUS_VIA_KITE)) {
                throw new Exception('User not found', 404);
            }

            $data = $this->eJabberd()->getPublicUserInfo($user->getExtension());
            // use data from ejabberd
            $pbx = $this->getPbxEs()->getPBXByDN($user->getPBXDn());

            $data['pbxHostname'] = $pbx->getHostname();
        } else {
            $data = $this->eJabberd()->getPublicUsersInfo($extension);
        }
        return $this->result($data, $response);
    }

    public function getPublicPresence(Request $request, Response $response, array $args): Response {
        $extension = explode(',', $args['extension']);
        if (count($extension) == 1) {
            $user = $this->getColleagues()->getUserByExtension($extension[0]);
            if (!$user || !$user->can(Colleague::SHARE_STATUS_VIA_KITE)) {
                throw new Exception('User not found', 404);
            }
            $pbx  = $this->getPbxEs()->getPBXByDN($user->getPBXDn());

            $data = $this->eJabberd()->getPublicUserPresence($user->getExtension());

            $data['pbx']         = $pbx->getSerial();
            $data['pbxHostname'] = $pbx->getHostname();
        } else {
            $data = $this->eJabberd()->getPublicUsersPresence($extension);
        }
        return $this->result($data, $response);
    }

    public function getPresence(Request $request, Response $response, array $args): Response {
        $extension = explode(',', $args['extension']);
        if (count($extension) == 1) {
            $user = $this->getColleagues()->getUserByExtension($extension[0]);
            if (!$user) {
                throw new Exception('User not found', 404);
            }
            $data = $this->eJabberd()->getUserPresence($user->getExtension());
        } else {
            $data = $this->eJabberd()->getUsersPresence($extension);
        }

        return $this->result($data, $response);
    }

    public function putPresence(Request $request, Response $response, array $args): Response {
        $params    = $request->getParsedBody();
        $user      = $this->getUser();

        $colleague = $this->getColleagues()->getUserByExtension($args['extension']);
        if (!$colleague) {
            throw new Exception('User not found', 404);
        }

        if ($user->getAuthMode() !== $user::AUTH_TYPE_SERVICE) {
            if (!$user->isAdministrator() && $user->canNot(Colleague::CHANGE_PRESENCE, $colleague->getLdapProperty('groupDn'))) {
                throw new Exception('No permission to edit a colleague', 403);
            }
        }

        $status  = trim($params['status']);
        $message = trim($params['message']);
        $until   = trim($params['until']);

        $pbx = $this->getPbxEs()->getPBXByDN($colleague->getPBXDn());
        if (!$pbx instanceof PBX) {
            throw new Exception('PBX not found', 404);
        }

        // User is not From This PBX!
        if (strtolower($this->getPbxModel()->getSerial()) != strtolower($pbx->getSerial())) {
            $data = [
                'status' => $status,
                'message' => $message,
                'until' => $until
            ];
            if ($pbx->queryApi('/api/v1/user/' . $args['extension'] . '/presence/', 'PUT', $data, $user)) {
                return $this->success($response);
            }

            throw new Exception(Error::InternalServerError);
        }

        if ($status == 'mur' && $colleague->getRole() !== 'room') {
            throw new Exception('Status "mur" can be installed only for users "Room"', 400);
        }
        $this->eJabberd()->setUserPresence($colleague->getExtension(), $status, $message, $until);
        return $this->success($response);
    }

    public function getLocation(Request $request, Response $response, array $args): Response {
        $user = $this->getColleagues()->getUserByExtension($args['extension']);
        if (!$user) {
            throw new Exception('User not found', 404);
        }
        $location = $this->eJabberd()->getUserLocationStatus($user->getExtension());
        return $this->result($location, $response);
    }

    private function getUserAvatarUrl($extension) {
        $user = $this->getColleagues()->getUserByExtension($extension);
        $host = $_SERVER['HTTP_HOST'];

        if ($user == null || !$user->can(Colleague::SHARE_STATUS_VIA_KITE)) {
            $url = 'http://' . $host . '/user_avatar/not_found.png';
        } else {
            $url = 'http://' . $host . '/user_avatar/' . $extension . '.png';
        }

        return $url;
    }

    private function eJabberd(): Ejabberd {
        return $this->getContainer()->get(Ejabberd::class);
    }

    private function getPbxModel(): PBXModel {
        return $this->getContainer()->get(PBXModel::class);
    }

    private function getPbxEs(): PBXes {
        return $this->getContainer()->get(PBXes::class);
    }

    private function getColleagues(): Colleagues {
        return $this->getContainer()->get(Colleagues::class);
    }
}
