<?php

namespace Controllers;

use Models\VideoConference as VideoConferenceModel;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\CallFiles;
use Wildix\Colleagues\Colleague;
use Wildix\InvalidDataException;
use Wildix\PBXes\PBXes;
use Wildix\Phonebooks\Contact;
use Wildix\Validator;

class VideoConference extends Controller {

    public function jabberInvitation(Request $request, Response $response, array $args): Response {
        $validator        = $this->getContainer()->get(Validator::class);
        $conferenceFields = ['room', 'pin', 'subject', 'participants', 'schedule'];
        $customData       = ['password', 'passwordPin', 'passwordChanged'];
        $conferenceKeys   = array_merge($conferenceFields, $customData);

        $conferenceValidator = $validator->withData(
            $request->getParsedBody(),
            $conferenceKeys
        );

        $conferenceValidator->rule('required', $conferenceFields);

        if (!$conferenceValidator->validate()) {
            throw new InvalidDataException(\Wildix\Error::BadRequest, $conferenceValidator->errors());
        }

        $this->inviteParticipants($conferenceValidator->data());

        return $this->success($response);
    }

    public function invite(Request $request, Response $response, array $args): Response {
        $params = $request->getParsedBody();
        $payload    = (string)($params['payload'] ?? '');
        $conference = json_decode($payload, true);

        if (!($conference && $conference['users'])) {
            return $this->error('Wrong invite list. There is no users', $response, 400);
        }

        $videoConference = new VideoConferenceModel();
        $videoConference->setUser($this->getUser());

        $localUsers  = [];
        $remoteUsers = [];
        $currentUser = $this->getUser();
        array_map(
            function ($el) use (&$localUsers, &$remoteUsers, $currentUser) {
                if ($el['type'] == 'email' || $el['type'] == 'contact') {
                    $remoteUsers[] = $el;
                } else {
                    // For new added contacts email not provided
                    if ($el['payload']['isPhantom']) {
                        $contacts = $this->serviceContacts()->getContacts(
                            [Contact::EXTENSION => $el['payload']['extension']],
                            null,
                            null,
                            null,
                            null,
                            null,
                            null,
                            $currentUser
                        );

                        if (count($contacts) > 0 && $contacts[0]->getEmail()) {
                            $el['payload']['email'] = $contacts[0]->getEmail();
                        }
                    }
                    $localUsers[] = $el;
                }
            },
            $conference['users']
        );

        $pbxes = $this->getPbxesFromUserList($localUsers);

        $remoteConference = $videoConference->createRoom($conference, $localUsers, $remoteUsers, $pbxes);

        foreach ($localUsers as $user) {
            if ($user['payload'] && ($user['payload']['isPhantom'] || $conference['reinvite'])) {
                try {
                    // Skip jabber invitation for online conference users
                    [$extension] = explode('@', $user['payload']['jid']);
                    if ($this->service()->memberPresence($remoteConference['room'], $extension) === 'available') {
                        continue;
                    }
                    $videoConference->inviteByJabber($user, true, $remoteConference['schedule']);
                } catch (\Exception $e) {
                    error_log($e->getMessage());
                }
            } else if ($remoteConference['schedule']['changed'] && $remoteConference['schedule']['enabled']) {
                $videoConference->inviteByJabber($user, true, $remoteConference['schedule']);
            }
        }

        return $this->result(
            [
                'room' => $remoteConference['room'],
                'pin' => ($remoteConference['pin'] ?? ''),
                'pbxHost' => $remoteConference['pbxHost'],
                'externalUrl' => $remoteConference['externalUrl'],
            ],
            $response
        );
    }

    public function createRoom(Request $request, Response $response, array $args): Response {
        /*
            {
             "name": "user1 famali Conference",
             "users": [{
                 "type": "colleague"
                 "name": "user1 famali",
                 "jid": "111@slawutich.wildixin.com",
                 "email": "",
                 "phone": ""
             },{
                 "type": "colleague",
                 "name": "user 300`with'apostrophe dfdff fddffgdf dfdfdfgg",
                 "jid": "100@wildix"
                 "email": "",
                 "phone": ""
             }],
             "schedule": {
                 "start": "17/10/2017 21:00",
                 "end": "17/10/2017 22:00",
                 "description": "descripton",
                 "enabled": false,
                 "country": "UA",
                 "timezone": "Europe/Kiev"
             }
            }
        */

        $params = $request->getParsedBody();

        $user         = $this->getUser();
        $subject      = $params['subject'] ?? '';
        $participants = $params['participants'] ?? '';
        $schedule     = $params['schedule'] ?? '';
        $owner        = $user->getJid();

        if (empty($participants)) {
            return $this->error('Wrong invite list. There is no users', $response, 400);
        }

        $params = [
                'subject'      => $subject,
                'owner'        => $owner,
                'participants' => $participants,
                'schedule'     => $schedule,
        ];

        $remoteConference = $this->service()->createRoom($params);

        return $this->result($remoteConference, $response);
    }

    public function call(Request $request, Response $response, array $args): Response {
        $roomId = $args['id'];
        $params = $request->getParsedBody();

        $user       = $this->getUser();
        $ownSerial  = $this->servicePbx()->getSerial();
        $roomNumber = $this->servicePbx()->getFeatureAccessCode('CONFERENCE') . $roomId;

        // POST parameter
        $extensions = $params['colleagues'] ?? [];
        $extensions = array_map('trim', $extensions);
        $extensions = array_filter($extensions, 'strlen');

        if (count($extensions) == 0) {
            return $this->error('Numbers is empty', $response, 404);
        }

        $roomInfo = $this->service()->getRoom($roomId);
        if ($roomInfo) {
            $colleagues = $this->serviceColleagues()->getUsers(
                [Colleague::EXTENSION => $extensions],
                null,
                null,
                null,
                null,
                null,
                null,
                $user
            );

            $externalColleaguesByPbx = [];
            $ownColleagues           = [];
            // separate own users and users from another pbxs
            foreach ($colleagues as $colleague) {
                if (!$colleague || !$colleague->getPBXDn()) {
                    continue;
                }

                $serial = $this->servicePbxes()->getPBXByDN($colleague->getPBXDn())->getSerial();
                if (!$serial) {
                    continue;
                }

                // own
                if (strtolower($serial) == strtolower($ownSerial)) {
                    array_push($ownColleagues, $colleague->getExtension());
                    continue;
                }
                // external
                if (!isset($externalColleaguesByPbx[$serial])) {
                    $externalColleaguesByPbx[$serial] = array();
                }
                array_push($externalColleaguesByPbx[$serial], $colleague->getExtension());
            }

            // Invite users from self pbx
            foreach ($ownColleagues as $extension) {
                $this->serviceCallFiles()->makeCall($extension, $roomNumber, $roomInfo['subject'], $roomId);
            }

            // invite external
            foreach ($externalColleaguesByPbx as $serial => $extensions) {
                $pbx = $this->servicePbxes()->getPBXBySerial($serial);
                if (!$pbx || strtolower(trim($serial)) == strtolower(trim($ownSerial))) {
                    continue;
                }
                $pbx->queryApi('/api/v1/VideoConference/' . $roomId . '/Call/', 'POST', array('colleagues' => $extensions), $this->getUser());
            }
        }

        return $this->success($response);
    }

    public function destroyRoom(Request $request, Response $response, array $args): Response {
        $roomId = (string)$args['id'];

        $videoConference = $this->service();

        if ($this->isConferenceOwner($roomId)) {
            unset($_SESSION['videoConferenceRooms'][$roomId]);

            return $this->result($videoConference->destroyRoom($roomId), $response);
        }

        if ($this->isConferenceParticipant($roomId)) {
            unset($_SESSION['videoConferenceParticipations'][$roomId]);

            $userJid = $this->getUser()->getExtension() . '@' . $this->servicePbx()->getSerial();

            return $this->result($videoConference->removeParticipant($roomId, $userJid), $response);
        }

        return $this->error('You have not permission to destroy this room', $response, 403);
    }

    public function kick(Request $request, Response $response, array $args): Response {
        $params = $request->getParsedBody();

        $roomId = (string)$params['room'] ?? '';
        $nick   = (string)$params['nick'] ?? '';

        if ($this->isConferenceOwner($roomId)) {
            return $this->result($this->service()->kickParticipant($roomId, $nick), $response);
        }

        return $this->error('You have not permission to kick anyone from this room', $response, 403);
    }

    public function pushHistory(Request $request, Response $response, array $args): Response {
        $params = $request->getParsedBody();

        $videoConference = $this->service();

        return $this->result(
            $videoConference->pushHistory(
                array(
                'room'      => $params['room'],
                'nick'      => $params['nick'],
                'name'      => $params['name'],
                'avatar'    => $params['avatar'],
                'email'     => $params['email'],
                'csessid'   => $params['csessid'],
                'jid'       => $params['jid'],
                'extension' => $params['extension'],
                )
            ),
            $response
        );
    }

    public function getRooms(Request $request, Response $response, array $args): Response {
        $userJid  = $this->getUser()->getExtension() . '@' . $this->servicePbx()->getSerial();
        $roomList = $this->service()->getRoomsByJid($userJid);

        $_SESSION['videoConferenceRooms']          = (array)$_SESSION['videoConferenceRooms'];
        $_SESSION['videoConferenceParticipations'] = (array)$_SESSION['videoConferenceParticipations'];
        if (!empty($roomList)) {
            foreach ($roomList as $values) {
                if (!in_array($values['room'], $_SESSION['videoConferenceRooms'])) {
                    if ($values['owner'] == $userJid) {
                        $_SESSION['videoConferenceRooms'][] = $values['room'];
                    } else {
                        $_SESSION['videoConferenceParticipations'][] = $values['room'];
                    }
                }
            }
        }

        return $this->result($roomList, $response);
    }

    public function schedule(Request $request, Response $response, array $args): Response {
        $schedule = $this->service()->getSchedule($args['id']);

        return $this->result($schedule, $response);
    }

    public function authorize(Request $request, Response $response, array $args): Response {
        $params = $request->getParsedBody();

        $data = $this->service()->authorizeInRoom(
            [
                'room' => (string)$params['room'],
                'name' => (string)$params['name'],
                'email' => (string)$params['email'],
                'extension' => (string)$params['extension']
            ]
        );

        return $this->result($data, $response);
    }

    private function getPbxesFromUserList(&$users) {
        $list              = [];
        $cachedCustomPorts = [];

        foreach ($users as $idx => $user) {
            if (strpos($user['payload']['jid'], 'web') === 0) {
                continue;
            }

            try {
                $user = $this->serviceColleagues()->getUserByExtension((int)$user['payload']['jid']);
            } catch (\Exception $e) {
                error_log($e->getMessage());
                continue;
            }

            if (!$user) {
                continue;
            }

            $pbx = $this->servicePbxes()->getPBXByDN($user->getPBXDn());

            if (!$pbx) {
                continue;
            }

            $serial = $pbx->getSerial();
            if (!$cachedCustomPorts[$serial]) {
                if (strtolower($serial) == strtolower($this->servicePbx()->getSerial())) {
                    $cachedCustomPorts[$serial] = $this->servicePbx()->getCustomHttpsPort() ?: '443';
                } else {
                    try {
                        $cachedCustomPorts[$serial] = $pbx->getCustomHttpsPort();
                    } catch (\Exception $e) {
                        $cachedCustomPorts[$serial] = '443';
                        error_log($e->getMessage());
                    }
                }
            }

            $users[$idx]['payload']['port'] = $cachedCustomPorts[$serial];

            $list[] = $serial;
            $users[$idx]['payload']['serial'] = $serial;
        }

        return array_unique($list);
    }

    private function isConferenceParticipant($room) {
        if ($_SESSION['videoConferenceParticipations'] && is_array($_SESSION['videoConferenceParticipations'])) {
            if (in_array($room, $_SESSION['videoConferenceParticipations'])) {
                return true;
            }
        }
    }

    private function isConferenceOwner($room) {
        if ($_SESSION['videoConferenceRooms'] && is_array($_SESSION['videoConferenceRooms'])) {
            if (in_array($room, $_SESSION['videoConferenceRooms'])) {
                return true;
            }
        }
        if ($_SESSION['videoConferenceParticipations'] && is_array($_SESSION['videoConferenceParticipations'])) {
            if (in_array($room, $_SESSION['videoConferenceParticipations'])) {
                return false;
            }
        }

        // Let's get info about room. If owner has same jid as our user we'll allow to destroy
        $info = $this->service()->getRoom($room);
        if (is_array($info)) {
            $ownerExtension = '';
            $description    = $info['description'];
            if ($description['owner'] && $this->getUser() != null) {
                // Owner field coms with extension & serial such as 31@22000000002b
                // pbx_owner_jid on VideoConference Service
                // Currently owner is "extension@pbxHost" where pbxHost get from Portal
                $ownerInfo = explode('@', $description['owner']);
                if ($this->getUser()->getExtension() == $ownerInfo[0]) {
                    return true;
                }
            }
        }

        return false;
    }

    private function inviteParticipants(array $data) :void {
        $videoConference = new VideoConferenceModel();
        $videoConference->setUser($this->getUser());
        $videoConference->setConference(
            [
                'room' => $data['room'],
                'pin' => $data['pin'],
                'subject' => $data['subject'],
                'password' => $data['password'] ?? '',
                'passwordPin' => $data['passwordPin'] ?? '',
            ]
        );

        $schedule            = $data['schedule'];
        $schedule['enabled'] = isset($schedule['enabled']) ? filter_var($schedule['enabled'], FILTER_VALIDATE_BOOLEAN) : false;
        $schedule['changed'] = isset($schedule['changed']) ? filter_var($schedule['changed'], FILTER_VALIDATE_BOOLEAN) : false;

        $passwordChanged = isset($data['passwordChanged']) ? filter_var($data['passwordChanged'], FILTER_VALIDATE_BOOLEAN) : false;

        foreach ($data['participants'] as $user) {
            if ($user['type'] !== 'colleague') {
                continue;
            }
            if ($user['isPhantom'] || $user['reinvite']) {
                try {
                    // Skip jabber invitation for online conference users
                    if ($this->service()->memberPresence($data['room'], $user['extension']) === 'available') {
                        continue;
                    }
                    $videoConference->inviteByJabber($user, true, $schedule);
                } catch (\Exception $e) {
                    $this->getLogger()->error(__LINE__, $e->getMessage());
                }
            } else if ($schedule['changed'] && $schedule['enabled'] || $passwordChanged) {
                $videoConference->inviteByJabber($user, true, $schedule);
            }
        }
    }

    private function service(): \Wildix\Services\VideoConference {
        return $this->getContainer()->get(\Wildix\Services\VideoConference::class);
    }

    private function serviceContacts(): \Wildix\Phonebooks\Contacts {
        return $this->getContainer()->get(\Wildix\Phonebooks\Contacts::class);
    }

    private function serviceColleagues(): \Wildix\Colleagues\Colleagues {
        return $this->getContainer()->get(\Wildix\Colleagues\Colleagues::class);
    }

    private function servicePbxes(): PBXes {
        return $this->getContainer()->get(PBXes::class);
    }

    private function servicePbx(): \Models\PBX {
        return $this->getContainer()->get(\Models\PBX::class);
    }

    private function serviceCallFiles(): CallFiles {
        return $this->getContainer()->get(CallFiles::class);
    }
}
