<?php
namespace Controllers;

use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\Colleagues\Colleague;
use Wildix\Error;
use Wildix\Exception;
use Wildix\VoiceMail\VoiceMails;

class VoiceMailController extends Controller {

    public function getItems(Request $request, Response $response, array $args): Response {
        $user = $this->getUser();

        return $this->_getVoiceMailListByUser($user, $request, $response);
    }

    public function update(Request $request, Response $response, array $args): Response {
        $user = $this->getUser();

        return $this->_updateVoiceMails($user, $args['id'], $request, $response);
    }

    public function delete(Request $request, Response $response, array $args): Response {
        $user = $this->getUser();

        return $this->_deleteVoiceMails($user, $args['id'], $request, $response);
    }

    public function getVoiceMailsByUser(Request $request, Response $response, array $args): Response {
        $colleague = $this->_getColleagueByExtension($args['extension']);
        $this->_checkPermissionSeeVoicemail($colleague);

        return $this->_getVoiceMailListByUser($colleague, $request, $response);
    }

    public function updateVoiceMailsByUser(Request $request, Response $response, array $args): Response {
        $colleague = $this->_getColleagueByExtension($args['extension']);
        $this->_checkPermissionSeeVoicemail($colleague);

        return $this->_updateVoiceMails($colleague, $args['id'], $request, $response);
    }

    public function deleteVoiceMailsByUser(Request $request, Response $response, array $args): Response {
        $colleague = $this->_getColleagueByExtension($args['extension']);
        $this->_checkPermissionSeeVoicemail($colleague);

        return $this->_deleteVoiceMails($colleague, $args['id'], $request, $response);
    }

    private function _getVoiceMailListByUser(Colleague $user, Request $request, Response $response): Response {
        $filter = $request->getQueryParams()['filter'] ?? [];

        if (!is_array($filter)) {
            $filter = [];
        }

        $result = $this->service()->getVoiceMails($filter, null, null, null, null, null, null, $user);

        return $this->result(
            [
                'total'   => count($result),
                'records' => $result
            ],
            $response
        );
    }

    private function _updateVoiceMails(Colleague $user, $id, Request $request, Response $response): Response {
        $data = $request->getParsedBody()['data'] ?? '';

        if (empty($data)) {
            return $this->error('Voicemail data cannot be empty', $response, 400);
        }

        $filter = [];
        if ($id !== null) {
            $filter['msg_id'] = $id;
        } else {
            $filter = $request->getQueryParams()['filter'] ?? [];
            if (!is_array($filter)) {
                $filter = [];
            }
        }

        $result = $this->service()->updateVoiceMails($data, $filter, null, null, null, null, null, null, $user);

        if (is_numeric($result)) {
            $result = ['changed' => $result];
        }

        return $this->result($result, $response);
    }

    private function _deleteVoiceMails(Colleague $user, $id, Request $request, Response $response): Response {
        if ($user->canNotUse($user::VOICEMAIL)) {
            return $this->error('Access to Voicemail is not allowed', $response, 403);
        }

        $filter = [];
        if ($id !== null) {
            $filter['msg_id'] = $id;
        } else {
            $filter = $request->getQueryParams()['filter'] ?? [];
            if (!is_array($filter)) {
                $filter = [];
            }
        }

        $result = $this->service()->deleteVoiceMails($filter, null, null, null, null, null, null, $user);

        if (is_numeric($result)) {
            $result = ['changed' => $result];
        }

        return $this->result($result, $response);
    }

    private function _getColleagueByExtension($extension) {
        $colleague = $this->getColleagueRepository()->getUserByExtension($extension);
        if ($colleague === null) {
            throw new Exception(Error::NotFound, 404);
        }

        return $colleague;
    }

    private function _checkPermissionSeeVoicemail($colleague) {
        if (!$this->getUser()->can(Colleague::SEE_VOICEMAIL, $colleague->getGroupDn())) {
            throw new \Wildix\Exception('Access to colleague Voicemail is not allowed', 403);
        }
    }

    /**
     * Returns \Wildix\Colleagues\Colleagues object.
     *
     * @return \Wildix\Colleagues\Colleagues
     */
    private function getColleagueRepository(): \Wildix\Colleagues\Colleagues {
        return $this->getContainer()->get(\Wildix\Colleagues\Colleagues::class);
    }

    private function service(): VoiceMails {
        return $this->getContainer()->get(VoiceMails::class);
    }
}
