<?php
declare(strict_types=1);

namespace Controllers;

use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\Error;
use Wildix\PBXes\Network\NetRepository;
use Wildix\PBXes\Network\NetworkServices;

class WmsNetwork extends Controller {
    public function init(Request $request, Response $response, array $args): Response {
        $this->service()->run($request->getParsedBody());

        return $this->success($response);
    }

    public function getSettings(Request $request, Response $response, array $args): Response {
        return $this->result($this->repository()->get(), $response);
    }

    public function getNtp(Request $request, Response $response, array $args): Response {
        return $this->result($this->service()->getNtp($request->getParsedBody()), $response);
    }

    public function checkCredentials(Request $request, Response $response, array $args): Response {
        $this->service()->checkCredentials($request->getParsedBody());

        return $this->success($response);
    }

    public function addClientToLdap(Request $request, Response $response, array $args): Response {
        $this->service()->addClientToLdap($request->getParsedBody());

        return $this->success($response);
    }

    public function sipPeerPrune(Request $request, Response $response, array $args): Response {
        $this->service()->sipPeerPrune($request->getQueryParams());

        return $this->success($response);
    }

    public function removeUser(Request $request, Response $response, array $args): Response {
        $this->service()->removeUser($request->getQueryParams());

        return $this->success($response);
    }

    public function ldapSlaveSyncStatus(Request $request, Response $response, array $args): Response {
        return $this->sendOutputData('/tmp/ldap_client_sync.out', $response);
    }

    public function ldapUnsetSlaveStatus(Request $request, Response $response, array $args): Response {
        return $this->sendOutputData(\multipbx::LDAP_UNSET_SLAVE_OUT, $response);
    }

    public function checkBackgroundUnsetSlaveState(Request $request, Response $response, array $args): Response {
        return $this->result(
            ['processRunning' => $this->isProcessRunning('ldap_unset_slave')],
            $response
        );
    }

    public function checkBackgroundRemovePbxState(Request $request, Response $response, array $args): Response {
        $result = ['processRunning' => $this->isProcessRunning('wms_network_remove_pbx')];

        if ($result['processRunning'] === false && is_file(\multipbx::REMOVE_PBX_OUT)) {
            $result['errors'] = (string)file_get_contents(\multipbx::REMOVE_PBX_OUT);
        }

        return $this->result($result, $response);
    }

    private function sendOutputData(string $filePath, Response $response): Response {
        if (!file_exists($filePath)) {
            return $this->error(Error::FileNotFound, $response, 404);
        }

        ini_set('zlib.output_compression', 'Off');

        return $response->withHeader('X-Accel-Redirect', sprintf('/inner-storage%s', $filePath));
    }

    /**
     * Checking if process is running.
     *
     * @param string $processName Script name.
     *
     * @return boolean
     */
    private function isProcessRunning(string $processName): bool {
        if ($processName === '') {
            return false;
        }

        exec(sprintf('ps axw | grep "[%s]%s"', substr($processName, 0, 1), substr($processName, 1)), $output);

        return !empty($output);
    }

    private function service(): NetworkServices {
        return $this->getContainer()->get(NetworkServices::class);
    }

    private function repository(): NetRepository {
        return $this->getContainer()->get(NetRepository::class);
    }
}
