<?php
namespace Services;

use Wildix\Databases\QueueDB;

/**
 * Class Helpers
 */
class Helpers {

    const YES = 'yes';
    const NO  = 'no';

    /**
     * Get data from ini file.
     *
     * @deprecated
     *
     * @param string $path    Path.
     * @param array  $default Default.
     *
     * @return array
     */
    public static function iniFileGet(string $path, array $default = []): array {
        $res = $default;

        if (self::isReadableFile($path)) {
            $res = parse_ini_file($path, true, INI_SCANNER_RAW);
            $res = is_array($res) ? $res : $default;
        }
        return (array)$res;
    }

    /**
     * Write ini file
     *
     * @deprecated
     *
     * @param string  $path      Path to ini file.
     * @param array   $data      Saved data.
     * @param boolean $quotes    Use quotes for string values.
     * @param string  $delimiter Delimiter.
     *
     * @return boolean
     */
    public static function iniFilePut(string $path, array $data = [], $quotes = true, $delimiter = '=') :bool {
        if (empty($path) || !is_dir(dirname($path)) || !is_writable(dirname($path))) {
            return false;
        }
        return file_put_contents($path, self::convertArrayToIniData($data, $quotes, $delimiter)) !== false;
    }

    /**
     * Convert array to ini file format.
     *
     * @param array   $data      Converted array.
     * @param boolean $quotes    Use quotes for string values.
     * @param string  $delimiter Delimiter.
     *
     * @return string
     */
    public static function convertArrayToIniData(array $data, $quotes = true, $delimiter = '='): string {
        $content  = '';
        $sections = '';
        foreach ($data as $key => $item) {
            if (is_array($item)) {
                $sections .= PHP_EOL . '[' . $key . ']' . PHP_EOL;
                foreach ($item as $k => $v) {
                    if (is_numeric($v) || is_bool($v) || !$quotes) {
                        $sections .= $k . $delimiter . $v . PHP_EOL;
                    } else {
                        $sections .= $k . $delimiter . '"' . $v . '"' . PHP_EOL;
                    }
                }
            } else {
                if (is_numeric($item) || is_bool($item) || !$quotes) {
                    $content .= $key . $delimiter . $item . PHP_EOL;
                } else {
                    $content .= $key . $delimiter . '"' . $item . '"' . PHP_EOL;
                }
            }
        }
        $content .= $sections;

        return $content;
    }

    /**
     * Format boolean value.
     *
     * @param string|integer|boolean $s       Input value.
     * @param array|null             $variant Variant output values.
     *
     * @return mixed
     */
    public static function formatBoolean($s, ?array $variant = [self::NO, self::YES]) {
        return filter_var($s, FILTER_VALIDATE_BOOLEAN) ? $variant[1] : $variant[0];
    }

    /**
     * Is readable file.
     *
     * @param string $filepath File path.
     *
     * @return boolean
     */
    public static function isReadableFile(string $filepath): bool {
        return !empty($filepath) && is_file($filepath) && is_readable($filepath);
    }

    /**
     * Is writable file.
     *
     * @param string $filepath File path.
     *
     * @return boolean
     */
    public static function isWritableFile(string $filepath): bool {
        return !empty($filepath) && is_file($filepath) && is_writable($filepath);
    }

    /**
     * Is IP address.
     *
     * @param string|null $str Input string.
     *
     * @return boolean
     */
    public static function isIp(?string $str): bool {
        return filter_var($str, FILTER_VALIDATE_IP) !== false;
    }


    /**
     * Get content file
     *
     * @param string $path    Path file.
     * @param string $default Default.
     *
     * @return string
     */
    public static function fileGetContent(string $path, string $default = ''): string {
        $res = $default;
        if (self::isReadableFile($path)) {
            $res = file_get_contents($path);
            $res = !empty($res) ? $res : $default;
        }
        return (string)$res;
    }

    public static function scanDirRecursive(string $path = ''): array {
        $res  = [];
        $dirs = [];
        if (is_dir($path)) {
            $items = self::scanDir($path);
            foreach ($items as $item) {
                if (is_file($path . $item)) {
                    $res[$path][] = $item;
                } else {
                    $dirs[] = $path . $item . '/';
                }
            }
            foreach ($dirs as $dir) {
                $res = array_merge($res, self::scanDirRecursive($dir));
            }
        }
        return $res;
    }

    public static function scanDir(string $path): array {
        $items = [];
        if (is_dir($path) && is_readable($path)) {
            $items = array_diff(scandir($path), ['.', '..']);
        }
        return $items;
    }

    public static function fileNamePrepare(string $name): string {
        $name   = preg_replace('/[^a-zA-Z0-9_.-]/', '', $name);
        $params = explode('.', $name);
        if (count($params) > 1) {
            $ext = end($params);
            array_pop($params);
            $name = implode('', $params) . '.' . $ext;
        }
        return $name;
    }

    /**
     * Explode string IDs by delimiter and cast to integer.
     *
     * @param string $str       Input string.
     * @param string $delimiter Delimiter.
     *
     * @return array
     */
    public static function explodeIdsByDelimiter(string $str, $delimiter = ','): array {
        return '' === $str ? [] : array_filter(array_map('intval', explode($delimiter, $str)));
    }

    /**
     * Add task to queue
     *
     * @param string $type   Type of task.
     * @param array  $params Params of task.
     *
     * @return mixed
     */
    public static function addTask(string $type, array $params = []) {
        try {
            return QueueDB::getInstance()->addJob($type, json_encode($params));
        } catch (\Exception $e) {
            return $e->getMessage();
        }
    }

    /**
     * Make picture url for colleague.
     *
     * @param string $picture
     * @param string $extension
     * @param string $name
     *
     * @return string
     */
    public static function makePictureUrl(string $picture, string $extension, string $name) :string {
        $result = '';

        if ($picture && $extension && $name) {
            $pictureParts = explode('?', $picture);
            $type         = $pictureParts[1] ?? '';
            $result       = "{$pictureParts[0]}?{$extension}&{$name}&{$type}";
        }

        return $result;
    }

    /**
     * Make pbx wms version by ldap field description.
     *
     * @param string $description
     *
     * @return string
     */
    public static function makePbxVersionByDescription(string $description): string {
        $pbxData = json_decode(trim($description, 'pbxData='));
        return property_exists($pbxData, 'wmsVersion') ? explode('~', $pbxData->wmsVersion)[0] : '';
    }
}
