<?php
namespace Services;

use Wildix\Validator;

/**
 * Class LuaService
 */
class LuaService {

    /**
     * Lua instance.
     *
     * @var \lua.
     */
    private $_lua;

    private $_models;

    /**
     * Move user.
     *
     * @param array $params Params [uid, number, pbxip].
     *
     * @return boolean
     */
    public function moveUser(array $params): bool {
        $luaSource = '
                function move_features(uid, number, pbxip)
                    return require"common.userFeatures.importUserFeatures":exec(uid, number, pbxip)
                end
        ';

        return $this->run('move_features', $luaSource, $params) !== false;
    }

    public function isWideband(string $ip): bool {
        $luaSource = '
                function getWideband(ip)
                    local h = require"common.helper"
	                return h.is_wideband(ip)
	            end
        ';
        $this->getLua()->eval($luaSource);
        return $this->getLua()->getWideband($ip);
    }

    public function checkPam(string $login, string $password): bool {
        $luaSource = '
                function check (name, pwd)
                    if require"luapam".lpm_auth(name, pwd) == "SUCCESS" then
                        return true
                    end
                    return false
                end
        ';
        $this->getLua()->eval($luaSource);
        return $this->getLua()->check($login, $password);
    }

    public function checkACL(string $groupDn, string $canType, string $what, ?string $destGroupDn, ?string $licenseType): bool {
        $luaSource = '
                function check (group_dn,can_type,what,dest_group_dn,license_type)
                    local wmsperm = require"common.permissions"
                    return wmsperm:check (group_dn,can_type,what,dest_group_dn,license_type)
                end
        ';
        $this->getLua()->eval($luaSource);
        return Validator::isTrue($this->getLua()->check($groupDn, $canType, $what, $destGroupDn, $licenseType));
    }

    public function saveUserFeatures(string $uid, array $features): bool {
        $luaSource = '
                function save_features(uid, features)
                    return require"common.userFeatures.userFeatures":save(uid, features)
                end
        ';
        $this->getLua()->eval($luaSource);
        return Validator::isTrue($this->getLua()->save_features($uid, $features));
    }

    public function getUserFeatures(?string $uid): array {
        $luaSource = '
                function get_features(uid)
                    return require"common.userFeatures.userFeatures":load(uid)
                end
        ';
        $this->getLua()->eval($luaSource);
        return $this->getLua()->get_features($uid);
    }

    public function getUserDefaultFeatures(?string $uid): array {
        $luaSource = '
                function get_default_features(uid)
				    return require"common.userFeatures.userFeatures":default(uid)
			    end
        ';
        $this->getLua()->eval($luaSource);
        return $this->getLua()->get_default_features($uid);
    }

    public function getCallClasses(): array {
        $luaSource = '
                function get_call_classes ()
				    return require"common.callClass.callClassImpl":getNamesList()
		        end
        ';
        $this->getLua()->eval($luaSource);
        return $this->getLua()->get_call_classes();
    }

    public function getFeatures(): array {
        $file = '/etc/callweaver/res_features.lua';
        $res  = [];
        if(file_exists($file)) {
            $this->getLua()->extensions = [];
            $this->getLua()->include($file);
            $res = $this->getLua()->gl_features;
        }
        return $res;
    }

    public function getModels() {
        if(!$this->_models) {
            $wildixfxs2 = '
		    function getModels()
		        return resources.wildixfxs2[\'models\']
		    end
		';
            $wildixfxsr2 = '
		    function getModels(foo, bar)
		        return resources.wildixfxsr2[\'models\']
		    end
		';
            $wildixgw   = '
		    function getModels()
		        return resources.wildixgw[\'models\']
		    end
		';
            $this->getLua()->include('/var/www/provision/resources/wildixfxs2.lua');
            $this->getLua()->include('/var/www/provision/resources/wildixfxsr2.lua');
            $this->getLua()->include('/var/www/provision/resources/wildixgw.lua');

            $this->getLua()->eval($wildixfxs2);
            $models = $this->getLua()->getModels();

            $this->getLua()->eval($wildixfxsr2);
            $fxsR2   = $this->getLua()->getModels();
            $models = array_merge($models, $fxsR2);

            $this->getLua()->eval($wildixgw);
            $other   = $this->getLua()->getModels();
            $models = array_merge($models, $other['FXS']);

            $this->_models = $models;
        }

        return $this->_models;
    }

    /**
     * Device verify.
     *
     * @param string $ip   IP address.
     * @param int    $port Port.
     *
     * @return array|null
     */
    public function deviceVerify(string $ip, int $port): ?array {
        $device    = [
            'allow' => false,
            'mac' => '',
            'model' => '',
            'address' => $ip,
            'port' => $port
        ];
        $luaSource = '
            package.path = package.path .. ";/var/www/provision/?.lua"
            local discoveryObj = require"discovery":new()

            function getData (args)
                discoveryObj:setArgs(args)
                local data = discoveryObj:run()
                return data
            end
        ';
        $this->getLua()->eval($luaSource);
        $content = $this->getLua()->getData([
            'host' => ['ip' => $ip],
            'port' => ['number' => $port]
        ]);

        if($content) {
            $item = array_values(json_decode( '{' . $content . '}', true))[0];
            preg_match('/(?:(?:[0-9a-f]{2}[\:]{1}){5}|(?:[0-9a-f]{2}[-]{1}){5}|(?:[0-9a-f]{2}){5})[0-9a-f]{2}/i', $item['mac_addr_str'], $matches);
                $device['mac']   = trim($matches[0]);
                $device['model'] = $item['model'];
                $device['allow'] = true;
        }

        return $device;
    }

    /**
     * Change status and delete voicemail.
     *
     * @param string $extension User extension.
     * @param string $messageId Message ID.
     * @param string $action    Action name.
     *
     * @return mixed
     */
    public function manageVoicemail(string $extension, string $messageId, string $action) {
        $luaSource = '
                function actionMsg (extn, msgID, action)
                    local vma = require"common.voiceMailApplication"
                    local params = {["mailbox"] = extn, ["msgId"] = msgID}
                    local success, err
                    local result = {["changed"] = 0, ["error"] = ""}
                    if action == "read" then
                        success, err = vma.markAsReadMsg(params)
                    elseif action == "delete" then
                        success, err = vma.deleteMsg(params)
                    else
                        return false
                    end
                    if success == true then
                        result["changed"] = 1
                    else
                        if err then
                            result["error"] = err
                        else
                            result["error"] = "Unknown"
                        end
                    end
                    vma.updateVM(extn)
                    return result
                end
        ';
        $this->getLua()->eval($luaSource);
        return $this->getLua()->actionMsg($extension, $messageId, $action);
    }

    public function manageVoicemails(string $extension, string $messageId, string $action) {
        $luaSource = '
                function actionMsg (extn, msgIDs, action)
                    local helper = require"common.helper"
                    local vma = require"common.voiceMailApplication"

                    local result = {["changed"] = 0, ["errors"]={}}
                    local arrId = helper.split(msgIDs, ",")

                    local params = {["mailbox"] = extn, ["msgId"] = ""}

                    for i=1, #arrId do
                        params["msgId"] = arrId[i]
                        local success, err
                        if action == "read" then
                            success, err = vma.markAsReadMsg(params)
                        elseif action == "delete" then
                            success, err = vma.deleteMsg(params)
                        end
                        if success == true then
                            result["changed"] = result["changed"]+1
                        else
                            if err then
                                result["errors"][params["msgId"]] = err
                            else
                                result["errors"][params["msgId"]] = "Unknown"
                            end

                        end
                    end

                    vma.updateVM(extn)

                    return result
                end
        ';
        $this->getLua()->eval($luaSource);
        return $this->getLua()->actionMsg($extension, $messageId, $action);
    }

    public function assignWairUser(string $targetExtension, string $name, string $currentExtension, string $mac, string $ipei, string $baseModel, string $contact): bool {
        $luaSource = '
                function assign_wair_user(target_extension, display_name, current_extension, wair_base_mac, headset_ipei, base_model, sipuri)
                    return require"common.helper".assign_wair_user(target_extension, display_name, current_extension, wair_base_mac, headset_ipei, base_model, sipuri)
                end
        ';
        $this->getLua()->eval($luaSource);
        return Validator::isTrue($this->getLua()->assign_wair_user($targetExtension, $name, $currentExtension, $mac, $ipei, $baseModel, $contact));
    }

    /**
     * Execute lua script.
     *
     * @param string $functionName Lua function name.
     * @param string $command      Lua script source.
     * @param array  $params       Params to lua source.
     *
     * @return mixed
     */
    private function run(string $functionName, string $command, ?array $params = null) {
        $this->getLua()->eval($command);
        return $this->getLua()->call($functionName, $params);
    }

    /**
     * Initialize or return Lua, not used DI to increase productivity.
     *
     * @return \lua
     */
    public function getLua() {
        if (!isset($this->_lua)) {
            $this->_lua = new \lua();
        }

        return $this->_lua;
    }
}
