<?php

namespace Wildix\Slim\Middleware\Authorization;

use Psr\Http\Message\ResponseInterface;
use Psr\Http\Message\ServerRequestInterface;
use Psr\Http\Server\MiddlewareInterface;
use Psr\Http\Server\RequestHandlerInterface;
use Slim\Psr7\Factory\ResponseFactory;
use Wildix\Auth\Api\ApiAuthenticator;
use Wildix\Exception\UnauthorizedException;
use Wildix\System\ADSingleSignOnService;

class Authorization extends BaseMiddleware implements MiddlewareInterface {

    private ApiAuthenticator $apiAuthenticator;

    /**
     * Authorization constructor
     *
     * @param ApiAuthenticator      $apiAuthenticator Auth class.
     * @param ADSingleSignOnService $singleSignOn     AD class.
     * @param ResponseFactory       $responseFactory  Response factory.
     */
    public function __construct(ApiAuthenticator $apiAuthenticator, ADSingleSignOnService $singleSignOn, ResponseFactory $responseFactory) {
        parent::__construct($singleSignOn, $responseFactory);
        $this->apiAuthenticator = $apiAuthenticator;
    }

    public function process(ServerRequestInterface $request, RequestHandlerInterface $handler): ResponseInterface {
        $user = $request->getAttribute('user');
        if ($user && $user->getAuthMode() === \User::AUTH_TYPE_SERVICE) {
            throw new UnauthorizedException();
        }

        if ($this->hasClientAppId() && !$this->authenticateApiAuth($request)) {
            return $this->sendErrorUnauthorizedResponse($request);
        }

        if (!$this->authorizedUserToApplication($request)) {
            return $this->sendErrorUnauthorizedResponse($request);
        }

        return $handler->handle($request);
    }

    /**
     * Check if the current request has HTTP_X_APP_ID header.
     *
     * @return boolean
     *
     * redundant check
     * @see    \Wildix\Auth\Api\ApiAuthenticator::isS2sAuth
     */
    private function hasClientAppId(): bool {
        return $this->apiAuthenticator->getAppId() !== null;
    }

    private function authenticateApiAuth(ServerRequestInterface $request): bool {
        $token   = $this->getBearerToken($request);
        $headers = $this->normalizeHeader($request->getHeaders());
        return $this->apiAuthenticator->authenticate(
            (string)$token,
            (string)$request->getMethod(),
            (string)$request->getAttribute('originUri'),
            (array)$headers,
            (array)$request->getQueryParams(),
            $request->getParsedBody()
        );
    }

    private function authorizedUserToApplication(ServerRequestInterface $request): bool {
        $user = $request->getAttribute('user');

        return $user && $user->isLogged();
    }

    private function normalizeHeader(array $getHeaders): array {
        $res = [];
        foreach ($getHeaders as $key => $value) {
            $res[$key] = count($value) == 1 ? $value[0] : $value;
        }

        return $res;
    }
}
