<?php

namespace Wildix\Slim\Middleware\Authorization;

use Psr\Http\Message\ResponseInterface;
use Psr\Http\Message\ServerRequestInterface;
use Slim\Psr7\Factory\ResponseFactory;
use Wildix\Core\Container;
use Wildix\System\ADSingleSignOnService;

class BaseMiddleware {

    protected ADSingleSignOnService $adSingleSignOn;
    protected ResponseFactory $responseFactory;

    public function __construct(ADSingleSignOnService $adSingleSignOn, ResponseFactory $responseFactory) {
         $this->adSingleSignOn  = $adSingleSignOn;
         $this->responseFactory = $responseFactory;
    }

    /**
     * Check need show basic authentication dialog
     *
     * @param ServerRequestInterface $request Server request.
     *
     * @return boolean
     */
    public function needShowBasicAuthDialog(ServerRequestInterface $request): bool {
        $request         = $request->getQueryParams();
        $callback        = $request['callback'];
        $isSkipBasicAuth = ($request['skip-basic-auth-header'] === 'true');

        return (!isset($callback) && !$isSkipBasicAuth);
    }

    /**
     * Send error unauthorized response
     *
     * @param ServerRequestInterface $request Server request.
     *
     * @return ResponseInterface
     */
    public function sendErrorUnauthorizedResponse(ServerRequestInterface $request): ResponseInterface {
        return $this->makeErrorResponse($request, 401, 'Unauthorized', null);
    }

    public function sendErrorResponse(ServerRequestInterface $request, int $httpCode, string $reason, ?array $errors): ResponseInterface {
        return $this->makeErrorResponse($request, $httpCode, $reason, $errors);
    }

    private function makeErrorResponse(ServerRequestInterface $request, int $httpCode, string $reason, ?array $errors): ResponseInterface {
        $response = $this->getResponseFactory()->createResponse($httpCode);
        if (isset($this->adSingleSignOn) && $this->adSingleSignOn->isEnabled()) {
            $response = $response->withHeader('WWW-Authenticate', 'Negotiate');
        } else {
            if ($this->needShowBasicAuthDialog($request)) {
                $response = $response->withHeader('WWW-Authenticate', 'Basic');
            }
        }

        $responseType = $request->getHeaderLine('responseType');
        $responseType = isset($responseType) ? strtolower($responseType) : '';

        return \Controllers\Controller::getResponse(
            [
                'type' => 'error',
                'reason' => $reason,
                'result' => !empty($errors) ? $errors : null,
            ],
            $response,
            $responseType
        );
    }

    /**
     * Get Bearer token or null.
     *
     * @param ServerRequestInterface $request Server request.
     *
     * @return string|null Token or null.
     */
    public function getBearerToken(ServerRequestInterface $request): ?string {
        $tokenMatcher = [];
        // request passed as parameter to middleware
        preg_match('/^Bearer ([a-z0-9-_]+\.[a-z0-9-_]+\.[a-z0-9-_]+)$/i', $request->getHeaderLine('Authorization'), $tokenMatcher);

        return count($tokenMatcher) === 2 ? $tokenMatcher[1] : null;
    }

    /**
     * Get responseFactory
     *
     * @return ResponseFactory
     */
    public function getResponseFactory(): ResponseFactory {
        return $this->responseFactory;
    }


    protected function getLog() {
        return Container::get(\Logger\Log::class);
    }
}
