<?php

namespace Wildix\Slim\Middleware\Authorization;

use Psr\Http\Message\ResponseInterface;
use Psr\Http\Message\ServerRequestInterface;
use Psr\Http\Server\MiddlewareInterface;
use Psr\Http\Server\RequestHandlerInterface;
use Slim\Psr7\Factory\ResponseFactory;
use Wildix\Auth\ServiceAuth\ServiceAuthenticator;
use Wildix\Exception\UnauthorizedException;

class Service implements MiddlewareInterface {

    protected ServiceAuthenticator $serviceAuthenticator;
    protected ResponseFactory $responseFactory;
    protected string $serviceName;
    private \Logger\Log $logger;

    /**
     * ServiceMiddleware constructor
     *
     * @param ServiceAuthenticator $serviceAuthenticator Authenticator.
     */
    public function __construct(ServiceAuthenticator $serviceAuthenticator, ResponseFactory $responseFactory, \Logger\Log $logger, string $serviceName) {
        $this->serviceAuthenticator = $serviceAuthenticator;
        $this->responseFactory      = $responseFactory;
        $this->serviceName          = $serviceName;
        $this->logger               = $logger;
    }

    public function process(ServerRequestInterface $request, RequestHandlerInterface $handler): ResponseInterface {
        $this->serviceAuthenticator->initServiceClass($this->serviceName);
        if (!$this->_checkServiceAuth($request)) {
            throw new UnauthorizedException();
        }

        return $handler->handle($request);
    }

    /**
     * Get Bearer token or null.
     *
     * @param ServerRequestInterface $request Server request.
     *
     * @return string|null Token or null.
     */
    public function getBearerToken(ServerRequestInterface $request): ?string {
        $tokenMatcher = array();
        // request passed as parameter to middleware
        preg_match('/^Bearer ([\w\.-]+)$/i', $request->getHeaderLine('Authorization'), $tokenMatcher);

        return count($tokenMatcher) === 2 ? $tokenMatcher[1] : null;
    }


    private function _checkServiceAuth($request): bool {
        $token = $this->getBearerToken($request);
        if (!isset($token)) {
            $this->getLog()->error(__LINE__, '[ServiceAuth] Empty JWT token');
            return false;
        }

        $service = $this->serviceAuthenticator->authenticate($token);
        if (!$service) {
            $this->getLog()->error(
                __LINE__,
                sprintf('[ServiceAuth] Failed authentication in service: %s', $this->serviceAuthenticator->getServiceName())
            );
            return false;
        }
        $this->getLog()->info(
            __LINE__,
            sprintf('[ServiceAuth] Success authentication in service: %s', $this->serviceAuthenticator->getServiceName())
        );
        $request->withAttribute('service', $service);

        return true;
    }

    private function getLog(): \Logger\Log {
        return $this->logger;
    }
}
