<?php

namespace Wildix\Slim\Middleware\Check;

use Psr\Container\ContainerInterface;
use Psr\Http\Message\ResponseInterface;
use Psr\Http\Message\ServerRequestInterface;
use Psr\Http\Server\MiddlewareInterface;
use Psr\Http\Server\RequestHandlerInterface;
use Slim\Psr7\Factory\ResponseFactory;
use Wildix\Auth\ServiceAuth\ServiceAuthenticator;
use Wildix\Auth\ServiceAuth\ServiceAuthenticatorFactory;
use Wildix\Auth\ServiceAuth\Services\BaseServiceAuthInterface;
use Wildix\Exception\UnauthorizedException;
use Wildix\Slim\Middleware\Authorization\BaseMiddleware;
use Wildix\System\ADSingleSignOnService;

class AllowForService extends BaseMiddleware implements MiddlewareInterface {

    /**
     * Service name
     *
     * @var string
     */
    private $serviceAuthenticatorName;

    /**
     * Container
     *
     * @var ContainerInterface
     */
    private $container;

    /**
     * Service
     *
     * @var BaseServiceAuthInterface|ServiceAuthenticator $serviceAuthenticator Service.
     */
    private $serviceAuthenticator;

    /**
     * Allow for service constructor.
     *
     * @param string                $serviceAuthenticatorName Service.
     * @param ADSingleSignOnService $adSingleSignOn           Service SingleSignOn.
     * @param ResponseFactory       $responseFactory          Response factory interface.
     * @param ContainerInterface    $container                Response factory interface.
     */
    public function __construct(
        $serviceAuthenticatorName,
        ADSingleSignOnService $adSingleSignOn,
        ResponseFactory $responseFactory,
        ContainerInterface $container
    ) {
        parent::__construct($adSingleSignOn, $responseFactory);
        $this->serviceAuthenticatorName = $serviceAuthenticatorName;
        $this->container                = $container;
    }

    public function process(ServerRequestInterface $request, RequestHandlerInterface $handler): ResponseInterface {
        if ($this->checkServiceAuth($request)) {
            return $handler->handle($request);
        }
        throw new UnauthorizedException();
    }

    private function checkServiceAuth(ServerRequestInterface $request): bool {
        $token = $this->getBearerToken($request);
        if (!isset($token)) {
            $this->getLog()->error(__LINE__, '[ServiceAuth] Empty JWT token');

            return false;
        }

        $service = $this->getServiceAuthenticator()->authenticate($token);
        if (!$service) {
            $this->getLog()->error(
                __LINE__,
                sprintf(
                    '[ServiceAuth] Failed authentication in service: %s',
                    $this->getServiceAuthenticator()->getServiceName()
                )
            );

            return false;
        }

        $this->getLog()->info(
            __LINE__,
            sprintf(
                '[ServiceAuth] Success authentication in service: %s',
                $this->getServiceAuthenticator()->getServiceName()
            )
        );
        $request->withAttribute('service', $service);

        return true;
    }

    private function getServiceAuthenticator(): ServiceAuthenticator|BaseServiceAuthInterface {
        if ($this->serviceAuthenticator === null) {
            $serviceAuthenticatorFactory = $this->getContainer()->get(ServiceAuthenticatorFactory::class);
            $this->serviceAuthenticator  = $serviceAuthenticatorFactory->buildService($this->serviceAuthenticatorName);
        }
        return $this->serviceAuthenticator;
    }

    /**
     * Get container
     *
     * @return ContainerInterface
     */
    public function getContainer(): ContainerInterface {
        return $this->container;
    }
}
