<?php
declare(strict_types=1);

namespace Wildix\Slim\Middleware;

use Psr\Http\Message\ResponseInterface;
use Psr\Http\Message\ServerRequestInterface;
use Psr\Http\Server\MiddlewareInterface;
use Psr\Http\Server\RequestHandlerInterface;
use Wildix\HttpSecurity\Service;

class CorsMiddleware implements MiddlewareInterface {

    private const DEFAULT_ALLOWED_HEADERS = [
        'accept',
        'authorization',
        'content-type',
    ];

    private Service $httpSecurity;

    public function __construct(Service $httpSecurity) {
        $this->httpSecurity = $httpSecurity;
    }

    public function process(ServerRequestInterface $request, RequestHandlerInterface $handler): ResponseInterface {
        $response = $handler->handle($request);

        $requestOrigin = $request->getHeaderLine('Origin');
        $requestHost   = $request->getHeaderLine('Host');
        if ($this->isCrossOrigin($requestOrigin, $requestHost)) {
            if ($request->hasHeader('Origin') && $this->httpSecurity->isOriginAllow($requestOrigin)) {
                $response = $response->withHeader('Access-Control-Allow-Origin', $requestOrigin);
            }

            $requestHeaders = implode(',', self::DEFAULT_ALLOWED_HEADERS);
            // @TODO check ChromeOS app usage / outdated but released in chrome store
            if ($this->httpSecurity->isWizyconfOrigin($requestOrigin)) {
                $requestHeaders = $request->getHeaderLine('Access-Control-Request-Headers');
            }

            $response = $response->withHeader('Access-Control-Allow-Methods', 'GET, POST, PUT, DELETE, OPTIONS');
            if ($requestHeaders !== '') {
                $response = $response->withHeader('Access-Control-Allow-Headers', $requestHeaders);
            }

            $response = $response->withHeader('Access-Control-Allow-Credentials', 'true');
        }

        return $response;
    }

    private function isCrossOrigin(string $requestOrigin, string $requestHost): bool {
        // no origin, no CORS
        if ($requestOrigin === '') {
            return false;
        }

        $parsedUrl         = parse_url($requestOrigin);
        $requestOriginHost = $parsedUrl['host'] ?? '';
        $isCrossOrigin     = false;
        if (strcasecmp($requestOriginHost, $requestHost) !== 0) {
            $isCrossOrigin = true;
        }

        return $isCrossOrigin;
    }
}
