<?php

declare(strict_types=1);

use Api\Support\Enums\UserCanTypes;
use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;
use Slim\App;
use Slim\Exception\HttpNotFoundException;
use Slim\Psr7\Factory\ResponseFactory;
use Slim\Routing\RouteCollectorProxy;
use Wildix\Auth\Api\ApiAuthenticator;
use Wildix\Auth\Devices\Authenticator as DevicesAuthenticator;
use Wildix\Auth\ServiceAuth\Services\KiteAuth;
use Wildix\Auth\ServiceAuth\Services\PhonebridgeAuth;
use Wildix\Auth\ServiceAuth\Services\WizywebinarAuth;
use Wildix\Auth\SmsService\SmsServiceAuthenticator;
use Wildix\Colleagues\Colleague;
use Wildix\Slim\Middleware\Authorization\Authorization;
use Wildix\Slim\Middleware\Authorization\Device;
use Wildix\Slim\Middleware\Authorization\Logical;
use Wildix\Slim\Middleware\Authorization\Portal;
use Wildix\Slim\Middleware\Authorization\ScreenSharing;
use Wildix\Slim\Middleware\Authorization\Sms;
use Wildix\Slim\Middleware\Check\AccessControlList;
use Wildix\Slim\Middleware\Check\AllowForAdmin;
use Wildix\Slim\Middleware\Check\AllowForRoot;
use Wildix\Slim\Middleware\Check\AllowForService;
use Wildix\Slim\Middleware\Check\UserCan;
use Wildix\Slim\Middleware\SetUserToRequestMiddleware;
use Wildix\Slim\Middleware\Support\ExtractUserDataFromSession;
use Wildix\Slim\Middleware\Support\NormalizeCallChannel;
use Wildix\System\ADSingleSignOnService;

return static function (App $app) {
    $app->options(
        '/{routes:.*}',
        function (Request $request, Response $response) {
        // CORS Pre-Flight OPTIONS Request Handler
            return $response;
        }
    );
    // paste all routes here
    // routes definitions
    // middlewares
    $responseFactory  = new ResponseFactory();
    $adSingleSignOn   = $app->getContainer()->get(ADSingleSignOnService::class);
    $user             = $app->getContainer()->get(User::class);
    $setUserToRequest = new SetUserToRequestMiddleware($user);
    $authorization    = new Authorization(
        $app->getContainer()->get(ApiAuthenticator::class),
        $adSingleSignOn,
        $responseFactory
    );

    $pbx                 = $app->getContainer()->get(\Models\PBX::class);
    $wmsPermissions      = WMS_Permissions::getInstance();
    $portalAuthorization = new Portal($responseFactory, $pbx);
    $smsAuthorization    = new Sms(
        $app->getContainer()->get(SmsServiceAuthenticator::class),
        $adSingleSignOn,
        $responseFactory
    );
    $deviceAuthorization = new Device(
        $app->getContainer()->get(DevicesAuthenticator::class),
        $adSingleSignOn,
        $responseFactory
    );
    // region User can
    $screenSharingAuthorization = new ScreenSharing($responseFactory, $pbx);
    $userCanUseVoiceMail        = new UserCan(UserCanTypes::CAN, Colleague::VOICEMAIL);
    $checkUserCanUseCDR         = new UserCan(UserCanTypes::CAN_USE, $user::CTI_CDR);
    $checkUserCanRemoveCalls    = new UserCan(UserCanTypes::CAN, $user::DELETE_CALLS);
    $userTwoFactorAuthOff       = new UserCan(UserCanTypes::CAN, Colleague::TWO_FACTOR_AUTH_OFF);
    $userCdrView                = new UserCan(UserCanTypes::CAN_USE, $user::CTI_CDR_VIEW);
    $checkUserCanFaxes          = new UserCan(UserCanTypes::CAN_USE, $user::FAX);
    // endregion
    // region AccessControlList
    $checkAclDialplanPagingGroups          = new AccessControlList($wmsPermissions, 'page', 'dialplan__paging_groups');
    $checkAclDialplanCallGroups            = new AccessControlList($wmsPermissions, 'page', 'dialplan__call_groups');
    $checkAclDialplanIvr                   = new AccessControlList($wmsPermissions, 'page', 'dialplan__ivr');
    $checkAclDialplanFeatureCode           = new AccessControlList($wmsPermissions, 'page', 'dialplan__feature_code');
    $checkAclDialplanTimetables            = new AccessControlList($wmsPermissions, 'page', 'dialplan__timetables');
    $checkAclDialplanGeneralSettings       = new AccessControlList($wmsPermissions, 'page', 'dialplan__general_settings');
    $checkAclSettingsUtilitesBackup        = new AccessControlList($wmsPermissions, 'page', 'settings__utilites__backup');
    $checkAclSettingsSystem                = new AccessControlList($wmsPermissions, 'page', 'settings__system');
    $checkAclSettingsUtilitesRemoteSupport = new AccessControlList($wmsPermissions, 'page', 'settings__utilites__remote_support');
    $checkAclSettingsUtilitesActivation    = new AccessControlList($wmsPermissions, 'page', 'settings__utilites__activation');
    $checkAclSettingsPbx                   = new AccessControlList($wmsPermissions, 'page', 'settings__pbx');
    $checkAclSettingsUtilitesUpgrade       = new AccessControlList($wmsPermissions, 'page', 'settings__utilites__upgrade');
    $checkAclUsersGroups                   = new AccessControlList($wmsPermissions, 'page', 'users__groups');
    $checkAclDialplanDialplan              = new AccessControlList($wmsPermissions, 'page', 'dialplan__dialplan');
    $checkAclUsersUsers                    = new AccessControlList($wmsPermissions, 'page', 'users__users');
    $checkAclAddRemoveUsers                = new AccessControlList($wmsPermissions, 'permissions', 'add_remove_users');
    $checkAclSettingsUtilitesGenerateTrace = new AccessControlList($wmsPermissions, 'page', 'settings__utilites__generate_trace');
    $checkAclTopControlHalt                = new AccessControlList($wmsPermissions, 'page', 'top_control__halt');
    $checkAclTrunksTrunks                  = new AccessControlList($wmsPermissions, 'page', 'trunks__trunks');
    $checkAclTrunksHuntGroups              = new AccessControlList($wmsPermissions, 'page', 'trunks__hunt_groups');
    $checkAclTrunksPricelists              = new AccessControlList($wmsPermissions, 'page', 'trunks__pricelists');

    $checkAclUsersPbxs = new AccessControlList($wmsPermissions, 'page', 'users__pbxs');
    $checkAclDevices   = new AccessControlList($wmsPermissions, 'page', 'devices');

    // endregion
    $normalizeCallChannel = new NormalizeCallChannel();
    $allowForAdmin        = new AllowForAdmin();
    $allowForRoot         = new AllowForRoot();

    $logicKite                = new Logical($adSingleSignOn, $responseFactory);
    $middlewareForServiceKite = new AllowForService(
        KiteAuth::SERVICE_KITE,
        $adSingleSignOn,
        $responseFactory,
        $app->getContainer()
    );
    $logicKite->addMiddleware($middlewareForServiceKite);
    $logicKite->addMiddleware($authorization);

    $logicPhonebridge                = new Logical($adSingleSignOn, $responseFactory);
    $middlewareForServicePhonebridge = new AllowForService(
        PhonebridgeAuth::SERVICE_PHONEBRIDGE,
        $adSingleSignOn,
        $responseFactory,
        $app->getContainer()
    );
    $logicPhonebridge->addMiddleware($middlewareForServicePhonebridge);
    $logicPhonebridge->addMiddleware($allowForAdmin);

    $logicWizywebinar                = new Logical($adSingleSignOn, $responseFactory);
    $middlewareForServiceWizywebinar = new AllowForService(
        WizywebinarAuth::SERVICE_WIZYWEBINAR,
        $adSingleSignOn,
        $responseFactory,
        $app->getContainer()
    );
    $logicWizywebinar->addMiddleware($middlewareForServiceWizywebinar);
    $logicWizywebinar->addMiddleware($authorization);

    // region Routes
    // public
    $app->get('/api/google/singlesignon/', '\Controllers\GoogleController:singleSignOn');
    $app->get('/api/google/callback/', '\Controllers\GoogleController:callback');

    $app->get('/api/microsoft/singlesignon/', '\Controllers\Microsoft:singleSignOn');
    $app->get('/api/microsoft/callback/', '\Controllers\Microsoft:callback');
    $app->get('/api/zohocrm/callback/', '\Controllers\CRM\ZohoCrm:callback');
    $app->get('/api/office365/callback/', '\Controllers\Office365:callback');

    $app->get('/api/infusionsoft/callback/', '\Controllers\CRM\InfusionsoftController:callback');
    $app->get('/api/v1/infusionsoft/authurl/', '\Controllers\CRM\InfusionsoftController:getAuthUrl')
        ->addMiddleware($allowForAdmin)
        ->addMiddleware($setUserToRequest);

    $app->get('/api/comeback/', '\Controllers\Comeback:endpoint')
        ->addMiddleware($authorization)
        ->addMiddleware($setUserToRequest);

    // region authorization
    $app->group(
        '/authorization',
        function (RouteCollectorProxy $app) use ($authorization) {
            $app->get('/oauth2/', '\Controllers\OAuth2\Auth:getAuthForm');
            $app->get('/oauth2grant/', '\Controllers\OAuth2\Auth:getGrantForm')->addMiddleware(
                $authorization
            );
            $app->post('/oauth2grant/', '\Controllers\OAuth2\Auth:grantConfirm')->addMiddleware(
                $authorization
            );
            $app->post('/oauth2token/', '\Controllers\OAuth2\Auth:getTokens');

            $app->post('/oauth2login/', '\Controllers\Personal:login');
            $app->get('/oauth2logout/', '\Controllers\Personal:logout');
        }
    )->addMiddleware($setUserToRequest);
    // endregion

    // region Group /api/v1
    $app->group(
        '/api/v1',
        function (RouteCollectorProxy $app) use (
            $wmsPermissions,
            $authorization,
            $screenSharingAuthorization,
            $portalAuthorization,
            $smsAuthorization,
            $deviceAuthorization,
            $allowForAdmin,
            $normalizeCallChannel,
            $allowForRoot,
            $adSingleSignOn,
            $responseFactory,
            $user,
            $userCanUseVoiceMail,
            $checkUserCanUseCDR,
            $checkUserCanRemoveCalls,
            $userTwoFactorAuthOff,
            $userCdrView,
            $checkUserCanFaxes,
            $checkAclDialplanPagingGroups,
            $checkAclDialplanCallGroups,
            $checkAclDialplanIvr,
            $checkAclDialplanFeatureCode,
            $checkAclDialplanTimetables,
            $checkAclDialplanGeneralSettings,
            $checkAclSettingsUtilitesBackup,
            $checkAclSettingsUtilitesRemoteSupport,
            $checkAclSettingsUtilitesActivation,
            $checkAclSettingsSystem,
            $checkAclSettingsPbx,
            $checkAclSettingsUtilitesUpgrade,
            $checkAclUsersGroups,
            $checkAclDialplanDialplan,
            $checkAclUsersUsers,
            $checkAclAddRemoveUsers,
            $checkAclDevices,
            $checkAclSettingsUtilitesGenerateTrace,
            $checkAclTopControlHalt,
            $checkAclTrunksTrunks,
            $checkAclTrunksHuntGroups,
            $checkAclTrunksPricelists,
            $checkAclUsersPbxs,
            $logicKite,
            $logicPhonebridge,
            $logicWizywebinar
        ) {
            $app->post('/chat/messages/', '\Controllers\Jabber:sendMessage')->addMiddleware($authorization);
            // region Group /personal
            $app->group(
                '/personal',
                function (RouteCollectorProxy $app) use ($userTwoFactorAuthOff) {
                // Private Personal API (user should have authorization on this pbx)
                    $app->post('/roster/', '\Controllers\Personal:postRoster');
                    $app->get('/acl/', '\Controllers\Personal:getAcl');
                    $app->delete('/roster/', '\Controllers\Personal:deleteRoster');
                    $app->put('/roster/', '\Controllers\Personal:putRoster');
                    $app->get('/info/', '\Controllers\Personal:getInfo');
                    $app->post('/settings/', '\Controllers\Personal:postSettings');
                    $app->put('/settings/', '\Controllers\Personal:postSettings');
                    $app->post('/settings/twofactor/', '\Controllers\Personal:postTwoFactor');
                    $app->delete('/settings/twofactor/', '\Controllers\Personal:deleteTwoFactor')->addMiddleware($userTwoFactorAuthOff);
                    $app->put('/settings/password/', '\Controllers\Personal:putPassword');
                    $app->get('/settings/', '\Controllers\Personal:getSettings');
                    $app->get('/logout/', '\Controllers\Personal:logout');
                    $app->put('/presence/', '\Controllers\Personal:putPresence');
                    $app->put('/presence/location/', '\Controllers\Personal:putPresenceLocation');
                    $app->get('/paginggroups/', '\Controllers\Personal:getPagingGroups');
                    $app->get('/token/', '\Controllers\Personal:getAuthToken');
                // Public Personal API
                    $app->post('/devices/', '\Controllers\Personal:addDevice');
                }
            )->addMiddleware($authorization);
            // endregion
            $app->post('/settings/twofactor/{secondFactor}/code/', '\Controllers\Personal:sendCodeTo2ndFactor')
                ->addMiddleware(new ExtractUserDataFromSession(User::getInstance()));
            $app->post('/personal/login/', '\Controllers\Personal:login');
            // Public PBX API
            $app->get('/pbx/ping/', '\Controllers\PBX:ping');
            // region Group /pbx
            $app->group(
                '/pbx',
                function (RouteCollectorProxy $app) use (
                    $checkAclSettingsUtilitesRemoteSupport,
                    $checkAclSettingsUtilitesActivation,
                    $checkAclTopControlHalt,
                    $wmsPermissions,
                    $allowForAdmin,
                    $allowForRoot,
                    $user,
                    $checkAclDialplanDialplan,
                    $checkAclUsersUsers,
                    $checkAclUsersGroups,
                    $checkAclSettingsUtilitesUpgrade,
                    $checkAclAddRemoveUsers,
                    $checkAclSettingsPbx,
                    $checkAclSettingsSystem,
                    $checkAclSettingsUtilitesBackup,
                    $checkAclDialplanCallGroups
                ) {
                    // Private PBX API
                    $app->get('/version/', '\Controllers\PBX:getVersion');
                    $app->get('/secureport/', '\Controllers\PBX:getSecurePort');
                    $app->get('/candidates/', '\Controllers\PBX:getCandidates');
                    $app->get('/icecandidates/', '\Controllers\PBX:getIceCandidates');
                    $app->get('/externallineprefix/', '\Controllers\PBX:getExternalLinePrefix');
                    $app->get('/{serial}/redirect/', '\Controllers\PBX:redirect');

                    $app->get('/ports/status/', '\Controllers\PBX:getPortsStatus');
                    $app->delete('/ports/status/', '\Controllers\PBX:recheckPortsStatus');

                    $app->get('/settings/smtp/', '\Controllers\Smtp:getSettings')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->put('/settings/smtp/', '\Controllers\Smtp:setSettings')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->post('/settings/smtp/test/', '\Controllers\Smtp:runTest')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->get('/settings/smtp/test/status/{testId}/', '\Controllers\Smtp:getTestStatus')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->get('/settings/dhcp/', '\Controllers\DhcpServer:getSettings')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->put('/settings/dhcp/', '\Controllers\DhcpServer:setSettings')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->get('/settings/dhcp/leases/', '\Controllers\DhcpServer:getLeases')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->put('/settings/dhcp/leases/', '\Controllers\DhcpServer:setLeases')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->delete('/settings/dhcp/leases/', '\Controllers\DhcpServer:deleteLeases')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->delete('/settings/dhcp/leases/purge/', '\Controllers\DhcpServer:purgeLeases')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);

                    $app->get('/settings/callqueues/', '\Controllers\Dialplans\CallGroupsQueueController:getItems')->addMiddleware($checkAclDialplanCallGroups);
                    $app->get('/settings/callqueues/{id:\d+}/', '\Controllers\Dialplans\CallGroupsQueueController:getItem')->addMiddleware($checkAclDialplanCallGroups);
                    $app->post(
                        '/settings/callqueues/{id:\d+}/members/dynamic/',
                        '\Controllers\Dialplans\CallGroupsQueueController:addMember'
                    )->addMiddleware($checkAclDialplanCallGroups);
                    $app->delete(
                        '/settings/callqueues/{id:\d+}/members/dynamic/{member:.+\/.+}/',
                        '\Controllers\Dialplans\CallGroupsQueueController:deleteMember'
                    )->addMiddleware($checkAclDialplanCallGroups);
                    $app->get('/settings/network/interfaces/', '\Controllers\Network:getInterfaces')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    // Private PBX remote support
                    $app->put('/settings/remotesupport/', '\Controllers\RemoteSupport:setType')->addMiddleware($checkAclSettingsUtilitesRemoteSupport)->addMiddleware($allowForAdmin);
                    $app->post('/settings/remotesupport/', '\Controllers\RemoteSupport:checkState')->addMiddleware($checkAclSettingsUtilitesRemoteSupport)->addMiddleware($allowForAdmin);
                    $app->post('/settings/remotesupport/location/', '\Controllers\RemoteSupport:getLocation')->addMiddleware($checkAclSettingsUtilitesRemoteSupport)->addMiddleware($allowForAdmin);
                    $app->put('/settings/remotesupport/location/', '\Controllers\RemoteSupport:setLocation')->addMiddleware($checkAclSettingsUtilitesRemoteSupport)->addMiddleware($allowForAdmin);

                    $app->get('/settings/faxsms/generalsettings/', '\Controllers\FaxSmsSettings:getSettings')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->put('/settings/faxsms/generalsettings/', '\Controllers\FaxSmsSettings:setSettings')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->get('/settings/faxsms/profiles/', '\Controllers\FaxSmsProfiles:getProfiles')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->post('/settings/faxsms/profiles/', '\Controllers\FaxSmsProfiles:createProfile')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->put('/settings/faxsms/profiles/{profileId}/', '\Controllers\FaxSmsProfiles:updateProfile')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->delete('/settings/faxsms/profiles/{profileIds}/', '\Controllers\FaxSmsProfiles:removeProfiles')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->get('/settings/faxsms/testprofile/{profileId}/', '\Controllers\FaxSmsProfiles:runFetchmailTest');
                    $app->get(
                        '/settings/faxsms/testprofilestatus/{profileId}/',
                        '\Controllers\FaxSmsProfiles:getFetchmailTestStatus'
                    );
                    $app->get('/settings/snmp/', '\Controllers\Snmp:getSettings')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->put('/settings/snmp/', '\Controllers\Snmp:setSettings')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);

                    $app->get('/settings/ldapradius/', '\Controllers\LdapRadius:getSettings')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->put('/settings/ldapradius/', '\Controllers\LdapRadius:setSettings')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);

                    $app->put('/settings/license/', '\Controllers\LicenseController:refresh')->addMiddleware($checkAclSettingsUtilitesActivation)->addMiddleware($allowForAdmin);
                    $app->delete('/settings/license/', '\Controllers\LicenseController:deactivate')->addMiddleware($checkAclSettingsUtilitesActivation)->addMiddleware($allowForAdmin);
                    $app->post('/settings/license/', '\Controllers\LicenseController:activate')->addMiddleware($checkAclSettingsUtilitesActivation)->addMiddleware($allowForAdmin);
                    $app->get('/settings/license/', '\Controllers\LicenseController:get')->addMiddleware($checkAclSettingsUtilitesActivation)->addMiddleware($allowForAdmin);
                    $app->put('/settings/logo/', '\Controllers\LicenseController:downloadLogo')->addMiddleware($checkAclSettingsUtilitesActivation)->addMiddleware($allowForAdmin);
                    $app->delete('/settings/logo/', '\Controllers\LicenseController:deleteLogo')->addMiddleware($checkAclSettingsUtilitesActivation)->addMiddleware($allowForAdmin);
                    $app->post('/settings/sslcertificate/', '\Controllers\SSLCertificate:renew')->addMiddleware($checkAclSettingsUtilitesActivation)->addMiddleware($allowForAdmin);
                    $app->get(
                        '/settings/sslcertificate/backgroundrenewstate/',
                        '\Controllers\SSLCertificate:checkBackgroundRenewState'
                    )->addMiddleware($checkAclSettingsUtilitesActivation)->addMiddleware($allowForAdmin);
                    $app->get('/settings/languageandregion/', '\Controllers\LanguageAndRegionController:getSettings')->addMiddleware($checkAclSettingsPbx);
                    $app->post('/settings/languageandregion/', '\Controllers\LanguageAndRegionController:setSettings')->addMiddleware($checkAclSettingsPbx);
                    $app->get('/settings/languageandregion/status/', '\Controllers\LanguageAndRegionController:getStatus')->addMiddleware($checkAclSettingsPbx);

                    // region Group /settings/storages
                    $app->group(
                        '/settings/storages',
                        function (RouteCollectorProxy $app) {
                            $app->get('/', '\Controllers\Storages:getData');

                            $app->post('/initusb/', '\Controllers\Storages:initUsb');
                            $app->post('/mountusb/', '\Controllers\Storages:mountUsb');
                            $app->post('/unmountusb/', '\Controllers\Storages:unMountUsb');
                            $app->post('/remountshare/', '\Controllers\Storages:remountShare');
                            $app->post('/unmountsmb/', '\Controllers\Storages:unMountSmb');
                            $app->post('/mountsmb/', '\Controllers\Storages:mountSmb');
                            $app->post('/mountnfs/', '\Controllers\Storages:mountNfs');
                            $app->post('/mounttoexistingshare/', '\Controllers\Storages:mountToExistingShare');
                        }
                    )->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    // endregion
                    // region Group /settings
                    $app->group(
                        '/settings',
                        function (RouteCollectorProxy $app) {
                            $app->get('/httpsecurity/', '\Controllers\HttpSecurity:getSettings');
                            $app->post('/httpsecurity/', '\Controllers\HttpSecurity:setSettings');
                            $app->get('/adsinglesignon/', '\Controllers\ADSingleSignOnController:getSettings');
                            $app->post('/adsinglesignon/', '\Controllers\ADSingleSignOnController:setSettings');
                            $app->delete('/adsinglesignon/', '\Controllers\ADSingleSignOnController:deleteKeyTab');
                            $app->get('/tls/', '\Controllers\Settings\Security\Tls:getSettings');
                            $app->put('/tls/', '\Controllers\Settings\Security\Tls:setSettings');
                        }
                    )->addMiddleware($checkAclSettingsSystem);
                    // endregion
                    // NTP
                    $app->put('/settings/ntp/', '\Controllers\Ntp:setSettings')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->get('/settings/ntp/', '\Controllers\Ntp:getSettings')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);

                    // Remote Syslog Server
                    $app->get('/settings/remotesyslog/', '\Controllers\RemoteSyslog:getSettings')->addMiddleware($checkAclSettingsSystem);
                    $app->put('/settings/remotesyslog/', '\Controllers\RemoteSyslog:setSettings')->addMiddleware($checkAclSettingsSystem);

                    // Http Proxy
                    $app->put('/settings/httpproxy/', '\Controllers\HttpProxy:setSettings')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    $app->get('/settings/httpproxy/', '\Controllers\HttpProxy:getSettings')->addMiddleware($checkAclSettingsSystem)->addMiddleware($allowForAdmin);
                    // Queue metrics
                    $app->get('/settings/qmetrics/', '\Controllers\QueueMetrics:getSettings');
                    $app->put('/settings/qmetrics/', '\Controllers\QueueMetrics:setSettings');

                    // Upgrade page
                    $app->get('/upgrade/settings/', '\Controllers\Upgrade:getSettings')->addMiddleware($checkAclSettingsUtilitesUpgrade)->addMiddleware($allowForAdmin);
                    $app->put('/upgrade/settings/', '\Controllers\Upgrade:setSettings')->addMiddleware($checkAclSettingsUtilitesUpgrade)->addMiddleware($allowForAdmin);

                    $app->post('/upgrade/check/', '\Controllers\Upgrade:runCheckUpgrade')->addMiddleware($checkAclSettingsUtilitesUpgrade)->addMiddleware($allowForAdmin);
                    $app->get('/upgrade/check/', '\Controllers\Upgrade:getCheckerStatus')->addMiddleware($checkAclSettingsUtilitesUpgrade)->addMiddleware($allowForAdmin);

                    $app->post('/upgrade/', '\Controllers\Upgrade:runUpgrade')->addMiddleware($checkAclSettingsUtilitesUpgrade)->addMiddleware($allowForAdmin);
                    $app->get('/upgrade/', '\Controllers\Upgrade:getUpgradeStatus')->addMiddleware($checkAclSettingsUtilitesUpgrade)->addMiddleware($allowForAdmin);

                    // Sip proxy
                    $app->get('/users/sip/registrations/', '\Controllers\Sipproxy:getUsersRegistrations')->addMiddleware($allowForAdmin);
                    $app->get('/users/{extensions}/sip/registrations/', '\Controllers\Sipproxy:getUsersRegistrations')->addMiddleware($allowForAdmin);

                    $app->get('/callhistory/', '\Controllers\CallHistory:getHistory')->addMiddleware($allowForAdmin);

                    // Check if FS has I/O errors
                    $app->get('/filesystem/status/', '\Controllers\FileSystem:isFileSystemWithErrors');
                    $app->get('/filesystem/status.log/', '\Controllers\FileSystem:getLogFile');

                    // Reboot / PowerOff PBX
                    $app->put('/system/reboot/', '\Controllers\SystemController:reboot')
                        ->addMiddleware($checkAclSettingsUtilitesUpgrade)
                        ->addMiddleware($checkAclSettingsUtilitesBackup)
                        ->addMiddleware($checkAclTopControlHalt);
                    $app->put('/system/poweroff/', '\Controllers\SystemController:powerOff')
                        ->addMiddleware($checkAclSettingsUtilitesUpgrade)
                        ->addMiddleware($checkAclSettingsUtilitesBackup)
                        ->addMiddleware($checkAclTopControlHalt);


                    // Analytics Integration
                    $app->get('/integration/analytics/', '\Controllers\Integrations\AnalyticsController:getItems')->addMiddleware($allowForAdmin);
                    $app->put('/integration/analytics/{service}/', '\Controllers\Integrations\AnalyticsController:update')->addMiddleware($allowForAdmin);

                    // Zoho CRM Integration
                    $app->get('/zohocrm/', '\Controllers\CRM\ZohoCrm:getStatus')->addMiddleware($allowForAdmin);
                    $app->post('/zohocrm/', '\Controllers\CRM\ZohoCrm:enable')->addMiddleware($allowForAdmin);
                    $app->delete('/zohocrm/', '\Controllers\CRM\ZohoCrm:disable')->addMiddleware($allowForAdmin);
                    $app->post('/zohocrm/confirm/', '\Controllers\CRM\ZohoCrm:confirm')->addMiddleware($allowForAdmin);

                    // Office365 Integration
                    $app->get('/office365/', '\Controllers\Office365:getStatus')->addMiddleware($allowForAdmin);
                    $app->get('/office365/adminconsenturl/', '\Controllers\Office365:adminConsentUrl')->addMiddleware($allowForAdmin);
                    $app->post('/office365/', '\Controllers\Office365:enable')->addMiddleware($allowForAdmin);
                    $app->delete('/office365/', '\Controllers\Office365:disable')->addMiddleware($allowForAdmin);

                    // x-bees integrations
                    $app->get('/integration/x-bees/', '\Controllers\Integrations\XBeesController:get')->addMiddleware($allowForAdmin);

                    // Auth applications config
                    $app->get('/applications/s2s/', '\Controllers\Applications\S2sController:getItems')->addMiddleware($allowForRoot);
                    $app->post('/applications/s2s/', '\Controllers\Applications\S2sController:create')->addMiddleware($allowForRoot);
                    $app->put('/applications/s2s/{id}/', '\Controllers\Applications\S2sController:update')->addMiddleware($allowForRoot);
                    $app->delete('/applications/s2s/{id}/', '\Controllers\Applications\S2sController:remove')->addMiddleware($allowForRoot);

                    $app->get('/applications/oauth2/', '\Controllers\Applications\OAuth2Controller:getItems')->addMiddleware($allowForRoot);
                    $app->post('/applications/oauth2/', '\Controllers\Applications\OAuth2Controller:create')->addMiddleware($allowForRoot);
                    $app->put('/applications/oauth2/{id}/', '\Controllers\Applications\OAuth2Controller:update')->addMiddleware($allowForRoot);
                    $app->delete('/applications/oauth2/{id}/', '\Controllers\Applications\OAuth2Controller:remove')->addMiddleware($allowForRoot);

                    $app->get('/applications/simpletoken/', '\Controllers\Applications\SimpleTokenController:getItems')->addMiddleware($allowForRoot);
                    $app->post('/applications/simpletoken/', '\Controllers\Applications\SimpleTokenController:create')->addMiddleware($allowForRoot);
                    $app->put('/applications/simpletoken/{id}/', '\Controllers\Applications\SimpleTokenController:update')->addMiddleware($allowForRoot);
                    $app->delete('/applications/simpletoken/{id}/', '\Controllers\Applications\SimpleTokenController:remove')->addMiddleware($allowForRoot);

                    // CRM settings
                    $app->get('/settings/crm/', '\Controllers\CrmSettings:getSettings')->addMiddleware($checkAclSettingsPbx)->addMiddleware($allowForAdmin);
                    $app->put('/settings/crm/', '\Controllers\CrmSettings:setSettings')->addMiddleware($checkAclSettingsPbx)->addMiddleware($allowForAdmin);

                    // WMS network
                    $app->get('/wmsnetwork/ldapslavesyncstatus/', '\Controllers\WmsNetwork:ldapSlaveSyncStatus');
                    $app->get('/wmsnetwork/ldapunsetslavestatus/', '\Controllers\WmsNetwork:ldapUnsetSlaveStatus');
                    $app->get('/wmsnetwork/checkbackgroundunsetslavestate/', '\Controllers\WmsNetwork:checkBackgroundUnsetSlaveState');
                    $app->get('/wmsnetwork/checkbackgroundremovepbxstate/', '\Controllers\WmsNetwork:checkBackgroundRemovePbxState');
                    $app->get('/wmsnetwork/', '\Controllers\WmsNetwork:getSettings')->addMiddleware($allowForAdmin);
                    $app->put('/wmsnetwork/', '\Controllers\WmsNetwork:init')->addMiddleware($allowForAdmin);
                    $app->put('/wmsnetwork/checkcredentials/', '\Controllers\WmsNetwork:checkCredentials');
                    $app->put('/wmsnetwork/addclienttoldap/', '\Controllers\WmsNetwork:addClientToLdap');
                    $app->get('/wmsnetwork/sippeerprune/', '\Controllers\WmsNetwork:sipPeerPrune');
                    $app->post('/wmsnetwork/ntp/', '\Controllers\WmsNetwork:getNtp');
                    $app->delete('/wmsnetwork/removeuser/', '\Controllers\WmsNetwork:removeUser');

                    // ACL
                    $app->get('/aclgroups/', '\Controllers\Groups:getManagedGroups')->addMiddleware($checkAclUsersGroups);
                    $app->post('/aclgroups/', '\Controllers\Groups:create')->addMiddleware($checkAclUsersGroups)->addMiddleware($allowForRoot);
                    $app->put('/aclgroups/{id}/', '\Controllers\Groups:update')->addMiddleware($checkAclUsersGroups);
                    $app->delete('/aclgroups/{ids:[\d,]+}/', '\Controllers\Groups:remove')->addMiddleware($checkAclUsersGroups)->addMiddleware($allowForRoot);
                    $app->get('/aclgroups/rules/', '\Controllers\Groups:getRules');

                    // Dialplans
                    $app->get('/dialplans/context/', '\Controllers\Dialplans\Context:getNames')->addMiddleware($checkAclDialplanDialplan);
                    $app->get('/dialplans/applications/', '\Controllers\Dialplans\RulesController:getApplications')->addMiddleware($checkAclDialplanDialplan);
                    $app->get('/dialplans/', '\Controllers\Dialplans\RulesController:getItems')->addMiddleware($checkAclDialplanDialplan);
                    $app->get('/dialplans/{id}/', '\Controllers\Dialplans\RulesController:getItem')->addMiddleware($checkAclDialplanDialplan);
                    $app->post('/dialplans/', '\Controllers\Dialplans\RulesController:create')->addMiddleware($checkAclDialplanDialplan);
                    $app->put('/dialplans/{id}/', '\Controllers\Dialplans\RulesController:update')->addMiddleware($checkAclDialplanDialplan);
                    $app->delete('/dialplans/{ids}/', '\Controllers\Dialplans\RulesController:remove')->addMiddleware($checkAclDialplanDialplan);

                    // Private Colleagues API
                    $app->get('/colleagues/', '\Controllers\Colleagues:getItemsAcl')->addMiddleware($checkAclUsersUsers);
                    $app->get('/colleagues/{id}/', '\Controllers\Colleagues:getItemAcl')->addMiddleware($checkAclUsersUsers);
                    $app->post('/colleagues/', '\Controllers\Colleagues:create')->addMiddleware($checkAclUsersUsers)->addMiddleware($checkAclAddRemoveUsers);
                    $app->put('/colleagues/{id}/', '\Controllers\Colleagues:update')->addMiddleware($checkAclUsersUsers);
                    $app->delete('/colleagues/{id:\d+}/', '\Controllers\Colleagues:remove')->addMiddleware($checkAclUsersUsers)->addMiddleware($checkAclAddRemoveUsers);
                    $app->delete('/colleagues/{ids:[\d,]+}/', '\Controllers\Colleagues:removeItems');

                    $app->get('/whitelabel/', '\Controllers\Settings\WhiteLabelController:getSettings');
                    $app->post('/whitelabel/', '\Controllers\Settings\WhiteLabelController:setSettings')->addMiddleware($allowForAdmin);
                    $app->delete('/whitelabel/', '\Controllers\Settings\WhiteLabelController:resetSettings')->addMiddleware($allowForAdmin);
                }
            )->addMiddleware($authorization);
            // endregion
            // Public User API (operated user should allow this operation by ACL)
            // region Group /user
            $app->group(
                '/user',
                function (RouteCollectorProxy $app) use ($logicWizywebinar, $logicPhonebridge, $logicKite, $authorization, $adSingleSignOn, $responseFactory, $allowForAdmin) {
                    $app->get('/{extension}/public/info/', '\Controllers\User:getPublicUserInfo')->addMiddleware($logicKite);
                    $app->get('/{extension}/public/presence/', '\Controllers\User:getPublicPresence')->addMiddleware($authorization);
                    $app->get('/{extension}/presence/', '\Controllers\User:getPresence')->addMiddleware($authorization);

                    $app->put('/{extension}/presence/', '\Controllers\User:putPresence')->addMiddleware($logicWizywebinar);
                    $app->get('/{extension}/location/', '\Controllers\User:getLocation')->addMiddleware($authorization);
                    $app->get('/{extension}/info/', '\Controllers\User:getUserInfo')->addMiddleware($authorization);
                    $app->put('/{extension}/', '\Controllers\Colleagues:updateColleagueByExtension')->addMiddleware($allowForAdmin)->addMiddleware($authorization);
                    $app->post('/{extension}/message/', '\Controllers\User:sendMessage')->addMiddleware($authorization);

                    $app->post('/{extension}/call/', '\Controllers\Astman:startCall')->addMiddleware($logicPhonebridge);
                    $app->get('/{extension}/voicemail/', '\Controllers\VoiceMailController:getVoiceMailsByUser')->addMiddleware($authorization);
                    $app->put('/{extension}/voicemail/{id}/', '\Controllers\VoiceMailController:updateVoiceMailsByUser')->addMiddleware($authorization);
                    $app->delete('/{extension}/voicemail/{id}/', '\Controllers\VoiceMailController:deleteVoiceMailsByUser')->addMiddleware($authorization);
                }
            );
            // endregion
        // Private Colleagues API
            $app->get('/colleagues/', '\Controllers\Colleagues:getItems')->addMiddleware($authorization);
            $app->get('/colleagues/{id}/', '\Controllers\Colleagues:getItem')->addMiddleware($authorization);
            // region Group / device
            $app->post('/device/login/', '\Controllers\Devices\DevicesController:login')->addMiddleware($allowForRoot);
            $app->post('/device/logout/', '\Controllers\Devices\DevicesController:logout')->addMiddleware($allowForRoot);
            $app->put('/device/{id}/', '\Controllers\Devices\DevicesController:update')->addMiddleware($authorization);
            $app->post('/device/resettoken/', '\Controllers\Devices\DevicesController:resetToken')->addMiddleware($authorization);
            $app->post('/device/sync/', '\Controllers\Devices\DevicesController:sync')->addMiddleware($allowForAdmin);

            // endregion
            // region Group /devices
            $app->group(
                '/devices',
                function (RouteCollectorProxy $app) use ($allowForAdmin, $wmsPermissions, $checkAclDevices) {
                    $app->get('/', '\Controllers\Devices\DevicesController:getItems')->addMiddleware($checkAclDevices);
                    $app->post('/', '\Controllers\Devices\DevicesController:create')->addMiddleware($checkAclDevices);
                    $app->post('/sync/', '\Controllers\Devices\DevicesController:synchronizationItems')->addMiddleware($checkAclDevices);
                    $app->post('/scan/', '\Controllers\Devices\DevicesController:scanAddress')->addMiddleware($checkAclDevices);
                    $app->get('/verify/', '\Controllers\Devices\DevicesController:verify')->addMiddleware($checkAclDevices);
                    $app->put('/{id}/', '\Controllers\Devices\DevicesController:update')->addMiddleware($checkAclDevices);
                    $app->post('/{id}/sync/', '\Controllers\Devices\DevicesController:synchronization')->addMiddleware($checkAclDevices);
                    $app->post('/{id}/connect/', '\Controllers\Devices\DevicesController:connect')->addMiddleware($checkAclDevices);
                    $app->delete('/{id}/connect/', '\Controllers\Devices\DevicesController:disconnect')->addMiddleware($checkAclDevices);
                    $app->delete('/{id}/', '\Controllers\Devices\DevicesController:remove')->addMiddleware($checkAclDevices);

                    $app->get('/gateways/', '\Controllers\Devices\GatewaysController:getItems');

                    $app->get('/networks/', '\Controllers\Devices\NetworksController:getItems');
                    $app->post('/networks/', '\Controllers\Devices\NetworksController:create');
                    $app->put('/networks/{id}/', '\Controllers\Devices\NetworksController:update');
                    $app->delete('/networks/{id}/', '\Controllers\Devices\NetworksController:remove');

                    $app->get('/wair/', '\Controllers\Devices\WairDevicesController:getItems')->addMiddleware($allowForAdmin);
                    $app->put('/wair/{id}/', '\Controllers\Devices\WairDevicesController:update')->addMiddleware($allowForAdmin);
                }
            )->addMiddleware($authorization);
            // endregion
            // Private ScreenSharing API
            $app->post('/screensharing/room/', '\Controllers\ScreenSharing:create')->addMiddleware($authorization);
            $app->get('/screensharing/room/{roomId}/', '\Controllers\ScreenSharing:getItem')->addMiddleware($authorization);
            $app->delete('/screensharing/room/{roomId}/', '\Controllers\ScreenSharing:remove')->addMiddleware($authorization);

            // Private Notes API
            $app->get('/notes/', '\Controllers\Notes:getItems')->addMiddleware($authorization);
            $app->delete('/notes/{noteId}/', '\Controllers\Notes:remove')->addMiddleware($authorization);
            // region Group /contacts
            $app->group(
                '/contacts',
                function (RouteCollectorProxy $app) {
                    $app->get('/{contactId}/notes/', '\Controllers\Notes:getNotesByContact');
                    $app->post('/{contactId}/notes/', '\Controllers\Notes:addNoteToContact');
                    $app->delete('/{contactId}/notes/{noteId}/', '\Controllers\Notes:deleteNoteFromContact');

                    // Private contacts API
                    $app->get('/{ids}/', '\Controllers\ContactsController:getItems');
                    $app->get('/', '\Controllers\ContactsController:getItems');
                    $app->post('/', '\Controllers\ContactsController:create');
                    $app->put('/{contactId}/', '\Controllers\ContactsController:update');
                    $app->delete('/{contactId}/', '\Controllers\ContactsController:remove');
                }
            )->addMiddleware($authorization);
            // endregion
            // region Group /phonebooks
            $app->group(
                '/phonebooks',
                function (RouteCollectorProxy $app) {
                    $app->get('/', '\Controllers\PhonebooksController:getItems');
                    $app->post('/', '\Controllers\PhonebooksController:create');
                    $app->get('/{id:\d+}/', '\Controllers\PhonebooksController:getItem');
                    $app->put('/[{id:\d+}/]', '\Controllers\PhonebooksController:update');
                    $app->delete('/{id:\d+}/', '\Controllers\PhonebooksController:remove');

                    $app->get('/{phonebookId}/contacts/', '\Controllers\ContactsController:getPhonebookContacts');
                    $app->post('/{phonebookId}/contacts/', '\Controllers\ContactsController:create');
                    $app->delete('/{phonebookId}/contacts/', '\Controllers\ContactsController:removePhonebookContacts');
                    $app->get('/{phonebookId}/contacts/{contactId}/', '\Controllers\ContactsController:getPhonebookContact');
                    $app->put('/{phonebookId}/contacts/{contactId}/', '\Controllers\ContactsController:updatePhonebookContact');
                    $app->delete('/{phonebookId}/contacts/{contactId}/', '\Controllers\ContactsController:removePhonebookContact');
                }
            )->addMiddleware($authorization);
            // endregion
        // Private calendar API
            // region Group /outlook
            $app->group(
                '/outlook',
                function (RouteCollectorProxy $app) {
                    $app->put('/calendar/', '\Controllers\OutlookCalendar:updateCalendar');
                    $app->get('/calendar/', '\Controllers\OutlookCalendar:getCalendar');
                    $app->put('/settings/', '\Controllers\OutlookCalendar:updateSettings');
                }
            )->addMiddleware($authorization);
            // endregion
            // region Group /cdrview
        // Private CDR-View API
            $app->group(
                '/cdrview',
                function (RouteCollectorProxy $app) {
                    $app->get('/settings/', '\Controllers\CDRViewController:getSettings');
                    $app->put('/settings/', '\Controllers\CDRViewController:setSettings');
                    $app->get('/filters/', '\Controllers\CDRViewController:getFilters');
                    $app->post('/filters/', '\Controllers\CDRViewController:addFilter');
                    $app->delete('/filters/{id}/', '\Controllers\CDRViewController:deleteFilter');
                    $app->get('/extraproperties/', '\Controllers\CDRViewController:getExtraProperties');
                    $app->get('/infoproperties/', '\Controllers\CDRViewController:getInfoProperties');
                    $app->get('/callshistory/', '\Controllers\CDRViewController:getCallsHistory');
                }
            )->addMiddleware($userCdrView)->addMiddleware($authorization);
            // endregion
        // Private Predictive dialer API
            $app->post('/predictivedialer/calls/', '\Controllers\PredictiveDialer:makeCall')->addMiddleware($authorization);
            $app->get('/predictivedialer/calls/', '\Controllers\PredictiveDialer:getCurrentCalls')->addMiddleware($authorization);

        // Events
        // Private Ios CTIconnect API (user should have authorization on this pbx)
            $app->get('/ios/config/', '\Controllers\Ios:getSettings')->addMiddleware($authorization);
        // Private Integration Utility CTIconnect API (user should have authorization on this pbx)
            $app->get('/integrationutility/config/', '\Controllers\Ios:getSettings')->addMiddleware($authorization);

            $app->get('/wizyconf/config/', '\Controllers\Devices\DevicesController:getWizyconfConfig')->addMiddleware($deviceAuthorization);

            $app->get('/groups/', '\Controllers\Groups:getItems')->addMiddleware($authorization);

        // Private Kite API
            $app->post('/kite/emailtranscript/', '\Controllers\Kite:emailTranscript');

            $app->get('/doc/[{file:.*}]', '\Controllers\Doc:file')->addMiddleware($allowForAdmin)->addMiddleware($authorization);

            $app->get('/dictionaries/', '\Controllers\Dictionaries:getSettings')->addMiddleware($authorization);

            $app->get('/departments/', '\Controllers\Departments:getItems')->addMiddleware($authorization);

        // Private Conference API (user should have authorization on this pbx)
            // region Group videoconference
            $app->group(
                '/videoconference',
                function (RouteCollectorProxy $app) {
                    $app->get('/rooms/', '\Controllers\VideoConference:getRooms');
                    $app->post('/rooms/', '\Controllers\VideoConference:createRoom');
                    $app->delete('/rooms/{id}/', '\Controllers\VideoConference:destroyRoom');
                    $app->post('/chatinvite/', '\Controllers\VideoConference:jabberInvitation');
                    // deprecated use videoConference/Rooms/ or VideoConference/Rooms/:roomId/
                    $app->post('/invite/', '\Controllers\VideoConference:invite');
                    $app->post('/{id}/call/', '\Controllers\VideoConference:call');
                    $app->post('/kick/', '\Controllers\VideoConference:kick');
                    $app->post('/pushhistory/', '\Controllers\VideoConference:pushHistory');
                    $app->get('/{id}/schedule/', '\Controllers\VideoConference:schedule');
                    $app->post('/authorize/', '\Controllers\VideoConference:authorize');
                }
            )->addMiddleware($authorization);
            // endregion
            // region Group /originate
        // Private Call API (user should have authorization on this pbx)
            $app->group(
                '/originate',
                function (RouteCollectorProxy $app) use ($allowForAdmin) {
                    $app->post('/', '\Controllers\Astman:postOriginate')->addMiddleware($allowForAdmin);
                    $app->post('/mobility/', '\Controllers\Astman:postMobilityCall');
                    $app->post('/call/', '\Controllers\Astman:postCall');
                    $app->post('/playback/', '\Controllers\Astman:postPlayback');
                    $app->post('/ringtone/', '\Controllers\Astman:postRingtone');
                    $app->post('/sms/', '\Controllers\Astman:sendSms');
                    $app->post('/alarmclocks/', '\Controllers\Astman:postAlarmClock')->addMiddleware($allowForAdmin);
                    $app->get('/alarmclocks/', '\Controllers\Astman:getAlarmClocks');
                    $app->put('/alarmclocks/{id}/', '\Controllers\Astman:editAlarmClock')->addMiddleware($allowForAdmin);
                    $app->delete('/alarmclocks/{id}/', '\Controllers\Astman:deleteAlarmClock')->addMiddleware($allowForAdmin);
                    $app->delete('/{type}/{id}/', '\Controllers\Astman:deleteCall');
                    $app->get('/{type}/{id}/', '\Controllers\Astman:getCall');
                }
            )->addMiddleware($authorization);
            // endregion
            // region Group /fax
            $app->group(
                '/fax',
                function (RouteCollectorProxy $app) {
                    // Private Fax API
                    $app->post('/', '\Controllers\FaxController:sendFax');
                    $app->post('/file/', '\Controllers\FaxController:uploadFax');
                    $app->get('/queue/', '\Controllers\FaxController:getQueue');
                }
            )->addMiddleware($checkUserCanFaxes)->addMiddleware($authorization);
            // endregion
            // region Group /callhistory
            $app->group(
                '/callhistory',
                function (RouteCollectorProxy $app) use ($checkUserCanUseCDR, $checkUserCanRemoveCalls) {
                    $app->get('/', '\Controllers\CallHistory:getItems')->addMiddleware($checkUserCanUseCDR);
                    $app->get('/{id}/', '\Controllers\CallHistory:getItem')->addMiddleware($checkUserCanUseCDR);
                    $app->put('/{id}/', '\Controllers\CallHistory:update')->addMiddleware($checkUserCanUseCDR);
                    $app->delete('/{id}/', '\Controllers\CallHistory:remove')->addMiddleware($checkUserCanUseCDR)->addMiddleware($checkUserCanRemoveCalls);
                    $app->put('/{id}/tags/', '\Controllers\CallHistory:updateTags');
                    $app->put('/{id}/voicemail/', '\Controllers\CallHistory:readVoicemail');
                    $app->delete('/{id}/voicemail/', '\Controllers\CallHistory:removeVoicemail');
                }
            )->addMiddleware($authorization);
            // endregion
            $app->get('/user/{extension}/callhistory/', '\Controllers\CallHistory:getItems')->addMiddleware($checkUserCanUseCDR)->addMiddleware($authorization);

            $app->post('/calls/{channel:.*\/\d*\-.*\_\d*}/answer/', '\Controllers\Astman:answerCall')->addMiddleware($logicPhonebridge)->addMiddleware($normalizeCallChannel);
            $app->delete('/calls/{channel:.*\/\d*\-.*\_\d*}/', '\Controllers\Astman:endCall')->addMiddleware($logicPhonebridge)->addMiddleware($normalizeCallChannel);

            // region Group /calls
            $app->group(
                '/calls',
                function (RouteCollectorProxy $app) {
                    $app->post('/', '\Controllers\Astman:sendCall');
                    $app->post('/{channel}/transfer/', '\Controllers\Astman:postTransfer');
                    $app->get('/{channel}/', '\Controllers\Astman:getCallStatus');
                    $app->put('/{channel}/tags/', '\Controllers\Astman:editTags');
                    $app->get('/{channel}/tags/', '\Controllers\Astman:getTags');
                    $app->get('/{channel}/recordings/', '\Controllers\Astman:getRecordings');
                    $app->post('/{channel}/update/', '\Controllers\Astman:sipUpdate');
                    $app->post('/{channel}/dtmf/', '\Controllers\Astman:playDtmf');
                }
            )->addMiddleware($authorization)->addMiddleware($normalizeCallChannel);
            // endregion
            // region Group /voicemail
            $app->group(
                '/voicemail',
                function (RouteCollectorProxy $app) {
                    $app->get('/', '\Controllers\VoiceMailController:getItems');
                    $app->put('/{id}/', '\Controllers\VoiceMailController:update');
                    $app->delete('/{id}/', '\Controllers\VoiceMailController:delete');
                }
            )->addMiddleware($userCanUseVoiceMail)->addMiddleware($authorization);
        // Private Dialplan DB API (user should have authorization on this pbx)
            // endregion
            // region Group /dialplan
            $app->group(
                '/dialplan',
                function (RouteCollectorProxy $app) use (
                    $allowForAdmin,
                    $wmsPermissions,
                    $checkAclDialplanPagingGroups,
                    $checkAclDialplanGeneralSettings,
                    $checkAclDialplanTimetables,
                    $checkAclDialplanFeatureCode,
                    $checkAclDialplanIvr,
                    $checkAclDialplanCallGroups,
                    $checkAclSettingsUtilitesBackup
                ) {
                    $app->get('/db/{family}/[{key:.*}/]', '\Controllers\Dialplans\DbController:getItems');
                    $app->post('/db/{family}/[{key:.*}/]', '\Controllers\Dialplans\DbController:create')->addMiddleware($allowForAdmin);
                    $app->put('/db/{family}/[{key:.*}/]', '\Controllers\Dialplans\DbController:update')->addMiddleware($allowForAdmin);
                    $app->delete('/db/{family}/[{key:.*}/]', '\Controllers\Dialplans\DbController:remove')->addMiddleware($allowForAdmin);

                    $app->get('/paginggroups/', '\Controllers\Dialplans\PagingGroupsController:getItems')->addMiddleware($checkAclDialplanPagingGroups);
                    $app->post('/paginggroups/', '\Controllers\Dialplans\PagingGroupsController:create')->addMiddleware($checkAclDialplanPagingGroups);
                    $app->put('/paginggroups/{id}/', '\Controllers\Dialplans\PagingGroupsController:update')->addMiddleware($checkAclDialplanPagingGroups);
                    $app->delete('/paginggroups/{id}/', '\Controllers\Dialplans\PagingGroupsController:remove')->addMiddleware($checkAclDialplanPagingGroups);

                    $app->get('/callgroups/', '\Controllers\Dialplans\CallGroupsController:getItems')->addMiddleware($checkAclDialplanCallGroups);
                    $app->get('/callgroups/{id}/stat/', '\Controllers\Dialplans\CallGroupsController:getInfoByItem');
                    $app->post('/callgroups/', '\Controllers\Dialplans\CallGroupsController:create')->addMiddleware($checkAclDialplanCallGroups);
                    $app->put('/callgroups/{id}/', '\Controllers\Dialplans\CallGroupsController:update')->addMiddleware($checkAclDialplanCallGroups);
                    $app->delete('/callgroups/{id}/', '\Controllers\Dialplans\CallGroupsController:remove')->addMiddleware($checkAclDialplanCallGroups);

                    $app->get('/ivr/', '\Controllers\Dialplans\IvrController:getItems')->addMiddleware($checkAclDialplanIvr);
                    $app->post('/ivr/', '\Controllers\Dialplans\IvrController:create')->addMiddleware($checkAclDialplanIvr);
                    $app->put('/ivr/{id}/', '\Controllers\Dialplans\IvrController:update')->addMiddleware($checkAclDialplanIvr);
                    $app->delete('/ivr/{id}/', '\Controllers\Dialplans\IvrController:remove')->addMiddleware($checkAclDialplanIvr);

                    $app->get('/featurecodes/', '\Controllers\Dialplans\FeatureCodesController:getSettings')->addMiddleware($checkAclDialplanFeatureCode);
                    $app->put('/featurecodes/', '\Controllers\Dialplans\FeatureCodesController:setSettings')->addMiddleware($checkAclDialplanFeatureCode);

                    $app->get('/timetables/', '\Controllers\Dialplans\TimeTablesController:getItems')->addMiddleware($checkAclDialplanTimetables);
                    $app->post('/timetables/', '\Controllers\Dialplans\TimeTablesController:create')->addMiddleware($checkAclDialplanTimetables);
                    $app->put('/timetables/{id}/', '\Controllers\Dialplans\TimeTablesController:update')->addMiddleware($checkAclDialplanTimetables);
                    $app->delete('/timetables/{id}/', '\Controllers\Dialplans\TimeTablesController:remove')->addMiddleware($checkAclDialplanTimetables);

                    $app->get('/switches/', '\Controllers\Dialplans\SwitchesController:getItems')->addMiddleware($checkAclDialplanTimetables);
                    $app->post('/switches/', '\Controllers\Dialplans\SwitchesController:create')->addMiddleware($checkAclDialplanTimetables);
                    $app->put('/switches/{id}/', '\Controllers\Dialplans\SwitchesController:update')->addMiddleware($checkAclDialplanTimetables);
                    $app->delete('/switches/{id}/', '\Controllers\Dialplans\SwitchesController:remove')->addMiddleware($checkAclDialplanTimetables);

                    $app->get('/generalsettings/', '\Controllers\Dialplans\GeneralSettingsController:getSettings')->addMiddleware($checkAclDialplanGeneralSettings);
                    $app->put('/generalsettings/', '\Controllers\Dialplans\GeneralSettingsController:setSettings')->addMiddleware($checkAclDialplanGeneralSettings);

                    // Private Dialplan Debug API (user should have authorization on this pbx)
                    $app->post('/debug/config/', '\Controllers\Dialplans\DebugController:setConfig');
                    $app->get('/debug/calllogs/', '\Controllers\Dialplans\DebugController:getItems');
                    // Private dialplan rules apI (user should have authorization on this pbx)
                    $app->post('/rules/import/', '\Controllers\Dialplans\RulesController:import')->addMiddleware($checkAclSettingsUtilitesBackup)->addMiddleware($allowForAdmin);
                    $app->post('/rules/export/', '\Controllers\Dialplans\RulesController:export')->addMiddleware($checkAclSettingsUtilitesBackup)->addMiddleware($allowForAdmin);
                    $app->get('/rules/restorebackup/', '\Controllers\Dialplans\RulesController:restoreBackup')->addMiddleware($checkAclSettingsUtilitesBackup)->addMiddleware($allowForAdmin);
                    $app->get('/rules/pingexport/', '\Controllers\Dialplans\RulesController:pingExport')->addMiddleware($checkAclSettingsUtilitesBackup)->addMiddleware($allowForAdmin);
                    $app->get('/rules/downloadbackup/', '\Controllers\Dialplans\RulesController:downloadBackup')->addMiddleware($checkAclSettingsUtilitesBackup)->addMiddleware($allowForAdmin);
                }
            )->addMiddleware($authorization);
            // endregion
            // region Group /events
            $app->group(
                '/events',
                function (RouteCollectorProxy $app) use ($screenSharingAuthorization, $portalAuthorization) {
                    $app->post('/license/', '\Controllers\Events:onChangedLicence')->addMiddleware($portalAuthorization);
                    $app->post('/domain/', '\Controllers\Events:onChangedDomain')->addMiddleware($portalAuthorization);
                    $app->post('/devices/', '\Controllers\Events:onChangedDevices')->addMiddleware($portalAuthorization);
                    $app->post('/ports/', '\Controllers\Events:onChangedPortsStatus')->addMiddleware($portalAuthorization);
                    $app->post('/upgrade/', '\Controllers\Events:onUpgradePerform')->addMiddleware($portalAuthorization);
                    $app->post('/screensharing/{event}/', '\Controllers\ScreenSharing:onReceiveEvent')->addMiddleware($screenSharingAuthorization);
                    $app->post('/apikeyspair/', '\Controllers\Events:onChangedApiKeysPair')->addMiddleware($portalAuthorization);
                    $app->post('/wizywebinar/sync/', '\Controllers\Integrations\WizyWebinarController:syncContacts')->addMiddleware($portalAuthorization);
                }
            )->addMiddleware($authorization);
            // endregion
            // region Group /sounds
            $app->group(
                '/sounds',
                function (RouteCollectorProxy $app) {
                    $app->post('/directory/', '\Controllers\SoundsController:createDirectory');
                    $app->put('/directory/{path:.*}/', '\Controllers\SoundsController:editDirectory');
                    $app->delete('/directory/{path:.*}/', '\Controllers\SoundsController:removeDirectory');
                    $app->post('/', '\Controllers\SoundsController:create');
                    $app->get('/check/{path:.*}/', '\Controllers\SoundsController:check');
                    $app->get('/[{path:.*}/]', '\Controllers\SoundsController:getItems');
                    $app->put('/[{path:.*}/]', '\Controllers\SoundsController:update');
                    $app->delete('/[{path:.*}/]', '\Controllers\SoundsController:remove');
                }
            )->addMiddleware($allowForAdmin)->addMiddleware($authorization);
        // GenerateTrace
            // endregion
            // region Group /utilities
            $app->group(
                '/utilities',
                function (RouteCollectorProxy $app) {
                    $app->get('/trace/status/', '\Controllers\Utilities\TraceController:status');
                    $app->post('/trace/start/', '\Controllers\Utilities\TraceController:start');
                    $app->post('/trace/stop/', '\Controllers\Utilities\TraceController:stop');
                    $app->post('/trace/remove/', '\Controllers\Utilities\TraceController:remove');
                }
            )->addMiddleware($checkAclSettingsUtilitesGenerateTrace)->addMiddleware($allowForAdmin)->addMiddleware($authorization);
            // endregion
            $app->get('/sms/chatmembers/', '\Controllers\SmsController:getColleagues')->addMiddleware($smsAuthorization);

            // region Group /trunks
            $app->group(
                '/trunks',
                function (RouteCollectorProxy $app) use ($checkAclTrunksTrunks, $checkAclTrunksPricelists, $checkAclTrunksHuntGroups) {
                    $app->get('/sip/', '\Controllers\Trunks\SipTrunksController:getItems')->addMiddleware($checkAclTrunksTrunks);
                    $app->post('/sip/', '\Controllers\Trunks\SipTrunksController:create')->addMiddleware($checkAclTrunksTrunks);
                    $app->put('/sip/{id}/', '\Controllers\Trunks\SipTrunksController:update')->addMiddleware($checkAclTrunksTrunks);
                    $app->delete('/sip/{id}/', '\Controllers\Trunks\SipTrunksController:remove')->addMiddleware($checkAclTrunksTrunks);

                    $app->get('/pstn/', '\Controllers\Trunks\PstnTrunksController:getItems')->addMiddleware($checkAclTrunksTrunks);
                    $app->post('/pstn/', '\Controllers\Trunks\PstnTrunksController:create')->addMiddleware($checkAclTrunksTrunks);
                    $app->put('/pstn/{id}/', '\Controllers\Trunks\PstnTrunksController:update')->addMiddleware($checkAclTrunksTrunks);
                    $app->delete('/pstn/{id}/', '\Controllers\Trunks\PstnTrunksController:remove')->addMiddleware($checkAclTrunksTrunks);

                    $app->get('/fxo/', '\Controllers\Trunks\FxoTrunksController:getItems')->addMiddleware($checkAclTrunksTrunks);
                    $app->put('/fxo/{id}/', '\Controllers\Trunks\FxoTrunksController:update')->addMiddleware($checkAclTrunksTrunks);
                    $app->delete('/fxo/{id}/', '\Controllers\Trunks\FxoTrunksController:remove')->addMiddleware($checkAclTrunksTrunks);

                    $app->get('/groups/', '\Controllers\Trunks\GroupsController:getItems')->addMiddleware($checkAclTrunksHuntGroups);
                    $app->post('/groups/', '\Controllers\Trunks\GroupsController:create')->addMiddleware($checkAclTrunksHuntGroups);
                    $app->put('/groups/{id}/', '\Controllers\Trunks\GroupsController:update')->addMiddleware($checkAclTrunksHuntGroups);
                    $app->delete('/groups/{id}/', '\Controllers\Trunks\GroupsController:remove')->addMiddleware($checkAclTrunksHuntGroups);

                    $app->get('/prices/', '\Controllers\Trunks\PriceListController:getItems')->addMiddleware($checkAclTrunksPricelists);
                    $app->post('/prices/', '\Controllers\Trunks\PriceListController:create')->addMiddleware($checkAclTrunksPricelists);
                    $app->get('/prices/{id}/export/', '\Controllers\Trunks\PriceListController:export')->addMiddleware($checkAclTrunksPricelists);
                    $app->post('/prices/{id}/', '\Controllers\Trunks\PriceListController:update')->addMiddleware($checkAclTrunksPricelists);
                    $app->delete('/prices/{id}/', '\Controllers\Trunks\PriceListController:remove')->addMiddleware($checkAclTrunksPricelists);
                }
            )->addMiddleware($authorization);
            // endregion
            $app->get('/wizywebinar/', '\Controllers\Integrations\WizyWebinarController:getConfig')->addMiddleware($authorization);
            $app->get('/wizywebinar/token/', '\Controllers\Integrations\WizyWebinarController:getAccessToken')->addMiddleware($authorization);
            $app->get('/network/pbxes/', '\Controllers\Network\NetworkPbxesController:getItems')->addMiddleware($authorization);
            $app->put('/network/pbx/{serial}/', '\Controllers\Network\NetworkPbxesController:update')->addMiddleware($authorization)->addMiddleware($checkAclUsersPbxs)->addMiddleware($allowForAdmin);
            $app->delete('/network/pbx/{serial}/', '\Controllers\Network\NetworkPbxesController:delete')
                ->addMiddleware($authorization)->addMiddleware($checkAclUsersPbxs)->addMiddleware($allowForAdmin);
        }
    )->addMiddleware($setUserToRequest);
    // endregion
    // endregion
    /*
     * Catch-all route to serve a 404 Not Found page if none of the routes match
     * NOTE: make sure this route is defined last
     */
    $app->map(
        ['GET', 'POST', 'PUT', 'DELETE', 'PATCH'],
        '/{routes:.+}',
        function ($request, $response) {
            throw (new HttpNotFoundException($request, 'Page Not Found'));
        }
    );
};
