<?php

namespace Api\Analytics\Caracal\Services;

use Api\Analytics\Caracal\Interfaces\CaracalConfigInterface;
use Api\Analytics\Caracal\Interfaces\CaracalServiceInterface;
use Api\Translation\Interfaces\TranslationStrategyInterface;
use Logger\Log;
use PDO;
use PDOException;
use Services\ShellHelper;
use Wildix\Error;
use Wildix\Exception;
use Wildix\InvalidDataException;
use Wildix\License;
use Wildix\Validator;

/**
 * Class XcaracalService
 */
class CaracalService implements CaracalServiceInterface {

    const CONFLICT_PACKAGE_NAME = 'asternic';

    const SERVICE_NAME = 'xcaracal';
    const PROCESS_NAME = 'x-caracal';
    const DEFAULT_DB   = [
        'host' => 'localhost',
        'name' => 'qstats',
        'user' => 'root',
        'password' => 'wil01dix',
    ];

    /**
     * Config class.
     *
     * @var CaracalConfigInterface $config
     */
    private $_config;
    private $_allowed;
    private $_isCloud;
    private $_logger;
    private $_db;
    private $_translation;

    public function __construct(CaracalConfigInterface $config, Log $logger, TranslationStrategyInterface $translation, License $license) {
        $this->_config      = $config;
        $this->_logger      = $logger;
        $this->_translation = $translation;
        $this->_allowed     = $license->haveLicense(License::TYPE_XCARACAL);
        $this->_isCloud     = $license->isCloud();
    }

    public function info(): array {
        return [
            'name'             => self::SERVICE_NAME,
            'enable'           => $this->_isEnabled(),
            'isCanBeActivated' => $this->_checkLicense(),
            'database' => [
                'host' => $this->_isCloud() ? '' : (string)$this->_config()->get('DB_HOST'),
                'name' => $this->_isCloud() ? '' : (string)$this->_config()->get('DB_NAME'),
                'user' => $this->_isCloud() ? '' : (string)$this->_config()->get('DB_USER'),
                'password' => $this->_isCloud() ? '' : trim($this->_config()->get('DB_PASS'), '\''),
            ]
        ];
    }

    public function update(array $data): bool {
        $data['database'] = $this->_isCloud() ? self::DEFAULT_DB : array_merge(self::DEFAULT_DB, $data['database']);

        $this->_validate($data);

        if ($data['enable'] && !$this->_isEnabled()) {
            $this->assertAsternicIsNotInstalled();
            $this->checkDbServerReachable($data)->checkDbExist($data);

            ShellHelper::installCaracal($data['database']['host'], $data['database']['name'], $data['database']['user'], $data['database']['password']);
            if (!$this->_isEnabled()) {
                throw new Exception(Error::XcaracalIsNotRunning);
            }
        } else if (!$data['enable'] && $this->_isEnabled()) {
            $serviceInfo = $this->_isCloud() ? $data : $this->info();
            try {
                $this->checkDbServerReachable($serviceInfo)->deleteDb($serviceInfo);
            } catch (\Exception $e) {
                $errorMessage = sprintf("Can not drop the x-caracal database '%s'. Error: %s", $serviceInfo['database']['name'], $e->getMessage());
                $this->getLogger()->warning(__LINE__, $errorMessage);
            }

            ShellHelper::deleteCaracal();
        }

        return true;
    }

    private function checkDbServerReachable(array $data): self {
        try {
            $pdo = new \PDO(sprintf('mysql:host=%s', $data['database']['host']), $data['database']['user'], $data['database']['password']);
            $pdo->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_EXCEPTION);
            $this->setDbConnection($pdo);
        } catch (PDOException $exception) {
            $this->getLogger()->error(__LINE__, sprintf('Can not connect to database. Error: %s', $exception->getMessage()));
            throw new Exception(Error::InvalidCredentialsDB);
        }

        return $this;
    }

    private function checkDbExist(array $data): self {
        $databases = $this->getDbConnection()->query('select schema_name from information_schema.schemata')->fetchAll(\PDO::FETCH_COLUMN);
        if (in_array($data['database']['name'], $databases, true)) {
            throw new Exception(Error::IsXcaracalDataBaseExist);
        }

        return $this;
    }

    private function assertAsternicIsNotInstalled(): void {
        if (ShellHelper::isInstalledService(sprintf('^ii\s+%s', static::CONFLICT_PACKAGE_NAME), true)) {
            throw new Exception(Error::IsXcaracalDataBaseExist);
        }
    }

    private function deleteDb(array $data): self {
        $this->getDbConnection()->query(sprintf('DROP DATABASE IF EXISTS %s;', $data['database']['name']));

        $this->getLogger()->info(__LINE__, 'The x-caracal database deleted');

        return $this;
    }

    private function _validate(array $data): CaracalService {
        $validator = new Validator($data);

        $validator->rule('boolean', 'enable');
        $validator->rule('required', 'enable');

        if (!$this->_isCloud()) {
            $validator->rule('DomainOrIp', 'database.host');
            $validator->rule('required', 'database.password');
            $validator->rule('required', 'database.user');
            $validator->rule('required', 'database.host');
            $validator->rule('required', 'database.name');
            $validator->rule('regex', ['database.password', 'database.user', 'database.name'], '/^[^$#]*$/')
                ->reason(sprintf($this->getTranslation()->t('strings.notAllowedCharsPresent'), 'dollar($), hashtag(#)'));
        }

        if (!$validator->validate()) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $validator->errorsTree());
        }

        return $this;
    }

    private function _isEnabled(): bool {
        return ShellHelper::isActiveService(self::PROCESS_NAME);
    }

    private function _config(): CaracalConfigInterface {
        return $this->_config;
    }

    private function _isCloud(): bool {
        return $this->_isCloud;
    }

    private function _checkLicense(): bool {
        return $this->_allowed;
    }

    private function getDbConnection(): \PDO {
        return $this->_db;
    }

    public function setDbConnection(\PDO $db): self {
        $this->_db = $db;

        return $this;
    }

    private function getLogger(): Log {
        return $this->_logger;
    }

    private function getTranslation(): TranslationStrategyInterface {
        return $this->_translation;
    }
}
