<?php

declare(strict_types=1);

namespace Api\Database;

use Api\Database\Contracts\ManagesTransactionsTrait;
use Api\Database\Query\QueryBuilder;
use Closure;
use Exception;
use Latitude\QueryBuilder\Engine\CommonEngine;
use LogicException;
use PDO;
use PDOStatement;

/**
 * Class Connection
 */
class Connection implements ConnectionInterface {
    use DetectsDeadlocksTrait,
        DetectsLostConnectionsTrait,
        ManagesTransactionsTrait;

    /**
     * The connection resolvers.
     *
     * @var array
     */
    protected static $_resolvers = [];
    /**
     * The database connection configuration options.
     *
     * @var array
     */
    protected $_config = [];
    /**
     * The name of the connected database.
     *
     * @var string
     */
    protected $_database;
    /**
     * The default fetch mode of the connection.
     *
     * @var int
     */
    protected $_fetchMode = PDO::FETCH_OBJ;
    /**
     * Indicates whether queries are being logged.
     *
     * @var bool
     */
    protected $_loggingQueries = false;
    /**
     * The active PDO connection.
     *
     * @var \PDO|\Closure
     */
    protected $_pdo;

    /**
     * Indicates if the connection is in a "dry run".
     *
     * @var bool
     */
    protected $_pretending = false;

    /**
     * All of the queries run against the connection.
     *
     * @var array
     */
    protected $_queryLog = [];
    /**
     * The active PDO connection used for reads.
     *
     * @var \PDO|\Closure
     */
    protected $_readPdo;
    /**
     * The reconnector instance for the connection.
     *
     * @var callable
     */
    protected $_reconnector;

    /**
     * The table prefix for the connection.
     *
     * @var string
     */
    protected $_tablePrefix = '';
    /**
     * The number of active transactions.
     *
     * @var int
     */
    protected $_transactions = 0;

    protected $_queryBuilder;
    /**
     * Create a new database connection instance.
     *
     * @param \PDO|\Closure $pdo         Pdo connection.
     * @param string        $database    Database name.
     * @param string        $tablePrefix Table prefix.
     * @param array         $config      Config configuration.
     *
     * @return void
     */
    public function __construct($pdo, string $database = '', string $tablePrefix = '', array $config = []) {
        $this->_pdo = $pdo;

        $this->_database = $database;

        $this->_tablePrefix = $tablePrefix;

        $this->_config = $config;
    }

    /**
     * Get the connection resolver for the given driver.
     *
     * @param string $driver Driver name.
     *
     * @return mixed
     */
    public static function getResolver(string $driver) {
        return static::$_resolvers[$driver] ?? null;
    }

    /**
     * Register a connection resolver.
     *
     * @param string   $driver   Driver name.
     * @param \Closure $callback Callback.
     *
     * @return void
     */
    public static function resolverFor(string $driver, Closure $callback) {
        static::$_resolvers[$driver] = $callback;
    }

    /**
     * Run a select statement against the database and returns a generator.
     *
     * @param string  $query      Query string.
     * @param array   $bindings   Query params.
     * @param boolean $useReadPdo Use read pdo.
     *
     * @return \Generator
     */
    public function cursor(string $query, array $bindings = [], bool $useReadPdo = true): \Generator {
        $statement = $this->run(
            $query,
            $bindings,
            function ($query, $bindings) use ($useReadPdo) {
                if ($this->pretending()) {
                    return [];
                }

                $statement = $this->prepared(
                    $this->getPdoForSelect($useReadPdo)
                        ->prepare($query)
                );

                $this->bindValues(
                    $statement,
                    $this->prepareBindings($bindings)
                );

                $statement->execute();

                return $statement;
            }
        );

        while ($record = $statement->fetch()) {
            yield $record;
        }
    }

    /**
     * Run a SQL statement and log its execution context.
     *
     * @param string   $query    Query string.
     * @param array    $bindings Query params.
     * @param \Closure $callback Callback.
     *
     * @return mixed
     *
     * @throws QueryException|Exception Error.
     */
    protected function run(string $query, array $bindings, Closure $callback) {
        $this->reconnectIfMissingConnection();

        $start = microtime();

        try {
            $result = $this->runQueryCallback($query, $bindings, $callback);
        } catch (QueryException $e) {
            $result = $this->handleQueryException(
                $e,
                $query,
                $bindings,
                $callback
            );
        }

        $this->logQuery(
            $query,
            $bindings,
            $this->getElapsedTime(intval($start))
        );

        return $result;
    }

    /**
     * Reconnect to the database if a PDO connection is missing.
     *
     * @return void
     */
    protected function reconnectIfMissingConnection() {
        if (is_null($this->_pdo)) {
            $this->reconnect();
        }
    }

    /**
     * Reconnect to the database.
     *
     * @return mixed
     *
     * @throws LogicException Error.
     */
    public function reconnect() {
        if (is_callable($this->_reconnector)) {
            return call_user_func($this->_reconnector, $this);
        }

        throw new LogicException('Lost connection and no re-connector available.');
    }

    /**
     * Run a SQL statement.
     *
     * @param string   $query    Query string.
     * @param array    $bindings Query params.
     * @param \Closure $callback Callback.
     *
     * @return mixed
     *
     * @throws QueryException Error
     */
    protected function runQueryCallback(string $query, array $bindings, Closure $callback) {
        try {
            $result = $callback($query, $bindings);
        } catch (Exception $e) {
            throw new QueryException(
                $query,
                $this->prepareBindings($bindings),
                $e
            );
        }

        return $result;
    }

    /**
     * Prepare the query bindings for execution.
     *
     * @param array $bindings Query bindings.
     *
     * @return array
     */
    public function prepareBindings(array $bindings): array {
        foreach ($bindings as $key => $value) {
            if ($value === false) {
                $bindings[$key] = 0;
            }
        }

        return $bindings;
    }

    /**
     * Handle a query exception.
     *
     * @param \Exception $e        Error.
     * @param string     $query    Query string.
     * @param array      $bindings Query params.
     * @param \Closure   $callback Callback.
     *
     * @return mixed
     * @throws \Exception Error.
     */
    protected function handleQueryException(Exception $e, string $query, array $bindings, Closure $callback) {
        if ($this->_transactions >= 1) {
            throw $e;
        }

        return $this->tryAgainIfCausedByLostConnection(
            $e,
            $query,
            $bindings,
            $callback
        );
    }

    /**
     * Handle a query exception that occurred during query execution.
     *
     * @param QueryException $e        Error.
     * @param string         $query    Query string.
     * @param array          $bindings Query params.
     * @param \Closure       $callback Callback.
     *
     * @return mixed
     *
     * @throws QueryException Error.
     */
    protected function tryAgainIfCausedByLostConnection(
        QueryException $e,
        string $query,
        array $bindings,
        Closure $callback
    ) {
        if ($this->causedByLostConnection($e->getPrevious())) {
            $this->reconnect();

            return $this->runQueryCallback($query, $bindings, $callback);
        }

        throw $e;
    }

    /**
     * Log a query in the connection's query log.
     *
     * @param string     $query    Query string.
     * @param array      $bindings Query params.
     * @param float|null $time     Time log.
     *
     * @return void
     */
    public function logQuery(string $query, array $bindings, float $time = null) {
        if ($this->_loggingQueries) {
            $this->_queryLog[] = compact('query', 'bindings', 'time');
        }
    }

    /**
     * Get the elapsed time since a given starting point.
     *
     * @param integer $start Elapsed Time.
     *
     * @return float
     */
    protected function getElapsedTime(int $start): float {
        return round((microtime(true) - $start) * 1000, 2);
    }

    /**
     * Determine if the connection in a "dry run".
     *
     * @return boolean
     */
    public function pretending(): bool {
        return $this->_pretending === true;
    }

    /**
     * Configure the PDO prepared statement.
     *
     * @param \PDOStatement $statement Pdo statement.
     *
     * @return \PDOStatement
     */
    protected function prepared(PDOStatement $statement): PDOStatement {
        $statement->setFetchMode($this->_fetchMode);

        return $statement;
    }

    /**
     * Get the PDO connection to use for a select query.
     *
     * @param boolean $useReadPdo Flag for use reading connection.
     *
     * @return \PDO
     */
    protected function getPdoForSelect(bool $useReadPdo = true) {
        return $useReadPdo ? $this->getReadPdo() : $this->getPdo();
    }

    /**
     * Get the current PDO connection used for reading.
     *
     * @return \PDO
     */
    public function getReadPdo() {
        if ($this->_transactions >= 1) {
            return $this->getPdo();
        }

        if ($this->_readPdo instanceof Closure) {
            $this->_readPdo = call_user_func($this->_readPdo);
            return $this->_readPdo;
        }

        return $this->_readPdo ?: $this->getPdo();
    }

    /**
     * Set the PDO connection used for reading.
     *
     * @param \PDO|\Closure|null $pdo Pdo connection.
     *
     * @return $this
     */
    public function setReadPdo($pdo): Connection {
        $this->_readPdo = $pdo;

        return $this;
    }

    /**
     * Get the current PDO connection.
     *
     * @return \PDO
     */
    public function getPdo() {
        if ($this->_pdo instanceof Closure) {
            $this->_pdo = call_user_func($this->_pdo);
            return $this->_pdo;
        }

        return $this->_pdo;
    }

    /**
     * Set the PDO connection.
     *
     * @param \PDO|\Closure|null $_pdo PDO connection.
     *
     * @return $this
     */
    public function setPdo($_pdo): Connection {
        $this->_transactions = 0;

        $this->_pdo = $_pdo;

        return $this;
    }

    /**
     * Bind values to their parameters in the given statement.
     *
     * @param \PDOStatement $statement Given statement.
     * @param array         $bindings  Bind values.
     *
     * @return void
     */
    public function bindValues(PDOStatement $statement, array $bindings) {
        foreach ($bindings as $key => $value) {
            $statement->bindValue(
                is_string($key) ? $key : $key + 1,
                $value,
                is_int($value) ? PDO::PARAM_INT : PDO::PARAM_STR
            );
        }
    }

    /**
     * Run a delete statement against the database.
     *
     * @param string $query    Query strings.
     * @param array  $bindings Query params.
     *
     * @return integer
     */
    public function delete(string $query, array $bindings = []): int {
        return $this->affectingStatement($query, $bindings);
    }

    /**
     * Run an SQL statement and get the number of rows affected.
     *
     * @param string $query    Query strings.
     * @param array  $bindings Query params.
     *
     * @return integer
     * @throws Exception Error.
     */
    public function affectingStatement(string $query, array $bindings = []): int {
        return $this->run(
            $query,
            $bindings,
            function ($query, $bindings) {
                if ($this->pretending()) {
                    return 0;
                }

                $statement = $this->getPdo()->prepare($query);

                $this->bindValues($statement, $this->prepareBindings($bindings));

                $statement->execute();

                return $statement->rowCount();
            }
        );
    }

    /**
     * Disable the query log on the connection.
     *
     * @return void
     */
    public function disableQueryLog() {
        $this->_loggingQueries = false;
    }

    /**
     * Disconnect from the underlying PDO connection.
     *
     * @return void
     */
    public function disconnect() {
        $this->setPdo(null)->setReadPdo(null);
    }

    /**
     * Clear the query log.
     *
     * @return void
     */
    public function flushQueryLog() {
        $this->_queryLog = [];
    }

    /**
     * Get the name of the connected database.
     *
     * @return string
     */
    public function getDatabaseName(): string {
        return $this->_database;
    }

    /**
     * Get an option from the configuration options.
     *
     * @param string|null $option Oprions.
     *
     * @return mixed
     */
    public function getConfig(string $option = null) {
        return $this->_config[$option] ?? null;
    }

    /**
     * Get the PDO driver name.
     *
     * @return string
     */
    public function getDriverName(): ?string {
        return $this->getConfig('driver');
    }

    /**
     * Get the database connection name.
     *
     * @return string|null
     */
    public function getName(): ?string {
        return $this->getConfig('name');
    }

    /**
     * Get the connection query log.
     *
     * @return array
     */
    public function getQueryLog(): array {
        return $this->_queryLog;
    }

    /**
     * Get the table prefix for the connection.
     *
     * @return string
     */
    public function getTablePrefix(): string {
        return $this->_tablePrefix;
    }

    /**
     * Set the table prefix in use by the connection.
     *
     * @param string $prefix Table prefix.
     *
     * @return void
     */
    public function setTablePrefix(string $prefix) {
        $this->_tablePrefix = $prefix;
    }

    /**
     * Run an insert statement against the database.
     *
     * @param string $query    Query strings.
     * @param array  $bindings Query params.
     *
     * @return boolean
     */
    public function insert(string $query, array $bindings = []): bool {
        return $this->statement($query, $bindings);
    }

    /**
     * Execute an SQL statement and return the boolean result.
     *
     * @param string $query    Query strings.
     * @param array  $bindings Query params.
     *
     * @return boolean
     */
    public function statement(string $query, array $bindings = []): bool {
        return $this->run(
            $query,
            $bindings,
            function ($query, $bindings) {
                if ($this->pretending()) {
                    return true;
                }

                $statement = $this->getPdo()->prepare($query);

                $this->bindValues($statement, $this->prepareBindings($bindings));

                return $statement->execute();
            }
        );
    }

    /**
     * Is Doctrine available?
     *
     * @return boolean
     */
    public function isDoctrineAvailable(): bool {
        return class_exists('Doctrine\DBAL\Connection');
    }

    /**
     * Determine whether we're logging queries.
     *
     * @return boolean
     */
    public function logging(): bool {
        return $this->_loggingQueries;
    }

    /**
     * Execute the given callback in "dry run" mode.
     *
     * @param \Closure $callback Callback.
     *
     * @return array
     */
    public function pretend(Closure $callback): array {
        return $this->withFreshQueryLog(
            function () use ($callback) {
                $this->_pretending = true;

                $callback($this);

                $this->_pretending = false;

                return $this->_queryLog;
            }
        );
    }

    /**
     * Execute the given callback in "dry run" mode.
     *
     * @param \Closure $callback Callback.
     *
     * @return array
     */
    protected function withFreshQueryLog(Closure $callback): array {
        $loggingQueries = $this->_loggingQueries;

        $this->enableQueryLog();

        $this->_queryLog = [];

        $result = $callback();

        $this->_loggingQueries = $loggingQueries;

        return $result;
    }

    /**
     * Enable the query log on the connection.
     *
     * @return void
     */
    public function enableQueryLog() {
        $this->_loggingQueries = true;
    }

    /**
     * Get a new raw query expression.
     *
     * @param mixed $value Query statement.
     *
     * @return string
     */
    public function raw($value): string {
        return (string)$value;
    }

    /**
     * Run a select statement against the database.
     *
     * @param string $query    Query strings.
     * @param array  $bindings Query params.
     *
     * @return array
     */
    public function selectFromWriteConnection(string $query, array $bindings = []): array {
        return $this->select($query, $bindings, false);
    }

    /**
     * Run a select statement against the database.
     *
     * @param string  $query      Query strings.
     * @param array   $bindings   Query params.
     * @param boolean $useReadPdo Query type.
     *
     * @return array
     */
    public function select(string $query, array $bindings = [], bool $useReadPdo = false): array {
        return $this->run(
            $query,
            $bindings,
            function ($query, $bindings) use ($useReadPdo) {
                if ($this->pretending()) {
                    return [];
                }

                $statement = $this->prepared(
                    $this->getPdoForSelect($useReadPdo)
                        ->prepare($query)
                );

                $this->bindValues($statement, $this->prepareBindings($bindings));

                $statement->execute();

                return $statement->fetchAll();
            }
        );
    }

    /**
     * Run a select statement and return a single result.
     *
     * @param string $query    Query string.
     * @param array  $bindings Query bindings.
     *
     * @return mixed
     */
    public function selectOne(string $query, array $bindings = []) {
        $useReadPdo = false;
        $records    = $this->select($query, $bindings, $useReadPdo);

        return array_shift($records);
    }

    /**
     * Set the name of the connected database.
     *
     * @param string $database Database name.
     *
     * @return Connection
     */
    public function setDatabaseName(string $database): Connection {
        $this->_database = $database;
        return $this;
    }

    /**
     * Set the reconnect instance on the connection.
     *
     * @param callable $_reconnector Connector.
     *
     * @return Connection
     */
    public function setReconnector(callable $_reconnector): Connection {
        $this->_reconnector = $_reconnector;

        return $this;
    }

    /**
     * Begin a fluent query against a database table.
     *
     * @param string $table Table name.
     *
     * @return QueryBuilder
     */
    public function table(string $table): QueryBuilder {
        return $this->query()->select('*')->from($table);
    }

    /**
     * Get a new query builder instance.
     *
     * @return QueryBuilder
     */
    public function query(): QueryBuilder {
        if (!isset($this->_queryBuilder)) {
            $engine              = new CommonEngine();
            $this->_queryBuilder = new QueryBuilder($this, $engine);
        }
        return $this->_queryBuilder;
    }

    /**
     * Call the given Closure with the given value then return the value.
     *
     * @param mixed         $value    Given value.
     * @param callable|null $callback Call the given Closure.
     *
     * @return mixed
     */
    private function tap($value, callable $callback = null) {
        if (is_null($callback)) {
            return new \Wildix\Exception('Not running callable');
        }

        $callback($value);

        return $value;
    }

    /**
     * Run a raw, unprepared query against the PDO connection.
     *
     * @param string $query Query string.
     *
     * @return boolean
     */
    public function unprepared(string $query): bool {
        return $this->run(
            $query,
            [],
            function ($query) {
                if ($this->pretending()) {
                    return true;
                }

                return !($this->getPdo()->exec($query) === false);
            }
        );
    }

    /**
     * Run an update statement against the database.
     *
     * @param string $query    Query strings.
     * @param array  $bindings Query params.
     *
     * @return integer
     */
    public function update(string $query, array $bindings = []): int {
        return $this->affectingStatement($query, $bindings);
    }
}
