<?php

declare(strict_types=1);

namespace Api\Database;

use Api\Database\Query\QueryBuilder;
use Closure;

/**
 * Interface ConnectionInterface
 */
interface ConnectionInterface {

    /**
     * Begin a fluent query against a database table.
     *
     * @param string $table Table name.
     *
     * @return QueryBuilder
     */
    public function table(string $table): QueryBuilder;

    /**
     * Get a new raw query expression.
     *
     * @param mixed $value Query string.
     *
     * @return string
     */
    public function raw($value): string;

    /**
     * Run a select statement and return a single result.
     *
     * @param string $query    Query string.
     * @param array  $bindings Query bindings.
     *
     * @return mixed
     */
    public function selectOne(string $query, array $bindings = []);

    /**
     * Run a select statement against the database.
     *
     * @param string $query    Query string.
     * @param array  $bindings Query bindings.
     *
     * @return array
     */
    public function select(string $query, array $bindings = []): array;

    /**
     * Run an insert statement against the database.
     *
     * @param string $query    Query string.
     * @param array  $bindings Query bindings.
     *
     * @return boolean
     */
    public function insert(string $query, array $bindings = []): bool;

    /**
     * Run an update statement against the database.
     *
     * @param string $query    Query string.
     * @param array  $bindings Query bindings.
     *
     * @return integer
     */
    public function update(string $query, array $bindings = []): int;

    /**
     * Run a delete statement against the database.
     *
     * @param string $query    Query string.
     * @param array  $bindings Query bindings.
     *
     * @return integer
     */
    public function delete(string $query, array $bindings = []): int;

    /**
     * Execute an SQL statement and return the boolean result.
     *
     * @param string $query    Query string.
     * @param array  $bindings Query bindings.
     *
     * @return boolean
     */
    public function statement(string $query, array $bindings = []): bool;

    /**
     * Run an SQL statement and get the number of rows affected.
     *
     * @param string $query    Query string.
     * @param array  $bindings Query bindings.
     *
     * @return integer
     */
    public function affectingStatement(string $query, array $bindings = []): int;

    /**
     * Run a raw, unprepared query against the PDO connection.
     *
     * @param string $query Query string.
     *
     * @return boolean
     */
    public function unprepared(string $query): bool;

    /**
     * Prepare the query bindings for execution.
     *
     * @param array $bindings Query bindings.
     *
     * @return array
     */
    public function prepareBindings(array $bindings): array;

    /**
     * Execute a Closure within a transaction.
     *
     * @param Closure $callback Closure.
     * @param integer $attempts Attempts.
     *
     * @return mixed
     *
     * @throws \Throwable Error.
     */
    public function transaction(Closure $callback, int $attempts = 1);

    /**
     * Start a new database transaction.
     *
     * @return void
     */
    public function beginTransaction();

    /**
     * Commit the active database transaction.
     *
     * @return void
     */
    public function commit();

    /**
     * Rollback the active database transaction.
     *
     * @return void
     */
    public function rollBack();

    /**
     * Get the number of active transactions.
     *
     * @return integer
     */
    public function transactionLevel(): int;

    /**
     * Execute the given callback in "dry run" mode.
     *
     * @param Closure $callback Callback.
     *
     * @return array
     */
    public function pretend(Closure $callback): array;

    /**
     * Get a new query builder instance.
     *
     * @return QueryBuilder
     */
    public function query(): QueryBuilder;
}
