<?php

declare(strict_types=1);

namespace Api\Database;

/**
 * Class ConnectionResolver
 */
class ConnectionResolver implements ConnectionResolverInterface {

    /**
     * All the registered connections.
     *
     * @var array
     */
    protected $_connections = [];

    /**
     * The default connection name.
     *
     * @var string
     */
    protected $_default;

    /**
     * Create a new connection resolver instance.
     *
     * @param array $connections Connection resolver.
     *
     * @return void
     */
    public function __construct(array $connections = []) {
        foreach ($connections as $name => $connection) {
            $this->addConnection($name, $connection);
        }
    }

    /**
     * Add a connection to the resolver.
     *
     * @param string              $name       Connection name.
     * @param ConnectionInterface $connection Connection to the resolver.
     *
     * @return void
     */
    public function addConnection(string $name, ConnectionInterface $connection) {
        $this->_connections[$name] = $connection;
    }

    /**
     * Get a database connection instance.
     *
     * @param string $name Database name.
     *
     * @return ConnectionInterface
     */
    public function connection($name = null): ConnectionInterface {
        if (is_null($name)) {
            $name = $this->getDefaultConnection();
        }

        return $this->_connections[$name];
    }

    /**
     * Get the default connection name.
     *
     * @return string
     */
    public function getDefaultConnection(): string {
        return $this->_default;
    }

    /**
     * Check if a connection has been registered.
     *
     * @param string $name Connection name.
     *
     * @return boolean
     */
    public function hasConnection(string $name): bool {
        return isset($this->_connections[$name]);
    }

    /**
     * Set the default connection name.
     *
     * @param string $name Default connection name.
     *
     * @return void
     */
    public function setDefaultConnection(string $name) {
        $this->_default = $name;
    }
}
