<?php

declare(strict_types=1);

namespace Api\Database\Connectors;

use Api\Database\Connection;
use Api\Database\MySqlConnection;
use Api\Database\SQLiteConnection;
use Api\Database\SqlServerConnection;
use InvalidArgumentException;
use Logger;
use PDOException;
use Wildix\DependencyInjection\Container;

/**
 * Class ConnectionFactory
 */
class ConnectionFactory {

    /**
     * The IoC container instance.
     *
     * @var Container
     */
    protected $_container;

    /**
     * Create a new connection factory instance.
     *
     * @param Container $container Application container.
     *
     * @return void
     */
    public function __construct(Container $container) {
        $this->_container = $container;
    }

    /**
     * Establish a PDO connection based on the configuration.
     *
     * @param array       $config Configuration array.
     * @param string|null $name   Config name.
     *
     * @return \Api\Database\Connection
     */
    public function make(array $config, string $name = null):Connection {
        $config = $this->prepareConfig($config, $name);

        if (isset($config['read'])) {
            return $this->createReadWriteConnection($config);
        }

        return $this->createSingleConnection($config);
    }

    /**
     * Prepare the database configuration.
     *
     * @param array  $config Database configuration.
     * @param string $name   Name configuration.
     *
     * @return array
     */
    protected function prepareConfig(array $config, string $name): array {
        $config['prefix'] = '';
        $config['name']   = $name;

        return $config;
    }

    /**
     * Create a single database connection instance.
     *
     * @param array $config Database configuration.
     *
     * @return Connection
     */
    protected function createSingleConnection(array $config):Connection {
        $pdo = $this->createPdoResolver($config);

        return $this->createConnection(
            $config['driver'],
            $pdo,
            $config['database'],
            $config['prefix'],
            $config
        );
    }

    /**
     * Create a single database connection instance.
     *
     * @param array $config Database configuration.
     *
     * @return Connection
     */
    protected function createReadWriteConnection(array $config): Connection {
        $connection = $this->createSingleConnection($this->getWriteConfig($config));

        return $connection->setReadPdo($this->createReadPdo($config));
    }

    /**
     * Create a new PDO instance for reading.
     *
     * @param array $config Database configuration.
     *
     * @return \Closure
     */
    protected function createReadPdo(array $config): \Closure {
        return $this->createPdoResolver($this->getReadConfig($config));
    }

    /**
     * Get the read configuration for a read / write connection.
     *
     * @param array $config Database configuration.
     *
     * @return array
     */
    protected function getReadConfig(array $config): array {
        return $this->mergeReadWriteConfig(
            $config,
            $this->getReadWriteConfig($config, 'read')
        );
    }

    /**
     * Get the read configuration for a read / write connection.
     *
     * @param array $config Database configuration.
     *
     * @return array
     */
    protected function getWriteConfig(array $config): array {
        return $this->mergeReadWriteConfig(
            $config,
            $this->getReadWriteConfig($config, 'write')
        );
    }

    /**
     * Get a read / write level configuration.
     *
     * @param array  $config Database configuration.
     * @param string $type   Type.
     *
     * @return array
     */
    protected function getReadWriteConfig(array $config, string $type): array {
        $randType = array_rand($config[$type], 2);

        return isset($config[$type][0]) ? $randType[0] : $config[$type];
    }

    /**
     * Merge a configuration for a read / write connection.
     *
     * @param array $config Database configuration.
     * @param array $merge  Database configuration for marge.
     *
     * @return array
     */
    protected function mergeReadWriteConfig(array $config, array $merge): array {
        unset($config['read'], $config['write'], $merge['read'], $merge['write']);

        return array_merge($config, $merge);
    }

    /**
     * Create a new Closure that resolves to a PDO instance.
     *
     * @param array $config Database configuration.
     *
     * @return \Closure
     */
    protected function createPdoResolver(array $config): \Closure {
        return array_key_exists('host', $config) ? $this->createPdoResolverWithHosts($config) : $this->createPdoResolverWithoutHosts($config);
    }

    /**
     * Create a new Closure that resolves to a PDO instance with a specific host or an array of hosts.
     *
     * @param array $config Database configuration.
     *
     * @return \Closure
     */
    protected function createPdoResolverWithHosts(array $config): \Closure {
        return function () use ($config) {
            $hosts = $this->parseHosts($config);
            shuffle($hosts);
            foreach ($hosts as $key => $host) {
                $config['host'] = $host;

                try {
                    return $this->createConnector($config)->connect($config);
                } catch (PDOException $e) {
                    if (count($hosts) - 1 === $key && $this->_container->has(Logger::class)) {
                        $this->_container->get(Logger::class)->error($e->getMessage());
                    }
                }
            }

            throw $e;
        };
    }

    /**
     * Parse the hosts configuration item into an array.
     *
     * @param array $config Database configuration.
     *
     * @throws InvalidArgumentException Error.
     * @return array
     */
    protected function parseHosts(array $config): array {
        $hosts = !is_array($config['host']) ? [$config['host']] : $config['host'];

        if (empty($hosts)) {
            throw new InvalidArgumentException('Database hosts array is empty.');
        }

        return $hosts;
    }

    /**
     * Create a new Closure that resolves to a PDO instance where there is no configured host.
     *
     * @param array $config Database configuration.
     *
     * @return \Closure
     */
    protected function createPdoResolverWithoutHosts(array $config): \Closure {
        return function () use ($config) {
            return $this->createConnector($config)->connect($config);
        };
    }

    /**
     * Create a connector instance based on the configuration.
     *
     * @param array $config Database configuration.
     *
     * @throws InvalidArgumentException Error.
     * @return ConnectorInterface
     */
    public function createConnector(array $config) {
        if (! isset($config['driver'])) {
            throw new InvalidArgumentException('A driver must be specified.');
        }
        if (! isset($config['database'])) {
            throw new InvalidArgumentException('A database must be specified.');
        }
        $key = sprintf('db.connector.%s.%s', $config['driver'], $config['database']);
        if ($this->_container->has($key)) {
            return $this->_container->get($key);
        }

        switch ($config['driver']) {
            case 'mysql':
                return new MySqlConnector;
            case 'sqlite':
                return new SQLiteConnector;
            case 'sqlsrv':
                return new SqlServerConnector;
            default:
                throw new InvalidArgumentException(sprintf('Unsupported driver [%s]', $config['driver']));
        }
    }

    /**
     * Create a new connection instance.
     *
     * @param string        $driver     Driver name.
     * @param \PDO|\Closure $connection Pdo connection.
     * @param string        $database   Path to database.
     * @param string        $prefix     Prefix for name table.
     * @param array         $config     Config configuration.
     *
     * @throws InvalidArgumentException Error.
     * @return Connection Connection class.
     */
    protected function createConnection(string $driver, $connection, string  $database, string $prefix = '', array $config = []) {
        $resolver = Connection::getResolver($driver);
        if ($resolver) {
            return $resolver($connection, $database, $prefix, $config);
        }

        switch ($driver) {
            case 'mysql':
                return new MySqlConnection($connection, $database, $prefix, $config);
            case 'sqlite':
                return new SQLiteConnection($connection, $database, $prefix, $config);
            case 'sqlsrv':
                return new SqlServerConnection($connection, $database, $prefix, $config);
            default:
            // return error
                break;
        }

        throw new InvalidArgumentException(sprintf('Unsupported driver [%s]', $driver));
    }
}
