<?php

declare(strict_types=1);

namespace Api\Database\Connectors;

use Api\Database\DetectsLostConnectionsTrait;
use Exception;
use PDO;


/**
 * Class Connector
 */
class Connector {

    use DetectsLostConnectionsTrait;

    /**
     * The default PDO connection options.
     *
     * @var array
     */
    protected $_options = [
        PDO::ATTR_CASE => PDO::CASE_NATURAL,
        PDO::ATTR_ERRMODE => PDO::ERRMODE_EXCEPTION,
        PDO::ATTR_ORACLE_NULLS => PDO::NULL_NATURAL,
        PDO::ATTR_STRINGIFY_FETCHES => false,
        PDO::ATTR_EMULATE_PREPARES => false,
    ];

    /**
     * Create a new PDO connection.
     *
     * @param string $dsn     DNS.
     * @param array  $config  Config params.
     * @param array  $options Config options.
     *
     * @return \PDO
     *
     * @throws Exception Exception.
     */
    public function createConnection(string $dsn, array $config, array $options): PDO {
        [$username, $password] = [
            $config['username'] ?? '', $config['password'] ?? '',
        ];

        try {
            return $this->createPdoConnection(
                $dsn,
                $username,
                $password,
                $options
            );
        } catch (Exception $e) {
            return $this->tryAgainIfCausedByLostConnection(
                $e,
                $dsn,
                $username,
                $password,
                $options
            );
        }
    }

    /**
     * Create a new PDO connection instance.
     *
     * @param string $dsn      DNS.
     * @param string $username Username for connection.
     * @param string $password Password  for connection.
     * @param array  $options  Options for connection.
     *
     * @return \PDO
     */
    protected function createPdoConnection(string $dsn, string $username, string $password, array $options):PDO {
        return new PDO($dsn, $username, $password, $options);
    }

    /**
     * Determine if the connection is persistent.
     *
     * @param array $options Connection options.
     *
     * @return boolean
     */
    protected function isPersistentConnection(array $options): bool {
        return isset($options[PDO::ATTR_PERSISTENT]) &&
               $options[PDO::ATTR_PERSISTENT];
    }

    /**
     * Handle an exception that occurred during connect execution.
     *
     * @param \Exception $e        Exception.
     * @param string     $dsn      DNS.
     * @param string     $username Username for connection.
     * @param string     $password Password  for connection.
     * @param array      $options  Options for connection.
     *
     * @return \PDO
     *
     * @throws \Exception Exception if errors.
     */
    protected function tryAgainIfCausedByLostConnection(Exception $e, string $dsn, string $username, string $password, array $options): PDO {
        if ($this->causedByLostConnection($e)) {
            return $this->createPdoConnection($dsn, $username, $password, $options);
        }

        throw $e;
    }

    /**
     * Get the PDO options based on the configuration.
     *
     * @param array $config Array options.
     *
     * @return array
     */
    public function getOptions(array $config): array {
        $options = $config['options'] ?? [];

        return array_diff_key($this->_options, $options) + $options;
    }

    /**
     * Get the default PDO connection options.
     *
     * @return array
     */
    public function getDefaultOptions(): array {
        return $this->_options;
    }

    /**
     * Set the default PDO connection options.
     *
     * @param array $options Options for connection.
     *
     * @return void
     */
    public function setDefaultOptions(array $options) {
        $this->_options = $options;
    }
}
