<?php

declare(strict_types=1);

namespace Api\Database\Connectors;

use Exception;
use PDO;

/**
 * Class MySqlConnector
 */
class MySqlConnector extends Connector implements ConnectorInterface {

    /**
     * Establish a database connection.
     *
     * @param array $config Database connection.
     *
     * @return PDO
     * @throws Exception Error.
     */
    public function connect(array $config): PDO {
        $dsn = $this->getDsn($config);

        $options = $this->getOptions($config);

        $connection = $this->createConnection($dsn, $config, $options);

        if (! empty($config['database'])) {
            $connection->exec(sprintf('use `%s`;', $config['database']));
        }

        $this->configureEncoding($connection, $config);

        $this->configureTimezone($connection, $config);

        $this->setModes($connection, $config);

        return $connection;
    }

    /**
     * Set the connection character set and collation.
     *
     * @param \PDO  $connection Pdo connection.
     * @param array $config     Config params.
     *
     * @return PDO|null
     */
    protected function configureEncoding(PDO $connection, array $config): ?PDO {
        if (! isset($config['charset'])) {
            return $connection;
        }

        $connection->prepare(
            sprintf("set names '%s'", $config['charset']) . $this->getCollation($config)
        )->execute();
    }

    /**
     * Get the collation for the configuration.
     *
     * @param array $config Config params.
     *
     * @return string
     */
    protected function getCollation(array $config): string {
        return isset($config['collation']) ? sprintf(" collate '%s'", $config['collation']) : '';
    }

    /**
     * Set the timezone on the connection.
     *
     * @param \PDO  $connection Connection.
     * @param array $config     Config params.
     *
     * @return void
     */
    protected function configureTimezone(PDO $connection, array $config) {
        if (isset($config['timezone'])) {
            $connection->prepare('set time_zone="' . $config['timezone'] . '"')->execute();
        }
    }

    /**
     * Create a DSN string from a configuration.
     *
     * Chooses socket or host/port based on the 'unix_socket' config value.
     *
     * @param array $config Config params.
     *
     * @return string
     */
    protected function getDsn(array $config): string {
        return $this->hasSocket($config) ? $this->getSocketDsn($config) : $this->getHostDsn($config);
    }

    /**
     * Determine if the given configuration array has a UNIX socket value.
     *
     * @param array $config Config params.
     *
     * @return boolean
     */
    protected function hasSocket(array $config): bool {
        return isset($config['unix_socket']) && ! empty($config['unix_socket']);
    }

    /**
     * Get the DSN string for a socket configuration.
     *
     * @param array $config Config params.
     *
     * @return string
     */
    protected function getSocketDsn(array $config): string {
        return sprintf('mysql:unix_socket=%s;dbname=%s', $config['unix_socket'], $config['database']);
    }

    /**
     * Get the DSN string for a host / port configuration.
     *
     * @param array $config Config params.
     *
     * @return string
     */
    protected function getHostDsn(array $config): string {
        extract($config, EXTR_SKIP);

        return isset($port) ? sprintf('mysql:host=%s;port=%s;dbname=%s', $host, $port, $database) : sprintf('mysql:host=%s;dbname=%s', $host, $database);
    }

    /**
     * Set the modes for the connection.
     *
     * @param \PDO  $connection Connection.
     * @param array $config     Config params.
     *
     * @return void
     */
    protected function setModes(PDO $connection, array $config) {
        if (isset($config['modes'])) {
            $this->setCustomModes($connection, $config);
        } else if (isset($config['strict'])) {
            if ($config['strict']) {
                $connection->prepare($this->strictMode())->execute();
            } else {
                $connection->prepare("set session sql_mode='NO_ENGINE_SUBSTITUTION'")->execute();
            }
        }
    }

    /**
     * Set the custom modes on the connection.
     *
     * @param \PDO  $connection Connection.
     * @param array $config     Config params.
     *
     * @return void
     */
    protected function setCustomModes(PDO $connection, array $config) {
        $modes = implode(',', $config['modes']);

        $connection->prepare(sprintf("set session sql_mode='%s'", $modes))->execute();
    }

    /**
     * Get the query to enable strict mode.
     *
     * @return string
     */
    protected function strictMode(): string {
        return "set session sql_mode='ONLY_FULL_GROUP_BY,STRICT_TRANS_TABLES,NO_ZERO_IN_DATE,NO_ZERO_DATE,ERROR_FOR_DIVISION_BY_ZERO,NO_AUTO_CREATE_USER,NO_ENGINE_SUBSTITUTION'";
    }
}
