<?php

declare(strict_types=1);

namespace Api\Database\Connectors;

use PDO;

/**
 * Class SqlServerConnector
 */
class SqlServerConnector extends Connector implements ConnectorInterface {

    /**
     * Establish a database connection.
     *
     * @param array $config Config database.
     *
     * @throws \Exception Error.
     *
     * @return \PDO
     */
    public function connect(array $config): PDO {
        $options = $this->getOptions($config);

        return $this->createConnection($this->getDsn($config), $config, $options);
    }

    /**
     * The PDO connection options.
     *
     * @var array
     */
    protected $_options = [
        PDO::ATTR_CASE => PDO::CASE_NATURAL,
        PDO::ATTR_ERRMODE => PDO::ERRMODE_EXCEPTION,
        PDO::ATTR_ORACLE_NULLS => PDO::NULL_NATURAL,
        PDO::ATTR_STRINGIFY_FETCHES => false,
    ];

    /**
     * Create a DSN string from a configuration.
     *
     * @param array $config Config database.
     *
     * @return string
     */
    protected function getDsn(array $config): string {
        if (in_array('dblib', $this->getAvailableDrivers())) {
            return $this->getDblibDsn($config);
        } else if ($this->prefersOdbc($config)) {
            return $this->getOdbcDsn($config);
        } else {
            return $this->getSqlSrvDsn($config);
        }
    }

    /**
     * Determine if the database configuration prefers ODBC.
     *
     * @param array $config Config database.
     *
     * @return boolean
     */
    protected function prefersOdbc(array $config): bool {
        return in_array('odbc', $this->getAvailableDrivers()) &&
               $config['odbc'] === true;
    }

    /**
     * Get the DSN string for a DbLib connection.
     *
     * @param array $config Config database.
     *
     * @return string
     */
    protected function getDblibDsn(array $config): string {
        return $this->buildConnectString(
            'dblib',
            array_merge(
                [
                'host' => $this->buildHostString($config, ':'),
                'dbname' => $config['database'],
                ],
                [$config['appname'], $config['charset']]
            )
        );
    }

    /**
     * Get the DSN string for an ODBC connection.
     *
     * @param array $config Config database.
     *
     * @return string
     */
    protected function getOdbcDsn(array $config): string {
        return isset($config['odbc_datasource_name']) ? 'odbc:' . $config['odbc_datasource_name'] : '';
    }

    /**
     * Get the DSN string for a SqlSrv connection.
     *
     * @param array $config Config database.
     *
     * @return string
     */
    protected function getSqlSrvDsn(array $config): string {
        $arguments = [
            'Server' => $this->buildHostString($config, ','),
        ];

        if (isset($config['database'])) {
            $arguments['Database'] = $config['database'];
        }

        if (isset($config['readonly'])) {
            $arguments['ApplicationIntent'] = 'ReadOnly';
        }

        if (isset($config['pooling']) && $config['pooling'] === false) {
            $arguments['ConnectionPooling'] = '0';
        }

        if (isset($config['appname'])) {
            $arguments['APP'] = $config['appname'];
        }

        if (isset($config['encrypt'])) {
            $arguments['Encrypt'] = $config['encrypt'];
        }

        if (isset($config['trust_server_certificate'])) {
            $arguments['TrustServerCertificate'] = $config['trust_server_certificate'];
        }

        if (isset($config['multiple_active_result_sets']) && $config['multiple_active_result_sets'] === false) {
            $arguments['MultipleActiveResultSets'] = 'false';
        }

        return $this->buildConnectString('sqlsrv', $arguments);
    }

    /**
     * Build a connection string from the given arguments.
     *
     * @param string $driver    Driver name.
     * @param array  $arguments Config database.
     *
     * @return string
     */
    protected function buildConnectString(string $driver, array $arguments): string {
        return $driver . ':' . implode(
            ';',
            array_map(
                function ($key) use ($arguments) {
                    return sprintf('%s=%s', $key, $arguments[$key]);
                },
                array_keys($arguments)
            )
        );
    }

    /**
     * Build a host string from the given configuration.
     *
     * @param array  $config    Config database.
     * @param string $separator Separator configs.
     *
     * @return string
     */
    protected function buildHostString(array $config, string $separator): string {
        if (isset($config['port']) && ! empty($config['port'])) {
            return $config['host'] . $separator . $config['port'];
        } else {
            return $config['host'];
        }
    }

    /**
     * Get the available PDO drivers.
     *
     * @return array
     */
    protected function getAvailableDrivers(): array {
        return PDO::getAvailableDrivers();
    }
}
