<?php

declare(strict_types=1);

namespace Api\Database\Contracts;

use Closure;
use Exception;
use Throwable;

/**
 * Trait ManagesTransactions
 */
trait ManagesTransactionsTrait {

    /**
     * Start a new database transaction.
     *
     * @throws \Exception Error.
     *
     * @return void
     */
    public function beginTransaction() {
        $this->createTransaction();

        ++$this->_transactions;
    }

    /**
     * Execute a Closure within a transaction.
     *
     * @param \Closure $callback Callback.
     * @param integer  $attempts Attempts.
     *
     * @throws Throwable Error.
     *
     * @return mixed
     */
    public function transaction(Closure $callback, int $attempts = 1) {
        for ($currentAttempt = 1; $currentAttempt <= $attempts; $currentAttempt++) {
            $this->beginTransaction();

            try {
                return $this->tap(
                    $callback($this),
                    function () {
                        $this->commit();
                    }
                );
            } catch (Exception $e) {
                $this->handleTransactionException(
                    $e,
                    $currentAttempt,
                    $attempts
                );
            } catch (Throwable $e) {
                $this->rollBack();

                throw $e;
            }
        }
    }

    /**
     * Handle an exception encountered when running a transacted statement.
     *
     * @param \Exception $e              Error.
     * @param integer    $currentAttempt Attempt.
     * @param integer    $maxAttempts    Max attempt.
     *
     * @throws \Exception Error.
     *
     * @return void
     */
    protected function handleTransactionException(Exception $e, int $currentAttempt, int $maxAttempts) {
        if ($this->causedByDeadlock($e)
            && $this->_transactions > 1
        ) {
            --$this->_transactions;

            throw $e;
        }

        $this->rollBack();

        if ($this->causedByDeadlock($e)
            && $currentAttempt < $maxAttempts
        ) {
            return;
        }

        throw $e;
    }

    /**
     * Create a transaction within the database.
     *
     * @throws Exception Error.
     *
     * @return void
     */
    protected function createTransaction() {
        if ($this->_transactions == 0) {
            try {
                $this->getPdo()->beginTransaction();
            } catch (Exception $e) {
                $this->handleBeginTransactionException($e);
            }
        } else if ($this->_transactions >= 1) {
            $this->createSavepoint();
        }
    }

    /**
     * Create a save point within the database.
     *
     * @return void
     */
    protected function createSavepoint() {
        $this->getPdo()->exec(
            'SAVEPOINT trans' . ($this->_transactions + 1)
        );
    }

    /**
     * Handle an exception from a transaction beginning.
     *
     * @param \Exception $e Error.
     *
     * @throws \Exception Error.
     *
     * @return void
     */
    protected function handleBeginTransactionException(Exception $e) {
        if ($this->causedByLostConnection($e)) {
            $this->reconnect();

            $this->_pdo->beginTransaction();
        } else {
            throw $e;
        }
    }

    /**
     * Commit the active database transaction.
     *
     * @return void
     */
    public function commit() {
        if ($this->_transactions == 1) {
            $this->getPdo()->commit();
        }

        $this->_transactions = max(0, $this->_transactions - 1);
    }

    /**
     * Rollback the active database transaction.
     *
     * @param integer|null $toLevel Level database transaction.
     *
     * @return void
     */
    public function rollBack(?int $toLevel = null) {
        $toLevel = is_null($toLevel) ? $this->_transactions - 1 : $toLevel;

        if ($toLevel < 0 || $toLevel >= $this->_transactions) {
            return;
        }

        $this->performRollBack($toLevel);

        $this->_transactions = $toLevel;
    }

    /**
     * Perform a rollback within the database.
     *
     * @param integer $toLevel Level database transaction.
     *
     * @return void
     */
    protected function performRollBack(int $toLevel) {
        if ($toLevel == 0) {
            $this->getPdo()->rollBack();
        } else {
            $this->getPdo()->exec(
                'ROLLBACK TO SAVEPOINT trans' . ($toLevel + 1)
            );
        }
    }

    /**
     * Get the number of active transactions.
     *
     * @return integer
     */
    public function transactionLevel(): int {
        return $this->_transactions;
    }
}
