<?php

declare(strict_types=1);

namespace Api\Database;

use Api\Database\Configs\ConfigInterface;
use Api\Database\Connectors\ConnectionFactory;
use InvalidArgumentException;
use PDO;

/**
 * Class DatabaseManager
 */
class DatabaseManager implements ConnectionResolverInterface {

    /**
     * Databases configuration.
     *
     * @var ConfigInterface
     */
    protected ConfigInterface $config;
    /**
     * The active connection instances.
     *
     * @var array
     */
    protected $_connections = [];
    /**
     * The custom connection resolvers.
     *
     * @var array
     */
    protected $_extensions = [];
    /**
     * The database connection factory instance.
     *
     * @var ConnectionFactory
     */
    protected $_factory;

    /**
     * Create a new database manager instance.
     *
     * @param ConfigInterface   $config  Databases configuration.
     * @param ConnectionFactory $factory ConnectionFactory.
     *
     * @return void
     */
    public function __construct(ConfigInterface $config, ConnectionFactory $factory) {
        $this->config   = $config;
        $this->_factory = $factory;
    }

    /**
     * Dynamically pass methods to the default connection.
     *
     * @param string $method     Method.
     * @param array  $parameters Parameters.
     *
     * @return mixed
     */
    public function __call(string $method, array $parameters) {
        return $this->connection()->$method(...$parameters);
    }

    /**
     * Get a database connection instance.
     *
     * @param string $name Database connection name.
     *
     * @return Connection
     */
    public function connection($name = null): ConnectionInterface {
        [$database, $type] = $this->parseConnectionName($name);

        $name = $name ?: $database;

        if (!isset($this->_connections[$name])) {
            $connection                = $this->makeConnection($database);
            $this->_connections[$name] = $this->configure(
                $connection,
                $type
            );
        }

        return $this->_connections[$name];
    }

    /**
     * Parse the connection into an array of the name and read / write type.
     *
     * @param string|null $name Connection of the name.
     *
     * @return array
     */
    protected function parseConnectionName(?string $name): array {
        $name = $name ?: $this->getDefaultConnection();

        return [$name, 'write'];
    }

    /**
     * Get the default connection name.
     *
     * @return string
     */
    public function getDefaultConnection(): string {
        return $this->config->get('defaultConnection');
    }

    /**
     * Make the database connection instance.
     *
     * @param string $name Database connection name.
     *
     * @return Connection
     */
    protected function makeConnection(string $name): Connection {
        $config = $this->configuration($name);

        if (isset($this->_extensions[$name])) {
            return call_user_func($this->_extensions[$name], $config, $name);
        }

        $driver = $config['driver'];
        if (isset($this->_extensions[$driver])) {
            return call_user_func($this->_extensions[$driver], $config, $name);
        }

        return $this->_factory->make($config, $name);
    }

    /**
     * Get the configuration for a connection.
     *
     * @param string $name Configuration name.
     *
     * @return array
     *
     * @throws InvalidArgumentException Error.
     */
    protected function configuration(string $name): array {
        $name = $name ?: $this->getDefaultConnection();

        $connections = $this->config->get('connections');

        $config = $connections[$name] ?? null;
        if (is_null($config)) {
            throw new InvalidArgumentException(sprintf('Database [%s] not configured.', $name));
        }

        return $config;
    }

    /**
     * Prepare the database connection instance.
     *
     * @param Connection $connection Database connection instance.
     * @param string     $type       Type connection.
     *
     * @return Connection
     */
    protected function configure(Connection $connection, string $type): Connection {
        $connection = $this->setPdoForType($connection, $type);

        $connection->setReconnector(
            function ($connection) {
                $this->reconnect($connection->getName());
            }
        );

        return $connection;
    }

    /**
     * Prepare the read / write mode for database connection instance.
     *
     * @param Connection  $connection Database connection instance.
     * @param string|null $type       Type connection.
     *
     * @return Connection
     */
    protected function setPdoForType(Connection $connection, string $type = null): Connection {
        if ($type == 'read') {
            $connection->setPdo($connection->getReadPdo());
        } else {
            if ($type == 'write') {
                $connection->setReadPdo($connection->getPdo());
            }
        }

        return $connection;
    }

    /**
     * Reconnect to the given database.
     *
     * @param string|null $name Database name.
     *
     * @return Connection
     */
    public function reconnect(string $name = null) {
        $name = $name ?: $this->getDefaultConnection();
        $this->disconnect($name);

        if (!isset($this->connections[$name])) {
            return $this->connection($name);
        }

        return $this->refreshPdoConnections($name);
    }

    /**
     * Disconnect from the given database.
     *
     * @param string|null $name Connection name.
     *
     * @return void
     */
    public function disconnect(string $name = null) {
        $name = $name ?: $this->getDefaultConnection();
        if (isset($this->connections[$name])) {
            $this->connections[$name]->disconnect();
        }
    }

    /**
     * Refresh the PDO connections on a given connection.
     *
     * @param string $name Connection name.
     *
     * @return Connection
     */
    protected function refreshPdoConnections(string $name): Connection {
        $fresh = $this->makeConnection($name);

        return $this->_connections[$name]
            ->setPdo($fresh->getPdo())
            ->setReadPdo($fresh->getReadPdo());
    }

    /**
     * Get all drivers that are actually available.
     *
     * @return array
     */
    public function availableDrivers(): array {
        return array_intersect(
            $this->supportedDrivers(),
            str_replace('dblib', 'sqlsrv', PDO::getAvailableDrivers())
        );
    }

    /**
     * Get all support drivers.
     *
     * @return array
     */
    public function supportedDrivers(): array {
        return ['mysql', 'sqlite', 'sqlsrv'];
    }

    /**
     * Register an extension connection resolver.
     *
     * @param string   $name     Connection name.
     * @param callable $resolver Connection resolver.
     *
     * @return void
     */
    public function extend(string $name, callable $resolver) {
        $this->_extensions[$name] = $resolver;
    }

    /**
     * Return all created connections.
     *
     * @return array
     */
    public function getConnections(): array {
        return $this->_connections;
    }

    /**
     * Disconnect from the given database and remove from local cache.
     *
     * @param string|null $name Connection name.
     *
     * @return void
     */
    public function purge(string $name = null) {
        $name = $name ?: $this->getDefaultConnection();

        $this->disconnect($name);

        unset($this->_connections[$name]);
    }
}
