<?php
declare(strict_types=1);

namespace Api\Database\Query;

use Api\Database\ConnectionInterface;
use Api\Support\BaseCollectObject;
use Latitude\QueryBuilder\CriteriaInterface;
use Latitude\QueryBuilder\EngineInterface;
use Latitude\QueryBuilder\Query\DeleteQuery;
use Latitude\QueryBuilder\Query\InsertQuery;
use Latitude\QueryBuilder\Query\SelectQuery;
use Latitude\QueryBuilder\Query\UpdateQuery;
use Latitude\QueryBuilder\StatementInterface;

/**
 * Class QueryBuilder
 */
class QueryBuilder {

    /**
     * Query engine
     *
     * @var EngineInterface
     */
    protected $_engine;

    /**
     * Connection
     *
     * @var ConnectionInterface
     */
    private $_connection;

    /**
     * Query builder
     *
     * @var DeleteQuery|InsertQuery|SelectQuery|UpdateQuery
     */
    private $_query;


    public function __construct(ConnectionInterface $connection, EngineInterface $engine) {
        $this->_engine     = $engine;
        $this->_connection = $connection;
    }

    public function addFrom(...$tables): QueryBuilder {
        $this->_query->addFrom(...$tables);
        return $this;
    }

    public function andWhere(CriteriaInterface $criteria): QueryBuilder {
        $this->_query->andWhere($criteria);

        return $this;
    }

    /**
     * Create a new DELETE query
     *
     * @param string $table Name table.
     *
     * @return QueryBuilder
     */
    public function delete(string $table): QueryBuilder {
        $this->_query = $this->getEngine()->makeDelete();
        $this->_query->from($table);

        return $this;
    }

    /**
     * Get engine
     *
     * @return EngineInterface
     */
    public function getEngine(): EngineInterface {
        return $this->_engine;
    }

    public function execute(): int {
        if ($this->_query instanceof UpdateQuery) {
            return $this->updateQuery();
        } else {
            return $this->deleteQuery();
        }
    }

    public function updateQuery(): int {
        $query = $this->_query->compile();
        return $this->_connection->update($query->sql(), $query->params());
    }

    public function deleteQuery(): int {
        $query = $this->_query->compile();
        return $this->_connection->delete($query->sql(), $query->params());
    }

    public function from($tables): QueryBuilder {
        $this->_query->from($tables);

        return $this;
    }

    /**
     * Execute the query
     *
     * @return BaseCollectObject
     */
    public function get(): BaseCollectObject {
        $results = $this->runSelect();

        return new BaseCollectObject($results);
    }

    /**
     * Run the query as a "select" statement against the connection.
     *
     * @return array
     */
    protected function runSelect(): array {
        $query = $this->_query->compile();

        return $this->_connection->select(
            $query->sql(),
            $query->params()
        );
    }

    /**
     * Create a new INSERT query
     *
     * @param string  $table          Table name.
     * @param array   $map            Params.
     * @param boolean $returnInsertId If true return id.
     *
     * @return boolean|integer
     */
    public function insert(string $table, array $map = [], bool $returnInsertId = false) {
        $this->_query = $this->_engine->makeInsert();
        $this->_query->into($table);
        if (empty($map) === false) {
            $this->_query->map($map);
        }
        if (!$returnInsertId) {
            return $this->_connection->insert($this->_query->compile()->sql(), array_values($map));
        } else {
            $id = $this->_connection->getPdo()->lastInsertId();

            return is_numeric($id) ? (int)$id : $id;
        }
    }

    public function into($table): QueryBuilder {
        $this->_query->into($table);

        return $this;
    }

    public function lastInsertId() {
        $id = $this->_connection->getPdo()->lastInsertId();

        return is_numeric($id) ? (int)$id : $id;
    }

    public function leftJoin($table, CriteriaInterface $criteria): QueryBuilder {
        $this->_query->join($table, $criteria, 'LEFT');

        return $this;
    }

    public function limit(?int $limit): QueryBuilder {
        if (isset($this->_query)) {
            $this->_query->limit($limit);
        }

        return $this;
    }

    public function offset(?int $offset): QueryBuilder {
        $this->_query->offset($offset);

        return $this;
    }

    public function orWhere(CriteriaInterface $criteria): QueryBuilder {
        $this->_query->orWhere($criteria);

        return $this;
    }

    public function orderBy($column, string $direction = ''): QueryBuilder {
        $this->_query->orderBy($column, $direction);

        return $this;
    }

    /**
     * Create a new SELECT DISTINCT query
     *
     * @param string|StatementInterface ...$columns Array columns.
     *
     * @return QueryBuilder
     */
    public function selectDistinct(...$columns): QueryBuilder {
        return $this->select(...$columns)->distinct();
    }

    public function distinct($state = true): QueryBuilder {
        $this->_query->distinct($state);

        return $this;
    }

    /**
     * Create a new SELECT query
     *
     * @param string|StatementInterface ...$columns Array columns.
     *
     * @return QueryBuilder
     */
    public function select(...$columns): QueryBuilder {
        $this->_query = $this->getEngine()->makeSelect();
        if (empty($columns) === false) {
            $this->_query->columns(...$columns);
        }

        return $this;
    }

    public function toSql(): string {
        $query       = $this->_query->compile();
        $queryString = $query->sql();
        $queryParams = $query->params();
        $i           = 0;
        while (($pos = @strpos($queryString, '?', $i)) !== false) {
            $currentVal = array_shift($queryParams);
            if ($currentVal === true || $currentVal === false) {
                $currentVal = intval($currentVal) ? 'true' : 'false';
            } else {
                $currentVal = "'" . $currentVal . "'";
            }
            $queryString = substr($queryString, 0, $pos) . $currentVal . substr(
                $queryString,
                $pos + 1,
                strlen($queryString)
            );
            $i           = $pos;
        }

        return $queryString;
    }

    /**
     * Create a new UPDATE query
     *
     * @param string|StatementInterface $table Table name.
     * @param array                     $map   Values array.
     *
     * @return QueryBuilder
     */
    public function update($table, array $map = []): QueryBuilder {
        $this->_query = $this->getEngine()->makeUpdate()->table($table);
        if (empty($map) === false) {
            $this->_query->set($map);
        }

        return $this;
    }

    public function values(array $params): QueryBuilder {
        $this->_query->values($params);
        return $this;
    }

    /**
     * Add where condition to query
     *
     * @param CriteriaInterface|array $criteria If we use an array it should consist of [feild, field, condition].
     *
     * @return $this
     */
    public function where($criteria): QueryBuilder {
        if ($criteria instanceof CriteriaInterface) {
            if (isset($this->_query)) {
                $this->_query->where($criteria);
            }
        } else if (is_array($criteria)) {
            foreach ($criteria as $item) {
                $itemCriteria = $this->buildCriteria($item);
                if ($itemCriteria) {
                    $this->_query->andWhere($criteria);
                }
            }
        }

        return $this;
    }

    private function buildCriteria(array $item): ?CriteriaInterface {
        $criteria = new QueryCriteriaBuilder($item);
        return $criteria->getCriteriaCondition();
    }
}
