<?php

declare(strict_types=1);

namespace Api\Database;

use PDOException;

/**
 * Class QueryException
 */
class QueryException extends PDOException {

    /**
     * The SQL for the query.
     *
     * @var string
     */
    protected $_sql;

    /**
     * The bindings for the query.
     *
     * @var array
     */
    protected $_bindings;

    /**
     * Create a new query exception instance.
     *
     * @param string     $sql      Query string.
     * @param array      $bindings Bindings.
     * @param \Exception $previous Error.
     *
     * @return void
     */
    public function __construct(string $sql, array $bindings, \Exception $previous) {
        parent::__construct('', 0, $previous);

        $this->_sql      = $sql;
        $this->_bindings = $bindings;
        $this->code      = $previous->getCode();
        $this->message   = $this->formatMessage($sql, $bindings, $previous);

        if ($previous instanceof PDOException) {
            $this->errorInfo = $previous->errorInfo;
        }
    }

    /**
     * Format the SQL error message.
     *
     * @param string     $sql      Query string.
     * @param array      $bindings Bindings.
     * @param \Exception $previous Error.
     *
     * @return string
     */
    protected function formatMessage(string $sql, array $bindings, \Exception $previous): string {
        return $previous->getMessage() . ' (SQL: ' . $this->replaceArray('?', $bindings, $sql) . ')';
    }

    /**
     * Get the SQL for the query.
     *
     * @return string
     */
    public function getSql(): string {
        return $this->_sql;
    }

    /**
     * Get the bindings for the query.
     *
     * @return array
     */
    public function getBindings(): array {
        return $this->_bindings;
    }

    /**
     * ReplaceArray
     *
     * @param string $find     Finding param char.
     * @param array  $bindings Binndings data.
     * @param string $sql      Result query with binding params.
     *
     * @return string
     */
    private function replaceArray(string $find, array $bindings, string $sql):string {
        $query = '';
        foreach (array_fill(0, count($bindings), $find) as $key => $wildcard) {
            $query = substr_replace($query, '"' . $bindings[$key] . '"', strpos($sql, $wildcard), strlen($wildcard));
        }
        return $query;
    }
}
