<?php
declare(strict_types=1);

namespace Api\Dialplans\Rules\Imports\Repository;

use Api\Dialplans\Rules\Imports\Interfaces\ImportRepositoryInterface;
use Api\Dialplans\Rules\Imports\Request\ImportRequest;
use Wildix\Databases\Database;

/**
 * Class ImportRepository
 */
class ImportRepository extends Database implements ImportRepositoryInterface {

    public const TRANSFORMS = [
        'GoToIfFaxDetected' => [
            'old' => '/^([^,]*),([^,]*),([^,]*),([^,]*)$/', 'new' => '$1,$2,$3,$4,""', 'appname' => 'app-faxdetect'
        ],
        'FaxToEmail' => [
            'old' => '/^([^,]+),([^,]*),([^,]*),e,([^,]*)$/', 'new' => '$1,$2,$3,e', 'appname' => 'app-fax2email'
        ],
        'Background' => [
            'old' => '/^([^,]*)$/', 'new' => '$1,""', 'appname' => 'app-playback-waitDigits'
        ],
        'Playback' => [
            'old' => '/^([^,]*)$/', 'new' => '$1,""', 'appname' => 'app-playback'
        ],
        'DialPhone' => [
            'old' => '/^([^,]*),([^,]*),([^,]*)$/', 'new' => '$1,$2,$3,""', 'appname' => 'app-dial-internalcall'
        ],
        'FeatureIntercom' => [
            'old' => '/^([^,]+),([^,]*),([^,]*)$/', 'new' => '$1,$2,$3,""', 'appname' => 'app-service-intercom'
        ],
        'DialTrunk' => [
            'old' => '/^([^,]*),([^,]*),([^,]*),([^,]*),([^,]*)$/', 'new' => '$1,$2,$3,$4,$5,"auto"', 'appname' => 'app-dial-trunk'
        ],
        'GotoIfUserStatus' => [
            'old' => '/^([^,]*),([^,]*),([^,]*),([^,]*),([^,]*)$/', 'new' => '"${EXTEN}",$1,$2,$3,$4,$5', 'appname' => 'app-goto-if-userstatus'
        ]
    ];


    public function __construct() {
        $path = self::WILDIX_WMSDB;
        parent::__construct($path);
    }

    public function dropBackup(): void {
        $queries = [
            '_bkp_dialplan_links',
            '_bkp_dialplan_rulenames',
            '_bkp_dialplan_rules',
            '_bkp_dialplan_rules_includes',
        ];
        foreach ($queries as $query) {
            $this->dropTable($query);
        }
    }

    public function importRulesTablesToDb(ImportRequest $request): bool {
        try {
            $output = [];
            $result = [];
            exec('zcat ' . $request->getFileTmpName() . ' | sqlite3 ' . self::WILDIX_WMSDB, $output, $result);
            if ($result != 0) {
                return false;
            }
        } catch (\Throwable $exception) {
            $this->getLog()->error(__LINE__, $exception->getMessage());
            return false;
        }
        return true;
    }

    public function updateFromNewTables($backup): bool {
        $table   = ($backup ? '_bkp_' : '') . 'dialplan_rules';
        $rows    = $this->getApplicationTransformsQuery($table);
        $update  = 'UPDATE ' . $table . " SET parameters='%s', dialplanappname='%s' where id = %d";
        $queries = [];
        foreach ($rows as $row) {
            $params    = preg_replace(self::TRANSFORMS[$row['application']]['old'], self::TRANSFORMS[$row['application']]['new'], $row['parameters']);
            $params    = $this->realEscapeString($params);
            $queries[] = sprintf($update, $params, self::TRANSFORMS[$row['application']]['appname'], $row['id']);
        }

        if (count($queries)) {
            foreach ($queries as $query) {
                $this->exec($query);
            }
        }
        return (bool)count($queries);
    }

    private function getApplicationTransformsQuery(string $table): array {
        $sql = 'SELECT * from ' . $table . " where application in ('" . implode("','", array_keys(self::TRANSFORMS)) . "')";
        return $this->getRows($sql);
    }

    public function getNamesRuleFromBackup(): array {
        try {
            $sqlQuery = 'SELECT id, rulename as name, ruledescription as desc FROM _bkp_dialplan_rulenames ORDER BY rulename;';
            return $this->getRows($sqlQuery);
        } catch (\Exception $e) {
            return [];
        }
    }
}
