<?php
declare(strict_types=1);

namespace Api\Dialplans\Rules\Imports\Services;

use Api\Dialplans\Rules\Imports\Interfaces\ImportRepositoryInterface;
use Api\Dialplans\Rules\Imports\Interfaces\ImportServiceInterface;
use Api\Dialplans\Rules\Imports\Objects\RuleName;
use Api\Dialplans\Rules\Imports\Objects\RuleNames;
use Api\Dialplans\Rules\Imports\Request\ImportRequest;
use Api\Dialplans\Rules\Imports\Response\RulesImportResponse;
use Wildix\Dialplan\IncomingCalls;

/**
 * Class ImportService
 */
class ImportService implements ImportServiceInterface {

    /**
     * Dialplan Incoming Calls
     *
     * @var IncomingCalls
     */
    private $_incomingCalls;
    private $_repository;

    public function __construct(IncomingCalls $incomingCalls, ImportRepositoryInterface $repository) {
        $this->_incomingCalls = $incomingCalls;
        $this->_repository    = $repository;
    }

    public const DEV_SHM_TMP_IMPORT = '/dev/shm/tmp/import_';

    public function validateBackupFile(ImportRequest $request): ImportRequest {
        $fileContent = file_get_contents('compress.zlib://' . $request->getFileTmpName());
        $statements  = preg_split('/;[\r\n|\r|\n]/', $fileContent);
        $isLegal     = true;
        foreach ($statements as $statement) {
            if (!trim($statement)) {
                continue;
            }
            if (preg_match('/PRAGMA|BEGIN TRANSACTION|COMMIT/i', $statement) === 1) {
                continue;
            }
            if (preg_match('/(CREATE TABLE|ALTER TABLE|INSERT INTO) (\w+)/i', preg_replace(array('/\W/', '/\s+/'), ' ', $statement), $matches)) {
                if (!preg_match('/_bkp_(dialplan_links|dialplan_rulenames|dialplan_rules|dialplan_rules_includes)/i', $matches[2])) {
                    $isLegal = false;
                    $request->setFileErrorMessage(__LINE__ . 'Not legal backup table detected ' . $statement);
                    break;
                }
                continue;
            }
            $isLegal = false;
            $request->setFileErrorMessage(__LINE__ . 'Not legal backup table detected ' . $statement);
            break;
        }
        $request->setIsDialplanFile($isLegal);
        return $request;
    }

    public function dropBackup(): void {
        $this->getImportRepository()->dropBackup();
    }

    public function moveUploadedImportFile(ImportRequest $request): bool {
        $tmpName = self::DEV_SHM_TMP_IMPORT . microtime(true);
        $src     = $request->getFileTmpName();
        if (move_uploaded_file($src, $tmpName)) {
            $request->setFileTmpName($tmpName);
            return true;
        }
        return false;
    }

    public function importRules(ImportRequest $request): bool {
        return $this->getImportRepository()->importRulesTablesToDb($request);
    }

    public function updateFromNewTables(bool $backup = false): bool {
        $result = $this->getImportRepository()->updateFromNewTables($backup);
        if (!$backup && $result) {
            $this->_incomingCalls->set_extensions_incomingcalls();
            $this->_incomingCalls->set_queues_conf();
        }
        return $result;
    }

    public function getAllNameFromBackup(): RulesImportResponse {
        $response = new RulesImportResponse();
        $result   = new RuleNames();
        $rows     = $this->getImportRepository()->getNamesRuleFromBackup();
        foreach ($rows as $row) {
            $ruleName = new RuleName();
            $ruleName
                ->setId($row['id'] ?? 0)
                ->setName($row['name'] ?? '')
                ->setDescription($row['desc'] ?? '');
            $result->push($ruleName);
        }

        return $response->setRuleNames($result);
    }

    public function restoreBackup($ids): array {
        return $this->_incomingCalls->restoreBackup($ids);
    }

    private function getImportRepository(): ImportRepositoryInterface {
        return $this->_repository;
    }
}
