<?php
declare(strict_types=1);

namespace Api\Dictionary\Services;

use Api\Dictionary\Enum\AttachFormatsEnum;
use Api\Dictionary\Enum\CallTypesEnum;
use Api\Dictionary\Enum\CauseCodesEnum;
use Api\Dictionary\Enum\ConditionExamplesEnum;
use Api\Dictionary\Enum\DialPlanSetKeysEnum;
use Api\Dictionary\Enum\DialStatusesEnum;
use Api\Dictionary\Enum\DialTonesEnum;
use Api\Dictionary\Enum\DictionaryTypeEnum;
use Api\Dictionary\Enum\HgStrategiesEnum;
use Api\Dictionary\Enum\ItcEnum;
use Api\Dictionary\Enum\LanguagesEnum;
use Api\Dictionary\Enum\SipCodecsEnum;
use Api\Dictionary\Enum\SrtpSecurityEnum;
use Api\Dictionary\Enum\TimeoutEnum;
use Api\Dictionary\Enum\UserStatusesEnum;
use Api\Dictionary\Interfaces\DialplanRulesRepositoryInterface;
use Api\Dictionary\Objects\AttachFormats;
use Api\Dictionary\Objects\CallGroups;
use Api\Dictionary\Objects\CallTypes;
use Api\Dictionary\Objects\CauseCodes;
use Api\Dictionary\Objects\CidNames;
use Api\Dictionary\Objects\CidNumber;
use Api\Dictionary\Objects\ClassesOfCalls;
use Api\Dictionary\Objects\ConditionExamples;
use Api\Dictionary\Objects\Contexts;
use Api\Dictionary\Objects\DialStatuses;
use Api\Dictionary\Objects\DirectRtp;
use Api\Dictionary\Objects\HgStrategies;
use Api\Dictionary\Objects\HuntGroups;
use Api\Dictionary\Objects\Itc;
use Api\Dictionary\Objects\IvrTrees;
use Api\Dictionary\Objects\Languages;
use Api\Dictionary\Objects\MohClasses;
use Api\Dictionary\Objects\PageGroups;
use Api\Dictionary\Objects\Phonebooks;
use Api\Dictionary\Objects\RemotePbxes;
use Api\Dictionary\Objects\Ringtones;
use Api\Dictionary\Objects\SelectKeyValueObject;
use Api\Dictionary\Objects\SipCodecs;
use Api\Dictionary\Objects\SrtpSecurity;
use Api\Dictionary\Objects\Timeout;
use Api\Dictionary\Objects\TimeTables;
use Api\Dictionary\Objects\Trunks;
use Api\Dictionary\Objects\UserStatuses;
use Api\Dictionary\Objects\WmsSwitches;
use Api\Dictionary\Objects\WmsThreeSwitches;
use Api\Dictionary\Repository\DictionaryRepository;
use Api\Translation\Objects\TranslationInterface;
use Api\Dictionary\Objects\DialTones;
use Wildix\License;
use Wildix\Dialplan\RulesService;
use Wildix\Sounds\SoundsService;
use WMS_Ldap;

/**
 * Class DictionaryService
 */
class DictionaryService {

    public function __construct(
        private DialplanRulesRepositoryInterface $repository,
        private WMS_Ldap $wmsLdap,
        private TranslationInterface $translation
    ) {
    }

    /**
     * Return Dialplan Service.
     *
     * @return DictionaryRepository
     */
    private function getDictionaryRepository(): DialplanRulesRepositoryInterface {
        return $this->repository;
    }

    private function getWmsLdap(): WMS_Ldap {
        return $this->wmsLdap;
    }

    private function getTranslateService(): TranslationInterface {
        return $this->translation;
    }

    /**
     * Get AttachFormats Collection
     *
     * @return AttachFormats
     */
    public function getAttachFormats(): AttachFormats {
        $result = new AttachFormats();
        foreach (AttachFormatsEnum::getEnumValues() as $key => $value) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setLabel($value)
                ->setValue($value);
            $result->push($keyValueObject);
        }
        return $result;
    }

    /**
     * Get CallGroups Collection
     *
     * @param boolean $isValueName Return id call group or name.
     *
     * @return CallGroups
     * @throws \Exception Error
     */
    public function getCallGroups(bool $isValueName = false): CallGroups {
        $result = new CallGroups();
        $rows   = $this->getDictionaryRepository()->getCallGroups();

        foreach ($rows as $row) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setLabel($this->getTranslateService()->t($row['call_groupname']))
                ->setValue($isValueName ? $row['call_groupname'] : $row['id']);
            $result->push($keyValueObject);
        }

        return $result;
    }

    /**
     * Get CallTypes Collection
     *
     * @return CallTypes
     */
    public function getCallTypes(): CallTypes {
        $result = new CallTypes();
        foreach (CallTypesEnum::getEnumValues() as $value) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setLabel($value)
                ->setValue($value);
            $result->push($keyValueObject);
        }
        return $result;
    }

    /**
     * Get CauseCodes Collection
     *
     * @return CauseCodes
     */
    public function getCauseCodes(): CauseCodes {
        $result = new CauseCodes();
        foreach (CauseCodesEnum::getEnumValues() as $value) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setLabel($value)
                ->setValue(CauseCodesEnum::getIdByValue($value));
            $result->push($keyValueObject);
        }
        return $result;
    }

    /**
     * Get ClassesOfCalls Collection
     *
     * @return ClassesOfCalls
     */
    public function getClassesOfCalls(): ClassesOfCalls {
        $result         = new ClassesOfCalls();
        $keyValueObject = new SelectKeyValueObject();
        $keyValueObject
            ->setLabel('Auto')
            ->setValue('auto');
        $result->push($keyValueObject);
        foreach (get_call_classes_no_assoc() as $k => $v) {
            if ((int)$k !== 20) {
                $keyValueObject = new SelectKeyValueObject();
                $keyValueObject
                    ->setLabel($this->getTranslateService()->t($v))
                    ->setValue($k);
                $result->push($keyValueObject);
            }
        }
        return $result;
    }

    /**
     * Get ConditionExamples Collection
     *
     * @return ConditionExamples
     */
    public function getConditionExamples(): ConditionExamples {
        $result = new ConditionExamples();
        foreach (ConditionExamplesEnum::getEnumValues() as $value) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setLabel($value)
                ->setValue($value);
            $result->push($keyValueObject);
        }
        return $result;
    }

    /**
     * Get Context Collection.
     *
     * @param string $topValue Key context.
     *
     * @return Contexts
     * @throws \Exception Error.
     */
    public function getContexts($topValue = ''): Contexts {
        $result = new Contexts();
        if ($topValue !== '') {
            $result[$topValue] = $topValue;
        }
        $rows = $this->getDictionaryRepository()->getContexts();
        foreach ($rows as $row) {
            $result->appendByFormat($row);
        }

        return $result;
    }

    /**
     * Get DialStatuses Collection
     *
     * @return DialStatuses
     */
    public function getDialStatuses(): DialStatuses {
        $result = new DialStatuses();
        foreach (DialStatusesEnum::getEnumValues() as $value) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setLabel($value)
                ->setValue($value);
            $result->push($keyValueObject);
        }
        return $result;
    }

    /**
     * Get HgStrategies Collection
     *
     * @return HgStrategies
     */
    public function getHgStrategies():HgStrategies {
        $result = new HgStrategies();
        foreach (HgStrategiesEnum::getEnumValues() as $key => $value) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setLabel($this->getTranslateService()->t($value))
                ->setValue(HgStrategiesEnum::getIdByValue($value));
            $result->push($keyValueObject);
        }
        return $result;
    }

    /**
     * Get IvrTree Collection
     *
     * @return IvrTrees
     * @throws \Exception Error
     */
    public function getIvrTrees(): IvrTrees {
        $result = new IvrTrees();
        $rows   = $this->getDictionaryRepository()->getIvrTrees();
        foreach ($rows as $row) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setLabel($row['number'])
                ->setValue($row['number']);
            $result->push($keyValueObject);
        }

        return $result;
    }

    /**
     * Get MohClasses Collection
     *
     * @return MohClasses
     */
    public function getMohClasses():MohClasses {
        $result = new MohClasses();
        foreach (SoundsService::getMohClasses() as $item) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setLabel($item)
                ->setValue($item);
            $result->push($keyValueObject);
        }

        return $result;
    }

    /**
     * Get CidNames Collection
     *
     * @return CidNames
     */
    public function getCidName():CidNames {
        $result         = new CidNames();
        $keyValueObject = new SelectKeyValueObject();
        $keyValueObject
            ->setLabel('ADD${CALLERID(name):REMOVE}')
            ->setValue('ADD${CALLERID(name):REMOVE}');
        $result->push($keyValueObject);

        return $result;
    }

    /**
     * Get Ringtones Collection
     *
     * @return Ringtones
     */
    public function getRingtones(): Ringtones {
        $result = new Ringtones();

        $ringtonesArray = SoundsService::getRingTones();
        foreach ($ringtonesArray as $fileWithPath => $fileFormats) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setLabel($fileWithPath)
                ->setValue($fileWithPath);
            $result->push($keyValueObject);
        }

        return $result;
    }

    /**
     * Return rule service
     *
     * @return RulesService
     */
    public function getRulesService(): RulesService {
        return $this->_rulesService;
    }

    /**
     * Get PageGroups Collection
     *
     * @return PageGroups
     * @throws \Exception Error
     */
    public function getPageGroups(): PageGroups {
        $result = new PageGroups();
        $rows   = $this->getDictionaryRepository()->getPageGroups();
        foreach ($rows as $row) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setValue($row['id'])
                ->setLabel($row['name']);
            $result->push($keyValueObject);
        }

        return $result;
    }

    /**
     * Get Phonebooks Collection
     *
     * @return Phonebooks
     */
    public function getPhoneBooks(): Phonebooks {
        $phonebook = new \phonebook();
        $result    = new Phonebooks();
        $new       = $phonebook->select_phonebooks();

        foreach ($new as $key => $value) {
            $new[$key]      = $value;
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setValue($key)
                ->setLabel($value);
            $result->push($keyValueObject);
        }

        return $result;
    }

    /**
     * Get RemotePbxes Collection
     *
     * @return RemotePbxes
     */
    public function getRemotePbxes(): RemotePbxes {
        $result = new RemotePbxes();
        $search = $this->getWmsLdap()->getEntries(\WMS_Ldap::PBX, '', array('pbxDomain', 'dn'));
        if (is_array($search)) {
            foreach ($search as $pbx) {
                $keyValueObject = new SelectKeyValueObject();
                $keyValueObject
                    ->setValue($pbx['dn'])
                    ->setLabel($pbx['pbxDomain']);
                $result->push($keyValueObject);
            }
        }

        return $result;
    }

    /**
     * Get TimeTables Collection
     *
     * @return TimeTables
     * @throws \Exception Error
     */
    public function getTimeTables(): TimeTables {
        $result = new TimeTables();
        $rows   = $this->getDictionaryRepository()->getTimeTables();
        foreach ($rows as $item) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setValue($item['id'])
                ->setLabel($item['timetable']);
            $result->push($keyValueObject);
        }

        return $result;
    }

    /**
     * Get DialPlanSetKeys
     *
     * @return array
     */
    public function getSetKeys(): array {
        $result = [];
        foreach (DialPlanSetKeysEnum::getEnumValues() as $field) {
            if ($field === DictionaryTypeEnum::SERVICE && !License::getInstance()->haveLicense(License::TYPE_XBEES)) {
                continue;
            }
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setLabel(DialPlanSetKeysEnum::getIdByValue($field))
                ->setValue($field);
            $result[] = $keyValueObject->toArray();
        }
        return $result;
    }

    /**
     * Get Trunks Collection
     *
     * @return Trunks
     * @throws \Exception Error
     */
    public function getTrunks(): Trunks {
        $result = new Trunks();
        $rows   = $this->getDictionaryRepository()->getTrunks();
        if ($rows) {
            foreach ($rows as $trank) {
                $result->appendByFormat($trank);
            }
        }
        $keyValueObject = new SelectKeyValueObject();
        $keyValueObject
            ->setValue('classound')
            ->setLabel('classound');
        $result->push($keyValueObject);

        return $result;
    }

    /**
     * Get UserStatuses Collection
     *
     * @return UserStatuses
     */
    public function getUserStatuses(): UserStatuses {
        $result = new UserStatuses();

        foreach (UserStatusesEnum::getEnumValues() as $value) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setValue($value)
                ->setLabel($value);
            $result->push($keyValueObject);
        }
        return $result;
    }

    /**
     * Get WmsThreeSwitches Collection
     *
     * @return WmsThreeSwitches WmsThreeSwitches Collection
     * @throws \Exception Error
     */
    public function getWms3Switches(): WmsThreeSwitches {
        $result = new WmsThreeSwitches();
        $rows   = $this->getDictionaryRepository()->getWms3Switches();
        if ($rows) {
            foreach ($rows as $row) {
                $keyValueObject = new SelectKeyValueObject();
                $keyValueObject
                    ->setValue($row['id'])
                    ->setLabel($row['switchname']);
                $result->push($keyValueObject);
            }
        }
        return $result;
    }

    /**
     * Get Dialplan WMS Switches Collection
     *
     * @return WmsSwitches Dialplan WMS Switches Collection
     * @throws \Exception Error
     */
    public function getWmsSwitches(): WmsSwitches {
        $result = new WmsSwitches();
        $rows   = $this->getDictionaryRepository()->getWmsSwitches();
        if ($rows) {
            foreach ($rows as $row) {
                $keyValueObject = new SelectKeyValueObject();
                $keyValueObject
                    ->setLabel($row['switchname'])
                    ->setValue($row['id']);
                $result->push($keyValueObject);
            }
        }

        return $result;
    }

    /**
     * Get Hunt Groups Collection
     *
     * @return HuntGroups Hunt Groups Collection
     * @throws \Exception Error
     */
    public function getHuntGroups(): HuntGroups {
        $result = new HuntGroups();
        $rows   = $this->getDictionaryRepository()->getHuntGroups();
        if ($rows) {
            foreach ($rows as $row) {
                $keyValueObject = new SelectKeyValueObject();
                $keyValueObject->setValue($row['id']);
                if (!empty($row['description'])) {
                    $keyValueObject->setLabel($row['groupname'] . ' - ' . $row['description']);
                } else {
                    $keyValueObject->setLabel($row['groupname']);
                }
                $result->push($keyValueObject);
            }
        }
        return $result;
    }

    /**
     * Get CidNumber Collection
     *
     * @return CidNumber
     */
    public function getCidNumber(): CidNumber {
        $result         = new CidNumber();
        $keyValueObject = new SelectKeyValueObject();
        $keyValueObject
            ->setValue('ADD${CALLERID(number):REMOVE}')
            ->setLabel('ADD${CALLERID(number):REMOVE}');
        $result->push($keyValueObject);
        return $result;
    }

    /**
     * Get DirectRtp Collection
     *
     * @return DirectRtp
     */
    public function getDirectrtp(): DirectRtp {
        $result = new DirectRtp();
        foreach (['yes', 'no'] as $item) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setValue($item)
                ->setLabel($item);
            $result->push($keyValueObject);
        }
        return $result;
    }

    /**
     * Get Itc Collection
     *
     * @return Itc
     */
    public function getItc(): Itc {
        $result = new Itc();

        foreach (ItcEnum::getEnumValues() as $item) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setValue($item)
                ->setLabel($item);
            $result->push($keyValueObject);
        }
        return $result;
    }

    /**
     * Get Language Collection
     *
     * @return Languages
     */
    public function getLanguages() {
        $result = new Languages();

        foreach (LanguagesEnum::getEnumValues() as $item) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setValue($item)
                ->setLabel($item);
            $result->push($keyValueObject);
        }
        return $result;
    }

    /**
     * Get SrtpSecurity Collection
     *
     * @return SrtpSecurity
     */
    public function getSrtpSecurity(): SrtpSecurity {
        $result = new SrtpSecurity();

        foreach (SrtpSecurityEnum::getEnumValues() as $item) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setValue($item)
                ->setLabel($item);
            $result->push($keyValueObject);
        }
        return $result;
    }

    /**
     * Get Timeout Collection
     *
     * @return Timeout
     */
    public function getTimeout(): Timeout {
        $result = new Timeout();

        foreach (TimeoutEnum::getEnumValues() as $item) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setValue((int)$item)
                ->setLabel($item);
            $result->push($keyValueObject);
        }
        return $result;
    }

    public function getSipCodecs(): SipCodecs {
        $result = new SipCodecs();

        foreach (SipCodecsEnum::getEnumValues() as $item) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setValue($item)
                ->setLabel($item);
            $result->push($keyValueObject);
        }
        return $result;
    }

    public function getTags(): array {
        return [];
    }

    public function getTrunkCallerNum(): array {
        return [];
    }

    public function getDialTones(): DialTones {
        $result = new DialTones();

        foreach (DialTonesEnum::getEnums() as $key => $item) {
            $keyValueObject = new SelectKeyValueObject();
            $keyValueObject
                ->setValue($item)
                ->setLabel($key);
            $result->push($keyValueObject);
        }

        return $result;
    }

    public function getServices(): CallGroups {
        return $this->getCallGroups(true);
    }
}
