<?php

namespace Api\Network\Services;

use Api\Network\Builder\NetworkParticipantBuilder;
use Api\Network\Builder\NetworkParticipantDataBuilder;
use Api\Network\Builder\NetworkParticipantPbxModelBuilder;
use Api\Network\Enum\NetworkParticipantResponseFieldsEnum;
use Api\Network\Exception\ForbiddenException;
use Api\Network\Interfaces\NetworkFailOverServiceInterface;
use Api\Network\Interfaces\NetworkRepositoryInterface;
use Api\Network\Interfaces\NetworkServiceInterface;
use Api\Network\Objects\NetworkParticipants;
use Api\Network\Request\DeleteRequest;
use Api\Network\Request\UpdateRequest;
use Api\Network\Response\NetworkInfoResponse;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\PBXes\PBXes;

/**
 * Class NetworkService
 */
class NetworkService implements NetworkServiceInterface {

    private $_repository;
    private $_networkFailOverService;
    private $_pbxes;

    public function __construct(
        NetworkRepositoryInterface $_networkRepository,
        NetworkFailOverServiceInterface $networkFailOverService,
        PBXes $pbxes
    ) {
        $this->_repository             = $_networkRepository;
        $this->_networkFailOverService = $networkFailOverService;
        $this->_pbxes                  = $pbxes;
    }

    public function getItems(): NetworkInfoResponse {
        $response = new NetworkInfoResponse();
        $result   = new NetworkParticipants();

        if ($this->repository()->getCurrentPbx()->isEnabledWmsNetwork()) {
            $participants = $this->repository()->getNetworkParticipantsInfo();

            foreach ($participants as $pbx) {
                $dataBuilder = new NetworkParticipantDataBuilder($pbx);
                $builder     = new NetworkParticipantBuilder();

                $result->push($builder->build($dataBuilder, $this->getNetworkFailOverService()));
            }
        } else {
            $participant = $this->repository()->getCurrentPbx();

            $dataBuilder = new NetworkParticipantPbxModelBuilder($participant, $this->getPbxes());
            $builder     = new NetworkParticipantBuilder();

            $result->push($builder->build($dataBuilder, $this->getNetworkFailOverService()));
        }

        return $response->setNetworkParticipants($result);
    }

    public function update(UpdateRequest $request): void {
        if (!$this->repository()->isMaster()) {
            throw new ForbiddenException(Error::OnlyMasterCanEditNetwork);
        }

        if ($request->getSerial() === $request->getData()->getFailOver()) {
            throw new InvalidDataException(
                Error::getMessage(Error::BadRequest),
                [
                    NetworkParticipantResponseFieldsEnum::FAIL_OVER => [
                        'code' => Error::FailoverMustBeDifferent,
                        'reason' => Error::getMessage(Error::FailoverMustBeDifferent)
                    ]
                ]
            );
        }

        $currentFailOver = $this->repository()->getCurrentFailOverDn($request->getSerial());

        if ($currentFailOver) {
            $this->repository()->deleteFailOver($currentFailOver);
        }

        $failOverSerial = $request->getData()->getFailOver();

        if ($failOverSerial) {
            $this->repository()->addFailOver($request->getSerial(), $failOverSerial);
        }
    }

    public function delete(DeleteRequest $request): void {
        if (!$this->repository()->isMaster()) {
            throw new ForbiddenException(Error::OnlyMasterCanEditNetwork);
        }

        if ($request->getSerial() === $this->getMasterSerial()) {
            throw new ForbiddenException(Error::CantDeleteMasterPbx);
        }

        $ownerDn = $this->repository()->getCurrentOwnerDn($request->getSerial());

        if ($ownerDn) {
            $ownerSerial = $this->getPbxes()->getPBXByDN($ownerDn)->getSerial();

            throw new InvalidDataException(
                Error::getMessage(Error::BadRequest),
                [
                    NetworkParticipantResponseFieldsEnum::FAIL_OVER => [
                        'code' => Error::CantDeleteItIsFailover,
                        'reason' => sprintf(Error::getMessage(Error::CantDeleteItIsFailover), $ownerSerial)
                    ]
                ]
            );
        }

        $this->repository()->deletePbx($request->getSerial(), $request->getMigrateToSerial());
    }

    public function getMasterSerial(): string {
        return $this->repository()->getMasterSerial();
    }

    private function repository(): NetworkRepositoryInterface {
        return $this->_repository;
    }

    private function getNetworkFailOverService(): NetworkFailOverServiceInterface {
        return $this->_networkFailOverService;
    }

    private function getPbxes(): PBXes {
        return $this->_pbxes;
    }
}
