<?php
declare(strict_types=1);

namespace Api\Support;

use Api\Support\Interfaces\ArrayableInterface;
use Api\Support\Interfaces\Jsonable;
use ArrayAccess;
use ArrayIterator;
use Traversable;

/**
 * Class BaseCollectObject
 */
class BaseCollectObject implements ArrayAccess, \IteratorAggregate {

    /**
     * Items collection
     *
     * @var array
     */
    protected $_items = [];


    /**
     * BaseCollectObject constructor.
     *
     * @param array $items Items collection.
     */
    public function __construct(array $items = []) {
        $this->_items = $this->getArrayableItems($items);
    }

    /**
     * Get by type items
     *
     * @param ArrayableInterface|Jsonable|Traversable|array $items Items collection.
     *
     * @return array|mixed
     */
    protected function getArrayableItems($items): array {
        if (is_array($items)) {
            return $items;
        }

        if ($items instanceof self) {
            return $items->all();
        }

        if ($items instanceof ArrayableInterface) {
            return $items->toArray();
        }

        if ($items instanceof Jsonable) {
            return json_decode($items->toJson(), true);
        }

        if ($items instanceof Traversable) {
            return iterator_to_array($items);
        }

        return (array)$items;
    }

    /**
     * Get all items
     *
     * @return array|mixed
     */
    public function all(): array {
        return $this->_items;
    }

    /**
     * Determine if the collection is empty or not.
     *
     * @return boolean
     */
    public function isEmpty(): bool {
        return empty($this->_items);
    }

    /**
     * Check exists key
     *
     * @param string $key Find key.
     *
     * @return boolean
     */
    public function offsetExists($key): bool {
        return array_key_exists($key, $this->_items);
    }

    /**
     * Get by key
     *
     * @param mixed $key Key.
     *
     * @return mixed
     */
    public function offsetGet($key): mixed {
        return $this->_items[$key] ?? null;
    }

    /**
     * Get the keys of the collection items.
     *
     * @return static
     */
    public function keys(): BaseCollectObject {
        return new static(array_keys($this->_items));
    }

    /**
     * Concatenate values of a given key as a string.
     *
     * @param string $glue Defaults to an empty string. This is not the preferred usage of implode as glue would be the second parameter and thus, the bad prototype would be used.
     *
     * @return string
     */
    public function implode(string $glue = ""): string {
        return implode($glue, $this->_items);
    }

    /**
     * * Delete by key
     *
     * @param mixed $key Key.
     *
     * @return void
     */
    public function offsetUnset($key):void {
        unset($this->_items[$key]);
    }

    /**
     * Push some value in collection
     *
     * @param mixed $value Mixed value.
     *
     * @return $this
     */
    public function push($value): BaseCollectObject {
        $this->offsetSet(null, $value);

        return $this;
    }

    /**
     * Set offset key
     *
     * @param mixed $key   Key.
     * @param mixed $value Value.
     *
     * @return void
     */
    public function offsetSet($key, $value): void {
        if (is_null($key)) {
            $this->_items[] = $value;
        } else {
            $this->_items[$key] = $value;
        }
    }

    /**
     * Put to collection
     *
     * @param mixed $key   Key.
     * @param mixed $value Value.
     *
     * @return $this
     */
    public function put($key, $value): BaseCollectObject {
        $this->offsetSet($key, $value);

        return $this;
    }

    /**
     * Map collection with callback function
     *
     * @param callable $callback Callback function.
     *
     * @return $this
     */
    public function map(callable $callback): BaseCollectObject {
        $keys = array_keys($this->_items);

        $items = array_map($callback, $this->_items, $keys);

        return new static(array_combine($keys, $items));
    }

    public function transform(callable $callback): BaseCollectObject {
        $this->_items = $this->map($callback)->all();

        return $this;
    }

    /**
     * Get the collection of items as a plain array.
     *
     * @return array
     */
    public function toArray(): array {
        return array_map(
            static function ($value) {
                return $value instanceof ArrayableInterface ? $value->toArray() : $value;
            },
            $this->_items
        );
    }

    public function keyBy($key): BaseCollectObject {
        $result = [];
        foreach ($this->_items as $value) {
            if (is_object($value)) {
                $valueArray  = $value->toArray();
                $resolvedKey = $valueArray[$key];
            } else {
                $resolvedKey = $value[$key];
            }

            $result[$resolvedKey] = $value;
        }

        return new static($result);
    }

    public function filter(callable $callback): BaseCollectObject {
        return new static(array_values(array_filter($this->_items, $callback)));
    }

    public function getIterator(): ArrayIterator {
        return new ArrayIterator($this->_items);
    }

    public function count() {
        return count($this->_items);
    }

    /**
     * Return the first element in an array passing a given truth test.
     *
     * @param callable|null $callback Callback.
     * @param mixed         $default  Default value.
     *
     * @return mixed
     */
    public function first(callable $callback = null, $default = null) {
        if (is_null($callback)) {
            if (empty($this->_items)) {
                return value($default);
            }

            foreach ($this->_items as $item) {
                return $item;
            }
        }

        foreach ($this->_items as $key => $value) {
            if (call_user_func($callback, $value, $key)) {
                return $value;
            }
        }

        return value($default);
    }
}
