<?php

declare(strict_types=1);

namespace Api\Support\DTOValidate;

use Api\Support\Enums\StatusCodes;
use Spatie\DataTransferObject\DataTransferObject;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\Validator;

/**
 * Class DTOValidate
 */
abstract class DTOValidate extends DataTransferObject {
    const DTO_RULE = 'dto';
    protected $ignoreMissing = true;
    /**
     * Property validate
     *
     * @var Validator
     */
    protected $_validator;

    public function __construct($parameters = []) {
        parent::__construct($parameters);

        $this->validate();
    }

    abstract public function rules(): array;

    final public function validate(): void {
        $this->createDefaultValidator();

        foreach ($this->rules() as $field => $rule) {
            if ($rule === self::DTO_RULE) {
                if (is_array($field)) {
                    foreach ($this->$field as $item) {
                        $item->validate();
                    }
                } else {
                    $this->$field->validate();
                }
            } else if (is_array($rule)) {
                foreach ($rule as $item) {
                    $this->addRuleToValidator($item, $field);
                }
            } else {
                $this->addRuleToValidator($rule, $field);
            }
        }

        $this->_validator = $this->_validator->withData($this->getValidationData());
        $this->_validator->validate();

        $errors = array_merge_recursive($this->_validator->errors(), $this->customValidation());
        if (!empty($errors)) {
            $this->error(Error::getMessage(Error::VALIDATION_ERROR), StatusCodes::HTTP_BAD_REQUEST, $errors);
        }
    }

    /**
     * Custom validation
     *
     * @return array Errors array
     */
    protected function customValidation(): array {
        return [];
    }

    protected function createDefaultValidator(array $data = []): void {
        $this->_validator = new Validator($data);
    }

    protected function error(string $message, int $httpCode = null, $errors): void {
        throw new InvalidDataException($message, $errors, $httpCode);
    }

    private function getValidationData(): array {
        return array_merge(
            array_intersect_key((array)$this, $this->rules()),
            // global implicit dependency !
            $_REQUEST
        );
    }

    private function addRuleToValidator($rule, $field): void {
        if (is_array($rule)) {
            $params = array_slice($rule, 1);
            $this->_validator->rule($rule[0], [$field], ...array_values($params));
        } else {
            $this->_validator->rule($rule, [$field]);
        }
    }
}
