<?php

declare(strict_types=1);

namespace Api\Support\Facades;

use RuntimeException;
use Wildix\Core\Container;

/**
 * Class Facade
 */
abstract class Facade {

    /**
     * The resolved object instances.
     *
     * @var array
     */
    protected static $_resolvedInstance;

    /**
     * Get the root object behind the facade.
     *
     * @return mixed
     */
    public static function getFacadeRoot() {
        return static::resolveFacadeInstance(static::getFacadeAccessor());
    }

    /**
     * Get the registered name of the component.
     *
     * @return string|void Name of the component.
     *
     * @throws RuntimeException Error.
     */
    protected static function getFacadeAccessor(): string {
        throw new RuntimeException('Facade does not implement getFacadeAccessor method.');
    }

    /**
     * Resolve the facade root instance from the container.
     *
     * @param string|object $name Resolved facade name.
     *
     * @return mixed
     */
    protected static function resolveFacadeInstance($name) {
        if (is_object($name)) {
            return $name;
        }

        if (isset(static::$_resolvedInstance[$name])) {
            return static::$_resolvedInstance[$name];
        }
        $instance = Container::get($name);

        static::$_resolvedInstance[$name] = $instance;

        return static::$_resolvedInstance[$name];
    }

    /**
     * Clear a resolved facade instance.
     *
     * @param string $name Resolved facade.
     *
     * @return void
     */
    public static function clearResolvedInstance(string $name) {
        unset(static::$_resolvedInstance[$name]);
    }

    /**
     * Clear all of the resolved instances.
     *
     * @return void
     */
    public static function clearResolvedInstances() {
        static::$_resolvedInstance = [];
    }

    /**
     * Handle dynamic, static calls to the object.
     *
     * @param string $method Method.
     * @param array  $args   Arguments.
     *
     * @return mixed
     *
     * @throws RuntimeException Error.
     */
    public static function __callStatic(string $method, array $args) {
        $instance = static::getFacadeRoot();

        if (! $instance) {
            throw new RuntimeException('A facade root has not been set.');
        }

        return $instance->$method(...$args);
    }
}
