<?php

declare(strict_types=1);

namespace Api\Trunks\Factory;

use Api\Device\Interfaces\DeviceRepositoryInterface;
use Api\Translation\Interfaces\TrunkInterface;
use Api\Trunks\Pstn\Objects\TrunkFxoStatus;
use Api\Trunks\Pstn\Objects\TrunkGsmStatus;
use Api\Trunks\Pstn\Objects\TrunkIsdnStatus;
use Api\Trunks\Sip\Interfaces\TrunkStatusInterface;
use Wildix\Devices\Vendors;
use Wildix\SipProxy;

/**
 * Class TrunkStatusFactory
 */
class TrunkStatusFactory {

    const PORT_STATUS_UNREGISTERED = [
        'title' => 'Event not received',
        'state' => 'unregistered'
    ];
    const PORT_STATUS_ON           = [
        'title' => 'active',
        'state' => 'on'
    ];
    const PORT_STATUS_OFF          = [
        'title' => 'Inactive',
        'state' => 'off'
    ];

    public static function make(TrunkInterface $trunk, SipProxy $sipProxy, DeviceRepositoryInterface $deviceRepository): TrunkStatusInterface {
        if ($trunk->isGsm()) {
            return self::makeGsm($trunk, $sipProxy, $deviceRepository);
        } else if ($trunk->isFxo()) {
            return self::makeFxo($trunk, $sipProxy);
        } else {
            return self::makeIsdn($trunk, $sipProxy, $deviceRepository);
        }
    }

    private static function makeGsm(TrunkInterface $trunk, SipProxy $sipProxy, DeviceRepositoryInterface $deviceRepository): TrunkStatusInterface {
        $userName = $trunk->getUsername() . '_0_0';
        $status   = new TrunkGsmStatus();
        $incoming = $sipProxy->getTrunksRegistrations();
        if (isset($incoming[$userName])) {
            $device = self::getDevice($trunk, $deviceRepository);
            if (!empty($device) && self::isWgsm($device)) {
                $title       = $incoming[$userName]['registrations'][0]['received'];
                if (self::isDeviceSupportSignalLevel($device)) {
                    $signalLevel = explode(' ', $incoming[$userName]['registrations'][0]['useragent']);
                    $signalLevel = (int)$signalLevel[4];
                    $status->setSignalLevel($signalLevel);
                    $status->setSignalLevelSupport(true);
                }
                $status->setTitle($title);
                $status->setWgsm(true);
            }
            $status->setStatus(true);
        }

        return $status;
    }

    private static function getDevice(TrunkInterface $trunk, DeviceRepositoryInterface $deviceRepository): array {
        $device = [];
        $trunkDeviceMap = $deviceRepository->getTrunkIdDeviceMap([$trunk->getId()]);
        if (!empty($trunkDeviceMap[$trunk->getId()]['device'])) {
            $device = json_decode($trunkDeviceMap[$trunk->getId()]['device'], true);
        }
        return $device;
    }

    private static function isWgsm(array $device): bool {
        return $device['vendor'] === Vendors::WILDIX;
    }

    private static function isDeviceSupportSignalLevel(array $device): bool {
        return $device['model'] !== Vendors::DAYSAVER;
    }

    private static function makeFxo(TrunkInterface $trunk, SipProxy $sipProxy): TrunkStatusInterface {
        $userName = $trunk->getUsername() . '_1';
        $status   = new TrunkFxoStatus();
        $incoming = $sipProxy->getTrunksRegistrations();
        if (isset($incoming[$userName]['registrations'][0])) {
            $registration = $incoming[$userName]['registrations'][0];
            $status->setStatus((bool)$registration['online']);
            $status->setTitle($registration['received'] ?? '');
        }

        return $status;
    }

    private static function makeIsdn(TrunkInterface $trunk, SipProxy $sipProxy, DeviceRepositoryInterface $deviceRepository): TrunkStatusInterface {
        $result = self::makeIsdnStatus($trunk, $sipProxy);
        $trunk->setStatus($result);
        self::makeIsdnPortsStatus($trunk, $deviceRepository);
        return $result;
    }

    private static function makeIsdnPortsStatus(TrunkInterface $trunk, DeviceRepositoryInterface $deviceRepository) {
        $mac          = $deviceRepository->getMacByTrunkId($trunk->getId());
        $portsMonitor = self::getPortStatus();

        $trunk->getSettingsMgw()->getPorts()->transform(
            function ($port) use ($portsMonitor, $mac, $trunk) {
                $status = [
                    'physical' => self::PORT_STATUS_UNREGISTERED,
                    'logic' => self::PORT_STATUS_UNREGISTERED
                ];

                if ($trunk->getStatus()->getIncoming()['active'] && isset($portsMonitor[$mac][$port['port']])) {
                    $monitorStatus      = $portsMonitor[$mac][$port['port']];
                    $status['physical'] = $monitorStatus['physical'] === 'up' ? self::PORT_STATUS_ON : self::PORT_STATUS_OFF;
                    $status['logic']    = $monitorStatus['signaling'] === 'up' ? self::PORT_STATUS_ON : self::PORT_STATUS_OFF;
                }

                $port['status'] = $status;

                return $port;
            }
        );
    }

    private static function makeIsdnStatus(TrunkInterface $trunk, SipProxy $sipProxy): TrunkStatusInterface {
        $userName = $trunk->getUsername() . '_0_0';
        $status   = new TrunkIsdnStatus();
        $incoming = $sipProxy->getTrunksRegistrations();
        if (isset($incoming[$userName])) {
            $status->setIncoming(
                [
                    'active' => true,
                    'title' => $incoming[$userName]['registrations'][0]['received'],
                ]
            );
        }

        return $status;
    }

    private static function getPortStatus(): array {
        $monitor = [];

        foreach (glob('/dev/shm/wmsdb.d/isdn-*') as $file) {
            if (preg_match('/^.*?isdn-(.*)$/', $file, $matches)) {
                if (!empty($matches[1])) {
                    $monitor[$matches[1]] = [];

                    $isdn = parse_ini_file($file, true, INI_SCANNER_RAW);
                    foreach ($isdn as $key => $value) {
                        if (preg_match('/([^\d]+)(\d+)$/', $key, $slotMatches)) {
                            $monitor[$matches[1]][$slotMatches[2]] = $value;
                        }
                    }
                }
            }
        }

        return $monitor;
    }
}
