<?php

declare(strict_types=1);

namespace Api\Trunks\Fxo\Service;

use Api\Support\BaseCollectObject;
use Api\Trunks\Factory\TrunkStatusFactory;
use Api\Trunks\Fxo\Factory\FxoTrunkFactory;
use Api\Trunks\Fxo\Interfaces\FxoTrunkServiceInterface;
use Api\Trunks\Fxo\Objects\FxoTrunk;
use Api\Trunks\Fxo\Objects\FxoTrunks;
use Api\Trunks\Pstn\Objects\Trunk;
use Api\Trunks\Service\TrunkService;
use PBX;
use Wildix\Devices\Devices;
use Wildix\Devices\Vendors;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\Validator;

/**
 * Class FxoService
 */
class FxoTrunkService extends TrunkService implements FxoTrunkServiceInterface {

    public function getItems(): FxoTrunks {
        return $this->_modifyTech($this->getTrunkRepository()->getFxoItems());
    }

    public function remove(int $id): bool {
        $existsTrunk = $this->getById($id);
        $this->canDelete($existsTrunk);

        return $this->getTrunkRepository()->removePstn($id);
    }

    public function update(int $id, array $requestData): bool {
        $requestData['id'] = $id;
        $this->validateData($requestData);
        $existsTrunk                 = $this->getById($id);
        $requestData['settings_pbx'] = $existsTrunk->getSettingsPbx()->toArray();
        $trunk = FxoTrunkFactory::makeFromRequest($requestData);
        $trunk->setId($id);
        $trunk->setUsername($existsTrunk->getUsername());
        $trunk->getSettingsPbx()->setUsername($existsTrunk->getSettingsPbx()->getUsername());
        $result = $this->getTrunkRepository()->updatePstn($id, $trunk);
        $this->updateConfigs();
        $this->rebootDeviceIfNeed($trunk, $existsTrunk);

        return $result;
    }

    private function validateData(array $data): void {
        $validator = new Validator($data);

        $validator->rule('integer', 'id');
        $validator->rule('string', 'username');
        $validator->rule('string', 'technology');
        $validator->rule('string', 'address');
// ?????? ip or not
        $validator->rule('string', 'model');
        $validator->rule('string', 'settingsMgw.toneZone');
        $validator->rule('countryCode', 'settingsMgw.countryCode')->message(Error::CountryCode);
        $validator->rule('integer', 'settingsMgw.impedance');
        $validator->rule('integer', 'settingsMgw.fskCidField');
        $validator->rule('integer', 'settingsMgw.fskCidField');
        $validator->rule('boolean', 'settingsMgw.dcImpedance');
        $validator->rule('boolean', 'settingsMgw.busyToneDetected');
        $validator->rule('boolean', 'settingsMgw.polarityNormalDetected');
        $validator->rule('boolean', 'settingsMgw.currentDetected');
        $validator->rule('boolean', 'settingsMgw.silenceDetectionEnabled');

        foreach ($data['settingsMgw']['ports'] as $key => $port) {
            $validator->rule('integer', ['settingsMgw.ports.' . $key . '.port']);
            $validator->rule('integer', ['settingsMgw.ports.' . $key . '.priceList']);
            $validator->rule('integer', ['settingsMgw.ports.' . $key . '.cidType']);
            $validator->rule('integer', ['settingsMgw.ports.' . $key . '.flashTime']);
            $validator->rule('min', ['settingsMgw.ports.' . $key . '.answerTime'], 1);
            $validator->rule('integer', ['settingsMgw.ports.' . $key . '.faxType']);
            $validator->rule('integer', ['settingsMgw.ports.' . $key . '.faxSpeed']);
            $validator->rule('integer', ['settingsMgw.ports.' . $key . '.codecInputGain']);
            $validator->rule('integer', ['settingsMgw.ports.' . $key . '.codecOutputGain']);
            $validator->rule('boolean', ['settingsMgw.ports.' . $key . '.preversalDetection']);
            $validator->rule('string', ['settingsMgw.ports.' . $key . '.context']);

            $validator->rule('numeric', ['settingsMgw.ports.' . $key . '.hotlineNumber']);
            $validator->rule('required', ['settingsMgw.ports.' . $key . '.hotlineNumber']);
            $validator->rule('boolean', ['settingsMgw.ports.' . $key . '.pstnRingCount']);
            $validator->rule('between', ['settingsMgw.ports.' . $key . '.destinationCode'], [1, 9999])->reason('Numeric between 1-9999');
        }
        $validator->rule('between', ['settingsMgw.silenceDetection'], [1, 120])->reason('Numeric between 1-120');
        if ($data['settingsMgw']['currentDetected'] === true) {
            $validator->rule('between', 'settingsMgw.currentDisconnectThreshold', [100, 2000])->reason('Please insert numeric between 100-2000');
        } else {
            $validator->rule('integer', 'settingsMgw.currentDisconnectThreshold');
        }

        if (!$validator->validate()) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $validator->errorsTree(), 400);
        }
    }

    private function _modifyTech(
        FxoTrunks $trunks
    ): FxoTrunks {
        $ids = [];
        $trunks->map(
            function (FxoTrunk $trunk) use (&$ids) {
                $ids[] = $trunk->getId();
            }
        );
        $trunkDeviceMap = $this->getDeviceRepository()->getTrunkIdDeviceMap($ids);
        $trunks->map(
            function (FxoTrunk $trunk) use ($trunkDeviceMap) {
                $trunk->setAddress('');
                if (!empty($trunkDeviceMap[$trunk->getId()]['device'])) {
                    $device = json_decode($trunkDeviceMap[$trunk->getId()]['device'], true);
                    if (!empty($device['vendor']) && $device['vendor'] === Vendors::WILDIX) {
                        $trunk->setAddress((string)$device['address']);
                        $trunk->setModel((string)$device['model']);
                    }
                }
                $status = TrunkStatusFactory::make($trunk, $this->getSipProxy(), $this->getDeviceRepository());
                $trunk->setStatus($status);
            }
        );

        return $trunks;
    }

    private function getById(
        int $id
    ): FxoTrunk {
        return $this->getTrunkRepository()->findOrFailFxo($id);
    }

    private function canDelete(
        Trunk $trunk
    ): void {
        $findLinkInDialplan = $this->getTrunkRepository()->findLinkInDialplan($trunk->getUsername());
        if ($findLinkInDialplan) {
            $find = new BaseCollectObject($findLinkInDialplan);
            $find = $find->keyBy('rulename');
            throw new InvalidDataException(
                sprintf(
                    'You can\'t remove %s. It\'s linked to %s',
                    $trunk->getTitle(),
                    $find->keys()->implode(', ')
                ),
                [],
                400
            );
        }
    }

    private function rebootDeviceIfNeed(FxoTrunk $trunk, ?FxoTrunk $existsTrunk): void {
        if ($trunk->getUsername()) {
            $fxoReboot = $this->isNeedRebootFxo($trunk, $existsTrunk);

            PBX::getInstance()->sip_notify_check_sync([$trunk->getUsername() . '_0'], $fxoReboot);
            $this->getLogger()->info(
                __LINE__,
                'Send check-sync ' . ($fxoReboot ? '(reboot) ' : '') . 'to ' . $trunk->getUsername() . '_0'
            );
        } else {
            $this->getLogger()->info(__LINE__, 'Failed settings[\'username\'] for trunk check-sync');
        }
    }

    private function isNeedRebootFxo(FxoTrunk $trunk, ?FxoTrunk $existsTrunk): bool {
        $deviceMac = $this->getDeviceRepository()->getMacByTrunkId($trunk->getId());
        if (isset($existsTrunk)
            && ($existsTrunk->getSettingsMgw()->getToneZone() !== $trunk->getSettingsMgw()->getToneZone()
            || $existsTrunk->getSettingsMgw()->getImpedance() !== $trunk->getSettingsMgw()->getImpedance())
        ) {
            return true;
        }

        if ($deviceMac) {
            $device = $this->getDevicesInstance()->getDeviceByMac($deviceMac);
            if (Vendors::isNeedsToReboot($device->getModel())) {
                return true;
            }
        }

        return false;
    }

    private function getDevicesInstance(): Devices {
        return Devices::getInstance();
    }
}
