<?php

namespace Api\Trunks\PriceList\Services;

use Api\Trunks\PriceList\Enum\PriceListColumnsEnum;
use Api\Trunks\PriceList\Interfaces\PriceListCardRepositoryInterface;
use Api\Trunks\PriceList\Interfaces\PriceListRepositoryInterface;
use Api\Trunks\PriceList\Interfaces\PriceListServiceInterface;
use Api\Trunks\PriceList\Objects\Price;
use Api\Trunks\PriceList\Objects\Prices;
use Exception;
use phpDocumentor\Reflection\Types\Integer;
use Slim\Psr7\UploadedFile;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\Validator;

/**
 * Class PriceListService
 */
class PriceListService implements PriceListServiceInterface {

    /**
     * Card Repository.
     *
     * @var PriceListCardRepositoryInterface
     */
    private $_cardRepository;

    /**
     * Prices Repository
     *
     * @var PriceListRepositoryInterface
     */
    private $_repository;

    /**
     * PriceListService constructor.
     *
     * @param PriceListCardRepositoryInterface $_cardRepository     Card interface.
     * @param PriceListRepositoryInterface     $priceListRepository Prices interface.
     */
    public function __construct(PriceListCardRepositoryInterface $_cardRepository, PriceListRepositoryInterface $priceListRepository) {
        $this->_cardRepository = $_cardRepository;
        $this->_repository     = $priceListRepository;
    }

    public function getItems(): Prices {
        return $this->_cardRepository->getAll()->map(
            function (Price $price) {
                return $price->setItems($this->_repository->get($price->getId()));
            }
        );
    }

    public function getItem(int $id): Price {
        return $this->_cardRepository->findOrFail($id)->setItems($this->_repository->get($id));
    }

    /**
     * Create new price list
     *
     * @param array $params Params for created.
     *
     * @return Price
     *
     * @throws InvalidDataException InvalidDataException.
     */
    public function create(array $params): Price {
        $this->validateData($params);
        $id = $this->_cardRepository->create((string)$params['name']);
        if ($id && $params['upload'] instanceof UploadedFile && $params['upload']->getFilePath()) {
            $this->import($id, $params);
        }

        return $this->getItem($id);
    }

    /**
     * Update price list
     *
     * @param integer $id     Id price.
     * @param array   $params Params for updated.
     *
     * @return Price
     *
     * @throws InvalidDataException InvalidDataException.
     */
    public function update(int $id, array $params): Price {
        $this->_cardRepository->findOrFail($id);
        $params['id'] = $id;
        $this->validateData($params);
        $this->_cardRepository->update($id, $params);
        if ($params['upload'] instanceof UploadedFile && $params['upload']->getFilePath()) {
            $this->import($id, $params);
        }
        return $this->getItem($id);
    }

    public function remove($id): bool {
        $this->_cardRepository->findOrFail($id);
        return $this->_repository->delete($id) && $this->_cardRepository->delete($id);
    }

    private function importPrice(string $filename, $id): bool {
        if (!is_file($filename)) {
            throw new InvalidDataException('Wrong import file');
        }
        $this->_repository->delete($id);
        exec('printf ".separator \',\'\n.mode csv\n.import ' . escapeshellarg($filename) . ' rate" | sqlite3 ' . $this->_cardRepository->getDbPatch(), $out);
        return !$this->_repository->get($id)->isEmpty();
    }

    private function import(int $id, array $params): bool {
        if (is_uploaded_file($params['upload']->getFilePath())) {
            $tmpPriceFileName = '/tmp/session/import_pricelist' . $id . '.csv~';
            try {
                $params['upload']->moveTo($tmpPriceFileName);
                // remove first line and add price list id
                exec('sed -i -e \'/^[[:cntrl:]]/d;1d; s/^/' . escapeshellarg($id) . ',/\' ' . $tmpPriceFileName);
                $result = $this->importPrice($tmpPriceFileName, $id);
            } catch (Exception $e) {
                $result = false;
            }
        }

        return $result;
    }

    private function isValidFile(string $filename): bool {
        $result['badlines'] = [];
        $handle             = fopen($filename, 'r');
        if ($handle !== false) {
            $previewCount = 0;

            // NOTE: brutal hack for Windows EOL, as enclosure HEX code (0x02) of not printable character.
            // NOTE: replaced with UTF symbol. Be carefull while save this file!
            $header = fgetcsv($handle, 1024, ',');
            if ($header !== false && count($header) == 10) {
                while (($data = fgetcsv($handle, 1024, ',')) !== false) {
                    if (count($data) !== 10) {
                        $result['badlines'][] = join(',', $data);
                    } else {
                        $tmpPreviewDataFlag = true;
                        $columns            = PriceListColumnsEnum::getEnumValues();
                        foreach ($data as $fieldId => $fieldValue) {
                            if (!$this->checkField($columns[$fieldId], $fieldValue)) {
                                $result['badlines'][] = htmlspecialchars(join(',', $data));
                                $tmpPreviewDataFlag   = false;
                                break;
                            }
                        }
                        if ($tmpPreviewDataFlag && $previewCount < 3) {
                            $previewCount++;
                        }
                    }
                    usleep(1);
                }
            } else {
                // bad header csv
                $result['badlines'][] = join(',', $header);
            }
            fclose($handle);
        }

        return empty($result['badlines']);
    }

    /**
     * Check price list field with regular expression
     * NOTE: regular expressions defined inside the function
     *
     * @param string $field Field.
     * @param string $value Value.
     *
     * @return boolean Result.
     */
    private function checkField(string $field, string $value): bool {
        $fields = [
            'dialprefix' => '/^[\*\#\+0-9]{1,10}\z/',
            'destination' => '/^[^<>]*\z/',
            'rate' => '/^([0\.]?[0-9]{1,2})|([0-9]{1,3}\.[0-9]{1,2})\z/',
            'min_duration' => '/^[0-9]{1,3}\z/',
            'billing_block' => '/^[1-9][0-9]{0,3}\z/',
            'connect_charge' => '/^([0\.]?[0-9]{1,2})|([0-9]{1,3}\.[0-9]{1,2})\z/',
            'start_day' => '/^[1-7]\z/',
            'stop_day' => '/^[1-7]\z/',
            'start_time' => '/^[0-9]{1,5}\z/',
            'stop_time' => '/^[0-9]{1,5}\z/'
        ];
        if (array_key_exists($field, $fields)) {
            return !(!preg_match($fields[$field], $value));
        }
        return false;
    }

    public function export(int $id) {
        $data = $this->_repository->get($id);
        $cols = strtolower(implode(',', PriceListColumnsEnum::getEnumValues()));
        echo $cols . PHP_EOL;
        foreach ($data->toArray() as $row) {
            echo strtolower(implode(',', $row)) . PHP_EOL;
        }
    }

    /**
     * Validate data
     *
     * @param array $params Params request.
     *
     * @throws InvalidDataException InvalidDataException.
     *
     * @return void
     */
    private function validateData(array $params): void {
        $errors = [];
        if (isset($params['upload']) && $params['upload']->getFilePath() && !$this->isValidFile($params['upload']->getFilePath())) {
            $errors['upload'] = [
                'code' => 400,
                'reason' => 'Not valid price list'
            ];
        }
        $form = new Validator($params);
        $form->rule('regex', ['name'], '/^[a-z0-9]{1,}[a-z0-9\-_ ]*$/i')->reason(Error::getMessage(Error::IncorrectFormat));
        $form->rule('IsFile', ['upload']);
        $form->rule(
            'NotIn',
            ['name'],
            $this->_cardRepository->getAllWithout($params['id'] ?? 0)->pluck('name')
        )->reason(Error::getMessage(Error::MustBeUnique));
        $form->rule('required', ['name']);
        $form->validate();
        $errors = array_merge($errors, $form->errors());
        if (!empty($errors)) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $errors);
        }
    }
}
