<?php

declare(strict_types=1);

namespace Api\Trunks\Pstn\Services;

use Api\Trunks\Pstn\Enum\TrunkTechnologyEnum;
use Api\Trunks\Pstn\Factory\TrunkRequestFactory;
use Api\Trunks\Pstn\Interfaces\PstnTrunkServiceInterface;
use Api\Trunks\Pstn\Objects\Trunk;
use Api\Trunks\Pstn\Objects\Trunks;
use Api\Trunks\Service\TrunkService;
use PBX;
use Wildix\Devices\Devices;
use Wildix\Devices\Vendors;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\Validator;

/**
 * Class TrunkService
 */
class PstnTrunkService extends TrunkService implements PstnTrunkServiceInterface {

    public function create(array $data): int {
        $this->_validate($data);

        $trunk = TrunkRequestFactory::make($data);

        $trunk->setUsername($this->getNewTrunkName($trunk->getTechnology()));
        if (!$trunk->getSettingsPbx()->getUsername()) {
            $trunk->getSettingsPbx()->setUsername($trunk->getUsername());
        }
        if ($this->isNeedChangeSmsCenters($trunk, null)) {
            $this->changeSmsCenters($trunk);
        }

        $id = $this->getTrunkRepository()->createPstn($trunk);
        $trunk->setId($id);
        $this->updateConfigs();
        $this->rebootDeviceIfNeed($trunk, null);

        return $id;
    }

    public function remove(int $id): bool {
        $this->getTrunkRepository()->findOrFailPstn($id);

        $this->getTrunkRepository()->removePstn($id);
        $this->getDeviceRepository()->deleteTrunk($id);

        $this->getPeerService()->delete($id);

        $this->getPeerService()->reloadPeers();
        $this->getPeerService()->storeIp();

        $this->updateConfigs();

        return true;
    }

    public function update(int $id, array $data): bool {
        $existsTrunk = $this->getTrunkRepository()->findOrFailPstn($id);
        $data['id']  = $id;
        $this->_validate($data);

        $trunk = TrunkRequestFactory::make($data);

        $trunk->setId($id);
        $trunk->setUsername($existsTrunk->getUsername());
        $trunk->getSettingsPbx()->setUsername($existsTrunk->getSettingsPbx()->getUsername());

        if ($this->isNeedChangeSmsCenters($trunk, $existsTrunk)) {
            $this->changeSmsCenters($trunk);
        }

        $this->getTrunkRepository()->updatePstn($id, $trunk);

        $this->updateConfigs();

        $this->rebootDeviceIfNeed($trunk, $existsTrunk);

        return true;
    }

    private function _validate(array $data): void {
        $validator = new Validator($data);

        $validator->rule('array', ['settingsPbx', 'settingsMgw']);
        $validator->rule('integer', 'id');
        $validator->rule('in', ['technology'], [TrunkTechnologyEnum::BRI, TrunkTechnologyEnum::PRI, TrunkTechnologyEnum::FXO, TrunkTechnologyEnum::GSM]);
        $validator->rule('string', 'settingsPbx.username');
        $validator->rule('required', ['settingsPbx', 'settingsMgw', 'name', 'technology', 'settingsMgw.ports']);

        if (!is_null($data['settingsMgw']['countryCode'])) {
            $validator->rule('countryCode', 'settingsMgw.countryCode')->message(Error::CountryCode);
        }
        if (strtolower($data['technology']) === TrunkTechnologyEnum::FXO) {
            $validator->rule('strictBoolean', 'settingsMgw.silenceDetectionEnabled')->reason('Parameter must be boolean');
        }

        $validator->rule('string', 'settingsMgw.dtmfDebounce');
        $validator->rule('string', 'settingsMgw.lineType');
        $validator->rule('string', 'settingsMgw.dialTone');
        $validator->rule('string', 'settingsMgw.toneZone');
        $validator->rule('StrictBoolean', 'settingsMgw.preserveOrigCid');
        $validator->rule('StrictInteger', 'settingsMgw.ports.callingNameMaxLength');

        $validator->rule('strictInteger', 'settingsMgw.ports.callingNameMaxLength');
        foreach ($data['settingsMgw']['ports'] as $key => $port) {
            if ($port['destinationCode']) {
                $validator->rule('required', ['settingsMgw.ports.' . $key . '.destinationCode']);
                $validator->rule('numeric', ['settingsMgw.ports.' . $key . '.destinationCode']);
                $validator->rule('between', ['settingsMgw.ports.' . $key . '.destinationCode'], [1, 9999])->reason('Numeric between 1-9999');
            }
            if ($data['technology'] === TrunkTechnologyEnum::GSM) {
                $validator->rule('required', ['settingsMgw.ports.' . $key . '.number']);
            }

            $validator->rule('min', 'settingsMgw.ports.' . $key . '.callingNameMaxLength', 0);
            $validator->rule('StrictInteger', ['settingsMgw.ports.' . $key . '.callingNameMaxLength']);

            $validator->rule('email', ['settingsMgw.ports.' . $key . '.email']);

            $validator->rule('min', 'settingsMgw.ports.' . $key . '.maximumFacilityWaitingDelay', 0);
            $validator->rule('StrictInteger', ['settingsMgw.ports.' . $key . '.maximumFacilityWaitingDelay']);
        }

        if (!$validator->validate()) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $validator->errorsTree());
        }
    }

    private function getNewTrunkName(string $technology): string {
        $names = [];

        foreach ($this->getTrunkRepository()->getTrunkNames($technology) as $name) {
            $names[] = $name['username'];
        }

        sort($names);
        foreach ($names as $key => $name) {
            if (strtolower($technology . $key) != $name) {
                return strtolower($technology . $key);
            }
        }

        return strtolower($technology . count($names));
    }

    public function getItems(): Trunks {
        $trunks = $this->getTrunkRepository()->getPstnItems();
        $this->setTrunkStatuses($trunks);

        return $this->_modifyTech($trunks);
    }

    private function isNeedChangeSmsCenters(Trunk $trunk, ?Trunk $existsTrunk): bool {
        return (!isset($existsTrunk)
                && isset($trunk->getSettingsMgw()->getPorts()->prepareForDb()['port_smscn']))
            || (isset($trunk->getSettingsMgw()->getPorts()->prepareForDb()['port_smscn'])
                && isset($existsTrunk->getSettingsMgw()->getPorts()->prepareForDb()['port_smscn'])
                && array_diff(
                    $trunk->getSettingsMgw()->getPorts()->prepareForDb()['port_smscn'],
                    $existsTrunk->getSettingsMgw()->getPorts()->prepareForDb()['port_smscn']
                )
            );
    }

    private function changeSmsCenters(Trunk $trunk): void {
        foreach ($trunk->getSettingsMgw()->getPorts()->prepareForDb()['port_smscn'] as $key => $value) {
            if (!empty($value)) {
                touch('/tmp/' . $trunk->getUsername() . '_' . $key . '_0.smscn');
            }
        }
    }

    private function rebootDeviceIfNeed(Trunk $trunk, ?Trunk $existsTrunk): void {
        if ($trunk->getUsername()) {
            if ($trunk->isFxo()) {
                $fxoReboot = $this->isNeedRebootFxo($trunk, $existsTrunk);
                $this->getPbxInstance()->sip_notify_check_sync([$trunk->getUsername() . '_0'], $fxoReboot);
                $this->getLogger()->info(
                    __LINE__,
                    'Send check-sync ' . ($fxoReboot ? '(reboot) ' : '') . 'to ' . $trunk->getUsername() . '_0'
                );
            } else {
                if ($trunk->isGsm()) {
                    $this->getPbxInstance()->sip_notify_check_sync(
                        [$trunk->getUsername() . '_0_0'],
                        $this->isNeedRebootGsm($trunk, $existsTrunk)
                    );
                    $this->getLogger()->info(__LINE__, 'Send check-sync to ' . $trunk->getUsername() . '_0_0');
                }
            }
        } else {
            $this->getLogger()->info(__LINE__, 'Failed settings[\'username\'] for trunk check-sync');
        }
    }

    private function isNeedRebootFxo(Trunk $trunk, ?Trunk $existsTrunk): bool {
        $deviceMac = $this->getDeviceRepository()->getMacByTrunkId($trunk->getId());

        if (isset($existsTrunk)
            && ($existsTrunk->getSettingsMgw()->getToneZone() !== $trunk->getSettingsMgw()->getToneZone()
            || $existsTrunk->getSettingsMgw()->getImpedance() !== $trunk->getSettingsMgw()->getImpedance())
        ) {
            return true;
        }

        if ($deviceMac) {
            $device = $this->getDevicesInstance()->getDeviceByMac($deviceMac);
            if (Vendors::isNeedsToReboot($device->getModel())) {
                return true;
            }
        }

        return false;
    }

    private function getDevicesInstance(): Devices {
        return Devices::getInstance();
    }

    private function getPbxInstance(): PBX {
        return PBX::getInstance();
    }

    private function isNeedRebootGsm(Trunk $trunk, ?Trunk $existsTrunk): bool {
        if (!isset($existsTrunk) || array_diff($trunk->toArray(), $existsTrunk->toArray())) {
            return true;
        }

        return false;
    }

    private function _modifyTech(Trunks $trunks): Trunks {
        $trunkDeviceMap = $this->getDeviceRepository()->getTrunkIdDeviceMap(array_keys($trunks->keyBy('id')->toArray()));
        $trunks->map(
            function (Trunk $trunk) use ($trunkDeviceMap) {
                $tech = $trunk->getTechnology();
                if ($trunk->isBri() || $trunk->isPri()) {
                    $trunk->setTechnology(TrunkTechnologyEnum::OTHER);
                    if (!empty($trunkDeviceMap[$trunk->getId()]['device'])) {
                        $device = json_decode($trunkDeviceMap[$trunk->getId()]['device'], true);
                        if (!empty($device['vendor']) && $device['vendor'] === Vendors::WILDIX) {
                            $trunk->setTechnology($tech);
                            $trunk->setAddress((string)$device['address']);
                        }
                    }
                } else if ($trunk->isGsm() && !empty($trunkDeviceMap[$trunk->getId()]['device'])) {
                    $device = json_decode($trunkDeviceMap[$trunk->getId()]['device'], true);
                    if (!empty($device['vendor'])
                        && ($device['vendor'] === Vendors::PORTECH || $device['vendor:additional'] == 'DWG')
                    ) {
                        $trunk->setAddress((string)$device['address']);
                    }
                } else if ($trunk->isFxo() && !empty($trunkDeviceMap[$trunk->getId()]['device'])) {
                    $device = json_decode($trunkDeviceMap[$trunk->getId()]['device'], true);
                    if (!empty($device['vendor']) && $device['vendor'] === Vendors::WILDIX) {
                        $trunk->setAddress((string)$device['address']);
                    }
                }
            }
        );
        return $trunks->filter(
            function (Trunk $item) {
                return !in_array($item->getTechnology(), [TrunkTechnologyEnum::OTHER, TrunkTechnologyEnum::FXO]);
            }
        );
    }
}
