<?php

declare(strict_types=1);

namespace Api\Trunks\Repository;

use Api\Translation\Interfaces\TrunkInterface;
use Api\Translation\Objects\TranslationInterface;
use Api\Trunks\Fxo\Enum\FxoTrunkColumnsEnum;
use Api\Trunks\Fxo\Factory\FxoTrunkFactory;
use Api\Trunks\Fxo\Objects\FxoTrunk;
use Api\Trunks\Fxo\Objects\FxoTrunks;
use Api\Trunks\Interfaces\TrunkRepositoryInterface;
use Api\Trunks\Pstn\Enum\SettingsMgwFieldsEnum;
use Api\Trunks\Pstn\Enum\SettingsPbxFieldsEnum;
use Api\Trunks\Pstn\Enum\TrunkColumnsEnum;
use Api\Trunks\Pstn\Factory\TrunkFactory;
use Api\Trunks\Pstn\Objects\Trunk as PstnTrunk;
use Api\Trunks\Pstn\Objects\Trunks as PstnTrunks;
use Api\Trunks\Sip\Factory\SipTrunkFactory;
use Api\Trunks\Sip\Objects\Trunk as SipTrunk;
use Api\Trunks\Sip\Objects\Trunks as SipTrunks;
use Wildix\Databases\Database;
use Wildix\Databases\WMSDB;
use Wildix\Dialplan\DialplanService;
use Wildix\Error;
use Wildix\InvalidDataException;

/**
 * Class TrunkRepository
 */
class TrunkRepository implements TrunkRepositoryInterface {

    const FXO_TECHNOLOGY_NAME = 'fxo';

    private $_translate;

    private $_db;

    public function __construct(TranslationInterface $translate, Database $db) {
        $this->_translate = $translate;
        $this->_db        = $db;
    }

    public function createPstn(PstnTrunk $trunk): int {
        $this->getDb()->insert(
            WMSDB::TABLE_PSTN_TRUNKS,
            [
                'username' => $trunk->getUsername(),
                'technology' => $trunk->getTechnology(),
                'ports' => $trunk->getPorts(),
                'settings_pbx' => json_encode($trunk->getSettingsPbx()->prepareForDb()),
                'settings_mgw' => json_encode($trunk->getSettingsMgw()->prepareForDb()),
            ]
        );

        return $this->getDb()->getLastInsertId();
    }

    /**
     * Get db
     *
     * @return Database
     */
    private function getDb(): Database {
        return $this->_db;
    }

    public function createSip(SipTrunk $trunk): int {
        $data = [
            'title' => $this->getDb()->realEscapeString($trunk->getTitle()),
            'username' => $this->getDb()->realEscapeString($trunk->getUsername()),
            'register' => $this->getDb()->realEscapeString($trunk->getRegister()),
            'settings_pbx' => $this->getDb()->realEscapeString($trunk->getSettingsPbx()->toJson()),
            'settings_wms' => $this->getDb()->realEscapeString($trunk->getSettingsWms()->toJson()),
        ];
        $this->getDb()->insert('trunks', $data);

        return $this->getDb()->getLastInsertId();
    }

    public function findOrFailPstn(int $id): PstnTrunk {
        $trunk = $this->getDb()->first(
            sprintf(
                'SELECT %s FROM %s WHERE `id`=%d',
                '*',
                WMSDB::TABLE_PSTN_TRUNKS,
                $id
            )
        );

        if (!$trunk) {
            throw new InvalidDataException(Error::NotFound, [], 404);
        }

        return TrunkFactory::make($trunk);
    }

    public function findOrFailSip(int $id): SipTrunk {
        $item = $this->getDb()->first(sprintf('SELECT * FROM %s where id= %s', WMSDB::TABLE_TRUNKS, $id));
        if (!$item) {
            throw new InvalidDataException(Error::NotFound, [], 404);
        }

        return SipTrunkFactory::make($item);
    }

    public function getPstnItems(): PstnTrunks {
        $trunks = $this->getDb()->getRows(
            sprintf(
                "SELECT * FROM %s ORDER BY username asc",
                WMSDB::TABLE_PSTN_TRUNKS
            ),
            TrunkColumnsEnum::ID
        );

        $result = new PstnTrunks();

        foreach ($trunks as $trunkData) {
            $result->push(TrunkFactory::make($trunkData));
        }

        return $result;
    }

    /**
     * Get all sip trunk
     *
     * @return SipTrunks
     * @throws \Exception Exception.
     */
    public function getSipItems(): SipTrunks {
        $items  = $this->getDb()->getRows(sprintf('SELECT * FROM %s order by username', WMSDB::TABLE_TRUNKS));
        $trunks = new SipTrunks();
        foreach ($items as $item) {
            $trunks->push(SipTrunkFactory::make($item));
        }

        return $trunks;
    }

    public function getFxoItems(): FxoTrunks {
        $trunks = $this->getDb()->getRows(
            sprintf(
                "SELECT * FROM %s WHERE `technology`='%s' ORDER BY username ASC",
                WMSDB::TABLE_PSTN_TRUNKS,
                self::FXO_TECHNOLOGY_NAME
            ),
            FxoTrunkColumnsEnum::ID
        );

        $result = new FxoTrunks();

        foreach ($trunks as $trunkData) {
            $result->push(FxoTrunkFactory::makeFromStorage($trunkData));
        }

        return $result;
    }

    public function getTrunkNames(string $technology): array {
        return $this->getDb()->getRows(
            sprintf('SELECT `username` FROM `%s` WHERE `technology`=\'%s\'', WMSDB::TABLE_PSTN_TRUNKS, $technology)
        );
    }

    public function isExistSip(array $settings): array {
        $isDinamyc = ($settings['settingsPbx']['host'] == 'dynamic');
        $sql       = "SELECT settings_pbx, username, title
                FROM trunks
                WHERE
                    (
                    title like '" . $this->getDb()->realEscapeString($settings['title']) . "' OR
                    " . ($isDinamyc ? "(username like '" . $this->getDb()->realEscapeString($settings['settingsPbx']['userName']) . "' AND settings_pbx like '%\"host\":\"dynamic\"%') OR" : '') . "
                    settings_pbx like '%\"peername\":\"" . $this->getDb()->realEscapeString($settings['settingsPbx']['peername']) . "\"%'
                    ) AND
                    id <> '" . $settings['id'] . "';";

        $items = $this->getDb()->first($sql);

        $errors = [];
        if ($items) {
            $settingsPbx = json_decode($items['settings_pbx'], true);

            if (strcasecmp((string)$items['title'], $settings['title']) == 0) {
                $errors['title']['reason'] = $this->getTranslate()->t('strings.AlreadyPresent');
                $errors['title']['code']   = 'AlreadyPresent';
            }

            if (!empty($settingsPbx['peername']) && strcasecmp($settingsPbx['peername'], $settings['settingsPbx']['peername']) == 0) {
                $errors['settingsPbx']['peername']['reason'] = $this->getTranslate()->t('strings.AlreadyPresent');
                $errors['settingsPbx']['peername']['code']   = 'AlreadyPresent';
            }

            if (strcasecmp($items['username'], $settings['settingsPbx']['userName'] ?? '') == 0) {
                $errors['settingsPbx']['userName']['reason'] = $this->getTranslate()->t('strings.AlreadyPresent');
                $errors['settingsPbx']['userName']['code']   = 'AlreadyPresent';
            }
        }

        return $errors;
    }

    /**
     * Get Translation
     *
     * @return TranslationInterface
     */
    public function getTranslate(): TranslationInterface {
        return $this->_translate;
    }

    public function removePstn(int $id): bool {
        return $this->getDb()->exec(
            sprintf(
                'DELETE FROM `%s` WHERE `id`=%d;',
                WMSDB::TABLE_PSTN_TRUNKS,
                $id
            )
        );
    }

    public function removeSip(int $id): bool {
        $query = sprintf('DELETE FROM %s WHERE id = %s', WMSDB::TABLE_TRUNKS, $id);

        return $this->getDb()->exec($query);
    }

    public function findLinkInDialplan(string $id): array {
        $sql = sprintf(
            "select dialplan_rulenames.id, dialplan_rulenames.rulename from dialplan_links
    left join dialplan_rulenames on dialplan_links.object_from_id = dialplan_rulenames.id
    where object_to_type = %s and object_to_id like '%s%%'",
            DialplanService::LINK_TRUNK,
            $id
        );

        return $this->getDb()->getRows($sql);
    }

    public function updatePstn(int $id, TrunkInterface $trunk): bool {
        $mgwSettings = $trunk->getSettingsMgw()->prepareForDb();
        $pbxSettings = $trunk->getSettingsPbx()->prepareForDb();
        if ($trunk->isBri() || $trunk->isPri()) {
            foreach (SettingsMgwFieldsEnum::$notForBRIPRI as $key) {
                unset($mgwSettings[$key]);
            }
            foreach (SettingsPbxFieldsEnum::$notForBRIPRI as $key) {
                unset($pbxSettings[$key]);
            }
        }
        if ($trunk->isGsm()) {
            foreach (SettingsMgwFieldsEnum::$notForGSM as $key) {
                unset($mgwSettings[$key]);
            }
            foreach (SettingsPbxFieldsEnum::$notForGSM as $key) {
                unset($pbxSettings[$key]);
            }
            $mgwSettings['busy_response']      = '503';
            $mgwSettings['dtmf_debounce']      = '80';
            $mgwSettings['port_sms_read']      = ['300'];
            $mgwSettings['port_initatcmd']     = ['WGSM'];
            $mgwSettings['sms_delivery_tries'] = '5';
        }

        if ($trunk->isBri()) {
            foreach (SettingsMgwFieldsEnum::$notForBRI as $key) {
                unset($mgwSettings[$key]);
            }
        }

        return $this->getDb()->exec(
            sprintf(
                "UPDATE `%s` SET `username`='%s', `technology`='%s', `ports`=%d, `settings_pbx`='%s', `settings_mgw`='%s' WHERE `id`=%d;",
                WMSDB::TABLE_PSTN_TRUNKS,
                $this->getDb()->realEscapeString($trunk->getUsername()),
                $this->getDb()->realEscapeString($trunk->getTechnology()),
                $trunk->getPorts(),
                $this->getDb()->realEscapeString(json_encode($pbxSettings)),
                $this->getDb()->realEscapeString(json_encode($mgwSettings)),
                $id
            )
        );
    }

    public function updateSip(int $id, SipTrunk $settings): bool {
        $query = sprintf(
            "UPDATE %s SET title='%s', username='%s', register='%s', settings_pbx='%s', settings_wms='%s' WHERE id=%s",
            WMSDB::TABLE_TRUNKS,
            $this->getDb()->realEscapeString($settings->getTitle()),
            $this->getDb()->realEscapeString($settings->getUsername()),
            $this->getDb()->realEscapeString($settings->getRegister()),
            $this->getDb()->realEscapeString($settings->getSettingsPbx()->toJson()),
            $this->getDb()->realEscapeString($settings->getSettingsWms()->toJson()),
            $id
        );

        return $this->getDb()->exec($query);
    }

    public function findOrFailFxo(int $id): FxoTrunk {
        $trunk = $this->getDb()->first(
            sprintf(
                'SELECT %s FROM %s WHERE `id`=%d',
                '*',
                WMSDB::TABLE_PSTN_TRUNKS,
                $id
            )
        );

        if (!$trunk) {
            throw new InvalidDataException(Error::NotFound, [], 404);
        }

        return FxoTrunkFactory::makeFromStorage($trunk);
    }
}
