<?php

declare(strict_types=1);

namespace Api\Trunks\Service;

use Api\Trunks\Adapter\JsonAdapter;
use Api\Trunks\Configs\TrunkConfigDnsFile;
use Api\Trunks\Configs\TrunkConfigExtraFile;
use Api\Trunks\Configs\TrunkConfigRegistrationFile;
use Api\Trunks\Interfaces\TrunkConfigCwServiceInterface;
use Api\Trunks\Pstn\Objects\Trunk;
use Api\Trunks\Pstn\Objects\Trunks as PstnTrunks;
use Api\Trunks\Sip\Objects\Trunks as SipTrunks;
use Wildix\System\Dhcp\ServerManager;

/**
 * Class TrunkConfigCwService
 */
class TrunkConfigCwService implements TrunkConfigCwServiceInterface {

    const DEVICE_MODEL_FXOR3 = 'w04fxor3';

    private $_configExtra;
    private $_configRegistration;
    private $_configTrunksDns;

    private $_outArr;
    private $_dnsMasq;

    public function __construct(TrunkConfigExtraFile $configExtraFile, TrunkConfigRegistrationFile $configRegistrationFile, TrunkConfigDnsFile $configTrunksDns) {
        $this->_configExtra        = $configExtraFile;
        $this->_configRegistration = $configRegistrationFile;
        $this->_configTrunksDns    = $configTrunksDns;
        $this->_outArr             = [];
        $this->_dnsMasq            = [];
    }


    /**
     * Update Configs
     *
     * @param PstnTrunks $pstnTrunks Trunk settings data.
     * @param SipTrunks  $sipTrunks  Sip settings data.
     *
     * @return void
     */
    public function update(PstnTrunks $pstnTrunks, SipTrunks $sipTrunks): void {
        $this->setSipCwData($sipTrunks);

        $trunkIdModelMap = $this->getTrunkIdModelMap($pstnTrunks);
        $this->setPstnCwData($pstnTrunks, $trunkIdModelMap);

        $this->getConfigExtra()->setDefault($this->_outArr);
        $this->getConfigExtra()->save();

        $this->getConfigRegistration()->setDefault($this->_outArr);
        $this->getConfigRegistration()->save();

        $this->getConfigTrunksDns()->setDefault($this->_dnsMasq);
        $this->getConfigTrunksDns()->save();

        if (! ServerManager::restartServer()) {
            $this->getConfigTrunksDns()->setDefault([]);
            $this->getConfigTrunksDns()->save();
        }
    }

    private function setPstnCwData($pstnTrunks, array $trunkIdModelMap): void {
        $pstnTrunks->map(
            function (Trunk $trunk) use (&$trunkIdModelMap) {
                // all trunks should be present in sip-extra.conf
                $settingsMgw = $trunk->getSettingsMgw()->toArray();
                $nLines      = 0;
                $ports       = count($settingsMgw['ports']);
                for ($i = 0; $i < $ports; ++$i) {
                    $devName              = $trunk->getUsername() . '_' . $i;
                    $username             = ($trunk->isFxo()) ? $devName : $devName . '_' . $nLines;
                    $line                 = [];
                    $line['register']     = '';
                    $line['settings_pbx'] = [];

                    $line['settings_pbx']['host'] = 'dynamic';
                    if ($trunk->isBri()) {
                        $line['settings_pbx']['call-limit'] = 2;
                    }

                    if ($trunk->isPri()) {
                        $line['settings_pbx']['call-limit'] = 30;
                    }
                    if ($trunk->isGsm() || $trunk->isFxo()) {
                        $line['settings_pbx']['call-limit'] = 1;
                    }

                    if ($settingsMgw['toneZone']) {
                        $line['settings_pbx']['tonezone'] = $settingsMgw['toneZone'];
                    }

                    $line['settings_pbx']['pricelist']     = $settingsMgw['ports'][$i]['priceList'] ?? '0';
                    $line['settings_pbx']['context']       = $settingsMgw['ports'][$i]['context'] ?? 'main';
                    $line['settings_pbx']['country_code']  = $settingsMgw['countryCode'] ? $settingsMgw['countryCode'] : '';
                    $line['settings_pbx']['insecure']      = 'invite,port';
                    $line['settings_pbx']['symmetric_rtp'] = 'yes';
                    $line['settings_pbx']['type']          = 'friend';

                    if ($trunk->isBri() || $trunk->isPri()) {
                        $line['settings_pbx']['preserve_orig_cid'] = $trunk->getSettingsMgw()->getPreserveOrigCid() ? 'yes' : 'no';
                    }
                    if ($trunk->isGsm()) {
                        $line['settings_pbx']['preserve_orig_cid'] = 'yes';
                    }

                    if ($trunk->isBri() || $trunk->isPri() || $trunk->isFxo() || $trunk->isGsm()) {
                        $line['settings_pbx']['dtmfmode']   = 'info';
                        $line['settings_pbx']['direct_rtp'] = 'yes';
                    } else {
                        $line['settings_pbx']['qualify']    = 'yes';
                        $line['settings_pbx']['direct_rtp'] = 'no';
                    }
                      $line['settings_pbx']['sendrpid'] = 'no';
                    if (!empty($settingsMgw['ports'][$i]['callingNameMaxLength'])
                        && $settingsMgw['ports'][$i]['callingNameMaxLength'] > 0
                    ) {
                         $line['settings_pbx']['cidheader'] = 'P-Asserted-Identity';
                    }
                    if ($trunk->isFxo()
                        && isset($trunkIdModelMap[$trunk->getId()])
                        && $trunkIdModelMap[$trunk->getId()] !== self::DEVICE_MODEL_FXOR3
                    ) {
                         $line['settings_pbx']['session-expires'] = 3600;
                    }

                      ksort($line);
                      $this->_outArr[$username] = $line;
                }
            }
        );
    }
    private function replaceSpacesToUrlFormat(string $string): string {
        return str_replace(' ', '%20', $string);
    }

    private function setSipCwData(SipTrunks $sipTrunks): void {
        foreach ($sipTrunks->all() as $trunk) {
            // all trunks should be present in sip-extra.conf
            $peername             = $this->replaceSpacesToUrlFormat($trunk->getUsername());
            $line                 = [];
            $line['register']     = $trunk->getRegister();
            $line['settings_pbx'] = json_decode($trunk->getSettingsPbx()->toJson(), true);
            $settingsWms          = json_decode($trunk->getSettingsWms()->toJson(), true);

            if (!empty($line['settings_pbx']['peername'])) {
                $peername = $this->replaceSpacesToUrlFormat($line['settings_pbx']['peername'] ?? $line['settings_pbx']['username']);
            }
            if (!empty($line['settings_pbx']['username'])) {
                $line['settings_pbx']['username'] = $this->replaceSpacesToUrlFormat($line['settings_pbx']['username']);
            }
            unset($line['settings_pbx']['peername']);
            if ($trunk->getSettingsPbx()->isDynamicHost()) {
                unset($line['settings_pbx']['secret']);
                $line['settings_pbx']['insecure'] = 'invite,port';
            }

            $line['settings_pbx']['sendrpid'] = 'no';

            if ($line['settings_pbx']['session_timer'] == 'no') {
                $line['settings_pbx']['rtptimeout']     = '60';
                $line['settings_pbx']['rtpholdtimeout'] = '1800';
                $line['settings_pbx']['session-timers'] = 'refuse';
                unset($line['settings_pbx']['session-expires']);
                unset($line['settings_pbx']['session-minse']);
            }

            unset($line['settings_pbx']['session_timer']);
            // delete variable
            if ($line['settings_pbx']['allowtransfer'] == 'no') {
                $line['settings_pbx']['hold_disallow'] = '1';
            } else {
                if ($line['settings_pbx']['allowtransfer'] == 'on') {
                    $line['settings_pbx']['allowtransfer'] = 'yes';
                    $line['settings_pbx']['hold_disallow'] = '0';
                }
            }
            if (!empty($line['settings_pbx']['force_srtp'])) {
                $line['settings_pbx']['force_srtp'] = ($line['settings_pbx']['force_srtp'] === 'yes') ? 'yes' : 'no';
            }


            if ($line['settings_pbx']['t38pt_udptl']) {
                if ($line['settings_pbx']['t38pt_udptl'] == 'no' && $line['settings_pbx']['t38pt_udptl'] == '') {
                    $line['settings_pbx']['t38pt_udptl'] = 'no';
                }
            } else {
                $line['settings_pbx']['t38pt_udptl'] = 'no';
            }

            if ($line['settings_pbx']['dtmfmode'] !== 'rfc2833') {
                if (isset($line['settings_pbx']['rfc2833payload'])) {
                    unset($line['settings_pbx']['rfc2833payload']);
                    // delete variable
                }
            } else {
                $line['settings_pbx']['rfc2833payload'] = trim($line['settings_pbx']['rfc2833payload']);
                if (!isset($line['settings_pbx']['rfc2833payload'])
                    || strlen($line['settings_pbx']['rfc2833payload']) == 0
                ) {
                    $line['settings_pbx']['rfc2833payload'] = 101;
                }
            }
            $line['settings_pbx']['fromnumber']         = $line['settings_pbx']['fromnumber'] ?? '';
            $line['settings_pbx']['fromname']           = $line['settings_pbx']['fromname'] ?? '';
            $line['settings_pbx']['cidheader']          = $line['settings_pbx']['cidheader'] ?? '';
            $line['settings_pbx']['cidbody']            = $line['settings_pbx']['cidbody'] ?? '';
            $line['settings_pbx']['privacyheader']      = $line['settings_pbx']['privacyheader'] ?? '';
            $line['settings_pbx']['diversion_header']   = $line['settings_pbx']['diversion_header'] ?? '';
            $line['settings_pbx']['historyinfo_header'] = $line['settings_pbx']['historyinfo_header'] ?? '';
            $line['settings_pbx']['contactnumber']      = $line['settings_pbx']['contactnumber'] ?? '';
            $line['settings_pbx']['rtpkeepalive']       = '1';

            if (!isset($line['settings_pbx']['qualify'])) {
                $line['settings_pbx']['qualify'] = strlen($line['register']) ? 'yes' : 'no';
            }

            $line['settings_pbx']['diversion_header']   = strtr(
                $line['settings_pbx']['diversion_header'],
                [';' => '\;']
            );
            $line['settings_pbx']['historyinfo_header'] = strtr(
                $line['settings_pbx']['historyinfo_header'],
                [';' => '\;']
            );

            if ($line['settings_pbx']['fromnumber'] != '') {
                unset($line['settings_pbx']['fromuser']);
            }

            $this->_outArr[$peername] = $line;

            if (!empty($settingsWms['dnsserver'])) {
                $host = $line['settings_pbx']['host'];
                if (!empty($settingsWms['outboundproxy'])) {
                    $host = $settingsWms['outboundproxy'];
                }
                if (!empty($host) && $host != 'dynamic') {
                    $this->_dnsMasq[] = sprintf('server=/%s/%s', $host, $settingsWms['dnsserver']);
                }
            }
        }
    }

    private function getTrunkIdModelMap(PstnTrunks $pstnTrunks): array {
        $trunkIdModelMap = [];
        $pstnTrunks->map(
            function (Trunk $trunk) use (&$trunkIdModelMap) {
                if (!empty($trunk->getDeviceMap())) {
                    $map = new JsonAdapter($trunk->getDeviceMap(), ['device']);
                    $trunkIdModelMap[$trunk->getId()] = $map->device['model'];
                }
            }
        );

        return $trunkIdModelMap;
    }

    /**
     * Get _configExtra
     *
     * @return TrunkConfigExtraFile
     */
    public function getConfigExtra(): TrunkConfigExtraFile {
        return $this->_configExtra;
    }

    /**
     * Get _configRegistration
     *
     * @return TrunkConfigRegistrationFile
     */
    public function getConfigRegistration(): TrunkConfigRegistrationFile {
        return $this->_configRegistration;
    }

    /**
     * Get _configTrunksDns
     *
     * @return TrunkConfigDnsFile
     */
    public function getConfigTrunksDns(): TrunkConfigDnsFile {
        return $this->_configTrunksDns;
    }
}
