<?php

declare(strict_types=1);

namespace Api\Trunks\Service;

use Api\Device\Interfaces\DeviceRepositoryInterface;
use Api\Trunks\Factory\TrunkStatusFactory;
use Api\Trunks\Interfaces\TrunkConfigCwServiceInterface;
use Api\Trunks\Interfaces\TrunkRepositoryInterface;
use Api\Trunks\PriceList\Interfaces\PriceListCardRepositoryInterface;
use Api\Trunks\Pstn\Interfaces\PeerServiceInterface;
use Api\Trunks\Pstn\Objects\Trunk;
use Api\Trunks\Pstn\Objects\Trunks;
use Logger\Log;
use PBX;
use Wildix\SipProxy;

/**
 * Class TrunkService
 */
class TrunkService {

    private $_trunkRepository;
    private $_deviceRepository;
    private $_priceListCardRepository;
    private $_peerService;
    private $_trunkConfigCwService;
    private $_sipProxy;
    private $_logger;
    private $_pbx;

    public function __construct(
        TrunkRepositoryInterface $_trunkRepository,
        DeviceRepositoryInterface $deviceRepository,
        PeerServiceInterface $peerService,
        TrunkConfigCwServiceInterface $trunkConfigCwService,
        PriceListCardRepositoryInterface $priceListCardRepository,
        SipProxy $sipProxy,
        Log $logger,
        PBX $pbx
    ) {
        $this->_trunkRepository         = $_trunkRepository;
        $this->_deviceRepository        = $deviceRepository;
        $this->_peerService             = $peerService;
        $this->_trunkConfigCwService    = $trunkConfigCwService;
        $this->_sipProxy                = $sipProxy;
        $this->_logger                  = $logger;
        $this->_priceListCardRepository = $priceListCardRepository;
        $this->_pbx = $pbx;
    }

    public function getLogger(): Log {
        return $this->_logger;
    }

    public function getPeerService(): PeerServiceInterface {
        return $this->_peerService;
    }

    private function setDevicePstnMap(Trunks $pstnItems): void {
        $map = $this->_deviceRepository->getTrunkIdDeviceMap($pstnItems->getIds());
        $pstnItems->map(
            function (Trunk $pstn) use ($map) {
                if ($pstn->getId() && isset($map[$pstn->getId()])) {
                    $pstn->setDeviceMap($map[$pstn->getId()]);
                }
            }
        );
    }

    public function updateConfigs(): void {
        $pstnItems = $this->getTrunkRepository()->getPstnItems();
        $sipItems  = $this->getTrunkRepository()->getSipItems();
        $this->setDevicePstnMap($pstnItems);
        $this->getTrunkConfigCwService()->update($pstnItems, $sipItems);
        $this->getPbxInstance()->put_cli_reload('sip');
    }

    public function getTrunkRepository(): TrunkRepositoryInterface {
        return $this->_trunkRepository;
    }

    public function getTrunkConfigCwService(): TrunkConfigCwServiceInterface {
        return $this->_trunkConfigCwService;
    }

    public function getSipProxy(): SipProxy {
        return $this->_sipProxy;
    }

    public function getDeviceRepository(): DeviceRepositoryInterface {
        return $this->_deviceRepository;
    }

    /**
     * Get _priceListCardRepository
     *
     * @return PriceListCardRepositoryInterface
     */
    public function getPriceListCardRepository(): PriceListCardRepositoryInterface {
        return $this->_priceListCardRepository;
    }

    private function getPbxInstance(): PBX {
        return $this->_pbx;
    }

    protected function setTrunkStatuses(Trunks $trunks): TrunkService {
        $trunks->map(
            function (Trunk $trunk) {
                $status = TrunkStatusFactory::make($trunk, $this->getSipProxy(), $this->getDeviceRepository());
                $trunk->setStatus($status);
            }
        );

        return $this;
    }
}
