<?php

declare(strict_types=1);

namespace Api\Trunks\Sip\Builder;

use Api\Trunks\Sip\Enum\SettingsPbxEnum;
use Api\Trunks\Sip\Objects\SettingsPbx;
use Api\Trunks\Sip\Objects\SettingsWms;
use Api\Trunks\Sip\Objects\Trunk;
use Wildix\Error;
use Wildix\InvalidDataException;

/**
 * Class SettingsPbxBuilder
 */
class SettingsBuilder {

    /**
     * Build data
     *
     * @param array $resultData Data from request.
     *
     * @throws InvalidDataException If there is invalid name for peername
     *
     * @return Trunk
     */
    public static function build(array $resultData): Trunk {
        if (empty(trim($resultData['settingsPbx']['peername']))) {
            $explodedPeername = explode('@', $resultData['settingsPbx']['userName']);
            $resultData['settingsPbx']['peername'] = $explodedPeername[0];
        }

        // check reserved trunk names like (bri|pri|gsm)X_X_X, fxoX_X and classound
        $regexPeername = '/((gsm|bri|pri|fxs)\d+\_\d+\_\d+|fxo\d+_\d+|classound)/i';
        if (preg_match($regexPeername, (string)$resultData['settingsPbx']['peername'])) {
            $error['settingsPbx']['peername']['reason'] = 'Protected name';
            $error['settingsPbx']['peername']['code']   = 'ProtectedName';
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $error);
        }
        if ($resultData['settingsPbx']['host'] == SettingsPbx::DYNAMIC_HOST) {
            $resultData['settingsWms']['enableReg'] = false;
        }

        if ($resultData['settingsWms']['enableReg']) {
            $isOverReg = $resultData['settingsWms']['overwriteReg'];
            $register  = self::getRegisterString(
                [
                    'peerName'   => (string)$resultData['settingsPbx']['peername'],
                    'expiry'     => (string)$resultData['settingsWms']['expiry'],
                    'userName'   => $isOverReg ? $resultData['settingsWms']['regUsername'] : $resultData['settingsPbx']['userName'],
                    'secret'     => $isOverReg ? $resultData['settingsWms']['regSecret'] : $resultData['settingsPbx']['secret'],
                    'host'       => $isOverReg ? $resultData['settingsWms']['regHost'] : $resultData['settingsPbx']['host'],
                    'port'       => $isOverReg ? $resultData['settingsWms']['regPort'] : $resultData['settingsPbx']['port'],
                    'fromUser'   => $isOverReg ? $resultData['settingsWms']['regFromUser'] : $resultData['settingsPbx']['fromUser'],
                    'fromDomain' => ($isOverReg ? '' : $resultData['settingsPbx']['fromDomain'])
                ]
            );
        } else {
            $register = '';
        }

        $trunk = new Trunk();

        return $trunk->setId((int)$resultData['id'])
            ->setTitle((string)$resultData['title'])
            ->setUsername((string)$resultData['settingsPbx']['userName'])
            ->setRegister($register)
            ->setSettingsPbx(self::buildPbxSettings($resultData))
            ->setSettingsWms(self::buildWmsSettings($resultData));
    }

    private static function getRegisterString($settings): string {
        $settings['secret'] = strtr($settings['secret'], [';' => '\;']);
        $secret             = (!empty($settings['secret']) ? sprintf(':%s', $settings['secret']) : '');

        $peerName = (!empty($settings['peerName']) ? self::escape($settings['peerName']) . '?' : '');
        $user     = (!empty($settings['userName']) ? self::escape($settings['userName']) : '');
        $authUser = ((!empty($user) && !empty($secret)) ? sprintf(':"%s"', $user) : '');
        $host     = (!empty($settings['host']) ? $settings['host'] : '');
        $port     = (!empty($settings['port']) ? sprintf(':%s', $settings['port']) : '');
        $expiry   = (!empty($settings['expiry']) ? sprintf('~%s', $settings['expiry']) : '');

        $fromDomain = (!empty($settings['fromDomain']) ? '@' . $settings['fromDomain'] : '');
        $fromUser   = (!empty($settings['fromUser']) ? $settings['fromUser'] : $user);

        return sprintf(
            '%s%s%s%s%s@%s%s%s',
            $peerName,
            $fromUser,
            $fromDomain,
            $secret,
            $authUser,
            $host,
            $port,
            $expiry
        );
    }

    private static function escape(string $str): string {
        $trans = [' ' => '%20'];

        return strtr($str, $trans);
    }

    private static function buildPbxSettings(array $data): SettingsPbx {
        $settingsPbx = new SettingsPbx();
        foreach ($data['settingsPbx'] as $k => $v) {
            if (in_array($k, SettingsPbxEnum::getEnumValues())) {
                $settingsArray[$k] = is_bool($v) ? $v : trim((string)$v);
            }
        }

        $settingsArray['disallow']      = 'all';
        $settingsArray['insecure']      = 'invite,port';
        $settingsArray['symmetric_rtp'] = 'yes';

        $audioCodecs = [];
        foreach ($data['settingsPbx']['audioCodecs'] as $codec) {
            if ($codec['enable']) {
                $audioCodecs[] = sprintf('%s:%s', $codec['key'], $codec['value']);
            }
        }
        $settingsArray['allow'] = implode(',', $audioCodecs) ?? 'g729:20,alaw:20,ulaw:20';

        if ($settingsArray['videoCodecs']) {
            $settingsArray['allow'] .= ',h264,h263,h263p';
        }
        if (!isset($settingsArray['qualify'])) {
            $settingsArray['qualify'] = isset($data['overwriteReg']) && $data['overwriteReg'];
        }

        $settingsArray['countryCode'] = $settingsArray['countryCode'] ? (int)$settingsArray['countryCode'] : null;

        if ($settingsArray['host'] == '') {
            $settingsArray['host'] = 'dynamic';
        }
        if (!is_bool($settingsArray['trustRpid']) && empty($settingsArray['trustRpid'])) {
            $settingsArray['trustRpid'] = true;
        }

        if (!isset($settingsArray['rfcPayload'])) {
            $settingsArray['rfcPayload'] = 101;
        }

        $settingsArray['thirtyEightPtErrorCorrection'] = $data['settingsPbx']['thirtyEightPtErrorCorrection'] ?? 'redundancy';
        $settingsArray['thirtyEightPtMaxDatagram']     = $data['settingsPbx']['thirtyEightPtMaxDatagram'] ?? 400;

        if ($settingsArray['sessionTimer']) {
            if (!$settingsArray['sessionMinse']) {
                $settingsArray['sessionMinse'] = 90;
            }
            if (!$settingsArray['sessionExpires']) {
                $settingsArray['sessionExpires'] = 600;
            }
        }

        if (!$settingsArray['allowTransfer']) {
            $settingsArray['playRemoteMoh'] = false;
        }

        $settingsPbx
            ->setPriceList(intval($settingsArray['priceList']))
            ->setPeerName((string)$settingsArray['peername'])
            ->setUserName((string)$settingsArray['userName'])
            ->setFromUser((string)$settingsArray['fromUser'])
            ->setFromDomain((string)$settingsArray['fromDomain'])
            ->setHost((string)$settingsArray['host'])
            ->setPort((string)$settingsArray['port'])
            ->setSecret((string)$settingsArray['secret'])
            ->setContext((string)$settingsArray['context'])
            ->setToneZone((string)$settingsArray['toneZone'])
            ->setCountryCode($settingsArray['countryCode'])
            ->setFromNumber((string)$settingsArray['fromNumber'])
            ->setFromName((string)$settingsArray['fromName'])
            ->setCidHeader((string)$settingsArray['cidHeader'])
            ->setCidBody((string)$settingsArray['cidBody'])
            ->setTrustRpid((bool)$settingsArray['trustRpid'])
            ->setPrivacyHeader((string)$settingsArray['privacyHeader'])
            ->setDiversionHeader((string)$settingsArray['diversionHeader'])
            ->setHistoryInfoHeader((string)$settingsArray['historyInfoHeader'])
            ->setPreserveOrigCid($settingsArray['preserveOrigCid'] ?? true)
            ->setAllowTransfer((bool)$settingsArray['allowTransfer'])
            ->setSessionTimer((bool)$settingsArray['sessionTimer'] ?? true)
            ->setSessionMinse((int)$settingsArray['sessionMinse'])
            ->setSessionExpires((int)$settingsArray['sessionExpires']);
        $settingsPbx
            ->setThirtyEightPtUdptl((bool)$settingsArray['thirtyEightPtUdptl'])
            ->setThirtyEightPtErrorCorrection((string)$settingsArray['thirtyEightPtErrorCorrection'])
            ->setThirtyEightPtMaxdatagram((int)$settingsArray['thirtyEightPtMaxDatagram'])
            ->setSsrcChange((bool)$settingsArray['ssrcChange'])
            ->setRport((string)$settingsArray['rport'])
            ->setForceSrtp((bool)$settingsArray['forceSrtp'])
            ->setTransportWildix((string)$settingsArray['transportWildix'])
            ->setDtmfMode((string)$settingsArray['dtmfMode'])
            ->setRfcPayload((int)$settingsArray['rfcPayload'])
            ->setType((string)$settingsArray['type'])
            ->setLanguage((string)$settingsArray['language'])
            ->setQualify((bool)$settingsArray['qualify'])
            ->setInsecure((string)$settingsArray['insecure'])
            ->setSymmetricRtp(true)
            ->setDisallow((string)$settingsArray['disallow'])
            ->setAllow((string)$settingsArray['allow'])
            ->setVideoCodecs((bool)$settingsArray['videoCodecs'])
            ->setPlayRemoteMoh((bool)$settingsArray['playRemoteMoh'])
            ->setContactNumber((string)$settingsArray['contactNumber']);

        return $settingsPbx;
    }

    private static function buildWmsSettings(array $data): SettingsWms {
        $settingsWms = new SettingsWms();

        if ($data['settingsWms']['overwriteReg']) {
            $settingsWms->setRegHost($data['settingsWms']['regHost'])
                ->setRegPort($data['settingsWms']['regPort'])
                ->setRegUsername($data['settingsWms']['regUsername'])
                ->setRegSecret($data['settingsWms']['regSecret'])
                ->setRegFromUser($data['settingsWms']['regFromUser']);
        }
        if ($data['settingsPbx']['host'] == 'dynamic') {
            $data['settingsWms']['enableReg'] = false;
        }
        $settingsWms->setEnableReg($data['settingsWms']['enableReg'] ?? false)
            ->setOverwriteReg($data['settingsWms']['overwriteReg'] ?? false)
            ->setOutboundProxy(trim((string)$data['settingsWms']['outboundProxy']) ?? '')
            ->setOutboundProxyPort((string)$data['settingsWms']['outboundProxyPort'] ?? '')
            ->setExpiry((string)($data['settingsWms']['expiry']) ?? '')
            ->setDnsServer($data['settingsWms']['dnsServer'] ?? '')
            ->setNatIp(trim((string)$data['settingsWms']['natIp']) ?? '');

        return $settingsWms;
    }
}
