<?php

declare(strict_types=1);

namespace Api\Trunks\Sip\Factory;

use Api\Support\BaseCollectObject;
use Api\Translation\Interfaces\TranslationStrategyInterface;
use Api\Translation\Objects\TranslationInterface;
use Api\Trunks\Sip\Objects\Trunk;
use Api\Trunks\Sip\Objects\TrunkRegistrationStatus;
use Api\Trunks\Sip\Objects\TrunkSipStatus;
use Wildix\CallweaverManager;
use Wildix\Core\Container;
use Wildix\SipProxy;

/**
 * Class SipStatusFactory
 */
class SipStatusFactory {

    public static function make(Trunk $sip): TrunkSipStatus {
        $sipStatus             = new TrunkSipStatus();
        $incomingRegistrations = self::getIncomingRegistrations();
        $outGoingRegistrations = self::getOutGoingRegistrations();
        $userName              = self::getUserNameByTrunk($sip);
        $peerName              = $sip->getSettingsPbx()->getPeerName();

        if ($incomingRegistrations->offsetExists($userName)) {
            self::changeStatusesByIncoming($incomingRegistrations->offsetGet($userName), $sipStatus, $sip);
        } else if ($outGoingRegistrations->offsetExists($peerName)) {
            self::changeStatusesByOutgoing($outGoingRegistrations->offsetGet($peerName), $sipStatus, $sip);
        } else if ($outGoingRegistrations->offsetExists($userName)) {
            self::changeStatusesByOutgoing($outGoingRegistrations->offsetGet($userName), $sipStatus, $sip);
        }

        return $sipStatus;
    }

    private static function changeStatusesByIncoming(TrunkRegistrationStatus $incoming, TrunkSipStatus $sipStatus, Trunk $sip): void {
        $registration = $incoming->getRegistration();
        if ($registration) {
            $sipStatus->setIncoming(
                [
                    'state' => TrunkSipStatus::STATE_ON,
                    'title' => $registration
                ]
            );
            $sipStatus->setOutgoing(self::getUnregisteredStatus($sip->getSettingsPbx()->getHost()));
        } else {
            self::setUnregisteredStatuses($sip, $sipStatus);
        }
    }

    private static function changeStatusesByOutgoing(array $outGoing, TrunkSipStatus $sipStatus, Trunk $sip): void {
        $statuses = [
            'Registered' => TrunkSipStatus::STATE_ON,
            'Reachable'  => TrunkSipStatus::STATE_ON,
            'Fail'       => TrunkSipStatus::STATE_OFF,
        ];

        if ($outGoing['host'] || !empty(self::getTranslateStateByKey($outGoing['status']))) {
            $sipStatus->setOutgoing(
                [
                    'state' => $statuses[$outGoing['status']] ?? TrunkSipStatus::STATE_UNREGISTERED,
                    'title' => $outGoing['status'] . ' ' . $sip->getSettingsPbx()->getHost()
                ]
            );
        } else {
            self::setUnregisteredStatuses($sip, $sipStatus);
        }
    }

    private static function setUnregisteredStatuses(Trunk $sip, TrunkSipStatus $sipStatus): void {
        $host = $sip->getSettingsPbx()->isDynamicHost() ? '' : $sip->getSettingsPbx()->getHost();
        $sipStatus->setOutgoing(self::getUnregisteredStatus($host));
        $sipStatus->setIncoming(self::getUnregisteredStatus($host));
    }

    private static function getUnregisteredStatus(string $host): array {
        return [
            'state' => TrunkSipStatus::STATE_UNREGISTERED,
            'title' => self::getTranslateStateByKey('Unregistered') . ' ' . $host
        ];
    }

    private static function getUserNameByTrunk(Trunk $sip): string {
        $userName = $sip->getUsername();
        if ($sip->getSettingsWms()->isOverwriteReg() && !empty($sip->getSettingsWms()->getRegUsername())) {
            $userName = $sip->getSettingsWms()->getRegUsername();
        }

        return $userName;
    }

    private static function getOutGoingRegistrations(): BaseCollectObject {
        $outgoing      = new BaseCollectObject();
        $outgoingInfos = CallweaverManager::getInstance()->getSipRegistrations();
        foreach ($outgoingInfos as $outgoingInfo) {
            $status = '';
            if (preg_match('/^\d+\s(\w+)/', $outgoingInfo['Refresh State'], $matches)) {
                if (!empty($matches[1])) {
                    $status = $matches[1];
                }
            }

            $regHost = '';
            if (preg_match('/^([^:]+):\d+$/', $outgoingInfo['Host'], $matches)) {
                if (!empty($matches[1])) {
                    $regHost = $matches[1];
                }
            }

            $key = rawurldecode($outgoingInfo['Peername']);
            $outgoing->put(
                $key,
                [
                    'host' => $regHost,
                    'status' => $status,
                ]
            );
        }

        return $outgoing;
    }

    private static function getIncomingRegistrations(): BaseCollectObject {
        $incoming = new BaseCollectObject();

        $incomingRegistrations = self::getSipProxy()->getTrunksRegistrations();
        foreach ($incomingRegistrations as $login => $value) {
            if (!empty($value) && !empty($value['registrations'])) {
                $trunkStatus = new TrunkRegistrationStatus();
                $trunkStatus
                    ->setLogin((string)$login)
                    ->setRegistration($value['registrations'][0]['received'])
                    ->setAgent($value['registrations'][0]['useragent'])
                    ->setStatus((bool)$value['registrations'][0]['online']);
                $incoming->put($login, $trunkStatus);
            }
        }

        return $incoming;
    }

    private static function getSipProxy(): SipProxy {
        return self::getDiContainer()->get(SipProxy::class);
    }

    private static function getDiContainer() {
        return $GLOBALS['diContainer'] ?? new Container();
    }

    private static function getTranslateStateByKey(string $key): string {
        $map = [
            'Fail' => self::getTranslate()->t('state.Fail'),
            'Reason' => self::getTranslate()->t('state.Fail.Reason'),
            'Timeout' => self::getTranslate()->t('state.Timeout'),
            'Expired' => self::getTranslate()->t('state.Expired'),
            'Rejected' => self::getTranslate()->t('state.Rejected'),
            'Auth Sent' => self::getTranslate()->t('state.AuthSent'),
            'Registered' => self::getTranslate()->t('state.Registered'),
            'Request Sent' => self::getTranslate()->t('state.RequestSent'),
            'Request' => self::getTranslate()->t('state.RequestSent'),
            'Unregistered' => self::getTranslate()->t('state.Unregistered'),
            'Reachable' => self::getTranslate()->t('state.Registered'),
            'Unreachable' => self::getTranslate()->t('state.RequestSent'),
        ];

        return array_key_exists($key, $map) ? $map[$key] : '';
    }

    private static function getTranslate(): TranslationInterface {
        return self::getDiContainer()->get(TranslationStrategyInterface::class);
    }
}
