<?php

declare(strict_types=1);

namespace Api\Trunks\Sip\Objects;

use Api\Support\Interfaces\ArrayableInterface;
use Api\Support\Interfaces\Jsonable;

/**
 * Class SettingsPbx
 */
class SettingsPbx implements Jsonable, ArrayableInterface {
    const DEFAULT_AUDIO_CODECS = [
        [
            'key'    => 'g729',
            'value'  => 20,
            'enable' => true,
        ],
        [
            'key'    => 'alaw',
            'value'  => 20,
            'enable' => true,
        ],
        [
            'key'    => 'ulaw',
            'value'  => 20,
            'enable' => true,
        ],
    ];
    const DNS_SRV              = 'dns-srv';
    const DYNAMIC_HOST         = 'dynamic';
    /**
     * Allow
     *
     * @var string
     */
    private $_allow;
    /**
     * Support Refer and Hold
     *
     * @var bool
     */
    private $_allowTransfer;
    /**
     * Audio codecs
     *
     * @var array
     */
    private $_audioCodecs;
    /**
     * Cid Body
     *
     * @var string
     */
    private $_cidBody;
    /**
     * Cid Header
     *
     * @var string
     */
    private $_cidHeader;
    /**
     * Dialplan
     *
     * @var string
     */
    private $_context;
    /**
     * Country Code
     *
     * @var int
     */
    private $_countryCode;
    /**
     * Disallow
     *
     * @var string
     */
    private $_disallow;
    /**
     * Diversion Header
     *
     * @var string
     */
    private $_diversionHeader;
    /**
     * DTMF mode
     *
     * @var string
     */
    private $_dtmfMode;
    /**
     * SDES-SRTP
     *
     * @var bool
     */
    private $_forceSrtp;
    /**
     * From domain
     *
     * @var string
     */
    private $_fromDomain;
    /**
     * From name
     *
     * @var string
     */
    private $_fromName;
    /**
     * From user
     *
     * @var string
     */
    private $_fromNumber;
    /**
     * From user
     *
     * @var string
     */
    private $_fromUser;
    /**
     * History-Info Header
     *
     * @var string
     */
    private $_historyInfoHeader;
    /**
     * Host Name
     *
     * @var string
     */
    private $_host;
    /**
     * Insecure
     *
     * @var string
     */
    private $_insecure;
    /**
     * Language
     *
     * @var string
     */
    private $_language;
    /**
     * Trunk name
     *
     * @var string
     */
    private $_peerName;
    /**
     * Play Remote Moh
     *
     * @var boolean
     */
    private $_playRemoteMoh = false;
    /**
     * Port
     *
     * @var string
     */
    private $_port;
    /**
     * Show original caller number
     *
     * @var boolean
     */
    private $_preserveOrigCid;
    /**
     * Price list
     *
     * @var int
     */
    private $_priceList;
    /**
     * Privacy Header
     *
     * @var string
     */
    private $_privacyHeader;
    /**
     * Qualify
     *
     * @var boolean
     */
    private $_qualify;
    /**
     * Payload 2833 rfc
     *
     * @var int
     */
    private $_rfcPayload;
    /**
     * Rport
     *
     * @var string
     */
    private $_rport;
    /**
     * Password
     *
     * @var string
     */
    private $_secret;
    /**
     * Session Timer max
     *
     * @var int
     */
    private $_sessionExpires;
    /**
     * Session Timer min
     *
     * @var int
     */
    private $_sessionMinse;
    /**
     * Session Timer enabled
     *
     * @var boolean
     */
    private $_sessionTimer;
    /**
     * Force static SSRC
     *
     * @var boolean
     */
    private $_ssrcChange;
    /**
     * Symmetric Rtp
     *
     * @var boolean
     */
    private $_symmetricRtp;
    /**
     * T38 error correction
     *
     * @var string
     */
    private $_thirtyEightPtErrorCorrection;
    /**
     *  T38 Maxdatagram
     *
     * @var int
     */
    private $_thirtyEightPtMaxdatagram;
    /**
     * T38 enabled
     *
     * @var boolean
     */
    private $_thirtyEightPtUdptl;
    /**
     * Tone Zone
     *
     * @var string
     */
    private $_toneZone;
    /**
     * Transport
     *
     * @var string
     */
    private $_transportWildix;
    /**
     * Incoming CID
     *
     * @var boolean
     */
    private $_trustRpid;
    /**
     * Registration Proxy enabled
     *
     * @var string
     */
    private $_type;
    /**
     * Auth Login
     *
     * @var string
     */
    private $_userName;
    /**
     * Allow
     *
     * @var boolean
     */
    private $_videoCodecs;
    /**
     * Contact Number
     *
     * @var string
     */
    private $_contactNumber;

    public function isDnsServer(): bool {
        return $this->getTransportWildix() === self::DNS_SRV;
    }

    /**
     * Get Transport
     *
     * @return string
     */
    public function getTransportWildix(): string {
        return $this->_transportWildix;
    }

    /**
     * Set Transport
     *
     * @param string $transportWildix Transport.
     *
     * @return SettingsPbx
     */
    public function setTransportWildix(string $transportWildix): SettingsPbx {
        if ($transportWildix == '') {
            $transportWildix = 'udp';
        }
        $this->_transportWildix = $transportWildix;

        return $this;
    }

    public function isDynamicHost(): bool {
        return $this->getHost() === self::DYNAMIC_HOST;
    }

    /**
     * Get Host Name
     *
     * @return string
     */
    public function getHost(): string {
        return $this->_host;
    }

    /**
     * Set Host Name
     *
     * @param string $host Host Name.
     *
     * @return SettingsPbx
     */
    public function setHost(string $host): SettingsPbx {
        $this->_host = $host;

        return $this;
    }

    /**
     * Set audio codecs from allow string
     *
     * @param string $allow Audio codecs.
     *
     * @return SettingsPbx
     */
    public function setAudioCodecsFromAllow(string $allow): SettingsPbx {
        $allowCodecs = [];

        preg_match_all('/\w+:\d+/', $allow, $audioCodecsStr);

        foreach ($audioCodecsStr[0] as $codec) {
            [$key, $value]     = explode(':', $codec);
            $allowCodecs[$key] = intval($value);
        }

        foreach (self::DEFAULT_AUDIO_CODECS as $codec) {
            if (array_key_exists($codec['key'], $allowCodecs)) {
                $codec['value'] = intval($allowCodecs[$codec['key']]);
            } else {
                $codec['enable'] = false;
            }

            $this->_audioCodecs[] = $codec;
        }

        return $this;
    }

    /**
     * Parse T38 string into params.
     *
     * @param string $string T38 string.
     *
     * @return SettingsPbx
     */
    public function setThirtyEightString(string $string): SettingsPbx {
        $params = explode(',', $string);

        $maxdatagram = (int)filter_var($params[1], FILTER_SANITIZE_NUMBER_INT) ?? 0;

        $this->setThirtyEightPtUdptl($params[0] === 'yes');
        $this->setThirtyEightPtMaxdatagram($maxdatagram);
        $this->setThirtyEightPtErrorCorrection($params[2] ?? '');

        return $this;
    }

    /**
     * Set Video codecs from allow string
     *
     * @param string $allow Allow string.
     *
     * @return SettingsPbx
     */
    public function setVideoCodecsFromAllow(string $allow): SettingsPbx {
        $this->_videoCodecs = (bool)preg_replace('/\w+:\d+,?/', '', $allow);

        return $this;
    }

    public function toArray(): array {
        return [
            'audioCodecs'                  => $this->getAudioCodecs(),
            'allowTransfer'                => $this->getAllowTransfer(),
            'cidBody'                      => $this->getCidBody(),
            'cidHeader'                    => $this->getCidHeader(),
            'context'                      => $this->getContext(),
            'countryCode'                  => $this->getCountryCode(),
            'disallow'                     => $this->getDisallow(),
            'diversionHeader'              => $this->getDiversionHeader(),
            'dtmfMode'                     => $this->getDtmfMode(),
            'forceSrtp'                    => $this->getForceSrtp(),
            'fromDomain'                   => $this->getFromDomain(),
            'fromName'                     => $this->getFromName(),
            'fromNumber'                   => $this->getFromNumber(),
            'fromUser'                     => $this->getFromUser(),
            'historyInfoHeader'            => $this->getHistoryInfoHeader(),
            'host'                         => $this->getHost(),
            'insecure'                     => $this->getInsecure(),
            'language'                     => $this->getLanguage(),
            'peername'                     => $this->getPeerName(),
            'port'                         => $this->getPort(),
            'preserveOrigCid'              => $this->getPreserveOrigCid(),
            'priceList'                    => $this->getPriceList(),
            'privacyHeader'                => $this->getPrivacyHeader(),
            'qualify'                      => $this->getQualify(),
            'rfcPayload'                   => $this->getRfcPayload(),
            'rport'                        => $this->getRport(),
            'secret'                       => $this->getSecret(),
            'sessionExpires'               => $this->getSessionExpires(),
            'sessionMinse'                 => $this->getSessionMinse(),
            'sessionTimer'                 => $this->getSessionTimer(),
            'ssrcChange'                   => $this->getSsrcChange(),
            'symmetricRtp'                 => $this->getSymmetricRtp(),
            'thirtyEightPtUdptl'           => $this->getThirtyEightPtUdptl(),
            'thirtyEightPtErrorCorrection' => $this->getThirtyEightPtErrorCorrection(),
            'thirtyEightPtMaxDatagram'     => $this->getThirtyEightPtMaxdatagram(),
            'toneZone'                     => $this->getToneZone(),
            'transportWildix'              => $this->getTransportWildix(),
            'trustRpid'                    => $this->getTrustRpid(),
            'type'                         => $this->getType(),
            'userName'                     => $this->getUserName(),
            'videoCodecs'                  => $this->getVideoCodecs(),
            'playRemoteMoh'                => $this->isPlayRemoteMoh(),
            'contactNumber'                => $this->getContactNumber(),
        ];
    }

    /**
     * Get Audio codecs
     *
     * @return array
     */
    public function getAudioCodecs(): array {
        return $this->_audioCodecs ?? self::DEFAULT_AUDIO_CODECS;
    }

    /**
     * Get Support Refer and Hold
     *
     * @return boolean
     */
    public function getAllowTransfer(): bool {
        return $this->_allowTransfer;
    }

    /**
     * Set Support Refer and Hold
     *
     * @param boolean $allowTransfer Support Refer and Hold.
     *
     * @return SettingsPbx
     */
    public function setAllowTransfer(bool $allowTransfer): SettingsPbx {
        $this->_allowTransfer = $allowTransfer;

        return $this;
    }

    /**
     * Get Cid Body
     *
     * @return string
     */
    public function getCidBody(): string {
        return $this->_cidBody ?? '';
    }

    /**
     * Set Cid Body
     *
     * @param string|null $cidBody Cid Body.
     *
     * @return SettingsPbx
     */
    public function setCidBody(?string $cidBody): SettingsPbx {
        $this->_cidBody = $cidBody;

        return $this;
    }

    /**
     * Get Cid Header
     *
     * @return string
     */
    public function getCidHeader(): string {
        return $this->_cidHeader ?? '';
    }

    /**
     * Set Cid Header
     *
     * @param string|null $cidHeader Cid Header.
     *
     * @return SettingsPbx
     */
    public function setCidHeader(?string $cidHeader): SettingsPbx {
        $this->_cidHeader = $cidHeader;

        return $this;
    }

    /**
     * Get Dialplan
     *
     * @return string
     */
    public function getContext(): string {
        return $this->_context;
    }

    /**
     * Set Dialplan
     *
     * @param string $context Dialplan.
     *
     * @return SettingsPbx
     */
    public function setContext(string $context): SettingsPbx {
        if ($context == '') {
            $context = 'main';
        }
        $this->_context = $context;

        return $this;
    }

    /**
     * Get Country Code
     *
     * @return integer
     */
    public function getCountryCode(): ?int {
        return $this->_countryCode;
    }

    /**
     * Set Country Code
     *
     * @param integer|null $countryCode Country Code.
     *
     * @return SettingsPbx
     */
    public function setCountryCode(?int $countryCode): SettingsPbx {
        $this->_countryCode = $countryCode;

        return $this;
    }

    /**
     * Get Disallow
     *
     * @return string
     */
    public function getDisallow(): string {
        return $this->_disallow;
    }

    /**
     * Set Disallow
     *
     * @param string $disallow Disallow.
     *
     * @return $this
     */
    public function setDisallow(string $disallow): SettingsPbx {
        $this->_disallow = $disallow;

        return $this;
    }

    /**
     * Get Diversion Header
     *
     * @return string
     */
    public function getDiversionHeader(): string {
        return $this->_diversionHeader ?? '';
    }

    /**
     * Set Diversion Header
     *
     * @param string|null $diversionHeader Diversion Header.
     *
     * @return SettingsPbx
     */
    public function setDiversionHeader(?string $diversionHeader): SettingsPbx {
        $this->_diversionHeader = $diversionHeader;

        return $this;
    }

    /**
     * Get DTMF mode
     *
     * @return string
     */
    public function getDtmfMode(): string {
        return $this->_dtmfMode ?? 'rfc2833';
    }

    /**
     * Set DTMF mode
     *
     * @param string|null $dtmfMode DTMF mode.
     *
     * @return SettingsPbx
     */
    public function setDtmfMode(?string $dtmfMode): SettingsPbx {
        if ($dtmfMode == '') {
            $dtmfMode = 'rfc2833';
        }
        $this->_dtmfMode = $dtmfMode;

        return $this;
    }

    /**
     * Get SDES-SRTP
     *
     * @return boolean
     */
    public function getForceSrtp(): bool {
        return $this->_forceSrtp;
    }

    /**
     * Set SDES-SRTP
     *
     * @param boolean $forceSrtp SDES-SRTP.
     *
     * @return SettingsPbx
     */
    public function setForceSrtp(bool $forceSrtp = false): SettingsPbx {
        $this->_forceSrtp = $forceSrtp;

        return $this;
    }

    /**
     * Get From domain
     *
     * @return string
     */
    public function getFromDomain(): string {
        return $this->_fromDomain ?? '';
    }

    /**
     * Set From domain
     *
     * @param string|null $fromDomain From domain.
     *
     * @return SettingsPbx
     */
    public function setFromDomain(?string $fromDomain): SettingsPbx {
        $this->_fromDomain = $fromDomain;

        return $this;
    }

    /**
     * Get From name
     *
     * @return string
     */
    public function getFromName(): string {
        return $this->_fromName ?? '';
    }

    /**
     * Set From name
     *
     * @param string|null $fromName From name.
     *
     * @return SettingsPbx
     */
    public function setFromName(?string $fromName): SettingsPbx {
        $this->_fromName = $fromName;

        return $this;
    }

    /**
     * Get From number
     *
     * @return string
     */
    public function getFromNumber(): string {
        return $this->_fromNumber ?? '';
    }

    /**
     * Set From number
     *
     * @param string|null $fromNumber From number.
     *
     * @return SettingsPbx
     */
    public function setFromNumber(?string $fromNumber): SettingsPbx {
        $this->_fromNumber = $fromNumber;

        return $this;
    }

    /**
     * Get From user
     *
     * @return string
     */
    public function getFromUser(): string {
        return $this->_fromUser ?? '';
    }

    /**
     * Set From user
     *
     * @param string|null $fromUser From user.
     *
     * @return SettingsPbx
     */
    public function setFromUser(?string $fromUser): SettingsPbx {
        $this->_fromUser = $fromUser;

        return $this;
    }

    /**
     * Get _historyInfoHeader
     *
     * @return string
     */
    public function getHistoryInfoHeader(): string {
        return $this->_historyInfoHeader ?? '';
    }

    /**
     * Set History-Info Header
     *
     * @param string|null $historyInfoHeader History-Info Header.
     *
     * @return SettingsPbx
     */
    public function setHistoryInfoHeader(?string $historyInfoHeader): SettingsPbx {
        $this->_historyInfoHeader = $historyInfoHeader;

        return $this;
    }

    /**
     * Get Insecure
     *
     * @return string
     */
    public function getInsecure(): string {
        return $this->_insecure;
    }

    /**
     * Set Insecure
     *
     * @param string $insecure Insecure.
     *
     * @return SettingsPbx
     */
    public function setInsecure(string $insecure): SettingsPbx {
        $this->_insecure = $insecure;

        return $this;
    }

    /**
     * Get Language
     *
     * @return string
     */
    public function getLanguage(): string {
        return $this->_language ?? '';
    }

    /**
     * Set Language
     *
     * @param string $language Language.
     *
     * @return SettingsPbx
     */
    public function setLanguage(?string $language): SettingsPbx {
        $this->_language = $language;

        return $this;
    }

    /**
     * Get Trunk name
     *
     * @return string
     */
    public function getPeerName(): string {
        return $this->_peerName;
    }

    /**
     * Set Trunk name
     *
     * @param string $peerName Trunk name.
     *
     * @return SettingsPbx
     */
    public function setPeerName(string $peerName): SettingsPbx {
        $this->_peerName = $peerName;

        return $this;
    }

    /**
     * Get Port
     *
     * @return string
     */
    public function getPort(): string {
        return $this->_port ?? '';
    }

    /**
     * Port
     *
     * @param string|null $port Port.
     *
     * @return SettingsPbx
     */
    public function setPort(?string $port): SettingsPbx {
        $this->_port = $port;

        return $this;
    }

    /**
     * Get Show original caller number
     *
     * @return boolean
     */
    public function getPreserveOrigCid(): bool {
        return $this->_preserveOrigCid;
    }

    /**
     * Set Show original caller number
     *
     * @param boolean $preserveOrigCid Show original caller number.
     *
     * @return SettingsPbx
     */
    public function setPreserveOrigCid(bool $preserveOrigCid = true): SettingsPbx {
        $this->_preserveOrigCid = $preserveOrigCid;

        return $this;
    }

    /**
     * Get Price list
     *
     * @return integer
     */
    public function getPriceList(): int {
        return $this->_priceList;
    }

    /**
     * Set Price list
     *
     * @param integer $priceList Price list.
     *
     * @return SettingsPbx
     */
    public function setPriceList(int $priceList): SettingsPbx {
        $this->_priceList = $priceList;

        return $this;
    }

    /**
     * Get Privacy Header
     *
     * @return string
     */
    public function getPrivacyHeader(): string {
        return $this->_privacyHeader ?? '';
    }

    /**
     * Set Privacy Header
     *
     * @param string|null $privacyHeader Privacy Header.
     *
     * @return SettingsPbx
     */
    public function setPrivacyHeader(?string $privacyHeader): SettingsPbx {
        $this->_privacyHeader = $privacyHeader;

        return $this;
    }

    /**
     * Get Qualify
     *
     * @return boolean
     */
    public function getQualify(): bool {
        return $this->_qualify;
    }

    /**
     * Set Qualify
     *
     * @param boolean $qualify Qualify.
     *
     * @return SettingsPbx
     */
    public function setQualify(bool $qualify = false): SettingsPbx {
        $this->_qualify = $qualify;

        return $this;
    }

    /**
     * Get Payload 2833 rfc
     *
     * @return integer
     */
    public function getRfcPayload(): int {
        return $this->_rfcPayload;
    }

    /**
     * Set Payload 2833 rfc
     *
     * @param integer $rfcPayload Payload 2833 rfc.
     *
     * @return SettingsPbx
     */
    public function setRfcPayload(int $rfcPayload): SettingsPbx {
        $this->_rfcPayload = $rfcPayload;

        return $this;
    }

    /**
     * Get Rport
     *
     * @return string
     */
    public function getRport(): string {
        return $this->_rport;
    }

    /**
     * Set Rport
     *
     * @param string $rPort Rport.
     *
     * @return SettingsPbx
     */
    public function setRport(string $rPort): SettingsPbx {
        if ($rPort == '') {
            $rPort = 'invite';
        }
        $this->_rport = $rPort;

        return $this;
    }

    /**
     * Get Password
     *
     * @return string
     */
    public function getSecret(): string {
        return $this->_secret ?? '';
    }

    /**
     * Set Password
     *
     * @param string|null $secret Password.
     *
     * @return SettingsPbx
     */
    public function setSecret(?string $secret): SettingsPbx {
        $this->_secret = $secret;

        return $this;
    }

    /**
     * Get Session Timer max
     *
     * @return integer
     */
    public function getSessionExpires(): int {
        return $this->_sessionExpires;
    }

    /**
     * Set Session Timer max
     *
     * @param integer $sessionExpires Session Timer max.
     *
     * @return SettingsPbx
     */
    public function setSessionExpires(int $sessionExpires): SettingsPbx {
        $this->_sessionExpires = $sessionExpires;

        return $this;
    }

    /**
     * Get Session Timer min
     *
     * @return integer
     */
    public function getSessionMinse(): int {
        return $this->_sessionMinse;
    }

    /**
     * Set Session Timer min
     *
     * @param integer $sessionMinse Session Timer min.
     *
     * @return SettingsPbx
     */
    public function setSessionMinse(int $sessionMinse): SettingsPbx {
        $this->_sessionMinse = $sessionMinse;

        return $this;
    }

    /**
     * Get Session Timer enabled
     *
     * @return boolean
     */
    public function getSessionTimer(): bool {
        return $this->_sessionTimer;
    }

    /**
     * Set Session Timer enabled
     *
     * @param boolean $sessionTimer Session Timer enabled.
     *
     * @return SettingsPbx
     */
    public function setSessionTimer(bool $sessionTimer = false): SettingsPbx {
        $this->_sessionTimer = $sessionTimer;

        return $this;
    }

    /**
     * Get Force static SSRC
     *
     * @return boolean
     */
    public function getSsrcChange(): bool {
        return $this->_ssrcChange;
    }

    /**
     * Set Force static SSRC
     *
     * @param boolean $ssrcChange Force static SSRC.
     *
     * @return SettingsPbx
     */
    public function setSsrcChange(bool $ssrcChange = false): SettingsPbx {
        $this->_ssrcChange = $ssrcChange;

        return $this;
    }

    /**
     * Get Symmetric Rtp
     *
     * @return boolean
     */
    public function getSymmetricRtp(): bool {
        return $this->_symmetricRtp;
    }

    /**
     * Set Symmetric Rtp
     *
     * @param boolean $symmetricRtp Symmetric Rtp.
     *
     * @return SettingsPbx
     */
    public function setSymmetricRtp(bool $symmetricRtp = true): SettingsPbx {
        $this->_symmetricRtp = $symmetricRtp;

        return $this;
    }

    /**
     * Get T38 enabled
     *
     * @return boolean
     */
    public function getThirtyEightPtUdptl(): bool {
        return $this->_thirtyEightPtUdptl;
    }

    /**
     * Set T38 enabled
     *
     * @param boolean $tThirtyEightPtUdptl T38 enabled.
     *
     * @return SettingsPbx
     */
    public function setThirtyEightPtUdptl(bool $tThirtyEightPtUdptl): SettingsPbx {
        $this->_thirtyEightPtUdptl = $tThirtyEightPtUdptl;

        return $this;
    }

    /**
     * Get _thirtyEightPtErrorCorrection
     *
     * @return string
     */
    public function getThirtyEightPtErrorCorrection(): string {
        return $this->_thirtyEightPtErrorCorrection;
    }

    /**
     * Set T38 error correction
     *
     * @param string $thirtyEightPtErrorCorrection T38 error correction.
     *
     * @return SettingsPbx
     */
    public function setThirtyEightPtErrorCorrection(string $thirtyEightPtErrorCorrection): SettingsPbx {
        if ($thirtyEightPtErrorCorrection == '') {
            $thirtyEightPtErrorCorrection = 'redundancy';
        }
        $this->_thirtyEightPtErrorCorrection = $thirtyEightPtErrorCorrection;

        return $this;
    }

    /**
     * Get T38 Maxdatagram
     *
     * @return integer
     */
    public function getThirtyEightPtMaxdatagram(): int {
        return $this->_thirtyEightPtMaxdatagram;
    }

    /**
     * Set T38 Maxdatagram
     *
     * @param integer $thirtyEightPtMaxdatagram T38 Maxdatagram.
     *
     * @return SettingsPbx
     */
    public function setThirtyEightPtMaxdatagram(int $thirtyEightPtMaxdatagram): SettingsPbx {
        $this->_thirtyEightPtMaxdatagram = $thirtyEightPtMaxdatagram;

        return $this;
    }

    /**
     * Get Tone Zone
     *
     * @return string
     */
    public function getToneZone(): string {
        return $this->_toneZone ?? '';
    }

    /**
     * Set Tone Zone
     *
     * @param string|null $toneZone Tone Zone.
     *
     * @return SettingsPbx
     */
    public function setToneZone(?string $toneZone): SettingsPbx {
        $this->_toneZone = $toneZone;

        return $this;
    }

    /**
     * Get Incoming CID
     *
     * @return boolean
     */
    public function getTrustRpid(): bool {
        return $this->_trustRpid;
    }

    /**
     * Set Incoming CID
     *
     * @param boolean $trustRpid Incoming CID.
     *
     * @return SettingsPbx
     */
    public function setTrustRpid(bool $trustRpid = false): SettingsPbx {
        $this->_trustRpid = $trustRpid;

        return $this;
    }

    /**
     * Get Registration Proxy enabled
     *
     * @return string
     */
    public function getType(): string {
        return $this->_type;
    }

    /**
     * Set Registration Proxy enabled
     *
     * @param string $type Registration Proxy enabled.
     *
     * @return SettingsPbx
     */
    public function setType(string $type): SettingsPbx {
        if ($type == '') {
            $type = 'friend';
        }
        $this->_type = $type;

        return $this;
    }

    /**
     * Get Auth Login
     *
     * @return string
     */
    public function getUserName(): string {
        return $this->_userName;
    }

    /**
     * Set Auth Login
     *
     * @param string $userName Auth Login.
     *
     * @return SettingsPbx
     */
    public function setUserName(string $userName): SettingsPbx {
        $this->_userName = $userName;

        return $this;
    }

    /**
     * Get Video codecs
     *
     * @return boolean
     */
    public function getVideoCodecs(): bool {
        return $this->_videoCodecs ?? false;
    }

    /**
     * Set Video codecs
     *
     * @param boolean $videoCodecs Video codecs.
     *
     * @return SettingsPbx
     */
    public function setVideoCodecs(bool $videoCodecs = false): SettingsPbx {
        $this->_videoCodecs = $videoCodecs;

        return $this;
    }

    /**
     * Get _playRemoteMoh
     *
     * @return boolean
     */
    public function isPlayRemoteMoh(): bool {
        return $this->_playRemoteMoh;
    }

    /**
     * Set playRemoteMoh
     *
     * @param boolean $playRemoteMoh Param playRemoteMoh true/false.
     *
     * @return SettingsPbx
     */
    public function setPlayRemoteMoh(bool $playRemoteMoh): SettingsPbx {
        $this->_playRemoteMoh = $playRemoteMoh;

        return $this;
    }

    /**
     * Get Contact Number
     *
     * @return string
     */
    public function getContactNumber(): string {
        return $this->_contactNumber;
    }

    /**
     * Set contactNumber
     *
     * @param string $contactNumber Param contactNumber.
     *
     * @return SettingsPbx
     */
    public function setContactNumber(string $contactNumber): SettingsPbx {
        $this->_contactNumber = $contactNumber;

        return $this;
    }

    public function toJson($options = 0): string {
        $params = array_filter(
            [
                'disallow'           => $this->getDisallow(),
                'allow'              => $this->getAllow(),
                'allowtransfer'      => $this->getAllowTransfer() ? 'on' : 'no',
                'cidbody'            => $this->getCidBody(),
                'cidheader'          => $this->getCidHeader(),
                'context'            => $this->getContext(),
                'diversion_header'   => $this->getDiversionHeader(),
                'dtmfmode'           => $this->getDtmfMode(),
                'force_srtp'         => $this->getForceSrtp() ? 'yes' : 'no',
                'fromdomain'         => $this->getFromDomain(),
                'fromname'           => $this->getFromName(),
                'fromnumber'         => $this->getFromNumber(),
                'fromuser'           => $this->getFromUser(),
                'historyinfo_header' => $this->getHistoryInfoHeader(),
                'host'               => $this->getHost(),
                'insecure'           => $this->getInsecure(),
                'language'           => $this->getLanguage(),
                'peername'           => $this->getPeername(),
                'port'               => $this->getPort(),
                'preserve_orig_cid'  => $this->getPreserveOrigCid() ? 'yes' : 'no',
                'privacyheader'      => $this->getPrivacyHeader(),
                'qualify'            => $this->getQualify() ? 'yes' : 'no',
                'rport'              => $this->getRport(),
                'secret'             => $this->getSecret(),
                'session-expires'    => (string)$this->getSessionExpires(),
                'session-minse'      => (string)$this->getSessionMinse(),
                'session_timer'      => $this->getSessionTimer() ? 'on' : 'no',
                'ssrcchange'         => $this->getSsrcChange() ? 'no' : 'yes',
                'symmetric_rtp'      => $this->getSymmetricRtp() ? 'yes' : 'no',
                't38pt_udptl'        => $this->getThirtyEightString(),
                'tonezone'           => $this->getToneZone(),
                'transport_wildix'   => $this->getTransportWildix(),
                'trustrpid'          => $this->getTrustRpid() ? 'yes' : 'no',
                'type'               => $this->getType(),
                'username'           => $this->getUserName(),
                'play_remote_moh'    => $this->isPlayRemoteMoh() ? 'on' : 'off',
                'contactnumber'      => $this->getContactNumber(),
            ]
        );

        if ($this->isDtmfModeRfc2033()) {
            $params['rfc2833payload'] = (string)$this->getRfcPayload();
        }

        $params['pricelist']    = (string)$this->getPriceList();
        $params['country_code'] = $this->getCountryCode() ? (string)$this->getCountryCode() : '';

        return json_encode($params, $options);
    }

    private function isDtmfModeRfc2033(): bool {
        return $this->getDtmfMode() === 'rfc2833';
    }

    /**
     * Get _allow
     *
     * @return string
     */
    public function getAllow(): string {
        return $this->_allow;
    }

    /**
     * Set Allow
     *
     * @param string $allow Allow.
     *
     * @return SettingsPbx
     */
    public function setAllow(string $allow): SettingsPbx {
        $this->_allow = $allow;

        return $this;
    }

    /**
     * Make T38 string
     *
     * @return string
     */
    public function getThirtyEightString(): string {
        return $this->getThirtyEightPtUdptl() ? sprintf('yes,maxdatagram=%s,%s', $this->getThirtyEightPtMaxdatagram(), $this->getThirtyEightPtErrorCorrection()) : 'no';
    }
}
