<?php

declare(strict_types=1);

namespace Api\Trunks\Sip\Objects;

use Api\Support\Interfaces\ArrayableInterface;
use Api\Support\Interfaces\Jsonable;
use Services\Helpers;

/**
 * Class SettingsWms
 */
class SettingsWms implements Jsonable, ArrayableInterface {

    /**
     * Host Name
     *
     * @var string
     */
    private $_regHost;

    /**
     * Host Port
     *
     * @var string
     */
    private $_regPort;

    /**
     * Username
     *
     * @var string
     */
    private $_regUsername;

    /**
     * Password
     *
     * @var string
     */
    private $_regSecret;

    /**
     * From user
     *
     * @var string
     */
    private $_regFromUser;

    /**
     * Enable registration
     *
     * @var boolean
     */
    private $_enableReg;

    /**
     * Overwrite registration
     *
     * @var boolean
     */
    private $_overwriteReg;

    /**
     * Outbound proxy
     *
     * @var string
     */
    private $_outboundProxy;

    /**
     * Outbound proxy port
     *
     * @var string
     */
    private $_outboundProxyPort;

    /**
     * Registration Expiry (sec)
     *
     * @var string
     */
    private $_expiry;

    /**
     * Custom DNS Server
     *
     * @var string
     */
    private $_dnsServer;

    /**
     * NAT IP
     *
     * @var string
     */
    private $_natIp;

    public function isOverwriteReg(): bool {
        return $this->getOverwriteReg() === Helpers::YES;
    }

    /**
     * Get Overwrite registration
     *
     * @return boolean
     */
    public function getOverwriteReg(): bool {
        return $this->_overwriteReg;
    }

    /**
     * Set Overwrite registration
     *
     * @param boolean $overwriteReg Overwrite registration.
     *
     * @return SettingsWms
     */
    public function setOverwriteReg(bool $overwriteReg = false): SettingsWms {
        $this->_overwriteReg = $overwriteReg;

        return $this;
    }

    public function toArray(): array {
        return [
            'dnsServer' => $this->getDnsServer(),
            'enableReg' => $this->getEnableReg(),
            'expiry' => $this->getExpiry(),
            'natIp' => $this->getNatIp(),
            'outboundProxy' => $this->getOutboundProxy(),
            'outboundProxyPort' => $this->getOutboundProxyPort(),
            'overwriteReg' => $this->getOverwriteReg(),
            'regFromUser' => $this->getRegFromUser(),
            'regHost' => $this->getRegHost(),
            'regPort' => $this->getRegPort(),
            'regSecret' => $this->getRegSecret(),
            'regUsername' => $this->getRegUsername(),

        ];
    }

    /**
     * Get Custom DNS Server
     *
     * @return string
     */
    public function getDnsServer(): string {
        return $this->_dnsServer ?? '';
    }

    /**
     * Set Custom DNS Server
     *
     * @param string $dnsServer Custom DNS Server.
     *
     * @return SettingsWms
     */
    public function setDnsServer(string $dnsServer): SettingsWms {
        $this->_dnsServer = $dnsServer;

        return $this;
    }

    /**
     * Get Enable registration
     *
     * @return boolean
     */
    public function getEnableReg(): bool {
        return $this->_enableReg ?? false;
    }

    /**
     * Set Enable registration
     *
     * @param boolean $enableReg Enable registration.
     *
     * @return SettingsWms
     */
    public function setEnableReg(bool $enableReg = true): SettingsWms {
        $this->_enableReg = $enableReg;

        return $this;
    }

    /**
     * Get Registration Expiry (sec)
     *
     * @return string
     */
    public function getExpiry(): string {
        return $this->_expiry ?? '';
    }

    /**
     * Set Registration Expiry (sec)
     *
     * @param string $expiry Registration Expiry (sec).
     *
     * @return SettingsWms
     */
    public function setExpiry(string $expiry): SettingsWms {
        $this->_expiry = $expiry;

        return $this;
    }

    /**
     * Get NAT IP
     *
     * @return string
     */
    public function getNatIp(): string {
        return $this->_natIp ?? '';
    }

    /**
     * Set NAT IP
     *
     * @param string $natIp NAT IP.
     *
     * @return SettingsWms
     */
    public function setNatIp(string $natIp): SettingsWms {
        $this->_natIp = $natIp;

        return $this;
    }

    /**
     * Get Outbound proxy
     *
     * @return string
     */
    public function getOutboundProxy(): string {
        return $this->_outboundProxy ?? '';
    }

    /**
     * Set Outbound proxy
     *
     * @param string $outboundProxy Outbound proxy.
     *
     * @return SettingsWms
     */
    public function setOutboundProxy(string $outboundProxy): SettingsWms {
        $this->_outboundProxy = $outboundProxy;

        return $this;
    }

    /**
     * Get Outbound proxy port
     *
     * @return string
     */
    public function getOutboundProxyPort(): string {
        return $this->_outboundProxyPort ?? '';
    }

    /**
     * Set Outbound proxy port
     *
     * @param string $outboundProxyPort Outbound proxy port.
     *
     * @return SettingsWms
     */
    public function setOutboundProxyPort(string $outboundProxyPort): SettingsWms {
        $this->_outboundProxyPort = $outboundProxyPort;

        return $this;
    }

    /**
     * Get From user
     *
     * @return string
     */
    public function getRegFromUser(): string {
        return $this->_regFromUser ?? '';
    }

    /**
     * Set From user
     *
     * @param string $regFromUser From user.
     *
     * @return SettingsWms
     */
    public function setRegFromUser(string $regFromUser): SettingsWms {
        $this->_regFromUser = $regFromUser;

        return $this;
    }

    /**
     * Get Reg Host
     *
     * @return string
     */
    public function getRegHost(): string {
        return $this->_regHost ?? '';
    }

    /**
     * Set Host Name
     *
     * @param string $regHost Host Name.
     *
     * @return SettingsWms
     */
    public function setRegHost(string $regHost): SettingsWms {
        $this->_regHost = $regHost;

        return $this;
    }

    /**
     * Get Host Port
     *
     * @return string
     */
    public function getRegPort(): string {
        return $this->_regPort ?? '';
    }

    /**
     * Set Host Port
     *
     * @param string $regPort Host Port.
     *
     * @return SettingsWms
     */
    public function setRegPort(string $regPort): SettingsWms {
        $this->_regPort = $regPort;

        return $this;
    }

    /**
     * Get Password
     *
     * @return string
     */
    public function getRegSecret(): string {
        return $this->_regSecret ?? '';
    }

    /**
     * Set Password
     *
     * @param string $regSecret Password.
     *
     * @return SettingsWms
     */
    public function setRegSecret(string $regSecret): SettingsWms {
        $this->_regSecret = $regSecret;

        return $this;
    }

    /**
     * Get Username
     *
     * @return string
     */
    public function getRegUsername(): string {
        return $this->_regUsername ?? '';
    }

    /**
     * Set Username
     *
     * @param string $regUsername Username.
     *
     * @return SettingsWms
     */
    public function setRegUsername(string $regUsername): SettingsWms {
        $this->_regUsername = $regUsername;

        return $this;
    }

    /**
     * Object to JSON
     *
     * @param integer $options Bitmask consisting.
     *
     * @return string
     */
    public function toJson($options = 0): string {
        return json_encode(
            array_merge(
                [
                'dnsserver' => $this->getDnsServer(),
                'enable_reg' => $this->getEnableReg() ? 'yes' : 'no',
                'expiry' => (string)$this->getExpiry(),
                'natip' => $this->getNatIp(),
                'outboundproxy' => $this->getOutboundProxy(),
                'outboundproxy_port' => $this->getOutboundProxyPort(),
                'overwrite_reg' => $this->getOverwriteReg() ? 'yes' : null,
                ],
                array_filter(
                    [
                    'reg_fromuser' => $this->getRegFromUser(),
                    'reg_host' => $this->getRegHost(),
                    'reg_port' => $this->getRegPort(),
                    'reg_secret' => $this->getRegSecret(),
                    'reg_username' => $this->getRegUsername(),
                    ]
                )
            ),
            $options
        );
    }
}
