<?php

declare(strict_types=1);

namespace Api\Trunks\Sip\Service;

use Api\Device\Interfaces\DeviceRepositoryInterface;
use Api\Translation\Interfaces\TranslationStrategyInterface;
use Api\Translation\Objects\TranslationInterface;
use Api\Trunks\Interfaces\TrunkConfigCwServiceInterface;
use Api\Trunks\Interfaces\TrunkRepositoryInterface;
use Api\Trunks\PriceList\Interfaces\PriceListCardRepositoryInterface;
use Api\Trunks\Pstn\Interfaces\PeerServiceInterface;
use Api\Trunks\Service\TrunkService;
use Api\Trunks\Sip\Builder\SettingsBuilder;
use Api\Trunks\Sip\Builder\SettingsKamailioBuilder;
use Api\Trunks\Sip\Interfaces\SipTrunkServiceInterface;
use Api\Trunks\Sip\Objects\SettingsKamailio;
use Api\Trunks\Sip\Objects\Trunk;
use Api\Trunks\Sip\Objects\Trunks;
use Logger\Log;
use PBX;
use SQLite3;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\SipProxy;
use Wildix\Slim\Slim;
use Wildix\Validator;

/**
 * Class SipService
 */
class SipTrunkService extends TrunkService implements SipTrunkServiceInterface {

    private $translation;

    public function __construct(
        TrunkRepositoryInterface $_trunkRepository,
        DeviceRepositoryInterface $deviceRepository,
        PeerServiceInterface $peerService,
        TrunkConfigCwServiceInterface $trunkConfigCwService,
        PriceListCardRepositoryInterface $priceListCardRepository,
        SipProxy $sipProxy,
        Log $logger,
        PBX $pbx,
        TranslationInterface $translation
    ) {
        parent::__construct(
            $_trunkRepository,
            $deviceRepository,
            $peerService,
            $trunkConfigCwService,
            $priceListCardRepository,
            $sipProxy,
            $logger,
            $pbx
        );

        $this->translation = $translation;
    }

    public function create(array $settings): int {
        $this->validateData($settings);

        $sip = SettingsBuilder::build($settings);

        $id = $this->getTrunkRepository()->createSip($sip);

        $this->createPeerBySip($id, $sip);

        $this->updateConfigs();

        return $id;
    }

    private function validateData(array $params): void {
        $validator = new Validator($params);

        $booleanFields = [
            'settingsPbx.qualify',
            'settingsWms.enableReg',
            'settingsPbx.videoCodecs',
            'settingsPbx.thirtyEightPtUdptl',
            'settingsPbx.trustRpid',
            'settingsPbx.preserveOrigCid',
            'settingsPbx.allowTransfer',
            'settingsPbx.playRemoteMoh',
            'settingsPbx.sessionTimer',
            'settingsPbx.ssrcChange',
            'settingsPbx.forceSrtp',
            'settingsWms.overwriteReg',
        ];
        foreach ($booleanFields as $field) {
            $validator->rule('strictBoolean', $field)->reason('Parameter must be boolean');
        }

        $validator->rule('required', ['settingsPbx', 'settingsWms']);

        $validator->rule('integer', 'id');

        $validator->rule('in', 'settingsPbx.priceList', array_merge($this->getExistPriceListId(), [0]))
            ->reason($this->getTranslate()->t('Price list not found'));

        $validator->rule('regex', 'title', '/^[a-zA-Zàèéêìíîòóôùú0-9!#\\$\\^&\\*\\(\\)_\\-\\+=`\\{\\}\\|\"\\?\\/\\. ]+$/')
            ->reason($this->getTranslate()->t('rules.enterValidValue'));
        $validator->rule('required', 'title');

        $validator->rule('regex', 'settingsPbx.userName', '/^[a-zA-Z0-9 !#@\\$\\^&\\*\\(\\)_\\-\\+=`\\{\\}\\|\\"\\?\\/\\.]+$/')
            ->reason($this->getTranslate()->t('rules.enterValidValue') . ' A-z0-9 _');
        $validator->rule('required', 'settingsPbx.userName');

        $validator->rule('notIn', 'settingsPbx.host', ['localhost', '127.0.0.1'])
            ->reason($this->getTranslate()->t('trunks.err.ThereIsNoProvideLocalhost'));

        $validator->rule('ipPort', 'settingsPbx.port');

        $validator->rule('regex', ['settingsPbx.secret'], '/^[^:\?]+$/')
            ->reason(sprintf($this->getTranslate()->t('strings.notAllowedCharsPresent'), '(:&#44; ?)'));

        $validator->rule('required', ['settingsPbx.context'])->message('Set valid dialplan');

        $validator->rule('countryCode', 'settingsPbx.countryCode')->message(Error::CountryCode);

        $validator->rule('array', 'settingsPbx.audioCodecs');
        $ruleAudioCodecs = function ($field, $codecs) {
            return (bool)array_filter(
                $codecs,
                function ($codec) {
                    return $codec['enable'];
                }
            );
        };
        $validator->rule($ruleAudioCodecs, 'settingsPbx.audioCodecs')
            ->reason($this->getTranslate()->t('trunks.err.selectCodec'))
            ->message(Error::FieldRequired);

        if ($params['settingsPbx']['thirtyEightPtUdptl']) {
            $validator->rule('in', ['settingsPbx.thirtyEightPtErrorCorrection'], ['redundancy', 'off', 'fec'])
                ->reason('Must contain "redundancy", "off", "fec"');

            $validator->rule('between', ['settingsPbx.thirtyEightPtMaxDatagram'], [0, 1500])
                ->reason('Numeric between 0-1500');
        }

        if ($params['settingsPbx']['sessionTimer']) {
            $sessionRanges = [
                'settingsPbx.sessionMinse',
                'settingsPbx.sessionExpires',
            ];
            $validator->rule('between', $sessionRanges, [90, 20000])->reason('Numeric between 90-20000');

            $maxMoreMin = (int)$params['settingsPbx']['sessionExpires'] < (int)$params['settingsPbx']['sessionMinse'];
            $validator->rule('boolean', $maxMoreMin)
                ->reason($this->getTranslate()->t('strings.enterValidValue') . ' : max > min');
        }

        $validator->rule('in', ['settingsPbx.rport'], ['invite', 'register,invite', 'off', 'all'])
            ->reason('Must contain "invite", "register,invite", "off"');

        $validator->rule('between', ['settingsWms.expiry'], [0, 86400])->reason('Numeric between 0-86400');

        $validator->rule('domainOrIp', 'settingsWms.dnsServer');

        $validator->rule('ipPort', 'settingsWms.outboundProxyPort');

        $validator->rule('ip', 'settingsWms.natIp');

        $validator->rule('in', ['settingsPbx.transportWildix'], ['udp', 'tcp', 'tls', 'dns-srv'])
            ->reason('Must contain "udp", "tcp", "tls", "dns-srv"');

        $validator->rule('in', ['settingsPbx.dtmfMode'], ['rfc2833', 'info', 'inband'])
            ->reason('Must contain "rfc2833","info","inband"');

        if ($params['settingsPbx']['dtmfMode'] == 'rfc2833') {
            $validator->rule('required', ['settingsPbx.rfcPayload']);
            $validator->rule('between', ['settingsPbx.rfcPayload'], [0, 127])->reason('Numeric between 0-127');
        }

        if ($params['settingsWms']['overwriteReg']) {
            $validator->rule('required', 'settingsWms.regUsername');
            $validator->rule('ipPort', 'settingsWms.regPort');
        }

        $errorsTrunkExist = $this->getTrunkRepository()->isExistSip($params);

        if (!$validator->validate() || $errorsTrunkExist) {
            $errors = array_merge_recursive($validator->errorsTree(), $errorsTrunkExist);
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $errors);
        }
    }

    public function createPeer(
        int $id,
        string $name,
        SettingsKamailio $settings,
        bool $isDynamicHost,
        bool $reloadPeersAndStoreIp = true
    ): void {
        $this->deletePeerById($id);

        $this->saveBasePeer($id, $settings, $this->replaceSpacesToUrlFormat($name), $isDynamicHost);

        $this->savePeerByUserName($id, $settings, $this->replaceSpacesToUrlFormat($name), $isDynamicHost);

        if ($isDynamicHost) {
            $this->savePeerByAuthName($id, $settings);
            $this->savePeerByNatIp($id, $settings);
        }

        if ($reloadPeersAndStoreIp) {
            $this->reloadPeersAndStoreIp();
        }
    }

    private function saveBasePeer(int $id, SettingsKamailio $settings, string $name, bool $isDynamicHost): void {
        $preparedSettings = $this->preparePeerSettings($settings->toArray());

        foreach ($preparedSettings as $key => $value) {
            $data = [
                $this->getDynamicPeerWmsId($id),
                SQLite3::escapeString($name . '::' . $key),
                SQLite3::escapeString($value),
            ];

            $this->saveToDb($data, $isDynamicHost);
        }
    }

    private function savePeerByUserName(int $id, SettingsKamailio $settings, string $name, bool $isDynamicHost): void {
        if (empty($settings->getUsername()) || $this->replaceSpacesToUrlFormat($settings->getUsername()) === $name) {
            return;
        }

        $userName         = $settings->getFromUser() ?: $settings->getUsername();
        $preparedSettings = $this->preparePeerSettings($settings->toArray());

        foreach ($preparedSettings as $key => $value) {
            $data = [
                $this->getDynamicPeerWmsId($id),
                SQLite3::escapeString($name . '::' . $userName . '::' . $key),
                SQLite3::escapeString($value),
            ];

            $this->saveToDb($data, $isDynamicHost);

            if ($key === 'outboundproxy') {
                $proxyData = [
                    $this->getDynamicPeerWmsId($id),
                    SQLite3::escapeString($value . '::' . $userName . '::' . $key),
                    SQLite3::escapeString($value),
                ];

                $this->saveToDb($proxyData, $isDynamicHost);
            }
        }
    }

    private function savePeerByAuthName(int $id, SettingsKamailio $settings): void {
        if ($settings->getPeerName() === $settings->getUsername()) {
            return;
        }

        $data = [
            $this->getDynamicPeerWmsId($id),
            SQLite3::escapeString($settings->getPeerName() . '::authname'),
            SQLite3::escapeString($settings->getUsername()),
        ];

        $this->getPeerService()->createDynamicPeer(...$data);
    }

    private function savePeerByNatIp(int $id, SettingsKamailio $settings): void {
        if ($settings->getNatIp()) {
            $data = [
                $this->getDynamicPeerWmsId($id),
                SQLite3::escapeString($settings->getPeerName() . '::natip'),
                $settings->getNatIp(),
            ];

            $this->getPeerService()->createDynamicPeer(...$data);
        }
    }

    private function preparePeerSettings(array $settings): array {
        $preparedSettings = array_filter($settings, 'strlen');

        array_walk(
            $preparedSettings,
            function (&$value, $key) {
                if ($key === 'authname' || $key === 'username' || $key === 'peername') {
                    $value = $this->replaceSpacesToUrlFormat($value);
                }
            }
        );

        return $preparedSettings;
    }

    private function saveToDb(array $data, bool $isDynamicHost): void {
        $isDynamicHost ? $this->getPeerService()->createDynamicPeer(...$data) : $this->getPeerService()->createStaticPeer(...$data);
    }

    private function replaceSpacesToUrlFormat(string $string): string {
        return str_replace(' ', '%20', $string);
    }

    private function getDynamicPeerWmsId(int $id): int {
        return intval('10' . $id);
    }

    private function reloadPeersAndStoreIp(): void {
        $this->getLogger()->info(__LINE__, 'Sip trunk reload Peers And Store Ip');
        exec('sudo /usr/sbin/kamcmd htable.reload dynamic_peers');
        exec('sudo /usr/sbin/kamcmd htable.reload static_peers');
        exec('curl -X POST --url http://127.0.0.1:5060/v1/trunks/static/ip/store/');
    }

    public function findOrFail(int $id): Trunk {
        return $this->getTrunkRepository()->findOrFailSip($id);
    }

    public function getItems(): Trunks {
        return $this->getTrunkRepository()->getSipItems();
    }

    public function remove(int $id): bool {
        $this->getLogger()->info(__LINE__, sprintf('Remove peers by id %s', $id));
        $sip = $this->getTrunkRepository()->findOrFailSip($id);

        $result = $this->getTrunkRepository()->removeSip($id);

        $this->deletePeerById($id, $sip->getSettingsPbx()->isDynamicHost());

        $this->reloadPeersAndStoreIp();

        $this->updateConfigs();

        return $result;
    }

    public function update(int $id, array $settings): bool {
        $this->getTrunkRepository()->findOrFailSip($id);

        $this->validateData($settings);

        $sip = SettingsBuilder::build($settings);

        $sipSettings = $this->getTrunkRepository()->updateSip($id, $sip);

        $this->createPeerBySip($id, $sip);

        $this->updateConfigs();

        return $sipSettings;
    }

    private function createPeerBySip(int $id, Trunk $sip): void {
        $kamailio = SettingsKamailioBuilder::build($sip->getSettingsPbx(), $sip->getSettingsWms());

        if ($sip->getSettingsPbx()->isDynamicHost()) {
            $this->createPeer($id, $sip->getSettingsPbx()->getUserName(), $kamailio, true);
        } else {
            $this->createPeer($id, $sip->getSettingsPbx()->getPeerName(), $kamailio, false);
        }
    }

    private function deletePeerById(int $id): void {
        $this->getPeerService()->deleteDynamicPeer($this->getDynamicPeerWmsId($id));
        $this->getPeerService()->deleteStaticPeer($this->getDynamicPeerWmsId($id));
    }

    private function getTranslate(): TranslationInterface {
        return $this->translation;
    }

    private function getExistPriceListId(): array {
        $priceList = $this->getPriceListCardRepository()->getAll();
        $result    = [];
        foreach ($priceList->all() as $priceItem) {
            $result[] = $priceItem->getId();
        }
        return $result;
    }

    public function isUseTrunk(array $dialPlanTrunk, int $id): bool {
        $trunk = $this->findOrFail($id);

        foreach ($dialPlanTrunk as $name => $dialPlan) {
            if ($trunk->getSettingsPbx()->getPeerName() == $dialPlan) {
                throw new InvalidDataException(sprintf("You can't remove %s. It's linked to %s", $trunk->getTitle(), $name), [], 400);
            }
        }

        return false;
    }
}
