<?php
class BackgroundTask{

	const LAUNCHED        	  = 'launched';
	const NOT_LAUNCHED        = 'not_launched';
	const DONE        		  = 'done';

	const PID_PATH = '/dev/shm/wmsdb.d/BackgroundTask';
	const OUTPUT_PATH = '/tmp/BackgroundTask';

	private $_logLevel = 1;
	private $_args = array();
	private $_id = null;
	private $_autoDelete = false;
	private $_autoDeletePid = false;

	public function __construct($worker = null) {
		if(!is_dir(self::PID_PATH)){
			mkdir(self::PID_PATH, 0777);
			chmod(self::PID_PATH, 0777);
		}

		if(!is_dir(self::OUTPUT_PATH)){
			mkdir(self::OUTPUT_PATH, 0777);
			chmod(self::OUTPUT_PATH, 0777);
		}

		register_shutdown_function(array($this, 'destroy'));

		if($worker){
			$this->setArgs(func_get_args());
		}
    }

    public function getStatus(){
    	if(file_exists($this->getPIDFile())){
    		return self::LAUNCHED;
    	}

    	$output = $this->getOutput();
    	if(is_array($output) && isset($output['result'])){
    		return self::DONE;
    	}
    	return self::NOT_LAUNCHED;
    }

    public function autoDelete(){
    	$this->_autoDelete = true;
    }

    public function delete(){
    	$this->deleteOutput();
    }

    private $_output = null;
    public function getOutput(){
    	if($this->_output === null){
    		$this->_output = array();
    		if(file_exists($this->getOutputFile())){
    			$content = file_get_contents($this->getOutputFile());
    			if($content != ''){
    				$this->_output = json_decode($content, true);
    				if($this->_output === null || $this->_output == ''){
    					$this->_output = array();
    				}
    			}
    		}
    	}
    	return $this->_output;
    }

    /**
     * run async
     * @return boolean
     */
    public function runAsync(){

    	$this->getLog()->info(__LINE__, 'Run background task "'. $this->_args[0].'"');

    	$status = $this->getStatus();
    	if($status == self::DONE){
    		$this->getLog()->info(__LINE__, 'Status of a background task "DONE"');
    		$this->deleteOutput();
    	}else if($status == self::LAUNCHED){
    		$this->getLog()->info(__LINE__, 'Status of a background task "DONE"');
    		return false;
    	}

    	$args = array_map('escapeshellarg', $this->_args);
    	$result = exec('php /usr/sbin/background_task.php '.join(' ', $args).' 1>/dev/null 2>/dev/null &');
    	if($result == ''){
    		return true;
    	}
    	return false;
    }

    /**
     * run sync
     */
    public function runSync($worker = null){
    	$this->setArgs(func_get_args());

    	$this->getLog()->info(__LINE__, 'Start progress background task "'. $this->_args[0].'"');

    	$status = $this->getStatus();
    	if($status == self::DONE){
    		$this->getLog()->info(__LINE__, 'Status of a background task "DONE"');
    		$this->deleteOutput();
    	}else if($status == self::LAUNCHED){
    		$this->getLog()->info(__LINE__, 'Status of a background task "LAUNCHED"');
    		return false;
    	}


    	$this->createPIDFile();

    	$class = null;
    	$method = null;
    	if(!empty($this->_args[0]) && strpos($this->_args[0], ':') > 0 ){
    		$exploded = explode(':',$this->_args[0]);
    		if(!empty($exploded[0]) && !empty($exploded[1])){
    			$class = $exploded[0];
    			$method = $exploded[1];
    		}
    	}

    	if($class == null || $method == null){
    		$msg = 'Bad command line syntax';
    		$this->getLog()->error(__LINE__, $msg);
    		die($msg);
    	}

    	if (!class_exists($class, true)) {
    		$msg = 'Сlass "'.$class.'" does not exist';
    		$this->getLog()->error(__LINE__, $msg);
    		die($msg);
    	}

    	$call_func = null;
    	if(method_exists($class, $method)){
    		// static method
    		$call_func = $class.'::'.$method;
    		$this->getLog()->info(__LINE__, 'Found static method "'.$call_func.'"');
    	}else {
    		// create object
    		$obj = null;
    		if(method_exists($class, 'getInstance')){
    			$this->getLog()->info(__LINE__, 'Found getInstance method');
    			$obj = $class::getInstance();
    		}else{
    			$this->getLog()->info(__LINE__, 'Сreating an object');
    			$obj = new $class;
    		}

    		if($obj && method_exists($obj, $method)){
    			$call_func = array($obj, $method);
    		}else{
    			$msg = 'Method "'.$class.':'.$method.'" does not exist';
    			$this->getLog()->error(__LINE__, $msg);
    			die($msg);
    		}
    	}

    	if(!empty($call_func)){
    		$params = array_slice($this->_args, 1);

    		$result = call_user_func_array($call_func, $params);
    		$this->getLog()->info(__LINE__, 'Result of a background task: '. print_r($result, true));
    		$this->setResult($result);
    	}

    	return true;
    }


    private $_log = null;
    private function getLog(){
    	if($this->_log === null){
    		$this->_log = new \Syslog('\BackgroundTask', $this->_logLevel);
    	}
    	return $this->_log;
    }

    private function setArgs($args){
    	$this->_args = $args;
    }

    private function getTaskId(){
		if($this->_id === null){
			$this->_id = md5($this->_args[0]);
		}
		return $this->_id;
    }

	private $_pidFile = null;
	private function getPIDFile(){
		if($this->_pidFile === null){
			$this->_pidFile = self::PID_PATH.'/background_task_'.$this->getTaskId().'.pid';
		}
		return $this->_pidFile;
	}

	private $_outputFile = null;
	private function getOutputFile(){
		if($this->_outputFile == null){
			$this->_outputFile = self::OUTPUT_PATH.'/background_task_'.$this->getTaskId().'.out';
		}
		return $this->_outputFile;
	}

	private function writeOutput(){
		$this->getLog()->info(__LINE__, 'Write output file: '.$this->getOutputFile());
		file_put_contents($this->getOutputFile(), json_encode($this->_output), LOCK_EX);
		chmod($this->getOutputFile(), 0666);
	}

	private function setResult($content){
		$this->_output['result'] = $content;
		$this->writeOutput();
	}

	private function deleteOutput(){
		if(file_exists($this->getOutputFile())){
			$this->getLog()->info(__LINE__, 'Remove output file: '.$this->getOutputFile());
			unlink($this->getOutputFile());
		}
	}

	private function createPIDFile(){
		if(!file_exists($this->getPIDFile())){
			$this->getLog()->info(__LINE__, 'Create PID file: '.$this->getPIDFile());
			if(!file_put_contents($this->getPIDFile(), getmypid())){
				$this->getLog()->error(__LINE__, 'PID file "'.$this->getPIDFile().'" not created');
			}else{
				$this->_autoDeletePid = true;
			}
		}
	}

	private function deletePIDFile(){
		if($this->_autoDeletePid){
			if(file_exists($this->getPIDFile())){
				$this->getLog()->info(__LINE__, 'Remove PID file: '.$this->getPIDFile());
				unlink($this->getPIDFile());
			}
		}
	}


	public function destroy(){
		$this->deletePIDFile();
		if($this->_autoDelete){
			$this->delete();
		}
	}

	public function __destruct() {
		$this->destroy();
	}

}