<?php
namespace Cache;

/**
 * Class FileCache
 */
class FileCache extends \Singleton implements FileCacheInterface {

    const DIR = '/tmp/FileCache/';

    const USER_REGISTRATIONS   = 'users_registrations';
    const TRUNK_REGISTRATIONS  = 'trunks_registrations';
    const USERS_ACTIVE_DEVICES = 'users_active_devices';

    private const KEY_EXPIRE      = 'expire';
    private const KEY_REAL_EXPIRE = 'realExpire';
    private const KEY_DATA        = 'data';
    private const KEY_CREATED     = 'created';

    private $_realTtl = 0;
    private $_lockTtl = 10;

    protected $_withCompressing = true;

    /**
     * Logger class.
     *
     * @var \Logger\Log $_logger
     */
    private $_logger;

    public function __construct() {
        $this->_makeDir(self::DIR);
    }

    public function useCompressing(bool $compress = true): void {
        $this->_withCompressing = $compress;
    }

    /**
     * Get logger class.
     *
     * @return \Logger\Log
     */
    private function _getLogger(): \Logger\Log {
        if ($this->_logger === null) {
            $this->_logger = \Logger\LoggerFactory::getInstance();
        }

        return $this->_logger;
    }

    /**
     * Set cache.
     *
     * @param string  $key        Cache key.
     * @param mixed   $data       Cached data.
     * @param integer $expiration Ttl for cache in second.
     * @param string  $group      Cache group.
     *
     * @return void
     */
    public function set(string $key, $data, int $expiration, string $group = ''): void {
        if (!$expiration) {
            return;
        }
        $file = $this->_getCachePath($group) . $this->sanitizeFileName($key);

        if (file_put_contents($file, $this->_packData($data, $expiration), LOCK_EX)) {
            chmod($file, 0776);
        }
        $this->removeLock($this->_getLockPath($key, $group));
    }

    /**
     * Get cache.
     *
     * @param string $key   Cache key.
     * @param string $group Cache group.
     *
     * @return boolean
     */
    public function get(string $key, string $group = '') {
        $cache = $this->_readCacheFile($key, $group);
        if ($cache === false) {
            return false;
        }

        $data = $cache[self::KEY_DATA] ?? false;
        if ($data !== false && $cache[self::KEY_EXPIRE] <= time()) {
            if ($this->setLock($key, $group) !== false) {
                $this->_getLogger()->info(__LINE__, sprintf('Set lock for cache %s', $file));
                $data = false;
            } else {
                $this->_getLogger()->info(__LINE__, sprintf('Impossible set lock. Return data from expired cache %s', $file));
            }
        }

        return $data;
    }

    public function getKeyRawData(string $key, string $group = '') {
        return $this->_readCacheFile($key, $group);
    }

    /**
     * Get created time for cache.
     *
     * @param string $key   Cache key.
     * @param string $group Cache group.
     *
     * @return false|integer
     */
    public function getCacheCreatedTime(string $key, string $group = '') {
        $data = $this->_readCacheFile($key, $group);
        if ($data === false) {
            return false;
        }

        return $data[self::KEY_CREATED] ?? false;
    }

    private function _readCacheFile(string $key, string $group = '') {
        $file  = $this->_getCachePath($group) . $this->sanitizeFileName($key);
        $cache = $this->_unpackData($file);
        if ($cache === false) {
            return false;
        }
        $cache['path'] = $file;

        return $cache;
    }

    public function delete(string $key, string $group = '') {
        $file = $this->_getCachePath($group) . $this->sanitizeFileName($key);

        return is_readable($file) && unlink($file);
    }

    private function _packData($data, int $ttl = 0) {
        $time = time();
        $ttl  = $time + $ttl;
        $data = [
            self::KEY_CREATED => $time,
            self::KEY_EXPIRE => $ttl,
            self::KEY_REAL_EXPIRE => ($this->_realTtl > 0) ? $time + $this->_realTtl : $ttl,
            self::KEY_DATA => $data
        ];

        return $this->compress(json_encode($data));
    }

    private function _unpackData(string $file) {
        if (!file_exists($file)) {
            return false;
        }

        $cache = $this->decompress(file_get_contents($file));
        if (empty($cache)) {
            return false;
        }

        $cache = json_decode($cache, true);
        if (empty($cache)
            || ($cache[self::KEY_REAL_EXPIRE] ?? 0) <= time()
            || !isset($cache[self::KEY_EXPIRE])
            || !isset($cache[self::KEY_DATA])
        ) {
            return false;
        }

        return $cache;
    }

    protected function compress($data) {
        if ($this->_withCompressing) {
            $data = gzcompress($data, 1);
        }

        return $data;
    }

    protected function decompress($data) {
        if ($this->_withCompressing) {
            $data = gzuncompress($data);
        }

        return $data;
    }

    protected function sanitizeFileName(string $name): string {
        return md5($name);
    }

    /**
     * Real TTL for cache.
     *
     * @param integer $realTtl Real TTL.
     *
     * @return void
     */
    public function setRealTtl(int $realTtl): void {
        $this->_realTtl = $realTtl;
    }

    /**
     * Try set lock for cache.
     *
     * @param string $key   Cache key.
     * @param string $group Cache group.
     *
     * @return boolean
     */
    public function setLock(string $key, string $group = ''): bool {
        $file = $this->_getLockPath($key, $group);

        if (file_exists($file)) {
            $expiration = (float)file_get_contents($file);
            if ($expiration <= time()) {
                $this->removeLock($file);
            }

            return false;
        }

        $ttl = time() + $this->_lockTtl;
        return file_put_contents($file, $ttl, LOCK_EX) !== false;
    }

    /**
     * Remove lock.
     *
     * @param string $file Cache file.
     *
     * @return void
     */
    public function removeLock(string $file): void {
        if (file_exists($file)) {
            unlink($file);
            $this->_getLogger()->info(__LINE__, sprintf('Remove lock for cache %s', $file));
        }
    }

    private function _getLockPath(string $key, string $group = ''): string {
        return $this->_getCachePath($group) . $this->sanitizeFileName($key) . '.lock';
    }

    public function clear(string $group): void {
        if (is_dir($this->_getCachePath($group))) {
            exec('rm -rf ' . $this->_getCachePath($group));
        }
    }

    private function _getCachePath(string $group): string {
        if (!empty($group)) {
            $group = preg_replace('/^a-z0-9_\-/i', '', $group) . '/';
        }
        $dir = self::DIR . $group;
        $this->_makeDir($dir);
        return $dir;
    }

    private function _makeDir(string $dir) {
        if (!is_dir($dir)) {
            mkdir($dir);
            chmod($dir, 0777);
        }
    }
}
