<?php

namespace Logger;

/**
 * Class Log
 */
class Log {

    const EMERGENCY = 1;
    const ERROR     = 2;
    const WARNING   = 3;
    const INFO      = 4;
    const DEBUG     = 5;

    /**
     * Log constructor.
     *
     * @param Writer\IWriter|null                $writer    Writer for Log.
     * @param Formatter\IFormatterInterface|null $formatter Formatter for Log.
     * @param Context|null                       $context   Log context.
     * @param integer                            $logLevel  Log level parameter.
     */
    public function __construct(Writer\IWriter $writer, Formatter\IFormatterInterface $formatter, Context $context, $logLevel = self::INFO) {
        $this->setLogLevel($logLevel);
        $this->setWriter($writer);
        $this->setContext($context);
        $this->setFormatter($formatter);
    }

    /**
     * Log level.
     *
     * @var integer
     */
    private $_logLevel = null;

    /**
     * Setter fo logLevel parameter.
     *
     * @param integer $logLevel Log level parameter.
     *
     * @return void
     */
    public function setLogLevel($logLevel) {
        $this->_logLevel = $logLevel;
    }

    /**
     * Writer log.
     *
     * @var Writer\IWriter
     */
    private $_writer = null;

    /**
     * Setter for _writer parameter.
     *
     * @param Writer\IWriter $writer Instance object of IWriter interface.
     *
     * @return void
     */
    public function setWriter(Writer\IWriter $writer) {
        $this->_writer = $writer;
    }

    /**
     * Context object.
     *
     * @var Context
     */
    private $_context = null;

    /**
     * Setter for _context parameter.
     *
     * @param Context $context Instance object of Context class.
     *
     * @return void.
     */
    public function setContext(Context $context) {
        $this->_context = $context;
    }

    /**
     * Getter for _context parameter
     *
     * @return Context
     */
    public function getContext() {
        return $this->_context;
    }

    /**
     * Formatter object.
     *
     * @var Formatter\IFormatterInterface
     */
    private $_formatter = null;

    /**
     * Setter fot _formatter parameter.
     *
     * @param Formatter\IFormatterInterface $formatter Instance object of IFormatter interface.
     *
     * @return void.
     */
    public function setFormatter(Formatter\IFormatterInterface $formatter) {
        $this->_formatter = $formatter;
    }

    /**
     * Formatting message for write to system log.
     *
     * @param array|string $message Message for format.
     *
     * @return string
     */
    private function getFormattedMessage($message) {
        if (!is_array($message)) {
            $message = array(
                'message' => $message
            );
        }

        if ($this->_context !== null && !empty($this->_context)) {
            $message['context'] = $this->_context;
        }

        return $this->_formatter->getMessage($message);
    }

    /**
     * Method for write to log.
     *
     * @param integer      $level   Log level.
     * @param integer      $line    Line number.
     * @param array|string $message Message for write to log.
     *
     * @return void
     */
    private function write($level, $line, $message) {
        if ($level <= $this->_logLevel) {
            $formattedMessage = $this->getFormattedMessage($message);

            switch ($level) {
                case self::EMERGENCY:
                    $this->_writer->emergency($line, $formattedMessage);
                    break;
                case self::ERROR:
                    $this->_writer->error($line, $formattedMessage);
                    break;
                case self::WARNING:
                    $this->_writer->warning($line, $formattedMessage);
                    break;
                case self::INFO:
                    $this->_writer->info($line, $formattedMessage);
                    break;
                case self::DEBUG:
                    $this->_writer->debug($line, $formattedMessage);
                    break;
                default:
                    $this->_writer->info($line, $formattedMessage);
                    break;
            }
        }
    }

    /**
     * Method for emergency messages.
     *
     * @param integer      $line    Line number.
     * @param array|string $message Message for write to log.
     *
     * @return void
     */
    public function emergency($line, $message) {
        $this->write(self::EMERGENCY, $line, $message);
    }

    /**
     * Method for error messages.
     *
     * @param integer      $line    Line number.
     * @param array|string $message Message for write to log.
     *
     * @return void
     */
    public function error($line, $message) {
        $this->write(self::ERROR, $line, $message);
    }

    /**
     * Method for warning messages.
     *
     * @param integer      $line    Line number.
     * @param array|string $message Message for write to log.
     *
     * @return void
     */
    public function warning($line, $message) {
        $this->write(self::WARNING, $line, $message);
    }

    /**
     * Method for info messages.
     *
     * @param integer      $line    Line number.
     * @param array|string $message Message for write to log.
     *
     * @return void
     */
    public function info($line, $message) {
        $this->write(self::INFO, $line, $message);
    }

    /**
     * Method for debug messages.
     *
     * @param integer      $line    Line number.
     * @param array|string $message Message for write to log.
     *
     * @return void
     */
    public function debug($line, $message) {
        $this->write(self::DEBUG, $line, $message);
    }
}
