<?php

namespace Logger\Writer;

/**
 * Class SysLog
 */
class SysLog implements IWriter {

    /**
     * Log identification parameter.
     *
     * @var null|string
     */
    private $_identification = null;

    /**
     * SysLog constructor.
     *
     * @param string      $identification    Log identification.
     * @param null|string $subIdentification Log sub identification.
     */
    public function __construct($identification, $subIdentification = null) {
        $this->_identification = $identification;
        $this->setSubIdentification($subIdentification);
    }

    /**
     * Log sub identification parameter.
     *
     * @var null|string
     */
    private $_subIdentification = null;

    /**
     * Setter for _subIdentification parameter.
     *
     * @param string $subIdentification Sub identification.
     *
     * @return void
     */
    public function setSubIdentification($subIdentification) {
        $this->_subIdentification = $subIdentification;
    }

    /**
     * Getter for Identification.
     *
     * @param null|integer $line Line number.
     *
     * @return null|string
     */
    private function getIdentification($line = null) {
        if ($line !== null) {
            return $this->_identification . '[' . $line . ']';
        }
        return $this->_identification;
    }

    /**
     * Write to log method.
     *
     * @param integer $priority Log level.
     * @param integer $line     Line number.
     * @param string  $message  Message which write to log.
     *
     * @return void
     */
    private function writeLog($priority, $line, $message) {
        openlog($this->getIdentification($line), LOG_PID, LOG_LOCAL0);

        $messageLines = preg_split("/(\n\r|\r\n|\n|\r)/", $message);

        foreach ($messageLines as $line) {
            syslog($priority, $line);
        }

        closelog();
    }

    /**
     * Method for emergency messages.
     *
     * @param integer $line    Line number.
     * @param string  $message Message for write to log.
     *
     * @return void
     */
    public function emergency($line, $message) {
        $this->writeLog(LOG_EMERG, $line, $message);
    }

    /**
     * Method for error messages.
     *
     * @param integer $line    Line number.
     * @param string  $message Message for write to log.
     *
     * @return void
     */
    public function error($line, $message) {
        $this->writeLog(LOG_ERR, $line, $message);
    }

    /**
     * Method for warning messages.
     *
     * @param integer $line    Line number.
     * @param string  $message Message for write to log.
     *
     * @return void
     */
    public function warning($line, $message) {
        $this->writeLog(LOG_WARNING, $line, $message);
    }

    /**
     * Method for info messages.
     *
     * @param integer $line    Line number.
     * @param string  $message Message for write to log.
     *
     * @return void
     */
    public function info($line, $message) {
        $this->writeLog(LOG_INFO, $line, $message);
    }

    /**
     * Method for debug messages.
     *
     * @param integer $line    Line number.
     * @param string  $message Message for write to log.
     *
     * @return void
     */
    public function debug($line, $message) {
        $this->writeLog(LOG_DEBUG, $line, $message);
    }
}
