<?php

namespace Microsoft;

/*
 * registration app https://apps.dev.microsoft.com/Disambiguation?ru=https%3a%2f%2fapps.dev.microsoft.com%2f
 *
 * https://msdn.microsoft.com/en-us/office/office365/api/contacts-rest-operations
 * https://msdn.microsoft.com/office/office365/APi/extended-properties-rest-operations
 *
 *
 *
 * https://msdn.microsoft.com/office/office365/APi/api-catalog
 * https://msdn.microsoft.com/office/office365/HowTo/examples-of-office-365-unified-api-calls
 *
 * https://msdn.microsoft.com/en-us/office/office365/howto/get-started-with-office-365-unified-api
 * https://graphexplorer2.azurewebsites.net/?UrlRequest=GET&text=https%3A%2F%2Fgraph.microsoft.com%2Fbeta%2Fme%2F
 *
 * List global contacts
 * https://graph.microsoft.com/beta/contacts
 * http://graph.microsoft.io/en-us/docs/api-reference/beta/api/orgcontact_list
 */

use Wildix\Services\Kite;

/**
 * Local implementation for Office365 Client.
 */
class Office365 extends \Singleton {

    private $_logLevel = 0;

    const TOKEN_VALID     = 1;
    const NO_ACCESS_TOKEN = 2;
    const NO_ID_TOKEN     = 3;
    const TOKEN_EXPIRED   = 4;

    private $_authUrl;
    private $_tokenUrl;

    private $_log = null;

    private $_baseUrl;
    private $_kiteService;

    public static function getInstance() {
        $className = self::class;
        if (!isset(self::$_Instances[$className])) {
            $baseUri   = env('MICROSOFT_AUTHORIZE_DOMAIN');
            // TODO config path
            $_authUrl  = $baseUri . '/common/oauth2/v2.0/authorize';
            $_tokenUrl = $baseUri . '/common/oauth2/v2.0/token';

            $kiteService = new \Wildix\Services\Kite();

            self::$_Instances[$className] = new Office365($kiteService, $baseUri, $_authUrl, $_tokenUrl);
        }
        return self::$_Instances[$className];
    }

    /**
     * Office365 constructor.
     *
     * @param Kite   $kiteService Service provided kite client.
     * @param string $baseUri     Base Url.
     * @param string $authUrl     Uri.
     * @param string $tokenUrl    Token.
     *
     * @throws \Exception Error.
     */
    public function __construct(Kite $kiteService, string $baseUri, string $authUrl, string $tokenUrl) {
        parent::__construct();
        $this->_kiteService = $kiteService;
        $this->_baseUrl     = $baseUri;
        $this->_authUrl     = $authUrl;
        $this->_tokenUrl    = $tokenUrl;
    }

    private function getLog() {
        if ($this->_log === null) {
            $this->_log = new \Syslog('\Microsoft\Office365', $this->_logLevel);
        }
        return $this->_log;
    }



    private $_microsoftIntagrationApp = null;
    private function getMicrosoftIntagrationApp($param) {
        if ($this->_microsoftIntagrationApp === null) {
            $this->_microsoftIntagrationApp = $this->_kiteService->getMicrosoftIntagrationApp();
        }

        return $this->_microsoftIntagrationApp[$param];
    }

    public function getAuthUrl($scope, $callback, $access_type = 'offline', $rePrompted = 'force') {
        $params = array(
                'client_id' => $this->getMicrosoftIntagrationApp('clientId'),
                'redirect_uri' => urlencode($this->getMicrosoftIntagrationApp('redirectUrl')),
                'response_type' => 'code',
                'scope' => urlencode($scope),
        // openid
                'state' => $callback
        );

        return $this->_authUrl . '?' . urldecode(http_build_query($params));
    }

    public function getAdminConsentUrl($callback) {
        // TODO check if need this addiction? could pass via params?
        $office365AppConfig = $this->_kiteService->getOffice365IntegrationApp();
        // copy-paste
        // TODO use oauth2 GenericProvider::getAuthorizationUrl
        $params = [
            'client_id' => $office365AppConfig['clientId'],
            'redirect_uri' => $office365AppConfig['redirectUrl'],
            'state' => $callback,
        ];

        // Have no idea why in other similar functions we use urldecode after http_build_query! weired...
        return $this->_baseUrl . '/common/adminconsent?' . http_build_query($params);
    }

    /**
     *
     * @var token
     * array(
     *     'received_token_time'=> time
     *     'access_token' => string
     *     'token_type' => string
     *     'expires_in' => int
     *     'refresh_token' => string
     * )
     */
    private $_token = null;
    public function getToken($code = null) {
        if ($this->_token === null && $code !== null) {
            $params       = array(
                    'client_id'     => $this->getMicrosoftIntagrationApp('clientId'),
                    'client_secret' => $this->getMicrosoftIntagrationApp('clientSecret'),
                    'redirect_uri'  => $this->getMicrosoftIntagrationApp('redirectUrl'),
                    'grant_type'    => 'authorization_code',
                    // 'resource'=> 'https://outlook.office365.com/',
                    'code'          => $code
            );
            $this->_token = $this->_postRequest($this->_tokenUrl, $params);
            $this->_token['received_token_time'] = time();
        }

        return $this->_token;
    }

    public function refreshToken($token) {
        $params = array(
                'client_id'     => $this->getMicrosoftIntagrationApp('clientId'),
                'client_secret' => $this->getMicrosoftIntagrationApp('clientSecret'),
                'refresh_token'  => $token['refresh_token'],
                'grant_type'    => 'refresh_token'
        );

        $newToken = $this->_postRequest($this->_tokenUrl, $params);
        $newToken['received_token_time'] = time();
        return $newToken;
    }

    public function isValidToken($token) {
        if (empty($token['access_token'])) {
            return self::NO_ACCESS_TOKEN;
        }
        if (empty($token['id_token'])) {
            return self::NO_ID_TOKEN;
        }
        if (!empty($token['received_token_time']) && !empty($token['expires_in'])) {
            if ($token['received_token_time'] + $token['expires_in'] < time()) {
                return self::TOKEN_EXPIRED;
            }
        }

        return self::TOKEN_VALID;
    }

    public function setToken($token) {
        $this->getLog()->info(__LINE__, 'setToken: ' . print_r($token, true));

        $result = $this->isValidToken($token);
        if ($result == self::TOKEN_VALID) {
            $this->_token = $token;
        } else if ($result = self::TOKEN_EXPIRED) {
            $token  = $this->refreshToken($token);
            $result = $this->isValidToken($token);
            if ($result == self::TOKEN_VALID) {
                $this->_token = $token;
            }
        }
        return $result;
    }


    private $_userInfo = null;
    public function getUserInfo($field = null) {
        if ($this->_userInfo === null) {
            $idToken = explode('.', $this->_token['id_token']);
            if (!empty($idToken[1])) {
                $userInfo = base64_decode($idToken[1]);
                if (!empty($userInfo)) {
                    $this->_userInfo = json_decode($userInfo, true);
                }
            }
        }

        if (!empty($field)) {
            return $this->_userInfo[$field];
        }
        return $this->_userInfo;
    }


    public function getPhonebooks($start = 0, $count = 100) {
        $params = array(
                '$top' => $count,
                '$skip' => $start
        );

        $response = $this->_getRequest('https://graph.microsoft.com/v1.0/me/contactFolders', $params);
        if ($response && !empty($response['error'])) {
            return $response;
        }

        $entry = $this->_prepareResponse($response);

        if (is_array($entry)) {
            array_unshift(
                $entry, array(
                    'id' => 'global',
                    'name' => 'Organizational contacts',
                    'parent' => null
                ), array(
                    'id' => 'default',
                    'name' => 'Contacts',
                    'parent' => null
                )
            );
        }

        return $entry;
    }

    private $_nextLinkPageContacts = null;
    public function getContacts($phonebookId = null, $start = 0, $count = 100) {
        $params = array(
                '$top' => $count,
                '$skip' => $start
        );

        $response = null;
        if (!$phonebookId || $phonebookId == 'default') {
            $response = $this->_getRequest('https://graph.microsoft.com/v1.0/me/contacts', $params);
        } else if ($phonebookId == 'global') {
            if ($start == 0) {
                $response = $this->_getRequest('https://graph.microsoft.com/beta/contacts', array('$top' => $count));
            } else if ($this->_nextLinkPageContacts !== null) {
                $response = $this->_getRequest($this->_nextLinkPageContacts);
            }
            if (!empty($response['@odata.nextLink'])) {
                $this->_nextLinkPageContacts = $response['@odata.nextLink'];
            } else {
                $this->_nextLinkPageContacts = null;
            }
        } else {
            $response = $this->_getRequest('https://graph.microsoft.com/v1.0/me/contactFolders/' . $phonebookId . '/contacts', $params);
        }

        if ($response && !empty($response['error'])) {
            return $response;
        }

        $entry = $this->_prepareResponse($response);

        return $entry;
    }

    private function _prepareResponse($response) {
        $result = array();
        if ($response && !empty($response)) {
            if (!empty($response['value']) && is_array($response['value'])) {
                foreach ($response['value'] as $value) {
                    if ($value['@odata.type'] == '#microsoft.graph.contact') {
                        $result[] = $this->_parseContact($value);
                    } else if ($value['@odata.type'] == '#microsoft.graph.contactFolder') {
                        $result[] = array(
                                'id' => $value['id'],
                                'name' => $value['displayName'],
                                'parent' => $value['parentFolderId']
                        );
                    } else if ($value['@odata.type'] == '#microsoft.graph.orgContact') {
                        $result[] = $this->_parseContact($value);
                    }
                }
            }
        }
        return $result;
    }

    private function _parseContact($contact) {
        $result = array(
                'id' => '',
                'created' => '',
                'modified' => '',
                'birthday' => '',
                'nickname' => '',
                'notes' => '',
                'name' => array(
                        'FirstName' => '',
                        'MiddleName' => '',
                        'LastName' => '',
                        'Title' => '',
                ),
                'website' => array(),
                'email' => array(),
                'phone' => array()
        );

        if (!empty($contact['id'])) {
            $result['id'] = $contact['id'];
        }

        if (!empty($contact['createdDateTime'])) {
            $result['created'] = $contact['createdDateTime'];
        }

        if (!empty($contact['lastModifiedDateTime'])) {
            $result['modified'] = $contact['lastModifiedDateTime'];
        }

        if (!empty($contact['birthday'])) {
            $result['birthday'] = $contact['birthday'];
        }

        if (!empty($contact['nickName'])) {
            $result['nickname'] = $contact['nickName'];
        }

        if (!empty($contact['personalNotes'])) {
            $result['notes'] = $contact['personalNotes'];
        }

        if (!empty($contact['givenName'])) {
            $result['name']['FirstName'] = $contact['givenName'];
        }
        if (!empty($contact['middleName'])) {
            $result['name']['MiddleName'] = $contact['middleName'];
        }
        if (!empty($contact['surname'])) {
            $result['name']['LastName'] = $contact['surname'];
        }

        if (!empty($contact['title'])) {
            $result['name']['Title'] = $contact['title'];
        }

        if (!empty($contact['businessHomePage'])) {
            $result['website'] = $contact['businessHomePage'];
        }

        if (!empty($contact['emailAddresses'])) {
            foreach ($contact['emailAddresses'] as $key => $value) {
                $result['email'][($key + 1)] = $value['address'];
            }
        } else if (!empty($contact['mail'])) {
            // from org contacts
            $result['email'][1] = $contact['mail'];
        }

        if (!empty($contact['homePhones'])) {
            foreach ($contact['homePhones'] as $key => $value) {
                $result['phone']['Home' . ($key + 1)] = $value;
            }
        }
        if (!empty($contact['businessPhones'])) {
            foreach ($contact['businessPhones'] as $key => $value) {
                $result['phone']['Business' . ($key + 1)] = $value;
            }
        }
        if (!empty($contact['mobilePhone'])) {
            $result['phone']['Mobile1'] = $contact['mobilePhone'];
        }

        if (!empty($contact['jobTitle'])) {
            $result['JobTitle'] = $contact['jobTitle'];
        }

        if (!empty($contact['companyName'])) {
            $result['company'] = $contact['companyName'];
        }

        if (!empty($contact['department'])) {
            $result['department'] = $contact['department'];
        }


        if (!empty($contact['officeLocation'])) {
            $result['OfficeLocation'] = $contact['officeLocation'];
        }

        if (!empty($contact['profession'])) {
            $result['profession'] = $contact['profession'];
        }

        if (!empty($contact['assistantName'])) {
            $result['AssistantName'] = $contact['assistantName'];
        }

        if (!empty($contact['manager'])) {
            $result['Manager'] = $contact['manager'];
        }

        if (!empty($contact['spouseName'])) {
            $result['SpouseName'] = $contact['spouseName'];
        }

        if (!empty($contact['generation'])) {
            $result['Generation'] = $contact['generation'];
        }

        if (!empty($contact['homeAddress'])) {
            $result['HomeAddress'] = $contact['homeAddress'];
        } else if (!empty($contact['city'])) {
            // from org contacts
            $result['HomeAddress'] = array(
                    'street' => $contact['streetAddress'],
                    'city' => $contact['city'],
                    'state' => $contact['state'],
                    'countryOrRegion' => $contact['country'],
                    'postalCode' => $contact['postalCode']
            );
        }

        if (!empty($contact['businessAddress'])) {
            $result['BusinessAddress'] = $contact['businessAddress'];
        }

        if (!empty($contact['otherAddress'])) {
            $result['OtherAddress'] = $contact['otherAddress'];
        }

        return $result;
    }

    public function getUsers($count = 100, $nextLink = null) {
        if (! $nextLink) {
            $params['$select'] = 'businessPhones,displayName,mail,givenName,jobTitle,mobilePhone,officeLocation,preferredLanguage,surname,userPrincipalName, department';
            if ($count) {
                $params['$top'] = $count;
            }
            $response = $this->_getRequest('https://graph.microsoft.com/v1.0/users', $params);
        } else {
            $response = $this->_getRequest($nextLink);
        }


        if ($response && !empty($response['error'])) {
            return $response;
        }

        $entry = $this->_parseResponseUsers($response);

        return $entry;
    }

    private function _parseResponseUsers($response) {
        $return = array(
            'users' => array()
        );
        if ($response && !empty($response) && !empty($response['value']) && is_array($response['value'])) {
            if ($response['@odata.nextLink']) {
                $return['nextLink'] = $response['@odata.nextLink'];
            }
            $response = $response['value'];
            for ($i = 0; $i < count($response); $i++) {
                $return['users'][] = array(
                        'id' => $response[$i]['id'],
                        'businessPhones' => $response[$i]['businessPhones'][0],
                        'displayName' => $response[$i]['displayName'],
                        'mail' => $response[$i]['mail'],
                        'givenName' => $response[$i]['givenName'],
                        'jobTitle' => $response[$i]['jobTitle'],
                        'mobilePhone' => $response[$i]['mobilePhone'],
                        'officeLocation' => $response[$i]['officeLocation'],
                        'preferredLanguage' => strtolower($response[$i]['preferredLanguage']),
                        'surname' => $response[$i]['surname'],
                        'userPrincipalName' => $response[$i]['userPrincipalName'],
                        'department' => $response[$i]['department']
                );
            }
        }
        return $return;
    }

    private function _getRequest($url, $params = null) {
        if ($params === null) {
            $params = array();
        }

        $token = $this->getToken(null);

        if (!empty($params)) {
            $url .= '?' . http_build_query($params);
        }

        $headers = array(
                'Authorization: Bearer ' . $token['access_token'],
                'Accept: application/json; odata.metadata=full',
        // none, minimal, full
        );

        $this->getLog()->info(__LINE__, 'GET: ' . $url);

        $curl = curl_init();
        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_HTTPGET, true);
        curl_setopt($curl, CURLOPT_CONNECTTIMEOUT, 5);
        curl_setopt($curl, CURLOPT_TIMEOUT, 120);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, true);

        curl_setopt($curl, CURLOPT_USERAGENT, 'Mozilla/5.0 (Windows; U; Windows NT 5.1; ru; rv:1.8.0.9) Gecko/20061206 Firefox/1.5.0.9');
        curl_setopt($curl, CURLOPT_HTTPHEADER, $headers);

        $result = curl_exec($curl);

        if ($result == false) {
            $result['error'] = array(
                    'code' => curl_errno($curl),
                    'message' => curl_error($curl)
            );
        } else {
            $result = json_decode($result, true);
        }
        curl_close($curl);
        $this->getLog()->info(__LINE__, 'Response: ' . print_r($result, true));
        return $result;
    }

    private function _postRequest($url, $postParams = null, $getParams = null) {
        if ($postParams === null) {
            $postParams = array();
        }

        $this->getLog()->info(__LINE__, 'POST: ' . $url);
        $this->getLog()->info(__LINE__, print_r($postParams, true));

        $curl = curl_init();
        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_POST, true);
        curl_setopt($curl, CURLOPT_POSTFIELDS, urldecode(http_build_query($postParams)));
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_TIMEOUT, 5);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, false);
        $result = curl_exec($curl);
        if ($result == false) {
            $result['error'] = array(
                    'code' => curl_errno($curl),
                    'message' => curl_error($curl)
            );
        } else {
            $result = json_decode($result, true);
        }
        curl_close($curl);
        $this->getLog()->info(__LINE__, 'Response: ' . print_r($result, true));
        return $result;
    }
}
