<?php
namespace Session;

use User;

// @TODO use https://github.com/ralouphie/getallheaders
if (!function_exists('getallheaders')) {
    function getallheaders() {
        $headers = [];
        foreach ($_SERVER as $name => $value) {
            if (substr($name, 0, 5) == 'HTTP_') {
                $headers[str_replace(' ', '-', ucwords(strtolower(str_replace('_', ' ', substr($name, 5)))))] = $value;
            }
        }
        return $headers;
    }
}

class Session extends \Singleton{

    const CSRF_TOKEN_HEADER = 'X-CSRF-Token';

    private $_logLevel = 0;

    /**
     * Constructor
     *
     * @param  array $settings
     */
    public function __construct(){
        /**
         * Session
         *
         * We must start a native PHP session to initialize the $_SESSION superglobal.
         * However, we won't be using the native session store for persistence, so we
         * disable the session cookie and cache limiter. We also set the session
         * handler to this class instance to avoid PHP's native session file locking.
        */

        session_set_save_handler(
            array($this, 'open'),
            array($this, 'close'),
            array($this, 'read'),
            array($this, 'write'),
            array($this, 'destroy'),
            array($this, 'garbageCollector')
        );
    }

    public $_isSupportSession = null;
    public function isSupportSession(){
        if($this->_isSupportSession === null){
            $this->_isSupportSession = php_sapi_name() != 'cli';
        }
        return $this->_isSupportSession;
    }

    public function setSupportSession($flag){
        $this->_isSupportSession = $flag;
    }


    private $_log = null;
    private function getLog(){
        if($this->_log === null){
            $this->_log = new \Syslog('Session', $this->_logLevel);
        }
        return $this->_log;
    }

    private $_sessionWrapper = null;
    private function getSessWrapper(){
        if($this->_sessionWrapper === null){
            switch ($this->getType()) {
                case 'sqlite': $this->_sessionWrapper = new Sqlite();
                    break;
                default: $this->_sessionWrapper = new File();
                    break;
            }
        }
        return $this->_sessionWrapper;
    }

    private $_type = null;
    public function setType($type){
        $this->_type = $type;
        return $this;
    }

    public function getType(){
        return $this->_type;
    }

    public function start(){
        if(!$this->isSupportSession()){
            return;
        }
        $this->_checkOauthMode();

        $this->getLog()->info(__LINE__, 'Start session');
        session_start();

        $cookie_lifetime = ini_get('session.cookie_lifetime');
        if(empty($cookie_lifetime) || $cookie_lifetime <= 0){
            // @TODO use constant
            // week = 7 days
            $cookie_lifetime = 604800;
        }

        $cookieOptions = [
            'expires' => time() + $cookie_lifetime,
            'path' => ini_get('session.cookie_path'),
            'domain' => ini_get('session.cookie_domain'),
            'secure' => array_key_exists('HTTPS', $_SERVER),
            'httponly' => ini_get('session.cookie_httponly'),
            'samesite' => 'None'
        ];
        setcookie(session_name(), session_id(), $cookieOptions);
    }

    private function _checkOauthMode() {
        $path = parse_url($_SERVER['REQUEST_URI'], PHP_URL_PATH);

        // check OAuth authorization API points
        $oauthPaths = [
            '/authorization/oauth2',
            '/authorization/oauth2Grant',
            '/authorization/oauth2Login',
            '/authorization/oauth2Logout',
        ];
        if (in_array($path, $oauthPaths, true)) {
            $this->getLog()->info(__LINE__, 'Session switch to OAuth2 mode');

            ini_set('session.cookie_path', '/authorization/');
            ini_set('session.name', 'OAUTHSESSID');
        }
    }

    public function addCsrfToken() {
        if (!isset($_SESSION[self::CSRF_TOKEN_HEADER])) {
            $_SESSION[self::CSRF_TOKEN_HEADER] = bin2hex(random_bytes(32));
        }
    }

    public static function getCsrfToken(): string {
        return $_SESSION[self::CSRF_TOKEN_HEADER] ?? '';
    }

    public static function isValidCsrfToken($token = ''): bool {
        if (User::getInstance()->getAuthMode() !== '') {
            return true;
        }
        $csrfToken = ($token === '' ? (array_change_key_case(getallheaders())[strtolower(self::CSRF_TOKEN_HEADER)] ?? '') : $token);
        return ($csrfToken !== '' && hash_equals(self::getCsrfToken(), $csrfToken));
    }

    public function writeClose(){
        if(!$this->isSupportSession()){
            return;
        }
        $this->getLog()->info(__LINE__, 'Write close session');
        session_write_close();
    }


    /********************************************************************************
     * Session Handler
    *******************************************************************************/

    public function open($savePath, $sessionName){
        $this->getLog()->info(__LINE__, 'Open session['.$sessionName.']: '.$savePath);
        return $this->getSessWrapper()->open($savePath, $sessionName);
    }

    public function close(){
        $this->getLog()->info(__LINE__, 'Close session');
        return $this->getSessWrapper()->close();
    }

    public function read($sess_id){
        $this->getLog()->info(__LINE__, 'Read session: '.$sess_id);
        return $this->getSessWrapper()->read($sess_id);
    }

    public function write($sess_id, $data){
        if(!$this->isSupportSession()){
            return true;
        }
        $this->getLog()->info(__LINE__, 'Write session['.$sess_id.']: '.$data);
        return $this->getSessWrapper()->write($sess_id, $data);
    }

    public function destroy($sess_id){
        $this->getLog()->info(__LINE__, 'Destroy session: '.$sess_id);

        return $this->getSessWrapper()->destroy($sess_id);
    }

    public function garbageCollector($sess_maxlifetime){
        $this->getLog()->info(__LINE__, 'Garbage collector: '.$sess_maxlifetime);

        return $this->getSessWrapper()->garbageCollector($sess_maxlifetime);
    }
}
