<?php

use PHPGangsta_GoogleAuthenticator as GoogleAuthenticator;
use Firebase\JWT\Key;
use Api\Network\Exception\ForbiddenException;
use Services\LuaService;
use Session\Session;
use Wildix\System\ADSingleSignOnService;
use \Wildix\Auth\Kerberos\Negotiate;
use \Wildix\Core\Container;
use Wildix\Utils\JWT;
use Wildix\Auth\AmazonCognito\AuthService;
use Wildix\Auth\Applications\OAuth2\Objects\Model as OAuth2ApplicationModel;
use Wildix\Auth\OAuth2\OAuth2Service;
use Wildix\Auth\SimpleToken\SimpleTokenService;
use Wildix\License;

/**
 * Class User
 *
 * Some kind of user manager and session interact
 */
class User extends \Wildix\Colleagues\Colleague {

    const JWT_ALG = 'HS256';

    const CACHE_TIME                        = 15; // seconds
    const DEFAULT_UPDATE_SESSION_CACHE_TIME = 300; // seconds

    const REMEMBER_COOKIE = '__rc';

    const AUTH_TYPE_GOOGLE         = 'google';
    const AUTH_TYPE_MICROSOFT      = 'microsoft';
    const AUTH_TYPE_PORTAL         = 'portal';
    const AUTH_TYPE_LOGIN_PASSWORD = 'login';
    const AUTH_TYPE_BASIC          = 'basic';
    const AUTH_TYPE_NEGOTIATE      = 'negotiate';
    const AUTH_TYPE_OAUTH2         = 'oauth2';
    const AUTH_TYPE_AMAZON_COGNITO = 'cognito';
    const AUTH_TYPE_S2S            = 's2s';
    const AUTH_TYPE_PERSONAL_TOKEN = 'personal';
    const AUTH_TYPE_WIZYCONF       = 'wizyconf';
    const AUTH_TYPE_SERVICE        = 'serviceAuth';
    const AUTH_TYPE_SIMPLE_TOKEN   = 'simpleToken';

    const SKIP_2FA_AUTH_MODES = [
        self::AUTH_TYPE_GOOGLE,
        self::AUTH_TYPE_MICROSOFT,
        self::AUTH_TYPE_AMAZON_COGNITO,
        self::AUTH_TYPE_OAUTH2,
        self::AUTH_TYPE_PERSONAL_TOKEN,
        self::AUTH_TYPE_WIZYCONF,
        self::AUTH_TYPE_SIMPLE_TOKEN,
        self::AUTH_TYPE_SERVICE,
        self::AUTH_TYPE_S2S,
    ];

    private static $_Instances = null;

    private $_authMode = '';

    private $_luaService;

    /**
     * @var \Wildix\DependencyInjection\Container $_diContainer
     */
    private $_diContainer;

    /**
     * Return singleton of user object.
     *
     * @return User
     */
    public static function getInstance() {
        if (self::$_Instances === null) {
            self::$_Instances = new self();
        }

        return self::$_Instances;
    }

    public function __construct() {
        parent::__construct(array());

        $this->restoreUser();
    }

    private $_log = null;
    private function getLog() {
        if ($this->_log === null) {
            $this->_log = \Logger\LoggerFactory::getInstance();
        }
        return $this->_log;
    }

    public function isValidPassword($password){
    	if($this->isConfiguredActiveDirectory()){
    		if($this->checkActiveDirectory($this->getLogin(), $password)){
				return true;
    		}
    	}

    	if($this->isOnlySingleSignOn() && $this->getName() !== 'admin'){
    		return false;
    	}

    	if($this->getPassword() == $password || $this->isEqualPasswords($this->getPassword(), $password)){
    		return true;
    	}

		return false;
    }

    private function fillSession($data){
    	if($data !== false){
    		$this->setPropertys($data['user']);

    		if(isset($data['withAdminPwd'])){
    			$_SESSION['withAdminPwd'] = $data['withAdminPwd'];
    		}else{
    			$_SESSION['withAdminPwd'] = false;
    		}

    		if(isset($data['twoFactorState'])){
    			$this->setTwoFactorState($data['twoFactorState']);
    		}

    		if(isset($data['type'])){
    			$this->setAuthType($data['type']);
    		}

    		if(isset($data['login'])){
    			$this->setLogin($data['login']);
    		}

            if (isset($data['portalUserInfo'])) {
                $this->setPortalUserInfo((array)$data['portalUserInfo']);
            }

    		$this->setSession();
    	}
    }

    /**
     * User authorization.
     *
     * @param string      $login    Login.
     * @param string      $password Password.
     * @param null|string $type     Type.
     *
     * @return array|bool
     */
    public function login($login, $password, $type = null, bool $isMsTeams = false) {
        $method = $type ? '; Method: ' . $type : '';

        $this->getLog()->info(__LINE__, 'Start authorization. Login: ' . $login . $method);

        switch ($type) {
            case self::AUTH_TYPE_MICROSOFT:
            case self::AUTH_TYPE_GOOGLE:
            case self::AUTH_TYPE_PORTAL:
                $result = $this->restoreFromOAuth($login, $password, $type);
                break;
            default:
                $result = $this->restoreFromLoginPassword($login, $password);
                break;
        }

        $licenseType = $result['user']['licenseType'] ?? '';
        if ($this->isNotAllowedLicenseType($licenseType) && $login !== 'admin' && !$this->isPhoneDevice()) {
            $this->getLog()->info(__LINE__, sprintf('Authorization deny for license type "%s". Login: %s', $licenseType, $login));
            return ['errorCode' => 403];
        }

        if ($isMsTeams && $result !== false && ! License::getInstance()->haveLicense(License::TYPE_CLASSOUND)) {
            $this->getLog()->info(__LINE__, sprintf('Authorization via MS Teams was rejected due to lack of CLASSOUND license. Login: %s', $login));
            return [
                'errorCode'    => 403,
                'errorMessage' => 'teams4Wildix integration requires CLASSOUND to be activated',
            ];
        }

        if ($result !== false) {
            $this->fillSession($result);
            $this->getLog()->info(__LINE__, 'Authorization success. Login: ' . $login . $method);

            return true;
        }

        $this->getLog()->error(__LINE__, 'Authorization fail: Login: ' . $login);
        return ['errorCode' => 401];
    }

    private function logout() {
        $this->getLog()->info(__LINE__, sprintf('Logout user: %s', $this->getName()));

        $this->_ldapProperties = null;
        $this->setPropertys([]);

        if (isset($_COOKIE[session_name()])) {
            setcookie(session_name(), '', -1, '/');
            unset($_COOKIE[session_name()]);
        }

        if (isset($_COOKIE[self::REMEMBER_COOKIE])) {
            setcookie(self::REMEMBER_COOKIE, '', -1, '/');
            unset($_COOKIE[self::REMEMBER_COOKIE]);
        }

        $_SESSION = [];
        session_write_close();
    }

    public function isLogged(){
    	$name = $this->getName();
    	if(!empty($name)){
    		$next = $this->getNextStage();
    		if($next === false){
    			return true;
    		}
   		}
        return false;
    }

    public function setAuthType($type){
    	$_SESSION['authType'] = $type;
    }

    public function getAuthType(){
    	if(isset($_SESSION['authType'])){
    		return $_SESSION['authType'];
    	}
		return '';
    }

    private function setLogin($type){
    	$_SESSION['login'] = $type;
    }

    private function setPortalUserInfo(array $data) {
    	$_SESSION['portalUserInfo'] = $data;
    }

    private function getLogin(){
    	if(isset($_SESSION['login'])){
    		return $_SESSION['login'];
    	}
    	return '';
    }

    public function setTwoFactorState($state){
    	$_SESSION['twoFactorState'] = $state;
    }

    public function getTwoFactorState(){
    	if(!empty($_SESSION) && !empty($_SESSION['twoFactorState'])){
    		return true;
    	}
    	return false;
    }

    public function getNextStage() {
        if (!in_array($this->getAuthType(), [self::AUTH_TYPE_GOOGLE, self::AUTH_TYPE_MICROSOFT], true)
            && !in_array($this->getAuthMode(), self::SKIP_2FA_AUTH_MODES, true)
        ) {
            $twoFactorState = $this->getTwoFactorState();
            if (!$twoFactorState) {
                $settings  = $this->getSettings();
                $twoFactor = $settings->getTwoFactor();
                if (!empty($twoFactor) && $twoFactor['enable'] == 1 && !empty($twoFactor['secret'])) {
                    $this->getLog()->info(__LINE__, sprintf('Need next stage authorization. App %s', $twoFactor['app']));

                    return $twoFactor['app'];
                }
            }
        }

        return false;
    }

    public function checkNextStage($code) {
        $next = $this->getNextStage();

        if ($next === false) {
            return true;
        }

        switch ($next) {
            case 'GoogleAuthenticator':
            case 'FreeOTPAuthenticator':
            case 'WindowsPhoneAuthenticator':
            case 'OtherAuthenticator':
                $discrepancy = 2;   // 2 => 2*30sec clock tolerance
                break;
            case \Wildix\Auth\TwoFactor\TwoFactorAuthenticatorFactory::FACTOR_SMS:
            case \Wildix\Auth\TwoFactor\TwoFactorAuthenticatorFactory::FACTOR_EMAIL:
                $discrepancy = 6;   // 6 => 6*30sec clock tolerance
                break;
            default:
                return false;
        }

        $settings    = $this->getSettings();
        $twoFactor   = $settings->getTwoFactor();
        $ga          = new GoogleAuthenticator();
        $checkResult = $ga->verifyCode($twoFactor['secret'], $code, $discrepancy);

        if (!$checkResult) {
            return false;
        }

        $this->setTwoFactorState(true);
        $this->getLog()->info(__LINE__, 'Two factor Authorization success. Login: ' . $this->getLogin() . '; Method: ' . $next. ';');

        return true;
    }


    /**
     * Encode JWT token.
     *
     * @param string  $secret   Secret.
     * @param array   $payload  Additional payload.
     * @param integer $lifetime Token lifetime time in seconds.
     *
     * @return string
     */
    public function getJwtAuthToken(string $secret, array $payload = [], int $lifetime = 30): string {
        $time   = time();
        $domain = \Models\PBX::getInstance()->getDomain();

        $user = $this->getExtension();
        if (empty($user)) {
            $user = $this->getLogin();
        }

        $defaultPayload = [
            'iss' => $domain,
            'sub' => 'PBX',
            'aud' => $domain,
            'iat' => $time,
            'exp' => $time + $lifetime,
            'user' => $user,
        ];

        return JWT::encode(array_merge($defaultPayload, $payload), $secret);
    }


    private $_adConf = null;
    private function getActiveDirectoryConfig($name = null){
        if($this->_adConf === null){
            $this->_adConf = false;
            if(file_exists("/rw2/etc/pbx/ad_connect.conf")){
                $this->_adConf = parse_ini_file("/rw2/etc/pbx/ad_connect.conf", false, INI_SCANNER_RAW);

                if ( !empty($this->_adConf['type']) ) {
                    switch ($this->_adConf['type']) {
                        case 'domino':
                            if( empty($this->_adConf['host']) ){
                                $this->_adConf = false;
                            }
                            break;

                        case 'activeD':
                            if(empty($this->_adConf['host']) || empty($this->_adConf['windomain'])){
                                $this->_adConf = false;
                            }
                            break;

                        case 'google' || 'office365':
                            break;

                        default:
                            $this->_adConf = false;
                            break;
                    }
                } elseif (empty($this->_adConf['host'])){
                        $this->_adConf = false;
                }
            }
        }

        if(!empty($name)){
            return $this->_adConf[$name];
        }
        return $this->_adConf;
    }



    private function isConfiguredActiveDirectory(){
    	$adConf = $this->getActiveDirectoryConfig();
    	if(!empty($adConf['host'])){
    		return true;
    	}
    	return false;
    }

    private function isOnlySingleSignOn(){
    	$adConf = $this->getActiveDirectoryConfig();
    	if(!empty($adConf) && !empty($adConf['signon_only']) && $adConf['signon_only'] == 'yes'){
    		return true;
    	}
    	return false;
    }


    private function getUserDataFromCacheOrLdap(){
        if ($this->hasSession()) {
            if ($this->hasSessionCache() && ((time() - $_SESSION['user']['time']) < self::CACHE_TIME)) {
    			$this->getLog()->debug(__LINE__, 'User data from cache');
    			return $_SESSION['user']['data'];
    		}else{
                $colleagues = \Wildix\Colleagues\Colleagues::getInstance();
                $filter = "(&(!(userKind=room))(!(userKind=park_orbit))(" . $colleagues->getLdapKey(\Wildix\Colleagues\Colleague::ID) . "={$_SESSION['userId']}))";

                return $colleagues->getUserByFilter($filter);
    		}
    	}
    	return null;
    }

    private function _compareUserDataWithSession($userData = []): bool {
        $sessionUserData = !empty($_SESSION['user']['data']) ? $_SESSION['user']['data'] : [];
        return $this->getUserHash($userData) === $this->getUserHash($sessionUserData);
    }

    private function getUserData($login){
    	$data = null;
    	if($login){
    		$filter = "(&(!(userKind=room))(!(userKind=park_orbit))(|(telephoneNumber={$login})(mail={$login})(userPrincipalName={$login})))";

            // admin search only by cn
            if ($login === 'admin') {
                $filter = "(&(!(userKind=room))(!(userKind=park_orbit))(cn={$login}))";
            }

    		$this->getLog()->info(__LINE__, 'Ldap filter: '.$filter);
    		$data = \Wildix\Colleagues\Colleagues::getInstance()->getUserByFilter($filter);
    		return $data;
    	}
    	return $data;
    }

    private function getLuaService(): LuaService {
        if (!$this->_luaService) {
            $this->_luaService = new LuaService();
        }
        return $this->_luaService;
    }

    private function checkPam(string $login, string $password): bool {
        return $this->getLuaService()->checkPam($login, $password);
    }

    private function checkActiveDirectory($login, $password){
    	if($this->isConfiguredActiveDirectory()){
    		$adConf = $this->getActiveDirectoryConfig();
    		$adUser = null; // SAMAccountName or pre-Windows 2000
    		$adUserUPN = null; // User Principal Name
    		if ( !empty($adConf['type']) ) {
    			switch ($adConf['type']) {
    				case 'domino':
    					$adUser = $login;
    					break;

    				case 'activeD':
                        // check @ and \ symbols in the login
                        if (preg_match('/^([^@]+)(@)([^@]+)$/', $login)) {
                            $adUserUPN = $login;
                        } else if (preg_match('/^([^\\\]+)(\\\)([^\\\]+)$/', $login)) {
                            $adUser = $login;
                        } else if (!empty($adConf['windomain'])) {
                            $this->getLog()->info(
                                __LINE__,
                                sprintf(
                                    'ActiveDirectory calculate SAM & UPN for login: %s, windomain: %s, fullDomain: %s',
                                    $login,
                                    $adConf['windomain'],
                                    $adConf['fullDomain']
                                )
                            );
                            if (strpos($login, '@') === false) {
                                $name = $login;
                            } else {
                                $name = explode('@', $login)[0];
                            }

                            $adUser    = sprintf('%s\\%s', $adConf['windomain'], $name);
                            $adUserUPN = sprintf('%s@%s', $name, $adConf['fullDomain']);
    					}
    					break;
    			}
    		}

    		if(!$adUser && !$adUserUPN){
    			return false;
    		}

    		$this->getLog()->info(__LINE__, 'ActiveDirectory SAM user: ' . $adUser);
    		$this->getLog()->info(__LINE__, 'ActiveDirectory UPN user: ' . $adUserUPN);

			try{
				$connect = (new Wildix\Ldap\Connector(
					$adConf['host'],
					$adConf['port'],
					filter_var($adConf['secure'],FILTER_VALIDATE_BOOLEAN)
				))->getConnection();
			} catch (Wildix\Ldap\Exception\ConnectException $e){
				$this->getLog()->warning(__LINE__, 'ActiveDirectory connection fail ('.$e->getCode().'): '.$e->getMessage());
			}

    		if($connect){
				//Connector already sets PROTOCOL_VERSION, REFERRALS, NETWORK_TIMEOUT options
    			ldap_set_option($connect, LDAP_OPT_TIMELIMIT, 180);
    			try{
    				if (!empty($adUser) && @ldap_bind($connect, $adUser, $password)) {
    					$this->getLog()->info(__LINE__, 'ActiveDirectory SAM login success. With user: ' . $adUser);
    					return true;
    				} else if (!empty($adUserUPN) && @ldap_bind($connect, $adUserUPN, $password)) {
    					$this->getLog()->info(__LINE__, 'ActiveDirectory UPN login success. With user: ' . $adUserUPN);
    					return true;
    				} else {
    					$this->getLog()->warning(__LINE__, 'ActiveDirectory login fail: '.ldap_error($connect));
    				}
    			}catch (\Exception $e){
    				$this->getLog()->warning(__LINE__, 'ActiveDirectory exception ('.$e->getCode().'): '.$e->getMessage());
    			}
    		}

    	}
    	return false;
    }

    private function restoreUser(){
        if ($this->hasSession()) {
            $this->getLog()->debug(__LINE__, 'Session authorization: ' . $_SESSION['username']);
            $userData = $this->getUserDataFromCacheOrLdap();
            if ($this->hasSessionCache() && $userData && !$this->_compareUserDataWithSession($userData)) {
                $this->getLog()->info(__LINE__, sprintf("The user's confidential data has been changed. Extension %s ", $userData['extension']));
                $this->logout();
                $userData = null;
            }
            $this->setPropertys($userData);
        } else {
            $token = $_GET['token'] ?? '';
            if ($token && SimpleTokenService::isSimpleToken($token)) {
                \Session\Session::getInstance()->setSupportSession(false);

                $this->getLog()->error(__LINE__, '[SimpleToken] authorization by token in parameter');
                $pbxUser = $this->_restoreUserBySimpleToken($token);
                $data    = $this->getUserData($pbxUser);
                if (empty($data)) {
                    $this->getLog()->error(__LINE__, sprintf('[SimpleToken] authorization by token parameter error. User "%s" not found', $pbxUser));
                } else {
                    $this->setPropertys($data);
                    $this->getLog()->error(__LINE__, '[SimpleToken] authorization by token parameter success');
                }
            } else if (!empty($_SERVER['PHP_AUTH_USER']) || !empty($_SERVER['PHP_AUTH_PW'])){

    			$login = $_SERVER['PHP_AUTH_USER'];
    			$password = $_SERVER['PHP_AUTH_PW'];

                if ($login === \Models\PBX::getInstance()->getSerial()) {
                    $this->getLog()->info(__LINE__, 'Skip user authorization, serial number is used as login: ' . $login);
                    return;
                }

                $this->getLog()->info(__LINE__, 'Basic authorization: ' . $login);
                $this->_setAuthMode(self::AUTH_TYPE_BASIC);

    			if(empty($_COOKIE)){
    				$this->getLog()->info(__LINE__, 'Cookie empty, not support session');
    				Session::getInstance()->setSupportSession(false);
    			}

    			if (!empty($login) && !empty($password) && !$this->isForbidden($login)) {
                    $loginStatus = $this->login($login, $password);
                    $this->getLog()->info(__LINE__, sprintf('Basic Authentication user: %s; status: %s', $login, $loginStatus === true ? 'true' : 'false'));

                    if (is_array($loginStatus) && $loginStatus['errorCode'] === 403) {
                        $this->logout();
                        if (defined('PBX_API')) {
                            throw new ForbiddenException();
                        }
                        return;
                    }

                    // try basic and fail -> register fail
                    if (is_array($loginStatus) && $loginStatus['errorCode'] === 401) {
                        $this->authFail($login);
                    }

    				$this->setTwoFactorState(true);
    			}

    		}else if (!empty($_SERVER['HTTP_AUTHORIZATION'])) {
    		    $authorization = $_SERVER['HTTP_AUTHORIZATION'];
    			if (strpos($authorization, 'Bearer ') === 0) {
                    $token = str_replace('Bearer ', '', $authorization);
                    $this->restoreFromJwtToken($token);
    			} else if (strpos($authorization, 'Negotiate ') === 0) {
                    // Kerberos(Negotiate) authorization
                    $this->_setAuthMode(self::AUTH_TYPE_NEGOTIATE);
                    $data = null;
                    $user = $this->getKerberosAuthenticatedUser();
                    if (!empty($user)) {
                        $data = $this->getUserData($user);
                    }
                    if (!empty($data) && is_array($data)) {
                        $this->setPropertys($data);
                    }
                }
            }
    	}


    	if(!$this->isLogged() && $this->getRememberData() != false){
    		$result = $this->restoreFromRememberCookie();
    		$this->fillSession($result);
    	}

    	if($this->isLogged()){
    		if($this->getRememberData() != false){
    			$this->setRememberCookie();
    		}
    		$this->setSession();

            $this->getLog()->info(__LINE__, 'Restore user success');
        } else {
            $this->getLog()->error(__LINE__, 'Restore user failed');
        }
    }

    /**
     * Authenticate via Kerberos(Negotiate)
     * @return false|string
     */
    private function getKerberosAuthenticatedUser() {
        $user           = false;
        $adSingleSignOn = Container::get(ADSingleSignOnService::class);;
        $kerberos       = Container::get(Negotiate::class);
        try {
            $user = $kerberos->auth();
        } catch (\Exception $exception) {
            $this->getLog()->error(__LINE__, $exception->getMessage());
        }
        return $user;
    }

    /**
     * @param $cryptPass - password hash from LDAP
     * @param $plainPass - plain text password
     * @return bool
     *
     * 1. Get salt from hashed password
     * 2. Convert plain text password to hash using salt and hash algorithm
     * 3. Compare two hashes
     */
    public function isEqualPasswords($cryptPass, $plainPass){

        $_passData = explode('$', $cryptPass);
        $_salt = $_passData[count($_passData)-2];
        $_pass = '{CRYPT}' . crypt($plainPass, '$6$' .$_salt. '$');

        if ($cryptPass == $_pass){
            return true;
        }

        return false;
    }

    public function updateSessionPassword(string $password) {
        $_SESSION['password'] = $password;
        $_SESSION['user']['data'][\Wildix\Colleagues\Colleague::PASSWORD] = $password;
    }

    public function updateSessionCache(int $newCacheTime) {
        $_SESSION['user']['time'] = $newCacheTime;
    }

    /**
     * @param $password  - plain text password
     * @return string - hashed password
     *
     *  Password hash format
     *  {CRYPT}$6$fH0QpwprRiJQR19Y$bXGOh7dIfOWpUb/Tuqr7yQVCqL3UkrJns9.7msfvMg4ZO/PsFC5Tbt32PXAw9qRFEBs1254aLimFeNM8YsYOv.
     *  |--0---|1|-------2--------|-----------------------------------------3--------------------------------------------|
     *
     *  0 - LDAP encryption method
     *  1 - hash algoritm ( 6 - sha512 )
     *  2 - salt (16 alphanumeric characters)
     *  3 - password hash
     */
    public static function cryptPassword($password){

        $_salt = substr(md5(time()), 0, 16);
        $_pass = '{CRYPT}' . crypt($password, '$6$' . $_salt . '$');

        return $_pass;
    }

    public static function makeUserPinFromPassword($password){
        $matchArray = array(
            'a' => 2,
            'b' => 2,
            'c' => 2,
            'd' => 3,
            'e' => 3,
            'f' => 3,
            'g' => 4,
            'h' => 4,
            'i' => 4,
            'j' => 5,
            'k' => 5,
            'l' => 5,
            'm' => 6,
            'n' => 6,
            'o' => 6,
            'p' => 7,
            'q' => 7,
            'r' => 7,
            's' => 7,
            't' => 8,
            'u' => 8,
            'v' => 8,
            'w' => 9,
            'x' => 9,
            'y' => 9,
            'z' => 9
        );

        // PIN length should be max 5 digits // Changed! PIN the same length like password
        $pass_length = strlen($password);
        $pin_length = ($pass_length > 0 AND $pass_length <= 5) ? $pass_length : 5;

        $pass = substr($password, 0, $pin_length);
        $pass = strtolower($pass);
        $pass = preg_replace("/[^0-9a-z]/", "*", $pass);
        $pin = strtr($pass, $matchArray);

        return $pin;

    }

    private function restoreFromLoginPassword($login, $password){
        // Allow user with hash password for remote access inside WMS-network
        $this->_setAuthMode(self::AUTH_TYPE_LOGIN_PASSWORD);
    	if ( strlen($password) > 60 ) {
            // check if remote ip present in LDAP as wPBX pbxIp or it's 127.0.0.1
            if (\Wildix\PBXes\PBXes::getInstance()->getPBXByIp($_SERVER['HTTP_X_PEER_ADDR']) || $_SERVER['HTTP_X_PEER_ADDR'] == '127.0.0.1') {
                $data = $this->getUserData($login);
                if ($data && $data['password'] == $password) {
                    return array (
                        'user' => $data,
                        'login' => $login
                    );
                }
            }

    		return false;
    	}

    	$data = $this->getUserDataFromCacheOrLdap();
    	if($data && $data['password'] == $password || $data && $this->isEqualPasswords($data['password'], $password)){
    		// This user is already logged in
    		$this->getLog()->info(__LINE__, 'Authorized from cache');

    		return array (
    				'user' => $data,
    				'login' => $login
    		);
    	}

    	if(!$data){
    		$data = $this->getUserData($login);
    	}

    	if(!empty($data) && is_array($data)){
    		if($this->isConfiguredActiveDirectory()){
    			if($this->checkActiveDirectory($login, $password)){
    				$adConf = $this->getActiveDirectoryConfig();
    				return array (
    						'user' => $data,
    						'login' => $login,
    						'type' => $adConf['type'],
    						'withAdminPwd' => false
    				);
    			}
    		}

    		//if ( $username == 'admin' || (empty($adConf) || empty($adConf['signon_only']) || $adConf['signon_only'] == 'no') ) {
    		/*
    		 * if ActiveDirectory not configured or not logged
    		 * try over PBX
    		 */
    		if($this->isOnlySingleSignOn() && $login !== 'admin'){
				return false;
    		}

    		if($data['password'] == $password || $this->isEqualPasswords($data['password'], $password)){
    			$this->getLog()->info(__LINE__, 'Authorized from PBX Ldap');
    			return array (
    					'user' => $data,
    					'login' => $login,
    					'withAdminPwd' => false
    			);
    		}

    	}else if((empty($data) || !is_array($data)) && $login == 'admin' ){
    		/*
    		 * if ldap broken or user not fount and login admin
    		 * try over PAM
    		 */
    		if($this->checkPam($login, $password)){
    			$this->getLog()->info(__LINE__, 'PAM Authorization successful');
    			$data = array(
    					self::NAME => $login,
    					self::PASSWORD => $password,
    					self::ROLE => self::ADMINISTRATOR,
    					self::LANGUAGE => 'en',
    					self::DIALPLAN => 'users',
    					self::FAX_DIALPLAN => 'users'
    			);
    			return array (
    					'user' => $data,
    					'login' => $login
    			);
    		}else{
    			$this->getLog()->error(__LINE__, 'PAM Authorization fail');
    		}
    	}
    	return false;
    }

    private function restoreFromOAuth($login, $token, $type){
        if($this->isOnlySingleSignOn()){
            $this->getLog()->info(__LINE__, 'WARNING: Login by OAuth is not permitted with Single SingOn only enabled.');
            return false;
        }
    	$email = '';
        $portalUserInfo = null;
    	$this->getLog()->info(__LINE__, 'Login by OAuth: '. $type);
    	if ($type === self::AUTH_TYPE_GOOGLE) {
    	    $this->_setAuthMode(self::AUTH_TYPE_GOOGLE);
    		$apiController = \GoogleAPI::getInstance();
    		if($apiController->setToken($token) == $apiController::TOKEN_VALID){
    			$userInfo = $apiController->getUserInfo();
    			if($userInfo['email_verified'] == 1){
    				$email = $userInfo['email'];
    			}
    		}
    	} else if($type === self::AUTH_TYPE_MICROSOFT) {
            $this->_setAuthMode(self::AUTH_TYPE_MICROSOFT);
    		$apiController = \Microsoft\Office365::getInstance();

    		if($apiController->setToken($token) == $apiController::TOKEN_VALID){
    			$userInfo = $apiController->getUserInfo();
    			$email = $userInfo['preferred_username'];
    		}
    	} else if ($type === self::AUTH_TYPE_PORTAL) {
            $this->_setAuthMode(self::AUTH_TYPE_PORTAL);
            try {
                $_portalApiUrl = env('SERVICE_PORTAL_API_URI');
                $_portalUrl    = env('SERVICE_UPGRADE_PORTAL_URI');
                $portalService = new \Wildix\Services\Portal(\Logger\Log::WARNING, $_portalApiUrl, $_portalUrl);

                $portalUserInfo = $portalService->singleSignOn($token);

                if (is_array($portalUserInfo)) {
                    $email = $portalUserInfo['loginAs'] ?? '';
                    $this->getLog()->info(__LINE__, array_merge($portalUserInfo, ['message' => 'PBX single sign on from portal is successful']));
                }

            } catch (\Exception $e) {
                $this->getLog()->error(__LINE__, $e->getMessage());
            }
        } else {
    		return false;
    	}

    	if(!empty($email) && $email == $login){
    		$this->getLog()->info(__LINE__, 'Find by login in ldap '. $login);
    		$colleague = \Wildix\Colleagues\Colleagues::getInstance()->getUserByLogin($login);

    		if(!$colleague){
    			$this->getLog()->info(__LINE__, 'Find by email in ldap '. $login);
    			$colleague = \Wildix\Colleagues\Colleagues::getInstance()->getUserByEmail($login);

                if (!$colleague && $type === self::AUTH_TYPE_PORTAL && $login === 'admin') {
                    $this->getLog()->info(__LINE__, 'Find by cn in ldap '. $login);
                    $colleague = \Wildix\Colleagues\Colleagues::getInstance()->getUserByName($login);
                }
    		}

    		if($colleague){
    			return array (
    					'user' => $colleague->getPropertys(),
    					'type' => $type,
    					'login' => $login,
    					'twoFactorState' => true,
    					'withAdminPwd' => false,
    					'portalUserInfo' => $portalUserInfo,
    			);
    		}
    	}
    	return false;
    }

    /**
     * Restore user from JWT token.
     *
     * @param string $token JWT token.
     *
     * @return void
     */
    private function restoreFromJwtToken(string $token): void {
        $this->getLog()->info(__LINE__, 'Restore user from JWT token. Token: ' . $token);

        Session::getInstance()->setSupportSession(false);
        $data  = null;
        $appId = \Wildix\Auth\Api\ApiAuthenticator::getAppId();

        $isSimpleToken = \Wildix\Auth\SimpleToken\SimpleTokenService::isSimpleToken($token);

        $oAuthHash    = '';
        $tokenSubject = '';

        if ($appId) {
            $this->_setAuthMode(self::AUTH_TYPE_S2S);

            $apiKeysPairData = new \Wildix\Auth\Api\ApiKeysPairData($this->getLog());
            $apiKeysPair     = $apiKeysPairData->getInfoByAppId($appId);
            $integrationName = $apiKeysPair[\Wildix\Auth\Api\ApiKeysPairData::KEY_APP] ?? '';
            $pbxUser         = $apiKeysPair[\Wildix\Auth\Api\ApiKeysPairData::KEY_PBX_USER] ?? '';

            $this->getLog()->info(__LINE__, sprintf('API JWT token authorization s2s.  APP: %s, user: %s', $integrationName, $pbxUser));
        } else if ($isSimpleToken) {
            $this->getLog()->error(__LINE__, '[SimpleToken] authorization by token in header');
            $pbxUser = $this->_restoreUserBySimpleToken($token);
        } else {
            $this->getLog()->info(__LINE__, 'API JWT start authorization');

            $tokenPayload = $this->decodeJwtToken($token, '', false);
            if (empty($tokenPayload)) {
                $this->getLog()->error(__LINE__, 'API JWT token is empty');
                return;
            }

            if ($this->_isServiceAuthToken($tokenPayload['sub'] ?? '')) {
                $this->_setAuthMode(self::AUTH_TYPE_SERVICE);
                return;
            }

            $data = null;
            if (isset($tokenPayload['custom:pbx_domain'])
                && isset($tokenPayload['iss'])
                && AuthService::isAmazonCognitoToken((string)$tokenPayload['iss'])
            ) {
                $this->getLog()->info(__LINE__, 'AWS JWT start authorization');
                $this->_setAuthMode(self::AUTH_TYPE_AMAZON_COGNITO);

                $iss       = (string)$tokenPayload['iss'];
                $pbxDomain = (string)($tokenPayload['custom:pbx_domain'] ?? '');

                if ($pbxDomain !== \Models\PBX::getInstance()->getDomain()) {
                    $this->getLog()->error(__LINE__, sprintf('AWS JWT incorrect PBX domain %s', $pbxDomain));
                    return;
                }

                $awsAuth = new AuthService();
                if (!$awsAuth->validate($iss, $token)) {
                    $this->getLog()->error(__LINE__, sprintf('AWS JWT validation error: %s', $awsAuth->getErrorMessage()));
                    if ($awsAuth->isInvalidSignature()) {
                        $this->getLog()->error(__LINE__, 'AWS JWT invalid signature. Need reload AWS public keys');
                        $res = $awsAuth->tryReloadPublicKeys($iss);
                        $this->getLog()->error(__LINE__, 'AWS JWT reload public keys status: ' . ($res ? 'true' : 'false'));
                    }
                    return;
                }
                $this->getLog()->info(__LINE__, 'AWS JWT validation success');

                $awsTokenData = $awsAuth->getTokenData();
                if (empty($awsTokenData['email'])) {
                    $this->getLog()->error(__LINE__, 'AWS JWT not found email in payload');
                    return;
                }

                $pbxUser = (string)$awsTokenData['email'];
                $this->getLog()->info(__LINE__, sprintf('AWS JWT found email: %s', $pbxUser));
            } else {
                if (isset($tokenPayload[OAuth2Service::OAUTH2_SESSION_KEY])) {
                    $this->getLog()->info(__LINE__, 'API JWT OAuth2 start authorization');
                    $this->_setAuthMode(self::AUTH_TYPE_OAUTH2);

                    $oauthData = (array)$tokenPayload[OAuth2Service::OAUTH2_SESSION_KEY];

                    $licenseToken = $this->_getOauthAppSecretKey($oauthData);
                    $oAuthHash    = (string)($oauthData[OAuth2Service::PARAM_HASH] ?? '');
                    if (!$licenseToken || !$oAuthHash) {
                        $this->getLog()->error(__LINE__, 'API JWT OAuth2 authorization break');
                        return;
                    }
                } else {
                    $licenseToken = License::getInstance()->getLimitations('token');

                    if ($this->_isPersonalToken($tokenPayload)) {
                        $this->_setAuthMode(self::AUTH_TYPE_PERSONAL_TOKEN);
                    }
                }

                $decodedToken = $this->decodeJwtToken($token, $licenseToken);

                if (empty($decodedToken) || !$this->validateJwtToken($decodedToken)) {
                    $this->getLog()->info(__LINE__, 'API JWT token validate failed');
                    return;
                }

                $tokenSubject = $decodedToken['sub'] ?? '';

                $pbxUser = $decodedToken['user'] ?? '';
                $data    = $this->getUserDataFromCacheOrLdap();

                if ($this->_isWizyconfToken($tokenSubject)) {
                    $this->_setAuthMode(self::AUTH_TYPE_WIZYCONF);
                }
            }
        }

        if (empty($data)) {
            $data = $this->getUserData($pbxUser);
        }

        if (!empty($data)
            && is_array($data)
            && $this->checkWizyconfUserCredentials($tokenSubject, $data)
            && $this->_checkOauthUserHash($oAuthHash, $data)
        ) {
            $this->getLog()->info(__LINE__, 'API JWT token authorization success. user: ' . $pbxUser);
            $this->setPropertys($data);
        } else {
            $this->getLog()->info(__LINE__, 'API JWT token authorization fail. user: ' . $pbxUser);
        }
    }

    private function _restoreUserBySimpleToken(string $token): string {
        $this->_setAuthMode(self::AUTH_TYPE_SIMPLE_TOKEN);

        $user = '';
        try {
            $service        = $this->_getDiContainer()->get(\Wildix\Auth\SimpleToken\SimpleTokenServiceInterface::class);
            $simpleTokenApp = $service->getSimpleTokenApp($token);
            $this->getLog()->info(
                __LINE__,
                sprintf(
                    '[SimpleToken] restore user success. Application: %s user: %s',
                    $simpleTokenApp->getId(),
                    $simpleTokenApp->getPbxUser()
                )
            );

            $user = $simpleTokenApp->getPbxUser();
        } catch (\Exception $e) {
            $this->getLog()->error(__LINE__, sprintf('[SimpleToken] restore user error. %s "%s"', $e->getMessage(), $token));
        }

        return $user;
    }

    private function _isPersonalToken($tokenPayload) {
        return !empty($tokenPayload['sub'])
            && !empty($tokenPayload['aud'])
            && $tokenPayload['sub'] === 'PBX'
            && $tokenPayload['aud'] === \Models\PBX::getInstance()->getDomain();
    }

    private function _isWizyconfToken($tokenSubject) {
        return stripos($tokenSubject, 'Wizyconf') !== false;
    }

    private function _isServiceAuthToken($tokenSubject) {
        return $tokenSubject === self::AUTH_TYPE_SERVICE;
    }

    /**
     * Check OAuth2 user hash if he not empty
     *
     * @param string $hash     User hash.
     * @param array  $userData User data.
     *
     * @return boolean
     */
    private function _checkOauthUserHash(string $hash, array $userData): bool {
        return $hash === '' || $hash === $this->getUserHash($userData);
    }

    /**
     * Get OAuth2 application secret
     *
     * @param array $oauthData OAuth access token payload data.
     *
     * @return string
     */
    private function _getOauthAppSecretKey(array $oauthData): string {
        $appSecret = '';

        if (!empty($oauthData)
            && !empty($oauthData[OAuth2Service::PARAM_CLIENT_ID])
            && isset($oauthData[OAuth2Service::PARAM_TOKEN_TYPE])
            && $oauthData[OAuth2Service::PARAM_TOKEN_TYPE] === \Wildix\Auth\OAuth2\OAuth2Service::PARAM_ACCESS_TOKEN
        ) {
            $oAuthAppId = (string)$oauthData[OAuth2Service::PARAM_CLIENT_ID];
            $this->getLog()->info(
                __LINE__, sprintf(
                    'API JWT OAuth2 authorization User=%s AppID=%s',
                    $tokenPayload['user'] ?? 'empty',
                    $oAuthAppId
                )
            );

            $appModel = $this->_getOauthAppById($oAuthAppId);
            if ($appModel !== null) {
                $appSecret = $appModel->getSecret();
            }
        } else {
            $this->getLog()->info(__LINE__, sprintf('API JWT OAuth2 invalid data %s', json_encode($oauthData)));
        }

        return $appSecret;
    }

    /**
     * Get OAuth2 application by ID
     *
     * @param string $oAuthAppId OAuth2 Application ID.
     *
     * @return OAuth2ApplicationModel|null
     */
    private function _getOauthAppById(string $oAuthAppId): ?OAuth2ApplicationModel {
        // YES we already have container in this MEGA class BUT use static access!!
        $appModel = \Wildix\Auth\Applications\OAuth2\Services\Service::getAppById($oAuthAppId);

        if (!empty($oAuthAppId) && $appModel === null) {
            $this->getLog()->error(__LINE__, sprintf('Not found OAuth2 application %s', $oAuthAppId));
        }

        return $appModel;
    }

    private function checkWizyconfUserCredentials(string $tokenSubject, array $userData) {
        if (stripos($tokenSubject, 'Wizyconf') !== false && $userData['licenseType'] !== License::TYPE_WIZYCONF) {
            $this->getLog()->error(
                __LINE__,
                \sprintf(
                    'API JWT token authorization fail. Wizyconf user has wrong license type. user: %s, license: %s',
                    $userData['extension'],
                    $userData['licenseType']
                )
            );
            return false;
        }

        return true;
    }

    private function restoreFromRememberCookie(){
    	$this->getLog()->info(__LINE__, 'Login by remember cookie.');

    	$cookie = $this->getRememberData();

    	if(empty($cookie)){
    		return false;
    	}

        $colleague = \Wildix\Colleagues\Colleagues::getInstance()->getUserByID($cookie['userId']);
        if($colleague === null){
            return false;
        }

    	$currentHash = $this->getHash($colleague, $cookie['expire']);
    	if($currentHash != $cookie['hash']){
    		return false;
    	}

    	return array (
    				'user' => $colleague->getPropertys(),
    				'type' => $cookie['authType'],
    				'twoFactorState' => true,
    				'login' => $cookie['login']
    	);
    }

    /**
     * Decode JWT token.
     *
     * @param string  $token  JWT token.
     * @param string  $secret Secret.
     * @param boolean $verify Verify JWT.
     *
     * @return array | null
     */
    private function decodeJwtToken(string $token, string $secret, bool $verify = true): ?array {
        try {
            if ($verify) {
                $payload = (array)JWT::decode($token, new Key($secret, self::JWT_ALG));
            } else {
                $payload = JWT::unpack($token)['payload'];
            }

            return $payload;
        } catch (\Exception $e) {
            $this->getLog()->error(
                __LINE__,
                \sprintf('JWT decode error: %s; token: %s', $e->getMessage(), $token)
            );
            return null;
        }
    }

    /**
     * Validate payload of JWT token.
     *
     * @param array $decodedToken Decoded JWT token.
     *
     * @return boolean
     */
    private function validateJwtToken(array $decodedToken): bool {
        if ($decodedToken['exp'] < time()) {
            $this->getLog()->error(
                __LINE__,
                \sprintf('Token expired; created: %s, expired: %s, now: %s',
                    $decodedToken['iat'],
                    $decodedToken['exp'],
                    time()
                )
            );
            return false;
        }

        if ($decodedToken['iss'] !== \Models\PBX::getInstance()->getDomain()) {
            $this->getLog()->error(
                __LINE__,
                \sprintf('Unknown token provider; iss: %s', $decodedToken['iss'])
            );
            return false;
        }

        if (empty($decodedToken['user'])) {
            $this->getLog()->error(
                __LINE__,
                'User extension is missing in token.'
            );
            return false;
        }

        return true;
    }

    private function getHash(\Wildix\Colleagues\Colleague $colleague, int $expireTime): string {
        $data = [
            $colleague->getName(),
            $colleague->getExtension(),
            $colleague->getID(),
            $colleague->getPassword(),
            $expireTime,
            self::$_salt
        ];
        return md5(implode('', $data));
    }

    private function setSession(){
        $this->getLog()->info(__LINE__, sprintf('Save user data to session uid: %s name: %s email: %s', $this->getID(), $this->getName(), $this->getEmail()));

        $_SESSION['user']['data'] = $this->getPropertys();
    	$_SESSION['user']['time'] = time();
    	$_SESSION['password'] = $this->getPassword();
    	$_SESSION['username'] = $this->getName();
    	$_SESSION['userId'] = $this->getID();
    	$_SESSION['email'] = $this->getEmail();
    	$_SESSION['language'] = \Language::getInstance()->getLanguageNameByCode($this->getLanguage());

        Session::getInstance()->addCsrfToken();
    }

    public function hasSession() {
        return (string)($_SESSION['username'] ?? '') !== '' && (string) ($_SESSION['password'] ?? '') !== '';
    }

    public function hasSessionCache() {
        return isset($_SESSION['user']['data']) && count($_SESSION['user']['data']) > 0 ;
    }

    public function clearSessionCache() {
        unset($_SESSION['user']);
    }

    private $_expireTime = null;
    private function getRememberCookieExpireTime(){
        if($this->_expireTime === null){
            $this->_expireTime = time()+60*60*24*15;
        }
        return $this->_expireTime;
    }

    public function setRememberCookie(){
        setcookie(
        '__rc',
        $this->makeRememberCookie(),
        $this->getRememberCookieExpireTime(),
        ini_get('session.cookie_path'),
        ini_get('session.cookie_domain'),
        (empty($_SERVER['HTTPS'])) ? false : true,
        ini_get('session.cookie_httponly')
        );
    }

    private $_rememberData = null;
    private function getRememberData(){
        if($this->_rememberData === null){
            $this->_rememberData = false;

            $cookie = isset($_COOKIE['__rc']) ? trim($_COOKIE['__rc']) : '';

            if(!empty($cookie)){
                $first = $cookie[0];
                $cookie[0] = $cookie[strlen($cookie)-3];
                $cookie[strlen($cookie)-3] = $first;
                $cookie = base64_decode($cookie);

                $cookie = explode('|', $cookie);

                if($cookie[3] >= time()){
                    $this->_rememberData = array(
                            'userId' => $cookie[0],
                    		'login' => $cookie[1],
                    		'authType' => $cookie[2],
                            'expire' => $cookie[3],
                            'hash' => $cookie[4]
                    );
                }
            }
        }

        return $this->_rememberData;
    }


    private static $_salt = 'ldVVmrn90GxSO8QJPamgTb5YGU8izLvV';
    private function makeRememberCookie(){
        $userUid = $this->getID();
    	$login = $this->getLogin();
        $cookie = $userUid.'|'.$login.'|'.$this->getAuthType().'|'.$this->getRememberCookieExpireTime().'|'.$this->getHash($this, $this->getRememberCookieExpireTime());
        $cookie = base64_encode($cookie);
        $first = $cookie[0];
        $cookie[0] = $cookie[strlen($cookie)-3];
        $cookie[strlen($cookie)-3] = $first;

        return $cookie;
    }


    /**
     * Checks if a device is a phone by user agent header.
     *
     * @return boolean
     */
    private function isPhoneDevice(): bool {
        $userAgent   = $_SERVER['HTTP_USER_AGENT'] ?? '';
        return preg_match('/^Android CTIconnect|(Android Collaboration (Vision|SuperVision|WP600A|WP600ACG).*)$/', $userAgent) === 1;
    }

    /**
     * Check if license type is not allowed.
     *
     * @param string $licenseType User license type.
     *
     * @return boolean
     */
    private function isNotAllowedLicenseType(string $licenseType): bool {
        return in_array($licenseType, [License::TYPE_BASIC, License::TYPE_PBX_SERVICE], true);
    }

    private function getUserIp() :string {
        $keys = [
            'HTTP_X_FORWARDED_FOR',
            'REMOTE_ADDR'
        ];

        foreach ($keys as $key) {
            $ip = $_SERVER[$key];
            if (isset($ip) && filter_var($ip, FILTER_VALIDATE_IP)) {
                return $ip;
            }
        }

        return 'unknown';
    }

    private function getForbiddenFile(string $login = null) {
        return sprintf('/tmp/session/lighty/auth_%s%s', $login === 'admin' ? 'admin_' : '', $this->getUserIp());
    }

    /**
     * Return Period in seconds for block user login.
     *
     * @param string|null $login User login.
     *
     * @return integer Period in seconds.
     */
    private function getForbiddenTime(string $login = null): int {
        return $login === 'admin' ? 3600 : 60;
    }

    /**
     * Return Period in seconds for checking failed attempts.
     *
     * @param string|null $login User login.
     *
     * @return integer Period in seconds.
     */
    private function getAttemptsPeriod(string $login = null): int {
        return $login === 'admin' ? 3600 : 60;
    }

    public function authFail(string $login = null) {
        file_put_contents($this->getForbiddenFile($login), time() . PHP_EOL,FILE_APPEND);
    }

    public function isForbidden(string $login = null) {
        $forbiddenFilename = $this->getForbiddenFile($login);
        if (file_exists($forbiddenFilename)) {
            $content = file($forbiddenFilename);

            if (trim($content[0]) == 'forbidden' && !empty($content[1])) {
                $expiration = ($content[1] + $this->getForbiddenTime($login));
                $time       = time();
                if ($expiration > $time) {
                    $minutes = ceil(($expiration - $time) / 60);
                    $this->getLog()->warning(__LINE__, sprintf('The user "%s" is still blocked for next %s minutes', $login, $minutes));
                    return true;
                } else {
                    unlink($forbiddenFilename);
                }
            } else if (!empty($content)) {
                $counter = 0;
                $limit   = (time() - $this->getAttemptsPeriod($login));

                foreach ($content as $key => $time) {
                    if ($time && ($time * 1) > $limit) {
                        $counter++;
                    }
                }

                if ($counter >= 5) {
                    file_put_contents($forbiddenFilename, 'forbidden' . PHP_EOL . time());
                    $this->getLog()->warning(__LINE__, sprintf('Blocking user "%s" after %s failed authentications', $login, $counter));
                    return true;
                }
            }
        }
        return false;
    }

    public function getUserHash(array $userData = []): string {
        $data = [
            $userData[self::NAME] ?? '',
            $userData[self::EXTENSION] ?? '',
            $userData[self::ID] ?? '',
            $userData[self::PASSWORD] ?? '',
            $userData[self::SIP_PASSWORD] ?? '',
            $userData[self::EMAIL] ?? '',
            self::$_salt,
        ];

        return md5(implode('', $data));
    }

    public function getAuthMode(): string {
        return $this->_authMode;
    }

    private function _setAuthMode(string $mode) {
        $this->_authMode = $mode;
        $this->getLog()->info(__LINE__, 'User Auth set mode: ' . $mode);
    }

    private function _getDiContainer(): \Wildix\DependencyInjection\Container {
        if (!$this->_diContainer) {
            $this->_diContainer = new \Wildix\DependencyInjection\Container();
        }

        return $this->_diContainer;
    }
}
