<?php

use Valitron\Validator as BaseValidator;

class Validator extends BaseValidator
{

    /**
     * Setup validation
     *
     * @param  array                     $data
     * @param  array                     $fields
     * @throws \InvalidArgumentException
     */
    public function __construct($data = array(), $fields = array())
    {
        // Allows filtering of used input fields against optional second array of field names allowed
        // This is useful for limiting raw $_POST or $_GET data to only known fields
        $this->_fields = !empty($fields) ? array_intersect_key($data, array_flip($fields)) : $data;

    }

    protected function validateUnsignedInteger($field, $value, $params) {
        return $this->validateInteger($field, $value, $params) && $value >= 0;
    }

    public function errorsTree(): array
    {
        $res = [];
        foreach ($this->_errors as $key => $value){
            $keys  = explode('.',$key);
            $val   = $value;
            $local = [];
            for ($i = count($keys) - 1; $i >= 0; $i--) {
                $local = [$keys[$i]=>$val];
                $val = $local;
            }
            $res = array_replace_recursive($res, $local);
        }

        return $res;
    }

    /**
     * Add an error to error messages array
     *
     * @param string $field
     * @param string $msg
     * @param array  $params
     * @param string $reason
     */
    public function error($field, $msg, array $params = array(), $reason = '')
    {
        $msg = $this->checkAndSetLabel($field, $msg, $params);
        $reason = $this->checkAndSetLabel($field, $reason, $params);

        $values = array();
        // Printed values need to be in string format
        foreach ($params as $param) {
            if (is_array($param)) {
                $param = "['" . implode("', '", $param) . "']";
            }
            if ($param instanceof \DateTime) {
                $param = $param->format('Y-m-d');
            } else {
                if (is_object($param)) {
                    $param = get_class($param);
                }
            }
            // Use custom label instead of field name if set
            if (is_string($params[0])) {
                if (isset($this->_labels[$param])) {
                    $param = $this->_labels[$param];
                }
            }
            $values[] = $param;
        }

        $this->_errors[$field] = array(
            'code' => $msg,
            'reason' => !empty($reason) ? $reason : \Wildix\Error::getMessage($msg)
        );
    }

    public function reason($msg)
    {
        $this->_validations[count($this->_validations) - 1]['reason'] = $msg;

        return $this;
    }

    /**
     * Run validations and return boolean result
     *
     * @return boolean
     */
    public function validate()
    {
        foreach ($this->_validations as $v) {
            foreach ($v['fields'] as $field) {
                list($values, $multiple) = $this->getPart($this->_fields, explode('.', $field));

                // Don't validate if the field is not required and the value is empty
                if ($this->hasRule('optional', $field) && isset($values)) {
                    //Continue with execution below if statement
                } elseif ($v['rule'] !== 'required' && !$this->hasRule('required', $field) && (! isset($values) || $values === '' || ($multiple && count($values) == 0))) {
                    continue;
                }

                // Callback is user-specified or assumed method on class
                $errors = $this->getRules();
                if (isset($errors[$v['rule']])) {
                    $callback = $errors[$v['rule']];
                } else {
                    $callback = array($this, 'validate' . ucfirst($v['rule']));
                }

                if (!$multiple) {
                    $values = array($values);
                }

                $result = true;
                foreach ($values as $value) {
                    $result = $result && call_user_func($callback, $field, $value, $v['params'], $this->_fields);
                }
                if (!$result) {
                    $this->error($field, $v['message'], $v['params'], (string)($v['reason'] ?? ''));
                }
            }
        }

        return count($this->errors()) === 0;
    }

    /**
     * Convenience method to add a single validation rule
     *
     * @param  string|callback           $rule
     * @param  array|string              $fields
     * @return $this
     * @throws \InvalidArgumentException
     */
    public function rule($rule, $fields)
    {
        // Get any other arguments passed to function
        $params = array_slice(func_get_args(), 2);

        if (is_callable($rule)
            && !(is_string($rule) && $this->hasValidator($rule)))
        {
            $name = $this->getUniqueRuleName($fields);
            $msg = isset($params[0]) ? $params[0] : null;
            $this->addInstanceRule($name, $rule, $msg);
            $rule = $name;
        }

        $errors = $this->getRules();
        if (!isset($errors[$rule])) {
            $ruleMethod = 'validate' . ucfirst($rule);
            if (!method_exists($this, $ruleMethod)) {
                throw new \InvalidArgumentException("Rule '" . $rule . "' has not been registered with " . __CLASS__ . "::addRule().");
            }
        }

        // Ensure rule has an accompanying message
        $msgs = $this->getRuleMessages();
        $message = isset($msgs[$rule]) ? $msgs[$rule] : self::ERROR_DEFAULT;

        $this->_validations[] = array(
            'rule' => $rule,
            'fields' => (array) $fields,
            'params' => (array) $params,
            'message' => $message
        );

        return $this;
    }

}
