<?php

namespace Wildix\Auth\AmazonCognito;

use Wildix\Utils\JWT;
use Firebase\JWT\SignatureInvalidException;
use Wildix\Services\Aws;

/**
 * Class AuthService
 */
class AuthService {

    const ALG = 'RS256';

    private $_decoded          = [];
    private $_error            = '';
    private $_invalidSignature = false;

    private static $_configStorage;

    /**
     * Time for reload Amazon public keys
     */
    const TIMEOUT_FOR_RELOAD_PUBLIC_KEY = 60 * 15;

    /**
     * AWS request service
     *
     * @var Aws $_awsRequestService
     */
    private $_awsRequestService;

    public function __construct() {
        JWT::$leeway = 60;
    }

    /**
     * Validate AWS JWT token
     *
     * @param string $iss Iss parameter.
     * @param string $jwt JWT token.
     *
     * @return boolean
     */
    public function validate(string $iss, string $jwt): bool {
        $keys = $this->getAwsPublicKeys($iss);

        $status = true;
        try {
            $decoded = JWT::decode($jwt, $keys, [self::ALG]);

            $this->_decoded = (array)$decoded;
        } catch (SignatureInvalidException $e) {
            $this->_error            = $e->getMessage();
            $this->_invalidSignature = true;

            $status = false;
        } catch (\Exception $e) {
            $this->_error = $e->getMessage();
            $status       = false;

            if ($this->_error === '"kid" invalid, unable to lookup correct key') {
                $this->_invalidSignature = true;
            }
        }

        return $status;
    }

    /**
     * Try reload Amazon public keys
     *
     * @param string $iss Iss parameter.
     *
     * @return boolean
     */
    public function tryReloadPublicKeys(string $iss): bool {
        $path = self::_getStorage()->getPathForStoredPublicKey($iss);

        if (file_exists($path)
            && filemtime($path) + self::TIMEOUT_FOR_RELOAD_PUBLIC_KEY < time()
        ) {
            $data = $this->_loadAndSavePublicKeys($iss);
            if (!empty($data)) {
                return true;
            }
        }

        return false;
    }

    /**
     * Get token payload data
     *
     * @return array
     */
    public function getTokenData(): array {
        return $this->_decoded;
    }

    /**
     * Get validation error message
     *
     * @return string
     */
    public function getErrorMessage(): string {
        return $this->_error;
    }

    /**
     * Check invalid signature
     *
     * @return boolean
     */
    public function isInvalidSignature(): bool {
        return $this->_invalidSignature;
    }

    /**
     * Is Amazon Cognito token
     *
     * @param string $iss Parameter iss from JWT token.
     *
     * @return boolean
     */
    public static function isAmazonCognitoToken(string $iss): bool {
        $pool = self::_getPoolByIss($iss);

        return $pool !== null;
    }

    private static function _getPoolByIss(string $iss): ?array {
        $config = self::getConfig();

        if ($config && $iss && isset($config['pools']) && is_array($config['pools'])) {
            foreach ($config['pools'] as $pool) {
                if (isset($pool[Storage::FIELD_ISS_URL]) && $pool[Storage::FIELD_ISS_URL] === $iss) {
                    return $pool;
                }
            }
        }

        return null;
    }

    /**
     * Get Amazon Cognito config
     *
     * @return array
     */
    public static function getConfig() {
        return self::_getStorage()->getConfig();
    }

    /**
     * Get AWS public keys for current ISS
     *
     * @param string $iss Iss parameter.
     *
     * @return array
     */
    public function getAwsPublicKeys(string $iss): array {
        $data = $this->_getStorage()->getStoredPublicKeys($iss);
        if (!empty($data)) {
            return $data;
        }

        return $this->_loadAndSavePublicKeys($iss);
    }

    private static function _getStorage() {
        if (self::$_configStorage === null) {
            self::$_configStorage = new Storage();
        }

        return self::$_configStorage;
    }

    private function _loadAndSavePublicKeys(string $iss): array {
        self::_getStorage()->deleteSavedPublicKeys($iss);

        $pool     = self::_getPoolByIss($iss);
        $keysData = $this->_getAwsRequestService()->getAwsPublicKeys($pool[Storage::FIELD_PUBLIC_TOKEN_URL]);

        self::_getStorage()->savePublicKeys($iss, $keysData);

        return $keysData;
    }

    private function _getAwsRequestService(): Aws {
        if ($this->_awsRequestService === null) {
            $this->_awsRequestService = new Aws();
        }

        return $this->_awsRequestService;
    }
}
