<?php

namespace Wildix\Auth\AmazonCognito;

use Cache\FileCache;
use Services\Helpers;
use Wildix\Services\GlobalConfig;

/**
 * Class Storage
 */
class Storage {

    const FIELD_ENABLE           = 'enable';
    const FIELD_ISS_URL          = 'iss_url';
    const FIELD_PUBLIC_TOKEN_URL = 'token_url';

    const CACHE_TTL   = 3600;
    const CACHE_GROUP = 'configs';

    private $_publickKeyPathTpl = '/rw2/etc/pbx/cognitoPublicKeys_%s.json';

    /**
     * Global config class
     *
     * @var GlobalConfig|null
     */
    private $_globalConfig;

    public function __construct(GlobalConfig $globalConfig = null, FileCache $cache = null) {
        if ($globalConfig === null) {
            $globalConfig = new GlobalConfig(2, env('SERVICE_CONFIGS_URI') . '/');
        }

        if ($cache === null) {
            $cache = FileCache::getInstance();
        }

        $this->_globalConfig = $globalConfig;
        $this->_cache        = $cache;
    }

    /**
     * Get AWS Cognito config
     *
     * @return null|array
     */
    public function getConfig(): ?array {
        $data = $this->_cache->get(GlobalConfig::TYPE_COGNITO, self::CACHE_GROUP);
        if ($data === false) {
            $data = $this->_getConfigFromServer();

            if ($data !== null) {
                $this->_cache->set(GlobalConfig::TYPE_COGNITO, $data, self::CACHE_TTL, self::CACHE_GROUP);
            }
        }

        return $data;
    }

    private function _getConfigFromServer() {
        try {
            $data = $this->_globalConfig->getConfig(GlobalConfig::TYPE_COGNITO);
        } catch (\Exception $e) {
            return null;
        }

        return $data;
    }

    /**
     * Get stored AWS public keys
     *
     * @param string $iss Iss parameter.
     *
     * @return array|null
     */
    public function getStoredPublicKeys(string $iss): ?array {
        $path = $this->getPathForStoredPublicKey($iss);

        if (!Helpers::isReadableFile($path)) {
            return null;
        }

        $data = json_decode(file_get_contents($path), true);
        if (empty($data)) {
            return null;
        }

        return $data;
    }

    /**
     * Save AWS public keys
     *
     * @param string $iss      Iss parameter.
     * @param array  $keysData Amazon cognito public keys.
     *
     * @return boolean
     */
    public function savePublicKeys(string $iss, array $keysData): bool {
        return file_put_contents($this->getPathForStoredPublicKey($iss), json_encode($keysData)) !== false;
    }

    /**
     * Delete saved AWS public keys
     *
     * @param string $iss Iss parameter.
     *
     * @return void
     */
    public function deleteSavedPublicKeys(string $iss): void {
        $path = $this->getPathForStoredPublicKey($iss);

        if (Helpers::isReadableFile($path)) {
            unlink($path);
        }
    }

    /**
     * Get path for stored public keys for current iss
     *
     * @param string $iss Iss parameter.
     *
     * @return string
     */
    public function getPathForStoredPublicKey(string $iss): string {
        return sprintf($this->_publickKeyPathTpl, md5($iss));
    }
}
