<?php

namespace Wildix\Auth\Api;

use Logger\Log;
use Wildix\Exception;
use Wildix\Utils\JWT;

/**
 * Class ApiAuthenticator
 */
class ApiAuthenticator {

    const         ALGORITHM       = 'HS256';
    private const HEADER_X_APP_ID = 'x-app-id';
    private const HEADER_HOST     = 'host';

    /**
     * Service parse api keys data.
     *
     * @var ApiKeysPairData
     */
    private $_apiKeysPairData;
    /**
     * Log service.
     *
     * @var Log
     */
    private $_log;

    /**
     * ApiAuthenticator constructor.
     *
     * @param ApiKeysPairData $apiKeysPairData Api keys pair data.
     * @param Log             $log             Logger.
     */
    public function __construct(
        ApiKeysPairData $apiKeysPairData,
        Log $log
    ) {
        $this->_apiKeysPairData = $apiKeysPairData;
        $this->_log             = $log;
    }

    public static function getAppId(): ?string {
        return $_SERVER['HTTP_X_APP_ID'] ?? null;
    }

    /**
     * Checks if is s2s authentication request.
     *
     * @return boolean
     */
    public static function isS2sAuth(): bool {
        return (self::getAppId() && \strpos((string)($_SERVER['HTTP_AUTHORIZATION'] ?? ''), 'Bearer ') === 0);
    }

    /**
     * Returns canonical request string.
     *
     * @param string $encodedToken  Encoded token.
     * @param string $requestMethod Request method.
     * @param string $requestUri    Request uri.
     * @param array  $headers       Request headers.
     * @param array  $getParams     Request GET params.
     * @param array  $postParams    Request POST params.
     *
     * @return boolean
     * @throws \Exception When token is not valid.
     */
    public function authenticate(
        string $encodedToken,
        string $requestMethod,
        string $requestUri,
        array $headers,
        array $getParams,
        array $postParams
    ): bool {
        if ($encodedToken === '') {
            return false;
        }

        try {
            $headers    = array_change_key_case($headers);
            $requestUri = (strpos($requestUri, '?') !== false) ? substr($requestUri, 0, strrpos($requestUri, '?')) : $requestUri;

            $this->checkRequiredHeaders($headers);

            $appId           = $headers[self::HEADER_X_APP_ID];
            $apiKeyPair      = $this->getApiKeyPair($appId);
            $secret          = $apiKeyPair[ApiKeysPairData::KEY_SECRET] ?? '';
            $integrationName = $apiKeyPair[ApiKeysPairData::KEY_APP] ?? '';

            $this->getLog()->info(__LINE__, 'API s2s JWT token authentication. APP: ' . $integrationName);
            $token              = $this->parseToken($encodedToken, $secret, $integrationName);
            $requestSign        = (array)$token['sign'];
            $hash               = isset($requestSign['hash']) && is_string($requestSign['hash']) ? $requestSign['hash'] : '';
            $alg                = isset($requestSign['alg']) && is_string($requestSign['alg']) ? $requestSign['alg'] : 'sha256';
            $signedHeadersNames = (array)($requestSign['headers'] ?? []);

            if (!in_array($alg, hash_algos())) {
                throw new Exception(sprintf('The specified encryption algorithm "%s" is unknown', $alg), 500);
            }

            $signedHeaders = array_intersect_key($headers, array_change_key_case(array_flip($signedHeadersNames)));
            ksort($signedHeaders);

            $canonicalRequest = $this->getCanonicalRequest($requestMethod, $requestUri, $getParams, $postParams, $signedHeaders);
            $hashedRequest    = hash($alg, $canonicalRequest);
            $comparisonResult = ($hash === $hashedRequest);

            if ($comparisonResult === false) {
                throw new Exception('Request signature does not match hash. APP: ' . $integrationName, 500);
            }

            $this->getLog()->info(__LINE__, sprintf('API s2s JWT token authentication success. APP: %s', $integrationName));

            return true;
        } catch (\Exception $e) {
            $this->getLog()->error(__LINE__, sprintf('API s2s JWT token authentication failed. Error: %s APP: %s', $e->getMessage(), $integrationName ?? 'unknown'));
            return false;
        }
    }

    /**
     * Returns canonical request string.
     *
     * @param string $requestMethod Request method.
     * @param string $requestUri    Request uri.
     * @param array  $getParams     Request GET params.
     * @param array  $postParams    Request POST params.
     * @param array  $signedHeaders Signed headers.
     *
     * @return string
     */
    private function getCanonicalRequest(
        string $requestMethod,
        string $requestUri,
        array $getParams,
        array $postParams,
        array $signedHeaders
    ): string {
        $canonicalHeaders  = $this->getCanonicalData($signedHeaders);
        $canonicalGetData  = $this->getCanonicalData($getParams);
        $canonicalPostData = $this->getCanonicalData($postParams);

        return ($requestMethod . $requestUri . $canonicalHeaders . $canonicalGetData . $canonicalPostData);
    }

    /**
     * Format data to canonical query string.
     *
     * @param array $data Canonical data.
     *
     * @return string
     */
    private function getCanonicalData(array $data): string {
        $canonicalData = '';
        $data          = array_change_key_case($data);
        ksort($data);

        foreach ($data as $key => $value) {
            $str = (is_array($value)) ? $this->getCanonicalData($value) : trim($value);

            $canonicalData .= $key . ':' . $str . ';';
        }

        return $canonicalData;
    }

    /**
     * Parse and returns decoded token.
     *
     * @param string $encodedToken Encoded token.
     * @param string $secret       Secret key.
     * @param string $issuer       Issuer (who created and signed this token).
     *
     * @return array
     * @throws \Exception When token is not valid.
     */
    private function parseToken(string $encodedToken, string $secret, string $issuer): array {
        if ($issuer === '') {
            throw new Exception('API integration name is not specified', 401);
        }

        $token = (array)JWT::decode($encodedToken, $secret, [self::ALGORITHM]);
        $sign  = (array)$token['sign'];

        if ($issuer !== $token['iss']) {
            throw new Exception('JWT token issuer not valid', 401);
        }
        if ($sign['alg'] !== 'sha256') {
            throw new Exception(sprintf('The specified encryption algorithm "%s" is unknown', $sign['alg']), 401);
        }

        return $token;
    }

    /**
     * Return api key pair by App ID.
     *
     * @param string $appId App ID.
     *
     * @return array
     * @throws \Exception When can not find api key pair by appId.
     */
    private function getApiKeyPair(string $appId): array {
        $apiKeyPair = $this->_apiKeysPairData->getInfoByAppId($appId);

        if ($apiKeyPair === null) {
            throw new Exception('Could not find api key pair for APP ID: "' . $appId . '"', 500);
        }

        return $apiKeyPair;
    }

    /**
     * Check required headers.
     *
     * @param array $headers Request headers.
     *
     * @return void
     * @throws \Exception When required header not exist.
     */
    private function checkRequiredHeaders(array $headers): void {
        if (empty($headers[self::HEADER_X_APP_ID])) {
            throw new Exception('Could not find header name "' . self::HEADER_X_APP_ID . '"', 500);
        }
        if (empty($headers[self::HEADER_HOST])) {
            throw new Exception('Could not find header name "' . self::HEADER_HOST . '"', 500);
        }
    }

    private function getLog(): Log {
        return $this->_log;
    }
}
