<?php

namespace Wildix\Auth\Api;

use Wildix\Exception;

/**
 * Class ApiKeysPairData
 */
class ApiKeysPairData {

    public const KEY_APP      = 'app';
    public const KEY_ID       = 'id';
    public const KEY_SECRET   = 'secret';
    public const KEY_PBX_USER = 'pbxUser';
    public const KEY_CREATED  = 'created';

    protected const FIELD_APP      = 'app';
    protected const FIELD_ID       = 'id';
    protected const FIELD_SECRET   = 'secret';
    protected const FIELD_PBX_USER = 'pbxUser';
    protected const FIELD_CREATED  = 'created';

    protected const API_KEYS_PAIR_DIR            = '/rw2/etc/api_keys_pair';
    protected const API_KEYS_PAIR_PORTAL_FILE    = self::API_KEYS_PAIR_DIR . '/portal_keys.ini';
    protected const API_KEYS_PAIR_CUSTOMERS_FILE = self::API_KEYS_PAIR_DIR . '/customers_keys.ini';
    protected const TMPFS_API_KEYS_PAIR_FILE     = '/dev/shm/api_keys_pair/all_keys.ini';

    protected const APP_FIELDS = [
        self::FIELD_ID,
        self::FIELD_APP,
        self::FIELD_SECRET,
        self::FIELD_PBX_USER,
        self::FIELD_CREATED,
    ];

    /**
     * Validator class.
     *
     * @var \Wildix\Validator
     */
    protected $_validator;

    /**
     * Api keys pairs.
     *
     * @var array|null
     */
    private $_apiKeyPairs;

    /**
     * Log service.
     *
     * @var \Logger\Log
     */
    private $_log;

    /**
     * ApiKeysPairData constructor.
     *
     * @param \Logger\Log $log Logger.
     */
    public function __construct(\Logger\Log $log) {
        $this->_log = $log;

        if (!is_file(self::TMPFS_API_KEYS_PAIR_FILE)) {
            $this->getLog()->info(__LINE__, 'Api keys pairs data file does not exist in tmpfs');
            $this->writeApiKeysPairToTmpFs();
        }
    }

    /**
     * Set validator.
     *
     * @param \Wildix\Validator $validator Validator class.
     *
     * @return void
     */
    public function setValidator(\Wildix\Validator $validator): void {
        $this->_validator = $validator;
    }

    /**
     * Returns key pair data by App ID.
     *
     * @param string $appId App ID.
     *
     * @return array|null
     */
    public function getInfoByAppId(string $appId): ?array {
        if ($this->_apiKeyPairs === null) {
            $this->_apiKeyPairs = $this->getApiKeysPair(self::TMPFS_API_KEYS_PAIR_FILE);
        }

        return $this->_apiKeyPairs[$appId] ?? null;
    }

    /**
     * Update api keys pair data.
     *
     * @param array  $keyPairsData Keys pairs data.
     * @param string $file         Keys file name.
     *
     * @throws Exception Not found validator class.
     *
     * @return void
     */
    protected function updateApiKeysPair(array $keyPairsData, string $file = self::API_KEYS_PAIR_CUSTOMERS_FILE): void {
        $this->getLog()->info(__LINE__, sprintf('Update api keys pairs data %s', $file));

        $prepareData = $this->prepareData($keyPairsData);
        $this->saveApiKeysPairs($file, $prepareData);

        if (is_file(self::API_KEYS_PAIR_PORTAL_FILE)) {
            $this->writeApiKeysPairToTmpFs();
        } else {
            $this->getLog()->error(__LINE__, sprintf('Can not update api keys pair file %s', $file));
        }
    }

    /**
     * Update portal keys.
     *
     * @param array $keyPairsData Keys pairs data.
     *
     * @throws Exception Not found validator class.
     *
     * @return void
     */
    public function updatePortalApiKeysPair(array $keyPairsData): void {
        $this->updateApiKeysPair($keyPairsData, self::API_KEYS_PAIR_PORTAL_FILE);
    }

    /**
     * Generate application ID and secret key.
     *
     * @param array $app Input application data.
     *
     * @return array
     */
    protected function generateAppIdAndSecret(array $app): array {
        $key  = 's2s-' . preg_replace('/[^a-z0-9]/u', '', mb_strtolower($app[self::KEY_APP]));
        $key .= '-' . preg_replace('/[^0-9]/', '', microtime());

        $app[self::KEY_ID]      = $key;
        $app[self::KEY_SECRET]  = \Wildix\Generator\StringGenerator::random(64);
        $app[self::KEY_CREATED] = time();

        return $app;
    }

    /**
     * Validate application config.
     *
     * @param mixed $data App config.
     *
     * @throws Exception Not found validator class.
     *
     * @return array
     */
    protected function validateApp($data): array {
        if ($this->_validator === null) {
            $this->getLog()->error(__LINE__, 'Not found validator class');
            throw new Exception(\Wildix\Error::UnknownError);
        }

        $this->_validator->reset();
        $this->_validator->rule(
            'required',
            self::APP_FIELDS
        );

        $errors = $this->_validator->validateData($data, self::APP_FIELDS);
        if (!empty($errors)) {
            $this->getLog()->error(__LINE__, sprintf('Validation error AppID %s %s', $data[self::FIELD_ID], json_encode($errors)));
        }

        return $errors;
    }

    /**
     * Prepare keys pair data to save.
     *
     * @param array $keyPairsData Keys pair data.
     *
     * @throws Exception Not found validator class.
     *
     * @return array
     */
    private function prepareData(array $keyPairsData): array {
        $result = [];
        foreach ($keyPairsData as $data) {
            $appId = (string)($data[self::FIELD_ID] ?? '');
            if ($appId === '') {
                $this->getLog()->error(__LINE__, sprintf('Not found key "%s" in api keys pair data', self::FIELD_ID));
                continue;
            }
            if (isset($result[$appId])) {
                $this->getLog()->error(__LINE__, sprintf('Duplicate APP ID "%s" in api keys pair data', $appId));
                continue;
            }

            $app = [
                self::KEY_APP => (string)($data[self::FIELD_APP] ?? ''),
                self::KEY_ID => $appId,
                self::KEY_SECRET => (string)($data[self::FIELD_SECRET] ?? ''),
                self::KEY_PBX_USER => (string)($data[self::FIELD_PBX_USER] ?? ''),
                self::KEY_CREATED => (int)($data[self::FIELD_CREATED] ?? time()),
            ];

            $errors = $this->validateApp($app);
            if (!empty($errors)) {
                continue;
            }

            $result[$appId] = $app;
        }

        return $result;
    }

    /**
     * Write WMP portal and customers api keys pairs to file to tmpfs.
     *
     * @return void
     */
    private function writeApiKeysPairToTmpFs(): void {
        $this->getLog()->info(__LINE__, 'Write api keys pairs data file to tmpfs');

        $this->saveApiKeysPairs(self::TMPFS_API_KEYS_PAIR_FILE, $this->mergeAllApiKeysPair());
    }

    /**
     * Merge WMP portal and customers api keys pairs.
     *
     * @return array
     */
    private function mergeAllApiKeysPair(): array {
        $this->getLog()->info(__LINE__, 'Read portal api keys pairs data file');
        $portalApiKeysPair = $this->getApiKeysPair(self::API_KEYS_PAIR_PORTAL_FILE);

        $this->getLog()->info(__LINE__, 'Read customers api keys pairs data file');
        $customersApiKeysPair = $this->getApiKeysPair(self::API_KEYS_PAIR_CUSTOMERS_FILE);

        $duplicatedAppId = array_intersect_key($portalApiKeysPair, $customersApiKeysPair);

        foreach ($duplicatedAppId as $id) {
            $this->getLog()->info(__LINE__, sprintf('Duplicate app id: %s', $id));
        }

        return array_merge($portalApiKeysPair, $customersApiKeysPair);
    }

    /**
     * Parse and return data from ini file.
     *
     * @param string $file Path to file.
     *
     * @return array
     */
    protected function getApiKeysPair(string $file): array {
        if (is_file($file) && is_readable($file)) {
            $data = parse_ini_file($file, true);
            return is_array($data) ? $data : [];
        } else {
            $this->getLog()->info(__LINE__, 'Api keys pairs file does not exist or does not readable');
        }

        return [];
    }

    /**
     * Seve api keys pair to file.
     *
     * @param string $file        File path.
     * @param array  $apiKeysPair Api keys pair.
     *
     * @return boolean
     */
    private function saveApiKeysPairs(string $file, array $apiKeysPair): bool {
        $this->createDir(dirname($file));
        $status = write_ini_file($file, $apiKeysPair);
        $this->changeMode($file);

        if ($status === false) {
            $this->getLog()->error(__LINE__, sprintf('Can not save api keys pair to file %s', $file));
        }
        return $status;
    }

    /**
     * Change owner, group and mode of file.
     *
     * @param string $path File path.
     *
     * @return void
     */
    private function changeMode(string $path): void {
        if (file_exists($path)) {
            $this->getLog()->info(__LINE__, 'Change user, group and mode for api keys pairs source');

            chmod($path, 0755);
            chown($path, 'wms');
            chgrp($path, 'wms');
        }
    }

    /**
     * Create directory.
     *
     * @param string $dir Directory path.
     *
     * @return void
     */
    private function createDir(string $dir): void {
        if (!is_dir($dir)) {
            mkdir($dir, true);
            $this->changeMode($dir);
        }
    }

    protected function getLog(): \Logger\Log {
        return $this->_log;
    }
}
