<?php

namespace Wildix\Auth\Api;

use Wildix\UserDataCleaner\Interfaces\UserDataCleanWorkerInterface;

/**
 * Class CustomerApplications
 */
class CustomerApplications extends ApiKeysPairData implements UserDataCleanWorkerInterface {

    protected const APP_FIELDS = [
        self::FIELD_APP,
        self::FIELD_PBX_USER,
    ];

    public function __construct(\Logger\Log $log, \Wildix\Validator $validator) {
        parent::__construct($log);
        $this->setValidator($validator);
    }

    /**
     * Get application items.
     *
     * @return array
     */
    public function getItems(): array {
        return $this->getApiKeysPair(self::API_KEYS_PAIR_CUSTOMERS_FILE);
    }

    public function buildItem($item): array {
        $item['created'] = (int)$item['created'];
        return $item;
    }

    /**
     * Create application.
     *
     * @param array $inputData Application data.
     *
     * @return array
     * @throws \Wildix\InvalidDataException Validation error.
     * @throws \Wildix\Exception            Not found validator class.
     */
    public function create(array $inputData): array {
        $this->validateInputData($inputData);

        $this->getLog()->info(__LINE__, sprintf('Create customer application %s', $inputData[self::KEY_APP]));
        $appData      = $this->generateAppIdAndSecret($inputData);
        $customerApps = $this->getItems();

        $customerApps[$appData[self::KEY_ID]] = $appData;

        $this->updateApiKeysPair($customerApps, self::API_KEYS_PAIR_CUSTOMERS_FILE);

        return $appData;
    }

    /**
     * Validate input application data.
     *
     * @param array $data Application data.
     *
     * @throws \Wildix\InvalidDataException Validation error.
     *
     * @return void
     */
    private function validateInputData(array $data): void {
        $this->_validator->rule(
            'required',
            self::APP_FIELDS
        );

        $errors = $this->_validator->validateData($data, self::APP_FIELDS);
        if (!empty($errors)) {
            $this->getLog()->error(__LINE__, sprintf('Validation error. Customer AppID %s %s', $data[self::KEY_ID] ?? '', json_encode($data)));
            throw new \Wildix\InvalidDataException(\Wildix\Error::getMessage(\Wildix\Error::VALIDATION_ERROR), $errors);
        }
    }

    /**
     * Update application.
     *
     * @param string $appId     Application ID.
     * @param array  $inputData Application data.
     *
     * @throws \Wildix\InvalidDataException Validation error.
     *
     * @return array
     */
    public function update(string $appId, array $inputData): array {
        $this->validateInputData($inputData);

        $this->getLog()->info(__LINE__, sprintf('Update customer application ID %s', $appId));
        if (!$this->_hasApplication($appId)) {
            throw new \Wildix\InvalidDataException(sprintf('Not existed application ID %s', $appId));
        }

        $customerApps = $this->getItems();
        $appData      = array_merge($customerApps[$appId], $inputData);

        $customerApps[$appId] = $appData;

        $this->updateApiKeysPair($customerApps, self::API_KEYS_PAIR_CUSTOMERS_FILE);

        return $appData;
    }

    /**
     * Delete application
     *
     * @param string $appId Application ID.
     *
     * @return boolean
     * @throws \Wildix\InvalidDataException Validation error.
     */
    public function remove(string $appId): bool {
        $this->getLog()->info(__LINE__, sprintf('Delete customer application ID %s', $appId));
        if (!$this->_hasApplication($appId)) {
            throw new \Wildix\InvalidDataException(sprintf('Not existed application ID %s', $appId));
        }

        $customerApps = $this->getItems();
        unset($customerApps[$appId]);

        $this->updateApiKeysPair($customerApps, self::API_KEYS_PAIR_CUSTOMERS_FILE);

        return true;
    }

    public function removeByExtension(string $extension): void {
        $apps = array_filter(
            $this->getItems(),
            function ($applications) use ($extension) {
                return $applications['pbxUser'] === $extension;
            }
        );

        foreach ($apps as $application) {
            $this->remove($application['id']);
        }
    }

    /**
     * Check application.
     *
     * @param string $appId Application ID.
     *
     * @return boolean
     */
    private function _hasApplication(string $appId): bool {
        return array_key_exists($appId, $this->getItems());
    }
}
