<?php

namespace Wildix\Auth\Applications\Abstractions;

use Wildix\Auth\Applications\Interfaces\ApplicationsInterface;
use Wildix\Auth\Applications\Interfaces\ModelInterface;
use Wildix\Auth\Applications\Interfaces\StorageInterface;
use Wildix\Exception;
use Wildix\Settings\ConfigFile;

/**
 * Class AbstractFileStorage
 */
abstract class AbstractFileStorage implements StorageInterface {

    const APP_CONFIG_FILE = '';

    /**
     * Config file class
     *
     * @var ConfigFile $_config
     */
    private $_configFile;

    /**
     * Get application items
     *
     * @return ApplicationsInterface
     */
    public function getItems(): ApplicationsInterface {
        return $this->fromStorageFormat($this->_getConfig()->get());
    }

    abstract protected function fromStorageFormat(array $items): ApplicationsInterface;

    abstract protected function toStorageFormat(array $items): array;

    /**
     * Get application
     *
     * @param string $appId Application ID.
     *
     * @return ModelInterface|null
     */
    public function getItem(string $appId): ?ModelInterface {
        return $this->getItems()->offsetGet($appId);
    }

    /**
     * Create application
     *
     * @param ModelInterface $data Input app data.
     *
     * @return ModelInterface Input app data.
     * @throws Exception Save error.
     */
    public function create(ModelInterface $data): ModelInterface {
        $appId = $data->getId();

        $items = $this->getItems();
        if ($items->offsetExists($appId)) {
            throw new Exception(\Wildix\Error::BadRequest);
        }

        $items->offsetSet($appId, $data);

        $this->_saveFormatted($items);

        return $data;
    }

    /**
     * Update application
     *
     * @param string         $appId Application ID.
     * @param ModelInterface $data  Input app data.
     *
     * @return void
     * @throws Exception Save error.
     */
    public function update(string $appId, ModelInterface $data): void {
        $items = $this->getItems();
        if (!$items->offsetExists($appId)) {
            throw new Exception(\Wildix\Error::NotFound, 400);
        }

        $items->offsetSet($appId, $data);

        $this->_saveFormatted($items);
    }

    /**
     * Delete application
     *
     * @param string $appId Application ID.
     *
     * @return void
     * @throws Exception Save error.
     */
    public function delete(string $appId): void {
        $items = $this->getItems();
        if (!$items->offsetExists($appId)) {
            throw new Exception(\Wildix\Error::NotFound, 404);
        }

        $items->offsetUnset($appId);

        $this->_saveFormatted($items);
    }

    private function _saveFormatted(ApplicationsInterface $items): void {
        $data = $this->toStorageFormat($items->toArray(true));
        $this->_getConfig()->setData($data);
        $this->_getConfig()->save();
    }

    private function _getConfig(): ConfigFile {
        if (static::APP_CONFIG_FILE === '') {
            throw new Exception('Empty file path for config file');
        }

        if (!$this->_configFile) {
            $this->_configFile = new ConfigFile(static::APP_CONFIG_FILE);
        }
        return $this->_configFile;
    }

    /**
     * Set config
     *
     * @param ConfigFile $config Config file class.
     *
     * @return void
     */
    public function setConfig(ConfigFile $config): void {
        $this->_configFile = $config;
    }
}
