<?php

namespace Wildix\Auth\Applications\OAuth2\Objects;

use Api\Support\Interfaces\ArrayableInterface;
use Wildix\Auth\Applications\Abstractions\Model as AbstractModel;

/**
 * Class Model
 */
class Model extends AbstractModel implements ArrayableInterface {

    const FIELD_REDIRECT_URI = 'redirectUri';
    const FIELD_ORIGINS      = 'origins';
    const FIELD_ACCESS_TTL   = 'accessTokenTtl';

    const FIELDS_EDITABLE = [
        self::FIELD_NAME,
        self::FIELD_REDIRECT_URI,
        self::FIELD_ACCESS_TTL,
    ];

    const FIELDS_SECURE = [
        self::FIELD_ID,
        self::FIELD_SECRET,
        self::FIELD_CREATED,
        self::FIELD_ORIGINS,
    ];

    const PREFIX_NAME = 'oauth2-';

    /**
     * Allowed TTL for access_token (minutes)
     */
    const ALLOWED_TTL = [1, 5, 10, 30, 60, 180, 360, 720, 1440, 4320, 10080, 20160, 40320];
    const DEFAULT_TTL = 1440;


    public function __construct(?array $properties) {
        parent::__construct($properties);
        $this->generateOrigins();
    }

    /**
     * Set redirect URIs
     *
     * @param array $uriArr URI list.
     *
     * @return self
     */
    public function setRedirectUri(array $uriArr): self {
        $this->_data[self::FIELD_REDIRECT_URI] = array_map('trim', $uriArr);

        return $this;
    }

    /**
     * Set access token TTL (minutes)
     *
     * @param integer $ttl TTL minute.
     *
     * @return self
     */
    public function setAccessTokenTtl(int $ttl = 10): self {
        if (!in_array($ttl, self::ALLOWED_TTL)) {
            $ttl = self::DEFAULT_TTL;
        }
        $this->_data[self::FIELD_ACCESS_TTL] = $ttl;

        return $this;
    }

    /**
     * Get application redirect URIs
     *
     * @return array
     */
    public function getRedirectUri(): array {
        return !empty($this->_data[self::FIELD_REDIRECT_URI]) ? $this->_data[self::FIELD_REDIRECT_URI] : [];
    }

    /**
     * Get access token TTL (minutes)]
     *
     * @return integer
     */
    public function getAccessTokenTtl(): int {
        return $this->_data[self::FIELD_ACCESS_TTL] ?? self::DEFAULT_TTL;
    }

    /**
     * Get application origins
     *
     * @return array
     */
    public function getOrigins(): array {
        return $this->_data[self::FIELD_ORIGINS] ?? [];
    }

    /**
     * Generate origins from redirect URIs
     *
     * @return void
     */
    public function generateOrigins(): void {
        $result = [];

        $uriArr = $this->getRedirectUri();
        foreach ($uriArr as $uri) {
            $data = parse_url($uri);
            if (empty($data['host']) || empty($data['scheme'])) {
                continue;
            }
            $origin = $data['scheme'] . '://' . $data['host'];
            if (isset($data['port'])) {
                $origin .= ':' . $data['port'];
            }
            $result[] = $origin;
        }

        $this->_data[self::FIELD_ORIGINS] = array_values(array_unique($result));
    }

    public function toArray(): array {
        return [
            self::FIELD_ID => $this->getProperty(self::FIELD_ID),
            self::FIELD_NAME => $this->getName(),
            self::FIELD_REDIRECT_URI => $this->getRedirectUri(),
            self::FIELD_SECRET => $this->getSecret(),
            self::FIELD_CREATED => $this->getCreated(),
            self::FIELD_ORIGINS => $this->getOrigins(),
            self::FIELD_ACCESS_TTL => $this->getAccessTokenTtl(),
        ];
    }
}
