<?php

namespace Wildix\Auth\Applications\OAuth2\Services;

use Wildix\Auth\Applications\Interfaces\ModelInterface;
use Wildix\Auth\Applications\Interfaces\OAuth2ServiceInterface;
use Wildix\Auth\Applications\Interfaces\StorageInterface;
use Wildix\Auth\Applications\OAuth2\Objects\Applications;
use Wildix\Auth\Applications\OAuth2\Objects\Model;
use Wildix\Auth\Applications\OAuth2\Requests\Request;
use Wildix\Auth\Applications\OAuth2\Storage\FileStorage;
use Wildix\Error;
use Wildix\InvalidDataException;

/**
 * Class Service
 */
class Service implements OAuth2ServiceInterface {

    /**
     * Application storage interface
     *
     * @var StorageInterface
     */
    private static $_storage;

    /**
     * Logger class
     *
     * @var \Logger\Log $_log
     */
    private $_log;

    public function __construct(
        StorageInterface $storage,
        \Logger\Log      $logger
    ) {
        self::$_storage = $storage;
        $this->_log     = $logger;
    }

    /**
     * Set validator.
     *
     * @param \Wildix\Validator $validator Validator class.
     *
     * @return void
     */
    public function setValidator(\Wildix\Validator $validator): void {
        $this->_validator = $validator;
    }

    /**
     * Get application list.
     *
     * @return Applications
     */
    public function getItems(): Applications {
        return $this->getStorage()->getItems();
    }

    /**
     * Get application item.
     *
     * @param string $appId Application ID.
     *
     * @return ModelInterface|null
     */
    public function getItem(string $appId): ?ModelInterface {
        return $this->getStorage()->getItem($appId);
    }

    /**
     * Get application by ID.
     *
     * @param string $appId Application ID.
     *
     * @return Model|null
     */
    public static function getAppById(string $appId): ?Model {
        if (self::$_storage === null) {
            self::$_storage = new FileStorage();
        }

        return self::$_storage->getItem($appId);
    }

    /**
     * Get all origins from OAuth2 applications
     *
     * @return array
     */
    public function getAllOrigins(): array {
        $originArr = [];
        if (self::$_storage === null) {
            self::$_storage = new FileStorage();
        }
        $apps = self::$_storage->getItems()->all();
        foreach ($apps as $app) {
            $originArr[] = $app->getOrigins();
        }

        if (!empty($originArr)) {
            $originArr = array_unique(array_merge(...$originArr));
        }

        return $originArr;
    }

    /**
     * Create application
     *
     * @param Request $data Input app data.
     *
     * @return Model
     * @throws \Wildix\InvalidDataException Invalid data.
     */
    public function create(Request $data): Model {
        $model = new Model($data->toArray());

        $this->getStorage()->create($model);

        $this->getLog()->info(__LINE__, sprintf('Created OAuth2 application %s', $model->getId()));

        return $model;
    }

    /**
     * Update application
     *
     * @param string  $appId Application ID.
     * @param Request $data  Input app data.
     *
     * @return Model
     * @throws \Wildix\InvalidDataException Invalid data.
     */
    public function update(string $appId, Request $data): Model {
        $item = $this->getItem($appId);
        if ($item === null) {
            throw new \Wildix\InvalidDataException(Error::getMessage(Error::NotFound), [], 404);
        }

        $item->setName($data->getName())
            ->setRedirectUri($data->getRedirectUri())
            ->setAccessTokenTtl($data->accessTokenTtl);

        $this->getStorage()->update($appId, $item);

        $this->getLog()->info(__LINE__, sprintf('Updated OAuth2 application %s', $item->getId()));

        return $item;
    }


    /**
     * Delete application
     *
     * @param string $appId Application ID.
     *
     * @return void
     * @throws InvalidDataException Invalid data.
     */
    public function remove(string $appId): void {
        $item = $this->getItem($appId);
        if ($item === null) {
            throw new InvalidDataException(sprintf('Not existed application ID %s', $appId), [], 400);
        }

        $this->getStorage()->delete($appId);

        $this->getLog()->info(__LINE__, sprintf('Removed OAuth2 application %s', $appId));
    }

    /**
     * Validate application
     *
     * @param Model $model Application model.
     *
     * @return void
     * @throws \Wildix\InvalidDataException Invalid data.
     */
    private function _validate(Model $model): void {
        $data      = $model->getAsArray();
        $allFields = array_merge(Model::FIELDS_EDITABLE, Model::FIELDS_SECURE);

        // stage 1 - validate all fields
        $this->_validator->rule(
            'required',
            $allFields
        );
        $this->_validator->rule(
            'array',
            Model::FIELD_REDIRECT_URI
        );

        $errors = $this->_validator->validateData($data, $allFields);
        $this->_validator->reset();

        // stage 2, validate array of URLs
        $this->_validator->rule(
            'url',
            Model::FIELD_REDIRECT_URI
        );
        foreach ($data[Model::FIELD_REDIRECT_URI] as $idx => $item) {
            $tmpErrors = $this->_validator->validateData([Model::FIELD_REDIRECT_URI => $item], [Model::FIELD_REDIRECT_URI]);
            if ($tmpErrors) {
                if (empty($errors[Model::FIELD_REDIRECT_URI])) {
                    $errors = $errors + $tmpErrors;
                }

                $errors[Model::FIELD_REDIRECT_URI]['rows'][] = [
                    'key'   => $idx,
                    'value' => $item,
                ];
            }
        }

        if (!empty($errors)) {
            $this->getLog()->error(__LINE__, sprintf('Validation error. Oauth2 AppID %s %s', $data[Model::FIELD_ID] ?? '', json_encode($data)));
            throw new \Wildix\InvalidDataException(\Wildix\Error::getMessage(\Wildix\Error::VALIDATION_ERROR), $errors);
        }
    }

    /**
     * Get storage
     *
     * @return StorageInterface
     */
    protected function getStorage(): StorageInterface {
        return self::$_storage;
    }

    /**
     * Get logger
     *
     * @return \Logger\Log
     */
    protected function getLog(): \Logger\Log {
        return $this->_log;
    }
}
