<?php

namespace Wildix\Auth\Applications\SimpleToken\Services;

use Wildix\Auth\Applications\Interfaces\ModelInterface;
use Wildix\Auth\Applications\Interfaces\SimpleTokenServiceInterface;
use Wildix\Auth\Applications\Interfaces\StorageInterface;
use Wildix\Auth\Applications\SimpleToken\Objects\Applications;
use Wildix\Auth\Applications\SimpleToken\Objects\Model;
use Wildix\Auth\Applications\SimpleToken\Requests\Request;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\UserDataCleaner\Interfaces\UserDataCleanWorkerInterface;

/**
 * Class Service
 */
class Service implements SimpleTokenServiceInterface, UserDataCleanWorkerInterface {

    /**
     * Application storage interface
     *
     * @var StorageInterface
     */
    private static $_storage;

    /**
     * Logger class
     *
     * @var \Logger\Log $_log
     */
    private $_log;

    public function __construct(
        StorageInterface $storage,
        \Logger\Log      $logger
    ) {
        self::$_storage = $storage;
        $this->_log     = $logger;
    }

    /**
     * Set validator.
     *
     * @param \Wildix\Validator $validator Validator class.
     *
     * @return void
     */
    public function setValidator(\Wildix\Validator $validator): void {
        $this->_validator = $validator;
    }

    /**
     * Get application list.
     *
     * @return Applications
     */
    public function getItems(): Applications {
        return $this->getStorage()->getItems();
    }

    /**
     * Get application item.
     *
     * @param string $appId Application ID.
     *
     * @return ModelInterface|null
     */
    public function getItem(string $appId): ?ModelInterface {
        return $this->getStorage()->getItem($appId);
    }

    /**
     * Create application
     *
     * @param Request $data Input app data.
     *
     * @return Model
     * @throws \Wildix\InvalidDataException Invalid data.
     */
    public function create(Request $data): Model {
        $model = new Model($data->toArray());

        $this->getStorage()->create($model);

        $this->getLog()->info(__LINE__, sprintf('Created Simple Token application %s', $model->getId()));

        return $model;
    }

    /**
     * Update application
     *
     * @param string  $appId Application ID.
     * @param Request $data  Input app data.
     *
     * @return Model
     * @throws InvalidDataException Invalid data.
     */
    public function update(string $appId, Request $data): Model {
        $item = $this->getItem($appId);
        if ($item === null) {
            throw new InvalidDataException(Error::getMessage(Error::NotFound), [], 404);
        }

        $item->setName($data->getName())
            ->setPbxUser($data->getPbxUser());

        $this->getStorage()->update($appId, $item);

        $this->getLog()->info(__LINE__, sprintf('Updated Simple Token application %s', $item->getId()));

        return $item;
    }


    /**
     * Delete application
     *
     * @param string $appId Application ID.
     *
     * @return void
     * @throws \Wildix\InvalidDataException Invalid data.
     */
    public function remove(string $appId): void {
        $item = $this->getItem($appId);
        if ($item === null) {
            throw new \Wildix\InvalidDataException(sprintf('Not existed application ID %s', $appId));
        }

        $this->getStorage()->delete($appId);

        $this->getLog()->info(__LINE__, sprintf('Removed Simple Token application %s', $appId));
    }

    public function removeByExtension(string $extension): void {
        $apps = $this->getItems()->filter(
            function ($app) use ($extension) {
                return $app->getPbxUser() === $extension;
            }
        );

        $apps->map(
            function ($application) {
                $this->remove($application->getId());
            }
        );
    }

    /**
     * Get storage
     *
     * @return StorageInterface
     */
    protected function getStorage(): StorageInterface {
        return self::$_storage;
    }

    /**
     * Get logger
     *
     * @return \Logger\Log
     */
    protected function getLog(): \Logger\Log {
        return $this->_log;
    }
}
