<?php

namespace Wildix\Auth\Devices;

use Wildix\Error;
use Wildix\Utils\JWT;
use Firebase\JWT\Key;
use Wildix\Exception;

/**
 * Class Authenticator
 */
class Authenticator {

    /**
     * Algorithm used to encode/decode JWT tokens;
     */
    const JWT_ALG = 'HS256';

    /**
     * Devices class.
     *
     * @var \Wildix\Devices\Devices
     */
    private $_devices;
    /**
     * PBX model.
     *
     * @var \Models\PBX
     */
    private $_pbx;
    /**
     * Log service.
     *
     * @var \Logger\Log
     */
    private $_log;

    public function __construct(\Wildix\Devices\Devices $devices, \Models\PBX $pbx, \Logger\Log $log) {
        $this->_devices = $devices;
        $this->_pbx     = $pbx;
        $this->_log     = $log;
    }

    /**
     * Authenticate device by user agent and JWT token.
     *
     * @param string      $ua    Device HTTP User-Agent.
     * @param string|null $token JWT token.
     *
     * @return \Wildix\Devices\Device | null
     * @throws \Wildix\Exception If device not found, device does not have assigned user, assigned user has wrong license or JWT token is wrong.
     */
    public function authenticateDevice(string $ua, ?string $token): ?\Wildix\Devices\Device {
        $device = $this->_devices->getByUserAgent($ua, true);

        if ($device === null) {
            throw new Exception(\Wildix\Error::DeviceNotFound, 401);
        }

        $assignedUsers = $device->getAssignedUsers();
        if (empty($assignedUsers)) {
            $this->_log->warning(__LINE__, 'Device does not have assigned user');
            throw new Exception(\Wildix\Error::NotFoundColleague, 401);
        }

        if ($assignedUsers[0]->getLicenseType() !== \Wildix\License::TYPE_WIZYCONF) {
            $this->_log->warning(
                __LINE__,
                \sprintf(
                    'Assigned user (%s) doesn\'t have a wizyconf license',
                    $assignedUsers[0]->getExtension()
                )
            );
            throw new Exception(\Wildix\Error::WrongLicenseType, 401);
        }

        if (empty($device->getToken())) {
            $this->_log->info(__LINE__, 'Skip checking access token validation for first provision');
            $deviceToken    = $this->_devices->generateToken();
            $updatedDevices = $this->_devices->updateDevice(['token' => $deviceToken], [\Wildix\Devices\Device::MAC => $device->getMac()]);

            if ($updatedDevices !== 1) {
                $this->_log->warning(
                    __LINE__,
                    sprintf('Failed to set token for processed device. Updated %d devices.', $updatedDevices)
                );
            }

            $device->setToken($deviceToken);
            return $device;
        }

        if (empty($token) || !$this->validateWizyconfToken($token, $device->getToken())) {
            $code = $this->_devices->generatePinForResetByMac($device->getMac());
            $this->_log->warning(
                __LINE__,
                \sprintf(
                    'Wizyconf device access token not valid; mac: %s, refresh code: %s, token: %s',
                    $device->getMac(),
                    $code,
                    $token
                )
            );
            $exception = new Exception($code, 401);
            $exception->setMessage(Error::getMessage(Error::UnknownError));
            throw $exception;
        }

        $this->_log->info(__LINE__, 'Device auth data checked');
        return $device;
    }

    /**
     * Validation Wizyconf refresh token.
     *
     * @param string $token  JWT token.
     * @param string $secret Device token from devdb.
     *
     * @return boolean
     */
    private function validateWizyconfToken(string $token, string $secret): bool {
        $decodedToken = null;

        try {
            $key          = new Key($secret, self::JWT_ALG);
            $decodedToken = (array)JWT::decode($token, $key);
        } catch (\Exception $e) {
            $this->_log->error(
                __LINE__,
                \sprintf('JWT decode error: %s; token: %s', $e->getMessage(), $token)
            );
            return false;
        }

        if ($decodedToken['exp'] < time()) {
            $this->_log->error(
                __LINE__,
                \sprintf(
                    'Token expired; created: %s, expired: %s, now: %s',
                    $decodedToken['iat'],
                    $decodedToken['exp'],
                    time()
                )
            );
            return false;
        }

        if ($decodedToken['iss'] !== $this->_pbx->getDomain()) {
            $this->_log->error(
                __LINE__,
                \sprintf('Unknown token provider; iss: %s', $decodedToken['iss'])
            );
            return false;
        }

        if (strpos($decodedToken['sub'], 'Wizyconf') === false) {
            $this->_log->error(
                __LINE__,
                \sprintf('Token not for Wizyconf; sub: %s', $decodedToken['sub'])
            );
            return false;
        }

        return true;
    }
}
