<?php

namespace Wildix\Auth\OAuth2;

use Wildix\Auth\Applications\OAuth2\Objects\Model;
use Wildix\Generator\StringGenerator;

/**
 * Class Generator
 */
class Generator {

    /**
     * User class
     *
     * @var \User $_user
     */
    private $_user;

    public function __construct(
        \User $user
    ) {
        $this->_user = $user;
    }

    /**
     * Generate authorization code
     *
     * @param integer $length Code length.
     *
     * @return string
     */
    public function generateAuthorizationCode(int $length = 64): string {
        return StringGenerator::random($length, true, true);
    }

    /**
     * Generate JWT access token
     *
     * @param Model   $app           Application model.
     * @param string  $userExtension User extension.
     * @param string  $userHash      User hash.
     * @param array   $oauthData     OAuth data.
     * @param integer $ttl           TTL for access token, default 1 hour.
     *
     * @return string
     */
    public function generateAccessToken(Model $app, string $userExtension, string $userHash, array $oauthData = [], int $ttl = 3600): string {
        return $this->_generateJwtToken(
            OAuth2Service::PARAM_ACCESS_TOKEN,
            $app,
            $userExtension,
            $userHash,
            $oauthData,
            $ttl
        );
    }


    /**
     * Generate refresh token
     *
     * @param Model   $app           Application model.
     * @param string  $userExtension User extension.
     * @param string  $userHash      User hash.
     * @param array   $oauthData     OAuth data.
     * @param integer $ttl           TTL for refresh token, default 7 day.
     *
     * @return string
     */
    public function generateRefreshToken(Model $app, string $userExtension, string $userHash, array $oauthData = [], int $ttl = 604800): string {
        return $this->_generateJwtToken(
            OAuth2Service::PARAM_REFRESH_TOKEN,
            $app,
            $userExtension,
            $userHash,
            $oauthData,
            $ttl
        );
    }

    /**
     * Generate JWT token
     *
     * @param string  $tokenType     Token type (access or refresh).
     * @param Model   $app           Application model.
     * @param string  $userExtension User extension.
     * @param string  $userHash      User hash.
     * @param array   $oauthData     OAuth data.
     * @param integer $ttl           TTL for access token, default 1 hour.
     *
     * @return string
     */
    private function _generateJwtToken(
        string $tokenType,
        Model  $app,
        string $userExtension,
        string $userHash,
        array  $oauthData = [],
        int    $ttl = 3600
    ) {
        $data = array_merge(
            $oauthData,
            [
                OAuth2Service::PARAM_CLIENT_ID  => $app->getId(),
                OAuth2Service::PARAM_TOKEN_TYPE => $tokenType,
                OAuth2Service::PARAM_HASH       => $userHash,
            ]
        );

        $payload = [
            OAuth2Service::OAUTH2_SESSION_KEY => $data,
            OAuth2Service::PARAM_USER         => $userExtension,
        ];

        return $this->_user->getJwtAuthToken($app->getSecret(), $payload, $ttl);
    }
}
