<?php

namespace Wildix\Auth\OAuth2;

use Firebase\JWT\Key;
use Wildix\Utils\JWT;
use Wildix\Auth\Applications\Interfaces\OAuth2ServiceInterface as OAuth2ApplicationService;
use Wildix\Auth\Applications\OAuth2\Objects\Model;
use Wildix\Colleagues\Colleagues;

/**
 * Class OAuth2Service
 */
class OAuth2Service implements OAuth2ServiceInterface {

    const JWT_ALG = 'HS256';

    const PARAM_CLIENT_ID     = 'client_id';
    const PARAM_CLIENT_SECRET = 'client_secret';
    const PARAM_RESPONSE_TYPE = 'response_type';
    const PARAM_REDIRECT_URI  = 'redirect_uri';
    const PARAM_SCOPE         = 'scope';
    const PARAM_STATE         = 'state';
    const PARAM_PROMPT        = 'prompt';
    const PARAM_GRANT_CONFIRM = 'grant_confirm';
    const PARAM_GRANT_TYPE    = 'grant_type';
    const PARAM_ACCESS_TOKEN  = 'access_token';
    const PARAM_REFRESH_TOKEN = 'refresh_token';
    const PARAM_CODE          = 'code';
    const PARAM_TOKEN_TYPE    = 'type';
    const PARAM_HASH          = 'hash';
    const PARAM_EXPIRES_IN    = 'expires_in';
    const PARAM_ERROR         = 'error';
    const PARAM_USER          = 'user';

    const RESPONSE_TYPE_CODE     = 'code';
    const RESPONSE_TYPE_TOKEN    = 'token';
    const ALLOWED_RESPONSE_TYPES = [self::RESPONSE_TYPE_CODE, self::RESPONSE_TYPE_TOKEN];

    const GRANT_TYPE_AUTHORIZATION_CODE = 'authorization_code';
    const GRANT_TYPE_REFRESH_TOKEN      = 'refresh_token';
    const ALLOWED_GRANT_TYPES           = [self::GRANT_TYPE_AUTHORIZATION_CODE, self::GRANT_TYPE_REFRESH_TOKEN];

    const OAUTH2_GRANT_URL = './oauth2Grant';

    const OAUTH2_SESSION_KEY = 'oauth2';

    const CONFIRM_YES       = 'yes';
    const TOKEN_TYPE_BEARER = 'Bearer';

    const PROMPT_NONE          = 'none';
    const PROMPT_CONSENT       = 'consent';
    const ALLOWED_PROMPT_TYPES = [self::PROMPT_NONE, self::PROMPT_CONSENT];

    /**
     * TTL for Authorization code 1 minute
     */
    const AUTHORIZATION_CODE_TTL = 60;

    /**
     * TTL for refresh token 7 day
     */
    const REFRESH_TOKEN_TTL = 604800;

    /**
     * Generator class
     *
     * @var Generator $_generator
     */
    private $_generator;

    /**
     * Application service class
     *
     * @var OAuth2ApplicationService
     */
    private $_appService;

    /**
     * Application storage class
     *
     * @var Storage $_storage
     */
    private $_storage;

    /**
     * User class
     *
     * @var \User $_user
     */
    private $_user;

    /**
     * Colleagues class
     *
     * @var Colleagues
     */
    private $_colleagues;

    /**
     * Logger class
     *
     * @var \Logger\Log $_log
     */
    private $_log;

    public function __construct(
        OAuth2ApplicationService $appService,
        Storage                  $storage,
        Generator                $generator,
        Colleagues               $colleagues,
        \User                    $user,
        \Logger\Log              $logger
    ) {
        $this->_appService = $appService;
        $this->_storage    = $storage;
        $this->_generator  = $generator;
        $this->_colleagues = $colleagues;
        $this->_user       = $user;
        $this->_log        = $logger;
    }

    /**
     * Handle authorization request
     *
     * @param array $data Input data.
     *
     * @return void
     * @throws OAuth2Exception Validation error.
     */
    public function handleAuthorizationRequest(array $data): void {
        $this->getLog()->info(
            __LINE__,
            sprintf(
                'OAuth2 handle Authorize Request appId=%s responseType=%s',
                $data[self::PARAM_CLIENT_ID] ?? '',
                $data[self::PARAM_RESPONSE_TYPE] ?? ''
            )
        );

        $app = $this->getAppById($data[self::PARAM_CLIENT_ID] ?? '');
        $this->_validateInputAuthorizationParams($app, $data);

        $this->_setSessionData($data);

        $this->getLog()->info(
            __LINE__,
            sprintf(
                'OAuth2 handle Authorize Request success. appId=%s responseType=%s',
                $data[self::PARAM_CLIENT_ID] ?? '',
                $data[self::PARAM_RESPONSE_TYPE] ?? ''
            )
        );
    }

    /**
     * Handle authorization grant request
     *
     * @param array $data Input data.
     *
     * @return Model
     * @throws OAuth2Exception Validation error.
     */
    public function handleAuthorizationGrantRequest(array $data): Model {
        $this->getLog()->info(
            __LINE__,
            sprintf(
                'OAuth2 handle Authorization Grant Request appId=%s',
                $data[self::PARAM_CLIENT_ID] ?? ''
            )
        );

        $app = $this->getAppById($data[self::PARAM_CLIENT_ID] ?? '');
        $this->_validateInputAuthorizationParams($app, $data);

        $this->_validateClientIdInSession($data[self::PARAM_CLIENT_ID]);

        return $app;
    }

    /**
     * Handle confirm request
     *
     * @param array $data Input data.
     *
     * @return boolean
     * @throws OAuth2Exception Validation error.
     */
    public function handleGrantConfirmRequest(array $data): bool {
        $this->getLog()->info(
            __LINE__,
            sprintf(
                'OAuth2 handle Grant Confirm Request appId=%s confirm=%s',
                $data[self::PARAM_CLIENT_ID] ?? '',
                $data[self::PARAM_GRANT_CONFIRM] ?? ''
            )
        );

        $app = $this->getAppById($data[self::PARAM_CLIENT_ID] ?? '');

        if (empty($data[self::PARAM_GRANT_CONFIRM])) {
            throw new OAuth2Exception('Empty grant confirm', OAuth2Exception::ERROR_INVALID_REQUEST);
        }

        $sessionData = $this->getSessionData();

        $this->_validateClientIdInSession($data[self::PARAM_CLIENT_ID]);

        $this->_validateRedirectUri($app, $data[self::PARAM_REDIRECT_URI] ?? '');

        $this->_validateResponseType($data[self::PARAM_RESPONSE_TYPE] ?? '');
        if ($data[self::PARAM_RESPONSE_TYPE] !== $sessionData[self::PARAM_RESPONSE_TYPE]) {
            throw new OAuth2Exception('Invalid OAuth2 response type in session', OAuth2Exception::ERROR_ACCESS_DENIED, 401);
        }

        return $data[self::PARAM_GRANT_CONFIRM] === self::CONFIRM_YES;
    }

    /**
     * Handle access token request
     *
     * @param array $inputData Input data.
     *
     * @return array
     * @throws OAuth2Exception Validation error.
     */
    public function handleAccessTokenRequest(array $inputData): array {
        $this->getLog()->info(
            __LINE__,
            sprintf(
                'OAuth2 handle Access Token Request appId=%s grantType=%s',
                $inputData[self::PARAM_CLIENT_ID] ?? '',
                $inputData[self::PARAM_GRANT_TYPE] ?? ''
            )
        );

        $app = $this->getAppById($inputData[self::PARAM_CLIENT_ID] ?? '');

        $this->_validateRedirectUri($app, $inputData[self::PARAM_REDIRECT_URI] ?? '');
        $this->_validateGrantType($inputData[self::PARAM_GRANT_TYPE] ?? '');
        $this->_validateApplicationSecret($app, $inputData[self::PARAM_CLIENT_SECRET] ?? '');

        $grantType = $inputData[self::PARAM_GRANT_TYPE];
        switch ($grantType) {
            case self::GRANT_TYPE_AUTHORIZATION_CODE:
                $result = $this->_handleAuthorizationCodeGrant($app, $inputData);
                break;
            case self::GRANT_TYPE_REFRESH_TOKEN:
                $result = $this->_handleRefreshTokenGrant($app, $inputData);
                break;
            default:
                throw new OAuth2Exception('Unsupported grant type', OAuth2Exception::ERROR_UNSUPPORTED_GRANT_TYPE);
        }

        return $result;
    }

    /**
     * Handle authorization code request
     *
     * @param Model $app       Application model.
     * @param array $inputData Input data.
     *
     * @return array
     * @throws OAuth2Exception Validation error.
     */
    private function _handleAuthorizationCodeGrant(Model $app, array $inputData): array {
        if (empty($inputData[self::PARAM_CODE])) {
            throw new OAuth2Exception('Empty authorization code', OAuth2Exception::ERROR_INVALID_REQUEST);
        }

        $savedCodeData = $this->_getStorage()->getAuthorizationCode($app->getId(), $inputData[self::PARAM_CODE]);
        if ($savedCodeData === null) {
            throw new OAuth2Exception('Invalid code', OAuth2Exception::ERROR_ACCESS_DENIED, 401);
        }

        if ($savedCodeData[self::PARAM_CLIENT_ID] !== $inputData[self::PARAM_CLIENT_ID]
            || $savedCodeData[self::PARAM_REDIRECT_URI] !== $inputData[self::PARAM_REDIRECT_URI]
        ) {
            throw new OAuth2Exception('Invalid code data', OAuth2Exception::ERROR_ACCESS_DENIED, 401);
        }

        $response = $this->_generateTokens($app, $savedCodeData);

        $this->_getStorage()->deleteAuthorizationCode($app->getId(), $inputData[self::PARAM_CODE]);
        $this->_deleteSessionData();

        return $response;
    }

    /**
     * Handle refresh token grant request
     *
     * @param Model $app       Application model.
     * @param array $inputData Input data.
     *
     * @return array
     * @throws OAuth2Exception Validation error.
     */
    private function _handleRefreshTokenGrant(Model $app, array $inputData): array {
        if (empty($inputData[self::PARAM_REFRESH_TOKEN])) {
            throw new OAuth2Exception('Empty refresh token', OAuth2Exception::ERROR_INVALID_REQUEST);
        }

        $savedTokenData = $this->_decodeAndValidateRefreshToken($inputData[self::PARAM_REFRESH_TOKEN], $app->getSecret());
        $oauthPayload   = $savedTokenData[self::OAUTH2_SESSION_KEY];

        if ($oauthPayload[Storage::FIELD_CLIENT_ID] !== $app->getId()) {
            throw new OAuth2Exception('Invalid refresh token data', OAuth2Exception::ERROR_INVALID_TOKEN, 401);
        }
        $oauthPayload[Storage::FIELD_USER_EXTENSION] = $savedTokenData[self::PARAM_USER];

        return $this->_generateTokens($app, $oauthPayload);
    }

    /**
     * Generate access and refresh tokens
     *
     * @param Model $app       Application model.
     * @param array $savedData Input data.
     *
     * @return array
     */
    private function _generateTokens(Model $app, array $savedData): array {
        $accessTokenTtl = $app->getAccessTokenTtl() * 60;

        $extension = $savedData[Storage::FIELD_USER_EXTENSION];
        $userHash  = $this->_user->getUserHash($this->_getUserDataByExtension($extension));

        $additionalPayload = array_intersect_key($savedData, array_flip([Storage::FIELD_SCOPE]));

        $accessToken  = $this->_generator->generateAccessToken($app, $extension, $userHash, $additionalPayload, $accessTokenTtl);
        $refreshToken = $this->_generator->generateRefreshToken($app, $extension, $userHash, $additionalPayload, self::REFRESH_TOKEN_TTL);

        $response = [
            self::PARAM_ACCESS_TOKEN  => $accessToken,
            self::PARAM_REFRESH_TOKEN => $refreshToken,
            self::PARAM_TOKEN_TYPE    => self::TOKEN_TYPE_BEARER,
            self::PARAM_EXPIRES_IN    => $accessTokenTtl,
        ];

        if (!empty($savedData[self::PARAM_STATE])) {
            $response[self::PARAM_STATE] = $savedData[self::PARAM_STATE];
        }

        return $response;
    }

    /**
     * Decode and validate refresh token
     *
     * @param string $token  Refresh token.
     * @param string $secret Secret key.
     *
     * @return array
     * @throws OAuth2Exception Validation error.
     */
    private function _decodeAndValidateRefreshToken(string $token, string $secret): array {
        try {
            $key          = new Key($secret, self::JWT_ALG);
            $decodedToken = (array)JWT::decode($token, $key);
        } catch (\Exception $e) {
            throw new OAuth2Exception('Invalid refresh token: ' . $e->getMessage(), OAuth2Exception::ERROR_INVALID_REQUEST);
        }

        $decodedToken[self::OAUTH2_SESSION_KEY] = (array)$decodedToken[self::OAUTH2_SESSION_KEY];
        if (empty($decodedToken)
            || $decodedToken['exp'] < time()
            || $decodedToken['iss'] !== \Models\PBX::getInstance()->getDomain()
            || !isset($decodedToken[self::PARAM_USER])
            || !isset($decodedToken[self::OAUTH2_SESSION_KEY][self::PARAM_HASH])
            || !isset($decodedToken[self::OAUTH2_SESSION_KEY][self::PARAM_TOKEN_TYPE])
            || $decodedToken[self::OAUTH2_SESSION_KEY][self::PARAM_TOKEN_TYPE] !== self::PARAM_REFRESH_TOKEN
        ) {
            throw new OAuth2Exception('Invalid refresh token data', OAuth2Exception::ERROR_INVALID_TOKEN, 401);
        }

        $this->_validateUserHash($decodedToken[self::PARAM_USER], $decodedToken[self::OAUTH2_SESSION_KEY][self::PARAM_HASH]);

        return $decodedToken;
    }

    /**
     * Get user data by extension or login
     *
     * @param string $extension User extension or admin login.
     *
     * @return array
     * @throws OAuth2Exception Not found user.
     */
    private function _getUserDataByExtension(string $extension): array {
        $userData = $this->_colleagues->getUserByExtensionOrAdmin($extension);
        if ($userData === null) {
            throw new OAuth2Exception(sprintf('Not found user %s', $extension), OAuth2Exception::ERROR_INVALID_TOKEN, 401);
        }

        return $userData;
    }

    /**
     * Validate Application ID in the session data.
     *
     * @param string $appId Application ID.
     *
     * @return void
     * @throws OAuth2Exception Client ID invalid.
     */
    private function _validateClientIdInSession(string $appId): void {
        $sessionData = $this->getSessionData();
        if ($sessionData === null || empty($sessionData[self::PARAM_CLIENT_ID])) {
            throw new OAuth2Exception('Empty session OAuth2 data', OAuth2Exception::ERROR_ACCESS_DENIED, 401);
        }

        if ($appId !== $sessionData[self::PARAM_CLIENT_ID]) {
            throw new OAuth2Exception('Invalid OAuth2 client ID in session', OAuth2Exception::ERROR_ACCESS_DENIED, 401);
        }
    }

    /**
     * Validate user hash
     *
     * @param string $extension User extension.
     * @param string $hash      User hash.
     *
     * @return void
     * @throws OAuth2Exception Hash is invalid.
     */
    private function _validateUserHash(string $extension, string $hash): void {
        if (empty($hash) || $hash !== $this->_user->getUserHash($this->_getUserDataByExtension($extension))) {
            throw new OAuth2Exception('Invalid user hash', OAuth2Exception::ERROR_INVALID_TOKEN, 401);
        }
    }

    /**
     * Redirect after grant response
     *
     * @param boolean $grantConfirm Grant confirm status.
     *
     * @return void
     * @throws OAuth2Exception Validation error.
     */
    public function redirectAfterGrant(bool $grantConfirm): void {
        $this->getLog()->info(
            __LINE__,
            sprintf(
                'OAuth2 redirect after grant. Confirm=%s',
                (int)$grantConfirm
            )
        );

        $sessionData = $this->getSessionData();
        if ($sessionData === null
            || empty($sessionData[self::PARAM_RESPONSE_TYPE])
            || empty($sessionData[self::PARAM_REDIRECT_URI])
        ) {
            throw new OAuth2Exception('Empty session OAuth2 data', OAuth2Exception::ERROR_ACCESS_DENIED, 401);
        }

        $responseType = $sessionData[self::PARAM_RESPONSE_TYPE];
        switch ($responseType) {
            case self::RESPONSE_TYPE_TOKEN:
                $fullUri = $this->_getUrlForImplicitResponse($grantConfirm, $sessionData);
                break;
            case self::RESPONSE_TYPE_CODE:
                $fullUri = $this->_getUrlForCodeResponse($grantConfirm, $sessionData);
                break;
            default:
                throw new OAuth2Exception('Unsupported response type', OAuth2Exception::ERROR_UNSUPPORTED_RESPONSE_TYPE);
        }

        $this->getLog()->info(
            __LINE__,
            sprintf(
                'OAuth2 grant redirect to url=%s responseType=%s confirm=%s',
                preg_replace('/(access_token=)(.*?)&/', '$1{ACCESS_TOKEN}&', $fullUri),
                $responseType,
                (int)$grantConfirm
            )
        );

        $this->redirectToUrl($fullUri);
    }

    /**
     * Get URL for code response mode
     *
     * @param boolean $grantConfirm Grant confirm status.
     * @param array   $savedData    Saved session data.
     *
     * @return string
     * @throws \Exception Save error.
     */
    private function _getUrlForCodeResponse(bool $grantConfirm, array $savedData): string {
        if ($grantConfirm) {
            $code = $this->_generator->generateAuthorizationCode();
            $this->_saveAuthorizationCode($code);

            $params = [
                self::PARAM_CODE      => $code,
                self::PARAM_CLIENT_ID => $savedData[self::PARAM_CLIENT_ID],
            ];
        } else {
            $params = [
                self::PARAM_ERROR => OAuth2Exception::ERROR_ACCESS_DENIED,
            ];
        }

        if (!empty($savedData[self::PARAM_STATE])) {
            $params[self::PARAM_STATE] = $savedData[self::PARAM_STATE];
        }

        return sprintf('%s?%s', $savedData[self::PARAM_REDIRECT_URI], \http_build_query($params));
    }

    /**
     * Get URL for token response mode (implicit)
     *
     * @param boolean $grantConfirm Grant confirm status.
     * @param array   $savedData    Saved session data.
     *
     * @return string
     * @throws OAuth2Exception Validation error.
     */
    private function _getUrlForImplicitResponse(bool $grantConfirm, array $savedData): string {
        if ($grantConfirm) {
            $app = $this->getAppById($savedData[self::PARAM_CLIENT_ID]);

            $accessTokenTtl = $app->getAccessTokenTtl() * 60;

            $extension = $this->_getUserExtensionOrLogin();
            $userHash  = $this->_user->getUserHash($this->_getUserDataByExtension($extension));

            $additionalPayload = [
                self::PARAM_PROMPT => $savedData[self::PARAM_PROMPT],
                self::PARAM_SCOPE  => $savedData[self::PARAM_SCOPE],
            ];

            $accessToken = $this->_generator->generateAccessToken($app, $extension, $userHash, $additionalPayload, $accessTokenTtl);

            $params = [
                self::PARAM_ACCESS_TOKEN => $accessToken,
                self::PARAM_EXPIRES_IN   => $accessTokenTtl,
            ];
        } else {
            $params = [
                self::PARAM_ERROR => OAuth2Exception::ERROR_ACCESS_DENIED,
            ];
        }

        if (!empty($savedData[self::PARAM_STATE])) {
            $params[self::PARAM_STATE] = $savedData[self::PARAM_STATE];
        }

        return sprintf('%s#%s', $savedData[self::PARAM_REDIRECT_URI], \http_build_query($params));
    }

    /**
     * Validate input authorization params
     *
     * @param Model $app  Application model.
     * @param array $data Input data.
     *
     * @return void
     * @throws OAuth2Exception  Validation error.
     */
    private function _validateInputAuthorizationParams(Model $app, array $data): void {
        $this->_validateRedirectUri($app, $data[self::PARAM_REDIRECT_URI] ?? '');
        $this->_validateResponseType($data[self::PARAM_RESPONSE_TYPE] ?? '');
        $this->_validatePromptType($data[self::PARAM_PROMPT] ?? '');
        $this->_validateScope($app, $data[self::PARAM_SCOPE] ?? '');
    }

    /**
     * Validate redirect URI
     *
     * @param Model  $app Application model.
     * @param string $uri Redirect URI.
     *
     * @return void
     * @throws OAuth2Exception Validation error.
     */
    private function _validateRedirectUri(Model $app, string $uri): void {
        if (empty($uri)) {
            throw new OAuth2Exception('Empty redirect URI', OAuth2Exception::ERROR_INVALID_REQUEST);
        }

        $appUri = $app->getRedirectUri();
        if (empty($appUri)) {
            throw new OAuth2Exception('In application not configured redirect URI', OAuth2Exception::ERROR_INVALID_URI);
        }

        if (!in_array($uri, $appUri, true)) {
            throw new OAuth2Exception('Unknown redirect URI', OAuth2Exception::ERROR_INVALID_URI);
        }
    }

    /**
     * Validate response type
     *
     * @param string $type Response type.
     *
     * @return void
     * @throws OAuth2Exception Validation error.
     */
    private function _validateResponseType(string $type): void {
        if (empty($type)) {
            throw new OAuth2Exception('Empty response type', OAuth2Exception::ERROR_UNSUPPORTED_RESPONSE_TYPE);
        }

        if (!in_array($type, self::ALLOWED_RESPONSE_TYPES)) {
            throw new OAuth2Exception('Invalid response type', OAuth2Exception::ERROR_UNSUPPORTED_RESPONSE_TYPE);
        }
    }

    /**
     * Validate prompt type
     *
     * @param string $type Prompt type.
     *
     * @return void
     * @throws OAuth2Exception Validation error.
     */
    private function _validatePromptType(string $type): void {
        if (empty($type)) {
            throw new OAuth2Exception('Empty prompt type', OAuth2Exception::ERROR_UNSUPPORTED_RESPONSE_TYPE);
        }

        if (!in_array($type, self::ALLOWED_PROMPT_TYPES)) {
            throw new OAuth2Exception('Invalid prompt type', OAuth2Exception::ERROR_UNSUPPORTED_RESPONSE_TYPE);
        }
    }

    /**
     * Validate scope
     *
     * @param Model  $app   Application model.
     * @param string $scope Scope data.
     *
     * @return void
     *
     * @todo need implement on next stage
     */
    private function _validateScope(Model $app, string $scope): void {
    }

    /**
     * Validate grant type
     *
     * @param string $type Grant type.
     *
     * @return void
     * @throws OAuth2Exception Validation error.
     */
    private function _validateGrantType(string $type): void {
        if (empty($type)) {
            throw new OAuth2Exception('Empty grant type', OAuth2Exception::ERROR_INVALID_GRANT);
        }

        if (!in_array($type, self::ALLOWED_GRANT_TYPES)) {
            throw new OAuth2Exception('unsupported_grant_type', OAuth2Exception::ERROR_INVALID_GRANT);
        }
    }

    /**
     * Validate application secret
     *
     * @param Model  $app    Application model.
     * @param string $secret Input secret.
     *
     * @return void
     * @throws OAuth2Exception Validation error.
     */
    private function _validateApplicationSecret(Model $app, string $secret): void {
        if ($secret !== '' && $app->getSecret() !== $secret) {
            throw new OAuth2Exception('Invalid application secret', OAuth2Exception::ERROR_ACCESS_DENIED, 401);
        }
    }

    /**
     * Get application by ID.
     *
     * @param string $clientId Application ID.
     *
     * @return Model
     * @throws OAuth2Exception Validation error.
     */
    public function getAppById(string $clientId): Model {
        if (empty($clientId)) {
            throw new OAuth2Exception('Empty client ID', OAuth2Exception::ERROR_INVALID_CLIENT, 401);
        }

        $appInfo = $this->_getAppService()->getItem($clientId);
        if ($appInfo === null) {
            throw new OAuth2Exception('Not existed client ID', OAuth2Exception::ERROR_INVALID_CLIENT, 401);
        }

        return $appInfo;
    }

    /**
     * Redirect to grant URL.
     *
     * @return void
     * @throws OAuth2Exception Validation error.
     */
    public function redirectToGrantForm(): void {
        $data = $this->getSessionData();
        if ($data === null || empty($data[self::PARAM_CLIENT_ID])) {
            throw new OAuth2Exception('Empty session OAuth2 data', OAuth2Exception::ERROR_ACCESS_DENIED, 401);
        }

        $this->redirectToUrl($this->generateGrantUrl($data));
    }

    /**
     * Generate grant URL
     *
     * @param array $params Request params.
     *
     * @return string
     */
    public function generateGrantUrl(array $params): string {
        return self::OAUTH2_GRANT_URL . '?' . \http_build_query($params);
    }

    /**
     * Redirect to URL
     *
     * @param string $url Url address.
     *
     * @return void
     */
    public function redirectToUrl(string $url): void {
        header('Location: ' . $url, true, 302);
        exit;
    }

    public function isOauthLogged(): bool {
        return $this->_user->isLogged();
    }

    /**
     * Save authorization code in DB
     *
     * @param string $code Authorization code.
     *
     * @return void
     * @throws \Exception DB error.
     */
    private function _saveAuthorizationCode(string $code): void {
        $sessionData = $this->getSessionData();

        $this->_getStorage()->saveAuthorizationCode(
            [
                Storage::FIELD_AUTHORIZATION_CODE => $code,
                Storage::FIELD_CLIENT_ID          => $sessionData[self::PARAM_CLIENT_ID],
                Storage::FIELD_USER_EXTENSION     => $this->_getUserExtensionOrLogin(),
                Storage::FIELD_USER_ID            => (int)$this->_user->getID(),
                Storage::FIELD_REDIRECT_URI       => $sessionData[self::PARAM_REDIRECT_URI],
                Storage::FIELD_STATE              => $sessionData[self::PARAM_STATE],
                Storage::FIELD_SCOPE              => $sessionData[self::PARAM_SCOPE],
                Storage::FIELD_PROMPT             => $sessionData[self::PARAM_PROMPT],
                Storage::FIELD_EXPIRES            => time() + self::AUTHORIZATION_CODE_TTL,
            ]
        );
    }

    /**
     * Get user extension or login
     *
     * @return string
     */
    private function _getUserExtensionOrLogin(): string {
        $data = $this->_user->getExtension();
        if (empty($data)) {
            $data = $this->_user->getName();
        }

        return $data;
    }

    /**
     * Save OAuth2 data to session
     *
     * @param array $data OAuth2 data.
     *
     * @return void
     */
    private function _setSessionData(array $data): void {
        $_SESSION[self::OAUTH2_SESSION_KEY] = $data;
    }

    /**
     * Get OAuth2 data from session
     *
     * @return array|null
     */
    public function getSessionData(): ?array {
        return $_SESSION[self::OAUTH2_SESSION_KEY] ?? null;
    }

    /**
     * Delete OAuth2 data from session
     *
     * @return void
     */
    private function _deleteSessionData(): void {
        if (isset($_SESSION[self::OAUTH2_SESSION_KEY])) {
            unset($_SESSION[self::OAUTH2_SESSION_KEY]);
        }
    }

    /**
     * Get OAuth2 application service
     *
     * @return \Wildix\Auth\Applications\OAuth2\Services\Service
     */
    private function _getAppService(): \Wildix\Auth\Applications\OAuth2\Services\Service {
        return $this->_appService;
    }

    /**
     * Get application storage
     *
     * @return Storage
     */
    private function _getStorage(): Storage {
        return $this->_storage;
    }

    /**
     * Get logger class
     *
     * @return \Logger\Log
     */
    protected function getLog(): \Logger\Log {
        return $this->_log;
    }
}
