<?php

namespace Wildix\Auth\OAuth2;

/**
 * Class Storage
 */
class Storage {

    const FIELD_CLIENT_ID          = 'client_id';
    const FIELD_USER_EXTENSION     = 'user_extension';
    const FIELD_USER_ID            = 'user_id';
    const FIELD_REDIRECT_URI       = 'redirect_uri';
    const FIELD_STATE              = 'state';
    const FIELD_SCOPE              = 'scope';
    const FIELD_PROMPT             = 'prompt';
    const FIELD_AUTHORIZATION_CODE = 'authorization_code';
    const FIELD_EXPIRES            = 'expires';

    const STORAGE_DIR = '/tmp/oauth2';

    /**
     * Logger class
     *
     * @var \Logger\Log $_log
     */
    private $_log;

    public function __construct(
        \Logger\Log $logger
    ) {
        $this->_log = $logger;

        if (!is_dir(self::STORAGE_DIR)) {
            mkdir(self::STORAGE_DIR, 0775);
        }
    }

    /**
     * Save authorization code data
     *
     * @param array $data Saved data.
     *
     * @return void
     */
    public function saveAuthorizationCode(array $data): void {
        $clientId = $data[self::FIELD_CLIENT_ID];
        $code     = $data[self::FIELD_AUTHORIZATION_CODE];
        file_put_contents($this->_getFilePath($clientId, $code), json_encode($data));

        $this->getLog()->info(__LINE__, sprintf('OAuth2 save authorization code for appId=%s', $clientId));
    }


    /**
     * Get data from saved authorization code
     *
     * @param string $clientId Application ID.
     * @param string $code     Authorization code.
     *
     * @return array|null
     */
    public function getAuthorizationCode(string $clientId, string $code): ?array {
        $path = $this->_getFilePath($clientId, $code);
        if (!is_file($path)) {
            return null;
        }

        $data = json_decode(file_get_contents($path), true);
        if (!isset($data[self::FIELD_EXPIRES]) || $data[self::FIELD_EXPIRES] < time()) {
            $this->deleteAuthorizationCode($clientId, $code);
            return null;
        }

        return $data;
    }

    /**
     * Delete authorization code.
     *
     * @param string $clientId Application ID.
     * @param string $code     Authorization code.
     *
     * @return void
     */
    public function deleteAuthorizationCode(string $clientId, string $code): void {
        $path = $this->_getFilePath($clientId, $code);
        if (!file_exists($path)) {
            return;
        }

        unlink($path);

        $this->getLog()->info(__LINE__, sprintf('OAuth2 delete authorization code appId=%s code=%s', $clientId, $code));
    }

    /**
     * Auth code file path
     *
     * @param string $clientId Application ID.
     * @param string $code     Authorization code.
     *
     * @return string
     */
    private function _getFilePath(string $clientId, string $code): string {
        return self::STORAGE_DIR . DIRECTORY_SEPARATOR . basename($clientId . '_' . $code . '.json');
    }

    /**
     * Get logger class
     *
     * @return \Logger\Log
     */
    protected function getLog(): \Logger\Log {
        return $this->_log;
    }
}
