<?php

namespace Wildix\Auth\ServiceAuth;

/**
 * Class PublicKeyRepository
 */
class PublicKeyRepository {

    /**
     * Logger class.
     *
     * @var \Logger\Log
     */
    private $_logger;

    /**
     * File cache.
     *
     * @var \Cache\FileCacheInterface
     */
    private $_cache;

    /**
     * Cache live in seconds (1 hour)
     *
     * @var integer
     */
    private $_cacheTtl = 3600;

    /**
     * Real TTL for cache (3 hours).
     *
     * @var integer
     */
    private $_cacheRealTtl = 3600 * 3;

    /**
     * Minimum interval for reload public key from server in seconds.
     *
     * @var integer
     */
    private $_minimumRenewInterval = 60;

    private $_domain            = '';
    private $_keyUrl            = '';
    private $_lastDownloadedKey = '';

    private $_cachePrefix = 'publicKeys';

    /**
     * HTTP client
     *
     * @var \Wildix\Services\AbstractServiceAuth
     */
    private $_httpClient;

    public function __construct(
        \Wildix\Services\AbstractServiceAuth $httpClient,
        \Cache\FileCacheInterface $cache,
        \Logger\Log $logger
    ) {
        $this->_httpClient = $httpClient;
        $this->_cache      = $cache;
        $this->_logger     = $logger;

        $this->_cache->setRealTtl($this->_cacheRealTtl);
    }

    /**
     * Get public key from cache or site
     *
     * @param string  $domain       Domain name or iss.
     * @param string  $publicKeyUrl Public key URL.
     * @param boolean $force        Force download from remote server.
     *
     * @return string
     */
    public function getPublicKey(string $domain, string $publicKeyUrl, bool $force = false): string {
        $this->_domain = $domain;
        $this->_keyUrl = $publicKeyUrl;

        $this->_logger->info(__LINE__, sprintf('[ServiceAuth] Get public key for domain: %s force: %s', $this->_domain, (int)$force));
        $publicKey = $this->_cache->get($domain, $this->_cachePrefix);
        if ($publicKey === false || $force) {
            $publicKey = $this->_downloadPublicKey();
        } else {
            $this->_logger->debug(__LINE__, '[ServiceAuth] Found public key in cache');
        }

        return $publicKey;
    }

    private function _downloadPublicKey() {
        $this->_logger->debug(__LINE__, sprintf('[ServiceAuth] Download public key from URL %s', $this->_keyUrl));
        $publicKey                = $this->_httpClient->get($this->_keyUrl);
        $this->_lastDownloadedKey = $publicKey;
        $this->_logger->info(__LINE__, sprintf('[ServiceAuth] Downloaded public key: %s', $publicKey));

        return $publicKey;
    }

    /**
     * Save downloaded public key.
     *
     * @return void
     */
    public function saveDownloadedKey(): void {
        if ($this->_lastDownloadedKey === '') {
            return;
        }

        $this->_logger->debug(__LINE__, sprintf('[ServiceAuth] Save public key for domain: %s', $this->_domain));
        $this->_cache->set($this->_domain, $this->_lastDownloadedKey, $this->_cacheTtl, $this->_cachePrefix);
        $this->_lastDownloadedKey = '';
    }

    /**
     * Initialize reload public key from remote server.
     *
     * @return boolean
     */
    public function tryReloadPublicKey(): bool {
        $this->_logger->debug(__LINE__, sprintf('[ServiceAuth] Try reload public key for domain: %s', $this->_domain));

        $created = $this->_cache->getCacheCreatedTime($this->_domain, $this->_cachePrefix);
        if ($created === false) {
            $this->_logger->info(__LINE__, '[ServiceAuth] Not found cached data');
            return false;
        }

        $diff = time() - $created;
        $this->_logger->info(__LINE__, sprintf('[ServiceAuth] Cache was created %d seconds ago', $diff));

        return $diff > $this->_minimumRenewInterval;
    }
}
