<?php

namespace Wildix\Auth\ServiceAuth;

use Wildix\Auth\ServiceAuth\Exceptions\IssuerException;
use Wildix\Auth\ServiceAuth\Exceptions\RecipientException;
use Wildix\Auth\ServiceAuth\Exceptions\UnsupportedServiceAuth;
use Wildix\Auth\ServiceAuth\Services\BaseServiceAuthInterface;
use Wildix\Utils\JWT;

/**
 * Class ServiceAuthenticator
 */
class ServiceAuthenticator {

    private $_supportedAsyncAlgorithm = [
        'RS256',
        'RS384',
        'RS512',
    ];

    private $_classTemplate = '\\Wildix\\Auth\\ServiceAuth\\Services\\%sAuth';

    /**
     * Logger class.
     *
     * @var \Logger\Log
     */
    private $_logger;

    private $_serviceName;

    private $_tokenPayload = [];

    /**
     * Authenticate class
     *
     * @var BaseServiceAuthInterface
     */
    private $_authClass;

    /**
     * Repository for public keys.
     *
     * @var PublicKeyRepository
     */
    private $_repository;

    public function __construct(
        string $pbxSerial,
        PublicKeyRepository $repository,
        \Logger\Log $logger
    ) {
        $this->_pbxSerial  = $pbxSerial;
        $this->_repository = $repository;
        $this->_logger     = $logger;

        JWT::$leeway = 30;
    }

    private $_retryCheckSign = 1;

    /**
     * Force download public key from remote server.
     *
     * @var boolean
     */
    private $_force = false;

    /**
     * Initialise service authorization class by name.
     *
     * @param string $serviceName  Auth service name.
     * @param string $returnObject Auth object.
     *
     * @return mixed
     * @throws UnsupportedServiceAuth Incorrect service auth
     */
    public function initServiceClass(string $serviceName, $returnObject = false) {
        if ($serviceName === '') {
            throw new UnsupportedServiceAuth('Empty auth service name');
        }

        $this->_logger->debug(__LINE__, sprintf('[ServiceAuth] init service name %s', $serviceName));
        $this->_serviceName = $serviceName;

        $className = sprintf($this->_classTemplate, ucfirst($this->_serviceName));

        $this->_logger->debug(__LINE__, sprintf('[ServiceAuth] Find class %s', $className));
        if (!class_exists($className)) {
            $this->_logger->error(__LINE__, sprintf('[ServiceAuth] Not found class %s', $className));
            throw new UnsupportedServiceAuth(sprintf('Unsupported service for auth %s', $className));
        }

        $this->_authClass = new $className;

        return $returnObject ? $this->_authClass : $this;
    }

    /**
     * Get auth service name.
     *
     * @return string
     */
    public function getServiceName(): string {
        return $this->_serviceName;
    }

    /**
     * Authenticate by JWT token.
     *
     * @param string $token JWT token.
     *
     * @return boolean
     * @throws IssuerException Incorrect iss parameter.
     */
    public function authenticate(string $token): bool {
        $this->_tokenPayload = JWT::unpack($token)['payload'];
        $iss                 = $this->_getIssFromPayload($this->_tokenPayload);

        $this->_assertIssNotEmpty($iss);
        $this->_assertTokenForMe($this->_tokenPayload);

        $this->_logger->info(__LINE__, sprintf('[ServiceAuth] Compare iss: %s in auth service: %s', $iss, $this->_serviceName));
        $compareResult = $this->_getAuthClass()->compareIss($iss);
        if (!$compareResult) {
            $this->_logger->debug(__LINE__, '[ServiceAuth] Check parameter iss failed');
            throw new IssuerException('Check parameter iss failed');
        }

        return $this->_checkSignature($token);
    }

    /**
     * Check iss parameter is not empty.
     *
     * @param string $iss Iss parameter.
     *
     * @throws IssuerException Incorrect iss parameter.
     * @return void
     */
    private function _assertIssNotEmpty(string $iss) {
        if ($iss === '') {
            $this->_logger->debug(__LINE__, '[ServiceAuth] Empty parameter iss in JWT token');
            throw new IssuerException('Empty iss parameter');
        }
    }

    /**
     * Check aud parameter. Token for me.
     *
     * @param array $payload JWT token payload.
     *
     * @throws RecipientException Incorrect aud parameter.
     * @return void
     */
    private function _assertTokenForMe(array $payload) {
        $aud = $this->_getAudFromPayload($payload);
        if (empty($aud)) {
            $this->_logger->debug(__LINE__, '[ServiceAuth] Empty parameter aud in JWT token');
            throw new RecipientException('Empty aud parameter');
        }
        if ($aud !== $this->_pbxSerial . '.wildixin.com') {
            $this->_logger->debug(__LINE__, sprintf('[ServiceAuth] Token not for me. Aud: %s', $aud));
            throw new RecipientException('Incorrect aud parameter');
        }
    }

    /**
     * Check JWT sign.
     *
     * @param string $token JWT token.
     *
     * @return boolean
     */
    private function _checkSignature(string $token): bool {
        $iss = $this->_getIssFromPayload($this->_tokenPayload);

        $publicKey = $this->_repository->getPublicKey($iss, $this->_getAuthClass()->getPublicKeyUrl($iss), $this->_force);
        $this->_logger->debug(__LINE__, sprintf('[ServiceAuth] Public key %s', $publicKey));

        try {
            JWT::decode($token, $publicKey, $this->_supportedAsyncAlgorithm);
        } catch (\Exception $e) {
            $this->_logger->error(__LINE__, sprintf('[ServiceAuth] Token validation error: %s', $e->getMessage()));
            $retryResult = false;
            if ($this->_retryCheckSign > 0 && $this->_repository->tryReloadPublicKey()) {
                $this->_retryCheckSign--;

                $this->_force = true;
                $retryResult  = $this->_checkSignature($token);
            }

            if ($this->_retryCheckSign < 0 || $retryResult === false) {
                return false;
            }
        }

        // if validation success save current key to cache
        $this->_repository->saveDownloadedKey();
        $this->_force = false;

        return true;
    }

    private function _getIssFromPayload($payload) {
        return (string)($payload['iss'] ?? '');
    }

    private function _getAudFromPayload($payload) {
        return (string)($payload['aud'] ?? '');
    }

    /**
     * Get auth class.
     *
     * @return BaseServiceAuthInterface
     * @throws UnsupportedServiceAuth Incorrect service auth
     */
    private function _getAuthClass(): BaseServiceAuthInterface {
        if ($this->_authClass === null) {
            throw new UnsupportedServiceAuth('Empty auth class');
        }

        return $this->_authClass;
    }
}
