<?php

namespace Wildix\Auth\ServiceAuth\Services;

use Wildix\Auth\ServiceAuth\Exceptions\UnsupportedCompareType;
use Wildix\Auth\ServiceAuth\Exceptions\UnsupportedServiceAuth;

/**
 * Class BaseServiceAuth
 */
abstract class BaseServiceAuth implements BaseServiceAuthInterface {

    const COMPARE_EQUAL = 'equal';
    const COMPARE_MATH  = 'math';

    protected $_issTemplate = '';
    protected $_compareType = self::COMPARE_EQUAL;

    /**
     * Template for public URL.
     *
     * @var string
     */
    private $_publicKeyUrlTemplate = 'https://:iss/public-key/';

    /**
     * Custom URL for public key for this auth service.
     *
     * @var string
     */
    protected $_customUrlForPublicKey = '';

    /**
     * Compare Iss parameter.
     *
     * @param string $iss Iss parameter.
     *
     * @return boolean
     * @throws UnsupportedCompareType If unsupported compare type.
     * @throws UnsupportedServiceAuth If unsupported auth service.
     */
    public function compareIss(string $iss): bool {
        if ($iss === '') {
            throw new UnsupportedServiceAuth('Empty iss parameter');
        }

        switch ($this->_compareType) {
            case self::COMPARE_EQUAL:
                $result = $this->_issTemplate === $iss;
                break;
            case self::COMPARE_MATH:
                $result = preg_match($this->_issTemplate, $iss);
                break;
            default:
                throw new UnsupportedCompareType('Unsupported compare type. Your configuration for authorization service is wrong.');
        }

        return $result && $this->additionalRules($iss);
    }

    /**
     * Additional rules for check iss.
     *
     * @param string $iss Iss parameter.
     *
     * @return boolean
     */
    protected function additionalRules(string $iss): bool {
        return true;
    }

    /**
     * Get URL for public key of this service.
     *
     * @param string $iss Iss parameter.
     *
     * @return string
     */
    public function getPublicKeyUrl(string $iss): string {
        return $this->_customUrlForPublicKey !== '' ? $this->_customUrlForPublicKey : $this->_makeUrlByIss($iss);
    }

    private function _makeUrlByIss(string $iss): string {
        return str_replace(':iss', $iss, $this->_publicKeyUrlTemplate);
    }
}
