<?php

namespace Wildix\Auth\TwoFactor;

use PHPGangsta_GoogleAuthenticator as GoogleAuthenticator;
use Wildix\Notifications\Email\Interfaces\EmailManagerInterface;

/**
 * Class TwoFactorAuthenticatorFactory
 */
class TwoFactorAuthenticatorFactory {
    const FACTOR_EMAIL = 'Email';
    const FACTOR_SMS   = 'Sms';

    private static $_mapFactorToFactory = array(
        self::FACTOR_EMAIL => '_createEmailAuthenticator',
        self::FACTOR_SMS => '_createSmsAuthenticator',
    );

    /**
     * Transport for email authenticator.
     *
     * @var EmailManagerInterface
     */
    private $_transportEmail;
    /**
     * Transport for sms authenticator.
     *
     * @var \Models\Astman
     */
    private $_transportSms;
    /**
     * Code generator for second factor.
     *
     * @var GoogleAuthenticator
     */
    private $_codeGenerator;

    /**
     * TwoFactorAuthenticatorFactory constructor.
     *
     * @param EmailManagerInterface $transportEmail Transport for sending emails.
     * @param \Models\Astman        $transportSms   Transport for sending sms.
     * @param GoogleAuthenticator   $codeGenerator  Code generator for authentication.
     */
    public function __construct(EmailManagerInterface $transportEmail, \Models\Astman $transportSms, GoogleAuthenticator $codeGenerator) {
        $this->_transportEmail = $transportEmail;
        $this->_transportSms   = $transportSms;
        $this->_codeGenerator  = $codeGenerator;
    }

    /**
     * Check availability an authenticator for second factor.
     *
     * @param string $factor Name of second factor.
     *
     * @return boolean
     */
    public static function hasAuthenticator($factor) {
        return isset(self::$_mapFactorToFactory[$factor]);
    }

    /**
     * Factory method for two factor authenticator.
     *
     * @param string $factor           Name of second factor.
     * @param string $currentPbxDomain Domain of current PBX.
     *
     * @return AbstractAuthenticator
     */
    public function createBy2ndFactor($factor, $currentPbxDomain) {
        return $this->{self::$_mapFactorToFactory[$factor]}($currentPbxDomain);
    }

    /**
     * Return an instance of email authenticator.
     *
     * @param string $currentPbxDomain Domain of current PBX.
     *
     * @return EmailAuthenticator
     */
    private function _createEmailAuthenticator($currentPbxDomain) {
        return new EmailAuthenticator($this->_codeGenerator, $this->_transportEmail, $currentPbxDomain);
    }

    /**
     * Return instance of sms authenticator.
     *
     * @param string $currentPbxDomain Domain of current PBX.
     *
     * @return SmsAuthenticator
     */
    private function _createSmsAuthenticator($currentPbxDomain) {
        return new SmsAuthenticator($this->_codeGenerator, $this->_transportSms, $currentPbxDomain);
    }
}
