<?php
namespace Wildix;

use Wildix\Colleagues\Colleague;
use Wildix\Databases\CTIconnectDB;
use Wildix\Settings\ConfigFile;
use Wildix\Settings\Env\EnvConfig;

class CDRView extends \Singleton {

    private const DEFAULT_DECIMAL = 2;

    /**
     * CdrView Config File
     *
     * @var ConfigFile
     */
    private $_config;
    private $_settings = [];

    public function getSettings(string $userId): array {
        if (!isset($this->_settings[$userId])) {
            $settings = [
                'id' => null,
                'period' => '7', // indicate last two month
                'sla_period' => '20',
                'miss_call_sec_period' => '0'
            ];

            $periods = CTIconnectDB::getInstance()->query('SELECT * FROM `cdr_periods` WHERE `uid`='.$userId);

            if ($periods) {
                if($period = $periods->fetchArray(SQLITE3_ASSOC)){
                    if ($period && !empty($period)) {
                        $settings['id'] = $period['id'];
                        $settings['period'] = $period['value'];
                        $settings['sla_period'] = $period['sla_value'];
                        $settings['miss_call_sec_period'] = $period['miss_call_sec_value'];
                    }
                }
            }
            $settings['grade_of_accuracy']        = $this->_getGradeOfAccuracy();
            $this->_settings[$userId] = $settings;
        }

        return $this->_settings[$userId];
    }

    /**
     *
     * @var array
     */
    private $_keyMap = array(
            'period' => 'value',
            'sla_period' => 'sla_value',
            'miss_call_sec_period' => 'miss_call_sec_value'
    );

    private function getRealKey($key){
        if(isset($this->_keyMap[$key])){
            return $this->_keyMap[$key];
        }
        return $key;
    }

    public function setSettings(string $userId, array $newSettings): bool {

        if (isset($newSettings)) {
            $this->_setGradeOfAccuracy((int)$newSettings['grade_of_accuracy']);
            $dirty    = false;
            $settings = $this->getSettings($userId);
            unset($settings['grade_of_accuracy']);
            foreach (array('period', 'sla_period', 'miss_call_sec_period') as $key){
                if(isset($newSettings[$key]) &&
                        $settings[$key]!=$newSettings[$key] &&
                        $newSettings[$key]!=''
                        && $newSettings[$key] != null ){
                    $dirty = true;
                    $settings[$key] = $newSettings[$key];
                }
            }

            if($dirty){
                $query = '';
                if($settings['id'] != null){
                    $params = array();
                    foreach ($settings as $key=>$value){
                        $params[] = "`".$this->getRealKey($key)."` = '".CTIconnectDB::getInstance()->realEscapeString($value)."'";
                    }

                    $query = 'UPDATE `cdr_periods` SET '.join(', ', $params).' WHERE `uid`='.$userId.';';

                }else{
                    $settings['uid'] = $userId;
                    unset($settings['id']);

                    $keys = array();
                    $values = array();
                    foreach ($settings as $key=>$value){
                        $keys[] = '`'.$this->getRealKey($key).'`';
                        $values[] = '\''.CTIconnectDB::getInstance()->realEscapeString($value).'\'';
                    }

                    $query = sprintf('INSERT INTO `cdr_periods` (%s) Values (%s)', join(', ', $keys), join(', ', $values));
                }
                $result = CTIconnectDB::getInstance()->exec($query);
                if ($result !== true) {
                    return $result;
                }
            }
        }

        return true;
    }

    /**
     *
     * @var array
     */
    private $_filterKeyMap = array(

    );

    private function getFilterRealKey($key){
        if(isset($this->_filterKeyMap[$key])){
            return $this->_filterKeyMap[$key];
        }
        return $key;
    }

    public function getCountFilters($filter, $search=null, $searchFields=null, \Wildix\Colleagues\Colleague $colleague = null){

        $query = 'SELECT count(id) as countRows
                                FROM `cdr_filters`
                                '.$this->createWhere($filter, $search, $searchFields, $colleague, CTIconnectDB::SELECT_QUERY).';';

        $rows = CTIconnectDB::getInstance()->query($query);
        if ($rows) {
            $row = $rows->fetchArray(SQLITE3_ASSOC);
            if($row){
                return $row['countRows'];
            }
        }

        return 0;
    }


    /**
     *
     * @param int $id
     * @param \Wildix\Colleagues\Colleague $colleague
     * @return <NULL, \Wildix\Phonebooks\Contact>
     */
    public function getFilterByID($id, \Wildix\Colleagues\Colleague $colleague = null) {
        $rows = $this->getFilters(array('id' => $id), null, null, null, 1, null, null, $colleague);
        return (count($rows) > 0) ? $rows[0] : null;
    }

    /**
     *
     * @param array $filter
     * @param string $search
     * @param array $searchFields
     * @param int $start
     * @param int $count
     * @param string $sort
     * @param string $dir (ASC | DESC)
     * @param \Wildix\Colleagues\Colleague $colleague
     * @return array <\Wildix\Phonebooks\Contact>
     */
    public function getFilters($filter = null, $search=null, $searchFields=null, $start = null, $count=null, $sort=null, $dir=null, \Wildix\Colleagues\Colleague $colleague = null) {
        $start = is_numeric($start) ? intval($start) : 0;
        $count = is_numeric($count) ? intval($count) : 100;
        $sort = $sort ? $sort : 'id';
        $dir = $dir ? $dir : 'ASC';

        $result = array();

        $query = 'SELECT id, uid, name, description, filter, shared
                    FROM `cdr_filters`
                    '.$this->createWhere($filter, $search, $searchFields, $colleague, CTIconnectDB::SELECT_QUERY).'
                    ORDER BY '.$sort.' '.$dir.'
                    LIMIT '.$start.', '.$count.';';

        $rows = CTIconnectDB::getInstance()->query($query);

        if ($rows) {
            while($row = $rows->fetchArray(SQLITE3_ASSOC)){
                if ($row && !empty($row)) {
                    $result[] = $row;
                }
            }
        }
        return $result;
    }


    /**
     *
     * @param array $contact
     * @param \Wildix\Colleagues\Colleague $colleague
     * @return int
     */
    public function addFilter(string $userId, $filter){
        $filter['uid'] = $userId;

        $keys = array();
        $values = array();
        foreach ($filter as $key=>$value){
            $keys[] = '`'.$this->getRealKey($key).'`';
            $values[] = '\''.CTIconnectDB::getInstance()->realEscapeString($value).'\'';
        }

        $query = sprintf('INSERT INTO `%s` (%s) Values (%s);', 'cdr_filters', join(', ', $keys), join(', ', $values));

        $result = CTIconnectDB::getInstance()->exec($query);
        if($result === true){
            return CTIconnectDB::getInstance()->getLastInsertId();
        }
        return $result;
    }


    public function deleteFilters($filter = null, $search=null, $searchFields=null, \Wildix\Colleagues\Colleague $colleague = null){

        if($filter && $filter['id']){
            $rowFilter = $this->getFilterByID($filter['id'], $colleague);
            if (!$rowFilter) {
                throw new Exception('Not found', 404);
            }
        }

        $query = 'DELETE FROM `cdr_filters`
                                '.$this->createWhere($filter, $search, $searchFields, $colleague, CTIconnectDB::DELETE_QUERY).';';

        $result = CTIconnectDB::getInstance()->exec($query);
        if ($result === true) {
            $result = CTIconnectDB::getInstance()->getCountChangesRows();
        }
        return $result;
    }

    /**
     *
     * @param array $filter
     * @param string $search
     * @param array $searchFields
     * @param \Wildix\Colleagues\Colleague $colleague
     * @return string
     */
    private function createWhere($filter = null, $search = null, $searchFields = null, \Wildix\Colleagues\Colleague $colleague = null, $queryType = null){
        $filter = is_array($filter) ? $filter : array();
        $search = $search ? trim($search) : '';
        $searchFields = $searchFields ? $searchFields : array('name', 'description');
        $colleague = is_object($colleague) ? $colleague : null;

        $where = array();
        if($filter && !empty($filter)){
            foreach ($filter as $key=>$value){
                if(is_scalar($value)){
                    $where[] = "`".$this->getFilterRealKey($key)."` = '".CTIconnectDB::getInstance()->realEscapeString($value)."'";
                }else if(is_array($value)){
                    $tmpWhere = array();
                    if(isset($value['not'])){
                        $tmpWhere[] = "`".$this->getFilterRealKey($key)."` <> '".CTIconnectDB::getInstance()->realEscapeString($value['not'])."'";
                    }
                    if(isset($value['or'])){
                        if(is_array($value['or'])){
                            foreach($value['or'] as $k=>$v){
                                $value['or'][$k] = CTIconnectDB::getInstance()->realEscapeString($v);
                            }
                            $tmpWhere[] = "`".$this->getFilterRealKey($key)."` IN ('".implode("','", $value['or'])."')";
                        }else{
                            $tmpWhere[] = "`".$this->getFilterRealKey($key)."` = '".CTIconnectDB::getInstance()->realEscapeString($value['or'])."'";
                        }
                    }

                    if(!empty($tmpWhere)){
                        $where[] = '('.join(' AND ', $tmpWhere).')';
                    }
                }
            }
        }

        if ($colleague){
            switch ($queryType){
                case CTIconnectDB::DELETE_QUERY:
                case CTIconnectDB::UPDATE_QUERY:
                    $where[] = '`'.$this->getFilterRealKey('uid').'`='.CTIconnectDB::getInstance()->realEscapeString($colleague->getID());
                    break;
                default: $where[] = '(`'.$this->getFilterRealKey('uid').'`='.CTIconnectDB::getInstance()->realEscapeString($colleague->getID()).' OR `'.$this->getFilterRealKey('shared').'`=1)';

            }
        }

        if($search && !empty($search) && count($searchFields)>0){
            $searchWhere = array();
            foreach ($searchFields as $key){
                $searchWhere[] = "`".$this->getFilterRealKey($key)."` LIKE '%".CTIconnectDB::getInstance()->realEscapeString($search)."%'";
            }
            if(!empty($searchWhere)){
                $where[] = '('.join(' OR ', $searchWhere).')';
            }
        }

        if(count($where) > 0){
            return 'WHERE '.join(' AND ', $where);
        }
        return '';
    }

    /**
     * Info about cdr batabase
     * @return array of backend, min_id, max_id, min_date, max_date
     * @throws \Exception
     */
    public function getCDRInfo() {
        //
        $backend = \Wildix\Databases\CDRDB::getInstance()->getInfoString();
        if (!$backend){
            throw new Exception('CDR View backend is not selected',400);
        }
        //
        $filter = array('pbxid'=>'*');
        $start = 0;
        $count = 1;
        $sort = \Wildix\CallHistory\Call::ID;
        $dir = 'ASC';
        $firstCalls = \Wildix\CallHistory\Calls::getInstance()->getCalls($filter, null, null, $start, $count, $sort, $dir, null);
        $dir = 'DESC';
        $lastCalls = \Wildix\CallHistory\Calls::getInstance()->getCalls($filter, null, null, $start, $count, $sort, $dir, null);
        return array(
            "backend"=> $backend,
            "skipped"=> 0,
            "timeout"=> 1,
            "exception"=> null,
            "min_id"=> count($firstCalls)?$firstCalls[0]->getID():'',
            "max_id"=> count($lastCalls)?$lastCalls[0]->getID():'',
            "min_date"=> count($firstCalls)?$firstCalls[0]->getStart():'',
            "max_date"=> count($lastCalls)?$lastCalls[0]->getStart():'',
            "events_limit"=> env('CDRVIEW_CDR_CHUNK_LIMIT', 3000),
        );
    }

    private $trunksPatern = null;
    private function getTrunksPattern() {
        if (!$this->trunksPatern){
            $trunkNames = \Wildix\Databases\WMSDB::getInstance()->getTrunksNamesList();
            # if the list was empty we create a pattern which matches "nothing"
            if (count($trunkNames)==0)
                $this->trunksPatern = "/a^/";
            else {
                array_walk($trunkNames, function(&$value){
                    $value = preg_quote($value, '/');
                });
                $this->trunksPatern =
                    sprintf( "/SIP\/(%s)\-/" , implode("|", $trunkNames)) ;
            }
        }
        return $this->trunksPatern;
    }

    private $userExtensionsList = array();
    private function getUserExtensionsList() {
        if (count($this->userExtensionsList)==0)
            foreach (\Wildix\Colleagues\Colleagues::getInstance()->getUsers() as $user)
                if ($user->getExtension())
                    $this->userExtensionsList[] = $user->getExtension();
        return $this->userExtensionsList;
    }

    private $usersPattern = null;
    private function getUsersPattern() {
        if (!$this->usersPattern){
            if (count($this->getUserExtensionsList())==0)
                $this->usersPattern = "a^";
            else
                $this->usersPattern = sprintf("/(%s)/",implode("|", $this->getUserExtensionsList()));
        }
        return $this->usersPattern;
    }

    private $userChannelsPattern = null;
    private function getUserChannelsPattern() {
        if (!$this->userChannelsPattern){
            if (count($this->getUserExtensionsList())==0)
                $this->userChannelsPattern = "a^";
            else
                $this->userChannelsPattern = sprintf("/SIP\/(%s)\-/",implode("|", $this->getUserExtensionsList()));
        }
        return $this->userChannelsPattern;
    }

    /**
     * Transform to the record which java front app capaible to store to the local client app
     * @param \Wildix\CallHistory\Call $call
     */
    private function getNormalizedCallRecord(\Wildix\CallHistory\Call $call) {
        $record = array(
            'id'=>$call->getID(),
            'from'=>$call->getFromNumber(),
            'fromChannel'=>$call->getChannel(),
            'to'=>$call->getToNumber(),
            'toChannel'=>$call->getDstchannel(),
            'type'=>$call->getApp(),
            'class'=>$call->getCallclass(),
            'destType'=>$call->getDestType(),
            'direction'=>$call->getDestType(),
            'date'=>$call->getStart(),
            'duration'=>$call->getDuration(),
            'billsec'=>$call->getBillsec(),
            'cost'=>$call->getCost(),
            'tags'=>$call->getLastdata(),
            'pbx'=>$call->getPbxid(),
            'userfield'=>$call->getUserfield(),
            'faxdata'=>$call->getFaxdata(),
            'fromName'=>$call->getFromName(),
            'toName'=>$call->getToName(),
        );
        if (!$call->getApp()){ // << app is always zero now ?
            if ($call->getLastapp() == 'VoiceMail')
                $record['type'] = 3; //if disposition is 16 - voicemail
            else if (strlen($call->getFaxdata())>0)
                $record['type'] = 2; //if faxdata is set - fax
            else
                $record['type'] = 1; //unknown
        }
        if (!$call->getCallclass()){
            $record['class'] = 1;
        }
        if (!$call->getDestType()){
            if (
                preg_match($this->getTrunksPattern() , $call->getDstchannel()) ||
                preg_match("/Local\/.+\@mobility\-/" , $call->getDstchannel()) ||
                (preg_match("/Local\/.+\@mobility\-/" , $call->getChannel() && $call->getDstchannel() == ''))
               ){
                // if destination channel is trunk
                // if destination channel is Local/+39234634634@mobility-1234,1
                // if source channel is is Local/+39234634634@mobility-1234,1 and destination channel is empty
                $record['direction'] = 3; //outgoing
            } else if (preg_match($this->getTrunksPattern() , $call->getChannel())){
                // if source channel is trunk
                $record['direction'] = 2; //incoming
            } else if (
                        preg_match("/Local\/cg\_\d+\@queuehelper\-/" , $call->getChannel()) ||
                        ($call->getChannel() == "**Unknown**" && $call->getToNumber() == "s" && $call->getDcontext() == "default") ||
                        ($call->getToNumber() == "" && $call->getDstchannel() == "") ||
                        ($call->getFromNumber() == "" && $call->getDstchannel() == "")
                   ){
                // source channel is like Local/cg_X@queuehelper
                // source channel is "**Unknown**" and dcontext is "default" and dst is "s"
                // dst is empty and dst channel is empty
                // src is empty and channel is empty
                $record['direction'] = 5; //ignore
            } else {
                $matchSrc = preg_match($this->getUsersPattern(), $call->getFromNumber()); // check src
                $matchDst = preg_match($this->getUsersPattern(), $call->getToNumber()); // check dst
                //$matchFrom = self.getUsersPattern()[0].match(inputRow[1]) # check c_from
                //$matchTo = self.getUsersPattern()[0].match(inputRow[3]) # check c_to
                $matchSrcChannel = preg_match($this->getUserChannelsPattern(), $call->getChannel()); // check source channel

                if (
                    ($matchSrc && $matchDst) ||
                    ($matchSrc && $call->getDstchannel() == "") ||
                    ($matchSrcChannel && $call->getDstchannel() == "") ||
                    ($call->getDcontext() == "unknowns" && $call->getDstchannel() == "")
                   )
                    // src and dst are local (c_from and c_to are local)
                    // src is local and dstchannel is not present
                    // source channel is local and dstchannel is not present
                    // dcontext is "unknowns" and dstchannel is not present
                    $record['direction'] = 4; // local
                else if ($matchSrc)
                    $record['direction'] = 3; // outgoing
                else if ($matchDst)
                    $record['direction'] = 2; // incoming
                else
                    $record['direction'] = 1; // unknown

            }
        }
        if (!$call->getStart()){
            $record['date'] = '0000-00-00 00:00:00';
        }
        return $record;
    }

    /**
     * Calls history for given filters
     * @param array $filter
     * @param type $start
     * @param type $count
     * @param type $sort
     * @param type $dir
     * @param Colleague $colleague
     * @param boolean $allowSeeCallRecordings Allow see callrecordings in the calls.
     * @return array of calls min and max call id and total record
     */
    public function getCallsHistory($filter,$start,$count,$sort,$dir, Colleague $colleague = null, bool $allowSeeCallRecordings = true) {
        $total = \Wildix\CallHistory\Calls::getInstance()->getCount($filter, null, null, null);

        $calls = [];
        if($total > 0){
            $calls = \Wildix\CallHistory\Calls::getInstance()->getCalls(
                $filter,
                null,
                null,
                $start,
                $count,
                $sort,
                $dir,
                $colleague,
                $allowSeeCallRecordings
            );
            $count = count($calls);
            for ($i = 0; $i < $count; $i++) {
                $calls[$i] = $this->getNormalizedCallRecord($calls[$i]);
            }
        }
        //
        $aggFilter = $filter;
        unset($aggFilter['exclude']);
        $firstCalls = \Wildix\CallHistory\Calls::getInstance()->getCalls($aggFilter, null, null, 0, 1, $sort, 'ASC', null);
        $lastCalls = \Wildix\CallHistory\Calls::getInstance()->getCalls($aggFilter, null, null, 0, 1, $sort, 'DESC', null);

        return array(
            'total'=>$total,
            "min_id"=> count($firstCalls)?$firstCalls[0]->getID():'',
            "max_id"=> count($lastCalls)?$lastCalls[0]->getID():'',
            'events'=>$calls
        );
    }

    public function getRemoteHost(): string {
        $config = $this->_getConfigParam('CDRVIEW_REMOTE_URI', '');
        return trim($config, '/');
    }

    private function _getGradeOfAccuracy(): int {
        return (int)$this->_getConfigParam('CDRVIEW_GRADE_OF_ACCURACY', self::DEFAULT_DECIMAL);
    }

    private function _getConfig(): ConfigFile {
        if (!$this->_config) {
            $this->_config = new EnvConfig();
        }
        return $this->_config;
    }

    private function _getConfigParam(string $field, $default) {
        $res = $default;
        if ($this->_getConfig()->has($field)) {
            $res = $this->_getConfig()->get($field);
        }

        return $res;
    }

    private function _setGradeOfAccuracy(int $value) {
        $value = ($value > 0 && $value <= 4) ? $value : self::DEFAULT_DECIMAL;
        $this->_getConfig()->set('CDRVIEW_GRADE_OF_ACCURACY', $value)->save();
    }
}
