<?php

namespace Wildix\Calendar;

use Psr\Http\Message\ServerRequestInterface as Request;
use Wildix\Ejabberd;

/**
 * Class CalendarService
 */
final class CalendarService {

    private $_eventRepo;
    private $_settingRepo;
    private $_ejabberdService;

    public function __construct(Events $eventRepo, Settings $settingRepo, Ejabberd $ejabberdService) {
        $this->_eventRepo       = $eventRepo;
        $this->_settingRepo     = $settingRepo;
        $this->_ejabberdService = $ejabberdService;
    }

    public function updateSettings(array $storedSettings, array $settings, $user, $app) {
        $needRemove = array_flip(array_keys($storedSettings));
        foreach ($settings as $key => $value) {
            unset($needRemove[$key]);
        }

        if (!empty($needRemove)) {
            $this->_settingRepo->deleteSettings($needRemove, $user, $app);
        }

        return $this->_settingRepo->addSettings($settings, $user, $app);
    }

    public function createEventsFromRequest(Request $request): array {
        $params = $request->getParsedBody();
        $events = (array)($params['data'] ?? []);
        foreach ($events as $key => $value) {
            $events[$key]['type'] = Events::EVENT_TYPE_CALENDAR;
            $events[$key]['app']  = Events::EVENT_APP_OUTLOOK;
        }

        return $events;
    }

    private function countDontWorkingMinutes($start, $end) {
        $dateTimeZone = new \DateTimeZone('UTC');
        $startDate    = new \DateTime('2009-10-11 ' . $start, $dateTimeZone);
        if ($start > $end) {
            // two day
            $endDate = new \DateTime('2009-10-12 ' . $end, $dateTimeZone);
        } else {
            $endDate = new \DateTime('2009-10-11 ' . $end, $dateTimeZone);
        }

        $interval = $startDate->diff($endDate);
        $minutes  = $interval->days * 24 * 60;
        $minutes += $interval->h * 60;
        $minutes += $interval->i;

        return (24 * 60) - $minutes;
    }

    public function generateWorkHoursEvents($settings, $app = null) {
        $result = [];

        $workDays = explode(',', $settings['workDays']);
        if (empty($workDays)) {
            return $result;
        }

        $dontWorkingMins = $this->countDontWorkingMinutes($settings['workHoursStart'], $settings['workHoursEnd']);

        $dateTimeZone = new \DateTimeZone('UTC');
        $date         = new \DateTime('@' . time(), $dateTimeZone);

        if ($settings['workHoursEnd'] < $settings['workHoursStart'] && $date->format('H:i') < $settings['workHoursStart']) {
            // if start working hours in the future
            $dateEnd   = new \DateTime($date->format('Y-m-d') . ' ' . $settings['workHoursStart']);
            $dateStart = clone $dateEnd;

            $event = array(
                'dateStart' => $dateStart->modify('-' . $dontWorkingMins . ' minutes')->format('Y-m-d\TH:i:s\Z'),
                'dateEnd' => $dateEnd->format('Y-m-d\TH:i:s\Z')
            );

            $id = md5($event['dateStart'] . '-' . $event['dateEnd']);

            $event['id']         = strtoupper($id . substr($id, 0, 16));
            $event['type']       = Events::EVENT_TYPE_WORKING_HOURS;
            $event['name']       = 'Off work';
            $event['status']     = 'OffWork';
            $event['importance'] = 'Normal';
            $event['private']    = '0';
            $event['options']    = '';
            $event['app']        = $app ?? Events::EVENT_APP_OUTLOOK;

            $result[] = $event;
        }

        $date->setTime(0, 0, 0);

        for ($i = 0; $i < 14; $i++) {
            $event = array();
            // calculate events for 2 weeks
            $dayOfWeek = $date->format('N');
            if (in_array($dayOfWeek, $workDays)) {
                // is working day
                if ($settings['workHoursEnd'] > $settings['workHoursStart']) {
                    // if is working time in the day
                    $dateStart = new \DateTime($date->format('Y-m-d') . ' ' . $settings['workHoursEnd']);

                    $event['dateStart'] = $dateStart->format('Y-m-d\TH:i:s\Z');
                    $event['dateEnd']   = $dateStart->modify('+' . $dontWorkingMins . ' minutes')->format('Y-m-d\TH:i:s\Z');
                } else {
                    // if is working time in the night
                    $dateEnd = new \DateTime($date->format('Y-m-d') . ' ' . $settings['workHoursStart']);
                    $dateEnd->modify('+1 day');

                    $dateStart = clone $dateEnd;

                    $event['dateStart'] = $dateStart->modify('-' . $dontWorkingMins . ' minutes')->format('Y-m-d\TH:i:s\Z');
                    $event['dateEnd']   = $dateEnd->format('Y-m-d\TH:i:s\Z');
                }
            } else {
                // is not working day
                $dateStart = new \DateTime($date->format('Y-m-d') . ' ' . $settings['workHoursStart']);

                $event['dateStart'] = $dateStart->format('Y-m-d\TH:i:s\Z');
                $event['dateEnd']   = $dateStart->modify('+1 day')->format('Y-m-d\TH:i:s\Z');
            }

            $id = md5($event['dateStart'] . '-' . $event['dateEnd']);

            $event['id']         = strtoupper($id . substr($id, 0, 16));
            $event['type']       = Events::EVENT_TYPE_WORKING_HOURS;
            $event['name']       = 'Off work';
            $event['status']     = 'OffWork';
            $event['importance'] = 'Normal';
            $event['private']    = '0';
            $event['options']    = '';
            $event['app']        = $app ?? Events::EVENT_APP_OUTLOOK;

            $result[] = $event;

            $date->modify('+1 day');
        }

        return $result;
    }

    public function updateEvents($storedEvents, $newEvents, $user, $app = null, $skipRemove = false) {
        $needRemove = $skipRemove ? [] : array_flip(array_column($storedEvents, 'id'));
        foreach ($newEvents as $key => $value) {
            // find the event in the stored
            $filteredEvents = array_filter(
                $storedEvents,
                function ($item) use (&$value) {
                    return $item['id'] == $value['id'];
                }
            );

            $storedEvent = reset($filteredEvents);
            if ($storedEvent && $storedEvent['processed'] == 1) {
                $diff = array_diff($storedEvent, $newEvents[$key]);
                if (!isset($diff['dateStart']) && !isset($diff['dateEnd']) && !isset($diff['name'])) {
                    // if processed event and not changed restore this flag
                    $newEvents[$key]['processed'] = 1;
                }
            }
            // no need to remove this event
            unset($needRemove[$value['id']]);
        }

        if (!empty($needRemove)) {
            foreach ($needRemove as $key => $value) {
                // find the event in the stored
                $filteredEvents = array_filter(
                    $storedEvents,
                    function ($item) use (&$value) {
                        return $item['id'] == $value;
                    }
                );

                $storedEvent = reset($filteredEvents);
                if ($storedEvent && $storedEvent['processed'] != 1) {
                    continue;
                }

                // if remove processed event need reset status
                $dateTimeZone = new \DateTimeZone('UTC');
                $date         = new \DateTime('now', $dateTimeZone);
                $currentDate  = $date->format('Y-m-d\TH:i:s\Z');
                if ($storedEvent && $currentDate >= $storedEvent['dateStart'] && $currentDate <= $storedEvent['dateEnd']) {
                    // if at this time
                    $userSettings     = $user->getSettings();
                    $userStatus       = $userSettings->getStatus();
                    $newStatus        = $this->_eventRepo->convertEventToStatus($storedEvent, $userSettings->getCalendar()['OUTLOOK']);
                    if ($newStatus['status'] == $userStatus['status']
                        && $newStatus['message'] == $userStatus['message']
                        && $newStatus['until'] == $userStatus['until']
                    ) {
                        $this->_ejabberdService->setUserPresence($user->getExtension(), 'available', '', '');
                    }
                }
            }

            $this->deleteEvents(array_keys($needRemove), $user);
        }

        return $this->_eventRepo->addEvents($newEvents, $user);
    }

    public function deleteEvents(array $eventsIds, $user) {
        $filter = ['id' => [ 'or' => $eventsIds]];
        return $this->_eventRepo->deleteEvents($filter, null, null, $user);
    }
}
