<?php
namespace Wildix\Calendar;

use Wildix\Databases\CalendarDB;
use Wildix\Exception;

class Events extends \Singleton {

	const TABLE_NAME = 'events';

	const EVENT_APP_OUTLOOK = 1;
	const EVENT_APP_OFFICE365_OUTLOOK = 2;

	const EVENT_TYPE_CALENDAR = 1;
	const EVENT_TYPE_WORKING_HOURS = 2;

	const EVENT_FIELD_ID = 'id';
	const EVENT_FIELD_TYPE = 'type';
	const EVENT_FIELD_OWNER = 'owner';
	const EVENT_FIELD_NAME = 'name';
	const EVENT_FIELD_DATE_START = 'dateStart';
	const EVENT_FIELD_DATE_END = 'dateEnd';
	const EVENT_FIELD_STATUS = 'status';
	const EVENT_FIELD_IMPORTANCE = 'importance';
	const EVENT_FIELD_PRIVATE = 'private';
	const EVENT_FIELD_PROCESSED = 'processed';
	const EVENT_FIELD_OPTIONS = 'options';
	const EVENT_FIELD_APP = 'app';

	private $_tableFields = '`id`, `type`, `owner`, `name`, `dateStart`, `dateEnd`, `status`, `importance`, `private`, `processed`, `options`, `app`';

    public function getUserEvents($user, bool $workingHours = false, $app = self::EVENT_APP_OUTLOOK): array {
        $eventFilter = $workingHours ? $this->_getWorkingHoursEventsFilter($app) : $this->_getCalendarEventsFilter($app);

        return $this->getEvents($eventFilter, null, null, 0, 99999, null, null, $user);
    }

    private function _getWorkingHoursEventsFilter($app): array {
        return [
            'type' => self::EVENT_TYPE_WORKING_HOURS,
            'app' => $app,
        ];
    }

    private function _getCalendarEventsFilter($app): array {
        return [
            'type' => self::EVENT_TYPE_CALENDAR,
            'app' => $app,
        ];
    }

	private function isValidData($data, \Wildix\Colleagues\Colleague $colleague = null, $queryType){

		if($queryType == CalendarDB::INSERT_QUERY){
			if(empty($data['id'])){
				return array(
						'code' => '0',
						'message' => 'The id cannot be empty'
				);
			}
			if(!isset($data['name'])){
				return array(
						'code' => '0',
						'message' => 'The name cannot be empty'
				);
			}

		}else if($queryType == CalendarDB::UPDATE_QUERY){

		}

		return true;
	}

	/**
	 *
	 * @var array
	 */
	private $_keyMap = array(

	);

	/**
	 *
	 * @param string $key
	 * @return string
	 */
	private function getRealKey($key){
		if(isset($this->_keyMap[$key])){
			return $this->_keyMap[$key];
		}
		return $key;
	}


	/**
	 *
	 * @param array $filter
	 * @param string $search
	 * @param array $searchFields
	 * @param int $start
	 * @param int $count
	 * @param string $sort
	 * @param string $dir (ASC | DESC)
	 * @param \Wildix\Colleagues\Colleague $colleague
	 * @return array
	 */
	public function getEvents($filter = null, $search=null, $searchFields=null, $start = null, $count=null, $sort=null, $dir=null, \Wildix\Colleagues\Colleague $colleague = null) {
        $start = is_numeric($start) ? intval($start) : 0;
        $count = is_numeric($count) ? intval($count) : 100;
		$sort = $sort ? $sort : self::EVENT_FIELD_DATE_START;
		$dir = $dir ? $dir : 'ASC';

		$events = array();

		$query = 'SELECT '.$this->_tableFields.'
                    FROM `'.self::TABLE_NAME.'`
                    '.$this->createWhere($filter, $search, $searchFields, $colleague, CalendarDB::SELECT_QUERY).'
                    ORDER BY '.$sort.' '.$dir.'
                    LIMIT '.$start.', '.$count.';';

		$rows = CalendarDB::getInstance()->query($query);

		if ($rows) {
			while($row = $rows->fetchArray(SQLITE3_ASSOC)){
				if ($row && !empty($row)) {
					$events[] = $row;
				}
			}
		}
		return $events;
	}

	/**
	 *
	 * @param array $filter
	 * @param string $search
	 * @param array $searchFields
	 * @param \Wildix\Colleagues\Colleague $colleague
	 * @return int
	 */
	public function getCount($filter, $search=null, $searchFields=null, \Wildix\Colleagues\Colleague $colleague = null){

		$query = 'SELECT count() as countRows
                                FROM `'.self::TABLE_NAME.'`
                                '.$this->createWhere($filter, $search, $searchFields, $colleague, CalendarDB::SELECT_QUERY).'';

		$rows = CalendarDB::getInstance()->query($query);
		if ($rows) {
			$row = $rows->fetchArray(SQLITE3_ASSOC);
			if($row){
				return $row['countRows'];
			}
		}
		return 0;
	}


	/**
	 *
	 * @param array $events
	 * @param \Wildix\Colleagues\Colleague $colleague
	 * @return int
	 */
	public function addEvents($events, \Wildix\Colleagues\Colleague $colleague = null){
		$result = array();


		CalendarDB::getInstance()->beginTransaction();

		foreach ($events as $key=>$value){
			$result[] = $this->addEvent($value, $colleague);
		}

		CalendarDB::getInstance()->endTransaction();

		return $result;
	}


	/**
	 *
	 * @param array $event
	 * @param \Wildix\Colleagues\Colleague $colleague
	 * @return int
	 */
	public function addEvent($event, \Wildix\Colleagues\Colleague $colleague = null){
		$event = array_map('trim', $event);

		$isValid = $this->isValidData($event, $colleague, CalendarDB::INSERT_QUERY);
		if($isValid !== true){
			throw new Exception($isValid['message'], $isValid['code']);
		}

		if($colleague){
			$event[self::EVENT_FIELD_OWNER] = $colleague->getID();
		}

		if($event[self::EVENT_FIELD_NAME] && mb_detect_encoding($event[self::EVENT_FIELD_NAME].' ') != 'UTF-8'){
			//Fix issue: event imported from Outlook with wrong encoding brakes websocket connection to edjabberd
			$event[self::EVENT_FIELD_NAME] = mb_convert_encoding($event[self::EVENT_FIELD_NAME], "UTF-8");
		}

		$keys = array();
		$values = array();
		foreach ($event as $key => $value){
			$keys[] = '`'.$this->getRealKey($key).'`';
			$values[] = '\''.CalendarDB::getInstance()->realEscapeString($value).'\'';
		}

		$query = sprintf('INSERT OR REPLACE INTO `%s` (%s) Values (%s)', self::TABLE_NAME, join(', ', $keys), join(', ', $values));

		$result = CalendarDB::getInstance()->exec($query);
		if($result === true){
			return $event[self::EVENT_FIELD_ID];
		}

		return $result;
	}

	/**
	 *
	 * @param $filter array
	 * @param string $search
	 * @param array $searchFields
	 * @param \Wildix\Colleagues\Colleague $colleague
	 * @return int
	 */
	public function deleteEvents($filter = null, $search=null, $searchFields=null, \Wildix\Colleagues\Colleague $colleague = null){

		$query = 'DELETE FROM `'.self::TABLE_NAME.'`
                                '.$this->createWhere($filter, $search, $searchFields, $colleague, CalendarDB::DELETE_QUERY).';';

		$result = CalendarDB::getInstance()->exec($query);
		if ($result === true) {
			$result = CalendarDB::getInstance()->getCountChangesRows();
		}
		return $result;
	}

	/**
	 *
	 * @param array $data
	 * @param $filter array
	 * @param string $search
	 * @param array $searchFields
	 * @param \Wildix\Colleagues\Colleague $colleague
	 * @return int
	 */
	public function updateEvents($event, $filter = null, $search=null, $searchFields=null, \Wildix\Colleagues\Colleague $colleague = null){

		$event = array_map('trim', $event);

        if(count($event) == 0){
            throw new Exception('Event data cannot be empty', 0);
        }

        $isValid = $this->isValidData($event, $colleague, CalendarDB::UPDATE_QUERY);
        if($isValid !== true){
            throw new Exception($isValid['message'], $isValid['code']);
        }

        $params = array();
        foreach ($event as $key=>$value){
            $params[] = "`".$this->getRealKey($key)."` = '".CalendarDB::getInstance()->realEscapeString($value)."'";
        }

        $query = 'UPDATE `'.self::TABLE_NAME.'` SET '.join(', ', $params).'
                                '.$this->createWhere($filter, $search, $searchFields, $colleague, CalendarDB::UPDATE_QUERY).';';

        $result = CalendarDB::getInstance()->exec($query);
        if ($result === true) {
            $result = CalendarDB::getInstance()->getCountChangesRows();
        }
        return $result;
	}

	/**
	 *
	 * @param array $filter
	 * @param string $search
	 * @param array $searchFields
	 * @param \Wildix\Colleagues\Colleague $colleague
	 * @return string
	 */
	private function createWhere($filter = null, $search = null, $searchFields = null, \Wildix\Colleagues\Colleague $colleague = null, $queryType = null){
		$filter = is_array($filter) ? $filter : array();
		$search = $search ? trim($search) : '';
		$searchFields = $searchFields ? $searchFields : array(self::EVENT_FIELD_NAME);
		$colleague = is_object($colleague) ? $colleague : null;

		$where = array();
		if($filter && !empty($filter)){
			foreach ($filter as $key=>$value){
				$keys = explode(",",$key);
				if(count($keys) > 1){
					$tmpWhere = array();
					foreach ($keys as $key){
						$tmpWhere[] = $this->filterCreate($key, $value);
					}
					if(!empty($tmpWhere)){
						$where[] = '('.join(' OR ', $tmpWhere).')';
					}
				}else{
					$where[] = $this->filterCreate($key, $value);
				}
			}
		}

		if ($colleague){
			$where[] = '`'.$this->getRealKey(self::EVENT_FIELD_OWNER).'` = '.$colleague->getID();
		}

		if($search && !empty($search) && count($searchFields)>0){
			$searchWhere = array();
			foreach ($searchFields as $key){
				$searchWhere[] = "`".$this->getRealKey($key)."` LIKE '%".CalendarDB::getInstance()->realEscapeString($search)."%'";
			}
			if(!empty($searchWhere)){
				$where[] = '('.join(' OR ', $searchWhere).')';
			}
		}

		if(count($where) > 0){
			return 'WHERE '.join(' AND ', $where);
		}
		return '';
	}


	private function filterCreate($key, $value){
		if(is_scalar($value)){
			return "`".$this->getRealKey($key)."` = '".CalendarDB::getInstance()->realEscapeString($value)."'";
		}else if(is_array($value)){
			$tmpWhere = array();
			if(!empty($value['from'])){
				$tmpWhere[] = "`".$this->getRealKey($key)."` >= '".CalendarDB::getInstance()->realEscapeString($value['from'])."'";
			}
			if(!empty($value['to'])){
				$tmpWhere[] = "`".$this->getRealKey($key)."` <= '".CalendarDB::getInstance()->realEscapeString($value['to'])."'";
			}
			if(isset($value['not'])){
				$tmpWhere[] = "`".$this->getRealKey($key)."` <> '".CalendarDB::getInstance()->realEscapeString($value['not'])."'";
			}
			if(isset($value['or'])){
				if(is_array($value['or'])){
					foreach($value['or'] as $k=>$v){
						$value['or'][$k] = CalendarDB::getInstance()->realEscapeString($v);
					}
					$tmpWhere[] = "`".$this->getRealKey($key)."` IN ('".implode("','", $value['or'])."')";
				}else{
					$tmpWhere[] = "`".$this->getRealKey($key)."` = '".CalendarDB::getInstance()->realEscapeString($value['or'])."'";
				}
			}

			if(!empty($tmpWhere)){
				return '('.join(' AND ', $tmpWhere).')';
			}
		}
		return '';
	}



	public function convertEventToStatus($event, $settings){

		$status = array(
				'until' => $event['dateEnd'],
				'message' => $event['name'],
				'status' => (isset($settings['status_map'][$event['status']])) ? $settings['status_map'][$event['status']]: 'available'
		);


		if($event['type'] == self::EVENT_TYPE_WORKING_HOURS){
			$status['status'] = $settings['working_hours']['status'];
			$status['message'] = $settings['working_hours']['message'];

		}else{
			if($settings['show_message'][$event['status']] != 1){
				$status['message'] = '';
			}
			if($event['private'] == 1 && $settings['hide_private'] == 1){
				$status['message'] = '';
			}
		}


		if($status['status'] === 'available'){
			// available status not have untill
			$status['until'] = '';
		}
		return $status;
	}
}
