<?php
namespace Wildix\CallHistory;

class Call
{
    const ID           = 'id';

    public const ID_DELIMITER = '_';

    const START             = 'start'; // date time
    const ANSWER            = 'answer'; // date time
    const END               = 'end'; // date time
    const FROM_NUMBER  = 'from_number';
    const TO_NUMBER    = 'to_number';
    const FROM_NAME    = 'from_name';
    const TO_NAME      = 'to_name';
    const BILLSEC      = 'billsec';
    const DISPOSITION  = 'disposition';
    const LASTAPP      = 'lastapp';
    const USERFIELD    = 'userfield';
    const FAXDATA      = 'faxdata';
    const LASTDATA     = 'lastdata';
    const COST         = 'cost';
    const ORGANIZATION = 'organization';
    const CHANNEL      = 'channel';
    const DSTCHANNEL   = 'dstchannel';
    const APP          = 'app';
    const CALLCLASS    = 'callclass';
    const DEST_TYPE    = 'dest_type';
    const DURATION     = 'duration';
    const DCONTEXT     = 'dcontext';
    const PBXID        = 'pbxid';

    const TYPE         = 'type';


    private $_data = null;
    private $_id   = null;

    public function __construct($properties) {
        if (is_array($properties)) {
            $this->_data = $properties;
        }

        if (!empty($properties[self::ID])) {
            $this->_id = $properties[self::ID];
        } else {
            $this->_id = null;
        }

        if (!empty($properties[self::FAXDATA])) {
            $this->_data[self::TYPE] = 'fax';
        } else {
            $this->_data[self::TYPE] = 'call';
        }
    }


    public function getID() {
        return $this->_id;
    }

    public function getStart() {
        return $this->getProperty(self::START);
    }

    public function getAnswer() {
        return $this->getProperty(self::ANSWER);
    }

    public function getEnd() {
        return $this->getProperty(self::END);
    }

    public function getFromNumber() {
        return $this->getProperty(self::FROM_NUMBER);
    }

    public function getToNumber() {
        return $this->getProperty(self::TO_NUMBER);
    }

    public function getFromName() {
        return $this->getProperty(self::FROM_NAME);
    }

    public function getToName() {
        return $this->getProperty(self::TO_NAME);
    }

    public function getBillsec() {
        return $this->getProperty(self::BILLSEC);
    }

    public function getDisposition() {
        return $this->getProperty(self::DISPOSITION);
    }

    public function getChannel() {
        return $this->getProperty(self::CHANNEL);
    }

    public function getDstchannel() {
        return $this->getProperty(self::DSTCHANNEL);
    }

    public function getFaxdata() {
        return $this->getProperty(self::FAXDATA);
    }

    public function getLastapp() {
        return $this->getProperty(self::LASTAPP);
    }

    public function getUserfield() {
        return $this->getProperty(self::USERFIELD);
    }

    public function getPbxid() {
        return $this->getProperty(self::PBXID);
    }

    public function getLastdata() {
        return $this->getProperty(self::LASTDATA);
    }

    public function getCost() {
        return $this->getProperty(self::COST);
    }

    public function getApp() {
        return $this->getProperty(self::APP);
    }

    public function getCallclass() {
        return $this->getProperty(self::CALLCLASS);
    }

    public function getDestType() {
        return $this->getProperty(self::DEST_TYPE);
    }

    public function getDuration() {
        return $this->getProperty(self::DURATION);
    }

    public function getDcontext() {
        return $this->getProperty(self::DCONTEXT);
    }

    public function getType() {
        return $this->getProperty(self::TYPE);
    }


    protected function getProperty($name) {
        if ($this->_data && !empty($this->_data) && isset($this->_data[$name])) {
            return $this->_data[$name];
        }
        return null;
    }

    public function getAsArray($fields = null) {
        if (empty($fields)) {
            return $this->_data;
        }
        return array_intersect_key($this->_data, array_flip($fields));
    }

    /**
     * Get recordings in call.
     *
     * @return array
     */
    public function getRecordings() {
        $recordings = [];
        $string     = $this->getUserfield();
        if (substr($string, 0, 2) === 'R:') {
            $recordStr = preg_replace('/^R:/', '', $string);
            $values    = explode(';', $recordStr);

            $arr = [];
            foreach ($values as $record) {
                $arr[] = $this->_parseCallRecordString($record);
            }
            $recordings = array_merge(...$arr);
        }

        return $recordings;
    }

    /**
     * Parse call record string.
     *
     * @param string $record Record string.
     *
     * @example R:2018.12.18/12/20/2018.12.18-12.20.16-11-00880-e5dab4edef396277c0bb-{n}.wav|1;
     * 2018.12.18/12/20/2018.12.18-12.20.18-00880-11-c3e54c7de4086eb81102-{n}.wav|1;2018.12.18/12/32/2018.12.18-12.32.32-00880-11-d01ce2ef4b70e108fc79-{n}.wav|1;
     *
     * @return array
     */
    private function _parseCallRecordString(string $record) {
        $recordings = [];

        if ('' !== $record) {
            $recordRegExp = '/^.*?\/(\d+)\.(\d+)\.(\d+)-(\d+)\.(\d+)\.(\d+)-([^-]+)-([^-]+)-.*?\|(\d+)$/i';
            $filename     = preg_replace('/\|\d+$/i', '', $record);

            if (preg_match($recordRegExp, $record, $match)) {
                for ($i = 0; $i < $match[9]; $i++) {
                    $recordings[] = [
                        'file' => preg_replace('/\{n\}/i', $i, $filename),
                        'date' => $match[3] . '/' . $match[2] . '/' . $match[1] . ' ' . $match[4] . ':' . $match[5] . ':' . $match[6],
                        'source' => $match[7],
                        'destination' => $match[8],
                    ];
                }
            }
        }

        return $recordings;
    }

    /**
     * Get fax
     *
     * @return null|array
     */
    public function getFax() {
        return $this->_parseFaxRecord($this->getFaxdata());
    }

    /**
     * Parse fax record string.
     *
     * @param string $string Fax string.
     *
     * @example OK|OTHERS/2018.12.19-10.30.53_1_127152610.pdf|00880|1
     * @example OK|F:00880/fax-00880-888888-1545211764767574296-1.pdf|888888|1|8031x7700|OK|WIL
     * @example OK|1010/2019.01.09-11.52.41_1_920115160.pdf|88|1
     * @example ERR|F:00880/fax-00880-11-1545212142966094850-1.pdf|11|The call dropped prematur
     *
     * @return null|array
     */
    private function _parseFaxRecord(string $string) {
        $faxInfo = null;

        $data = explode('|', $string);
        if (empty($string) || !isset($data[1])) {
            return $faxInfo;
        }

        $pathToFax       = $data[1];
        $pathToFaxReport = null;

        if (strpos($pathToFax, 'F:') === 0) {
            $pathToFax = substr($pathToFax, 2);
        }

        if (strpos($pathToFax, 'fax-') !== false) {
            $pathToFaxReport = str_replace('fax-', 'report-', $pathToFax);
        }

        $faxInfo = [
            'fax' => $pathToFax,
            'report' => $pathToFaxReport,
        ];

        return $faxInfo;
    }

    public static function normalizeChannel(string $channel) {
        $timestampPos = strrpos($channel, self::ID_DELIMITER);
        if (false !== $timestampPos && preg_match('/\d{10}$/', substr($channel, $timestampPos))) {
            $channel = substr($channel, 0, $timestampPos);
        }

        return $channel;
    }
}
