<?php
namespace Wildix\CallHistory;

use Wildix\Colleagues\Colleague;
use Wildix\Databases\CDRDB;

class Calls extends \Singleton
{
    const ATTACHMENTS     = 'attachments';
    const ATTACHMENTS_DIR = 'spoolview/';

    /**
     * @var \Wildix\Databases\CDRDB
     */
    private $_cdrDb;

    /**
     * @var \Models\PBX $_pbx Model PBX class.
     */
    private $_pbx;

    /**
     * @var \Wildix\CallRecordings  Call recordings class.
     */
    private $_callRecordings;

    /**
     * @var \Wildix\CallRecordings|\Wildix\Faxes\Faxes Faxes class.
     */
    private $_faxes;

    private $_tableFields = '`rowid` as id, `start`, `answer`, `end`, `src`, `dst`, `c_from` as from_number, `c_to` as to_number, `from_name`, `to_name`, `billsec`, `disposition`, `lastapp`, `userfield`, `faxdata`, `lastdata`, `cost`, `organization`, `channel`, `dstchannel`, `app`, `callclass`, `dest_type`, `duration`, `dcontext`, `pbxid` ';
    private $_tableFieldsMsSql = '[rowid] as id, CONVERT(varchar(20), [start], 20) as [start], CONVERT(varchar(20), [answer], 20) as [answer], CONVERT(varchar(20), [end], 20) as [end], [src], [dst], [c_from] as from_number, [c_to] as to_number, [from_name], [to_name], [billsec], [disposition], [lastapp], [userfield], [faxdata], [lastdata], [cost], [organization], [channel], [dstchannel], [app], [callclass], [dest_type], [duration], [dcontext], [pbxid] ';
    private $_allowedFilterTableFields = [
        'rowid',
        'start',
        'answer',
        'end',
        'src',
        'dst',
        'c_from',
        'c_to',
        'from_name',
        'to_name',
        'billsec',
        'disposition',
        'lastapp',
        'userfield',
        'faxdata',
        'lastdata',
        'cost',
        'organization',
        'channel',
        'dstchannel',
        'app',
        'callclass',
        'dest_type',
        'duration',
        'dcontext',
        'pbxid',
    ];
    private $_allowedFilterAdditionalFields = [
        'direction',
        'archived',
        'type',
        'archive',
        'exclude',
        self::ATTACHMENTS,
    ];

    private $_keyMap = array(
        'id' => 'rowid',
        'from_number' => 'c_from',
        'to_number' => 'c_to'
    );


    private $_logLevel = 0;

    private $_log = null;

    /**
     * Calls constructor.
     *
     * @param \Wildix\Databases\CDRDB|null $cdrDb CDRDB class.
     * @param \Models\PBX|null             $pbx   Pbx class.
     */
    public function __construct(
        \Models\PBX $pbx = null,
        \Wildix\CallRecordings $callRecordings = null,
        \Wildix\Faxes\Faxes $faxes = null
    ) {
        $this->_pbx   = $pbx !== null ? $pbx : \Models\PBX::getInstance();
        $this->_callRecordings = $callRecordings !== null ? $callRecordings : new \Wildix\CallRecordings($this->getLog());
        $this->_faxes = $faxes !== null ? $faxes : new \Wildix\Faxes\Faxes($this->getLog());
    }

    public function setCDRDB(CDRDB $sdrDB) {
        $this->_cdrDb = $sdrDB;
    }

    private function getCDRDB() {
        if(!$this->_cdrDb) {
            $this->_cdrDb = CDRDB::getInstance();
        }

        return $this->_cdrDb;
    }

    private function getLog(){
        if($this->_log === null){
            $this->_log = new \Syslog('Calls', $this->_logLevel);
        }
        return $this->_log;
    }
    private function getTableFields(){
        if ($this->getCDRDB()->isMsSql()) {
            return $this->_tableFieldsMsSql;
        }
        return $this->_tableFields;
    }

    private function getTable(){
        return $this->escapeKey($this->getCDRDB()->getCdrTableName());
    }

    private function escapeKey($key) {
        if ($this->getCDRDB()->isMsSql()) {
            return '[' . $key . ']';
        }

        return '`' . str_replace('`', '``', $key) . '`';
    }

    /**
     *
     * @param int $id
     * @param Colleague $colleague
     * @param boolean $allowSeeCallRecordings Allow see callrecordings in the calls.
     * @return <NULL, \Wildix\Phonebooks\Contact>
     */
    public function getByID($id, Colleague $colleague = null, bool $allowSeeCallRecordings = true) {
        $calls = $this->getCalls(
            array(Call::ID => $id),
            null,
            null,
            null,
            1,
            null,
            null,
            $colleague,
            $allowSeeCallRecordings
        );
        return (count($calls) > 0) ? $calls[0] : null;
    }

    /**
     *
     * @param array $filter
     * @param string $search
     * @param array $searchFields
     * @param int $start
     * @param int $count
     * @param string $sort
     * @param string $dir (ASC | DESC)
     * @param Colleague $colleague
     * @param boolean $allowSeeCallRecordings Allow see callrecordings in the calls.
     * @return array <\Wildix\Phonebooks\Contact>
     */
    public function getCalls(
        $filter = null,
        $search=null,
        $searchFields=null,
        $start = null,
        $count=null,
        $sort=null,
        $dir=null,
        ?Colleague $colleague = null,
        bool $allowSeeCallRecordings = true
    ) {
        $start = is_numeric($start) ? intval($start) : 0;
        $count = is_numeric($count) ? intval($count) : 100;
        $sort = $sort ? $sort : Call::ID;
        $dir = $dir ? strtoupper($dir) : 'DESC';

        if ($start < 0) {
            throw new \Wildix\Exception('Parameter "start" must be greater or equal than 0');
        }

        if ($count < 0) {
            throw new \Wildix\Exception('Parameter "count" must be greater or equal than 0');
        }

        if ($dir != 'DESC' && $dir != 'ASC') {
            throw new \Wildix\Exception('Parameter "dir" must be "ASC" or "DESC"');
        }

        $sortColumns = explode(',', $sort);
        $sortColumns = array_map(function ($val) {
            return $this->_keyMap[$val] ?? $val;
        }, $sortColumns);

        $this->_checkAllowedFields($sortColumns);

        $orderBy = '';
        foreach ($sortColumns as $sortColumn) {
            $orderBy .= ($orderBy ? ', ' : '');
            $orderBy .= $this->escapeKey($sortColumn) . ' ' . $dir;
        }

        $calls = array();

        if($this->getCDRDB()->isMsSql()){
            if($start == 0){
                $query = 'SELECT TOP '.$count.' '.$this->getTableFields().' ';
                $query .= 'FROM '.$this->getTable().' ';
                $query .= $this->createWhere($filter, $search, $searchFields, $colleague).' ';
                $query .= 'ORDER BY '.$orderBy.' ';
            }else if(false){
                // if MS SQL 2012
                $query = 'SELECT '.$this->getTableFields().' ';
                $query .= 'FROM '.$this->getTable().' ';
                $query .= $this->createWhere($filter, $search, $searchFields, $colleague).' ';
                $query .= 'ORDER BY '.$orderBy.' ';
                $query .= 'OFFSET '.$start.' ROWS ';
                $query .= 'FETCH NEXT '.$count.' ROWS ONLY ';
            }else{
                // if MS SQL2008

                $originalQuery = 'SELECT * ';
                $originalQuery.= 'FROM '.$this->getTable().' ';
                $originalQuery.= $this->createWhere($filter, $search, $searchFields, $colleague).' ';

                $numeredQuery = 'SELECT ROW_NUMBER() OVER (ORDER BY '.$orderBy.') AS [RowNum] , * ';
                $numeredQuery.= 'FROM ('.$originalQuery.') as t';

                $query = 'SELECT '.$this->getTableFields().' ';
                $query.= 'FROM ('.$numeredQuery.') as t1 ';
                $query.= 'WHERE [RowNum] BETWEEN '.($start+1).' AND '.($start+$count).' ';
            }
        }else{
            // SQLite & MySQL
            $query = sprintf(
                'SELECT %s FROM %s %s ORDER BY %s LIMIT %d, %d',
                $this->getTableFields(),
                $this->getTable(),
                $this->createWhere($filter, $search, $searchFields, $colleague),
                $orderBy,
                $start,
                $count
            );
        }

        $this->getLog()->info(__LINE__, $query);
        $rows = $this->getCDRDB()->query($query);

        if ($rows) {
            while($row = $rows->fetch()){
                if ($row && !empty($row)) {
                    if (!$allowSeeCallRecordings) {
                        $row[Call::USERFIELD] = '';
                    }
                    $calls[] = new Call($row);
                }
            }
        }
        return $calls;
    }

    /**
     *
     * @param array $filter
     * @param string $search
     * @param array $searchFields
     * @param Colleague $colleague
     * @return int
     */
    public function getCount($filter, $search=null, $searchFields=null, Colleague $colleague = null){

        $query = 'SELECT count(rowid) as countRows ';
        $query .= 'FROM '.$this->getTable().' ';
        $query .= $this->createWhere($filter, $search, $searchFields, $colleague).';';
        $this->getLog()->info(__LINE__, $query);
        $rows = $this->getCDRDB()->query($query);
        if ($rows) {
            $row = $rows->fetch();
            if($row){
                return (int)$row['countRows'];
            }
        }

        return 0;
    }

    /**
     *
     * @param array $data
     * @param array $filter
     * @param string $search
     * @param array $searchFields
     * @param Colleague $colleague
     * @return boolean
     */
    public function updateCall($data, $filter = null, $search=null, $searchFields=null, Colleague $colleague = null){

        $data = array_map('trim', $data);

        if(empty($data)){
            return 0;
        }

        $params = array();
        foreach ($data as $key=>$value){
            $key = isset($this->_keyMap[$key])? $this->_keyMap[$key] : $key;
            $params[] = "".$this->escapeKey($key)." = ".$this->getCDRDB()->realEscapeString($value)."";
        }

        $query = 'UPDATE '.$this->getTable().' SET '.join(', ', $params).' '.$this->createWhere($filter, $search, $searchFields, $colleague).';';
        $this->getLog()->info(__LINE__, $query);
        $result = $this->getCDRDB()->exec($query);
        if ($result === true) {
            $result = $this->getCDRDB()->getCountChangesRows();
        }
        return $result;
    }

    /**
     *
     * @param array $filter
     * @param string $search
     * @param array $searchFields
     * @param Colleague $colleague
     * @return boolean
     */
    public function deleteCall($filter = null, $search=null, $searchFields=null, Colleague $colleague = null){

        if($filter && $filter[Call::ID]){
            $call = $this->getByID($filter[Call::ID],$colleague);
            if(!$call){
                throw new \Wildix\Exception('Not found call', 400);
            }
        }

        $where = $this->createWhere($filter, $search, $searchFields, $colleague);

        $this->_searchAndDeleteAttachments($where);

        $query = 'DELETE FROM ' . $this->getTable() . ' ' . $where . ';';

        $result = $this->getCDRDB()->exec($query);
        if ($result === true) {
            $result = $this->getCDRDB()->getCountChangesRows();
        }
        return $result;
    }

    /**
     * Search and delete recordings and faxes in call
     *
     * @param string $where WHERE sql.
     *
     * @return void
     */
    private function _searchAndDeleteAttachments(string $where) {
        $count  = 1000;
        $page   = 1;
        $offset = 0;
        do {
            $query        = 'SELECT * FROM ' . $this->getTable(
                ) . ' ' . $where . ' order by rowid  limit ' . $count . ' offset ' . $offset . ';';
            $rows         = $this->getCDRDB()->query($query);
            $countResults = $rows->rowCount();
            if ($countResults == 0) {
                break;
            }
            while ($row = $rows->fetch()) {
                if (!empty($row)) {
                    $call = new Call($row);

                    $recordings = $call->getRecordings();
                    $this->_callRecordings->deleteCallRecordings($recordings);

                    $faxInfo = $call->getFax();
                    if ($faxInfo !== null) {
                        $this->_faxes->deleteFax($faxInfo);
                    }
                    unset($call);
                }
            }
            $page++;
            $offset = $count * $page;
        } while ($countResults <= $count);
    }

    /**
     * Check allowed fields.
     *
     * @param array   $fields         Fields names.
     * @param boolean $withAdditional With additional field list (not existed in CDR table).
     *
     * @throws \Exception If field not allowed.
     * @return void
     */
    private function _checkAllowedFields(array $fields, bool $withAdditional = false): void {
        $allowedFields = $this->_allowedFilterTableFields;
        if ($withAdditional) {
            $allowedFields = array_merge($allowedFields, $this->_allowedFilterAdditionalFields);
        }

        $diff = array_diff($fields, $allowedFields);
        if (!empty($diff)) {
            throw new \Wildix\Exception(sprintf('Not allowed fields %s', implode(', ', $diff)));
        }
    }

    /**
     *
     * @param array $filter
     * @param string $search
     * @param array $searchFields
     * @param Colleague $colleague
     * @return string
     */
    private function createWhere($filter = null, $search = null, $searchFields = null, Colleague $colleague = null){
        $filterDirty = is_array($filter) ? $filter : array();
        $search = $search ? trim($search) : '';
        $searchFields = is_array($searchFields) ? $searchFields : array(Call::FROM_NUMBER, Call::TO_NUMBER, Call::FROM_NAME, Call::TO_NAME);

        $keyMap = $this->_keyMap;
        $searchFields = array_map(function($val) use (&$keyMap) {
            return isset($keyMap[$val])? $keyMap[$val] : $val;
        }, $searchFields);
        $this->_checkAllowedFields($searchFields, false);

        $filter = [];
        foreach ($filterDirty as $key => $val) {
            if (array_key_exists($key, $keyMap)) {
                $key = $keyMap[$key];
            }
            $filter[$key] = $val;
        }
        $this->_checkAllowedFields(array_keys($filter), true);

        $colleague = is_object($colleague) ? $colleague : null;

        $where = array();

        $from = false;
        $to = false;

        if ($colleague){
            $from = true;
            $to = true;
            if($filter && !empty($filter) && !empty($filter['direction'])){
                switch (strtolower($filter['direction'])){
                    case 'outgoing': $to = false;
                        break;
                    case 'incoming': $from = false;
                        break;
                }
            }
            $tmpWhere = array();
            $archived = ($filter && !empty($filter) && array_key_exists('archived', $filter)) ? intval($filter['archived']) : null;

            $mobilePhone = $colleague->getMobilePhone();
            if($to){
                $toQuery = $this->escapeKey('c_to')." = '".$colleague->getExtension()."'";
                if ($mobilePhone) {
                    $toQuery = "(" . $toQuery . " OR " . $this->escapeKey('c_to')." = '".$mobilePhone."')";
                }

                $tmpWhere[] = "(".$toQuery."  AND ".$this->escapeKey('callclass')." <> '5' ".(($archived !== null)? " AND ".$this->escapeKey('to_archived')."='".$archived."'" : "").")";
            }
            if($from){
                $fromQuery = $this->escapeKey('c_from')." = '".$colleague->getExtension()."'";
                if ($mobilePhone) {
                    $fromQuery = "(" . $fromQuery . " OR " . $this->escapeKey('c_from')." = '".$mobilePhone."')";
                }

                $tmpWhere[] = "(".$fromQuery." ".(($archived !== null)? " AND ".$this->escapeKey('from_archived')."='".$archived."'" : "").")";
            }

            $where[] = "(".implode(" OR ", $tmpWhere).")";
            $tmpWhere = null;
        }
        $this->getLog()->info(__LINE__, print_r($filter,1));
        if($filter && !empty($filter)){
            foreach ($filter as $key=>$value){
                if(is_string($value) || is_integer($value)){
                    if($key == 'type'){
                        switch (strtolower($value)){
                            case 'call':
                                $tmpWhere = array();
                                if($from){
                                    $tmpWhere[] = "(".$this->escapeKey('userfield')." LIKE 'VM:%' AND ".$fromQuery.")";
                                }
                                $tmpWhere[] = "(".$this->escapeKey('userfield')." NOT LIKE 'VM:%')";

                                $where[] = $this->escapeKey('faxdata')." = '' AND (".implode(" OR ", $tmpWhere).")";
                                $tmpWhere = null;
                                break;
                            case 'fax': $where[] = $this->escapeKey('faxdata')." <> ''";
                                break;
                            case 'voicemail': $where[] = '('.$this->escapeKey('userfield')." LIKE 'VM:%' AND ".$toQuery.")";
                                break;
                            default: $where[] = $this->escapeKey('rowid')." = NULL";
                                break;
                        }

                    }else if($key == 'disposition'){
                        $where[] = $this->escapeKey($key)." = ".$this->getCDRDB()->realEscapeString(strtoupper($value));
                    }else if($key == 'pbxid'){
                        if ($value != '*')
                            $where[] = $this->escapeKey($key)." = ".$this->getCDRDB()->realEscapeString(strtoupper($value));
                    }else if($key != 'direction' && $key != 'archive'){ //direction, archive considered above
                        $where[] = $this->escapeKey($key)." = ".$this->getCDRDB()->realEscapeString($value);
                    }

                }else if(is_array($value)){

                    $tmpWhere = array();

                    if (in_array($key, array('start', 'answer', 'end'))){
                        if(!empty($value['from'])){
                            if ($this->getCDRDB()->isMsSql()) {
                                $tmpWhere[] = $this->escapeKey($key)." >= CONVERT(DATETIME, ".$this->getCDRDB()->realEscapeString(date('Y-m-d H:i:s', strtotime(str_replace('/','-',$value['from'])))).", 20)";
                            } else {
                                $tmpWhere[] = $this->escapeKey($key)." >= ".$this->getCDRDB()->realEscapeString(date('Y-m-d H:i:s', strtotime(str_replace('/','-',$value['from']))));
                            }
                        }
                        if(!empty($value['to'])){
                            if ($this->getCDRDB()->isMsSql()) {
                                $tmpWhere[] = $this->escapeKey($key)." <= CONVERT(DATETIME, ".$this->getCDRDB()->realEscapeString(date('Y-m-d H:i:s', strtotime(str_replace('/','-',$value['to'])))).", 20)";
                            } else {
                                $tmpWhere[] = $this->escapeKey($key)." <= ".$this->getCDRDB()->realEscapeString(date('Y-m-d H:i:s', strtotime(str_replace('/','-',$value['to']))));
                            }
                        }
                    } else {
                        if(!empty($value['from'])){
                            $tmpWhere[] = $this->escapeKey($key)." >= ".$this->getCDRDB()->realEscapeString($value['from']);
                        }
                        if(!empty($value['to'])){
                            $tmpWhere[] = $this->escapeKey($key)." <= ".$this->getCDRDB()->realEscapeString($value['to']);
                        }
                    }

                    if(isset($value['not'])){
                        if($key == 'type'){
                            switch (strtolower($value['not'])){
                                case 'call':
                                    $tmpWhere2 = array();
                                    $tmpWhere2[] = $this->escapeKey('faxdata')." <> ''";
                                    if($to){
                                        $tmpWhere2[] = "(".$this->escapeKey('userfield')." LIKE 'VM:%' AND ".$toQuery.")";
                                    }

                                    $where[] = "(".implode(" OR ", $tmpWhere2).")";
                                    break;
                                case 'fax': $where[] = $this->escapeKey('faxdata')." = ''";
                                    break;
                                case 'voicemail':
                                    $tmpWhere2 = array();
                                    if($from){
                                        $tmpWhere2[] = "(".$this->escapeKey('userfield')." LIKE 'VM:%' AND ".$fromQuery.")";
                                    }
                                    $tmpWhere2[] = $this->escapeKey('userfield')." NOT LIKE 'VM:%'";

                                    $where[] = "(".implode(" OR ", $tmpWhere2).")";
                                    break;
                                default: $where[] = $this->escapeKey('rowid')." = NULL";
                                    break;
                            }
                        }else{
                            $tmpWhere[] = $this->escapeKey($key)." <> ".$this->getCDRDB()->realEscapeString($value['not']);
                        }

                    }
                    if(isset($value['or'])){
                        if(is_array($value['or'])){
                            foreach($value['or'] as $k=>$v){
                                $value['or'][$k] = $this->getCDRDB()->realEscapeString($v);
                            }
                            $tmpWhere[] = $this->escapeKey($key)." IN (".implode(",", $value['or']).")";
                        }else{
                            $tmpWhere[] = $this->escapeKey($key)." = ".$this->getCDRDB()->realEscapeString($value['or']);
                        }
                    }
                    if($key == 'exclude'){
                        //here goes array of id ranges which will be excluded
                        foreach ($value as $idsToExclude) {
                            if (!empty($idsToExclude['from']) && !empty($idsToExclude['to']))
                                $tmpWhere[] = sprintf('(rowid NOT BETWEEN %s AND %s)',(int)$idsToExclude['from'],(int)$idsToExclude['to']);
                        }
                    }

                    if(!empty($tmpWhere)){
                        $where[] = '('.join(' AND ', $tmpWhere).')';
                    }
                }
            }
        }

        if(!$filter || empty($filter['pbxid'])){
            $where[] = "(" . $this->escapeKey('pbxid') . " = '" . $this->_pbx->getSerial() . "' OR ".$this->escapeKey('pbxid') ." = ''" .
                //do not restrict for fax & voicemail
                " OR " . $this->escapeKey('lastapp') . " = 'VoiceMail' OR " . $this->escapeKey('faxdata') . " <> '')";

        }

        if($search && !empty($search)){
            $searchWhere = array();

            $search = $this->getCDRDB()->realEscapeString('%'.$search.'%');

            $countFields = count($searchFields);
            for($i=0; $i < $countFields; $i++){
                $searchWhere[] = $this->escapeKey($searchFields[$i])." LIKE {$search}";
            }
            if(!empty($searchWhere)){
                $where[] = '('.join(' OR ', $searchWhere).')';
            }
        }

        if(count($where) > 0){
            $where = 'WHERE '.join(' AND ', $where);
        }else{
            $where = '';
        }

        return $where;
    }

    /**
     * Format attachments data.
     *
     * @param array  $row           CDR row.
     * @param string $userExtension User extension.
     *
     * @return array
     */
    public function formatAttachments(array  $row, string $userExtension = ''): array {
        $vm         = [];
        $recordings = [];
        $faxes      = [];

        $incoming = (string)($row[Call::FROM_NUMBER] ?? '') !== $userExtension;

        if (!empty($row[Call::USERFIELD])) {
            $recordings = $this->_parseRecordings($row[Call::USERFIELD]);
            $vm         = $this->_parseVoiceMail($row[Call::USERFIELD]);
        }

        if (!empty($row[Call::FAXDATA])) {
            $faxes = $this->_parseFaxes($row[Call::FAXDATA], $incoming, $userExtension);
        }

        return [
            'vm' => $vm,
            'recordings' => $recordings,
            'faxes' => $faxes,
        ];
    }

    /**
     * Parse recordings string.
     *
     * @param string $s Recordings string.
     *
     * @example R:2018.12.21/13/01/2018.12.21-13.01.09-11-00880-8f8899e7a614914761bb-{n}.wav|1;2018.12.21/13/01/2018.12.21-13.01.11-00880-11-319c0338d246230db3c2-{n}.wav|1;
     *
     * @return array
     */
    private function _parseRecordings(string $s) {
        if (!preg_match('/^R:(.*)$/', $s, $match)) {
            return [];
        }

        $data = [];
        $arr  = explode(';', $match[1]);
        foreach ($arr as $item) {
            $item = trim($item);
            if (empty($item)) {
                continue;
            }

            preg_match('/^(.*?\/(\d+)\.(\d+)\.(\d+)-(\d+)\.(\d+)\.(\d+)-([^-]+)-([^-]+)-.*?)\|(\d+)$/', $item, $match);
            if ($match) {
                $fileName = $match[1];
                if (strpos($fileName, '-{n}') !== false) {
                    $fileName = self::ATTACHMENTS_DIR . 'recordings/' . str_replace('{n}', 0, $fileName);
                } else {
                    $fileName = self::ATTACHMENTS_DIR . 'recordings/' . $fileName . '-0.wav';
                }

                $data[] = [
                    'url' => $fileName,
                    'time' => strtotime(sprintf('%s-%s-%s %s:%s:%s', $match[2], $match[3], $match[4], $match[5], $match[6], $match[7])),
                    'source' => $match[8],
                    'destination' => $match[9],
                ];
            }
        }

        return $data;
    }

    /**
     * Parse voice mail string.
     *
     * @param string $s Input voice mail string.
     *
     * @example VM:/11/INBOX/msg1564733747-0864070165.mp3
     *
     * @return array
     */
    private function _parseVoiceMail(string $s) {
        if (!preg_match('/^VM:(.*)$/', $s, $match)) {
            return [];
        }

        return ['url' => self::ATTACHMENTS_DIR . 'voicemail/' . ltrim($match[1], '/')];
    }

    /**
     * Parse fax string.
     *
     * @param string  $s             Input Fax string.
     * @param boolean $isIncoming    Is incoming fax.
     * @param string  $userExtension User extension.
     *
     * @example ERR||00380661234588|Fax transfer was not successful
     * @example OK|F:18/fax-18-888888-1545299734594801565-1.pdf|888888|1|8031x7700|OK|WILDIXFAX|141
     * @example OK|OTHERS/2018.12.20-10.56.59_1_914106691.pdf|18|1
     * @example OK|460/2019.11.05-11.53.29_1_632583233.pdf|460|1
     * @example OK|F:460/fax-460-12345-1572947528262759379-1.pdf|12345|1|8031x7700|OK|WILDIXFAX|136
     * @example OK|OTHERS/2018.12.20-16.07.05_2_1203826071.pdf|460000|2
     * @example OK|F:25000/fax-25000-777-1560435512580788236-1.pdf|777|1|8031x7700|OK|WILDIXFAX|80
     * @example ERR|F:25000/fax-25000-25001-1560435600142893824-1.pdf|25001|CHANUNAV
     * @example ERR|F:345/report-345-3453453454-1407328941444897990-14.pdf|3453453454|CHANUNAVA
     *
     * @return array
     */
    private function _parseFaxes(string $s, bool $isIncoming, string $userExtension): array {
        $inputData = explode('|', $s);
        if (empty($s) || empty($inputData[1])) {
            return [];
        }

        $faxStatus = trim($inputData[0]);
        $faxUrl    = trim($inputData[1]);

        $folder     = $isIncoming ? 'faxin' : 'faxout';
        $pathToFile = '';
        if (strpos($faxUrl, 'F:') === 0) {
            $pathToFile = substr($faxUrl, 2);
        } else if ($faxStatus === 'OK' || $faxStatus === 'ERR') {
            preg_match('/\|fax-([0-9]+)-/', $faxUrl, $matchFax);
            if ($folder === 'faxout' && !empty($matchFax[1])) {
                $pathToFile = $matchFax[1] . '/' . $faxUrl;
            } else {
                preg_match('/(tmp\/)/', $faxUrl, $matchTmp);
                if ($folder === 'faxin' && !empty($matchTmp[1])) {
                    $faxUrl = str_replace('/tmp/', $userExtension, $faxUrl);
                }
                $pathToFile = $faxUrl;
            }
        }

        if (empty($pathToFile)) {
            return [];
        }

        return ['url' => self::ATTACHMENTS_DIR . $folder . '/' . $pathToFile];
    }
}
