<?php
namespace Wildix;

/**
 * Class CallweaverManager
 */
class CallweaverManager extends \AGI_AsteriskManager {
    const CALLWEAVER_MANAGER_CONFIG     = '/etc/callweaver/manager.conf';
    const CALLWEAVER_NETWORK_SIP_CONFIG = '/etc/callweaver/sip-wms-network.conf';

    const CONTENT_NETWORK_SIP_CONFIG = '[WMS_NETWORK_PEER]
username=%s
secret=%s
host=127.0.0.1
insecure=invite,port
allow=opus:20,g729:60,alaw:20,ulaw:20,g722:20,h264,h263
type=friend
dtmfmode=info
session-expires=7200';

    private static $_instance = null;

    public static function getInstance() {
        if (self::$_instance === null) {
            self::$_instance = new self();
        }

        return self::$_instance;
    }

    public function __construct() {
        parent::__construct(null, $this->getConfig());
    }

    private $_config = null;

    private function getConfig() {
        if ($this->_config === null) {
            if (file_exists(self::CALLWEAVER_MANAGER_CONFIG)) {
                $iniContent = file_get_contents(self::CALLWEAVER_MANAGER_CONFIG);
                $iniContent = preg_replace('/^#/', ';', $iniContent);
                $config     = parse_ini_string($iniContent, true, INI_SCANNER_RAW);
                $username   = '';

                foreach ($config as $key => $value) {
                    if (!empty($value) && !empty($value['secret'])) {
                        $username = $key;
                        break;
                    }
                }

                $this->_config = array(
                    'server'   => $config['general']['bindaddr'],
                    'port'     => $config['general']['port'],
                    'username' => $username,
                    'secret'   => $config[$username]['secret']
                );
            } else {
                $this->_config = array(
                        'server' => '127.0.0.1',
                        'port' => 5039,
                        'username' => 'root',
                        'secret' => 'wildix'
                );
            }
        }
        return $this->_config;
    }

    private $_connected = null;

    public function isConnected() {
        if ($this->_connected === null) {
            $this->_connected = true;

            if (!parent::connect()) {
                $this->_connected = null;
                throw new Exception('Failed connect to the PBX engine');
            }
        }

        return $this->_connected;
    }


    private function parseResult($data) {
        $result = array();

        $data = explode("\n", trim($data));
        foreach ($data as $value) {
            $clearValue = trim($value);

            $pos = strpos($clearValue, ':');

            if ($pos !== false) {
                $key = trim(substr($clearValue, 0, $pos));

                $result[$key] = '';

                if (isset($clearValue[$pos + 1])) {
                    $result[$key] = trim(substr($clearValue, $pos + 1));
                }
            } else {
                $result['Body'] .= $value;
            }
        }

        return $result;
    }


    /**
     * Deletes an entry in the Asterisk database for a given family and key.
     *
     * @param string $family Family.
     * @param string $key    Key.
     *
     * @return boolean
     *
     * @link http://www.voip-info.org/wiki-database+del
     */
    public function databaseDel($family, $key) {
        if ($this->isConnected()) {
            $data = $this->Command('DATABASE DEL "' . $family . '" "' . $key . '"');
            if (!empty($data) && isset($data['data'])) {
                $result = $this->parseResult($data['data']);
                if (isset($result['Body']) && !empty($result['Body']) && strpos($result['Body'], 'Database entry removed') !== false) {
                    return true;
                }
            }
        }

        return false;
    }

    /**
     * Deletes a family or specific keytree within a family in the Asterisk database.
     *
     * @param string $family  Family.
     * @param string $keyTree Key tree.
     *
     * @return boolean
     *
     * @link http://www.voip-info.org/wiki-database+deltree
     */
    public function databaseDelTree($family, $keyTree = '') {
        if ($this->isConnected()) {
            $cmd = 'DATABASE DELTREE "' . $family . '"';
            if ($keyTree != '') {
                $cmd .= ' "' . $keyTree . '"';
            }

            $data = $this->Command($cmd);
            if (!empty($data) && isset($data['data'])) {
                $result = $this->parseResult($data['data']);
                if (isset($result['Body']) && !empty($result['Body']) && strpos($result['Body'], 'database entries removed') !== false) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Retrieves an entry in the Asterisk database for a given family and key.
     *
     * @param string $family Family.
     * @param string $key    Key.
     *
     * @return string, false
     *
     * @link http://www.voip-info.org/wiki-database+get
     */
    public function databaseGet($family, $key) {
        if ($this->isConnected()) {
            $data = $this->Command('DATABASE GET "' . $family . '" "' . $key . '"');
            if (!empty($data) && isset($data['data'])) {
                $result = $this->parseResult($data['data']);
                if (isset($result['Value'])) {
                    return $result['Value'];
                }
            }
        }
        return false;
    }

    public function databaseGetFamilyTree($family) {
        if ($this->isConnected()) {
            $data = $this->Command("DATABASE QUERY \"SELECT * FROM astdb WHERE key LIKE '/$family/%'\"");
            if (!empty($data) && isset($data['data'])) {
                $data   = explode("\n\n", trim($data['data']));
                $result = array();

                foreach ($data as $key => $value) {
                    $tmp = $this->parseResult($value);
                    if ($tmp && isset($tmp['key']) && isset($tmp['value'])) {
                        $result[] = array(
                            'key' => $tmp['key'],
                            'value' => $tmp['value']
                        );
                    }
                }
                return $result;
            }
        }
        return false;
    }

    /**
     * Adds or updates an entry in the Asterisk database for a given family, key, and value.
     *
     * @param string $family Family.
     * @param string $key    Key.
     * @param string $value  Value.
     *
     * @return boolean
     */
    public function databasePut($family, $key, $value) {
        if ($this->isConnected()) {
            $value = str_replace("\n", '\n', addslashes($value));
            $data  = $this->Command('DATABASE PUT "' . $family . '" "' . $key . '" "' . $value . '"');

            if (!empty($data) && isset($data['data'])) {
                $result = $this->parseResult($data['data']);
                if (isset($result['Body']) && !empty($result['Body']) && strpos($result['Body'], 'success') !== false) {
                    return true;
                }
            }
        }

        return false;
    }

    public function getCallStatus($channel) {
        if ($this->isConnected()) {
            $this->channels = array();
            $result         = array();

            $params['ActionID'] = 122;
            $params['Channel']  = $channel;

            $this->add_event_handler('Status', array($this, 'StatusHandler'));
            $this->add_event_handler('StatusComplete', array($this, 'StatusCompleteHandler'));

            $response = $this->send_request('Status', $params);

            if ($response['Response'] == 'Success') {
                if ($response['Message'] == 'Channel status will follow') {
                    $this->wait_response();
                }

                if (isset($this->channels[0])) {
                    $result = $this->channels[0];
                    unset($result['Event']);
                    unset($result['ActionID']);
                }

                return $result;
            } else if ($response['Response'] == 'Error' && !empty($response['Message'])) {
                throw new Exception($response['Message']);
            }

            throw new Exception('Unknown response from PBX engine');
        }
    }

    private function getTagsByChannel($channel) {
        $result = $this->GetVar($channel, 'CDR(tags)');

        if ($result && $result['Response'] == 'Success') {
            if (!empty($result['Value'])) {
                $variable = $result['Value'];
                if (strpos($variable, 'tags:') >= 0) {
                    return trim(preg_replace('/^tags:/i', '', $variable));
                }
            }
            return '';
        } else if ($result['Response'] == 'Error') {
            if (!empty($result['Message'])) {
                throw new Exception($result['Message']);
            }
        }

        throw new Exception('Unknown response from PBX engine');
    }

    private function setTagsByChannel($channel, $tags) {
        $result = $this->SetVar($channel, 'CDR(tags)', 'tags:' . $tags);

        if ($result && $result['Response'] == 'Success') {
            return true;
        } else if ($result['Response'] == 'Error') {
            if (!empty($result['Message'])) {
                throw new Exception($result['Message']);
            }
        }

        throw new Exception('Unknown response from PBX engine');
    }

    public function getTags($channels) {
        if ($this->isConnected()) {
            $result = array();

            foreach (explode(',', $channels) as $channel) {
                $tags = $this->getTagsByChannel($channel);
                if (!empty($tags)) {
                    $result[$channel] = explode(',', $tags);
                }
            }

            return $result;
        }
    }

    public function setTags($channels, $tags) {
        if ($this->isConnected()) {
            foreach (explode(',', $channels) as $channel) {
                $this->setTagsByChannel($channel, $tags);
            }
            return true;
        }
    }

    public function getRecordings($channel) {
        if ($this->isConnected()) {
            $result = $this->GetVar($channel, 'CDR(userfield)');

            if ($result && $result['Response'] == 'Success') {
                $records = array();
                if (!empty($result['Value'])) {
                    if (!empty($result['Value']) && substr($result['Value'], 0, 2) === 'R:') {
                        $recordStr = preg_replace('/^R:/', '', $result['Value']);
                        $values    = explode(';', $recordStr);

                        foreach ($values as $record) {
                            if (!empty($record)) {
                                $recordRegExp = '/^.*?\/(\d+)\.(\d+)\.(\d+)-(\d+)\.(\d+)\.(\d+)-([^-]+)-([^-]+)-.*?\|(\d+)$/i';
                                $filename     = preg_replace('/\|\d+$/i', '', $record);

                                if (preg_match($recordRegExp, $record, $m)) {
                                    for ($i = 0; $i < $m[9]; $i++) {
                                        $records[] = array(
                                            'file' => preg_replace('/\{n\}/i', $i, $filename),
                                            'date' => $m[3] . '/' . $m[2] . '/' . $m[1] . ' ' . $m[4] . ':' . $m[5] . ':' . $m[6],
                                            'source' => $m[7],
                                            'destination' => $m[8]
                                        );
                                    }
                                }
                            }
                        }
                    }
                }
                return $records;
            } else if ($result['Response'] == 'Error') {
                if (!empty($result['Message'])) {
                    throw new Exception($result['Message']);
                }
            }
            throw new Exception('Unknown response from PBX engine');
        }
    }

    public function sendRefer($from, $to, $uaid) {
        if ($this->isConnected()) {
            $parameters = array(
                'Dest' => $from,
                'To' => $to,
                'UaID' => $uaid
            );

            $result = $this->send_request('SendRefer', $parameters);
            $this->log('Send Refer. Result - ' . print_r($result, true));

            if ($result && $result['Response'] == 'Success') {
                return true;
            } else if ($result['Response'] == 'Error') {
                if (!empty($result['Message'])) {
                    throw new Exception($result['Message']);
                }
            }
            throw new Exception('Unknown response from PBX engine');
        }
        return false;
    }

    public function originate($channel,
                       $exten=NULL, $context=NULL, $priority=NULL,
                       $application=NULL, $data=NULL,
                       $timeout=NULL, $callerid=NULL, $variable=NULL, $account=NULL, $async=NULL, $actionid=NULL) {
        if ($this->isConnected()) {
            $result = call_user_func_array(array('parent', 'Originate'), func_get_args());

            if ($result && $result['Response'] == 'Success') {
                return true;
            } else if ($result['Response'] == 'Error') {
                if (!empty($result['Message'])) {
                    throw new Exception($result['Message']);
                }
            }

            throw new Exception('Unknown response from PBX engine');
        }
    }

    public function hangup($channel) {
        if ($this->isConnected()) {
            $result = call_user_func_array(array('parent', 'Hangup'), func_get_args());

            if ($result && $result['Response'] == 'Success') {
                return true;
            } else if ($result['Response'] == 'Error') {
                if (!empty($result['Message'])) {
                    throw new Exception($result['Message']);
                }
            }

            throw new Exception('Unknown response from PBX engine');
        }
    }

    /**
     * Try answer the call.
     *
     * @param string $channel Call channel. E.g.: SIP/111-00000001.
     * @param string $uaid    Device User-Agent. E.g.: sip:111@10.100.1.100:47950;transport=tls.
     *
     * @return boolean
     *
     * @throws Exception If error occurred.
     */
    public function answer(string $channel, string $uaid): bool {
        if ($channel === '' || $uaid === '' || ! $this->isConnected()) {
            return false;
        }

        $parameters = [
            'Channel' => $channel,
            'UaID'    => $uaid,
            'Event'   => 'talk',
        ];

        $result = $this->send_request('SIPNotifyBroadsoft', $parameters);
        if ($result && $result['Response'] == 'Success') {
            return true;
        } else if ($result['Response'] == 'Error') {
            if (!empty($result['Message'])) {
                throw new Exception($result['Message']);
            }
        }

        throw new Exception('Unknown response from PBX engine');
    }

    /**
     * Send DTMF to channel
     *
     * @param string  $channel Call channel. E.g.: SIP/11-00000001.
     * @param string  $digit   DTMF code. E.g. 0-9#* .
     * @param boolean $bridged Send to current or bridged channel.
     *
     * @return boolean
     * @throws Exception If error occurred.
     */
    public function playDtmf(string $channel, string $digit, bool $bridged = false): bool {
        if (!$channel || $digit === '' || !$this->isConnected()) {
            return false;
        }

        $parameters = [
            'Channel' => $channel,
            'Digit'    => $digit,
            'Direction' => $bridged ? 'Other' : 'Current',
        ];

        $result = $this->send_request('PlayDTMF', $parameters);
        if ($result && $result['Response'] === 'Success') {
            return true;
        } else if ($result['Response'] === 'Error') {
            if (!empty($result['Message'])) {
                throw new Exception($result['Message']);
            }
        }

        throw new Exception('Unknown response from PBX engine');
    }

    public function redirect($channel, $extrachannel, $exten, $context, $priority) {
        if ($this->isConnected()) {
            $result = call_user_func_array(array('parent', 'Redirect'), func_get_args());
            if ($result && $result['Response'] == 'Success') {
                return true;
            } else if ($result['Response'] == 'Error') {
                if (!empty($result['Message'])) {
                    throw new Exception($result['Message']);
                }
            }
            throw new Exception('Unknown response from PBX engine');
        }
    }

    /**
     * Bridge.
     * Action: Bridge
     * Channel1: SIP/333-00000136
     * Channel2: SIP/0222-00000138
     *
     * @param string $channel1 Channel 1.
     * @param string $channel2 Channel 2.
     *
     * @return mixed
     *
     * @throws Exception Error.
     *
     * @link https://wiki.asterisk.org/wiki/display/AST/ManagerAction_Bridge
     */
    public function bridge($channel1, $channel2) {
        if ($this->isConnected()) {
            $result = $this->send_request('Bridge', array('Channel1' => $channel1, 'Channel2' => $channel2));
            if ($result && $result['Response'] == 'Success') {
                return true;
            } else if ($result['Response'] == 'Error') {
                if (!empty($result['Message'])) {
                    throw new Exception($result['Message']);
                }
            }
            throw new Exception('Unknown response from PBX engine');
        }
    }

    public function sipUpdate($channel, $data) {
        if ($this->isConnected()) {
            $data   = array_merge(array('Channel' => $channel), $data);
            $result = $this->send_request('SIPUpdatePAI', $data);

            if ($result && $result['Response'] == 'Success') {
                return true;
            } else if ($result['Response'] == 'Error') {
                if (!empty($result['Message'])) {
                    throw new Exception($result['Message']);
                }
            }
            throw new Exception('Unknown response from PBX engine');
        }
    }

    /**
     * Clear sip peer cache.
     *
     * @param string $extension Sip extension.
     *
     * @return bool Success or not.
     * @throws Exception
     */
    public function sipPeerPrune(string $extension) {
        if ('' === $extension || !$this->isConnected()) {
            return false;
        }

        $peerPrune = $this->Command(sprintf('sip prune realtime peer %s', $extension));

        return isset($peerPrune['data']) && false !== strpos($peerPrune['data'], sprintf("Peer '%s' pruned.", $extension));
    }

    public function getSipRegistrations() {
        if ($this->isConnected()) {
            $result = array();

            $data = $this->Command('sip show registry');
            if (!empty($data) && isset($data['data'])) {
                $dirtyRegistrations = trim($data['data']);
                $dirtyRegistrations = explode("\n", $dirtyRegistrations);
                $headersRange       = array();

                // parse sip registrations
                $headers = array(
                            'Host',
                            'dnsmgr',
                            'Username',
                            'Peername',
                            'Refresh State',
                            'Reg.Time',
                );

                while ($row = array_shift($dirtyRegistrations)) {
                    if (strpos($row, $headers[0]) === 0) {
                        foreach ($headers as $key => $header) {
                            $start = strpos($row, $header);

                            $headersRange[$header] = array(
                                'start' => $start
                            );

                            if (!empty($headers[$key + 1])) {
                                $nextStart = strpos($row, $headers[$key + 1]);
                                $headersRange[$headers[$key + 1]] = array(
                                    'start' => $nextStart
                                );

                                $headersRange[$header]['end'] = $nextStart;
                            } else {
                                $headersRange[$header]['end'] = strlen($row);
                            }
                        }
                        break;
                    }
                }

                foreach ($dirtyRegistrations as $row) {
                    $rowData = array();
                    foreach ($headersRange as $param => $value) {
                        $rowData[$param] = trim(substr($row, $value['start'], $value['end'] - $value['start']));
                        $rowData[$param] = urldecode($rowData[$param]);
                    }

                    if (!empty($rowData['Username'])) {
                        if ($rowData['Peername'] == 'unknown') {
                            $rowData['Peername'] = $rowData['Username'];
                        }

                        if (preg_match('/^(\d+)\s(\w+)/', $rowData['Refresh State'], $matches)) {
                            if (!empty($matches[1])) {
                                $rowData['Code']  = trim($matches[1]);
                                $rowData['State'] = trim($matches[2]);
                            }
                        }
                        $result[] = $rowData;
                    }
                }
            }
            return $result;
        }
        return false;
    }

    public function setSipWmsNetworkConf(array $data) {
        if (file_put_contents(self::CALLWEAVER_NETWORK_SIP_CONFIG, sprintf(self::CONTENT_NETWORK_SIP_CONFIG, $data['username'], $data['secret']))) {
            $this->reloadModule('sip');
        } else {
            throw new Exception(\Wildix\Error::IsNotWriteable);
        }
    }

    public function reloadModule($what) {
        if (!$this->isConnected()) {
            return false;
        }

        switch ($what) {
            case 'wmspbx':
                $command = 'reload';
                break;

            case 'sip':
                $command = 'sip reload';
                break;

            case 'moh':
                $command = 'moh reload';
                break;

            case 'pbx_lua':
                $command = 'reload pbx_lua.so';
                break;

            case 'pbx_config':
                // same as dialplan reload
                $command = 'reload pbx_config.so';
                break;

            case 'ael':
            case 'dnsmgr':
            case 'iax2':
            case 'logger':
            case 'rtp':
            case 'features':
                $command = 'reload ' . $what;
                break;

            default:
                throw new Exception(\Wildix\Error::ModuleNotFound);
        }

        $this->Command($command);

        return true;
    }

    public function QueueStatus($params = array(), $actionid=1)
    {
        $this->queueParams = array();
        $this->queueMembers = array();
        $params['ActionID'] = $actionid;

        $this->add_event_handler('QueueMember', array($this, 'QueueMemberHandler'));
        $this->add_event_handler('QueueParams', array($this, 'QueueParamsHandler'));
        $this->add_event_handler('QueueStatusComplete', array($this, 'QueueStatusCompleteHandler'));

        $response = $this->send_request('QueueStatus', $params);


        if ($response['Message'] == 'Queue status will follow') {
           $this->wait_response();
        }

        $result['params'] = $this->queueParams;
        $result['members'] = $this->queueMembers;

        return $result;
    }

    private $queueMembers = array();
    private $queueParams = array();

    public function QueueMemberHandler ($e, $parameters, $b, $c)
    {
        $this->queueMembers[] = $parameters;
        return false;
    }

    public function QueueParamsHandler ($e, $parameters, $b, $c)
    {
        $this->queueParams[] = $parameters;
        return false;
    }

    public function QueueStatusCompleteHandler ($e, $parameters, $b, $c)
    {
        return true;
    }

    private $callGroups = null;

    public function getAllCallGroups($resetCache = false) {
        if ($resetCache == true) {
            $this->callGroups = null;
        }

        if ($this->callGroups == null) {
            $response = $this->QueueStatus(array(), 124);

            foreach ($response['params'] as $groupParams) {
                $this->callGroups[$groupParams['Queue']]['params'] = $groupParams;
            }

            foreach ($response['members'] as $groupMembers) {
                if (isset($this->callGroups[$groupMembers['Queue']])) {
                    $this->callGroups[$groupMembers['Queue']]['members'][] = $groupMembers;
                }
            }
        }
    }

    public function getCallGroupsByExtension($extension, $resetCache = false) {
        $result = array();

        if ($resetCache == true || $this->callGroups == null) {
            $this->getAllCallGroups(true);
        }

        if($this->callGroups) {
            foreach ($this->callGroups as $groupId => $value) {
                foreach ($value['members'] as $member) {
                    if (strpos($member['Name'], $extension . '@') !== false) {
                        $result[$member['Membership']][$groupId] = $this->callGroups[$groupId];
                    }
                }
            }
        }

        return $result;
    }

    private $channels = array();

    public function ChannelsStatus($actionId = 1) {

        $this->channels = array();

        $params['ActionID'] = $actionId;

        $this->add_event_handler('Status', array($this, 'StatusHandler'));
        $this->add_event_handler('StatusComplete', array($this, 'StatusCompleteHandler'));

        $response = $this->send_request('Status', $params);

        if ($response['Response'] == 'Success') {
            if ($response['Message'] == 'Channel status will follow') {
                $this->wait_response();
            }
            return $this->channels;
        } else if ($response['Response'] == 'Error' && !empty($response['Message'])) {
            throw new Exception($response['Message']);
        }

        throw new Exception('Unknown response from PBX engine');
    }

    public function StatusHandler($e, $parameters, $b, $c) {
        $this->channels[] = $parameters;
        return false;
    }

    public function StatusCompleteHandler($e, $parameters, $b, $c) {
        return true;
    }
}
