<?php

namespace Wildix\Certificate;

use Logger\Log;
use Models\PBX;
use Wildix\Error;
use Wildix\Exception;
use Wildix\License;
use Wildix\Notifications\Email\Interfaces\EmailManagerInterface;
use Wildix\Services\SSL;

/**
 * Class CertificateRenewal
 */
class SSLCertificate {

    /**
     * Let’s Encrypt recommend renewing certificates automatically when they have a third of their total lifetime left.
     * For Let’s Encrypt’s current 90-day certificates, that means renewing 30 days before expiration.
     * https://letsencrypt.org/docs/integration-guide/
     * 30 days (in seconds).
     *
     * @var integer
     */
    const TIME_TO_EXPIRATION = 2592000;

    const DEFAULT_REQUEST_INTERVAL = 300;
    const DEFAULT_ATTEMPTS_COUNT   = 10;

    private const CERTIFICATE_DIR           = '/etc/wildix/ssl';
    private const CERTIFICATE_NAME          = 'wildixin.com.crt';
    private const PRIVATE_KEY_NAME          = 'wildixin.com.key';
    private const CUSTOM_CERTIFICATE_NAME   = 'tproxy.crt';
    private const EJABBERD_CERTIFICATE_NAME = 'ejabberd.pem';

    private const TEMP_PREFIX     = 'temp-';
    private const BACKUP_PREFIX   = 'backup-';
    private const ORIGINAL_PREFIX = 'original-';

    private const STORAGE = '/tmp/ssl_certificate_renewal.data';

    private const ERROR_MESSAGES_FILEPATH = '/tmp/ssl_certificate_error_messages.out';

    /**
     * SSL service.
     *
     * @var \Wildix\Services\SSL
     */
    private $_sslService;
    /**
     * PBX.
     *
     * @var \Models\PBX
     */
    private $_pbx;
    /**
     * License.
     *
     * @var \Wildix\License
     */
    private $_license;
    /**
     * Email manager.
     *
     * @var EmailManagerInterface
     */
    private $_mailer;
    /**
     * WMS Ldap.
     *
     * @var \WMS_Ldap
     */
    private $_wmsLdap;
    /**
     * Logger.
     *
     * @var \Logger\Log
     */
    private $_log;

    private $_domains;
    private $_errorMessages = [];
    private $_zone          = 'wildixin.com';

    /**
     * SSLCertificate constructor.
     *
     * @param SSL                   $service SSL Service.
     * @param PBX                   $pbx     PBX.
     * @param License               $license License.
     * @param EmailManagerInterface $mailer  Mailer.
     * @param \WMS_Ldap             $wmsLdap WMS Ldap.
     * @param Log                   $log     Logger.
     */
    public function __construct(SSL $service, PBX $pbx, License $license, EmailManagerInterface $mailer, \WMS_Ldap $wmsLdap, Log $log) {
        $this->_sslService = $service;
        $this->_pbx        = $pbx;
        $this->_license    = $license;
        $this->_mailer     = $mailer;
        $this->_wmsLdap    = $wmsLdap;
        $this->_log        = $log;

        $this->_domains = $this->getDomainsList();
    }

    /**
     * Run renewal certificate script in background
     *
     * @param boolean $forced          Forced update.
     * @param integer $attemptsCount   Maximum attempts count for getting certificate.
     * @param integer $requestInterval Interval between attempts.
     * @param boolean $saveErrors      Write error messages to file.
     *
     * @return void
     */
    public static function runBackgroundRenew(
        bool $forced = false,
        int $attemptsCount = self::DEFAULT_ATTEMPTS_COUNT,
        int $requestInterval = self::DEFAULT_REQUEST_INTERVAL,
        bool $saveErrors = false
    ): void {
        $command = \sprintf(
            'nohup php /usr/sbin/ssl_certificate_renewal.php %s %s %s %s >/dev/null 2>&1 &',
            $forced ? '-f' : '',
            '-c ' . $attemptsCount,
            '-i ' . $requestInterval,
            $saveErrors ? '-s' : ''
        );

        exec('whoami', $output);

        if (isset($output[0]) && $output[0] === 'root') {
            $command = 'sudo -u wms ' . $command;
        }

        exec($command);
    }

    public static function isFailedToRenew(): bool {
        return is_file(self::STORAGE);
    }

    public static function getRemainingDaysCount(): int {
        if (is_file(self::STORAGE)) {
            return (int)file_get_contents(self::STORAGE);
        }
        return 0;
    }

    public function getSavedErrors(): string {
        if (is_file(self::ERROR_MESSAGES_FILEPATH)) {
            return (string)file_get_contents(self::ERROR_MESSAGES_FILEPATH);
        }
        return '';
    }

    public function isBackgroundRenewRunning(): bool {
        exec('ps axw | grep "[s]sl_certificate_renewal"', $output);
        return !empty($output);
    }

    public function getFormattedErrors(): string {
        $formattedText = '';

        $errors = \array_unique($this->_errorMessages);

        if (\count($errors)) {
            $formattedText .= 'The following errors occurred while trying to get a certificate:<br>';
            $formattedText .= \implode('<br>', $errors);
        }

        return $formattedText;
    }

    /**
     * Renew certificate.
     *
     * @param boolean $forced          Forced update.
     * @param integer $attemptsCount   Maximum attempts count for getting certificate.
     * @param integer $requestInterval Interval between attempts.
     * @param boolean $saveErrors      Write error messages to file.
     *
     * @return void
     * @throws \Wildix\Exception When can not renew certificate.
     */
    public function renew(
        bool $forced = false,
        int $attemptsCount = self::DEFAULT_ATTEMPTS_COUNT,
        int $requestInterval = self::DEFAULT_REQUEST_INTERVAL,
        bool $saveErrors = false
    ): void {
        if (PHP_SAPI !== 'cli') {
            $this->getLog()->warning(__LINE__, 'Script is running in not CLI mode. Current mode: ' . PHP_SAPI);
            return;
        }

        if (!$this->_license->isActivated()) {
            $this->getLog()->warning(__LINE__, 'License is not activated so the certificate will not be renewed');
            return;
        }

        $certInfo = null;

        try {
            $this->getLog()->info(__LINE__, 'Renew certificate domains: ' . json_encode($this->_domains));

            $certInfo = $this->getCertificateInfo($this->getCertificatePath());

            if (!$forced && $this->isCurrentCertificateValid($certInfo)) {
                $this->getLog()->info(__LINE__, 'Certificate does not need renewal');
                $this->removeFile(self::STORAGE);
                $this->removeFile(self::ERROR_MESSAGES_FILEPATH);
                return;
            }

            if (!is_writable(self::CERTIFICATE_DIR)) {
                $errorMessage = \sprintf('Certificate directory "%s" is not writable', self::CERTIFICATE_DIR);
                $this->getLog()->error(__LINE__, $errorMessage);
                throw new \Exception($errorMessage, 500);
            }

            $this->backupOriginalCertificate();

            if ($this->renewCertificate($forced, $attemptsCount, $requestInterval)) {
                return;
            }
        } catch (\Exception $e) {
            $this->_errorMessages[] = $e->getMessage();
            $this->getLog()->error(__LINE__, \sprintf('Failed to renew certificate. Error: "%s"', $e->getMessage()));

            if ($saveErrors) {
                $this->saveErrorMessages($this->getFormattedErrors());
                return;
            }

            $customCertificatePath = self::CERTIFICATE_DIR . DIRECTORY_SEPARATOR . self::CUSTOM_CERTIFICATE_NAME;

            $customCertInfo = $this->getCertificateInfo($customCertificatePath);

            // if the custom certificate is loaded and not expired, we do not show notifications
            if (\is_array($customCertInfo) && !$this->isExpiredCertificate($customCertInfo, 0)) {
                return;
            }

            $remainingDays = $this->calculateRemainingDays($certInfo);

            $this->saveRemainingDaysInfo($remainingDays);
            $this->notifyAboutUnsuccessfulRenew($this->getAdminUser(), $remainingDays);
        }
    }

    /**
     * Renew certificate.
     *
     * @param boolean $forced          Forced update.
     * @param integer $attemptsCount   Maximum attempts count for getting certificate.
     * @param integer $requestInterval Interval between attempts.
     *
     * @return boolean
     * @throws \Exception When can not renew certificate.
     */
    private function renewCertificate(bool $forced, int $attemptsCount, int $requestInterval): bool {
        if ($forced) {
            $this->getLog()->info(__LINE__, 'Certificate force renewal');
        }

        $currentAttempt  = 1;
        $certificateData = null;

        do {
            try {
                $this->getLog()->info(__LINE__, \sprintf('Trying renew certificate. Attempt number: %d', $currentAttempt));

                $certificateData = $this->_sslService->getCertificate($this->_domains, $forced);

                break;
            } catch (\Exception $e) {
                $this->_errorMessages[] = $e->getMessage();
                $this->getLog()->error(
                    __LINE__,
                    \sprintf('Failed to request certificate from service. Attempt number: %d, exception: %s', $currentAttempt, $e->getMessage())
                );

                \sleep($requestInterval);
            }

            $currentAttempt++;
        } while ($currentAttempt <= $attemptsCount);

        if ($certificateData === null) {
            throw new \Exception('Failed to request certificate from service', 500);
        }

        if (is_file($this->getCertificatePath())) {
            $this->saveNewCertificate($certificateData['fullchainPem'], $certificateData['privateKey'], self::TEMP_PREFIX);
            $this->backupOldCertificate();
            $this->replaceOldWithNewCertificate();
        } else {
            $this->saveNewCertificate($certificateData['fullchainPem'], $certificateData['privateKey']);
        }

        $this->reloadServices();
        $this->removeFile(self::STORAGE);
        $this->removeFile(self::ERROR_MESSAGES_FILEPATH);

        $this->getLog()->info(__LINE__, 'Certificate successfully renewed');

        return true;
    }

    /**
     * Returns parsed certificate info.
     *
     * @param string $path Path to certificate file.
     *
     * @return array|null
     * @throws \Exception When can not parse certificate.
     */
    private function getCertificateInfo(string $path): ?array {
        if (!is_file($path)) {
            $this->getLog()->info(__LINE__, 'Certificate file not found');
            return null;
        }

        if (!is_readable($path)) {
            $errorMessage = \sprintf('Certificate file "%s" is not readable', $path);
            $this->getLog()->error(__LINE__, $errorMessage);
            throw new \Exception($errorMessage, 500);
        }

        $certInfo = openssl_x509_parse(file_get_contents($path));

        if (!\is_array($certInfo)) {
            $errorMessage = \sprintf('Fail to parse certificate with error: %s', openssl_error_string());
            $this->getLog()->error(__LINE__, $errorMessage);
            throw new \Exception($errorMessage, 500);
        }

        return $certInfo;
    }

    /**
     * Validate certificate.
     *
     * @param array|null $certInfo Parsed certificate info.
     *
     * @return boolean
     * @throws \Exception When certificate is not valid.
     */
    private function isCurrentCertificateValid(?array $certInfo): bool {
        if (\is_array($certInfo)) {
            return !$this->isExpiredCertificate($certInfo) && $this->isSameDomainsList($certInfo);
        }
        return false;
    }

    /**
     * Checks certificate expiration.
     *
     * @param array   $certInfo         Parsed certificate info.
     * @param integer $timeToExpiration Time to expiration.
     *
     * @return boolean
     * @throws \Exception When missing expected key in the certificate info.
     */
    private function isExpiredCertificate(array $certInfo, int $timeToExpiration = self::TIME_TO_EXPIRATION): bool {
        if (!isset($certInfo['validTo_time_t'])) {
            $message = 'Missing expected key "validTo_time_t" in certificate';
            $this->getLog()->error(__LINE__, $message);
            throw new \Exception($message, 500);
        }

        $validToDate   = new \DateTime('@' . $certInfo['validTo_time_t']);
        $remainingTime = $validToDate->format('U') - time();

        if ($remainingTime > $timeToExpiration) {
            $this->getLog()->info(
                __LINE__,
                \sprintf('Certificate not expired and does not need renewal. Valid to date %s', $validToDate->format('Y-m-d H:i:s'))
            );
            return false;
        }

        $this->getLog()->info(__LINE__, \sprintf('Certificate expires. Valid to date %s', $validToDate->format('Y-m-d H:i:s')));
        return true;
    }

    /**
     * Compares the domain list.
     *
     * @param array $certInfo Parsed certificate info.
     *
     * @return boolean
     * @throws \Exception When missing expected key in the certificate info.
     */
    private function isSameDomainsList(array $certInfo): bool {
        if (!isset($certInfo['extensions']['subjectAltName'])) {
            $message = 'Missing expected key "extensions.subjectAltName" in certificate';
            $this->getLog()->error(__LINE__, $message);
            throw new \Exception($message, 500);
        }

        // DNS:*.005056a7048coo.wildixin.com, DNS:*.221100000b53.wildixin.com, DNS:005056a7048coo.wildixin.com, DNS:221100000b53.wildixin.com
        $subjectAltName     = \str_replace(['DNS:', '*.', ' '], '', $certInfo['extensions']['subjectAltName']);
        $subjectDomains     = \array_unique(\explode(',', $subjectAltName));
        $certificateDomains = [];

        foreach ($this->_domains as $domain) {
            $certificateDomains[] = $domain . '.' . $this->_zone;
        }

        \sort($subjectDomains);
        \sort($certificateDomains);

        if ($certificateDomains === $subjectDomains) {
            $this->getLog()->info(__LINE__, 'Certificate has same domain list');
            return true;
        }

        $this->getLog()->info(__LINE__, 'Certificate has different domain list and need renewal');
        return false;
    }

    /**
     * Notify about unsuccessful certificate renewal.
     *
     * @param array   $user          User data.
     * @param integer $remainingDays Certificate expire days.
     *
     * @return void
     */
    private function notifyAboutUnsuccessfulRenew(array $user, int $remainingDays): void {
        if (empty($user)) {
            return;
        }

        $message  = \sprintf('TLS Certificate for domain %s will expire in %d days.<br>', $this->_pbx->getDomain(), $remainingDays);
        $message .= $this->getFormattedErrors();

        if (!empty($user['mail'])) {
            $this->sendEmail($user['mail'], $message);
        } else {
            $name = $user['cn'] ?? 'user';
            $this->getLog()->info(__LINE__, \sprintf('The "%s" email not set', $name));
        }
    }

    /**
     * Saves certificate data to temporary files.
     *
     * @param string $certificate Certificate PEM.
     * @param string $primaryKey  Primary key.
     * @param string $prefix      Prefix.
     *
     * @return void
     * @throws Exception When can not save temporary files.
     */
    private function saveNewCertificate(string $certificate, string $primaryKey, string $prefix = ''): void {
        $tempCertificateFile = $this->getCertificatePath($prefix);
        $tempPrivateKeyFile  = $this->getPrivateKeyPath($prefix);
        $tempEjabberdPemFile = $this->getEjabberdPemPath($prefix);

        $this->saveFile($tempCertificateFile, $certificate);
        $this->saveFile($tempPrivateKeyFile, $primaryKey);
        $this->saveFile($tempEjabberdPemFile, $primaryKey . $certificate);

        $filesSaved = is_file($tempCertificateFile) && is_file($tempPrivateKeyFile) && is_file($tempEjabberdPemFile);

        if (!$filesSaved) {
            $this->getLog()->error(__LINE__, 'Can not save new certificate files');
            throw new Exception(Error::FileNotFound);
        }
    }

    /**
     * Backups old certificate files.
     *
     * @return void
     * @throws Exception When can not backup old certificate files.
     */
    private function backupOldCertificate(): void {
        $copyCertificateStatus = $this->copyFile($this->getCertificatePath(), $this->getCertificatePath(self::BACKUP_PREFIX));
        $copyPrivateKeyStatus  = $this->copyFile($this->getPrivateKeyPath(), $this->getPrivateKeyPath(self::BACKUP_PREFIX));

        if (!$copyCertificateStatus || !$copyPrivateKeyStatus) {
            $this->getLog()->error(__LINE__, 'Can not save backup for old certificate');
            throw new Exception(Error::ErrorCreatingBackup);
        }
    }

    /**
     * Backup original certificate file if not exist.
     *
     * @return void
     * @throws Exception When can not backup original certificate file.
     */
    private function backupOriginalCertificate(): void {
        $certificatePath         = $this->getCertificatePath();
        $privateKeyPath          = $this->getPrivateKeyPath();
        $originalCertificatePath = $this->getCertificatePath(self::ORIGINAL_PREFIX);
        $originalPrivateKeyPath  = $this->getPrivateKeyPath(self::ORIGINAL_PREFIX);

        if (is_file($originalCertificatePath) && is_file($originalPrivateKeyPath)) {
            $this->getLog()->info(__LINE__, 'Original certificate and private key exist, backup is not needed');
            return;
        }

        if (!is_file($certificatePath) || !is_file($privateKeyPath)) {
            $this->getLog()->info(__LINE__, 'Not found certificate file for save original backup');
            return;
        }

        $copyCertificateStatus = $this->copyFile($certificatePath, $originalCertificatePath);
        $copyPrivateKeyStatus  = $this->copyFile($privateKeyPath, $originalPrivateKeyPath);

        if (!$copyCertificateStatus || !$copyPrivateKeyStatus) {
            $this->getLog()->error(__LINE__, 'Can not save original certificate');
            throw new Exception(Error::ErrorCreatingBackup);
        }
    }

    /**
     * Replace old certificate files with new ones.
     *
     * @return void
     * @throws \Wildix\Exception When can not replace old certificate files with new ones.
     */
    private function replaceOldWithNewCertificate(): void {
        $renameCertificateStatus = $this->renameFile($this->getCertificatePath(self::TEMP_PREFIX), $this->getCertificatePath());
        $renamePrivateKeyStatus  = $this->renameFile($this->getPrivateKeyPath(self::TEMP_PREFIX), $this->getPrivateKeyPath());
        $renameEjabberdPemStatus = $this->renameFile($this->getEjabberdPemPath(self::TEMP_PREFIX), $this->getEjabberdPemPath());

        if (!$renameCertificateStatus || !$renamePrivateKeyStatus || !$renameEjabberdPemStatus) {
            $message = 'Can not rename certificate files';
            $this->getLog()->error(__LINE__, $message);
            throw new \Wildix\Exception($message, 500);
        }
    }

    /**
     * Calculate remaining days for certificate.
     *
     * @param array|null $certInfo Parsed certificate info.
     *
     * @return integer
     * @throws \Exception When missing expected key in the certificate info.
     */
    private function calculateRemainingDays(?array $certInfo): int {
        if (!\is_array($certInfo)
            || !isset($certInfo['validTo_time_t'])
            || (isset($certInfo['extensions']['subjectAltName']) && !$this->isSameDomainsList($certInfo))
        ) {
            return 0;
        }

        $oneDayInSeconds = 86400;
        $remainingDays   = \floor(($certInfo['validTo_time_t'] - time()) / $oneDayInSeconds);

        return $remainingDays > 0 ? $remainingDays : 0;
    }

    /**
     * Return admin user data.
     *
     * @return array
     */
    private function getAdminUser(): array {
        $result = $this->_wmsLdap->getEntry(
            \WMS_Ldap::PERSONS,
            '(&(cn=admin))',
            ['cn', 'mail', 'mobile']
        );

        if (\is_array($result)) {
            return $result;
        }

        if (\is_string($result)) {
            $this->getLog()->error(__LINE__, $result);
        }

        return [];
    }

    /**
     * Send email to user.
     *
     * @param string $email   Email.
     * @param string $message Message.
     *
     * @return void
     */
    private function sendEmail(string $email, string $message): void {
        if (empty($email)) {
            return;
        }

        $params = [
            'subject' => 'Error obtaining of certificate',
            'to' => $email,
            'body' => $message,
        ];

        $this->getLog()->info(__LINE__, \sprintf('Send email to admin user %s', $email));

        $this->_mailer->send($params);
    }

    /**
     * Returns a list of domains for which need to generate a certificate.
     *
     * @return array
     */
    private function getDomainsList(): array {
        $serial     = $this->_pbx->getSerial();
        $domains    = [$serial, $serial . '--vpn'];
        $domainName = $this->_pbx->getDomain();

        if (!empty($domainName)) {
            $domains[] = \str_replace('.' . $this->_zone, '', $domainName);
        }

        return \array_unique($domains);
    }

    /**
     * Restart services depended from certificates
     *
     * @return void
     */
    private function reloadServices(): void {
        $this->getLog()->info(__LINE__, 'Reload tproxyd and ejabberd services');
        exec('sudo /etc/init.d/tproxyd reload');
        exec('sudo systemctl reload-or-restart ejabberd.service');
        $this->getLog()->info(__LINE__, 'Update certificate revocation list (CRL) for Kamailio');
        exec('sudo /usr/sbin/kamcmd tls.reload');
    }

    private function saveRemainingDaysInfo($data): bool {
        return $this->saveFile(self::STORAGE, $data);
    }

    private function saveErrorMessages(string $errors): bool {
        return $this->saveFile(self::ERROR_MESSAGES_FILEPATH, $errors);
    }

    private function getLog(): Log {
        return $this->_log;
    }

    private function getCertificatePath(string $prefix = ''): string {
        return self::CERTIFICATE_DIR . DIRECTORY_SEPARATOR . $prefix . self::CERTIFICATE_NAME;
    }

    private function getPrivateKeyPath(string $prefix = ''): string {
        return self::CERTIFICATE_DIR . DIRECTORY_SEPARATOR . $prefix . self::PRIVATE_KEY_NAME;
    }

    private function getEjabberdPemPath(string $prefix = ''): string {
        return self::CERTIFICATE_DIR . DIRECTORY_SEPARATOR . $prefix . self::EJABBERD_CERTIFICATE_NAME;
    }

    private function saveFile(string $filename, string $data): bool {
        $this->getLog()->info(__LINE__, sprintf('Save file: %s', $filename));

        $status = file_put_contents($filename, $data);
        $this->changeMode($filename);
        return $status;
    }

    private function copyFile(string $filename, string $dest): bool {
        $this->getLog()->info(__LINE__, sprintf('Copy file: %s to %s', $filename, $dest));

        $status = copy($filename, $dest);
        $this->changeMode($filename);
        $this->changeMode($dest);

        return $status;
    }

    private function renameFile(string $filename, string $dest): bool {
        $this->getLog()->info(__LINE__, sprintf('Rename file: %s to %s', $filename, $dest));

        return rename($filename, $dest);
    }

    private function removeFile(string $file): void {
        if (is_file($file)) {
            $this->getLog()->info(__LINE__, sprintf('Delete file: %s', $file));
            unlink($file);
        }
    }

    private function changeMode(string $file): void {
        if (is_file($file)) {
            $this->getLog()->info(__LINE__, sprintf('Change user, group and mode for file: %s', $file));

            chmod($file, 0775);
            chown($file, 'wms');
            chgrp($file, 'wms');
        }
    }
}
