<?php
namespace Wildix\Colleagues;

use Api\Dictionary\Enum\LanguagesEnum;
use Services\Helpers;
use Services\LuaService;
use Wildix\Core\Container;
use Wildix\Generator\StringGenerator;

/**
 * Model Colleague
 */
class Colleague {
    const DN           = 'dn';
    const ID           = 'id';
    const EXTENSION    = 'extension';
    const JID          = 'jid';
    const OFFICE_PHONE = 'officePhone';
    const MOBILE_PHONE = 'mobilePhone';
    const FAX_NUMBER   = 'faxNumber';
    const NAME         = 'name';
    const EMAIL        = 'email';
    const PBX_DN       = 'pbxDn';
    const ROLE         = 'role';
    const GROUP_DN     = 'groupDn';
    const PASSWORD     = 'password';
    const SIP_PASSWORD = 'sipPassword';
    const USER_PIN     = 'userPIN';
    const LANGUAGE     = 'language';
    const DIALPLAN     = 'dialplan';
    const FAX_DIALPLAN = 'faxDialplan';
    const DEPARTMENT   = 'department';
    const PICTURE      = 'picture';
    const LOGIN        = 'login';
    const SOURCE_ID    = 'sourceId';
    const LICENSE_TYPE = 'licenseType';
    const GROUP_NAME   = 'groupName';
    const PBX          = 'pbx';

    /**
     * Role
     */
    const ADMINISTRATOR = 'admin';
    const ROLE_USER = 'user';
    const ROLE_FAX = 'fax';
    const ROLE_PARK_ORBIT = 'park_orbit';
    const ROLE_ROOM = 'room';

    protected $_ldapProperties = null;
    private $_uid              = null;

    public function __construct($properties) {
        $this->setPropertys($properties);
    }

    public function setPropertys($properties) {
        if (is_array($properties)) {
            $this->_ldapProperties = $properties;

            if (empty($this->_ldapProperties[self::LICENSE_TYPE])) {
                $this->_ldapProperties[self::LICENSE_TYPE] = \Wildix\License::getInstance()->getMaximumAvailableLicenseType();
            }
        }

        if (!empty($properties[self::ID])) {
            $this->_uid = $properties[self::ID];
        } else {
            $this->_uid = null;
        }
    }

    public function getPropertys() {
        return $this->_ldapProperties;
    }

    public function getDn() {
        return $this->getLdapProperty(self::DN);
    }

    public function getID() {
        return $this->_uid;
    }

    public function getExtension() {
        return $this->getLdapProperty(self::EXTENSION);
    }

	public function getJid() {
        $pbxDomain = $this->getPbxDomain();

        return (strlen($pbxDomain) > 0) ? ($this->getLdapProperty(self::EXTENSION) . '@' . $pbxDomain) : '';
    }

    public function getOfficePhone() {
        return $this->getLdapProperty(self::OFFICE_PHONE);
    }

    public function getMobilePhone() {
        return $this->getLdapProperty(self::MOBILE_PHONE);
    }

    public function setMobilePhone($number) {
        return $this->setLdapProperty(self::MOBILE_PHONE, $number);
    }

    public function getFaxNumber() {
        return $this->getLdapProperty(self::FAX_NUMBER);
    }

    public function getName() {
        return $this->getLdapProperty(self::NAME);
    }

    public function getEmail() {
        return $this->getLdapProperty(self::EMAIL);
    }

    public function setEmail($email) {
        return $this->setLdapProperty(self::EMAIL, $email);
    }

    public function getPBXDn() {
        return $this->getLdapProperty(self::PBX_DN);
    }

    public function getRole() {
        return $this->getLdapProperty(self::ROLE);
    }

    public function getGroupDn() {
        return $this->getLdapProperty(self::GROUP_DN);
    }

    public function getPassword() {
        return $this->getLdapProperty(self::PASSWORD);
    }

    public function setPassword($password) {
        return $this->setLdapProperty(self::PASSWORD, $password);
    }

    public function getUserPIN(){
        return $this->getLdapProperty(Colleague::USER_PIN);
    }

    public function setUserPIN($pin){
        return $this->setLdapProperty(Colleague::USER_PIN, $pin);
    }

    public function getSipPassword() {
        return $this->getLdapProperty(self::SIP_PASSWORD);
    }

    public function setSipPassword(string $password) {
        return $this->setLdapProperty(self::SIP_PASSWORD, $password);
    }

    public function getLanguage() {
        return $this->getLdapProperty(self::LANGUAGE) ?: LanguagesEnum::EN;
    }

    public function getDialplan() {
        return $this->getLdapProperty(self::DIALPLAN);
    }

    public function getFaxDialplan() {
        return $this->getLdapProperty(self::FAX_DIALPLAN);
    }

    public function getDepartment() {
        return $this->getLdapProperty(self::DEPARTMENT);
    }

    public function getPicture() {
        return Helpers::makePictureUrl($this->getLdapProperty(self::PICTURE), $this->getExtension(), $this->getName());
    }

    public function getSourceId() {
        return $this->getLdapProperty(self::SOURCE_ID);
    }

    public function getLicenseType() {
        return $this->getLdapProperty(self::LICENSE_TYPE);
    }

    public function getPBX() {
        $dn = $this->getPBXDn();
        return strtoupper(substr($dn, 6, 12));
    }

    public function getPbxDomain() {
        $pbx = \Wildix\PBXes\PBXes::getInstance()->getPBXByDN($this->getPBXDn());

        return isset($pbx) ? $pbx->getDomain() : '';
    }

    private $_settings = null;

    /**
     * Access to user settings
     *
     * @return \Wildix\Colleagues\Settings
     */
    public function getSettings() {
        if ($this->_settings === null) {
            $this->_settings = new \Wildix\Colleagues\Settings($this);
        }
        return $this->_settings;
    }

    public function isAdministrator() {
        return $this->getRole() === self::ADMINISTRATOR;
    }

    /**
     * @return bool
     */
    public function isRoot() {
        return $this->isAdministrator() && $this->getName() == 'admin';
    }


    public function getAsArray($fields = null) {
        if (!empty($this->_ldapProperties)) {
            $arr                = array_diff_key($this->_ldapProperties, array(self::PASSWORD => 1, self::SIP_PASSWORD => 1, self::USER_PIN => 1));
            $arr[self::JID]     = $this->getJid();
            $arr[self::PICTURE] = $this->getPicture();
            if (empty($fields)) {
                return $arr;
            }
            return array_intersect_key($arr, array_flip($fields));
        }
        return array();
    }

    public function getLdapProperty($name) {
        if ($this->_ldapProperties === null && $this->_uid !== null) {
            $this->_ldapProperties = Colleagues::getInstance()->getUserByID($this->_uid);
        }

        if (!empty($this->_ldapProperties) && isset($this->_ldapProperties[$name])) {
            return $this->_ldapProperties[$name];
        }

        return '';
    }

    public function setLdapProperty($name, $value) {
        if ($this->_ldapProperties !== null && $this->_uid !== null) {
            if ($this->_ldapProperties[$name] != $value) {
                if (!Colleagues::getInstance()->updateUserProperty(array($name => $value), $this)) {
                    return false;
                }
                $this->_ldapProperties[$name] = $value;
            }
            return true;
        }
        return false;
    }

    // @codingStandardsIgnoreStart

    /**
     * Start ACL methods
     **/

    /**
     * Group: can/cannot
     **/
    const CALL                       = 'CALL';
    const SEND_DOCUMENTS             = 'IM_FAX';
    const CHANGE_PRESENCE            = 'CHANGE_PRESENCE';
    const INTERCOM                   = 'INTERCOM';
    const INTRUSION                  = 'SUPERVISION';
    const CALL_PICKUP                = 'PICKUP';
    const MODIFY_PUBLIC_PHONEBOOKS   = 'EDIT_PP';
    const DELETE_CALLS               = 'DELETE_CALLS';
    const VIEW                       = 'SEE';
    const SHARE_STATUS_VIA_KITE      = 'SHARE_STATUS_VIA_KITE';
    const SHARE_STATUS_MESS_VIA_KITE = 'SHARE_STATUS_MESS_VIA_KITE';
    const SHARE_GEO_VIA_KITE         = 'SHARE_GEO_VIA_KITE';
    const VIEW_GEO                   = 'VIEW_GEO';
    const CREATE_CONFERENCE          = 'CREATE_CONFERENCE';
    const SEE_CDRVIEW_DIGITS         = 'SEE_CDRVIEW_DIGITS';
    const HOTEL_ACCESS_ALARM_CALL_OF = 'ACCESS_ALARM_CALL_OF';
    const HOTEL_ADMIN                = 'HOTEL_ADMIN';
    const HOTEL_EVENTS               = 'HOTEL_EVENTS';
    const MANAGE_CALLCENTER          = 'MANAGE_CALLCENTER';
    const DIRECTORY_LOOKUP           = 'DIRECTORY_LOOKUP';
    const TWO_FACTOR_AUTH_OFF        = 'TWO_FACTOR_AUTH_OFF';
    const ENABLE_VIDEO_CALL          = 'ENABLE_VIDEO_CALL';
    const SEE_EXTENSIONS             = 'SEE_EXTENSIONS';
    const SEE_VOICEMAIL              = 'SEE_VOICEMAIL';
    const SEE_CALLS_OF_USERS         = 'SEE_CALLS';
    const SEE_CALL_RECORDINGS        = 'SEE_CALL_RECORDINGS';
    public const CAN_CANNOT_TO_GROUP = [
        self::CALL,
        self::SEND_DOCUMENTS,
        self::CHANGE_PRESENCE,
        self::INTERCOM,
        self::INTRUSION,
        self::CALL_PICKUP,
        self::VIEW,
        self::VIEW_GEO,
        self::SEE_VOICEMAIL,
        self::SEE_CALLS_OF_USERS,
    ];
    /**
     * ACL list without Groups!
     **/
    public const CAN_CANNOT_COMMON = [
        self::MODIFY_PUBLIC_PHONEBOOKS,
        self::DELETE_CALLS,
        self::SEE_CDRVIEW_DIGITS,
        self::HOTEL_ACCESS_ALARM_CALL_OF,
        self::HOTEL_ADMIN,
        self::HOTEL_EVENTS,
        self::SHARE_STATUS_VIA_KITE,
        self::SHARE_STATUS_MESS_VIA_KITE,
        self::SHARE_GEO_VIA_KITE,
        self::MANAGE_CALLCENTER,
        self::DIRECTORY_LOOKUP,
        self::CREATE_CONFERENCE,
        self::TWO_FACTOR_AUTH_OFF,
        self::ENABLE_VIDEO_CALL,
        self::SEE_EXTENSIONS,
        self::SEE_CALL_RECORDINGS
    ];

    /**
     * Group: can set/cannot set
     **/
    const DO_NOT_DISTURB                 = 'DND';
    const BUSY_CALL_FORWARD              = 'BCF';
    const UNAVAILABLE_CALL_FORWARD       = 'UCF';
    const FORCE_CALL_FORWARD             = 'FCF';
    const CALL_WAITING                   = 'CW';
    const MOBILITY_EXTENSION_MANAGEMENT  = 'MEXT';
    const DIAL_TIMEOUT                   = 'TOUT';
    const TELEPHONE_BLOCKED              = 'CALLBLOCK';
    const RING_ONLY_ACTIVE_DEVICE        = 'ROAD';
    const MOBILITY_CONFIRMATION          = 'MOBCONFIRMATION';
    const FUNCTION_KEYS                  = 'FKEYS';
    const TIMETABLE                      = 'STIME';
    const THREE_STATE_SWITCH             = '3SWITCH';
    const SWITCH_                        = 'SWITCH';
    const PHONEBOOKS                     = 'PHONEBOOK';
    const PERSONAL_INFORMATION           = 'P_INFO';
    const ADVANCED_STATUS                = 'ADVANCED_STATUS';
    const FAXSERVERSET                   = 'FAXSERVERSET';
    const EMAIL_MISSED_CALL_NOTIFICATION = 'EMAILMCN';
    const SMS_MISSED_CALL_NOTIFICATION   = 'SMSMCN';
    const RINGTONE                       = 'RING';
    public const CAN_CANNOT_SET          = [
        self::DO_NOT_DISTURB,
        self::BUSY_CALL_FORWARD,
        self::UNAVAILABLE_CALL_FORWARD,
        self::FORCE_CALL_FORWARD,
        self::CALL_WAITING,
        self::MOBILITY_EXTENSION_MANAGEMENT,
        self::DIAL_TIMEOUT,
        self::TELEPHONE_BLOCKED,
        self::RING_ONLY_ACTIVE_DEVICE,
        self::MOBILITY_CONFIRMATION,
        self::FUNCTION_KEYS,
        self::TIMETABLE,
        self::THREE_STATE_SWITCH,
        self::SWITCH_,
        self::PHONEBOOKS,
        self::PERSONAL_INFORMATION,
        self::ADVANCED_STATUS,
        self::FAXSERVERSET,
        self::EMAIL_MISSED_CALL_NOTIFICATION,
        self::SMS_MISSED_CALL_NOTIFICATION,
        self::RINGTONE,
    ];

    /**
     * Group: can use/cannot use
     **/
    const CTI_PRO                  = 'CTI_PRO';
    const CTI_PO                   = 'CTI_PO';
    const CTI_CDR                  = 'CTI_CDR';
    const CTI_CDR_VIEW             = 'CTI_CDR_VIEW';
    const SPEED_DIAL               = 'SHORTNUM';
    const PBX_RECORD               = 'PBXRECORD';
    const PERSONAL_RECORD          = 'PERSONALRECORD';
    const SMS                      = 'SMS';
    const FAX                      = 'FAX';
    const PAGING                   = 'PAGING';
    const PRE_ANSWER_SERVICES      = 'PRE_ANSWER_SERVICES';
    const PRE_ANSWER_SERVICES_FULL = 'PRE_ANSWER_SERVICES_FULL';
    const PHONE_SETTING_MENU       = 'PHONE_SETTING_MENU';
    const ADV_PHONE_SETTINGS_MENU  = 'ADVANCED_PHONE_SETTINGS_MENU';
    const BROWSER_PHONE            = 'BROWSER_PHONE';
    const VOICEMAIL                = 'VOICEMAIL';
    const VOICEMAIL_WITHOUT_PIN    = 'VOICEMAIL_WITHOUT_PIN';
    const CONTACT_CENTER           = 'CONTACT_CENTER';
    const TRUNK_TO_TRUNK_TRANSFER  = 'TRUNK_TO_TRUNK_TRANSFER';
    const TRUNK_TO_TRUNK_FORWARD   = 'TRUNK_TO_TRUNK_FORWARD';
    public const CAN_CANNOT_USE    = [
        self::CTI_PRO,
        self::CTI_PO,
        self::CTI_CDR,
        self::CTI_CDR_VIEW,
        self::SPEED_DIAL,
        self::PBX_RECORD,
        self::PERSONAL_RECORD,
        self::SMS,
        self::FAX,
        self::PAGING,
        self::PRE_ANSWER_SERVICES,
        self::PRE_ANSWER_SERVICES_FULL,
        self::PHONE_SETTING_MENU,
        self::ADV_PHONE_SETTINGS_MENU,
        self::BROWSER_PHONE,
        self::VOICEMAIL,
        self::VOICEMAIL_WITHOUT_PIN,
        self::CONTACT_CENTER,
        self::TRUNK_TO_TRUNK_TRANSFER,
        self::TRUNK_TO_TRUNK_FORWARD,
    ];

    const ADVANCED_PHONE_SETTINGS_MENU = 'ADVANCED_PHONE_SETTINGS_MENU';

    /**
     * Group: can call/cannot call
     **/
    const INTERNAL      = '1';
    const LOCAL         = '2';
    const NATIONAL      = '3';
    const MOBILE        = '4';
    const EMERGENCY     = '5';
    const FREE          = '6';
    const PREMIUM1      = '7';
    const PREMIUM2      = '8';
    const PREMIUM3      = '9';
    const PREMIUM4      = '10';
    const NORTH_AMERICA = '11';
    const AFRICA        = '12';
    const EUROPE1       = '13';
    const EUROPE2       = '14';
    const SOUTH_AMERICA = '15';
    const OCEANIA       = '16';
    const RUSSIA        = '17';
    const ASIA1         = '18';
    const ASIA2         = '19';
    const SHORT_NUMBERS = '20';
    const INTERNATIONAL = '21';
    const ALL           = 'all';
    public const CAN_CANNOT_CALL = [
        self::INTERNAL,
        self::LOCAL,
        self::NATIONAL,
        self::MOBILE,
        self::EMERGENCY,
        self::FREE,
        self::PREMIUM1,
        self::PREMIUM2,
        self::PREMIUM3,
        self::PREMIUM4,
        self::NORTH_AMERICA,
        self::AFRICA,
        self::EUROPE1,
        self::EUROPE2,
        self::SOUTH_AMERICA,
        self::OCEANIA,
        self::RUSSIA,
        self::ASIA1,
        self::ASIA2,
        self::SHORT_NUMBERS,
        self::INTERNATIONAL,
        self::ALL,
    ];

    // @codingStandardsIgnoreEnd

    /**
     * This method verifies whether the user can
     * CALL
     * SEND_DOCUMENTS
     * INTERCOM
     * INTRUSION
     * CALL_PICKUP
     * MODIFY_PUBLIC_PHONEBOOKS
     * DELETE_CALLS
     * SEE_MOBILE_NUMBER
     * VIEW
     * SHARE_STATUS_VIA_KITE
     * CREATE_CONFERENCE
     * TWO_FACTOR_AUTH_OFF
     * ENABLE_VIDEO_CALL
     * SEE_VOICEMAIL
     * SEE_CALLS_OF_USERS
     * SEE_CALL_RECORDINGS
     *
     * Examples: $result = $user->can($user::CALL);
     *
     * @param string $what        Acl name.
     * @param string $destGroupDn GroupDn name.
     *
     * @return boolean
     */
    public function can($what, $destGroupDn = null) {
        return $this->checkACL('yesno', $what, $destGroupDn);
    }

    public function canNot($what, $destGroupDn = null) {
        return !$this->can($what, $destGroupDn);
    }

    /**
     * This method checks whether the user can set
     * DO_NOT_DISTURB
     * BUSY_CALL_FORWARD
     * UNAVAILABLE_CALL_FORWARD
     * FORCE_CALL_FORWARD
     * CALL_WAITING
     * MOBILITY_EXTENSION_MANAGEMENT
     * DIAL_TIMEOUT
     * TELEPHONE_BLOCKED
     * MOBILITY_CONFIRMATION
     * FUNCTION_KEYS
     * TIMETABLE
     * THREE_STATE_SWITCH
     * SWITCH_
     * PHONEBOOKS
     * PERSONAL_INFORMATION
     * EMAIL_MISSED_CALL_NOTIFICATION
     * SMS_MISSED_CALL_NOTIFICATION
     * RINGTONE
     * ALL
     *
     * Examples: $result = $user->canSet($user::DO_NOT_DISTURB);
     *
     * @param string $what Acl name.
     *
     * @return boolean
     */
    public function canSet($what) {
        return $this->checkACL('canset', $what);
    }

    public function canNotSet($what) {
        return !$this->canSet($what);
    }

    /**
     * This method checks whether the user can use
     * CTI_PRO
     * CTI_PO
     * CTI_CDR
     * CTI_CDR_VIEW
     * SPEED_DIAL
     * PBX_RECORD
     * PERSONAL_RECORD
     * FAX
     * PAGING
     * ALL
     *
     * Examples: $result = $user->canUse($user::CTI_PRO);
     *
     * @param string $what Acl name.
     *
     * @return boolean
     */
    public function canUse($what) {
        return $this->checkACL('canuse', $what);
    }

    public function canNotUse($what) {
        return !$this->canUse($what);
    }


    /**
     * This method checks whether the user can call
     * LOCAL
     * NATIONAL
     * MOBILE
     * EMERGENCY
     * FREE
     * PREMIUM1
     * PREMIUM2
     * PREMIUM3
     * PREMIUM4
     * NORTH_AMERICA
     * AFRICA
     * EUROPE1
     * EUROPE2
     * SOUTH_AMERICA
     * OCEANIA
     * RUSSIA
     * ASIA1
     * ASIA2
     * SHORT_NUMBERS
     * All
     *
     * Examples: $result = $user->canCall($user::LOCAL);
     *
     * @param string $what Acl name.
     *
     * @return boolean
     */
    public function canCall($what) {
        return $this->checkACL('cancall', $what);
    }

    public function canNotCall($what) {
        return !$this->canCall($what);
    }

    public function checkACL($canType, $what, $destGroupDn = null): bool {
        $luaService = Container::get(LuaService::class);
        return $luaService->checkACL($this->getGroupDn(), $canType, $what, $destGroupDn, $this->getLicenseType());
    }

    /**
     * End ACL methods
     **/

    /**
     * Admin permissions
     *
     * @param string $page Page to check.
     *
     * @return boolean
     **/
    public function isAdminPageAllowed($page) {
        return \WMS_Permissions::getInstance()->is_allowed('page', $page);
    }

    /**
     * Check is internal or imported user
     *
     * @return boolean
     */
    public function isInternalUser(): bool {
        return empty($this->getSourceId()) || $this->getSourceId() === $this->getID();
    }

    public function getAclByFilter($filter) {
        $aclList = array();

        if (!$filter || !is_array($filter)) {
            return $aclList;
        }

        if (isset($filter['yesno'])) {
            $aclList['yesno'] = $this->aclFilterYesNo($filter['yesno']);
        }

        if (isset($filter['canset'])) {
            $aclList['canset'] = $this->aclFilterByType($filter['canset'], 'canSet');
        }
        if (isset($filter['canuse'])) {
            $aclList['canuse'] = $this->aclFilterByType($filter['canuse'], 'canUse');
        }
        if (isset($filter['cancall'])) {
            $aclList['cancall'] = $this->aclFilterByType($filter['cancall'], 'canCall');
        }

        return $aclList;
    }

    private function aclFilterByType($filter, $type) {
        $pos    = strpos(',', $filter);
        $aclArr = array();

        if ($pos !== 'false') {
            $options = explode(',', $filter);
            foreach ($options as $option) {
                $aclArr[] = trim($option);
            }
        } else {
            $aclArr[] = $filter;
        }

        $result = array();
        foreach ($aclArr as $acl) {
            $result[$acl] = $this->$type($acl);
        }

        return $result;
    }

    private function aclFilterYesNo($filter) {
        $result = array();

        $pos    = strpos(',', $filter);
        $aclArr = array();

        if ($pos !== 'false') {
            $options = explode(',', $filter);
            foreach ($options as $option) {
                $aclArr[] = trim($option);
            }
        } else {
            $aclArr[] = $filter;
        }

        $groups = \Wildix\Colleagues\Groups::getInstance()->getGroups(\User::getInstance());

        foreach ($aclArr as $acl) {
            if ($groups && !in_array($acl, self::CAN_CANNOT_COMMON)) {
                foreach ($groups as $key => $group) {
                    $result[$acl][$group->getDn()] = $this->can($acl, $group->getDn());
                }
            } else {
                $result[$acl] = $this->can($acl);
            }
        }

        return $result;
    }

    public function resetPassword(): void {
        $password = StringGenerator::random();

        $this->setPassword(\User::getInstance()->cryptPassword($password));
        $this->setUserPIN(\User::getInstance()->makeUserPinFromPassword($password));

        $this->setSipPassword(StringGenerator::random());
    }
}
