<?php

namespace Wildix\Colleagues;

use Wildix\Error;
use Wildix\Generator\StringGenerator;
use Wildix\InvalidDataException;
use Wildix\License;
use Wildix\Normalizer\PhoneNumberNormalizer;

/**
 * Class ColleagueFactory
 */
class ColleagueFactory {

    /**
     * Values of these properties is phone number
     *
     * @var array
     */
    private static $_phonesProperties = array(Colleague::FAX_NUMBER, Colleague::MOBILE_PHONE, Colleague::OFFICE_PHONE);
    /**
     * Properties which cannot be changed
     * Remove sip password from list WMS-5931
     *
     * @var array
     */
    private static $_unmanagedProperties = array(
                                            Colleague::ID,
                                            Colleague::DN,
                                            Colleague::PBX_DN,
                                            Colleague::USER_PIN,
    );

    /**
     * Instance of validator
     *
     * @var Validator
     */
    private $_validator;
    /**
     * Instance of normalizer for phone numbers
     *
     * @var PhoneNumberNormalizer
     */
    private $_phoneNumber;
    /**
     * Base LDAP DN
     *
     * @var string
     */
    private $_baseDN;
    /**
     * Current PBX LDAP DN
     *
     * @var string
     */
    private $_pbxDN;
    /**
     * Current PBX LDAP DN
     *
     * @var string
     */
    private $_availableGroupDn;

    /**
     * ColleagueFactory constructor.
     *
     * @param Validator             $validator             Instance of validator.
     * @param PhoneNumberNormalizer $phoneNumberNormalizer Instance of normalizer for phone numbers.
     * @param string                $baseDN                Base LDAP DN.
     * @param string                $pbxDN                 Current PBX LDAP DN.
     * @param string                $availableGroupDn      Available group DN.
     */
    public function __construct(
        Validator $validator,
        PhoneNumberNormalizer $phoneNumberNormalizer,
        string $baseDN,
        string $pbxDN,
        string $availableGroupDn
    ) {
        $this->_validator        = $validator;
        $this->_phoneNumber      = $phoneNumberNormalizer;
        $this->_baseDN           = $baseDN;
        $this->_pbxDN            = $pbxDN;
        $this->_availableGroupDn = $availableGroupDn;
    }

    /**
     * Build a colleague instance from data
     *
     * @param array $data Colleague data.
     *
     * @throws InvalidDataException If data not passed validation
     * @return Colleague
     */
    public function build(array $data) {
        $this->assertDataIsValid($data, array(Colleague::EXTENSION, Colleague::NAME));

        $normalizedProperties = $this->normalizeData($data);
        $completedProperties  = array_merge($normalizedProperties, $this->generateNewUserProperties($normalizedProperties));

        return new Colleague($completedProperties);
    }

    /**
     * Update colleague instance by data
     *
     * @param Colleague $colleague Instance for update.
     * @param array     $data      New data.
     *
     * @throws InvalidDataException If data not passed validation
     * @return void
     */
    public function update(Colleague $colleague, array $data) {
        $this->assertDataIsValid($data);
        $this->assertDataCanBeApplied($colleague, $data);

        $normalizedProperties = $this->normalizeData($data);

        $secrets = isset($normalizedProperties[Colleague::PASSWORD]) ? $this->buildPasswordWithPin($normalizedProperties[Colleague::PASSWORD]) : array();

        $colleague->setPropertys(array_merge($colleague->getPropertys(), $normalizedProperties, $secrets));
    }

    /**
     * Checking that data is valid
     *
     * @param array      $data           Data for validation.
     * @param array|null $requiredFields Fields that must be present in the data.
     *
     * @throws InvalidDataException If data not passed validation
     * @return void
     */
    private function assertDataIsValid(array $data, array $requiredFields = null) {
        $errors = $this->_validator->validate($data);

        if (isset($requiredFields)) {
            foreach ($requiredFields as $field) {
                if (isset($data[$field])) {
                    continue;
                }

                $errors[$field] = array(
                    'code' => Error::FieldRequired,
                    'reason' => Error::getMessage(Error::FieldRequired),
                );
            }
        }

        if ($errors && count($errors) > 0) {
            // Error should be like - Invalid data
            throw new InvalidDataException(Error::getMessage(Error::UnknownError), $errors);
        }
    }

    /**
     * Normalizing data
     *
     * @param array $data Data for normalize.
     *
     * @return array
     */
    private function normalizeData(array $data) {
        $result = array();

        foreach ($data as $property => $value) {
            if (in_array($property, self::$_phonesProperties)) {
                $result[$property] = $this->_phoneNumber->normalize($value);

                continue;
            }

            $result[$property] = $value;
        }

        return $result;
    }

    /**
     * Generate data for new user
     *
     * @param array $normalizedProperties A user's normalized properties.
     *
     * @return array
     */
    private function generateNewUserProperties(array $normalizedProperties) {
        $id     = $this->generateId();
        $passwd = isset($normalizedProperties[Colleague::PASSWORD]) ? $normalizedProperties[Colleague::PASSWORD] : StringGenerator::random(10);

        $uncontrolledData = [
            Colleague::ID => $id,
            Colleague::DN => $this->buildDn($id),
            Colleague::PBX_DN => $this->_pbxDN,
        ];

        $licenseTypeKey = Colleague::LICENSE_TYPE;

        $sipPassword = !empty($normalizedProperties[Colleague::SIP_PASSWORD]) ? $normalizedProperties[Colleague::SIP_PASSWORD] : StringGenerator::random(10);
        $language    = !empty($normalizedProperties[Colleague::LANGUAGE]) ? $normalizedProperties[Colleague::LANGUAGE] : 'en';
        $groupDn     = !empty($normalizedProperties[Colleague::GROUP_DN]) ? $normalizedProperties[Colleague::GROUP_DN] : $this->_availableGroupDn;
        $licenseType = !empty($normalizedProperties[$licenseTypeKey]) ? $normalizedProperties[$licenseTypeKey] : License::getDefaultLicenseType();

        $defaultData = [
            Colleague::SIP_PASSWORD => $sipPassword,
            Colleague::LANGUAGE => $language,
            Colleague::GROUP_DN => $groupDn,
            Colleague::LICENSE_TYPE => $licenseType,
        ];

        return $uncontrolledData + $defaultData + $this->buildPasswordWithPin($passwd);
    }

    /**
     * Generate new id
     *
     * @return integer
     */
    private function generateId() {
        return mt_rand(1000000, 10000000);
    }

    /**
     * Build LDAP user DN
     *
     * @param integer $id User id.
     *
     * @return string
     */
    private function buildDn($id) {
        return 'uid=' . $id . ',o=' . \WMS_Ldap::ORGANIZATION_NAME . ',' . $this->_baseDN;
    }

    private function buildPasswordWithPin($openPassword) {
        return array(
            Colleague::PASSWORD => \User::cryptPassword($openPassword),
            Colleague::USER_PIN => \User::makeUserPinFromPassword($openPassword),
        );
    }

    /**
     * Checking that data can be applied to the colleague instance
     *
     * @param Colleague $colleague Instance for applying data.
     * @param array     $data      Data which should be applied.
     *
     * @throws InvalidDataException If data cannot be applied
     * @return void
     */
    private function assertDataCanBeApplied(Colleague $colleague, array $data) {
        $unmanagedPropertiesInData = array_intersect(array_keys($data), self::$_unmanagedProperties);
        $errors                    = array();

        if (count($unmanagedPropertiesInData) > 0) {
            $errors += array_reduce(
                $unmanagedPropertiesInData,
                function (array $result, $property) {
                    $result[$property] = array(
                        // Error - field cannot be changed
                        'code' => Error::UnknownError,
                        'reason' => Error::getMessage(Error::UnknownError),
                    );

                    return $result;
                },
                array()
            );
        }

        if (isset($data[Colleague::NAME])) {
            if ($data[Colleague::NAME] != 'admin' && $colleague->isRoot()) {
                $errors[Colleague::NAME] = array(
                    'code' => Error::UnknownError,
                    'reason' => 'The administrator name can not be changed',
                );
            } else if ($data[Colleague::NAME] == 'admin' && !$colleague->isRoot()) {
                $errors[Colleague::NAME] = array(
                    'code' => Error::UnknownError,
                    'reason' => 'This name is not available',
                );
            }
        }

        if (count($errors) > 0) {
            // Error should be like - This data cannot be applied for this user
            throw new InvalidDataException(Error::getMessage(Error::UnknownError), $errors);
        }
    }
}
