<?php
namespace Wildix\Colleagues;

use Wildix\Error;
use Wildix\Exception;
use Wildix\InvalidDataException;

/**
 * Model Colleagues
 */
class Colleagues extends \Singleton {
    const FILE_GLOBAL_ROSTER_CACHE = '/dev/shm/groster.lua';

     const SEARCH_STRATEGY_STARTS_WITH = 'startsWith';
     const SEARCH_STRATEGY_CONTAIN     = 'contain';

    private $_keyMap = array(
                    Colleague::DN           => 'dn',
                    Colleague::ID           => 'uid',
                    Colleague::NAME         => 'cn',
                    Colleague::LOGIN        => 'userPrincipalName',
                    Colleague::EXTENSION    => 'telephoneNumber',
                    Colleague::OFFICE_PHONE => 'wOfficePhone',
                    Colleague::MOBILE_PHONE => 'mobile',
                    Colleague::FAX_NUMBER   => 'facsimileTelephoneNumber',
                    Colleague::EMAIL        => 'mail',
                    Colleague::PBX_DN       => 'pbxowner',
                    Colleague::PBX          => 'pbx',
                    Colleague::ROLE         => 'userKind',
                    Colleague::GROUP_NAME   => 'groupName',
                    Colleague::GROUP_DN     => 'owner',
                    Colleague::PASSWORD     => 'userPassword',
                    Colleague::SIP_PASSWORD => 'wuserPasswordSip',
                    Colleague::USER_PIN     => 'userPIN',
                    Colleague::LANGUAGE     => 'preferredLanguage',
                    Colleague::DIALPLAN     => 'userDialplan',
                    Colleague::FAX_DIALPLAN => 'userFaxDialplan',
                    Colleague::DEPARTMENT   => 'ou',
                    Colleague::PICTURE      => 'imageUrl',
                    Colleague::SOURCE_ID    => 'sourceId',
                    Colleague::LICENSE_TYPE => 'licenseType',
                );

    private $_log = null;
    private function getLog() {
        if ($this->_log === null) {
            $this->_log = new \Syslog('\Collegues', 4);
        }
        return $this->_log;
    }

    public function getLdapKey($key) {
        return $this->_keyMap[$key] ? $this->_keyMap[$key] : $key;
    }

    public function getLdapKeys() {
        return array_values($this->_keyMap);
    }

    private $_keyMapFlip = null;
    public function getRevertKey($key) {
        if ($this->_keyMapFlip === null) {
            $this->_keyMapFlip = array_flip($this->_keyMap);
        }
        return $this->_keyMapFlip[$key] ? $this->_keyMapFlip[$key] : $key;
    }

    public function getUserByFilter($filter) {
        $data = \WMS_Ldap::getInstance()->getEntry(
            \WMS_Ldap::PERSONS,
            $filter,
            $this->getLdapKeys()
        );

        if ($data) {
            $us = array();
            foreach ($this->getLdapKeys() as $key) {
                $us[$this->getRevertKey($key)] = isset($data[$key]) ? $data[$key] : '';
            }

            if (empty($us[\Wildix\Colleagues\Colleague::LICENSE_TYPE])) {
                $us[\Wildix\Colleagues\Colleague::LICENSE_TYPE] = \Wildix\License::getInstance()->getMaximumAvailableLicenseType();
            }
            return $us;
        }
        return false;
    }

    public function getUserByID($uid, Colleague $colleague = null) {
        return $this->getUserByProperty(Colleague::ID, $uid, $colleague);
    }

    public function getUsersByIds(array $uids, Colleague $colleague = null) {
        $filter = array(Colleague::ID => $uids);

        return $this->getUsers($filter, null, null, 0, count($uids), null, null, $colleague);
    }

    public function getUserByLogin($login, Colleague $colleague = null) {
        return $this->getUserByProperty(Colleague::LOGIN, $login, $colleague);
    }

    public function getUserByName($name, Colleague $colleague = null) {
        return $this->getUserByProperty(Colleague::NAME, $name, $colleague);
    }

    public function getUserByEmail($email, Colleague $colleague = null) {
        return $this->getUserByProperty(Colleague::EMAIL, $email, $colleague);
    }

    public function getUserByExtension($extension, Colleague $colleague = null) {
        return $this->getUserByProperty(Colleague::EXTENSION, $extension, $colleague);
    }

    private function getUserByProperty($name, $value, Colleague $colleague = null) {
        $filter[$name] = $value;
        $user          = $this->getUsers($filter, null, null, 0, 1, null, null, $colleague);
        if ($user && !empty($user)) {
            return $user[0];
        }
        return null;
    }

    public function getUsersByDepartmentAndPbx($department, $pbxDn, Colleague $colleague = null, $limit = null, $sort = Colleague::NAME) {
        if (empty($department)) {
            return array();
        }
        $filter[Colleague::DEPARTMENT] = $department;
        $filter[Colleague::PBX_DN]     = $pbxDn;

        return $this->getUsers($filter, null, null, null, $limit, $sort, null, $colleague);
    }

    public function getUsersByGroup(\Wildix\Colleagues\Group $group) {
        $filter[Colleague::GROUP_DN] = $group->getDn();
        return $this->getUsers($filter);
    }

    public function getUsersByGroupAndPbx(\Wildix\Colleagues\Group $group, $pbxDn, $limit = null, $sort = Colleague::NAME) {
        $filter[Colleague::GROUP_DN] = $group->getDn();
        $filter[Colleague::PBX_DN]   = $pbxDn;

        return $this->getUsers($filter, null, null, null, $limit, $sort);
    }

    public function getUserByLdapAttributes($props) {
        $set = array();
        foreach ($this->getLdapKeys() as $key) {
            $set[$this->getRevertKey($key)] = $props[$key] ? $props[$key] : '';
        }
        return new Colleague($set);
    }

    /**
     * Get user data by extension or admin login
     *
     * @param string $term User extension or login.
     *
     * @return array|null
     */
    public function getUserByExtensionOrAdmin(string $term): ?array {
        $filter  = '(&(!(userKind=room))(!(userKind=park_orbit))';
        $filter .= $term === 'admin' ? '(cn=%s))' : '(telephoneNumber=%s))';
        $filter  = sprintf($filter, $term);

        $data = $this->getUserByFilter($filter);

        return $data !== false ? $data : null;
    }

    private $_total = 0;
    public function getTotalLastQuery() {
        return $this->_total;
    }

    public function getUsers(
        $filter = null,
        $search = null,
        $searchFields = null,
        $start = null,
        $count = null,
        $sort = null,
        $dir = null,
        Colleague $colleague = null,
        $searchStrategy = self::SEARCH_STRATEGY_CONTAIN
    ) {
        $start          = $start ? $start : 0;
        $count          = ($count && $count < 16001) ? $count : 1500;
        $dir            = $dir ? strtolower($dir) : 'asc';
        $defaultFields  = [
            Colleague::NAME,
            Colleague::EXTENSION,
            Colleague::OFFICE_PHONE,
            Colleague::MOBILE_PHONE,
            Colleague::FAX_NUMBER,
            Colleague::EMAIL,
        ];
        $searchFields   = is_array($searchFields) ? $searchFields : $defaultFields;

        if ($start < 0) {
            throw new Exception('Parameter "start" must be greater or equal than 0', 400);
        }

        if ($count < 0) {
            throw new Exception('Parameter "count" must be greater or equal than 0', 400);
        }

        if ($dir != 'desc' && $dir != 'asc') {
            throw new Exception('Parameter "dir" must be "asc" or "desc"', 400);
        }

        if (! in_array($searchStrategy, [self::SEARCH_STRATEGY_CONTAIN, self::SEARCH_STRATEGY_STARTS_WITH], true)) {
            $exceptionMessage = sprintf('Parameter "searchStrategy" must be "%s" or "%s"', self::SEARCH_STRATEGY_CONTAIN, self::SEARCH_STRATEGY_STARTS_WITH);
            throw new Exception($exceptionMessage, 400);
        }

        $users = [];
        try {
            if ($colleague && empty($filter[Colleague::GROUP_DN])) {
                $groups = Groups::getInstance()->getGroups($colleague);
                if (empty($groups)) {
                    return $users;
                }
                foreach ($groups as $group) {
                    $filter[Colleague::GROUP_DN][] = $group->getDn();
                }
            }

            $ldapFilter = '';

            if ($sort) {
                $sort = $this->getLdapKey($sort);
            }

            if ($filter) {
                foreach ($filter as $key => $val) {
                    if (is_array($val)) {
                        $ldapFilter .= '(|(' . $this->getLdapKey($key) . '=' . implode(
                                ')(' . $this->getLdapKey($key) . '=',
                                $val
                            ) . '))';
                    } else {
                        $ldapFilter .= '(' . $this->getLdapKey($key) . '=' . $val . ')';
                    }
                }
            }

            if ($search) {
                // if($searchFields[0] == '!'){}
                $ldapFilter .= '(|';
                foreach ($searchFields as $key) {
                    if (is_array($search)) {
                        foreach ($search as $searchValue) {
                            $ldapFilter .= $this->buildSearchQuery($key, $searchValue, $searchStrategy);
                        }
                    } else {
                        $ldapFilter .= $this->buildSearchQuery($key, $search, $searchStrategy);
                    }
                }
                $ldapFilter .= ')';
            }

            if ($ldapFilter != '') {
                $ldapFilter = '(&' . $ldapFilter . ')';
            }
        } catch (\Throwable $exception) {
            throw new Exception(Error::BadRequest, 400);
        }

        $ldapKeys = $this->getLdapKeys();
        $rows     = \WMS_Ldap::getInstance()->getEntriesSortLimit(\WMS_Ldap::PERSONS, $ldapFilter, $ldapKeys, $start, $count, $sort, $dir);
        if ($rows && is_array($rows)) {
            $this->_total = $rows['count'];
            unset($rows['count']);
            $rowsCount = count($rows);
            for ($i = 0; $i < $rowsCount; $i++) {
                if (empty($rows[$i])) {
                    continue;
                }
                if (empty($rows[$i]['pbxowner'])) {
                    $this->getLog()->info(__LINE__, 'Skip user (' . $rows[$i]['dn'] . ') with empty pbxowner');
                    continue;
                }
                $us = array();
                foreach ($ldapKeys as $key) {
                    $us[$this->getRevertKey($key)] = $rows[$i][$key] ? $rows[$i][$key] : '';
                }
                $users[] = new Colleague($us);
            }
        }
        return $users;
    }

    private function buildSearchQuery($key, $search, $searchStrategy = self::SEARCH_STRATEGY_CONTAIN): string {
        // allow to use full wildcard for these parameters
        $ldapSubstringsMatch = [
            Colleague::EXTENSION,
            Colleague::OFFICE_PHONE,
            Colleague::MOBILE_PHONE,
            Colleague::FAX_NUMBER,
            Colleague::EMAIL,
        ];
        // do not allow to use of wildcard for these ldap parameters
        $notWildcardParams = [
            Colleague::GROUP_DN,
            Colleague::PBX_DN,
        ];

        $result = '';
        if ($searchStrategy === self::SEARCH_STRATEGY_STARTS_WITH) {
            $searchExpression = ($key === Colleague::MOBILE_PHONE) ? '+' . trim($search) : $search;

            $result .= '(' . $this->getLdapKey($key) . '=' . $searchExpression . '*)';
            if (in_array($key, $notWildcardParams, true)) {
                $result .= '(' . $this->getLdapKey($key) . '=' . $search . ')';
            } else if (!in_array($key, $ldapSubstringsMatch, true)) {
                $result .= '(' . $this->getLdapKey($key) . '=* ' . $search . '*)';
            }
        } else {
            if (in_array($key, $notWildcardParams, true)) {
                $result .= '(' . $this->getLdapKey($key) . '=' . $search . ')';
            } else {
                $result .= '(' . $this->getLdapKey($key) . '=*' . $search . '*)';
            }
        }
        return $result;
    }

    public function updateUserProperty($prop, Colleague $colleague) {
        $set = array();
        foreach ($prop as $key => $val) {
            $set[$this->getLdapKey($key)] = $val;
        }
        $result = \WMS_Ldap::getInstance()->update($colleague->getDN(), $set);
        if (!is_bool($result)) {
            throw new Exception($result.': '.print_r($prop, true));
        }
        return $result;
    }

    /**
     * Persist a new user into WMS Ldap
     *
     * @param Colleague $colleague A new colleague.
     *
     * @throws InvalidDataException If user data already in use
     * @return void
     */
    public function addUser(Colleague $colleague) {
        $attributes = $this->buildColleagueLdapAttributes($colleague);
        $this->assertAttributesNotInUse($attributes);
        $this->assertCnNotAdmin($attributes['cn']);

        $result = \WMS_Ldap::getInstance()->addUser($attributes);

        if ($result !== true) {
            if (is_string($result)) {
                $this->getLog()->error(__LINE__, $result);
            }

            throw new Exception(Error::InternalServerError);
        }

        $this->clearGlobalRosterCache();
    }

    /**
     * Persist a user after change
     *
     * @param Colleague $colleague          A changed colleague.
     *
     * @throws InvalidDataException If user data already in use
     * @return void
     */
    public function updateUser(Colleague $colleague) {
        $attributes = $this->buildColleagueLdapAttributes($colleague);
        unset($attributes[Colleague::DN]);
        unset($attributes[Colleague::PBX]);
        unset($attributes[Colleague::GROUP_NAME]);

        $this->assertAttributesNotInUse($attributes);
        $this->assertCnAdminIsUsedByAdmin($colleague->getDn(), $colleague->getName());

        $result = \WMS_Ldap::getInstance()->update($colleague->getDn(), $attributes);

        if ($result !== true) {
            if (is_string($result)) {
                $this->getLog()->error(__LINE__, $result);
            }

            throw new Exception(Error::InternalServerError);
        }

        $this->clearGlobalRosterCache();
    }

    public function removeUser(Colleague $colleague, $currentPbxDn) {
        if ($colleague->isRoot()) {
            throw new Exception(Error::AccessError, 403);
        }

        $result = \WMS_Ldap::getInstance()->deleteEntryByDn($colleague->getDn());

        if ($result !== true) {
            if (is_string($result)) {
                $this->getLog()->error(__LINE__, $result);
            }

            throw new Exception($result);
        }

        $this->deleteLdapAttributeMember(\WMS_Ldap::GROUPS, $colleague->getDn());
        $this->deleteLdapAttributeMember(\WMS_Ldap::PBX, $colleague->getDn());

        $this->clearGlobalRosterCache();
        $this->removeUserFiles($colleague, (string)$currentPbxDn);
    }

    /**
     * Build array with Ldap attributes of colleague
     *
     * @param Colleague $colleague A colleague.
     *
     * @return array
     */
    private function buildColleagueLdapAttributes(Colleague $colleague) {
        $attributes = array();

        foreach ($colleague->getPropertys() as $key => $val) {
            $attributes[$this->getLdapKey($key)] = $val;
        }

        $attributes['o'] = \WMS_Ldap::ORGANIZATION_NAME;

        return $attributes;
    }

    /**
     * Check attributes are not in use
     *
     * @param array $attributes Attributes array.
     *
     * @throws InvalidDataException If some of array item already in use
     * @return void
     */
    private function assertAttributesNotInUse(array $attributes) {
        $attributeInUse = $this->isDataInUse(
            $attributes['uid'],
            $attributes[LDAP_EXTEN_NUMBER],
            isset($attributes['userPrincipalName']) ? $attributes['userPrincipalName'] : '',
            isset($attributes['mobile']) ? $attributes['mobile'] : '',
            isset($attributes['mail']) ? $attributes['mail'] : ''
        );

        if ($attributeInUse !== false) {
            throw new InvalidDataException(
                // Error message should be like - Data already in use
                Error::getMessage(Error::MustBeUnique),
                array(
                    $this->getRevertKey($attributeInUse) => array(
                        'code' => Error::MustBeUnique,
                        'reason' => Error::getMessage(Error::MustBeUnique)
                    )
                )
            );
        }
    }

    /**
     * Check New user name not equal "admin".
     *
     * @param string $cn Entered User name.
     *
     * @throws InvalidDataException If user name equal "admin".
     *
     * @return void
     */
    private function assertCnNotAdmin(string $cn) {
        if (mb_strtolower($cn) === 'admin') {
            throw new InvalidDataException(
                Error::getMessage(Error::InvalidData),
                array(
                    $this->getRevertKey('cn') => array(
                        'code' => Error::InvalidData,
                        'reason' => 'You can not add user with name "admin".',
                    )
                )
            );
        }
    }

    /**
     * Check if name(cn) "admin" could be set upped for User.
     *
     * @param string $dn User DN in Ldap.
     * @param string $cn Entered User name.
     *
     * @throws InvalidDataException If user name equal "admin".
     *
     * @return void
     */
    private function assertCnAdminIsUsedByAdmin(string $dn, string $cn) {
        if (mb_strtolower($cn) !== 'admin') {
            return;
        }

        $user          = \WMS_Ldap::getInstance()->getEntryByDn($dn, ['cn']);
        $userCurrentCn = $user['cn'] ?? '';

        if ($userCurrentCn !== 'admin') {
            throw new InvalidDataException(
                Error::getMessage(Error::InvalidData),
                array(
                    $this->getRevertKey('cn') => array(
                        'code' => Error::InvalidData,
                        'reason' => "You can not setup name 'admin' for user.",
                    )
                )
            );
        }
    }

    /**
     * Check cn or extenNumber present
     *
     * @param string $uid               A unique id.
     * @param string $extenNumber       A extension.
     * @param string $userPrincipalName A login.
     * @param string $mobile            A mobile phone.
     * @param string $mail              A mail address.
     *
     * @return mixed string 'extenNumber' or 'userPrincipalName' or 'mobile' or 'mail' if present or boolean false
     * @author Igor Kuts <igor.kuts@wildix.com>
     */
    private function isDataInUse($uid, $extenNumber, $userPrincipalName, $mobile, $mail) {
        $uidExpr     = !empty($uid) ? '(!(uid=' . $uid . '))' : '';
        $extenCnExpr = '(|';

        if ($extenNumber != '') {
            $extenCnExpr .= '(' . LDAP_EXTEN_NUMBER . '=' . $extenNumber . ')';
        }

        if ($userPrincipalName != '') {
            $extenCnExpr .= '(userPrincipalName=' . $userPrincipalName . ')';
        }

        if ($mobile != '') {
            $extenCnExpr .= '(mobile=' . $mobile . ')';
        }

        if ($mail != '') {
            $extenCnExpr .= '(mail=' . $mail . ')';
        }

        $extenCnExpr .= ')';
        $search       = \WMS_Ldap::getInstance()->getEntries(
            \WMS_Ldap::PERSONS,
            $uidExpr . $extenCnExpr,
            array('cn', LDAP_EXTEN_NUMBER, 'userPrincipalName', 'mobile', 'mail')
        );

        if ($search !== false) {
            if ($extenNumber == $search[0][LDAP_EXTEN_NUMBER]) {
                return LDAP_EXTEN_NUMBER;
            } else if ($userPrincipalName != '' && $userPrincipalName == $search[0]['userPrincipalName']) {
                return 'userPrincipalName';
            } else if ($mobile != '' && $mobile == $search[0]['mobile']) {
                return 'mobile';
            } else if ($mail != '' && $mail == $search[0]['mail']) {
                return 'mail';
            }
        }

        return false;
    }

    private function clearGlobalRosterCache() {
        if (file_exists(self::FILE_GLOBAL_ROSTER_CACHE)) {
            unlink(self::FILE_GLOBAL_ROSTER_CACHE);
        }
    }

    /**
     * Remove colleague's files
     *
     * @param Colleague $colleague A colleague for remove data.
     *
     * @return void
     */
    public function removeUserAvatar(Colleague $colleague) {
        $colleagueDir  = get_users_settings_dir();
        $avatarFilePNG = $colleagueDir . 'avatars/user_avatar_' . $colleague->getID() . '.png';
        $avatarFileJPG = $colleagueDir . 'user_avatar_' . $colleague->getID() . '.jpg';

        if (file_exists($avatarFilePNG)) {
            unlink($avatarFilePNG);
        } else if (file_exists($avatarFileJPG)) {
            unlink($avatarFileJPG);
        }
    }

    private function removeUserSettingsFile(Colleague $colleague): void {
        $settingsFile = sprintf('%s%s.lua', get_users_settings_dir(), $colleague->getID());

        if (file_exists($settingsFile)) {
            if (!unlink($settingsFile)) {
                $this->getLog()->error(__LINE__, sprintf('Can\'t delete user settings file %s', $settingsFile));
            }
        }
    }

    /**
     * Delete attribute member with specific value from Ldap
     *
     * @param string $ldapObjectClass Ldap object class for remove attribute.
     * @param string $member          Specific member value for delete.
     *
     * @return void
     */
    private function deleteLdapAttributeMember($ldapObjectClass, $member) {
        $memberFilter = '(|(member=' . $member . '))';
        $entries      = \WMS_Ldap::getInstance()->getEntries($ldapObjectClass, $memberFilter, array('dn'));

        if (is_array($entries)) {
            $attributes = array('member' => $member);

            foreach ($entries as $entry) {
                \WMS_Ldap::getInstance()->deleteAttrs($entry['dn'], $attributes);
            }
        }
    }

    /**
     * Removed user files from local or remote PBX.
     *
     * @param Colleague $colleague    A colleague instance which files should be deleted.
     * @param string    $currentPbxDn Dn of current PBX.
     *
     * @throws \Wildix\Services\Wildix\Exception\ServiceException Failed remove a user on a remote PBX.
     * @return void
     */
    private function removeUserFiles(Colleague $colleague, $currentPbxDn) {
        if ($colleague->getPBXDn() === $currentPbxDn) {
            $this->removeUserAvatar($colleague);
            $this->removeUserSettingsFile($colleague);

            return;
        }

        try {
            $this->buildServicePbx($colleague)->removeUser($colleague->getID());
        } catch (\ErrorException $e) {
            $this->getLog()->error(__LINE__, $e->getMessage());
        }
    }

    /**
     * Return PBX service by a colleague
     *
     * @param Colleague $colleague Instance of colleague.
     *
     * @return \Wildix\Services\PBX
     * @throws \ErrorException If PBX not found
     */
    private function buildServicePbx(Colleague $colleague) {
        $pbx = \Wildix\PBXes\PBXes::getInstance()->getPBXByDN($colleague->getPBXDn());

        if (!isset($pbx)) {
            throw new \ErrorException(sprintf('Not found PBX by dn "%s" from Colleague "%d"', $colleague->getPBXDn(), $colleague->getID()));
        }

        return new \Wildix\Services\PBX($pbx);
    }
}
