<?php

namespace Wildix\Colleagues;

/**
 * Class Groups
 */
class Groups extends \Singleton {

    public const FIELD_DN            = 'dn';
    public const FIELD_ID            = 'id';
    public const FIELD_NAME          = 'name';
    public const FIELD_WCGRP         = 'wcgrp';
    public const FIELD_INHERITS      = 'inherits';
    public const FIELD_INHERITS_DN   = 'dn';
    public const FIELD_INHERITS_NAME = 'name';
    public const FIELD_RULES         = 'rules';
    public const FIELD_RULES_RULE    = 'rule';
    public const FIELD_RULES_ABILITY = 'ability';
    public const FIELD_RULES_GROUP   = 'group';

    public const RULES_SET_ABILITY      = 'ability';
    public const RULES_SET_ACCESS       = 'access';
    public const RULES_SET_ACCESS_KEY   = 'key';
    public const RULES_SET_ACCESS_GROUP = 'group';

    public const RULE_ABILITY_OPTIONS_YES  = [
        'yes',
        'no',
    ];
    public const RULE_ABILITY_OPTIONS_SET  = [
        'set',
        'noset',
    ];
    public const RULE_ABILITY_OPTIONS_USE  = [
        'use',
        'nouse',
    ];
    public const RULE_ABILITY_OPTIONS_CALL = [
        'call',
        'nocall',
    ];

    private const NORMALIZED_MAP = [
        self::FIELD_DN => Group::DN,
        self::FIELD_ID => Group::ID,
        self::FIELD_NAME => Group::NAME,
        self::FIELD_WCGRP => Group::WCGRP,
    ];

    private const VALIDATION_RULES_GROUP = [
        'groupName' => self::FIELD_NAME,
        'uniqueLdapProp' => [
            [
                self::FIELD_WCGRP,
                \WMS_Ldap::GROUPS,
                Group::WCGRP,
                self::FIELD_DN,
            ],
            [
                self::FIELD_NAME,
                \WMS_Ldap::GROUPS,
                Group::NAME,
                self::FIELD_DN,
            ],
        ],
        'existsLdapObj' => self::FIELD_INHERITS,
        'integer' => self::FIELD_WCGRP,
        'array' => self::FIELD_RULES,
        'aclToG' => self::FIELD_RULES,
        'required' => self::FIELD_NAME,
    ];

    /**
     * Base LDAP DN.
     *
     * @var string
     */
    private $_baseDn;
    /**
     * LDAP DN of current PBX.
     *
     * @var string
     */
    private $_currentPbxDn;
    /**
     * Connection to LDAP.
     *
     * @var \WMS_Ldap
     */
    private $_ldap;
    /**
     * Colleagues repository.
     *
     * @var \Wildix\Colleagues\Colleagues
     */
    private $_colleagues;
    /**
     * The license service.
     *
     * @var \Wildix\License
     */
    private $_license;
    /**
     * Validator.
     *
     * @var \Wildix\Validator
     */
    private $_validator;

    private $_groups;

    /**
     * @var \WMS_Permissions
     */
    private $_permissions;

    public function setBaseDn(string $baseDn): void {
        $this->_baseDn = $baseDn;
    }

    public function setCurrentPbxDn(string $currentPbxDn): void {
        $this->_currentPbxDn = $currentPbxDn;
    }

    public function setLdap(\WMS_Ldap $ldap): void {
        $this->_ldap = $ldap;
    }

    public function setColleagues(Colleagues $colleagues): void {
        $this->_colleagues = $colleagues;
    }

    public function setLicense(\Wildix\License $license): void {
        $this->_license = $license;
    }

    public function setValidator(\Wildix\Validator $validator): void {
        $this->_validator = $validator;
    }

    public function setPermissions(\WMS_Permissions $permissions): void {
        $this->_permissions = $permissions;
    }

    /**
     * Returns list of all groups.
     *
     * @param Colleague|null $colleague Colleague.
     *
     * @return array|Group[]|null
     */
    public function getGroups(\Wildix\Colleagues\Colleague $colleague = null) {
        if ($this->_groups === null) {
            $this->_groups = array();

            $allGroups = $this->getGroupsByProperty(null);
            if ($colleague) {
                foreach ($allGroups as $key => $group) {
                    if ($colleague->can(Colleague::VIEW, $group->getDn())) {
                        $this->_groups[] = $group;
                    }
                }
            } else {
                $this->_groups = $allGroups;
            }
        }

        return $this->_groups;
    }

    /**
     * Returns list of managed groups for current user.
     *
     * @param array $fields
     *
     * @return array
     */
    public function getManagedGroups(array $fields = []): array {
        $result = [];
        $groups = (array)$this->getGroups();
        foreach ($groups as $group) {
            if (!$this->_getPermissions()->is_allowed('group', $group->getDn())) {
                continue;
            }

            $normalizedData = $this->normalizeGroupData($group->getGroupAsArray());

            if (!empty($fields)) {
                $normalizedData = array_intersect_key($normalizedData, array_flip($fields));
            }

            $result[] = $normalizedData;
        }

        return $result;
    }

    public function getGroupByDn(string $dn, ?array $fields = null): ?Group {
        $group = $this->_getLdap()->getEntryByDn($dn, $fields ?? Group::getFields(), [Group::ACL_WG2G, Group::ACL_WG2E]);

        if ($group && !empty($group)) {
            return new Group($group);
        }

        return null;
    }

    /**
     * Returns second group item in list or a group which name contains string 'default'.
     *
     * @return string
     */
    public function getAvailableGroupDn(): string {
        $groupsList = $this->getGroups();

        if (is_array($groupsList)) {
            foreach ($groupsList as $group) {
                if ((int)$group->getWcgrp() === 2 || (preg_match('/^default.*/i', $group->getName()) === 1)) {
                    return $group->getDn();
                }
            }
        }

        return '';
    }

    public function getGroupById(int $groupId, array $fields): ?Group {
        $data = $this->_getLdap()->getEntry(\WMS_Ldap::GROUPS, '(' . Group::ID . '=' . $groupId . ')', $fields);

        if (is_array($data)) {
            return new Group($data);
        }

        return null;
    }

    /**
     * Normalizes a data of group for cross-application communication.
     *
     * @param array $data A data of group.
     *
     * @return array
     */
    public function normalizeGroupData(array $data): array {
        $inherits = null;
        $rules    = [];

        if (!empty($data[Group::OWNER])) {
            $inherits = [
                self::FIELD_INHERITS_DN => $data[Group::OWNER],
                self::FIELD_INHERITS_NAME => $this->_getGroupName($data[Group::OWNER]),
            ];
        }

        if (!empty($data[Group::ACL_WG2G])) {
            $rules = array_map([__CLASS__, '_destructAcl'], $data[Group::ACL_WG2G]);
        }

        return [
            self::FIELD_DN => $data[self::NORMALIZED_MAP[self::FIELD_DN]] ?? '',
            self::FIELD_ID => $data[self::NORMALIZED_MAP[self::FIELD_ID]] ?? '',
            self::FIELD_NAME => $data[self::NORMALIZED_MAP[self::FIELD_NAME]] ?? '',
            self::FIELD_WCGRP => $data[self::NORMALIZED_MAP[self::FIELD_WCGRP]] ?? '',
            self::FIELD_INHERITS => $inherits,
            self::FIELD_RULES => $rules,
        ];
    }

    /**
     * Adds a data of a new group or updates a data of exists group if the data contains property "dn".
     *
     * To unset properties "wcgrp", "inherits" or "rules" they should be an empty string,
     * other ways if they were not set nosing happened.
     *
     * @param array $data A data of group.
     *
     * @return string
     * @throws \Wildix\InvalidDataException if the data is invalid.
     */
    public function saveGroup(array $data): string {
        $validator = $this->_validator->withData($data);
        $validator->rules(self::VALIDATION_RULES_GROUP);

        if (!$validator->validate()) {
            throw new \Wildix\InvalidDataException(\Wildix\Error::getMessage(\Wildix\Error::InvalidData), $validator->errors());
        }

        $emptyData = [
            self::FIELD_DN => null,
            self::FIELD_NAME => null,
            self::FIELD_WCGRP => null,
            self::FIELD_INHERITS => null,
            self::FIELD_RULES => null,
        ];
        [
            self::FIELD_DN => $dn,
            self::FIELD_NAME => $name,
            self::FIELD_WCGRP => $wcgrp,
            self::FIELD_INHERITS => $inherits,
            self::FIELD_RULES => $rules,
        ]          = array_merge($emptyData, $data);

        $wcgrp    = isset($wcgrp) ? (int)$wcgrp : null;
        $owner    = $inherits['dn'] ?? $inherits ?? null;
        $aclWgToG = isset($rules) ? $this->_buildAclWgToG($rules) : null;

        if (!isset($dn)) {
            $dn = $this->_create($name, $wcgrp, $owner, $aclWgToG);
        } else {
            $this->_update($dn, $name, $wcgrp, $owner, $aclWgToG);
        }

        return $dn;
    }

    public function removeGroup(\Wildix\Colleagues\Group $group): void {
        $users = $this->_colleagues->getUsersByGroup($group);

        if (count($users) > 0) {
            throw new \Wildix\Exception(\Wildix\Error::ACL_GROUP_IS_NOT_EMPTY, 405);
        }

        $groupDn        = $group->getDn();
        $childrenGroups = $this->_getChildren($groupDn);

        foreach ($childrenGroups as $child) {
            $this->_unsetOwner($child->getDn(), $groupDn);
        }

        $this->_delete($groupDn);
    }

    /**
     * Returns a list available ACL rules for groups.
     *
     * @return array
     */
    public function getRules(): array {
        $rulesSetCanCannotToGroup = $this->_buildRulesSet(\Wildix\Colleagues\Colleague::CAN_CANNOT_TO_GROUP, true);
        $rulesSetCanCannotCommon  = $this->_buildRulesSet(\Wildix\Colleagues\Colleague::CAN_CANNOT_COMMON, false);
        $rulesSetCanCannotSet     = $this->_buildRulesSet(\Wildix\Colleagues\Colleague::CAN_CANNOT_SET, false);
        $rulesSetCanCannotUse     = $this->_buildRulesSet(\Wildix\Colleagues\Colleague::CAN_CANNOT_USE, false);
        $rulesSetCanCannotCall    = $this->_buildRulesSet(\Wildix\Colleagues\Colleague::CAN_CANNOT_CALL, false);

        return [
            [
                self::RULES_SET_ABILITY => self::RULE_ABILITY_OPTIONS_YES,
                self::RULES_SET_ACCESS => array_merge($rulesSetCanCannotCommon, $rulesSetCanCannotToGroup),
            ],
            [
                self::RULES_SET_ABILITY => self::RULE_ABILITY_OPTIONS_SET,
                self::RULES_SET_ACCESS => $rulesSetCanCannotSet,
            ],
            [
                self::RULES_SET_ABILITY => self::RULE_ABILITY_OPTIONS_USE,
                self::RULES_SET_ACCESS => $rulesSetCanCannotUse,
            ],
            [
                self::RULES_SET_ABILITY => self::RULE_ABILITY_OPTIONS_CALL,
                self::RULES_SET_ACCESS => $rulesSetCanCannotCall,
            ],
        ];
    }

    /**
     * Returns a default rules for groups.
     *
     * @param boolean $destruct Indicates that result items should be destructed into arrays. Default FALSE.
     *
     * @return array
     */
    public function getDefault(bool $destruct = false): array {
        $isHotel = $this->_license->haveLicense(\Wildix\License::LICENSE_HOTEL);

        $aclRulesForHotels      = ['018|all|ACCESS_ALARM_CALL_OF|no', '019|all|HOTEL_ADMIN|yes', '020|all|HOTEL_EVENTS|yes'];
        $aclRulesForConferences = [($isHotel ? '021' : '018') . '|all|CREATE_CONFERENCE|no'];

        $default = array_merge(
            [
                '001|all|SUPERVISION|no',
                '002|all|INTERCOM|no',
                '003||MANAGE_CALLCENTER|no',
                '004||CTI_CDR_VIEW|nouse',
                '005|all|SEE_VOICEMAIL|no',
                '006||PBXRECORD|nouse',
                '007||PERSONALRECORD|nouse',
                '008||all|nocall',
                '009||2|call',
                '010||3|call',
                '011||4|call',
                '012||5|call',
                '013||13|call',
                '014||14|call',
                '015|all|EDIT_PP|no',
                '016|all|DELETE_CALLS|no',
                '017|all|CHANGE_PRESENCE|no',
            ],
            $isHotel ? $aclRulesForHotels : [],
            $this->_license->haveLicense(\Wildix\License::LICENSE_CONFERENCE) ? $aclRulesForConferences : []
        );

        if ($destruct) {
            $default = array_map([__CLASS__, '_destructAcl'], $default);
        }

        return $default;
    }

    private static function _destructAcl(string $acl): array {
        [, $group, $rule, $ability] = explode('|', $acl);

        return [
            self::FIELD_RULES_RULE => $rule,
            self::FIELD_RULES_ABILITY => $ability,
            self::FIELD_RULES_GROUP => $group,
        ];
    }

    /**
     * Creates a group with given properties and stores it in LDAP.
     *
     * @param string       $name     The group name.
     * @param integer|null $wcgrp    The group WCGRP.
     * @param string|null  $owner    The group owner "dn".
     * @param array|null   $aclWgToG The group ACL rules.
     *
     * @return string The new group "dn".
     * @throws \RuntimeException if failed add the group to LDAP.
     */
    private function _create(string $name, ?int $wcgrp, ?string $owner, ?array $aclWgToG): string {
        $organization = 'com0';
        $attributes   = [
            Group::DN => sprintf('gid=%s,o=%s,%s', random_int(1000000, 10000000), $organization, $this->_baseDn),
            Group::ORGANIZATION => $organization,
            Group::NAME => $name,
            Group::WCGRP => $wcgrp,
            Group::OWNER => $owner,
            Group::ACL_WG2E => [
                '001|all|noepbx',
                '002|' . $this->_currentPbxDn . '|epbx',
                '003|all|noegroup',
                '004|add_remove_users|noeperm',
                '005|all|noepage',
                '006|users__phonebooks3|epage',
                '007|dialplan__call_groups|epage',
                '008|dialplan__timetables|epage',
                '009|dialplan__ivr|epage',
                '010|settings__utilites__backup|epage',
            ],
        ];

        if (!isset($aclWgToG)) {
            $attributes[Group::ACL_WG2G] = $this->getDefault();
        } else if (count($aclWgToG) > 0) {
            $attributes[Group::ACL_WG2G] = $aclWgToG;
        }

        $result = $this->_getLdap()->addGroup($attributes);

        if ($result !== true) {
            throw new \RuntimeException(
                sprintf(
                    'Failed add the Group [name: %s | wcgrp: %d | owner: %s] to LDAP. Error: %s',
                    $name,
                    $wcgrp,
                    $owner,
                    (is_string($result) ? $result : 'Unknown')
                )
            );
        }

        $this->_removeCacheFiles();

        return $attributes[Group::DN];
    }

    /**
     * Changed a group name, wcgrp, owner and wg2gAcl properties.
     *
     * @param string       $dn       A group 'dn' for update.
     * @param string|null  $name     A group new name if NULL is not be changed.
     * @param integer|null $wcgrp    A group new WCGRP if NULL is not changed. To erase should be zero.
     * @param string|null  $owner    A group new owner 'dn' (a parent group) if NULL is not changed. To erase should be an empty string.
     * @param array|null   $aclWgToG A group new ACL rules if NULL is not changed. To erase should be an empty array.
     *
     * @return void
     * @throws \RuntimeException if failed store a new data in LDAP.
     */
    private function _update(string $dn, ?string $name, ?int $wcgrp, ?string $owner, ?array $aclWgToG): void {
        $data       = $this->_getLdap()->getEntryByDn($dn, [Group::NAME, Group::OWNER, Group::WCGRP, Group::ACL_WG2G], [Group::ACL_WG2G]);
        $attributes = [
            Group::NAME => $name ?? $data[Group::NAME],
            Group::OWNER => $owner ?? ($data[Group::OWNER] ?? ''),
            Group::WCGRP => $wcgrp ?? $data[Group::WCGRP],
            Group::ACL_WG2G => $aclWgToG ?? $data[Group::ACL_WG2G],
        ];

        $result = $this->_getLdap()->update($dn, $attributes);

        if ($result !== true) {
            throw new \RuntimeException(
                sprintf(
                    'Failed update the Group %s [name: %s | wcgrp: %d | owner: %s] to LDAP. Error: %s',
                    $dn,
                    $name,
                    $wcgrp,
                    $owner,
                    (is_string($result) ? $result : 'Unknown')
                )
            );
        }

        if ($data[Group::NAME] !== $name) {
            // Modify user rosters
            $userFeaturesSyncArgs = escapeshellarg(
                json_encode(
                    [
                        'type' => 'roster',
                        'action' => 'modify',
                        'parameters' => [$data[Group::NAME], $name],
                    ]
                )
            );

            exec('/usr/sbin/user_features_sync ' . $userFeaturesSyncArgs);
        }

        $this->_removeCacheFiles();
    }

    /**
     * Returns all groups contain specified 'dn' as owner property.
     *
     * @param string $groupDn A group 'dn'.
     *
     * @return array
     */
    private function _getChildren(string $groupDn): array {
        $groups     = [];
        $dataGroups = $this->_getLdap()->getEntries(\WMS_Ldap::GROUPS, '(owner=' . $groupDn . ')', [Group::DN]);

        if (!is_array($dataGroups)) {
            return $groups;
        }

        foreach ($dataGroups as $dataGroup) {
            $groups[] = new Group($dataGroup);
        }

        return $groups;
    }

    /**
     * Removes the owner property from a group by it 'dn'.
     *
     * @param string $dn    A target group 'dn'.
     * @param string $owner A value of the owner property to removed.
     *
     * @return void
     * @throws \RuntimeException if failed delete the property from LDAP.
     */
    private function _unsetOwner(string $dn, string $owner): void {
        $result = $this->_getLdap()->deleteAttrs($dn, [Group::OWNER => $owner]);

        if ($result !== true) {
            throw new \RuntimeException($result);
        }
    }

    private function _delete(string $dn): void {
        $result = $this->_getLdap()->deleteEntryByDn($dn);

        if ($result !== true) {
            throw new \RuntimeException($result);
        }

        $this->_removeCacheFiles();
    }

    private function _getGroupName(string $dn): ?string {
        static $groupsMap;

        if (!isset($groupsMap)) {
            $groups = (array)$this->getGroups();

            foreach ($groups as $group) {
                $groupsMap[$group->getDn()] = $group->getName();
            }
        }

        return $groupsMap[$dn] ?? null;
    }

    private function _buildAclWgToG(array $rules): array {
        $result = [];

        usort($rules, [__CLASS__, '_compareAcl']);

        foreach ($rules as $index => $rule) {
            if (in_array($rule[self::FIELD_RULES_RULE], \Wildix\Colleagues\Colleague::CAN_CANNOT_COMMON)) {
                $rule[self::FIELD_RULES_GROUP] = 'all';
            }

            $normalizedIndex  = str_pad((string)($index + 1), 3, '0', STR_PAD_LEFT);
            $ruleWithoutIndex = $rule[self::FIELD_RULES_GROUP] . '|' . $rule[self::FIELD_RULES_RULE] . '|' . $rule[self::FIELD_RULES_ABILITY];

            $result[$ruleWithoutIndex] = $normalizedIndex . '|' . $ruleWithoutIndex;
        }

        return array_values($result);
    }

    private static function _compareAcl(array $ruleA, array $ruleB): int {
        $calculateAbilityWeight = static function (string $ability): int {
            if (in_array($ability, ['yes', 'no'], true)) {
                return 1;
            }

            if (in_array($ability, ['use', 'nouse'], true)) {
                return 2;
            }

            if (in_array($ability, ['set', 'noset'], true)) {
                return 3;
            }

            return 4;
        };
        $compareGroups          = static function (string $groupA, string $groupB): int {
            if ($groupA === $groupB || ($groupA !== 'all' && $groupB !== 'all')) {
                return 0;
            }

            return ($groupA === 'all') ? -1 : 1;
        };
        $compareName            = static function (string $nameA, string $nameB): int {
            // Name 'all' can relate only to call rules
            if ($nameA === $nameB || ($nameA !== 'all' && $nameB !== 'all')) {
                return 0;
            }

            return ($nameA === 'all') ? -1 : 1;
        };

        $weightRuleA = $calculateAbilityWeight($ruleA[self::FIELD_RULES_ABILITY]);
        $weightRuleB = $calculateAbilityWeight($ruleB[self::FIELD_RULES_ABILITY]);

        if ($weightRuleA !== $weightRuleB) {
            return ($weightRuleA < $weightRuleB) ? -1 : 1;
        }

        if ($weightRuleA === 1) {
            // no/yes
            return $compareGroups($ruleA[self::FIELD_RULES_GROUP], $ruleB[self::FIELD_RULES_GROUP]);
        }

        // all others
        return $compareName($ruleA[self::FIELD_RULES_RULE], $ruleB[self::FIELD_RULES_RULE]);
    }

    private function _buildRulesSet(array $items, bool $relatesToGroup): array {
        $set = [];

        foreach ($items as $item) {
            $set[] = [
                self::RULES_SET_ACCESS_KEY => $item,
                self::RULES_SET_ACCESS_GROUP => $relatesToGroup,
            ];
        }

        return $set;
    }

    private function _removeCacheFiles(): void {
        exec('rm /dev/shm/wg*.acl');

        // Remove global roster cache file
        if (file_exists(Colleagues::FILE_GLOBAL_ROSTER_CACHE)) {
            unlink(Colleagues::FILE_GLOBAL_ROSTER_CACHE);
        }
    }

    private function getGroupsByProperty($name, $value = '') {
        $filter = '';
        if (is_array($value)) {
            $filter = "(|({$name}=" . join(")({$name}=", $value) . '))';
        } else if (!empty($name)) {
            $filter = "({$name}={$value})";
        }

        $ldapGroups = $this->_getLdap()->getEntries(
            \WMS_Ldap::GROUPS,
            $filter,
            [Group::DN, Group::ORGANIZATION, Group::ID, Group::NAME, Group::WCGRP, Group::OWNER, Group::ACL_WG2G, Group::ACL_WG2E],
            [Group::ACL_WG2G, Group::ACL_WG2E]
        );

        if ($ldapGroups && !empty($ldapGroups) && is_array($ldapGroups)) {
            $groups = array();
            foreach ($ldapGroups as $key => $group) {
                $groups[] = new Group($group);
            }

            return $groups;
        }

        return null;
    }

    private function _getLdap(): \WMS_Ldap {
        return $this->_ldap ?? \WMS_Ldap::getInstance();
    }

    private function _getPermissions(): \WMS_Permissions {
        if (is_null($this->_permissions)) {
            throw new \RuntimeException('Not found permission class');
        }
        return $this->_permissions;
    }
}
