<?php
namespace Wildix\Colleagues;

use Api\Dictionary\Services\DictionaryService;
use Api\Trunks\Fxo\Interfaces\FxoTrunkServiceInterface;
use Api\Trunks\Pstn\Interfaces\PstnTrunkServiceInterface;
use Api\Trunks\Sip\Interfaces\SipTrunkServiceInterface;
use Api\Trunks\Sip\Objects\Trunk as SipTrunk;
use Services\LuaService;
use Wildix\CallweaverManager;
use Wildix\ContactCenter\ContactCenter;
use Wildix\DependencyInjection\Container;
use Wildix\Dialplan\CallGroups\CallGroup;
use Wildix\Dialplan\CallGroups\CallGroupsService;
use Wildix\Dialplan\IncomingCalls;
use Wildix\Ejabberd;

/**
 * Settings
 */
class Settings {

    private $_settings     = null;
    private $_diffSettings = array();

    private $_colleague;

    private $_luaService;

    private $_container;

    private $_allowedCallGroups;
    private $_staticCallGroups;
    private $_dynamicCallGroups;

    public function __construct(Colleague $colleague) {
        $this->_colleague = $colleague;
        // global implicit dependency !!!
        // @TODO rewrite
        $this->_container = new Container();

        $this->getSettings();
    }

    /**
     * GetInternalClassFeatures
     *
     * @return array
     */
    public function getInternalClassFeatures() {
        return $this->getSettings('internal');
    }

    /**
     * SetInternalClassFeatures
     *
     * @param array $features Internal features.
     *
     * @return void
     */
    public function setInternalClassFeatures(array $features) {
        $this->setClassFeatures('internal', $features);
    }

    public function getExternalClassFeatures() {
        return $this->getSettings('external');
    }

    public function setExternalClassFeatures(array $features) {
        $this->setClassFeatures('external', $features);
    }

    public function getBlacklistClassFeatures() {
        return $this->getSettings('blacklist');
    }

    public function setBlacklistClassFeatures(array $features) {
        $this->setClassFeatures('blacklist', $features);
    }

    public function getWhitelistClassFeatures() {
        return $this->getSettings('whitelist');
    }

    public function setWhitelistClassFeatures(array $features) {
        $this->setClassFeatures('whitelist', $features);
    }

    public function getOptions() {
        return base64_decode($this->getSettings('CTIOPTIONS'));
    }

    public function getStatus() {
        $status          = $this->getSettings('STATUS');
        $location        = explode('],[', base64_decode($status['location']));
        $decodedLocation = array(
                'address' => '',
                'lat'     => '',
                'lng'     => ''
        );

        $trans = array(
                ']' => '',
                '[' => ''
        );
        if (!empty($location[0])) {
            $decodedLocation['address'] = strtr($location[0], $trans);
        }
        if (!empty($location[1])) {
            $decodedLocation['lat'] = strtr($location[1], $trans);
        }
        if (!empty($location[2])) {
            $decodedLocation['lng'] = strtr($location[2], $trans);
        }

        foreach ($decodedLocation as $key => $value) {
            $splited = explode(',', $decodedLocation[$key]);
            $text    = '';
            foreach ($splited as $val) {
                $text .= chr(intval($val));
            }
            $decodedLocation[$key] = $text;
        }


        return array(
            'status' => ($status['value'] === '') ? 'available' : $status['value'],
            'message' => base64_decode($status['message']),
            'until' => base64_decode($status['untilTime']),
            'location' => $decodedLocation
        );
    }

    public function getCalendar() {
        return $this->getSettings('CALENDAR');
    }

    public function setCalendar(array $calendar) {
        $this->_settings['CALENDAR']     = $calendar;
        $this->_diffSettings['CALENDAR'] = $calendar;
    }

    public function getWebApp() {
        return $this->getSettings('WEBAPP');
    }

    public function getPhonebooks() {
        $phoneBooks = str_replace('new.', '', $this->_settings['PHONEBOOK']);
        return array_filter(array_map('trim', explode(',', $phoneBooks)), 'strlen');
    }

    public function getRoster() {
        return $this->getSettings('ROSTER');
    }

    public function setRoster(array $newRoster) {
        $this->_settings['ROSTER']     = $newRoster;
        $this->_diffSettings['ROSTER'] = $newRoster;
    }


    public function getSharedVoiceMail() {
        return $this->getSettings('SHAREDVM');
    }

    public function getCallWaitingTone() {
        return $this->getSettings('CWTONE');
    }


    public function getTwoFactor() {
        return $this->getSettings('TWOFACTOR');
    }

    public function getCallCenter() {
        $callCenterSettings = $this->getSettings('CALLCENTER');

        unset($callCenterSettings['CALLCENTER_MODE']);

        $callCenterSettings['enable']     = isset($callCenterSettings['enable']) && $callCenterSettings['enable'] === '1';
        $callCenterSettings['statusSync'] = isset($callCenterSettings['statusSync']) && $callCenterSettings['statusSync'] === '1';

        $callCenterSettings['callGroups'] = $this->buildCallCenterCallGroups();

        return $callCenterSettings;
    }

    public function setCallCenter(array $settings): void {
        $settings = $this->getPreparedSettings($settings);

        if (count(array_diff_assoc_recursive($settings, $this->_settings['CALLCENTER']))) {
            $this->_diffSettings['CALLCENTER'] = $settings;
        }

        unset($this->_settings['CALLCENTER']['callGroups']);
        $this->_settings['CALLCENTER'] = array_merge($this->_settings['CALLCENTER'], $settings);
    }

    public function setTwoFactor($new) {
        $this->_settings['TWOFACTOR']     = $new;
        $this->_diffSettings['TWOFACTOR'] = $new;
    }

    /**
     * SetClassFeatures
     *
     * @param string $className ClassName.
     * @param array  $features  Features.
     *
     * @return void
     */
    public function setClassFeatures($className, array $features) {
        $newFeatures = array_merge_recursive_distinct($this->_settings[$className], $features);
        $diff        = array_diff_assoc_recursive($newFeatures, $this->_settings[$className]);

        if (!empty($diff)) {
            if (!$this->_diffSettings[$className]) {
                $this->_diffSettings[$className] = array();
            }
            $this->_diffSettings[$className] = array_merge_recursive_distinct($this->_diffSettings[$className], $diff);
        }
        $this->_settings[$className] = $newFeatures;
    }

    public function getDiffSettings() {
        return $this->_diffSettings;
    }

    public function getFunctionKeys() {
        $fkeys = $this->getSettings('FKEYS');
        if (!empty($fkeys)) {
            $fkeys = explode(',', $fkeys);
            foreach ($fkeys as $key => $value) {
                $fkey = explode(' ', $value);
                if (strpos($fkey[0], '---') === 0) {
                    $fkeys[$key] = array(
                            'key' => $fkey[1],
                            'value' => $fkey[2],
                            'title' => ''
                    );
                } else {
                    $fkeys[$key] = array(
                            'key' => $fkey[0],
                            'value' => ((!empty($fkey[1])) ? $fkey[1] : '')
                    );
                    if ($fkey[0] != 'QCALL') {
                        $fkeys[$key]['title'] = (!empty($fkey[2])) ? $fkey[2] : '';
                    } else {
                        $fkeys[$key]['title'] = '';
                    }

                    if (in_array($fkey[0], array('EXT', 'PORBIT'))) {
                        $fkeys[$key]['viewCalls']         = isset($fkey[5]) ? $fkey[5] : '';
                        $fkeys[$key]['audioNotification'] = isset($fkey[4]) ? $fkey[4] : '';
                        $fkeys[$key]['directTransfer']    = isset($fkey[3]) ? $fkey[3] : '';
                    }

                    if ($fkey[0] === 'SD') {
                        $fkeys[$key]['directTransfer'] = isset($fkey[3]) ? $fkey[3] : '';
                    }
                    if ($fkey[0] === 'QCALL') {
                        $fkeys[$key]['viewCalls']         = isset($fkey[4]) ? $fkey[4] : '';
                        $fkeys[$key]['audioNotification'] = isset($fkey[3]) ? $fkey[3] : '';
                        $fkeys[$key]['directTransfer']    = isset($fkey[2]) ? $fkey[2] : '';
                    }
                }
            }
        }
        return $fkeys;
    }

    public function save() {
        if (!empty($this->_diffSettings)) {
            if (isset($this->_settings['POPUPURL']) && !empty($this->_settings['POPUPURL']['url'])) {
                $this->_settings['POPUPURL']['url'] = htmlspecialchars($this->_settings['POPUPURL']['url'], ENT_QUOTES);
            }

            if ($this->_diffSettings['CALLCENTER'] && $this->_colleague->getExtension()) {
                $contactCenter = new ContactCenter(CallweaverManager::getInstance());

                if ($this->_settings['CALLCENTER']['enable'] == '1') {
                    $contactCenter->enable($this->_colleague->getExtension(), $this->_settings['CALLCENTER']['callGroups'], 0);
                } else if ($this->_settings['CALLCENTER']['enable'] == '0') {
                    $contactCenter->disable($this->_colleague->getExtension());
                }

                Ejabberd::getInstance()->setCallcenterMode($this->_colleague->getExtension(), $this->_settings['CALLCENTER']['statusSync']);
            }

            if ($this->getLuaService()->saveUserFeatures($this->_colleague->getID(), $this->_settings)) {
                $this->_diffSettings = array();
                return true;
            }
            return false;
        }
        return true;
    }

    public function getSettings($key = null) {
        if (!$this->_settings) {
            $this->_settings = $this->getLuaService()->getUserFeatures($this->_colleague->getID());
        }

        if (isset($this->_settings['POPUPURL']) && !empty($this->_settings['POPUPURL']['url'])) {
            $this->_settings['POPUPURL']['url'] = htmlspecialchars_decode($this->_settings['POPUPURL']['url'], ENT_QUOTES);
        }

        if ($key !== null) {
            return $this->_settings[$key];
        }
        return $this->_settings;
    }

    public function groupByFields(array $fields, $isOldVersion = false) {
        $result = array();
        $need   = array_flip($fields);

        if (isset($need['personal'])) {
            $result['personal'] = array(
                'email' => $this->_colleague->getEmail(),
                'mobilePhone' => $this->_colleague->getMobilePhone(),
            );
        }

        if (isset($need['ringtones'])) {
            $result['ringtones'] = \Wildix\Sounds\SoundsService::getRingTones();
        }

        if (isset($need['fkeys'])) {
            $result['fkeys'] = array(
                'keys' => $this->getFunctionKeys(),
                'list' => $this->getFunctionKeysList(),
            );
        }

        if (isset($need['features'])) {
            $result['features'] = array(
                'internal' => $this->getInternalClassFeatures(),
                'external' => $this->getExternalClassFeatures(),
                'blacklist' => $this->getBlacklistClassFeatures(),
                'whitelist' => $this->getWhitelistClassFeatures(),
            );

            if ($isOldVersion) {
                foreach ($result['features'] as $key => $val) {
                    $result['features'][$key] = $val['default'];
                }
            }
        }

        if (isset($need['sharedvm'])) {
            $result['sharedvm'] = $this->getSharedVoiceMail();
        }

        if (isset($need['cwtone'])) {
            $result['cwtone'] = $this->getCallWaitingTone();
        }

        if (isset($need['contactCenter'])) {
            $result['contactCenter'] = $this->getCallCenter();
        }

        return $result;
    }

    public function getFunctionKeysList() {
        $dialplanClass = $this->_container->get(IncomingCalls::class);

        $result = array(
            'STIME' => $dialplanClass->selectfrom_timetables(),
            'TRUNK' => array(),
            'SWITCH' => $dialplanClass->selectfrom_wmsswitches(),
            '3SWITCH' => $dialplanClass->selectfrom_wms3switches(),
            'GPICKUP' => array(),
            'QPICKUP' => array(),
            'QCALL' => array(),
        );

        // Groups
        $groups = \Wildix\Colleagues\Groups::getInstance()->getGroups($this->_colleague);

        foreach ($groups as $key => $group) {
            $result['GPICKUP'][$group->getWcgrp()] = $group->getName();
        }

        // Call groups
        foreach ($dialplanClass->getAllCallGroups() as $record) {
            $result['QPICKUP'][$record['cgroup_id']] = $record['call_groupname'];
        }

        $result['QCALL'] = $result['QPICKUP'];

        // TRUNKS
        $trunkServices = [
            $this->_container->get(SipTrunkServiceInterface::class),
            $this->_container->get(PstnTrunkServiceInterface::class),
            $this->_container->get(FxoTrunkServiceInterface::class),
        ];

        foreach ($trunkServices as $service) {
            foreach ($service->getItems() as $trunk) {
                if ($trunk instanceof SipTrunk) {
                    $result['TRUNK'][] = $trunk->getUsername();
                } else {
                    array_push($result['TRUNK'], ...$trunk->getAuthNamesByPorts());
                }
            }
        }

        return $result;
    }

    private function getLuaService(): LuaService {
        if (!$this->_luaService) {
            $this->_luaService = new LuaService();
        }
        return $this->_luaService;
    }

    private function getCallGroupsService(): CallGroupsService {
        return $this->_container->get(CallGroupsService::class);
    }

    private function getStaticCallGroups(): array {
        if (!$this->_colleague->getExtension()) {
            $this->_staticCallGroups = [];
        }

        if ($this->_staticCallGroups === null) {
            $this->_staticCallGroups = array_map(
                function (CallGroup $callGroup) {
                    return $callGroup->getId();
                },
                $this->getCallGroupsService()->getItemsByExtension($this->_colleague->getExtension())
            );
        }

        return $this->_staticCallGroups;
    }

    private function getDictionaryService(): DictionaryService {
        return $this->_container->get(DictionaryService::class);
    }

    private function getAllowedCallGroups(): array {
        if ($this->_allowedCallGroups === null) {
            $callCenterWhitelistSettings = $this->getSettings('CALLCENTER_WHITELIST');

            if ($callCenterWhitelistSettings['enable'] === '1') {
                $this->_allowedCallGroups = array_map(
                    function ($callGroup) {
                        return (int)$callGroup['id'];
                    },
                    $callCenterWhitelistSettings['callGroups']
                );

                $this->_allowedCallGroups = array_merge($this->_allowedCallGroups, $this->getStaticCallGroups());
            } else {
                $this->_allowedCallGroups = $this->getDictionaryService()
                    ->getCallGroups()
                    ->map(
                        function ($callGroup) {
                            return $callGroup->getValue();
                        }
                    )
                    ->toArray();
            }
        }

        return $this->_allowedCallGroups;
    }

    private function getPreparedSettings(array $settings): array {
        if (isset($settings['callGroups']) && $settings['callGroups']) {
            foreach ($settings['callGroups'] as $key => $callGroup) {
                if ($callGroup['id']
                    && ($this->isStaticCallGroup($callGroup['id'])
                    || !$this->isAllowedCallGroup($callGroup['id'])
                    || !$callGroup['active'])
                ) {
                    unset($settings['callGroups'][$key]);
                    continue;
                }

                $settings['callGroups'][$key] = ['id' => (string)$callGroup['id']];
            }

            if (count($settings['callGroups'])) {
                $settings['callGroups'] = array_combine(
                    range(1, count($settings['callGroups'])),
                    $settings['callGroups']
                );
            }
        }

        return $settings;
    }

    private function getDynamicCallGroups(): array {
        if ($this->_dynamicCallGroups === null) {
            $callCenterSettings = $this->getSettings('CALLCENTER');

            $this->_dynamicCallGroups = array_map(
                function ($callGroup) {
                    return (int)$callGroup['id'];
                },
                $callCenterSettings['callGroups']
            );
        }

        return $this->_dynamicCallGroups;
    }

    private function isDynamicCallGroup(int $callGroupId): bool {
        return in_array($callGroupId, $this->getDynamicCallGroups(), true);
    }

    private function isStaticCallGroup(int $callGroupId): bool {
        return in_array($callGroupId, $this->getStaticCallGroups(), true);
    }

    public function isAllowedCallGroup(int $callGroupId): bool {
        return in_array($callGroupId, $this->getAllowedCallGroups(), true);
    }

    private function buildCallCenterCallGroups(): array {
        $result = [];

        if ($this->_colleague->getExtension()) {
            $allCallGroups = $this->getDictionaryService()->getCallGroups();

            foreach ($allCallGroups->all() as $callGroup) {
                if (!$this->isAllowedCallGroup($callGroup->getValue())) {
                    continue;
                }

                if ($this->isDynamicCallGroup($callGroup->getValue())) {
                    $result[] = [
                        'id' => $callGroup->getValue(),
                        'isStatic' => false,
                        'title' => $callGroup->getLabel(),
                        'active' => true,
                    ];
                } else if ($this->isStaticCallGroup($callGroup->getValue())) {
                    $result[] = [
                        'id' => $callGroup->getValue(),
                        'isStatic' => true,
                        'title' => $callGroup->getLabel(),
                        'active' => true,
                    ];
                } else {
                    $result[] = [
                        'id' => $callGroup->getValue(),
                        'isStatic' => false,
                        'title' => $callGroup->getLabel(),
                        'active' => false,
                    ];
                }
            }
        }

        return $result;
    }
}
