<?php

namespace Wildix\Colleagues;

use Wildix\Error;

/**
 * Class Validator
 */
class Validator {

    /**
     * Available data fields of colleague instance
     *
     * @var array
     */
    public static $availableFields = array(
                                      Colleague::DN,
                                      Colleague::ID,
                                      Colleague::EXTENSION,
                                      Colleague::OFFICE_PHONE,
                                      Colleague::MOBILE_PHONE,
                                      Colleague::FAX_NUMBER,
                                      Colleague::NAME,
                                      Colleague::EMAIL,
                                      Colleague::ROLE,
                                      Colleague::GROUP_DN,
                                      Colleague::LANGUAGE,
                                      Colleague::DIALPLAN,
                                      Colleague::FAX_DIALPLAN,
                                      Colleague::DEPARTMENT,
                                      Colleague::LOGIN,
                                      Colleague::PASSWORD,
                                      Colleague::SIP_PASSWORD,
                                      Colleague::LICENSE_TYPE,
    );
    /**
     * Validation rules
     *
     * @var array
     */
    private static $_rules = array(
        'optional' => array(
            array(Colleague::ROLE),
            array(Colleague::DIALPLAN),
            array(Colleague::FAX_DIALPLAN),
            array(Colleague::LANGUAGE),
            array(Colleague::GROUP_DN),
            array(Colleague::ID),
            array(Colleague::DN),
            array(Colleague::NAME),
            array(Colleague::EXTENSION),
            array(Colleague::PASSWORD),
        ),
        'in' => array(
            array(
                Colleague::ROLE,
                array(Colleague::ADMINISTRATOR, Colleague::ROLE_USER, Colleague::ROLE_ROOM, Colleague::ROLE_FAX, Colleague::ROLE_PARK_ORBIT),
            ),
        ),
        'min' => array(array(Colleague::ID, 1000000)),
        'max' => array(array(Colleague::ID, 10000000)),
        'lengthMin' => array(array(Colleague::DN, 3), array(Colleague::GROUP_DN, 3)),
        'commonName' => Colleague::NAME,
        'principalName' => Colleague::LOGIN,
        'extenNumber' => Colleague::EXTENSION,
        'phoneNumber' => array(array(Colleague::FAX_NUMBER), array(Colleague::OFFICE_PHONE), array(Colleague::MOBILE_PHONE)),
        'email' => Colleague::EMAIL,
        'regex' => [
            [
                Colleague::DEPARTMENT, '/^[^@&<>]{1,250}\z/i',
            ],
        ],
        'languageCode' => Colleague::LANGUAGE,
        'dialplan' => array(array(Colleague::DIALPLAN), array(Colleague::FAX_DIALPLAN)),
        'password' => Colleague::PASSWORD,
        'sipPassword' => Colleague::SIP_PASSWORD,
        'isEmptyOrDifferent' => array(
            array(Colleague::MOBILE_PHONE, Colleague::OFFICE_PHONE),
        ),
    );

    /**
     * Instance of Wildix validator
     *
     * @var \Wildix\Validator
     */
    private $_wildixValidator;

    /**
     * Validator constructor.
     *
     * @param \Wildix\Validator $wildixValidator Instance of Wildix validator.
     * @param array|Group[]     $groups          Array of available groups.
     */
    public function __construct(\Wildix\Validator $wildixValidator, array $groups) {
        $this->_wildixValidator = $wildixValidator;

        $this->addGroupDnRules($groups);
        $this->addLicenseTypeRules();
        $this->_wildixValidator->rules(self::$_rules);
    }

    /**
     * Validate the data and return array of validation errors
     *
     * @param array $data Data for validation.
     *
     * @return array
     */
    public function validate(array $data) {
        $validator   = $this->_wildixValidator->withData($data, self::$availableFields);
        $extraFields = array_diff(array_keys($data), self::$availableFields);

        if (count($extraFields) > 0) {
            array_walk(
                $extraFields,
                function ($field) use ($validator) {
                    if (in_array($field, self::$availableFields)) {
                        $validator->error($field, Error::UnknownError);
                    }
                }
            );
        }

        return $validator->validate() ? array() : $validator->errors();
    }

    /**
     * Add rules for field groupDn
     *
     * @param array $groups Array of available groups.
     *
     * @return void
     */
    private function addGroupDnRules(array $groups): void {
        $availableGroups = [];

        foreach ($groups as $group) {
            if ($group instanceof Group) {
                $availableGroups[] = $group->getDn();
            }
        }

        self::$_rules['in'][] = [Colleague::GROUP_DN, $availableGroups];
    }

    /**
     * Add rules for field licenseType.
     *
     * @return void
     */
    private function addLicenseTypeRules(): void {
        self::$_rules['in'][] = [Colleague::LICENSE_TYPE, \Wildix\License::getLicenseTypes()];
    }
}
