<?php
namespace Wildix\Databases;

use PDO;
use PDOStatement;
use Wildix\Exception;

/**
 * CDRDB Class
 */
class CDRDB extends \PDO {

    private $_logLevel = 0;

    const CDR_CONF_FILE        = '/etc/callweaver/cdr.conf';
    const CDR_MYSQL_CONF_FILE  = '/etc/callweaver/cdr_mysql.conf';
    const CDR_SQLITE_CONF_FILE = '/etc/callweaver/cdr_sqlite3_custom.conf';
    const CDR_MSSQL_CONF_FILE  = '/etc/callweaver/cdr_tds.conf';

    private $_cdrConf = array();
    private $_dbConf  = array();
    private $_dsn     = null;

    private static $_instances = array();

    public static function getInstance() {
        $className = get_called_class();
        if (!isset(self::$_instances[$className])) {
            self::$_instances[$className] = new $className();
        }
        return self::$_instances[$className];
    }

    private $_log = null;
    private function getLog() {
        if ($this->_log === null) {
            $this->_log = new \Syslog('CDRDB', $this->_logLevel);
        }
        return $this->_log;
    }

    public function isExistTable($tableName) {
        $sql = sprintf("SHOW TABLES FROM `%s` LIKE '%s';", $this->_dbConf['dbname'], $tableName);
        return $this->query($sql)->rowCount();
    }

    public function __construct() {
        $className = get_called_class();
        if (isset(self::$_instances[$className])) {
            throw new Exception("Cannot define 2nd instance of Singleton '" . $className . "'");
        }

        if (file_exists(self::CDR_CONF_FILE)) {
            $this->_cdrConf = \CDR::parse_ini_file_quotes_safe(self::CDR_CONF_FILE);

            if ($this->_cdrConf && $this->_cdrConf['general'] && !empty($this->_cdrConf['general']['ctibackend'])) {
                $this->_cdrConf['general']['ctibackend'] = trim($this->_cdrConf['general']['ctibackend']);

                if ($this->isMsSql()) {
                    $this->_dbConf = parse_ini_file(self::CDR_MSSQL_CONF_FILE, true, INI_SCANNER_RAW);
                    $this->_dbConf = $this->_dbConf['global'];

                    $port = '';
                    if (isset($this->_dbConf['port']) && $this->_dbConf['port']) {
                        $port = ':' . $this->_dbConf['port'];
                    }

                    $dsn = 'dblib:host=' . $this->_dbConf['hostname'] . $port . ';dbname=' . $this->_dbConf['dbname'];
                    parent::__construct(
                        $dsn,
                        $this->_dbConf['user'],
                        $this->_dbConf['password']
                    );
                } else if ($this->isMySql()) {
                    $this->_dbConf = parse_ini_file(self::CDR_MYSQL_CONF_FILE, true, INI_SCANNER_RAW);
                    $this->_dbConf = $this->_dbConf['global'];
                    if (isset($this->_dbConf['timeout']) && is_numeric($this->_dbConf['timeout'])) {
                        $timeout = $this->_dbConf['timeout'];
                    } else {
                        $timeout = 10;
                    }

                    $dsn  = 'mysql:host=' . $this->_dbConf['hostname'] . ';port=' . $this->_dbConf['port'];
                    $dsn .= ';dbname=' . $this->_dbConf['dbname'] . ';charset=utf8';
                    parent::__construct(
                        $dsn,
                        $this->_dbConf['user'],
                        $this->_dbConf['password'],
                        array(
                            \PDO::ATTR_TIMEOUT => $timeout,
                            \PDO::ATTR_ERRMODE => \PDO::ERRMODE_EXCEPTION
                        )
                    );
                } else if ($this->isSqlite()) {
                    $this->_dbConf = parse_ini_file(self::CDR_SQLITE_CONF_FILE, true, INI_SCANNER_RAW);
                    $this->_dbConf = $this->_dbConf['cdrdb'];
                    parent::__construct('sqlite:' . $this->_dbConf['dbdir'] . '/' . $this->_dbConf['dbfile']);
                } else {
                    throw new Exception('CDR backend not selected');
                }

                $this->setAttribute(\PDO::ATTR_DEFAULT_FETCH_MODE, \PDO::FETCH_ASSOC);
            } else {
                throw new Exception('CDR backend not selected');
            }
        } else {
            throw new Exception('CDR backend not selected');
        }
        self::$_instances[$className] = $this;
    }

    public function getInfoString() {
        if (!$this->isEnabled()) {
            return null;
        } else if ($this->isSqlite()) {
            return sprintf(
                '%s::%s::%s::%s',
                $this->_cdrConf['general']['ctibackend'],
                $this->_dbConf['dbdir'],
                $this->_dbConf['dbfile'],
                $this->_dbConf['table']
            );
        } else {
            return sprintf(
                '%s::%s::%s::%s:%s',
                $this->_cdrConf['general']['ctibackend'],
                $this->_dbConf['hostname'],
                $this->_dbConf['port'],
                $this->_dbConf['dbname'],
                $this->_dbConf['table']
            );
        }
    }

    public function isEnabled() {
        if ($this->isMsSql()) {
            if (($this->_dbConf['enabled'] == 'yes' || $this->_dbConf['enabled'] == '1')
                && !empty($this->_dbConf['hostname'])
                && !empty($this->_dbConf['dbname'])
                && !empty($this->_dbConf['table'])
            ) {
                return true;
            }
        } else if ($this->isMySql()) {
            if (($this->_dbConf['enabled'] == 'yes' || $this->_dbConf['enabled'] == '1')
                && !empty($this->_dbConf['hostname'])
                && !empty($this->_dbConf['dbname'])
                && !empty($this->_dbConf['table'])
            ) {
                return true;
            }
        } else if ($this->isSqlite()) {
            $this->getLog()->info(__LINE__, gettype($this->_dbConf['enabled']));
            if (($this->_dbConf['enabled'] == 'yes' || $this->_dbConf['enabled'] == '1')
                && !empty($this->_dbConf['dbdir'])
                && !empty($this->_dbConf['dbfile'])
            ) {
                return true;
            }
        }
        return false;
    }

    public function isMsSql() {
        if ($this->_cdrConf
            && $this->_cdrConf['general']
            && !empty($this->_cdrConf['general']['ctibackend'])
            && $this->_cdrConf['general']['ctibackend'] == 'tds'
        ) {
            return true;
        }
        return false;
    }

    public function isMySql() {
        if ($this->_cdrConf
            && $this->_cdrConf['general']
            && !empty($this->_cdrConf['general']['ctibackend'])
            && $this->_cdrConf['general']['ctibackend'] == 'mysql'
        ) {
            return true;
        }
        return false;
    }

    public function isSqlite() {
        if ($this->_cdrConf
            && $this->_cdrConf['general']
            && !empty($this->_cdrConf['general']['ctibackend'])
            && $this->_cdrConf['general']['ctibackend'] == 'sqlite3'
        ) {
            return true;
        }
        return false;
    }


    public function getCdrTableName() {
        return $this->_dbConf['table'];
    }

    public function getNotesTableName() {
        return $this->_dbConf['table'] . '_contact_notes';
    }

    public function getItems(string $query, array $params = []): array {
        $result = $this->prepare($query);
        $result->execute($params);
        $res = $result->fetchAll();
        if (!is_array($res)) {
            $res = [];
        }

        return $res;
    }

    public function query($statement, $mode = PDO::ATTR_DEFAULT_FETCH_MODE, ...$params): PDOStatement {
        if ($this->isEnabled()) {
            $sth = $this->prepare($statement);

            if ($sth) {
                if ($sth->execute($params)) {
                    return $sth;
                } else {
                    throw new Exception(join(', ', $sth->errorInfo()));
                }
            } else {
                throw new Exception(join(', ', $this->errorInfo()));
            }
            return false;
        } else {
            throw new Exception('CDR engine disabled');
        }
    }
    // phpcs:enable Squiz.NamingConventions.ValidVariableName.NotCamelCaps
    // phpcs:enable Zend.NamingConventions.ValidVariableName.NotCamelCaps
    public function exec($statement) {
        if ($this->isEnabled()) {
            $res = parent::exec($statement);
            if ($res === false) {
                // The database disk image is malformed
                throw new Exception('Error: ' . join(', ', $this->errorInfo()));
            }
            return $res;
        } else {
            throw new Exception('CDR engine disabled');
        }
    }

    public function beginTransaction():bool {
        if ($this->isEnabled()) {
            if (!$this->isMsSql()) {
                return parent::beginTransaction();
            }
        } else {
            throw new Exception('CDR engine disabled');
        }
        return false;
    }

    public function commit(): bool {
        if ($this->isEnabled()) {
            if (!$this->isMsSql()) {
                return parent::commit();
            }
        } else {
            throw new Exception('CDR engine disabled');
        }
        return false;
    }

    public function lastInsertId($name = null) {
        if ($this->isEnabled()) {
            if ($this->isMsSql()) {
                $sth = $this->prepare('SELECT SCOPE_IDENTITY() AS id');
                if ($sth) {
                    if ($sth->execute()) {
                        $result = $sth->fetch();
                        return $result['id'];
                    } else {
                        throw new Exception('Error: ' . join(', ', $sth->errorInfo()));
                    }
                } else {
                    throw new Exception('Error: ' . join(', ', $this->errorInfo()));
                }
            } else {
                return parent::lastInsertId($name);
            }
        } else {
            throw new Exception('CDR engine disabled');
        }
    }
    public function realEscapeString($str) {
        return $this->quote($str);
    }

    /**
     * Should be deleted CDR older than this quantity of months
     *
     * @param integer $number Quantity of months.
     *
     * @return boolean
     * @throws \Exception Error
     */
    public function deleteOlderThan($number) {
        return \CDR::getInstance()->clearDatabases((int)$number);
    }
}
