<?php
namespace Wildix\Databases;

use Logger\Log;
use Logger\LoggerFactory;
use SQLite3Result;
use Wildix\Exception;

/**
 * Database Class
 */
class Database extends \SQLite3 {

    private static $_instances = [];
    private $_table            = [];
    private $_dbPatch          = null;
    private $_log              = null;


    const SCHEMA_PATH = BASEPATH . 'api/libs/Wildix/Databases/schemas/';
    // configuration location is not that important. its structure is more important.
    const CONFIG_PATH = BASEPATH . 'api/libs/Api/Database/Configs/database.php';

    const WILDIX_WMSDB         = '/rw2/var/lib/wmsdb.d/wmsdb';
    const WILDIX_QUEUEDB       = '/rw2/var/queue/main.db';
    const WILDIX_CTICONNECT_DB = '/mnt/cticonnect/cticonnectdb';
    const WILDIX_CALENDAR_DB   = '/mnt/cticonnect/calendardb';
    const WILDIX_DEVDB         = '/rw2/var/lib/wmsdb.d/devdb';
    const WILDIX_USRLOC        = '/dev/shm/wmsdb.d/kamailio/usrloc.db';
    const WILDIX_HTABLES       = '/rw2/var/kamailio/htables.db';
    const WILDIX_REGSTATEDB    = '/dev/shm/wmsdb.d/regstatedb';

    const SELECT_QUERY = 'select';
    const INSERT_QUERY = 'insert';
    const DELETE_QUERY = 'delete';
    const UPDATE_QUERY = 'update';

    public static function getInstance($path) {
        $className = get_called_class();
        if (!isset(self::$_instances[$className])) {
            self::$_instances[$className] = new $className($path);
        }
        return self::$_instances[$className];
    }

    public function __construct($path) {
        $className = get_called_class();
        if (isset(self::$_instances[$className])) {
            throw new Exception("Cannot define 2nd instance of Singleton '" . $className . "'");
        }
        $this->_dbPatch = $path;
        $this->connect();

        self::$_instances[$className] = $this;
    }

    /**
     * Get logger class
     *
     * @return Log
     */
    protected function getLog(): Log {
        if ($this->_log === null) {
            $this->_log = LoggerFactory::getInstance();
        }
        return $this->_log;
    }

    public function connect() {
        if ($this->_dbPatch) {
            $this->open($this->_dbPatch);
            $this->busyTimeout(1000);
        }
    }

    public function disconnect() {
        $this->close();
    }

    public function query($sql): SQLite3Result|false {
        $res = @parent::query($sql);
        if ($this->lastErrorCode() > 0) {
            // The database disk image is malformed
            if ($this->lastErrorCode() == 14 && $this->removeJournal()) {
                // unable to open database file
                $res = parent::query($sql);
            } else {
                throw new Exception('Error in(' . $this->_dbPatch . '): ' . $this->lastErrorMsg(), $this->lastErrorCode());
            }
        }
        return $res;
    }

    public function exec($sql): bool {
        $res = @parent::exec($sql);
        if (!$res) {
            // The database disk image is malformed
            if ($this->lastErrorCode() == 14 && $this->removeJournal()) {
                // unable to open database file
                $res = parent::exec($sql);
            } else {
                throw new Exception('Error in(' . $this->_dbPatch . '): ' . $this->lastErrorMsg(), $this->lastErrorCode());
            }
        }
        return $res;
    }

    public function beginTransaction() {
        $this->exec('BEGIN TRANSACTION;');
    }

    public function endTransaction() {
        $this->exec('COMMIT TRANSACTION;');
    }

    public function rollbackTransaction() {
        $this->exec('ROLLBACK;');
    }

    /**
     * Attempt to make a transaction and rollback changes in case the transaction was failed.
     *
     * @param array|string $queries Result-less queries / query to be executed.
     *
     * @return void
     * @throws \Exception If rollback changes was failed.
     */
    public function doTransaction($queries) {
        if (empty($queries)) {
            return;
        }

        $queries = is_array($queries) ? $queries : array($queries);

        try {
            $this->beginTransaction();

            foreach ($queries as $query) {
                $this->exec($query);
            }

            $this->endTransaction();
        } catch (\Exception $e) {
            try {
                $this->rollbackTransaction();
            } catch (\Exception $e) {
                throw new Exception('Error in(' . $this->_dbPatch . '): ' . $this->lastErrorMsg(), $this->lastErrorCode());
            }

            $this->getLog()->error(__LINE__, 'Transaction in (' . $this->_dbPatch . ') failed and the changes were rollback');
            return;
        }
    }

    private function removeJournal() {
        if (file_exists($this->_dbPatch . '-journal')) {
            // Used sudo rm due to sometimes not enough rights
            exec('sudo rm ' . $this->_dbPatch . '-journal', $out, $ret);
            $this->getLog()->error(__LINE__, 'Sqlite journal file exists: ' . $this->_dbPatch . '-journal - ' . ($ret == 0 ? 'deleted' : 'can not delete it'));
            return $ret == 0;
        }
        return false;
    }

    public function getLastInsertId() {
        return $this->lastInsertRowID();
    }

    public function getCountChangesRows() {
        return $this->changes();
    }

    public function realEscapeString($str) {
        return \SQLite3::escapeString($str);
    }

    public function getTableInfo($table) {
        if (empty($this->_table[$table])) {
            $this->_table[$table] = array();

            $result = $this->query('PRAGMA table_info(' . $table . ');');
            while ($row = $result->fetchArray(SQLITE3_ASSOC)) {
                $this->_table[$table][$row['name']] = $row;
            }
        }
        return $this->_table[$table];
    }

    public function dropTable($table) {
        return $this->exec('DROP TABLE IF EXISTS `' . $table . '`;');
    }

    /**
     * Auto disconnect
     */
    public function __destruct() {
        $this->disconnect();
    }

    /**
     * Get first element in database.
     *
     * @param string $sql SQL query.
     *
     * @return array|null
     * @throws \Exception Exception.
     */
    public function first(string $sql): ?array {
        $rows = $this->query($sql);
        $item = $rows->fetchArray(SQLITE3_INTEGER);
        $rows->finalize();
        return $item ? $item : null;
    }

    /**
     * Get list items.
     *
     * @param string $sql      Sql query.
     * @param string $keyField Key field.
     * @param object $class    Class.
     *
     * @return array
     * @throws \Exception Exception.
     */
    //phpcs:disable
    public function getRows(string $sql, ?string $keyField = null, $class = null): array {
        $items = [];
        $rows  = $this->query($sql);
        while ($row = $rows->fetchArray(SQLITE3_ASSOC)) {
            if ($keyField && $class) {
                $items[$row[$keyField]] = new $class($row);
            } else if ($keyField) {
                $items[$row[$keyField]] = $row;
            } else if ($class) {
                $items[] = new $class($row);
            } else {
                $items[] = $row;
            }
        }
        $rows->finalize();
        return $items;
    }
// phpcs:enable
    /**
     *  Insert to database.
     *
     * @param string $table  Table name.
     * @param array  $fields Array fields to value.
     *
     * @return boolean
     * @throws \Exception Exception.
     */
    public function insert(string $table, array $fields): bool {
        $keys   = '';
        $values = '';
        foreach ($fields as $key => $field) {
            if ($keys !== '' && $values !== '') {
                $keys   .= ', ';
                $values .= ', ';
            }
            $keys .= '\'' . $this->realEscapeString(trim($key)) . '\'';
            if (is_int($field)) {
                $values .= $field;
            } else if (is_array($field)) {
                $values .= '\'' . $this->realEscapeString(json_encode($field)) . '\'';
            } else {
                $values .= '\'' . $this->realEscapeString(trim($field)) . '\'';
            }
        }
        $sql = 'INSERT INTO `' . $table . '` (' . $keys . ') VALUES (' . $values . ')';
        return $this->exec($sql);
    }

    public function getDbPatch(): string {
        return $this->_dbPatch;
    }

    public function setDbPatch(string $dbPatch): Database {
        $this->_dbPatch = $dbPatch;
        return $this;
    }
}
