<?php
namespace Wildix\Devices;

use Wildix\Core\BaseModel;
/**
 * Class Device
 */
class Device extends BaseModel {
    const ID    = 'id';
    const PARAM = 'param';
    const VALUE = 'value';
    const FLAG  = 'flag';
    const MAC   = 'mac';
    const TOKEN = 'token';
    const USER  = 'username';

    const ASSIGNED_USERS = 'assignedUsers';

    private const ALIASES = [
        'WP490GR4' => 'WelcomeConsole',
        'WP480GR4' => 'WorkForce',
        'WP410R3' => 'Start'
    ];

    public function __construct($properties) {
        parent::__construct($properties);
        if (!empty($this->_data['value'])) {
            $this->_data['value'] = json_decode($this->_data['value'], true);
        }

        if (!empty($this->_data['settings'])) {
            $this->_data['settings'] = json_decode($this->_data['settings'], true);
        }
    }

    public function getParam() {
        return $this->getProperty(self::PARAM);
    }

    public function getValue() {
        return $this->getProperty(self::VALUE);
    }

    public function getIpAddress(): string {
        $value = (array)$this->getProperty(self::VALUE);
        return $value['address'] ?? '';
    }

    public function getModel(): string {
        return $this->_data['value']['model'] ?? Vendors::UNKNOWN;
    }

    public function getModelAlias(): string {
        $model = $this->getModel();
        if (isset(self::ALIASES[strtoupper($model)])) {
            $model = self::ALIASES[strtoupper($model)];
        }
        return $model;
    }

    public function isWizyconf(): bool {
        return strcasecmp($this->getModel(), \Wildix\Devices\Vendors::WIZYCONF) === 0;
    }

    public function isFxsFamily(): bool {
        return in_array(mb_strtoupper($this->getModel()), Vendors::FXS_FAMILY);
    }

    public function isWprTwoFamily(): bool {
        return in_array(mb_strtoupper($this->getModel()), Vendors::WPR2_FAMILY);
    }

    public function isWpaFamily(): bool {
        return in_array(mb_strtoupper($this->getModel()), Vendors::WPA_FAMILY);
    }

    public function getFlag() {
        return $this->getProperty(self::FLAG);
    }

    public function getMac() {
        return $this->getProperty(self::MAC);
    }

    public function getToken() {
        return $this->getProperty(self::TOKEN);
    }

    public function setToken($value) {
        return $this->setProperty(self::TOKEN, $value);
    }

    public function getUser(): string {
        $isTechnologyFXS    = isset($this->_data['technology']) && $this->_data['technology'] === 'FXS';
        $isAdditionalVendor = false;

        if (!isset($this->_data[self::USER]) || !$this->allowAssign()) {
            return '';
        }
        if (isset($this->_data['vendor:additional'])) {
            $isAdditionalVendor = in_array($this->_data['vendor:additional'], Vendors::FXS_FAMILY) || $this->getModel() === 'w02fxsr2';
        }

        if (!$isTechnologyFXS && !$isAdditionalVendor && strpos($this->_data[self::USER], 'unknown') !== false) {
            $this->_data[self::USER] = str_replace('unknown', 'login', $this->_data[self::USER]);
        }
        return (string)$this->_data['username'];
    }

    public function allowAssign(): bool {
        return in_array(strtolower($this->getModel()), Vendors::MODELS_ALLOW_ASSIGN);
    }

    public function getAssignedUsers() {
        return $this->getProperty(self::ASSIGNED_USERS);
    }

    public function setAssignedUsers($value) {
        return $this->setProperty(self::ASSIGNED_USERS, $value);
    }

    public function isAllowMultipleUsers() {
        return in_array(strtoupper($this->getModel()), Vendors::FXS_FAMILY);
    }

    public function getSettings(): array {
        return isset($this->_data['settings']) ? (array)$this->_data['settings'] : [];
    }

    public function getDescription(): string {
        return isset($this->_data['description']) ? (string)$this->_data['description'] : '';
    }

    public function getFirmware(): string {
        $value = $this->getValue() ?? [];
        return isset($value['fw:version']) ? (string)$value['fw:version'] : '';
    }

    public function isProvision(): bool {
        return $this->getMac() && $this->getIpAddress() && $this->getFirmware();
    }
}
