<?php
namespace Wildix\Devices;

use Api\Trunks\Service\TrunkConfigCwService;
use CIDR;
use Logger\Log;
use PBX;
use PHPGangsta_GoogleAuthenticator as GoogleAuthenticator;
use Services\LuaService;
use Services\TProxyService;
use Wildix\Colleagues\Colleagues;
use Wildix\Databases\DEVDB;
use Wildix\Databases\USRLOC;
use Wildix\Databases\WMSDB;
use Wildix\Dialplan\DialplanService;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\SipProxy;
use Wildix\System\HttpProxy;
use Wildix\System\Network;
use Wildix\System\Upgrade\UpgradeSettings;
use Wildix\TaskQueue\QueueManager;
use Wildix\Trunks\PstnTrunks\Trunk as PstnTrunk;
use Wildix\Utils\JWT;
use WMS_Ldap;

/**
 * Class Devices
 */
class Devices extends \Singleton {

    /**
     *  Firmware local ttl.
     */
    const FIRMWARE_INFO_LOCAL_TTL = 86400;

    /**
     * Firmware local cache path.
     */
    const FIRMWARE_LOCAL_CACHE_PATH = '/dev/shm/.fwversions.json';

    /**
     *  Firmware host.
     */
    const FIRMWARE_DEFAULT_HOST = 'firmwares.wildix.com';

    /**
     * Local directory firmware.
     */
    const LOCAL_DIR_FIRMWARE = '/var/www/firmwares/';

    const DEVICE_SYNC_CONFIG = '/rw2/etc/pbx/device_sync.conf';

    const DEVICE_SYNC_MANUAL = 'disable_sync_manual';

    const DEVICE_SYNC_PORTAL = 'disable_sync_portal';

    /**
     * Firmware port.
     */
    private const FIRMWARE_DEFAULT_PORT = 443;

    /**
     * Firmware path.
     */
    const FIRMWARE_DEFAULT_PATH = '/remoteupgrade/fwversions-5.0.json';

    /**
     *  Firmware protocol.
     */
    private const FIRMWARE_DEFAULT_SCHEME = 'https://';

    /**
     * Firmware timeout.
     */
    private const FIRMWARE_FETCH_TIMEOUT = 3;

    /**
     * Default user password.
     */
    private const DEFAULT_PASS = '1234';

    /**
     * Default user extension.
     */
    private const DEFAULT_EXT = '123456798';

    /**
     *  Device sync command asynchronously.
     */
    private const SYNC_DEVICE = 'nohup sudo wmsbind -o \'%s\' %s %s >/dev/null 2>&1 &';

    /**
     *  Device sync command synchronously.
     */
    private const SYNC_DEVICE_SYNCHRONOUSLY = 'nohup sudo wmsbind -o \'%s\' %s %s >/dev/null 2>&1';

    /**
     * Device base fields.
     *
     * @var string
     */
    private $_tableFields = '`id`, `param`, `value`, `flag`, `mac`, `token`';

    /**
     * Device key map.
     *
     * @var array
     */
    private $_keyMap = [];

    /**
     * Generator a temporary PIN code.
     *
     * @var GoogleAuthenticator
     */
    private $_googleAuthenticator;

    /**
     * Registered devices
     *
     * @var array
     */
    private $_registeredItems;

    /**
     * PBX.
     *
     * @var \Models\PBX
     */
    private $_pbx;

    /**
     * SipProxy
     *
     * @var SipProxy
     */
    private $_sipProxy;

    /**
     * HttpProxy.
     *
     * @var HttpProxy
     */
    private $_httpProxy;

    /**
     * Upgrade.
     *
     * @var UpgradeSettings
     */
    private $_upgrade;

    /**
     * Colleagues.
     *
     * @var Colleagues
     */
    private $_colleagues;

    /**
     * Log service.
     *
     * @var \Logger\Log
     */
    private $_log;

    /**
     * LuaService.
     *
     * @var LuaService
     */
    private $_luaService;

    /**
     * TProxyService.
     *
     * @var TProxyService
     */
    private $_tProxyService;

    /**
     * Models from lua
     *
     * @var array
     */
    private $_models = [];

    /**
     * Remote firmwares.
     *
     * @var array
     */
    private $_remoteFirmwares;

    /**
     * Local firmwares.
     *
     * @var array
     */
    private $_localFirmwares = [];

    /**
     * TProxy devices.
     *
     * @var array
     */
    private $_tProxyDevices = [];

    /**
     * Set Authenticator.
     *
     * @param GoogleAuthenticator $googleAuthenticator Authenticator.
     *
     * @return void
     */
    public function setGoogleAuthenticator(GoogleAuthenticator $googleAuthenticator): void {
        $this->_googleAuthenticator = $googleAuthenticator;
    }

    /**
     * Set PBX.
     *
     * @param \Models\PBX $pbx PBX model.
     *
     * @return void
     */
    public function setPbx(\Models\PBX $pbx): void {
        $this->_pbx = $pbx;
    }

    /**
     * Set HttpProxy instance.
     *
     * @param HttpProxy $httpProxy HttpProxy instance.
     *
     * @return void;
     */
    public function setHttpProxy(HttpProxy $httpProxy): void {
        $this->_httpProxy = $httpProxy;
    }

    /**
     * Get HttpProxy instance.
     *
     * @return HttpProxy
     */
    private function getHttpProxy(): HttpProxy {
        return $this->_httpProxy;
    }

    /**
     * Set SipProxy instance.
     *
     * @param SipProxy $sipProxy SipProxy instance.
     *
     * @return void;
     */
    public function setSipProxy(SipProxy $sipProxy): void {
        $this->_sipProxy = $sipProxy;
    }

    /**
     * Get SipProxy instance.
     *
     * @return SipProxy
     */
    private function getSipProxy(): SipProxy {
        return $this->_sipProxy;
    }

    /**
     * Set Upgrade instance.
     *
     * @param UpgradeSettings $upgrade Upgrade instance.
     *
     * @return void
     */
    public function setUpgradeSettings(UpgradeSettings $upgrade): void {
        $this->_upgrade = $upgrade;
    }

    /**
     * Get Upgrade instance.
     *
     * @return UpgradeSettings
     */
    private function getUpgradeSettings(): UpgradeSettings {
        return $this->_upgrade;
    }

    /**
     * Set Colleagues.
     *
     * @param Colleagues $colleagues Colleagues.
     *
     * @return void
     */
    public function setColleagues(Colleagues $colleagues): void {
        $this->_colleagues = $colleagues;
    }

    /**
     * Set logger.
     *
     * @param Log $log Logger.
     *
     * @return void
     */
    public function setLog(Log $log): void {
        $this->_log = $log;
    }

    /**
     * Set lua service.
     *
     * @param LuaService $service Lua service.
     *
     * @return void
     */
    public function setLuaService(LuaService $service): void {
        $this->_luaService = $service;
    }

    /**
     * Set TProxy service.
     *
     * @param TProxyService $service TProxy service.
     *
     * @return void
     */
    public function setTproxyService(TProxyService $service): void {
        $this->_tProxyService = $service;
    }

    /**
     * Get lua service.
     *
     * @return LuaService Lua service.
     */
    private function getLuaService(): LuaService {
        return $this->_luaService;
    }

    /**
     * Log.
     *
     * @return Log
     */
    private function getLog(): Log {
        if ($this->_log === null) {
            $this->_log = \Logger\LoggerFactory::getInstance();
        }
        return $this->_log;
    }

    /**
     * Generate token for devdb device record.
     *
     * @return string
     * @throws \Exception If secret length is not valid or no source of secure random.
     */
    public function generateToken(): string {
        return $this->_googleAuthenticator->createSecret(16);
    }

    /**
     * Encode JWT token for Wizyconf device.
     *
     * @param string  $mac      A device mac address.
     * @param string  $secret   Secret.
     * @param integer $lifetime Token lifetime time in seconds.
     * @param array   $payload  Additional payload.
     *
     * @return string
     */
    public function createWizyconfAccessToken(string $mac, string $secret, int $lifetime, array $payload = []): string {
        $time           = time();
        $domain         = $this->_pbx->getDomain();
        $defaultPayload = [
            'iss'    => $domain,
            'sub'    => 'Wizyconf ' . $mac,
            'aud'    => $domain,
            'iat'    => $time,
            'exp'    => $time + $lifetime
        ];

        return JWT::encode(array_merge($defaultPayload, $payload), $secret);
    }

    /**
     * Get device from DB by User-Agent.
     *
     * @param string  $ua                A device HTTP User-Agent.
     * @param boolean $withAssignedUsers Add to result info about assigned users.
     *
     * @return Device|null
     * @throws InvalidDataException InvalidDataException
     */
    public function getByUserAgent(string $ua, bool $withAssignedUsers = false): ?Device {
        if (!preg_match('/\((Wizyconf)\s(.+?)\s([0-9a-fA-F:]{12,17})\)/', $ua, $matches) || empty($matches[3])) {
            $this->getLog()->warning(__LINE__, 'Cannot parse device mac');
            return null;
        }

        $mac   = $this->_clearMac($matches[3]);
        $model = Vendors::getDeviceInfoByMac($mac)['modelLine'];

        if ($model === Vendors::UNKNOWN) {
            $this->getLog()->warning(__LINE__, sprintf('For device with mac %s not found model', $mac));
            return null;
        }

        $device = $this->getDeviceByMac($mac);

        if ($device === null) {
            $this->getLog()->warning(__LINE__, 'Mac isn\'t present in database');
            return null;
        }

        if (!$withAssignedUsers) {
            return $device;
        }

        $assignedUsers = $this->getAssignedUsers($mac);

        $device->setAssignedUsers($assignedUsers);
        return $device;
    }

    /**
     * Get devices list from DB.
     *
     * @param array   $filter       Filter.
     * @param string  $search       Search query.
     * @param array   $searchFields Search fields.
     * @param integer $start        Offset.
     * @param integer $count        Limit items.
     * @param string  $sort         Field ordering.
     * @param string  $dir          Ordering (ASC | DESC).
     *
     * @return array
     */
    public function getDevices(array $filter = null, string $search = null, array $searchFields = null, $start = null, $count = null, $sort = null, $dir = null) :array {
        $start = is_numeric($start) ? intval($start) : 0;
        $count = is_numeric($count) ? intval($count) : 100;
        $sort  = $sort ? $sort : Device::ID;
        $dir   = $dir ? $dir : 'ASC';

        $devices = array();

        $query = 'SELECT ' . $this->_tableFields . '
                                FROM `' . DEVDB::TABLE_AUTOPROV . '`
                                ' . $this->createWhere($filter, $search, $searchFields, DEVDB::SELECT_QUERY) . '
                                ORDER BY ' . $sort . ' ' . $dir . '
                                LIMIT ' . $start . ', ' . $count . ';';

        $rows = DEVDB::getInstance()->query($query);
        if ($rows) {
            while ($row = $rows->fetchArray(SQLITE3_ASSOC)) {
                if ($row && !empty($row)) {
                    $devices[] = new Device($row);
                }
            }
        }
        return $devices;
    }

    /**
     * Get count items.
     *
     * @param array  $filter       Filter.
     * @param string $search       Search query.
     * @param array  $searchFields Search fields.
     *
     * @return integer
     */
    public function getCount(array $filter = null, string $search = null, array $searchFields = null) :int {
        $query = 'SELECT count(id) as countRows
                                FROM `' . DEVDB::TABLE_AUTOPROV . '`
                                ' . $this->createWhere($filter, $search, $searchFields, DEVDB::SELECT_QUERY) . ';';

        $rows = DEVDB::getInstance()->query($query);
        if ($rows) {
            $row = $rows->fetchArray(SQLITE3_ASSOC);
            if ($row) {
                return $row['countRows'];
            }
        }

        return 0;
    }

    /**
     * Get users assigned for device
     *
     * @param string $mac A device mac address.
     *
     * @return array
     */
    public function getAssignedUsers(string $mac): array {
        $users = [];
        $query = \sprintf("SELECT value as userExt FROM `%s` WHERE `param` = 'username' AND `mac` = '%s';", DEVDB::TABLE_AUTOPROV, DEVDB::getInstance()->realEscapeString($mac));

        $rows = DEVDB::getInstance()->getRows($query);

        foreach ($rows as $row) {
            $user = $this->_colleagues->getUserByExtension($row['userExt']);
            if ($user !== null) {
                $users[] = $user;
            }
        }
        return $users;
    }

    /**
     * Create where.
     *
     * @param array  $filter       Filter.
     * @param string $search       Search.
     * @param array  $searchFields Search fields.
     * @param string $queryType    Type query.
     *
     * @return string
     */
    private function createWhere(array $filter = null, string $search = null, array $searchFields = null, $queryType) {
        $filter = is_array($filter) ? $filter : array();
        if ($queryType != DEVDB::DELETE_QUERY) {
            $filter[Device::PARAM] = 'device';
        }

        $search       = $search ? trim($search) : '';
        $searchFields = $searchFields ? $searchFields : array();

        $where = array();
        if ($filter && !empty($filter)) {
            foreach ($filter as $key => $value) {
                $keys = explode(',', $key);
                if (count($keys) > 1) {
                    $tmpWhere = array();
                    foreach ($keys as $key) {
                        $tmpWhere[] = $this->filterCreate($key, $value);
                    }
                    if (!empty($tmpWhere)) {
                        $where[] = '(' . join(' OR ', $tmpWhere) . ')';
                    }
                } else {
                    $where[] = $this->filterCreate($key, $value);
                }
            }
        }

        if ($search && !empty($search)) {
            $searchWhere = array();
            $search      = DEVDB::getInstance()->realEscapeString($search);

            $countFields = count($searchFields);
            for ($i = 0; $i < $countFields; $i++) {
                $searchWhere[] = $searchFields[$i] . ' LIKE \'%' . $search . '%\'';
            }
            if (!empty($searchWhere)) {
                $where[] = '(' . join(' OR ', $searchWhere) . ')';
            }
        }

        if (count($where) > 0) {
            $where = 'WHERE ' . join(' AND ', $where);
        } else {
            $where = '';
        }

        return $where;
    }

    private function filterCreate($key, $value) {
        if (is_scalar($value)) {
            return '`' . $this->getRealKey($key) . "` = '" . DEVDB::getInstance()->realEscapeString($value) . "'";
        } else if (is_array($value)) {
            $tmpWhere = array();
            if (!empty($value['from'])) {
                $tmpWhere[] = '`' . $this->getRealKey($key) . "` >= '" . DEVDB::getInstance()->realEscapeString($value['from']) . "'";
            }
            if (!empty($value['to'])) {
                $tmpWhere[] = '`' . $this->getRealKey($key) . "` <= '" . DEVDB::getInstance()->realEscapeString($value['to']) . "'";
            }
            if (isset($value['not'])) {
                $tmpWhere[] = '`' . $this->getRealKey($key) . "` <> '" . DEVDB::getInstance()->realEscapeString($value['not']) . "'";
            }
            if (isset($value['or'])) {
                if (is_array($value['or'])) {
                    foreach ($value['or'] as $k => $v) {
                        $value['or'][$k] = DEVDB::getInstance()->realEscapeString($v);
                    }
                    $tmpWhere[] = '`' . $this->getRealKey($key) . "` IN ('" . implode("','", $value['or']) . "')";
                } else {
                    $tmpWhere[] = '`' . $this->getRealKey($key) . "` = '" . DEVDB::getInstance()->realEscapeString($value['or']) . "'";
                }
            }

            if (!empty($tmpWhere)) {
                return '(' . join(' AND ', $tmpWhere) . ')';
            }
        }
        return '';
    }

    /**
     * Validate params.
     *
     * @param array $data Data.
     *
     * @return void
     * @throws InvalidDataException InvalidDataException.
     */
    private function validate(array $data): void {
        $res = [];

        if (!isset($data['user'])
            && !isset($data['multipleUsers'])
            && !isset($data['device'])
            && !isset($data['settings'])
            && !isset($data['description'])
        ) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $res);
        }

        if (isset($data['user']) && empty($data['user'])) {
            $res['user'] = [
                'code' => 400,
                'reason' => Error::getMessage(Error::FieldCannotBeEmpty)
            ];
        }

        if (!empty($data['user']) && !$this->isEntry($data['user'], 'unknown') && !$this->_colleagues->getUserByExtension($data['user'])) {
            $res['user'] = [
                'code' => 400,
                'reason' => Error::getMessage(Error::USER_EXTENSION)
            ];
        }

        if (isset($data['multipleUsers']) && empty($data['multipleUsers'])) {
            $res['multipleUsers'] = [
                'code' => 400,
                'reason' => Error::getMessage(Error::FieldCannotBeEmpty)
            ];
        }

        if (isset($data['multipleUsers'])) {
            foreach ($data['multipleUsers'] as $key => $item) {
                $stringItem = (string)$item;
                if (!$this->isEntry($stringItem, 'unknown') && !$this->_colleagues->getUserByExtension($item)) {
                    $res['multipleUsers'][] = [
                        'code' => 400,
                        'reason' => $item . ' ' . Error::getMessage(Error::USER_EXTENSION)
                    ];
                }
            }
        }

        if (!empty($res)) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $res);
        }
    }

    /**
     * Get real key field.
     *
     * @param string $key Field.
     *
     * @return string
     */
    private function getRealKey($key) {
        if (isset($this->_keyMap[$key])) {
            return $this->_keyMap[$key];
        }
        return $key;
    }

    public function updateDevice($data, $filter = null, $search = null, $searchFields = null) {
        array_walk_recursive(
            $data,
            function (&$value) {
                $stringValue = (string)$value;
                $value       = trim($stringValue);
            }
        );

        if (empty($data)) {
            return 0;
        }

        if (isset($data['value'])) {
            $data['value'] = json_encode($data['value']);
        }

        if (isset($filter[Device::ID])) {
            unset($data[Device::ID]);
        }

        $params = array();
        foreach ($data as $key => $value) {
            $params[] = '`' . $this->getRealKey($key) . "` = '" . DEVDB::getInstance()->realEscapeString($value) . "'";
        }

        $query = 'UPDATE `' . DEVDB::TABLE_AUTOPROV . '` SET ' . join(', ', $params) . '
                                ' . $this->createWhere($filter, $search, $searchFields, DEVDB::UPDATE_QUERY) . ';';

        $result = DEVDB::getInstance()->exec($query);
        if ($result === true) {
            $result = DEVDB::getInstance()->getCountChangesRows();
        }
        return $result;
    }

    /**
     * Return a pin code for calling method resetByPin.
     *
     * @param string $mac A device mac address.
     *
     * @return string A pin code.
     * @throws \Wildix\Exception If device not found.
     * @throws \LogicException   Wrong Devices instance.
     */
    public function generatePinForResetByMac(string $mac): string {
        if (!isset($this->_googleAuthenticator)) {
            throw new \LogicException('Wrong Devices instance. It doesn\'t contain google authenticator.');
        }

        $device = $this->getDeviceByMac($mac);

        if (!isset($device)) {
            throw new \Wildix\Exception(\Wildix\Error::DeviceNotFound, 404);
        }

        return $this->_googleAuthenticator->getCode($device->getToken());
    }

    /**
     * Reset a device token by PIN code.
     *
     * @param string $pin    PIN code for device.
     * @param array  $fields List fields for reset.
     *
     * @return void
     * @throws \Wildix\Exception if device not found by a PIN code.
     */
    public function resetByPin(string $pin, array $fields): void {
        $devicesPage            = 1;
        $devicesPerPage         = 10;
        $createGetDevicesParams = function () use (&$devicesPage, $devicesPerPage) {
            return [
            ['token' => ['not' => '']], 'wizyconf', ['value'], ($devicesPage - 1) * $devicesPerPage, $devicesPerPage,
            ];
        };
        $devices                = call_user_func_array([$this, 'getDevices'], $createGetDevicesParams());

        while (!empty($devices)) {
            $device = current(
                array_filter(
                    $devices,
                    function (Device $device) use ($pin) {
                        return $this->_googleAuthenticator->verifyCode($device->getToken(), $pin, 10);
                    }
                )
            );

            if (!$device instanceof Device) {
                $devicesPage++;
                $devices = call_user_func_array([$this, 'getDevices'], $createGetDevicesParams());

                continue;
            }

            $deviceData = $device->getAsArray();

            foreach ($fields as $field) {
                if (!isset($deviceData[$field])) {
                    continue;
                }

                $deviceData[$field] = '';
            }

            $updatedDevices = $this->updateDevice($deviceData, [Device::ID => $device->getId()]);

            if ($updatedDevices !== 1) {
                $this->getLog()->warning(
                    __LINE__,
                    sprintf('Reset by PIN affected not a specific device with mac "%s". Updated %d devices.', $device->getMac(), $updatedDevices)
                );
            }

            return;
        }

        throw new \Wildix\Exception(\Wildix\Error::DeviceNotFound, 404);
    }

    /**
     * Build base query.
     *
     * @param string  $pattern    Query pattern.
     * @param array   $conditions Conditions.
     * @param integer $count      Limit items.
     * @param integer $start      Offset items.
     *
     * @return string
     */
    private function buildBaseQuery(string $pattern, array $conditions = [], int $count = 0, int $start = 0): string {
        $fields = 't1.mac, t1.id AS id, t1.value AS value, t1.token AS token';
        $joins  = '';

        if ($count > 0) {
            $pattern .= ' LIMIT ' . $count;
        }

        if ($start > 0) {
            $pattern .= ' OFFSET ' . $start;
        }

        if (in_array('user', $conditions)) {
            $joins  .= 'LEFT JOIN autoprov AS t2 ON (lower(t1.mac) = lower(t2.mac) AND t2.param=\'username\') ';
            $fields .= ', t2.value AS username';
        }

        if (in_array('description', $conditions)) {
            $joins  .= 'LEFT JOIN autoprov AS t3 ON (lower(t1.mac) = lower(t3.mac) AND t3.param=\'description\') ';
            $fields .= ', t3.value AS description';
        }

        if (in_array('settings', $conditions)) {
            $joins  .= 'LEFT JOIN autoprov AS t4 ON (lower(t1.mac) = lower(t4.mac) AND t4.param=\'settings\') ';
            $fields .= ', t4.value AS settings';
        }

        return sprintf($pattern, $fields, DEVDB::TABLE_AUTOPROV, $joins);
    }

    /**
     * Get devices.
     *
     * @param integer $start Offset items.
     * @param integer $count Limit items.
     *
     * @return array
     */
    public function getItems(int $start, int $count) :array {
        $devices = [];
        $items   = DEVDB::getInstance()
            ->getRows(sprintf('SELECT * FROM %s WHERE param="device" LIMIT %s OFFSET %s', DEVDB::TABLE_AUTOPROV, $count, $start), 'mac');
        if (!empty($items)) {
            $macs     = '\'' . strtolower(implode('\', \'', array_keys($items))) . '\'';
            $pattern  = 'SELECT mac, value, param FROM %s WHERE lower(mac) IN (%s) AND param="%s"';
            $comments = DEVDB::getInstance()->getRows(sprintf($pattern, DEVDB::TABLE_AUTOPROV, $macs, 'description'), 'mac');
            $settings = DEVDB::getInstance()->getRows(sprintf($pattern, DEVDB::TABLE_AUTOPROV, $macs, 'settings'), 'mac');
            $users    = DEVDB::getInstance()->getRows(sprintf($pattern, DEVDB::TABLE_AUTOPROV, $macs, 'username'), 'mac');
        }

        foreach ($items as $key => $item) {
            $item['description'] = isset($comments[$key]) ? $comments[$key]['value'] : '';
            $item['settings']    = isset($settings[$key]) ? $settings[$key]['value'] : '{}';
            $item['username']    = isset($users[$key]) ? $users[$key]['value'] : '';
            $devices[$key]       = new Device($item);
        }

        return $this->assignedUsersToDevices($devices);
    }

    /**
     * Get device by mac address.
     *
     * @param string  $macAddress Mac address.
     * @param array   $conditions Conditions.
     * @param boolean $check      Check device.
     *
     * @return Device|null
     * @throws InvalidDataException InvalidDataException.
     */
    public function getDeviceByMac(string $macAddress, array $conditions = [], $check = false) {
        $macAddress = $this->_clearMac($macAddress);
        $macAddress = DEVDB::getInstance()->realEscapeString($macAddress);
        $pattern    = 'SELECT %s FROM %s as t1 %s WHERE t1.param="device" AND lower(t1.mac)="' . $macAddress . '"';

        $res = DEVDB::getInstance()->first($this->buildBaseQuery($pattern, $conditions));

        if ($check && empty($res)) {
            $this->getLog()->error(__LINE__, Error::getMessage(Error::NOT_FOUND_DEVICE) . ' ' . $macAddress);
            throw new InvalidDataException(Error::getMessage(Error::NOT_FOUND_DEVICE), [], 404);
        }

        if (!empty($res)) {
            $res = new Device($res);
        }

        return $res;
    }

    /**
     * Create devices by macs.
     *
     * @param array  $macAddresses Mac address list.
     * @param string $server       IP address server.
     * @param string $address      IP address device.
     *
     * @return array
     * @throws InvalidDataException No valid data.
     */
    public function create(array $macAddresses = [], string $server = '', $address = ''): array {
        $result  = [];
        $noValid = [];
        if (empty($macAddresses)) {
            throw new InvalidDataException(
                Error::getMessage(Error::VALIDATION_ERROR),
                ['macs' => Error::getMessage(Error::FieldCannotBeEmpty)]
            );
        }

        foreach ($macAddresses as $key => $item) {
            $item               = $this->_clearMac($item);
            $macAddresses[$key] = $item;
            $stringItem         = (string)$item;
            if (!Network::isMac($stringItem)) {
                $noValid[$key] = [
                    'code' => Error::MAC_ADDRESS,
                    'reason' => Error::getMessage(Error::MAC_ADDRESS)
                ];
            }
        }

        if (!empty($noValid)) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $noValid);
        }

        $readyDevices = $this->findReadyDevices($macAddresses);

        foreach ($macAddresses as $mac) {
            $info    = Vendors::getDeviceInfoByMac($mac, $server);
            $address = Network::isIp($address) ? $address : '';
            $device  = [
                'mac' => $mac,
                'param' => 'device',
                'value' => [
                    'mac' => $mac,
                    'model' => $info['modelLine'],
                    'vendor' => Vendors::WILDIX,
                    'address' => $address
                ]
            ];


            if (!in_array($mac, $readyDevices) && (strtoupper($info['modelLine']) !== strtoupper(Vendors::UNKNOWN) || Network::isIp($address))) {
                DEVDB::getInstance()->insert(DEVDB::TABLE_AUTOPROV, $device);
            }
            if (in_array($mac, $readyDevices)) {
                $info['url'] = 'Device already exists';
            }
            $device['value']['url'] = $info['url'];
            $result[$mac]           = $device;
        }

        return array_values($result);
    }

    /**
     * Update device by mac address.
     *
     * @param string $macAddress Mac address.
     * @param array  $data       Array data.
     *
     * @return boolean
     * @throws InvalidDataException Invalid Data Exception.
     */
    public function update(string $macAddress = '', array $data = []): bool {
        $macAddress = $this->_clearMac($macAddress);
        if (!Network::isMac($macAddress)) {
            $res['mac'] = [
                'code' => 400,
                'reason' => Error::getMessage(Error::MAC_ADDRESS)
            ];
            throw new InvalidDataException(Error::getMessage(Error::MAC_ADDRESS), $res);
        }

        $device = $this->getDeviceByMac($macAddress, [], true);

        if (!$device->allowAssign()) {
            unset($data['user']);
        }

        if (!$device->isAllowMultipleUsers()) {
            unset($data['multipleUsers']);
        }

        $preparedData = $this->prepareToSaveDevice($data);

        $this->validate($preparedData);

        if (isset($preparedData['device']) && $preparedData['device']['token'] === '') {
            $query = 'UPDATE %s SET token=\'\' WHERE mac=\'%s\' AND param=\'device\'';
            DEVDB::getInstance()->exec(sprintf($query, DEVDB::TABLE_AUTOPROV, $device->getMac()));
        }

        if (isset($data['settings'])) {
            $this->updateSettings($device->getMac(), $preparedData['settings']);
        }

        if (isset($preparedData['description'])) {
            $this->updateDescription($device->getMac(), $preparedData['description']);
        }

        if (isset($preparedData['user']) && !$device->isAllowMultipleUsers()) {
            $this->updateUser($device->getMac(), $preparedData['user']);
        } else if (isset($preparedData['multipleUsers']) && $device->isAllowMultipleUsers()) {
            $this->setMultipleUsers($device->getMac(), $data['multipleUsers']);
        }

        if (Vendors::isNeedsToReboot($device->getModel())) {
            $peer = 'unknown';

            if ($device->getModel() === TrunkConfigCwService::DEVICE_MODEL_FXOR3) {
                $trunkId = $this->getTrunkId($device->getMac());
                if ($trunkId) {
                    $peer = $this->getTrunkName($trunkId) . '_0';
                }
            } else {
                $device = $this->assignedUsersToDevices([$device->getMac() => $device])[$device->getMac()];

                $assignedUsers = $device->getAssignedUsers();
                sort($assignedUsers);
                $peer = $assignedUsers[0] ?? 'unknown';
            }

            PBX::getInstance()->sip_notify_check_sync([$peer], true);
        }

        return true;
    }

    /**
     * Remove device by mac address.
     *
     * @param string $macAddress Mac address.
     *
     * @return boolean
     * @throws InvalidDataException InvalidDataException.
     */
    public function remove(string $macAddress): bool {
        $macAddress = $this->_clearMac($macAddress);
        if (!Network::isMac($macAddress)) {
            throw new InvalidDataException(Error::getMessage(Error::MAC_ADDRESS), []);
        }

        $this->getDeviceByMac($macAddress, [], true);

        $query   = 'DELETE FROM %s WHERE lower(mac)="%s"';
        $queries = [
            sprintf($query, DEVDB::TABLE_AUTOPROV, $macAddress),
            sprintf($query, DEVDB::TABLE_WAIR_USERS, $macAddress),
            sprintf($query, DEVDB::TABLE_WAIR_GATEWAYS, $macAddress),
        ];
        DEVDB::getInstance()->doTransaction($queries);
        $addresses = $this->getListAddressDevices();
        $devices   = $this->getTproxyDevices();
        $this->saveTproxyDevices($devices, $addresses);
        return true;
    }

    public function findLinkInDialplan($trunkName): array {
        $sql = sprintf(
            "select dialplan_rulenames.id, dialplan_rulenames.rulename from dialplan_links
    left join dialplan_rulenames on dialplan_links.object_from_id = dialplan_rulenames.id
    where object_to_type = %s and object_to_id like '%s%%'",
            DialplanService::LINK_TRUNK,
            $trunkName
        );

        return WMSDB::getInstance()->getRows($sql);
    }

    /**
     * Bind device.
     *
     * @param string  $macAddress Mac address.
     * @param string  $address    Ip address.
     * @param boolean $async      Async mode (default = true).
     *
     * @return boolean
     * @throws InvalidDataException Invalid data Exception
     */
    public function synchronizationItem(string $macAddress, string $address, bool $async = true): bool {
        $this->getLog()->info(__LINE__, sprintf('Synchronization device start: mac=%s, ip=%s', $macAddress, $address));
        $macAddress = $this->_clearMac($macAddress);
        if ($this->isDisableSync(self::DEVICE_SYNC_MANUAL)) {
            return false;
        }

        if (!Network::isMac($macAddress)) {
            throw new InvalidDataException(Error::getMessage(Error::MAC_ADDRESS), []);
        }

        if (!Network::isIp($address)) {
            throw new InvalidDataException(Error::getMessage(Error::InvalidIpAddress), []);
        }

        $device = $this->getDeviceByMac($macAddress);

        if (empty($device)) {
            $this->create([$macAddress], '', $address);
            $device = $this->getDeviceByMac($macAddress);
        }

        if (in_array($address, $this->getNatIps())) {
            $this->syncDeviceViaSipNotify($device);
        } else {
            $this->executeActionProvisionSync($device, $async);
        }

        return true;
    }

    /**
     * Execute command sync with action provision.
     *
     * @param Device  $device Device.
     * @param boolean $async  Async mode (default = true).
     *
     * @return void
     */
    private function executeActionProvisionSync(Device $device, bool $async = true): void {
        $item = [
            'mac' => $device->getMac(),
            'address' => $device->getIpAddress()
        ];
        $this->executeCommandSync('action=provision', [$item], $async);
    }

    /**
     * Bind devices.
     *
     * @param array $items Mac address and Ip address.
     *
     * @return void
     */
    public function synchronizationItems(array $items): void {
        $items = array_filter($items, [$this, 'isValidBindItem']);

        foreach ($items as $item) {
            \Services\Helpers::addTask(
                'deviceSync',
                [
                    'ip' => $item['address'],
                    'mac' => $item['mac']
                ]
            );
        }
        QueueManager::runQueueManager();
    }

    /**
     * Get ready macs address.
     *
     * @return array
     */
    public function getReadyMacs(): array {
        $rows = DEVDB::getInstance()->getRows('SELECT distinct mac FROM ' . DEVDB::TABLE_AUTOPROV, 'mac');
        return array_keys($rows);
    }

    /**
     * Update device settings.
     *
     * @param string $mac      Mac address.
     * @param array  $settings Settings.
     *
     * @return void
     * @throws InvalidDataException InvalidDataException
     */
    private function updateSettings(string $mac, array $settings) :void {
        $device = $this->getDeviceByMac($mac, ['settings'])->getAsArray();
        if ($device['settings'] === null) {
            $this->createDeviceRow($mac, json_encode($settings), 'settings');
        } else {
            $currentSettings = $device['settings'];
            $settings        = json_encode(array_merge($currentSettings, $settings));
            $this->updateDeviceRow($mac, $settings, 'settings');
        }
    }

    /**
     * Update device description.
     *
     * @param string $mac         Mac address.
     * @param string $description Description.
     *
     * @throws InvalidDataException Invalid Data Exception.
     * @return void
     */
    private function updateDescription(string $mac, string $description) :void {
        $device = $this->getDeviceByMac($mac, ['description'])->getAsArray();
        if ($device['description'] === null) {
            $this->createDeviceRow($mac, $description, 'description');
        } else {
            $this->updateDeviceRow($mac, $description, 'description');
        }
    }

    /**
     * Update user device.
     *
     * @param string $mac          Mac address.
     * @param string $newExtension New extension.
     *
     * @return void
     * @throws \Exception Exception.
     */
    private function updateUser(string $mac, string $newExtension) :void {
        $device = $this->getDeviceByMac($mac, ['user']);

        $user = $this->getDeviceUser($newExtension, $device->getUser());

        if (empty($device->getUser())) {
            $this->createDeviceRow($device->getMac(), $user['newExtension'], 'username');
        } else {
            $this->updateDeviceRow($device->getMac(), $user['newExtension'], 'username');
        }

        if (!$device->isWizyconf() && $device->isProvision()) {
            $this->setUserToDevice($user, $device);
        }
    }

    /**
     * Set multiple users.
     *
     * @param string $mac   Mac address.
     * @param array  $users Users extension.
     *
     * @return void
     * @throws \Exception Exception.
     */
    public function setMultipleUsers(string $mac, array $users) :void {
        $device   = $this->getDeviceByMac($mac);
        $queries  = [];
        $pattern  = 'INSERT OR REPLACE INTO autoprov (id, param, value, flag, mac) ';
        $pattern .= 'VALUES ((SELECT id FROM autoprov WHERE param="username" AND mac="%s" AND flag="%s"), "username", "%s", "%s", "%s")';

        for ($i = 0; $i < $this->getPortCount($device->getModel()); $i++) {
            $port = $i + 1;
            if (isset($users[$i])) {
                $user = $users[$i];
            } else {
                $user = $this->getFreeUnknownAccount($this->getUnknownAccounts());
            }
            $queries[] = sprintf($pattern, $device->getMac(), $port, DEVDB::getInstance()->realEscapeString($user), $port, $device->getMac());
        }

        DEVDB::getInstance()->doTransaction($queries);

        $syncUsers = [];
        $info      = Vendors::getDeviceInfoByMac($mac);
        if ($info['modelLine'] == 'MGW1' || $info['modelLine'] == 'MGW') {
            $registrations = USRLOC::getInstance()->getRows(
                'SELECT username FROM ' . USRLOC::TABLE_LOCATION . ' WHERE user_agent LIKE "%' . $device->getMac() . '%"'
            );
            foreach ($registrations as $registration) {
                $syncUsers[0] = $registration['username'];
                if (!$this->isEntry($syncUsers[0], 'unknown')) {
                    break;
                }
            }
        }

        if (Network::isIp($device->getIpAddress())) {
            $this->syncDeviceViaSipNotify($device, $syncUsers);
            $this->executeActionProvisionSync($device);
        }
    }

    /**
     * Set user to device.
     *
     * @param array  $user   User data.
     * @param Device $device Model device.
     *
     * @return void
     */
    private function setUserToDevice(array $user, Device $device) :void {
        $devicesNatIp = $this->getNatIps();
        $wpa          = false;
        $wprTwo       = false;
        if ($device->isWprTwoFamily()) {
            $wprTwo = true;
        } else if ($device->isWpaFamily()) {
            $wpa = true;
        }
        if (!in_array($device->getIpAddress(), $devicesNatIp) && !$wprTwo && !$wpa) {
            if ($device->isFxsFamily()) {
                $arg = 'action=provision';
            } else {
                $arg = 'action=setuser,username=' . $user['newExtension'] . ',password="' . base64_encode($user['password']) . '",userid=' . $user['id'];
            }
            $item = [
                'mac' => $device->getMac(),
                'address' => $device->getIpAddress()
            ];
            $this->executeCommandSync($arg, [$item]);
        } else if (!empty($user['oldExtension'])) {
            $syncUser = str_replace('login', 'unknown', $user['oldExtension']);
            $this->syncDeviceViaSipNotify($device, [$syncUser]);
            $this->getLog()->info(__LINE__, 'Send check-sync (reboot ' . $device->isWpaFamily() . ') to ' . $device->getModel() . ' (' . $syncUser . ')');
        }
    }

    /**
     * Update device param in database.
     *
     * @param string $mac   Mac address.
     * @param string $value Value.
     * @param string $row   Type value device.
     *
     * @return void
     */
    private function updateDeviceRow(string $mac, string $value, string $row) :void {
        $query = 'UPDATE %s SET value=\'%s\' WHERE mac=\'%s\' AND param=\'%s\'';
        DEVDB::getInstance()
            ->exec(sprintf($query, DEVDB::TABLE_AUTOPROV, DEVDB::getInstance()->realEscapeString($value), $mac, $row));
    }

    /**
     * Add device param to database.
     *
     * @param string $mac   Mac address.
     * @param string $value Value.
     * @param string $row   Type value device.
     *
     * @return boolean
     */
    private function createDeviceRow(string $mac, string $value, string $row) :bool {
        return DEVDB::getInstance()->insert(
            DEVDB::TABLE_AUTOPROV,
            [
                'param' => $row,
                'value' => $value,
                'mac' => $mac,
                ]
        );
    }

    /**
     * Prepare device to api returned.
     *
     * @param Device $item Array data from database.
     *
     * @return array
     */
    public function buildItem(Device $item) :array {
        $device   = $item->getAsArray();
        $value    = $device['value'] ?? [];
        $settings = $device['settings'] ?? [];

        $res = [
            'id' => (int)$item->getId(),
            'mac' => (string)$item->getMac(),
            'model' => $item->getModelAlias(),
            'provision' => $item->isProvision(),
            'registrar' => [
                'address' => $value['registrar'] ?? '',
                'port' => $value['registrar:port'] ?? '',
            ],
            'vendor' => [
                'additional' => $value['vendor:additional'] ?? '',
                'name' => $value['vendor'] ?? '',
            ],
            'description' => (string)($device['description'] ?? ''),
            'firmware' => $item->getFirmware(),
            'alternatePort' => (string)($value['alternate_port'] ?? ''),
            'settings' => $this->buildSettings($settings, $item->getModel()),
            'token' => (string)($device['token'] ?? ''),
            'address' => $item->getIpAddress(),
            'allowAssign' => $item->allowAssign(),
            'user' => '',
            'qtyPorts' => $this->getPortCount($item->getModel()),
            'multipleUsers' => [],
            'ready' => $this->getState($item),
            'isOnline' => $this->isOnline($item),
            'connect' => $this->getConnectUrl($item->getIpAddress()),
            'ssl' => \Wildix\Validator::isTrue($device['ssl'] ?? ''),
            'received' => $this->getReceived($item->getMac()),
            'isLocal' => $this->isLocalAddress($item->getIpAddress())
        ];

        if ($item->isAllowMultipleUsers()) {
            $res['multipleUsers'] = array_values($device['assignedUsers']);
            $res['allowAssign']   = true;
        } else {
            $res['user'] = $item->getUser();
        }

        return $res;
    }

    /**
     * Prepare device to save
     *
     * @param array $data Device.
     *
     * @return array
     */
    private function prepareToSaveDevice(array $data) :array {
        $device = [];
        if (array_key_exists('token', $data) && $data['token'] === '') {
            $device['device'] = [
                'token' => ''
            ];
        }

        if (!empty($data['settings'])) {
            $device['settings'] = $this->prepareSettingsToSave($data['settings']);
        }

        if (isset($data['description'])) {
            $device['description'] = (string)$data['description'] ?: '';
        }

        if (isset($data['user'])) {
            $device['user'] = (string)$data['user'];
        }

        if (isset($data['multipleUsers'])) {
            $device['multipleUsers'] = (array)$data['multipleUsers'];
        }

        return $device;
    }

    /**
     * Prepare settings to save.
     *
     * @param array $data     Settings.
     * @param array $settings Settings.
     *
     * @return void
     */
    private function prepareDnsSettingToSave(array $data, array &$settings): void {
        if (isset($data['dns'])) {
            $settings['dns_server']        = $data['dns']['primary'] ?? '';
            $settings['second_dns_server'] = $data['dns']['secondary'] ?? '';
        }
    }

    /**
     * Prepare settings to save.
     *
     * @param array $data Settings.
     *
     * @return array
     */
    private function prepareSettingsToSave(array $data):array {
        $settings = [];
        $this->prepareDnsSettingToSave($data, $settings);
        if (isset($data['voiceVLAN'])) {
            $settings['vid_voice'] = $data['voiceVLAN']['voice'] ?? '';
            $settings['qos_voice'] = $data['voiceVLAN']['data'] ?? '';
        }
        if (isset($data['dataVLAN'])) {
            $settings['vid_data'] = $data['dataVLAN']['voice'] ?? '';
            $settings['qos_data'] = $data['dataVLAN']['data'] ?? '';
        }
        if (isset($data['ntp'])) {
            $settings['ntp_server']   = $data['ntp']['server'] ?? '';
            $settings['ntp_timezone'] = $data['ntp']['timezone'] ?? '';
        }
        if (isset($data['toneZone'])) {
            $settings['tonezone'] = $data['toneZone'] ?? '';
        }
        if (isset($data['snmp'])) {
            $settings['snmp'] = $data['snmp'] ?? '';
        }
        if (isset($data['useStaticIp'])) {
            $settings['use_static_ip'] = \Wildix\Validator::isTrue($data['useStaticIp']) ? 1 : 0;
        }
        if (isset($data['useDst'])) {
            $settings['use_dst'] = \Wildix\Validator::isTrue($data['useDst']) ? 'yes' : 'no';
        }

        if (isset($data['syslogServer'])) {
            $settings['syslog_server'] = $data['syslogServer'] ?? '';
        }

        return $settings;
    }

    /**
     * Prepare setting to api.
     *
     * @param array  $settings Settings.
     * @param string $model    Model.
     *
     * @return array
     */
    private function buildSettings(array $settings, string $model): array {
        $res = [
            'dns' => [
                'primary'   => (string)($settings['dns_server'] ?? ''),
                'secondary' => (string)($settings['second_dns_server'] ?? ''),
            ],
            'voiceVLAN' => [
                'voice' => (string)($settings['vid_voice'] ?? ''),
                'data'  => (int)($settings['qos_voice'] ?? 0),
            ],
            'dataVLAN'  => [
                'voice' => (string)($settings['vid_data'] ?? ''),
                'data'  => (int)($settings['qos_data'] ?? 0),
            ],
            'ntp' => [
                'server'   => (string)($settings['ntp_server'] ?? ''),
                'timezone' => (string)($settings['ntp_timezone'] ?? ''),
            ],
            'useStaticIp'  => \Wildix\Validator::isTrue($settings['use_static_ip'] ?? ''),
            'useDst'       => \Wildix\Validator::isTrue($settings['use_dst'] ?? 'yes'),
            'syslogServer' => (string)($settings['syslog_server'] ?? ''),
        ];

        if (Vendors::hasToneZone($model)) {
            $res['toneZone'] = (string)($settings['tonezone'] ?? '');
        }

        if (Vendors::hasSnmp($model)) {
            $res['snmp'] = (string)($settings['snmp'] ?? '');
        }

        return $res;
    }

    /**
     * Get user device.
     *
     * @param string      $newExtension     New user extension.
     * @param string|null $currentExtension Current user extension.
     *
     * @return array
     * @throws \Exception Error.
     */
    private function getDeviceUser(string $newExtension, ?string $currentExtension) :array {
        $user = [
            'oldExtension' => $currentExtension ?? ''
        ];

        if ($this->isEntry($newExtension, 'unknown')) {
            if (!$this->isEntry($user['oldExtension'], 'login')) {
                $user['newExtension'] = $this->getFreeUnknownAccount($this->getUnknownAccounts());
                $user['password']     = self::DEFAULT_PASS;
                $user['id']           = self::DEFAULT_EXT;
            }
        } else {
            $colleague = $this->_colleagues->getUserByExtension($newExtension);
            $pbx       = \Wildix\PBXes\PBXes::getInstance()->getPBXBySerial($this->_pbx->getSerial());

            $user['id']           = $colleague->getID();
            $user['dn']           = $colleague->getDn();
            $user['pbxDN']        = $colleague->getPBXDn();
            $user['newExtension'] = $newExtension;
            $user['password']     = $colleague->getSipPassword() ?: self::DEFAULT_PASS;

            if (!empty($user['pbxDN']) && $pbx->getDn() != $colleague->getPBXDn()) {
                $this->getLuaService()->moveUser([$user['id'], $user['newExtension'], $pbx->getIp()]);

                // update location
                $res = WMS_Ldap::getInstance()->moveUser($user['dn'], $user['pbxDN'], $pbx->getDn());
                if ($res !== true) {
                    return array('Destination User' => 'cant not move from remote PBX');
                }

                // clear local sip peer cache - devices present only on local PBX
                \Wildix\CallweaverManager::getInstance()->sipPeerPrune($user['newExtension']);
            }
        }

        return $user;
    }

    /**
     * Get ip addresses devices.
     *
     * @return array
     */
    private function getListAddressDevices(): array {
        $result = [];
        $query  = 'SELECT value FROM %s WHERE param=\'device\'';
        $rows   = DEVDB::getInstance()->getRows(sprintf($query, DEVDB::TABLE_AUTOPROV));
        foreach ($rows as $row) {
            if (!empty($row['value'])) {
                $device = json_decode($row['value'], true);
                $result[($device['address']) ?? ''] = ($device['address'] ?? '');
            }
        }

        return array_values($result);
    }

    /**
     * Find ready devices in system.
     *
     * @param array $macs Macs.
     *
     * @return array
     */
    private function findReadyDevices(array $macs): array {
        return array_values(array_intersect($this->getReadyMacs(), $macs));
    }

    /**
     * Get free account unknown user.
     *
     * @param array $accounts List unknown accounts.
     *
     * @return string
     */
    private function getFreeUnknownAccount(array $accounts = []) :string {
        $account = 'unknown1';
        $loop    = true;
        $i       = 0;
        while ($loop) {
            $i++;
            $account = 'unknown' . $i;
            if (!isset($accounts[$account]) || $i > 10000) {
                $loop = false;
            }
        }
        return $account;
    }

    /**
     * Get unknown accounts.
     *
     * @return array
     */
    private function getUnknownAccounts(): array {
        $result = [];
        $sql    = 'SELECT value, mac FROM ' . DEVDB::TABLE_AUTOPROV . ' WHERE param=\'username\' AND value like \'unknown%\'';
        $items  = DEVDB::getInstance()->getRows($sql);
        foreach ($items as $item) {
            $result[$item['value']] = $item['mac'];
        }
        return $result;
    }

    /**
     * Is entry string.
     *
     * @param string $base   Base string.
     * @param string $search Sub string.
     *
     * @return boolean
     */
    private function isEntry(string $base, string $search): bool {
        return strpos($base, $search) !== false;
    }

    /**
     * Load and assigned user to devices.
     *
     * @param array $devices Devices.
     *
     * @return array
     */
    private function assignedUsersToDevices(array $devices): array {
        $pattern = 'SELECT mac, value, flag as port FROM %s WHERE mac IN (\'%s\') AND param=\'username\'';
        $users   = DEVDB::getInstance()->getRows(sprintf($pattern, DEVDB::TABLE_AUTOPROV, implode('\', \'', array_keys($devices))));

        foreach ($devices as $key => $device) {
            $deviceUsers = array_filter(
                $users,
                function ($item) use ($device) {
                    return $item['mac'] === $device->getMac();
                }
            );

            $sortColumn = array_column($deviceUsers, 'port');
            array_multisort($sortColumn, SORT_ASC, $deviceUsers);

            $device->setAssignedUsers(
                array_map(
                    function ($item) {
                        return $item['value'];
                    },
                    $deviceUsers
                )
            );

            $devices[$key] = $device;
        }

        return $devices;
    }

    /**
     * Execute Command sync device.
     *
     * @param string  $action Action.
     * @param array   $items  Mac address and Ip address.
     * @param boolean $async  Async mode (default = true).
     *
     * @return void
     */
    private function executeCommandSync(string $action, array $items, bool $async = true): void {
        $addresses = implode(',', array_column($items, 'address'));
        $macs      = implode(',', array_column($items, 'mac'));
        $command   = sprintf(($async === true ? self::SYNC_DEVICE : self::SYNC_DEVICE_SYNCHRONOUSLY), escapeshellcmd($action), $addresses, $macs);
        $this->getLog()->info(__LINE__, $command);
        exec($command);
    }

    /**
     * Execute command scanning ip range.
     *
     * @param string $startIpAddress Start ip range.
     * @param string $endIpAddress   End ip range.
     *
     * @return string
     */
    private function executeCommandScanning(string $startIpAddress, string $endIpAddress): string {
        $output = [];
        $start  = ip2long($startIpAddress);
        $end    = ip2long($endIpAddress);
        $range  = implode(' ', array_map('long2ip', range($start, $end)));
        $opts   = '-sS';

        if ($this->isRawScan()) {
            $opts = '--unprivileged -sT';
        }

        $cmd = 'nohup sudo nmap %s -Pn -n -T4 --max-rate 100 --open -p443,80 -oN - %s| grep "for \|MAC" -A5';
        exec(sprintf($cmd, $opts, $range), $output);

        return implode("\n", $output);
    }

    /**
     * Connect device.
     *
     * @param string $macAddress Mac address.
     *
     * @return array
     * @throws InvalidDataException Invalid data exception.
     */
    public function connect(string $macAddress) :array {
        $macAddress = $this->_clearMac($macAddress);
        if (!Network::isMac($macAddress)) {
            throw new InvalidDataException(Error::getMessage(Error::MAC_ADDRESS), []);
        }
        $current = $this->getDeviceByMac($macAddress, [], true);

        $address = $current->getIpAddress();

        if (empty($address)) {
            $this->getLog()->error(__LINE__, Error::getMessage(Error::InvalidIpAddress . $macAddress));
            throw new InvalidDataException(Error::getMessage(Error::InvalidIpAddress), []);
        }
        $devices           = $this->getTproxyDevices();
        $devices[$address] = [
            'name' => 'http',
            'ssl' => '1',
            'host' => $address,
            'port' => '443',
            'dump' => '0',
            'opt' => sprintf('%s.%s', rand(100000, 999999), $this->_pbx->getDomain())
        ];

        $listDevices = $this->getListAddressDevices();
        $this->saveTproxyDevices($devices, $listDevices);

        return $devices[$address];
    }

    /**
     * Disconnect device.
     *
     * @param string $macAddress Mac address.
     *
     * @return void
     * @throws InvalidDataException Invalid data exception.
     */
    public function disconnect(string $macAddress) :void {
        $macAddress = $this->_clearMac($macAddress);
        if (!Network::isMac($macAddress)) {
            throw new InvalidDataException(Error::getMessage(Error::MAC_ADDRESS), []);
        }
        $current = $this->getDeviceByMac($macAddress, [], true);
        $devices = $this->getTproxyDevices();
        unset($devices[$current->getIpAddress()]);
        $listDevices = $this->getListAddressDevices();

        $this->saveTproxyDevices($devices, $listDevices);
    }

    /**
     * Parse ip range.
     *
     * @param string $range Range ip.
     *
     * @return array
     * @throws \Wildix\InvalidDataException Invalid data exception.
     */
    public function parseIpRange(string $range) :array {
        $ranges = [
        'start' => '', 'end' => ''
        ];

        if (Network::isIp($range)) {
            $ranges['start'] = trim($range);
            $ranges['end']   = trim($range);
            return $ranges;
        }


        $range = explode('-', $range);
        if (Network::isIp(trim($range[0])) && Network::isIp(trim($range[1]))) {
                $ranges['start'] = trim($range[0]);
                $ranges['end']   = trim($range[1]);
        } else {
            $tmp   = explode('.', $range[0]);
            $endIp = (isset($range[1]) && isset($tmp[3])) ? implode('.', [$tmp[0], $tmp[1], $tmp[2], $range[1]]) : '';
            if (Network::isIp(trim($range[0]))
                && Network::isIp(trim($endIp))
                && $tmp[3] < $range[1]
            ) {
                $ranges['start'] = trim($range[0]);
                $ranges['end']   = trim($endIp);
            } else {
                throw new \Wildix\InvalidDataException(
                    \Wildix\Error::getMessage(\Wildix\Error::InvalidData),
                    [
                        'range' => [
                            'code' => \Wildix\Error::InvalidIpAddress,
                            'reason' => \Wildix\Error::getMessage(\Wildix\Error::InvalidIpAddress)
                        ]
                    ]
                );
            }
        }

        return $ranges;
    }

    /**
     * Scanning ip address range.
     *
     * @param string $startIpAddress Start ip range.
     * @param string $endIpAddress   End ip range.
     *
     * @return array
     */
    public function scanIpAddressRange(string $startIpAddress, string $endIpAddress) :array {
        $items    = [];
        $scanning = $this->executeCommandScanning($startIpAddress, $endIpAddress);

        $readyItems = array_filter(
            $this->getItems(0, 1000),
            function ($item) {
                return $item->isProvision();
            }
        );
        $readyMacs  = array_map(
            function ($item) {
                return $item->getMac();
            },
            $readyItems
        );
        $readyMacs  = array_flip($readyMacs);

        $output = explode('Nmap scan report for ', $scanning);
        foreach ($output as $key => $item) {
            $ipAddress = '';
            $port      = '';
            $mac       = '';
            $model     = '';
            preg_match_all('/(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})|([0-9A-Fa-f\:]{17})|(80)\/tcp|(443)\/tcp/si', $item, $matches);

            foreach ($matches[0] as $tmp) {
                $intTmp = (int)$tmp;
                if (Network::isIp($tmp)) {
                    $ipAddress = $tmp;
                } else if ($intTmp === 80 || $intTmp === 443) {
                    $port = $intTmp;
                } else if (Network::isMac($this->_clearMac($tmp))) {
                    $mac    = $this->_clearMac($tmp);
                    $device = Vendors::getDeviceInfoByMac($tmp);
                    $model  = $device['modelLine'] !== Vendors::UNKNOWN ? $device['modelLine'] : '';
                }
            }
            if (!empty($ipAddress)
                && !empty($port)
                && !isset($readyMacs[$mac])
                && (                (!empty($mac) && !empty($model))
                || empty($mac) && empty($model))
            ) {
                $items[] = [
                    'address' => $ipAddress,
                    'port' => $port,
                    'mac' => $mac,
                    'model' => strtolower($model)
                ];
            }
        }

        return $items;
    }

    /**
     * Verification device.
     *
     * @param string  $ip   Ip address.
     * @param integer $port Port.
     *
     * @return array|null
     * @throws InvalidDataException Invalid data exception.
     */
    public function verifyDevices(string $ip, int $port) :?array {
        if (!Network::isIp($ip) || ($port !== 80 && $port !== 443)) {
            throw new InvalidDataException(Error::getMessage(Error::InvalidData), []);
        }
        $device = $this->getLuaService()->deviceVerify($ip, $port);

        $item = $this->getDeviceByMac($device['mac']);

        if (!empty($device) && !empty($device['mac']) && !empty($item) && $item->isProvision()) {
            $device['allow'] = false;
        }
        return $device;
    }

    /**
     * Is raw scan.
     *
     * @return boolean Raw scan.
     */
    private function isRawScan() :bool {
        $provConfFile = '/rw2/etc/provision.conf';
        $provConf     = is_file($provConfFile) ? parse_ini_file($provConfFile) : null;
        return is_array($provConf) && $provConf['RawScan'] && $provConf['RawScan'] === 'disable';
    }

    /**
     * Get TProxy Device.
     *
     * @return array
     */
    private function getTproxyDevices() :array {
        $config = $this->_tProxyService->getConfig();
        foreach ($config as $line) {
            $host = $line['host'];
            $opt  = $line['opt'];
            if (!empty($host) && !empty($opt)) {
                $opt         = explode(':', $opt)[1];
                $line['opt'] = trim($opt);
                $this->_tProxyDevices[$host] = $line;
            }
        }

        return $this->_tProxyDevices;
    }

    /**
     * Save Tproxy Device
     *
     * @param array $devices       Devices.
     * @param array $syncAddresses Sync addresses.
     *
     * @return array
     */
    private function saveTproxyDevices(array $devices, array $syncAddresses) :array {
        $allDevices = $this->getTproxyDevices();

        foreach ($allDevices as $key => $device) {
            $this->_tProxyService->removeHttpProxying($device['host']);
        }

        foreach ($devices as $key => $device) {
            if (!in_array($key, $syncAddresses)) {
                continue;
            }

            $this->_tProxyService->addHttpProxying($device['host'], 'Host: ' . $device['opt']);
        }

        $this->_tProxyService->saveAndReload();

        return $devices;
    }

    /**
     * Get connect url.
     *
     * @param string $ipAddress IP A\address.
     *
     * @return string
     */
    private function getConnectUrl(string $ipAddress) :string {
        $connect = '';
        $tProxy  = $this->getTproxyDevices();
        if (isset($tProxy[$ipAddress])) {
            $connect = $tProxy[$ipAddress]['opt'] ?? '';
        }
        return $connect;
    }

    /**
     * Get port count.
     *
     * @param string $model Device Model.
     *
     * @return integer
     */
    private function getPortCount(string $model) :int {
        if (empty($this->_models)) {
            $this->_models = $this->getLuaService()->getModels();
        }

        return $this->_models[strtoupper($model)] ?? 1;
    }

    /**
     * Get device state.
     *
     * @param Device $device Entity device.
     *
     * @return string
     * @throws \Exception Exception
     */
    private function getState(Device $device) :string {
        $remote = $this->getRemoteFirmwares();
        $local  = $this->getLocalFirmwares();
        $state  = 'on';
        $value  = $device->getValue();

        if (!empty($value['vendor:additional']) && isset($remote[$value['vendor:additional']])) {
            $state = 'fw';
        } else if (!empty($local[$device->getModel()])) {
            if ($value['fw:version']
                && $value['fw:version'] !== $local[$value['model']]
                && !in_array($local[$value['model']], ['2.07l', '2.07q'])
            ) {
                $state = 'fwlocal';
            }
        }

        if ($value['vendor'] === 'wildix' && !empty($value['firmware']) && $this->isEntry($value['firmware'], '.3.10.5')) {
            $state = 'phunsup';
        }
        return $state;
    }

    /**
     * Ger remote firmwares.
     *
     * @return array
     * @throws \Exception Exception
     */
    public function getRemoteFirmwares(): array {
        if (!is_array($this->_remoteFirmwares)) {
            $data            = [];
            $upgradeSettings = $this->getUpgradeSettings()->getSettings();
            $port            = self::FIRMWARE_DEFAULT_PORT;
            $firmwareHost    = self::FIRMWARE_DEFAULT_HOST;
            $scheme          = self::FIRMWARE_DEFAULT_SCHEME;
            if (env('SERVICE_FIRMWARE_URI') !== null) {
                $customServer = parse_url(env('SERVICE_FIRMWARE_URI'));
                $scheme       = isset($customServer['scheme']) ? $customServer['scheme'] . '://' : $scheme;
                $firmwareHost = $customServer['host'] ?? $firmwareHost;
                $port         = $customServer['port'] ?? $port;
            }
            if (is_file(self::FIRMWARE_LOCAL_CACHE_PATH)) {
                $data = json_decode(file_get_contents(self::FIRMWARE_LOCAL_CACHE_PATH), true);
            }

            if (isset($data['cacheDate']) && (time() - $data['cacheDate']) <= self::FIRMWARE_INFO_LOCAL_TTL) {
                $this->_remoteFirmwares = $data;
                return $this->_remoteFirmwares;
            }

            if (filter_var($upgradeSettings['customFirmwareServer']['enabled'], FILTER_VALIDATE_BOOLEAN)) {
                $customServer = parse_url($upgradeSettings['customFirmwareServer']['url']);
                $scheme       = isset($customServer['scheme']) ? $customServer['scheme'] . '://' : $scheme;
                $firmwareHost = $customServer['host'] ?? $firmwareHost;
                $port         = $customServer['port'] ?? $port;
            }
            $context = $this->getStreamContext();
            $url     = sprintf('%s%s%s%s', $scheme, $firmwareHost, $port === '' ? '' : ':' . $port, self::FIRMWARE_DEFAULT_PATH);
            $data    = file_get_contents($url, false, $context);
            $this->getLog()->info(__LINE__, 'Generate firmware info url:' . $url);
            if ($data === false) {
                $error = error_get_last();
                $this->getLog()->error(__LINE__, sprintf('Can not fetch firmware info from url "%s" error: %s', $url, $error ? $error['message'] : 'Unknown errors'));
                $this->_remoteFirmwares = [];

                return $this->_remoteFirmwares;
            }
            $data                   = json_decode($data, true);
            $data['cacheDate']      = time();
            $this->_remoteFirmwares = $data;
        }

        return $this->_remoteFirmwares;
    }

    /**
     * Get stream context
     *
     * @return null|resource
     * @throws \RuntimeException If no ssl extension
     */
    private function getStreamContext() {
        $proxySettings  = $this->getHttpProxy()->getSettings();
        $contextOptions = [
            'http' => [
                'timeout' => self::FIRMWARE_FETCH_TIMEOUT
            ]
        ];

        if (!isset($proxySettings) || !\Wildix\Validator::isTrue($proxySettings['enable'])) {
            return stream_context_create($contextOptions);
        }

        $proxyUrl  = parse_url($proxySettings['host']);
        $proxyUrl  = isset($proxyUrl['scheme']) ? $proxyUrl['scheme'] . '://' : '';
        $proxyUrl .= $proxySettings['host'];

        if (!empty($proxySettings['port'])) {
            $proxyUrl .= ':' . $proxySettings['port'];
        } else if (strpos($proxyUrl, 'http://') === 0) {
            $proxyUrl .= ':80';
        } else if (strpos($proxyUrl, 'https://') === 0) {
            $proxyUrl .= ':443';
        }

        $proxyUrl = str_replace(array('http://', 'https://'), array('tcp://', 'ssl://'), $proxyUrl);

        if (strpos($proxyUrl, 'ssl:') === 0 && !extension_loaded('openssl')) {
            throw new \RuntimeException('You must enable the openssl extension to use a proxy over https');
        }

        $contextOptions['http']['proxy']           = $proxyUrl;
        $contextOptions['http']['request_fulluri'] = true;

        // proxy auth if needed
        if (!empty($proxySettings['login'])) {
            $auth = $proxySettings['login'];
            if (!empty($proxySettings['password'])) {
                $auth .= ':' . $proxySettings['password'];
            }
            $contextOptions['http']['header'] = sprintf('Proxy-Authorization: Basic %s\r\n', base64_encode($auth));
        }

        return stream_context_create($contextOptions);
    }

    /**
     * Get locale firmwares.
     *
     * @return array
     */
    public function getLocalFirmwares(): array {
        if (empty($this->_localFirmwares) && is_readable(self::LOCAL_DIR_FIRMWARE)) {
            $configs = glob(self::LOCAL_DIR_FIRMWARE . '*/firmwares.json');
            $configs = array_filter($configs, 'is_file');
            $out     = [];
            foreach ($configs as $config) {
                foreach (json_decode(file_get_contents($config), true) as $model => $fwData) {
                    $out[$model] = $fwData['version'];
                }
            }
            $this->_localFirmwares = $out;
        }
        return $this->_localFirmwares;
    }

    /**
     * Get NAT ips.
     *
     * @return array
     */
    private function getNatIps() :array {
        return PBX::getInstance()->getNatIps();
    }

    /**
     * Sync device via sip.
     *
     * @param Device $device Device.
     * @param array  $users  Users.
     *
     * @return void
     */
    private function syncDeviceViaSipNotify(Device $device, array $users = []): void {
        if (empty($users)) {
            $this->assignedUsersToDevices([$device->getMac() => $device]);
        } else {
            $device->setAssignedUsers($users);
        }

        $res           = false;
        $assignedUsers = $device->getAssignedUsers();

        if ($assignedUsers) {
            $res = PBX::getInstance()->sip_notify_check_sync($assignedUsers, $device->isWpaFamily());
        } else {
            $this->getLog()->info(__LINE__, 'Send SIP notify check-sync. No assigned users');
            $this->executeActionProvisionSync($device);
        }

        $this->getLog()->info(
            __LINE__,
            sprintf(
                'Sync devices via sip notify: res=%s, mac=%s, users=%s',
                $res,
                $device->getMac(),
                implode(',', $assignedUsers)
            )
        );
    }

    public function isDisableSync(string $key): bool {
        $config = new \Wildix\Settings\ConfigFile(self::DEVICE_SYNC_CONFIG);
        $res    = (boolean)$config->getBoolean($key);
        if ($res) {
            $this->getLog()
                ->info(
                    __LINE__,
                    sprintf('Device sync is disabled by option "%s" in config "%s".', self::DEVICE_SYNC_MANUAL, self::DEVICE_SYNC_CONFIG)
                );
        }
        return $res;
    }

    /**
     * Get Received.
     *
     * @param string $mac Device mac addr.
     *
     * @return string
     */
    private function getReceived(string $mac) :string {
        $registeredItems = $this->getRegisteredItems();

        if (isset($registeredItems[$mac])) {
            return implode(' ', $registeredItems[$mac]['received']);
        }
        return '';
    }

    /**
     * Is online device.
     *
     * @param Device $device Device object.
     *
     * @return boolean
     */
    private function isOnline(Device $device) :bool {
        $registeredItems = $this->getRegisteredItems();
        return $device->getIpAddress() && $device->getMac() && isset($registeredItems[$device->getMac()]);
    }

    /**
     * Get registered devices
     *
     * @return array
     */
    private function getRegisteredItems() :array {
        if (!is_array($this->_registeredItems)) {
            $res     = [];
            $pattern = [
                'extends' => [],
                'userAgent' => '',
                'received' => [],
            ];
            $trunks  = $this->getSipProxy()->getTrunksRegistrations();
            $devices = $this->getSipProxy()->getUsersRegistrations();


            foreach ($trunks as $userExtension => $item) {
                if (!empty($item)
                    && !empty($item['registrations'])
                    && preg_match('/^(.*\s)+([0-9a-fA-F]{12}).*$/', $item['registrations'][0]['useragent'], $matches)
                    && $matches
                    && $matches[2]
                ) {
                    $mac = $this->_clearMac($matches[2]);
                    if (!isset($res[$mac])) {
                        $res[$mac] = $pattern;
                    }
                    foreach ($item['registrations'] as $regInfo) {
                        $res[$mac]['extends'][]  = $userExtension;
                        $res[$mac]['userAgent']  = $regInfo['useragent'];
                        $res[$mac]['received'][] = $regInfo['received'];
                    }
                }
            }

            foreach ($devices as $userExtension => $item) {
                $items = isset($item['registrations']) ? $item['registrations'] : [];
                array_map(
                    function ($item) use (&$res, $userExtension, $pattern) {
                        if (preg_match('/^(.*\s)+([0-9a-fA-F]{12}).*$/', $item['useragent'], $matches) && $matches && $matches[2]) {
                            $mac = $this->_clearMac($matches[2]);
                            if (!isset($res[$mac])) {
                                $res[$mac] = $pattern;
                            }
                            $res[$mac]['extends'][]  = $userExtension;
                            $res[$mac]['userAgent']  = $item['useragent'];
                            $res[$mac]['received'][] = $item['received'];
                        }
                        return $item;
                    },
                    $items
                );
            }
            $this->_registeredItems = $res;
        }

        return $this->_registeredItems;
    }

    /**
     * Is local address.
     *
     * @param string $address Ip address.
     *
     * @return boolean
     */
    private function isLocalAddress(string $address) :bool {
        if (CIDR::IPisWithinCIDR($address, '10.0.0.0/8')
            || CIDR::IPisWithinCIDR($address, '172.16.0.0/12')
            || CIDR::IPisWithinCIDR($address, '192.168.0.0/16')
        ) {
            return true;
        }
        return false;
    }

    /**
     * Clearing delimiter characters in mac address.
     *
     * @param string $mac Dirty mac address.
     *
     * @return string
     */
    private function _clearMac(string $mac): string {
        return strtolower(str_replace([':', '-'], '', $mac));
    }

    private function removeTrunkByDeviceMac(string $macAddress): void {
        $trunkId          = $this->getTrunkId($macAddress);
        if ($trunkId > 0) {
            $deleteTrunkQuery = sprintf('DELETE FROM %s WHERE id = %s', WMSDB::TABLE_PSTN_TRUNKS, $trunkId);
            WMSDB::getInstance()->doTransaction([$deleteTrunkQuery]);
        }
    }

    /**
     * Check validation bind item.
     *
     * @param array $item Bind item.
     *
     * @return boolean
     */
    private function isValidBindItem(array $item) {
        return $item['mac'] && $item['address'] && Network::isMac($item['mac']) && Network::isIp($item['address']);
    }

    public function getTrunkId(string $mac): int {
        $query = DEVDB::getInstance()->first(
            sprintf(
                'SELECT `%s` FROM `%s` WHERE `%s`=\'%s\' and mac=\'%s\'',
                Device::VALUE,
                DEVDB::TABLE_AUTOPROV,
                Device::PARAM,
                'trunkid',
                $mac
            )
        );
        return $query ? (int)$query[Device::VALUE] : 0;
    }

    public function getTrunkName(int $id): string {
        $query = WMSDB::getInstance()->first(
            sprintf(
                'SELECT `%s` FROM `%s` WHERE `%s`=%d',
                PstnTrunk::USERNAME,
                WMSDB::TABLE_PSTN_TRUNKS,
                PstnTrunk::ID,
                $id
            )
        );

        return $query ? $query[PstnTrunk::USERNAME] : '';
    }

    public function unassignDevicesByExtension(string $extension): void {
        DEVDB::getInstance()->exec('update autoprov set value="unknown" where param="username" and value = "' . $extension . '"');
        PBX::getInstance()->sip_notify_check_sync([$extension]);
    }
}
