<?php
namespace Wildix\Devices\Networks;

use Wildix\Databases\DEVDB;
use Wildix\Databases\USRLOC;
use Wildix\Devices\Device;

/**
 * Class NetworkStorage
 */
class NetworkStorage {
    // phpcs:disable Generic.Files.LineLength.TooLong
    private const NETWORKS        = 'SELECT * FROM ' . DEVDB::TABLE_WAIR_NETWORKS;
    private const AVAILABLE_GW    = 'SELECT * FROM `' . DEVDB::TABLE_AUTOPROV . '` WHERE param=\'device\'';
    private const ALL_GATEWAYS    = 'SELECT * FROM `' . DEVDB::TABLE_WAIR_GATEWAYS . '`';
    private const GATEWAYS        = self::ALL_GATEWAYS . ' AS g INNER JOIN `' . DEVDB::TABLE_AUTOPROV . '` AS d ON (lower(g.mac) = lower(d.mac) AND d.param=\'device\')';
    private const LOCATION_USERS  = 'SELECT * FROM ' . USRLOC::TABLE_LOCATION . '%s';
    private const INSERT_GATEWAYS = 'INSERT INTO ' . DEVDB::TABLE_WAIR_GATEWAYS . ' (`network_id`,`mac`,`type`) VALUES (%s, "%s", %s)';
    private const USERS_UPDATE    = 'UPDATE ' . DEVDB::TABLE_WAIR_USERS . ' SET "mac"="%s" WHERE "mac"="%s"';
    private const NETWORK_UPDATE  = 'UPDATE ' . DEVDB::TABLE_WAIR_NETWORKS . ' SET %s WHERE id=%s';
    private const REMOVE_PATTERN  = 'DELETE FROM %s WHERE %s="%s"';
    // phpcs:enable Generic.Files.LineLength.TooLong

    private const W_AIR_FAMILY = [
        'W-AIR',
        'W-AIR-SP',
    ];

    private $_db;
    private $_usrDb;

    private function _db(): DEVDB {
        if (!$this->_db) {
            $this->_db = DEVDB::getInstance();
        }
        return $this->_db;
    }

    private function _usrDb(): USRLOC {
        if (!$this->_usrDb) {
            $this->_usrDb = USRLOC::getInstance();
        }
        return $this->_usrDb;
    }

    public function getNetwork(int $identity): ?Network {
        $query = self::NETWORKS . ' WHERE id=' . $identity;
        $item  = $this->_db()->first($query);
        if ($item) {
            $item = new Network($item);
        }
        return $item;
    }

    public function getNetworks(): array {
        return $this->_db()->getRows(self::NETWORKS, null, Network::class);
    }

    public function getAvailableGateways(bool $onlyNotUsed = true): array {
        $available = [];
        $query     = self::AVAILABLE_GW;
        if ($onlyNotUsed) {
            $gateways = $this->_db()->getRows(self::ALL_GATEWAYS, 'mac');

            if (!empty($gateways)) {
                $query .= sprintf(' AND mac NOT IN (\'%s\')', implode('\', \'', array_keys($gateways)));
            }
        }
        $items = $this->_db()->getRows($query, null, Device::class);
        foreach ($items as $item) {
            $vendor = $item->getValue();
            if (in_array(strtoupper($item->getModel()), self::W_AIR_FAMILY) && $vendor['vendor:additional'] !== 'W-AIR SB') {
                $available[] = $item;
            }
        }
        return $available;
    }

    public function getGateways(): array {
        return $this->_db()->getRows(self::GATEWAYS, 'mac', Device::class);
    }

    public function getGatewaysUsers(array $macs): array {
        $where  = '';
        $filter = [];
        $result = [];
        if (!empty($macs)) {
            foreach ($macs as $mac) {
                $filter[] = 'user_agent LIKE "%' . $mac . '%"';
            }
            $where .= ' WHERE ' . implode(' OR ', $filter);
            $query  = sprintf(self::LOCATION_USERS, $where);
            $result = $this->_usrDb()->getRows($query);
        }

        return $result;
    }

    public function create(array $network, $gateways = []): void {
        $this->_db()->insert(DEVDB::TABLE_WAIR_NETWORKS, $network);
        $this->_updateGateways(
            (int)$this->_db()->getLastInsertId(),
            $gateways
        );
    }

    public function update(int $identity, array $network, $gateways = []): void {
        $params = [];
        foreach ($network as $key => $value) {
            $params[] = '"' . $this->_db()->realEscapeString($key) . '"="' . (string)$this->_db()->realEscapeString($value) . '"';
        }
        $this->_db()->exec(sprintf(self::NETWORK_UPDATE, implode(', ', $params), $identity));
        $this->_updateGateways($identity, $gateways);
    }

    public function remove(int $identity): void {
        $queries = [
            sprintf(self::REMOVE_PATTERN, DEVDB::TABLE_WAIR_GATEWAYS, 'network_id', $identity),
            sprintf(self::REMOVE_PATTERN, DEVDB::TABLE_WAIR_NETWORKS, 'id', $identity),
        ];
        $this->_db()->doTransaction($queries);
    }

    private function _updateGateways(int $networkId, array $gateways): void {
        $oldPrimary = $this->_getOldPrimary($networkId);
        $newPrimary = isset($gateways[0]) ? $gateways[0]['gwid'] : '';
        $queries    = [sprintf(self::REMOVE_PATTERN, DEVDB::TABLE_WAIR_GATEWAYS, 'network_id', $networkId)];
        if (!empty($newPrimary)) {
            $queries[] = sprintf(self::REMOVE_PATTERN, DEVDB::TABLE_WAIR_GATEWAYS, 'mac', $newPrimary);
        }

        foreach ($gateways as $gateway) {
            $queries[] = sprintf(self::INSERT_GATEWAYS, $networkId, $gateway['gwid'], $gateway['type']);
        }
        if (!empty($oldPrimary) && !empty($newPrimary) && $newPrimary != $oldPrimary) {
            $queries[] = sprintf(self::REMOVE_PATTERN, DEVDB::TABLE_WAIR_USERS, 'mac', $newPrimary);
            $queries[] = sprintf(self::USERS_UPDATE, $newPrimary, $oldPrimary);
        }

        $this->_db()->doTransaction($queries);
    }

    private function _getOldPrimary(int $networkId): string {
        $res = $this->_db()->first(sprintf(self::GATEWAYS . ' WHERE `network_id`=%s AND type=1', $networkId));
        return !empty($res) ? $res['mac'] : '';
    }
}
