<?php
namespace Wildix\Devices\Networks;

use JsonSchema\Exception\ValidationException;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\Validator;

/**
 * Class NetworksService
 */
class NetworksService {

    private $_storage;

    public function __construct(NetworkStorage $storage) {
        $this->_storage = $storage;
    }

    /**
     * Network data storage.
     *
     * @return NetworkStorage
     */
    private function _storage(): NetworkStorage {
        return $this->_storage;
    }

    /**
     * Get networks list.
     *
     * @return array
     */
    public function getItems(): array {
        $networks = $this->_storage()->getNetworks();
        $gateways = $this->_storage()->getGateways();
        $users    = $this->_storage()->getGatewaysUsers(array_keys($gateways));
        foreach ($networks as $network) {
            $network->setGateways($this->_filteredGatewaysByNetworkId($network->getId(), $gateways));
            $network->setUsers($this->_filteredUsersByGateways($network->getGateways(), $users));
        }
        return $networks;
    }

    /**
     * Get network.
     *
     * @param integer $identity Identity.
     * @param boolean $check    Has exeption if empty.
     *
     * @return array|null Network.
     * @throws InvalidDataException InvalidDataException.
     */
    public function getItem(int $identity, bool $check = false): ?Network {
        $network = $this->_storage()->getNetwork($identity);
        if (empty($network) && $check) {
            throw new InvalidDataException(Error::NOT_FOUND_DEVICE, []);
        }
        return $network;
    }

    /**
     * Create network.
     *
     * @param array $data Network data.
     *
     * @return void
     * @throws InvalidDataException InvalidDataException.
     */
    public function create(array $data = []): void {
        $this->_validate($data);
        $data = $this->_prepareToSave($data);
        $this->_storage()->create($data['network'], $data['gateways']);
    }

    /**
     * Update network by id.
     *
     * @param integer $identity Network id.
     * @param array   $data     Network data.
     *
     * @return void
     * @throws InvalidDataException InvalidDataException.
     */
    public function update(int $identity, array $data = []): void {
        $this->getItem($identity, true);
        $this->_validate($data);
        $data = $this->_prepareToSave($data);
        $this->_storage()->update($identity, $data['network'], $data['gateways']);
    }

    /**
     * Remove network.
     *
     * @param integer $identity Network id.
     *
     * @return boolean
     * @throws InvalidDataException InvalidDataException.
     */
    public function remove(int $identity): bool {
        $this->getItem($identity, true);
        $this->_storage()->remove($identity);
        return true;
    }

    /**
     * Prepare network to api returned.
     *
     * @param Network $network Array data from database.
     *
     * @return array
     */
    public function buildItem(Network $network): array {
        $users    = [];
        $gateways = [];
        foreach ($network->getGateways() as $gateway) {
            $gateways[] = [
                'mac' => (string)$gateway['mac'],
                'isMaster' => (boolean)$gateway['type'],
                'address' => (string)$gateway['address'],
                'model' => (string)$gateway['model']
            ];
        }
        foreach ($network->getUsers() as $user) {
            $users[] = [
                'extension' => (string)$user['username'],
                'register' => true,
                'iPai' => trim(substr($user['user_agent'], -14))
            ];
        }
        return [
            'id' => $network->getId(),
            'title' => $network->getTitle(),
            'code' => $network->getCode(),
            'proxy' => $network->getProxy(),
            'autoTree' => $network->isAutoTreeDetect(),
            'syncInternet' => $network->isSyncInternet(),
            'baseIp' => $network->getBaseIp(),
            'gateways' => $gateways,
            'users' => $users,
        ];
    }

    /**
     * Prepare data to save
     *
     * @param array $data Data.
     *
     * @return array
     */
    private function _prepareToSave(array $data): array {
        $res = [
            'network' => [],
            'gateways' => []
        ];

        $res['network'] = [
            'name' => (string)$data['title'] ?? '',
            'code' => (string)$data['code'],
            'sipproxy' => (string)$data['proxy'] ?? '',
            'primarybase' => (string)$data['baseIp'] ?? '',
            'autodecttree' => (integer)$data['autoTree'] ?? 0,
            'sync_plus_via_ethernet' => (integer)$data['syncInternet'] ?? 0
        ];

        if (isset($data['gateways']) && is_array($data['gateways'])) {
            foreach ($data['gateways'] as $key => $gateway) {
                $res['gateways'][] = [
                    'gwid' => $gateway['mac'],
                    'type' => (integer)($key === 0)
                ];
            }
        }

        return $res;
    }

    /**
     * Filtered gateways by network id.
     *
     * @param integer $networkId Network id.
     * @param array   $gateways  Gateways.
     *
     * @return array
     */
    private function _filteredGatewaysByNetworkId(int $networkId, array $gateways): array {
        $result = [];
        foreach ($gateways as $gateway) {
            $data = $gateway->getAsArray();
            if ($networkId === $data['network_id']) {
                $result[] = [
                    'mac' => $gateway->getMac(),
                    'address' => $gateway->getIpAddress(),
                    'type' => $data['type'],
                    'model' => strtoupper($gateway->getModel()),
                ];
            }
        }
        return $result;
    }

    /**
     * Filtered users by gateways.
     *
     * @param array $gateways Gateways.
     * @param array $users    Users.
     *
     * @return array
     */
    private function _filteredUsersByGateways(array $gateways, array $users): array {
        $res = [];
        foreach ($gateways as $gateway) {
            $users = array_filter(
                $users,
                function ($user) use ($gateway) {
                    return stristr($user['user_agent'], $gateway['mac']) !== false;
                }
            );

            if (!empty($users)) {
                array_push($res, ...$users);
            }
        }
        return $res;
    }

    /**
     * Validate
     *
     * @param array $data Data.
     *
     * @return void
     *
     * @throws InvalidDataException InvalidDataException
     */
    private function _validate(array $data): void {
        $validator = new Validator($data);
        $validator->rule('required', ['title', 'code']);
        $validator->rule('length', 'code', 4)->reason('{field} must be 4 digits');
        $validator->rule('min', 'code', 0);
        $validator->rule('max', 'code', 9999);
        $validator->rule('numeric', 'code');
        $validator->rule('array', ['gateways']);

        if (isset($data['proxy']) && !\Wildix\System\Network::isIp($data['proxy'])) {
            $validator->rule('url', ['proxy']);
        }

        if (isset($data['gateways']) && is_array($data['gateways'])) {
            foreach ($data['gateways'] as $key => $gateway) {
                $validator->rule('macAddress', ['gateways.' . $key . '.mac']);
            }
        }

        $validator->validate();

        if (!empty($validator->errors())) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $validator->errors());
        }
    }
}
