<?php
declare(strict_types=1);

namespace Wildix;

use Cache\FileCache;
use Logger\Log;
use Wildix\Auth\Api\CustomerApplications;
use Wildix\Auth\Applications\Interfaces\SimpleTokenServiceInterface;
use Wildix\CallHistory\Calls;
use Wildix\Databases\WMSDB;
use Wildix\DependencyInjection\ConfigInterface;
use Wildix\DependencyInjection\Container;
use Wildix\Dialplan\CallGroups\CallGroupsService;
use Wildix\Dialplan\PagingGroups\PagingGroupsService;
use Wildix\Phonebooks\Phonebooks;
use Wildix\VoiceMail\VoiceMails;

/**
 * Class DiConfig
 */
final class DiConfig implements ConfigInterface {

    public static function configure(Container $container): void {
        $container->{Validator::class} = function () {
            return new Validator();
        };
        $container->singleton(
            \WMS_Ldap::class,
            static function (): \WMS_Ldap {
                return \WMS_Ldap::getInstance();
            }
        );
        $container->singleton(
            License::class,
            function (Container $container) {
                $license = License::getInstance();
                $license->setWmsLdap($container->get(\WMS_Ldap::class));
                return $license;
            }
        );

        $container->singleton(
            ChatHistoryCleaner::class,
            function () {
                return new ChatHistoryCleaner();
            }
        );

        $container->singleton(
            UserDataCleaner::class,
            function (Container $container) {
                return new UserDataCleaner(
                    $container->get(Calls::class),
                    $container->get(ChatHistoryCleaner::class),
                    $container->get(VoiceMails::class),
                    $container->get(Phonebooks::class),
                    $container->get(CallweaverManager::class),
                    $container->get(Devices\Devices::class),
                    $container->get(Colleagues\Colleagues::class),
                    $container->get(CallGroupsService::class),
                    $container->get(PagingGroupsService::class),
                    $container->get(SimpleTokenServiceInterface::class),
                    $container->get(CustomerApplications::class),
                    $container->get(Log::class)
                );
            }
        );

        $container->singleton(
            Ejabberd::class,
            function (Container $container) {
                $ejabberd = Ejabberd::getInstance();
                $ejabberd->setLog($container->get(Log::class));

                return $ejabberd;
            }
        );

        $container->singleton(
            CompanyGlobalConfig::class,
            function () {
                return new CompanyGlobalConfig();
            }
        );

        $container->singleton(
            CallweaverManager::class,
            function () {
                return CallweaverManager::getInstance();
            }
        );

        $container->singleton(
            Dictionaries::class,
            function (Container $container) {
                return new Dictionaries(
                    $container->get(\Wildix\Dialplan\ContextRepository::class)
                );
            }
        );

        $container->singleton(
            CallRecordings::class,
            function () {
                $syslog = new \Syslog('Faxes');
                return new CallRecordings($syslog);
            }
        );

        $container->singleton(
            Smtp::class,
            function (Container $container) {
                return new Smtp(
                    $container->get(WMSDB::class),
                    $container->get(Log::class),
                    env('SERVICE_SMTP_URI'),
                    $container->get(WhiteLabel::class)->getPbxName()
                );
            }
        );

        $container->singleton(
            WhiteLabel::class,
            function () {
                return WhiteLabel::getInstance();
            }
        );

        $container->singleton(
            LanguageAndRegion::class,
            function (Container $container) {
                return new LanguageAndRegion(
                    $container->get(WMSDB::class),
                    $container->get(Log::class),
                );
            }
        );

        $container->singleton(
            SipProxy::class,
            function (Container $container) {
                return new SipProxy($container->get(FileCache::class));
            }
        );

        $container->singleton(
            CDRView::class,
            function () {
                return CDRView::getInstance();
            }
        );

        $container->singleton(
            SipRtp::class,
            function () {
                return SipRtp::getInstance();
            }
        );

        $container->singleton(
            CallFiles::class,
            function () {
                return CallFiles::getInstance();
            }
        );
    }
}
