<?php

namespace Wildix\Dialplan;

use \PBX;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\Validator;

abstract class BaseDialplansService {

    const MAX_COUNT   = null;
    const STATUS_KEYS = [
        0 => 'NOT_INUSE',
        1 => 'INUSE',
        2 => 'RINGING',
        'default' => 'NOT_INUSE'
    ];

    protected $items = [];

    protected $_storage;

    protected $_server;

    public function __construct($storage, PBX $server = null) {
        $this->_storage = $storage;
        $this->_server  = $server;
    }

    protected function storage() {
        return $this->_storage;
    }

    protected function server(): PBX {
        return $this->_server;
    }

    protected function setState(string $type, int $id, $status): void {
        $state = static::STATUS_KEYS[$status] ?? static::STATUS_KEYS['default'];
        $this->_server->set_devstate($type, sprintf('*%s*1', $id), $state);
    }

    protected function getAvailableId(): int {
        $res = 0;
        $i   = 0;
        $ids = array_keys($this->getList());

        while (!$res) {
            $i++;
            if (!in_array($i, $ids)) {
                $res = $i;
            }
        }

        return $res;
    }

    protected function getList(): array {
        if (empty($this->items)) {
            foreach ($this->storage()->getItems() as $item) {
                $this->items[$item->getId()] = $item->getName();
            }
        }

        return $this->items;
    }

    public function getItems(): array {
        $items = $this->storage()->getItems();
        return $this->storage()->loadRelations($items);
    }

    public function getItem(int $identity, bool $check = false) {
        $item = $this->storage()->getItem($identity);

        if (empty($item) && $check) {
            throw new InvalidDataException(Error::NotFound, [], 404);
        }

        return $item;
    }

    public function remove(int $identity): bool {
        $item   = $this->getItem($identity, true);
        $linked = $this->storage()->getLinked($identity);

        if (count($linked)) {
            throw new InvalidDataException(sprintf('You can’t remove %s. It’s linked to %s ', $item->getName(), implode(', ', $linked)), []);
        }

        $this->storage()->remove($identity);

        return true;
    }

    protected function validate(array $data, int $id = null): void {
        $errors    = [];
        $items     = $this->getList();
        $validator = $this->makeValidator($data);
        if ($id) {
            unset($items[$id]);
        }

        if (in_array((string)$data['title'], $items)) {
            $errors['title'] = [
                'code' => 'AlreadyPresent',
                'reason' => 'Already present'
            ];
        }

        if (!$id && static::MAX_COUNT && count($items) >= static::MAX_COUNT) {
            $errors['limit'] = [
                'code' => 'LimitReached',
                'reason' => 'Limit items ' . static::MAX_COUNT
            ];
        }

        if (!$validator->validate()) {
            $errors = array_merge($errors, $validator->errorsTree());
        }

        if (!empty($errors)) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $errors);
        }
    }

    protected function makeValidator(array $data): Validator {
        return new Validator($data);
    }
}
