<?php
declare(strict_types=1);

namespace Wildix\Dialplan\CallGroups;

use Cache\FileCache;
use Logger\Log;
use PBX;
use Wildix\Dialplan\BaseDialplansService;
use Wildix\Error;
use Wildix\Exception;
use Wildix\InvalidDataException;
use Wildix\License;
use Wildix\Validator;

/**
 * Class CallGroupsService
 */
class CallGroupsService extends BaseDialplansService {

    public const ALLOWED_QUEUE_MANAGER_LICENSE_TYPES = [
        License::TYPE_BASIC,
        License::TYPE_ESSENTIAL,
        License::TYPE_PREMIUM,
        License::TYPE_BUSINESS,
    ];
    const HTTP_URI_FOR_QUEUE_INFO                    = 'http://127.0.0.1:5280/api/get_queue_status?id=';
    const QUEUE_INFO_CACHE_EXPIRATION                = 10;

    const DEFAULT_SETTINGS = [
        'timeout'                     => 60,
        'queue-manager'               => '',
        'queue-cid'                   => '',
        'queue-youarenext'            => 'queue-youarenext',
        'queue-thereare'              => 'queue-thereare',
        'queue-callswaiting'          => 'queue-callswaiting',
        'queue-holdtime'              => 'queue-holdtime',
        'queue-minutes'               => 'queue-minutes',
        'queue-seconds'               => 'queue-seconds',
        'queue-thankyou'              => 'queue-thankyou',
        'queue-reporthold'            => 'queue-reporthold',
        'announce-frequency'          => 0,
        'announce-round-seconds'      => 5,
        'announce-holdtime'           => 'yes',
        'retry'                       => 1,
        'wrapuptime'                  => 0,
        'maxlen'                      => 5,
        'servicelevel'                => '',
        'strategy'                    => 'ringall',
        'joinempty'                   => 'invalid',
        'leavewhenempty'              => 'invalid',
        'eventmemberstatus'           => true,
        'eventwhencalled'             => true,
        'reportholdtime'              => 0,
        'memberdelay'                 => 0,
        'weight'                      => 0,
        'periodic-announce'           => 'queue-periodic-announce',
        'periodic-announce-frequency' => '',
        'setqueuevar'                 => true,
        'autopause'                   => 'no',
        'autopausebusy'               => false,
        'autopauseunavail'            => false,
        'autopausetime'               => 5,
        'autopausedelay'              => 0,
        'ringinuse'                   => false,
    ];

    const MAX_MEMBERS_IN_PENALTY_GROUP = 10;
    const DEFAULT_TIMEOUT              = 15;

    /**
     * Call group storage
     *
     * @var CallGroupsStorage
     */
    protected $_storage;

    /**
     * Entities Config class
     *
     * @var ConfigFiles
     */
    private $_configFiles;

    /**
     * Logger class.
     *
     * @var Log $_logger
     */
    private $_logger;

    private $_cache;

    private $_maxMembers = 0;
    private $_colleagues;

    public function __construct(
        $storage,
        PBX $server,
        ConfigFiles $configFiles,
        int $maxMembers,
        Log $logger,
        FileCache $cache
    ) {
        $this->_configFiles = $configFiles;
        $this->_maxMembers  = $maxMembers + 2;
        $this->_logger      = $logger;
        $this->_cache       = $cache;

        parent::__construct($storage, $server);
    }

    public function create(array $data): void {
        $this->validate($data);
        $this->storage()->create($this->_prepareToSave($data));
        $this->updateConfig();
        $this->_logger->info(__LINE__, 'Dialplan create Call Group');
    }

    public function update(int $identity, array $data): void {
        $item = $this->getItem($identity, true);
        $this->validate($data, $identity);
        $this->storage()->update($identity, $this->_prepareToSave($data, $item));
        $this->updateConfig();
        $this->_logger->info(__LINE__, sprintf('Dialplan update Call Group %d', $identity));
    }

    public function remove(int $identity): bool {
        parent::remove($identity);
        $this->updateConfig();
        $this->_logger->info(__LINE__, sprintf('Dialplan delete Call Group %d', $identity));

        return true;
    }

    public function buildItem(CallGroup $item) {
        $settings = $item->getSettings();
        return [
            'id'       => $item->getId(),
            'title'    => $item->getName(),
            'members'  => $item->getUsers(),
            'settings' => [
                'queueManager'     => (string)($settings['queue-manager'] ?? ''),
                'cid'              => (string)($settings['queue-cid'] ?? ''),
                'autoPause'        => $settings['autopause'] ?? 'no',
                'autoPauseTime'    => (int)$settings['autopausetime'],
                'autoPauseDelay'   => (int)$settings['autopausedelay'],
                'autoPauseBusy'    => Validator::isTrue($settings['autopausebusy']),
                'autoPauseUnAvail' => Validator::isTrue($settings['autopauseunavail']),
                'maxLen'           => (int)$settings['maxlen'],
                'retry'            => (int)$settings['retry'],
                'ringInUse'        => Validator::isTrue($settings['ringinuse']),
                'strategy'         => $settings['strategy'],
                'timeout'          => (int)($settings['timeout'] ?? 0),
                'weight'           => (int)($settings['weight'] ?? 0),
                'wrapUpTime'       => (int)$settings['wrapuptime'],
                'rule'             => (string)($settings['context'] ?? ''),
                'announce'         => [
                    'frequency'    => (int)$settings['announce-frequency'],
                    'holdTime'     => $settings['announce-holdtime'],
                    'roundSeconds' => (int)$settings['announce-round-seconds'],
                ],
                'exitAllMembers'   => [
                    'busy'    => Validator::isTrue($settings['exit_all_members_inuse'] ?? false),
                    'paused'  => Validator::isTrue($settings['exit_all_members_paused'] ?? false),
                    'ringing' => Validator::isTrue($settings['exit_all_members_ringing'] ?? false),
                ],
            ],
        ];
    }

    protected function makeValidator(array $data): Validator {
        $validator = new Validator($data);
        $validator->rule('required', ['title']);
        $validator->rule('regex', ['title'], '/^[0-9a-zA-Z-._]*$/')->reason(Error::getMessage(Error::IncorrectFormat));
        if (isset($data['members'])) {
            $validator->rule('array', ['members']);
            $validator->rule('max', ['members'], $this->_maxMembers)->reason('Max number of users is ' . $this->_maxMembers);
        }
        if (isset($data['settings'])) {
            $validator->rule('array', ['settings']);
            $validator->rule(
                'boolean',
                [
                    'settings.ringInUse',
                    'settings.exitAllMembers.paused',
                    'settings.exitAllMembers.busy',
                    'settings.exitAllMembers.ringing',
                ]
            );
            $validator->rule(
                'integer',
                [
                    'settings.weight',
                    'settings.timeout',
                    'settings.retry',
                    'settings.wrapUpTime',
                    'set                                                                                                                tings.wrapUpTime',
                    'settings.autoPauseTime',
                    'settings.autoPauseDelay',
                    'settings.maxLen',
                    'settings.announce.frequency',
                    'settings.announce.roundSeconds',
                ]
            );
            $validator->rule(
                'min',
                [
                    'settings.wrapUpTime',
                    'settings.autoPauseTime',
                    'settings.autoPauseDelay',
                    'settings.maxLen',
                ],
                0
            )->reason('Please enter a valid (min: 0)');
            $validator->rule('min', ['settings.retry'], 1)->reason('Please enter a valid (min: 1)');
            $validator->rule('between', ['settings.weight'], [0, 50])->reason('Please enter a valid value (0-50)');
            $validator->rule('between', ['settings.timeout'], [0, 120])->reason('Please enter a valid value (0-120)');
            $validator->rule('in', ['settings.announce.holdTime'], ['yes', 'no', 'once']);
            $validator->rule('in', ['settings.autoPause'], ['yes', 'no', 'all']);
            $validator->rule('in', ['settings.strategy'], ['ringall', 'roundrobin', 'wringall', 'leastrecent', 'fewestcalls', 'linear']);
            $validator->rule('extenNumber', ['settings.queueManager']);
            $validator->rule('checkUserLicenseType', ['settings.queueManager'], self::ALLOWED_QUEUE_MANAGER_LICENSE_TYPES);
            $validator->rule('hasUser', ['settings.queueManager']);
            $validator->rule(
                'required',
                [
                    'settings.retry',
                    'settings.wrapUpTime',
                    'settings.autoPauseTime',
                    'settings.autoPauseDelay',
                    'settings.maxLen',
                ]
            );
            $validator->rule('PhoneNumber', ['settings.cid']);
        }

        return $validator;
    }

    private function _prepareToSave(array $data, CallGroup $item = null): array {
        $settings = $item ? $item->getSettings() : self::DEFAULT_SETTINGS;

        $res = [
            'name'     => $data['title'],
            'settings' => $settings,
        ];

        if (isset($data['settings'])) {
            $res['settings'] = $this->_mergeSettings($settings, $data['settings']);

            if ($res['settings']['queue-manager']) {
                $this->_logger->info(
                    __LINE__,
                    sprintf(
                        'Missed calls managers is set for feature notifications for call groups, CG name: %s, CG id: %d, manager extension %s',
                        $res['name'],
                        $data['id'] ?? 0,
                        $res['settings']['queue-manager']
                    )
                );
            }
        }

        if (isset($data['members'])) {
            $res['members'] = array_map(
                function ($item) {
                    return stripos($item, '@') === false ? $item . '@internalcalls' : $item;
                },
                $data['members']
            );
        }

        return $res;
    }

    private function _mergeSettings(array $settings, array $data): array {
        $settings['queue-manager']            = (string)$data['queueManager'];
        $settings['queue-cid']                = (string)$data['cid'];
        $settings['autopause']                = (string)$data['autoPause'] ?? 'no';
        $settings['autopausetime']            = (int)$data['autoPauseTime'];
        $settings['autopausedelay']           = (int)$data['autoPauseDelay'];
        $settings['maxlen']                   = (int)$data['maxLen'];
        $settings['retry']                    = (int)$data['retry'];
        $settings['ringinuse']                = Validator::isTrue($data['ringInUse']);
        $settings['strategy']                 = (string)$data['strategy'];
        $settings['timeout']                  = (int)$data['timeout'];
        $settings['weight']                   = (int)$data['weight'];
        $settings['wrapuptime']               = (int)$data['wrapUpTime'];
        $settings['context']                  = (string)$data['rule'];
        $settings['announce-frequency']       = (int)$data['announce']['frequency'];
        $settings['announce-holdtime']        = (string)$data['announce']['holdTime'];
        $settings['exit_all_members_inuse']   = Validator::isTrue($data['exitAllMembers']['busy']);
        $settings['exit_all_members_paused']  = Validator::isTrue($data['exitAllMembers']['paused']);
        $settings['exit_all_members_ringing'] = Validator::isTrue($data['exitAllMembers']['ringing']);

        if ($settings['strategy'] !== 'ringall' && $settings['strategy'] !== 'wringall') {
            $settings['autofill'] = 'yes';
        } else {
            unset($settings['autofill']);
        }

        $exitArr = ['invalid'];
        if ($settings['exit_all_members_inuse'] === true) {
            $exitArr[] = 'inuse';
        }
        unset($settings['exit_all_members_inuse']);

        if ($settings['exit_all_members_paused'] === true) {
            $exitArr[] = 'paused';
        }
        unset($settings['exit_all_members_paused']);

        if ($settings['exit_all_members_ringing'] === true) {
            $exitArr[] = 'ringing';
        }
        unset($settings['exit_all_members_ringing']);

        $settings['joinempty']      = join(',', $exitArr);
        $settings['leavewhenempty'] = join(',', $exitArr);

        if ($settings['autopause'] !== 'no') {
            $settings['autopauseunavail'] = true;
        } else {
            $settings['autopausebusy']    = false;
            $settings['autopauseunavail'] = false;
        }

        return $settings;
    }

    public function updateConfig() {
        $items = $this->storage()->getItems();
        $items = $this->storage()->loadRelations($items, true);

        $this->_configFiles->saveConfigs($items);
        $this->server()->Command('reload app_queue.so');
    }

    public function loadMembersByIds(array $ids = []) {
        return $this->storage()->getMembersByIds($ids);
    }

    public function getItemsByExtension(string $ext) {
        $validator = new Validator(['extension' => $ext]);
        $validator->rule('extenNumber', ['extension']);
        $validator->rule('required', ['extension']);
        if (!$validator->validate()) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $validator->errors());
        }
        return $this->storage()->getItemsByExtension($ext);
    }

    public function getAgentsAvailabilityInfo(int $identity):array {
        $item            = $this->getItem($identity, true);
        $queueInfo       = $this->getQueueInfoById($identity);
        $result['id']    = $item->getId();
        $result['title'] = $item->getName();

        return array_merge($result, $queueInfo);
    }

    public function getQueueInfoById(int $identity):array {
        $cacheKey  = 'EjabberdQueueInfo_' . $identity;
        $fromCache = $this->_cache->get($cacheKey);

        if ($fromCache) {
            return json_decode($fromCache, true);
        }

        $url = self::HTTP_URI_FOR_QUEUE_INFO . $identity;
        $ch  = curl_init($url);
        curl_setopt($ch, CURLOPT_USERAGENT, 'PBX Api');
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_TIMEOUT, 5);
        curl_setopt($ch, CURLOPT_AUTOREFERER, true);
        curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, false);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);

        $response = curl_exec($ch);
        if (!$response) {
            throw new Exception(sprintf('Queue statistic data by id % is not found', $identity) . curl_error($ch));
        }

        $parsed = json_decode($response, true);
        if (!$parsed) {
            throw new Exception('Cannot parse json: ' . json_last_error());
        }
        curl_close($ch);
        $result = [];
        if (!isset($parsed['status'])) {
            $this->_cache->set($cacheKey, $response, self::QUEUE_INFO_CACHE_EXPIRATION);
            $result = $parsed;
        }
        return $result;
    }

    public function clearUserByExtension(string $extension): void {
        $this->_storage->clearUserByExtension($extension);
        $this->updateConfig();
    }

    public function clearQueueManagerByExtension(string $extension): void {
        $callGroups = array_filter(
            $this->_storage->getItems(),
            function ($callGroup) use ($extension) {
                return $callGroup->getSettings()['queue-manager'] === $extension;
            }
        );

        foreach ($callGroups as $callGroup) {
            $this->_storage->updateQueueManager($callGroup->getId());
        }
        $this->updateConfig();
    }
}
