<?php
declare(strict_types=1);

namespace Wildix\Dialplan\CallGroups;

use Wildix\Databases\WMSDB;

/**
 * Class CallGroupsStorage
 */
class CallGroupsStorage {

    const DIALPLAN_LINK = 1;
    const LINK_GROUP    = 3;

    const PARAM_SETTINGS = 'settings';
    const PARAM_MEMBERS  = 'members';

    private const GET_ITEMS = 'SELECT * FROM ' . WMSDB::TABLE_CALL_GROUPS;

    private $_db;

    private function _db(): WMSDB {
        if (!$this->_db) {
            $this->_db = WMSDB::getInstance();
        }
        return $this->_db;
    }

    public function getItems(): array {
        return $this->_db()->getRows(self::GET_ITEMS, null, CallGroup::class);
    }

    public function getItem(int $identity): ?CallGroup {
        $item = $this->_db()->first(self::GET_ITEMS . ' WHERE id=' . $identity);
        if ($item) {
            $item = new CallGroup($item);
        }
        return $item;
    }

    public function create(array $data): void {
        $this->_db()->beginTransaction();
        $this->_db()->insert(
            WMSDB::TABLE_CALL_GROUPS,
            [
                'call_groupname'   => $data['name'],
                'queue_parameters' => json_encode($data[self::PARAM_SETTINGS]),
            ]
        );
        if (isset($data[self::PARAM_MEMBERS])) {
            $this->_insertUsers($this->_db()->getLastInsertId(), $data[self::PARAM_MEMBERS]);
        }
        $this->_db()->endTransaction();
    }

    public function update(int $identity, array $data): void {
        $this->_db()->beginTransaction();
        $this->_db()->exec(
            sprintf(
                'UPDATE %s SET call_groupname=\'%s\', queue_parameters=\'%s\' WHERE id=%d',
                WMSDB::TABLE_CALL_GROUPS,
                $this->_db()->realEscapeString($data['name']),
                json_encode($data[self::PARAM_SETTINGS]),
                $identity
            )
        );

        if (isset($data[self::PARAM_MEMBERS])) {
            $this->_insertUsers($identity, $data[self::PARAM_MEMBERS]);
        }

        $this->_db()->endTransaction();
    }

    public function remove(int $identity): void {
        $this->_db()->beginTransaction();
        $this->_clearUsers($identity);
        $this->_db()->exec(sprintf('DELETE FROM %s WHERE id=%d', WMSDB::TABLE_CALL_GROUPS, $identity));
        $this->_db()->endTransaction();
    }

    public function getLinked(int $identity): array {
        // phpcs:disable Generic.Files.LineLength.TooLong
        $res = $this->_db()->getRows(
            sprintf(
                'SELECT rn.rulename FROM %s as rl LEFT JOIN %s as rn ON rn.id=rl.object_from_id WHERE rl.object_from_type=%s AND rl.object_to_type=%s AND rl.object_to_id=%d',
                WMSDB::TABLE_RULES_LINKS,
                WMSDB::DIALPLAN_RULE_NAME,
                self::DIALPLAN_LINK,
                self::LINK_GROUP,
                $identity
            )
        );
        // phpcs:enable Generic.Files.LineLength.TooLong
        return $res ? array_column($res, 'rulename') : [];
    }

    public function loadRelations(array $groups = [], bool $fullName = false): array {
        $items = [];
        foreach ($groups as $group) {
            $items[$group->getId()] = $group;
        }

        $users = $this->_getUsers(array_keys($items));

        foreach ($users as $user) {
            if (isset($items[$user['call_group_id']])) {
                $items[$user['call_group_id']]->addUser(($fullName ? $user['group_membername'] : $user['group_member']));
            }
        }

        return array_values($items);
    }

    public function getMembersByIds(array $ids) {
        return array_column($this->_getUsers($ids), 'group_member');
    }

    public function getItemsByExtension(string $ext) {
        $sql = 'SELECT * FROM ' . WMSDB::TABLE_CALL_GROUPS_USERS . ' WHERE group_member LIKE \'' . $ext . '@%\'';
        $ids = array_keys($this->_db()->getRows($sql, 'call_group_id'));

        return $this->_db()->getRows(sprintf(self::GET_ITEMS . ' WHERE id IN (%s)', implode(', ', $ids)), null, CallGroup::class);
    }

    public function clearUserByExtension(string $extension): void {
        $this->_db()->exec(sprintf('DELETE FROM %s WHERE group_member=\'%s\'', WMSDB::TABLE_CALL_GROUPS_USERS, sprintf('%s@internalcalls', $extension)));
    }

    private function _getUsers(array $ids = []): array {
        $items = [];
        if ($ids) {
            $items = $this->_db()->getRows(sprintf('SELECT * FROM %s WHERE call_group_id IN (%s)', WMSDB::TABLE_CALL_GROUPS_USERS, implode(', ', $ids)));
        }

        return $items;
    }

    private function _clearUsers(int $groupId) {
        $this->_db()->exec(sprintf('DELETE FROM %s WHERE call_group_id=%d', WMSDB::TABLE_CALL_GROUPS_USERS, $groupId));
    }

    private function _insertUsers(int $groupId, array $members): void {
        $this->_clearUsers($groupId);
        $members = array_unique($members);
        foreach ($members as $member) {
            $this->_db()->insert(
                WMSDB::TABLE_CALL_GROUPS_USERS,
                [
                    'call_group_id' => $groupId,
                    'group_member'  => $member,
                    'group_membername'  => $this->_calcMemberName($member),
                ]
            );
        }
    }

    private function _calcMemberName(string $member): string {
        return (strpos($member, '@') !== false ? 'Local/' : 'SIP/' ) . $member;
    }

    public function updateQueueManager(int $identity, string $queueManager = ''): void {
        $queueParams = $this->_db()->first(
            sprintf(
                'SELECT queue_parameters FROM %s WHERE id=%d',
                WMSDB::TABLE_CALL_GROUPS,
                $identity
            )
        )['queue_parameters'];

        $queueParams = json_decode($queueParams, true);

        $queueParams['queue-manager'] = $queueManager;

        $this->_db()->exec(
            sprintf(
                'UPDATE %s SET queue_parameters=\'%s\' WHERE id=%d',
                WMSDB::TABLE_CALL_GROUPS,
                json_encode($queueParams),
                $identity
            )
        );
    }
}
