<?php

namespace Wildix\Dialplan\CallGroups;

use Services\Helpers;
use Wildix\Dialplan\CallGroups\Config\CallGroupsEntitiesConfig;
use Wildix\Dialplan\CallGroups\Config\CallGroupsRulesConfig;

/**
 * Class ConfigFiles
 */
class ConfigFiles {

    /**
     * Entities Config class
     *
     * @var CallGroupsEntitiesConfig
     */
    private $_entitiesConfig;

    /**
     * Rules Config class
     *
     * @var CallGroupsRulesConfig
     */
    private $_rulesConfig;

    private $_booleanParams = [
        'autopausebusy',
        'autopauseunavail',
        'eventmemberstatus',
        'eventwhencalled',
        'setqueuevar',
        'ringinuse',
    ];

    private $_removedParams = [
        'exit_all_members_inuse',
        'exit_all_members_paused',
        'exit_all_members_ringing',
    ];

    public function __construct(
        CallGroupsEntitiesConfig $entitiesConfig,
        CallGroupsRulesConfig $rulesConfig
    ) {
        $this->_entitiesConfig = $entitiesConfig;
        $this->_rulesConfig    = $rulesConfig;
    }

    /**
     * Save call group config files
     *
     * @param array $items Call group items.
     *
     * @return void
     */
    public function saveConfigs(array $items): void {
        $data = $this->_prepareToConfigFiles($items);

        $this->_rulesConfig->setData($data['rules']);
        $this->_rulesConfig->save();

        $this->_entitiesConfig->setData($data['entities']);
        $this->_entitiesConfig->save();
    }

    private function _prepareToConfigFiles(array $queueItems): array {
        $queueEntries = [];
        $queueRules   = [];

        foreach ($queueItems as $item) {
            $queueParams = [];
            $itemId      = $item->getId();

            $settings = $item->getSettings();
            $members  = $item->getUsers();

            $queueEntries[$itemId]['queue-name'] = $item->getAsArray()['call_groupname'];

            foreach ($settings as $key => $val) {
                if (in_array($key, $this->_removedParams, true)) {
                    continue;
                }

                if ($val !== '') {
                    if (in_array($key, $this->_booleanParams, true)) {
                        $val = Helpers::formatBoolean($val);
                    }

                    $queueEntries[$itemId][$key] = $val;
                }
                if ($key === 'strategy' || $key === 'timeout') {
                    $queueParams[$key] = $val;
                }
            }

            if (!isset($queueParams['strategy']) || $queueParams['strategy'] === '') {
                $queueParams['strategy'] = 'ringall';
            }
            if (!isset($queueParams['timeout']) || $queueParams['timeout'] === '') {
                $queueParams['timeout'] = CallGroupsService::DEFAULT_TIMEOUT;
            }

            $isRingAll  = $queueParams['strategy'] === 'ringall';
            $isWringAll = $queueParams['strategy'] === 'wringall';

            if ($isRingAll && count($members) > CallGroupsService::MAX_MEMBERS_IN_PENALTY_GROUP) {
                $queueEntries[$itemId]['defaultrule'] = $itemId;
            }

            $n       = 0;
            $penalty = 1;
            $timeout = (int)$queueParams['timeout'];

            foreach ($members as $member) {
                if ($isRingAll && ++$n > CallGroupsService::MAX_MEMBERS_IN_PENALTY_GROUP) {
                    $n = 0;

                    $queueRules[$itemId]['penaltychange'][] = ($timeout * $penalty - 2) . ',' . ($penalty + 1) . ',' . ($penalty + 1);
                    ++$penalty;
                }

                if ($isWringAll && preg_match('~/(\d+)@internalcalls~', $member, $math)) {
                    $queueEntries[$itemId]['members'][] = 'SIP/' . $math[1] . ',1';
                } else {
                    $queueEntries[$itemId]['members'][] = $member . ',' . $penalty;
                }
            }
        }

        return [
            'entities' => $queueEntries,
            'rules'    => $queueRules,
        ];
    }
}
