<?php
namespace Wildix\Dialplan;

use Api\Translation\Interfaces\TranslationStrategyInterface;
use Psr\Container\ContainerInterface;
use Wildix\Databases\WMSDB;


/**
 * Repository class for dialplan contexts
 */
class ContextRepository {

    protected ContainerInterface $_container;

    /**
     * WMS database.
     *
     * @var WMSDB $_db.
     */
    private $_db;

    /**
     * ContextRepository constructor.
     *
     * @param WMSDB $db Databases.
     */
    public function __construct(WMSDB $db, ContainerInterface $container) {
        $this->_db        = $db;
        $this->_container = $container;
    }

    /**
     * Gets all context names registered in WMS.
     *
     * @param string $topValue Top value.
     *
     * @return array Context names & incoming rule names.
     * @throws \Exception DB exception.
     */
    public function getNames($topValue = ''): array {
        $result = array();
        if ($topValue !== '') {
            $result[$topValue] = $topValue;
        }

        $translateService = $this->_container->get(TranslationStrategyInterface::class);

        $dialplanWithPredefinedDescription = array(
                                              'internalcalls',
                                              'pbxfeatures',
                                              'pbxservices',
                                              'pbxinternal',
                                              'users',
                                              'main',
                                              'classound',
        );

        $sql = "SELECT context, ruledescription FROM
            (SELECT DISTINCT rulename AS context, ruledescription AS ruledescription FROM dialplan_rulenames
            UNION SELECT 'pbxinternal' AS context, '' AS ruledescription
            UNION SELECT 'pbxfeatures' AS context, '' AS ruledescription
            UNION SELECT 'pbxservices' AS context, '' AS ruledescription
            UNION SELECT 'internalcalls' AS context, '' AS ruledescription)
            ORDER BY context ASC;";

        $rows = $this->_db->query($sql);
        if ($rows) {
            do {
                $row = $rows->fetchArray(SQLITE3_ASSOC);
                if ($row) {
                    if (in_array($row['context'], $dialplanWithPredefinedDescription)) {
                        $result[$row['context']] = $row['context'] . ' (' . $translateService->t('dialplan.rule.' . $row['context']) . ')';
                    } else if ($row['ruledescription'] != '') {
                        $result[$row['context']] = $row['context'] . ' (' . $row['ruledescription'] . ')';
                    } else {
                        $result[$row['context']] = $row['context'];
                    }
                }
            } while ($row);
            $rows->finalize();
        }

        return $result;
    }

    /**
     * Get contexts list of id => name format.
     *
     * @return array
     * @throws \Exception DB exception.
     */
    public function getContexts(): array {
        $contexts       = [];
        $contexts['-1'] = 'pbxfeatures';
        $contexts['-2'] = 'pbxservices';
        $contexts['-3'] = 'internalcalls';
        $contexts['-4'] = 'pbxinternal';

        $sql  = 'SELECT id, rulename as context FROM ' . \Wildix\Dialplan\DialplanService::DIALPLAN_NAMES_TABLE . ' ORDER BY context;';
        $rows = $this->_db->getRows($sql);
        foreach ($rows as $row) {
            $contexts[$row['id']] = $row['context'];
        }

        return $contexts;
    }

    public function getDialplanNames($topValue = '') {
        $result = [];

        $sql = "SELECT context, ruledescription FROM
            (SELECT DISTINCT rulename AS context, ruledescription AS ruledescription FROM dialplan_rulenames
            UNION SELECT 'pbxinternal' AS context, '' AS ruledescription
            UNION SELECT 'pbxfeatures' AS context, '' AS ruledescription
            UNION SELECT 'pbxservices' AS context, '' AS ruledescription
            UNION SELECT 'internalcalls' AS context, '' AS ruledescription)
            ORDER BY context ASC;";

        $rows = $this->_db->getRows($sql);
        foreach ($rows as $row) {
            $result[] = [
                'context' => $row['context'],
                'ruledescription' => $row['ruledescription'] ?? '',
                'key' => 'dialplan.rule.' . $row['context']
            ];
        }

        return $result;
    }
}
