<?php
declare(strict_types=1);

namespace Wildix\Dialplan;

use Logger\Log;
use PBX;
use Wildix\CallweaverManager;
use Wildix\Databases\WMSDB;
use Wildix\DependencyInjection\ConfigInterface;
use Wildix\DependencyInjection\Container;
use Wildix\Dialplan\Switches\SwitchesService;
use Wildix\Dialplan\TimeTables\TimeTablesService;
use Wildix\Dictionaries;
use WMS_Ldap;

/**
 * Class DiConfig
 */
final class DiConfig implements ConfigInterface {

    public static function configure(Container $container): void {
        $container->singleton(
            ExtensionsVariables::class,
            function () {
                return ExtensionsVariables::getInstance();
            }
        );

        $container->singleton(
            ContextRepository::class,
            function (Container $container) {
                return new ContextRepository($container->get(WMSDB::class), $container);
            }
        );

        $container->singleton(
            ApplicationFactory::class,
            function (Container $container) {
                $appConfigDir = '';
                return new ApplicationFactory(
                    $container->get(Log::class),
                    $container->get(DialplanFormatter::class),
                    $appConfigDir
                );
            }
        );

        $container->singleton(
            ApplicationService::class,
            function (Container $container) {
                return new ApplicationService(
                    $container->get(ApplicationFactory::class)
                );
            }
        );

        $container->singleton(
            Rules::class,
            function () {
                return new Rules();
            }
        );

        $container->singleton(
            IncomingCalls::class,
            function (Container $container) {
                $applicationsPath = BASEPATH . 'dialplan/applications/';
                $timeTables       = $container->get(TimeTablesService::class);
                $switchesService  = $container->get(SwitchesService::class);
                $pbx              = PBX::getInstance();
                return new IncomingCalls($applicationsPath, $timeTables, $switchesService, $pbx);
            }
        );

        $container->singleton(
            RulesValidator::class,
            function () {
                return new RulesValidator();
            }
        );

        $container->singleton(
            DialplanService::class,
            function (Container $container) {
                return new DialplanService(
                    $container->get(ApplicationService::class),
                    $container->get(RulesService::class),
                    $container->get(WMSDB::class),
                    $container->get(CallweaverManager::class),
                    $container->get(DialplanFormatter::class),
                    $container->get(WMS_Ldap::class),
                    $container->get(Log::class)
                );
            }
        );

        $container->singleton(
            RulesService::class,
            function (Container $container) {
                return new RulesService(
                    $container->get(ApplicationService::class),
                    $container->get(DialplanFormatter::class),
                    $container->get(WMSDB::class),
                    $container->get(Dictionaries::class),
                    $container->get(RulesValidator::class),
                    $container->get(Log::class)
                );
            }
        );

        $container->singleton(
            DialplanFormatter::class,
            function (Container $container) {
                return new DialplanFormatter($container->get(Dictionaries::class));
            }
        );
    }
}
