<?php

namespace Wildix\Dialplan;

use Wildix\Dictionaries;

/**
 * Class DialplanFormatter
 */
class DialplanFormatter {

    /**
     * Dictionaries class.
     *
     * @var Dictionaries
     */
    private $_dictionaries;

    public function __construct(\Wildix\Dictionaries $dictionaries) {
        $this->_dictionaries = $dictionaries;
    }

    /**
     * Replace strings for application Set.
     *
     * @var array
     */
    private $_replaceForAppSet = [
        'LANGUAGE()' => 'language',
        'CDR(tags)' => 'tags',
        'SECURITY' => 'srtp_security',
        'SIP_CODECS' => 'sip_codecs',
        'MUSICCLASS()' => 'mohclass',
        'RINGTONE()' => 'ringtone',
        '__CALLTYPE' => 'calltype',
        'CALLERID(name)' => 'cidname',
        'CALLERID(number)' => 'cidnumber',
        '__TRUNK_CALLER_NUM' => 'trunk_caller_num',
        'DIRECTRTP' => 'directrtp',
        'TIMEOUT' => 'timeout',
        'ITC' => 'itc',
        'Service' => 'service',
    ];

    /**
     * Mapping for application param names.
     * https://docs.google.com/spreadsheets/d/1MHiS3Tu5WBWm254xwDbARjEUJWyNJ0tzU62veqaXH1M/edit?usp=sharing
     *
     * @var array
     */
    private $_appParamNamesMapping = [
        'AuthenticateUser' => ['destination' => 'number'],
        'DialRemotePbx' => [
            'exten' => 'number',
            'context' => 'dialplan',
        ],
        'CorrectOutgoingNumber' => [
            'out_prefix' => 'outPrefix',
            'context' => 'dialplan',
        ],
        'DialPhone' => [
            'exten' => 'number',
            'filename' => 'msgForOperator',
        ],
        'DialTrunk' => [
            'exten' => 'number',
            'maxcalls' => 'maxCalls',
            'class_of_call' => 'classOfCall',
        ],
        'Hangup' => ['causecode' => 'causeCode'],
        'DialHuntGroup' => [
            'exten' => 'number',
            'hunt_group' => 'huntGroup',
            'class_of_call' => 'classOfCall',
            'hg_strategy' => 'hgStrategy',
        ],
        'Queue' => [
            'queuename' => 'queueId',
            'announceoverride' => 'msgForOperator',
            'custom_mohclass' => 'mohClass',
        ],
        'Conference' => ['confno' => 'roomNumber'],
        'FaxToEmail' => ['localstationid' => 'localstationId'],
        'VoiceMail' => ['flags' => 'options'],
        'FeatureIntercom' => [
            'exten' => 'number',
            'filename' => 'msgForOperator',
        ],
        'FeaturePaging' => ['dest' => 'groupId'],
        'MuxMon' => ['email' => 'destination'],
        'VoiceMailMain' => ['destination' => 'number'],
        'GoToIfFaxDetected' => [
            'waitdur' => 'detectTime',
            'exten' => 'number',
            'context' => 'dialplan',
        ],
        'Goto' => [
            'exten' => 'number',
            'context' => 'dialplan',
        ],
        'GotoIf' => [
            'exten' => 'number',
            'context' => 'dialplan',
        ],
        'GotoIf3SwitchIsOnOrBlinking' => [
            'switchname' => 'switchId',
            'exten_on' => 'numberOn',
            'context_on' => 'dialplanOn',
            'exten_bl' => 'numberBlinking',
            'context_bl' => 'dialplanBlinking',
        ],
        'GotoIfCallType' => [
            'verify' => 'callType',
            'exten' => 'number',
            'context' => 'dialplan',
        ],
        'GotoIfBusyCongested' => [
            'dialstatus' => 'dialStatus',
            'exten' => 'number',
            'context' => 'dialplan',
        ],
        'GotoIfSwitchIsOn' => [
            'switchname' => 'switchId',
            'exten' => 'number',
            'context' => 'dialplan',
        ],
        'GotoIfUserStatus' => [
            'userstatus' => 'userStatus',
            'exten' => 'number',
            'context' => 'dialplan',
        ],
        'VerifyCaller' => [
            'verify' => 'contains',
            'exten' => 'number',
            'context' => 'dialplan',
        ],
        'GotoIfGroupCount' => [
            'groupcount' => 'callsCount',
            'exten' => 'number',
            'context' => 'dialplan',
        ],
        'Timetable' => [
            'timetablename' => 'timetableId',
            'exten' => 'number',
            'context' => 'dialplan',
        ],
        'LookUpDestCidNameInPhonebook' => [
            'destination' => 'number',
            'phonebook' => 'phonebookId',
        ],
        'LookUpCidNameInPhonebook' => [
            'destination' => 'number',
            'phonebook' => 'phonebookId',
        ],
        'MobilityExtensionLookUp' => ['destination' => 'number'],
        'IVR' => ['ivrname' => 'ivrName'],
        'Playback' => ['noanswer' => 'noAnswer'],
        'Background' => ['noanswer' => 'noAnswer'],
        'Set' => [
            'set_l_key' => 'key',
            'set_l_val' => 'value',
        ],
        'CustomApp' => ['scriptname' => 'cwApp'],
    ];


    /**
     * Convert app param names to API format.
     *
     * @param string $appName   Application name.
     * @param string $paramName Param name.
     *
     * @return string
     */
    public function convertParamNameConfigToApi(string $appName, string $paramName): string {
        if (!array_key_exists($appName, $this->_appParamNamesMapping)
            || !array_key_exists($paramName, $this->_appParamNamesMapping[$appName])
        ) {
            return $paramName;
        }

        return $this->_appParamNamesMapping[$appName][$paramName];
    }

    /**
     * Convert app param names from API format.
     *
     * @param string $appName   Application name.
     * @param string $paramName Param name.
     *
     * @return string
     */
    public function convertParamNameApiToConfig(string $appName, string $paramName): string {
        // phpcs:disable Squiz.PHP.DisallowMultipleAssignments.FoundInControlStructure
        if (!array_key_exists($appName, $this->_appParamNamesMapping)
            || ($key = array_search($paramName, $this->_appParamNamesMapping[$appName])) === false
        ) {
            return $paramName;
        }
        // phpcs:enable Squiz.PHP.DisallowMultipleAssignments.FoundInControlStructure

        return $key;
    }

    /**
     * Get saved values in format key=>value.
     *
     * @param ApplicationModel $app         Application.
     * @param string           $savedString Saved in string params.
     *
     * @example "_{_exten_}","_{_pbx_}","_{_context_}","_{_timeout_}","_{_options_}"
     * @example "33","pbxid=221100000a55\,dc=wildix","users5","2","m"
     *
     * @return array
     */
    public function savedStringToParamsArray(ApplicationModel $app, string $savedString): array {
        $appName      = $app->getName();
        $template     = $app->getTemplateForOutput();
        $paramsConfig = $app->getParams(false);

        $result          = [];
        $matchesReplace  = [];
        $matchesParams   = [];
        $sortedParamKeys = [];

        $beforeReplaceFrom = [
            '|',
            '\;',
            '\"',
            '\,',
            '\\',
            ];
        $beforeReplaceTo   = [
            '-w-pipe-w-',
            '-w-semicolon-w-',
            '-w-doublequote-w-',
            '-w-comma-w-',
            '-w-backslash-w-',
            ];

        $afterReplaceFrom = [
            '-w-pipe-w-',
            '-w-comma-w-',
            '-w-semicolon-w-',
            '-w-doublequote-w-',
            '-w-backslash-w-',
            ];
        $afterReplaceTo   = [
            '|',
            ',',
            ';',
            '"',
            '\\',
            ];

        $regexpStr = $template;
        $regexpStr = preg_replace('/([.*+?^=!:${}()|[\]\/\\\])/', '\\\$1', $regexpStr);

        $paramNames = array_column($paramsConfig, 'name');
        foreach ($paramNames as $idx => $paramName) {
            $regexpStr = preg_replace('/_\\\{_' . $paramName . '_\\\}/', '(.*)', $regexpStr);

            $matchesReplace[$idx] = '$' . ($idx + 1);
            $matchesParams[$idx]  = '\_\{\_([^\{]+)\_\}';
        }
        $matchesReplace = implode('|', $matchesReplace);
        $matchesParams  = implode('.*', $matchesParams);

        preg_match_all('/' . $matchesParams . '/', $template, $match);
        foreach ($match as $idx => $val) {
            if ($idx === 0) {
                continue;
            }
            $sortedParamKeys[$idx - 1] = $val[0];
        }

        $mayContainPipeApps = [
            'CustomApp',
            'FaxToEmail',
            'GotoIf',
            'Set',
            'FastAGI',
            ];
        if (in_array($appName, $mayContainPipeApps)) {
            $savedString = str_replace('|', '-w-pipe-w-', $savedString);
        }

        if ($appName !== 'CustomApp' && $appName !== 'FaxToEmail' && is_string($savedString)) {
            $savedString = str_replace($beforeReplaceFrom, $beforeReplaceTo, $savedString);
        }

        if ($appName === 'MobilityExtensionLookUp') {
            $savedString = $this->implodeOptions($savedString);
        }

        $dirtyParamValues = explode('|', preg_replace('/' . $regexpStr . '/', $matchesReplace, $savedString));

        $needClearDoubleEscaping = [
            'Set',
            'FastAGI',
            'GotoIf',
            'Playback',
            'Background',
            'DialPhone',
            'Queue',
            'FeatureIntercom',
        ];

        foreach ($sortedParamKeys as $idx => $param) {
            if ($appName === 'Set' && $param === 'set_l_key') {
                if (array_key_exists($dirtyParamValues[$idx], $this->_replaceForAppSet)) {
                    $dirtyParamValues[$idx] = $this->_replaceForAppSet[$dirtyParamValues[$idx]];
                }
            }

            if ($appName === 'FaxToEmail') {
                $dirtyParamValues[$idx] = str_replace('-w-pipe-w-', ' ', $dirtyParamValues[$idx]);
            } else {
                $dirtyParamValues[$idx] = str_replace('-w-pipe-w-', '|', $dirtyParamValues[$idx]);
            }

            // rename params by mapping
            $mappedParamName = $this->convertParamNameConfigToApi($appName, $param);
            if (in_array($mappedParamName, $app->getActiveParamsList())) {
                $result[$mappedParamName] = str_replace($afterReplaceFrom, $afterReplaceTo, $dirtyParamValues[$idx]);
                if (in_array($appName, $needClearDoubleEscaping, true)) {
                    $result[$mappedParamName] = stripslashes($result[$mappedParamName]);
                }
            }
        }

        return $result;
    }

    private function implodeOptions(string $savedString): string {
        $explodedSavedString   = explode('","', $savedString);
        $secondPartSavedString = implode('', array_slice($explodedSavedString, 1));
        return implode('","', [$explodedSavedString[0], $secondPartSavedString]);
    }

    /**
     * Format input parameter value.
     *
     * @param string $appName         Application name.
     * @param string $paramName       Parameter name.
     * @param string $value           Parameter value.
     * @param array  $appConfigParams Application config.
     *
     * @return string
     */
    public function formatInputParamValue(string $appName, string $paramName, string $value, array $appConfigParams): string {
        $search  = [
            '\\',
            '"',
            ';',
            ];
        $replace = [
            '\\\\',
            '\\"',
            '\\;',
            ];

        if ($value === '-') {
            $value = '';
        }

        if ($appName === 'VoiceMail' && $paramName === 'destination' && strncmp($value, 'email:', 6) === 0) {
            $value = substr($value, 6);
            $value = str_replace($search, $replace, $value);
        } else if ($appName === 'Set' && $paramName === 'set_l_key') {
            $flippedArray = array_flip($this->_replaceForAppSet);
            if (array_key_exists($value, $flippedArray)) {
                $value = $flippedArray[$value];
            }
        } else if ($appName !== 'CustomApp') {
            $value = str_replace($search, $replace, $value);
            if ($appName === 'FaxToEmail' && $paramName === 'destination') {
                $value = preg_replace('/\s*[,|\s]\s*/', '|', $value);
            } else {
                $value = str_replace(',', '\,', $value);

                if ($appName === 'MobilityExtensionLookUp' && $paramName === 'options') {
                    $values = str_split($value);
                    $values = array_map(
                        function ($option) use ($values) {
                            return in_array($option['name'], $values) ? $option['name'] : '';
                        },
                        $appConfigParams[$paramName]['option']
                    );
                    $value  = implode('","', $values);
                }
            }
        }

        return $value;
    }

    /**
     * Formatting input number.
     *
     * @param string $number Number of dialplan.
     *
     * @return string
     */
    public function formatInputNumber(string $number): string {
        $dDotPos = strpos($number, ':');
        $spPos   = strpos($number, ' ');
        $pos     = false;

        if ($dDotPos !== false && $spPos !== false) {
            $pos = min($dDotPos, $spPos);
        } else if ($dDotPos !== false) {
            $pos = $dDotPos;
        } else if ($spPos !== false) {
            $pos = $spPos;
        }

        if ($pos !== false) {
            $number = substr($number, 0, $pos);
        }

        return $number;
    }

    /**
     * Return string for Lua config file.
     *
     * @param array $savedRules   Rules list.
     * @param array $includesData Includes list.
     *
     * @return string
     */
    public function getLuaConfigStr(array $savedRules, array $includesData): string {
        $rulesArr = [];

        $contexts = $this->_dictionaries->getDictionary('contexts');

        foreach ($savedRules as $row) {
            $contexts[$row['drule_id']] = $row['context'];
            $number = $row['number'];

            if (preg_match('/[XNZ\[\]\.\!]/', $number)) {
                $number = '_' . $number;
            }

            if (!isset($rulesArr[$row['drule_id']])) {
                $rulesArr[$row['drule_id']] = [];
            }

            if (!isset($rulesArr[$row['drule_id']][$number])) {
                $rulesArr[$row['drule_id']][$number] = [];
            }

            if ($row['app'] == 'CustomApp' && !empty($row['appdata'])) {
                $row['appdata'] = strtr($row['appdata'], ['\\' => '\\\\', '"' => '\\"']);
                $rulesArr[$row['drule_id']][$number][] = '		require_once"services":start_call("app-custom","' . $row['appdata'] . '");';
                continue;
            }

            $appArgs = empty($row['appdata']) ? [] : explode(',', $row['appdata']);

            if (in_array($row['app'], ['Background', 'Playback', 'SendDTMF'])) {
                $rulesArr[$row['drule_id']][$number][] = '		app.' . $row['app'] . '(' . join(',', $appArgs) . ');';
                continue;
            }

            if ($row['app'] === 'Answer') {
                $rulesArr[$row['drule_id']][$number][] = '		app.Answer(",nocdr");';
                continue;
            }

            if ($row['app'] === 'WaitExten') {
                $rulesArr[$row['drule_id']][$number][] = '		app.WaitExten(' . join(',', $appArgs) . ');';
                continue;
            }

            if (in_array($row['app'], ['app-dial-internalcall', 'app-queue', 'app-dial-trunk', 'app-dial-hunt-group'])) {
                $appArgs[] = 1;
            }

            if ($row['app'] === 'app-set' && isset($appArgs[0]) && $appArgs[0] === '"CDR(tags)"') {
                $tags = count($appArgs) > 1 ? trim(join(',', array_slice($appArgs, 1)), '"') : '';
                if (!empty($tags)) {
                    $rulesArr[$row['drule_id']][$number][] = '		require_once"services":start_call("' . $row['app'] . '","CDR(tags)","tags:' . $tags . '");';
                }
            } else {
                $s = '		require_once"services":start_call("' . $row['app'] . '"' . (!empty($appArgs) ? ',' . join(',', $appArgs) : '') . ');';

                $rulesArr[$row['drule_id']][$number][] = $s;
            }
        }

        // WMS-4496 - delete # from dial the user / intercom app
        foreach ($rulesArr as $contextId => $context) {
            if (array_key_exists('i', $context)) {
                continue;
            }

            $isWaitDigRuleExist = false;
            foreach ($context as $number => $rules) {
                foreach ($rules as $rule) {
                    if (strpos($rule, 'app-playback-waitDigits') !== false || strpos($rule, 'app.WaitExten') !== false) {
                        $isWaitDigRuleExist = true;
                    }
                }
            }

            if ($isWaitDigRuleExist) {
                $s = '		i_exten = channel.INVALID_EXTEN:get() or ""
		if "#" == string.sub(i_exten, -1) then
			require_once"services":start_call("app-goto-general",string.sub(i_exten, 1, #i_exten - 1),channel.CONTEXT:get() or "","1","");
		else
			require_once"services":start_call("app-hangup","","1");
		end';
                $rulesArr[$contextId]['i'] = [$s];
            }
        }

        $rulesArr = $this->_fixParamInLastRule($rulesArr);

        $includes = $this->_getIncludes($includesData, $contexts);

        return $this->_generateLuaContent($rulesArr, $includes, $contexts);
    }

    /**
     * Generate content for lua file.
     *
     * @param array $outArr   Array of rules.
     * @param array $includes Array of includes.
     * @param array $contexts Array of contexts.
     *
     * @return string
     */
    private function _generateLuaContent(array $outArr, array $includes, array $contexts): string {
        $outputStr = '';
        foreach ($contexts as $contextId => $contextName) {
            if ($contextId <= 0 || strcasecmp($contextName, 'ctivideo') === 0) {
                continue;
            }

            $outputStr .= PHP_EOL;
            $outputStr .= 'extensions["' . $contexts[$contextId] . '"] = {' . PHP_EOL;
            if (isset($outArr[$contextId])) {
                foreach ($outArr[$contextId] as $number => $rules) {
                    $outputStr .= '	["' . str_replace('"', '\"', $number) . '"] = function (c, e)' . PHP_EOL;
                    foreach ($rules as $rule) {
                        $outputStr .= $rule . PHP_EOL;
                    }
                    $outputStr .= '	end;' . PHP_EOL;
                }
            }

            if (!empty($includes[$contextId])) {
                $outputStr .= '	include = {';
                foreach ($includes[$contextId] as $include) {
                    $outputStr .= '"' . $include . '",';
                }
                $outputStr  = substr($outputStr, 0, -1);
                $outputStr .= '};' . PHP_EOL;
            }
            $outputStr .= '}' . PHP_EOL;
        }

        return $outputStr;
    }

    /**
     * Change last param 1 => 0 for last rule in list.
     *
     * @param array $outArr Rules array.
     *
     * @return array
     */
    private function _fixParamInLastRule(array $outArr): array {
        foreach ($outArr as $contextId => $context) {
            foreach ($context as $number => $rules) {
                $keyUpdate  = false;
                $rulesCount = count($rules);
                if ($rulesCount > 0) {
                    $key = $rulesCount - 1;
                    if (strpos($rules[$key], '"app-dial-internalcall",') !== false
                        || strpos($rules[$key], '"app-queue",') !== false
                        || strpos($rules[$key], '"app-dial-trunk",') !== false
                        || strpos($rules[$key], '"app-dial-hunt-group",') !== false
                    ) {
                        $keyUpdate = $key;
                    }
                    if ($keyUpdate !== false) {
                        $outArr[$contextId][$number][$keyUpdate] = substr($outArr[$contextId][$number][$keyUpdate], 0, -3) . '0);';
                    }
                }
            }
        }

        return $outArr;
    }

    /**
     * Get list of includes.
     *
     * @param array $includesData Includes input.
     * @param array $contexts     Contexts.
     *
     * @return array
     */
    private function _getIncludes(array $includesData, array $contexts): array {
        $includes = [];
        foreach ($includesData as $row) {
            if (isset($contexts[$row['rule_id']]) && isset($contexts[$row['include_rule_id']])) {
                if (!isset($includes[$row['rule_id']])) {
                    $includes[$row['rule_id']] = [];
                }
                $includes[$row['rule_id']][] = $contexts[$row['include_rule_id']];
            }
        }

        foreach ($contexts as $contextId => $contextName) {
            if (!isset($includes[$contextId])) {
                $includes[$contextId] = [];
            }
        }

        return $includes;
    }
}
