<?php

namespace Wildix\Dialplan;

use Wildix\Exception;

/**
 * Class DialplanService
 */
class DialplanService {

    const DIALPLAN_NAMES_TABLE    = 'dialplan_rulenames';
    const DIALPLAN_RULES_TABLE    = 'dialplan_rules';
    const DIALPLAN_LINKS_TABLE    = 'dialplan_links';
    const DIALPLAN_INCLUDES_TABLE = 'dialplan_rules_includes';

    const PARAM_ID          = 'id';
    const PARAM_NAME        = 'name';
    const PARAM_DESCRIPTION = 'description';
    const PARAM_NUMBERS     = 'numbers';
    const PARAM_INCLUDES    = 'includedDialplans';
    const PARAM_RULES       = 'rules';

    const PARAM_INCLUDED_DIALPLAN_ID = 'dialplanId';

    const LUA_CONFIG_NAME = 'extensions-incomingcalls.lua';

    /**
     * The links can be the following:
     * Dialplan -> Dialplan
     * Dialplan -> Hunt group
     * Dialplan <-> Trunk
     * Dialplan -> PBX
     * Dialplan -> Page group
     * Dialplan -> Call group
     * Dialplan -> Switch
     * Dialplan -> Timetable
     * Dialplan -> Phonebook
     * Dialplan -> IVR
     * User -> Dialplan
     */
    const LINK_DIALPLAN   = 1;
    const LINK_TRUNK      = 2;
    const LINK_CALL_GROUP = 3;
    const LINK_USERS      = 4;
    const LINK_HUNT_GROUP = 5;
    const LINK_PBX        = 6;
    const LINK_PAGE_GROUP = 7;
    const LINK_SWITCH     = 8;
    const LINK_TIMETABLE  = 9;
    const LINK_PHONEBOOK  = 10;
    const LINK_IVR        = 11;

    private $_notAllowedDialplanNames = [
        'internalcalls',
        'pbxfeatures',
        'pbxinternal',
        'pbxservices',
        'outgoingfax',
        'faxout',
        'unknowns',
        'ldap_q',
        'queuehelper',
        'parkedcalls',
        'cti-hotel-wake-up-call',
        'ctivideo',
        'mobility',
        'wmsdisa-callback',
        'wmsdisa',
        'cticonnect',
        'predial',
        'hdlr',
    ];

    /**
     * Application service.
     *
     * @var ApplicationService
     */
    private $_appService;

    /**
     * Rules service.
     *
     * @var RulesService
     */
    private $_rulesService;

    /**
     * WMS Database.
     *
     * @var \Wildix\Databases\WMSDB
     */
    private $_db;

    /**
     * Dialplan formatter.
     *
     * @var DialplanFormatter
     */
    private $_formatter;

    /**
     * WMS LDAP class.
     *
     * @var \WMS_Ldap
     */
    private $_wmsLdap;

    /**
     * Logger class.
     *
     * @var \Logger\Log $_logger
     */
    private $_logger;

    /**
     * Errors list.
     *
     * @var array
     */
    private $_errors = [];

    /**
     * Not allowed for delete.
     *
     * @var array $_notAllowedForDelete
     */
    private $_notAllowedForDelete = [
        'users',
        'main',
        'classound',
    ];

    /**
     * Dialplan Service constructor.
     *
     * @param \Wildix\Dialplan\ApplicationService $appService   Application service class.
     * @param \Wildix\Dialplan\RulesService       $rulesService Dialplan rules service class.
     * @param \Wildix\Databases\WMSDB             $wmsDb        WMS Db class.
     * @param \Wildix\CallweaverManager           $cwManager    Callweaver Manager class.
     * @param \Wildix\Dialplan\DialplanFormatter  $formatter    Dialplan formatter class.
     * @param \WMS_Ldap                           $wmsLdap      Wms Ldap class.
     * @param \Logger\Log                         $logger       Logger class.
     */
    public function __construct(
        \Wildix\Dialplan\ApplicationService $appService,
        \Wildix\Dialplan\RulesService $rulesService,
        \Wildix\Databases\WMSDB $wmsDb,
        \Wildix\CallweaverManager $cwManager,
        \Wildix\Dialplan\DialplanFormatter $formatter,
        \WMS_Ldap $wmsLdap,
        \Logger\Log $logger
    ) {
        $this->_appService   = $appService;
        $this->_rulesService = $rulesService;
        $this->_db           = $wmsDb;
        $this->_cwManager    = $cwManager;
        $this->_formatter    = $formatter;
        $this->_wmsLdap      = $wmsLdap;
        $this->_logger       = $logger;
    }

    public function getDialplanTrunkUse():array {
        $result = $this->_db->getRows(
            sprintf(
                'select parameters, rulename from dialplan_rules join dialplan_rulenames on dialplan_rulenames.id = rule_id
                        where  dialplanappname = \'%s\'',
                'app-dial-trunk'
            )
        );
        $arr    = [];
        $result = array_map(
            function ($row) use (&$arr) {
                $arr[$row['rulename']] = explode('","', $row['parameters'])[1] ?? '';
            },
            $result
        );

        return array_filter($arr);
    }

    /**
     * Get dialplan list.
     *
     * @return array
     * @throws \Exception Error.
     */
    public function getDialplans() {
        $result = $this->_db->getRows(
            sprintf(
                'SELECT %s, rulename AS %s, ruledescription AS %s FROM %s',
                self::PARAM_ID,
                self::PARAM_NAME,
                self::PARAM_DESCRIPTION,
                self::DIALPLAN_NAMES_TABLE
            )
        );
        // replace null in description
        $result = array_map(
            function ($row) {
                $row[self::PARAM_DESCRIPTION] = $row[self::PARAM_DESCRIPTION] ? $row[self::PARAM_DESCRIPTION] : '';
                return $row;
            },
            $result
        );

        return $result;
    }

    /**
     * Get dialplan info by ID.
     *
     * @param integer $id Dialplan ID.
     *
     * @return null|array
     * @throws \Exception Error.
     */
    public function getDialplanById(int $id): ?array {
        return $this->_db->first(
            sprintf(
                'SELECT id as %s, rulename AS %s, ruledescription AS %s FROM %s WHERE id = %d',
                self::PARAM_ID,
                self::PARAM_NAME,
                self::PARAM_DESCRIPTION,
                self::DIALPLAN_NAMES_TABLE,
                $id
            )
        );
    }

    /**
     * Get dialplan info by name.
     *
     * @param string $name Dialplan name.
     *
     * @return null|array
     * @throws \Exception Error.
     */
    public function getDialplanByName(string $name): ?array {
        return $this->_db->first(
            sprintf(
                'SELECT id as %s, rulename AS %s, ruledescription AS %s FROM %s WHERE rulename=\'%s\'',
                self::PARAM_ID,
                self::PARAM_NAME,
                self::PARAM_DESCRIPTION,
                self::DIALPLAN_NAMES_TABLE,
                $this->_db->realEscapeString($name)
            )
        );
    }

    /**
     * Get additional data (numbers and includes) for dialplan.
     *
     * @param integer $dialplanId Dialplan ID.
     *
     * @throws \Exception Error.
     * @return array
     */
    public function getAdditionalData(int $dialplanId) {
        return [
            self::PARAM_NUMBERS => $this->_rulesService->getRulesByDialplan($dialplanId),
            self::PARAM_INCLUDES => $this->_getIncludesByDialplanId($dialplanId)
        ];
    }

    /**
     * Get list of includes for dialplan.
     *
     * @param integer $dialplanId Dialplan ID.
     *
     * @throws \Exception Error.
     * @return array
     */
    private function _getIncludesByDialplanId(int $dialplanId): array {
        return $this->_db->getRows(
            sprintf(
                'SELECT include_rule_id as %s FROM %s WHERE rule_id = %d',
                self::PARAM_INCLUDED_DIALPLAN_ID,
                self::DIALPLAN_INCLUDES_TABLE,
                $dialplanId
            )
        );
    }

    /**
     * Create new dialplan
     *
     * @param string $name        Dialplan name.
     * @param string $description Dialplan description.
     * @param array  $numbers     Dialplan numbers data.
     * @param array  $includes    Dialplan includes data.
     *
     * @return integer ID of created dialplan.
     * @throws \Wildix\InvalidDataException Error.
     */
    public function createDialplan(string $name, string $description, array $numbers, array $includes): int {
        $this->_checkCreatedName($name);

        try {
            $this->_db->beginTransaction();

            $this->_db->query(
                sprintf(
                    'INSERT INTO %s (rulename, ruledescription) VALUES (\'%s\', \'%s\')',
                    self::DIALPLAN_NAMES_TABLE,
                    $this->_db->realEscapeString($name),
                    $this->_db->realEscapeString($description)
                )
            );
            $dialplanId = $this->_db->getLastInsertId();

            $this->saveDialplanNumbersAndIncludes($dialplanId, $numbers, $includes);

            $this->_db->endTransaction();
            $this->_logger->info(__LINE__, sprintf('Created dialplan id=%d', $dialplanId));
        } catch (\Exception $e) {
            $this->_db->rollbackTransaction();
            throw new \Wildix\InvalidDataException($e->getMessage(), $e->getErrors());
        }

        $this->saveLuaConfig();

        return $dialplanId;
    }

    /**
     * Check allowed dialplan name.
     *
     * @param string $name Dialplan name.
     *
     * @throws \Wildix\InvalidDataException Error in dialplan name.
     * @return void
     */
    private function _checkCreatedName(string $name): void {
        if (empty($name)) {
             throw new \Wildix\InvalidDataException('', ['name' => \Wildix\Error::getMessage(\Wildix\Error::DialplanNameEmpty)]);
        }

        if (!preg_match('/^[a-z0-9]+[0-9a-z \-\.\_\+]*[a-z0-9\-\.\_\+]+$/ui', $name)) {
            throw new \Wildix\InvalidDataException('', ['name' => _t('error.diaplanInvalidName')]);
        }

        if (in_array($name, $this->_notAllowedDialplanNames)) {
            throw new \Wildix\InvalidDataException('', ['name' => _t('error.diaplanNameNotAllowed', $name)]);
        }

        $dialplan = $this->getDialplanByName($name);
        if ($dialplan !== null) {
            throw new \Wildix\InvalidDataException('', ['name' => \Wildix\Error::getMessage(\Wildix\Error::DialplanNameUnique)]);
        }
    }

    /**
     * Update dialplan.
     *
     * @param integer $dialplanId  Dialplan ID.
     * @param string  $description Dialplan description.
     * @param array   $numbers     Dialplan numbers data.
     * @param array   $includes    Dialplan includes data.
     *
     * @return boolean.
     * @throws Exception Bad data.
     * @throws \Wildix\InvalidDataException Bad data.
     */
    public function updateDialplan(int $dialplanId, string $description, array $numbers, array $includes): bool {
        $dialplan = $this->getDialplanById($dialplanId);
        if ($dialplan === null) {
            throw new Exception(\Wildix\Error::DialplanNotFound, 404);
        }

        try {
            $this->_db->beginTransaction();

            $this->_db->query(
                sprintf(
                    'UPDATE %s SET ruledescription=\'%s\' WHERE id=%d',
                    self::DIALPLAN_NAMES_TABLE,
                    $this->_db->realEscapeString($description),
                    $dialplanId
                )
            );

            $this->_deleteOldData($dialplanId);
            $this->saveDialplanNumbersAndIncludes($dialplanId, $numbers, $includes);

            $this->_db->endTransaction();
            $this->_logger->info(__LINE__, sprintf('Update dialplan success. ID=%d', $dialplanId));
        } catch (\Exception $e) {
            $this->_db->rollbackTransaction();
            throw new \Wildix\InvalidDataException($e->getMessage(), $e->getErrors());
        }
        $this->saveLuaConfig();

        return true;
    }

    /**
     * Save dialplan numbers and includes.
     *
     * @param integer $dialplanId Dialplan ID.
     * @param array   $numbers    Numbers data.
     * @param array   $includes   Includes data.
     *
     * @return void
     *
     * @throws \Wildix\InvalidDataException Bad data.
     */
    private function saveDialplanNumbersAndIncludes(int $dialplanId, array $numbers, array $includes): void {
        if (!empty($numbers)) {
            $saveSuccess = $this->_rulesService->saveRules($dialplanId, $numbers);
            if (!$saveSuccess) {
                throw new \Wildix\InvalidDataException('', $this->_rulesService->getErrors());
            }
        }

        if (!empty($includes)) {
            $this->_saveIncludes($dialplanId, $includes);
        }
    }

    /**
     * Delete previous dialplan data in DB.
     *
     * @param integer $dialplanId Dialplan ID.
     *
     * @return void
     * @throws \Exception Error.
     */
    private function _deleteOldData(int $dialplanId): void {
        $this->_db->query(
            sprintf('DELETE FROM %s WHERE rule_id = %d;', self::DIALPLAN_RULES_TABLE, $dialplanId)
        );
        $this->_db->query(
            sprintf('DELETE FROM %s WHERE rule_id = %d;', self::DIALPLAN_INCLUDES_TABLE, $dialplanId)
        );
        $this->_db->query(
            sprintf(
                'DELETE FROM %s WHERE object_from_type = %d and object_from_id = %d;',
                self::DIALPLAN_LINKS_TABLE,
                self::LINK_DIALPLAN,
                $dialplanId
            )
        );
    }


    /**
     * Save dialplan includes.
     *
     * @param integer $dialplanId Dialplan ID.
     * @param array   $data       Includes data.
     *
     * @return void
     *
     * @throws \Wildix\InvalidDataException Bad data.
     */
    private function _saveIncludes(int $dialplanId, array $data): void {
        $ids = [];
        foreach ($data as $row) {
            if (!empty($row['dialplanId']) && intval($row['dialplanId']) !== 0) {
                $ids[] = (int)$row['dialplanId'];
            }
        }
        $ids = array_unique($ids);

        $priority = 0;
        foreach ($ids as $id) {
            $context = $this->_rulesService->getContextById($id);
            if ($context === null) {
                throw new \Wildix\InvalidDataException(_t('error.diaplanNotExist', $id));
            }
            if ($id === $dialplanId) {
                throw new \Wildix\InvalidDataException(_t('error.diaplanNotAllowedCyclically', $id));
            }

            $this->_db->query(
                sprintf(
                    'INSERT INTO %s (rule_id, priority, include_rule_id) VALUES (%d, %d, %d);',
                    self::DIALPLAN_INCLUDES_TABLE,
                    $dialplanId,
                    $priority,
                    $id
                )
            );
            $priority++;
        }
    }

    /**
     * Delete dialplans.
     *
     * @param array $dialplanIds Dialplan IDs list.
     *
     * @throws \Exception Error.
     * @return boolean
     */
    public function deleteDialplans(array $dialplanIds): bool {
        if ($this->_checkAllowRemoveDialplan($dialplanIds) !== true) {
            return false;
        }
        $idsStr = implode(',', $dialplanIds);

        $queries   = [];
        $queries[] = sprintf(
            'DELETE FROM %s WHERE id IN (%s);',
            self::DIALPLAN_NAMES_TABLE,
            $idsStr
        );

        $queries[] = sprintf(
            'DELETE FROM %s WHERE rule_id IN (%s);',
            self::DIALPLAN_RULES_TABLE,
            $idsStr
        );

        $queries[] = sprintf(
            'DELETE FROM %s WHERE rule_id IN (%s) OR include_rule_id IN (%s);',
            self::DIALPLAN_INCLUDES_TABLE,
            $idsStr,
            $idsStr
        );

        $queries[] = sprintf(
            'DELETE FROM %s WHERE object_from_type = %d AND object_from_id IN (%s);',
            self::DIALPLAN_LINKS_TABLE,
            self::LINK_DIALPLAN,
            $idsStr
        );

        $this->_db->doTransaction($queries);

        $this->saveLuaConfig();

        $this->_logger->info(__LINE__, sprintf('Delete dialplans: %s', $idsStr));

        return true;
    }

    /**
     * Check is allow remove dialplans.
     *
     * @param array $dialplanIds Dialplan IDs.
     *
     * @return boolean
     *
     * @throws \Wildix\InvalidDataException Bad data.
     */
    private function _checkAllowRemoveDialplan(array $dialplanIds): bool {
        $removedNames = [];
        $idsStr       = implode(',', $dialplanIds);

        $data = $this->_db->getRows(
            sprintf(
                'SELECT id, rulename FROM %s WHERE id IN (%s);',
                self::DIALPLAN_NAMES_TABLE,
                $idsStr
            )
        );

        $notExistingDialplans = array_diff($dialplanIds, array_column($data, 'id'));
        if (!empty($notExistingDialplans)) {
            throw new \Wildix\InvalidDataException(_t('error.diaplanNotFoundDelete', implode(',', $notExistingDialplans)));
        }

        foreach ($data as $item) {
            $removedNames[$item['id']] = $item['rulename'];
        }

        $this->_checkLinkedDialplans($removedNames);
        $this->_checkLinkedToOtherObjects($removedNames);
        $this->_checkLinkedToUsers($removedNames);
        $this->_checkLinkedToTrunk($removedNames);

        if (!empty($this->_errors)) {
            throw new \Wildix\InvalidDataException('', $this->_errors);
        }

        return true;
    }

    /**
     * Check link this dialplan to other dialplans.
     *
     * @param array $removedNames Removed dialplan names.
     *
     * @throws \Exception Error.
     * @return void
     */
    private function _checkLinkedDialplans(array $removedNames): void {
        foreach ($removedNames as $removedId => $removedName) {
            $linkNames      = [];
            $linkedDialplan = $this->_db->getRows(
                sprintf(
                    'SELECT rn.rulename FROM %s as rl LEFT JOIN %s as rn ON rn.id = rl.object_from_id
                    WHERE rl.object_from_type = %d
                    AND rl.object_to_type = %d  AND rl.object_to_id = \'%s\' AND rl.object_from_id <> %d',
                    self::DIALPLAN_LINKS_TABLE,
                    self::DIALPLAN_NAMES_TABLE,
                    self::LINK_DIALPLAN,
                    self::LINK_DIALPLAN,
                    $this->_db->realEscapeString($removedName),
                    $removedId
                )
            );
            foreach ($linkedDialplan as $item) {
                $linkNames[] = $item['rulename'];
            }
            if (!empty($linkNames)) {
                $this->_errors[] = _t('error.diaplanCantRemoveLinked', $removedName, join(', ', $linkNames));
            } else if (in_array($removedName, $this->_notAllowedForDelete, true)) {
                $this->_errors[] = _t('error.diaplanCantRemove', $removedName);
            }
        }
    }

    /**
     * Check link this dialplan to other objects.
     *
     * @param array $removedNames Removed dialplan names.
     *
     * @throws \Exception Error.
     * @return void
     */
    private function _checkLinkedToOtherObjects(array $removedNames): void {
        $linkedOther = $this->_db->first(
            sprintf(
                'SELECT count(*) as links_count FROM %s WHERE object_from_type <> %d AND object_to_type = %d AND object_to_id in (%s)',
                self::DIALPLAN_LINKS_TABLE,
                self::LINK_DIALPLAN,
                self::LINK_DIALPLAN,
                '\'' . join('\',\'', array_map([$this->_db, 'realEscapeString'], $removedNames)) . '\''
            )
        );

        if ($linkedOther && $linkedOther['links_count'] > 0) {
            $this->_errors[] = _t('error.diaplanNotAllowedObjects');
        }
    }

    /**
     * Check link this dialplan to users.
     *
     * @param array $removedNames Removed dialplan names.
     *
     * @throws \Exception Error.
     * @return void
     */
    private function _checkLinkedToUsers(array $removedNames): void {
        $filter = '(|';
        foreach ($removedNames as $name) {
            $filter .= '(userDialplan=' . ldap_escape($name) . ')(userFaxDialplan=' . ldap_escape($name) . ')';
        }
        $filter .= ')';

        $users = $this->_wmsLdap->getEntries($this->_wmsLdap::PERSONS, $filter, ['dn']);
        if (!empty($users)) {
            $this->_errors[] = _t('error.diaplanNotAllowedUsers');
        }
    }

    /**
     * Check link this dialplan to trunks.
     *
     * @param array $removedNames Removed dialplan names.
     *
     * @throws \Exception Error.
     * @return void
     */
    private function _checkLinkedToTrunk(array $removedNames): void {
        foreach ($removedNames as $name) {
            $linkTrunks = $this->_db->first(
                sprintf(
                    'SELECT count(*) as links_count FROM trunks WHERE settings_pbx LIKE \'%%context":"%s%%\'',
                    $this->_db->realEscapeString($name)
                )
            );
            if ($linkTrunks && $linkTrunks['links_count'] > 0) {
                $this->_errors[] = _t('error.diaplanNotAllowedTrunks');
            }

            // TODO we will have the problems here for all multiport trunks
            $linkMultiPortTrunks = $this->_db->first(
                sprintf(
                    'SELECT count(*) as links_count FROM pstn_trunks WHERE settings_pbx LIKE \'%%context":"%s%%\'',
                    $this->_db->realEscapeString($name)
                )
            );
            if ($linkMultiPortTrunks && $linkMultiPortTrunks['links_count'] > 0) {
                $this->_errors[] = _t('error.diaplanNotAllowedTrunks');
            }
        }
    }

    /**
     * Save config to Lua file.
     *
     * @return boolean
     * @throws Exception Error.
     */
    public function saveLuaConfig(): bool {
        $config = $this->_getFormattedLuaConfig();

        $tmpFilePath  = '/tmp/' . self::LUA_CONFIG_NAME;
        $destFilePath = PATH_TO_CW_CONF . self::LUA_CONFIG_NAME;

        $tmpFile = fopen($tmpFilePath, 'w');
        if ($config !== '' && $tmpFile) {
            fputs($tmpFile, $config);
        }
        fclose($tmpFile);

        $lockFile = '/tmp/extensions.lock';
        if (file_exists($lockFile)) {
            $handle = fopen($lockFile, 'r+');
            if (flock($handle, LOCK_EX)) {
                rename($tmpFilePath, $destFilePath);
                $this->_logger->info(__LINE__, sprintf('Saved dialplan config %s', $destFilePath));
                flock($handle, LOCK_UN);
            } else {
                $this->_logger->error(__LINE__, sprintf('Error flock file %s', $lockFile));
            }
            fclose($handle);
        } else {
            $this->_logger->info(__LINE__, sprintf('Not found file for lock %s', $lockFile));
            rename($tmpFilePath, $destFilePath);
            $this->_logger->info(__LINE__, sprintf('Saved dialplan config %s', $destFilePath));
        }

        $this->_applyChanges();

        return true;
    }

    /**
     * Reload callweaver and apply changes.
     *
     * @throws Exception Error.
     * @return void
     */
    private function _applyChanges(): void {
        $this->_cwManager->reloadModule('pbx_lua');
        $this->_logger->info(__LINE__, 'Reload callweaver after change dialplans');
    }

    /**
     * Return formatted config for Lua file.
     *
     * @throws \Exception Error.
     * @return string
     */
    private function _getFormattedLuaConfig(): string {
        $rules    = $this->_db->getRows(
            sprintf(
                'SELECT rn.id AS drule_id, rn.rulename AS context, r.number, r.priority AS priority, r.dialplanappname AS app, r.parameters AS appdata
                FROM %s AS r LEFT JOIN %s AS rn ON (rn.id=r.rule_id)
                ORDER BY `context`, `number`, `priority`',
                self::DIALPLAN_RULES_TABLE,
                self::DIALPLAN_NAMES_TABLE
            )
        );
        $includes = $this->_db->getRows(
            sprintf(
                'SELECT rule_id, include_rule_id FROM %s ORDER BY priority',
                self::DIALPLAN_INCLUDES_TABLE
            )
        );
        return $this->_formatter->getLuaConfigStr($rules, $includes);
    }
}
