<?php
declare(strict_types=1);

namespace Wildix\Dialplan\FeatureCodes;

use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\Validator;
use \PBX;

/**
 * Class FeatureCodesService
 */
class FeatureCodesService {

    private $_storage;
    private $_server;

    public function __construct(FeatureCodesStorage $storage, PBX $server) {
        $this->_storage = $storage;
        $this->_server  = $server;
    }

    private function _storage(): FeatureCodesStorage {
        return $this->_storage;
    }

    private function _reloadCli(): void {
        $this->_server->put_cli_reload('features');
        $this->_server->put_cli_reload('pbx_lua');
    }

    public function getSettings(): array {
        $data = $this->_storage()->getSettings();
        return $this->_buildSettings($data);
    }

    private function _buildSettings(array $data): array {
        $res = [
            'preferences' => [],
            'features' => [],
            'services' => [],
            'inCallFeatures' => [],
        ];

        foreach (FeatureCodesStorage::FEATURES as $item) {
            $res['features'][$item] = $data[$item];
        }
        foreach (FeatureCodesStorage::PREFERENCES as $item) {
            $res['preferences'][$item] = $data[$item];
        }
        foreach (FeatureCodesStorage::SERVICES as $item) {
            if (isset($data[$item])) {
                $res['services'][$item] = $data[$item];
            }
        }
        foreach (FeatureCodesStorage::IN_CALL_FEATURES as $item) {
            $res['inCallFeatures'][$item] = $data[$item];
        }

        return $res;
    }

    public function setSettings(array $data): void {
        $this->_validate($data);
        $this->_validateNotDuplicate($this->getSettings(), $data);
        $this->_storage()->setSettings($this->buildToSave($data));
        $this->_reloadCli();
    }

    private function buildToSave($data): array {
        $res = [];

        if (isset($data['preferences']) && is_array($data['preferences'])) {
            $res = array_merge($res, $this->filterFields($data['preferences'], FeatureCodesStorage::PREFERENCES));
        }

        if (isset($data['features']) && is_array($data['features'])) {
            $res = array_merge($res, $this->filterFields($data['features'], FeatureCodesStorage::FEATURES));
        }

        if (isset($data['inCallFeatures']) && is_array($data['inCallFeatures'])) {
            $res = array_merge($res, $this->filterFields($data['inCallFeatures'], FeatureCodesStorage::IN_CALL_FEATURES));
        }

        if (isset($data['services']) && is_array($data['services'])) {
            $res = array_merge($res, $this->filterFields($data['services'], FeatureCodesStorage::SERVICES));
        }

        return $res;
    }

    private function filterFields(array $data, array $fields): array {
        return array_filter(
            $data,
            function ($key) use ($fields) {
                return in_array($key, $fields);
            },
            ARRAY_FILTER_USE_KEY
        );
    }

    private function _validate(array $data) {
        $validator = new Validator($data);
        $validator->rule('array', FeatureCodesStorage::GROUPS);
        foreach (FeatureCodesStorage::PREFERENCES as $item) {
            if (isset($data['preferences']) && $data['preferences'][$item]) {
                $validator->rule('regex', [sprintf('preferences.%s', $item)], '/^[*#0-9]{0,}$/')
                    ->reason(Error::getMessage(Error::IncorrectFormat));
            }
        }

        foreach (FeatureCodesStorage::FEATURES as $item) {
            if (isset($data['features']) && isset($data['features'][$item])) {
                $validator->rule('required', [sprintf('features.%s', $item)])->reason(Error::getMessage(Error::FieldCannotBeEmpty));
                $validator->rule('lengthMax', [sprintf('features.%s', $item)], 8)->reason('Value is to long');
                $validator->rule('featureCode', [sprintf('features.%s', $item)]);
            }
        }

        foreach (FeatureCodesStorage::IN_CALL_FEATURES as $item) {
            if (isset($data['inCallFeatures']) && isset($data['inCallFeatures'][$item])) {
                $validator->rule('required', [sprintf('inCallFeatures.%s', $item)])->reason(Error::getMessage(Error::FieldCannotBeEmpty));
                $validator->rule('lengthMax', [sprintf('inCallFeatures.%s', $item)], 8)->reason('Value is to long');
                $validator->rule('featureCode', [sprintf('inCallFeatures.%s', $item)]);
            }
        }

        foreach (FeatureCodesStorage::SERVICES as $item) {
            if (isset($data['services']) && isset($data['services'][$item])) {
                $validator->rule('required', [sprintf('services.%s', $item)])->reason(Error::getMessage(Error::FieldCannotBeEmpty));
                $validator->rule('lengthMax', [sprintf('services.%s', $item)], 8)->reason('Value is to long');
                $validator->rule('featureCode', [sprintf('services.%s', $item)]);
            }
        }

        if (!$validator->validate()) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $validator->errorsTree());
        }
    }

    private function _validateNotDuplicate(array $current, array $data): void {
        $res     = [];
        $allData = [];
        foreach ($data as $group => $value) {
            if (isset($current[$group]) && is_array($current[$group])) {
                $allData[$group] = array_merge($current[$group], $data[$group]);
            }
        }
        $inCallFeatures = $allData['inCallFeatures'];
        unset($allData['inCallFeatures']);

        foreach ($allData as $group => $items) {
            foreach ($items as $index => $value) {
                array_walk_recursive(
                    $allData,
                    function ($val, $key) use (&$res, $group, $index, $value) {
                        if ($val === $value && $key !== $index) {
                            $res[$group][$index] = [
                            'code' => 'alreadyUsed',
                            'reason' => 'Already used'
                            ];
                        }
                    }
                );
            }
        }

        foreach ($inCallFeatures as $index => $value) {
            array_walk_recursive(
                $inCallFeatures,
                function ($val, $key) use (&$res, $index, $value) {
                    if ($val === $value && $key !== $index) {
                        $res['inCallFeatures'][$index] = [
                        'code' => 'alreadyUsed',
                        'reason' => 'Already used'
                        ];
                    }
                }
            );
        }

        if (!empty($res)) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $res);
        }
    }
}
