<?php

namespace Wildix\Dialplan\GeneralSettings;

use \PBX;
use Wildix\Dialplan\ExtensionsVariables;
use Wildix\Error;
use Wildix\Validator;
use Wildix\InvalidDataException;

/**
 * Class GeneralSettingsService
 */
class GeneralSettingsService {

    private $_storage;
    private $_server;

    public function __construct(GeneralSettingsStorage $storage, PBX $server) {
        $this->_storage = $storage;
        $this->_server  = $server;
    }

    private function _storage(): GeneralSettingsStorage {
        return $this->_storage;
    }

    private function _reloadCli(): void {
        $this->_server->put_cli_reload('features');
        $this->_server->put_cli_reload('pbx_lua');
        $this->_server->put_cli_reload('sip');
        $this->_server->put_cli_reload('pbx_config');
    }

    public function getSettings(): array {
        $data = $this->_storage()->getSettings();
        return $this->_buildSettings($data);
    }

    private function _buildSettings(array $data): array {
        return [
            'faxLang' => $data[ExtensionsVariables::FAXLANG],
            'parkTimeout' => $data[ExtensionsVariables::PARK_TIMEOUT],
            'internationalPrefix' => $data[ExtensionsVariables::INTERNATIONAL_PREFIX],
            'nationalPrefix' => $data[ExtensionsVariables::NATIONAL_PREFIX],
            'intDialTimeout' => $data[ExtensionsVariables::INTDIAL_TIMEOUT],
            'endOfDigit' => $data[ExtensionsVariables::END_OF_DIGIT],
            'buttonSound' => $data[ExtensionsVariables::BUTTON_SOUND] === 'yes',
            'attachFormatVm' => $data[ExtensionsVariables::ATTACH_FORMAT_VM],
            'attachFormatRec' => $data[ExtensionsVariables::ATTACH_FORMAT_REC],
            'recNotify' => $data[ExtensionsVariables::REC_NOTIFY] === 'yes',
            'recAttach' => $data[ExtensionsVariables::REC_ATTACH] === 'yes',
            'describeVoiceMail' => $data[ExtensionsVariables::DESCRIBE_VOICEMAIL] === 'yes',
            'ctiConnectCt3PhoneBooks' => $data[ExtensionsVariables::CTICONNECT3_PHONEBOOKS] === 'yes',
            'userVariables' => $data[ExtensionsVariables::USER_VARIABLES] ?? '',
            'trunkRegistrationNotification' => (bool)$data['trunkRegistrationNotification'],
            'voiceMailMessageMoreInfo' => (bool)$data['voicemail_message_more_info'],
            'quickDialPatterns' => $data['quick_dial_patterns'],
            'prefixExtLine' => $data['prefixextline'],
            'transferDigitTimeout' => $data['transfer_digit_timeout'],
            'transferInterDigitTimeout' => $data['transfer_interdigit_timeout'],
        ];
    }

    public function setSettings(array $data): void {
        $this->_validate($data);
        $this->_storage()->setSettings($this->_buildToSave($data));
        $this->_reloadCli();
    }

    private function _buildToSave($data): array {
        return [
            ExtensionsVariables::FAXLANG => $data['faxLang'],
            ExtensionsVariables::PARK_TIMEOUT => $data['parkTimeout'],
            ExtensionsVariables::INTERNATIONAL_PREFIX => $data['internationalPrefix'],
            ExtensionsVariables::NATIONAL_PREFIX => $data['nationalPrefix'],
            ExtensionsVariables::INTDIAL_TIMEOUT => $data['intDialTimeout'],
            ExtensionsVariables::END_OF_DIGIT => $data['endOfDigit'],
            ExtensionsVariables::BUTTON_SOUND => $this->_yesNo($data['buttonSound']),
            ExtensionsVariables::ATTACH_FORMAT_VM => $data['attachFormatVm'],
            ExtensionsVariables::ATTACH_FORMAT_REC => $data['attachFormatRec'],
            ExtensionsVariables::REC_NOTIFY => $this->_yesNo($data['recNotify']),
            ExtensionsVariables::REC_ATTACH => $this->_yesNo($data['recAttach']),
            ExtensionsVariables::DESCRIBE_VOICEMAIL => $this->_yesNo($data['describeVoiceMail']),
            ExtensionsVariables::CTICONNECT3_PHONEBOOKS => $this->_yesNo($data['ctiConnectCt3PhoneBooks']),
            ExtensionsVariables::USER_VARIABLES => $data['userVariables'],
            'quick_dial_patterns' => $data['quickDialPatterns'],
            'prefixextline' => $data['prefixExtLine'],
            'transfer_digit_timeout' => $data['transferDigitTimeout'],
            'transfer_interdigit_timeout' => $data['transferInterDigitTimeout'],
            'trunkRegistrationNotification' => $data['trunkRegistrationNotification'],
            'voicemail_message_more_info' => $this->_yesNo($data['voiceMailMessageMoreInfo']),
        ];
    }

    private function _yesNo(bool $data): string {
        return $data ? 'yes' : 'no';
    }

    private function _validate(array $data) {
        $validator = new Validator($data);

        $fieldsMinRule      = [
            'prefixExtLine',
            'transferDigitTimeout',
            'internationalPrefix',
            'nationalPrefix',
        ];
        $fieldsNumericRule  = [
            'prefixExtLine',
            'transferDigitTimeout',
            'transferInterDigitTimeout',
            'parkTimeout',
            'intDialTimeout',
            'internationalPrefix',
            'nationalPrefix',
        ];
        $fieldsRequiredRule = [
            'transferDigitTimeout',
            'transferInterDigitTimeout',
        ];

        $validator->rule('min', $fieldsMinRule, 0);
        $validator->rule('numeric', $fieldsNumericRule);
        $validator->rule('required', $fieldsRequiredRule);
        $validator->rule('between', 'transferInterDigitTimeout', [1, 12])->reason('Numeric between 1-12');
        $validator->rule('between', 'parkTimeout', [0, 32767])->reason('Numeric between 0-32767');
        $validator->rule('between', 'intDialTimeout', [10, 600])->reason('Numeric between 10-600');

        if (!$validator->validate()) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $validator->errorsTree());
        }
    }
}
