<?php

namespace Wildix\Dialplan\GeneralSettings;

use Wildix\Databases\DEVDB;
use Wildix\Databases\WMSDB;
use Wildix\Dialplan\ExtensionsVariables;
use Wildix\Dialplan\FeatureCodes\FeatureConfigFile;

/**
 * Class GeneralSettingsStorage
 */
class GeneralSettingsStorage {

    private $_featureConfig;
    private $_voiceMailConfig;
    private $_trunkNotificationConfig;
    private $_extensionsVariables;

    const DEFAULT_DIALPLAN_PARAMS = [
        'GINTERNATIONAL_PREFIX' => '',
        'GNATIONAL_PREFIX' => '',
        'GINTDIAL_TIMEOUT' => '',
        'GPARK_TIMEOUT' => '',
        'GUSER_VARIABLES' => '',
        'GFAXLANG' => '',
        'GCTICONNECT3_PHONEBOOKS' => '',
        'GRECORD_FORMAT' => '',
        'GREC_NOTIFY' => 'yes',
        'GREC_ATTACH' => 'yes',
        'DESCRIBE_VOICEMAIL' => 'yes',
        'GATTACH_FORMAT_VM' => 'mp3',
        'GATTACH_FORMAT_REC' => 'mp3',
        'GEOD' => '2',
        'GBUTTON_SOUND' => 'yes',
        'GPBXIVRTIMEOUT' => '',
        'DYNAMIC_FEATURES' => 'a_callswitch#a_callrec#a_callrecpause',
    ];

    const UNUSED_OPTIONS = [
        'GPREANSWER',
        'GCOUNTRY_CODE',
    ];

    const OLD_ATTACH_FORMAT = 'wv';
    const NEW_ATTACH_FORMAT = 'mp3';

    public function __construct(
        FeatureConfigFile $featureConfig,
        VoiceMailConfigFile $voiceMailConfig,
        TrunkNotificationConfig $trunkNotificationConfig,
        ExtensionsVariables $extensionsVariables
    ) {
        $this->_featureConfig           = $featureConfig;
        $this->_voiceMailConfig         = $voiceMailConfig;
        $this->_trunkNotificationConfig = $trunkNotificationConfig;
        $this->_extensionsVariables     = $extensionsVariables;
    }

    public function getSettings(): array {
        return array_merge(
            $this->getDialplanVariables(),
            $this->getFeatureConfigTimeouts(),
            [
                'prefixextline' => $this->getPrefixExtLine(),
                'quick_dial_patterns' => $this->getQuickDialPatterns(),
                'voicemail_message_more_info' => $this->getVoiceMailMessageMoreInfo(),
                'trunkRegistrationNotification' => $this->getStatusTrunkRegNotification()
            ]
        );
    }

    public function setSettings(array $data): void {
        $this->updateDialplanVariables($data);
        $this->updatePrefixExtLine($data['prefixextline']);
        $this->updateQuickDialPatterns($data['quick_dial_patterns']);
        $this->updateVoiceMailMessageMoreInfo($data['voicemail_message_more_info']);
        $this->updateStatusTrunkRegNotification($data['trunkRegistrationNotification']);
        $this->updateFeatureConfigTimeouts($data);
    }

    private function getFeatureConfigTimeouts(): array {
        $featuresConfig = $this->_featureConfig->get('general');

        return [
            'transfer_digit_timeout' => $featuresConfig['transferdigittimeout'],
            'transfer_interdigit_timeout' => $featuresConfig['transferinterdigittimeout'],
        ];
    }

    private function getDialplanVariables(): array {
        $dialplanParameters = self::DEFAULT_DIALPLAN_PARAMS;

        $extensionsVariables = array_diff_key($this->_extensionsVariables->get(), array_flip(self::UNUSED_OPTIONS));

        foreach ($extensionsVariables as $option => $value) {
            if (isset(self::DEFAULT_DIALPLAN_PARAMS[$option])) {
                $dialplanParameters[$option] = $value;
            } else {
                if (!$this->_extensionsVariables->isInclude($option)) {
                    $option = $option . '=' . $value;
                }
                $dialplanParameters[ExtensionsVariables::USER_VARIABLES] .= $option . PHP_EOL;
            }
        }

        if (is_numeric($dialplanParameters[ExtensionsVariables::PARK_TIMEOUT])) {
            $dialplanParameters[ExtensionsVariables::PARK_TIMEOUT] /= 1000;
        }

        // Update format, wav49 is deprecated
        $dialplanParameters[ExtensionsVariables::ATTACH_FORMAT_VM]  = $this->updateAttachFormat($dialplanParameters[ExtensionsVariables::ATTACH_FORMAT_VM]);
        $dialplanParameters[ExtensionsVariables::ATTACH_FORMAT_REC] = $this->updateAttachFormat($dialplanParameters[ExtensionsVariables::ATTACH_FORMAT_REC]);

        return $dialplanParameters;
    }

    private function updateAttachFormat(string $format): string {
        if ($format === self::OLD_ATTACH_FORMAT) {
            $format = self::NEW_ATTACH_FORMAT;
        }

        return $format;
    }

    private function getPrefixExtLine(): string {
        return WMSDB::getInstance()->getSystemVariable(WMSDB::SYSTEM_EXTERNAL_LINE_PREFIX) ?? '0';
    }

    private function getQuickDialPatterns(): string {
        $quickDialPatterns = DEVDB::getInstance()->getSystemVariable(DEVDB::SYSTEM_QUICK_DIAL_PATTERNS, '');
        return str_replace('|', "\n", $quickDialPatterns);
    }

    private function getVoiceMailMessageMoreInfo(): bool {
        $voiceMailConfig = $this->_voiceMailConfig->get();
        return ($voiceMailConfig['saycid'] === 'yes' && $voiceMailConfig['envelope'] === 'yes');
    }

    private function getStatusTrunkRegNotification(): string {
        return $this->_trunkNotificationConfig->get('start', 'general');
    }

    private function updateDialplanVariables(array $data): void {
        $paramsToSave = [];
        foreach (self::DEFAULT_DIALPLAN_PARAMS as $option => $value) {
            $paramsToSave[$option] = $data[$option] ?? '';
        }

        $paramsToSave[ExtensionsVariables::PARK_TIMEOUT]    *= 1000;
        $paramsToSave[ExtensionsVariables::COUNTRY_CODE]     = $this->_extensionsVariables->get(ExtensionsVariables::COUNTRY_CODE);
        $paramsToSave[ExtensionsVariables::DYNAMIC_FEATURES] = self::DEFAULT_DIALPLAN_PARAMS['DYNAMIC_FEATURES'];

        $userVariables = $this->_prepareCustomVariables($paramsToSave[ExtensionsVariables::USER_VARIABLES]);
        unset($paramsToSave[ExtensionsVariables::USER_VARIABLES]);

        $paramsToSave = array_merge($paramsToSave, $userVariables);

        $this->_extensionsVariables->setData($paramsToSave);
        $this->_extensionsVariables->save();
    }

    private function _prepareCustomVariables(string $variables): array {
        $userVariables = [];

        if ($variables) {
            $variablesExp   = array_filter(explode(PHP_EOL, $variables));
            $dialplanParams = array_merge(self::DEFAULT_DIALPLAN_PARAMS, array_flip(self::UNUSED_OPTIONS));

            foreach ($variablesExp as $variableStr) {
                if (!$this->_extensionsVariables->isInclude($variableStr)) {
                    $variableStr = str_replace("\r", '', $variableStr);
                }

                $variableStr = explode('=', $variableStr);
                $key         = trim($variableStr[0], ' ');
                $value       = trim($variableStr[1] ?? '', ' ');

                if (array_key_exists($key, $dialplanParams)) {
                    continue;
                }

                $userVariables[$key] = $value;
            }
        }

        return $userVariables;
    }

    private function updatePrefixExtLine(string $code): void {
        WMSDB::getInstance()->setSystemVariable(WMSDB::SYSTEM_EXTERNAL_LINE_PREFIX, $code);
    }

    private function updateQuickDialPatterns(string $patterns): void {
        $patterns = str_replace(' ', '', $patterns);
        $patterns = str_replace("\n", '|', $patterns);
        DEVDB::getInstance()->setSystemVariable(DEVDB::SYSTEM_QUICK_DIAL_PATTERNS, $patterns);
    }

    private function updateVoiceMailMessageMoreInfo(string $status): void {
        $data['saycid']   = $status;
        $data['envelope'] = $status;

        $this->_voiceMailConfig->setData($data);
        $this->_voiceMailConfig->save();

        exec('callweaver -rx "voicemail reload"');
    }

    private function updateStatusTrunkRegNotification(bool $status): void {
        $this->_trunkNotificationConfig->set('start', (int)$status, 'general');
        $this->_trunkNotificationConfig->save();
    }

    private function updateFeatureConfigTimeouts(array $timeouts): void {
        $this->_featureConfig->set('transferinterdigittimeout', $timeouts['transfer_interdigit_timeout'], 'general');
        $this->_featureConfig->set('transferdigittimeout', $timeouts['transfer_digit_timeout'], 'general');
        $this->_featureConfig->save();
    }
}
