<?php

namespace Wildix\Dialplan\Ivr;

use \Wildix\Core\BaseModel;

/**
 * Class Ivr
 */
class Ivr extends BaseModel {

    const FIELD_ID        = 'id';
    const FIELD_PARAMS    = 'parameters';
    const FIELD_NUMBER    = 'number';
    const FIELD_COMMAND   = 'command';
    const FIELD_PARENT_ID = 'parent_id';
    const FIELD_TREE_ID   = 'treeid';
    const FIELD_ITEMS     = 'items';
    const FIELD_TYPE      = 'type';
    const FIELD_NAME      = 'title';

    const TYPE_MENU = 'menu';
    const TYPE_ITEM = 'item';

    const ALLOWED_TYPES = [
        self::TYPE_MENU,
        self::TYPE_ITEM,
    ];

    const PARAM_SOUND    = 'sound';
    const PARAM_TIMES    = 'times';
    const PARAM_TIMEOUT  = 'timeout';
    const PARAM_BACKCODE = 'backcode';
    const PARAM_CONTEXT  = 'context';
    const PARAM_EXTEN    = 'exten';
    const PARAM_PRIORITY = 'priority';

    const PARAMS_DELIMITER = '|';

    /**
     * 0-3 only menu params
     * 4-6 menu & item params
     */
    const PARAMS_MAP = [
        0 => self::PARAM_SOUND,
        1 => self::PARAM_TIMES,
        2 => self::PARAM_TIMEOUT,
        3 => self::PARAM_BACKCODE,
        4 => self::PARAM_CONTEXT,
        5 => self::PARAM_EXTEN,
        6 => self::PARAM_PRIORITY,
    ];

    const INTEGER_PARAMS = [
        self::PARAM_TIMES,
        self::PARAM_TIMEOUT,
    ];

    public function __construct(?array $properties) {
        parent::__construct($properties);

        if (!empty($properties[self::FIELD_PARAMS])) {
            $this->setParameters($this->_parametersStrToArr($properties[self::FIELD_PARAMS]));
            $this->_setType($this->calcType());
        }

        if ($this->_data && !isset($this->_data[self::FIELD_ITEMS])) {
            $this->setItems([]);
        }
    }

    public function getName(): string {
        return $this->getProperty(self::FIELD_NUMBER, '');
    }

    public function getCommand(): string {
        return $this->getProperty(self::FIELD_COMMAND, '');
    }

    public function getType(): string {
        return $this->getProperty(self::FIELD_TYPE, self::TYPE_MENU);
    }

    public function getParameters(): array {
        return $this->getProperty(self::FIELD_PARAMS, []);
    }

    public function getParentId(): int {
        return $this->getProperty(self::FIELD_PARENT_ID, 0);
    }

    public function getTreeId(): int {
        return $this->getProperty(self::FIELD_TREE_ID, 0);
    }

    public function getItems(): array {
        return $this->getProperty(self::FIELD_ITEMS, []);
    }

    public function setItems(array $data = []): void {
        $this->setProperty(self::FIELD_ITEMS, $data);
    }

    public function setParameters(array $data = []): void {
        $this->setProperty(self::FIELD_PARAMS, $data);
    }

    private function _setType(string $type): void {
        $this->setProperty(self::FIELD_TYPE, $type);
    }

    /**
     * Parse saved string to params
     *
     * @param string $str Saved string.
     *
     * @return array
     */
    private function _parametersStrToArr(string $str): array {
        $params = explode(self::PARAMS_DELIMITER, $str);

        $data = [];
        foreach (self::PARAMS_MAP as $k => $v) {
            $data[$v] = isset($params[$k]) ? trim($params[$k]) : '';
            if (in_array($v, self::INTEGER_PARAMS)) {
                $data[$v] = (int)$data[$v];
            }
        }

        return $data;
    }

    /**
     * Calculate model type (menu|item)
     *
     * @return string
     */
    protected function calcType(): string {
        $params = $this->getParameters();

        return (
            empty($params[self::PARAM_SOUND])
            && empty($params[self::PARAM_TIMES])
            && empty($params[self::PARAM_TIMEOUT])
            && empty($params[self::PARAM_BACKCODE])
        ) ? self::TYPE_ITEM : self::TYPE_MENU;
    }

    public function isMenu(): bool {
        return $this->calcType() === self::TYPE_MENU;
    }
}
