<?php

namespace Wildix\Dialplan\Ivr;

use Wildix\Dialplan\BaseDialplansService;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\Validator;

/**
 * Class IvrService
 */
class IvrService extends BaseDialplansService {

    /**
     * IVR validator.
     *
     * @var IvrValidator $_validator
     */
    private $_validator;

    public function setValidator(IvrValidator $validator) {
        $this->_validator = $validator;
    }

    /**
     * Create IVR item
     *
     * @param array $data IVR item data.
     *
     * @throws InvalidDataException If error validation.
     * @return void
     */
    public function create(array $data = []): void {
        $this->validate($data);
        $data['id'] = $this->getAvailableId();
        $this->storage()->create($data);
    }

    /**
     * Update IVR item
     *
     * @param integer $identity Item id.
     * @param array   $data     IVR data.
     *
     * @throws InvalidDataException If error validation.
     * @return void
     */
    public function update(int $identity, array $data = []): void {
        $this->getItem($identity, true);
        $this->validate($data, $identity);
        $this->storage()->update($identity, $data);
    }

    /**
     * Remove IVR item
     *
     * @param integer $identity IVR tree id.
     *
     * @return boolean
     *
     * @throws InvalidDataException InvalidDataException.
     */
    public function remove(int $identity): bool {
        $item   = $this->getItem($identity, true);
        $linked = $this->storage()->getLinked($item->getName());

        if (count($linked)) {
            throw new InvalidDataException(sprintf('You can’t remove %s. It’s linked to %s ', $item->getName(), implode(', ', $linked)), []);
        }

        $this->storage()->remove($identity);

        return true;
    }

    /**
     * Build IVR item for API response
     *
     * @param Ivr     $item   IVR model.
     * @param boolean $isRoot Is root level.
     *
     * @return array
     */
    public function buildItem(Ivr $item, $isRoot = true) {
        if ($isRoot) {
            $data = [
                Ivr::FIELD_ID   => $item->getTreeId(),
                Ivr::FIELD_NAME => $item->getName(),
            ];
        } else {
            $data = [Ivr::FIELD_NUMBER => $item->getName()];
        }
        $data[Ivr::FIELD_TYPE] = $item->getType();

        if ($item->getType() === Ivr::TYPE_MENU) {
            $data[Ivr::FIELD_PARAMS] = $item->getParameters();

            $child = [];
            foreach ($item->getItems() as $childItem) {
                $child[] = $this->buildItem($childItem, false);
            }
            $data[Ivr::FIELD_ITEMS] = $child;
        } else {
            $data[Ivr::FIELD_PARAMS] = array_intersect_key($item->getParameters(), array_flip([Ivr::PARAM_CONTEXT, Ivr::PARAM_EXTEN]));
        }

        return $data;
    }

    /**
     * Validate IVR data
     *
     * @param array   $data IVR input data.
     * @param integer $id   IVR id.
     *
     * @return void
     * @throws InvalidDataException InvalidDataException
     */
    protected function validate(array $data, int $id = null): void {
        $this->_validator->init($data, $id);
        $items = $this->getListTreeId();
        if ($id) {
            unset($items[$id]);
        }

        if (in_array((string)$data['title'], $items)) {
            $this->_validator->addUniqueNameError();
        }

        if (!$this->_validator->validate()) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $this->_validator->errors());
        }
    }

    protected function getAvailableId(): int {
        return $this->storage()->getMaxTreeId() + 1;
    }

    private function getListTreeId(): array {
        $items = [];
        foreach ($this->storage()->getItems() as $item) {
            $items[$item->getTreeId()] = $item->getName();
        }

        return $items;
    }
}
