<?php

namespace Wildix\Dialplan\Ivr;

use Wildix\Databases\WMSDB;

/**
 * Class IvrStorage
 */
class IvrStorage {

    const DIALPLAN_LINK = 1;
    const LINK_GROUP    = 11;

    private const GET_ITEMS = 'SELECT * FROM ' . WMSDB::TABLE_IVR;

    private $_db;

    private function _db(): WMSDB {
        if (!$this->_db) {
            $this->_db = WMSDB::getInstance();
        }
        return $this->_db;
    }

    /**
     * Get Ivr array items
     *
     * @return array Model items.
     *
     * @throws \Exception Error read database.
     */
    public function getItems(): array {
        return $this->_db()->getRows(self::GET_ITEMS . ' WHERE parent_id = 0 ORDER BY `treeid`', null, Ivr::class);
    }

    private function _buildTree(array $root, array $child) {
        $tree         = [];
        $childByParen = [];

        foreach ($child as $childItem) {
            $childByParen[$childItem->getParentId()][] = $childItem;
        }

        foreach ($root as $item) {
            $tree[$item->getTreeId()] = $this->_treeAddChild($item, $childByParen);
        }

        return $tree;
    }

    private function _treeAddChild(Ivr $item, array $childByParent) {
        $data = isset($childByParent[$item->getId()]) ? $childByParent[$item->getId()] : [];

        $child = [];
        foreach ($data as $childItem) {
            $child[] = $this->_treeAddChild($childItem, $childByParent);
        }
        $item->setItems($child);

        return $item;
    }

    /**
     * Get Ivr Model item
     *
     * @param integer $identity Ivr ID.
     *
     * @return Ivr|null
     * @throws \Exception Error read database.
     */
    public function getItem(int $identity): ?Ivr {
        $item = $this->_db()->first(self::GET_ITEMS . ' WHERE treeid=' . $identity);
        if ($item) {
            $item = new Ivr($item);
        }
        return $item;
    }

    /**
     * Create IVR tree
     *
     * @param array $data Tree data.
     *
     * @return void
     */
    public function create(array $data): void {
        $this->_db()->beginTransaction();
        $this->_saveItem($data, 0, $data['id']);
        $this->_db()->endTransaction();
    }

    private function _saveItem(array $data, int $parentId = 0, int $treeId = 0) {
        $isMenu = $data[Ivr::FIELD_TYPE] === Ivr::TYPE_MENU;

        $saveData = [
            Ivr::FIELD_PARENT_ID => (int)$parentId,
            Ivr::FIELD_COMMAND   => $isMenu ? 'node' : '',
            Ivr::FIELD_PARAMS    => $this->_buildParams($data[Ivr::FIELD_PARAMS]),
            Ivr::FIELD_NUMBER    => trim((string)($parentId === 0 ? $data[Ivr::FIELD_NAME] : $data[Ivr::FIELD_NUMBER])),
            Ivr::FIELD_TREE_ID   => (int)$treeId,
        ];

        $this->_db()->insert(WMSDB::TABLE_IVR, $saveData);
        $thisId = $this->_db()->getLastInsertId();

        if ($isMenu && !empty($data[Ivr::FIELD_ITEMS])) {
            foreach ($data[Ivr::FIELD_ITEMS] as $item) {
                $this->_saveItem($item, $thisId, $treeId);
            }
        }
    }

    private function _buildParams(array $params): string {
        $data = [];
        foreach (Ivr::PARAMS_MAP as $key) {
            $data[] = trim((string)($params[$key] ?? ''));
        }

        return implode(Ivr::PARAMS_DELIMITER, $data);
    }

    /**
     * Update Ivr tree
     *
     * @param integer $identity Ivr tree ID.
     * @param array   $data     Ivr data.
     *
     * @return void
     */
    public function update(int $identity, array $data): void {
        $this->_db()->beginTransaction();
        $this->_removeTree($identity);
        $this->_saveItem($data, 0, $identity);
        $this->_db()->endTransaction();
    }

    public function getLinked($identity): array {
        // phpcs:disable Generic.Files.LineLength.TooLong
        $res = $this->_db()->getRows(
            sprintf(
                'SELECT rn.rulename FROM %s as rl LEFT JOIN %s as rn ON rn.id=rl.object_from_id WHERE rl.object_from_type=%s AND rl.object_to_type=%s AND rl.object_to_id="%s"',
                WMSDB::TABLE_RULES_LINKS,
                WMSDB::DIALPLAN_RULE_NAME,
                self::DIALPLAN_LINK,
                self::LINK_GROUP,
                $identity
            )
        );
        // phpcs:enable Generic.Files.LineLength.TooLong
        return $res ? array_column($res, 'rulename') : [];
    }

    /**
     * Delete Ivr tree
     *
     * @param integer $identity Ivr tree ID.
     *
     * @return void
     */
    public function remove(int $identity): void {
        $this->_db()->beginTransaction();
        $this->_removeTree($identity);
        $this->_db()->endTransaction();
    }

    private function _removeTree(int $treeId) {
        $this->_db()->exec(sprintf('DELETE FROM %s WHERE treeid=%d', WMSDB::TABLE_IVR, $treeId));
    }

    /**
     * Get max Ivr tree ID in database.
     *
     * @return integer
     * @throws \Exception DB exception.
     */
    public function getMaxTreeId(): int {
        $row = $this->_db()->first('SELECT MAX(treeid) as id FROM ' . WMSDB::TABLE_IVR);

        return $row['id'] ?? 0;
    }

    public function loadRelations(array $items) {
        $child = $this->_db()->getRows(self::GET_ITEMS . ' ORDER BY `parent_id`, `number`', null, Ivr::class);

        return $this->_buildTree($items, $child);
    }
}
