<?php

namespace Wildix\Dialplan\Ivr;

use Wildix\Error;
use Wildix\Sounds\Storage;
use Wildix\Validator;

/**
 * Class IvrValidator
 */
class IvrValidator {

    const AVAILABLE_NUMBERS = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '*', '#'];

    /**
     * Validator class
     *
     * @var Validator $_validator
     */
    private $_validator;

    /**
     * Sound storage
     *
     * @var Storage $_soundStorage
     */
    private $_soundStorage;

    private $_data = [];

    private $_id;

    private $_allowedContexts = [];

    public function setContexts(array $contexts) {
        $this->_allowedContexts = $contexts;
    }

    public function setSoundStorage(Storage $storage) {
        $this->_soundStorage = $storage;
    }

    /**
     * Sound storage class
     *
     * @return Storage
     */
    private function _sound(): Storage {
        return $this->_soundStorage;
    }

    /**
     * Init validator with data
     *
     * @param array        $data IVR input data.
     * @param integer|null $id   IVR id.
     *
     * @return void
     */
    public function init(array $data, int $id = null): void {
        $this->_validator = new Validator($data);

        $this->_data = $data;
        $this->_id   = $id;
        $this->_buildRules();
    }

    /**
     * Validate data
     *
     * @return boolean
     */
    public function validate(): bool {
        return $this->_validator->validate();
    }

    /**
     * Add error unique name
     *
     * @return void
     */
    public function addUniqueNameError(): void {
        $this->_validator->addCustomError(Ivr::FIELD_NAME, Error::MustBeUnique, 'Already present');
    }

    public function errors() {
        return $this->_validator->errors();
    }

    private function _buildRules() {
        $this->_addItemRule($this->_data, '');
    }

    private function _addItemRule($data, $path = '') {
        if (empty($path)) {
            $this->_addRulesForRootElement($path);
        } else {
            $this->_addRulesForChildElement($path);
            $this->_validator->rule(
                'NotIn',
                $this->_p($path, Ivr::FIELD_NUMBER),
                [$data[Ivr::FIELD_PARAMS][Ivr::PARAM_BACKCODE]]
            )->reason('It cannot be named backcode');
        }
        $this->_validator->rule(
            'in',
            Ivr::FIELD_TYPE,
            [
                Ivr::TYPE_MENU,
                Ivr::TYPE_ITEM,
            ]
        );
        $this->_validator->rule(
            'array',
            [
                $this->_p($path, Ivr::FIELD_ITEMS),
                $this->_p($path, Ivr::FIELD_PARAMS),
            ]
        );
        $this->_validator->rule('required', [$this->_p($path, Ivr::FIELD_PARAMS)]);
        $this->_validator->rule('required', [Ivr::FIELD_TYPE]);

        $type = (!empty($data[Ivr::FIELD_TYPE]) && $data[Ivr::FIELD_TYPE] === Ivr::TYPE_MENU) ? Ivr::TYPE_MENU : Ivr::TYPE_ITEM;

        $this->_addRulesForParams($data[Ivr::FIELD_PARAMS] ?? [], $type, $path);

        if ($type === Ivr::TYPE_MENU) {
            $this->_validator->rule('required', [$this->_p($path, Ivr::FIELD_ITEMS)]);

            $this->_addRulesForChild($data, $path);
        }
    }

    private function _addRulesForRootElement($path) {
        $this->_validator->rule('in', $this->_p($path, Ivr::FIELD_TYPE), Ivr::ALLOWED_TYPES);
        $this->_validator->rule(
            'regex',
            [$this->_p($path, Ivr::FIELD_NAME)],
            '/^([\w\.-])+$/i'
        )->reason('Only 0-9, a-z, A-Z, -, _, . characters allowed');
        $this->_validator->rule(
            'required',
            [
                $this->_p($path, Ivr::FIELD_NAME),
                $this->_p($path, Ivr::FIELD_TYPE),
                $this->_p($path, Ivr::FIELD_PARAMS),
            ]
        );
    }

    private function _addRulesForChildElement($path) {
        $this->_validator->rule('in', $this->_p($path, Ivr::FIELD_TYPE), Ivr::ALLOWED_TYPES);
        $this->_validator->rule('in', $this->_p($path, Ivr::FIELD_NUMBER), self::AVAILABLE_NUMBERS);
        $this->_validator->rule(
            'required',
            [
                $this->_p($path, Ivr::FIELD_NUMBER),
                $this->_p($path, Ivr::FIELD_TYPE),
                $this->_p($path, Ivr::FIELD_PARAMS),
            ]
        );
    }

    private function _addRulesForParams($paramsData, $type, $path) {
        if (!empty($paramsData[Ivr::PARAM_CONTEXT])) {
            $this->_validator->rule('in', $this->_p($path, Ivr::FIELD_PARAMS . '.' . Ivr::PARAM_CONTEXT), $this->_allowedContexts);
        }

        if ($type === Ivr::TYPE_MENU) {
            $requiredFields = [];

            if ($this->_id) {
                $this->_validateSound((string)($paramsData[Ivr::PARAM_SOUND] ?? ''), $path);
                $requiredFields[] = $this->_p($path, Ivr::FIELD_PARAMS . '.' . Ivr::PARAM_SOUND);
            }

            $this->_validator->rule('unsignedInteger', $this->_p($path, Ivr::FIELD_PARAMS . '.' . Ivr::PARAM_TIMES));
            $this->_validator->rule('between', $this->_p($path, Ivr::FIELD_PARAMS . '.' . Ivr::PARAM_TIMES), [1, 30])
                ->reason('Number is not in range 1-30');

            if (!empty($paramsData[Ivr::PARAM_TIMEOUT])) {
                $this->_validator->rule('unsignedInteger', $this->_p($path, Ivr::FIELD_PARAMS . '.' . Ivr::PARAM_TIMEOUT));
            }

            if (!empty($paramsData[Ivr::PARAM_BACKCODE])) {
                $this->_validator->rule('in', $this->_p($path, Ivr::FIELD_PARAMS . '.' . Ivr::PARAM_BACKCODE), self::AVAILABLE_NUMBERS)
                    ->reason('Only 0-9,*,# characters allowed');
            }

            $requiredFields[] = $this->_p($path, Ivr::FIELD_PARAMS . '.' . Ivr::PARAM_TIMES);
            $this->_validator->rule('required', $requiredFields);
        }
    }

    private function _validateSound(string $sound, string $path): void {
        $sound = str_replace('00000/', '', $sound);
        $item  = $this->_sound()->getItem($sound);
        if (empty($item)) {
            $this->_validator->addCustomError(
                $this->_p($path, Ivr::FIELD_PARAMS . '.' . Ivr::PARAM_SOUND),
                Error::NotFound,
                Error::getMessage(Error::NotFound)
            );
        }
    }

    private function _addRulesForChild($data, $path) {
        $items = !empty($data[Ivr::FIELD_ITEMS]) ? $data[Ivr::FIELD_ITEMS] : [];

        foreach ($items as $k => $item) {
            $this->_addItemRule($item, $this->_p($path, Ivr::FIELD_ITEMS . '.' . $k));
            $this->_validator->rule(
                'NotIn',
                $this->_p($path, Ivr::FIELD_ITEMS . '.' . $k . '.' . Ivr::FIELD_NUMBER),
                [$data[Ivr::FIELD_PARAMS][Ivr::PARAM_BACKCODE]]
            )->reason('It cannot be equal to parent node backcode.');
        }

        $this->_checkDistinctNumbers($items, $path);
    }

    private function _checkDistinctNumbers($items, $path) {
        $numbers = [];
        foreach ($items as $k => $item) {
            $number = (string)($item[Ivr::FIELD_NUMBER] ?? '');
            if (!in_array($number, self::AVAILABLE_NUMBERS, true)) {
                continue;
            }
            $numbers[$number][] = $k;
        }

        foreach ($numbers as $items) {
            if (count($items) > 1) {
                foreach ($items as $item) {
                    $this->_validator->addCustomError(
                        $this->_p($path, Ivr::FIELD_ITEMS . '.' . $item . '.' . Ivr::FIELD_NUMBER),
                        Error::MustBeUnique,
                        'Already present in the node'
                    );
                }
            }
        }
    }

    private function _p(string $path, string $name) {
        return $path ? $path . '.' . $name : $name;
    }
}
