<?php

namespace Wildix\Dialplan;

use Api\Dictionary\Enum\DictionaryTypeEnum;
use Wildix\License;

/**
 * Class RulesService
 */
class RulesService {

    const PARAM_APPS = 'apps';

    const PARAM_NUMBER = 'number';

    const PARAM_APP_NAME = 'name';

    const PARAM_APP_PARAMS = 'params';

    const FIELD_NUMBER = 'number';

    const FIELD_APPLICATION = 'application';

    const FIELD_PARAMETERS = 'parameters';

    const FIELD_PRIORITY = 'priority';

    const FIELD_RULE_ID = 'rule_id';

    const FIELD_DIALPLAN_APP_NAME = 'dialplanappname';

    const DEFAULT_NUMBER_IN_API      = 'default';
    const DEFAULT_NUMBER_IN_DIALPLAN = 's';

    /**
     * Application service.
     *
     * @var ApplicationService $_appService
     */
    private $_appService;

    /**
     * WMS Database.
     *
     * @var \Wildix\Databases\WMSDB $_db
     */
    private $_db;

    /**
     * Dialplan formatter.
     *
     * @var DialplanFormatter
     */
    private $_formatter;

    /**
     * Validator rule params.
     *
     * @var \Wildix\Dialplan\RulesValidator
     */
    private $_validator;

    /**
     * Logger class.
     *
     * @var \Logger\Log $_logger
     */
    private $_logger;

    /**
     * Dictionaries class.
     *
     * @var Dictionaries
     */
    private $_dictionaries;

    private $_links = [];

    private $_errors = [];

    public function __construct(
        \Wildix\Dialplan\ApplicationService $appService,
        \Wildix\Dialplan\DialplanFormatter $formatter,
        \Wildix\Databases\WMSDB $wmsDb,
        \Wildix\Dictionaries $dictionaries,
        \Wildix\Dialplan\RulesValidator $validator,
        \Logger\Log $logger
    ) {
        $this->_appService   = $appService;
        $this->_formatter    = $formatter;
        $this->_db           = $wmsDb;
        $this->_dictionaries = $dictionaries;
        $this->_validator    = $validator;
        $this->_logger       = $logger;
    }

    /**
     * Get rules list for dialplan.
     *
     * @param integer $dialplanId Dialplan ID.
     *
     * @return array
     * @throws \Exception Error.
     */
    public function getRulesByDialplan(int $dialplanId) {
        $result = [];

        $rows = $this->_db->query(
            sprintf(
                'SELECT %s, %s, %s FROM %s WHERE rule_id=%d ORDER BY %1$s, %s',
                self::FIELD_NUMBER,
                self::FIELD_APPLICATION,
                self::FIELD_PARAMETERS,
                \Wildix\Dialplan\DialplanService::DIALPLAN_RULES_TABLE,
                $dialplanId,
                self::FIELD_PRIORITY
            )
        );

        while ($row = $rows->fetchArray(SQLITE3_ASSOC)) {
            $app = $this->_appService->getApplication($row[self::FIELD_APPLICATION]);
            if (!$app) {
                continue;
            }

            // fix WMS-7290 change number 's' to 'default'
            if (strtolower($row[self::FIELD_NUMBER]) === self::DEFAULT_NUMBER_IN_DIALPLAN) {
                $row[self::FIELD_NUMBER] = self::DEFAULT_NUMBER_IN_API;
            }

            $savedString = $row[self::FIELD_PARAMETERS] ? $row[self::FIELD_PARAMETERS] : '';

            $result[$row[self::FIELD_NUMBER]][self::PARAM_NUMBER] = $row[self::FIELD_NUMBER];
            $result[$row[self::FIELD_NUMBER]][self::PARAM_APPS][] = [
                self::PARAM_APP_NAME => $app->getName(),
                self::PARAM_APP_PARAMS => $this->_formatter->savedStringToParamsArray($app, $savedString),
            ];
        }
        $rows->finalize();

        return array_values($result);
    }

    /**
     * Save rules for dialplan.
     *
     * @param integer $dialplanId Dialplan ID.
     * @param array   $data       Numbers and rules.
     *
     * @return boolean
     */
    public function saveRules(int $dialplanId, array $data): bool {
        $addedNumbers = [];
        foreach ($data as $idx => $numbers) {
            if (isset($numbers[self::PARAM_NUMBER]) && is_string($numbers[self::PARAM_NUMBER]) && $numbers[self::PARAM_NUMBER] !== '') {
                $number = trim($numbers[self::PARAM_NUMBER]);

                // fix WMS-7290 change number 'default' to 's'
                if (strtolower($number) === self::DEFAULT_NUMBER_IN_API) {
                    $number = self::DEFAULT_NUMBER_IN_DIALPLAN;
                }
            } else {
                $this->_addError(_t('error.ruleEmptyNumber', $idx), __LINE__);
                continue;
            }

            $rules = !empty($numbers[self::PARAM_APPS]) ? $numbers[self::PARAM_APPS] : [];
            if (!is_array($rules)) {
                $this->_addError(_t('error.ruleNumberFormat', $number), __LINE__);
                continue;
            }

            if (in_array($number, $addedNumbers, true)) {
                $this->_addError(_t('error.ruleNumberDuplicate', $number), __LINE__);
                continue;
            }

            $addedNumbers[] = $number;
            $priority       = 0;
            foreach ($rules as $rule) {
                $inputParams = !empty($rule[self::PARAM_APP_PARAMS]) ? $rule[self::PARAM_APP_PARAMS] : [];

                if (!is_array($rule)
                    || !is_array($inputParams)
                    || empty($rule[self::PARAM_APP_NAME])
                    || !is_string($rule[self::PARAM_APP_NAME])
                ) {
                    $this->_addError(_t('error.ruleFormat', $idx), __LINE__);
                    continue;
                }

                $rule[self::PARAM_APP_PARAMS] = $this->_restoreOriginalParamNames($rule[self::PARAM_APP_NAME], $inputParams);

                if ($this->_prepareRuleQuery($dialplanId, $number, $rule, $priority)) {
                    $this->_prepareLinks($dialplanId, $rule);
                    $priority++;
                }
            }
        }

        $this->_addUniqueLinks();

        return empty($this->_errors);
    }

    /**
     * Get context by ID.
     *
     * @param integer $id Context ID.
     *
     * @return string|null
     */
    public function getContextById(int $id): ?string {
        $contexts = $this->_dictionaries->getDictionary('contexts');

        return $contexts[$id] ?? null;
    }

    /**
     * Restore original input param names for current application.
     *
     * @param string $appName     Application name.
     * @param array  $inputParams Param name.
     *
     * @return array
     */
    private function _restoreOriginalParamNames(string $appName, array $inputParams): array {
        $params = [];
        foreach ($inputParams as $key => $val) {
            $params[$this->_formatter->convertParamNameApiToConfig($appName, $key)] = $val;
        }

        return $params;
    }

    /**
     * Prepare rule SQL query.
     *
     * @param integer $dialplanId Dialplan ID.
     * @param string  $number     Dialplan number.
     * @param array   $rule       Rule data.
     * @param integer $priority   Priority.
     *
     * @return boolean
     */
    private function _prepareRuleQuery(int $dialplanId, string $number, array $rule, int $priority): bool {
        $appName = (string)$rule[self::PARAM_APP_NAME];

        $app = $this->_appService->getApplication($appName);
        if ($app === null) {
            $this->_addError(_t('error.ruleInvalidApplication', $appName ? $appName : 'EMPTY', $number), __LINE__);
            return false;
        }

        $template = $this->_checkParamsAndPrepareTemplateToDbFormat($app, $number, $rule[self::PARAM_APP_PARAMS]);
        if (count($this->_errors) !== 0) {
            return false;
        }

        $this->_db->query(
            sprintf(
                'INSERT INTO %s (%s, %s, %s, %s, %s, %s) VALUES (%d, \'%s\', %d, \'%s\', \'%s\', \'%s\');',
                \Wildix\Dialplan\DialplanService::DIALPLAN_RULES_TABLE,
                self::FIELD_RULE_ID,
                self::FIELD_NUMBER,
                self::FIELD_PRIORITY,
                self::FIELD_APPLICATION,
                self::FIELD_PARAMETERS,
                self::FIELD_DIALPLAN_APP_NAME,
                $dialplanId,
                $this->_db->realEscapeString($this->_formatter->formatInputNumber($number)),
                $priority,
                $appName,
                $this->_db->realEscapeString($template),
                $app->getDialplanAppName()
            )
        );

        return true;
    }

    /**
     * Check list of params for current application and prepare rule template for save in DB.
     *
     * @param ApplicationModel $app         Application.
     * @param string           $number      Rule number.
     * @param array            $inputParams Rule input params.
     *
     * @return string
     */
    private function _checkParamsAndPrepareTemplateToDbFormat(ApplicationModel $app, string $number, array $inputParams): string {
        $appName         = $app->getName();
        $template        = $app->getTemplateForDb();
        $appConfigParams = $app->getParams(false);
        $activeParams    = $app->getActiveParamsList(false);

        if ($appName === 'Set' && $inputParams['set_l_key'] === DictionaryTypeEnum::SERVICE && !License::getInstance()->haveLicense(License::TYPE_XBEES)) {
            $this->_addError(
                _t(
                    'error.ruleNoXbeesLicence',
                    $appName,
                    $inputParams['set_l_key']
                ),
                __LINE__
            );
            return '';
        }

        if ($appName === 'Set' && preg_match("/\"/", $inputParams['set_l_key'])) {
            $this->_addError(
                _t(
                    'error.ruleQuotesNotAllowed',
                    $appName,
                    $inputParams['set_l_key']
                ),
                __LINE__
            );
            return '';
        }

        foreach ($appConfigParams as $appParamData) {
            $paramName = (string)$appParamData['name'];

            // check param isset in request data
            if (!array_key_exists($paramName, $inputParams) && in_array($paramName, $activeParams)) {
                $this->_addError(
                    _t(
                        'error.ruleInvalidParams',
                        $this->_formatter->convertParamNameConfigToApi($appName, $paramName),
                        $number,
                        $appName
                    ),
                    __LINE__
                );
                return '';
            }

            $paramValue = isset($inputParams[$paramName]) ? (string)$inputParams[$paramName] : '';

            // generate random value if empty
            if ($paramValue === ''
                && (($appName === 'GotoIfGroupCount' && $paramName === 'groupname') || ($appName === 'DialHuntGroup' && $paramName === 'uniqueval'))
            ) {
                $paramValue = rand(10000000, 99999999);
            }

            if (!$this->_validator->validate($app, $paramName, $paramValue)) {
                $this->_addError(
                    _t(
                        'error.ruleInvalidValueParams',
                        $this->_formatter->convertParamNameConfigToApi($appName, $paramName),
                        $number,
                        $appName
                    ),
                    __LINE__
                );
                return '';
            }

            $value    = $this->_formatter->formatInputParamValue($appName, $paramName, $paramValue, $appConfigParams);
            $template = str_replace('_{_' . $paramName . '_}', $value, $template);
        }

        if ($appName === 'AsrIvr') {
            $template = '"i"';
        }

        return $template;
    }

    // phpcs:disable Generic.Metrics.CyclomaticComplexity.MaxExceeded

    /**
     * Prepare dialplan links.
     *
     * @param integer $dialplanId Dialplan ID.
     * @param array   $rule       Rule data.
     *
     * @return void
     */
    private function _prepareLinks(int $dialplanId, array $rule): void {
        $params = $rule[self::PARAM_APP_PARAMS];

        switch ($rule[self::PARAM_APP_NAME]) {
            case 'DialHuntGroup':
                $this->_addLink($dialplanId, DialplanService::LINK_HUNT_GROUP, $params['hunt_group']);
                break;
            case 'DialRemotePbx':
                $this->_addLink($dialplanId, DialplanService::LINK_PBX, $params['pbx']);
                break;
            case 'DialTrunk':
                $this->_addLink($dialplanId, DialplanService::LINK_TRUNK, $params['trunk']);
                break;
            case 'FeaturePaging':
                $this->_addLink($dialplanId, DialplanService::LINK_PAGE_GROUP, $params['dest']);
                break;
            case 'Queue':
                $this->_addLink($dialplanId, DialplanService::LINK_CALL_GROUP, $params['queuename']);
                break;
            case 'GotoIf3SwitchIsOnOrBlinking':
                $this->_addLink($dialplanId, DialplanService::LINK_SWITCH, $params['switchname']);
                $this->_addLink($dialplanId, DialplanService::LINK_DIALPLAN, $params['context_on']);
                $this->_addLink($dialplanId, DialplanService::LINK_DIALPLAN, $params['context_bl']);
                break;
            case 'CorrectOutgoingNumber':
            case 'Goto':
            case 'GotoIfBusyCongested':
            case 'GoToCallCostLimit':
            case 'GotoIfCallType':
            case 'GoToIfFaxDetected':
            case 'GotoIfGroupCount':
            case 'VerifyCaller':
                $this->_addLink($dialplanId, DialplanService::LINK_DIALPLAN, $params['context']);
                break;
            case 'GotoIfSwitchIsOn':
                $this->_addLink($dialplanId, DialplanService::LINK_SWITCH, $params['switchname']);
                $this->_addLink($dialplanId, DialplanService::LINK_DIALPLAN, $params['context']);
                break;
            case 'Timetable':
                $this->_addLink($dialplanId, DialplanService::LINK_TIMETABLE, $params['timetablename']);
                $this->_addLink($dialplanId, DialplanService::LINK_DIALPLAN, $params['context']);
                break;
            case 'LookUpCidNameInPhonebook':
            case 'LookUpDestCidNameInPhonebook':
                $this->_addLink($dialplanId, DialplanService::LINK_PHONEBOOK, $params['phonebook']);
                break;
            case 'IVR':
                $this->_addLink($dialplanId, DialplanService::LINK_IVR, $params['ivrname']);
                break;
            default:
                // default not add link
                break;
        }
    }
    // phpcs:enable Generic.Metrics.CyclomaticComplexity.MaxExceeded

    /**
     * Add dialplan link.
     *
     * @param integer $dialplanId   Dialplan ID.
     * @param integer $typeId       Type ID.
     * @param string  $linkedObject Linked object.
     *
     * @return void
     */
    private function _addLink(int $dialplanId, int $typeId, string $linkedObject): void {
        $this->_links[] = [
            'object_from_id' => $dialplanId,
            'object_to_type' => $typeId,
            'object_to_id' => $linkedObject,
        ];
    }

    /**
     * Add unique links to bath SQL queries.
     *
     * @return void
     */
    private function _addUniqueLinks(): void {
        $links = array_unique($this->_links, SORT_REGULAR);
        foreach ($links as $link) {
            $this->_db->query(
                sprintf(
                    'INSERT INTO %s (object_from_type, object_from_id, object_to_type, object_to_id) VALUES (%d, %d, %d, \'%s\');',
                    \Wildix\Dialplan\DialplanService::DIALPLAN_LINKS_TABLE,
                    \Wildix\Dialplan\DialplanService::LINK_DIALPLAN,
                    $link['object_from_id'],
                    $link['object_to_type'],
                    $this->_db->realEscapeString($link['object_to_id'])
                )
            );
        }
    }

    /**
     * Add error.
     *
     * @param string  $errorText Error text.
     * @param integer $line      Line number.
     *
     * @return void
     */
    private function _addError(string $errorText, int $line): void {
        $this->_errors[] = $errorText;
        $this->_logger->error($line, $errorText);
    }

    /**
     * Return errors.
     *
     * @return array
     */
    public function getErrors(): array {
        return $this->_errors;
    }
}
