<?php
namespace Wildix\Dialplan\TimeTables;

use Wildix\Dialplan\BaseDialplansService;
use Wildix\Error;
use Wildix\Validator;

/**
 * Class TimeTablesService
 */
class TimeTablesService extends BaseDialplansService {

    const STATUS_KEYS = [
        0 => 'NOT_INUSE',
        1 => 'RINGING',
        2 => 'INUSE',
        'default' => 'NOT_INUSE'
    ];

    public function create(array $data): void {
        $this->validate($data);
        $data['id'] = $this->getAvailableId();
        $this->storage()->create($data);
        $this->setState('stime', $data['id'], $data['state']);
    }

    public function update(int $identity, array $data = []): void {
        $this->getItem($identity, true);
        $this->validate($data, $identity);
        $this->storage()->update($identity, $data);
        $this->setState('stime', $identity, $data['state']);
    }

    public function buildItem(TimeTable $item) {
        return [
            'id' => $item->getId(),
            'title' => $item->getName(),
            'state' => $item->getState(),
            'items' => array_map(function ($item) {
                return [
                    'from' => [
                        'day' => (int)$item['day_id_from'],
                        'dayOfWeek' => (int)$item['day_of_week_from']
                    ],
                    'to' => [
                        'day' => (int)$item['day_id_to'],
                        'dayOfWeek' => (int)$item['day_of_week_to']
                    ],
                    'time' => [
                        'from' => $item['start_daytime'],
                        'to' => $item['end_daytime'],
                    ],
                    'month' => (int)$item['month_id'],
                    'year' => (int)$item['year_id']
                ];
            }, $item->getItems())
        ];
    }

    protected function makeValidator($data): Validator {
        $validator = new Validator($data);
        $validator->rule('required', ['title', 'state']);
        $validator->rule('regex', ['title'], '/^[0-9a-zA-Z-._]*$/')->reason(Error::getMessage(Error::IncorrectFormat));
        $validator->rule('array', ['items']);
        $validator->rule('integer', ['state']);
        $validator->rule('min', ['state'], 0);
        $validator->rule('max', ['state'], 2);

        if (isset($data['items']) && is_array($data['items'])) {
            foreach ($data['items'] as $key => $item) {
                $maxDays = ($item['month'] > 0 && $item['month'] <= 12) ?  date('t', mktime(0, 0, 0, $item['month'], 1, (int)$item['year'])) : 31;
                $validator->rule('array', ['items.' . $key . '.from']);
                $validator->rule('required', ['items.' . $key . '.from.day']);
                $validator->rule('between', ['items.' . $key . '.from.day'], [0, $maxDays]);

                $validator->rule('required', ['items.' . $key . '.from.dayOfWeek']);
                $validator->rule('between', ['items.' . $key . '.from.dayOfWeek'], [0, 7]);

                $validator->rule('array', ['items.' . $key . '.to']);
                $validator->rule('required', ['items.' . $key . '.to.day']);
                $validator->rule('between', ['items.' . $key . '.to.day'], [0, $maxDays]);

                $validator->rule('required', ['items.' . $key . '.to.dayOfWeek']);
                $validator->rule('between', ['items.' . $key . '.to.dayOfWeek'], [0, 7]);

                $validator->rule('array', ['items.' . $key . '.time']);
                $validator->rule('required', ['items.' . $key . '.time']);
                $validator->rule('time', ['items.' . $key . '.time.from']);

                $validator->rule('required', ['items.' . $key . '.time']);
                $validator->rule('time', ['items.' . $key . '.time.to']);

                $validator->rule('required', ['items.' . $key . '.month']);
                $validator->rule('between', ['items.' . $key . '.month'], [0, 12]);
                $validator->rule('required', ['items.' . $key . '.year']);
                $validator->rule('integer', ['items.' . $key . '.year']);
            }
        }

        return $validator;
    }
}
