<?php

namespace Wildix\Dialplan\TimeTables;

use Wildix\Databases\WMSDB;

/**
 * Class TimeTablesStorage
 */
class TimeTablesStorage {

    const DIALPLAN_LINK      = 1;
    private const LINK_GROUP = 9;

    private const GET_ITEMS = 'SELECT * FROM ' . WMSDB::TABLE_TIME_TABLES;

    private $_db;

    private function _db(): WMSDB {
        if (!$this->_db) {
            $this->_db = WMSDB::getInstance();
        }
        return $this->_db;
    }

    public function getItems(): array {
        return $this->_db()->getRows(self::GET_ITEMS, null, TimeTable::class);
    }

    public function getItem(int $identity): ?TimeTable {
        $item = $this->_db()->first(self::GET_ITEMS . ' WHERE id=' . $identity);
        if ($item) {
            $item = new TimeTable($item);
        }
        return $item;
    }

    public function create(array $data): void {
        $this->_db()->beginTransaction();
        $this->_db()->insert(WMSDB::TABLE_TIME_TABLES, $this->_prepareToSave($data));
        $items = isset($data['items']) && !empty($data['items']) ? $data['items'] : [];
        $this->_insertSettings($data['id'], $items);
        $this->_db()->endTransaction();
    }

    public function update(int $identity, array $data): void {
        $this->_db()->beginTransaction();
        $this->_db()->exec(sprintf('UPDATE %s SET timetable="%s", state=%d WHERE id=%d', WMSDB::TABLE_TIME_TABLES, $data['title'], $data['state'], $identity));
        $this->_removeSettings($identity);
        $this->_insertSettings($identity, $data['items']);
        $this->_db()->endTransaction();
    }

    public function remove(int $identity): void {
        $this->_db()->beginTransaction();
        $this->_db()->exec(sprintf('DELETE FROM %s WHERE id=%d', WMSDB::TABLE_TIME_TABLES, $identity));
        $this->_removeSettings($identity);
        $this->_db()->endTransaction();
    }

    public function loadRelations(array $tables = []): array {
        $items = [];
        foreach ($tables as $table) {
            $items[$table->getId()] = $table;
        }

        $settings = $this->_getSettings(array_keys($items));

        foreach ($settings as $setting) {
            if (isset($items[$setting['parent_id']])) {
                $items[$setting['parent_id']]->addItem($setting);
            }
        }

        return array_values($items);
    }

    public function getLinked($identity): array {
        // phpcs:disable Generic.Files.LineLength.TooLong
        $res = $this->_db()->getRows(
            sprintf(
                'SELECT rn.rulename FROM %s as rl LEFT JOIN %s as rn ON rn.id=rl.object_from_id WHERE rl.object_from_type=%s AND rl.object_to_type=%s AND rl.object_to_id=%d',
                WMSDB::TABLE_RULES_LINKS,
                WMSDB::DIALPLAN_RULE_NAME,
                self::DIALPLAN_LINK,
                self::LINK_GROUP,
                $identity
            )
        );
        // phpcs:enable Generic.Files.LineLength.TooLong
        return $res ? array_column($res, 'rulename') : [];
    }

    private function _getSettings(array $ids = []): array {
        $items = [];
        if ($ids) {
            $items = $this->_db()->getRows(sprintf('SELECT * FROM %s WHERE parent_id IN (%s)', WMSDB::TABLE_TIME_SETTING, implode(', ', $ids)));
        }
        return $items;
    }

    private function _prepareToSave(array $data): array {
        return [
            'id' => $data['id'],
            'timetable' => $data['title'],
            'state' => $data['state'],
            'desc' => ''
        ];
    }

    private function _prepareSetting(int $identity, array $data = []) {

        $res = [
            'parent_id' => $identity,
            'day_of_week_from' => 0,
            'day_of_week_to' => 0,
            'day_id_from' => 0,
            'day_id_to' => 0,
            'month_id' => 0,
            'start_daytime' => '00:00:00',
            'end_daytime' => '23:59:59',
            'year_id' => 0,
            'enabled_id' => 1,
        ];

        if (isset($data['from']) && isset($data['from']['dayOfWeek'])) {
            $res['day_of_week_from'] = $data['from']['dayOfWeek'];
        }

        if (isset($data['from']) && isset($data['from']['day'])) {
            $res['day_id_from'] = $data['from']['day'];
        }

        if (isset($data['to']) && isset($data['to']['dayOfWeek'])) {
            $res['day_of_week_to'] = $data['to']['dayOfWeek'];
        }

        if (isset($data['to']) && isset($data['to']['day'])) {
            $res['day_id_to'] = $data['to']['day'];
        }

        if (isset($data['time']) && isset($data['time']['from'])) {
            $res['start_daytime'] = $data['time']['from'];
        }

        if (isset($data['time']) && isset($data['time']['to'])) {
            $res['end_daytime'] = $data['time']['to'];
        }

        if (isset($data['month'])) {
            $res['month_id'] = $data['month'];
        }

        if (isset($data['year'])) {
            $res['year_id'] = $data['year'];
        }

        return $res;
    }

    private function _insertSettings(int $id, array $items = []): void {
        if (empty($items)) {
            $row = $this->_prepareSetting($id, []);
            $this->_db()->insert(WMSDB::TABLE_TIME_SETTING, $row);
        }

        foreach ($items as $item) {
            $row = $this->_prepareSetting($id, $item);
            $this->_db()->insert(WMSDB::TABLE_TIME_SETTING, $row);
        }
    }

    private function _removeSettings(int $parentId) {
        $this->_db()->exec(sprintf('DELETE FROM %s WHERE parent_id=%d', WMSDB::TABLE_TIME_SETTING, $parentId));
    }
}
