<?php
namespace Wildix;

use Throwable;
use Wildix\Core\Http;

/**
 * Ejabberd Class.
 */
class Ejabberd extends \Singleton {
    const URL_API           = 'http://127.0.0.1:5280/api/';
    const CONF_PATH         = '/rw2/etc/ejabberd/ejabberd.yml';
    const CONF_PATH_MOD_MAM = '/rw2/etc/ejabberd/ejabberd_mod_mam.yml';
    const CONF_PATH_DEFINES = '/rw2/etc/ejabberd/ejabberd_defines.yml';

    private const CONF_PATH_MOD_ASTMAN = '/rw2/etc/ejabberd/ejabberd_mod_wildix_astman.yml';

    private $_log;
    private $_settings;

    public function setLog(\Logger\Log $logger): void {
        $this->_log = $logger;
    }

    private function getLog(): \Logger\Log {
        if (!isset($this->_log)) {
            $this->_log = \Logger\LoggerFactory::getInstance();
        }

        return $this->_log;
    }

    /**
     * Returns all information of user presence status
     * Possible values are:
     * <code>
     * $status = array(
     *     'online' => true // Or false
     *     'show' => false // Or string
     *     'device_show' => false // Or string
     *     'status' => false // Or string
     *     'location' => array
     * )
     * </code>
     *
     * @param string $extension Extension.
     *
     * @return array
     */
    public function getUserPresence(string $extension): array {
        $data           = $this->getData('get_user_status', ['user' => $extension]);
        $data['online'] = $data['online'] === 'true';

        return $data;
    }

    public function getPublicUserPresence(string $extension) {
        $data              = $this->getData('get_public_user_presence', ['user' => $extension]);
        $data['online']    = $data['online'] === 'true';
        $data['extension'] = $extension;

        return $data;
    }

    public function getPublicUsersPresence(array $extensions): array {
        $data = $this->getData('get_public_users_presence', ['users' => implode(',', $extensions)]);
        $data = array_map(
            function ($item) {
                $item['extension'] = $item['user'];
                $item['online']    = $item['online'] === 'true';
                unset($item['user']);

                return $item;
            },
            $data
        );

        return $data;
    }

    /**
     * Returns short presence information of users
     * Possible values are:
     * <code>
     * $status = array(
     *     'user' => 111 // extension
     *     'status' => 'dnd' // "online", "away", "dnd" ...
     * )
     * </code>
     *
     * @param array $extensions Extension.
     *
     * @return array
     */
    public function getUsersPresence(array $extensions): array {
        $res = [];

        $items = $this->getPublicUsersPresence($extensions);
        foreach ($items as $item) {
            $status = $item['online'] ? $item['show'] : 'offline';
            $res[]  = [
                'user'   => $item['extension'],
                'status' => $status
            ];
        }

        return $res;
    }

    public function getPublicUserInfo(string $extension): array {
        $item = $this->getData('get_public_user_info', ['user' => $extension]);

        return [
            "extension" => $item['user'],
            "uid" => $item['uid'],
            "name" => $item['name'],
            "pbx" => $item['pbx'],
            "avatar" => $item['avatar'],
            "presence" => [
                "online" => $item['online'] === 'true',
                "show" => $item['show'],
                "device_show" => $item['device_show'],
                "status" => $item['status'],
                "location" => $item['location'],
                "until" => $item['until']
              ]
        ];
    }

    public function getPublicUsersInfo(array $extensions) {
        $res   = [];
        $items = $this->getData('get_public_users_info', ['users' => implode(',', $extensions)]);

        foreach ($items as $item) {
            $res[] = [
                "extension" => $item['user'],
                "uid" => $item['uid'],
                "name" => $item['name'],
                "pbx" => $item['pbx'],
                "avatar" => $item['avatar'],
                "presence" => [
                    "online" => $item['online'] === 'true',
                    "show" => $item['show'],
                    "device_show" => $item['device_show'],
                    "status" => $item['status'],
                    "location" => $item['location'],
                    "until" => $item['until']
                ]
            ];
        }

        return $res;
    }

    public function setUserPresence(string $extensions, string $status, string $message, $until): void {
        $zone = new \DateTimeZone('UTC');
        $date = \DateTime::createFromFormat('d/m/Y H:i', $until, $zone);

        if ($until !== '' && $date === false) {
            throw new InvalidDataException('Until date has bad format', [], 400);
        }

        if ($until !== '' && $date->getTimestamp() < time()) {
            throw new InvalidDataException('Until date must be greater than the current', [], 400);
        }

        if (!in_array($status, ['available', 'away', 'dnd', 'mur'])) {
            throw new InvalidDataException('Status can be: available, away, dnd, mur', [], 400);
        }

        $params = [
            'user' => $extensions,
            'show' => $status,
            'status' => $message,
            'until' => $date ? $date->format('Y-m-d\TH:i:00\Z') : '',
        ];

        $this->setData('set_user_status', $params);
    }

    public function setCallcenterMode(string $extension, $mode) {
        $value  = $mode == 1 ? 'true' : 'false';
        $params = [
            'user' => $extension,
            'name' => 'callcenter_mode',
            'value' => $value
        ];

        $this->setData('set_user_option', $params);
    }

    public function setUnreadMessagesViaEmail(string $extension, int $mode): void {
        $value  = $mode === 1 ? 'true' : 'false';
        $params = [
            'user' => $extension,
            'name' => 'unread_messages_via_email',
            'value' => $value
        ];

        $this->setData('set_user_option', $params);
    }

    public function setUserPresenceLocation(string $extension, string $lat, string $lng, string $address): void {
        $params = [
            'user' => $extension,
            'lat' => $lat,
            'lon' => $lng,
            'address' => $address
        ];
        $this->setData('set_user_location', $params);
    }

    /**
     * Returns a show presence of current user.
     *
     * @param string $extension User.
     *
     * @return boolean True when user is online and false when not.
     */
    public function getUserOnlineStatus(string $extension): bool {
        $data = $this->getUserPresence($extension);
        return $data['online'];
    }

    /**
     * Returns a device_show presence of current user.
     * Possible values are:
     * <code>
     * $location = array(
     *     'address' => false // Or string when address is available
     *     'lat' => false // Or string when lat is available
     *     'lng' => false // Or string when lng is available
     * )
     * </code>
     *
     * @param string $extension User.
     *
     * @return array
     */
    public function getUserLocationStatus(string $extension) {
        $data = $this->getUserPresence($extension);

        return $data['location'];
    }

    public function sendCustomMessage(string $from, string $to, string $xmlns, string $type, string $message): void {
        $params = [
            'from' => $from,
            'to' => $to,
            'xmlns' => $xmlns,
            'type' => $type,
            'message' => $message,
            'id' => 'CL_' . time() . substr(microtime(), 2, 3)
        ];
        $this->setData('send_custom_message', $params);
    }

    public function sendMessage(string $from, string $to, string $type, string $message): void {
        $params = [
            'from' => $from,
            'to' => $to,
            'subject' => '',
            'type' => $type,
            'body' => $message
        ];
        $this->setData('send_message', $params);
    }

    public function moveUsers(array $userExtensions, string $fromPbxDomain, string $toPbxDomain): void {
        $params = [
            'users' => $userExtensions,
            'from'  => $fromPbxDomain,
            'to'    => $toPbxDomain
        ];

        $this->setData('users_moved', $params);
    }

    private function changeHost(string $from, string $to): void {
        $params = [
                'from' => $from,
                'to' => $to
            ];

        $this->setData('host_changed', $params);
    }

    public function getSettings() {
        if ($this->_settings === null) {
            $this->_settings = array(
                    'host' => ''
            );

            $ejabberdConf = file_get_contents(self::CONF_PATH);
            if ($ejabberdConf) {
                // Parse stored hostname
                // Hosts: "ejabberd15-master.wildixin.com"
                if (preg_match('/^hosts: \["([^"]+)"\].*?$/m', $ejabberdConf, $matches)) {
                    if (!empty($matches[1])) {
                        $this->_settings['host'] = $matches[1];
                    }
                }
            }

            $this->_settings['sql'] = $this->getSqlSettings();
        }

        return $this->_settings;
    }

    private function getSqlSettings() {
        $ejabberdModMamConf = file_get_contents(self::CONF_PATH_MOD_MAM);
        $sqlSettings        = array(
            'enabled' => false,
            'server' => '',
            'type' => '',
            'database' => ''
        );

        if ($ejabberdModMamConf) {
            // parse stored sql_type
            // sql_type: "mysql | pgsql | odbc | mssql | sqlite"
            if (preg_match('/^(#{0,1}) *sql_type: (.*?)$/m', $ejabberdModMamConf, $matches)) {
                if (isset($matches[2]) && !empty($matches[2])) {
                    $sqlSettings['type'] = $matches[2];
                }
            }

            if (preg_match('/^(#{0,1}) *sql_database: "(.*?)"$/m', $ejabberdModMamConf, $matches)) {
                if (isset($matches[2]) && !empty($matches[2])) {
                    $sqlSettings['database'] = stripslashes($matches[2]);
                }
            }

            // parse stored sql_server
            // sql_server: "DSN=database;UID=ejabberd;PWD=password" // for MSSQL
            // sql_server: "DSN=chat_history" // for another
            if (preg_match('/^(#{0,1}) *sql_server: "([^"]+)".*?$/m', $ejabberdModMamConf, $matches)) {
                if (isset($matches[1]) && $matches[1] != '#') {
                    $sqlSettings['enabled'] = true;
                }
                if (isset($matches[2]) && !empty($matches[2])) {
                    $sqlSettings['server'] = stripslashes($matches[2]);
                }
            }

            if (preg_match('/^(#{0,1}) *sql_port: ([\d]+).*?$/m', $ejabberdModMamConf, $matches)) {
                if (isset($matches[2]) && !empty($matches[2])) {
                    $sqlSettings['port'] = $matches[2];
                }
            }
        }

        return $sqlSettings;
    }

    public function setSettings($newSettings, $restart = true) {
        if (is_writable(self::CONF_PATH)) {
            $oldSettings  = $this->getSettings();
            $newSettings  = array_replace_recursive($oldSettings, $newSettings);
            $diffSettings = array_diff_assoc_recursive($newSettings, $oldSettings);

            if (!empty($diffSettings)) {
                $this->_settings = $newSettings;

                if (isset($diffSettings['sql']['enabled'])
                    || !empty($diffSettings['sql']['type'])
                    || !empty($diffSettings['sql']['server'])
                    || !empty($diffSettings['sql']['db'])
                ) {
                    $comment = '';
                    if ($newSettings['sql']['enabled'] == false) {
                        $comment = '# ';
                    }

                    if ($newSettings['sql']['type'] == 'sqlite' && !empty($newSettings['sql']['server'])) {
                        $newSettings['sql']['server'] = '';
                    }

                    $sql = $newSettings['sql'];

                    $ejabberdModMamConf  = $comment . 'sql_type: ' . $sql['type'] . "\n";
                    $ejabberdModMamConf .= isset($sql['server']) ? $comment . 'sql_server: "' . addslashes($sql['server']) . "\"\n" : '';
                    $ejabberdModMamConf .= (!empty($sql['port'])) ? $comment . 'sql_port: ' . $sql['port'] . "\n" : '';
                    $ejabberdModMamConf .= isset($sql['db']) ? $comment . 'sql_database: "' . addslashes($sql['db']) . "\"\n" : '';
                    $ejabberdModMamConf .= isset($sql['username']) ? $comment . 'sql_username: "' . addslashes($sql['username']) . "\"\n" : '';
                    $ejabberdModMamConf .= isset($sql['password']) ? $comment . 'sql_password: "' . addslashes($sql['password']) . "\"\n" : '';
                    $ejabberdModMamConf .= "\n";
                    $ejabberdModMamConf .= $comment . "modules:\n";
                    $ejabberdModMamConf .= $comment . "  mod_mam:\n";
                    $ejabberdModMamConf .= $comment . "    db_type: 'STORAGE_DB'\n";
                    $ejabberdModMamConf .= $comment . '    default: always';

                    file_put_contents(self::CONF_PATH_MOD_MAM, $ejabberdModMamConf);
                }

                $ejabberdConf = file_get_contents(self::CONF_PATH);
                if ($ejabberdConf) {
                    if (!empty($diffSettings['host'])) {
                        $ejabberdConf = preg_replace('/^hosts: \["[^"]+"\].*?$/m', 'hosts: ["' . $newSettings['host'] . '"]', $ejabberdConf);

                        if (file_put_contents(self::CONF_PATH, $ejabberdConf) !== false) {
                            $this->changeHost($oldSettings['host'], $newSettings['host']);
                        }
                    }
                }
            }

            // defines config added later and must still be re-saved even if ModMamConf has not changed
            $oldDefines = $this->getDefines();
            $newDefines = array();
            if ($newSettings['sql']['enabled'] == true) {
                $newDefines['STORAGE_DB'] = 'sql';
            } else {
                $newDefines['STORAGE_DB'] = 'mnesia';
            }

            $newDefines = array_replace_recursive($oldDefines, $newDefines);
            $this->setDefines($newDefines);

            if ($restart) {
                $this->restart();
            }
        }
        return false;
    }

    /**
     * Set module astman max_lua_workers param.
     *
     * @param integer $count Count of lua workers.
     *
     * @return boolean
     */
    private function setAstmanMaxLuaWorkers(int $count): bool {
        if (!is_writable(self::CONF_PATH_MOD_ASTMAN)) {
            $this->getLog()->error(__LINE__, sprintf('Config file "%s" is not writable ', self::CONF_PATH_MOD_ASTMAN));
            return false;
        }

        $ejabberdModAstmanConf = file(self::CONF_PATH_MOD_ASTMAN, FILE_IGNORE_NEW_LINES);

        if (\count($ejabberdModAstmanConf)) {
            foreach ($ejabberdModAstmanConf as $row => $setting) {
                if ((\strpos($setting, 'max_lua_workers') !== false)) {
                    $ejabberdModAstmanConf[$row] = \preg_replace('/\d+/', $count, $setting);
                }
            }
        }

        return file_put_contents(self::CONF_PATH_MOD_ASTMAN, \implode("\n", $ejabberdModAstmanConf)) !== false;
    }

    /**
     * Configure max count of lua workers.
     *
     * @param integer $totalMemory         PBX RAM memory in bytes.
     * @param boolean $needIncreaseWorkers Set "true" if need increase workers or "false" other way.
     *
     * @return void
     */
    public function configureLuaWorkers(int $totalMemory, bool $needIncreaseWorkers): void {
        // RAM memory in MB
        $totalMemory = \ceil($totalMemory / 1024 / 1024);

        // by default max lua workers set to 0 (disabled).
        $maxLuaWorkers = 0;

        if ($needIncreaseWorkers) {
            $this->getLog()->info(__LINE__, 'Has enabled integrations');
            if ($totalMemory <= 500) {
                $maxLuaWorkers = 3;
            } else if ($totalMemory > 500 && $totalMemory <= 1900) {
                $maxLuaWorkers = 5;
            } else if ($totalMemory > 1900) {
                $maxLuaWorkers = 10;
            }
        }

        $this->getLog()->info(__LINE__, \sprintf('Set max lua workers: %s, PBX RAM %s MB', $maxLuaWorkers, $totalMemory));
        $this->setAstmanMaxLuaWorkers($maxLuaWorkers);

        $this->restart();
    }

    private $_defines = null;
    private function getDefines() {
        if ($this->_defines === null) {
            $this->_defines = array(
                    'STORAGE_DB' => 'mnesia'
            );

            $definesConf = file_get_contents(self::CONF_PATH_DEFINES);
            if ($definesConf) {
                if (preg_match('/^ *\'STORAGE_DB\': (.*?)$/m', $definesConf, $matches)) {
                    if (isset($matches[1]) && !empty($matches[1])) {
                        $this->_defines['STORAGE_DB'] = $matches[1];
                    }
                }
            }
        }
        return $this->_defines;
    }

    private function setDefines($newDefines) {
        $ejabberdDefinesConf  = "define_macro:\n";
        $ejabberdDefinesConf .= "  'STORAGE_DB': " . $newDefines['STORAGE_DB'] . "\n";

        $oldDefines = file_get_contents(self::CONF_PATH_DEFINES);
        if ($oldDefines != $ejabberdDefinesConf) {
            file_put_contents(self::CONF_PATH_DEFINES, $ejabberdDefinesConf);
        }
    }

    public function restart() {
        exec('sudo /etc/init.d/ejabberd restart > /dev/null &');
        return true;
    }

    public function cleanEjabberdSlave(): void {
        exec('(sudo /etc/init.d/ejabberd stop; sudo /bin/rm -f /rw/var/lib/ejabberd/*; sudo /etc/init.d/ejabberd start) > /dev/null &');
    }

    public function sendPostIt($from, $to, $message) {
        $params = [
            'from' => $from,
            'to' => $to,
            'message' => $message
        ];
        $this->setData('send_postit', $params);
    }

    private function getData(string $action, array $query = []): array {
        try {
            $res  = Http::client()->get(self::URL_API . $action, ['query' => $query]);
            $data = json_decode($res->getBody()->getContents(), JSON_UNESCAPED_UNICODE);
        } catch (Throwable $e) {
            throw new Exception('Request fault: ' . $e->getCode());
        }

        return $data;
    }

    private function setData(string $action, array $params = []): void {
        Http::client()->post(self::URL_API . $action, ['body' => json_encode($params)]);
    }
}
