<?php
namespace Wildix\Faxes;

/**
 * Class FaxSmsSettings
 */
class FaxSmsSettings {
    const COMPANIES_DIRECTORY = '/etc/companies.d/';

    private $_companyGlobalConfig = null;

    /**
     * FaxSmsSettings constructor.
     *
     * @param \Wildix\CompanyGlobalConfig $companyGlobalConfig Company global config.
     */
    public function __construct(\Wildix\CompanyGlobalConfig $companyGlobalConfig) {
        $this->_companyGlobalConfig = $companyGlobalConfig;
    }

    private function getCompanyGlobalConfig() {
        return $this->_companyGlobalConfig;
    }

    private function validateSettings($params) {
        $form = new \Wildix\Validator(
            $params,
            array('mailbox', 'smsPbx', 'faxPdfReport', 'curlSmsServerEnable')
        );

        $form->rule(
            'email',
            array('mailbox')
        );

        $form->rule(
            'boolean',
            array('faxPdfReport', 'curlSmsServerEnable')
        );

        $form->rule(
            'pbxDn',
            array('smsPbx')
        );

        if (!$form->validate()) {
            throw new \Wildix\InvalidDataException('Validation error', $form->errors());
        }
    }

    private $_options = null;

    public function saveSettings(array $settings) {
        $this->validateSettings($settings);

        $configVariables = array(
            'GSMSTEXT' => base64_encode($settings['smsText']),
            'GBKMAILBOX' => $settings['mailbox'],
        );
        // this check is set here because the field can be disabled on frontend and therefore it does not come in request
        if (isset($settings['curlSmsServer'])) {
            $configVariables['CURLSMSSERVER'] = base64_encode($settings['curlSmsServer']);
        }
        if ($settings['smsPbx']) {
            $configVariables['GSMSPBX'] = $settings['smsPbx'];
        } else {
            $configVariables['GSMSPBX'] = \Wildix\PBXes\PBXes::getInstance()->getPBXBySerial(\Models\PBX::getInstance()->getSerial())->getDn();
        }
        $configVariables['GFAXPDFREPORT'] = $settings['faxPdfReport'];

        $httpScriptEnable = 0;
        if ($settings['curlSmsServerEnable'] == 'on') {
            $httpScriptEnable = 1;
        }
        \Wildix\System\Sms::getInstance()->update($httpScriptEnable);

        $this->getCompanyGlobalConfig()->setVariables($configVariables);

        \Wildix\CallweaverManager::getInstance()->reloadModule('pbx_config');

        return $this->getSettings();
    }

    /**
     * Retrieve Fax/SMS options
     *
     * @param boolean $force Force reload from file (after saving).
     *
     * @return array
     */
    public function getSettings($force = false) {
        if ($this->_options === null || $force === true) {
            $isSlave      = \Models\PBX::getInstance()->isSlave();
            $isWmsNetwork = \Models\PBX::getInstance()->isEnabledWmsNetwork();
            $pbxForSel    = array();
            $pbxForSelObj = \Wildix\PBXes\PBXes::getInstance()->getPBXes();
            foreach ($pbxForSelObj as $PBX) {
                $pbxForSel[] = $PBX->getAsArray();
            }
            $currentPbxDn = \Wildix\PBXes\PBXes::getInstance()->getPBXBySerial(\Models\PBX::getInstance()->getSerial())->getDn();

            $mailbox       = $this->getCompanyGlobalConfig()->getVariable('GBKMAILBOX');
            $smsText       = base64_decode($this->getCompanyGlobalConfig()->getVariable('GSMSTEXT'));
            $curlSmsServer = base64_decode($this->getCompanyGlobalConfig()->getVariable('CURLSMSSERVER'));
            $smsPbx        = $this->getCompanyGlobalConfig()->getVariable('GSMSPBX');
            $smsPbx        = $smsPbx ? $smsPbx : $currentPbxDn;

            if ($this->getCompanyGlobalConfig()->getVariable('GFAXPDFREPORT') == 'on') {
                $faxPdfReport = true;
            } else {
                $faxPdfReport = false;
            }

            $curlSmsServerEnable = \Wildix\System\Sms::getInstance()->get('HTTP_SCRIPT_ENABLE') ? true : false;

            $settings = array(
                'mailbox'             => $mailbox,
                'smsText'             => $smsText,
                'smsPbx'              => $smsPbx,
                'faxPdfReport'        => $faxPdfReport,
                'curlSmsServerEnable' => $curlSmsServerEnable,
                'curlSmsServer'       => $curlSmsServer,
                'isSlave'             => $isSlave,
                'isWmsNetwork'        => $isWmsNetwork,
                'pbxForSel'           => $pbxForSel,
            );

            $this->_options = $settings;
        }

        return $this->_options;
    }
}
