<?php
namespace Wildix\Faxes;

use Models\PBX;
use Psr\Http\Message\UploadedFileInterface;
use Slim\Exception\HttpBadRequestException;
use Slim\Psr7\Request;
use Wildix\Exception;
use Wildix\InvalidDataException;
use Wildix\Validator;

/**
 * Class Faxes
 */
class Faxes {
    const FAXCOVER_PATH        = '/tmp/session/';
    const MAX_UPLOAD_FILE_SIZE = 5242880;
    const FAX_QUEUE            = '/var/www/spoolview/faxqueue/0/';
    const FAX_FILES_PATH       = '/mnt/faxes';

    /**
     * Logger class.
     *
     * @var \Syslog Log.
     */
    private $_log;

    /**
     * Validator class.
     *
     * @var Validator Validator.
     */
    private $_validator = null;

    /**
     * Faxes constructor.
     *
     * @param \Syslog|null $log Logger class.
     */
    public function __construct(\Syslog $log = null) {
        $this->_log       = $log ? $log : new \Syslog('Faxes');
        $this->_validator = new Validator();
    }

    /**
     * Method sends fax
     *
     * @param array|mixed|null             $faxNumbers Extensions where send fax.
     * @param string|null                  $postpone   Postpone date if not null.
     * @param array                        $faxCover   Fax cover parameters.
     * @param array|mixed|null             $fileName   Fax file array.
     * @param \Wildix\Colleagues\Colleague $user       This user.
     * @param Request                      $request    Request.
     *
     * @return boolean|string
     * @throws HttpBadRequestException Parameter comment.
     */
    public function sendFax($faxNumbers, $postpone, array $faxCover, $fileName, \Wildix\Colleagues\Colleague $user, Request $request) {
        $fileName       = self::FAXCOVER_PATH . $this->getSafeFilename($fileName);
        $recipients     = array();
        $recipientsName = array();

        if (!is_array($faxNumbers) || empty($faxNumbers)) {
            throw new HttpBadRequestException($request, 'Extension number is empty');
        }

        foreach ($faxNumbers as $number) {
            $this->validateFaxNumber($number['number']);
            $recipients[] = $number['number'];
            $recipientsName[$number['number']] = $number['name'];
        }

        $addFaxData = '';

        if ($faxCover['faxCover'] == true) {
            $filter['fax']['or'] = $recipients;
            $contacts            = \Wildix\Phonebooks\Contacts::getInstance()->getContacts($filter, null, null, null, null, null, null, $user);
            $usersData           = array();

            if ($contacts) {
                foreach ($contacts as $contact) {
                    if (empty($recipientsName[$contact->getFax()])) {
                        $recipientsName[$contact->getFax()] = $contact->getName();
                    }
                    if ($recipientsName[$contact->getFax()] == $contact->getName()) {
                        $usersData[$contact->getFax()] = array(
                            'username' => $contact->getName(),
                            'phone' => $contact->getPhone(),
                            'mobile' => $contact->getMobile(),
                            'fax' => $contact->getFax(),
                            'organization' => $contact->getOrganisation()
                        );
                    }
                }
            }

            $path       = $this->getFaxcoverPath($user);
            $file       = $path . 'faxcover_data_' . $user->getID() . '_' . rand(1000, 9999999) . '.txt';
            $addFaxData = "FAXCOVER_ENABLE='true' FAXCOVER_FILE='" . $file . "'";

            $fileJsonFaxCoverData['FAXCOVER_FROM']             = ($faxCover['from']);
            $fileJsonFaxCoverData['FAXCOVER_FROM_COMPANY']     = ($faxCover['fromCompany']);
            $fileJsonFaxCoverData['FAXCOVER_FROM_PHONE']       = ($faxCover['fromPhone']);
            $fileJsonFaxCoverData['FAXCOVER_FROM_FAX']         = ($faxCover['fromFax']);
            $fileJsonFaxCoverData['FAXCOVER_SUBJECT']          = ($faxCover['subject']);
            $fileJsonFaxCoverData['FAXCOVER_NOTES']            = ($faxCover['notes']);
            $fileJsonFaxCoverData['FAXCOVER_CHECKBOX_URGENT']  = ($faxCover['checkboxUrgent']);
            $fileJsonFaxCoverData['FAXCOVER_CHECKBOX_REVIEW']  = ($faxCover['checkboxReview']);
            $fileJsonFaxCoverData['FAXCOVER_CHECKBOX_COMMENT'] = ($faxCover['checkboxComment']);
            $fileJsonFaxCoverData['FAXCOVER_CHECKBOX_REPLY']   = ($faxCover['checkboxReply']);
            $fileJsonFaxCoverData['FAXCOVER_CHECKBOX_FORWARD'] = ($faxCover['checkboxForward']);
            $fileJsonFaxCoverData['FAXCOVER_IMAGE']            = $path . 'faxcover_user_' . $user->getID() . '.jpg';
            $fileJsonFaxCoverData['FAXCOVER_TO']               = ($faxCover['to']);
            $fileJsonFaxCoverData['FAXCOVER_TO_PHONE']         = ($faxCover['toPhone']);
            $fileJsonFaxCoverData['FAXCOVER_TO_COMPANY']       = ($faxCover['toCompany']);
            $fileJsonFaxCoverData['FAXCOVER_TO_DATA']          = $usersData;

            file_put_contents($file, json_encode($fileJsonFaxCoverData));
        }

        $addSchedule = '';
        if ($postpone) {
            $scheduleTime = date('Y-m-d H:i', strtotime(str_replace('/', '-', $postpone)));
            $addSchedule  = 'SCHEDULE="' . $scheduleTime . '"';
        }

        $context         = $user->getFaxDialplan();
        $lang            = $user->getLanguage();
        $extensionNumber = $user->getExtension();
        $mail            = trim($user->getEmail());

        if ($mail != '') {
            $mail = ' -m ' . $mail;
        }

        $numbers  = implode(',', $recipients);
        $command  = $addSchedule . ' CONTEXT="' . $context . '" LANGUAGE="' . $lang . '" FROMNUMBER="' . $extensionNumber;
        $command .= '" FROMUSER="' . $extensionNumber . '" ' . $addFaxData . ' /usr/sbin/wms_makefax -n' . $numbers . ' -a"' . $fileName . '" -c ' . $mail;

        $res = exec(escapeshellcmd($command), $output, $backcode);

        if (empty($res)) {
            $res = true;
            exec('sudo /etc/init.d/transmitfax restart', $outputTransmit);
        }

        return $res;
    }

    /**
     * Methods return fax cover path.
     *
     * @param \Wildix\Colleagues\Colleague $user This user.
     *
     * @return string
     */
    public function getFaxcoverPath(\Wildix\Colleagues\Colleague $user) {
        $faxCoverPath  = self::FAXCOVER_PATH;
        $faxCoverShare = PBX::FAXCOVER_SHARE_PATH;

        if (PBX::getInstance()->isCustomShareMounted($faxCoverShare)) {
            // construct faxcover path
            $faxCoverPath = $faxCoverShare . '/';

            // add exten to path
            $extensionNumber = $user->getExtension();

            if (intval($extensionNumber) != 0) {
                $faxCoverPath .= $extensionNumber . '/';
            }

            // create path if not exists
            if (!file_exists($faxCoverPath)) {
                mkdir($faxCoverPath);
            }
        }

        return $faxCoverPath;
    }

    public function isFaxcoverShareMounted() {
        if (!PBX::getInstance()->isCustomShareMounted(PBX::FAXCOVER_SHARE_PATH)) {
            return false;
        }
        return true;
    }

    /**
     * Method return default fax cover path.
     *
     * @return string
     */
    public function faxcoverPath() {
        return self::FAXCOVER_PATH;
    }

    /**
     * Method return max upload size.
     *
     * @return integer
     */
    public function getMaxUploadSize() {
        return self::MAX_UPLOAD_FILE_SIZE;
    }

    /**
     * Method uploads file and return path to file
     *
     * @param UploadedFileInterface $file File array.
     *
     * @throws \Exception Error.
     *
     * @return array
     */
    public function uploadFax(UploadedFileInterface $file): array {
        if ($file->getSize() > $this->getMaxUploadSize()) {
            throw new Exception('limit');
        }

        if (!preg_match('/(tif|tiff|pdf)/i', $file->getClientMediaType())) {
            throw new Exception('bad-file');
        } else if (!$this->isFaxcoverShareMounted() && str_contains($file->getClientMediaType(), 'pdf')) {
            throw new Exception('share_not_mounted_allowed_only_tiff');
        }

        $userFilename = $this->getSafeFilename($file->getClientFilename());
        $extension    = pathinfo($userFilename, PATHINFO_EXTENSION);
        $filename     = md5($userFilename) . ($extension ? '.' . $extension : '');
        $file->moveTo($this->faxcoverPath() . $filename);

        return [
            'userFileName' => $userFilename,
            'filename' => $filename
        ];
    }

    /**
     * Get list of the fax queue
     *
     * @param string $dir Name of fax queue directory.
     *
     * @return array
     */
    public function getQueueFiles($dir = self::FAX_QUEUE): array {
        $result  = [];
        $dirList = scandir($dir);

        foreach ($dirList as $file) {
            $result[] = array(
                'filename' => $file,
                'date' => filectime($dir . $file) * 1000
            );
        }
        return $result;
    }

    /**
     * Delete fax file & report
     *
     * @param array $faxInfo Fax info.
     *
     * @return void
     */
    public function deleteFax(array $faxInfo) {
        try {
            $this->_deleteFileFromStorage($faxInfo['fax']);
            $this->_log->info(__LINE__, sprintf('Removed fax file %s', $faxInfo['fax']));

            if ($faxInfo['report'] !== null) {
                $this->_deleteFileFromStorage($faxInfo['report']);
                $this->_log->info(__LINE__, sprintf('Removed fax report file %s', $faxInfo['report']));
            }
        } catch (\Exception $e) {
            $this->_log->warning(__LINE__, $e->getMessage());
        }
    }

    /**
     * Delete file from storage.
     *
     * @param string $file File name.
     *
     * @return void
     * @throws \Exception Exception.
     */
    private function _deleteFileFromStorage(string $file) {
        if ($file === '') {
            return;
        }

        $cmd = '/bin/rm ' . self::FAX_FILES_PATH . DIRECTORY_SEPARATOR . $file;
        exec($cmd, $out, $result);
        if ($result !== 0) {
            throw new Exception(sprintf('Cannot remove fax file  %s', $file));
        }
    }

    /**
     * Validate fax number
     *
     * @param string $number Fax number.
     *
     * @return void
     * @throws InvalidDataException Invalid Data Exception.
     */
    private function validateFaxNumber(string $number): void {
        $this->_validator->reset();
        $this->_validator->rule('required', ['number']);
        $this->_validator->rule('PhoneNumber', ['number']);
        $errors = $this->_validator->validateData(['number' => $number], ['number']);

        if ($errors) {
            throw new InvalidDataException(\Wildix\Error::getMessage(\Wildix\Error::VALIDATION_ERROR), $errors);
        }
    }

    public function getSafeFilename(string $filename) : string {
        $replacementSymbols = array(
            ' ' => '_',
            '%' => '_',
            '[' => '_',
            ']' => '_',
            '`' => '',
            '$' => '',
            ';' => '',
            ':' => '',
            "'" => '',
            '"' => ''
        );

        return basename(strtr($filename, $replacementSymbols));
    }
}
