<?php
namespace Wildix\HttpSecurity;

use Wildix\Auth\Applications\Interfaces\OAuth2ServiceInterface;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\Validator;

class Service {
    private const WIZYCONF_ORIGIN_FIELD = 'wiziconf';

    private $_permanentAllowedOrigins;

    /**
     * Instance of validator.
     *
     * @var Validator
     */
    private $_validator;

    /**
     * Rules for validate data.
     *
     * @var array
     */
    private static $_validationRules = array(
        'SchemeAndDomain' => ModelSettings::FIELD_ORIGIN,
        'Alpha' => 'chromeExtension',
    );

    /**
     * Instance of repository.
     *
     * @var Repository
     */
    private $_repository;

    private OAuth2ServiceInterface $oAuth2Service;

    /**
     * Service constructor.
     *
     * @desc Initialize instances factory, repository and validator
     */
    public function __construct(Validator $validator, Repository $repository, OAuth2ServiceInterface $oAuth2Service, $origins = []) {
        $this->_validator               = $validator;
        $this->_repository              = $repository;
        $this->oAuth2Service            = $oAuth2Service;
        $this->_permanentAllowedOrigins = $origins;

        $this->_validator->rules(self::$_validationRules);
    }

    public function getSettings() {
        return $this->_repository->getSettings();
    }

    public function save(array $data) {
        if(isset($data[ModelSettings::FIELD_ORIGIN]) && is_array($data[ModelSettings::FIELD_ORIGIN])){
            foreach($data[ModelSettings::FIELD_ORIGIN] as $key=>$val) {
                if($val == '') {
                    unset($data[ModelSettings::FIELD_ORIGIN][$key]);
                    continue;
                }
                $data[ModelSettings::FIELD_ORIGIN][$key] = rtrim($val, '/');
            }
            $data[ModelSettings::FIELD_ORIGIN] = array_values(array_unique($data[ModelSettings::FIELD_ORIGIN]));
        }

        $this->validateAllowOriginArray($data);

        $data[ModelSettings::FIELD_ORIGIN] = array_values(array_unique(array_merge($data[ModelSettings::FIELD_ORIGIN] ?? [], $this->getPermanentAllowedOrigins())));

        $add = $this->getSettings();
        $add->update($data);

        $this->_repository->save($add);
    }

    private function validateAllowOriginArray($data) {
        $validateErrors = null;
        // to validate the data, replaced the template '*' on 'aaa'
        if (isset($data[ModelSettings::FIELD_ORIGIN]) && is_array($data[ModelSettings::FIELD_ORIGIN])) {
            foreach ($data[ModelSettings::FIELD_ORIGIN] as $key => $val) {
                if (str_starts_with($val, 'chrome-extension://')) {
                    $checkingValue = str_replace('chrome-extension://', '', $val);
                    $field         = 'chromeExtension';
                } else {
                    $checkingValue = str_replace('*', 'aaa', $val);
                    $field         = ModelSettings::FIELD_ORIGIN;
                }

                $errors = $this->_validator->validateData([$field => $checkingValue], [$field]);
                if ($errors) {
                    if ($validateErrors === null) {
                        $validateErrors = $errors;
                    }

                    $validateErrors[$field]['rows'][] = [
                        'key' => $key,
                        'value' => $checkingValue
                    ];
                }
            }
        }

        if ($validateErrors !== null) {
            throw new InvalidDataException(Error::getMessage(Error::InvalidData), $validateErrors);
        }
    }

    public function isOriginAllow(string $origin) {
        $originAllow  = $this->getSettings()->getOriginAllow();
        $oauthOrigins = $this->oAuth2Service->getAllOrigins();
        // stored config could miss default
        $predefined  = $this->getPermanentAllowedOrigins();
        $originAllow = array_merge($originAllow, $oauthOrigins, $predefined);
        $originAllow = array_unique($originAllow);
        foreach($originAllow as $allow) {
            if(strpos($allow, '*') !== false) {
                //if it contains a pattern
                $allow = addcslashes($allow, '[/:-.');
                $allow = str_replace('*', '[a-zA-Z\d\-]*', $allow);
                if(preg_match("/^".$allow."$/i", $origin)) {
                    return true;
                }
            }elseif ($allow == $origin){
                return true;
            }
        }

        return false;
    }
    //CSRF protection function, works only if remote host sent Origin header
    //Warning, Firefox don't send origin header from form, POST requests issue
    // Firefox 70+ supports Origin
    public static function isCsrf($headersArray) {
        $origin = $headersArray['origin'];
        $host = $headersArray['host'];

        if(!$origin && $headersArray['method'] != "GET" && $headersArray['referer']) {
            $origin = parse_url($headersArray['referer']);
            $origin = $origin['scheme']."://".$origin['host'] . (isset($origin['port']) ? ':'.$origin['port'] : '');
        }

        // TODO check
        if ($origin && $origin != env('SERVICE_PBXS_URI') && preg_replace('/http[s]?:\/\//', '', $origin) != $host) {
            return false;
        }
        return true;
    }

    /**
     * Get permanent allowed origins
     *
     * @return string[]
     */
    public function getPermanentAllowedOrigins(): array {
        return $this->_permanentAllowedOrigins;
    }

    /**
     * Returns allowed origins array without permanent origins
     *
     * @return array
     */
    public function getAllowedOrigins(): array {
        return [
            ModelSettings::FIELD_ORIGIN => array_values(array_diff($this->getSettings()->toArray()[ModelSettings::FIELD_ORIGIN], $this->getPermanentAllowedOrigins()))
        ];
    }

    public function isWizyconfOrigin(string $origin): bool {
        $wiziconfOrigin = $this->getPermanentAllowedOrigins()[self::WIZYCONF_ORIGIN_FIELD] ?? '';

        return $wiziconfOrigin === $origin;
    }
}
