<?php

namespace Wildix\Import;

use Wildix\Colleagues\Colleagues;
use Wildix\Services\Exception;
use Wildix\Services\Kite;

/**
 * Class Infusionsoft
 */
class Infusionsoft extends Import {

    const URI_PATTERN = '%s://%s/api/Infusionsoft/Callback/?callback=callbackInfusionsoftFunctionContacts';
    /**
     * Authentication url.
     *
     * @var string
     */
    private $_authUrl;

    /**
     * Url to work with api.
     *
     * @var string
     */
    private $_apiUrl;

    /**
     * Url for to get or refresh token.
     *
     * @var string
     */
    private $_accessTokenUrl;

    /**
     * Access Token must be in format.
     *      {
     *          "access_token":        "{access_token}",
     *          "token_type":          "bearer",
     *          "expires_in":          86400,
     *          "refresh_token":       "{refresh_token}",
     *          "received_token_time": {timestamp},
     *          "scope":               "full|{application_name}" //non required
     *      }
     *
     * @var null
     */
    private $_accessToken = null;

    /**
     * List of colleagues of authorized user.
     *
     * @var array
     */
    private $_users = [];

    /**
     * Number of users received from api.
     *
     * @var integer
     */
    private $_usersOffset = 0;

    /**
     * Shows whether we got all users from api.
     *
     * @var boolean
     */
    private $_allUsersReceived = false;

    /**
     * Infusionsoft constructor.
     *
     * @param Kite    $kite     Kite class.
     * @param string  $authUrl  Auth url.
     * @param string  $apiUrl   Api url.
     * @param integer $logLevel Log class.
     */
    public function __construct(Kite $kite, string $authUrl, string $apiUrl, $logLevel = \Syslog::ERROR) {
        parent::__construct($kite, $logLevel);
        $this->_authUrl        = $authUrl . '/app/oauth/authorize';
        $this->_apiUrl         = $apiUrl . '/crm/rest/v1';
        $this->_accessTokenUrl = $apiUrl . '/token';
        $this->setIntegrationAppCredentials(Kite::INFUSIONSOFT_INTEGRATION_APP_KEY, Kite::INFUSIONSOFT_INTEGRATION_APP_URL);
    }

    /**
     * Getting application name from token scope field received from infusionsoft.
     *
     * @param array $token Token array.
     *
     * @return string Application name or empty string.
     */
    private function getAppFromToken(array $token) {
        if (empty($token['scope'])) {
            return '';
        }

        $scopeAsArray = explode('|', $token['scope']);
        return $scopeAsArray[count($scopeAsArray) - 1];
    }

    /**
     * Getting headers for curl request that contain bearer authentication.
     *
     * @return array Headers.
     */
    private function getHeadersWithBearerAuth(): array {
        if ($this->_accessToken === null) {
            return [];
        }
        return [
            'Accept: application/json',
            'Authorization: Bearer ' . $this->_accessToken['access_token'],
        ];
    }

    /**
     * Getting user by id from infusionsoft colleagues list.
     *
     * @param integer|string $userId User id.
     *
     * @return array|null Return user from infusionsoft colleagues list or null if not found.
     */
    private function getUserByIdFromInfusionsoftUsersList($userId) {
        if (isset($this->_users[$userId])) {
            return $this->_users[$userId];
        }

        if ($this->_allUsersReceived) {
            return null;
        }

        $foundUser = null;
        $limit     = 1000;

        do {
            $users = $this->getUsers($this->_accessToken, $limit, $this->_usersOffset);

            if (!isset($users)) {
                $this->getLog()->info(__LINE__, 'Users not found');
                break;
            }

            $count = $users['count'];
            $users = $users['users'];

            foreach ($users as $user) {
                $id                = $user['id'];
                $userEntry         = ['email' => $user['email_address']];
                $this->_users[$id] = $userEntry;

                if ($userId === $id) {
                    $foundUser = $userEntry;
                }
            }

            $this->_usersOffset += $limit;

            if ($this->_usersOffset >= $count) {
                $this->_allUsersReceived = true;
            }
        } while ($foundUser === null && !$this->_allUsersReceived);

        return $foundUser;
    }

    /**
     * Getting user dn from ldap colleagues by user email.
     *
     * @param string $email User email.
     *
     * @return string User dn or empty string if user not exist.
     */
    private function getUserDnByEmail($email) {
        $user = Colleagues::getInstance()->getUserByEmail($email);
        return $user ? $user->getDn() : '';
    }

    /**
     * Getting dn from ldap for user owner.
     *
     * @param integer|string $ownerId User owner infusionsoft id.
     *
     * @return string Dn or empty string if owner not found in ldap.
     */
    private function getLdapDnForUserOwner($ownerId) {
        if (empty($ownerId)) {
            return '';
        }

        $owner = $this->getUserByIdFromInfusionsoftUsersList($ownerId);

        if ($owner === null) {
            return '';
        }

        if (isset($owner['dn'])) {
            return $owner['dn'];
        }

        $ownerDn = $this->getUserDnByEmail($owner['email']);
        $this->_users[$ownerId]['dn'] = $ownerDn;

        return $ownerDn;
    }

    /**
     * Getting flat array of user phone/fax numbers.
     *
     * @param mixed $items Assoc array of user phone/fax numbers.
     *
     * @return array Flat array of user numbers.
     */
    private function getContactPhoneFaxNumbers($items): array {
        $numbers = [];

        if (!is_array($items)) {
            return $numbers;
        }

        foreach ($items as $item) {
            $key       = ucfirst(strtolower($item['field']));
            $type      = isset($item['type']) ? trim(strtolower($item['type'])) : '';
            $number    = isset($item['number']) ? $item['number'] : '';
            $extension = isset($item['extension']) ? $item['extension'] : '';

            switch ($type) {
                case 'work':
                    $numbers['work' . $key]          = $number;
                    $numbers['workExtension' . $key] = $extension;
                    break;
                case 'home':
                    $numbers['home' . $key]          = $number;
                    $numbers['homeExtension' . $key] = $extension;
                    break;
                case 'mobile':
                    $numbers['mobile' . $key]          = $number;
                    $numbers['mobileExtension' . $key] = $extension;
                    break;
                case 'other':
                    $numbers['other' . $key]          = $number;
                    $numbers['otherExtension' . $key] = $extension;
                    break;
                default:
                    $numbers[$key]               = $number;
                    $numbers['extension' . $key] = $extension;
                    break;
            }
        }

        return $numbers;
    }

    /**
     * Getting flat array of user addresses.
     *
     * @param mixed $items Assoc array of user addresses.
     *
     * @return array Flat array of user addresses.
     */
    private function getContactAddresses($items): array {
        $addresses = [];

        if (!is_array($items)) {
            return $addresses;
        }

        $addressesMap = [
            'CountryCode' => 'country_code',
            'StreetAddress1' => 'line1',
            'StreetAddress2' => 'line2',
            'City' => 'locality',
            'State' => 'region',
        ];

        foreach ($items as $key => $addressInfo) {
            $addressType = isset($addressInfo['field']) ? trim(strtolower($addressInfo['field'])) : '';

            foreach ($addressesMap as $addressKey => $iAddressKey) {
                $addresses[$addressType . $addressKey] = isset($addressInfo[$iAddressKey]) ? $addressInfo[$iAddressKey] : '';
            }

            if (isset($addressInfo['zip_code'])) {
                $zipFour = !empty($addressInfo['zip_four']) ? '-' . $addressInfo['zip_four'] : '';
                $addresses[$addressType . 'PostalCode'] = $addressInfo['zip_code'] . $zipFour;
            }
        }

        return $addresses;
    }

    /**
     * Method which generated authentication url.
     *
     * @param string $scope       Scope.
     * @param string $callbackUrl After get auth code from infusionsoft, kite redirect to pbx by this url.
     *
     * @return string Auth url.
     */
    public function getAuthUrl($scope, $callbackUrl) {
        $params = [
            'client_id'     => $this->getIntegrationApp('clientId'),
            'redirect_uri'  => urlencode($this->getIntegrationApp('redirectUrl')),
            'response_type' => 'code',
            'scope'         => urlencode($scope),
            'state'         => $callbackUrl,
        ];

        return $this->_authUrl . '?' . urldecode(http_build_query($params));
    }

    /**
     * Getting token, see $this->_accessToken structure from doc.
     *
     * @param null|string|number $code Authorization code received from infusionsoft.
     *
     * @return array|null Token or null if token is not set.
     */
    public function getToken($code = null) {
        if ($this->_accessToken !== null || $code === null) {
            return $this->_accessToken;
        }

        $params = [
            'client_id'     => $this->getIntegrationApp('clientId'),
            'client_secret' => $this->getIntegrationApp('clientSecret'),
            'redirect_uri'  => $this->getIntegrationApp('redirectUrl'),
            'grant_type'    => 'authorization_code',
            'code'          => $code
        ];

        try {
            $this->_accessToken = $this->sendRequest('POST', $this->_accessTokenUrl, $params);
        } catch (\Exception $e) {
            $this->getLog()->error(__LINE__, 'Failed to get access token: ' . $e->getCode() . ' ' . $e->getMessage());
            return [];
        }

        $this->_accessToken['received_token_time'] = time();

        return $this->_accessToken;
    }

    /**
     * Refresh token.
     *
     * @param array $token Token, see $this->_accessToken structure from doc.
     *
     * @return array Updated token, {refresh_token} and {access_token} fields both updated.
     */
    public function refreshToken(array $token): array {
        $params = [
            'refresh_token' => $token['refresh_token'],
            'grant_type'    => 'refresh_token',
        ];

        $this->enableBasicAuth();

        try {
            $newToken = $this->sendRequest('POST', $this->_accessTokenUrl, $params);
            $this->disableBasicAuth();
        } catch (\Exception $e) {
            $this->getLog()->error(__LINE__, 'Failed to refresh access token: ' . $e->getCode() . ' ' . $e->getMessage());
            $this->disableBasicAuth();
            return [];
        }

        $newToken['received_token_time'] = time();

        return $newToken;
    }

    /**
     * Check token state.
     *
     * @param array $token Token, see $this->_accessToken structure from doc.
     *
     * @return integer Token state by const.
     */
    public function isValidToken(array $token) {
        if (empty($token['access_token']) || empty($token['received_token_time']) || empty($token['expires_in'])) {
            return self::NO_ACCESS_TOKEN;
        }

        if ($token['received_token_time'] + $token['expires_in'] - time() > 3 * 60 * 60) {
            return self::TOKEN_VALID;
        }

        $refreshTokenExpiresIn = 180 * 24 * 60 * 60;
        if ($token['received_token_time'] + $refreshTokenExpiresIn - time() > 0) {
            return self::TOKEN_EXPIRED;
        }

        return self::REFRESH_TOKEN_EXPIRED;
    }

    /**
     * Set token.
     *
     * @param array $token Token, see $this->_accessToken structure from doc.
     *
     * @return void
     * @throws \Wildix\Exception If token expired and we cannot refresh it.
     */
    public function setToken(array $token): void {
        // @codingStandardsIgnoreStart
        $this->getLog()->info(__LINE__, 'setToken: ' . print_r($token, true));
        // @codingStandardsIgnoreEnd
        $result = $this->isValidToken($token);

        if ($result === self::TOKEN_VALID) {
            $this->_accessToken = $token;
            return;
        }

        if ($result !== self::TOKEN_EXPIRED) {
            // @codingStandardsIgnoreStart
            throw new \Wildix\Exception('Validation of the token returned ' . $result);
            // @codingStandardsIgnoreEnd
        }

        $token  = $this->refreshToken($token);
        $result = $this->isValidToken($token);

        if ($result !== self::TOKEN_VALID) {
            // @codingStandardsIgnoreStart
            throw new \Wildix\Exception('Can not refresh token');
            // @codingStandardsIgnoreEnd
        }

        $this->_accessToken = $token;
    }

    /**
     * Get info about user by token.
     *
     * @param array $token Token, see $this->_accessToken structure from doc.
     *
     * @return array|null Info about user or null if something went wrong.
     */
    public function getUserInfo(array $token) {
        try {
            $this->setToken($token);
        } catch (\Exception $e) {
            $this->getLog()->error(__LINE__, 'Failed to set access token: ' . $e->getMessage());
            return null;
        }

        $params = ['access_token' => $this->_accessToken['access_token']];

        try {
            $userInfo = $this->sendRequest('GET', $this->_apiUrl . '/oauth/connect/userinfo', $params);
        } catch (\Exception $e) {
            $this->getLog()->error(__LINE__, 'Failed to get user info: ' . $e->getCode() . ' ' . $e->getMessage());
            return null;
        }

        $userInfo['app'] = $this->getAppFromToken($this->_accessToken);

        return $userInfo;
    }

    /**
     * Getting user contacts list.
     *
     * @param array   $token  Token, see $this->_accessToken structure from doc.
     * @param integer $limit  The number of contacts that must be requested through api.
     * @param integer $offset The number of contacts that have already been requested through api.
     * @param string  $order  Result list will be sorted by this field.
     *
     * @return array|null Contacts list or null if something went wrong.
     */
    public function getContacts(array $token, int $limit = 100, int $offset = 0, string $order = 'id') {
        try {
            $this->setToken($token);
        } catch (\Exception $e) {
            $this->getLog()->error(__LINE__, 'Failed to set access token: ' . $e->getMessage());
            return null;
        }

        $params = [
            'limit' => $limit,
            'offset' => $offset,
            'order' => $order,
        ];

        try {
            $contacts = $this->sendRequest('GET', $this->_apiUrl . '/contacts', $params, false, $this->getHeadersWithBearerAuth());
        } catch (\Exception $e) {
            $this->getLog()->error(__LINE__, 'Failed to get list of contacts: ' . $e->getCode() . ' ' . $e->getMessage());
            return null;
        }

        return $contacts;
    }

    /**
     * Getting users list.
     *
     * @param array   $token           Token, see $this->_accessToken structure from doc.
     * @param integer $limit           The number of users that must be requested through api.
     * @param integer $offset          The number of users that have already been requested through api.
     * @param boolean $includeInactive Include inactive users to result.
     * @param boolean $includePartners Include partners to result.
     *
     * @return array|null Users list or null if something went wrong.
     */
    public function getUsers(array $token, int $limit = 100, int $offset = 0, bool $includeInactive = true, bool $includePartners = true) {
        try {
            $this->setToken($token);
        } catch (\Exception $e) {
            $this->getLog()->error(__LINE__, 'Failed to set access token: ' . $e->getMessage());
            return null;
        }

        $params = [
            'limit' => $limit,
            'offset' => $offset,
            'include_inactive' => $includeInactive,
            'include_partners' => $includePartners,
        ];

        try {
            $users = $this->sendRequest('GET', $this->_apiUrl . '/users', $params, false, $this->getHeadersWithBearerAuth());
        } catch (\Exception $e) {
            $this->getLog()->error(__LINE__, 'Failed to get list of users: ' . $e->getCode() . ' ' . $e->getMessage());
            return null;
        }

        return $users;
    }

    /**
     * Transform infusionsoft api contact from assoc to flat array.
     *
     * @param array $iContact Contact received from infusionsoft api.
     *
     * @return array Contact as flat array.
     */
    public function parseContact(array $iContact): array {
        $contact            = [];
        $defaultStringValue = '';

        $contactFieldsMap = [
            'id' => 'id',
            'created' => 'date_created',
            'updated' => 'last_updated',
            'firstName' => 'given_name',
            'middleName' => 'middle_name',
            'lastName' => 'family_name',
            'preferredName' => 'preferred_name',
            'namePrefix' => 'prefix',
            'nameSuffix' => 'suffix',
            'birthday' => 'birthday',
            'notes' => 'notes',
            'owner' => 'owner_id',
            'website' => 'website',
            'contactType' => 'contact_type',
        ];

        foreach ($contactFieldsMap as $contactField => $iContactField) {
            $contact[$contactField] = isset($iContact[$iContactField]) ? $iContact[$iContactField] : $defaultStringValue;
        }

        if (isset($iContact['company']['company_name'])) {
            $contact['company'] = $iContact['company']['company_name'];
        }

        if (!empty($iContact['email_addresses'])) {
            $emails = $iContact['email_addresses'];
            foreach ($emails as $item) {
                $email           = strtolower($item['field']);
                $contact[$email] = $item['email'];
            }
        }

        if (!empty($iContact['phone_numbers'])) {
            $phones  = $this->getContactPhoneFaxNumbers($iContact['phone_numbers']);
            $contact = array_merge($contact, $phones);
        }

        if (!empty($iContact['fax_numbers'])) {
            $faxes   = $this->getContactPhoneFaxNumbers($iContact['fax_numbers']);
            $contact = array_merge($contact, $faxes);
        }

        if (!empty($iContact['addresses'])) {
            $addresses = $this->getContactAddresses($iContact['addresses']);
            $contact   = array_merge($contact, $addresses);
        }

        $contact['owner'] = $this->getLdapDnForUserOwner($contact['owner']);
        $contact['name']  = trim($contact['firstName']) . ' ' . trim($contact['middleName']) . ' ' . trim($contact['lastName']);

        return $contact;
    }

    protected function getBasicAuthCredentials() {
        $clientId     = $this->getIntegrationApp('clientId');
        $clientSecret = $this->getIntegrationApp('clientSecret');

        return $clientId . ':' . $clientSecret;
    }

    protected function parse($data) {
        $error = $data['error'];

        if (!empty($error['reason'])) {
            throw new Exception\ServiceException($error['reason'], $error['code']);
        }

        $result   = $data['result'];
        $httpCode = $data['info']['http_code'];

        if ($httpCode != 200) {
            throw new Exception\ServiceException($result, $httpCode);
        }

        $decodedResult = json_decode($result, true);

        if (!$decodedResult || !is_array($decodedResult)) {
            throw new Exception\ServiceException($result, 'RESPONSE_DECODE_ERROR');
        }

        return $decodedResult;
    }
}
