<?php

namespace Wildix\Import\ZohoClient;

use Wildix\Exception;

/**
 * Class ZohoClient
 *
 * Token array must be in format
 *      {
 *          "access_token":        "{access_token}",
 *          "refresh_token":       "{refresh_token}",
 *          "expires_in_sec":      3600,
 *          "api_domain":          "https://www.zohoapis.com",
 *          "token_type":          "Bearer",
 *          "expires_in":          3600000,
 *          "received_token_time": 4534566,
 *          "location":            "eu"
 *      }
 */
class ZohoClient {

    const TOKEN_VALID     = 1;
    const NO_ACCESS_TOKEN = 2;
    const NO_ID_TOKEN     = 3;
    const TOKEN_EXPIRED   = 4;

    /**
     * Token url without domain.
     *
     * @var string
     */
    private $_tokenUrl = '/oauth/v2/token';

    /**
     * Authentication url.
     *
     * @var string
     */
    private $_authUrl;

    /**
     * Revoke url for com domain url.
     *
     * @var string
     */
    private $_revokeUrl = '/oauth/v2/token/revoke';

    /**
     * Logger instance.
     *
     * @var \Logger\Log
     */
    private $_log;

    /**
     * Kite instance.
     *
     * @var \Wildix\Services\Kite
     */
    private $_kite;

    public function __construct(\Logger\Log $logger, \Wildix\Services\Kite $kite, string $_authUrl) {
        $this->_log     = $logger;
        $this->_kite    = $kite;
        $this->_authUrl = $_authUrl . '/oauth/v2/auth';
    }

    /**
     * Method which generated authentication url.
     *
     * @param string $scope        Scope.
     * @param string $callbackLink Callback url.
     *
     * @return string
     */
    public function getAuthUrl($scope, $callbackLink) {
        $params = array(
            'scope'         => urlencode($scope),
            'client_id'     => $this->getZohoProductsIntegrationApp('clientId'),
            'response_type' => 'code',
            'access_type'   => 'offline',
            'redirect_uri'  => urlencode($this->getZohoProductsIntegrationApp('redirectUrl')),
            'state'         => $callbackLink,
            'prompt'        => 'consent'
        );

        return $this->_authUrl . '?' . urldecode(http_build_query($params));
    }

    /**
     * Token.
     *
     * @var null
     */
    private $_tokenArray = null;

    /**
     * Getting Token.
     *
     * @param string      $accountsServer Domain url.
     * @param string      $location       Location.
     * @param null|string $code           Returned code.
     *
     * @return mixed|null
     */
    public function getToken($accountsServer, $location, $code = null) {
        if ($this->_tokenArray === null && $code !== null) {
            $params = array(
                'client_id'     => $this->getZohoProductsIntegrationApp('clientId'),
                'client_secret' => $this->getZohoProductsIntegrationApp('clientSecret'),
                'redirect_uri'  => $this->getZohoProductsIntegrationApp('redirectUrl'),
                'grant_type'    => 'authorization_code',
                'code'          => $code
            );

            try {
                $this->_tokenArray = $this->_postRequest($accountsServer . $this->_tokenUrl, $params);
            } catch (\Exception $e) {
                return array(
                    'error' => array(
                        'code' => $e->getCode(),
                        'message' => $e->getMessage()
                    )
                );
            }

            $this->_tokenArray['received_token_time'] = time();
            $this->_tokenArray['location']            = $location;
            $this->_tokenArray['accounts_server']     = $accountsServer;
        }

        return $this->_tokenArray;
    }

    /**
     * Revoke token.
     *
     * @param string $apiDomain    API Domain Url.
     * @param string $refreshToken Refresh token.
     *
     * @return array
     **/
    public function revokeToken($apiDomain, $refreshToken) {
        $params = array(
            'token' => $refreshToken
        );

        try {
            $response = $this->_getRequest($apiDomain . $this->_revokeUrl, $params);
        } catch (\Exception $e) {
            return array(
                'error' => array(
                    'code' => $e->getCode(),
                    'message' => $e->getMessage()
                )
            );
        }

        return $response;
    }

    /**
     * Current user information.
     *
     * @var null
     */
    private $_userInfo = null;

    /**
     * Getting information about current user.
     *
     * @param array $tokenArray Token array.
     *
     * @return mixed|null
     */
    public function getUserInfo(array $tokenArray) {
        try {
            $this->setToken($tokenArray);
        } catch (\Exception $e) {
            return array(
                'error' => array(
                    'code' => $e->getCode(),
                    'message' => $e->getMessage()
                )
            );
        }

        $params = array(
            'type' => 'CurrentUser'
        );

        try {
            $response = $this->_getRequest($tokenArray['api_domain'] . '/crm/v2/users', $params);
        } catch (\Exception $e) {
            return array(
                'error' => array(
                    'code' => $e->getCode(),
                    'message' => $e->getMessage()
                )
            );
        }

        $this->_userInfo = $response['users'][0];

        return $this->_userInfo;
    }



    private $_moreRecords = true;

    /**
     * Getting all users.
     *
     * @param array   $tokenArray Token array.
     * @param integer $perPage    Count records per one page.
     * @param integer $count      Number of current page.
     *
     * @return array|boolean
     */
    public function getUsers(array $tokenArray, $perPage = 200, $count = 1) {
        try {
            $this->setToken($tokenArray);
        } catch (\Exception $e) {
            return array(
                'error' => array(
                    'code' => $e->getCode(),
                    'message' => $e->getMessage()
                )
            );
        }

        $params = array(
            'type' => 'ActiveUsers',
            'page' => $count,
            'per_page' => $perPage
        );

        try {
            $response = $this->_getRequest($tokenArray['api_domain'] . '/crm/v2/users', $params);
        } catch (\Exception $e) {
            return array(
                'error' => array(
                    'code' => $e->getCode(),
                    'message' => $e->getMessage(),
                ),
            );
        }

        if (isset($response['status']) && $response['status'] === 'error') {
            return array(
                'error' => array(
                    'code' => isset($response['code']) ? $response['code'] : 'UNKNOWN',
                    'message' => isset($response['message']) ? $response['message'] : 'Unknown error',
                ),
            );
        }

        $result = $this->parseUsers($response);

        if ($this->_moreRecords === true) {
            if ($result['info']['more_records'] == false) {
                $this->_moreRecords = false;
            }

            return $result;
        }

        return false;
    }

    /**
     * Prepare users array for import.
     *
     * @param array $users Users array.
     *
     * @return array
     */
    private function parseUsers(array $users) {
        if (!empty($users) && !empty($users['users'])) {
            foreach ($users['users'] as &$user) {
                $user['role']        = $user['role']['name'];
                $user['profile']     = $user['profile']['name'];
                $user['dateOfBirth'] = $user['dob'];
                $user['firstName']   = $user['first_name'];
                $user['lastName']    = $user['first_name'];
                $user['fullName']    = $user['full_name'];
                $user['language']    = substr($user['language'], 0, 2);
            }
        }

        return $users;
    }

    /**
     * Refresh token.
     *
     * @param array $tokenArray Token array.
     *
     * @return mixed
     */
    private function refreshToken(array $tokenArray) {
        $params = array(
            'client_id' => $this->getZohoProductsIntegrationApp('clientId'),
            'client_secret' => $this->getZohoProductsIntegrationApp('clientSecret'),
            'refresh_token' => $tokenArray['refresh_token'],
            'grant_type' => 'refresh_token'
        );

        try {
            $newToken = $this->_postRequest($tokenArray['accounts_server'] . $this->_tokenUrl . '?' . urldecode(http_build_query($params)));
        } catch (\Exception $e) {
            return array(
                'error' => array(
                    'code' => $e->getCode(),
                    'message' => $e->getMessage()
                )
            );
        }

        $newToken['received_token_time'] = time();

        return $newToken;
    }

    /**
     * Token validation.
     *
     * @param array $tokenArray Token array.
     *
     * @return integer
     */
    private function isValidToken(array $tokenArray) {
        if (empty($tokenArray['access_token'])) {
            return self::NO_ACCESS_TOKEN;
        }

        if (!empty($tokenArray['received_token_time']) && !empty($tokenArray['expires_in'])) {
            if ($tokenArray['received_token_time'] + $tokenArray['expires_in'] / 1000 < time() - 120) {
                return self::TOKEN_EXPIRED;
            }
        }

        return self::TOKEN_VALID;
    }

    /**
     * Setting token.
     *
     * @param array $tokenArray Token array.
     *
     * @return integer
     * @throws \Exception If can not refresh token.
     */
    public function setToken(array $tokenArray) {
        // @codingStandardsIgnoreStart
        $this->_log->info(__LINE__, 'setToken: ' . print_r($tokenArray, true));
        // @codingStandardsIgnoreEnd
        $result = $this->isValidToken($tokenArray);
        if ($result == self::TOKEN_VALID) {
            $this->_tokenArray = $tokenArray;
        } else if ($result == self::TOKEN_EXPIRED) {
            $tokenArray = $this->refreshToken($tokenArray);
            $result     = $this->isValidToken($tokenArray);

            if ($result == self::TOKEN_VALID) {
                $this->_tokenArray = $tokenArray;
            } else {
                throw new Exception('Can not refresh token');
            }
        }
        return $result;
    }

    /**
     * Array with SecretCode and Client ID.
     *
     * @var null
     */
    private $_zohoProductsIntegrationApp = null;

    /**
     * Get ZohoCRM parameters.
     *
     * @param string $param Parameter name.
     *
     * @return mixed
     */
    private function getZohoProductsIntegrationApp($param) {
        if ($this->_zohoProductsIntegrationApp === null) {
            $this->_zohoProductsIntegrationApp = $this->_kite->getZohoProductsIntegrationApp();
        }

        return $this->_zohoProductsIntegrationApp[$param];
    }

    /**
     * Sending post request.
     *
     * @param string $url        Url.
     * @param array  $postParams Post parameters.
     *
     * @return mixed
     * @throws \Exception If curl returns false.
     */
    private function _postRequest($url, array $postParams = array()) {
        $this->_log->info(__LINE__, 'POST: ' . $url);
        // @codingStandardsIgnoreStart
        $this->_log->info(__LINE__, print_r($postParams, true));
        // @codingStandardsIgnoreEnd
        $curl = curl_init();

        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_POST, true);
        curl_setopt($curl, CURLOPT_POSTFIELDS, urldecode(http_build_query($postParams)));
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_TIMEOUT, 5);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, false);

        $result = curl_exec($curl);

        if ($result == false) {
            throw new Exception(curl_error($curl), curl_errno($curl));
        } else {
            $result = json_decode($result, true);
        }

        curl_close($curl);

        // phpcs:disable Squiz.PHP.DiscouragedFunctions.Discouraged
        $this->_log->debug(__LINE__, 'Response: ' . print_r($result, true));
        // phpcs:enable

        return $result;
    }

    /**
     * Sending get request.
     *
     * @param string $url    Url.
     * @param array  $params Parameters array.
     *
     * @return mixed
     * @throws \Exception If curl returns false.
     */
    private function _getRequest(string $url, array $params = array()) {
        if (!empty($params)) {
            $url .= '?' . http_build_query($params);
        }

        $headers = array(
            'Authorization: Bearer ' . $this->_tokenArray['access_token'],
            'Accept: application/json; odata.metadata=full',
        );

        $this->_log->info(__LINE__, 'GET: ' . $url);

        $curl = curl_init();
        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_HTTPGET, true);
        curl_setopt($curl, CURLOPT_CONNECTTIMEOUT, 5);
        curl_setopt($curl, CURLOPT_TIMEOUT, 120);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, true);

        curl_setopt($curl, CURLOPT_USERAGENT, 'Mozilla/5.0 (Windows; U; Windows NT 5.1; ru; rv:1.8.0.9) Gecko/20061206 Firefox/1.5.0.9');
        curl_setopt($curl, CURLOPT_HTTPHEADER, $headers);

        $result = curl_exec($curl);

        if ($result == false) {
            throw new Exception(curl_error($curl), curl_errno($curl));
        } else {
            $result = json_decode($result, true);
        }

        curl_close($curl);

        // @codingStandardsIgnoreStart
        $this->_log->debug(__LINE__, 'Response: ' . print_r($result, true));
        // @codingStandardsIgnoreEnd
        return $result;
    }
}
