<?php

namespace Wildix\Integration\XBees\Repository;

use Aws\CognitoIdentityProvider\CognitoIdentityProviderClient;
use Aws\CognitoIdentityProvider\Exception\CognitoIdentityProviderException;
use Aws\Result;
use Aws\Sdk;
use Models\PBX;
use Wildix\Error;
use Wildix\Exception;
use Wildix\Generator\StringGenerator;
use Wildix\Integration\XBees\Interfaces\AwsRepositoryInterface;
use Wildix\Integration\XBees\Objects\CpaConfig;
use Wildix\Services\GlobalConfig;

/**
 * Class AwsRepository
 */
class AwsRepository implements AwsRepositoryInterface {

    private $_globalConfig;
    private $_cpaConfig;
    private $_pbx;
    private $_awsSdk;

    public function __construct(GlobalConfig $globalConfig, PBX $pbx, Sdk $sdk) {
        $this->_globalConfig = $globalConfig;
        $this->_pbx          = $pbx;
        $this->_awsSdk       = $sdk;

        $this->_cpaConfig = $this->getCpaConfig();
    }

    public function respondToAuthChallenge(string $session): Result {
        return $this->getCognitoIdentityProvider()->respondToAuthChallenge(
            [
                'ChallengeName' => 'CUSTOM_CHALLENGE',
                'ChallengeResponses' => [
                    'USERNAME' => $this->getPbx()->getSerial(),
                    'ANSWER' => $this->getPbx()->getPbxKey()
                ],
                'ClientId' => $this->getCpaConfig()->getAppClient(),
                'Session' => $session
            ]
        );
    }

    public function getAuth(): Result {
        $result = null;

        try {
            $result = $this->initiateAuth();
        } catch (CognitoIdentityProviderException $e) {
            if ($e->getAwsErrorCode() === 'UserNotFoundException') {
                $this->signUp();
                $result = $this->initiateAuth();
            }
        }

        return $result;
    }

    private function signUp(): void {
        $this->getCognitoIdentityProvider()->signUp(
            [
                'ClientId' => $this->getCpaConfig()->getAppClient(),
                'Password' => StringGenerator::random(),
                'Username' => $this->getPbx()->getSerial(),
                'ValidationData' => [
                    [
                        'Name' => 'key',
                        'Value' => $this->getPbx()->getPbxKey(),
                    ],
                ],
            ]
        );
    }

    private function initiateAuth(): Result {
        return $this->getCognitoIdentityProvider()->initiateAuth(
            [
                'AuthFlow' => 'CUSTOM_AUTH',
                'AuthParameters' => [
                    'USERNAME' => $this->getPbx()->getSerial()
                ],
                'ClientId' => $this->getCpaConfig()->getAppClient()
            ]
        );
    }

    private function getCognitoIdentityProvider(): CognitoIdentityProviderClient {
        return $this->getAwsSdk()->createCognitoIdentityProvider(
            [
                'region' => $this->getCpaConfig()->getRegion(),
                'version' => 'latest',
                'credentials' => false
            ]
        );
    }

    private function getAwsSdk(): Sdk {
        return $this->_awsSdk;
    }

    private function getGlobalConfig(): GlobalConfig {
        return $this->_globalConfig;
    }

    private function getCpaConfig(): CpaConfig {
        if (!$this->_cpaConfig) {
            $config = $this->getGlobalConfig()->getConfig(GlobalConfig::TYPE_COGNITO);

            if (!isset($config['cpa'])) {
                throw new Exception(Error::CognitoPBXAuthDataNotPresent, 500);
            }

            $this->_cpaConfig = new CpaConfig($config['cpa']);
        }

        return $this->_cpaConfig;
    }

    private function getPbx(): PBX {
        return $this->_pbx;
    }
}
