<?php
namespace Wildix;

use Logger\Log;
use Wildix\Databases\WMSDB;

/**
 * Class LanguageAndRegion
 */
class LanguageAndRegion {

    const CONFIG_TONEZONE         = '/etc/callweaver/indications.conf';
    const KEY_INSTALLED_LANGUAGES = 'INSTALLED_LANGUAGES_SOUNDS';
    const LOG_INSTALL_FILE        = '/tmp/session/install_sounds_pakcs.log';

    const DEFAULT_LANG = 'en';

    private $_defaultInstalledLanguages = [self::DEFAULT_LANG];

    private $_log;
    private $_db;
    private $_listTonezones = null;

    public function __construct(WMSDB $db, Log $log) {
        $this->_db  = $db;
        $this->_log = $log;
    }

    private function getLog(): Log {
        return $this->_log;
    }

    private function getWmsDb(): WMSDB {
        return $this->_db;
    }

    public function getListTonezones() {
        if (!$this->_listTonezones) {
            $tmp = parse_ini_file(self::CONFIG_TONEZONE, true, INI_SCANNER_RAW);
            foreach ($tmp as $key => $val) {
                if ($key != 'general') {
                    $this->_listTonezones[$key] = $val['description'];
                }
            }
        }
        return $this->_listTonezones;
    }

    private $_listLanguagesForInstall = null;
    public function getListLanguagesForInstall() {
        if (!$this->_listLanguagesForInstall) {
            exec('sudo apt-get update > /dev/null 2>&1; apt-cache search --names-only "^wildix-sounds-[a-z\-]{2,5}$"', $out);
            $this->getLog()->info(__LINE__, 'List packs langs from repository: ' . join("\n", $out));

            foreach ($out as $packageLine) {
                if (preg_match('/wildix-sounds-([a-z\-]{2,5})\s/', $packageLine, $matches)) {
                    $this->_listLanguagesForInstall[] = $matches[1] === 'us' ? 'en-us' : $matches[1];
                }
            }
        }
        return $this->_listLanguagesForInstall;
    }

    public function initInstalledLanguages() {
        if (empty($this->_findSysVarInstalledLanguages())) {
            $this->setLanguages($this->_defaultInstalledLanguages);
        } else {
            $this->syncLangsFromDbToPbx();
        }
    }

    public function getInstalledLanguages() {
        $value = $this->_findSysVarInstalledLanguages();

        if ($value) {
            return json_decode($value, true);
        } else {
            return $this->_defaultInstalledLanguages;
        }
    }

    public function setLanguages(array $langs = []) {
        if (!in_array(self::DEFAULT_LANG, $langs, true)) {
            $langs[] = self::DEFAULT_LANG;
        }

        if (!$this->getWmsDb()->setSystemVariable(self::KEY_INSTALLED_LANGUAGES, json_encode($langs))) {
            throw new Exception('Error install sounds');
        }
        return $this->syncLangsFromDbToPbx();
    }

    public function syncLangsFromDbToPbx() {
        if (file_exists(self::LOG_INSTALL_FILE)) {
            unlink(self::LOG_INSTALL_FILE);
        }

        $langsFromDb = $this->getInstalledLanguages();

        $wrongUsPackage = array_search('en-us', $langsFromDb);
        if ($wrongUsPackage !== false) {
            $langsFromDb[$wrongUsPackage] = 'us';
        }


        exec('dpkg -l | grep -E "^ii\s+wildix-sounds-[a-z]{2}\s"', $out);
        $this->getLog()->info(__LINE__, 'List packs langs installed to pbx: ' . join("\n", $out));
        $langsFromPbx = array();

        $countLine = count($out);
        for ($i = 0; $i < $countLine; $i++) {
            if (preg_match('/wildix-sounds-([a-z]{2})/', $out[$i], $matches)) {
                $langsFromPbx[] = $matches[1];
            }
        }

        $install        = array_diff($langsFromDb, $langsFromPbx);
        $uninstall      = array_diff($langsFromPbx, $langsFromDb);
        $countInstall   = count($install);
        $countUninstall = count($uninstall);

        if ($countInstall == 0 && $countUninstall == 0) {
            return true;
        }

        if (($countInstall - $countUninstall) > 0) {
            $count = $this->getCountInstallLangsForMemory();
            if (($countInstall - $countUninstall) > $count) {
                $this->getWmsDb()->setSystemVariable(self::KEY_INSTALLED_LANGUAGES, json_encode($langsFromPbx));
                throw new Exception('No space left in memory.');
            }
        }

        $command  = '(nohup ';
        $command .= 'sudo /bin/mount -o remount,rw,async /rw2; ';
        $command .= 'sudo /bin/mount -o remount,async,rw /; ';

        foreach ($uninstall as $lang) {
            $pckg     = $this->getListPackagesSoundsForLang($lang);
            $command .= 'sudo DEBIAN_FRONTEND=noninteractive /usr/bin/apt-get -y -q purge ' . $pckg . '; ';
        }
        foreach ($install as $lang) {
            $pckg     = $this->getListPackagesSoundsForLang($lang);
            $command .= 'sudo DEBIAN_FRONTEND=noninteractive /usr/bin/apt-get -y --allow-unauthenticated -q install ' . $pckg . '; ';
        }

        $command .= 'sudo /usr/bin/apt-get clean; ';
        $command .= 'sudo /usr/bin/apt-get -y autoremove; ';
        $command .= 'sudo /bin/mount -o remount,rw,sync /rw2; ';
        $command .= "echo 'end_work_exec' ;";
        $command .= ')  > ' . self::LOG_INSTALL_FILE . ' 2>&1 &';

        $this->getLog()->info(__LINE__, 'Command install packs: ' . $command);
        exec($command);
        return true;
    }

    public function getListPackagesSoundsForLang($lang) {
        return 'wildix-sounds-' . $lang;
    }

    public function getStatus() {
        if (file_exists(self::LOG_INSTALL_FILE)) {
            $log = file_get_contents(self::LOG_INSTALL_FILE);
            if (strpos($log, 'end_work_exec') === false) {
                return true;
            }
        }
        return false;
    }

    public function getCountInstallLangsForMemory() {
        // size one lang(4 packs) ~ 5.5M
        $usg = disk_free_space('/rw2');
        return intval($usg / 1024 / 1024 / 5.5);
    }

    private function _findSysVarInstalledLanguages() {
        return $this->getWmsDb()->getSystemVariable(self::KEY_INSTALLED_LANGUAGES);
    }
}
