<?php
namespace Wildix\Ldap;

use Wildix\Ldap\Exception\ConnectException;
use Wildix\Ldap\Exception\BindException;

/**
 * Ldap connector
 */
class Connector {

    private $_host     = null;
    private $_port     = null;
    private $_secure   = null;
    private $_user     = null;
    private $_password = null;

    /**
     * Ldap Connector constructor
     *
     * @param string  $host     Ldap host or ip.
     * @param string  $port     Ldap port, can be empty.
     * @param boolean $secure   Use ldaps protocol or no.
     * @param string  $user     Ldap user, can be empty.
     * @param string  $password Ldap password.
     */
    public function __construct($host, $port, $secure, $user = null, $password = null) {
        $this->_host     = $host;
        $this->_port     = $port;
        $this->_secure   = $secure;
        $this->_user     = $user;
        $this->_password = $password;
    }

    public function getConnectString() {
        $conStr  = $this->_secure ? 'ldaps://' : 'ldap://';
        $conStr .= $this->_host;
        $conStr .= $this->_port ? ':' . $this->_port : '';
        return $conStr;
    }

    /**
     * Retrieve resource of ldap connection, will do also bind in
     * case "user" was passed to constructor
     *
     * @return resource
     * @throws BindException    throws in real connection problems or invalid user/password
     * @throws ConnectException throws if connect string cannot be parsed
     */
    public function getConnection() {
        // Always skip checking untrasted certificates - look comment in WMS-4510 for details
        // Available since 7.0.0, need to be global due to per connection will
        // reuse previous context if lunched via php-fmp and LDAP_OPT_X_TLS_NEWCTX - not implemented
        ldap_set_option(null, LDAP_OPT_X_TLS_REQUIRE_CERT, LDAP_OPT_X_TLS_ALLOW);

        $connect = ldap_connect($this->getConnectString());
        if (!is_resource($connect)) {
            throw new ConnectException('Ldap connection failed using: ' . $this->getConnectString());
        }

        ldap_set_option($connect, LDAP_OPT_PROTOCOL_VERSION, 3);
        ldap_set_option($connect, LDAP_OPT_REFERRALS, 0);
        ldap_set_option($connect, LDAP_OPT_NETWORK_TIMEOUT, 5);

        if ($this->_user != null) {
            if (!ldap_bind($connect, $this->_user, $this->_password)) {
                throw new BindException('Ldap bind failed: ' . ldap_error($connect), ldap_errno($connect));
            }
        }

        return $connect;
    }
}
