<?php
namespace Wildix;

/**
 * License Class
 */
class License extends \Singleton {

    const SMART_FILENAME          = '/rw2/etc/sf2';
    const ROSTER_MAX_USERS        = 600;
    const ROSTER_DIFF_FOR_WARNING = 50;
    const ROSTER_DEFAULT_SIZE     = 50;

    public const LICENSE_HOTEL      = 'hotel';
    public const LICENSE_CONFERENCE = 'cti_conference';

    public const TYPE_BASIC         = 'basic';
    public const TYPE_ESSENTIAL     = 'essential';
    public const TYPE_BUSINESS      = 'business';
    public const TYPE_PREMIUM       = 'premium';
    public const TYPE_WIZYCONF      = 'wizyconf';
    public const TYPE_PBX_SERVICE   = 'pbxService';
    public const TYPE_WIZYWEBINAR   = 'wizyWebinar';
    public const TYPE_XCARACAL      = 'x-caracal';
    public const TYPE_CLASSOUND     = 'classound';
    public const TYPE_MSPHONESYSTEM = 'msphonesystem';
    public const TYPE_XBEES         = 'x-bees';

    public const LIMITATION_LICENSE_TYPE    = 'licenseTypes';
    public const LIMITATION_NETWORK_LICENSE = 'networkLicense';

    public static $availableUserLicense = [
        self::TYPE_BASIC,
        self::TYPE_ESSENTIAL,
        self::TYPE_BUSINESS,
        self::TYPE_PREMIUM,
        self::TYPE_WIZYCONF,
        self::TYPE_PBX_SERVICE,
    ];
    const ERROR_LICENSE_NETWORK_INCORRECT = 'LICENSE_NETWORK_INCORRECT';

    private $_default = array(
                            'phones' => 0,
                            'trunks' => 0,
                            'calls' => 0,
                            'rooms' => 0,
                            self::LIMITATION_LICENSE_TYPE => [
                                self::TYPE_BASIC         => 0,
                                self::TYPE_ESSENTIAL     => 0,
                                self::TYPE_BUSINESS      => 0,
                                self::TYPE_PREMIUM       => 0,
                                self::TYPE_WIZYCONF      => 0,
                                self::TYPE_PBX_SERVICE   => 0,
                                self::TYPE_WIZYWEBINAR   => 0,
                                self::TYPE_XCARACAL      => 0,
                                self::TYPE_CLASSOUND     => 0,
                                self::TYPE_MSPHONESYSTEM => 0,
                            ],
                            'faxes' => array(
                                    'state' => false,
                                    'until' => '*'
                            ),
                            'redundancy' => array(
                                    'state' => false,
                                    'until' => '*'
                            ),
                            // 'state' means perUser (1) or perService (0). 'until' means subscription (license expiration date) or lifetime license (*)
                            'perUser' => array(
                                    'state' => false,
                                    'until' => '*'
                            ),
                            'support' => array(
                                    'state' => false,
                                    'until' => '*'
                            ),
                            'cloud' => array(
                                    'state' => false,
                                    'until' => '*'
                            ),
                            'cti_cdr_view' => array(
                                    'state' => false,
                                    'until' => '*'
                            ),
                            'cti_connect_pro' => array(
                                    'state' => false,
                                    'until' => '*'
                            ),
                            'cti_operator_panel' => array(
                                    'state' => false,
                                    'until' => '*'
                            ),
                            'hotel' => array(
                                    'state' => false,
                                    'until' => '*'
                            ),
                            'tapi' => array(
                                    'state' => false,
                                    'until' => '*'
                            ),
                            'cti_conference' => array(
                                    'state' => false,
                                    'until' => '*'
                            ),
                            'upgrade_2.0_3.0' => array(
                                    'state' => false,
                                    'until' => '*'
                            ),
                            'license_update' => array(
                                    'state' => false,
                                    'until' => '*'
                            ),
                            'token' => '',
                            'serial' => '',
                            'pbxKey' => '',
                            self::LIMITATION_NETWORK_LICENSE => 0,
                            self::TYPE_XBEES => 0,
                        );

    private $_logLevel = \Syslog::ERROR;

    private $_log = null;

    /**
     * WMS_ldap.
     *
     * @var \WMS_ldap
     */
    private $_ldap;

    /**
     * Maximum available license type.
     *
     * @var string
     */
    private $_maxAvailableLicenseType;
    /**
     * Info about user usage licenses.
     *
     * @var array
     */
    private $_licenseUsageData;

    private string $_failedMessage;

    public function getUpdateHasFailedMessage():string {
        return $this->_failedMessage ?? _t('license.failed.message');
    }

    private function setFailedMessage(string $message): void {
        $this->_failedMessage = $message;
    }

    /**
     * Have is error after callweaver try update
     *
     * @param array|null $output Get json after exec command.
     *
     * @return boolean
     */
    public function isLicenseUpdateHasFailed(array $output = null): bool {
        $cmd = "/usr/sbin/callweaver -rx 'license get_error'";
        exec($cmd, $output);

        $str = $output[0] ?? '';
        if (empty($str)) {
            return false;
        }

        $data = json_decode($str, true);

        $hasError = !empty($data['error']) || (json_last_error() !== JSON_ERROR_NONE);
        if ($hasError) {
            $this->getLog()->error(__LINE__, 'Checking license ' . $cmd . ' has failed result:' . json_encode($output));

            if (!empty($data['code']) && $data['code'] === self::ERROR_LICENSE_NETWORK_INCORRECT) {
                $this->setFailedMessage(_t('license.NetworkLicenseIncorrect'));
            }
        }

        return $hasError;
    }

    private function getLog() {
        if ($this->_log === null) {
            $this->_log = new \Syslog('\Wildix\License', $this->_logLevel);
        }
        return $this->_log;
    }

    /**
     * Returns available license types.
     *
     * @return array
     */
    public static function getLicenseTypes(): array {
        return [
            self::TYPE_BASIC,
            self::TYPE_ESSENTIAL,
            self::TYPE_BUSINESS,
            self::TYPE_PREMIUM,
            self::TYPE_WIZYCONF,
            self::TYPE_PBX_SERVICE,
            self::TYPE_WIZYWEBINAR,
            self::TYPE_XCARACAL,
            self::TYPE_CLASSOUND,
            self::TYPE_MSPHONESYSTEM,
        ];
    }

    /**
     * Returns default license type.
     *
     * @return string
     */
    public static function getDefaultLicenseType(): string {
        return self::TYPE_ESSENTIAL;
    }

    /**
     * Set ldap.
     *
     * @param \WMS_Ldap $ldap WMS_Ldap.
     *
     * @return void
     */
    public function setWmsLdap(\WMS_Ldap $ldap): void {
        $this->_ldap = $ldap;
    }

    /**
     * Have License
     *
     * @param string $licenseName Name.
     *
     * @return boolean
     */
    public function haveLicense($licenseName) {
        $licensesList = [
            self::TYPE_WIZYWEBINAR => self::TYPE_WIZYWEBINAR,
            self::TYPE_XCARACAL    => self::TYPE_XCARACAL,
            self::TYPE_CLASSOUND   => self::TYPE_CLASSOUND,
        ];

        if ($this->isActivated()) {
            if (in_array($licenseName, $licensesList)) {
                $limitation = $this->getLimitations(self::LIMITATION_LICENSE_TYPE);

                return (isset($limitation[$licensesList[$licenseName]]) && $limitation[$licensesList[$licenseName]] > 0);
            } else {
                $param = $this->getLimitations($licenseName);
                if ($licenseName === self::TYPE_XBEES) {
                    return $param === 1;
                }
                if ($param['state'] === true) {
                    return true;
                }
            }
        }
        return false;
    }


    private $_oldLimitations = null;
    private $_curMtime       = null;
    private $_startMtime     = null;
    private $_existFile      = null;
    private function startMonitoring() {
        $this->_existFile = file_exists(self::SMART_FILENAME);
        if ($this->_existFile) {
            clearstatcache(false, self::SMART_FILENAME);
            $this->_curMtime = filemtime(self::SMART_FILENAME);
        } else {
            $this->_curMtime = time() - 1 * 60 * 60;
        }

        $this->_startMtime = $this->_curMtime;

        $this->_oldLimitations = $this->getLimitations();

        // phpcs:disable Squiz.PHP.DiscouragedFunctions.Discouraged
        $this->getLog()->info(__LINE__, 'oldLimitations: ' . print_r($this->_oldLimitations, true));
        // phpcs:enable
    }

    private $_newLimitations = null;
    private function waitChangeLicense() {
        $sleepSeconds = 1 / 4;
        $totalSteps   = 40;
        $step         = 0;

        $this->getLog()->info(__LINE__, 'waitChangeLicense: start');
        do {
            $step++;
            usleep($sleepSeconds * 1000000);

            $this->getLog()->info(__LINE__, 'waitChangeLicense: step ' . $step . ' ');

            if (file_exists(self::SMART_FILENAME)) {
                clearstatcache(false, self::SMART_FILENAME);
                $this->_curMtime = filemtime(self::SMART_FILENAME);

                if ($this->_curMtime != $this->_startMtime) {
                    $this->getLog()->info(__LINE__, 'waitChangeLicense: file changed');
                }
            } else {
                if ($this->_existFile) {
                    $this->getLog()->info(__LINE__, 'waitChangeLicense: file removed');
                    break;
                }
            }
        } while ($step < $totalSteps && $this->_curMtime == $this->_startMtime);

        $this->getLog()->info(__LINE__, 'waitChangeLicense: end');

        $this->_resetCache();
        $this->_newLimitations = $this->getLimitations();

        // phpcs:disable Squiz.PHP.DiscouragedFunctions.Discouraged
        $this->getLog()->info(__LINE__, 'newLimitations: ' . print_r($this->_newLimitations, true));
        // phpcs:enable
        $diffLimitations = array_diff_assoc_recursive($this->_newLimitations, $this->_oldLimitations);
        return $diffLimitations;
    }

    public function reloadLicense() {
        $this->getLog()->info(__LINE__, 'Reload license');

        $this->startMonitoring();

        exec("/usr/sbin/callweaver -rx 'license reload' > /dev/null &", $return, $ret);

        $diffLimitations = $this->waitChangeLicense();

        return true;
    }

    public function deactivateLicense() {
        $this->getLog()->info(__LINE__, 'deactivate the license');

        $this->startMonitoring();

        exec("/usr/sbin/callweaver -rx 'license deactivate' > /dev/null &", $output);

        $diffLimitations = $this->waitChangeLicense();

        return true;
    }

    public function activateByKey(string $key) {
        $this->getLog()->info(__LINE__, 'activate license by pbxKey');

        if (!$this->isValidLicenseCodesSerial($key)) {
            $this->getLog()->error(__LINE__, 'invalid pbx key upon activating');
            throw new Exception('Invalid pbx key');
        }

        $this->startMonitoring();

        $cmd = escapeshellcmd("/usr/sbin/callweaver -rx 'license activate pbxKey=" . $key . "' > /dev/null &");

        $this->getLog()->info(__LINE__, $cmd);

        exec($cmd, $output);

        $diffLimitations = $this->waitChangeLicense();

        return true;
    }

    public function activateByFile($file) {
        if (!is_array($file) || count($file) < 16 || count($file) > 100) {
            throw new Exception('Bad license file.', 500);
        }

        $this->getLog()->info(__LINE__, 'Activation by file');

        $this->startMonitoring();
        $tmpFile = self::SMART_FILENAME . '.tmp';
        file_put_contents($tmpFile, join('', $file));

        $cmd = "/usr/sbin/callweaver -rx 'license activate file=" . $tmpFile . "' > /dev/null &";

        $this->getLog()->info(__LINE__, $cmd);

        exec($cmd, $output);

        $diffLimitations = $this->waitChangeLicense();

        return true;
    }

    public function isCloud() {
        return $this->haveLicense('cloud');
    }

    public function isActivated() {
        return ($this->getLicenseStatus() == 'active' && $this->getLimitations('phones') > 0);
    }

    public function isBadResident() {
        return $this->getLicenseStatus() != 'active';
    }

    public function getActiveDay($license) {
        $license = $this->getLimitations($license);
        if ($license['until'] == '*') {
            return 365;
        }

        return floor(($license['until'] - time()) / 60 / 60 / 24);
    }


    /**
     * Parse license file.
     *
     * @param array|boolean $content License lines.
     *
     * @see block License File https://wildix.atlassian.net/wiki/spaces/PBX/pages/18357319/Licensing
     *
     * @return array
     */
    public function parseLicense($content): array {
        $parsedLicense = $this->_default;
        if (!empty($content)) {
            $parsedLicense['phones'] = trim($content[0]);

            $parsedLicense['trunks'] = floor(trim($content[0]) * 2 / 3);
            if ($parsedLicense['trunks'] == '666') {
                $parsedLicense['trunks'] = '665';
            }

            $parsedLicense['calls'] = trim($content[1]);
            $parsedLicense['rooms'] = trim($content[2]);
            if ($parsedLicense['rooms'] > 0) {
                $parsedLicense['hotel'] = array(
                        'state' => true,
                        'until' => '*'
                );
            }

            // global license all others licenses should be less than this
            $activation = $this->parseItem($content[14]);
            if ($activation['until'] != '*' && floor(($activation['until'] - time()) / 60 / 60 / 24) > 7) {
                $activation['until'] = '*';
            }

            $parsedLicense['faxes']      = $this->parseItem($content[3], $activation);
            $parsedLicense['redundancy'] = $this->parseItem($content[4], $activation);

            // there is a date (or *) in this line, so don't use date from 'activation' line
            $parsedLicense['perUser']            = $this->parseItem($content[5]);
            $parsedLicense['support']            = $this->parseItem($content[6], $activation);
            $parsedLicense['cloud']              = $this->parseItem($content[7], $activation);
            $parsedLicense['cti_cdr_view']       = $this->parseItem($content[8], $activation);
            $parsedLicense['cti_connect_pro']    = $this->parseItem($content[9], $activation);
            $parsedLicense['cti_operator_panel'] = $this->parseItem($content[10], $activation);
            $parsedLicense['tapi']               = $this->parseItem($content[11], $activation);
            $parsedLicense['cti_conference']     = $this->parseItem($content[12], $activation);
            $parsedLicense['upgrade_2.0_3.0']    = $this->parseItem($content[13]);
            $parsedLicense['license_update']     = $this->parseItem($content[14]);

            if (count($content) > 18) {
                // token is the reverse base64 encoded string
                $parsedLicense['token'] = trim($content[15]);

                $parsedLicense['serial'] = trim($content[16]);
                $parsedLicense['pbxKey'] = trim($content[17]);
            }

            /*
                * inversion for backward compatibility
                * if set true then not cloud pbx
                * if false then cloud pbx
            */

            $parsedLicense['cloud']['state'] = !$parsedLicense['cloud']['state'];

            if (!empty($content[18])) {
                $licenseTypes = explode('|', $content[18]);

                $parsedLicense[self::LIMITATION_LICENSE_TYPE][self::TYPE_BASIC]         = (int)($licenseTypes[0] ?? 0);
                $parsedLicense[self::LIMITATION_LICENSE_TYPE][self::TYPE_ESSENTIAL]     = (int)($licenseTypes[1] ?? 0);
                $parsedLicense[self::LIMITATION_LICENSE_TYPE][self::TYPE_BUSINESS]      = (int)($licenseTypes[2] ?? 0);
                $parsedLicense[self::LIMITATION_LICENSE_TYPE][self::TYPE_PREMIUM]       = (int)($licenseTypes[3] ?? 0);
                $parsedLicense[self::LIMITATION_LICENSE_TYPE][self::TYPE_WIZYCONF]      = (int)($licenseTypes[4] ?? 0);
                $parsedLicense[self::LIMITATION_LICENSE_TYPE][self::TYPE_PBX_SERVICE]   = (int)($licenseTypes[5] ?? 0);
                $parsedLicense[self::LIMITATION_LICENSE_TYPE][self::TYPE_WIZYWEBINAR]   = (int)($licenseTypes[6] ?? 0);
                $parsedLicense[self::LIMITATION_LICENSE_TYPE][self::TYPE_XCARACAL]      = (int)($licenseTypes[7] ?? 0);
                $parsedLicense[self::LIMITATION_LICENSE_TYPE][self::TYPE_MSPHONESYSTEM] = (int)($licenseTypes[8] ?? 0);
                // $licenseTypes 8 It's MS-PhoneSystem will be implemented in the future
                $parsedLicense[self::LIMITATION_LICENSE_TYPE][self::TYPE_CLASSOUND]     = (int)($licenseTypes[9] ?? 0);
            }

            if (!empty($content[19])) {
                $parsedLicense[self::LIMITATION_NETWORK_LICENSE] = (int)$content[19];
            }

            if (!empty($content[20])) {
                $parsedLicense[self::TYPE_XBEES] = (int)$content[20];
            }
        }

        return $parsedLicense;
    }

    private $_limitations = null;
    /**
     * Get Limitations
     *
     * @param string $limitationName Name.
     *
     * @return array|boolean|number
     */
    public function getLimitations($limitationName = null) {
        if ($this->_limitations === null) {
            $this->getLog()->info(__LINE__, 'get limitations from file');

            if (file_exists(self::SMART_FILENAME)) {
                $this->_limitations = $this->parseLicense(file(self::SMART_FILENAME));
            } else {
                $this->_limitations = $this->_default;
            }
        } else {
            $this->getLog()->info(__LINE__, 'get limitations from cache');
        }
        if ($limitationName && isset($this->_limitations[$limitationName])) {
            return $this->_limitations[$limitationName];
        }
        return $this->_limitations;
    }

    private function _resetCache() {
        $this->getLog()->info(__LINE__, 'resetCache: ');
        $this->_limitations   = null;
        $this->_licenseStatus = null;
    }

    private $_licenseStatus = null;
    /**
     * Get License Status
     *
     * @return string
     */
    private function getLicenseStatus() {
        if ($this->_licenseStatus === null) {
            $this->_licenseStatus = false;
            $this->getLog()->info(__LINE__, 'get license status');
            exec("/usr/sbin/callweaver -rx 'core waitfullybooted' > /dev/null 2>&1 && /usr/sbin/callweaver -rx 'license status'", $output);

            if (is_array($output) && isset($output[0])) {
                $this->_licenseStatus = trim($output[0]);
            }
        }

        return $this->_licenseStatus;
    }

    private function parseItem($license, $activation = null) {
        $result = array(
                'state' => false,
                'until' => '*'
        );

        $licenseDetail = explode('|', trim($license));

        if (isset($licenseDetail[0])) {
            if (isset($licenseDetail[1])) {
                $result['until'] = $licenseDetail[1];
            }
            if ($licenseDetail[0] == 1 && isset($licenseDetail[1]) && ($licenseDetail[1] == '*' || $licenseDetail[1] > time())) {
                $result['state'] = true;
            }
        }

        $licenseDetail = explode('|', trim($license));

        if (!empty($activation) && $activation['until'] != '*') {
            if ($activation['state'] == false) {
                $result['state'] = false;
            }
            if ($result['until'] == '*' || $activation['until'] < $result['until']) {
                $result['until'] = $activation['until'];
            }
        }

        return $result;
    }

    public function getMaxRosterUsers() {
        return self::ROSTER_MAX_USERS;
    }

    public function isPerUser() {
        $perUserSubscriptionLimitation = $this->getLimitations('perUser');
        if ($perUserSubscriptionLimitation['state'] == 1) {
            return true;
        } else {
            return false;
        }
    }

    public function isPerService() {
        $perUserSubscriptionLimitation = $this->getLimitations('perUser');
        if ($perUserSubscriptionLimitation['state'] == 0) {
            return true;
        } else {
            return false;
        }
    }

    public function isSubscription() {
        $perUserSubscriptionLimitation = $this->getLimitations('perUser');
        if ($perUserSubscriptionLimitation['until'] != '*') {
            return true;
        } else {
            return false;
        }
    }

    public function isLifetime() {
        $perUserSubscriptionLimitation = $this->getLimitations('perUser');
        if ($perUserSubscriptionLimitation['until'] == '*') {
            return true;
        } else {
            return false;
        }
    }

    public function isBasic(): bool {
        return $this->isAvailableLicenseType(self::TYPE_BASIC);
    }

    public function isEssential(): bool {
        return $this->isAvailableLicenseType(self::TYPE_ESSENTIAL);
    }

    public function isBusiness(): bool {
        return $this->isAvailableLicenseType(self::TYPE_BUSINESS);
    }

    public function isPremium(): bool {
        return $this->isAvailableLicenseType(self::TYPE_PREMIUM);
    }

    /**
     * Returns user license usage.
     *
     * @return array
     */
    public function getUserLicenseUsage(): array {
        if ($this->_licenseUsageData === null) {
            $this->_licenseUsageData = [
                self::TYPE_BASIC         => $this->getLicenseTypeCount(self::TYPE_BASIC),
                self::TYPE_ESSENTIAL     => $this->getLicenseTypeCount(self::TYPE_ESSENTIAL),
                self::TYPE_BUSINESS      => $this->getLicenseTypeCount(self::TYPE_BUSINESS),
                self::TYPE_PREMIUM       => $this->getLicenseTypeCount(self::TYPE_PREMIUM),
                self::TYPE_WIZYCONF      => $this->getLicenseTypeCount(self::TYPE_WIZYCONF),
                self::TYPE_PBX_SERVICE   => $this->getLicenseTypeCount(self::TYPE_PBX_SERVICE),
                self::TYPE_WIZYWEBINAR   => $this->getLicenseTypeCount(self::TYPE_WIZYWEBINAR),
                self::TYPE_XCARACAL      => $this->getLicenseTypeCount(self::TYPE_XCARACAL),
                self::TYPE_MSPHONESYSTEM => $this->getLicenseTypeCount(self::TYPE_MSPHONESYSTEM),
                self::TYPE_CLASSOUND     => $this->getLicenseTypeCount(self::TYPE_CLASSOUND),
            ];
        }

        return $this->_licenseUsageData;
    }

    /**
     * Check if license limit exceeded.
     *
     * @return boolean
     */
    public function isLicenseLimitExceeded(): bool {
        $licenseAvailable = $this->getLimitations(self::LIMITATION_LICENSE_TYPE);
        $licenseUsed      = $this->getUserLicenseUsage();

        return ($licenseUsed[self::TYPE_BASIC] > $licenseAvailable[self::TYPE_BASIC]
            || $licenseUsed[self::TYPE_ESSENTIAL] > $licenseAvailable[self::TYPE_ESSENTIAL]
            || $licenseUsed[self::TYPE_BUSINESS] > $licenseAvailable[self::TYPE_BUSINESS]
            || $licenseUsed[self::TYPE_PREMIUM] > $licenseAvailable[self::TYPE_PREMIUM]
            || $licenseUsed[self::TYPE_WIZYCONF] > $licenseAvailable[self::TYPE_WIZYCONF]
            || $licenseUsed[self::TYPE_PBX_SERVICE] > $licenseAvailable[self::TYPE_PBX_SERVICE]
            || $licenseUsed[self::TYPE_WIZYWEBINAR] > $licenseAvailable[self::TYPE_WIZYWEBINAR]
            || $licenseUsed[self::TYPE_XCARACAL] > $licenseAvailable[self::TYPE_XCARACAL]
            || $licenseUsed[self::TYPE_CLASSOUND] > $licenseAvailable[self::TYPE_CLASSOUND]
        );
    }

    /**
     * Returns list of license with info about available license types count and used licenses count.
     *
     * @return array
     */
    public function getLicensesUsageInfo(): array {
        $licenseAvailable = $this->getLimitations(self::LIMITATION_LICENSE_TYPE);
        $licenseUsed      = $this->getUserLicenseUsage();
        $result           = [];

        foreach (self::getLicenseTypes() as $licenseType) {
            if (isset($licenseUsed[$licenseType]) && isset($licenseAvailable[$licenseType])) {
                $result[$licenseType] = [
                    'used' => $licenseUsed[$licenseType],
                    'available' => $licenseAvailable[$licenseType],
                ];
            }
        }

        return $result;
    }

    /**
     * Returns maximum available license type.
     * The default license is "essential".
     *
     * @return string
     */
    public function getMaximumAvailableLicenseType(): string {
        if ($this->_maxAvailableLicenseType === null) {
            if ($this->isPremium()) {
                $this->_maxAvailableLicenseType = self::TYPE_PREMIUM;
            } else if ($this->isBusiness()) {
                $this->_maxAvailableLicenseType = self::TYPE_BUSINESS;
            } else if ($this->isEssential()) {
                $this->_maxAvailableLicenseType = self::TYPE_ESSENTIAL;
            } else if ($this->isBasic()) {
                $this->_maxAvailableLicenseType = self::TYPE_BASIC;
            } else {
                $this->_maxAvailableLicenseType = self::getDefaultLicenseType();
            }
        }

        return $this->_maxAvailableLicenseType;
    }

    /**
     * Returns count used licenses type.
     *
     * @param string $licenseType License type.
     *
     * @return integer
     */
    private function getLicenseTypeCount(string $licenseType): int {
        $networkLicense = $this->getLimitations(self::LIMITATION_NETWORK_LICENSE);
        $filter         = '(licenseType=' . $licenseType . ')';

        // consider users without a license type as the maximum available license
        if ($this->getMaximumAvailableLicenseType() === $licenseType) {
            $filter = '(|(!(licenseType=*))' . $filter . ')';
        }

        $filter .= '(!(cn=admin))';

        $currentPbxDn = $this->getCurrentPbxDn();
        if ($networkLicense === 0 && $currentPbxDn) {
            $filter .= '(pbxowner=' . $currentPbxDn . ')';
        }

        return $this->_ldap->getCount(\WMS_Ldap::PERSONS, '(&' . $filter . ')');
    }

    /**
     * Return LDAP DN of current PBX
     *
     * @return string
     */
    private function getCurrentPbxDn(): ?string {
        $serial     = (string)$this->getLimitations('serial');
        $currentPBX = $this->_ldap->getEntry(\WMS_Ldap::PBX, '(pbxSerial=' . $serial . ')', ['dn']);

        return $currentPBX['dn'] ?? null;
    }

    /**
     * Checks if license type is available.
     *
     * @param string $licenseType License type.
     *
     * @return boolean
     */
    private function isAvailableLicenseType(string $licenseType): bool {
        return ($this->getLimitations(self::LIMITATION_LICENSE_TYPE)[$licenseType] > 0);
    }

    private function isValidLicenseCodesSerial(string $data) {
        return preg_match('/^[a-z0-9\-]+$/ui', $data);
    }

    public function getLimitExceededMessage(): string {
        $licensesUsageInfo = array_intersect_key($this->getLicensesUsageInfo(), array_flip(self::$availableUserLicense));
        $licensesUsageList = [];
        foreach ($licensesUsageInfo as $licenseType => $data) {
            $licensesUsageList[] = sprintf('%s: %d / %d', _t('license.type.' . $licenseType), $data['used'], $data['available']);
        }

        return _t('license.licenseLimitExceeded') . '.<br>' . implode('<br>', $licensesUsageList);
    }
}
