<?php

namespace Wildix\Microsoft;

use Beta\Microsoft\Graph\Model as BetaModel;
use Cache\FileCache;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\GuzzleException;
use League\OAuth2\Client\Provider\GenericProvider;
use Logger\Log;
use Microsoft\Graph\Exception\GraphException;
use Microsoft\Graph\Graph;
use Microsoft\Graph\Http\GraphCollectionRequest;
use Microsoft\Graph\Model\Event;
use Microsoft\Graph\Model\EventType;
use Microsoft\Graph\Model\MailboxSettings;
use Microsoft\Graph\Model\Sensitivity;
use Microsoft\Graph\Model\TimeZoneBase;
use Microsoft\Graph\Model\User;
use Microsoft\TimeZones\TimeZones;
use Services\Helpers;
use Throwable;
use UnexpectedValueException;
use Wildix\Calendar\CalendarService;
use Wildix\Calendar\Events;
use Wildix\Calendar\Settings;
use Wildix\Colleagues\Colleague;
use Wildix\Colleagues\Colleagues;
use Wildix\Databases\WMSDB;
use Wildix\Exception\SystemdException;
use Wildix\Phonebooks\Phonebook;
use Wildix\Phonebooks\PhonebooksService;
use Wildix\Phonebooks\Sync\SyncSettings;
use Wildix\Services\Kite;
use Wildix\System\Systemd;
use Wildix\TaskQueue\QueueManager;

/**
 * Class Office365Service
 */
final class Office365Service {
    public const STATUS  = 'INTEGRATION_OFFICE365';
    public const TENANT  = 'INTEGRATION_OFFICE365_TENANT';
    public const MODULES = 'INTEGRATION_OFFICE365_MODULES';

    public const MODULE_CALENDAR  = 'calendar';
    public const MODULE_PHONEBOOK = 'phonebook';

    // Full resync personal phonebooks every day
    private const FULL_RESYNC_PHONEBOOKS_EVERY_MINUTES = 60 * 24;

    private const USER_SETTINGS_LIST = [
        'workHoursStart',
        'workHoursEnd',
        'workDays',
    ];

    private const STATUS_MAP = [
        'free' => 'Free',
        'tentative' => 'Tentative',
        'busy' => 'Busy',
        'oof' => 'OutOfOffice',
        'workingElsewhere' => 'WorkingElsewhere',
        'unknown' => 'Free',
    ];

    private $_storage;
    private $_cache;
    private $_kiteService;
    private $_baseUri;
    private $_calendarService;
    private $_phonebooksService;
    private $_systemdCalendarsTimer;
    private $_systemdPhonebooksTimer;
    private $_logger;

    /**
     * Office365Service constructor.
     *
     * @param WMSDB             $storage                Configuration storage.
     * @param FileCache         $cache                  Cache implementation.
     * @param Kite              $kiteService            Kite Service.
     * @param string            $baseUri                Base client url.
     * @param CalendarService   $calendarService        Calendar Service.
     * @param PhonebooksService $phonebooksService      Phonebooks Service.
     * @param Systemd           $systemdCalendarsTimer  Systemd Calendar Events Service.
     * @param Systemd           $systemdPhonebooksTimer Systemd Phonebooks Service.
     * @param Log               $logger                 Logger instance.
     */
    public function __construct(
        WMSDB $storage,
        FileCache $cache,
        Kite $kiteService,
        string $baseUri,
        CalendarService $calendarService,
        PhonebooksService $phonebooksService,
        Systemd $systemdCalendarsTimer,
        Systemd $systemdPhonebooksTimer,
        Log $logger
    ) {
        $this->_storage                = $storage;
        $this->_cache                  = $cache;
        $this->_kiteService            = $kiteService;
        $this->_baseUri                = $baseUri;
        $this->_calendarService        = $calendarService;
        $this->_phonebooksService      = $phonebooksService;
        $this->_systemdCalendarsTimer  = $systemdCalendarsTimer;
        $this->_systemdPhonebooksTimer = $systemdPhonebooksTimer;
        $this->_logger                 = $logger;
    }

    public function isEnabledCalendarIntegration(): bool {
        return filter_var($this->_storage->getSystemVariable(self::STATUS), FILTER_VALIDATE_BOOLEAN)
            && $this->_systemdCalendarsTimer->isEnabled()
            && $this->_systemdCalendarsTimer->isActive()
            && in_array(self::MODULE_CALENDAR, $this->enabledModules());
    }

    public function isEnabledPhonebookIntegration(): bool {
        return filter_var($this->_storage->getSystemVariable(self::STATUS), FILTER_VALIDATE_BOOLEAN)
            && $this->_systemdPhonebooksTimer->isEnabled()
            && $this->_systemdPhonebooksTimer->isActive()
            && in_array(self::MODULE_PHONEBOOK, $this->enabledModules());
    }

    public function enabledModules() {
        $modules = $this->_storage->getSystemVariable(self::MODULES);
        if ($modules) {
            $modules = explode(',', $modules);
        } else {
            $modules = [];
        }

        return $modules;
    }

    /**
     * Enable integration Office365
     *
     * @param string $tenant  User Identifier for office 365.
     * @param array  $modules Office365 modules (calendar or phonebook).
     *
     * @throws SystemdException Error enabled system timer.
     *
     * @return void
     */
    public function enableIntegration(string $tenant, array $modules): void {
        $this->_storage->setSystemVariable(self::STATUS, true);
        $this->_storage->setSystemVariable(self::TENANT, $tenant);
        $this->_storage->setSystemVariable(self::MODULES, implode(',', $modules));
        try {
            if (in_array(self::MODULE_CALENDAR, $modules, true)) {
                $this->_systemdCalendarsTimer->enable();
                $this->_systemdCalendarsTimer->start();
            }
            if (in_array(self::MODULE_PHONEBOOK, $modules, true)) {
                $this->_systemdPhonebooksTimer->enable();
                $this->_systemdPhonebooksTimer->start();
            }
        } catch (SystemdException $systemdException) {
            $this->_logger->error(__LINE__, 'Microsoft365 Integration has not been enabled');
            $this->_logger->error(__LINE__, 'Systemd output:' . json_encode($systemdException->getSystemOutput()));
            throw $systemdException;
        }
        $this->_logger->info(__LINE__, 'Microsoft365 Integration has been enabled');
    }

    public function disableIntegration(): void {
        $this->_storage->setSystemVariable(self::STATUS, false);
        $this->_storage->setSystemVariable(self::TENANT, '');
        $this->_storage->setSystemVariable(self::MODULES, '');

        $this->_systemdCalendarsTimer->stop();
        $this->_systemdCalendarsTimer->disable();

        $this->_systemdPhonebooksTimer->stop();
        $this->_systemdPhonebooksTimer->disable();

        $this->_logger->info(__LINE__, 'Microsoft365 Integration has been disabled');
    }

    public function syncCalendars(): void {
        if ($this->isEnabledCalendarIntegration() === false) {
            $this->_logger->info(__LINE__, 'Sync Microsoft365 calendar events is disabled');
            return;
        }

        $office365Users = $this->_fetchOfficeUsers();
        $localUsers     = $this->_fetchLdapUsers($office365Users);
        // todo import repo
        $settingsRepo = Settings::getInstance();
        foreach ($localUsers as $user) {
            try {
                // update user settings if changed or new
                $userSettings  = $settingsRepo->getSettings(Settings::SETTINGS_APP_OFFICE365_OUTLOOK, $user);
                $updateResults = $this->_updateUserSettings($user, $userSettings);

                // update user events if changed or new or deleted
                $userEvents = $this->_getCalendarView($user, $userSettings);
            } catch (ClientException $exception) {
                $this->_logger->warning(__LINE__, sprintf('Error occurred during processing user "%s" events: %s', $user->getName(), $exception->getMessage()));
                $userEvents = [];
            } catch (\Exception $exception) {
                $this->_logger->error(__LINE__, sprintf('Error %s with trace %s', $exception->getMessage(), $exception->getTraceAsString()));
            }

            // todo omit it
            if (count($userEvents) === 0) {
                continue;
            }

            // todo move store into catch
            $this->_storeEvents($user, $userEvents);
        }
    }

    public function syncPhonebooks($fullSyncForce = false): void {
        if ($this->isEnabledPhonebookIntegration() === false) {
            $this->_logger->info(__LINE__, '[office365phonebooks] Sync Microsoft365 personal phonebooks is disabled');
            return;
        }

        $cacheKey         = 'office365syncPhonebooks';
        $lastFullSyncTime = $this->_cache->get($cacheKey);
        $fullSync         = ($lastFullSyncTime === false || $fullSyncForce);

        $this->_logger->info(__LINE__, '[office365phonebooks] Start sync Microsoft365 personal phonebooks. Mode Full=' . (int)$fullSync);

        $office365Users = $this->_fetchOfficeUsers();
        $localUsers     = $this->_fetchLdapUsers($office365Users);

        foreach ($localUsers as $user) {
            Helpers::addTask(
                'office365PhonebookSync',
                [
                    'userId' => $user->getID(),
                    'fullSync' => $fullSync
                ]
            );
        }

        if ($fullSync) {
            $this->_cache->set($cacheKey, ['time' => time()], self::FULL_RESYNC_PHONEBOOKS_EVERY_MINUTES * 60);
            Helpers::addTask('office365PhonebookCleaner');
        }

        QueueManager::runQueueManager();

        $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Finished add tasks for sync Microsoft365 personal phonebooks. Tasks count %s', count($localUsers)));
    }

    public function processUserPhonebooks(Colleague $user, bool $fullSync): void {
        $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Get contact folders for user %s msId %s', $user->getName(), $user->getSourceId()));
        try {
            $contactFoldersCollection = $this->_fetchContactFolders($user->getSourceId());
        } catch (\Exception $e) {
            $this->_logger->error(__LINE__, sprintf('[office365phonebooks] Error received contact folders for user %s: %s', $user->getName(), $e->getMessage()));
            return;
        }

        $contactFolders = $contactFoldersCollection->getItems();
        $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Found %s contact folders for user %s', count($contactFolders), $user->getName()));
        foreach ($contactFolders as $contactFolder) {
            try {
                $phonebook = $this->_phonebooksService->processContactFolder($contactFolder, $user);
                $this->_phonebooksService->initPhonebook($phonebook);

                $syncSettings = new SyncSettings($phonebook->getSyncParams());
                if ($fullSync === false && $syncSettings->getBackend() === 'office365' && !empty($syncSettings->getDeltaLink())) {
                    $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Delta mode for phonebook "%s"', $phonebook->getName()));
                    $contactsIterator = $this->_syncContactsDeltaMode($phonebook, $syncSettings);
                } else {
                    $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Full mode for phonebook "%s"', $phonebook->getName()));
                    $contactsIterator = $this->_syncContactsFullMode($phonebook, $syncSettings, $user, $contactFolder);
                }

                $this->_phonebooksService->saveContactFolderDeltaLink($phonebook, $contactsIterator->getDeltaLink());
            } catch (\Exception $e) {
                $this->_logger->error(__LINE__, $e->getMessage());
            }
        }
    }

    private function _syncContactsDeltaMode(
        Phonebook $phonebook,
        SyncSettings $syncSettings
    ): GraphCollectionRequest {
        $contactsIterator = $this->_getContactsIterator($syncSettings->getDeltaLink());

        while (!$contactsIterator->isEnd()) {
            $contacts = $contactsIterator->getPage();
            $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Delta mode %s %s', $phonebook->getName(), count($contacts)));
            $this->_phonebooksService->processContactsDelta($contacts, $phonebook, $syncSettings);
        }

        return $contactsIterator;
    }

    private function _syncContactsFullMode(
        Phonebook $phonebook,
        SyncSettings $syncSettings,
        Colleague $user,
        BetaModel\ContactFolder $contactFolder
    ): GraphCollectionRequest {
        $contactsIterator = $this->_getContactsIterator($this->_buildEndpointUrlForUser($user->getSourceId(), $contactFolder->getId()));

        while (!$contactsIterator->isEnd()) {
            $contacts = $contactsIterator->getPage();
            $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Full mode  %s %s', $phonebook->getName(), count($contacts)));
            $this->_phonebooksService->processContacts($contacts, $phonebook, $user, $syncSettings);
        }
        $this->_phonebooksService->deleteNotSyncedContacts();

        return $contactsIterator;
    }

    public function clearNotSyncedPhonebooks(): void {
        $allSyncedPhonebooks = $this->_phonebooksService->getPersonalPhonebooks();

        $office365Users = $this->_fetchOfficeUsers();
        $localUsers     = $this->_fetchLdapUsers($office365Users);

        $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Start delete not synced local phonebooks. All local phonebooks count %s', count($allSyncedPhonebooks)));
        foreach ($localUsers as $user) {
            try {
                $contactFoldersCollection = $this->_fetchContactFolders($user->getSourceId());
            } catch (\Exception $e) {
                continue;
            }

            $contactFolders = $contactFoldersCollection->getItems();
            foreach ($contactFolders as $contactFolder) {
                if (array_key_exists($contactFolder->getId(), $allSyncedPhonebooks)) {
                    unset($allSyncedPhonebooks[$contactFolder->getId()]);
                }
            }
        }

        $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Found %s local phonebooks for delete', count($allSyncedPhonebooks)));
        if (!empty($allSyncedPhonebooks)) {
            // need remove in the local db deleted phonebooks
            $this->_phonebooksService->removePhonebooks($allSyncedPhonebooks);
        }
    }

    private function _fetchContactFolders(string $userId): GraphCollection {
        $graph = $this->_getGraphClient(true);

        $queryParams = [
            '$select' => 'id,displayName',
            '$count' => 'true',
        ];
        $endpoint    = sprintf('/users/%s/contactFolders/delta?', $userId) . http_build_query($queryParams);

        $foldersCollectionIterator = $graph->createCollectionRequest('GET', $endpoint)
            ->setReturnType(BetaModel\ContactFolder::class);

        $contactFolders = [];
        while (!$foldersCollectionIterator->isEnd()) {
            $contactFolders[] = $foldersCollectionIterator->getPage();
        }

        return new GraphCollection(
            [
                GraphCollection::ITEMS => array_merge(...$contactFolders),
                GraphCollection::DELTA => $foldersCollectionIterator->getDeltaLink()
            ]
        );
    }

    private function _buildEndpointUrlForUser(string $userId, string $folderId = null): string {
        $queryParams = ['$count' => 'true'];

        if ($folderId !== null) {
            $endpoint = sprintf('/users/%s/contactFolders/%s/contacts/delta?', $userId, $folderId) . http_build_query($queryParams);
        } else {
            $endpoint = sprintf('/users/%s/contacts/delta?', $userId) . http_build_query($queryParams);
        }

        return $endpoint;
    }

    private function _getContactsIterator($endpoint): GraphCollectionRequest {
        $this->_logger->debug(__LINE__, sprintf('ENDPOINT %s', $endpoint));

        $graph = $this->_getGraphClient(true);
        $contactsCollectionIterator = $graph->createCollectionRequest('GET', $endpoint)
            ->setReturnType(BetaModel\Contact::class)
            ->addHeaders(['Prefer' => 'odata.maxpagesize=100']);

        return $contactsCollectionIterator;
    }

    /**
     * Return users from office365.
     *
     * @todo repo feature?
     *
     * @throws GuzzleException If error occurred during request to API.
     * @throws GraphException  If some issue in Graph client.
     *
     * @return User[]
     */
    private function _fetchOfficeUsers(): array {
        $graph = $this->_getGraphClient();

        $queryParams = [
            '$filter' => 'accountEnabled eq true',
            '$select' => 'id,displayName,mail,userPrincipalName,accountEnabled,assignedLicenses',
        ];
        $endpoint    = '/users?' . http_build_query($queryParams);

        $userCollectionIterator = $graph->createCollectionRequest('GET', $endpoint)
            ->setReturnType(User::class);

        // fetch all users
        // filter valid mailbox or/and licence
        $users = [];
        while (!$userCollectionIterator->isEnd()) {
            $usersPage = $userCollectionIterator->getPage();
            $users[]   = $this->_filterUsersWithLicences($usersPage);
        }

        return array_merge(...$users);
    }

    /**
     * Return local users related to remote.
     *
     * @param User[] $users Users from office365 server.
     *
     * @return Colleague[]
     *
     * @throws \Exception If cannot fetch local ?
     */
    private function _fetchLdapUsers(array $users) {
        // map/filter with ldap
        $sourceIds = array_map(
            static function (User $user) {
                return $user->getId();
            },
            $users
        );

        return Colleagues::getInstance()->getUsers([Colleague::SOURCE_ID => $sourceIds]);
    }

    /**
     * Try filter not valid users.
     * Better fetch(fail if not active) mailboxSettings but it require additional request!
     *
     * @param User[] $users Office365 users.
     *
     * @todo repo feature?
     *
     * @return User[]
     */
    private function _filterUsersWithLicences(array $users): array {
        return array_filter(
            $users,
            static function (User $user) {
                return count($user->getAssignedLicenses() ?? []) > 0 && (string)$user->getMail() !== '';
            }
        );
    }

    private function _fetchUserMailboxSettings(Colleague $user): array {
        $graph    = $this->_getGraphClient();
        $endpoint = sprintf('/users/%s/mailboxSettings', $user->getSourceId());

        return $this->_convertToLocalSettings(
            $graph->createRequest('GET', $endpoint)
                ->setReturnType(MailboxSettings::class)
                ->execute()
        );
    }

    private function _getCalendarView($user, $settings) {
        $graph = $this->_getGraphClient();

        // if NO delta in calendar Settings - initial request
        // or settings is not set!
        // or new period of time.
        $timezone = new \DateTimeZone('UTC');
        $start    = new \DateTimeImmutable('now', $timezone);

        $newSettings = [];

        $startOfInitialSyncPeriod = new \DateTimeImmutable($settings['initialDateTime'], $timezone);
        $endOfInitialSyncPeriod   = $startOfInitialSyncPeriod->modify('+1 day');

        $isInitial = !isset($settings['deltaLink'], $settings['initialDateTime'])
            || (string)$settings['deltaLink'] === ''
            || ($endOfInitialSyncPeriod <= $start);
        if ($isInitial) {
            $end = new \DateTimeImmutable('now +1 day +1 hour', $timezone);

            $queryParams = [
                'startDateTime' => $start->format(\DateTimeInterface::ATOM),
                'endDateTime' => $end->format(\DateTimeInterface::ATOM),
            ];

            // TODO timestamp?
            $newSettings['initialDateTime'] = $start->format(\DateTimeInterface::ATOM);

            $getEventsUrl = sprintf('/users/%s/calendar/calendarView/delta?%s', $user->getSourceId(), http_build_query($queryParams));
        } else {
            // delta
            $getEventsUrl = $settings['deltaLink'];
        }

        $calendarViewIterator = $graph->createCollectionRequest('GET', $getEventsUrl)
            ->setReturnType(Event::class);

        $events = [];
        while (!$calendarViewIterator->isEnd()) {
            $events[] = $calendarViewIterator->getPage();
        }

        $events = array_merge(...$events);

        // if not enough data in occurrence merge from seriesMaster
        array_walk(
            $events,
            static function (Event $event, $index, $events) {
                if ($event->getProperties()['@removed'] ?? false) {
                    return;
                }

                if ($event->getType()->is(EventType::OCCURRENCE)) {
                    foreach ($events as $originEvent) {
                        // find main event seriesMaster for this occurrence
                        if ($originEvent->getId() === $event->getSeriesMasterId()) {
                            $event->setOrganizer($originEvent->getOrganizer());
                            $event->setSubject($originEvent->getSubject());
                            $event->setShowAs($originEvent->getShowAs());
                            $event->setImportance($originEvent->getImportance());
                            $event->setSensitivity($originEvent->getSensitivity());
                        }
                    }
                }
            },
            $events
        );

        // remove seriesMaster / use only occurrence
        $events = array_filter(
            $events,
            static function (Event $event) {
                // left removed to process later?
                if ($event->getProperties()['@removed'] ?? false) {
                    return true;
                }

                return !$event->getType()->is(EventType::SERIES_MASTER);
            }
        );

        // new delta link after each fetch
        $newSettings['deltaLink'] = $calendarViewIterator->getDeltaLink();
        $this->_calendarService->updateSettings([], $newSettings, $user, Settings::SETTINGS_APP_OFFICE365_OUTLOOK);

        return $events;
    }

    private function _storeEvents(\Wildix\Colleagues\Colleague $user, array $events) {
        // map/convert events from office365 to local
        $convertedEvents = array_map([$this, '_convertToLocalEvent'], $events);

        // filter removed or canceled
        $removedEvents = array_filter(
            $convertedEvents,
            static function (array $event) {
                return (array_key_exists('removed', $event) && $event['removed'] === true) || array_key_exists('canceled', $event);
            }
        );
        if (count($removedEvents) > 0) {
            $removed = $this->_calendarService->deleteEvents(array_column($removedEvents, 'id'), $user);
        }

        $addOrUpdateEvents = array_filter(
            $convertedEvents,
            static function (array $event) {
                return ! (array_key_exists('removed', $event) || array_key_exists('canceled', $event));
            }
        );

        $userEvents = Events::getInstance()->getUserEvents($user, false, Events::EVENT_APP_OFFICE365_OUTLOOK);
        $updatedIds = $this->_calendarService->updateEvents($userEvents, $addOrUpdateEvents, $user, Events::EVENT_APP_OFFICE365_OUTLOOK, true);
    }

    private function _convertToLocalEvent(Event $event) {
        if ($event->getProperties()['@removed'] ?? false) {
            // all what we have for removed event
            return [
                'id' => $event->getId(),
                // better idea ?!
                'removed' => true,
            ];
        }

        if ($event->getIsCancelled()) {
            return [
                'id' => $event->getId(),
                // better idea ?!
                'canceled' => true,
            ];
        }

        $timezone = new \DateTimeZone('UTC');

        $dateStart = new \DateTime($event->getStart()->getDateTime(), $timezone);
        $dateEnd   = new \DateTime($event->getEnd()->getDateTime(), $timezone);

        return [
            'id' => $event->getId(),
            'name' => $event->getSubject(),
            'dateStart' => $dateStart->format('Y-m-d\TH:i:00\Z'),
            'dateEnd' => $dateEnd->format('Y-m-d\TH:i:00\Z'),
            'status' => self::STATUS_MAP[$event->getShowAs()->value()],
            'importance' => ucfirst($event->getImportance()->value()),
            'private' => (int)$event->getSensitivity()->is(Sensitivity::GRAPHPRIVATE),
            'options' => '',
            'type' => Events::EVENT_TYPE_CALENDAR,
            'app' => Events::EVENT_APP_OFFICE365_OUTLOOK,
        ];
    }

    /**
     * Get Microsoft graph client
     *
     * @param boolean $useBeta Use beta version API.
     *
     * @return Graph
     *
     * @todo use factory to build graph client
     */
    private function _getGraphClient($useBeta = false): Graph {
        // Get the access token from the cache
        $accessToken = $this->_getAccessToken();
        $this->_logger->debug(__LINE__, sprintf('TOKEN %s', $accessToken));

        // Create a Graph client
        $graph = new Graph();
        $graph->setAccessToken($accessToken);

        if ($useBeta) {
            $graph->setApiVersion('beta');
        }

        return $graph;
    }

    private function _getAccessToken() {
        $cachedAccessToken = $this->_cache->get($this->_getAccessTokenCacheKey());
        if ($cachedAccessToken === false) {
            $tenant      = $this->_storage->getSystemVariable(self::TENANT);
            $oauthClient = $this->_getOauth2Client($tenant);
            $accessToken = $oauthClient->getAccessToken('client_credentials', ['scope' => 'https://graph.microsoft.com/.default']);

            $cachedAccessToken = $accessToken->getToken();

            // reduce cache time for 10 minutes for time correction & 10 minutes gap between background job run
            $ttl = $accessToken->getExpires() - time() - (10 * 60);
            $this->_cache->set($this->_getAccessTokenCacheKey(), $cachedAccessToken, $ttl);
        }

        return $cachedAccessToken;
    }

    private function _getAccessTokenCacheKey() {
        return 'microsoft_graph_access_token_client_credentials';
    }

    /**
     * Get oauth2 client.
     *
     * @param string $tenant Microsoft tenant id or name.
     *
     * @return GenericProvider
     *
     * @throws UnexpectedValueException If tenant is empty.
     *
     * @todo use factory to build oauth2 client
     */
    private function _getOauth2Client(string $tenant): GenericProvider {
        if ($tenant === '') {
            throw new UnexpectedValueException('Tenant is not defined');
        }

        $office365AppConfig = $this->_kiteService->getOffice365IntegrationApp();

        return new GenericProvider(
            [
                'clientId'                => $office365AppConfig['clientId'] ?? '',
                'clientSecret'            => $office365AppConfig['clientSecret'] ?? '',
                // need? not user for daemon flow
                'redirectUri'             => $office365AppConfig['redirectUri'] ?? '',
                // todo use config for urls
                'urlAuthorize'            => sprintf('%s/%s%s', $this->_baseUri, $tenant, '/oauth2/v2.0/authorize'),
                'urlAccessToken'          => sprintf('%s/%s%s', $this->_baseUri, $tenant, '/oauth2/v2.0/token'),
                'urlResourceOwnerDetails' => '',
            ]
        );
    }

    private function _convertToLocalSettings(MailboxSettings $userMailboxSettings): array {
        $settings = [];
        try {
            $workingHours = $userMailboxSettings->getWorkingHours();
            if ($workingHours === null || !isset($workingHours->getProperties()['startTime'])) {
                $this->_logger->debug(__LINE__, sprintf('Working hours are not specified in the mailbox settings for integration with Office 365: %s', $userMailboxSettings->getId()));
                return $settings;
            }

            $mailboxTimezone = $workingHours->getTimeZone();
            // can not use getStartTime or getEndTime
            // https://github.com/microsoftgraph/msgraph-sdk-php/issues/387
            // https://github.com/microsoftgraph/MSGraph-SDK-Code-Generator/issues/524
            $settings['workHoursStart'] = $this->_convertToUtc($workingHours->getProperties()['startTime'], $mailboxTimezone);
            $settings['workHoursEnd']   = $this->_convertToUtc($workingHours->getProperties()['endTime'], $mailboxTimezone);
            // can not use getDaysOfWeek method due to issie
            // https://github.com/microsoftgraph/msgraph-sdk-php/issues/76
            $settings['workDays'] = array_reduce(
                $workingHours->getProperties()['daysOfWeek'] ?? [],
                static function ($carry, $item) {
                    $delimiter = $carry === null ? '' : ',';

                    return sprintf('%s%s%s', $carry, $delimiter, date('N', strtotime($item)));
                },
                null
            );
        } catch (Throwable $exception) {
            $this->_logger->error(__LINE__, sprintf('Error when _convertToLocalSettings: %s', $exception->getMessage()));
        }
        return $settings;
    }

    private function _convertToUtc(string $time, TimeZoneBase $windowsTimezone = null): string {
        $timezoneName = isset($windowsTimezone) ? $windowsTimezone->getName() : 'UTC';
        $timezone     = TimeZones::getTzFromWindows($timezoneName);
        $dateTime     = new \DateTime($time, $timezone);

        $utcTimezone = new \DateTimeZone('UTC');
        $dateTime->setTimezone($utcTimezone);

        return $dateTime->format('H:i');
    }

    private function _updateUserSettings(Colleague $user, array $userSettings) {
        // define is setting need to be updated
        $userMailboxSettings = $this->_fetchUserMailboxSettings($user);
        $areSettingsChanged  = $this->_areSettingsChanged($userSettings, $userMailboxSettings);
        if (! $areSettingsChanged) {
            return [];
        }

        $this->_calendarService->updateSettings($userSettings, $userMailboxSettings, $user, Settings::SETTINGS_APP_OFFICE365_OUTLOOK);
        // todo encapsulate event update
        $generatedWorkHoursEvents = $this->_calendarService->generateWorkHoursEvents($userMailboxSettings, Events::EVENT_APP_OFFICE365_OUTLOOK);
        // todo events repo
        $userWorkingHoursEvents = Events::getInstance()->getUserEvents($user, true, Events::EVENT_APP_OFFICE365_OUTLOOK);

        return $this->_calendarService->updateEvents($userWorkingHoursEvents, $generatedWorkHoursEvents, $user, Events::EVENT_APP_OFFICE365_OUTLOOK);
    }

    private function _areSettingsChanged(array $userSettings, array $convertedSettings): bool {
        $localSettings  = array_intersect_key($userSettings, array_flip(self::USER_SETTINGS_LIST));
        $remoteSettings = array_intersect_key($convertedSettings, array_flip(self::USER_SETTINGS_LIST));

        // yep array_diff compare one way
        return count(array_merge(array_diff($remoteSettings, $localSettings), array_diff($localSettings, $remoteSettings))) > 0;
    }
}
