<?php

namespace Wildix\Normalizer;

/**
 * Class PhoneNumberNormalizer
 */
class PhoneNumberNormalizer {

    /**
     * International phone prefix
     *
     * @var string
     */
    private $_internationalPrefix;
    /**
     * National phone prefix
     *
     * @var string
     */
    private $_nationalPrefix;
    /**
     * Country phone code
     *
     * @var string
     */
    private $_countryCode;

    /**
     * PhoneNumberNormalizer constructor.
     *
     * @param string $internationalPrefix International phone prefix.
     * @param string $nationalPrefix      National phone prefix.
     * @param string $countryCode         Country phone code.
     */
    public function __construct($internationalPrefix, $nationalPrefix, $countryCode) {
        $this->_internationalPrefix = (string)$internationalPrefix;
        $this->_nationalPrefix      = (string)$nationalPrefix;
        $this->_countryCode         = (string)$countryCode;
    }

    /**
     * Normalize phone number
     *
     * @param string  $phoneNumber            Value for normalize.
     * @param integer $numberLength           Length number for normalize.
     * @param boolean $ignoreStartsFromLetter Is TRUE then numbers starts from letter will be skipped.
     *
     * @return string
     */
    public function normalize($phoneNumber, $numberLength = 6, $ignoreStartsFromLetter = false) {
        if ($ignoreStartsFromLetter && !ctype_digit(mb_substr($phoneNumber, 0, 1))) {
            return $phoneNumber;
        }

        $cidLen = mb_strlen((string)$phoneNumber);

        if ($cidLen <= $numberLength || $phoneNumber[0] == '+') {
            return $phoneNumber;
        }

        $internationalPrefLen = mb_strlen($this->_internationalPrefix);
        $nationalPrefLen      = mb_strlen($this->_nationalPrefix);
        $firstChar            = mb_substr($phoneNumber, 0, 1);

        if ($internationalPrefLen > 0 && mb_substr($phoneNumber, 0, $internationalPrefLen) == $this->_internationalPrefix) {
            return '+' . mb_substr($phoneNumber, $internationalPrefLen);
        }

        if ($nationalPrefLen > 0 && mb_substr($phoneNumber, 0, $nationalPrefLen) == $this->_nationalPrefix) {
            $result  = !empty($this->_countryCode) ? '+' . $this->_countryCode : '';
            $result .= ($this->_countryCode == '39') ? $phoneNumber : mb_substr($phoneNumber, $nationalPrefLen);

            return $result;
        }

        if ($this->_countryCode == '39' && ($cidLen == 10 || in_array($firstChar, array('8', '7', '4', '3', '1')))) {
            // '8' -- Special rate numbers
            // '7' -- Internet service numbers
            // '4' -- Internal phone net service numbers
            // '3' -- Mobile phones
            // '1' -- Service and emergency numbers
            return '+39' . $phoneNumber;
        }

        if ($this->_countryCode == '380' && mb_substr($phoneNumber, 0, 2) == '80') {
            return '+3' . $phoneNumber;
        }

        if ($this->_countryCode == '39') {
            return '+' . $this->_countryCode . $this->_nationalPrefix . $phoneNumber;
        }

        return !empty($this->_countryCode) ? '+' . $this->_countryCode . $phoneNumber : $phoneNumber;
    }
}
