<?php

namespace Wildix\Notifications\Email;

use Wildix\License;
use Wildix\Notifications\Email\Interfaces\EmailTemplateProcessorInterface;
use Wildix\Services\SpeechToText;
use Wildix\WhiteLabel;

class EmailTemplateProcessor implements EmailTemplateProcessorInterface {

    private $_templateDir   = '/rw2/etc/mailTemplates';
    private $_voicemailText = false;
    private $_license;
    private $describeVoicemail;
    private $stt;
    private $whiteLabel;
    private $pbxDomain;

    private $_defaultParams = array(
        'body' => '',
        'body64' => '',
        'to' => '',
        'from' => '',
        'replyto' => '',
        'cc' => '',
        'bcc' => '',
        'template' => 'default',
        'type' => 'default',
        'subject' => '',
        'subject64' => '',
        'attach' => ''
    );

    public function __construct(
        License $_license,
        string $describeVoicemail,
        SpeechToText $stt,
        WhiteLabel $whiteLabel,
        string $pbxDomain
    ) {
        $this->_license          = $_license;
        $this->describeVoicemail = $describeVoicemail;
        $this->stt               = $stt;
        $this->whiteLabel        = $whiteLabel;
        $this->pbxDomain         = $pbxDomain;
    }

    public function process(array $data): array {
        $data = $this->convertParams($data);

        $data['body'] = $this->fillTemplate($data);

        return $data;
    }

    private function convertParams($params): array {
        $params = array_merge($this->_defaultParams, $params);

        // body section: body always wins over the body64
        if (empty($params['body']) && !empty($params['body64'])) {
            $params['body'] = base64_decode($params['body64']);
        }

        if (empty($params['name']) && !empty($params['name64'])) {
            $params['name'] = base64_decode($params['name64']);
        }

        $params['language'] = !empty($params['language']) ? strtolower(trim($params['language'])) : 'en';

        if (!empty($params['timestamp'])) {
            $timestamp = intval($params['timestamp']);
            [$params['time'], $params['date']] = $this->getDateTimeParams($timestamp, $params['language']);
        }

        $isVoicemail = (!empty($params['type']) && $params['type'] === 'voicemail' && !empty($params['attach']));

        $isVoicemailAble = ($this->getLicense()->isPremium() && $this->getLicense()->isSubscription());

        $describeVoicemail = !($this->getDescribeVoicemail() == 'no');

        $isTranscriptAllowed = filter_var($params['transcript'] ?? 'yes', FILTER_VALIDATE_BOOLEAN);
        if ($isVoicemail && $isVoicemailAble && $describeVoicemail && $isTranscriptAllowed) {
            $this->getVoiceMailText($params);

            if ($this->_voicemailText) {
                $params['body'] = $params['body'] . PHP_EOL . PHP_EOL . 'Transcript: ' . $this->_voicemailText;
            }
        }

        $params['subject'] = $this->getSubject($params);

        return $params;
    }

    private function getDateTimeParams($timestamp, $lang): array {
        $timezone = explode('/', date_default_timezone_get());

        if ($lang === 'en-us' || ($lang === 'en' && $timezone[0] === 'America')) {
            $timeFormat = 'g:i a';
            $dateFormat = 'n/j/Y';
        } else {
            $timeFormat = 'H:i';
            $dateFormat = 'd/m/Y';
        }

        return [
            date($timeFormat, $timestamp),
            date($dateFormat, $timestamp),
        ];
    }

    private function getVoiceMailText($params): void {
        $language = 'en-GB';

        if (!empty($params['asrLanguage'])) {
            $language = $params['asrLanguage'];
        }

        if (!empty($params['asrFile']) && file_exists($params['asrFile'])) {
            $this->_voicemailText = $this->getStt()->getTextFromVoiceMail($params['asrFile'], $language);
        } else {
            $this->_voicemailText = $this->getStt()->getTextFromVoiceMail($params['attach'], $language);
        }
    }

    private function fillTemplate($params) {
        if ($params['type'] == 'welcome') {
            return preg_replace('/"/', "'", $params['body']);
        }
        // get template
        $tpl = $this->getTemplate($params['type'], $params['template'], $params['language']);

        $updateData = [
            'collaboration' => $this->getWhiteLabel()->getCollaboration(),
            'brand' => $this->getWhiteLabel()->getBrand(),
            'emailLogo' => $this->getWhiteLabel()->getEmailLogo(),
            'site' => $this->getWhiteLabel()->getSite(),
            'helpLink' => $this->getWhiteLabel()->getHelpLink(),
            'whitespaceLine' => '',
        ];

        $whitespaceLineLength = 150;
        while (--$whitespaceLineLength > 0) {
            $updateData['whitespaceLine'] .= '&nbsp;&zwnj;';
        }

        if (!empty($params['body'])) {
            $params['body'] = $this->replaceNewLine($params['body']);
        }

        $updateData = array_change_key_case(array_merge($updateData, $params), CASE_UPPER);

        return $this->replaceVariablesInTemplate($tpl, $updateData);
    }

    public function getTemplate($type = 'default', $templateName = 'default', $lang = 'en'): string {
        if (!file_exists($this->_templateDir . '/' . $templateName)) {
            $templateName = 'default';
        }

        $header = $this->getTemplateComponent($templateName, $type, $lang, 'header.html');
        $body   = $this->getTemplateComponent($templateName, $type, $lang, 'body.html');
        $footer = $this->getTemplateComponent($templateName, $type, $lang, 'footer.html');

        return file_get_contents($header) . file_get_contents($body) . file_get_contents($footer);
    }

    private function getTemplateComponent($name, $type, $lang, $component) {
        $file = $this->searchFileByLanguage($name, $type, $lang, $component);
        if ($file) {
            return $file;
        }
        if ($name !== 'default') {
            $file = $this->searchFileByLanguage('default', $type, $lang, $component);
            if ($file) {
                return $file;
            }
        }
        $file = $this->searchFileByLanguage($name, 'default', $lang, $component);

        return $file ? $file : $this->searchFileByLanguage('default', 'default', $lang, $component);
    }

    private function searchFileByLanguage($templateName, $templateType, $templateLang, $templateComponent) {
        $file = $this->getFile($templateName, $templateType, $templateLang, $templateComponent);
        if (!$file && $templateLang !== 'en') {
            $file = $this->getFile($templateName, $templateType, 'en', $templateComponent);
        }

        return $file;
    }

    private function getFile($templateName, $templateType, $templateLang, $templateComponent) {
        $file = $this->_templateDir . '/' . $templateName . '/' . $templateType . '/' . $templateLang . '/' . $templateComponent;

        return file_exists($file) ? $file : false;
    }

    private function replaceNewLine($text) {
        return preg_replace('/[\n\r]/', '<br>', $text);
    }

    private function replaceVariablesInTemplate($template, array $variables) {
        return preg_replace_callback(
            '#{_(.*?)_}#',
            function ($match) use ($variables) {
                $match[1] = str_replace('_', '', trim($match[1], '$'));
                return $variables[$match[1]];
            },
            ' ' . $template . ' '
        );
    }

    private function getSubject($params): string {
        $subject = $params['subject'];

        // check if subject defined: subject always wins over the subject64
        if (empty($params['subject']) && empty($params['subject64'])) {
            $subject = sprintf('Wildix PBX %s', $this->getPbxDomain());
        }
        if (empty($params['subject']) && !empty($params['subject64'])) {
            $subject = base64_decode($params['subject64']);
        }
        if ($params['type'] === 'missedCall') {
            $subject = _t('sendEmail.subject.missedCall', $params['name'], $params['number'], $params['time'], $params['date']);
        }

        return $subject;
    }

    private function getLicense(): License {
        return $this->_license;
    }

    private function getDescribeVoicemail(): string {
        return $this->describeVoicemail;
    }

    private function getStt(): SpeechToText {
        return $this->stt;
    }

    private function getWhiteLabel(): WhiteLabel {
        return $this->whiteLabel;
    }

    private function getPbxDomain(): string {
        return $this->pbxDomain;
    }
}
