<?php
declare(strict_types=1);

namespace Wildix\PBXes\Network;

use Services\Helpers;
use Services\KamailioService;
use Services\LdapService;
use Services\TProxyService;
use Wildix\CallweaverManager;
use Wildix\Databases\AvatarCache;
use Wildix\Ejabberd;
use Wildix\Error;
use Wildix\Exception;
use Wildix\Generator\StringGenerator;
use Wildix\InvalidDataException;
use Wildix\License;
use Wildix\PBXes\PBX;
use Wildix\PBXes\PBXes;
use Wildix\Services\AbstractService;
use Wildix\Services\Exception\ServiceException;
use Wildix\System\Dhcp\ServerManager;
use Wildix\System\LdapRadius;
use Wildix\System\Network;
use Wildix\System\Ntp;
use Wildix\Validator;

/**
 * Class NetworkServices.
 */
class NetworkServices extends AbstractService {

    const DEFAULT_PUBLIC_PORT = 443;

    const LDAP_UNSET_CLIENT_OUT = '/tmp/ldap_unset_slave.out';

    const URL_SERVICE_COMMUNICATE = '/api/v1/PBX/WmsNetwork/';

    const USER_SETTINGS_DIR = '/rw2/etc/user.d';

    const ACTIONS = [
        'startServer' => 'startServer',
        'stopServer' => 'stopServer',
        'startClient' => 'startClient',
        'stopClient' => 'stopClient',
        'reInitLdapClient' => 'reInitLdapClient',
        'updatePbxSettings' => 'updatePbxSettings'
    ];

    /**
     * Server class.
     *
     * @var Server $_server
     */
    private $_server;

    /**
     * Client class.
     *
     * @var Client $_client
     */
    private $_client;

    /**
     * Ntp class.
     *
     * @var Ntp $_ntp
     */
    private $_ntp;

    /**
     * WMS_Ldap class.
     *
     * @var \WMS_Ldap $_wmsLdap
     */
    private $_wmsLdap;

    /**
     * PBXes class.
     *
     * @var PBXes $_PBXes
     */
    private $_PBXes;

    /**
     * KamailioService class.
     *
     * @var KamailioService $_kamailioService
     */
    private $_kamailioService;

    /**
     * TProxyService class.
     *
     * @var TProxyService $_tProxyService
     */
    private $_tProxyService;

    /**
     * Ejabberd class.
     *
     * @var Ejabberd $_ejabberd
     */
    private $_ejabberd;

    /**
     * LdapService class.
     *
     * @var LdapService $_ldapService
     */
    private $_ldapService;

    /**
     * AvatarCache class.
     *
     * @var AvatarCache $_avatarCacheDB
     */
    private $_avatarCacheDB;

    /**
     * ServerManager class.
     *
     * @var ServerManager $_serverManager
     */
    private $_serverManager;

    /**
     * CallweaverManager class.
     *
     * @var CallweaverManager $_callweaverManager
     */
    private $_callweaverManager;

    /**
     * NetRepository class.
     *
     * @var NetRepository $_netRepository
     */
    private $_netRepository;

    /**
     * LdapRepository class.
     *
     * @var LdapRepository $_ldapRepository
     */
    private $_ldapRepository;

    private $_ldapRadius;

    public function __construct(
        Server $server,
        Client $client,
        Ntp $ntp,
        \WMS_Ldap $wmsLdap,
        PBXes $PBXes,
        KamailioService $kamailioService,
        Ejabberd $ejabberd,
        LdapService $ldapService,
        AvatarCache $avatarCacheDB,
        ServerManager $serverManager,
        CallweaverManager $callweaverManager,
        NetRepository $netRepository,
        LdapRepository $ldapRepository,
        TProxyService $tProxyService,
        LdapRadius $ldapRadius,
        $logLevel = 5
    ) {
        parent::__construct($logLevel);

        $this->_server            = $server;
        $this->_client            = $client;
        $this->_ntp               = $ntp;
        $this->_wmsLdap           = $wmsLdap;
        $this->_PBXes             = $PBXes;
        $this->_kamailioService   = $kamailioService;
        $this->_ejabberd          = $ejabberd;
        $this->_ldapService       = $ldapService;
        $this->_avatarCacheDB     = $avatarCacheDB;
        $this->_serverManager     = $serverManager;
        $this->_callweaverManager = $callweaverManager;
        $this->_netRepository     = $netRepository;
        $this->_ldapRepository    = $ldapRepository;
        $this->_tProxyService     = $tProxyService;
        $this->_ldapRadius        = $ldapRadius;
    }

    private function startServer(array $data = []) {
        if ($this->getPbx()->isEnabledWmsNetwork()) {
            throw new Exception(Error::PBXAlreadyInNetwork);
        }
        $this->getLog()->info(__LINE__, 'Network: Initializing Server');
        $this->validateServerData($data);
        if (!License::getInstance()->isActivated()) {
            throw new Exception(Error::WMSIsNotActivated);
        }

        $this->_tProxyService->addLdapProxying();
        $this->_tProxyService->saveAndReload();

        $entry    = $this->_PBXes->getPBXBySerial($this->getPbx()->getSerial());
        $publicIp = $entry->getPublicIp() ? $entry->getPublicIp() : '';

        $dataNetConf = [
            'STARTUP'    => 2,
            'SRV_IP'     => '127.0.0.1',
            'NCOMMUNITY' => $data['slaves_login'],
            'NSECRET'    => $data['slaves_password'],
        ];
        $dateSipConf = [
            'username' => $this->getPbx()->getSerial(),
            'secret'   => $data['slaves_password'],
        ];

        $this->setNetConfig($dataNetConf);
        $this->_callweaverManager->setSipWmsNetworkConf($dateSipConf);
        $this->_kamailioService->reloadConf($data['slaves_password']);

        $insertLdapData = [
            'serial'       => $this->getPbx()->getSerial(),
            'hostname'     => gethostname(),
            'masterIp'     => $publicIp,
            PBX::IS_MASTER => true,
        ];

        $this->getLog()->info(__LINE__, 'Network: Insert pbx(server) to LDAP');
        $this->_server->insertPbxToLdap($insertLdapData);
        $this->getLog()->info(__LINE__, 'Network: Init server ldapcmd');
        $this->_ldapService->init('server');

        $currentParams = $this->_ldapRepository->get();
        if (!$this->_ldapRepository->save(array_merge($currentParams, ['host_rw' => 'localhost']))) {
            $this->getLog()->error(__LINE__, Error::getMessage(Error::IsNotWriteable) . ' ' . $this->_ldapRepository::STORAGE_PATH_LDAP);
            throw new Exception(Error::IsNotWriteable);
        }

        $this->getLog()->info(__LINE__, 'Network: save dhcp hosts config');
        $this->_serverManager->saveDhcpHostsConfig();
        $this->getLog()->info(__LINE__, 'Network: server initialization done');
    }


    public function stopServer(array $data = []) {
        if (!$this->getPbx()->isEnabledWmsNetwork()) {
            throw new Exception(Error::PBXNotInNetwork);
        }
        if (!$this->_netRepository->save(['STARTUP' => 0])) {
            $this->getLog()->error(__LINE__, Error::getMessage(Error::IsNotWriteable) . ' ' . $this->_netRepository::STORAGE_PATH_NET);
            throw new Exception(Error::IsNotWriteable);
        }
        $entry = $this->_PBXes->getPBXBySerial($this->getPbx()->getSerial());

        if ($this->_PBXes->updatePBXProperty([PBX::IS_MASTER => false], $entry) !== true) {
            throw new Exception(Error::PBXPropertyCannotUpdateInLdap);
        }

        if (!$this->_ldapRadius->isEnabled()) {
            $this->_tProxyService->removeLdapProxying();
            $this->_tProxyService->saveAndReload();
        }

        $this->_kamailioService->restart();
        $this->_serverManager->clearDhcpHostsConfigFile();
        $this->_ldapService->init('config');
    }

    private function startClient(array $data = []) {
        if ($this->getPbx()->isEnabledWmsNetwork()) {
            throw new Exception(Error::PBXAlreadyInNetwork);
        }
        $this->getLog()->info(__LINE__, 'Network: Initializing Client');
        $this->validateClientData($data);

        $this->assertRemoteHostAvailable($data['master_ip'], (int)$data['master_https_port'], 5);

        $this->getLog()->info(__LINE__, sprintf('Network: Connecting to Master %s:%s', $data['master_ip'], $data['master_https_port']));

        $requestParams = [
            'master_connect_login'    => $data['master_connect_login'],
            'master_connect_password' => $data['master_connect_password']
        ];

        $serverPort = $data['master_https_port'] !== '443' ? ':' . $data['master_https_port'] : '';

        $ntpData = $this->requestToServer('POST', 'ntp', $data['master_ip'] . $serverPort, $requestParams);

        $entry      = $this->_PBXes->getPBXBySerial($this->getPbx()->getSerial());
        $publicPort = $entry->getPublicPort() ? $entry->getPublicPort() : self::DEFAULT_PUBLIC_PORT;
        $publicIp   = $entry->getPublicIp() ? $entry->getPublicIp() : '';

        $dataNetConf = [
            'STARTUP'        => 1,
            'SRV_IP'         => $data['master_ip'],
            'NCOMMUNITY'     => $data['master_connect_login'],
            'NSECRET'        => $data['master_connect_password'],
            'SRV_HTTPS_PORT' => $data['master_https_port'],
        ];
        $dateSipConf = [
            'username' => $this->getPbx()->getSerial(),
            'secret'   => $data['master_connect_password']
        ];

        $this->setNetConfig($dataNetConf);
        $this->_callweaverManager->setSipWmsNetworkConf($dateSipConf);
        $this->_kamailioService->reloadConf($data['master_connect_password']);

        $this->_client->configuringNtpServer($ntpData['ntp']);
        $this->getLog()->info(__LINE__, 'Network: Adding the PBX to the Network');

        $ldapParams = [
            'master_connect_login'    => $data['master_connect_login'],
            'master_connect_password' => $data['master_connect_password'],
            'ip'                      => $publicIp,
            'serial'                  => $this->getPbx()->getSerial(),
            'hostname'                => gethostname(),
            'pbxDomain'               => $this->getPbx()->getDomain(),
            'publicPort'              => $publicPort
        ];
        $this->requestToServer('PUT', 'addClientToLdap', $data['master_ip'] . $serverPort, $ldapParams);
        if ($this->_ldapService->init('client') === 1) {
            throw new Exception(Error::InvalidConnectToServer);
        }

        $this->getLog()->info(__LINE__, 'Network: Setting LDAP configuration');
        $this->_client->usersSyncDisableRecurrent();
        $currentParams = $this->_ldapRepository->get();
        if (!$this->_ldapRepository->save(array_merge($currentParams, ['host_rw' => $data['master_ip'], 'port_rw' => $data['master_https_port']]))) {
            $this->getLog()->error(__LINE__, Error::getMessage(Error::IsNotWriteable) . ' ' . $this->_ldapRepository::STORAGE_PATH_LDAP);
            throw new Exception(Error::IsNotWriteable);
        }
        $this->_ejabberd->cleanEjabberdSlave();
        $this->_kamailioService->restart();
        $this->_client->fixAuth();
    }

    public function stopClient(array $data = []) {
        if (!$this->getPbx()->isEnabledWmsNetwork()) {
            throw new Exception(Error::PBXNotInNetwork);
        }
        if (is_file(self::LDAP_UNSET_CLIENT_OUT)) {
            unlink(self::LDAP_UNSET_CLIENT_OUT);
        }

        if (!$this->_netRepository->save(['STARTUP' => 0])) {
            $this->getLog()->error(__LINE__, Error::getMessage(Error::IsNotWriteable) . ' ' . $this->_netRepository::STORAGE_PATH_NET);
            throw new Exception(Error::IsNotWriteable);
        }
        $this->_ldapService->init('config');

        // WMS-5801 Moving a large number of users can take more than a minute. Running this process in background.
        exec('nohup /usr/bin/php /usr/sbin/ldap_unset_slave  >/dev/null &');
        sleep(1);
    }

    private function reInitLdapClient(array $data = []) {
        if (!$this->getPbx()->isEnabledWmsNetwork()) {
            throw new Exception(Error::PBXNotInNetwork);
        }

        $this->_checkVersions();

        if ($this->_ldapService->init('client') === 1) {
            throw new Exception(Error::InvalidConnectToServer);
        }
        try {
            $this->getLog()->info(__LINE__, 'Try to clear all avatar cache');
            $this->_avatarCacheDB->clearAllCache();
        } catch (\Exception $e) {
            $this->getLog()->error(__LINE__, sprintf('Can\'t clear avatar cache. Error[%s]: %s', $e->getCode(), $e->getMessage()));
        }
    }

    private function _checkVersions(): void {
        if ($this->isVersionDiffers()) {
            $this->getLog()->info(__LINE__, 'Network: problem the client, different versions pbx wms network');
            throw new InvalidDataException(sprintf(Error::getMessage(Error::VersionDiffers), $this->_server->getVersion()));
        }
    }

    private function assertRemoteHostAvailable(string $srvIp, int $servHttpsPort, float $timeout): void {
        if (!$this->_server->isHostAvailable($srvIp, $servHttpsPort, $timeout)) {
            throw new Exception(Error::CannotConnectToHost);
        }
    }

    /**
     * This method is temporary, it will need to be removed when all pbxes has actual the version
     *
     * @throws InvalidDataException Invalid data.
     *
     * @return void
     */
    public function updateClientIp(): void {
        if (!$this->getPbx()->isMaster()) {
            throw new InvalidDataException('The pbx is doesn\'t a server');
        }

        $this->_checkVersions();

        $search = $this->_wmsLdap->getEntries($this->_wmsLdap::PBX, '', ['pbxIp', 'pbxPublicIp', 'pbxSerial', 'pbxIsMaster']);

        foreach ($search as $pbx) {
            if (strripos($pbx['pbxIp'], '5.15.') !== false) {
                $this->_PBXes->updatePBXProperty([PBX::IP => $pbx['pbxPublicIp']], $this->_PBXes->getPBXBySerial($pbx['pbxSerial']));
            }
        }
    }

    public function getIp(): string {
        $netParams = $this->_netRepository->get();
        if (!$this->_server->isHostAvailable($netParams['SRV_IP'], (int)$netParams['SRV_HTTPS_PORT'], 5)) {
            return '';
        }

        $requestParams = [
            'master_connect_login'    => $netParams['NCOMMUNITY'],
            'master_connect_password' => $netParams['NSECRET']
        ];

        $serverPort = $netParams['SRV_HTTPS_PORT'] !== '443' ? ':' . $netParams['SRV_HTTPS_PORT'] : '';

        $data = $this->requestToServer('POST', 'ntp', $netParams['SRV_IP'] . $serverPort, $requestParams);

        return $data['ip'];
    }

    private function updatePbxSettings(array $data = []) {
        if (!$this->getPbx()->isEnabledWmsNetwork()) {
            throw new Exception(Error::PBXNotInNetwork);
        }

        $this->_checkVersions();

        $this->validateUpdateSettingsData($data);

        if ($this->getPbx()->isSlave()) {
            $netParams = $this->_netRepository->get();

            $this->assertRemoteHostAvailable($netParams['SRV_IP'], (int)$netParams['SRV_HTTPS_PORT'], 5);

            $serverPort    = $netParams['SRV_HTTPS_PORT'] !== '443' ? ':' . $netParams['SRV_HTTPS_PORT'] : '';
            $requestParams = [
                'master_connect_login'    => $data['login'],
                'master_connect_password' => $data['password']
            ];
            $this->requestToServer('PUT', 'checkCredentials', $netParams['SRV_IP'] . $serverPort, $requestParams);
        }

        $dataNetConf = [
            'NCOMMUNITY'     => $data['login'],
            'NSECRET'        => $data['password']
        ];
        $dateSipConf = [
            'username' => $this->getPbx()->getSerial(),
            'secret'   => $data['password']
        ];

        $currentNetParams = $this->_netRepository->get();
        $this->setNetConfig(array_merge($currentNetParams, $dataNetConf));
        $this->_callweaverManager->setSipWmsNetworkConf($dateSipConf);
        $this->_kamailioService->reloadConf($data['password']);
        $this->_ldapService->init('update_credentials');
        $this->_ldapService->init('config');
        $this->_ldapService->init('restart');
        $this->_wmsLdap->closeConnect();
        $entry = $this->_PBXes->getPBXBySerial($this->getPbx()->getSerial());

        if (empty($entry)) {
            throw new Exception(Error::EntryIsNotPresent);
        }

        $properties = [
            PBX::HOSTNAME => gethostname()
        ];
        if ($this->_PBXes->updatePBXProperty($properties, $entry) !== true) {
            throw new Exception(Error::PBXPropertyCannotUpdateInLdap);
        }
    }

    private function validateServerData(array $data) {
        $validator = new Validator($data);

        $validator->rule('required', 'slaves_login');
        $validator->rule('lengthMin', 'slaves_login', 13);
        $validator->rule('regex', 'slaves_login', '/^[0-9a-zA-Z]+$/');

        $validator->rule('required', 'slaves_password');
        $validator->rule('regex', 'slaves_password', '/^[0-9a-zA-Z]+$/');
        $validator->rule('NetworkPassword', 'slaves_password');

        if (!$validator->validate()) {
            throw new Exception(Error::InvalidData);
        }
    }

    private function validateClientData(array $data) {
        $validator = new Validator($data);
        $validator->rule('required', 'master_ip');

        $validator->rule('required', 'master_https_port');
        $validator->rule('integer', 'master_https_port');
        $validator->rule('between', 'master_https_port', [1, 65535]);

        $validator->rule('required', 'master_connect_login');
        $validator->rule('lengthMin', 'master_connect_login', 13);
        $validator->rule('regex', 'master_connect_login', '/^[0-9a-zA-Z]+$/');

        $validator->rule('required', 'master_connect_password');
        $validator->rule('regex', 'master_connect_password', '/^[0-9a-zA-Z]+$/');
        $validator->rule('NetworkPassword', 'master_connect_password');

        if (!$validator->validate()) {
            throw new Exception(Error::InvalidData);
        }
    }

    private function validateUpdateSettingsData(array $data) {
        $validator = new Validator($data);

        $validator->rule('required', 'login');
        $validator->rule('lengthMin', 'login', 13);
        $validator->rule('regex', 'login', '/^[0-9a-zA-Z]+$/');

        $validator->rule('required', 'password');
        $validator->rule('regex', 'password', '/^[0-9a-zA-Z]+$/');
        $validator->rule('NetworkPassword', 'password');

        if (!$validator->validate()) {
            throw new Exception(Error::InvalidData);
        }
    }

    private function validateCredentialsData(array $data) {
        $validator = new Validator($data);

        $validator->rule('required', 'master_connect_login');
        $validator->rule('lengthMin', 'master_connect_login', 13);
        $validator->rule('regex', 'master_connect_login', '/^[0-9a-zA-Z]+$/');

        $validator->rule('required', 'master_connect_password');
        $validator->rule('regex', 'master_connect_password', '/^[0-9a-zA-Z]+$/');
        $validator->rule('NetworkPassword', 'master_connect_password');

        if (!$validator->validate()) {
            throw new InvalidDataException(Error::getMessage(Error::InvalidData), $validator->errors());
        }
    }

    public function run(array $params) {
        if (isset(self::ACTIONS[$params['action']])) {
            $this->{self::ACTIONS[$params['action']]}($params);
        } else {
            throw new Exception(Error::NotFound, 404);
        }
    }

    public function isVersionDiffers(): bool {
        $search = $this->_wmsLdap->getEntries($this->_wmsLdap::PBX, '', ['description']);

        foreach ($search as $pbx) {
            if (!key_exists('description', $pbx) || version_compare($this->_server->getVersion(), Helpers::makePbxVersionByDescription($pbx['description'])) > 0) {
                return true;
            }
        }

        return false;
    }

    public function getNtp(array $params): array {
        $this->validateCredentialsData($params);
        if (!$this->getPbx()->isMaster()) {
            throw new Exception(Error::NotAserver);
        }

        $credentials = $this->getMasterCredentials();
        if ($credentials['master_login'] !== $params['master_connect_login'] || $credentials['master_password'] !== $params['master_connect_password']) {
            throw new Exception(Error::InvalidCredentials);
        }

        $this->_checkVersions();

        $netParams = $this->_netRepository->get();

        return [
            'ntp'        => $this->_ntp->getConfig(),
            'ip'         => (array_key_exists('REMOTE_ADDR', $_SERVER) ? $_SERVER['REMOTE_ADDR'] : ''),
            'listenPort' => $netParams['PORT'] ?? ''
        ];
    }

    public function checkCredentials(array $params): void {
        $this->validateCredentialsData($params);
        if (!$this->getPbx()->isMaster()) {
            throw new Exception(Error::NotAserver);
        }

        $credentials = $this->getMasterCredentials();
        if ($credentials['master_login'] !== $params['master_connect_login'] || $credentials['master_password'] !== $params['master_connect_password']) {
            throw new Exception(Error::InvalidCredentials);
        }
    }

    public function addClientToLdap(array $params) {
        $this->validateCredentialsData($params);
        if (!$this->getPbx()->isMaster()) {
            throw new Exception(Error::NotAserver);
        }

        $credentials = $this->getMasterCredentials();
        if ($credentials['master_login'] !== $params['master_connect_login'] || $credentials['master_password'] !== $params['master_connect_password']) {
            throw new Exception(Error::InvalidCredentials);
        }

        $entry = $this->_PBXes->getPBXBySerial($params['serial']);

        if (!$entry) {
            $attributes = [
                PBX::ID          => $params['serial'],
                PBX::SERIAL      => $params['serial'],
                PBX::HOSTNAME    => $params['hostname'],
                PBX::IP          => $params['ip'],
                PBX::DOMAIN      => $params['pbxDomain'],
                PBX::PUBLIC_PORT => $params['publicPort'],
                PBX::IS_MASTER   => false
            ];

            if ($this->_wmsLdap->addPbx($attributes) !== true) {
                throw new Exception(Error::PBXNotAddedToLdap);
            }
            $this->_kamailioService->restart();
        } else {
            $result = [
                PBX::HOSTNAME    => $params['hostname'],
                PBX::IP          => $params['ip'],
                PBX::DOMAIN      => $params['pbxDomain'],
                PBX::PUBLIC_PORT => $params['publicPort'],
                PBX::IS_MASTER   => false
            ];

            if ($this->_PBXes->updatePBXProperty($result, $entry) !== true) {
                throw new Exception(Error::PBXPropertyCannotUpdateInLdap);
            }
            $this->_kamailioService->restart();
        }

        $this->getLog()->info(__LINE__, 'Network: Writing DNSMasq parameters');
        $this->_serverManager->saveDhcpHostsConfig();
    }

    public function sipPeerPrune(array $params) {
        if (array_key_exists('extension', $params)) {
            $this->_callweaverManager->sipPeerPrune($params['extension']);
        }
    }

    public function removeUser(array $params) {
        $userId  = $params['userId'];
        $userDir = $this->getUsersSettingsDir();
        if (empty($userId)) {
            throw new Exception(Error::InvalidUserId);
        }
        if (!$userDir) {
            throw new Exception(Error::DirectoryNotFound);
        }
        $filename          = $userDir . $userId . '.lua';
        $filenameAvatarPng = $userDir . 'avatars/user_avatar_' . $userId . '.png';
        $filenameAvatarJpg = $userDir . 'avatars/user_avatar_' . $userId . '.jpg';

        if (file_exists($filename)) {
            unlink($filename);
        }

        if (file_exists($filenameAvatarPng)) {
            unlink($filenameAvatarPng);
        }

        if (file_exists($filenameAvatarJpg)) {
            unlink($filenameAvatarJpg);
        }
    }

    private function setNetConfig(array $dateNet): void {
        if (!$this->_netRepository->save($dateNet)) {
            $this->getLog()->error(__LINE__, Error::getMessage(Error::IsNotWriteable) . ' ' . $this->_netRepository::STORAGE_PATH_NET);
            throw new Exception(Error::IsNotWriteable);
        }
    }

    public function requestToServer(string $type, string $action, string $serverIp, array $params) {
        try {
            $this->disableBasicAuth();
            if (Network::isIp(explode(':', $serverIp)[0])) {
                $this->disableSslVerify();
            }

            $host    = 'https://' . $serverIp . self::URL_SERVICE_COMMUNICATE . $action;
            $request = $this->sendRequest($type, $host, $params);
        } catch (ServiceException $e) {
            if ($e->getCode() === 'NOT_FOUND') {
                throw new InvalidDataException(Error::getMessage(Error::VersionServerIsDifferentFromClient));
            } else {
                throw new InvalidDataException($e->getMessage());
            }
        }

        return $request;
    }

    private function getUsersSettingsDir() {
        return file_exists(self::USER_SETTINGS_DIR) ? self::USER_SETTINGS_DIR . '/' : false;
    }

    private function getMasterCredentials(): array {
        $params = $this->_netRepository->get();
        $result = [
            'master_login'    => 'w' . substr(md5((string)time()), 0, 12),
            'master_password' => StringGenerator::random(24, true, true)
        ];
        if (!$this->getPbx()->isEnabledWmsNetwork()) {
            return $result;
        }
        if (isset($params['NCOMMUNITY'])) {
            $result['master_login'] = $params['NCOMMUNITY'];
        }
        if (isset($params['NSECRET'])) {
            $result['master_password'] = $params['NSECRET'];
        }
        return $result;
    }
}
