<?php
namespace Wildix\PBXes;

use Wildix\Exception;

class PBX
{
    const DN           = 'dn';
    const ID           = 'pbxid';
    const NAME         = 'cn';
    const IP           = 'pbxIp';
    const SERIAL       = 'pbxSerial';
    const DOMAIN       = 'pbxDomain';
    const HOSTNAME     = 'pbxHostname';
    const PUBLIC_IP    = 'pbxPublicIp';
    const PUBLIC_PORT  = 'pbxPublicPort';
    const DESCRIPTION  = 'description';
    const IS_MASTER    = 'pbxIsMaster';

    protected $_ldapProperties = null;
    private $pbxid = null;

    public function __construct($properties) {
        $this->setPropertys($properties);
    }

    public function setPropertys($properties){
        if (is_array($properties)) {
            $this->_ldapProperties = $properties;
        }

        if(!empty($properties[PBX::ID])){
            $this->pbxid = $properties[PBX::ID];
        }else{
            $this->pbxid = null;
        }
    }

    public function getDn() {
        return $this->getLdapProperty(PBX::DN);
    }

    public function getID() {
        return $this->pbxid;
    }

    public function getName() {
        return $this->getLdapProperty(PBX::NAME);
    }

    public function getIp() {
        return $this->getLdapProperty(PBX::IP);
    }

    public function getIsMaster() {
        return !empty($this->getLdapProperty(PBX::IS_MASTER));
    }

    public function getSerial() {
        return $this->getLdapProperty(PBX::SERIAL);
    }

    public function getHostname() {
        return $this->getLdapProperty(PBX::HOSTNAME);
    }

    public function getDomain() {
        return $this->getLdapProperty(PBX::DOMAIN);
    }

    public function getPublicIp() {
    	return $this->getLdapProperty(PBX::PUBLIC_IP);
    }

    public function getPublicPort() {
    	return $this->getLdapProperty(PBX::PUBLIC_PORT);
    }

    public function getDescription() {
    	$return = array();
    	$desc = $this->getLdapProperty(PBX::DESCRIPTION);
    	if($desc){
    		preg_match_all("/([^\{=]*)=(\{[^=]*\})/", $desc, $arr);
    		if($arr){
    			foreach($arr[1] as $key=>$val){
    				$return[$val] = json_decode($arr[2][$key], true);
    			}
    		}
    	}
    	return $return;
    }

    public function getAsArray(){
        if(!empty($this->_ldapProperties)){
            return array(
                    'dn' => $this->getDn(),
                    'pbxid' => $this->getID(),
                    'ip' => $this->getIp(),
                    'serial' => $this->getSerial(),
                    'domain' => $this->getDomain(),
            		'publicIp' => $this->getPublicIp(),
            		'publicPort' => $this->getPublicPort(),
            		'description' => $this->getDescription()
            );
        }
        return array();
    }

    static public function getPBXFields(){
        return array(PBX::DN,
                    PBX::ID,
                    PBX::NAME,
                    PBX::IP,
                    PBX::SERIAL,
                    PBX::DOMAIN,
                    PBX::HOSTNAME,
        			PBX::PUBLIC_IP,
        			PBX::PUBLIC_PORT,
        			PBX::DESCRIPTION
        );
    }


    public function getLdapProperty($name) {
        if ($this->_ldapProperties === null && $this->pbxid !== null) {
            $this->_ldapProperties = \WMS_Ldap::getInstance()->getEntry(
                    \WMS_Ldap::PBX,
                    '('.PBX::ID."={$this->pbxid})",
                    PBX::getPBXFields()
            );
        }

        if ($this->_ldapProperties && !empty($this->_ldapProperties) && isset($this->_ldapProperties[$name])) {
            return $this->_ldapProperties[$name];
        }

        return '';
    }

    public function getCustomHttpsPort($user = null) {
        if ($user == null){
            $user = \User::getInstance();
        }
        $portResponse = $this->queryApi('/api/v1/PBX/secureport/', false, array(), $user);
        return ($portResponse && $portResponse['port']) ? $portResponse['port'] : null;
    }

    public function queryApi($api, $method = false, $params = array(), $user = null) {
        $url = 'http://' . $this->getIp() . $api . (!$method ? (strpos($api, '?') > 0 ? '&' : '?') . http_build_query($params) : '');
        $ch = curl_init($url);
        if ($method) {
            curl_setopt($ch, CURLOPT_CUSTOMREQUEST, $method);
            curl_setopt($ch, CURLOPT_POSTFIELDS, http_build_query($params));
        }
        if ($user) {
            curl_setopt($ch, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
            if ($user->getExtension())
                curl_setopt($ch, CURLOPT_USERPWD, $user->getExtension() . ':' . $user->getPassword());
            else
                curl_setopt($ch, CURLOPT_USERPWD, $user->getName() . ':' . $user->getPassword());
        }
        curl_setopt($ch, CURLOPT_USERAGENT, 'PBX Api');
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_TIMEOUT, 5);
        curl_setopt($ch, CURLOPT_AUTOREFERER, true);
        curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, false);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);

        $response = curl_exec($ch);
        if (!$response) {
            throw new Exception('Cannot connect to API or wrong response: ' . curl_error($ch));
        }

        $parsed = json_decode($response, true);
        if (!$parsed) {
            throw new Exception('Cannot parse json: ' . json_last_error());
        }

        if (isset($parsed['type']) && $parsed['type'] == 'error') {
            throw new Exception(isset($parsed['reason']) && $parsed['reason'] ? $parsed['reason'] : 'API returned unknown error');
        }

        curl_close($ch);

        return isset($parsed['result']) ? $parsed['result'] : null;
    }
}
