<?php
namespace Wildix\Phonebooks;

use Api\Database\ConnectionInterface;
use Api\Support\Facades\DB;
use Wildix\Colleagues\Colleague;
use Wildix\Databases\CTIconnectDB;
use Wildix\Databases\Database;
use Wildix\Exception;
use Wildix\Phonebooks\Exception\AvailablePhonebooksException;

class Contacts extends \Singleton {
    const TABLE_NAME          = 'cti_contacts';
    const TABLE_NAME_FULLTEXT = 'cti_contacts_fts4';
    const TABLE_FIELDS        = '`rid` as id, `username` as name, `email`, `extension`, `phone`, `office`, `home`, `home_mobile`,
                            `fax`, `mobile`, `organization`, `image_url` as picture, `shortcut`, `phonebook_id`, `user_class` as type,
                            `ringtone`, `note`, `address`, `vat_id`, `name_prefix`, `user_passport_type` as document_type,
                            `user_passport_id` as document_id, `pin`, `credit`, `hash`, `country`, `date_birth`, `born_in`,
                            `postal_code`, `city`, `province`, `license_plate`, `document_released_on` as document_released,
                            `document_from`, `last_update`, `group_id`, `source_id`, `owner_id`';
    const SEARCH_FIELDS       = [
        Contact::NAME,
        Contact::EMAIL,
        Contact::PHONE,
        Contact::FAX,
        Contact::MOBILE,
        Contact::ORGANIZATION,
        Contact::SHORTCUT,
        Contact::EXTENSION,
        Contact::OFFICE,
        Contact::HOME,
        Contact::HOME_MOBILE,
    ];

    public function getByID($id, Colleague $colleague = null) {
        $contacts = $this->getContacts(array(Contact::ID => $id), null, null, null, 1, null, null, $colleague);
        return (count($contacts) > 0) ? $contacts[0] : null;
    }

    public function getContacts($filter = null, $search = null, $searchFields = null, $start = null, $count = null, $sort = null, $dir = null, Colleague $colleague = null): array {
        $start = is_numeric($start) ? intval($start) : 0;
        $count = is_numeric($count) ? intval($count) : 100;
        $sort  = $sort ?: Contact::NAME;
        $dir   = $dir ?: 'ASC';

        try {
            $where = $this->createWhere($filter, $search, $searchFields, $colleague, Database::SELECT_QUERY);
        } catch (AvailablePhonebooksException $e) {
            return [];
        }

        $items = $this->db()
            ->query()
            ->select('*')
            ->from(sprintf('(SELECT %s FROM %s %s)', self::TABLE_FIELDS, self::TABLE_NAME, $where))
            ->offset($start)
            ->limit($count)
            ->orderBy($sort, $dir)
            ->get();

        return $items->map(
            function ($item) {
                return new Contact((array)$item);
            }
        )->toArray();
    }

    public function getCount($filter, string $search = null, $searchFields = null, Colleague $colleague = null): int {
        try {
            $where = $this->createWhere($filter, $search, $searchFields, $colleague, CTIconnectDB::SELECT_QUERY);
        } catch (AvailablePhonebooksException $e) {
            return 0;
        }

        $query = 'SELECT count() as countRows FROM `' . self::TABLE_NAME . '` ' . $where;
        $query = !empty($where) ? $query . ' OR ""="abc"' : $query . '';

        $rows = CTIconnectDB::getInstance()->query($query);
        if ($rows) {
            $row = $rows->fetchArray(SQLITE3_ASSOC);
            $rows->finalize();
            if($row){
                return $row['countRows'];
            }
        }
        return 0;
    }

    public function addContact(array $contact, Colleague $colleague = null) {
        $contact = $this->filterData($contact);

        $isValid = $this->isValidData($contact, $colleague, CTIconnectDB::INSERT_QUERY);
        if ($isValid !== true) {
            throw new Exception($isValid['message'], $isValid['code']);
        }

        if (empty($contact[Contact::HASH])) {
            $contact[Contact::HASH] = md5(join('', $contact));
        }

        if (empty($contact[Contact::SOURCE_ID])) {
            $contact[Contact::SOURCE_ID] = 'cti_' . time();
        }
        if (!isset($contact[Contact::COUNTRY])) {
            $contact[Contact::COUNTRY] = '';
        }

        $keys   = [];
        $values = [];
        foreach ($contact as $key => $value) {
            $keys[]   = '`' . $this->getRealKey($key) . '`';
            $values[] = '\'' . CTIconnectDB::getInstance()->realEscapeString($value) . '\'';
        }

        $query  = sprintf('INSERT INTO `%s` (%s) Values (%s)', self::TABLE_NAME, join(', ', $keys), join(', ', $values));
        $result = CTIconnectDB::getInstance()->exec($query);
        if($result === true){
            return CTIconnectDB::getInstance()->getLastInsertId();
        }
        return $result;
    }

    /**
     *
     * @param array $phonebook
     * @param Colleague $colleague
     * @return int
     */
    public function updateContacts($data, $filter = null, $search=null, $searchFields=null, Colleague $colleague = null){

        $data = $this->filterData($data);

        if(empty($data)){
            return 0;
        }

        if(isset($filter['id'])){
            $data['rid'] = $filter['id'];
        }

        $isValid = $this->isValidData($data, $colleague, CTIconnectDB::UPDATE_QUERY);
        if($isValid !== true){
            throw new Exception($isValid['message'], $isValid['code']);
        }

        if(isset($filter['id'])){
            unset($data['rid']);
        }

        $params = array();
        foreach ($data as $key=>$value){
            $params[] = "`".$this->getRealKey($key)."` = '".CTIconnectDB::getInstance()->realEscapeString($value)."'";
        }

        $query = 'UPDATE `'.self::TABLE_NAME.'` SET '.join(', ', $params).'
                                '.$this->createWhere($filter, $search, $searchFields, $colleague, CTIconnectDB::UPDATE_QUERY).';';

        $result = CTIconnectDB::getInstance()->exec($query);
        if ($result === true) {
            $result = CTIconnectDB::getInstance()->getCountChangesRows();
        }
        return $result;
    }

    /**
     *
     * @param $filter array
     * @param string $search
     * @param array $searchFields
     * @param Colleague $colleague
     * @return int
     */
    public function deleteContacts($filter = null, $search=null, $searchFields=null, Colleague $colleague = null){

        if($filter && $filter[Contact::PHONEBOOK_ID]){
            Phonebooks::getInstance()->getByID($filter[Contact::PHONEBOOK_ID], $colleague);
            if ($colleague) {
                if (!in_array($filter[Contact::PHONEBOOK_ID], $this->getAvailablePhonebooks($colleague, CTIconnectDB::DELETE_QUERY))) {
                    throw new Exception('No access delete contacts in this phonebook', 400);
                }
            }
        }

        if ($filter && isset($filter[Contact::ID])) {
            $contact = $this->getByID($filter[Contact::ID],$colleague);
            if(!$contact){
                throw new Exception('Not found contact', 400);
            }
            if($colleague){
                if( ! in_array($contact->getPhonebookID(),$this->getAvailablePhonebooks($colleague, CTIconnectDB::DELETE_QUERY))){
                    throw new Exception('No access delete contacts in this phonebook', 400);
                }
            }
        }
        $query = 'DELETE FROM `'.self::TABLE_NAME.'`
                                '.$this->createWhere($filter, $search, $searchFields, $colleague, CTIconnectDB::DELETE_QUERY).';';

        $result = CTIconnectDB::getInstance()->exec($query);
        if ($result === true) {
            $result = CTIconnectDB::getInstance()->getCountChangesRows();
        }
        return $result;
    }


    private function isValidData($data, Colleague $colleague = null, $queryType) {

        if (isset($data[Contact::NAME]) && $data[Contact::NAME] !== '' && mb_strlen($data[Contact::NAME], 'UTF-8') < Contact::NAME_MIN_LENGTH) {
            return [
                'code' => 400,
                'message' => 'The name is not valid (min ' . Contact::NAME_MIN_LENGTH . ' symbol)'
            ];
        }

        if (isset($data[Contact::OWNER_ID]) && $data[Contact::OWNER_ID] !== '') {
            $userData = \WMS_Ldap::getInstance()->getEntryByDn($data[Contact::OWNER_ID]);

            if (!is_array($userData)) {
                return [
                    'code' => 404,
                    'message' => 'Owner not found'
                ];
            }
        }

        if($queryType == CTIconnectDB::INSERT_QUERY){
            if(empty($data[Contact::NAME])){
                return [
                        'code' => 400,
                        'message' => 'The name cannot be empty'
                ];
            }
            if(empty($data[Contact::PHONEBOOK_ID])){
                return [
                        'code' => 400,
                        'message' => 'The phone book id cannot be empty'
                ];
            }
            if(!empty($data[Contact::SHORTCUT]) && $this->getCount( array("shortcut" => $data[Contact::SHORTCUT]), null, null, $colleague) > 0){
                return [
                    'code' => 400,
                    'message' => 'Shortcut already exists'
                ];
            }

        }else if($queryType == CTIconnectDB::UPDATE_QUERY){
            if(isset($data[Contact::NAME]) && empty($data[Contact::NAME])){
                return [
                        'code' => 400,
                        'message' => 'The name cannot be empty'
                ];
            }
            if(isset($data[Contact::PHONEBOOK_ID]) && empty($data[Contact::PHONEBOOK_ID])){
                return [
                        'code' => 400,
                        'message' => 'The phone book id cannot be empty'
                ];
            }
            if(!empty($data[Contact::SHORTCUT]) &&
               ($contact = $this->getContacts( array(Contact::SHORTCUT => $data[Contact::SHORTCUT]), null, null, null, 2, null, null, $colleague)) > 0 &&
               (count($contact) > 1 || (count($contact) == 1 && $contact[0]->getID() != $data['rid']))){
                return [
                    'code' => 200,
                    'message' => 'Shortcut already exists'
                ];
            }
        }

        return true;
    }

    /**
     *
     * @var array
     */
    private $_keyMap = array(
            Contact::ID => 'rid',
            Contact::NAME => 'username',
            Contact::DOCUMENT_TYPE => 'user_passport_type',
            Contact::DOCUMENT_ID => 'user_passport_id',
            Contact::DOCUMENT_RELEASED => 'document_released_on',
            Contact::TYPE => 'user_class',
            Contact::PICTURE => 'image_url'
    );

    /**
     *
     * @param string $key
     * @return string
     */
    private function getRealKey($key){
        if(isset($this->_keyMap[$key])){
            return $this->_keyMap[$key];
        }
        return $key;
    }


    /**
     *
     * @var array
     */
    private $_availablePhoneBooks = array();

    /**
     *
     * @param Colleague $colleague
     * @param string $queryType
     */
    private function getAvailablePhonebooks(Colleague $colleague, $queryType){
        if(!isset($this->_availablePhoneBooks[$queryType])){
            $this->_availablePhoneBooks[$queryType] = array();
        }

        if(!isset($this->_availablePhoneBooks[$queryType][$colleague->getID()])){
            $this->_availablePhoneBooks[$queryType][$colleague->getID()] = array();

            $phonebooks = array();
            if($queryType == CTIconnectDB::SELECT_QUERY){
                $phonebooks = \Wildix\Phonebooks\Phonebooks::getInstance()->getPhonebooks(null, null, null, null, 1000, null, null, $colleague);
            }else{
                $filter = array(
                    \Wildix\Phonebooks\Phonebook::TYPE => array(
                        'not' => '3'
                    )
                );
                $phonebooks = \Wildix\Phonebooks\Phonebooks::getInstance()->getPhonebooks($filter, null, null, null, 1000, null, null, $colleague);
            }

            foreach ($phonebooks as $key=>$phonebook){
                $this->_availablePhoneBooks[$queryType][$colleague->getID()][] = $phonebook->getID();
            }
        }
        return $this->_availablePhoneBooks[$queryType][$colleague->getID()];
    }

    private function createWhere($filter = null, $search = null, $searchFields = null, Colleague $colleague = null, $queryType = null) {
        $filter        = is_array($filter) ? $filter : array();
        $search        = $search ? trim($search) : '';
        $searchFields  = $searchFields ?: self::SEARCH_FIELDS;
        $colleague     = is_object($colleague) ? $colleague : null;
        $fulltextQuery = '';

        // optimization phonebook query
        if ($colleague) {
            $availablePhonebooks = $this->getAvailablePhonebooks($colleague, $queryType);
            if(isset($filter['phonebook_id'])){
                if(is_scalar($filter['phonebook_id'])){
                    if(in_array($filter['phonebook_id'], $availablePhonebooks)){
                        $availablePhonebooks = array($filter['phonebook_id']);
                        unset($filter['phonebook_id']);
                    }else{
                        throw new Exception('Not found phonebook', 400);
                    }
                }elseif($filter['phonebook_id']['not']){
                    $pos = array_search(intval($filter['phonebook_id']['not']), $availablePhonebooks, true);
                    if($pos !== false){
                        unset($filter['phonebook_id']);
                        unset($availablePhonebooks[$pos]);
                    }
                }
            }
        }

        $where = array();
        if($filter && !empty($filter)){
            foreach ($filter as $key=>$value){
                $keys = explode(",",$key);
                if(count($keys) > 1){
                    $tmpWhere = array();
                    foreach ($keys as $key){
                        $tmpWhere[] = $this->filterCreate($key, $value);
                    }
                    if(!empty($tmpWhere)){
                        $where[] = '('.join(' OR ', $tmpWhere).')';
                    }
                }else{
                    $where[] = $this->filterCreate($key, $value);
                }
            }
        }

        if ($colleague){
            if (empty($availablePhonebooks)) {
                throw new AvailablePhonebooksException(sprintf('The user %d does not have any phonebooks available.', $colleague->getID()));
            }

            $where[] = '`'.$this->getRealKey(Contact::PHONEBOOK_ID).'` IN('.join(',', $availablePhonebooks).')';

            $groups = \Wildix\Colleagues\Groups::getInstance()->getGroups($colleague);
            $availGroups[] = '';
            if($groups){
                foreach($groups as $key => $group){
                    $availGroups[] = $group->getDn();
                }
            }
            if(count($availGroups)>0){
                $where[] = ' `group_id` IN("'.join('","', $availGroups).'")';
            }
        }

        if ($search && !empty($search) && count($searchFields)>0) {

            $notNumberSearch = preg_match('/[^0-9\*\#\+]/',$search);

            if ($notNumberSearch && $queryType == CTIconnectDB::SELECT_QUERY) {
                $fulltextQuery = $this->createWhereForFulltextTable($search, $searchFields);
            } else {
                $searchWhere = array();
                foreach ($searchFields as $key){
                    if($notNumberSearch && in_array($key, array(Contact::PHONE, Contact::FAX, Contact::MOBILE, Contact::SHORTCUT,
                            Contact::EXTENSION, Contact::OFFICE, Contact::HOME, Contact::HOME_MOBILE))){
                                continue;
                    }

                    if ($key == Contact::NAME) {
                        // search by words
                        $words = explode(' ', $search);
                        $words = array_map('trim', $words);
                        $words = array_filter($words, 'strlen');
                        $words = array_map(
                            function ($item) {
                                return '%'.CTIconnectDB::getInstance()->realEscapeString($item).'%';
                            }, $words);

                        $nameKey = $this->getRealKey(Contact::NAME);
                        $searchWhere[] =  "( `{$nameKey}` LIKE '".join("' AND `{$nameKey}` LIKE '", $words)."')";
                    }else{

                        $searchWhere[] = "`".$this->getRealKey($key)."` LIKE '%".CTIconnectDB::getInstance()->realEscapeString($search)."%'";
                    }
                }
                if(!empty($searchWhere)){
                    $where[] = '('.join(' OR ', $searchWhere).')';
                }
            }

        }

        if(count($where) > 0){
            return $fulltextQuery.' WHERE '.join(' AND ', $where);
        }
        return '';
    }

    private function createWhereForFulltextTable($search = null, $searchFields = null): string {
        $return = '';
        if ($search) {
            $searchFields = array_intersect(array(Contact::NAME, Contact::EMAIL, Contact::ORGANIZATION), $searchFields);
            if (count($searchFields) > 0) {
                $where  = [];
                $search = preg_replace('/[\\/\'\"\-\*\^\)\(]/', ' ', $search);
                foreach ($searchFields as $key) {
                    if ($this->getRealKey($key) === 'username') {
                        // WMS-4209. use multi phrase search. The 'trim' to avoid double asterisk in the end
                        $username = preg_replace('/ /', '*', trim($search));
                        $where[]  = sprintf('%s:%s*', $this->getRealKey($key), $username);
                    } else {
                        $where[] = sprintf('%s:%s*', $this->getRealKey($key), $search);
                    }
                }
                $return = ' JOIN (SELECT `rowid`
                            FROM `'.self::TABLE_NAME_FULLTEXT.'`
                            WHERE `'.self::TABLE_NAME_FULLTEXT.'` MATCH "'.implode(" OR ", $where).'"
                            ) as fts4 on fts4.`rowid` == `'.self::TABLE_NAME.'`.`rowid` ';
            }

        }
        return $return;
    }

    private function filterCreate($key, $value){
        if(is_scalar($value)){
            return "`".$this->getRealKey($key)."` = '".CTIconnectDB::getInstance()->realEscapeString($value)."'";
        }else if(is_array($value)){
            $tmpWhere = array();
            if(!empty($value['from'])){
                $tmpWhere[] = "`".$this->getRealKey($key)."` >= '".CTIconnectDB::getInstance()->realEscapeString($value['from'])."'";
            }
            if(!empty($value['to'])){
                $tmpWhere[] = "`".$this->getRealKey($key)."` <= '".CTIconnectDB::getInstance()->realEscapeString($value['to'])."'";
            }
            if(isset($value['not'])){
                $tmpWhere[] = "`".$this->getRealKey($key)."` <> '".CTIconnectDB::getInstance()->realEscapeString($value['not'])."'";
            }
            if(isset($value['or'])){
                if(is_array($value['or'])){
                    foreach($value['or'] as $k=>$v){
                        $value['or'][$k] = CTIconnectDB::getInstance()->realEscapeString($v);
                    }
                    $tmpWhere[] = "`".$this->getRealKey($key)."` IN ('".implode("','", $value['or'])."')";
                }else{
                    $tmpWhere[] = "`".$this->getRealKey($key)."` = '".CTIconnectDB::getInstance()->realEscapeString($value['or'])."'";
                }
            }

            if(!empty($tmpWhere)){
                return '('.join(' AND ', $tmpWhere).')';
            }
        }
        return '';
    }

    /**
     * Cast incoming data fields in concrete type
     *
     * @param array $data Incoming data.
     *
     * @return array
     */
    private function castNumberFieldsInData(array $data): array {
        $numericFields = [
            Contact::ID,
            Contact::PHONEBOOK_ID,
            Contact::PIN,
            Contact::CREDIT,
        ];

        foreach ($numericFields as $fieldToCast) {
            if (isset($data[$fieldToCast])) {
                $data[$fieldToCast] = intval($data[$fieldToCast]);
            }
        }

        return $data;
    }

    /**
     * Filter incoming data for add or update contact
     *
     * @param array $data Incoming data.
     *
     * @return array
     */
    private function filterData(array $data): array {
        $data = array_map('trim', $data);
        $data = array_intersect_key($data, array_fill_keys(Contact::ALLOWED_FIELDS, null));


        $data['mobile']      = $this->preparePhoneNumber($data['mobile'] ?? '');
        $data['home_mobile'] = $this->preparePhoneNumber($data['home_mobile'] ?? '');
        $data['office']      = $this->preparePhoneNumber($data['office'] ?? '');
        $data['home']        = $this->preparePhoneNumber($data['home'] ?? '');
        $data['fax']         = $this->preparePhoneNumber($data['fax'] ?? '');
        $data['postal_code'] = $this->preparePostCode($data['postal_code'] ?? '');

        return $this->castNumberFieldsInData($data);
    }

    private function preparePhoneNumber(string $phone = ''): string {
        return preg_replace("/[^+a-zA-Z\d]/", '', $phone);
    }

    private function preparePostCode(string $code): string {
        return preg_replace("/[^a-zA-Z\-\d]/", '', $code);
    }

    private function db(): ConnectionInterface {
        return DB::connection(CTIconnectDB::class);
    }
}

