<?php

namespace Wildix\Phonebooks;

use Wildix\Exception;
use Wildix\Normalizer\PhoneNumberNormalizer;

/**
 * Class ImportContactBase
 */
abstract class ImportContactAbstract {

    /**
     * @var \Wildix\Normalizer\PhoneNumberNormalizer
     */
    private $_phoneNormalizer;

    public function __construct(PhoneNumberNormalizer $phoneNormalizer) {
        $this->_phoneNormalizer = $phoneNormalizer;
    }

    abstract public function parseContact($contact): array;

    public function mapFields(array $data, array $map): array {
        $contact = [
            'id'             => $data['id'],
            'birthday'       => $data['birthday'],
            'notes'          => $data['notes'],
            'name'           => trim($data['name']['FirstName']) . ' ' . trim($data['name']['MiddleName']) . ' ' . trim($data['name']['LastName']),
            'firstName'      => $data['name']['FirstName'],
            'middleName'     => $data['name']['MiddleName'],
            'lastName'       => $data['name']['LastName'],
            'nameTitle'      => $data['name']['Title'],
            'website'        => $data['website'],
            'email1'         => $data['email'][1] ?? '',
            'email2'         => $data['email'][2] ?? '',
            'email3'         => $data['email'][3] ?? '',
            'phoneHome1'     => $data['phone']['Home1'] ?? '',
            'phoneHome2'     => $data['phone']['Home2'] ?? '',
            'phoneBusiness1' => $data['phone']['Business1'] ?? '',
            'phoneBusiness2' => $data['phone']['Business2'] ?? '',
            'phoneMobile1'   => $data['phone']['Mobile1'] ?? '',
            'company'        => $data['company'] ?? '',
            'addressHome'    => trim($data['HomeAddress']['street'] ?? ''),
            'cityHome'       => trim($data['HomeAddress']['city'] ?? ''),
            'stateHome'      => trim($data['HomeAddress']['state'] ?? ''),
            'regionHome'     => trim($data['HomeAddress']['countryOrRegion'] ?? ''),
            'postcodeHome'   => trim($data['HomeAddress']['postalCode'] ?? ''),
            'addressWork'    => trim($data['BusinessAddress']['street'] ?? ''),
            'cityWork'       => trim($data['BusinessAddress']['city'] ?? ''),
            'stateWork'      => trim($data['BusinessAddress']['state'] ?? ''),
            'regionWork'     => trim($data['BusinessAddress']['countryOrRegion'] ?? ''),
            'postcodeWork'   => trim($data['BusinessAddress']['postalCode'] ?? ''),
            'addressOther'   => trim($data['OtherAddress']['street'] ?? ''),
            'cityOther'      => trim($data['OtherAddress']['city'] ?? ''),
            'stateOther'     => trim($data['OtherAddress']['state'] ?? ''),
            'regionOther'    => trim($data['OtherAddress']['countryOrRegion'] ?? ''),
            'postcodeOther'  => trim($data['OtherAddress']['postalCode'] ?? ''),
        ];

        $externalRecord = [];
        foreach ($map as $internal => $external) {
            // fix fields for contacts class
            if ($internal === 'username') {
                $internal = 'name';
            }

            if (!empty($external)) {
                $value = $this->processMappingSyntax($contact, $external);

                if (!empty($value)) {
                    $value = preg_replace('/\s{2,}/', ' ', $value);
                }
                if (strtolower($internal) !== 'note') {
                    $value = $this->removeNewLines($value);
                }

                $externalRecord[$internal] = $this->normalizeValue($internal, $value);
            }
        }

        if (!empty($contact['birthday'])) {
            $birthday = new \DateTime($contact['birthday']);
            if ($birthday !== false) {
                $externalRecord['date_birth'] = $birthday->format('m/d/Y');
            }
        } else {
            $externalRecord['date_birth'] = '';
        }

        return $this->validateInputRecord($this->normalizeRecordData($externalRecord));
    }

    /**
     * Find value for the provided $sourceColumnName in a $sourceRow with ability to use special syntax
     *
     * @param array  $sourceRow        Imported row.
     * @param string $sourceColumnName Column name from imported backend, can contain special syntax with + and |.
     * @param array  $sourceHeaders    Appliable for csv, when col name have to be calculated by its position.
     *
     * @return null|string - returns value for the provided coulumn name or null if $sourceColumnName was not found
     */
    protected function processMappingSyntax($sourceRow, $sourceColumnName, $sourceHeaders = []) {
        $result             = null;
        $processOr          = true;
        $processConcat      = true;
        $selectedColumnName = null;

        // Applicable for csv
        foreach ($sourceHeaders as $sourceHeaderPos => $sourceHeaderName) {
            if (isset($sourceRow[$sourceHeaderPos])) {
                $sourceRow[$sourceHeaderName] = trim($sourceRow[$sourceHeaderPos]);
            }
        }

        // If no special syntax symbols then just return value
        if ($sourceColumnName && strpos($sourceColumnName, '+') === false && strpos($sourceColumnName, '|') === false) {
            if (isset($sourceRow[$sourceColumnName])) {
                $result = trim($sourceRow[$sourceColumnName]);
            }
            return $result;
        }

        // Split for OR syntax
        $executionStack = $processOr ? explode('|', $sourceColumnName) : [$sourceColumnName];

        // Split for CONCAT syntax
        foreach ($executionStack as $keyOr => $columnNameOr) {
            $executionStack[$keyOr] = $processConcat ? explode('+', $columnNameOr) : [$columnNameOr];

            $result = '';
            foreach ($executionStack[$keyOr] as $keyConcat => $columnNameConcat) {
                $columnNameConcat = trim($columnNameConcat);
                if (array_search(preg_replace('/\s+/', ' ', $columnNameConcat), ["' '", '" "'])) {
                    $result .= ' ';
                } else if (isset($sourceRow[$columnNameConcat])) {
                    if (!empty($result) && substr($result, -1) !== ' ') {
                        $result .= ' ';
                    }
                    $result .= $sourceRow[$columnNameConcat];

                    if (trim($sourceRow[$columnNameConcat])) {
                        // Selected column name used for applying city, postal code , so here just do not loose important column name
                        if (!array_search($selectedColumnName, ['addressHome', 'addressWork', 'addressBusiness', 'addressOther'])) {
                            $selectedColumnName = $columnNameConcat;
                        }
                    }
                }
            }
            $result = trim($result);
            if (!empty($result)) {
                break;
            }
        }

        return $result;
    }

    protected function normalizeValue($key, $value) {
        if (in_array($key, ['phone', 'fax', 'mobile', 'extension', 'office', 'home', 'home_mobile'])) {
            return $this->_phoneNormalizer->normalize($this->filterTheNumber($value), true);
        } else if ($value && mb_detect_encoding($value . ' ') !== 'UTF-8') {
            // During mb_detect_encoding we add space to the end of string because result will wrong if multibite symbol at the end of string!
            return mb_convert_encoding($value, 'UTF-8');
        }

        return $value;
    }

    protected function filterTheNumber($number) {
        return preg_replace('/[^\+0-9a-zA-Z]/', '', $number);
    }

    protected function removeNewLines($value) {
        return trim(preg_replace('/\s+/', ' ', $value));
    }

    protected function normalizeRecordData(array $data): array {
        foreach ($data as $_key => &$_item) {
            switch ($_key) {
                case \Wildix\Phonebooks\Contact::NOTE:
                    // remove all control chars except LF(line feed), CR(carriage return)
                    $_item = preg_replace('/[\x{0}-\x{9}\x{0B}\x{0C}\x{0E}-\x{1F}]/', ' ', $_item);
                    break;
                default:
                    $_item = preg_replace('/[\x{0}-\x{1F}]/', ' ', $_item);
                    break;
            }
        }

        return $data;
    }

    protected function validateInputRecord($externalRecord) {
        $checkFields = [
            'phone'       => '/^([\+]?[0-9\*\#]{2,50}\z)?/',
            'office'      => '/^([\+]?[0-9\*\#]{2,50}\z)?/',
            'home'        => '/^([\+]?[0-9\*\#]{2,50}\z)?/',
            'extension'   => '/^([\+]?[0-9\*\#]{2,50}\z)?/',
            'home_mobile' => '/^([\+]?[0-9\*\#]{2,50}\z)?/',
            'fax'         => '/^([\+]?[0-9\*\#]{2,50}\z)?/',
            'email'       => '/^([^\W].*\@.*\.[a-z]{2,4}\z)?/i',
            'mobile'      => '/^([\+]?[0-9\*\#]{2,50}\z)?/',
        ];

        if (empty($externalRecord['name'])) {
            throw new Exception('Empty username');
        }

        foreach ($checkFields as $field => $pattern) {
            if (isset($externalRecord[$field]) && !preg_match($pattern, $externalRecord[$field])) {
                unset($externalRecord[$field]);
            }
        }

        return $externalRecord;
    }
}
