<?php
namespace Wildix\Phonebooks;

use Wildix\Colleagues\Colleague;
use Wildix\Databases\CTIconnectDB;
use Wildix\Error;
use Wildix\InvalidDataException;

class Phonebooks extends \Singleton
{
    const TABLE_NAME = 'cti_phonebooks';

    private $_tableFields = '`rid` as id, `name`, `type`, `hash`, `owner_id`, `sync_set`, `sync_interval`, `sync_params`';

    /**
     *
     * @param int            $id
     * @param Colleague|null $colleague
     *
     * @return Phonebook
     * @throws InvalidDataException
     */
    public function getByID(int $id, Colleague $colleague = null): Phonebook {
        $phonebooks = $this->getPhonebooks(array(Phonebook::ID => $id), null, null, null, 1, null, null, $colleague);

        if (empty($phonebooks)) {
            throw new InvalidDataException(Error::NotFound, [], 404);
        }
        return $phonebooks[0];
    }

    /**
     *
     * @param null           $filter array
     * @param null           $search
     * @param null           $searchFields
     * @param null           $start
     * @param null           $count
     * @param null           $sort
     * @param null           $dir    (ASC | DESC)
     * @param Colleague|null $colleague
     *
     * @return array <\Wildix\Phonebooks\Phonebook>
     */
    public function getPhonebooks($filter = null, $search=null, $searchFields=null, $start = null, $count=null, $sort=null, $dir= null, Colleague $colleague = null){
        $start = is_numeric($start) ? intval($start) : 0;
        $count = is_numeric($count) ? intval($count) : 100;
        $sort = $sort ?: Phonebook::NAME;
        $dir = $dir ?: 'ASC';

        $phonebooks = array();

        $query = 'SELECT '.$this->_tableFields.'
                                FROM `'.self::TABLE_NAME.'`
                                '.$this->createWhere($filter, $search, $searchFields, $colleague, CTIconnectDB::SELECT_QUERY).'
                                ORDER BY '.$sort.' '.$dir.'
                                LIMIT '.$start.', '.$count.';';

        $rows = CTIconnectDB::getInstance()->query($query);
        if ($rows) {
            while($row = $rows->fetchArray(SQLITE3_ASSOC)){
                if ($row && !empty($row)) {
                    $phonebooks[] = new Phonebook($row);
                }
            }
        }
        return $phonebooks;
    }

    /**
     *
     * @param array     $filter
     * @param string    $search
     * @param array     $searchFields
     * @param Colleague $colleague
     *
     * @return int
     */
    public function getCount($filter, $search=null, $searchFields = null, Colleague $colleague = null){

        $query = 'SELECT count(rid) as countRows
                                FROM `'.self::TABLE_NAME.'`
                                '.$this->createWhere($filter, $search, $searchFields, $colleague, CTIconnectDB::SELECT_QUERY).';';

        $rows = CTIconnectDB::getInstance()->query($query);
        if ($rows) {
            $row = $rows->fetchArray(SQLITE3_ASSOC);
            if($row){
                return $row['countRows'];
            }
        }

        return 0;
    }

    /**
     *
     * @param array $phonebook
     * @param Colleague $colleague
     *
     * @return int
     */
    public function addPhonebook($data, Colleague $colleague = null, $system = false){
        $data = array_map('trim', $data);

        if(!isset($data[Phonebook::TYPE])){
            $data[Phonebook::TYPE] = Phonebook::TYPE_PERSONAL;
        }

        $isValid = $this->isValidData($data, $colleague, CTIconnectDB::INSERT_QUERY);
        if($isValid !== true){
            throw new \Wildix\Exception($isValid['message'], $isValid['code']);
        }

        if($colleague){
            $data[Phonebook::OWNER_ID] = $colleague->getID();
        }

        if(!isset($data[Phonebook::HASH])){
            $data[Phonebook::HASH] = md5($data[Phonebook::NAME]);
        }

        $data = $this->_prepareData($data, $colleague, $system);

        $keys = array();
        $values = array();
        foreach ($data as $key=>$value){
            $keys[] = '`'.$this->getRealKey($key).'`';
            $values[] = '\''.CTIconnectDB::getInstance()->realEscapeString($value).'\'';
        }

        $query = sprintf('INSERT INTO `%s` (%s) Values (%s)', self::TABLE_NAME, join(', ', $keys), join(', ', $values));

        $result = CTIconnectDB::getInstance()->exec($query);

        if($result === true){
            return CTIconnectDB::getInstance()->getLastInsertId();
        }

        return $result;
    }

    private function _prepareData(array $data, Colleague $colleague = null, bool $system = false): array {
        if ($system === true) {
            return $data;
        }

        if ($colleague === null || !$colleague->isAdministrator()) {
            $removedFields = [
                Phonebook::SYNC,
                Phonebook::SYNC_INTERVAL,
                Phonebook::SYNC_PARAMS,
            ];
            foreach ($removedFields as $key) {
                if (isset($data[$key])) {
                    unset($data[$key]);
                }
            }
        }

        return $data;
    }

    /**
     *
     * @param $filter array
     * @param string $search
     * @param array $searchFields
     * @param Colleague $colleague
     *
     * @return int
     */
    public function deletePhonebooks($filter = null, $search = null, $searchFields=null, Colleague $colleague = null){
        $result = array(
                'phonebooks'=>0,
                'contacts' => 0
        );

        $phonebooks = $this->getPhonebooks($filter, $search, $searchFields, 0, $count=99999, null, null, $colleague);
        if(count($phonebooks) == 0){
            throw new \Wildix\Exception('Not found phonebook', 400);
        }

        foreach ($phonebooks as $phonebook){
            $contactFilter = array(
                    \Wildix\Phonebooks\Contact::PHONEBOOK_ID => $phonebook->getID()
            );
            $result['contacts'] += \Wildix\Phonebooks\Contacts::getInstance()->deleteContacts($contactFilter, null, null, $colleague);
        }

        $query = 'DELETE FROM `'.self::TABLE_NAME.'`
                                '.$this->createWhere($filter, $search, $searchFields, $colleague, CTIconnectDB::DELETE_QUERY).';';

        $resultQuery = CTIconnectDB::getInstance()->exec($query);
        if ($resultQuery === true) {
            $result['phonebooks'] = CTIconnectDB::getInstance()->getCountChangesRows();
        }
        return $result;
    }

    /**
     *
     * @param array     $data
     * @param Colleague $colleague
     *
     * @return boolean
     */
    public function updatePhonebooks($data, $filter = null, $search = null, $searchFields=null, Colleague $colleague = null, $system = false){
        $data = array_map('trim', $data);

        $data = $this->_prepareData($data, $colleague, $system);

        if(count($data) == 0){
            throw new \Wildix\Exception('Phonebook data cannot be empty', 0);
        }

        $isValid = $this->isValidData($data, $colleague, CTIconnectDB::UPDATE_QUERY);
        if($isValid !== true){
            throw new \Wildix\Exception($isValid['message'], $isValid['code']);
        }

        if($filter && $filter[Phonebook::ID]){
            $ph = $this->getByID($filter[Phonebook::ID],$colleague);

            if($colleague && $ph->getOwner() != $colleague->getID()){
                throw new \Wildix\Exception('No access for edit this phonebook', 400);
            }
        }

        $params = array();
        foreach ($data as $key=>$value){
            $params[] = "`".$this->getRealKey($key)."` = '".CTIconnectDB::getInstance()->realEscapeString($value)."'";
        }

        $query = 'UPDATE `'.self::TABLE_NAME.'` SET '.join(', ', $params).'
                                '.$this->createWhere($filter, $search, $searchFields, $colleague, CTIconnectDB::UPDATE_QUERY).';';

        $result = CTIconnectDB::getInstance()->exec($query);
        if ($result === true) {
            $result = CTIconnectDB::getInstance()->getCountChangesRows();
        }
        return $result;
    }

    private function isValidData($data, Colleague $colleague = null, $queryType){

        if($queryType == CTIconnectDB::INSERT_QUERY){
            if(!in_array($data[Phonebook::TYPE], array(Phonebook::TYPE_PERSONAL,
                    Phonebook::TYPE_GLOBAL,
                    Phonebook::TYPE_COLLEAGUE))){
                return array(
                        'code' => 400,
                        'message' => 'The type can be 1 (PERSONAL), 2 (GLOBAL) or 3 (COLLEAGUE)'
                );
            }
            if($data[Phonebook::TYPE] != Phonebook::TYPE_PERSONAL && !$colleague->isAdministrator()){
                return array(
                        'code' => 403,
                        'message' => 'No permission to add a global phone book'
                );
            }

            if($data[Phonebook::TYPE] == Phonebook::TYPE_COLLEAGUE){
                return array(
                        'code' => 403,
                        'message' => 'No permission to add a LDAP phone book'
                );
            }

            if(empty($data[Phonebook::NAME])){
                return array(
                        'code' => 400,
                        'message' => 'The name cannot be empty'
                );
            }

        }else if($queryType == CTIconnectDB::UPDATE_QUERY){
            if(isset($data[Phonebook::TYPE])){
                if(!in_array($data[Phonebook::TYPE], array(Phonebook::TYPE_PERSONAL,
                        Phonebook::TYPE_GLOBAL,
                        Phonebook::TYPE_COLLEAGUE))){
                    return array(
                            'code' => 400,
                            'message' => 'The type can be 1 (PERSONAL), 2 (GLOBAL) or 3 (COLLEAGUE)'
                    );
                }

                if($data[Phonebook::TYPE] != Phonebook::TYPE_PERSONAL && !$colleague->isAdministrator()){
                    return array(
                            'code' => 400,
                            'message' => 'No permission to update a global phone book'
                    );
                }

                if($data[Phonebook::TYPE] == Phonebook::TYPE_COLLEAGUE){
                    return array(
                            'code' => 403,
                            'message' => 'No permission to update a LDAP phone book'
                    );
                }
            }
        }

        return true;
    }


    /**
     *
     * @var array
     */
    private $_keyMap = array(
            Phonebook::ID => 'rid'
    );

    /**
     *
     * @param string $key
     * @return string
    */
    private function getRealKey($key){
        if(isset($this->_keyMap[$key])){
            return $this->_keyMap[$key];
        }
        return $key;
    }

    /**
     *
     * @param array     $filter
     * @param string    $search
     * @param array     $searchFields
     * @param Colleague $colleague
     *
     * @return string
     */
    private function createWhere($filter = null, $search = null, $searchFields = null, Colleague $colleague = null, $queryType = null){

        $filter = is_array($filter) ? $filter : array();
        $search = $search ? trim($search) : '';
        $searchFields = $searchFields ? $searchFields : array(Phonebook::NAME);
        $colleague = is_object($colleague) ? $colleague : null;

        $where = array();
        if($filter && !empty($filter)){
            foreach ($filter as $key=>$value){
                if(is_scalar($value)){
                    $where[] = "`".$this->getRealKey($key)."` = '".CTIconnectDB::getInstance()->realEscapeString($value)."'";
                }else if(is_array($value)){
                    $tmpWhere = array();
                    if(!empty($value['from'])){
                        $tmpWhere[] = "`".$this->getRealKey($key)."` >= '".CTIconnectDB::getInstance()->realEscapeString($value['from'])."'";
                    }
                    if(!empty($value['to'])){
                        $tmpWhere[] = "`".$this->getRealKey($key)."` <= '".CTIconnectDB::getInstance()->realEscapeString($value['to'])."'";
                    }
                    if(isset($value['not'])){
                        $tmpWhere[] = "`".$this->getRealKey($key)."` <> '".CTIconnectDB::getInstance()->realEscapeString($value['not'])."'";
                    }
                    if(isset($value['in'])){
                        $in = array_map(function ($item) {
                            return CTIconnectDB::getInstance()->realEscapeString($item);
                        }, $value['in']);

                        $tmpWhere[] = "`" . $this->getRealKey($key) . "` in  ('". implode("','", $in) . "')";
                    }

                    if(!empty($tmpWhere)){
                        $where[] = '('.join(' AND ', $tmpWhere).')';
                    }
                }
            }
        }

        if ($colleague){
            $ownerKey = $this->getRealKey(Phonebook::OWNER_ID);
            $typeKey = $this->getRealKey(Phonebook::TYPE);
            $idKey = $this->getRealKey(Phonebook::ID);

            if($queryType == CTIconnectDB::SELECT_QUERY){
                $phonebooksId = $colleague->getSettings()->getPhonebooks();
                if($colleague->canSet($colleague::PHONEBOOKS) || count($phonebooksId) == 0){
                    //if in features are not selected phone books then show all
                    $where[] = '((`'.$ownerKey.'` = '.$colleague->getID().' AND `'.$typeKey.'` = '.Phonebook::TYPE_PERSONAL.') OR `'.$typeKey.'` = '.Phonebook::TYPE_COLLEAGUE.' OR `'.$typeKey.'` = '.Phonebook::TYPE_GLOBAL.')';
                }else{
                    //if in features are selected phone books then show all selected and LDAP
                    $phonebooksId = join(',',$phonebooksId);
                    $where[] = '((`'.$ownerKey.'` = '.$colleague->getID().' AND `'.$typeKey.'` = '.Phonebook::TYPE_PERSONAL.') OR `'.$typeKey.'` = '.Phonebook::TYPE_COLLEAGUE.' OR `'.$idKey.'` IN('.$phonebooksId.'))';
                }
            }else{
                if($colleague->isAdministrator()){
                    $where[] = '(`'.$ownerKey.'` = '.$colleague->getID().' AND (`'.$typeKey.'` = '.Phonebook::TYPE_PERSONAL.' OR `'.$typeKey.'` = '.Phonebook::TYPE_GLOBAL.'))';
                }else{
                    $where[] = '(`'.$ownerKey.'` = '.$colleague->getID().' AND `'.$typeKey.'` = '.Phonebook::TYPE_PERSONAL.')';
                }

            }
        }

        if($search && !empty($search) && count($searchFields) > 0){
            $searchWhere = array();

            foreach($searchFields as $key){
                $searchWhere[] = '`'.$this->getRealKey($key)."` LIKE '%".CTIconnectDB::getInstance()->realEscapeString($search)."%'";
            }
            if(!empty($searchWhere)){
                $where[] = '('.join(' OR ', $searchWhere).')';
            }
        }

        if(count($where) > 0){
            return 'WHERE '.join(' AND ', $where);
        }

        return '';
    }
}
