<?php

namespace Wildix\Phonebooks;

use Logger\Log;
use Wildix\Colleagues\Colleague;
use Wildix\Colleagues\Colleagues;
use Wildix\Microsoft\GraphCollection;
use Beta\Microsoft\Graph\Model as BetaModel;
use Wildix\Microsoft\ImportContactOffice365;
use Wildix\Phonebooks\Sync\SyncSettings;

/**
 * Class PhonebooksService
 */
class PhonebooksService {

    const SYNC_WITH_ADMIN_CONSENT_MODE = 2;

    /**
     * @var Phonebooks
     */
    private $_phonebooksClass;

    /**
     * @var Contacts
     */
    private $_contactsClass;

    /**
     * @var ImportContactOffice365
     */
    private $_importContact;

    /**
     * @var Log
     */
    private $_logger;

    private $_phonebookTemplateName = 'MS365/%s';

    private $_localContacts = [];

    public function __construct(Phonebooks $phonebooks, Contacts $contacts, ImportContactOffice365 $import, Log $logger) {
        $this->_phonebooksClass = $phonebooks;
        $this->_contactsClass   = $contacts;
        $this->_importContact   = $import;
        $this->_logger          = $logger;
    }

    public function processContactFolder(BetaModel\ContactFolder $contactFolder, Colleague $user): Phonebook {
        $msFolderId    = $contactFolder->getId();
        $phonebookName = $this->_makePhonebookName($contactFolder->getDisplayName());

        $localPhonebook = $this->getPersonalPhonebookByMsId($msFolderId, $user);
        if ($localPhonebook === null) {
            // add phonebook
            $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Add new phonebook "%s" for user %s', $phonebookName, $user->getName()));

            $syncSettings = new SyncSettings();
            $syncSettings->setSettings(
                [
                    'phonebook' => $msFolderId,
                    'msUserId' => $user->getSourceId(),
                    'userId' => $user->getID(),
                    'updateTime' => time(),
                    'deltaLink' => '',
                ]
            );

            $data        = [
                Phonebook::TYPE => Phonebook::TYPE_PERSONAL,
                Phonebook::NAME => $phonebookName,
                Phonebook::HASH => $msFolderId,
                Phonebook::SYNC => self::SYNC_WITH_ADMIN_CONSENT_MODE,
                Phonebook::SYNC_INTERVAL => 1,
                Phonebook::SYNC_PARAMS => json_encode($syncSettings->asArray())
            ];
            $phonebookId    = $this->_phonebooksClass->addPhonebook($data, $user, true);
            $localPhonebook = $this->_phonebooksClass->getByID($phonebookId); // TODO:: check
        } else {
            // update name
            $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Update phonebook "%s" for user %s', $phonebookName, $user->getName()));

            $this->_phonebooksClass->updatePhonebooks([Phonebook::NAME => $phonebookName], [Phonebook::ID => $localPhonebook->getId()]);
        }

        return $localPhonebook;
    }

    public function saveContactFolderDeltaLink(Phonebook $phonebook, string $link) {
        $syncSettings = new SyncSettings($phonebook->getSyncParams());
        $syncSettings->setDeltaLink($link)
            ->setUpdateTime();

        $this->_phonebooksClass->updatePhonebooks(
            [Phonebook::SYNC_PARAMS => json_encode($syncSettings->asArray())],
            [Phonebook::ID => $phonebook->getID()],
            null,
            null,
            null,
            true
        );
    }

    public function initPhonebook(Phonebook $phonebook): void {
        $this->_localContacts = $this->getContactsInPersonalPhonebook($phonebook->getID());
    }

    public function deleteNotSyncedContacts() {
        $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Delete not synced contacts %s', count($this->_localContacts)));
        if (!empty($this->_localContacts)) {
            // need remove in the local db deleted contacts
            foreach ($this->_localContacts as $contact) {
                $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Remove in the local db deleted contact "%s"', $contact->getName()));
                $this->_contactsClass->deleteContacts([Contact::ID => $contact->getId()]);
            }
        }
    }

    public function processContacts($contacts, Phonebook $phonebook, Colleague $user, SyncSettings $syncSettings) {
        $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Start processing contacts "%s"', count($contacts)));
        foreach ($contacts as $msContact) {
            $this->processContact($msContact, $phonebook, $user, $syncSettings->getMapping());
            if (array_key_exists($msContact->getId(), $this->_localContacts)) {
                unset($this->_localContacts[$msContact->getId()]);
            }
        }
    }

    public function processContactsDelta($contacts, Phonebook $phonebook, SyncSettings $syncSettings) {
        $user = Colleagues::getInstance()->getUserByID($syncSettings->getUserId());
        if ($user === null) {
            throw new \Wildix\Exception(sprintf('[office365phonebooks] Not found user %s', $syncSettings->getUserId()));
        }

        $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Start processing contacts delta "%s"', count($contacts)));
        foreach ($contacts as $msContact) {
            /** @var BetaModel\Contact $msContact */
            if (array_key_exists('@removed', $msContact->getProperties())) {
                $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Delete contact "%s" in phonebook "%s"', $msContact->getId(), $phonebook->getName()));
                $this->_contactsClass->deleteContacts(
                    [
                        Contact::SOURCE_ID => $msContact->getId(),
                        Contact::PHONEBOOK_ID => $phonebook->getID()
                    ]
                );
            } else {
                $this->processContact($msContact, $phonebook, $user, $syncSettings->getMapping());
            }
        }
    }

    public function processContact(
        BetaModel\Contact $msContact,
        Phonebook $phonebook,
        Colleague $user,
        array $mapping
    ) {
        $localContact = $this->getContactBySourceId($msContact->getId(), $this->_localContacts);

        $data = $this->_buildContactData($msContact, $phonebook, $user, $mapping);

        if ($localContact) {
            $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Update contact "%s" in the phonebook "%s"', $msContact->getDisplayName(), $phonebook->getID()));
            $this->_contactsClass->updateContacts(
                $data,
                [
                    Contact::ID => $localContact->getId(),
                    Contact::PHONEBOOK_ID => $phonebook->getID(),
                ]
            );
        } else {
            $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Add new contact "%s" to the phonebook "%s"', $msContact->getDisplayName(), $phonebook->getID()));
            $this->_contactsClass->addContact($data, $user);
        }
    }

    private function _buildContactData(BetaModel\Contact $msContact, Phonebook $phonebook, Colleague $user, array $mapping): array {
        $contact = $this->_importContact->mapFields(
            $this->_importContact->parseContact($msContact),
            $mapping
        );

        $contact[Contact::SOURCE_ID]    = $msContact->getId();
        $contact[Contact::PHONEBOOK_ID] = $phonebook->getId();
        $contact[Contact::OWNER_ID]     = sprintf('uid=%s,o=com0,dc=wildix', $user->getID());
        $this->_logger->debug(__LINE__, sprintf('[office365phonebooks] Contact "%s"', json_encode($contact)));

        return $contact;
    }

    public function getPersonalPhonebooks(Colleague $user = null): array {
        $data = $this->_phonebooksClass->getPhonebooks(
            [
                Phonebook::TYPE => Phonebook::TYPE_PERSONAL,
                Phonebook::SYNC => self::SYNC_WITH_ADMIN_CONSENT_MODE,
            ],
            null,
            null,
            null,
            null,
            null,
            null,
            $user
        );

        $items = [];
        foreach ($data as $item) {
            $items[$item->getHash()] = $item;
        }

        return $items;
    }

    private function _makePhonebookName(string $name): string {
        return sprintf($this->_phonebookTemplateName, $name);
    }

    public function getPersonalPhonebookByMsId(string $msPhonebookId, Colleague $user): ?Phonebook {
        $localPhonebooks = $this->getPersonalPhonebooks($user);

        return array_key_exists($msPhonebookId, $localPhonebooks) ? $localPhonebooks[$msPhonebookId] : null;
    }

    public function getContactsInPersonalPhonebook(int $phonebookId, Colleague $user = null): array {
        $contacts = [];

        $items = $this->_contactsClass->getContacts(
            [Contact::PHONEBOOK_ID => $phonebookId],
            null,
            null,
            null,
            1000000,
            null,
            null,
            $user
        );

        foreach ($items as $contact) {
            $contacts[$contact->getSourceID()] = $contact;
        }

        return $contacts;
    }

    public function getContactBySourceId(string $sourceId, $contacts): ?Contact {
        return array_key_exists($sourceId, $contacts) ? $contacts[$sourceId] : null;
    }

    public function removePhonebooks(array $phonebooks) {
        $ids = array_map(
            function ($item) {
                return $item->getId();
            },
            $phonebooks
        );

        $this->_logger->info(__LINE__, sprintf('[office365phonebooks] Remove in the local db deleted phonebooks %s', implode(', ', $ids)));
        $this->_phonebooksClass->deletePhonebooks([Phonebook::ID => ['in' => $ids]]);
    }
}
