<?php
namespace Wildix\Services;

/**
 * Service implementing the basic logic of working with wildix micro services
 */
abstract class AbstractService {

    private $_basicAuth = true;
    private $_sslVerify = true;
    /**
     * Logger instance
     *
     * @var \Syslog|null
     */
    private $_log;
    /**
     * User-Agent
     *
     * @var string
     */
    private $_userAgent;
    /**
     * Value for CURLOPT_CONNECTTIMEOUT
     *
     * @var integer
     */
    private $_connectTimeout;
    /**
     * Value for CURLOPT_TIMEOUT
     *
     * @var integer
     */
    private $_timeout = 120;

    private $_logLevel;

    protected $_httpProxy;

    protected $_useHttpProxy = true;

    public function __construct($logLevel = \Logger\Log::DEBUG, \Wildix\System\HttpProxy $httpProxy = null) {
        $this->_logLevel = $logLevel;

        $this->_httpProxy = isset($httpProxy) ? $httpProxy : new \Wildix\System\HttpProxy();

        $this->setUserAgent('Wildix PBX (' . WMS_VERSION_FULL . ')');
        $this->setConnectTimeout(3);
    }

    public function enableBasicAuth() {
        $this->_basicAuth = true;
    }

    public function disableBasicAuth() {
        $this->_basicAuth = false;
    }

    public function disableSslVerify() {
        $this->_sslVerify = false;
    }

    protected function setUserAgent($userAgent) {
        $this->_userAgent = $userAgent;
    }

    protected function setConnectTimeout($timeout) {
        $this->_connectTimeout = $timeout;
    }

    protected function setTimeout($timeout) {
        $this->_timeout = (int)$timeout;
    }

    protected function getBasicAuthCredentials() {
        return $this->getPbx()->getSerial() . ':' . $this->getPbx()->getPbxKey();
    }

    protected function getLog() {
        if ($this->_log === null) {
            $this->_log = new \Syslog(get_called_class(), $this->_logLevel);
        }
        return $this->_log;
    }

    public function getPbx() {
        return \Models\PBX::getInstance();
    }

    protected function makeRequest($type, $url, $params, $formData, $headers, $curlOptions = []) {
        $params = $params ?? [];
        $curl   = curl_init();
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, $this->_sslVerify);
        curl_setopt($curl, CURLOPT_SSL_VERIFYHOST, $this->_sslVerify ? 2 : 0);
        curl_setopt($curl, CURLOPT_USERAGENT, $this->_userAgent);
        if ($this->_basicAuth) {
            curl_setopt($curl, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
            curl_setopt($curl, CURLOPT_USERPWD, $this->getBasicAuthCredentials());
        }
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_CONNECTTIMEOUT, $this->_connectTimeout);
        curl_setopt($curl, CURLOPT_TIMEOUT, $this->_timeout);
        curl_setopt($curl, CURLOPT_AUTOREFERER, true);

        if (is_array($curlOptions) && count($curlOptions)) {
            curl_setopt_array($curl, $curlOptions);
        }

        $this->addProxySettings($curl);

        switch ($type) {
            case 'POST':
                curl_setopt($curl, CURLOPT_URL, $url);
                curl_setopt($curl, CURLOPT_POST, true);
                if ($formData) {
                    curl_setopt($curl, CURLOPT_POSTFIELDS, $params);
                } else {
                    curl_setopt($curl, CURLOPT_POSTFIELDS, http_build_query($params));
                }
                break;
            case 'PUT':
                curl_setopt($curl, CURLOPT_URL, $url);
                curl_setopt($curl, CURLOPT_CUSTOMREQUEST, 'PUT');
                if ($formData) {
                    curl_setopt($curl, CURLOPT_POSTFIELDS, $params);
                } else {
                    curl_setopt($curl, CURLOPT_POSTFIELDS, http_build_query($params));
                }
                break;
            case 'DELETE':
                curl_setopt($curl, CURLOPT_URL, $url);
                curl_setopt($curl, CURLOPT_CUSTOMREQUEST, 'DELETE');
                break;
            default:
                curl_setopt($curl, CURLOPT_URL, $url . '?' . http_build_query($params));
                curl_setopt($curl, CURLOPT_HTTPGET, true);
                break;
        }

        $this->getLog()->info(
            __LINE__,
            sprintf('Service %s request %s', get_class($this), $url)
        );

        curl_setopt($curl, CURLOPT_HTTPHEADER, $headers);
        $response = array(
            'result' => curl_exec($curl),
            'info' => curl_getinfo($curl),
            'error' => array(
                'reason' => curl_error($curl),
                'code' => curl_errno($curl)
            )
        );
        curl_close($curl);

        if (isset($response['info']['http_code']) && $response['info']['http_code'] !== 200) {
            $this->getLog()->error(
                __LINE__,
                sprintf('Service %s response error %s %s', get_class($this), json_encode($response['info']), json_encode($response['error']))
            );
        }

        return $response;
    }

    protected function sendRequest($type, $url, $params = null, $formData = false, $headers = null) {
        if (!is_array($params)) {
            $params = array();
        }

        if (empty($headers)) {
            $headers = array('Accept: application/json');
        }

        $this->getLog()->info(__LINE__, $type . ': ' . $url);

        $response = $this->makeRequest($type, $url, $params, $formData, $headers);

        if (is_array($response['info']) && array_key_exists('http_code', $response['info']) && $response['info']['http_code'] === 404) {
            throw new Exception\ServiceException($response['error']['reason'], 'NOT_FOUND');
        }

        if ($response['result'] === false) {
            throw new Exception\ServiceException($response['error']['reason'], 'CONNECTION_ERROR');
        }

        $this->getLog()->info(__LINE__, 'Dirty response: ' . $response['result']);

        $result = $this->parse($response['result']);

        // @codingStandardsIgnoreLine
        $this->getLog()->info(__LINE__, 'Response: ' . print_r($result, true));
        return $result;
    }

    protected function parse($data) {
        $decodedResult = json_decode($data, true);

        if (!$decodedResult || !is_array($decodedResult)) {
            throw new Exception\ServiceException($data, 'RESPONSE_DECODE_ERROR');
        }

        if ($decodedResult['type'] == 'error') {
            throw new Exception\ServiceException($decodedResult['reason'], $decodedResult['code']);
        }

        return $decodedResult['result'];
    }

    protected function addProxySettings(&$curl) {
        if ($this->_useHttpProxy && !($this instanceof \Wildix\SipProxy)) {
            $proxySettings = $this->_httpProxy->getSettings();

            if (\Wildix\Validator::isTrue($proxySettings['enable'])) {
                curl_setopt($curl, CURLOPT_PROXY, $proxySettings['host'] . ':' . $proxySettings['port']);

                if ($proxySettings['auth_method'] !== 'none') {
                    curl_setopt($curl, CURLOPT_PROXYUSERPWD, $proxySettings['login'] . ':' . $proxySettings['password']);
                    curl_setopt($curl, CURLOPT_PROXYAUTH, ($proxySettings['auth_method'] === 'ntlm') ? CURLAUTH_NTLM : CURLAUTH_BASIC);
                }
            }
        }
    }
}
