<?php

namespace Wildix\Services;

use Wildix\Exception;
use Wildix\Services\Exception\ServiceException;

/**
 * Class GlobalConfig
 */
class GlobalConfig extends AbstractService {

    const CONFIG_FILE_EXTENSION = '.json';

    const TYPE_COGNITO = 'cognito';
    const TYPE_CONFIG  = 'config';

    const ALLOWED_CONFIGS = [
        self::TYPE_COGNITO,
        self::TYPE_CONFIG,
    ];

    private $_configSite;

    /**
     * GlobalConfig constructor.
     *
     * @param integer $logLevel    LogLevel.
     * @param string  $_configSite ConfigSite.
     */
    public function __construct(int $logLevel, string $_configSite) {
        parent::__construct($logLevel);
        $this->_configSite = $_configSite;
    }


    /**
     *  Get global config by name
     *
     * @param string $configName Global config name.
     *
     * @return array
     *
     * @throws \Exception Error load config.
     */
    public function getConfig(string $configName): array {
        if (!in_array($configName, self::ALLOWED_CONFIGS)) {
            throw new Exception(sprintf('Unknown config %s', $configName), 500);
        }

        $data = $this->_getConfigFromServer($configName);
        if (!isset($data[$configName])) {
            throw new Exception(sprintf('Incorrect config data. Not found key %s in %s', $configName, json_encode($data)), 500);
        }

        return $data[$configName];
    }

    private function _getConfigFromServer(string $configName): array {
        $this->getLog()->info(__LINE__, sprintf('Load from remote server global config %s', $configName));

        return $this->sendRequest('GET', $this->_getUrlByName($configName));
    }

    private function _getUrlByName(string $name): string {
        return $this->_configSite . $name . self::CONFIG_FILE_EXTENSION;
    }

    protected function parse($data) {
        $decodedResult = json_decode($data, true);

        if (!$decodedResult || !is_array($decodedResult)) {
            throw new ServiceException($data, 'RESPONSE_DECODE_ERROR');
        }

        return $decodedResult;
    }
}
