<?php
namespace Wildix\Services;

use Wildix\System\HttpProxy;

/**
 * Class Group Chat
 */
class GroupChat extends AbstractService {

    public const GROUP_CHAT_ENABLED = 'enabled';
    public const GROUP_CHAT_DOMAIN  = 'domain';
    public const SMS_ENABLED        = 'smsEnabled';
    public const SMS_DOMAIN         = 'smsBotDomain';

    private $_defaultConfig;

    private $_config;
    private $_license;


    public function __construct(\Wildix\License $license, $_defaultConfig, $logLevel = \Logger\Log::DEBUG, HttpProxy $httpProxy = null) {
        parent::__construct($logLevel, $httpProxy);
        $this->_license       = $license;
        $this->_defaultConfig = $_defaultConfig;
    }

    /**
     * Parse config with group chat domain.
     *
     * @param string $name Parameter name.
     *
     * @return string
     */
    public function getConfig(string $name): string {
        // do not return the config in case of disabled by license
        if (!$this->isCloudGroupChatAllowedByLicense()) {
            return '';
        }
        if ($this->_config === null) {
            $this->_config = $this->_defaultConfig;
        }

        return $this->_config[$name] ?? '';
    }

    /**
     * Return the group chat domain if enabled.
     *
     * @return string
     */
    public function getGroupChatDomain(): string {
        $groupChatEnabled = $this->getConfig(self::GROUP_CHAT_ENABLED);
        $groupChatDomain  = $this->getConfig(self::GROUP_CHAT_DOMAIN);

        if (filter_var($groupChatEnabled, FILTER_VALIDATE_BOOLEAN) && filter_var($groupChatDomain, FILTER_VALIDATE_DOMAIN)) {
            return $groupChatDomain;
        }

        return '';
    }

    /**
     * Return the SMS bot domain if group chats are enabled.
     *
     * @return string
     */
    public function getSmsBotDomain(): string {
        $groupChatEnabled = $this->getConfig(self::GROUP_CHAT_ENABLED);
        $smsEnabled       = $this->getConfig(self::SMS_ENABLED);
        $smsBotDomain     = $this->getConfig(self::SMS_DOMAIN);

        if (filter_var($groupChatEnabled, FILTER_VALIDATE_BOOLEAN)
            && filter_var($smsEnabled, FILTER_VALIDATE_BOOLEAN)
            && filter_var($smsBotDomain, FILTER_VALIDATE_DOMAIN)
        ) {
            return $smsBotDomain;
        }

        return '';
    }

    /**
     * Parse license and check if PBX allowed to use cloud group chat.
     *
     * @return boolean
     */
    private function isCloudGroupChatAllowedByLicense() {
        return !$this->_license->isLifetime();
    }
}
