<?php

namespace Wildix\Services;

/**
 * Class PhoneBridge
 */
class PhoneBridge {
    const SERVICE_ID_ZOHO_CRM = 'zoho-crm';

    private $_phonebridgeUrl;

    /**
     * Instance of logger.
     *
     * @var \Logger\Log
     */
    private $_logger;

    /**
     * PhoneBridge constructor.
     *
     * @param \Logger\Log $logger          Param.
     * @param string      $_phoneBridgeUrl Param.
     */
    public function __construct(\Logger\Log $logger, $_phoneBridgeUrl) {
        $this->_logger         = $logger;
        $this->_phonebridgeUrl = $_phoneBridgeUrl . '/zohov3';
    }

    /**
     * Enable PhoneBridge integration for external service.
     *
     * @param string $serviceId        External service identifier.
     * @param string $comebackLocation Location for return after confirmation.
     *
     * @return boolean|string If Zoho CRM account has no confirmation of access for PhoneBridge return link to confirmation page.
     */
    public function enable($serviceId, $comebackLocation) {
        if (!$comebackLocation) {
            return false;
        }

        $ch = curl_init();
        curl_setopt($ch, CURLOPT_USERAGENT, 'Wildix PBX (' . WMS_VERSION_FULL . ')');
        curl_setopt($ch, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
        curl_setopt($ch, CURLOPT_USERPWD, \Models\PBX::getInstance()->getSerial() . ':' . \Models\PBX::getInstance()->getPbxKey());
        curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 2);
        curl_setopt($ch, CURLOPT_TIMEOUT, 30);
        curl_setopt($ch, CURLOPT_URL, sprintf('%s/api/v1/Integration/%s/', $this->getPhonebridgeUrl(), $serviceId));
        curl_setopt($ch, CURLOPT_AUTOREFERER, false);
        curl_setopt($ch, CURLOPT_POSTFIELDS, http_build_query(array('comebackLocation' => $comebackLocation)));
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);

        // phpcs:disable Generic.CodeAnalysis.UnusedFunctionParameter.Found
        $redirectUrl = null;
        curl_setopt(
            $ch,
            CURLOPT_HEADERFUNCTION,
            function ($ch, $header) use (&$redirectUrl) {
                $headerParts = explode(': ', $header);

                if (strtolower($headerParts[0]) === 'location') {
                    $redirectUrl = implode(': ', array_slice($headerParts, 1));
                }

                return strlen($header);
            }
        );
        // phpcs:enable

        curl_exec($ch);
        $info   = curl_getinfo($ch);
        $reason = curl_error($ch);
        $code   = curl_errno($ch);

        curl_close($ch);

        if ($info['http_code'] === 200) {
            return true;
        }

        $this->_logger->error(__LINE__, \sprintf('PhoneBridge enable integration error code: %s, reason: %s', $code, $reason));

        return isset($redirectUrl) ? $redirectUrl : false;
    }

    public function disable($serviceId) {
        $ch = curl_init();
        curl_setopt($ch, CURLOPT_USERAGENT, 'Wildix PBX (' . WMS_VERSION_FULL . ')');
        curl_setopt($ch, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
        curl_setopt($ch, CURLOPT_USERPWD, \Models\PBX::getInstance()->getSerial() . ':' . \Models\PBX::getInstance()->getPbxKey());
        curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 2);
        curl_setopt($ch, CURLOPT_TIMEOUT, 30);
        curl_setopt($ch, CURLOPT_URL, sprintf('%s/api/v1/Integration/%s/', $this->getPhonebridgeUrl(), $serviceId));
        curl_setopt($ch, CURLOPT_CUSTOMREQUEST, 'DELETE');
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);

        curl_exec($ch);
        $info   = curl_getinfo($ch);
        $reason = curl_error($ch);
        $code   = curl_errno($ch);

        curl_close($ch);

        if ($info['http_code'] !== 200) {
            $this->_logger->error(__LINE__, \sprintf('PhoneBridge disable integration error code: %s, reason: %s', $code, $reason));
        }

        return ($info['http_code'] === 200);
    }

    public function sync($serviceId) {
        $usersFilter   = array(
            \Wildix\Colleagues\Colleague::EXTENSION => '*',
            \Wildix\Colleagues\Colleague::SOURCE_ID => '*',
        );
        $users         = \Wildix\Colleagues\Colleagues::getInstance()->getUsers($usersFilter);
        $usersData     = $this->_collectUsersDataForPhoneBridge($users);
        $usersDataSize = count($usersData);

        if ($usersDataSize === 0) {
            $this->_logger->info(__LINE__, 'The PBX does not have users to synchronize with PhoneBridge.');

            return true;
        }

        $isErrorOccurred = false;
        $batchLimit      = 50;
        $page            = 0;

        do {
            $offset          = $batchLimit * $page;
            $batch           = array_slice($usersData, $offset, $batchLimit);
            $isErrorOccurred = ($isErrorOccurred || !$this->_sendUsersData($serviceId, $batch));
            $page++;
        } while ($offset < $usersDataSize - $batchLimit);

        return !$isErrorOccurred;
    }

    /**
     * Convert users to data in PhoneBridge format.
     *
     * @param array $users List of users.
     *
     * @return array
     */
    private function _collectUsersDataForPhoneBridge(array $users) {
        return array_reduce(
            $users,
            function (array $data, \Wildix\Colleagues\Colleague $user) {
                if ($user->isInternalUser()) {
                    return $data;
                }

                $data[] = array(
                    'id' => $user->getID(),
                    'extension' => $user->getExtension(),
                    'sourceId' => $user->getSourceId(),
                    'pbxDomain' => $user->getPbxDomain(),
                );

                return $data;
            },
            array()
        );
    }

    private function _sendUsersData($serviceId, array $batch) {
        $ch = curl_init();
        curl_setopt($ch, CURLOPT_USERAGENT, 'Wildix PBX (' . WMS_VERSION_FULL . ')');
        curl_setopt($ch, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
        curl_setopt($ch, CURLOPT_USERPWD, \Models\PBX::getInstance()->getSerial() . ':' . \Models\PBX::getInstance()->getPbxKey());
        curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 3);
        curl_setopt($ch, CURLOPT_TIMEOUT, 120);
        curl_setopt($ch, CURLOPT_AUTOREFERER, false);
        curl_setopt($ch, CURLOPT_URL, sprintf('%s/api/v1/Integration/%s/sync/', $this->getPhonebridgeUrl(), $serviceId));
        curl_setopt($ch, CURLOPT_POSTFIELDS, http_build_query(array('users' => $batch)));
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);

        curl_exec($ch);
        $info = curl_getinfo($ch);
        curl_close($ch);

        return ($info['http_code'] === 200);
    }

    public function getPhonebridgeUrl(): string {
        return $this->_phonebridgeUrl;
    }
}
