<?php
namespace Wildix\Services;

use Wildix\Exception;

/**
 * Class Portal
 */
class Portal extends AbstractService {

    private $_portalApiUrl;
    private $_deviceUrl;
    private $_portsUrl;
    private $_interfacesUrl;
    private $_urlDownloadCustomLogo;
    private $_servicesLogUrl;
    private $_upgradeUrl;
    private $_credentialsUrl;
    private $_wmsDomainGet;
    private $_wmsDomainSet;
    private $_wmsDomainCheck;
    private $_wmsSetIpPort;
    private $_singleSignOn;

    private $_remoteSupportConfigUrl;

    private $_portsStatusCacheFile = '/tmp/portsStatusCache';
    private $_interfacesCacheFile  = '/tmp/interfacesCache';

    private $_portalUrl;

    private $_releasesPath = 'wmsupgrade/versions';

    private $_upgradeErrorResponse = [
        'errorCode' => 500,
        'error' => ''
    ];

    public function __construct($logLevel, $_portalApiUrl, $_portalUrl) {
        parent::__construct($logLevel);
        $this->_portalApiUrl           = $_portalApiUrl;
        $this->_remoteSupportConfigUrl = $this->_portalApiUrl . '/v2/PBX/RemoteSupport/Config';
        $this->_deviceUrl              = $this->_portalApiUrl . '/v2/PBX/Devices/';
        $this->_portsUrl               = $this->_portalApiUrl . '/v2/PBX/Ports/';
        $this->_interfacesUrl          = $this->_portalApiUrl . '/v2/PBX/Interfaces/';
        $this->_urlDownloadCustomLogo  = $this->_portalApiUrl . '/v2/PBX/Logo';
        $this->_servicesLogUrl         = $this->_portalApiUrl . '/v2/PBX/Logs/Services/';
        $this->_upgradeUrl             = $this->_portalApiUrl . '/v2/PBX/Upgrade/Settings/';
        $this->_credentialsUrl         = $this->_portalApiUrl . '/v2/PBX/Credentials/';
        $this->_wmsDomainGet           = $this->_portalApiUrl . '/v2/PBX/WmsDomainGet';
        $this->_wmsDomainSet           = $this->_portalApiUrl . '/v2/PBX/WmsDomainSet';
        $this->_wmsDomainCheck         = $this->_portalApiUrl . '/v2/PBX/WmsDomainCheck';
        $this->_wmsSetIpPort           = $this->_portalApiUrl . '/v2/PBX/WmsSetIp';
        $this->_singleSignOn           = $this->_portalApiUrl . '/v2/PBX/SSO/';
        $this->_portalUrl              = $_portalUrl;
    }


    /**
     * Set path for releases Url.
     *
     * @param string $path Script path.
     *
     * @return void
     */
    public function setPath(string $path): void {
        $this->_releasesPath = $path;
    }

    private function getPortsStatusFromCache($force) {
        if (!$force) {
            if (!file_exists($this->_portsStatusCacheFile)) {
                $force = true;
            } else {
                $cacheTime     = filemtime($this->_portsStatusCacheFile);
                $cacheLifeTime = (time() - $cacheTime) / 60 / 60 / 24;
                if ($cacheLifeTime > 1) {
                    $force = true;
                }
            }
        }
        return !$force;
    }

    private function needUploadInterface($interfaces) {
        $needUpload = false;

        if (!file_exists($this->_interfacesCacheFile)) {
            $needUpload = true;
        } else {
            $cache = file_get_contents($this->_interfacesCacheFile);
            if ($cache != json_encode($interfaces)) {
                $needUpload = true;
            }
        }

        return $needUpload;
    }

    protected function sendRequest($type, $url, $params = null, $formData = false, $headers = null) {
        if (!is_array($params)) {
            $params = array();
        }

        $this->getLog()->info(__LINE__, $type . ': ' . $url);
        // @codingStandardsIgnoreLine
        $this->getLog()->info(__LINE__, 'Params: ' . print_r($params, true));
        $this->getLog()->debug(__LINE__, 'Auth: ' . \Models\PBX::getInstance()->getSerial() . ':' . \Models\PBX::getInstance()->getPbxKey());

        if (empty($headers)) {
            $headers = array('Accept: application/json');
        }

        $response = $this->makeRequest($type, $url, $params, $formData, $headers);
        $result   = $this->parse($response);

        // @codingStandardsIgnoreLine
        $this->getLog()->info(__LINE__, 'Response: ' . print_r($result, true));
        return $result;
    }

    protected function parse($response) {
        if ($response['result'] && $response['info']['content_type'] === 'image/png') {
            return array(
                'type' => 'result',
                'result' => $response['result']
            );
        }

        $decodedResult = json_decode($response['result'], true);
        if (!$decodedResult || !is_array($decodedResult)) {
            return array(
                'type' => 'error',
                'result' => $response['error']
            );
        }

        return $decodedResult;
    }

    public function getDomain($onlyMy = true, $serial = false) {
        $params = ['type' => $onlyMy ? 0 : 1];

        if ($serial) {
            $params['serial'] = is_array($serial) ? implode(';', $serial) : $serial;
        }

        $response = $this->sendRequest('POST', $this->_wmsDomainGet, $params);
        if (!$response
            || empty($response['type'])
            || $response['type'] !== 'result'
            || empty($response['result']['domains'])
            || !is_array($response['result']['domains'])
        ) {
            return [];
        }

        $domains = [];
        foreach ($response['result']['domains'] as $key => $val) {
            $domains[strtolower($key)] = $val;
        }

        return $domains;
    }

    public function checkDomain($domain) {
        $response  = $this->sendRequest('POST', $this->_wmsDomainCheck, ['domain' => $domain]);
        $isSuccess = $this->_isSuccessResponse($response);

        $return['success'] = $isSuccess;
        if (!$isSuccess && !empty($response['result'])) {
            $return['error']['code'] = intval($response['result']['code']);
            switch ($response['result']['code']) {
                case '301':
                    $return['error']['message'] = 'EMPTY SERIAL OR DOMAIN';
                    break;
                case '302':
                    $return['error']['message'] = 'WRONG SERIAL';
                    break;
                case '303':
                    $return['error']['message'] = 'DOMAIN IS NOT CORRECT';
                    break;
                case '304':
                    $return['error']['message'] = 'PBX NOT FOUND';
                    break;
                case '305':
                    $return['error']['message'] = 'DOMAIN IS NOT AVAILABLE';
                    break;
                default:
                    $return['error']['message'] = $response['result']['reason'] ?? '';
                    break;
            }
        } else if ($response['result']['success'] === false) {
            $return['error'] = [
                'code' => 305,
                'message' => 'DOMAIN IS NOT AVAILABLE',
            ];
        }

        // @codingStandardsIgnoreLine
        $this->getLog()->info(__LINE__, 'Return checkDomain(): ' . print_r($return, true));
        return $return;
    }

    public function setDomain($domain) {
        $response = $this->sendRequest('POST', $this->_wmsDomainSet, ['domain' => $domain]);

        return $this->_isSuccessResponse($response);
    }

    public function setPort($port) {
        $params = [
            'pbxkey' => \Models\PBX::getInstance()->getPbxKey(),
            'port' => $port,
        ];

        $response = $this->sendRequest('POST', $this->_wmsSetIpPort, $params);

        return $this->_isSuccessResponse($response);
    }

    public function setIp() {
        $params = [
            'pbxkey' => \Models\PBX::getInstance()->getPbxKey()
        ];
        $response = $this->sendRequest('POST', $this->_wmsSetIpPort, $params);

        if ($this->_isSuccessResponse($response)) {
            $this->getLog()->info(__LINE__, 'IP PBX successfully updated on the portal');
        } else {
            $this->getLog()->warning(__LINE__, 'Update IP PBX on the portal failed');
        }
    }

    public function setExternalIp() {
        $params = ['pbxkey' => \Models\PBX::getInstance()->getPbxKey()];

        $array = array( 15, 30, 60, 120, 240 );

        for ($i = 0; $i < 100; ++$i) {
            $timeout  = $i < count($array) ? $array[$i] : end($array);
            $response = $this->sendRequest('POST', $this->_wmsSetIpPort, $params);
            if ($this->_isSuccessResponse($response)) {
                $this->getLog()->warning(__LINE__, 'Send ip to portal: OK, retry number ' . $i);
                return true;
            }

            $code = intval($response['result']['code']);

            if ($code < 200 || $code > 499 && $code < 505) {
                $message  = 'Could not connect to portal: try number ' . $i;
                $message .= ', error code ' . $code;
                $message .= ', next retry in ' . $timeout . 'sec';
                $this->getLog()->warning(__LINE__, $message);
                usleep($timeout * 1000000);
            } else {
                $this->getLog()->warning(__LINE__, 'Could not set ip on portal: http code ' . $code . ', retry number ' . $i);
                break;
            }

            if ($i === 99) {
                $this->getLog()->warning(__LINE__, 'Error while connect to portal with 100 retries');
            }
        }

        return false;
    }

    private function _isSuccessResponse($response) {
        return (!empty($response['type']) && $response['type'] === 'result')
            && (!empty($response['result']['success']) && $response['result']['success'] === true);
    }

    public function getDevices($start = null, $count = null) {
        $return = false;
        $params = array(
            'start' => ($start !== null) ? $start : 0,
            'count' => ($count !== null) ? $count : 100
        );

        $response = $this->sendRequest('GET', $this->_deviceUrl, $params);

        if (!$response) {
            return $return;
        }

        if ($response['type'] == 'error' && is_array($response['result'])) {
            throw new Exception($response['result']['reason'], $response['result']['code']);
        } else if ($response['type'] == 'result' && is_array($response['result']) && is_array($response['result']['records'])) {
            $return = $response['result'];
            if (!empty($return['records'])) {
                foreach ($return['records'] as $key => $value) {
                    if (!empty($value['serial'])) {
                        $return['records'][$key]['serial'] = strtolower($value['serial']);
                    }
                    if (!empty($value['model'])) {
                        $return['records'][$key]['model'] = strtolower($value['model']);
                    }
                    if (!empty($value['pbx'])) {
                        $return['records'][$key]['pbx'] = strtolower($value['pbx']);
                    }
                }
            }
        }

        return $return;
    }

    public function setPorts($ports, $force = false) {
        $params = array(
            'ports' => $ports,
            'force' => ($force) ? 1 : 0
        );

        $response = $this->sendRequest('PUT', $this->_portsUrl, $params);

        if (!$response) {
            return false;
        }

        if ($response['type'] == 'error' && is_array($response['result'])) {
            throw new Exception($response['result']['reason'], $response['result']['code']);
        } else if ($response['type'] == 'result' && !empty($response['result'])) {
            return $response['result'];
        } else {
            throw new Exception('Bad response', 500);
        }
    }

    public function getPortsStatus($force = false) {
        $return = false;

        if ($this->getPortsStatusFromCache($force)) {
            $cache = file_get_contents($this->_portsStatusCacheFile);
            if (!empty($cache)) {
                $return = json_decode($cache, true);
            }
        } else {
            // @codingStandardsIgnoreLine
            if ($response = $this->sendRequest('GET', $this->_portsUrl)) {
                if ($response['type'] == 'error' && is_array($response['result'])) {
                    throw new Exception($response['result']['reason'], $response['result']['code']);
                } else if ($response['type'] == 'result' && is_array($response['result']) && is_array($response['result']['records'])) {
                    $return = $response['result'];
                } else {
                    throw new Exception('Bad response', 500);
                }
            }

            if ($return) {
                $portsChecked = true;
                foreach ($return['records'] as $key => $value) {
                    if ($value['status'] == null || $value['status'] == 'null') {
                        $portsChecked = false;
                        break;
                    }
                }

                if ($portsChecked) {
                    file_put_contents($this->_portsStatusCacheFile, json_encode($return));
                }
            }
        }

        return $return;
    }

    public function setInterfaces($interfaces) {
        $return = false;

        if (!$this->needUploadInterface($interfaces)) {
            return $return;
        }

        $params = array('interfaces' => $interfaces);
        // @codingStandardsIgnoreLine
        if ($response = $this->sendRequest('PUT', $this->_interfacesUrl, $params)) {
            if ($response['type'] == 'error' && is_array($response['result'])) {
                throw new Exception($response['result']['reason'], $response['result']['code']);
            } else if ($response['type'] == 'result' && !empty($response['result'])) {
                $return = $response['result'];
            } else {
                throw new Exception('Bad response', 500);
            }
        }

        // update cache
        file_put_contents($this->_interfacesCacheFile, json_encode($interfaces));

        return $return;
    }

    public function setServicesLog($log) {
        $return = false;

        $params = array(
            'serial' => \Models\PBX::getInstance()->getSerial(),
            'log' => base64_encode($log)
        );

        // @codingStandardsIgnoreLine
        if ($response = $this->sendRequest('POST', $this->_servicesLogUrl, $params)) {
            if ($response['type'] == 'error' && is_array($response['result'])) {
                throw new Exception($response['result']['reason'], $response['result']['code']);
            } else if ($response['type'] == 'result' && !empty($response['result'])) {
                $return = $response['result'];
            } else {
                throw new Exception('Bad response', 500);
            }
        }

        return $return;
    }

    public function getUpgradeParameters() {
        $response = $this->sendRequest('GET', $this->_upgradeUrl);

        if (!$response) {
            return false;
        }

        if ($response['type'] == 'error' && is_array($response['result'])) {
            throw new Exception($response['result']['reason'], $response['result']['code']);
        } else if ($response['type'] == 'result' && is_array($response['result'])) {
            return $response['result'];
        } else {
            throw new Exception('Bad response', 500);
        }
    }

    public function getPartnerCustomLogo() {
        $response = $this->sendRequest('GET', $this->_urlDownloadCustomLogo);

        if ($response['type'] == 'error' && is_array($response['result'])) {
            throw new Exception($response['result']['reason'], $response['result']['code']);
        } else if ($response['type'] == 'result' && !empty($response['result'])) {
            $imgResult = $response['result'];
            if (!empty($imgResult)) {
                return $imgResult;
            }
        } else {
            throw new Exception('Bad response', 500);
        }

        return false;
    }

    /**
     * Return configuration of OpenVPN for Remote Support.
     *
     * @return array
     * @throws \ErrorException if failed send request to the Portal or got error in response
     */
    public function getRemoteSupportConfigs(): array {
        $response = $this->sendRequest('GET', $this->_remoteSupportConfigUrl);
        $this->_assertResponseIsSuccess($response);

        return $response['result'];
    }

    /**
     * Returns list of keys pairs for accessing to WMS API.
     *
     * @return array
     * @throws \Exception if failed send request to the Portal or got error in response
     */
    public function getApiKeysPair(): array {
        $response = $this->sendRequest('GET', $this->_credentialsUrl);

        if (!$response) {
            return [];
        }

        if ($response['type'] === 'error' && is_array($response['result'])) {
            throw new Exception($response['result']['reason'], $response['result']['code']);
        } else if ($response['type'] === 'result' && is_array($response['result'])) {
            return $response['result'];
        }

        throw new Exception('Bad response', 500);
    }

    /**
     * Assert that response from the Portal is success.
     *
     * @param array $response Response data.
     *
     * @return void
     * @throws \ErrorException if response has error.
     */
    private function _assertResponseIsSuccess(array $response): void {
        if ($response['type'] !== 'error') {
            return;
        }

        $error     = $response['code'] ?? $response['result']['reason'];
        $errorCode = $response['result']['code'] ?? 0;

        throw new \ErrorException($error, $errorCode);
    }

    /**
     * Get all versions of releases from portal.
     *
     * @return array
     */
    public function getReleasesList(): array {
        $response = $this->makeRequest('GET', $this->_portalUrl . '/' . $this->_releasesPath, [], null, []);
        if ($response['info']['http_code'] !== 200 || $response['error']['code'] !== 0) {
            return [];
        }

        $releases = parse_ini_string($response['result'], true, INI_SCANNER_RAW);

        return $releases !== false ? $releases : [];
    }

    /**
     * Check on portal has denied upgrade this PBX.
     *
     * @param string $url Url address.
     * @param string $mac Pbx mac.
     *
     * @return array
     */
    public function checkUpgradeIsDenied(string $url, string $mac): array {
        $curlOpt = [
            CURLOPT_USERAGENT => 'WMS Activation Script',
            CURLOPT_TIMEOUT => 10,
            CURLOPT_AUTOREFERER => true,
        ];

        $url .= '?check=1';

        $response = $this->makeRequest('POST', $url, ['mac' => $mac], false, [], $curlOpt);
        if ($response['info']['http_code'] !== 200) {
            $info          = $this->_upgradeErrorResponse;
            $info['error'] = 'Cannot connect upgrade server ' . $url;

            return $info;
        }

        return $this->_processUpgradePermissionsResponse($response['result']);
    }

    /**
     * Verify JWT token n portal and returns authorized user info.
     *
     * @param string $token JWT token.
     *
     * @return array
     * @throws \Exception When token not verified
     */
    public function singleSignOn(string $token): array {
        $response = $this->sendRequest('POST', $this->_singleSignOn, ['token' => $token]);

        if ($response['type'] === 'error' && is_array($response['result'])) {
            $this->getLog()->error(__LINE__, \sprintf('Portal single sign on. Code %s, error: %s', $response['result']['code'], $response['result']['reason']));
            throw new Exception($response['result']['reason'], $response['result']['code']);
        } else if ($response['type'] == 'result' && !empty($response['result'])) {
            return $response['result'];
        } else {
            $this->getLog()->error(__LINE__, 'Portal single sign on. Bad response: ' . json_encode($response));
            throw new Exception('Bad response', 500);
        }
    }

    /**
     * Process portal response.
     *
     * @param string $response Portal response.
     *
     * @return array
     */
    private function _processUpgradePermissionsResponse(string $response): array {
        if ($response === 'UPGRADE') {
            return [
                'errorCode' => 0,
                'error' => ''
            ];
        }


        $out = $this->_upgradeErrorResponse;
        if (preg_match('/ERROR ([0-9]+)/', $response, $matches)) {
            if (in_array(
                $matches[1],
                [
                    '301',
                    '302',
                    '303',
                    '304',
                    '308',
                    '401',
                ]
            )
            ) {
                /*
                    301 - Request without POST parameters
                    302 - Wrong serial length
                    303 - Gateway not found by serial
                    304 - Upgrade is not allowed for the company
                    308 - Wrong identity length
                    401 - PBX is found but with identity mismatch
                    301 302 308 - the request is not accepted by upgrade server
                    303 401 - PBX is not found on upgrade server
                    304 - upgrade procedure is disabled
                */

                $out['errorCode'] = (int)$matches[1];
                $out['error']     = _t('upgrade.UpgradeServerError' . $matches[1]);
            } else {
                $out['error'] = _t('upgrade.UpgradeServerResponseIs', $response);
            }
        } else {
            $out['error'] = _t('upgrade.UpgradeServerUnknownResponse');
        }

        return $out;
    }
}
