<?php
namespace Wildix\Services;

/**
 * Class Kite
 */
class VideoConference extends AbstractService {

    const CONF_FILE = '/rw2/etc/pbx/conference.conf';

    const GET_ROOMS          = 'https://:domain/api/v1/Rooms/';
    const GET_ROOM           = 'https://:domain/api/room/info';
    const CREATE_ROOM        = 'https://:domain/api/v1/Rooms/';
    const DELETE_ROOM        = 'https://:domain/api/room/destroy';
    const REMOVE_PARTICIPANT = 'https://:domain/api/room/uninvite';
    const KICK_PARTICIPANT   = 'https://:domain/api/room/kick';
    const GET_SCHEDULE       = 'https://:domain/api/room/:roomId/schedule';
    const AUTHORIZE_IN_ROOM  = 'https://:domain/api/conference/authorize';
    const PUSH_HISTORY       = 'https://:domain/history/push';
    const MEMBER_PRESENCE    = 'https://:domain/api/room/:roomId/members/:extension/presence/';


    private $_videoHost;

    private $_config = null;

    /**
     * VideoConference constructor.
     *
     * @param integer $logLevel   Log level.
     * @param string  $_videoHost Host.
     */
    public function __construct($logLevel, string $_videoHost) {
        parent::__construct($logLevel);
        $this->_videoHost = parse_url($_videoHost, PHP_URL_HOST);
    }

    /**
     * Parse config with host.
     *
     * @param string $name Parameter name.
     *
     * @return array
     */
    public function getConfig($name) {
        if ($this->_config == null) {
            if (!$this->_config) {
                $this->_config = array(
                        'host' => $this->_videoHost
                );
            }
        }

        return $this->_config[$name] ?? '';
    }

    /**
     * Returns url.
     *
     * @param string      $url       Url.
     * @param null|string $roomId    Room id.
     * @param null|string $extension User extension.
     *
     * @return string
     */
    private function getUrl($url, $roomId = null, $extension = null): string {
        $url = str_replace(':domain', $this->getConfig('host'), $url);
        if ($roomId !== null) {
            $url = str_replace(':roomId', $roomId, $url);
        }
        if ($extension !== null) {
            $url = str_replace(':extension', $extension, $url);
        }
        return $url;
    }

    /**
     * Get rooms by jid.
     *
     * @param string $jid JID of onwer user.
     *
     * @return array
     */
    public function getRoomsByJid($jid) {
        $params = array(
                'filter[participant]' => $jid
        );
        return $this->sendRequest('GET', $this->getUrl(self::GET_ROOMS), $params);
    }

    public function createRoom($params) {
        return $this->sendRequest('POST', $this->getUrl(self::CREATE_ROOM), $params);
    }

    /**
     * Destroy room.
     *
     * @param string $roomId Room id.
     *
     * @return array
     */
    public function destroyRoom($roomId) {
        $params = array(
                'room'  => $roomId
        );
        return $this->sendRequest('POST', $this->getUrl(self::DELETE_ROOM), $params);
    }

    public function getRoom($roomId) {
        $params = array(
                'room'  => $roomId
        );

        $response = $this->sendRequest('POST', $this->getUrl(self::GET_ROOM), $params);
        if (!empty($response['description'])) {
            $response['description'] = json_decode($response['description'], true);
        }
        return $response;
    }

    public function getSchedule($roomId) {
        return $this->sendRequest('GET', $this->getUrl(self::GET_SCHEDULE, $roomId));
    }

    /**
     * Remove user from the member list.
     *
     * @param string $roomId Room id.
     * @param string $jid    JID of the user.
     *
     * @return array
     */
    public function removeParticipant($roomId, $jid) {
        $params = array(
                'room'  => $roomId,
                'participant' => $jid
        );
        return $this->sendRequest('POST', $this->getUrl(self::REMOVE_PARTICIPANT), $params);
    }

    /**
     * Kick participant from the room.
     *
     * @param string $roomId Room id.
     * @param string $nick   Nick of the user.
     *
     * @return array
     */
    public function kickParticipant($roomId, $nick) {
        $params = array(
                'room'  => $roomId,
                'nick'  => $nick,
        );
        return $this->sendRequest('POST', $this->getUrl(self::KICK_PARTICIPANT), $params);
    }

    public function authorizeInRoom($params) {
        return $this->sendRequest('POST', $this->getUrl(self::AUTHORIZE_IN_ROOM), $params);
    }

    public function pushHistory($params) {
        return $this->sendRequest('POST', $this->getUrl(self::PUSH_HISTORY), $params);
    }

    /**
     * Return user presence in video conference.
     * Presence statuses - "available", "unavailable", "not-found"
     *
     * @param string $roomId    Room id.
     * @param string $extension User extension.
     *
     * @return string
     * @throws Exception\ServiceException When response with error.
     */
    public function memberPresence(string $roomId, string $extension) {
        $response = $this->sendRequest('GET', $this->getUrl(self::MEMBER_PRESENCE, $roomId, $extension), ['serial' => $this->getPbx()->getSerial()]);

        return !empty($response) ? $response : 'not-found';
    }
}
