<?php

namespace Wildix\Services;

use Wildix\Exception;

/**
 * Class WizyWebinar
 */
class WizyWebinar extends AbstractService {

    private $_wizyWebinarContactsUrl;
    private $_wizyWebinarTokenUrl;

    public function __construct($logLevel, string $wizyWebinarUri) {
        parent::__construct($logLevel);
        $this->_wizyWebinarContactsUrl = $wizyWebinarUri . '/api/v1/contacts/';
        $this->_wizyWebinarTokenUrl    = $wizyWebinarUri . '/api/v1/token/';
    }

    /**
     * Get WizyWebinar contacts
     *
     * @return array
     *
     * @throws \Exception Error get response from wizywebinar
     */
    public function getWizyWebinarContacts(): array {
        $response = $this->sendRequest('GET', $this->_wizyWebinarContactsUrl);

        if (!empty($response['type']) && $response['type'] === 'result' && is_array($response['result']) && is_array($response['result']['contacts'])) {
            return $response['result']['contacts'];
        }

        if (!empty($response['type'])
            && $response['type'] === 'error'
            && !empty($response['result']['reason'])
            && !empty($response['result']['code'])
        ) {
            throw new Exception($response['result']['reason'], $response['result']['code']);
        }

        throw new Exception('Bad response', 500);
    }

    /**
     * Get access token for WizyWebinar
     *
     * @param string $email User email.
     *
     * @return string
     *
     * @throws \Exception Error get response from wizywebinar
     */
    public function wizyWebinarAccessToken(string $email): string {
        $data = [
            'login' => $email,
            'serial' => \Models\PBX::getInstance()->getSerial(),
            'extension' => \User::getInstance()->getExtension()
        ];

        $response = $this->sendRequest('POST', $this->_wizyWebinarTokenUrl, $data);

        if (!empty($response['type'])
            && $response['type'] === 'result'
            && !empty($response['result']['token'])
        ) {
            return (string)$response['result']['token'];
        }

        if (!empty($response['type'])
            && $response['type'] === 'error'
            && !empty($response['result']['reason'])
            && !empty($response['result']['code'])
        ) {
            throw new Exception($response['result']['reason'], $response['result']['code']);
        }

        throw new Exception('Bad response', 500);
    }

    protected function sendRequest($type, $url, $params = null, $formData = false, $headers = null) {
        if (!is_array($params)) {
            $params = array();
        }

        $this->getLog()->info(__LINE__, $type . ': ' . $url);
        // @codingStandardsIgnoreLine
        $this->getLog()->info(__LINE__, 'Params: ' . print_r($params, true));
        $this->getLog()->debug(__LINE__, 'Auth: ' . \Models\PBX::getInstance()->getSerial() . ':' . \Models\PBX::getInstance()->getPbxKey());

        if (empty($headers)) {
            $headers = array('Accept: application/json');
        }

        $response = $this->makeRequest($type, $url, $params, $formData, $headers);
        $result   = $this->parse($response);

        // @codingStandardsIgnoreLine
        $this->getLog()->info(__LINE__, 'Response: ' . print_r($result, true));
        return $result;
    }

    protected function parse($response) {
        $decodedResult = json_decode($response['result'], true);
        if (!$decodedResult || !is_array($decodedResult)) {
            return array(
                'type' => 'error',
                'result' => $response['error']
            );
        }

        return $decodedResult;
    }
}
