<?php
namespace Wildix\Settings;

/**
 * Class ConfigFile
 */
class ConfigFile extends \Singleton {

    private $_file           = null;
    private $_defaultSection = null;
    private $_noQuotes       = null;
    protected $_delimiter    = null;
    protected $_config       = null;
    private $_logger         = null;
    private $_fileOwner      = null;
    private $_fileGroup      = null;
    private $_default        = [];

    public function __construct($file, $defaultSection = null, $noQuotes = false, $delimiter = '=', $fileOwner = 'wms', $fileGroup = 'wms') {
        $this->_file           = $file;
        $this->_defaultSection = $defaultSection;
        $this->_noQuotes       = $noQuotes;
        $this->_delimiter      = $delimiter;
        $this->_fileOwner      = $fileOwner;
        $this->_fileGroup      = $fileGroup;
        $this->_logger         = \Logger\LoggerFactory::getInstance();
    }

    public function setDefault(array $data) {
        $this->_default = $data;
    }

    public function getDefault(): array {
        return $this->_default;
    }

    protected function parseNotIniFormat($iniContent) {
        $config = array();

        $sectionName = '';
        if (preg_match_all('/\s\[([^\]]+)\]\s/i', $iniContent, $matches)) {
            // if have sections, set the first section as a global
            $sectionName = 'global';
        }
        $iniContent = preg_split("/\\r\\n|\\r|\\n/", $iniContent);
        foreach ($iniContent as $line) {
            $line = trim($line);
            if (strpos($line, ';') === 0) {
                continue;
            }
            if (preg_match('/^\[([^\]]+)\]$/i', $line, $matches)) {
                // if section name
                if (!empty($matches[1])) {
                    $sectionName = $matches[1];
                    continue;
                }
            }

            $exploded = explode('=', $line);
            $option   = trim($exploded[0]);
            unset($exploded[0]);

            $value = trim(join('=', $exploded));

            if (!empty($option)) {
                if ($sectionName != '') {
                    $config[$sectionName][$option] = $value;
                } else {
                    $config[$option] = $value;
                }
            }
        }
        return $config;
    }

    protected function parseIniFile(string $iniContent) {
        return parse_ini_string($iniContent, true, INI_SCANNER_RAW);
    }

    public function get($param = null, $section = null) {
        if ($section === null) {
            $section = $this->_defaultSection;
        }

        if ($this->_config === null) {
            $this->_config = array();

            $iniContent = is_readable($this->_file) ? file_get_contents($this->_file) : false;

            if ($iniContent) {
                $iniContent = preg_replace('/^#/m', ';', $iniContent);
                $iniContent = preg_replace('/ => /', ' = ', $iniContent);
                $iniContent = preg_replace('/^[ \t]*=.*/m', '', $iniContent);
                $file       = $this->parseIniFile($iniContent);
                if ($file === false) {
                    $file = $this->parseNotIniFormat($iniContent);
                }

                $this->_config = $file;
            }
        }

        $sectionContent = $this->_config;

        if ($param !== null || $section !== null) {
            if ($section !== null) {
                if (isset($this->_config[$section])) {
                    $sectionContent = $this->_config[$section];
                } else {
                    $sectionContent = null;
                }
            }
            if ($param !== null) {
                if ($sectionContent && isset($sectionContent[$param])) {
                    return $sectionContent[$param];
                } else {
                    return null;
                }
            }
        }
        return $sectionContent;
    }

    public function getBoolean($param = null, $section = null) {
        $val = $this->get($param, $section);

        if ($val === 'no' || $val === '0' || $val === 0 || $val === '' || $val === null) {
            // if desabled
            return false;
        }
        return true;
    }

    public function set($param, $value, $section = null) {
        if ($this->_config === null) {
            $this->get();
        }

        if ($section === null) {
            $this->_config[$param] = $value;
        } else {
            $this->_config[$section][$param] = $value;
        }
        return $this;
    }

    public function setData(array $data) {
        $this->_config = $data;
    }

    public function has($param, $section = null): bool {
        $res = $this->get($param, $section);
        return boolval($res);
    }

    /**
     * Save only INI file
     *
     * @return boolean
     */
    public function save() {
        $content = $this->prepareContent();

        if (!is_writable($this->_file)) {
            exec('sudo /bin/chown ' . $this->_fileOwner . '.' . $this->_fileGroup . ' ' . $this->_file);
        }

        $handle = fopen($this->_file, 'w');
        if ($handle) {
            if (fwrite($handle, $content) === false) {
                $this->_logger->error(__LINE__, 'Cannot write config file: ' . $this->_file);
                return false;
            }
            fclose($handle);
            return true;
        }

        return false;
    }

    /**
     * Prepare content for save
     *
     * @return string
     */
    protected function prepareContent() {
        $content    = '';
        $parameters = $this->get();

        foreach ($parameters as $key => $item) {
            if (is_array($item)) {
                // write section
                $content .= PHP_EOL . '[' . $key . ']' . PHP_EOL;
                // write key: value
                foreach ($item as $subKey => $subItem) {
                    $content .= $this->createContentRow($subKey, $subItem, $this->_delimiter);
                }
            } else {
                $content .= $this->createContentRow($key, $item, $this->_delimiter);
            }
        }

        return $content;
    }

    protected function createContentRow($key, $item, $delimiter) {
        // write key: value
        $quote = '"';
        if (is_numeric($item) || is_bool($item) || $this->_noQuotes) {
            $quote = '';
        }

        if ($item === true) {
            $item = 'true';
        } else if ($item === false) {
            $item = 'false';
        }

        return $key . $delimiter . $quote . $item . $quote . PHP_EOL;
    }

    public function getPath(): string {
        return (string)$this->_file;
    }
}
