<?php

namespace Wildix\Settings;

use Wildix\System\Systemd;

/**
 * Class RsyslogLogging
 */
class RsyslogLogging {
    const RULE_REMOTE_SERVER = 'remote-server';

    private static $_locationRsyslogD          = '/rw2/etc/rsyslog.d';
    private static $_templateConfigurationRule = '# This file is managed from the WMS interface.
# Changes made in manual mode can be lost at any time.

%s %s';
    private static $_rulesUnderControl         = array(
                                                  self::RULE_REMOTE_SERVER,
    );

    /**
     * Instance of rsyslog daemon.
     *
     * @var Systemd
     */
    private $_rsyslogDaemon;

    /**
     * Syslog constructor.
     *
     * @param Systemd $rsyslogDaemon Instance of rsyslog daemon.
     */
    public function __construct(Systemd $rsyslogDaemon) {
        $this->_rsyslogDaemon = $rsyslogDaemon;
    }

    /**
     * Remove the rsyslog configuration file.
     *
     * @param string $ruleName This name is define the rsyslog configuration file.
     *
     * @return void
     * @throws \InvalidArgumentException If the rule isn't under control.
     */
    public function removeRule($ruleName) {
        $this->_assertRuleIsAvailable($ruleName);

        $configurationFileLocation = $this->_createConfigurationFileLocation($ruleName);

        if (!file_exists($configurationFileLocation)) {
            return;
        }

        unlink($configurationFileLocation);

        $this->_rsyslogDaemon->reloadOrRestart();
    }

    /**
     * Changing the selections and action of the rule in a configuration file.
     *
     * @param string $ruleName  This name is define the rsyslog configuration file.
     * @param string $selectors Message selectors for which to apply the action.
     * @param string $action    Destination for recording messages on specified selectors.
     *
     * @return void
     * @throws \InvalidArgumentException If the rule isn't under control.
     */
    public function updateRule($ruleName, $selectors, $action) {
        $this->_assertRuleIsAvailable($ruleName);

        file_put_contents($this->_createConfigurationFileLocation($ruleName), sprintf(self::$_templateConfigurationRule, $selectors, $action));

        $this->_rsyslogDaemon->reloadOrRestart();
    }

    private function _assertRuleIsAvailable($ruleName) {
        if (!in_array($ruleName, self::$_rulesUnderControl)) {
            throw new \InvalidArgumentException(sprintf('The rule name "%s" is illegal', $ruleName));
        }
    }

    private function _createConfigurationFileLocation($ruleName) {
        return sprintf('%s/%s.conf', self::$_locationRsyslogD, $ruleName);
    }
}
