<?php

namespace Wildix\Settings;

/**
 * Class TurnServer
 */
class TurnServer {

    public const TURN_SERVER_CONF = '/etc/turnserver/turnserv.conf';

    public const MIN_PORT = 'min-port';
    public const MAX_PORT = 'max-port';

    /**
     * Turn server system service.
     *
     * @var \Wildix\System\Systemd
     */
    private $_serviceTurnServer;
    /**
     * Logger.
     *
     * @var \Logger\Log
     */
    private $_log;
    /**
     * Config params.
     *
     * @var array
     */
    private $_config;

    public function __construct(\Wildix\System\Systemd $serviceTurnServer, \Logger\Log $log) {
        $this->_serviceTurnServer = $serviceTurnServer;
        $this->_log               = $log;

        $this->_config = $this->parseConfig();
    }

    public function getConfig(): array {
        return $this->_config;
    }

    /**
     * Get parameter value.
     *
     * @param string $param Param name.
     *
     * @return mixed|null
     */
    public function getParam(string $param) {
        return $this->_config[$param] ?? null;
    }

    /**
     * Set parameter value.
     *
     * @param string $param Param name.
     * @param mixed  $value Param value.
     *
     * @return void
     */
    public function setParam(string $param, $value): void {
        $this->_config[$param] = $value;
    }

    /**
     * Save config and restart turnserver service.
     *
     * @return void
     */
    public function saveConfig(): void {
        if (!is_file(self::TURN_SERVER_CONF) || !is_readable(self::TURN_SERVER_CONF)) {
            $this->getLog()->error(__LINE__, 'Turnserver config not exist or not readable: ' . self::TURN_SERVER_CONF);
            return;
        }

        $turnServerConf = file(self::TURN_SERVER_CONF);

        $this->getLog()->info(
            __LINE__,
            [
                'message' => 'Save turnserver configuration',
                'turnserverParams' => \json_encode($this->_config),
            ]
        );

        $minPort = $this->_config[self::MIN_PORT] ?? 0;
        $maxPort = $this->_config[self::MAX_PORT] ?? 0;

        $minPortRegExp = $this->getPortRegExp(self::MIN_PORT);
        $maxPortRegExp = $this->getPortRegExp(self::MAX_PORT);

        foreach ($turnServerConf as $lineNum => $line) {
            if (\preg_match($minPortRegExp, $line) && $minPort > 0) {
                $turnServerConf[$lineNum] = \sprintf(self::MIN_PORT . "=%d\n", $minPort);
            } else if (\preg_match($maxPortRegExp, $line) && $maxPort > 0) {
                $turnServerConf[$lineNum] = \sprintf(self::MAX_PORT . "=%d\n", $maxPort);
            }
        }

        file_put_contents(self::TURN_SERVER_CONF, implode('', $turnServerConf));

        $this->changeOwner(self::TURN_SERVER_CONF);

        $this->getLog()->info(__LINE__, 'Restart turnserver service');
        $this->_serviceTurnServer->restart();
    }

    private function getLog(): \Logger\Log {
        return $this->_log;
    }

    /**
     * Parse turn server config file.
     *
     * Config example:
     *
     * no-udp
     * no-tls
     * denied-peer-ip=0.0.0.0-0.255.255.255
     * denied-peer-ip=127.0.0.0-127.255.255.255
     * denied-peer-ip=169.254.0.0-169.254.255.255
     * denied-peer-ip=240.0.0.0-255.255.255.255
     * no-tcp-relay
     * no-multicast-peers
     * no-loopback-peers
     * fingerprint
     * lt-cred-mech
     * realm=127.0.0.1
     * min-port=60000
     * max-port=60999
     * listening-ip=127.0.0.1
     * listening-port=3478
     * relay-ip=0.0.0.0
     * relay-threads=1
     * wildix-ldap="url=ldap://localhost:389 base_dn=dc=wildix dn=cn=admin,dc=wildix pwd=wildix filter_ptrn=(&(objectClass=wPerson)(telephoneNumber=%s))"
     * pidfile=/var/run/turnserver.pid
     *
     * @return array
     */
    private function parseConfig(): array {
        $config = [];

        if (!is_file(self::TURN_SERVER_CONF) || !is_readable(self::TURN_SERVER_CONF)) {
            $this->getLog()->error(__LINE__, 'Turnserver config not exist or not readable: ' . self::TURN_SERVER_CONF);
            return $config;
        }

        $configParams = file(self::TURN_SERVER_CONF, FILE_IGNORE_NEW_LINES | FILE_SKIP_EMPTY_LINES);

        $portRegexp = '/^(' . self::MIN_PORT . '|' . self::MAX_PORT . ')\s*=\s*(\d*)\s*$/';

        foreach ($configParams as $line) {
            $line    = \ltrim($line);
            $matches = [];

            if (\preg_match($portRegexp, $line, $matches)) {
                $config[\trim($matches[1])] = (int)\trim($matches[2]);
            }
        }

        $this->getLog()->info(
            __LINE__,
            [
                'message' => 'Parse turnserver config',
                'parsedParams' => \json_encode($config),
            ]
        );

        return $config;
    }

    /**
     * Get regexp for port parameters.
     *
     * @param string $name Port parameter name.
     *
     * @return string
     */
    private function getPortRegExp(string $name): string {
        return '/^\s*' . $name . '\s*=\s*\d*/';
    }

    /**
     * Change owner and group for file.
     *
     * @param string $file File path.
     *
     * @return void
     */
    private function changeOwner(string $file): void {
        if (is_file($file)) {
            $this->getLog()->info(__LINE__, \sprintf('Change user and group for file: %s', $file));

            chown($file, 'wms');
            chgrp($file, 'wms');
        }
    }
}
