<?php

namespace Wildix;

use Cache\FileCache;

/**
 * SipProxy Class.
 */
class SipProxy extends \Wildix\Services\AbstractService {

    private const USERS_ACTIVE_DEVICE_CACHE_TIME = 60;

    private $_sipProxyGetAllUsersUrl                = 'http://127.0.0.1:5060/v1/users/';
    private $_sipProxyGetAllDynamicTrunksUrl        = 'http://127.0.0.1:5060/v1/trunks/dynamic/';
    private $_sipProxyUsersActiveDevicesUrlTemplate = 'http://127.0.0.1:5060/v1/users/%s/devices/active/';
    private $_communicateScriptUserAgent            = 'WMSCommunicate Script';
    private $_sipProxyGetWairDevicesUrl             = 'http://127.0.0.1:5060/v1/wair_devices/';

    private $_cacheAction;
    private $_cacheTime     = 0;
    private $_requestParams = '';
    private $_showOffline;
    /**
     * Instance of \Cache\FileCache.
     *
     * @var \Cache\FileCache
     */
    private $_fileCache;

    /**
     * SipProxy constructor.
     *
     * @param \Cache\FileCache $cache Dependency.
     *
     * @return void
     */
    public function __construct(\Cache\FileCache $cache) {
        // for SipProxy class logLevel is 0.
        parent::__construct(0);
        $this->_fileCache = $cache;
        $this->setUserAgent($this->_communicateScriptUserAgent);
    }

    /**
     * Get Registrations from cache or SipProxy server.
     *
     * @param string $url Api Url to SipProxy server.
     *
     * @return array
     */
    private function getRegistrations(string $url) : array {
        $fromCache = $this->_fileCache->get($this->_cacheAction . $this->_requestParams . (int)$this->_showOffline);

        if ($fromCache) {
            return $fromCache;
        }

        try {
            $serviceRequest = $this->sendRequest('GET', $url . $this->_requestParams);
        } catch (\Exception $e) {
            $serviceRequest = [];
        }

        return $this->parseSipProxyResponse($serviceRequest);
    }

    /**
     * Parse response from SipProxy server.
     *
     * @param array $response Response from SipProxy server.
     *
     * @return array
     */
    private function parseSipProxyResponse(array $response) : array {
        $return = [];

        foreach ($response as $value) {
            $return[$value['extension']]['registrations'] = $value['registrations'];
        }

        // show only registrations with "online" => 1
        if (!$this->_showOffline) {
            $result = [];
            foreach ($return as $key => $value) {
                $registrationsCount = count($value['registrations']);
                for ($i = 0; $i < $registrationsCount; $i++) {
                    if ($value['registrations'][$i]['online'] == 1) {
                        $result[$key]['registrations'][$i] = $value['registrations'][$i];
                    }
                }
            }
            $return = $result;
        }

        $this->_fileCache->set($this->_cacheAction . $this->_requestParams . (int)$this->_showOffline, $return, $this->_cacheTime);

        // @codingStandardsIgnoreLine
        $this->getLog()->info(__LINE__, $this->_cacheAction . 'parseSipProxyResponse return array: ' . print_r($return, true));

        return $return;
    }

    /**
     * Sets requestParams for future request.
     *
     * @param array $params Array of params which will be glued with comma.
     *
     * @return void
     */
    private function setRequestParams(array $params) : void {
        if (!empty($params)) {
            $this->_requestParams = join(',', $params);
        }
    }

    /**
     * Get users registrations.
     *
     * @param array   $extensions  Array of extensions.
     * @param boolean $showOffline Flag for showing offline registrations.
     *
     * @return array
     */
    public function getUsersRegistrations(array $extensions = [], bool $showOffline = false) : array {
        $this->setRequestParams($extensions);
        $this->_cacheAction = FileCache::USER_REGISTRATIONS;
        $this->_showOffline = $showOffline;

        return $this->getRegistrations($this->_sipProxyGetAllUsersUrl);
    }

    /**
     * Get trunk registrations.
     *
     * @param array   $login       Array of login.
     * @param boolean $showOffline Flag for showing offline registrations.
     *
     * @return array
     */
    public function getTrunksRegistrations(array $login = [], bool $showOffline = false) : array {
        $this->setRequestParams($login);
        $this->_cacheAction = FileCache::TRUNK_REGISTRATIONS;
        $this->_showOffline = $showOffline;

        return $this->getRegistrations($this->_sipProxyGetAllDynamicTrunksUrl);
    }

    /**
     * Get users active devices
     *
     * @param array $extensions List of extensions.
     *
     * @throws Services\Exception\ServiceException Failed get users active devices.
     * @return array
     */
    public function getUsersActiveDevices(array $extensions) : array {
        $cache = $this->_fileCache->get(FileCache::USERS_ACTIVE_DEVICES);

        if (!is_array($cache)) {
            $cache = [];
        }

        $cacheMisses = array_diff($extensions, array_keys($cache));

        if (count($cacheMisses) > 0) {
            $response = $this->sendRequest('GET', sprintf($this->_sipProxyUsersActiveDevicesUrlTemplate, implode(',', $cacheMisses)));

            foreach ($response as $device) {
                $cache[$device['extension']] = $device;
            }
        }

        $this->_fileCache->set(FileCache::USERS_ACTIVE_DEVICES, $cache, self::USERS_ACTIVE_DEVICE_CACHE_TIME);

        return array_intersect_key($cache, array_flip($extensions));
    }

    public function getWairDevices(): array {
        try {
            $result = $this->sendRequest('GET', $this->_sipProxyGetWairDevicesUrl);
        } catch (\Exception $e) {
            $result = [];
        }

        return $result;
    }
}
