<?php
declare(strict_types=1);

namespace Wildix\Sms;

use Wildix\Dialplan\CallGroups\CallGroupsService;
use Wildix\Colleagues\Colleague;
use Wildix\Colleagues\Colleagues;
use Wildix\Dialplan\DialplanService;
use Wildix\Dialplan\RulesService;
use Wildix\License;
use Wildix\Validator;

/**
 * Class SmsService
 */
class SmsService {

    const SMS_DIALPLAN = 'kite-sms';
    const SMS_APP_NAME = 'Queue';

    const RESULT_MEMBERS = 'members';
    const RESULT_SUBJECT = 'subject';

    const DEFAULT_SUPPORTED_COUNTRY_CODES = ['1', '44'];

    /**
     * Colleagues class
     *
     * @var Colleagues
     */
    private $_colleagues;

    /**
     * Dialplan service.
     *
     * @var DialplanService
     */
    private $_dialplanService;

    /**
     * Dialplan rules array
     *
     * @var array
     */
    private $_dialplanRules;

    /**
     * Callgroups array
     *
     * @var array
     */
    private $_callgroups = [];

    /**
     * Dialplan rules service
     *
     * @var RulesService
     */
    private $_dialplanRulesService;

    /**
     * Validator class
     *
     * @var Validator
     */
    private $_validator;

    /**
     * Validator class
     *
     * @var CallGroupsService
     */
    private $_callGroups;

    /**
     * Destination phone number
     *
     * @var string
     */
    private $_dstNumber;

    /**
     * SmsService constructor.
     *
     * @param Colleagues        $colleagues   Colleagues class.
     * @param DialplanService   $dialplan     Dialplan service.
     * @param RulesService      $rulesService Rules service.
     * @param Validator         $validator    Validator class.
     * @param CallGroupsService $callGroups   CallGroupsService class.
     */
    public function __construct(
        Colleagues $colleagues,
        DialplanService $dialplan,
        RulesService $rulesService,
        Validator $validator,
        CallGroupsService $callGroups
    ) {
        $this->_colleagues           = $colleagues;
        $this->_dialplanService      = $dialplan;
        $this->_dialplanRulesService = $rulesService;
        $this->_validator            = $validator;
        $this->_callGroups           = $callGroups;
    }

    /**
     * Returns colleagues from dialplan rule or by office number.
     *
     * @param string $src Source number.
     * @param string $dst Destination number.
     *
     * @return array
     *
     * @throws \Wildix\InvalidDataException Invalid data
     */
    public function getColleagues(string $src, string $dst) : array {
        $validateData = [
            'src' => $src,
            'dst' => $dst
        ];

        $this->_validator->rule('required', 'src');
        $this->_validator->rule('required', 'dst');

        $this->_validator->rule('PhoneNumber', 'src');
        $this->_validator->rule('PhoneNumber', 'dst');

        $validateErrors = $this->_validator->validateData($validateData, ['src', 'dst']);
        if ($validateErrors) {
            throw new \Wildix\InvalidDataException(\Wildix\Error::getMessage(\Wildix\Error::InvalidData));
        }

        $src              = $this->formatPhoneNumber($src);
        $this->_dstNumber = $this->formatPhoneNumber($dst);

        $result = $this->getColleaguesFromDialplan();
        if ($result === []) {
            $result = $this->getColleaguesByOfficeNumber();
        }

        if ($result !== []) {
            $result = [
                self::RESULT_MEMBERS => $result,
                self::RESULT_SUBJECT => $src,
            ];
        }
        return $result;
    }

    /**
     * Returns colleagues from dialplan.
     *
     * @return array
     *
     * @throws \Exception Error in dialplan service.
     */
    private function getColleaguesFromDialplan() : array {
        $result = [];

        $smsDialplan = $this->_dialplanService->getDialplanByName(self::SMS_DIALPLAN);
        if ($smsDialplan === null) {
            return $result;
        }

        $this->_dialplanRules = $this->_dialplanRulesService->getRulesByDialplan($smsDialplan[DialplanService::PARAM_ID]);
        $this->filterDialplanNumbers();

        if ($this->_dialplanRules === []) {
            return $result;
        }

        array_map([$this, 'fillCallgroups'], $this->_dialplanRules);

        if ($this->_callgroups === []) {
            return $result;
        }

        $callgroupMembers = $this->_callGroups->loadMembersByIds($this->_callgroups);

        if ($callgroupMembers === []) {
            return $result;
        }

        $colleagues = $this->_colleagues->getUsers(
            [
                Colleague::EXTENSION => explode(',', str_replace('@internalcalls', '', implode(',', $callgroupMembers))),
                Colleague::LICENSE_TYPE => [
                    License::TYPE_PREMIUM,
                    License::TYPE_BUSINESS,
                ],
            ]
        );

        return array_map(
            function ($item) {
                return $item->getJid();
            },
            $colleagues
        );
    }

    /**
     * Returns colleagues by office number.
     *
     * @return array
     *
     * @throws \Exception Error in Colleagues class.
     */
    private function getColleaguesByOfficeNumber() : array {
        $colleagues = $this->_colleagues->getUsers(
            [
                Colleague::OFFICE_PHONE => $this->_dstNumber,
                Colleague::LICENSE_TYPE => [
                    License::TYPE_PREMIUM,
                    License::TYPE_BUSINESS,
                ],
            ]
        );

        return array_map(
            function ($item) {
                return $item->getJid();
            },
            $colleagues
        );
    }

    /**
     * Filter founded in dialplan numbers for needed office number
     *
     * @return void
     */
    private function filterDialplanNumbers() : void {
        $this->_dialplanRules = array_filter(
            $this->_dialplanRules,
            function ($item) {
                return $item[RulesService::PARAM_NUMBER] === $this->_dstNumber;
            }
        );
    }

    /**
     * Fill callgroups array from application rule
     *
     * @param array $rule Application rule.
     *
     * @return void
     */
    private function fillCallgroups(array $rule) {
        foreach ($rule[RulesService::PARAM_APPS] as $app) {
            if ($app[RulesService::PARAM_APP_NAME] === self::SMS_APP_NAME) {
                $this->_callgroups[] = $app[RulesService::PARAM_APP_PARAMS]['queueId'];
                // only first founded callgroup (temporary)
                break;
            }
        }
    }

    /**
     * Format phone number to e164
     * Adding "+" is it needs
     *
     * @param string $number Phone number.
     *
     * @return string
     */
    private function formatPhoneNumber(string $number) {
        return (strpos($number, '+') !== 0 ? '+' : '') . $number;
    }
}
