<?php

namespace Wildix\Sms;

use Wildix\Databases\WMSDB;
use Wildix\Exception;
use Wildix\License;
use Wildix\Sms\Dto\SmsViaGsmDto;
use Wildix\Sms\Exception\SmsViaGsmQueueException;
use Wildix\System\Sms;

/**
 * Class SmsViaGsmService
 */
class SmsViaGsmService {

    /**
     * User instance.
     *
     * @var \User
     */
    private $_user;

    /**
     * SmsViaGsmService constructor.
     *
     * @param \User|null $user Active User.
     */
    public function __construct(?\User $user) {
        $this->_user = $user;
    }

    public function sendSms(SmsViaGsmDto $smsViaGsmDto) {
        $smsOptions = $smsViaGsmDto->getOptions();

        $user = $this->_user;

        $license = License::getInstance();
        if (! $license->haveLicense('faxes')) {
            throw new Exception(_t('error.noSmsLicence'), 403);
        }

        if ($user !== null && $user->canNotUse($user::SMS)) {
            throw new Exception(_t('error.noSmsAccess'), 403);
        }

        $fromNumber = trim($smsOptions['fromNumber']);
        // extension for regular users. otherwise(cli or root admin) use fromNumber option.
        if ($user !== null && ! $user->isRoot()) {
            $fromNumber = trim($user->getExtension());
        }

        $toNumbers = $smsOptions['number'];
        $message   = $smsOptions['message'] ?? base64_decode($smsOptions['message64']);


        $postpone   = trim($smsOptions['postpone'] ?? '');
        $gsmGateway = trim($smsOptions['gsm']);

        $setting = parse_ini_file('/rw2/etc/companies.d/0/variables.conf', false, INI_SCANNER_RAW);

        if (! $smsOptions['notHeaderMessage'] && $setting['GSMSTEXT_0']) {
            $message = base64_decode($setting['GSMSTEXT_0']) . $message;
        }

        $locPbx = \Wildix\PBXes\PBXes::getInstance()->getPBXBySerial(\Models\PBX::getInstance()->getSerial())->getDn();
        $netPbx = false;
        if ($setting['GSMSPBX_0'] && $setting['GSMSPBX_0'] != $locPbx && \Models\PBX::getInstance()->isEnabledWmsNetwork()) {
            $netPbx = \WMS_Ldap::getInstance()->getEntryByDn($setting['GSMSPBX_0'], array('pbxIp'));
        }

        if ($netPbx && !empty($netPbx['pbxIp'])) {
            // phpcs:ignore Generic.CodeAnalysis.EmptyStatement.DetectedIf
            if ($user === null) {
                // warning: can not send request to pbx in wms network from cli sapi
                // TODO WMS-10041
            } else {
                $this->sendFromRemotePbx($fromNumber, $toNumbers, $message, $postpone, $gsmGateway, $netPbx, $user);
            }

            return;
        }

        $device = null;
        if (!empty($gsmGateway)) {
            // DaySaver / W01GSM
            $device = \Wildix\Devices\Devices::getInstance()->getDeviceByMac($gsmGateway);
            if (!$device) {
                throw new Exception(_t('error.noGatewayFound'), 404);
            }
        }

        $gsmTrunksNotAvailable        = WMSDB::getInstance()->getCountGsmTrunks() === 0;
        $httpScriptDisabled           = ! Sms::getInstance()->get('HTTP_SCRIPT_ENABLE');
        $classoundLicenseNotAvailable = ! $license->haveLicense(License::TYPE_CLASSOUND);
        if ($gsmTrunksNotAvailable && $httpScriptDisabled && $classoundLicenseNotAvailable) {
            throw new Exception(_t('error.noGateway'), 400);
        }

        if (!empty($postpone)) {
            $postpone = \DateTime::createFromFormat('d/m/Y H:i', $postpone);
            if ($postpone === false) {
                throw new Exception(_t('error.badDateFormat'), 400);
            }

            $nowDate = new \DateTime();
            if ($postpone->getTimestamp() <= $nowDate->getTimestamp()) {
                throw new Exception(_t('error.dateInFuture'), 400);
            }
        }

        $result = \Models\Astman::getInstance()->sendSMS($fromNumber, $toNumbers, $message, $postpone, $device);
        if (is_array($result)) {
            $queueException = new SmsViaGsmQueueException(_t('error.cantAddSmsQueue'));
            $queueException->setFailedRecipient($result);
            throw $queueException;
        }
    }

    private function sendFromRemotePbx($fromNumber, $toNumbers, $message, $postpone, $gsmGateway, $pbx, $user) {
        if ($_SERVER['REMOTE_ADDR'] == $pbx['pbxIp']) {
            throw new Exception(_t('error.recursiveSmsServer'));
        }

        $data = array(
            'number' => $toNumbers,
            'message' => $message,
            'postpone' => $postpone,
            'fromNumber' => $fromNumber,
            'gsm' => $gsmGateway,
            'notHeaderMessage' => 1
        );

        $patch = 'http://' . $pbx['pbxIp'] . '/api/v1/Originate/sms';

        $ch = curl_init($patch);

        curl_setopt($ch, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
        curl_setopt($ch, CURLOPT_USERPWD, $user->getExtension() . ':' . $user->getPassword());
        curl_setopt($ch, CURLOPT_HEADER, true);
        curl_setopt($ch, CURLOPT_POST, true);
        curl_setopt($ch, CURLOPT_POSTFIELDS, http_build_query($data));
        curl_setopt($ch, CURLOPT_USERAGENT, 'PBX Api');
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_TIMEOUT, 10);
        curl_setopt($ch, CURLOPT_AUTOREFERER, true);

        $response     = curl_exec($ch);
        $responseInfo = curl_getinfo($ch);
        $response     = substr($response, $responseInfo['header_size']);

        curl_close($ch);

        if (!empty($response)) {
            $response = json_decode($response, true);
            if ($response['type'] === 'error') {
                throw new Exception($response['reason']);
            }

            if ($response['type'] === 'warning') {
                $queueException = new SmsViaGsmQueueException($response['reason']);
                $queueException->setFailedRecipient($response['result']);
                throw $queueException;
            }
        }
    }
}
