<?php
namespace Wildix;

use Logger\Log;
use Wildix\Databases\WMSDB;

class Smtp {

    const COMPANIES_DIRECTORY = '/etc/companies.d/';
    const SMTP_TEST_DIRECTORY = '/tmp/session/';
    const SMTP_TEST_END_FLAG  = '--#smtp test end#--';
    const DEFAULT_FROM_EMAIL  = 'no-reply@wildix.com';

    private $_log;
    private $_defaultSmtpHost;
    private $_pbxName;
    private $_wmsdb;

    private $knownOptions = [
            'account'      => 'default',
            'from'         => '',
            'host'         => '',
            'user'         => '',
            'password'     => '',
            'domain'       => 'localhost',
            'auth'         => 'off',
            'tls'          => 'off',
            'tls_starttls' => 'off',
            'timeout'      => '15',
            'port'         => '25',
            'syslog'       => 'on',
            'is_default_settings' => 'off'
// Note: is_default_settings is not part of msmtp config
    ];

    /**
     * Smtp constructor.
     *
     * @param Log    $logger          Logger class.
     * @param string $smtpDefaultHost Default SMTP HOST.
     */
    public function __construct(
        WMSDB $wmsdb,
        Log    $logger,
        string $smtpDefaultHost = '',
        string $pbxName = ''
    ) {
        $this->_wmsdb           = $wmsdb;
        $this->_log             = $logger;
        $this->_pbxName         = $pbxName;
        $this->_defaultSmtpHost = parse_url($smtpDefaultHost, PHP_URL_HOST);
    }

    /**
     * Retrieve msmtp daemon config path - /etc/companies.d/0/msmtprcphp
     *
     * @param  integer $profileId
     * @return string
     */
    public static function getMsmtpConfigPath($profileId = 0) {
        return self::COMPANIES_DIRECTORY . "$profileId/msmtprc";
    }

    /**
     * Retrieve msmtp options
     *
     * @param  integer $profileId
     * @return array|mixed
     */
    public function getMailOptions($profileId = 0) {
        $result = [];

        $value = $this->_getWmsDb()->getSystemVariable(WMSDB::SYSTEM_SMTP_SETTINGS);
        if (!is_null($value)) {
            $result = json_decode($value, true);
        } else {
            $result = $this->getMailOptionsFromConfig($profileId);

            $result['is_default_settings'] = (($result['host'] ?? '') === $this->_defaultSmtpHost) ? 'on' : 'off';
        }

        return $result;
    }

    /**
     * Save msmtp options to corresponding $profileId file.
     * Note: For now is only one profile so "account" will not transmitted and always be "default"
     *
     * @param  $options array of msmtp options
     * @param  integer                        $profileId
     * @return boolean
     */
    public function setMailOptions($options, $profileId = 0) {
        // Save only known options of msmtp, other left with default values
        $saveOptions = $this->knownOptions;
        foreach ($saveOptions as $key => $optValue) {
            if (isset($options[$key])) {
                $saveOptions[$key] = $options[$key];
            }
        }

        if ($saveOptions['host'] === $this->_defaultSmtpHost) {
            $saveOptions['is_default_settings'] = Validator::ON;
        }

        if (Validator::isTrue($saveOptions['is_default_settings'])) {
            $saveOptions['host'] = $this->_defaultSmtpHost;
            $saveOptions['from'] = self::DEFAULT_FROM_EMAIL;
        }

        if (Validator::isTrue($saveOptions['tls'])) {
            $saveOptions['tls_certcheck'] = 'off';
        }

        $this->_getWmsDb()->setSystemVariable(WMSDB::SYSTEM_SMTP_SETTINGS, json_encode($saveOptions));

        // Do not save option "is_default_settings" to msmtp config file
        unset($saveOptions['is_default_settings']);

        // "from" save muttrc & msmtprc
        // "realname" = get_company
        // include "source /etc/Muttrc" as template
        if (! file_exists($muttrc = self::getMuttConfigPath($profileId))) {
            @file_put_contents($muttrc, 'source /etc/Muttrc');
        }
        if (! file_exists($msmtprc = self::getMsmtpConfigPath($profileId))) {
            @touch($msmtprc);
        }

        if (!is_writable($muttrc)) {
            throw new Exception($muttrc . ' is not writeable',500);
        }
        if (!is_writable($msmtprc)) {
            throw new Exception($msmtprc . ' is not writeable',500);
        }

        $s2file = '';
        foreach ($saveOptions as $key => $value) {
            if (trim($value) !== '') {
                $s2file .= $key . ' ' . trim($value) . PHP_EOL;
            }
        }

        if ($saveOptions['account'] != 'default') {
            $s2file .= 'account default: ' . $saveOptions['account'] . PHP_EOL;
        }

        // to /etc/msmtprc
        file_put_contents($msmtprc, $s2file);
        chmod($msmtprc, 0600);

        // to Muttrc
        file_put_contents(
            $muttrc,
            [
                "set sendmail=\"/usr/bin/msmtp-enqueue.sh $profileId\"\n",
                "set use_from=yes\n",
                "set from={$saveOptions['from']}\n",
                "set realname=\"" . $this->getPbxName() . "\"\n",
                "set copy=no\n",
                "set charset=\"utf-8\"\n",
            ]
        );
        chown($muttrc, 'wms');
        chgrp($muttrc, 'wms');
        chmod($muttrc, 0655);

        return $this->getMailOptions($profileId);
    }

    public function runTest($data, $profileId = 0) {
        $mailTo   = $data['mailto'];
        $smtpFrom = $data['from'];

        $testId = 'default';

        $currentDate = date('r');
        $hostName    = file_get_contents('/etc/hostname');
        $messageId   = substr(md5(time() . rand()), 0, 15) . '@' . $hostName;

        $mailFile = tempnam('/tmp/session', 'mail_');
        $endFlag  = self::SMTP_TEST_END_FLAG;

        file_put_contents(
            $mailFile,
            [
                "Date: $currentDate" . PHP_EOL,
                "From: {$this->getPbxName()} ($hostName) <$smtpFrom>" . PHP_EOL,
                "To: $mailTo" . PHP_EOL,
                "Subject: Test message from {$this->getPbxName()}" . PHP_EOL,
                "Message-ID: <$messageId>" . PHP_EOL,
                "Mime-Version: 1.0" . PHP_EOL,
                "Content-Type: text/plain; charset=us-ascii" . PHP_EOL,
                "Content-Disposition: inline" . PHP_EOL,
                "User-Agent: Mutt/1.5.9i" . PHP_EOL,
                "" . PHP_EOL,
                $data['body'],
            ]
        );

        $resultTestOutFilename = self::getSmtpTestPath($testId) . '_out';
        $resultTestErrFilename = self::getSmtpTestPath($testId) . '_err';
        file_put_contents($resultTestOutFilename, "Starting sendmail... " . PHP_EOL);

        $comConfigFile = self::getMsmtpConfigPath($profileId);

        if (!file_exists($mailFile)) {
            return ['queued' => false];
        }

        // Sanitize before pass to shell
        $mailTo        = escapeshellcmd($mailTo);
        $comConfigFile = escapeshellcmd($comConfigFile);

        $resultTestOutFilename = escapeshellcmd($resultTestOutFilename);
        $resultTestErrFilename = escapeshellcmd($resultTestErrFilename);
        // verbose; removed --timeout=4 (should taken from config)
        $taskCmd = "nohup bash -c \"{ msmtp --file=$comConfigFile -d -t $mailTo < $mailFile >>$resultTestOutFilename 2>$resultTestErrFilename; rm $mailFile; echo '$endFlag' >> $resultTestOutFilename ; } & \" > /dev/null 2>/dev/null & ";
        $this->getLog()->info(__LINE__, 'Run smtp test comand: ' . $taskCmd);
        exec($taskCmd, $output, $result);

        return [
            'queued' => $result === 0,
            'testId' => $testId,
        ];
    }

    public function getTestStatus($testId) {
        $resultTestOutFilename = self::getSmtpTestPath($testId) . '_out';
        $resultTestErrFilename = self::getSmtpTestPath($testId) . '_err';
        if (!file_exists($resultTestOutFilename) || !check_read_file($resultTestOutFilename)) {
            return [
                'queued' => false,
                'testId' => $testId,
            ];
        }
        $resultContent = file_get_contents($resultTestOutFilename);
        if (file_exists($resultTestErrFilename) && check_read_file($resultTestErrFilename)) {
            $resultContent .= file_get_contents($resultTestErrFilename);
        }
        $isEnded = strpos($resultContent, self::SMTP_TEST_END_FLAG) !== false;
        if ($isEnded) {
            $resultContent = str_replace(self::SMTP_TEST_END_FLAG, '', $resultContent);
            unlink($resultTestOutFilename);
            unlink($resultTestErrFilename);
        }
        return [
            'content' => $resultContent,
            'queued' => !$isEnded,
            'testId' => $testId,
        ];
    }

    public function isConfigSynchronized(): bool {
        $value = $this->_getWmsDb()->getSystemVariable(WMSDB::SYSTEM_SMTP_SETTINGS);
        if (!is_null($value)) {
            $dbOptions = json_decode($value, true);
            unset($dbOptions['is_default_settings']);
        } else {
            return false;
        }

        $confOptions = $this->getMailOptionsFromConfig();
        return count(array_diff_assoc($dbOptions, $confOptions)) == 0 &&
            count(array_diff_assoc($confOptions, $dbOptions)) == 0;
    }

    public function isDefaultConfig(): bool {
        $options = $this->getMailOptions();

        return $options['host'] === $this->_defaultSmtpHost || $options['host'] === '';
    }

    /**
     * Get logger class.
     *
     * @return Log
     */
    private function getLog(): Log {
        return $this->_log;
    }

    private static function getMuttConfigPath($profileId = 0) {
        return self::COMPANIES_DIRECTORY . "$profileId/Muttrc";
    }

    private static function getSmtpTestPath($testId) {
        return self::SMTP_TEST_DIRECTORY . "smtp_test_" . $testId;
    }

    private function getMailOptionsFromConfig($profileId = 0) {
        $options = $this->knownOptions;
        unset($options['is_default_settings']);

        // get options from /etc/msmtprc
        $configContent = @file(self::getMsmtpConfigPath($profileId));

        if ($configContent !== false) {
            foreach ($configContent as $line) {
                if (strpos($line, 'account ') !== false && strpos($line, ':') !== false) {
                    continue;
                }

                $tmp = array_map('trim', explode(' ', trim($line)));
                if (isset($options[$tmp[0]])) {
                    $options[$tmp[0]] = $tmp[1];
                }
            }
        }

        return $options;
    }

    private function getPbxName(): string {
        return $this->_pbxName;
    }

    private function _getWmsDb(): WMSDB {
        return $this->_wmsdb;
    }
}
