<?php
declare(strict_types=1);

namespace Wildix\Sounds;

use Services\Helpers;
use Services\ShellHelper;
use Wildix\Error;
use Wildix\InvalidDataException;
use Wildix\Services\TextToSpeech;

/**
 * Class SoundsService
 */
class SoundsService {
    public const REAL_PATH              = '/var/lib/callweaver/sounds/00000/';
    private const RING_TONES_PATH       = self::REAL_PATH . 'RingTones/';
    public const WEB_PATH               = '/spoolview/sounds/00000/';
    private const MAX_UPLOAD_FILE_SIZE  = 10485760;
    private const EXCLUDE_CUSTOM_SOUNDS = '/var/lib/wms/exclude_custom_sounds.txt';
    public const FILE_EXTENSIONS        = [
        '.mp3',
        '.alaw',
        '.ulaw',
        '.wav',
        '.wv',
        '.gsm',
        '.g729',
        '.opus',
    ];

    private $_textToSpeechService;
    private $_defaultSounds;
    private $_storage;


    public function __construct(Storage $storage) {
        $this->_storage = $storage;
    }

    /**
     * Set TextToSpeech service.
     *
     * @param TextToSpeech $service TextToSpeech.
     *
     * @return void
     */
    public function setTextToSpeech(TextToSpeech $service): void {
        $this->_textToSpeechService = $service;
    }

    /**
     * Get Audio from params.
     *
     * @param array $params Params.
     *
     * @return mixed
     */
    private function _getAudio(array $params) {
        return $this->_textToSpeechService->getAudio($params);
    }


    /**
     * Get storage.
     *
     * @return Storage
     */
    private function _storage(): Storage {
        return $this->_storage;
    }

    /**
     * Get items.
     *
     * @param string $path Path.
     *
     * @return array
     * @throws InvalidDataException Exception.
     */
    public function getItems(string $path = ''): array {
        $result = [];
        $this->_checkDirectory($path);
        foreach ($this->_storage()->getItems($path) as $item) {
            $result[] = $this->_buildItem($item);
        }
        return $result;
    }

    /**
     * Build item.
     *
     * @param array $files Files.
     *
     * @return array
     */
    private function _buildItem(array $files): array {
        $item = [
            'name' => $files['name'],
            'path' => $files['path'],
            'size' => $files['size'],
            'isDir' => $files['isDir'],
            'isDefault' => $this->_isDefaultItem($files['path']),
            'modified'  => $files['modified'],
            'formats' => [],
        ];
        if (!$files['isDir']) {
            foreach ($files['formats'] as $file) {
                $item['formats'][] = [
                    'type' => strtoupper($file['type']),
                    'path' => self::WEB_PATH . $file['path']
                ];
            }
        }
        return $item;
    }

    /**
     * Create item from params.
     *
     * @param array $params Params.
     *
     * @return void
     *
     * @throws InvalidDataException InvalidDataException.
     */
    public function create(array $params): void {
        $dir = rtrim($params['directory'], '/') . '/';
        $this->_checkEmptyParams($params);
        $this->_checkDirectory($dir);
        if (!empty($params['text'])) {
            $params['name'] = Helpers::fileNamePrepare($params['name'] . '.' . $params['type']);
            $file           = $this->_getSpeechTextFile($params, $dir);
        } else {
            $params['name'] = Helpers::fileNamePrepare($params['file']->getClientFileName());
            $file           = $this->_getUploadedFile($params, $dir);
        }
        $destination = $dir . $params['name'];
        $this->_storage()->create($file, $destination);
        [$filename] = explode('.', $params['name']);
        $output     = $dir . $filename;
        $this->_convertFiles($this->realPath($destination), $this->realPath($output));
        if (strpos($destination, 'MusicOnHold') !== false) {
            \PBX::getInstance()->put_cli_reload('moh');
        }
    }

    /**
     * Edit item.
     *
     * @param string $path   Item path.
     * @param array  $params Data.
     *
     * @return void.
     *
     * @throws InvalidDataException InvalidDataException
     */
    public function update(string $path, array $params): void {
        $pathParams = explode('/', $path);
        $basename   = end($pathParams);
        [$fileName] = explode('.', $basename);
        $newPath    = Helpers::fileNamePrepare($params['name']) ?: $fileName;
        array_pop($pathParams);
        if (!empty($params['move'])) {
            $pathParams = [];
            $newPath    = $params['move'] . '/' . $newPath;
        }
        $pathParams[] = $newPath;
        $newPath      = str_replace('//', '/', implode('/', $pathParams));
        $this->_validateFileUpdate($path, $newPath);
        $this->_storage()->update($path, $newPath);
    }

    /**
     * Remove item.
     *
     * @param string $path Path.
     *
     * @return void.
     * @throws InvalidDataException Exception.
     */
    public function remove(string $path): void {
        $this->_storage()->remove($path);
    }

    /**
     * Speech text.
     *
     * @param array  $params      Params.
     * @param string $destination Destionation.
     *
     * @return string
     * @throws InvalidDataException Exception.
     */
    private function _getSpeechTextFile(array $params, string $destination): string {
        $this->_validateFileCreate($params['name'], $destination);
        return $this->_getAudio($params);
    }

    /**
     * Upload file.
     *
     * @param array  $params      File.
     * @param string $destination Destination.
     *
     * @return string
     * @throws InvalidDataException Exception.
     */
    private function _getUploadedFile(array $params, string $destination): string {
        $this->_validateFileCreate($params['name'], $destination);

        if (filesize($params['file']->getFilePath()) > self::MAX_UPLOAD_FILE_SIZE) {
            throw new InvalidDataException('Max upload file size', []);
        }
        return file_get_contents($params['file']->getFilePath());
    }

    public function checkItem(string $path): void {
        if (!empty($this->_storage()->getItem($path))) {
            throw new InvalidDataException('File with this name is already exist', []);
        }
    }

    /**
     * Convert files.
     *
     * @param string $input  Input file.
     * @param string $output Output File.
     *
     * @return void
     */
    private function _convertFiles(string $input, string $output): void {
        $wavFilePath = $this->_getWavFile($input, $output);
        $tempFiles   = [];
        if ($this->_isRingTones($output)) {
            $this->_convertRingTones($wavFilePath, $output);
        } else {
            $this->_convertSounds($wavFilePath, $output);
            $tempFiles[] = $output . '.wav';
        }
        $tempFiles[] = $output . '.mp3';
        foreach ($tempFiles as $tempFile) {
            if (is_file($tempFile)) {
                unlink($tempFile);
            }
        }
    }

    /**
     * Convert Sounds
     *
     * @param string $input  Input.
     * @param string $output Output.
     *
     * @return void.
     */
    private function _convertSounds(string $input, string $output): void {
        exec(sprintf('/usr/bin/sox %s -r8000 -c1 -e a-law -t al %s.alaw', $input, $output));
        exec(sprintf('opusenc --cvbr --bitrate 64 --quiet %s %s.opus', $input, $output));
        exec(sprintf('callweaver -rx \'file convert %s.alaw %s.g729\'', $output, $output));
    }

    /**
     * Convert RingTones
     *
     * @param string $input  Input.
     * @param string $output Output.
     *
     * @return void.
     */
    private function _convertRingTones(string $input, string $output): void {
        exec(sprintf('/usr/bin/sox %s -n trim 0 25', $input));
        exec(sprintf('/usr/bin/sox %s -e mu-law -t wav -r8000 -c1 %s.ulaw', $input, $output));
        exec(sprintf('opusenc --cvbr --bitrate 64 --quiet %s.wav %s.opus', $output, $output));
    }

    /**
     * Is RingTones.
     *
     * @param string $path Path.
     *
     * @return boolean
     */
    private function _isRingTones(string $path): bool {
        return strpos($path, 'RingTones/') !== false;
    }

    /**
     * Get wav file
     *
     * @param string $input  Input.
     * @param string $output Output.
     *
     * @return string
     */
    private function _getWavFile(string $input, string $output): string {
        preg_match('/^(.+)\.((wav|mp3|alaw|opus)+)$/i', $input, $matches);
        $path      = $output . '.wav';
        $extension = strtolower($matches[2]);
        switch ($extension) {
            case 'alaw':
                exec(sprintf('/usr/bin/sox -t al %s -t wav -e signed-integer -b16 -r48000 -c1 %s', $input, $path));
                break;
            case 'opus':
                $tempFile = '/tmp/session/tempfile.wav';
                exec(sprintf('opusdec %s %s', $input, $tempFile));
                exec(sprintf('/usr/bin/sox -t wav %s -t wav -e signed-integer -b16 -r48000 -c1 %s', $tempFile, $path));
                unlink($tempFile);
                break;
            default:
                $tempFile = '/tmp/session/tempfile.wav';
                exec(sprintf('/usr/bin/sox %s -t wav -e signed-integer -b16 -r48000 -c1 %s', $input, $tempFile));
                unlink($input);
                rename($tempFile, $path);
                break;
        }
        return $path;
    }

    /**
     * Check directory.
     *
     * @param string $path Path.
     *
     * @return void
     * @throws InvalidDataException Exception.
     */
    private function _checkDirectory(string $path): void {
        if (!is_dir($this->realPath($path)) || !is_readable($this->realPath($path)) || !is_writable($this->realPath($path))) {
            throw new InvalidDataException(Error::getMessage(Error::DirectoryNotFound), []);
        }
    }

    private function _checkEmptyParams($params) {
        if (empty($params['text']) && empty($params['file'])) {
            throw new InvalidDataException(Error::getMessage(Error::FieldCannotBeEmpty), []);
        }
    }

    /**
     * Is default item
     *
     * @param string $path Path.
     *
     * @return boolean
     */
    private function _isDefaultItem(string $path): bool {
        $path = str_replace(self::REAL_PATH, '', $path);
        if (!$this->_defaultSounds) {
            $content              = explode("\n", Helpers::fileGetContent(self::EXCLUDE_CUSTOM_SOUNDS));
            $content[]            = 'MusicOnHold';
            $content[]            = 'RingTones';
            $this->_defaultSounds = array_flip($content);
        }
        return isset($this->_defaultSounds[$path]);
    }

    /**
     * Validate file
     *
     * @param string $name Name.
     * @param string $dir  Directory.
     *
     * @return void
     * @throws InvalidDataException Exception
     */
    private function _validateFileCreate(string $name, string $dir): void {
        preg_match('/^(.+)\.((wav|mp3|alaw|opus)+)$/i', $name, $matches);
        if (!isset($matches[1])) {
            throw new InvalidDataException('Wrong file extension.', []);
        }
        $this->checkItem($dir . $name);
    }

    /**
     * Validate file update
     *
     * @param string $path    Path.
     * @param string $newPath New path.
     *
     * @return void
     * @throws InvalidDataException InvalidDataException.
     */
    private function _validateFileUpdate(string $path, string $newPath): void {
        if ($this->_isDefaultItem($path)) {
            throw new InvalidDataException(Error::getMessage(Error::AccessError), []);
        }
        $this->checkItem($newPath);
    }

    /**
     * Get Moh Classes.
     *
     * @return array
     */
    public static function getMohClasses(): array {
        return ShellHelper::getMohClasses();
    }

    /**
     * Get RingTones.
     *
     * @return array
     */
    public static function getRingTones(): array {
        $result = [];
        $dirs   = Helpers::scanDirRecursive(self::RING_TONES_PATH);

        foreach ($dirs as $dir => $files) {
            array_map(
                function ($file) use ($dir, &$result) {
                    $fullPath             = $dir . $file;
                    $info                 = pathinfo($fullPath);
                    $shortPath            = str_replace(self::RING_TONES_PATH, '', $dir . $info['filename']);
                    $result[$shortPath][] = $info['basename'];
                },
                $files
            );
        }
        return $result;
    }

    public function createDirectory(string $data): void {
        $path   = '';
        $params = explode('/', $data);
        foreach ($params as $param) {
            $path .= '/' . Helpers::fileNamePrepare($param);
        }
        $path = $this->realPath($path);
        if (is_dir($path) || !mkdir($path, 0777, true)) {
            throw new InvalidDataException('Cannot create directory', []);
        }
    }

    public function editDirectory(string $path, array $params): void {
        $realPath   = $this->realPath($path);
        $pathParams = explode('/', $path);
        $name       = end($pathParams);
        array_pop($pathParams);
        if (!empty($params['name'])) {
            $name = Helpers::fileNamePrepare($params['name']);
        }
        if (!empty($params['move'])) {
            $pathParams = explode('/', $params['move']);
        }
        $pathParams[] = $name;
        $newPath      = $this->realPath(implode('/', $pathParams));
        if (!is_dir($realPath) || is_dir($newPath) || !rename($realPath, $newPath)) {
            throw new InvalidDataException('Cannot update directory', []);
        }
    }

    public function removeDirectory(string $path): void {
        $this->_checkDirectory($path);
        if (!$this->_isDefaultItem($this->realPath($path))) {
            exec('rm -f -R ' . $this->realPath($path));
        } else {
            throw new InvalidDataException(Error::getMessage(Error::DirectoryNotFound), []);
        }
    }

    public function realPath(string $path): string {
        $params = explode('/', preg_replace('/(\/)|(\.){2,}/', '$1', $path));
        $params = array_filter(
            $params,
            function ($item) {
                return !empty($item) && $item !== '.';
            }
        );
        return self::REAL_PATH . implode('/', $params);
    }
}
