<?php
declare(strict_types=1);
namespace Wildix\Sounds;

use Services\Helpers;
use Wildix\InvalidDataException;
use Wildix\Error;

/**
 * Class Storage
 */
class Storage {

    private $_context;

    public function __construct(string $context) {
        $this->_context = $context;
    }

    /**
     * Get items.
     *
     * @param string $path Path.
     *
     * @return array
     */
    public function getItems(string $path): array {
        $items     = [];
        $result    = [
            'files' => [],
            'dirs'  => []
        ];
        $directory = $this->_context . $path;
        if (!empty($path)) {
            $pathParams = explode('/', $path);
            array_pop($pathParams);
            $items[] = [
                'name' => '..',
                'path' => implode('/', $pathParams) ?: '/',
                'size' => 0,
                'modified' => 0,
                'isDir' => true,
                'formats' => []
            ];
        }
        foreach (Helpers::scanDir($directory) as $item) {
            $itemPath = trim($path . '/' . $item, '/');
            $pathInfo = pathinfo($this->_context . $itemPath);
            $name     = $pathInfo['filename'];
            $ext      = $pathInfo['extension'] ?? '';

            if (!empty($ext)) {
                $result['files'][$name]['size'] = isset($result['files'][$name]['size']) ? $result['files'][$name]['size'] : 0;

                $result['files'][$name]['name']      = $name;
                $result['files'][$name]['path']      = $itemPath;
                $result['files'][$name]['isDir']     = false;
                $result['files'][$name]['modified']  = filemtime($this->_context . $itemPath);
                $result['files'][$name]['size']     += filesize($this->_context . $itemPath);
                $result['files'][$name]['formats'][] = [
                    'path' => $itemPath,
                    'type' => $ext
                ];
            } else {
                $result['dirs'][$name]['name']     = $name;
                $result['dirs'][$name]['path']     = $itemPath;
                $result['dirs'][$name]['size']     = 0;
                $result['dirs'][$name]['isDir']    = true;
                $result['dirs'][$name]['formats']  = [];
                $result['dirs'][$name]['modified'] = filemtime($this->_context . $itemPath);
            }
        }
        return array_merge($items, array_values($result['dirs']), array_values($result['files']));
    }

    /**
     * Get item files.
     *
     * @param string  $path  Path.
     * @param boolean $check Check exist files.
     *
     * @return array
     * @throws InvalidDataException Exception.
     */
    public function getItem(string $path, bool $check = false): array {
        $result   = [];
        $pathInfo = pathinfo($this->_context . $path);
        $filePath = str_replace($this->_context, '', $pathInfo['dirname'] . '/' . $pathInfo['filename']);
        foreach (SoundsService::FILE_EXTENSIONS as $ext) {
            if (is_file($this->_context . $filePath . $ext)) {
                $result[] = [
                    'path' => $filePath,
                    'type' => $ext
                ];
            }
        }
        if (empty($result) && $check) {
            throw new InvalidDataException(Error::getMessage(Error::FileNotFound), []);
        }
        return $result;
    }

    /**
     * Save item to filesystem
     *
     * @param string $file        File.
     * @param string $destination File destination.
     *
     * @return void
     * @throws InvalidDataException Exception.
     */
    public function create(string $file, string $destination): void {
        if (!file_put_contents($this->_context . $destination, $file)) {
            throw new InvalidDataException(\Wildix\Error::getMessage(\Wildix\Error::AccessError), []);
        }
    }

    /**
     * Update item
     *
     * @param string $path    File path.
     * @param string $newPath New file path.
     *
     * @return void
     * @throws InvalidDataException Exception.
     */
    public function update(string $path, string $newPath): void {
        $files = $this->getItem($path, true);
        foreach ($files as $file) {
            $newFile = $this->_context . $newPath . $file['type'];
            if (!rename($this->_context . $file['path'] . $file['type'], $newFile)) {
                throw new InvalidDataException(Error::getMessage(Error::AccessError), []);
            }
        }
    }

    /**
     * Remove item.
     *
     * @param string $path Path to item.
     *
     * @return void.
     * @throws InvalidDataException InvalidDataException.
     */
    public function remove(string $path): void {
        $files = $this->getItem($path, true);
        foreach ($files as $file) {
            if (!unlink($this->_context . $file['path'] . $file['type'])) {
                throw new InvalidDataException(\Wildix\Error::getMessage(\Wildix\Error::AccessError, []));
            }
        }
    }
}
