<?php

namespace Wildix\System;

use Exception;
use Slim\Psr7\UploadedFile;
use Wildix\Databases\WMSDB;

/**
 * Active Directory Single Sign On Class.
 */
class ADSingleSignOnService {
    const CONFIG_FILE = '/rw2/etc/pbx/adSSO_connect.conf';
    const AD_SSO      = 'ADSSO';
    const KEYTAB_PATH = '/rw2/etc/pbx/';
    // 10 Mb max upload size
    const MAX_UPLOAD_FILESIZE = 10485760;

    private $_config        = null;
    private $_wmsDbInstance = null;
    private $_defaultParams = array(
        'enabled' => 'false',
        'keyTab' => '',
        'fqdn' => '',
    );

    public function __construct(\Wildix\Settings\ConfigFile $config) {
        $this->_config = $config;
        $this->_logger = \Logger\LoggerFactory::getInstance();
    }

    /**
     * Make full path to KeyTab file
     *
     * @param string $fileName File name.
     *
     * @return string Return full path to file.
     */
    private function makeKeyTabPath($fileName) {
        return self::KEYTAB_PATH . basename($fileName);
    }

    /**
     * Get WMSDB instance
     *
     * @return null|WMSDB
     */
    private function getWmsDbInstance() {
        if ($this->_wmsDbInstance === null) {
            $this->_wmsDbInstance = WMSDB::getInstance();
        }
        return $this->_wmsDbInstance;
    }

    /**
     * Delete not more used keytab file.
     *
     * @param string $newFilePath Full path to keyTab file on disk.
     *
     * @return void
     */
    private function cleanUpOldFile($newFilePath) {
        $filePath = $this->getKeyTabPath();

        if (is_file($filePath) && $filePath !== $newFilePath) {
            unlink($filePath);
        }
    }

    private function uploadKeyTab(UploadedFile $file) {
        if (empty($file->getClientFilename()) || empty($file->getFilePath())) {
            $this->_logger->error(__LINE__, 'KeyTab file cannot be uploaded');
            return false;
        }
        // remove not allowed characters
        $name = preg_replace('/[^a-zA-Z0-9\._\-@]/', '', $file->getClientFilename());

        $keyTabPath = $this->makeKeyTabPath($name);
        $tmpName    = $file->getFilePath();

        if (filesize($tmpName) > self::MAX_UPLOAD_FILESIZE) {
            $this->_logger->error(__LINE__, 'KeyTab file size much than ' . self::MAX_UPLOAD_FILESIZE / 1024 / 1024 . 'MB');
            return false;
        }

        try {
            $file->moveTo($keyTabPath);
        } catch (Exception $e) {
            $this->_logger->error(__LINE__, 'KeyTab file cannot be uploaded');
            return false;
        }

        // delete old file
        $this->cleanUpOldFile($keyTabPath);

        return $keyTabPath;
    }

    /**
     * Chack if integration is enabled.
     *
     * @return boolean
     */
    public function isEnabled() {
        if ($this->_config->get('enabled') === 'true') {
            if (file_exists($this->getKeyTabPath())) {
                return true;
            }
        }
        return false;
    }

    /**
     * Get full path to keyTab file.
     *
     * @return string|null
     */
    public function getKeyTabPath() {
        return $this->_config->get('keyTab');
    }

    /**
     * Save configuration
     *
     * @param array $data Data for saving.
     *
     * @return array|false Array of settings in success false otherwise
     */
    public function setSettings(array $data) {
        $currentConfig = $this->getSettings();
        $toSave        = array_merge($currentConfig, $data);

        if ($toSave['uploadData'] instanceof UploadedFile) {
            $toSave['keyTab'] = $this->uploadKeyTab($toSave['uploadData']);
        }

        if ($toSave['enabled'] !== 'false' && !$toSave['keyTab']) {
            return false;
        }

        // remove no more needed data
        unset($toSave['uploadData']);

        // save to DB
        $this->getWmsDbInstance()->setSystemVariable(self::AD_SSO, json_encode($toSave));

        // save to file
        foreach ($toSave as $key => $val) {
            $this->_config->set($key, $val);
        }
        $this->_config->save();

        // check if the FQDN has been changed. Need update DNS
        if ($currentConfig['fqdn'] !== $toSave['fqdn']) {
            exec('sudo /etc/network/if-up.d/hosts move');
        }

        return $this->getSettings();
    }

    /**
     * Get configuration from config file.
     *
     * @return array
     */
    public function getSettings() {
        $dataFromFile = $this->_config->get();
        return array_merge($this->_defaultParams, $dataFromFile);
    }

    public function isConfigSynchronized() {
        $fromDB = $this->getWmsDbInstance()->getSystemVariable(self::AD_SSO);
        if (empty($fromDB)) {
            return false;
        }
        $dataFromDB     = json_decode($fromDB, true);
        $dataFromConfig = $this->getSettings();
        if (array_diff($dataFromDB, $dataFromConfig) === array_diff($dataFromConfig, $dataFromDB)) {
            return true;
        }
        return false;
    }

    /**
     * Delete keyTab file from disk
     *
     * @return array Result configuration.
     */
    public function deleteKeyTab() {
        // delete file from disk
        $this->cleanUpOldFile('');

        $data = array(
            'enabled' => 'false',
            'keyTab' => '',
            'uploadData' => ''
        );

        $result = $this->setSettings($data);
        return $result;
    }
}
