<?php

namespace Wildix\System\Dhcp;

/**
 * Class Converter
 */
class Converter {

    /**
     * Convert saved format to config DHCP
     *
     * @param array $input Saved data.
     *
     * @return array
     */
    public function convertFileToConfig(array $input) {
        $result = array(
            'router' => isset($input['options']['router'][0]) ? $input['options']['router'][0] : '',
            'dnsServerPrimary' => isset($input['options']['dns-server'][0]) ? $input['options']['dns-server'][0] : '',
            'dnsServerSecondary' => isset($input['options']['dns-server'][1]) ? $input['options']['dns-server'][1] : '',
            'ntpServerPrimary' => isset($input['options']['ntp-server'][0]) ? $input['options']['ntp-server'][0] : '',
            'ntpServerSecondary' => isset($input['options']['ntp-server'][1]) ? $input['options']['ntp-server'][1] : '',
            'netBiosNsPrimary' => isset($input['options']['netbios-ns'][0]) ? $input['options']['netbios-ns'][0] : '',
            'netBiosNsSecondary' => isset($input['options']['netbios-ns'][1]) ? $input['options']['netbios-ns'][1] : '',
            'domainName' => $input['options']['domain-name'],
            'dhcpLeaseMax' => (int)$input['general']['dhcp-lease-max'],
            'deny' => (isset($input['general']['dhcp-ignore']) && $input['general']['dhcp-ignore'] === '#known') ? true : false,
        );

        list (
            $result['start'],
            $result['end'],
            $result['networkMask'],
            $result['defaultLeaseTime']
            ) = $this->_parseDhcpRangeFromString($input['general']['dhcp-range']);

        return $result;
    }

    /**
     * Parse DHCP range from config row
     *
     * @param string $range String dhcp-range from config.
     *
     * @return array
     */
    private function _parseDhcpRangeFromString(string $range) {
        $start            = '';
        $end              = '';
        $networkMask      = '';
        $defaultLeaseTime = '216000';

        if (trim($range) !== '') {
            list ($start, $end, $networkMask, $defaultLeaseTime,) = explode(',', $range);

            $r = substr($defaultLeaseTime, -1);
            // convert to minutes
            switch ($r) {
                case 'h':
                    $defaultLeaseTime = rtrim($defaultLeaseTime, 'h') * 60;
                    break;
                case 'm':
                    $defaultLeaseTime = rtrim($defaultLeaseTime, 'm');
                    break;
                default:
                    $defaultLeaseTime = $defaultLeaseTime / 60;
                    break;
            }
        }

        return array($start, $end, $networkMask, (int)$defaultLeaseTime);
    }


    /**
     * Prepare config from to save.
     *
     * @param array $values           Input config data.
     * @param array $activeInterfaces Active interfaces.
     *
     * @return array
     */
    public function convertConfigToFile(array $values, array $activeInterfaces) {
        $interfaces = array_keys($activeInterfaces);

        $paramNoDhcp = [];
        foreach ($interfaces as $interface) {
            if ($interface !== $values['currentInterface']) {
                $paramNoDhcp[] = $interface;
            }
        }

        if (!$values['dnsServerPrimary'] && $values['dnsServerSecondary']) {
            $values['dnsServerPrimary']   = $values['dnsServerSecondary'];
            $values['dnsServerSecondary'] = '';
        }

        if (!$values['ntpServerPrimary'] && $values['ntpServerSecondary']) {
            $values['ntpServerPrimary']   = $values['ntpServerSecondary'];
            $values['ntpServerSecondary'] = '';
        }

        if (!$values['netBiosNsPrimary'] && $values['netBiosNsSecondary']) {
            $values['netBiosNsPrimary']   = $values['netBiosNsSecondary'];
            $values['netBiosNsSecondary'] = '';
        }

        $config = array(
            'general' => array(
                'dhcp-range'        => $values['start'] . ',' . $values['end'] . ',' . $values['networkMask'] . ',' . $values['defaultLeaseTime'] . 'm',
                'dhcp-lease-max'    => $values['dhcpLeaseMax'],
                'interface'         => implode(',', $interfaces),
                'no-dhcp-interface' => implode(',', $paramNoDhcp),
            ),
            'options' => array(
                'domain-name' => $values['domainName'],
                'netbios-ns'  => array($values['netBiosNsPrimary'], $values['netBiosNsSecondary']),
                'dns-server'  => array($values['dnsServerPrimary'], $values['dnsServerSecondary']),
                'router'      => ($values['router'] === '') ? array($values['interfaceIP']) : array($values['router']),
                'ntp-server'  => array($values['ntpServerPrimary'], $values['ntpServerSecondary']),
            ),
        );

        // Ignore any clients which are specified in dhcp-host lines
        // or /etc/ethers. Equivalent to ISC "deny unkown-clients".
        // This relies on the special "known" tag which is set when
        // a host is matched.
        if ($values['deny'] === \Wildix\System\Dhcp\Server::ON) {
            $config['general']['dhcp-ignore'] = '#known';
        }

        return $config;
    }
}
