<?php

namespace Wildix\System\Dhcp;

/**
 * Class Leases
 */
class Leases {

    const LEASES_DYNAMIC = 'dynamic';
    const LEASES_STATIC  = 'static';

    /**
     * Logger class instance.
     *
     * @var \Logger\Log
     */
    private $_logger;

    /**
     * Dynamic leases
     *
     * @var array
     */
    private $_dynamicLeases = [];

    /**
     * Static leases
     *
     * @var array
     */
    private $_staticLeases = [];

    /**
     * Changed leases config files
     *
     * @var boolean
     */
    private $_leasesChanged = false;

    /**
     * Leases constructor.
     *
     * @param \Logger\Log $logger Logger class.
     */
    public function __construct(\Logger\Log $logger) {
        $this->_logger = $logger;

        $this->_fixStaticLeasesEthernetAddresses();

        $this->_dynamicLeases = $this->_getCachedLeases(self::LEASES_DYNAMIC);
        $this->_staticLeases  = $this->_getCachedLeases(self::LEASES_STATIC);
    }

    /**
     * Fix static leases ethernet addresses.
     *
     * @return void
     */
    private function _fixStaticLeasesEthernetAddresses() {
        $currentStaticLeases = $this->_fileToArray(\Wildix\System\Dhcp\ServerManager::CONFIG_STATIC_LEASES);
        if (!is_array($currentStaticLeases) || !count($currentStaticLeases)) {
            return;
        }

        $fixedArray = [];
        foreach ($currentStaticLeases as $line) {
            $line = trim($line);
            if ($line === '') {
                continue;
            }

            $arr         = explode(' ', $line);
            $mac         = isset($arr[0]) ? $arr[0] : '';
            $address     = isset($arr[1]) ? $arr[1] : '';
            $nextAddress = isset($arr[2]) ? $arr[2] : '';

            if (!\Wildix\System\Network::isIp($address)) {
                $currentMac     = $mac;
                $currentAddress = substr($address, 0, strlen($address) - 17);

                $nextMac = substr($address, -17);

                $fixedArray[] = $currentMac . ' ' . $currentAddress;
                $fixedArray[] = $nextMac . ' ' . $nextAddress;
            } else {
                $fixedArray[] = $line;
            }
        }

        if (count($fixedArray) !== count($currentStaticLeases)) {
            $this->_writeLeasesConfig(\Wildix\System\Dhcp\ServerManager::CONFIG_STATIC_LEASES, implode("\n", $fixedArray) . "\n");
            $this->_logger->info(__LINE__, 'Fixed static leases');
        }
    }

    /**
     * Read file to array
     *
     * @param string $file Filename.
     *
     * @return array|null
     */
    private function _fileToArray(string $file) {
        if (!is_readable($file)) {
            $this->_logger->warning(__LINE__, sprintf('File not readable %s', $file));
            return null;
        }

        return file($file);
    }

    /**
     * Get cached leases
     *
     * @param string $type Leases type.
     *
     * @return array
     */
    private function _getCachedLeases(string $type) {
        if ($type == self::LEASES_STATIC) {
            $file = \Wildix\System\Dhcp\ServerManager::CONFIG_STATIC_LEASES;
        } else {
            $file = \Wildix\System\Dhcp\ServerManager::CONFIG_DYNAMIC_LEASES;
        }

        $data = $this->_fileToArray($file);
        if ($data === null) {
            return [];
        }

        $result = [];
        foreach ($data as $line) {
            $line = trim($line);
            if ($line === '') {
                continue;
            }

            $lease = ($type === self::LEASES_STATIC) ? $this->_parseStaticLease($line) : $this->_parseDynamicLease($line);
            if ($lease) {
                $result[] = $lease;
            }
        }

        return $result;
    }

    /**
     * Parse static lease config row.
     *
     * @param string $line Config row.
     *
     * @return array|boolean
     */
    private function _parseStaticLease(string $line) {
        list ($mac, $address) = explode(' ', $line);

        if (!\Wildix\System\Network::isIp($address)) {
            return false;
        }

        $device = \Wildix\Devices\Vendors::getDeviceByMac(trim($mac));

        return [
            'ip'     => trim($address),
            'mac'    => trim($mac),
            'vendor' => $device['vendor'],
            'device' => $device['type'],
        ];
    }

    /**
     * Parse dynamic lease config row.
     *
     * @param string $line Config row.
     *
     * @return array|boolean
     */
    private function _parseDynamicLease(string $line) {
        list ($expTime, $mac, $address, $hostname) = explode(' ', $line);

        if (!\Wildix\System\Network::isIp($address)) {
            return false;
        }

        $device = \Wildix\Devices\Vendors::getDeviceByMac(trim($mac));

        return [
            'ip'        => trim($address),
            'mac'       => trim($mac),
            'vendor'    => $device['vendor'],
            'device'    => $device['type'],
            'hostname'  => trim($hostname),
            'leasetime' => date('G:i:s d/m/y', trim($expTime)),
        ];
    }

    /**
     * Search leases by type
     *
     * @param string $type   Lease type.
     * @param string $search Search param.
     *
     * @return array
     */
    private function _searchLeasesByType(string $type, string $search) {
        $leases = ($type === self::LEASES_STATIC) ? $this->_staticLeases : $this->_dynamicLeases;
        if ($search === null) {
            return $leases;
        }

        $result = [];
        foreach ($leases as $lease) {
            if (array_search($search, $lease, true) !== false) {
                $result[] = $lease;
            }
        }

        return $result;
    }

    /**
     * Check leases in text
     *
     * @param string $contents Text from config file.
     * @param string $ip       IP address.
     *
     * @return boolean|array
     */
    private function _checkLease(string $contents, string $ip) {
        $pattern = '/.*(' . $ip . '.*)\n/';
        if (preg_match($pattern, $contents, $matches)) {
            return $matches;
        }

        $pattern = '/.*(' . $ip . '.*)/';
        if (preg_match($pattern, $contents, $matches)) {
            return $matches;
        }

        return false;
    }

    /**
     * Write leases config to file
     *
     * @param string $file     Filename.
     * @param string $contents Content.
     *
     * @return boolean
     */
    private function _writeLeasesConfig(string $file, string $contents) {
        if (!is_writable($file)) {
            $this->_logger->error(__LINE__, sprintf('Can not write to config file %s', $file));
            return false;
        }

        try {
            $fp = fopen($file, 'w');
            fputs($fp, $contents);
            fclose($fp);
        } catch (\Exception $e) {
            $this->_logger->error(__LINE__, $e->getMessage());
            return false;
        }

        return true;
    }

    /**
     * Remove lease
     *
     * @param string $mac  Mac address.
     * @param string $type Lease type.
     *
     * @return boolean
     *
     * @throws \Wildix\Exception Remove exception.
     */
    private function _removeLease(string $mac, string $type) {
        $leases = $this->_searchLeasesByType($type, $mac);
        if (!count($leases)) {
            return false;
        }

        if ($type === self::LEASES_STATIC) {
            $file = \Wildix\System\Dhcp\ServerManager::CONFIG_STATIC_LEASES;
        } else {
            $file = \Wildix\System\Dhcp\ServerManager::CONFIG_DYNAMIC_LEASES;
        }

        if (!is_writable($file)) {
            throw new \Wildix\Exception(\Wildix\Error::IsNotWriteable);
        }

        $contents = file_get_contents($file);

        $matches = $this->_checkLease($contents, $leases[0]['ip']);
        if (!$matches) {
            return false;
        }

        $toFile = str_replace(trim($matches[0]), '', $contents);
        if (strlen($contents) !== strlen($toFile)) {
            $contents = explode("\n", $toFile);
            $contents = arem_empty($contents);

            $this->_writeLeasesConfig($file, implode("\n", $contents));

            $this->_logger->info(__LINE__, sprintf('Remove %s lease %s', $type, $mac));

            if ($type === self::LEASES_STATIC) {
                $this->_staticLeases = $this->_getCachedLeases($type);
            } else {
                $this->_dynamicLeases = $this->_getCachedLeases($type);
            }

            $this->_leasesChanged = true;
        }

        return true;
    }

    /**
     * Set static lease
     *
     * @param string $ip  IP address.
     * @param string $mac Mac address.
     *
     * @return boolean
     */
    private function _setLeaseStatic(string $ip, string $mac) {
        if (!\Wildix\System\Network::isIp($ip)
            || !\Wildix\System\Network::isMac(\Wildix\System\Network::delim2mac($mac))
        ) {
            return false;
        }

        $lease = $this->_searchLeasesByType(self::LEASES_STATIC, $ip);
        if (count($lease)) {
            $this->_logger->error(
                __LINE__,
                sprintf(
                    'Lease with such ip = %s now is busy. Owner is hostname = %s with mac = %s',
                    $ip,
                    isset($lease[0]['hostname']) ? $lease[0]['hostname'] : '',
                    isset($lease[0]['mac']) ? $lease[0]['mac'] : ''
                )
            );
            return false;
        }

        $lease = $this->_searchLeasesByType(self::LEASES_STATIC, $mac);
        if (count($lease)
            && substr_count(\Wildix\System\Network::delim2mac($mac), '0') !== 12
        ) {
            $this->_logger->error(
                __LINE__,
                sprintf(
                    'Lease with such mac = %s now is busy. Owner is hostname = %s with ip = %s',
                    $lease[0]['mac'],
                    $lease[0]['hostname'],
                    $ip
                )
            );

            return false;
        }

        $mac = \Wildix\System\Network::mac2delim($mac);

        $contents   = $this->_fileToArray(\Wildix\System\Dhcp\ServerManager::CONFIG_STATIC_LEASES);
        $contents[] = $mac . ' ' . $ip . "\n";

        $this->_writeLeasesConfig(\Wildix\System\Dhcp\ServerManager::CONFIG_STATIC_LEASES, implode('', $contents));

        $this->_logger->info(__LINE__, sprintf('Add static lease %s', $mac));

        $this->_leasesChanged = true;

        $this->_staticLeases = $this->_getCachedLeases(self::LEASES_STATIC);

        return true;
    }



    /**
     * Set leases
     *
     * @param array  $leases   Leases macs for save status.
     * @param string $toStatus New status for leases.
     *
     * @throws \Wildix\Exception Exception save file.
     *
     * @return void
     */
    public function setLeasesStatus(array $leases, string $toStatus) {
        if (is_array($leases) && count($leases)) {
            foreach ($leases as $lease) {
                $mac = \Wildix\System\Network::normaliseMac($lease);
                if (!$mac) {
                    $this->_logger->error(__LINE__, sprintf('Not MAC format %s', $lease));
                } else {
                    $this->_changeStatus($mac, $toStatus);
                }
            }
        } else {
            $this->_logger->info(__LINE__, 'Leases not found ');
        }
    }

    /**
     * Change status for lease
     *
     * @param string $mac      Mac address.
     * @param string $toStatus New status.
     *
     * @throws \Wildix\Exception Exception save file.
     *
     * @return void
     */
    private function _changeStatus(string $mac, string $toStatus) {
        switch ($toStatus) {
            case self::LEASES_STATIC:
                if (count($this->_searchLeasesByType(self::LEASES_STATIC, $mac)) !== 0) {
                    $this->_logger->error(__LINE__, sprintf('MAC %s already in the static leases list', $mac));
                }

                $dynamicLeases = $this->_searchLeasesByType(self::LEASES_DYNAMIC, $mac);
                if (is_array($dynamicLeases) && count($dynamicLeases)) {
                    $data = array_pop($dynamicLeases);
                    $this->_setLeaseStatic($data['ip'], \Wildix\System\Network::delim2mac($data['mac']));
                } else {
                    $this->_logger->error(__LINE__, sprintf('MAC %s not found in dynamic leases list', $mac));
                }
                break;
            case self::LEASES_DYNAMIC:
                if (!count($this->_searchLeasesByType(self::LEASES_STATIC, $mac))) {
                    $this->_logger->error(__LINE__, sprintf('MAC %s not found in static leases list', $mac));
                } else {
                    $this->_removeLease($mac, self::LEASES_STATIC);
                }
                break;
            default:
                $this->_logger->error(__LINE__, sprintf('Not allowed status %s', $toStatus));
                break;
        }
    }

    /**
     * Remove leases
     *
     * @param array $macs Mac to remove.
     *
     * @throws \Wildix\Exception Exception save file.
     *
     * @return void
     */
    public function removeLeases(array $macs) {
        if (is_array($macs) && count($macs)) {
            foreach ($macs as $mac) {
                $this->_removeLease($mac, self::LEASES_DYNAMIC);
                $this->_removeLease($mac, self::LEASES_STATIC);
            }
        }
    }

    /**
     * Get leases
     *
     * @param array $serverConfig Dhcp server actual config.
     *
     * @throws \Wildix\Exception Exception read file.
     *
     * @return array
     */
    public function getLeases(array $serverConfig) {
        return array_merge($this->_getActualLeasesDynamic(), $this->_getActualLeasesStatic($serverConfig));
    }

    /**
     * Get actual dynamic leases.
     *
     * @throws \Wildix\Exception Exception read file.
     *
     * @return array
     */
    private function _getActualLeasesDynamic() {
        $result = [];
        foreach ($this->_dynamicLeases as $lease) {
            $found = false;
            foreach (array_keys($this->_staticLeases) as $key) {
                if ($lease['mac'] === $this->_staticLeases[$key]['mac']) {
                    $found = true;

                    if (ip2long($lease['ip']) !== ip2long($this->_staticLeases[$key]['ip'])) {
                        $this->_removeLease($this->_staticLeases[$key]['mac'], self::LEASES_STATIC);
                        unset($this->_staticLeases[$key]);

                        $found = false;
                    }
                    break;
                }
            }

            if ($found) {
                $lease['status']    = self::LEASES_STATIC;
                $lease['leasetime'] = '-';
            } else {
                $lease['status'] = self::LEASES_DYNAMIC;
            }

            if ($lease['hostname'] === '' || $lease['hostname'] === '*') {
                $lease['hostname'] = $lease['vendor'] === \Wildix\Devices\Vendors::UNKNOWN ? '-' : $lease['vendor'];
            }

            // for wildix devices
            $lease = $this->_renameWildixDevices($lease);

            $lease['binip'] = ip2long($lease['ip']);

            $result[] = $lease;
        }

        return $result;
    }

    /**
     * Get static leases
     *
     * @param array $serverConfig Dhcp server config.
     *
     * @throws \Wildix\Exception Exception read file.
     *
     * @return array
     */
    private function _getActualLeasesStatic(array $serverConfig) {
        $result = [];

        foreach ($this->_staticLeases as $lease) {
            $subnet = \Wildix\System\Network::generateNetwork($lease['ip'], $serverConfig['networkMask']);
            if (ip2long($subnet['network']) !== ip2long($serverConfig['ipAddress'])
                && $serverConfig['ipaddress'] !== '0.0.0.0'
            ) {
                $this->_removeLease($lease['mac'], self::LEASES_STATIC);
            } else {
                if (!array_value_exists($this->_dynamicLeases, $lease['mac'], 'mac')) {
                    $lease['status']    = self::LEASES_STATIC;
                    $lease['leasetime'] = '-';
                    $lease['hostname']  = $lease['vendor'] === \Wildix\Devices\Vendors::UNKNOWN ? '-' : $lease['vendor'];

                    $lease = $this->_renameWildixDevices($lease);

                    $lease['binip'] = ip2long($lease['ip']);

                    $result[] = $lease;
                }
            }
        }

        return $result;
    }

    /**
     * Rename hostname for Wildix devices
     *
     * @param array $lease Lease data.
     *
     * @return mixed
     */
    private function _renameWildixDevices(array $lease) {
        if ($lease['hostname'] === \Wildix\Devices\Vendors::WILDIX) {
            if (substr($lease['mac'], 0, 8) === '00:90:f8') {
                $lease['hostname'] = 'Wildix MG';
            } else if (substr($lease['mac'], 0, 8) === '00:15:65') {
                $lease['hostname'] = 'Wildix P';
            } else if (substr($lease['mac'], 0, 8) === '9c:75:14') {
                $lease['hostname'] = 'Wildix P';
            }
        }

        return $lease;
    }

    /**
     * Purge all leases
     *
     * @return boolean
     */
    public function purgeLeases() {
        return $this->_purgeFile(\Wildix\System\Dhcp\ServerManager::CONFIG_DYNAMIC_LEASES)
            && $this->_purgeFile(\Wildix\System\Dhcp\ServerManager::CONFIG_STATIC_LEASES);
    }

    /**
     * Purge file.
     *
     * @param string $name Filename.
     *
     * @return boolean
     */
    private function _purgeFile(string $name) {
        $filename = '/rw2' . $name;
        if (is_writable($filename)) {
            exec(': > ' . $filename . ' 2> /dev/null');
            return true;
        }

        return false;
    }

    /**
     * Leases changed
     *
     * @return boolean
     */
    public function isLeasesChanged() {
        return $this->_leasesChanged;
    }
}
