<?php

namespace Wildix\System\Dhcp;

/**
 * Class DHCPServer
 */
class Server {

    const ON  = 'on';
    const OFF = 'off';

    /**
     * Current interface
     *
     * @var string $_currentInterface
     */
    private $_currentInterface;

    /**
     * Active interfaces
     *
     * @var array $_activeInterfaces
     */
    private $_activeInterfaces;

    /**
     * Interfaces allowed for DHCP server.
     *
     * @var array $_allowedInterfaces
     */
    private $_allowedInterfaces;

    /**
     * DHCP default config
     *
     * @var array $_defaultConfig
     */
    private $_defaultConfig = array(
        'general' => array(
            'interface'         => 'eth0,eth1,eth2,eth3',
            'dhcp-range'        => '10.135.0.10,10.135.0.99,255.255.255.0,3600m',
            'dhcp-lease-max'    => '50',
            'no-dhcp-interface' => 'eth0,eth2,eth3',
            'dhcp-leasefile'    => '/rw2' . \Wildix\System\Dhcp\ServerManager::CONFIG_DYNAMIC_LEASES,
        ),
        'options' => array(
            'domain-name' => '',
            'netbios-ns'  => '',
            'dns-server'  => '',
            'router'      => '',
            'ntp-server'  => '',
        ),
    );

    /**
     *  DHCP config
     *
     * @var array $_config
     */
    private $_config;


    /**
     * Network class instance
     *
     * @var \Wildix\System\Network $_network
     */
    private $_network;

    /**
     * Converter.
     *
     * @var \Wildix\System\Dhcp\Converter
     */
    private $_converter;

    /**
     * A storage for caching dhcp server settings.
     *
     * @var \Wildix\Databases\WMSDB
     */
    private $_configStorage = null;

    /**
     * Logger class instance.
     *
     * @var \Logger\Log
     */
    private $_logger = null;

    /**
     * DHCP option 42, Network Time Protocol Servers
     * http://www.networksorcery.com/enp/protocol/bootp/option042.htm
     *
     * @var int $_RFCNumberNtpProtocol Ntp Protocol number.
     */
    private $_RFCNumberNtpProtocol = 42;

    /**
     * WMS_Ldap class.
     *
     * @var \WMS_Ldap $_wmsLdap
     */
    private $_wmsLdap;

    /**
     * Server constructor.
     *
     * @param \Wildix\System\Network        $network   Network class.
     * @param \Wildix\Databases\WMSDB       $storage   Storage class.
     * @param \Wildix\System\Dhcp\Converter $converter Converter class.
     * @param \WMS_Ldap                     $wmsLdap   WMS_Ldap class.
     * @param \Logger\Log                   $logger    Logger class.
     */
    public function __construct(
        \Wildix\System\Network $network,
        \Wildix\Databases\WMSDB $storage,
        \Wildix\System\Dhcp\Converter $converter,
        \WMS_Ldap $wmsLdap,
        \Logger\Log $logger
    ) {
        $this->_network       = $network;
        $this->_configStorage = $storage;
        $this->_converter     = $converter;
        $this->_wmsLdap       = $wmsLdap;
        $this->_logger        = $logger;

        $this->_initData();
    }

    /**
     * Init config and interfaces data
     *
     * @return void
     */
    private function _initData() {
        $configFromStorage = $this->_readConfigFromStorage();
        if ($configFromStorage !== null) {
            $this->_config = array_merge($this->_defaultConfig, $configFromStorage);
        } else {
            $this->_config = $this->_defaultConfig;
            $this->_logger->warning(__LINE__, 'Not found DHCP settings in DB-storage');
        }
        array_walk_recursive($this->_config, 'trim');

        $this->_activeInterfaces = $this->_network->getActiveNetworkInterfaces();
        $this->_currentInterface = (string)$this->_calculateCurrentInterface();
    }

    /**
     * Get current interface
     *
     * @return string
     */
    public function getCurrentInterface() {
        return $this->_currentInterface;
    }

    /**
     * Calculate current interface
     *
     * @return string
     */
    private function _calculateCurrentInterface() {
        $currentInterface = '';

        $bootInterfaces = $this->_network->getBootNetworkInterfaces();

        $ignoreInterfaces = $this->_calculateIgnoredInterfaces($bootInterfaces);

        $this->_allowedInterfaces = $this->_calculateAllowedInterfaces($ignoreInterfaces);

        // predefine
        if ($this->_config['general']['interface'] === '') {
            $this->_config['general']['interface'] = implode(',', $this->_allowedInterfaces);
        }

        if ($this->_config['general']['interface'] !== '' && $this->_config['general']['no-dhcp-interface'] !== '') {
            $generalInterface       = explode(',', $this->_config['general']['interface']);
            $generalNoDHCPInterface = explode(',', $this->_config['general']['no-dhcp-interface']);

            $diff = array_diff($generalInterface, $generalNoDHCPInterface);
            if (is_array($diff) && count($diff)) {
                $i = array_pop($diff);
                if (isset($this->_activeInterfaces[$i]) && array_search($i, $this->_allowedInterfaces) !== false) {
                    $currentInterface = $i;
                }
            }
        }

        if (trim($currentInterface) === '') {
            if (isset($this->_allowedInterfaces[0]) && $this->_allowedInterfaces[0] !== '') {
                $currentInterface = $this->_allowedInterfaces[0];
            } else {
                $currentInterface = 'eth1';
            }
        }

        return $currentInterface;
    }

    /**
     * Calculate ignored interfaces.
     *
     * @param array $bootInterfaces Boot interfaces.
     *
     * @return array
     */
    private function _calculateIgnoredInterfaces(array $bootInterfaces) {
        $ignoreInterfaces = [];
        foreach (array_keys($bootInterfaces) as $key) {
            if ($bootInterfaces[$key]['type'] === 'dhcp'
                || !isset($this->_activeInterfaces[$key]['link'])
                || $this->_activeInterfaces[$key]['link'] === 'unplug'
            ) {
                $ignoreInterfaces[] = $key;
            }
        }

        return $ignoreInterfaces;
    }

    /**
     * Calculate allowed interfaces.
     *
     * @param array $ignoreInterfaces Ignored interfaces.
     *
     * @return array
     */
    private function _calculateAllowedInterfaces(array $ignoreInterfaces) {
        $allowedInterfaces = [];
        foreach (array_keys($this->_activeInterfaces) as $key) {
            if (strpos($key, ':') !== false
                || !isset($this->_activeInterfaces[$key]['activate'])
                || $this->_activeInterfaces[$key]['activate'] === 'DOWN'
                || in_array($key, $ignoreInterfaces)
                || ($this->_activeInterfaces[$key]['address'] === ''
                && $this->_activeInterfaces[$key]['netmask'] === '')
            ) {
                continue;
            }
            $allowedInterfaces[] = $key;
        }

        return $allowedInterfaces;
    }

    /**
     * Read DHCP settings from database
     *
     * @return array
     */
    private function _readConfigFromStorage(): array {
        $variables = $this->_configStorage->getSystemVariable(\Wildix\System\Dhcp\ServerManager::DHCP_SETTINGS_STORAGE) ?: '{}';
        return json_decode($variables, true);
    }

    /**
     * Get allowed interfaces without tun.
     *
     * @return array
     */
    private function _getInterfacesWithoutVtun() {
        $interfaces = [];

        foreach ($this->_allowedInterfaces as $key) {
            if (strpos($key, 'wtun') !== false) {
                continue;
            }
            $interfaces[] = $key;
        }

        return $interfaces;
    }

    /**
     * Get DHCP config.
     *
     * @return array
     */
    public function getSavedConfig() {
        return $this->_convertSavedDataToConfigFormat();
    }

    /**
     * Convert saved data to config format
     *
     * @return array
     */
    private function _convertSavedDataToConfigFormat() {
        $result = $this->_converter->convertFileToConfig($this->_config);

        if (isset($this->_activeInterfaces[$this->_currentInterface])) {
            $result['interfaceIP'] = (string)$this->_activeInterfaces[$this->_currentInterface]['address'];
        } else {
            $result['interfaceIP'] = '';
        }

        $result['currentInterface'] = $this->_currentInterface;

        $subnet = $this->_network->getSubnet($result['interfaceIP'], $result['networkMask']);

        $result['last']  = $subnet['last'];
        $result['first'] = $subnet['first'];

        $netmask = isset($this->_activeInterfaces[$this->_currentInterface]) ? $this->_activeInterfaces[$this->_currentInterface]['netmask'] : '';
        $link    = isset($this->_activeInterfaces[$this->_currentInterface]) ? $this->_activeInterfaces[$this->_currentInterface]['link'] : '';

        $tmp = $this->_network->generateNetwork($result['start'], $netmask);

        $result['ipAddress']  = $tmp['network'];
        $result['interfaces'] = $this->_getInterfacesWithoutVtun();
        $result['plugged']    = $link === 'plug';

        return $result;
    }

    /**
     * Prepare config from to save.
     *
     * @param array $values Input config data.
     *
     * @return array
     */
    private function _prepareConfigToSave(array $values) {
        $this->_currentInterface = (string)$values['currentInterface'];

        return $this->_converter->convertConfigToFile($values, $this->_activeInterfaces);
    }

    /**
     * Fill config data to local variable $this->_config.
     *
     * @param array $store Config data.
     *
     * @return void
     */
    private function _fillConfigDataToLocalVariable(array $store) {
        if (!empty($store)) {
            if (!array_key_exists('dhcp-ignore', $store['general'])) {
                unset($this->_config['general']['dhcp-ignore']);
            } else {
                $this->_config['general']['dhcp-ignore'] = $store['general']['dhcp-ignore'];
            }
        }

        if (isset($store['general'])) {
            foreach ($store['general'] as $key => $value) {
                if (array_key_exists($key, $this->_config['general'])) {
                    $this->_config['general'][$key] = $value;
                }
            }
        }

        if (isset($store['options'])) {
            foreach ($store['options'] as $key => $value) {
                if (array_key_exists($key, $this->_config['options'])) {
                    $this->_config['options'][$key] = $value;
                }
            }
        }
    }

    /**
     *  Write DHCP setting data to storage
     *
     * @return void
     */
    private function _writeDhcpSettingsToStorage() {
        $this->_configStorage->setSystemVariable(\Wildix\System\Dhcp\ServerManager::DHCP_SETTINGS_STORAGE, json_encode($this->_config));
    }

    /**
     *  Save DHCP config to file
     *
     * @throws \Wildix\Exception Exception save file.
     *
     * @return void
     */
    private function _saveDhcpConfigFile() {
        $configGeneral = $this->_processConfigGeneral();
        $configOptions = $this->_processConfigOptions();
        $configDevices = $this->_processConfigDevices();

        $result = array_merge($configGeneral, $configOptions, $configDevices);

        $this->_writeConfigToFile(\Wildix\System\Dhcp\ServerManager::CONFIG_DHCP, implode("\n", $result) . "\n");
        $this->_logger->info(__LINE__, 'DHCP server enabled');
    }

    /**
     * Process general section in config
     *
     * @return array
     */
    private function _processConfigGeneral() {
        $result = [];

        $result[] = '# General';
        foreach ($this->_config['general'] as $key => $value) {
            if (in_array($key, array('read-ethers', 'enable-tftp', 'log-dhcp'))) {
                $result[] = $key;
            } else {
                $result[] = $key . '=' . $value;
            }
        }

        return $result;
    }

    /**
     * Process configs section in config
     *
     * @return array
     */
    private function _processConfigOptions() {
        $result = [];

        $result[] = '# Options';
        foreach ($this->_config['options'] as $option => $value) {
            if (!is_array($value) && trim($value) === '') {
                continue;
            }

            if (!is_array($value)) {
                $value = array($value);
            }

            $resVals = [];
            foreach ($value as $v) {
                if (trim($v) !== '') {
                    if ($option === 'ntp-server') {
                        $option = $this->_RFCNumberNtpProtocol;
                    }
                    $resVals[] = $v;
                }
            }

            if (count($resVals) === 0) {
                continue;
            }

            $result[] = 'dhcp-option=' . (is_numeric($option) ? $option : 'option:' . $option) . ',' . implode(',', $resVals);
        }

        return $result;
    }

    /**
     * Process devices section in config
     *
     * @return array
     */
    private function _processConfigDevices() {
        /*
            * set specific device options
             *'43' => "\"http://{$values['ipiface']}/getsettings.php?action=setaps&mac={mac}\""
             * 'dhcp-boot' => '"getsettings.php?action=setaps&mac={mac}"',
             *
             * AS:
             * # Snom
             * dhcp-mac=00:04:13:*:*:*,net:snom
             * dhcp-option=net:snom,66,"http://10.135.0.1"
             * dhcp-boot=net:snom,"getsettings.php?action=setaps&mac={mac}"
             *
             * # Thomson
             * dhcp-mac=00:0e:50:*:*:*,net:thomson
             * dhcp-mac=00:14:7f:*:*:*,net:thomson
             * dhcp-mac=00:18:f6:*:*:*,net:thomson
             * dhcp-option=net:thomson,43,"http://10.135.0.1/getsettings.php?action=setaps&mac={mac}"
             *
        */

        $result = [];

        $ipAddress = $this->_activeInterfaces[$this->_currentInterface]['address'];
        if ($this->_network->isIp($ipAddress)) {
            $deviceMacs = \Wildix\Devices\Vendors::getDeviceMacs();

            // Wildix
            $part     = \Wildix\Devices\Vendors::WILDIX;
            $result[] = "\n#" . $part;
            foreach ($deviceMacs['hardphones'][$part] as $mac) {
                $result[] = 'dhcp-mac=' . $part . ',' . rtrim(chunk_split($mac, 2, ':'), ':') . str_repeat(':*', 3);
            }

            // wp400 does not support https in 18.0.10.3 firmware
            $result[] = 'dhcp-option=net:' . $part . ',43,"http://' . $ipAddress . '/?action=setaps&mac="';
            $result[] = 'dhcp-option=net:' . $part . ',128,"http://' . $ipAddress . '/?action=setaps&mac="';

            // Snom
            $part     = \Wildix\Devices\Vendors::SNOM;
            $result[] = "\n#" . $part;
            foreach ($deviceMacs['hardphones'][$part] as $mac) {
                $result[] = 'dhcp-mac=' . $part . ',' . rtrim(chunk_split($mac, 2, ':'), ':') . str_repeat(':*', 3);
            }
            $result[] = 'dhcp-option=net:' . $part . ',66,"https://' . $ipAddress . ':443"';
            $result[] = 'dhcp-boot=net:' . $part . ',"?action=setaps&mac={mac}"';

            // Thomson
            $part     = \Wildix\Devices\Vendors::THOMSON;
            $result[] = "\n#" . $part;
            foreach ($deviceMacs['hardphones'][$part] as $mac) {
                $result[] = 'dhcp-mac=' . $part . ',' . rtrim(chunk_split($mac, 2, ':'), ':') . str_repeat(':*', 3);
            }
            $result[] = 'dhcp-option=net:' . $part . ',43,"https://' . $ipAddress . ':443/?action=setaps&mac"';
        }

        return $result;
    }

    /**
     * Save empty DHCP config file
     *
     * @throws \Wildix\Exception Exception save file.
     *
     * @return void
     */
    private function _saveEmptyDhcpConfigFile() {
        $this->_writeConfigToFile(\Wildix\System\Dhcp\ServerManager::CONFIG_DHCP, '');
        $this->_logger->info(__LINE__, 'DHCP server disabled');
    }

    /**
     * Save empty DHCP hosts config file
     *
     * @throws \Wildix\Exception Exception save file.
     *
     * @return void
     */
    public function saveEmptyDhcpHostsConfigFile() {
        $this->_writeConfigToFile(\Wildix\System\Dhcp\ServerManager::CONFIG_HOSTS, '');
        $this->_logger->info(__LINE__, 'Cleared DHCP server hosts config');
    }

    /**
     * Write dnsmasq config
     *
     * @param string $status Status enabled DHCP.
     *
     * @throws \Wildix\Exception Exception save file.
     *
     * @return void
     */
    private function _writeDnsmasqConfig($status) {
        $interface = $this->_network->getBootNetworkInterfaces($this->_currentInterface);
        $output[]  = 'DMASQINTERFACE=' . $this->_currentInterface;
        $output[]  = 'DMASQADDR=' . $interface['address'] . '/' . $interface['netmask'];

        $routes = [];

        if (isset($interface['gateway'])) {
            foreach ($interface['gateway'] as $route) {
                $routes[] = implode('/', $route);
            }
        }

        $output[] = 'DMASQROUTE="' . implode(' ', $routes) . '"';
        $output[] = 'DMASQENABLE=' . ($status === self::ON ? '1' : '0');

        $this->_writeConfigToFile(\Wildix\System\Dhcp\ServerManager::CONFIG_DNSMASQ, implode(PHP_EOL, $output) . PHP_EOL);
    }

    /**
     * Write config data to file
     *
     * @param string $file Filename.
     * @param string $data Saved data.
     *
     * @throws \Wildix\Exception Exeption save file.
     *
     * @return void
     */
    private function _writeConfigToFile(string $file, string $data = '') {
        if (file_put_contents($file, $data) === false) {
            $this->_logger->error(__LINE__, 'Cannot write to file: ' . $file);
            throw new \Wildix\Exception(\Wildix\Error::IsNotWriteable);
        } else {
            $this->_logger->info(__LINE__, 'Write settings to file: ' . $file);
        }
    }

    /**
     * Set default value if not exist
     *
     * @param array $values Server config array.
     *
     * @return array
     */
    private function _setDefaultValues(array $values) {
        if (!isset($values['status'])) {
            $values['status'] = self::OFF;
        }
        if (!isset($values['deny'])) {
            $values['deny'] = self::OFF;
        }

        return $values;
    }

    /**
     * Validate input data and save config
     *
     * @param array|null                    $values    Input values.
     * @param \Wildix\System\Dhcp\Validator $validator Dhcp validator.
     *
     * @throws \Wildix\Exception            Save exception.
     * @throws \Wildix\InvalidDataException Invalid data exception.
     *
     * @return void
     */
    public function checkAndSaveConfig(?array $values, \Wildix\System\Dhcp\Validator $validator) {
        if ($values === null) {
            $this->_saveEmptyDhcpConfigFile();
        } else {
            $values = $this->_setDefaultValues($values);

            $correctSubnetRange = $this->_network->getSubnet($values['ipAddress'], $values['networkMask']);
            $validator->validateDhcpSettings($values, $correctSubnetRange, $this->_allowedInterfaces);

            $config = $this->_prepareConfigToSave($values);

            $this->_fillConfigDataToLocalVariable($config);

            $this->_writeDhcpSettingsToStorage();

            if ($values['status'] === self::ON) {
                $this->_saveDhcpConfigFile();
            } else {
                $this->_saveEmptyDhcpConfigFile();
            }

            $this->_writeDnsmasqConfig($values['status']);
        }
    }

    /**
     * Save hosts config.
     *
     * @throws \Wildix\Exception Exception save file.
     *
     * @return void
     */
    public function saveHosts() {
        $entry = $this->_wmsLdap->getEntries($this->_wmsLdap::PBX, '', array( 'pbxid', 'pbxIp'));

        $content = '';
        if (is_array($entry)) {
            foreach ($entry as $onePBX) {
                $content .= sprintf('%s %s.conf.wld%s', $onePBX['pbxIp'], $onePBX['pbxid'], PHP_EOL);
            }
        }
        $this->_writeConfigToFile(\Wildix\System\Dhcp\ServerManager::CONFIG_HOSTS, $content);
    }
}
