<?php

namespace Wildix\System\Dhcp;

/**
 * Class ServerManager
 */
class ServerManager {

    const DHCP_SETTINGS_STORAGE = 'DHCP_SERVER_SETTINGS';
    const CONFIG_DHCP           = '/etc/dnsmasq.d/dhcp.conf';
    const CONFIG_DNSMASQ        = '/etc/dnsmasq.d/config';

    const CONFIG_HOSTS = '/etc/dnsmasq.d/hosts';

    const CONFIG_DYNAMIC_LEASES = '/etc/dnsmasq.leases';
    const CONFIG_STATIC_LEASES  = '/etc/ethers';

    /**
     * DHCP server
     *
     * @var \Wildix\System\Dhcp\Server
     */
    private $_server;

    /**
     * DHCP leases
     *
     * @var \Wildix\System\Dhcp\Leases
     */
    private $_leases;

    /**
     * Messages
     *
     * @var array $_messages
     */
    private $_messages = [];

    /**
     * Warning reason
     *
     * @var string $_reason
     */
    private $_reason = '';

    /**
     * Active interfaces
     *
     * @var array $_activeInterfaces
     */
    private $_activeInterfaces;

    /**
     * Network class instance
     *
     * @var \Wildix\System\Network $_network
     */
    private $_network;

    /**
     * Validator class instance
     *
     * @var \Wildix\System\Dhcp\Validator $_validator
     */
    private $_validator;

    /**
     * ServerManager constructor.
     *
     * @param \Wildix\System\Dhcp\Server    $server    Dhcp server.
     * @param \Wildix\System\Dhcp\Leases    $leases    Dhcp leases.
     * @param \Wildix\System\Network        $network   Network class.
     * @param \Wildix\System\Dhcp\Validator $validator Dhcp validator.
     * @param \Logger\Log                   $logger    Logger.
     */
    public function __construct(
        \Wildix\System\Dhcp\Server $server,
        \Wildix\System\Dhcp\Leases $leases,
        \Wildix\System\Network $network,
        \Wildix\System\Dhcp\Validator $validator,
        \Logger\Log $logger
    ) {
        $this->_server    = $server;
        $this->_leases    = $leases;
        $this->_network   = $network;
        $this->_validator = $validator;
        $this->_logger    = $logger;

        $this->_activeInterfaces = $this->_network->getActiveNetworkInterfaces();
    }

    /**
     * Restart DHCP server
     *
     * @return boolean
     */
    public static function restartServer() {
        $execStatus = exec('sudo /etc/init.d/dnsmasq restart');
        if (strpos($execStatus, 'failed') !== false) {
            $execStatus = exec('sudo /etc/init.d/dnsmasq restart');
        }

        return (strpos($execStatus, 'failed') !== false) ? false : true;
    }

    /**
     * Restart DHCP server and save result in log
     *
     * @return boolean
     */
    private function _restartServerAndSaveLog() {
        $status = self::restartServer();
        if ($status) {
            $this->_logger->info(__LINE__, 'DHCP server restarted');
        } else {
            $this->_logger->error(__LINE__, 'Error restart DHCP server');
        }

        return $status;
    }

    /**
     * Get current DHCP status.
     *
     * @return boolean
     */
    public static function getServerStatus() {
        return (int)exec("netstat -alnp 2>/dev/null | grep -e '0.0.0.0:67 ' | [ ! -z \"`awk '{ print $6 }'`\" ] && echo 1 || echo 0") === 1;
    }

    /**
     * Get actual config DHCP server, fixed if network changed
     *
     * @return array
     */
    public function getActualServerConfig() {
        $config = $this->_server->getSavedConfig();

        $currentInterface = $this->_server->getCurrentInterface();

        $networkChanged = $this->_checkNetworkChanged($currentInterface, $config['ipAddress'], $config['networkMask']);
        $ipInRange      = $this->_checkIpInDhcpRange($currentInterface, $config['start'], $config['end']);

        if ($networkChanged || $ipInRange) {
            $config = $this->_actualizeConfig($config, $currentInterface, $networkChanged);
        }

        $config = $this->_fixEmptyRouter($config);

        $config['status'] = self::getServerStatus();

        return $config;
    }

    /**
     * Fix empty router address.
     *
     * @param array $config Not fixed config.
     *
     * @return mixed
     */
    private function _fixEmptyRouter(array $config) {
        if ($config['router'] === '') {
            $config['router'] = $config['interfaceIP'];
        }

        return $config;
    }

    /**
     * Actualize config with current network settings
     *
     * @param array   $config           Saved config.
     * @param string  $currentInterface Current interface.
     * @param boolean $networkChanged   Network changed.
     *
     * @return array
     */
    private function _actualizeConfig(array $config, string $currentInterface, $networkChanged) {
        if ($networkChanged) {
            $this->purgeLeases();
            $this->_restartServerAndSaveLog();
        }

        $this->_reason = \Wildix\Error::getMessage(\Wildix\Error::SaveNewChanges);

        $subnet = $this->_network->getSubnet(
            $this->_activeInterfaces[$currentInterface]['address'],
            $this->_activeInterfaces[$currentInterface]['netmask']
        );

        $config['ipAddress'] = $subnet['subnet'];

        $config['last']  = $subnet['last'];
        $config['first'] = $subnet['first'];

        $config['start'] = '';
        $config['end']   = '';

        $config['router'] = (string)$this->_network->getDefaultGateway($currentInterface);

        $config['networkMask'] = isset($this->_activeInterfaces[$currentInterface]) ? (string)$this->_activeInterfaces[$currentInterface]['netmask'] : '';

        return $config;
    }

    /**
     * Save server config.
     *
     * @param array $settings Input settings.
     *
     * @return boolean
     *
     * @throws \Wildix\Exception            Save Exception
     * @throws \Wildix\InvalidDataException Validate Exception
     */
    public function saveConfig(array $settings) {
        $this->_server->checkAndSaveConfig($settings, $this->_validator);
        $this->_logger->info(__LINE__, 'Save DHCP config');

        return $this->_restartServerAndSaveLog();
    }

    /**
     * Save hosts config.
     *
     * @return boolean
     *
     * @throws \Wildix\Exception            Save Exception
     * @throws \Wildix\InvalidDataException Validate Exception
     */
    public function saveDhcpHostsConfig() {
        $this->_server->saveHosts();
        $this->_logger->info(__LINE__, 'Save DHCP hosts config');

        return $this->_restartServerAndSaveLog();
    }

    /**
     * Stop DHCP server
     *
     * @throws \Wildix\Exception            Save exception.
     * @throws \Wildix\InvalidDataException Invalid data exception.
     *
     * @return boolean
     */
    public function stopDhcpServer() {
        $this->_server->checkAndSaveConfig(null, $this->_validator);
        return $this->_restartServerAndSaveLog();
    }

    /**
     * Clear Dhcp hosts config file
     *
     * @throws \Wildix\Exception            Save exception.
     *
     * @return boolean
     */
    public function clearDhcpHostsConfigFile() {
        $this->_server->saveEmptyDhcpHostsConfigFile();
        return $this->_restartServerAndSaveLog();
    }

    /**
     * Get messages
     *
     * @return array
     */
    public function getMessages() {
        return [
            'reason' => $this->_reason,
            'warnings' => $this->_messages,
        ];
    }

    /**
     * Check changed network IP or mask.
     *
     * @param string $interface   Interface name.
     * @param string $ipAddress   Network IP address.
     * @param string $networkMask Network mask.
     *
     * @return boolean
     */
    private function _checkNetworkChanged($interface, $ipAddress, $networkMask) {
        $changed = false;

        $currentNetworkIp = $this->_activeInterfaces[$interface]['network'];
        $plugged          = (isset($this->_activeInterfaces[$interface]['link']) && $this->_activeInterfaces[$interface]['link'] === 'plug');

        if (ip2long($currentNetworkIp) !== ip2long($ipAddress)
            && $plugged
        ) {
            if ($currentNetworkIp !== '' && $ipAddress !== '0.0.0.0') {
                $this->_addMessage(
                    \Wildix\Error::getMessage(\Wildix\Error::IpWasChanged),
                    'net.dhcp.msg.IPWasChanged',
                    [
                        'ipAddress' => $ipAddress,
                        'currentNetworkIp' => $currentNetworkIp
                    ]
                );
            }
            $changed = true;
        }

        if (ip2long($this->_activeInterfaces[$interface]['netmask']) !== ip2long($networkMask)
            && $plugged
        ) {
            if ($this->_activeInterfaces[$interface]['netmask'] !== '' && $networkMask !== '') {
                $this->_addMessage(
                    \Wildix\Error::getMessage(\Wildix\Error::MaskWasChanged),
                    'net.dhcp.msg.NetmaskWasChanged',
                    [
                        'oldMask' => $networkMask,
                        'currentMask' => $this->_activeInterfaces[$interface]['netmask']
                    ]
                );
            }
            $changed = true;
        }

        return $changed;
    }

    /**
     * Check network IP in DHCP range
     *
     * @param string $interface Checked interface.
     * @param string $start     DHCP range start IP.
     * @param string $end       DHCP range end IP.
     *
     * @return boolean
     */
    private function _checkIpInDhcpRange($interface, $start, $end) {
        // ?? [MANTIS(1675)] Show an error if the ip of the pbx is in the dhcp range.
        if (ip2long($this->_activeInterfaces[$interface]['address']) >= ip2long($start)
            && ip2long($this->_activeInterfaces[$interface]['address']) <= ip2long($end)
            && $start !== '' && $end !== ''
        ) {
            $this->_addMessage(
                \Wildix\Error::getMessage(\Wildix\Error::IpMustNotInRange),
                'net.dhcp.msg.IPMustNotInRange',
                [
                    'currentAddress' => $this->_activeInterfaces[$interface]['address'],
                    'start' => $start,
                    'end' => $end,
                ]
            );

            return true;
        }

        return false;
    }

    /**
     * Add message to response.
     *
     * @param string     $reason Message text.
     * @param string     $code   Message code.
     * @param array|null $data   Additional data for message.
     *
     * @return void
     */
    private function _addMessage($reason, $code, $data) {
        $this->_messages[] = [
            'reason' => $reason,
            'code' => $code,
            'data' => $data
        ];
    }

    /**
     * Get leases
     *
     * @throws \Wildix\Exception Exception
     *
     * @return array
     */
    public function getActualLeases() {
        $leases = $this->_leases->getLeases($this->getActualServerConfig());
        $this->_checkNeedRestartDhcpServer();

        return $leases;
    }

    /**
     * Check changed leases configs and restart DHCP server
     *
     * @return void
     */
    private function _checkNeedRestartDhcpServer() {
        if ($this->_leases->isLeasesChanged()) {
            $this->_restartServerAndSaveLog();
        }
    }

    /**
     * Purge all leases
     *
     * @return boolean
     */
    public function purgeLeases() {
        return $this->_leases->purgeLeases();
    }

    /**
     * Remove leases
     *
     * @param array $macs Mac to remove.
     *
     * @throws \Wildix\Exception Exception save file.
     *
     * @return boolean
     */
    public function removeLeases(array $macs) {
        $this->_leases->removeLeases($macs);
        return $this->_restartServerAndSaveLog();
    }

    /**
     * Set leases
     *
     * @param array  $leases   Leases macs for save status.
     * @param string $toStatus New status for leases.
     *
     * @throws \Wildix\Exception Exception save file.
     *
     * @return boolean
     */
    public function setLeasesStatus(array $leases, string $toStatus) {
        $this->_leases->setLeasesStatus($leases, $toStatus);
        return $this->_restartServerAndSaveLog();
    }

    /**
     * Remove old temporary config file.
     *
     * @return void
     */
    public function removeOldConfigFile() {
        $file = self::CONFIG_DHCP . '_';
        if (is_file($file)) {
            exec('sudo rm ' . $file);
        }
    }
}
