<?php

namespace Wildix\System\Dhcp;

/**
 * Class DHCP config Validator
 */
class Validator extends \Wildix\Validator {

    /**
     * Validate new DHCP settings.
     *
     * @param array $params             New settings array.
     * @param array $correctSubnetRange Correct subnet range for this params.
     * @param array $allowedInterfaces  Allowed interfaces.
     *
     * @throws \Wildix\InvalidDataException If validation errors.
     *
     * @return void
     */
    public function validateDhcpSettings(array $params, array $correctSubnetRange, array $allowedInterfaces): void {
        if (!\Wildix\Validator::isTrue($params['status'])) {
            return;
        }

        // phpcs:disable Generic.CodeAnalysis.UnusedFunctionParameter.Found
        $this->rule(
            function ($field, $value, $params, $fields) {
                return in_array($value, $params[0]);
            },
            'currentInterface',
            $allowedInterfaces
        )->message(\Wildix\Error::InterfaceNotAllowed);

        $this->rule(
            function ($field, $value, $params, $fields) {
                if (ip2long($value) >= ip2long($fields['start']) && ip2long($value) <= ip2long($fields['end'])) {
                    return false;
                }
                return true;
            },
            'router'
        )->message(\Wildix\Error::IpMustNotBeInRange);

        $this->rule(
            function ($field, $value, $params, $fields) {
                if (ip2long($value) > ip2long($fields['end'])) {
                    return false;
                }
                return true;
            },
            'start'
        )->message(\Wildix\Error::DhcpInvalidRange);

        $this->rule(
            function ($field, $value, $params, $fields) {
                $correctRange = $params[0];

                if (ip2long($value) < ip2long($correctRange['first'])) {
                    return false;
                }
                return true;
            },
            'start',
            $correctSubnetRange
        )->message(\Wildix\Error::IpMustBeBigger);

        $this->rule(
            function ($field, $value, $params, $fields) {
                $correctRange = $params[0];

                if (ip2long($value) > ip2long($correctRange['last'])) {
                    return false;
                }
                return true;
            },
            'start',
            $correctSubnetRange
        )->message(\Wildix\Error::IpMustBeSmaller);

        $this->rule(
            function ($field, $value, $params, $fields) {
                $correctRange = $params[0];

                if (ip2long($value) > ip2long($correctRange['last'])) {
                    return false;
                }
                return true;
            },
            'end',
            $correctSubnetRange
        )->message(\Wildix\Error::IpMustBeSmaller);

        $this->rule(
            function ($field, $value, $params, $fields) {
                $correctRange = $params[0];

                if (ip2long($value) < ip2long($correctRange['first'])) {
                    return false;
                }
                return true;
            },
            'end',
            $correctSubnetRange
        )->message(\Wildix\Error::IpMustBeBigger);

        $this->rule(
            function ($field, $value, $params, $fields) {
                if (ip2long($fields['start']) + (int)$value - 1 > ip2long($fields['end'])) {
                    return false;
                }
                return true;
            },
            'dhcpLeaseMax',
            $correctSubnetRange
        )->message(\Wildix\Error::DhcpLeaseMaxTooBig);
        // phpcs:enable

        $this->rule(
            'ip',
            array(
                'dnsServerPrimary',
                'dnsServerSecondary',
                'netBiosNsPrimary',
                'netBiosNsSecondary',
                'ntpServerPrimary',
                'ntpServerSecondary',
                'router',
                'start',
                'end',
                'ipAddress',
                'networkMask',
            )
        );

        $this->rule(
            'domainNameSuffix',
            'domainName'
        );

        $this->rule(
            'boolean',
            array(
                'status',
                'deny',
            )
        );

        $this->rule(
            'min',
            array(
                'defaultLeaseTime',
                'dhcpLeaseMax',
            ),
            1
        );

        $this->rule(
            'required',
            array(
                'start',
                'end',
                'ipAddress',
                'networkMask',
                'defaultLeaseTime',
                'dhcpLeaseMax',
            )
        );

        $validateErrors = $this->validateData(
            $params,
            array(
                'dnsServerPrimary',
                'dnsServerSecondary',
                'netBiosNsPrimary',
                'netBiosNsSecondary',
                'ntpServerPrimary',
                'ntpServerSecondary',
                'router',
                'domainName',
                'defaultLeaseTime',
                'dhcpLeaseMax',
                'start',
                'end',
                'interfaceIP',
                'currentInterface',
                'deny',
                'status',
                'ipAddress',
                'networkMask',
            )
        );

        if ($validateErrors) {
            throw new \Wildix\InvalidDataException('Validation error', $validateErrors);
        }
    }
}
