<?php

namespace Wildix\System;

use Wildix\Error;
use Wildix\Exception;

/**
 * HttpProxy Class to work with Proxy settings
 */
class HttpProxy {

    private static $_httpConfPath  = '/rw2/etc/pbx/httpproxy.ini';
    private static $_proxyFilename = '/rw2/etc/openvpn/httpproxy';
    private static $_passFilename  = '/rw2/etc/openvpn/passfile';
    private static $_aptFilename   = '/etc/apt/apt.conf.d/10proxy';
    private static $_cwFilename    = '/rw2/etc/callweaver/licensing.conf';

    public function validateSupport($values) {
        $command = 'curl --proxy ' . $values['host'] . ':' . $values['port'] . ' --proxytunnel ';
        if ($values['auth_method'] == 'basic' || $values['auth_method'] == 'ntlm') {
            $command .= ' --proxy-user ' . escapeshellarg($values['login'] . ':' . $values['password']) . ' --proxy-' . $values['auth_method'] . ' ';
        }
        $command .= ' --connect-timeout 3 --max-time 5 -S -I vpn4.wildix.com:443';
        exec($command, $output, $execReturn);
        if (empty($output)) {
            // check return codes
            if ($execReturn == '5' || $execReturn == '7' || $execReturn == '28') {
                throw new Exception(Error::WrongHostOrPort);
            } else {
                throw new Exception(Error::UnknownError);
            }
        } else {
            // analyze first header
            if (!preg_match('/^HTTP\/1.[0-1]\s+([0-9]+)(\s+.*|)$/', trim($output[0]), $matches)) {
                throw new Exception(Error::IsNotReachable);
            } else if ($matches[1] != '200') {
                switch ($matches[1]) {
                    case '407':
                        // HTTP/1.0 407 Proxy Authentication Required: wrong auth params
                        throw new Exception(Error::Authentication);

                    case '403':
                        // HTTP/1.0 403 Forbidden: wrong port of dest server
                        throw new Exception(Error::IsNotReachable);

                    default:
                        throw new Exception(Error::UnknownError);
                }
            }
        }

        return true;
    }

    public function disable() {
        $settings           = $this->getSettings();
        $settings['enable'] = 0;

        $this->setSettings($settings);
    }

    public function setSettings($settings) {
        if (!file_exists(self::$_httpConfPath)) {
            touch(self::$_httpConfPath);
            chgrp(self::$_httpConfPath, 'wms');
            chown(self::$_httpConfPath, 'wms');
        }
        if (!is_writable(self::$_httpConfPath)) {
            throw new Exception(Error::AccessError);
        }
        write_ini_file(self::$_httpConfPath, $settings);

        $this->generateDependency($settings);
    }

    public function createDependencyFiles() {
        if (!file_exists(self::$_proxyFilename)) {
            touch(self::$_proxyFilename);
            chgrp(self::$_proxyFilename, 'wms');
            chown(self::$_proxyFilename, 'wms');
        }
        if (!file_exists(self::$_passFilename)) {
            touch(self::$_passFilename);
            chgrp(self::$_passFilename, 'wms');
            chown(self::$_passFilename, 'wms');
        }
        if (!file_exists(self::$_aptFilename)) {
            touch(self::$_aptFilename);
            chgrp(self::$_aptFilename, 'wms');
            chown(self::$_aptFilename, 'wms');
        }
        if (!file_exists(self::$_cwFilename)) {
            touch(self::$_cwFilename);
            chgrp(self::$_cwFilename, 'wms');
            chown(self::$_cwFilename, 'wms');
        }
    }

    public function generateDependency($values = false) {
        if (!$values) {
            $values = $this->getSettings();
        }

        $this->createDependencyFiles();

        if (!is_writable(self::$_proxyFilename)
            || !is_writable(self::$_passFilename)
            || !is_writable(self::$_aptFilename)
            || !is_writable(self::$_cwFilename)
        ) {
            throw new Exception(Error::AccessError);
        }

        if (!$values['enable']) {
            file_put_contents(self::$_proxyFilename, '');
            file_put_contents(self::$_passFilename, '');
            file_put_contents(self::$_aptFilename, '');
            file_put_contents(self::$_cwFilename, '');
            return;
        }

        $oldVpnHttpProxy = parse_ini_file(self::$_proxyFilename, false, INI_SCANNER_RAW);
        $newSettings     = array(
            'proxy_enable' => $values['enable'],
            'proxy_host' => $values['host'],
            'proxy_port' => $values['port'],
            'proxy_auth_method' => $values['auth_method'],
        );

        if ($oldVpnHttpProxy != $newSettings) {
            write_ini_file(self::$_proxyFilename, $newSettings);
        }

        $oldPassContent = file_get_contents(self::$_passFilename);
        $passContent    = $values['login'] . "\n" . $values['password'] . "\n";

        if ($oldPassContent != $passContent) {
            file_put_contents(self::$_passFilename, $passContent);
        }

        $this->updateAptProxySettings(self::$_aptFilename, $values);

        $proxy = array(
            'proxy_enable' => $values['enable'] ? 'yes' : 'no',
            'proxy_auth_method' => $values['auth_method'],
            'proxy_host' => $values['host'],
            'proxy_port' => $values['port'],
        );
        if ($values['auth_method'] != 'none') {
            $proxy['login']    = $values['login'];
            $proxy['password'] = $values['password'];
        }

        $oldCwContent = parse_ini_file(self::$_cwFilename, false, INI_SCANNER_RAW);
        if ($oldCwContent != $proxy) {
            write_ini_file(self::$_cwFilename, array('proxy' => $proxy), true, ' = ');
        }
    }

    private function updateAptProxySettings($aptFilename, $values) {
        if (file_exists($aptFilename)) {
            $aptContent = '';
            if ($values['enable'] && $values['host']) {
                $aptContent = $values['host'] . ':' . $values['port'];
                if ($values['login']) {
                    $aptContent = $values['login'] . ':' . $values['password'] . '@' . $aptContent;
                }
                $aptContent = 'Acquire::https::proxy "http://' . $aptContent . '/";';
            }

            $oldAptContent = file_get_contents($aptFilename);
            if ($oldAptContent != $aptContent) {
                file_put_contents($aptFilename, $aptContent);
            }
        }
    }

    public function getSettings() {
        $return = array(
            'enable' => false,
            'host' => '',
            'port' => '',
            'auth_method' => 'none',
            'login' => '',
            'password' => '',
        );

        if (!file_exists(self::$_httpConfPath)) {
            return $return;
        }
        $httpConf         = parse_ini_file(self::$_httpConfPath, false, INI_SCANNER_RAW);
        $return           = array_merge($return, $httpConf);
        $return['enable'] = \Wildix\Validator::isTrue($return['enable']);

        return $return;
    }
}
