<?php

namespace Wildix\System;

use Logger\Log;
use Models\PBX;
use Services\TProxyService;
use Wildix\Exception;
use WMS_Ldap;

/**
 * LdapRadius Options
 */
class LdapRadius {
    const LDAP_DN = 'cn=netreader,dc=wildix';

    /**
     * WMS_Ldap class
     *
     * @var WMS_Ldap
     */
    private $_wmsLdap;

    /**
     * PBX singleton
     *
     * @var PBX
     */
    private $_pbx;

    /**
     * Logger
     *
     * @var Log
     */
    private $_logger;

    /**
     * LdapRadius constructor.
     *
     * @param WMS_Ldap      $wmsLdap WMS_Ldap class.
     * @param PBX           $pbx     PBX class.
     * @param TProxyService $tProxy  TProxyService class.
     * @param Log           $logger  Logger class.
     */
    public function __construct(WMS_Ldap $wmsLdap, PBX $pbx, TProxyService $tProxy, Log $logger) {
        $this->_wmsLdap = $wmsLdap;
        $this->_pbx     = $pbx;
        $this->_tProxy  = $tProxy;
        $this->_logger  = $logger;
    }

    public function isEnabled(): bool {
        return $this->getOptions()['connection_enable'] === 'on';
    }

    /**
     * Retrieve LDAP options.
     *
     * @return array
     * @throws Exception Exception
     */
    public function getOptions() {
        $options = [
            'connection_enable' => 'off',
            'connection_pass' => ''
        ];

        $entry = $this->getEntry();
        if (is_array($entry)) {
            $options['connection_pass'] = $entry['userPassword'];
            if ($options['connection_pass'] !== '') {
                $options['connection_enable'] = 'on';
            }
        }

        return $options;
    }

    /**
     * Save LDAP options.
     *
     * @param array $options Array of LDAP options.
     *
     * @return boolean
     * @throws Exception Exception in save options
     */
    public function setOptions(array $options) {
        // Should be in master or single pbx mode
        if (isset($options['connection_enable']) && $options['connection_enable'] == 'on') {
            $this->_tProxy->addLdapProxying();
            $this->_tProxy->saveAndReload();

            // If PBX is not in slave mode
            if ($this->_pbx->isSlave() == false) {
                // Create new
                if (!is_array($this->getEntry())) {
                    $attributes = array(
                        'objectClass' => array(
                            'simpleSecurityObject',
                            'organizationalRole',
                        ),
                        'dn' => self::LDAP_DN,
                        'cn' => 'netreader',
                        'userPassword' => $options['connection_pass']
                    );

                    $res = $this->_wmsLdap->addExternalUser($attributes);
                    if (!$res) {
                        $this->_logger->error(__LINE__, 'Not added LDAP account');
                        throw new Exception('InternalServerError');
                    }
                } else {
                    $res = $this->_wmsLdap->update(self::LDAP_DN, array('userPassword' => $options['connection_pass']));
                    if (!$res) {
                        $this->_logger->error(__LINE__, 'Not updated LDAP account');
                        throw new Exception('InternalServerError');
                    }
                }

                $serverHost = '';
            } else {
                $this->_logger->error(__LINE__, 'Not saved LDAP options. PBX in Slave Mode');
                throw new Exception('InternalServerError');
            }
        } else {
            $res = $this->_wmsLdap->deleteEntryByDn(self::LDAP_DN);
            if ($res !== true) {
                $this->_logger->info(__LINE__, $res);
            }

            $serverHost = '';

            if (!$this->_pbx->isEnabledWmsNetwork()) {
                $this->_tProxy->removeLdapProxying();
                $this->_tProxy->saveAndReload();
            }
        }

        $this->updateAllowedHost($serverHost);

        return true;
    }

    /**
     * Get LDAP entry by dn
     *
     * @return mixed
     */
    private function getEntry() {
        return $this->_wmsLdap->getEntryByDn(
            self::LDAP_DN,
            array(
                'cn',
                'userPassword',
                'objectClass',
                'dn',
            ),
            array('objectClass')
        );
    }

    /**
     * Get allowed host from /etc/hosts.allow
     *
     * @return string
     * @throws Exception Exception
     */
    private function getAllowedHost() {
        $cmd = "cat /etc/hosts.allow | grep slapd | grep ALLOW | sed 's/slapd: 5.15.0.0\\/255.255.255.0 127.0.0.1//' | sed 's/: ALLOW//'";
        exec($cmd, $output, $returnValue);

        if ($returnValue !== 0 || !isset($output[0])) {
            $this->_logger->error(__LINE__, sprintf('Error get allowed host. Error code %d', $returnValue));
            throw new Exception('InternalServerError');
        }

        return trim($output[0]);
    }

    /**
     * Update allowed host in file /etc/hosts.allow
     *
     * @param string $host Host or IP.
     *
     * @return void
     * @throws Exception Exception
     */
    private function updateAllowedHost(string $host = '') {
        $cmd  = 'fname=/etc/hosts.allow; filetmp=`mktemp` && sudo /bin/cat $fname | sed \'/slapd/d;\' > ';
        $cmd .= '$filetmp && echo "slapd: 5.15.0.0/255.255.255.0 127.0.0.1 ' . $host;
        $cmd .= ' : ALLOW" >> $filetmp && echo -e "slapd: ALL : DENY\n" >> $filetmp && sudo /bin/cp $filetmp $fname && rm $filetmp';
        exec($cmd, $output, $returnValue);

        if ($returnValue !== 0) {
            $this->_logger->error(__LINE__, sprintf('Error set allowed host. Error code %d', $returnValue));
            throw new Exception('InternalServerError');
        }
    }
}
