<?php

namespace Wildix\System;

use Models\PBX;
use Wildix\System\RemoteSupport;
use Wildix\System\Storage\Storage;

/**
 * Mount FS class
 */
class Mount {

    private $_mounts = null;

    public $availableMps;
    public $sambaMounts;
    public $nfsMounts;
    public $errors = array();

    const SMB_VERSIONS = [
        '3.0' => 'min: Win 8, WinServer 2012',
        '2.1' => 'min: Win 7, WinServer 2008 R2',
        '2.0' => 'min: Vista SP1, WinServer 2008',
        '1.0' => 'min: Win 95, NT 4.0',
    ];

    const DEFAULT_SMB_VERSION = '2.0';

    public static function getMountPoints($forInit = false) {
        if ($forInit) {
            return array(
                realpath('/var/spool/callweaver/monitor')   => 1,
                realpath('/var/spool/callweaver/voicemail') => 1,
                '/mnt/cticonnect'                           => 1,
                '/mnt/faxes'                                => 1,
                '/mnt/cdr'                                  => 1,
                '/mnt/cdrcsv'                               => 1,
            );
        } else {
            return array(
                realpath('/var/spool/callweaver/monitor')   => _t('devices.CallRecording'),
                realpath('/var/spool/callweaver/voicemail') => _t('devices.Voicemail'),
                realpath('/var/spool/callweaver/outgoing')  => _t('devices.CallGeneration'),
                '/mnt/cticonnect'                           => _t('devices./mnt/cticonnect'),
                '/mnt/backups'                              => _t('devices./mnt/backups'),
                '/mnt/faxes'                                => _t('devices.FaxesMount'),
                '/mnt/cdr'                                  => _t('devices./mnt/cdr'),
                '/mnt/cdrcsv'                               => _t('devices./mnt/cdrcsv'),
            );
        }
    }

    public function getMounts() {
        if ($this->_mounts === null) {
            $this->updateMountsList();
        }
        return $this->_mounts;
    }

    /**
     * Get mounts by type
     *
     * @param mixed $type Accepts: 'cifs' (Samba), 'nfs' or false to get all.
     *
     * @return array
     */
    public function getMountsByType($type = false) {
        $mounts = array();
        $mount  = $this->getMounts();

        foreach ($mount as $parts) {
            $matches = preg_split('! !', $parts);
            if (!$type || trim($matches[2]) === $type) {
                $mounts[trim($matches[1])] = trim(str_replace('\040', ' ', $matches[0]));
            }
        }
        return $mounts;
    }

    public function getMountInfoFromConf($type, $path, $folder) {
        $assocArr = false;
        $return   = array();
        if (file_exists(Storage::$fileName)) {
            $assocArr = parse_ini_file(Storage::$fileName, true, INI_SCANNER_RAW);
        }
        if ($assocArr && $assocArr['Devices']) {
            foreach ($assocArr['Devices'] as $deviceString) {
                $deviceArr = explode('|', $deviceString);

                if ($deviceArr[0] === $type && $deviceArr[1] === $path && $deviceArr[2] === $folder) {
                    $return = array(
                        'type' => $type,
                        'servername' => $deviceArr[1],
                        'serverpath' => $deviceArr[2]
                    );

                    if (isset($deviceArr[3])) {
                        $return['domain'] = $deviceArr[3];
                    }
                    if (isset($deviceArr[4])) {
                        $return['username'] = $deviceArr[4];
                    }
                    if (isset($deviceArr[5])) {
                        $return['password'] = $deviceArr[5];
                    }
                    if (isset($deviceArr[6])) {
                        $return['smbversion'] = $deviceArr[6];
                    }
                }
            }
        }

        return $return;
    }

    public function getMountPointsForSmb() {
        $ampm = array();
        $this->updateSmbMounts();
        $mpmeanings = self::getMountPoints(false);
        foreach ($this->availableMps as $amp => $value) {
            if (!in_array($amp, array('/mnt/cdr', '/mnt/cticonnect', '/mnt/users'))) {
                $ampm[$amp] = $mpmeanings[$amp];
            }
        }

        if (isset($this->availableMps['/dev/shm/var/spool/callweaver/outgoing'])) {
            unset($this->availableMps['/dev/shm/var/spool/callweaver/outgoing']);
        }

        return $ampm;
    }

    public function getMountedServerName($parameters) {
        $serverName = gethostbyname($parameters['servername']);

        // checking for bad chars in serverpath
        if (!preg_match('/^[\/A-Za-z0-9_\. -]+$/', $parameters['serverpath'])) {
            $this->errors[] = _t('devices.err.WrongDirPath');
            return false;
        }

        if ($parameters['type'] == 'smb') {
            // checking if we can connect to the host
            if (!RemoteSupport::getInstance()->testHostPort($serverName, '139', '4') && !RemoteSupport::getInstance()->testHostPort($serverName, '445', '4')) {
                $this->errors[] = _t('devices.err.CannotConnectToHost', $parameters['servername']);
                return false;
            }
        } else if ($parameters['type'] == 'nfs') {
            // checking if we can connect to the host
            if (!RemoteSupport::getInstance()->testHostPort($serverName, '111', '4') && !RemoteSupport::getInstance()->testHostPort($serverName, '2049', '4')) {
                $this->errors[] = _t('devices.err.CannotConnectToHost', $parameters['servername']);
                return false;
            }
        }

        return $serverName;
    }


    public function mountUsb($name, $mountPoint) {
        $name = '/dev/' . $name;
        if (\Wildix\License::getInstance()->isCloud()) {
            $exec = 'sudo mount --bind /media ' . escapeshellarg($mountPoint) . ' 2>&1; sudo chmod -R a=rwx ' . escapeshellarg($mountPoint) . ' 2>&1';
        } else {
            // That trick is done to provide write permissions surely
            exec('sudo blkid -c /dev/null -o value -s TYPE ' . escapeshellarg($name), $fstype);
            $exec = '';

            // FAT 12/16/32
            if ($fstype[0] === 'vfat') {
                $wmsFile = parse_ini_file(Storage::$fileName);
                $isError = false;

                foreach ($wmsFile as $key => $value) {
                    if ($key === $mountPoint) {
                        $isError = false;
                        $exec    = 'sudo mount -t vfat -o umask=000,noatime,nodiratime ' . escapeshellarg($name) . ' ' . escapeshellarg($mountPoint) . ' 2>&1';
                        $rt      = exec($exec, $mount, $execRet);

                        if ($execRet !== 0) {
                            $this->errors[] = $rt;
                            return false;
                        }
                    } else {
                        $isError = true;
                    }
                }

                if ($isError) {
                    $this->errors[] = _t('devices.err.VfatFileSystem', $mountPoint);
                    return false;
                }
            } else if ($fstype[0] === 'ext3' || $fstype[0] === 'ext2' || $fstype[0] === 'ext4') {
                $exec = 'sudo mount -o noatime,nodiratime ' . escapeshellarg($name) . ' ' . escapeshellarg($mountPoint) . ' 2>&1; sudo chmod -R a=rwx ' . escapeshellarg($mountPoint) . ' 2>&1';
            }
        }

        $rt = exec($exec, $mount, $execRet);

        if ($execRet === 0) {
            return true;
        }

        $this->errors[] = $rt;
        return false;
    }

    public function mountShare($serverName, $mountPoint, $parameters) {
        $mountPoint = rtrim($mountPoint, '/') . '/';

        if ($parameters['type'] == 'smb') {
            $username = $parameters['username'];
            $password = $parameters['password'];

            $domainOption = trim($parameters['domain']) ? 'domain=' . escapeshellarg(trim($parameters['domain'])) . ',' : '';
            $exec         = 'sudo mount -t cifs -o noperm,noserverino,' . $domainOption;
            $exec        .= $username !== '' ? 'username=' . escapeshellarg($username) . ',password=' . escapeshellarg($password) . '' : 'guest';
            $exec        .= ',vers=' . escapeshellarg($parameters['smbversion']);
            $exec        .= ',rw,uid=wms,gid=wms,file_mode=0755,dir_mode=0755 ';
            $exec        .= escapeshellarg('//' . $serverName . '/' . $parameters['serverpath']);
            $exec        .= ' ' . escapeshellarg($mountPoint) . ' 2>&1';
        } else if ($parameters['type'] == 'nfs') {
            // sudo mount -t nfs -o rw "192.168.1.5:/Public" /media/
            $server = $serverName . ':/' . $parameters['serverpath'];
            $exec   = 'sudo mount -t nfs -o rw,vers=3,rsize=32768,wsize=32768,hard,intr,proto=tcp,timeo=600,';
            $exec  .= 'retrans=2,sec=sys,async,nolock,addr=' . escapeshellarg($serverName) . ' ' . escapeshellarg($server) . ' ' . escapeshellarg($mountPoint) . ' 2>&1';
        }

        if (!isset($exec)) {
            return false;
        }

        exec($exec, $result, $execRet);

        if ($execRet === 0) {
            return true;
        }

        $fp = fopen('/tmp/w_gui_mount_err', 'w');
        fputs($fp, $exec . "\n" . implode("\n", $result));
        fclose($fp);

        $this->errors[] = _t('devices.err.smbMountParamsError');
        return false;
    }

    public function unMount($mountPoint) {
        $rt = exec('sudo umount -lf ' . escapeshellarg($mountPoint) . ' 2>&1', $mount, $execRet);

        if ($execRet !== 0) {
            $this->errors[] = $rt;
            return false;
        }

        return true;
    }

    public function updateMountsList() {
        $this->_mounts = array();
        if (file_exists('/proc/mounts')) {
            $this->_mounts = file('/proc/mounts');
            $this->_mounts = array_map('trim', $this->_mounts);
        }
    }

    public function updateSmbMounts() {
        $systemMounts       = $this->getMountsByType();
        $this->sambaMounts  = $this->getMountsByType('cifs');
        $this->nfsMounts    = $this->getMountsByType('nfs');
        $mountPointsNoLink  = self::getMountPoints(false);
        $this->availableMps = array_diff_key($mountPointsNoLink, $systemMounts, $this->nfsMounts);
    }

    public function validateMountPoint($mountPoint, $returnValueOnAnotherDeviceIsMounted) {
        if (!is_dir($mountPoint)) {
            if (!mkdir($mountPoint, 0775)) {
                $this->errors[] = _t('devices.err.CannotCreateDirectory', $mountPoint);
                return false;
            }
        }
        $mountPoint = rtrim($mountPoint, '/');
        $realPath   = realpath($mountPoint);

        if ($realPath != false) {
            $mountPoint = rtrim($realPath, '/');
        }

        $systemMountPoints = $this->getMountsByType();

        if (array_key_exists($mountPoint, $systemMountPoints)) {
            $this->errors[] = _t('devices.err.AnotherDeviceMounted', $mountPoint);
            return $returnValueOnAnotherDeviceIsMounted;
        }

        return $mountPoint;
    }

    public function checkMountSmbOrMountNfs($mountPoint, $params) {
        $errors = array();

        if ($params['servername'] != '') {
            if (!preg_match('/^[^`;\(\)\[\]\{\}\|]+$/', $params['servername'])) {
                $errors['servername'] = _t('devices.err.CharsNotRequired');
            }
        }
        if ($params['serverpath'] == '') {
            $errors['serverpath'] = _t('devices.err.PleaseInputStorage');
        } else {
            if (!preg_match('/^[^`;\(\)\[\]\{\}\|]+$/', $params['serverpath'])) {
                $errors['serverpath'] = _t('devices.err.CharsNotRequired');
            }
        }

        $belongsToNetworkMounts = in_array('//' . $params['servername'] . '/' . $params['serverpath'], $this->sambaMounts) ||
            in_array('' . $params['servername'] . ':/' . $params['serverpath'], $this->nfsMounts);

        if ($belongsToNetworkMounts && $mountPoint == '/dev/shm/var/spool/callweaver/outgoing/') {
            $errors['serverpath'] = _t('devices.err.CallGenerationInUse');
        }

        if ($params['type'] === 'smb' && !array_key_exists($params['smbversion'], self::SMB_VERSIONS)) {
            $errors['smbversion'] = _t('devices.err.InvalidSmbVersion');
        }

        return (empty($errors)) ? true : $errors;
    }
}
