<?php

namespace Wildix\System;

/**
 * Class Network
 */
class Network extends \Singleton {

    public static function bin2ip($bin) {
        $output = '';
        preg_match('/([01]{8})([01]{8})([01]{8})([01]{8})/', $bin, $match);
        if ($match) {
            for ($i = 1; $i <= 4; $i++) {
                $output .= base_convert($match[$i], 2, 10) . (($i < 4) ? '.' : '');
            }
        }
        return $output;
    }

    private $_httpsOnlyConf = '/rw2/etc/pbx/https-only.conf';

    public function isHttpsOnlyEnabled() {
        // always use https for cloud PBXs
        if (\Wildix\License::getInstance()->isCloud()) {
            return true;
        }
        // check config file
        $data = null;
        if (file_exists($this->_httpsOnlyConf)) {
            $data = parse_ini_file($this->_httpsOnlyConf);
        }

        if ($data) {
            return $data['https_only'] == 'on' ? true : false;
        }
        return false;
    }

    public function forceToHttps() {
        // get custom secure port
        $port       = '';
        $securePort = \Models\PBX::getInstance()->getCustomHttpsPort();
        if (!empty($securePort) && $securePort !== 443) {
            $port = ':' . $securePort;
        }

        // check if https-only is configured. WMS-3733
        if ($this->isExternalRequest() && $this->isHttpsOnlyEnabled() && empty($_SERVER['HTTPS'])) {
            header(sprintf('location: https://%s%s%s', $_SERVER['SERVER_NAME'], $port, $_SERVER['REQUEST_URI']), true, 301);
            exit;
        }
    }

    /**
     * Get subnet information.
     *
     * @param null|string $ip   IP address.
     * @param null|string $mask Mask.
     *
     * @return array
     */
    public static function getSubnet(?string $ip, ?string $mask) {
        $aIp   = explode('.', $ip);
        $ipBit = '';
        foreach ($aIp as $value) {
            $b = base_convert(intval($value), 10, 2);
            if (strlen($b) < 8) {
                $b = str_repeat('0', 8 - strlen($b)) . $b;
            }
            $ipBit .= $b;
        }

        $aMask   = explode('.', $mask);
        $maskBit = '';
        foreach ($aMask as $value) {
            $b = base_convert(intval($value), 10, 2);
            if (strlen($b) < 8) {
                $b = str_repeat('0', 8 - strlen($b)) . $b;
            }
            $maskBit .= $b;
        }
        $maskL = strpos($maskBit, '0');

        $networkIdBit = substr($ipBit, 0, $maskL);
        $hostIdBitLen = 32 - strlen($networkIdBit);

        $bcastB = $networkIdBit . str_repeat('1', $hostIdBitLen);
        $firstB = $maskL == 31 ? $networkIdBit . '0' : $networkIdBit . str_repeat('0', $hostIdBitLen - 1) . '1';
        $lastB  = $maskL == 31 ? $networkIdBit . '1' : $networkIdBit . str_repeat('1', $hostIdBitLen - 1) . '0';

        $a['ip']     = $ip;
        $a['ip_bit'] = $ipBit;

        $a['mask']     = self::bin2ip($maskBit);
        $a['mask_bit'] = $maskBit;
        $a['mask_l']   = $maskL;

        $a['bcast']     = self::bin2ip($bcastB);
        $a['bcast_bit'] = $bcastB;

        $a['first']     = self::bin2ip($firstB);
        $a['first_bit'] = $firstB;

        $a['last']     = self::bin2ip($lastB);
        $a['last_bit'] = $lastB;

        $a['subnet']     = self::bin2ip($ipBit & $maskBit);
        $a['subnet_bit'] = ($ipBit & $maskBit);

        $a['count'] = base_convert($lastB, 2, 10) - base_convert($firstB, 2, 10);

        return $a;
    }

    /**
     * Check string is IP address.
     *
     * @param null|string $ip String IP.
     *
     * @return boolean
     */
    public static function isIp(?string $ip) {
        $long = ip2long($ip);
        return ($long != -1 && $long !== false);
    }

    /**
     * Check string is MAC.
     *
     * @param null|string $mac String MAC address.
     *
     * @return boolean
     */
    public static function isMac(?string $mac) {
        return (preg_match('/^[0-9A-Fa-f]{12}$/', $mac) === 1);
    }

    /**
     * Add delimiter for mac address
     *
     * @param null|string $mac       Mac address without delimiter.
     * @param string      $delimiter Delimiter.
     *
     * @return boolean|string
     */
    public static function mac2delim(?string $mac, string $delimiter = ':') {
        return (self::isMac($mac)) ? rtrim(chunk_split($mac, 2, $delimiter), $delimiter) : false;
    }

    /**
     * Remove delimiters from mac address
     *
     * @param null|string $mac       Mac address with delimiters.
     * @param string      $delimiter Delimiter.
     *
     * @return boolean|string
     */
    public static function delim2mac(?string $mac, string $delimiter = ':') {
        $mac = str_replace($delimiter, '', $mac);

        return (self::isMac($mac)) ? $mac : false;
    }

    /**
     * Normalise mac format
     *
     * @param null|string $mac       Mac address.
     * @param string      $delimiter Delimiter.
     *
     * @return boolean|string
     */
    public static function normaliseMac(?string $mac, string $delimiter = ':') {
        $mac = strtolower(str_replace([':', '-'], '', $mac));

        return self::mac2delim($mac, $delimiter);
    }

    /**
     * Generate network from address & netmask
     *
     * @param null|string $address IP address.
     * @param null|string $netmask IP Mask.
     *
     * @return array
     */
    public static function generateNetwork(?string $address, ?string $netmask) {
        $subnet = self::getSubnet($address, $netmask);

        return array(
            'broadcast' => $subnet['bcast'],
            'network'   => $subnet['subnet']
        );
    }

    /**
     * Get active network interfaces
     *
     * @param string|null $interface Interface name.
     *
     * @return array
     */
    public function getActiveNetworkInterfaces($interface = null) {
        // TODO: move implementation from old class.
        return \network::getInstance()->get_active_network_interfaces($interface);
    }

    /**
     * Get boot network interfaces
     *
     * @param string|null $interface Interface name.
     *
     * @return array
     */
    public function getBootNetworkInterfaces($interface = null) {
        // TODO: move implementation from old class.
        return \network::getInstance()->get_boot_network_interfaces($interface);
    }

    /**
     * Get default gateway
     *
     * @param string|null $interface Interface name.
     *
     * @return string
     */
    public function getDefaultGateway($interface = null) {
        // TODO: move implementation from old class.
        return \network::getInstance()->get_default_gateway($interface);
    }

    private function isExternalRequest(): bool {
        $localhostInterfaces = [
            '127.0.0.1',
            '::1',
        ];

        return in_array($_SERVER['REMOTE_ADDR'], $localhostInterfaces, true) === false;
    }
}
