<?php

namespace Wildix\System;

use Wildix\Error;
use Wildix\Exception;
use Wildix\InvalidDataException;
use Wildix\Settings\ConfigFile;
use Wildix\System\Ntp\NtpConfig;

/**
 * Class Ntp
 */
class Ntp {

    const INI_FILE    = '/rw2/etc/pbx/ntpserver.ini';
    const CONFIG_FILE = '/etc/ntp.conf';
    const TZ_DIR      = '/usr/share/zoneinfo/';

    const DEFAULT_NTP_CONF = [
        'restrict default noquery nomodify',
        'restrict 127.0.0.1',
        'server 127.127.1.0',
        'fudge 127.127.1.0 stratum 10',
        'driftfile /tmp/ntp.drift',
    ];

    const DEFAULT_INI_CONF = [
        'SERVER1'   => 'time.google.com',
        'SERVER2'   => 'pool.ntp.org',
        'PROVTIME'  => 'eth1',
        'SYNCEVERY' => 1,
        'SYNCCAL'   => 'hours',
    ];

    public $options;

    private $_iniFile;
    private $_configFile;

    public function __construct() {
        $this->options = self::DEFAULT_INI_CONF;
    }

    public function setIniFile(ConfigFile $file) {
        $this->_iniFile = $file;
    }

    private function _getIniFile(): ConfigFile {
        if (!$this->_iniFile) {
            $this->_iniFile = new ConfigFile(self::INI_FILE);
        }
        return $this->_iniFile;
    }

    public function setConfigFile(NtpConfig $file) {
        $this->_configFile = $file;
    }

    private function _getConfigFile(): NtpConfig {
        if (!$this->_configFile) {
            $this->_configFile = new NtpConfig(self::CONFIG_FILE);
        }
        return $this->_configFile;
    }

    public function changeGroupAndOwner($group, $owner) {
        if (!fileowner(self::INI_FILE) || !filegroup(self::INI_FILE)) {
            chgrp(self::INI_FILE, $group);
            chown(self::INI_FILE, $owner);
        }
    }

    public function setConfig($options = []) {
        $dirtyOptions  = array_merge($this->getConfig(), $options);
        $this->options = array_intersect_key($dirtyOptions, self::DEFAULT_INI_CONF);

        $this->_getIniFile()->setData($this->options);
        $res = $this->_getIniFile()->save();

        $this->updateNtpConf();

        $this->restart();

        return $res;
    }

    public function getConfig($option = null) {
        $options       = $this->_getIniFile()->get();
        $this->options = array_merge($this->options, $options);

        if (is_numeric($this->options['SYNCCAL'])) {
            $this->options['SYNCCAL'] = array_pop(
                array_slice(
                    array(
                     'hours',
                     'dayes',
                     'weeks',
                     'mounts',
                    ),
                    $this->options['SYNCCAL'],
                    1
                )
            );
        }

        if (is_numeric($this->options['PROVTIME'])) {
            $this->options['PROVTIME'] = $this->options['PROVTIME'] == 1 ? 'eth1' : '';
        }

        return $option !== null ? $this->options[$option] : $this->options;
    }

    public function getTimezone() {
        // get current system timezone
        $localtime = '/etc/localtime';
        $prfx      = '/rw2';

        if (is_link($prfx . $localtime)) {
            $localtime = $prfx . $localtime;
        }
        if (!is_link($localtime)) {
            return false;
        }
        $tmp  = str_replace(self::TZ_DIR, '', readlink($localtime));
        $area = substr($tmp, 0, strpos($tmp, '/'));
        $zone = str_replace($area . '/', '', $tmp);

        return array(
            'area' => $area,
            'zone' => $zone,
            );
    }

    public function updateNtpConf() {
        $this->options = $this->getConfig();

        $servers = [
            'server ' . $this->options['SERVER1'] . ' true',
            'server ' . $this->options['SERVER2'] . ' true',
        ];
        $this->_getConfigFile()->setData(array_merge(self::DEFAULT_NTP_CONF, $servers));
        return (bool)$this->_getConfigFile()->save();
    }

    public function restart() {
        return exec(
            'sudo /bin/systemctl restart ntp'
        );
    }

    public function stop() {
        return exec(
            '
            sudo /bin/systemctl stop ntp;
            sudo /bin/systemctl disable ntp;
            '
        );
    }

    public function start() {
        return exec(
            '
            sudo /bin/systemctl enable ntp;
            sudo /bin/systemctl start ntp;
            '
        );
    }

    public function updateSettings($values) {
        $this->_validate($values);
        $toConf = [
            'SERVER1' => $values['pntp'],
            'SERVER2' => $values['sntp']
        ];

        if (!$this->setConfig($toConf)) {
            throw new Exception(Error::AccessError);
        }

        return true;
    }

    public function getSettings() {
        $conf = $this->getConfig();

        return [
            'time' => date('d/m/Y H:i:s', time()),
            'pntp' => $conf['SERVER1'],
            'sntp' => $conf['SERVER2'],
        ];
    }

    public function hasTimezone($region, $zone) {
        $list = $this->getListTimezones();
        return isset($list[$region]) && in_array($zone, $list[$region]);
    }

    public function getListTimezones($withfilenames = false) {
        if (!is_dir(self::TZ_DIR)) {
            return false;
        }

        $skipPatterns = array(
            'posix',
            'right',
            'Mideast',
            'Mexico',
            'Chile',
            'Brazil',
            'Arctic',
            'Antarctica',
            'SystemV',
            'GMT0',
            'GMT-0',
            'GMT\+0',
        );

        // option '-follow' of the util find is deprecated and should be replaced by option '-L', but used version 4.7.0 is not support it
        exec(
            'find ' . self::TZ_DIR . ' -follow -type f -print | awk \'!/' . implode('|', $skipPatterns) . '/ {print $1}\' | sort -V',
            $files
        );
        $result = array();
        $cnt    = count($files);
        for ($i = 0; $i < $cnt; ++$i) {
            $tmp = str_replace(self::TZ_DIR, '', $files[$i]);
            $key = substr($tmp, 0, strpos($tmp, '/'));
            if (strpos($tmp, '/') === false) {
                continue;
            }
            $keyTwo = str_replace($key . '/', '', $tmp);
            if ($withfilenames) {
                $result[$key][$keyTwo] = $files[$i];
            } else {
                $result[$key][] = $keyTwo;
            }
        }

        return $result;
    }

    public function setTimezone($area, $zone) {
        $tzones   = $this->getListTimezones(true);
        $filename = $tzones[$area][$zone];
        if (!file_exists($filename)) {
            return false;
        }
        $timezone = str_replace(self::TZ_DIR, '', $filename);

        exec(
            '
            rm -f /rw2/etc/localtime && ln -sf ' . $filename . ' /rw2/etc/localtime;
            echo ' . $timezone . '> /rw2/etc/timezone;
            '
        );

        return true;
    }

    private function _validate($params): void {
        $form = new \Wildix\Validator(
            $params,
            ['pntp', 'sntp']
        );

        $form->rule('required', ['pntp', 'sntp']);
        $form->rule('DomainOrIp', ['pntp', 'sntp']);
        $form->validate();

        if (!empty($form->errors())) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $form->errors());
        }
    }
}
