<?php

namespace Wildix\System;

use Wildix\Exception;
use Wildix\Error;
use Singleton;

/**
 * Class RemoteSupport
 */
class RemoteSupport extends Singleton {

    const OPEN_VPN_CLIENT_KEY_PATH = '/etc/openvpn/client.key';

    const REMOTE_SUPPORT_CONFIG_PATH = '/rw2/etc/pbx/remote_support.config';

    const OPEN_VPN_MONIT_PATH = '/etc/monit/openvpn.conf';

    const OPEN_VPN_CLIENT_CRT_PATH = '/etc/openvpn/client.crt';

    const OPEN_VPN_SERVER_CRT_PATH = '/etc/openvpn/ca.crt';

    const OPEN_VPN_WILDIX_CLIENT_CONFIG_PATH = '/etc/openvpn/wildixclient.conf';

    const OPEN_VPN_PASS_FILE_PATH = '/rw2/etc/openvpn/passfile';

    const OPEN_VPN_HTTP_PROXY_PATH = '/rw2/etc/openvpn/httpproxy';

    const MIN_VALID_YEAR = 2010;

    const OPERATION_START = 'start';

    const OPERATION_STOP = 'stop';

    const OPERATION_RESTART = 'restart';

    const PROXY_AUTH_BASIC = 'basic';

    const PROXY_AUTH_NTLM = 'ntlm';

    private static $_configFiles = array(
        self::OPEN_VPN_SERVER_CRT_PATH,
        self::OPEN_VPN_CLIENT_CRT_PATH,
        self::OPEN_VPN_CLIENT_KEY_PATH,
        self::OPEN_VPN_WILDIX_CLIENT_CONFIG_PATH,
    );

    private $_httpProxy;
    private $_license;
    private $_portal;
    private $_log;

    /**
     * RemoteSupport constructor.
     *
     * @param HttpProxy               $httpProxy HTTP Proxy Settings.
     * @param \Wildix\License         $license   License service.
     * @param \Wildix\Services\Portal $portal    API of service Portal.
     * @param \Logger\Log             $log       Log service.
     */
    public function __construct(HttpProxy $httpProxy = null, \Wildix\License $license = null, \Wildix\Services\Portal $portal = null, \Logger\Log $log = null) {
        if (!isset($httpProxy)) {
            parent::__construct();

            return;
        }

        $this->_httpProxy = $httpProxy;
        $this->_license   = $license;
        $this->_portal    = $portal;
        $this->_log       = $log;
    }

    public function getHttpProxy() {
        if (!$this->_httpProxy) {
            $this->_httpProxy = new HttpProxy();
        }

        return $this->_httpProxy;
    }

    private function _getLicense(): \Wildix\License {
        if (!isset($this->_license)) {
            $this->_license = \Wildix\License::getInstance();
        }

        return $this->_license;
    }

    private function _getPortal(): \Wildix\Services\Portal {
        if (!isset($this->_portal)) {
            $_portalApiUrl = env('SERVICE_PORTAL_API_URI');
            $_portalUrl    = env('SERVICE_UPGRADE_PORTAL_URI');
            $this->_portal = new \Wildix\Services\Portal(\Logger\Log::INFO, $_portalApiUrl, $_portalUrl);
        }

        return $this->_portal;
    }

    private function _getLog(): \Logger\Log {
        if (!isset($this->_log)) {
            $this->_log = \Logger\LoggerFactory::getInstance();
            $this->_log->setLogLevel(\Logger\Log::INFO);
        }

        return $this->_log;
    }

    /**
     * Test Host with port
     *
     * @param string  $host    Host.
     * @param integer $port    Port.
     * @param integer $timeout Timeout.
     *
     * @return boolean
     */
    public function testHostPort($host, $port, $timeout) {
        $errno    = null;
        $errstr   = null;
        $resource = fsockopen($host, $port, $errno, $errstr, $timeout);

        $isResource = is_resource($resource);

        if ($isResource) {
            fclose($resource);
        } else {
            $this->_getLog()->debug(__LINE__, sprintf('Failed open socket on %s:%d | Error: %d - %s', $host, $port, $errno, $errstr));
        }

        return $isResource;
    }

    /**
     * Get current state
     *
     * @return integer
     */
    public function getState() {
        $exec = exec("ps afx | grep 'openvpn.*wildixclient' | grep -v grep | wc -l");

        if ($exec !== '1') {
            return 3;
        }

        $serverIp = $this->getServerIp();

        if ($serverIp === '') {
            return 1;
        }

        $isHostAvailable = $this->testHostPort($serverIp, 443, 5);

        return $isHostAvailable ? 0 : 1;
    }

    /**
     * Get server IP
     *
     * @return string
     */
    private function getServerIp() {
        $clientIp = explode('.', $this->getIp());
        $serverIp = '';

        // 172.16.0.8 - default gateway for RemoteSupport
        if (count($clientIp) == 4) {
            $clientIp[2] = '0';
            $clientIp[3] = ($clientIp[0] === '172' && $clientIp[1] === '16') ? '8' : '1';
            $serverIp    = implode('.', $clientIp);
        }

        return $serverIp;
    }

    /**
     * Get IP
     *
     * @return string
     */
    public function getIp() {
        $output = null;
        $ip     = '';
        $cmd    = "[ -d /sys/class/net/tun0/ ] && /sbin/ifconfig tun0 | grep -E 'inet\s.*\s[5|172].*.0.[1|8]' | awk '{print($2)}'";

        exec($cmd, $output);

        if (is_array($output) && isset($output[0]) && is_ip($output[0])) {
            $ip = $output[0];
        }

        return $ip;
    }

    /**
     * Init operation
     *
     * @param string $type Operation type.
     *
     * @throws Exception Error.
     *
     * @return array
     */
    public function initOperation($type) {
        $data = array(
            'success' => 1,
            'error'   => '',
            'verbose' => '',
        );

        try {
            switch ($type) {
                case self::OPERATION_RESTART:
                    $this->operationRestart();
                    break;
                case self::OPERATION_STOP:
                    $state = $this->operationStop();

                    if ($state) {
                        $data['state'] = $state;
                    }
                    break;
                case self::OPERATION_START:
                    $this->operationStart();
                    break;
                default:
                    throw new Exception('Invalid type');
            }
        } catch (Exception $e) {
            $data['error'] = $e->getCode();
        }

        return $data;
    }

    /**
     * Check valid year
     *
     * @throws Exception Error.
     *
     * @return void.
     */
    private function _checkValidYear() {
        if (date('Y') < self::MIN_VALID_YEAR) {
            throw new Exception(Error::DateLooksWrong);
        }
    }

    /**
     * Operation restart.
     *
     * @throws Exception Error.
     *
     * @return void.
     */
    protected function operationRestart() {
        $this->_checkValidYear();
        $this->stop();
        $this->actualizeConfiguration();
        $this->start();
    }

    /**
     * Operation stop.
     *
     * @return integer
     */
    protected function operationStop() {
        $this->stop();
        return $this->getState();
    }

    /**
     * Operation start.
     *
     * @throws Exception Error.
     *
     * @return void
     */
    protected function operationStart() {
        $this->_checkValidYear();
        $this->stop();

        $this->actualizeConfiguration();
        $this->_updateProxySettings();

        $this->start();
    }

    private function _updateProxySettings(): void {
        $file               = file_get_contents(self::OPEN_VPN_WILDIX_CLIENT_CONFIG_PATH);
        $openVpnOptions     = preg_split('/$\R?^/m', $file);
        $proxySettings      = $this->_getProxySettings();
        $proxyString        = null;
        $proxyOptionsString = null;

        if ($proxySettings['proxy'] && $proxySettings['proxyOptions']) {
            $proxyString        = $proxySettings['proxy'];
            $proxyOptionsString = $proxySettings['proxyOptions'];
        }

        $pSFound  = null;
        $posFound = null;

        $openVpnOptions = array_filter($openVpnOptions);

        foreach ($openVpnOptions as $key => $option) {
            $pos = strpos($option, 'http-proxy ');

            if ($pos !== false) {
                $pSFound              = true;
                $openVpnOptions[$key] = $proxyString;
            }

            $pos = strpos($option, 'http-proxy-option ');

            if ($pos !== false) {
                $posFound             = true;
                $openVpnOptions[$key] = $proxyOptionsString;
            }
        }

        if (!$pSFound && $proxyString) {
            $openVpnOptions[] = $proxyString;
        }

        if (!$posFound && $proxyOptionsString) {
            $openVpnOptions[] = $proxyOptionsString . PHP_EOL;
        }

        file_put_contents(
            self::OPEN_VPN_WILDIX_CLIENT_CONFIG_PATH,
            implode(PHP_EOL, $openVpnOptions)
        );
    }

    /**
     * Get proxy settings.
     *
     * @return array
     */
    private function _getProxySettings() {
        $proxySetting = $this->getHttpProxy()->getSettings();
        $settings     = array(
            'proxy'        => '',
            'proxyOptions' => '',
        );

        if ($proxySetting) {
            if ($proxySetting['host'] && $proxySetting['port'] && $proxySetting['enable']) {
                $settings['proxy']        = 'http-proxy ' . $proxySetting['host'] . ' ' . $proxySetting['port'];
                $settings['proxyOptions'] = 'http-proxy-option AGENT Mozilla/5.0+(Windows;+U;+Windows+NT+5.0;+en-GB;+rv:1.7.6)+Gecko/20050226+Firefox/1.0.1';
            }

            if ($proxySetting['auth_method'] && $proxySetting['auth_method'] != 'none' && $proxySetting['enable']) {
                $settings['proxy'] .= ' /rw2/etc/openvpn/passfile ' . $proxySetting['auth_method'];
            }
        }

        return $settings;
    }

    /**
     * Stop
     *
     * @return void.
     */
    public function stop() {
        $this->autoStartList('del');
        exec('sudo /bin/systemctl stop openvpn > /dev/null 2>&1');
        exec('sudo /bin/systemctl disable openvpn > /dev/null 2>&1');
    }

    /**
     * Start
     *
     * @return void.
     */
    public function start() {
        $this->autoStartList('add');
        exec('sudo /bin/systemctl enable openvpn > /dev/null 2>&1');
        exec('sudo /bin/systemctl start openvpn > /dev/null 2>&1');
    }

    /**
     * Auto start list
     *
     * @param string $action Action.
     *
     * @return void
     */
    public function autoStartList($action) {
        // @codingStandardsIgnoreLine
        $config = @parse_ini_file(self::REMOTE_SUPPORT_CONFIG_PATH);

        if ($action == 'add') {
            $config['autostart'] = 1;
            $replace             = '';
        } else {
            $config['autostart'] = 0;
            $replace             = '#';
        }

        write_ini_file(self::REMOTE_SUPPORT_CONFIG_PATH, $config);

        if (file_exists(self::OPEN_VPN_MONIT_PATH)) {
            $config = file(self::OPEN_VPN_MONIT_PATH);
            $config = preg_replace('/^#?(.*)$/', $replace . '$1', $config);

            file_put_contents(self::OPEN_VPN_MONIT_PATH, implode('', $config));
        }

        exec('sudo /bin/systemctl restart monit');
    }

    /**
     * Check and create configs.
     *
     * @param string $fileName File name.
     *
     * @throws Exception Error.
     *
     * @return void.
     */
    private function _checkConfig($fileName) {
        if (!file_exists($fileName)) {
            touch($fileName);
        }

        if (!file_exists($fileName) || !is_readable($fileName)) {
            throw new Exception(Error::IsNotWriteable . ' ' . $fileName);
        }
    }

    /**
     * Opens file in writable mode.
     *
     * @param string $fileName File name.
     *
     * @return resource
     *
     * @throws Exception Throws exception if file is not writable.
     */
    private function _getWritableFile($fileName) {
        // @codingStandardsIgnoreLine
        $resource = @fopen($fileName, 'w');

        if (!is_resource($resource)) {
            throw new Exception(Error::IsNotWriteable . ' ' . $fileName);
        }

        return $resource;
    }

    /**
     * Write config.
     *
     * @param array $vars Config.
     *
     * @return void.
     *
     * @throws Exception Error.
     */
    public function writeConfigs(array $vars) {
        $clientKey = $this->_getWritableFile(self::OPEN_VPN_CLIENT_KEY_PATH);
        $clientCrt = $this->_getWritableFile(self::OPEN_VPN_CLIENT_CRT_PATH);
        $serverCrt = $this->_getWritableFile(self::OPEN_VPN_SERVER_CRT_PATH);

        $this->_checkConfig(self::OPEN_VPN_WILDIX_CLIENT_CONFIG_PATH);
        $this->_checkConfig(self::OPEN_VPN_PASS_FILE_PATH);
        $this->_checkConfig(self::OPEN_VPN_HTTP_PROXY_PATH);

        $httpProxyRow = '';
        $proxyOpts    = parse_ini_file(self::OPEN_VPN_HTTP_PROXY_PATH, false, INI_SCANNER_RAW);

        if (is_array($proxyOpts) && $proxyOpts['proxy_enable'] == '1') {
            $httpProxyRow  = 'http-proxy ' . $proxyOpts['proxy_host'] . ' ';
            $httpProxyRow .= $proxyOpts['proxy_port'] . ' ';
            $httpProxyRow .= self::OPEN_VPN_PASS_FILE_PATH . ' ';
            $httpProxyRow .= $proxyOpts['proxy_auth_method'] . "\n";
            $httpProxyRow .= $vars['httpProxy']['addOptions'];
        }

        $vars['client']['config'] = str_replace('${wms_BASEPATH}', BASEPATH, $vars['client']['config']);
        $vars['client']['config'] = str_replace('${wms_http_proxy}', $httpProxyRow, $vars['client']['config']);
        $vars['client']['config'] = preg_replace('/\$\{wms_(.*?)\}/', '', $vars['client']['config']);

        $fp = fopen(self::OPEN_VPN_WILDIX_CLIENT_CONFIG_PATH, 'w');
        fwrite($fp, $vars['client']['config']);
        fclose($fp);

        fwrite($clientKey, $vars['client']['key']);
        fclose($clientKey);

        fwrite($clientCrt, $vars['client']['crt']);
        fclose($clientCrt);

        fwrite($serverCrt, $vars['server']['crt']);
        fclose($serverCrt);
    }

    /**
     * Check state
     *
     * @return array
     */
    public function checkState() {
        $data = array(
            'success' => 1,
            'error'   => '',
            'verbose' => '',
            'state'   => $this->getState(),
        );

        if ($data['state'] == 0) {
            $data['vtun_ip'] = $this->getIp();
        }

        return $data;
    }

    /**
     * Update OpenVPN configuration for Remote Support if needed.
     *
     * @return void
     * @throws \LogicException if license is not active.
     */
    public function actualizeConfiguration(): void {
        if (!$this->_isConfigurationExists()) {
            $this->_updateConfiguration();

            return;
        }

        if ($this->_isCrtsExpired()) {
            // After the expiration of the certificate, the configuration may not be updated.
            // Therefore, it is previously deleted, in order to avoid errors in the OpenVPN service.
            $this->_removeConfiguration();
            $this->_updateConfiguration();

            return;
        }

        if ($this->_isCrtsNeedUpdate()) {
            $this->_updateConfiguration();

            return;
        }

        $this->_getLog()->debug(__LINE__, 'OpenVPN configuration of Remote Support is not need update');
    }

    private function _isConfigurationExists(): bool {
        foreach (self::$_configFiles as $file) {
            if (!file_exists($file)) {
                return false;
            }
        }

        return true;
    }

    private function _updateConfiguration(): void {
        if (!$this->_getLicense()->isActivated()) {
            throw new \LogicException('License isn\'t activated.');
        }

        $configurationData = $this->_getConfigurationFromPortal();
        $this->writeConfigs($configurationData);

        $this->_getLog()->info(__LINE__, 'New OpenVPN configuration was stored');
    }

    private function _isCrtsExpired(): bool {
        $serverCrt = $this->_getCrt(self::OPEN_VPN_SERVER_CRT_PATH);
        $clientCrt = $this->_getCrt(self::OPEN_VPN_CLIENT_CRT_PATH);

        return ($this->_isCrtExpired($serverCrt) || $this->_isCrtExpired($clientCrt));
    }

    private function _getCrt($path): array {
        return openssl_x509_parse(file_get_contents($path));
    }

    private function _isCrtExpired(array $crt): bool {
        return (isset($crt['validTo_time_t']) && $this->_isTimeLessThanOrEqualsTo($crt['validTo_time_t'], 'now'));
    }

    /**
     * Compare timestamp with formatted time string.
     *
     * @param integer $time          A timestamp.
     * @param string  $formattedTime A date/time string. Valid formats are explained in Date and Time Formats.
     *
     * @return boolean
     */
    private function _isTimeLessThanOrEqualsTo(int $time, string $formattedTime): bool {
        $utcTimeZone           = new \DateTimeZone('UTC');
        $dateTimeForComparison = new \DateTimeImmutable($formattedTime, $utcTimeZone);

        return ($time <= $dateTimeForComparison->getTimestamp());
    }

    private function _removeConfiguration(): void {
        array_map([$this, '_removeFile'], self::$_configFiles);
    }

    private function _removeFile(string $filePath): void {
        if (file_exists($filePath)) {
            unlink($filePath);
        }
    }

    private function _isCrtsNeedUpdate(): bool {
        $serverCrt = $this->_getCrt(self::OPEN_VPN_SERVER_CRT_PATH);
        $clientCrt = $this->_getCrt(self::OPEN_VPN_CLIENT_CRT_PATH);

        return ($this->_isCrtShouldBeReplaced($serverCrt) || $this->_isCrtCloseToExpire($serverCrt) || $this->_isCrtCloseToExpire($clientCrt));
    }

    private function _isCrtShouldBeReplaced(array $crt): bool {
        // Certificates that expired in May 2019 was created before change OpenVPN configuration in WMS-4917 and should be replaced.
        return (!isset($crt['validTo_time_t']) || date('Y-m', $crt['validTo_time_t']) === '2019-05');
    }

    private function _isCrtCloseToExpire(array $crt): bool {
        return (isset($crt['validTo_time_t']) && $this->_isTimeLessThanOrEqualsTo($crt['validTo_time_t'], '+6 months'));
    }

    private function _getConfigurationFromPortal(): array {
        $configurationData = $this->_getPortal()->getRemoteSupportConfigs();
        $this->_assertConfigurationDataIsValid($configurationData);

        return $configurationData;
    }

    /**
     * Assert that a configuration data is valid.
     *
     * @param mixed $configurationData Should be array with next structure
     *                                 {client: {key: <string>, crt: <string>, config: <string>},
     *                                 server: {crt: <string>},
     *                                 httpProxy: {addOptions: <string>}}.
     *
     * @return void
     * @throws \ErrorException if configuration data isn't valid.
     */
    private function _assertConfigurationDataIsValid($configurationData): void {
        if (!\is_array($configurationData)
            || empty($configurationData['client']['key'])
            || empty($configurationData['client']['crt'])
            || empty($configurationData['client']['config'])
            || empty($configurationData['server']['crt'])
            || empty($configurationData['httpProxy']['addOptions'])
        ) {
            throw new \ErrorException('This is not valid OpenVPN configuration.');
        }
    }
}
