<?php

namespace Wildix\System\RemoteSyslogSettings;

use Wildix\Databases\WMSDB;
use Wildix\Error;
use Wildix\Exception;
use Wildix\InvalidDataException;

/**
 * Class Repository
 */
class Repository {

    const STORAGE_PLACE = 'REMOTE_SYSLOG_SETTINGS';

    /**
     * Instance of WMSDB.
     *
     * @var WMSDB
     */
    private $_storage;
    /**
     * Instance of settings factory.
     *
     * @var Factory
     */
    private $_settingsFactory;

    /**
     * Repository constructor.
     *
     * @param WMSDB   $storage         A storage for persist settings data.
     * @param Factory $settingsFactory A factory for build an instance of settings.
     */
    public function __construct(WMSDB $storage, Factory $settingsFactory) {
        $this->_storage         = $storage;
        $this->_settingsFactory = $settingsFactory;
    }

    /**
     * Find settings data in storage and return instance of settings.
     *
     * @return Settings
     * @throws Exception If failed create instance of settings from persisted data.
     */
    public function find() {
        $data = $this->loadDataFromStorage();

        if (!isset($data)) {
            return null;
        }

        try {
            $settings = $this->_settingsFactory->create(json_decode($data, true));
        } catch (InvalidDataException $e) {
            throw new Exception(Error::UnknownError, 500, $e);
        }

        return $settings;
    }

    /**
     * Saves the settings in the storage.
     *
     * @param Settings $settings Instance of settings for saving.
     *
     * @return void
     */
    public function save(Settings $settings) {
        $this->_storage->setSystemVariable(self::STORAGE_PLACE, json_encode($settings->toArray()));
    }

    /**
     * Loads data from storage.
     *
     * @return string|null
     */
    private function loadDataFromStorage() {
        return $this->_storage->getSystemVariable(self::STORAGE_PLACE);
    }
}
