<?php

namespace Wildix\System\RemoteSyslogSettings;

use Wildix\Settings\RsyslogLogging;

/**
 * Class Settings
 */
class Settings {

    const FIELD_ENABLED   = 'enabled';
    const FIELD_FILTER    = 'filter';
    const FIELD_TRANSPORT = 'transport';
    const FIELD_FORMAT    = 'format';
    const FIELD_HOST      = 'host';
    const FIELD_PORT      = 'port';

    const TRANSPORT_TCP = 'tcp';
    const TRANSPORT_UDP = 'udp';

    const PORT_DEFAULT = 514;

    const FORMAT_RSYSLOG_FORWARD             = 'RSYSLOG_ForwardFormat';
    const FORMAT_RSYSLOG_TRADITIONAL_FORWARD = 'RSYSLOG_TraditionalForwardFormat';
    const FORMAT_RSYSLOG_SYSLOG_PROTOCOL_23  = 'RSYSLOG_SyslogProtocol23Format';

    public static $fields     = array(
        self::FIELD_ENABLED,
        self::FIELD_FILTER,
        self::FIELD_TRANSPORT,
        self::FIELD_FORMAT,
        self::FIELD_HOST,
        self::FIELD_PORT,
        );
    public static $transports = array(
        self::TRANSPORT_TCP,
        self::TRANSPORT_UDP,
        );
    public static $formats    = array(
        self::FORMAT_RSYSLOG_FORWARD,
        self::FORMAT_RSYSLOG_TRADITIONAL_FORWARD,
        self::FORMAT_RSYSLOG_SYSLOG_PROTOCOL_23,
    );

    /**
     * Is the server enabled.
     *
     * @var boolean
     */
    private $_enabled;
    /**
     * A message filter.
     *
     * @var string
     */
    private $_filter;
    /**
     * The transport protocol name.
     *
     * @var string
     */
    private $_transport;
    /**
     * The format used for a message.
     *
     * @var string
     */
    private $_format;
    /**
     * The server address.
     *
     * @var string
     */
    private $_host;
    /**
     * The server port.
     *
     * @var integer
     */
    private $_port;

    /**
     * Settings constructor.
     *
     * @param array $data Settings in array format.
     */
    public function __construct(array $data) {
        $this->_enabled   = (bool)$data[self::FIELD_ENABLED];
        $this->_filter    = $data[self::FIELD_FILTER];
        $this->_transport = $data[self::FIELD_TRANSPORT];
        $this->_format    = $data[self::FIELD_FORMAT];
        $this->_host      = $data[self::FIELD_HOST];
        $this->_port      = isset($data[self::FIELD_PORT]) ? (int)$data[self::FIELD_PORT] : self::PORT_DEFAULT;
    }

    /**
     * Convert object data to an array
     *
     * @return array
     */
    public function toArray() {
        return array(
            'enabled' => $this->_enabled,
            'filter' => $this->_filter,
            'transport' => $this->_transport,
            'format' => $this->_format,
            'host' => $this->_host,
            'port' => $this->_port,
        );
    }

    /**
     * Change the rsyslog remote logging configuration file with own data or remove it if settings is disabled.
     *
     * @param RsyslogLogging $settingsRsyslogLogging Logging settings manipulator for rsyslog.
     *
     * @return void
     */
    public function apply(RsyslogLogging $settingsRsyslogLogging) {
        if ($this->_enabled) {
            $settingsRsyslogLogging->updateRule(RsyslogLogging::RULE_REMOTE_SERVER, $this->_filter, $this->_buildActionForConfiguration());
        } else {
            $settingsRsyslogLogging->removeRule(RsyslogLogging::RULE_REMOTE_SERVER);
        }
    }

    private function _buildActionForConfiguration() {
        $action  = ($this->_transport === self::TRANSPORT_TCP) ? '@@' : '@';
        $action .= $this->_host . ':' . $this->_port . ';' . $this->_format;

        return $action;
    }
}
