<?php

namespace Wildix\System;

/**
 * Snmp is a class working with SNMP Server
 * http://www.net-snmp.org/docs/man/snmpd.examples.html
 */
class Snmp {
    const CONFIG_FILE         = '/rw2/etc/snmp/snmpd.conf';
    const DEFAULT_CONFIG_FILE = array(
        'syslocation  wildix',
        'syscontact  support@wildix.com',
        'sysservices 78',
        '%s',
        'view all    included  .1                               80',
        'view system included  .iso.org.dod.internet.mgmt.mib-2.system',
        'pass .1.3.6.1.4.1.7777.1 /usr/bin/php /usr/sbin/snmp-trunks',
    );

    const TRUSTED_ADDR = 'trustedAddr';

    private $_defaultOptions = array(
        'enable'       => 'off',
        'port'         => 161,
        'community'    => 'public',
        self::TRUSTED_ADDR => [],
    );

    private $_options = null;
    private $_systemd = null;

    private function getSystemd() {
        if ($this->_systemd === null) {
            $this->_systemd = new \Wildix\System\Systemd('snmpd');
        }
        return $this->_systemd;
    }

    /**
     * Retrieve SNMP options
     *
     * @param string $file File that we will parse.
     *
     * @return array
     */
    public function getOptions($file = self::CONFIG_FILE) {
        if ($this->_options === null) {
            if (!file_exists($file) || !is_readable($file)) {
                return $this->_defaultOptions;
            }

            $this->_options = $this->_defaultOptions;

            $this->_options['enable'] = ($this->isEnabled()) ? 'on' : 'off';

            // get community & trusted addresses
            $content = file_get_contents($file);
            if (preg_match_all('/rocommunity\s(\w+)\s([0-9.\/]*)/', $content, $matches)) {
                if (isset($matches[1][1])) {
                    $this->_options['community'] = $matches[1][1];
                }

                foreach ($matches[2] as $addr) {
                    if ($addr == '127.0.0.1') {
                        continue;
                    }
                    if ($addr != '') {
                        $this->_options[self::TRUSTED_ADDR][] = trim($addr);
                    }
                }
            }

            // get listened port
            if (preg_match('/agentaddress (\w+):([0-9.\/]+)/', $content, $matches)) {
                if (!empty($matches) && !empty($matches[2])) {
                    $this->_options['port'] = (int)$matches[2];
                }
            }
        }

        return $this->_options;
    }

    private function _validate($params) {
        $errors = [];

        $form = new \Wildix\Validator(
            $params,
            [
                'enable',
                'enable',
                'community',
                'port',
                self::TRUSTED_ADDR,
            ]
        );

        $form->rule(
            'ipPort',
            ['port']
        );

        $form->rule(
            'in',
            'enable',
            [
                'on',
                'off',
                ]
        );

        $form->rule(
            'alpha',
            ['community']
        );

        $form->rule(
            'ipPort',
            ['port']
        );

        $form->rule(
            'boolean',
            ['enable']
        );

        $form->rule(
            'alpha',
            ['community']
        );

        $form->rule(
            'required',
            array(
                'enable',
                'community',
                'port',
                self::TRUSTED_ADDR,
            )
        );

        if (!$form->validate()) {
            $errors = $form->errors();
        }

        $form = new \Wildix\Validator();
        $form->rule(
            function ($field, $value, $params, $fields) {
                if (!is_string($value) || empty($value)) {
                    return false;
                }

                $arr = explode('/', $value);
                if (empty($arr[0]) || !filter_var($arr[0], FILTER_VALIDATE_IP, FILTER_FLAG_IPV4)) {
                    return false;
                }


                if (isset($arr[1]) && (!preg_match('/^\d+$/', $arr[1]) || $arr[1] < 0 || $arr[1] > 32)) {
                    return false;
                }

                if (isset($arr[2])) {
                    return false;
                }

                return true;
            },
            self::TRUSTED_ADDR
        )->message(\Wildix\Error::InvalidIpAddress);

        foreach ($params[self::TRUSTED_ADDR] as $idx => $item) {
            $tmpErrors = $form->validateData([self::TRUSTED_ADDR => $item], [self::TRUSTED_ADDR]);
            if ($tmpErrors) {
                if (empty($errors[self::TRUSTED_ADDR])) {
                    $errors = $errors + $tmpErrors;
                }

                $errors[self::TRUSTED_ADDR]['rows'][] = [
                    'key' => $idx,
                    'value' => $item,
                ];
            }
        }

        if (!empty($errors)) {
            throw new \Wildix\InvalidDataException('Validation error', $errors);
        }
    }

    private function generateRoCommunity($snmpParameters) {
        $rocommunity = array();

        foreach ($snmpParameters[self::TRUSTED_ADDR] as $addr) {
            $addrArr = explode('/', $addr);
            $address = $addrArr[0];

            if (trim($address) !== '') {
                $cidr = isset($addrArr[1]) ? (int)$addrArr[1] : 32;
                if ($cidr == 0) {
                    $addr = '0.0.0.0/' . $cidr;
                } else if ($cidr >= 1 && $cidr <= 15) {
                    $addr = preg_replace('/(.*)\.(.*)\.(.*)\.(.+)$/', '${1}.0.0.0/' . $cidr, $address);
                } else if ($cidr >= 16 && $cidr <= 23) {
                    $addr = preg_replace('/(.*)\.(.*)\.(.*)\.(.+)$/', '${1}.${2}.0.0/' . $cidr, $address);
                } else if ($cidr >= 24 && $cidr <= 31) {
                    $addr = preg_replace('/(.*)\.(.*)\.(.*)\.(.+)$/', '${1}.${2}.${3}.0/' . $cidr, $address);
                } else if ($cidr == 32) {
                    $addr = $address . '/' . $cidr;
                }
            }

            if (trim($addr) !== '') {
                $rocommunity[] = 'rocommunity ' . $snmpParameters['community'] . ' ' . trim($addr);
            }
        }

        if (empty($rocommunity)) {
            $rocommunity[] = 'rocommunity ' . $snmpParameters['community'];
        }
        $rocommunity[] = 'rocommunity ' . $snmpParameters['community'] . ' 127.0.0.1';

        return $rocommunity;
    }

    public function isEnabled() {
        return $this->getSystemd()->isEnabled();
    }
    /**
     * Save snmp options to corresponding config files.
     *
     * @param array  $snmpParameters Array of snmp options.
     * @param string $file           File that we will parse.
     *
     * @return boolean
     */
    public function setOptions(array $snmpParameters, $file = self::CONFIG_FILE) {
        $somethingUpdated = false;

        $this->_validate($snmpParameters);

        if (!file_exists($file) || !is_writable($file)) {
            return false;
        }

        $snmpParameters['community']        = trim($snmpParameters['community']);
        $snmpParameters[self::TRUSTED_ADDR] = array_unique($snmpParameters[self::TRUSTED_ADDR]);

        // generate rocommunity <community name> <trusted IP>
        $rocommunity = $this->generateRoCommunity($snmpParameters);

        $output   = array();
        $output[] = 'agentaddress udp:' . $snmpParameters['port'] . "\n";
        $output[] = implode("\n", $rocommunity);

        $state = ($this->isEnabled()) ? 'on' : 'off';
        if ($state != $snmpParameters['enable']) {
            $somethingUpdated = true;
        }

        $defaultContent = implode("\n", self::DEFAULT_CONFIG_FILE);
        $newContent     = sprintf($defaultContent, implode($output));
        if ($defaultContent != $newContent) {
            $somethingUpdated = true;
            file_put_contents($file, $newContent);
        }

        if ($somethingUpdated) {
            if ($snmpParameters['enable'] == 'on') {
                $this->restart();
            } else {
                $this->disable();
            }
        }
        $this->_options = null;
        return true;
    }

    public function restart() {
        if (!$this->isEnabled()) {
            $this->getSystemd()->enable();
        }

        $this->getSystemd()->restart();
    }

    public function disable() {
        if ($this->getSystemd()->isActive()) {
            $this->getSystemd()->stop();
        }
        $this->getSystemd()->disable();
    }
}
