<?php
namespace Wildix\System\Storage;

use Wildix\License;
use Wildix\System\FileSystem;
use Wildix\System\Mount;

/**
 * Class Storage
 */
class Storage {

    /**
     * File for writing to for reboot mounts.
     *
     * @var string
     **/
    public static $fileName = '/etc/wms.conf';
    /**
     * File section for writing to
     *
     * @var string
     **/
    public static $fileSection = 'Devices';

    private $_mount;
    private $_storageDevices;
    private $_errors;

    public function __construct(Mount $mount, StorageDevices $storageDevices) {
        $this->_mount          = $mount;
        $this->_storageDevices = $storageDevices;

        $this->_errors = array();
    }

    public function getStorageData() {
        $this->_mount->updateMountsList();
        return $this->_storageDevices->getStorages($this->_mount->getMountsByType());
    }

    public function getErrors() {
        return array_merge($this->_errors, $this->_mount->errors);
    }

    public function getMounted($params) {
        $mount   = $this->_mount->getMounts();
        $mounted = array();

        foreach ($params as $dev => $arr) {
            foreach ($mount as $line) {
                if (substr_count($line, '/dev/' . $dev)) {
                    $matches = preg_split('!(/dev/' . $dev . ' | | )!', $line);
                    if (!empty($matches[1]) && trim($matches[1]) != '/') {
                        $mounted[$dev][] = trim($matches[1]);
                    }
                }
            }
        }

        return $mounted;
    }

    private function changeIpWithNames($mounts) {
        $config = parse_ini_file(self::$fileName, true, INI_SCANNER_RAW);
        if (!isset($config[self::$fileSection])
            || !is_array($config[self::$fileSection])
            || empty($config[self::$fileSection])
        ) {
            return $mounts;
        }
        $devices = $config[self::$fileSection];
        foreach ($mounts as $mpoint => $path) {
            foreach ($devices as $devMpoint => $mountRow) {
                if ($devMpoint == $mpoint) {
                    $params          = explode('|', $mountRow);
                    $mounts[$mpoint] = preg_replace('~//\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/~', '//' . $params[1] . '/', $path);
                    break;
                }
            }
        }
        return $mounts;
    }

    public function getData() {
        $this->_mount->updateSmbMounts();

        if (isset($this->_mount->availableMps['/dev/shm/var/spool/callweaver/outgoing'])) {
            unset($this->_mount->availableMps['/dev/shm/var/spool/callweaver/outgoing']);
        }

        $storage = $this->getStorageData();

        $mountPoints = $this->getMounted($storage['params']);
        // ------------------------ USB mounts!
        $storageDeviceData = $this->_storageDevices->getDeviceData($storage['usbExi'], $storage['params'], $mountPoints, $this->_mount->availableMps);
        // ------------------------ Array of available mounts for SMB
        $availableRemoteSharesMountPoints = $this->_mount->getMountPointsForSmb();
        // ------------------------ Samba partitions
        $sambaMounts = $this->changeIpWithNames($this->_mount->sambaMounts);
        // ------------------------ Nfs partitions
        $nfsMounts = $this->changeIpWithNames($this->_mount->nfsMounts);
        // ------------------------  Not mounted partitions from wms.conf
        $notMountedDevices = $this->_storageDevices->getNotMountedDevicesFromConf();

        return array(
            'usbMounts' => $storageDeviceData,
            'availableRemoteSharesMountPoints' => $availableRemoteSharesMountPoints,
            'sambaMounts' => $sambaMounts,
            'nfsMounts' => $nfsMounts,
            'notMounted' => $notMountedDevices
        );
    }

    /**
     * Remount SMB device
     *
     * @param string $mPoint MPoint.
     * @param string $type   Type.
     *
     * @return boolean
     */
    final public function remountShare($mPoint, $type) {
        $notMountedSMBDevices = $this->_storageDevices->getNotMountedDevicesFromConf($type);

        if (array_key_exists($mPoint, $notMountedSMBDevices)) {
            return $this->mountShare($mPoint, $notMountedSMBDevices[$mPoint]);
        }
        return false;
    }

    public function initUsb($block) {
        $storage = $this->getStorageData();

        // check if some of partition of the initialized device is already mounted
        foreach (array_keys($this->getMounted($storage['params'])) as $key) {
            if ($block == substr($key, 0, strlen($block)) && isset($storage['params'][$key])) {
                return false;
            }
        }

        if (!isset($storage['usbExi'][$block])) {
            return false;
        }

        $mountPoints = Mount::getMountPoints(true);
        // need to mount only not mounted mountpoints
        $systemMounts = $this->_mount->getMountsByType();

        foreach ($systemMounts as $systemMount => $tVal) {
            if (isset($mountPoints[$systemMount])) {
                unset($mountPoints[$systemMount]);
            }
        }

        if (empty($mountPoints) && empty($systemMounts['/mnt/backups'])) {
            $mountPoints = array('/mnt/backups' => 1);
        }

        if (!$this->_storageDevices->initUsb($block)) {
            $this->_errors[] = 'Could not prepare USB key';
            return false;
        }

        $this->_mount->updateMountsList();

        foreach ($mountPoints as $mountpoint => $val) {
            $this->mountUsb($block . '1', $mountpoint);
        }

        return true;
    }

    public function mountAllDevices($devicesToMount, $params) {
        foreach ($params as $dev => $dparams) {
            $this->mountDevices($devicesToMount['usb'], $dev, 'usb', $dparams);
        }

        $this->mountDevices($devicesToMount['smb']);
    }

    private function needReloadCdr(&$settings, $mountPoint) {
        $reloadSqlite = false;
        $reloadCsv    = false;

        if ($mountPoint == '/mnt/cdr' && License::getInstance()->getLimitations('phones') > 0) {
            $settings['sqlite3']['dbfile'] = 'cdrdb';
            $settings['sqlite3']['dbdir']  = '/mnt/cdr';
            $settings['sqlite3']['table']  = 'cdr';

            if (License::getInstance()->getLimitations('phones') <= 40) {
                $settings['sqlite3']['enabled'] = 'yes';
                \CDR::write_cdr_conf($settings);

                if (\CDR::getInstance()->SetSqliteDB() !== true) {
                    $settings['sqlite3']['enabled'] = 'no';
                } else {
                    $temp = \CDR::getInstance()->getSettings();
                    if ($temp['general']['ctibackend'] == ''
                        || ($temp['mysql']['enabled'] != 'yes' && $temp['tds']['enabled'] != 'yes')
                    ) {
                        $settings['general']['enabled']    = 'yes';
                        $settings['general']['ctibackend'] = 'sqlite3';
                    }
                }
            }

            $reloadSqlite = true;

            // cdr csv part
            if ($settings['csv']['logdir'] == '/mnt/cdr' && $settings['csv']['master'] != '') {
                $settings['csv']['enabled'] = 'yes';

                $reloadCsv = true;
            }
        }
        if ($mountPoint == '/mnt/cdrcsv') {
            if ($settings['csv']['logdir'] == '/mnt/cdrcsv' && $settings['csv']['master'] != '') {
                $settings['csv']['enabled'] = 'yes';

                $reloadCsv = true;
            }
        }

        return array(
            'sqlite3' => $reloadSqlite,
            'csv'     => $reloadCsv
        );
    }

    public function setCdrMount($mountPoint) {
        if ($mountPoint != '/mnt/cdr' && $mountPoint != '/mnt/cdrcsv') {
            return false;
        }

        $settings = \CDR::getInstance()->getSettings();
        if (!$settings) {
            return false;
        }

        $needReload = $this->needReloadCdr($settings, $mountPoint);

        \CDR::write_cdr_conf($settings);

        if ($needReload['sqlite3']) {
            if ($settings['sqlite3']['enabled'] == 'yes') {
                \PBX::getInstance()->Command('module unload cdr_sqlite3_custom.so');
                \PBX::getInstance()->Command('module load cdr_sqlite3_custom.so');
            } else {
                \PBX::getInstance()->Command('module unload cdr_sqlite3_custom.so');
            }
        }
        if ($needReload['csv']) {
            \PBX::getInstance()->Command('reload cdr_csv.so');
        }

        return true;
    }

    public function unsetCdrMount($mountPoint) {
        $settings = \CDR::getInstance()->getSettings();
        if (!$settings) {
            return;
        }
        if ($settings['sqlite3']['enabled'] == 'no' && $settings['csv']['enabled'] == 'no') {
            return;
        }

        if ($mountPoint == '/mnt/cdr') {
            $unsetCdrCsv = false;

            if (!file_exists('/mnt/cticonnect')) {
                if (isset($settings['csv']['logdir']) && $settings['csv']['logdir'] == '/mnt/cdr') {
                    $unsetCdrCsv = true;
                }

                $settings['sqlite3']['enabled'] = 'no';

                if ($unsetCdrCsv) {
                    $settings['csv']['enabled'] = 'no';
                }
            } else {
                $settings['sqlite3']['dbdir'] = '/mnt/cticonnect';
            }

            \CDR::write_cdr_conf($settings);
            \PBX::getInstance()->Command('reload cdr_sqlite3.so');

            if ($unsetCdrCsv) {
                \PBX::getInstance()->Command('reload cdr_csv.so');
            }
        }

        if ($mountPoint == '/mnt/cdrcsv' && isset($settings['csv']['logdir']) && $settings['csv']['logdir'] == '/mnt/cdrcsv') {
            $settings['csv']['enabled'] = 'no';
            \CDR::write_cdr_conf($settings);
            \PBX::getInstance()->Command('reload cdr_csv.so');
        }
    }

    private function delFromConf($key) {
        if (!file_exists(self::$fileName)) {
            return false;
        }
        $assocArr = parse_ini_file(self::$fileName, true, INI_SCANNER_RAW);
        if (isset($assocArr[self::$fileSection][$key])) {
            unset($assocArr[self::$fileSection][$key]);
        }
        write_ini_file(self::$fileName, $assocArr);
        return true;
    }

    private function putToConf($key, $item) {
        if (file_exists(self::$fileName)) {
            $assocArr = parse_ini_file(self::$fileName, true, INI_SCANNER_RAW);
        }
        $assocArr[self::$fileSection][$key] = $item;
        write_ini_file(self::$fileName, $assocArr);
    }

    private function postmount($mountpoint) {
        $mountpoint = rtrim($mountpoint, '/');

        if ($mountpoint == '/mnt/cdr' || $mountpoint == '/mnt/cdrcsv') {
            $this->setCdrMount($mountpoint);
        }

        if ($mountpoint == '/mnt/cticonnect') {
            $ctiDbJournal = '/mnt/cticonnect/cticonnectdb-journal';
            exec('[ `ls -la ' . $ctiDbJournal . ' 2>&1 | grep -c \'Input/output error\'` -ne 0 ] && sudo rm -f ' . $ctiDbJournal);
        }

        if ($mountpoint == '/mnt/cdr') {
            $cdrDbJournal = '/mnt/cdr/cdrdb-journal';
            exec('[ `ls -la ' . $cdrDbJournal . ' 2>&1 | grep -c \'Input/output error\'` -ne 0 ] && sudo rm -f ' . $cdrDbJournal);
        }

        if ($mountpoint == '/mnt/cticonnect' && !\CDR::getInstance()->setCticonnectDb()) {
            $this->_errors[] = _t('strings.CanNotCreateCTIDB');
            $this->unMount($mountpoint, 'usb');
            return false;
        }

        // Perform backup operations
        if ($mountpoint == '/mnt/backups') {
            $out = FileSystem::getInstance()->moveBackupsToShare();
            if ($out != 'OK') {
                $this->_errors[] = _t($out);
                $this->unMount($mountpoint, 'usb');
                return false;
            }
        }

        return true;
    }

    public function mountAll() {
        $mountPoints                 = Mount::getMountPoints(true);
        $mountPoints['/mnt/backups'] = 1;

        foreach ($mountPoints as $mountPoint => $val) {
            // mount not existing disk (for use mount in /media directory)
            $this->mountUsb('bda1', $mountPoint);
        }
    }

    public function mountUsb($name, $mountPoint, $returnValueOnAnotherDeviceIsMounted = false) {
        $mountPoint = $this->_mount->validateMountPoint($mountPoint, $returnValueOnAnotherDeviceIsMounted);

        if (!is_string($mountPoint)) {
            return $mountPoint;
        }

        if (!$this->_mount->mountUsb($name, $mountPoint)) {
            return false;
        }

        // postmount hook
        $this->postmount($mountPoint);
        // mounted succesfully
        $this->_mount->updateMountsList();

        $params = $this->_storageDevices->getPartitionParams($name);
        $string = implode('|', array($params['type'], $params['name'], $params['uuid'], $params['serial']));
        $this->putToConf($mountPoint, $string);

        return true;
    }

    public function mountShare($mountPoint, $parameters, $returnValueOnAnotherDeviceIsMounted = false) {
        if ($mountPoint == '' || $mountPoint == '/') {
            return false;
        }
        $parameters['servername'] = trim($parameters['servername'], '/');
        $parameters['serverpath'] = trim($parameters['serverpath'], '/');

        $serverName = $this->_mount->getMountedServerName($parameters);

        if (!$serverName) {
            return false;
        }

        $mountPoint = $this->_mount->validateMountPoint($mountPoint, $returnValueOnAnotherDeviceIsMounted);

        if (!is_string($mountPoint)) {
            return $mountPoint;
        }

        if (!$this->_mount->mountShare($serverName, $mountPoint, $parameters)) {
            return false;
        }

        $string = implode('|', $parameters);
        $this->putToConf($mountPoint, $string);

        $this->_mount->updateMountsList();

        if (!$this->_storageDevices->isWritablePartition($mountPoint)) {
            $this->_errors[] = _t('devices.err.DeviceIsNotWritable', $mountPoint);
            $this->unMount($mountPoint, 'smb');
            return false;
        }

        // postmount hook
        $this->postmount($mountPoint);

        return true;
    }

    public function mountDevices($devices, $partition = '', $type = 'smb', $dparams = null) {
        foreach ($devices as $mpoint => $mparams) {
            if (isset($dparams) && $dparams['uuid'] !== $mparams['uuid']) {
                continue;
            }

            if (isset($dparams)) {
                echo 'Mounting /dev/' . $partition . ' on ' . $mpoint . "\t";
            }

            if ($type === 'smb') {
                echo 'Mounting //' . $mparams['servername'] . '/' . $mparams['serverpath'] . ' on ' . $mpoint . "\t";
                $mounted = $this->mountShare($mpoint, $mparams, true);
            } else {
                $mounted = $this->mountUsb($partition, $mpoint, true);
            }

            if ($mounted) {
                echo 'OK' . "\n";
            } else {
                if ($mpoint == '/mnt/cdr' || $mpoint == '/mnt/cdrcsv') {
                    $this->unsetCdrMount($mpoint);
                }
                echo 'ERROR' . "\n";
                echo implode("\n", $this->_errors) . "\n";
                $this->_errors = array();
            }
        }
    }

    public function mountToExistingShare($mountPoint, $shareType, $sharePath, $shareFolder) {
        $share = $this->_mount->getMountInfoFromConf($shareType, $sharePath, $shareFolder);
        return $this->mountShare($mountPoint, $share);
    }

    /**
     * Mounts usb/smb devices
     *  using information in wms.conf
     *  verbose goes to console
     *
     * @return mixed
     */
    public function mountAfterReboot() {
        $devicesToMount = $this->_storageDevices->getDevicesToMount();

        if (!$devicesToMount) {
            return;
        }

        $storages = $this->_storageDevices->getStorages($this->_mount->getMountsByType());
        $this->mountAllDevices($devicesToMount, $storages['params']);
    }

    public function unMount($mountPoint, $type) {
        if ($type === 'usb' || array_key_exists($mountPoint, $this->_mount->getMountsByType())) {
            if ($type === 'usb' && $mountPoint === '/mnt/users') {
                // copy all users settings from the share
                $command = '[ -d /rw2/etc/user.d/ ] && cp /mnt/users/[0-9]*.lua /rw2/etc/user.d/;';
                exec($command, $output);
            }

            $unmounted = $this->_mount->unMount($mountPoint);

            if ($mountPoint === '/mnt/cdr' || $mountPoint === '/mnt/cdrcsv') {
                $this->unsetCdrMount($mountPoint);
            }
            if (!$unmounted) {
                return false;
            }
        }

        $this->delFromConf($mountPoint);
        $this->_mount->updateMountsList();

        return true;
    }

    public function validateShareMount($mountPoint, $params) {
        $this->_mount->updateSmbMounts();
        return $this->_mount->checkMountSmbOrMountNfs($mountPoint, $params);
    }
}
