<?php
namespace Wildix\System\Storage;

use Wildix\License;
use Wildix\System\Mount;

/**
 * Class Storage Devices
 */
class StorageDevices {

    private function humanBytes($size) {
        $filesizename = array(
                         '',
                         'K',
                         'M',
                         'G',
                         'T',
                         'P',
                         'E',
                         'Z',
                         'Y',
                        );

        $result = '0';
        if ($size) {
            $i      = floor(log($size, 1024));
            $result = round($size / pow(1024, $i)) . $filesizename[$i];
        }

        return $result;
    }

    private function getStorageInfo($storage, $partition) {
        $size      = disk_total_space('/');
        $freeSize  = disk_free_space('/');
        $usedSize  = $size - $freeSize;
        $usageSize = round(100 * $usedSize / $size);
        $fstype    = '-';
        exec('/usr/sbin/timeout 3 df -T / | grep -F "%" | head -n2 | tail -n1', $output);
        if (!empty($output[0])) {
            $output = preg_replace('/\s+/', '|', $output[0]);
            $output = explode('|', $output);

            if (!empty($output[1])) {
                $fstype = $output[1];
            }
        }

        return array(
            'type' => $storage['type'],
            'name' => $partition,
            'number' => '1',
            'serial' => $storage['serial'],
            'model' => $storage['model'],
            'vendor' => $storage['vendor'],
            'uuid' => 'unknown',
            'fstype' => (empty($fstype) || $fstype == '-') ? 'ext2' : $fstype,
            'dfsize' => $this->humanBytes($size),
            'used' => $this->humanBytes($usedSize),
            'avail' => $this->humanBytes($freeSize),
            'usage' => $usageSize . '%'
        );
    }

    private function getStorageParams($storages) {
        $params = array();

        if (!is_array($storages)) {
            return $params;
        }

        foreach ($storages as $storageName => $storage) {
            if (License::getInstance()->isCloud()) {
                reset($storages);
                $firstStorage = key($storages);
                $storage      = $storages[$firstStorage];

                if (empty($storage['partitions'])) {
                    break;
                }

                reset($storage['partitions']);
                $firstPartitions = key($storage['partitions']);
                $partition       = $storage['partitions'][$firstPartitions];

                $params[$partition] = $this->getStorageInfo($storage, $partition);
            } else {
                if (empty($storage['type']) || !isset($storage['partitions'])) {
                    continue;
                }

                foreach ($storage['partitions'] as $partition) {
                    preg_match('/(\d+)$/', $partition, $matches);
                    $device = array(
                        'type' => $storage['type'],
                        'name' => $partition,
                        'number' => $matches[1],
                        'serial' => $storage['serial'],
                        'model' => $storage['model'],
                        'vendor' => $storage['vendor']
                    );
                    exec('sudo blkid -c /dev/null -o value -s UUID -s TYPE /dev/' . $partition, $fstype);
                    $device['uuid']   = $fstype[0];
                    $device['fstype'] = $fstype[1];
                    unset($fstype);

                    $usg = exec('/usr/sbin/timeout 5 df -h 2>/dev/null | grep "^/dev/' . $partition . '" | awk \'OFS="|" {print($2,$3,$4,$5)}\'');

                    $dfparts          = explode('|', $usg);
                    $device['dfsize'] = $dfparts[0];
                    $device['used']   = $dfparts[1];
                    $device['avail']  = $dfparts[2];
                    $device['usage']  = $dfparts[3];

                    $params[$partition] = $device;
                }
            }
        }

        return $params;
    }

    private function getPartitions() {
        $partitions = file('/proc/partitions', FILE_IGNORE_NEW_LINES);
        unset($partitions[0], $partitions[1]);

        return $partitions;
    }

    private function prepareStorageList() {
        $storages   = array();
        $partitions = $this->getPartitions();

        foreach ($partitions as $partition) {
            $pArr      = explode(' ', $partition);
            $partition = array_pop($pArr);

            if (preg_match('/^[a-z]+$/', $partition)) {
                if (!isset($storages[$partition])) {
                    $storages[$partition] = array();
                }
                // hack to use Xen drivers
                if (strpos($partition, 'xvd') === 0) {
                    $storages[$partition]['vendor'] = 'XEN';
                    $storages[$partition]['model']  = $partition;
                    $storages[$partition]['serial'] = '';
                    $storages[$partition]['type']   = 'hdd';
                } else {
                    if (file_exists('/sys/block/' . $partition . '/device/vendor')) {
                        $storages[$partition]['vendor'] = trim(file_get_contents('/sys/block/' . $partition . '/device/vendor'));
                    } else {
                        $storages[$partition]['vendor'] = '';
                    }

                    if (file_exists('/sys/block/' . $partition . '/device/model')) {
                        $storages[$partition]['model'] = trim(file_get_contents('/sys/block/' . $partition . '/device/model'));
                    } else {
                        $storages[$partition]['model'] = '';
                    }
                    $storages[$partition]['serial'] = 'unknown';
                    $storages[$partition]['type']   = 'hdd';
                }
            } else {
                if (!empty($partition)) {
                    $mainPartition = preg_replace('/\d+$/', '', $partition);
                    if (!isset($storages[$mainPartition])) {
                        $storages[$mainPartition] = array();
                    }
                    if (!isset($storages[$mainPartition]['partitions'])) {
                        $storages[$mainPartition]['partitions'] = array();
                    }
                    $storages[$mainPartition]['partitions'][] = $partition;
                }
            }
        }

        return $storages;
    }

    /**
     * Get not mounted devices from 'wms.conf'
     *
     * @param mixed $type Accepts: 'usb', 'smb' or 'hdd' or false to get all.
     *
     * @return mixed array or false.
     */
    final public function getNotMountedDevicesFromConf($type = false) {
        if (!file_exists(Storage::$fileName)) {
            return false;
        }

        $config = parse_ini_file(Storage::$fileName, true, INI_SCANNER_RAW);
        if (!isset($config[Storage::$fileSection])) {
            return false;
        }

        $devices = $config[Storage::$fileSection];
        if (!is_array($devices) || empty($devices)) {
            return false;
        }

        $tMounts = file('/etc/mtab');
        if (!is_array($tMounts)) {
            $tMounts = array();
        }

        $tDevices = $devices;
        foreach ($tDevices as $mpoint => $params) {
            foreach ($tMounts as $tMountRow) {
                if (strstr($tMountRow, $mpoint)) {
                    unset($devices[$mpoint]);
                }
            }
        }

        $notMountedDevices = $this->filterDevicesByType($devices, $type);
        $return            = array_merge($notMountedDevices['usb'], $notMountedDevices['smb']);
        return !empty($return) ? $return : false;
    }

    public function getStorages($sysMounts) {
        $storages = $this->prepareStorageList();

        // remove storage which have /boot
        if (isset($sysMounts['/boot'])) {
            $systemDevice = preg_replace(array('/\d+$/', '~/dev/~'), '', $sysMounts['/boot']);
            unset($storages[$systemDevice]);
        }

        $dir = null;
        if (file_exists('/dev/disk/by-id')) {
            $dir = opendir('/dev/disk/by-id');
        }
        if ($dir) {
            // @codingStandardsIgnoreLine
            while (($file = readdir($dir)) !== false) {
                $linkVal = exec('readlink "/dev/disk/by-id/' . $file . '"');
                $tmpArr  = explode('/', $linkVal);
                $device  = array_pop($tmpArr);

                if (!isset($storages[$device])) {
                    continue;
                }

                // get possible serial
                $tmpArr    = explode('_', $file);
                $tmpSerial = array_pop($tmpArr);
                // get device type
                $tmpArr = explode('-', $file);
                $type   = $tmpArr[0];
                // usb device has personal serial parser
                // bypass old serial format for hdd type
                if ($type !== 'usb' && !preg_match('/^[0-9a-f]{1,20}$/i', $tmpSerial)) {
                    continue;
                }

                if ($type === 'usb') {
                    $storages[$device]['type'] = $type;
                    $tmpSerial                 = preg_replace('/-\d+:\d+$/', '', $tmpSerial);
                }

                $storages[$device]['serial'] = $tmpSerial;
            }

            closedir($dir);
        }

        return array(
            'usbExi' => $storages,
            'params' => $this->getStorageParams($storages)
        );
    }

    public function getPartitionMounts($partitions, $params, $mountPoints, $availableMps) {
        $partitionMounts = array();

        if (!is_array($partitions)) {
            return $partitionMounts;
        }

        foreach ($partitions as $partition) {
            $partitionIsMounted = false;

            if (isset($mountPoints[$partition])) {
                $partitionMountpoints = $mountPoints[$partition];
                foreach ($partitionMountpoints as $pmpoint) {
                    $partitionIsMounted = true;
                    $partitionMounts[]  = array(
                        'name' => $params[$partition]['number'],
                        'geometry' => _t(
                            'devices.PartitionGeometry',
                            $params[$partition]['fstype'],
                            $params[$partition]['dfsize'] . 'B',
                            $params[$partition]['avail'] . 'B',
                            $params[$partition]['usage']
                        ),
                        'geoInfo' => array(
                            'fstype' => $params[$partition]['fstype'],
                            'dfsize' => $params[$partition]['dfsize'] . 'B',
                            'avail' => $params[$partition]['avail'] . 'B',
                            'usage' => $params[$partition]['usage'],
                        ),
                        'pmpoint' => $pmpoint,
                        'partition' => $partition,
                        'mounted' => $partitionIsMounted,
                        'points' => $availableMps
                    );
                }
            }

            $partitionFsType = $params[$partition]['fstype'];
            if ($partitionFsType === 'vfat' || $partitionFsType === 'ext2' || $partitionFsType === 'ext3' || $partitionFsType === 'ext4') {
                if (empty($availableMps)) {
                    if ($partitionIsMounted) {
                        continue;
                    }
                }
                $partitionMounts[] = array(
                    'name' => $params[$partition]['number'],
                    'geometry' => _t('devices.Type', $params[$partition]['fstype']),
                    'geoInfo' => array(
                        'fstype' => $params[$partition]['fstype']
                    ),
                    'pmpoint' => '',
                    'partition' => $partition,
                    'mounted' => $partitionIsMounted,
                    'points' => $availableMps
                );
            }
        }

        return $partitionMounts;
    }

    public function getPartitionParams($partition) {
        $mainPartition = preg_replace('/\d+$/', '', $partition);

        $params           = [];
        $params['type']   = 'hdd';
        $params['name']   = $partition;
        $params['serial'] = strpos($mainPartition, 'xvd') === 0 ? '' : 'unknown';

        if (!License::getInstance()->isCloud()) {
            exec('sudo blkid -c /dev/null -o value -s UUID /dev/' . $partition, $uuid);
            $params['uuid'] = $uuid[0];

            $dir = null;
            if (file_exists('/dev/disk/by-id')) {
                $dir = opendir('/dev/disk/by-id');
            }
            if ($dir) {
                // @codingStandardsIgnoreLine
                while (($file = readdir($dir)) !== false) {
                    $linkVal = exec('readlink "/dev/disk/by-id/' . $file . '"');
                    $tmpArr  = explode('/', $linkVal);
                    $device  = array_pop($tmpArr);

                    if ($device !== $mainPartition) {
                        continue;
                    }

                    // get possible serial
                    $tmpArr    = explode('_', $file);
                    $tmpSerial = array_pop($tmpArr);
                    // get device type
                    $tmpArr = explode('-', $file);
                    $type   = $tmpArr[0];
                    // usb device has personal serial parser
                    // bypass old serial format for hdd type
                    if ($type !== 'usb' && !preg_match('/^[0-9a-f]{1,20}$/i', $tmpSerial)) {
                        continue;
                    }

                    if ($type === 'usb') {
                        $params['type'] = $type;
                        $tmpSerial      = preg_replace('/-\d+:\d+$/', '', $tmpSerial);
                    }

                    $params['serial'] = $tmpSerial;
                }
                closedir($dir);
            }
        } else {
            $params['uuid'] = 'unknown';
        }

        return $params;
    }

    public function getDeviceData($usbExi, $params, $mountPoints, $availableMps) {
        $deviceData = array();

        foreach ($usbExi as $device => $devData) {
            $deviceData[$device] = array();
            // do not allow autoinit for usb < 256 M
            $size = (int)exec('sudo sfdisk -s /dev/' . $device);

            $isAutoinitRejected   = $size < 262144;
            $notAllowedForUsbInit = $this->ifNotAllowedForUsbInit($mountPoints);

            $deviceData[$device]['name']        = _t(
                'devices.usb.SerialNumber',
                $devData['vendor'],
                $devData['model'],
                $devData['serial']
            ) . ' ' . bytes($size * 1024);
            $deviceData[$device]['initEnabled'] = !($isAutoinitRejected || isset($notAllowedForUsbInit[$device]));

            $deviceData[$device]['model']  = $devData['model'];
            $deviceData[$device]['vendor'] = $devData['vendor'];
            $deviceData[$device]['serial'] = $devData['serial'];
            $deviceData[$device]['size']   = bytes($size * 1024);

            $deviceData[$device]['partitionMounts'] = $this->getPartitionMounts($devData['partitions'], $params, $mountPoints, $availableMps);
        }

        return $deviceData;
    }

    public function getDevicesToMount() {
        $config = parse_ini_file(Storage::$fileName, true, INI_SCANNER_RAW);
        if (!isset($config[Storage::$fileSection])) {
            return false;
        }

        $devicesToMount = $config[Storage::$fileSection];
        if (!is_array($devicesToMount) || empty($devicesToMount)) {
            return false;
        }

        return $this->filterDevicesByType($devicesToMount);
    }

    public function filterDevicesByType($devices, $type = false) {
        $usbToMount = array();
        $smbToMount = array();

        foreach ($devices as $mpoint => $params) {
            $params          = explode('|', $params);
            $needAddToResult = !$type || $type === $params[0];

            if (in_array($params[0], array('usb', 'hdd')) && $needAddToResult) {
                $usbToMount[$mpoint]['type']   = $params[0];
                $usbToMount[$mpoint]['name']   = $params[1];
                $usbToMount[$mpoint]['uuid']   = $params[2];
                $usbToMount[$mpoint]['serial'] = $params[3];
            } else if ($params[0] == 'smb' && $needAddToResult) {
                $smbToMount[$mpoint]['type']       = 'smb';
                $smbToMount[$mpoint]['servername'] = $params[1];
                $smbToMount[$mpoint]['serverpath'] = $params[2];
                $smbToMount[$mpoint]['domain']     = $params[3];
                $smbToMount[$mpoint]['username']   = $params[4];
                $smbToMount[$mpoint]['password']   = $params[5];
                $smbToMount[$mpoint]['smbversion'] = $params[6] ?? Mount::DEFAULT_SMB_VERSION;
            } else if ($params[0] == 'nfs' && $needAddToResult) {
                $smbToMount[$mpoint]['type']       = 'nfs';
                $smbToMount[$mpoint]['servername'] = $params[1];
                $smbToMount[$mpoint]['serverpath'] = $params[2];
            }
        }

        return array(
            'usb' => $usbToMount,
            'smb' => $smbToMount
        );
    }

    public function ifNotAllowedForUsbInit($mountPoints) {
        $result = array();
        foreach ($mountPoints as $tDev => $value) {
            $dev          = preg_replace('/\d/', '', $tDev);
            $result[$dev] = 1;
        }

        return $result;
    }

    public function isNotAllowedDevice($device): bool {
        // Get rootfs UUID from /proc/cmdline
        $tmpUuid     = explode(' ', file_get_contents('/proc/cmdline'));
        $cmdlineUUID = explode('root=UUID=', $tmpUuid[1]);

        $rootPartition = null;
        $rootDevice    = null;
        if (isset($cmdlineUUID[1]) && file_exists('/dev/disk/by-uuid/' . $cmdlineUUID[1])) {
            // Get root device and root partition by UUID
            $linkVal = readlink('/dev/disk/by-uuid/' . $cmdlineUUID[1]);
            $tmpArr  = explode('/', $linkVal);

            $rootPartition = array_pop($tmpArr);
            $rootDevice    = preg_split('/[\d]+/', $rootPartition);
        }

        // Get fstype and blkUUID from device
        exec('sudo /sbin/blkid -c /dev/null -o value -s TYPE /dev/' . $device, $fstype);
        exec('sudo /sbin/blkid -c /dev/null -o value -s UUID /dev/' . $device, $blkUUID);

        // If device is equal to root_partition => THIS IS NOT ALLOWED: IT'S ROOT DEVICE
        if ((isset($rootDevice) && $device == $rootDevice[0]) || $device == $rootPartition) {
            return true;
            // If empty $blkUUID or $fstype => THIS IS NOT ALLOWED: NOT FOUND IN BLKID
        } else if (empty($blkUUID) || empty($fstype)) {
            return true;
            // If $fstype != "ntfs, ext2,3,4 or vfat" => THIS IS NOT ALLOWED: WRONG FS
        } else if (!($fstype[0] === 'vfat' || $fstype[0] === 'ext3' || $fstype[0] === 'ext2' || $fstype[0] === 'ext4' || $fstype[0] === 'ntfs')) {
            return true;
            // If $blkUUID == $cmdlineUUID => THIS IS NOT ALLOWED: IT'S ROOT DEVICE
        } else if (isset($cmdlineUUID[1]) && $blkUUID[0] == $cmdlineUUID[1]) {
            return true;
        }

        return false;
    }

    public function isWritablePartition($mountPoint) {
        $writable  = true;
        $tFilename = $mountPoint . '/wms_' . substr(md5(time()), 0, 8) . '.tmp';
        $fp        = fopen($tFilename, 'w');
        if (!$fp) {
            $writable = false;
        }
        fclose($fp);
        unlink($tFilename);

        return $writable;
    }

    public function initUsb($block) {
        $exec = 'nohup sudo /usr/sbin/init_usb -i /dev/' . $block . ' -s "full" > /dev/null';
        exec($exec, $output, $exitCode);

        if ($exitCode != 0) {
            return false;
        }

        exec('nohup sudo /sbin/udevadm trigger');

        return true;
    }
}
