<?php

namespace Wildix\System;

use Wildix\Exception\SystemdException;

/**
 * Class Systemd
 * https://www.freedesktop.org/software/systemd/man/systemctl.html
 */
class Systemd {

    /**
     * Daemon name.
     *
     * @var string
     */
    private $_unit;
    /**
     * Daemon properties.
     *
     * @var array
     */
    private $_properties = array();

    /**
     * Systemd constructor.
     *
     * @param string $unit Service|Timer name.
     */
    public function __construct(string $unit) {
        $this->_unit = $unit;
    }

    public function getUnitProperties(array $properties = array('LoadState', 'ActiveState', 'SubState', 'UnitFileState')) {
        $undefinedProperties = array_diff($properties, array_keys($this->_properties));

        if (count($undefinedProperties) > 0) {
            $this->_updatePropertiesCache($undefinedProperties);
        }

        return array_intersect_key($this->_properties, array_flip($properties));
    }

    private function clearPropertiesCache() {
        $this->_properties = array();
    }

    public function isEnabled() {
        $properties = $this->getUnitProperties(array('UnitFileState'));
        if ($properties && $properties['UnitFileState'] == 'enabled') {
            return true;
        }
        return false;
    }

    public function isActive() {
        $properties = $this->getUnitProperties(array('ActiveState'));
        if ($properties && $properties['ActiveState'] == 'active') {
            return true;
        }
        return false;
    }

    public function enable() {
        $output     = [];
        $resultCode = null;
        $result     = $this->exec('sudo systemctl enable ' . $this->_unit . ' 2>&1', $output, $resultCode);
        if ($resultCode !== 0) {
            $exception = new SystemdException(sprintf('Service %s has not been enabled and return error: %s', $this->_unit, $result));
            $exception->setSystemOutput($output);
            throw $exception;
        }
        $this->clearPropertiesCache();
    }

    public function start() {
        $output     = [];
        $resultCode = null;
        $result     = $this->exec('sudo systemctl start ' . $this->_unit . ' 2>&1', $output, $resultCode);
        if ($resultCode !== 0) {
            $exception = new SystemdException(sprintf('Service %s has not been started and return error: %s', $this->_unit, $result));
            $exception->setSystemOutput($output);
            throw $exception;
        }
        $this->clearPropertiesCache();
    }

    public function restart() {
        $this->exec('sudo systemctl restart ' . $this->_unit);
        $this->clearPropertiesCache();
    }

    public function reloadOrRestart() {
        $this->exec('sudo systemctl reload-or-restart ' . $this->_unit);
        $this->clearPropertiesCache();
    }

    public function stop() {
        $this->exec('sudo systemctl stop ' . $this->_unit);
        $this->clearPropertiesCache();
    }

    public function disable() {
        $this->exec('sudo systemctl disable ' . $this->_unit);
        $this->clearPropertiesCache();
    }

    private function _updatePropertiesCache(array $properties) {
        $this->exec('sudo systemctl show ' . $this->_unit . ' --property ' . implode(',', $properties), $output);

        if (!empty($output)) {
            $this->_properties += parse_ini_string(implode("\n", $output), false, INI_SCANNER_TYPED);
        }
    }

    private function exec($cmd, &$output = null, &$returnVar = null) {
        return exec($cmd, $output, $returnVar);
    }
}
