<?php

namespace Wildix\System;

class TraceService {

    const TYPE_IP     = 1;
    const TYPE_CUSTOM = 2;

    const PARAM_TRACE_NAME = 'traceName';
    const PARAM_TRACE_TYPE = 'traceType';
    const PARAM_IP_ADDRESS = 'ipAddress';
    const PARAM_COMMAND    = 'command';
    const RTP_CONFIG       = '/rw2/etc/wudpecho.conf';
    const ERROR_FILE       = '/mnt/backups/traces/trace.err';
    const TRACES_DIR       = '/mnt/backups/traces';

    // minimal free disc space in bytes
    private const MIN_FREE_DISC_SPACE = 10 * 1024 * 1024;
    private const RTP_PORT_START      = 'START';
    private const RTP_PORT_END        = 'END';

    private $_rtpPorts = [
        self::RTP_PORT_START => '10000',
        self::RTP_PORT_END => '15000',
    ];

    private $_command;
    private $_processData = null;

    public function __construct(private \Logger\Log $_logger) {
        $this->_command = exec('which tcpdump');

        $this->initRtpPorts();

        $this->createTraceDir(self::TRACES_DIR, 0777);
    }

    /**
     * Start generate process
     *
     * @param array $params Params.
     *
     * @throws \ErrorException If process couldn't run
     *
     * @return void
     */
    public function start(array $params): void {
        $this->validateParams($params);

        $name      = trim($params[self::PARAM_TRACE_NAME]);
        $type      = $params[self::PARAM_TRACE_TYPE];
        $ipAddress = isset($params[self::PARAM_IP_ADDRESS]) ? trim($params[self::PARAM_IP_ADDRESS]) : '';
        $command   = isset($params[self::PARAM_COMMAND]) ? $params[self::PARAM_COMMAND] : '';

        $this->removeErrorLog();

        $cliCommand = $this->generateTraceCommand($name, $type, $ipAddress, $command);
        $this->removeAllTraces();

        if (!$this->isHaveFreeSpace()) {
            throw new \ErrorException('No free disk space in backups share (less than 10 MB)');
        }

        $this->_logger->info(__LINE__, 'Generate trace custom command: ' . $cliCommand);

        exec($cliCommand, $output, $exitStatus);

        // 0 - Successful completion.
        // 1 - An error occurred.
        if ($exitStatus !== 0) {
            $errorMessage  = 'Failed to run the command, return code: ' . $exitStatus;
            $errorMessage .= "\nOutput: " . json_encode($output);

            $this->_logger->error(__LINE__, $errorMessage);
            throw new \ErrorException('Failed to run the command');
        }
    }

    /**
     * Stop generate process
     *
     * @param null|string $pid Process ID.
     *
     * @return void
     * @throws \ErrorException If can not stop generate trace
     */
    public function stop(string $pid = null): void {
        if (empty($pid)) {
            $processData = $this->getProcessData();

            if ($processData['processIsRunning'] === true && !empty($processData['pid'])) {
                $pid = $processData['pid'];
            }
        }

        if (!empty($pid)) {
            exec('sudo /bin/kill ' . $pid, $output, $exitStatus);

            if ($exitStatus !== 0) {
                $errorMessage  = 'Failed to stop trace, return code: ' . $exitStatus;
                $errorMessage .= "\nOutput: " . json_encode($output);

                $this->_logger->error(__LINE__, $errorMessage);
                throw new \ErrorException('Fail stop generate trace');
            }
        }

        $this->clearProcessData();
        $this->removeErrorLog();
    }

    /**
     * Remove generated trace files
     *
     * @return boolean
     */
    public function remove(): bool {
        return $this->removeAllTraces();
    }

    /**
     * Return information about generate process
     *
     * @return array
     * @throws \ErrorException If found error at run time
     */
    public function getStatus(): array {
        $status = $this->getProcessData();
        $error  = $this->getErrorData();

        if (count($error)) {
            try {
                $this->stop();
                $status['processIsRunning'] = false;
            } catch (\ErrorException $e) {
                $error[] = $e->getMessage();
            }

            $status['traceErrors'] = $error;
        }

        if (empty($status['traceName']) && $status['processIsRunning'] === false) {
            $files = glob(self::TRACES_DIR . '/*.pcap');

            if (is_array($files) && count($files) > 0) {
                $status['traceName'] = str_replace(self::TRACES_DIR . '/', '', $files[0]);
                $status['fileSize']  = filesize($files[0]);
            }
        }

        if (!empty($status['traceName'])) {
            $file = self::TRACES_DIR . '/' . $status['traceName'];
            if (file_exists($file)) {
                $status['traceSize'] = filesize($file);
            }
        }

        return $status;
    }

    /**
     * Validate params
     *
     * @param array $params Array of params.
     *
     * @return void
     * @throws \Wildix\InvalidDataException If params not valid
     */
    private function validateParams(array $params): void {
        $data = new \Wildix\Validator($params, [self::PARAM_TRACE_NAME, self::PARAM_TRACE_TYPE, self::PARAM_IP_ADDRESS, self::PARAM_COMMAND]);

        // Add required params
        $data->rule('required', [self::PARAM_TRACE_NAME, self::PARAM_TRACE_TYPE]);

        // Validate name
        $data->rule('regex', [self::PARAM_TRACE_NAME], '/^[-.\w]+$/');

        // Allow only valid types
        $data->rule('in', [self::PARAM_TRACE_TYPE], [self::TYPE_IP, self::TYPE_CUSTOM]);

        $paramType = isset($params[self::PARAM_TRACE_TYPE]) ? $params[self::PARAM_TRACE_TYPE] : null;

        switch ($paramType) {
            case self::TYPE_IP:
                if (!empty($params[self::PARAM_IP_ADDRESS])) {
                    // For ip type, param 'ipAddress' should only be valid ip address
                    $data->rule('ip', [self::PARAM_IP_ADDRESS]);
                }
                break;
            case self::TYPE_CUSTOM:
                // For custom type param 'command' is required
                $data->rule('required', [self::PARAM_COMMAND]);
                break;

            default:
                // Do nothing
                break;
        }

        if (!$data->validate()) {
            throw new \Wildix\InvalidDataException('Validation error', $data->errors());
        }
    }

    /**
     * Find running process and return information about this process
     *
     * @return array
     * @throws \ErrorException If can not stop generate trace.
     */
    private function getProcessData(): array {
        if ($this->_processData === null) {
            $this->_processData = [
                'processIsRunning' => false,
                'pid' => '',
                'traceName' => '',
                'traceParam' => '',
                'traceType' => self::TYPE_IP,
                'traceErrors' => [],
            ];

            $cmd = "ps axo pid,args | grep " . escapeshellarg($this->_command) . " | grep -v " . escapeshellarg('grep ' . $this->_command);
            exec($cmd, $output);

            $portRange = $this->getRtpPortByKey(self::RTP_PORT_START) . '-' . $this->getRtpPortByKey(self::RTP_PORT_END);

            foreach ($output as $value) {
                // 24053 /usr/sbin/tcpdump -ni any -s 0 -v (port 5060 or portrange 10000-15000) -w /mnt/backups/traces/dfdffd.pcap
                if (preg_match('#^(\d+) ' . $this->_command . ' (.*) -W 1 -C \d+ -w ' . self::TRACES_DIR . '/(.*)$#i', trim($value), $matches)) {
                    // If trace file not exists and process still running, then kill that process
                    if (!file_exists(self::TRACES_DIR . '/' . $matches[3])) {
                        if (!empty($matches[1])) {
                            $this->stop($matches[1]);
                        }
                        continue;
                    }

                    $this->_processData['processIsRunning'] = true;
                    $this->_processData['pid']              = $matches[1];
                    $this->_processData['traceName']        = $matches[3];

                    $params  = $matches[2];
                    $pattern = '#^-ni any -s 0 -v (host (.+) and )?\(port 5060 or portrange ' . $portRange . '\)#i';

                    if (preg_match($pattern, $params, $paramsMatches)) {
                        $this->_processData['traceType']  = self::TYPE_IP;
                        $this->_processData['traceParam'] = (!empty($paramsMatches[2])) ? $paramsMatches[2] : '';
                    } else {
                        $this->_processData['traceParam'] = $params;
                        $this->_processData['traceType']  = self::TYPE_CUSTOM;
                    }
                }
            }
        }

        return $this->_processData ?? [];
    }

    /**
     * Remove error log file
     *
     * @return void
     */
    private function removeErrorLog(): void {
        if (is_file(self::ERROR_FILE) && !unlink(self::ERROR_FILE)) {
            $this->_logger->error(__LINE__, 'Failed to delete trace error file: ' . self::ERROR_FILE);
        }
    }

    /**
     * Parse and set rtp ports from configuration file if exist
     *
     * @return void
     */
    private function initRtpPorts(): void {
        $conf = parse_ini_file(self::RTP_CONFIG);
        if ($conf !== false) {
            if (!empty($conf[self::RTP_PORT_START])) {
                $this->_rtpPorts[self::RTP_PORT_START] = $conf[self::RTP_PORT_START];
            }
            if (!empty($conf[self::RTP_PORT_END])) {
                $this->_rtpPorts[self::RTP_PORT_END] = $conf[self::RTP_PORT_END];
            }
        }
    }

    /**
     * Return rtp ports by key
     *
     * @param string $key Key.
     *
     * @return string
     */
    private function getRtpPortByKey(string $key): string {
        return $this->_rtpPorts[$key] ?? '';
    }

    /**
     * Add the extension if needed.
     *
     * @param string $name Trace file name.
     *
     * @return string
     */
    private function traceNameAddExtension(string $name): string {
        if (!preg_match('/^.+\.pcap$/i', $name)) {
            return $name . '.pcap';
        }
        return $name;
    }

    /**
     * Remove all trace files
     *
     * @return boolean
     * @throws \ErrorException Exception.
     */
    private function removeAllTraces(): bool {
        $traceFiles = glob(self::TRACES_DIR . '/*.pcap');

        if ($traceFiles === false) {
            $this->_logger->error(__LINE__, 'Failed to find trace files');
            return false;
        }

        foreach ($traceFiles as $_file) {
            if (!unlink($_file)) {
                $this->_logger->error(__LINE__, 'Failed to delete trace file: ' . $_file);
                throw new \ErrorException('Remove trace fail');
            }
        }

        return true;
    }

    /**
     * Parse the trace error file and return found errors
     *
     * @return array
     */
    private function getErrorData(): array {
        $error = [];
        if (file_exists(self::ERROR_FILE)) {
            $errorFileContent = file(self::ERROR_FILE);
            foreach ($errorFileContent as $_value) {
                if (!preg_match(
                    '/(^tcpdump: listening on (eth[0-9]+|any)|^tcpdump: data link type LINUX_SLL2|^tcpdump: WARNING: Promiscuous mode not supported on the \"any\" device|^Got \d)/',
                    $_value,
                    $matches
                )
                ) {
                    $error[] = $_value;
                }
            }
        }
        return $error;
    }

    /**
     * Check disk space
     *
     * @return boolean
     */
    private function isHaveFreeSpace(): bool {
        return $this->getFreeSpace() > self::MIN_FREE_DISC_SPACE;
    }

    /**
     * Return available disk space
     *
     * @return float
     */
    private function getFreeSpace(): float {
        $diskFreeSpace = disk_free_space(self::TRACES_DIR);
        return $diskFreeSpace !== false ? $diskFreeSpace : 0;
    }

    /**
     * Generate trace command
     *
     * $type == self::TYPE_IP
     * /usr/sbin/tcpdump -ni any -s 0 -v (port 5060 or portrange 10000-15000) -w /mnt/backups/traces/test.pcap
     * /usr/sbin/tcpdump -ni any -s 0 -v host 8.8.8.8 and (port 5060 or portrange 10000-15000) -w /mnt/backups/traces/test.pcap
     *
     * $type == self::TYPE_CUSTOM
     * /usr/sbin/tcpdump -v -s0 -iany port 5060 -w /mnt/backups/traces/test.pcap
     * custom tcpdump = -v -s0 -iany port 5060
     *
     * @param string  $name      Trace file name.
     * @param integer $type      Type 'ip' or 'custom'.
     * @param string  $ipAddress Ip address.
     * @param string  $command   Command.
     *
     * @return string
     *
     * @throws \ErrorException If custom tcpdump command empty.
     */
    private function generateTraceCommand(string $name, int $type, string $ipAddress = '', string $command = ''): string {
        $resultCommand = [
            'sudo',
            $this->_command,
        ];

        if ($type === self::TYPE_IP) {
            $resultCommand[] = '-ni any -s 0 -v';

            $filter = '';
            if (!empty($ipAddress)) {
                $filter = 'host ' . $ipAddress . ' and ';
            }

            $portRange       = $this->getRtpPortByKey(self::RTP_PORT_START) . '-' . $this->getRtpPortByKey(self::RTP_PORT_END);
            $filter         .= '(port 5060 or portrange ' . $portRange . ')';
            $resultCommand[] = "'" . $filter . "'";
        } else if ($type === self::TYPE_CUSTOM) {
            if (!empty($command)) {
                $resultCommand[] = $command;
            } else {
                throw new \ErrorException('Custom tcpdump command empty');
            }
        }

        $freeSpace = $this->getFreeSpace();

        if ($freeSpace <= 0) {
            throw new \ErrorException('No free disk space in backups share (less than 10 MB)');
        }

        $maxFileSize = (int)($freeSpace / 1000 / 1000) - 10;

        if ($maxFileSize > 50) {
            $maxFileSize = 50;
        }

        // trace output file
        $resultCommand[] = '-W 1 -C ' . $maxFileSize . ' -w ' . self::TRACES_DIR . '/' . $this->traceNameAddExtension($name);
        $resultCommand[] = '1>/dev/null 2>' . self::ERROR_FILE . ' &';

        return join(' ', $resultCommand);
    }

    /**
     * Clear process data
     *
     * @return void
     */
    private function clearProcessData() {
        $this->_processData = null;
    }

    /**
     * Create directory for trace files
     *
     * @param string  $path Path of the directory to be created.
     * @param integer $mode The permission to be set for the created directory.
     *
     * @return void
     * @throws \ErrorException If the directory could not be created.
     */
    private function createTraceDir(string $path, int $mode = 0775): void {
        if (is_dir($path)) {
            return;
        }

        if (!mkdir($path, $mode)) {
            $this->_logger->error(__LINE__, "Failed to create directory '" . $path . "': " . $path);
            throw new \ErrorException('Failed to create directory for trace');
        }

        if (!chmod($path, $mode)) {
            $this->_logger->error(__LINE__, "Failed to change permissions for directory '" . $path . "': " . $path);
            throw new \ErrorException('Failed to change permissions for directory');
        }
    }
}
