<?php

namespace Wildix\System\Upgrade;

/**
 * Class AptGetService
 */
class AptGetService {

    /**
     * Verbose logger class.
     *
     * @var \Wildix\System\Upgrade\VerboseLogger
     */
    private $_verboseLogger;

    public function __construct(
        \Wildix\System\Upgrade\VerboseLogger $verbose
    ) {
        $this->_verboseLogger = $verbose;
    }

    /**
     * Get verbose logger class.
     *
     * @return VerboseLogger
     */
    public function getLogger(): VerboseLogger {
        return $this->_verboseLogger;
    }

    /**
     * Apt-get update command.
     *
     * @return boolean
     */
    public function aptGetUpdate(): bool {
        $this->getLogger()->info(_t('upgrade.ExecutingCommand', _t('upgrade.updatePackagesList')));

        if (!is_dir('/var/cache/apt/archives/partial')) {
            exec('sudo mkdir -p /var/cache/apt/archives/partial');
        }
        exec('nohup sudo /usr/bin/apt-get update >> ' . $this->getLogger()->getVerboseFilePath() . ' 2>&1', $output, $return);

        if ($return !== 0) {
            $this->getLogger()->error(_t('upgrade.ErrorRisedWhileExecutingCommand', _t('upgrade.updatePackagesList')));
            return false;
        }

        return true;
    }

    /**
     * Apt-get clean command.
     *
     * @return boolean
     */
    public function aptGetClean(): bool {
        $this->getLogger()->info(_t('upgrade.ExecutingCommand', _t('upgrade.cleanUp')));
        $command = sprintf(
            'nohup sudo /usr/bin/apt-get clean >> %s 2>&1',
            $this->getLogger()->getVerboseFilePath()
        );
        exec($command, $output, $return);

        if ($return !== 0) {
            $this->getLogger()->error(_t('upgrade.ErrorRisedWhileExecutingCommand', _t('upgrade.cleanUp')));

            return false;
        }

        return true;
    }

    /**
     * Apt-get autoremove command.
     *
     * @return boolean
     */
    public function aptGetAutoremove(): bool {
        $this->getLogger()->info(_t('upgrade.ExecutingCommand', _t('upgrade.cleanUp')));
        $command = sprintf(
            'nohup sudo /usr/bin/apt-get -y autoremove >> %s 2>&1',
            $this->getLogger()->getVerboseFilePath()
        );
        exec($command, $output, $return);

        if ($return !== 0) {
            $this->getLogger()->error(_t('upgrade.ErrorRisedWhileExecutingCommand', _t('upgrade.cleanUp')));

            return false;
        }

        return true;
    }

    /**
     * Apt-get dist-upgrade command.
     *
     * @param string  $params    Dist-upgrade params.
     * @param boolean $customLog Move output to verbose file or not.
     *
     * @return mixed
     */
    public function getAptGetDistUpgrade($params = '', $customLog = false) {
        $this->getLogger()->info(_t('upgrade.CheckingNewPackages'), false);
        exec(
            sprintf(
                '%s sudo %s /usr/bin/apt-get %s dist-upgrade %s',
                $customLog ? 'nohup' : '',
                $customLog ? 'DEBIAN_FRONTEND=noninteractive' : '',
                $params . ($customLog ? ' -y --allow-unauthenticated -q' : ''),
                $customLog ? sprintf('>> %s 2>&1 ', $this->getLogger()->getVerboseFilePath()) : ''
            ),
            $output,
            $return
        );
        if ($return !== 0) {
            $this->getLogger()->error(PHP_EOL . _t('upgrade.ErrorRisedWhileExecutingCommand', _t('upgrade.upgradePackages')));

            return false;
        }
        $this->getLogger()->info(_t('upgrade.spaceOk'));

        return $output;
    }

    public function aptGetCheck(): bool {
        exec(
            sprintf(
                'nohup sudo /usr/bin/apt-get -q check 2>&1 </dev/null'
            ),
            $output,
            $return
        );
        if ($return !== 0) {
            // no error: check could be failed to run [dpkg --configure -a]
            $this->getLogger()->info(PHP_EOL . 'Packages check failed [apt-get check]: ' . array_pop($output));

            return false;
        }

        return true;
    }

    public function dpkgConfigurePackages(): bool {
        exec(
            sprintf(
                'nohup sudo /usr/bin/dpkg --force-confdef,confold,overwrite --configure -a >> %s 2>&1 </dev/null',
                $this->getLogger()->getVerboseFilePath()
            ),
            $output,
            $return
        );
        if ($return !== 0) {
            $this->getLogger()->error(PHP_EOL . 'Error during fix packages[dpkg --configure -a]: ' . array_pop($output));

            return false;
        }

        return true;
    }
}
