<?php

namespace Wildix\System\Upgrade;

use Exception;
use Logger\Log;
use Models\PBX;
use Wildix\InvalidDataException;
use Wildix\Services\Portal;
use Wildix\System\FileSystem;

/**
 * Class AutoUpgrade
 */
class AutoUpgrade extends UpgradeService {

    /**
     * Notification sender class.
     *
     * @var NotificationSender;
     */
    private $_notificationSender;

    public function __construct(
        UpgradeSettings $upgradeSettingsService,
        Portal $portal,
        PBX $pbx,
        AptGetService $aptGetService,
        FileSystem $fileSystem,
        VerboseLogger $verbose,
        NotificationSender $notificationSender,
        string $pbxDomain
    ) {
        $this->_upgradeSettingsService = $upgradeSettingsService;
        $this->_portal                 = $portal;
        $this->_pbx                    = $pbx;
        $this->_aptGetService          = $aptGetService;
        $this->_fileSystem             = $fileSystem;
        $this->_verboseLogger          = $verbose;
        $this->_notificationSender     = $notificationSender;

        $this->getLogger()->setLogLevel(Log::DEBUG);

        $this->setMode(self::MODE_UPGRADE);
        $this->_notificationSender->setPbxDomain($pbxDomain);
        $this->_notificationSender->setVerboseFile($this->getLogger()->getVerboseFilePath());
    }

    /**
     * Get notification sender class.
     *
     * @throws Exception Error found class.
     *
     * @return NotificationSender
     */
    public function getNotificationSender(): NotificationSender {
        return $this->_notificationSender;
    }

    /**
     * Check need autoupgrade.
     *
     * @param null|string $repo  Autoupgrade repository.
     * @param string      $email Notification email.
     *
     * @throws InvalidDataException Error.
     *
     * @return boolean
     */
    private function _checkNeedAutoUpgrade(?string $repo = null, string $email = ''): bool {
        if ($this->_fileSystem->checkFileSystem()) {
            $this->getNotificationSender()->send(NotificationSender::FILESYSTEM_WITH_ERRORS, $email);
            return false;
        }

        $upgradeSettings = $this->_upgradeSettingsService->getAutoUpgradeConfig();

        $upgradeSettings[UpgradeSettings::FIELD_REPOSITORY] = $repo;
        $upgradeSettings[UpgradeSettings::FIELD_ENABLED]    = true;
        $this->_upgradeSettingsService->updateSettings(
            [
                UpgradeSettings::SECTION_WMS_RELEASE => $upgradeSettings
            ]
        );

        if (!$this->checkMajorUpdateAllowed()) {
            $this->getNotificationSender()->send(NotificationSender::NOT_ALLOWED_MAJOR_UPGRADE, $email);

            $this->prepareToUpgrade();
            return false;
        }

        if (!$this->checkUpgradeIni()) {
            $this->getNotificationSender()->send(NotificationSender::NOT_VALID_REPO, $email);

            $this->prepareToUpgrade();
            return false;
        }

        $this->checkNewVersion();

        $this->checkSourcesList();
        $this->aptGetUpdate();
        $this->calculatePackagesForUpgrade();

        return $this->isAvailableUpgrade();
    }

    /**
     * Make auto upgrade.
     *
     * @param null|string $repo  Autoupgrade repository.
     * @param string      $email Notification email.
     *
     * @throws Exception Error autoupgrade.
     *
     * @return boolean
     */
    public function makeAutoUpgrade(?string $repo = null, string $email = '') {
        if (!$this->_checkNeedAutoUpgrade($repo, $email)) {
            $this->getLogger()->info(_t('upgrade.AutoUpgradeNoNewVersion'));

            $this->prepareToUpgrade();

            return true;
        }

        if (!$this->_createBackup($email)) {
            $this->getLogger()->error('Error create backup for autoupgrade');
            return false;
        }

        $isRebootRequired = false;

        $result = $this->runUpgradeScript(false);
        if (!$result) {
            $this->getLogger()->error('Not runing upgrade script');
        }

        $logFile = $this->getOutputLog();
        preg_match('/dpkg returned an error code|Error|\sE:/', $logFile, $matches);

        if (!empty($matches)) {
            $this->getNotificationSender()->send(NotificationSender::UPGRADE_FINISHED_WITH_ERROR, $email);
        } else {
            $this->getNotificationSender()->send(NotificationSender::UPGRADE_FINISHED, $email);
            $isRebootRequired = true;
        }

        // wait 2 minutes
        sleep(120);

        $this->prepareToUpgrade();

        sleep(30);

        // legacy reboot verification: check "WMS-interface successfully installed!" in logs.
        if ($isRebootRequired) {
            $this->getLogger()->info('Reboot system after autoupgrade');
            $this->_reboot();
        }

        return true;
    }

    /**
     * Create wms backup before autoupgrade.
     *
     * @param string $email Email for send backup notification.
     *
     * @throws Exception Error create backup.
     *
     * @return boolean
     */
    private function _createBackup(string $email) {
        if ($this->getPbx()->isCustomShareMounted('/mnt/backups')) {
            exec('sudo /usr/sbin/wms_backup_create with_host no_verbose', $output, $return);
            if ($return !== 0 && strpos($output[0], '10mb') === false && strpos($output[0], 'Users settings') === false) {
                $this->getNotificationSender()->send(NotificationSender::BACKUP_CREATED, $email);
                return false;
            }
        } else {
            $this->getNotificationSender()->send(NotificationSender::BACKUP_NOT_CREATED, $email);
            return false;
        }

        $this->getNotificationSender()->send(NotificationSender::BACKUP_CREATED, $email);

        return true;
    }

    /**
     * Reboot PBX
     *
     * @return void
     */
    private function _reboot(): void {
        if (system('sudo /sbin/reboot 2>&1') === false) {
            echo 'Cannot reboot.';
        }
    }
}
