<?php

namespace Wildix\System\Upgrade;

/**
 * Class Checker
 */
class Checker extends UpgradeService {

    const ACTUAL_SECONDS = 30;

    const CHECKER_SCRIPT = '/usr/sbin/upgrade_checker';

    /**
     * Wms Ldap class.
     *
     * @var \WMS_Ldap
     */
    protected $_wmsLdap;

    public function __construct(
        UpgradeSettings $upgradeSettingsService,
        \Wildix\Services\Portal $portal,
        \Models\PBX $pbx,
        \Wildix\System\Upgrade\AptGetService $aptGetService,
        \Wildix\System\FileSystem $fileSystem,
        \WMS_Ldap $wmsLdap,
        \Wildix\System\Upgrade\VerboseLogger $verbose
    ) {
        $this->_upgradeSettingsService = $upgradeSettingsService;
        $this->_portal                 = $portal;
        $this->_pbx                    = $pbx;
        $this->_aptGetService          = $aptGetService;
        $this->_fileSystem             = $fileSystem;
        $this->_wmsLdap                = $wmsLdap;
        $this->_verboseLogger          = $verbose;

        $this->getLogger()->setLogLevel(\Logger\Log::DEBUG);

        $this->setMode(self::MODE_CHECK);
    }

    /**
     * Get Wms Ldap class.
     *
     * @return \WMS_Ldap
     */
    private function _getWmsLdap(): \WMS_Ldap {
        return $this->_wmsLdap;
    }

    /**
     * Get new version of packages info.
     *
     * @return array
     */
    public function getNewPackagesInfo(): array {
        if ($this->_newPackages === null) {
            $this->calculatePackagesForUpgrade(false);
        }
        return $this->_newPackages;
    }

    /**
     * Get new version of wms-interface info.
     *
     * @throws \Exception Error.
     * @return array
     */
    public function getNewVersionInfo(): array {
        if ($this->_newVersionInfo === null) {
            $this->checkNewVersion();
        }
        return $this->_newVersionInfo;
    }

    /**
     * Run checking upgrade script.
     *
     * @return array
     *
     * @throws \Exception Error.
     */
    public function runCheckerScript(): array {
        if (!$this->searchRunningProcess(UpgradeService::PROCESS_CHEKER)['isRunning']) {
            exec(self::CHECKER_SCRIPT . ' ' . escapeshellarg($_SESSION['language']) . ' > /dev/null &', $output, $ret);
        }
        return $this->searchRunningProcess(UpgradeService::PROCESS_CHEKER);
    }

    /**
     * Get upgrade checker status.
     *
     * @throws \Exception Error.
     *
     * @return array
     */
    public function getCheckerStatus(): array {
        if ($this->_needReCheck()) {
            $this->runCheckerScript();
        }

        $checkLog         = $this->getOutputLog();
        $checkIsRunning   = $this->searchRunningProcess(UpgradeService::PROCESS_CHEKER)['isRunning'];
        $upgradeIsRunning = $this->searchRunningProcess(UpgradeService::PROCESS_UPGRADE)['isRunning'];

        return [
            'repository' => $this->getCurrentRepo(),
            'check' => [
                'isRunning' => $checkIsRunning,
                'withoutError' => !$checkIsRunning ? strpos($checkLog, 'upgrade_no_error') !== false : false,
                'fileSystemStatus' => !$this->_fileSystem->checkFileSystem(),
                'needBackupMessage' => $this->_getNeedBackupMessage(),
                'log' => $this->formatLog($checkLog),
            ],
            'upgrade' => ['isRunning' => $upgradeIsRunning],
            'newVersion' => !$checkIsRunning ? $this->_getNewVersionInfoFromLog($checkLog) : '',
            'newPackages' => !$checkIsRunning ? strpos($checkLog, 'new_packages') !== false : false,
        ];
    }


    /**
     * Get new version wms-interface from log file.
     *
     * @param null|string $log Log file data.
     *
     * @return string
     */
    private function _getNewVersionInfoFromLog(?string $log = null): string {
        if ($log === null) {
            $log = $this->getOutputLog();
        }
        preg_match('/.*new_wms_version\:([^\:]*)\:.*/', $log, $matches);

        return isset($matches[1]) ? $matches[1] : '';
    }

    /**
     * Get need backup message string.
     *
     * @return string
     */
    private function _getNeedBackupMessage(): string {
        $result = $this->_getWmsLdap()->getEntriesSortLimit(\WMS_Ldap::PERSONS, '(' . LDAP_EXTEN_NUMBER . '=*)', ['cn'], 0, 1);
        if ($result['count'] <= 1) {
            return '';
        }

        if (file_exists(self::BACKUP_FILE_REPORT)) {
            $fileTime = filemtime(self::BACKUP_FILE_REPORT) + 86400;
            if ($fileTime > time()) {
                return '';
            }

            return _t('upgrade.NeedBackup');
        } else {
            $ldapAdmin = $this->_getWmsLdap()->getEntry(\WMS_Ldap::PERSONS, '(cn=admin)', ['userPassword']);
            if ($ldapAdmin['userPassword'] == 'wildix') {
                return '';
            }

            return _t('upgrade.NeedBackupAfterUpdate');
        }
    }

    /**
     * Get "need re-check" information.
     *
     * @return boolean
     */
    private function _needReCheck(): bool {
        $path = BASEPATH . 'upgrade' . DIRECTORY_SEPARATOR . VerboseLogger::LOG_CHECK;
        if (!file_exists($path)) {
            return true;
        }

        $fileTime = filemtime($path);

        return $fileTime === false || $fileTime + self::ACTUAL_SECONDS < time();
    }
}
