<?php
declare(strict_types=1);

namespace Wildix\System\Upgrade;

use Logger\LoggerFactory;
use Wildix\Notifications\Email\Interfaces\EmailManagerInterface;
use Wildix\Settings\Env\EnvConfig;
use Wildix\System\Systemd;
use Wildix\System\Upgrade\Objects\UpgradeSettingsData;

/**
 * Class DiConfig
 */
final class DiConfig implements \Wildix\DependencyInjection\ConfigInterface {

    public static function configure(\Wildix\DependencyInjection\Container $container): void {
        $container->singleton(
            VerboseLogger::class,
            function (\Wildix\DependencyInjection\Container $container) {
                return new VerboseLogger(
                    $container->get(\Logger\Log::class)
                );
            }
        );

        $container->singleton(
            AptGetService::class,
            function (\Wildix\DependencyInjection\Container $container) {
                return new AptGetService(
                    $container->get(\Wildix\System\Upgrade\VerboseLogger::class)
                );
            }
        );

        $container->singleton(
            UpgradeSettings::class,
            function (\Wildix\DependencyInjection\Container $container) {
                return new UpgradeSettings(
                    $container->get(EnvConfig::class),
                    $container->get(Systemd::class . ':nginx'),
                    $container->get(\Logger\Log::class)
                );
            }
        );

        $container->singleton(
            NotificationSender::class,
            function (\Wildix\DependencyInjection\Container $container) {
                return new NotificationSender(
                    $container->get(EmailManagerInterface::class)
                );
            }
        );

        $container->singleton(
            Checker::class,
            function (\Wildix\DependencyInjection\Container $container) {
                return new Checker(
                    $container->get(UpgradeSettings::class),
                    $container->get(\Wildix\Services\Portal::class),
                    $container->get(\Models\PBX::class),
                    $container->get(AptGetService::class),
                    $container->get(\Wildix\System\FileSystem::class),
                    $container->get(\WMS_Ldap::class),
                    $container->get(\Wildix\System\Upgrade\VerboseLogger::class)
                );
            }
        );

        $container->singleton(
            ManuallyUpgrade::class,
            function (\Wildix\DependencyInjection\Container $container) {
                return new ManuallyUpgrade(
                    $container->get(UpgradeSettings::class),
                    $container->get(\Wildix\Services\Portal::class),
                    $container->get(\Models\PBX::class),
                    $container->get(AptGetService::class),
                    $container->get(\Wildix\System\FileSystem::class),
                    $container->get(\Wildix\System\Upgrade\VerboseLogger::class)
                );
            }
        );

        $container->singleton(
            AutoUpgrade::class,
            function (\Wildix\DependencyInjection\Container $container) {
                $pbxDomain = '';
                try {
                    $pbxDomain = $container->get(\Wildix\PBXes\PBXes::class)
                        ->getPBXBySerial($container->get(\Models\PBX::class)->getSerial())->getDomain();
                } catch (\Exception $e) {
                    LoggerFactory::getInstance()->error(__LINE__, 'Not possible obtain PBX domain name');
                }

                return new AutoUpgrade(
                    $container->get(UpgradeSettings::class),
                    $container->get(\Wildix\Services\Portal::class),
                    $container->get(\Models\PBX::class),
                    $container->get(AptGetService::class),
                    $container->get(\Wildix\System\FileSystem::class),
                    $container->get(\Wildix\System\Upgrade\VerboseLogger::class),
                    $container->get(NotificationSender::class),
                    $pbxDomain
                );
            }
        );
    }
}
