<?php

namespace Wildix\System\Upgrade;

/**
 * Class ManuallyUpgrade
 */
class ManuallyUpgrade extends UpgradeService {

    private const PROGRESS_PERCENT_DIST_UPGRADE_START  = 9;
    private const PROGRESS_PERCENT_DIST_UPGRADE_FINISH = 95;

    public function __construct(
        UpgradeSettings $upgradeSettingsService,
        \Wildix\Services\Portal $portal,
        \Models\PBX $pbx,
        \Wildix\System\Upgrade\AptGetService $aptGetService,
        \Wildix\System\FileSystem $fileSystem,
        \Wildix\System\Upgrade\VerboseLogger $verbose
    ) {
        $this->_upgradeSettingsService = $upgradeSettingsService;
        $this->_portal                 = $portal;
        $this->_pbx                    = $pbx;
        $this->_aptGetService          = $aptGetService;
        $this->_fileSystem             = $fileSystem;
        $this->_verboseLogger          = $verbose;

        $this->getLogger()->setLogLevel(\Logger\Log::DEBUG);
    }

    /**
     * Make upgrade.
     *
     * @return boolean
     * @throws \Exception Error upgrade.
     */
    public function makeUpgrade(): bool {
        if ($this->hasErrors()) {
            return false;
        }

        $this->_setProgress(0);
        $this->getLogger()->info('PBX upgrade is started', false);

        if (!$this->_checkAllowStartUpgrade()) {
            return false;
        }

        $this->aptGetUpdate();

        $this->cleaningDiskSpaceIfNeed();

        // looks like it should be in _checkAllowStartUpgrade,
        // but we need "apt update" for actual packages list
        $this->checkDiskSpaceQuota();

        if ($this->hasErrors()) {
            return false;
        }
        $this->_setProgress(6);

        $this->calculatePackagesForUpgrade();

        $this->_setProgress(7);

        if (!$this->isAvailableUpgrade()) {
            $this->getLogger()->info('PBX does not have new packages for upgrade');

            return false;
        }

        $this->_setProgress(8);

        $this->getLogger()->info(sprintf('PBX have new packages %d', count($this->_newPackages)));

        if ($this->hasErrors()) {
            return false;
        }

        $this->getLogger()->info(sprintf('Upgrade packages from repo %s', $this->getCurrentRepo()), false);

        if ($this->getAptGetService()->aptGetCheck() === false) {
            $this->getAptGetService()->dpkgConfigurePackages();
        }

        if ($this->hasErrors()) {
            return false;
        }

        $this->_monitInitOperation('stop');

        $this->_setProgress(self::PROGRESS_PERCENT_DIST_UPGRADE_START);

        $this->getAptGetService()->getAptGetDistUpgrade('', true);

        $this->_setProgress(self::PROGRESS_PERCENT_DIST_UPGRADE_FINISH);

        // clean downloaded packages
        $this->getAptGetService()->aptGetClean();
        $this->_setProgress(96);
        if (!$this->hasErrors()) {
            $this->getAptGetService()->aptGetAutoremove();
            $this->_setProgress(97);
        } else {
            $this->getLogger()->error('PBX upgrade was failed.');
            $this->_monitInitOperation('restart');

            return false;
        }

        $this->_monitInitOperation('restart');
        $this->_setProgress(98);

        return true;
    }

    /**
     * Check allow start upgrade.
     *
     * @return boolean
     * @throws \Exception Error read config.
     */
    private function _checkAllowStartUpgrade(): bool {
        $this->_setProgress(1);

        $this->checkUpgradeIni();

        if ($this->hasErrors()) {
            return false;
        }
        $this->_setProgress(2);

        $this->checkSourcesList();

        if ($this->hasErrors()) {
            return false;
        }
        $this->_setProgress(3);

        $this->checkAllowingUpgradePermissions();

        if ($this->hasErrors()) {
            return false;
        }
        $this->_setProgress(4);

        if (!$this->_backupVarLibDpkg()) {
            return false;
        }
        $this->_setProgress(5);

        return true;
    }

    /**
     * Make backup before upgrade.
     *
     * @return boolean
     */
    private function _backupVarLibDpkg(): bool {
        $command = 'tar cPphzf /rw2/var/backup/dpkg.tgz /var/lib/dpkg 2>&1 && cp /etc/apt/sources.list /rw2/var/backup/sources.list 2>&1 && sudo sync 2>&1 ';
        exec($command, $output, $return);
        if ($return !== 0) {
            $this->getLogger()->error(implode(PHP_EOL, $output));
            $this->getLogger()->error(_t('upgrade.ErrorRisedWhileExecutingCommand', _t('upgrade.backUpDpkg')) . PHP_EOL);

            return false;
        }

        return true;
    }

    /**
     * Get upgrade status.
     *
     * @throws \Exception Error.
     *
     * @return array
     */
    public function getUpgradeStatus(): array {
        $upgradeLog       = $this->getOutputLog();
        $upgradeIsRunning = $this->searchRunningProcess(self::PROCESS_UPGRADE)['isRunning'];

        return [
            'isRunning' => $upgradeIsRunning,
            'completed' => strpos($upgradeLog, self::UPGRADE_COMPLETED) !== false,
            'progress' => $this->_getProgress(),
            'log' => $this->formatLog($upgradeLog),
        ];
    }

    /**
     * Run monit operation.
     *
     * @param string $mode Monit mode.
     *
     * @return boolean
     */
    private function _monitInitOperation(string $mode): bool {
        $command = sprintf('/etc/init.d/monit %s >> %s 2>&1', $mode, $this->getLogger()->getVerboseFilePath());
        exec($command, $output, $return);
        if ($return === 2) {
            $this->getLogger()->error(_t('upgrade.ErrorRisedWhileExecutingCommand', 'monitoring service ' . $mode) . PHP_EOL);

            return false;
        }

        return true;
    }

    /**
     * Get upgrade progress value.
     *
     * @return integer
     */
    private function _getProgress(): int {
        $progress = 0;
        if (\Services\Helpers::isReadableFile(self::PROGRESS_FILE)) {
            $progress = (int)(file_get_contents(self::PROGRESS_FILE));
        }

        // need dynamically count progress based on log file wmsupgrade.out
        if ($progress >= self::PROGRESS_PERCENT_DIST_UPGRADE_START
            && $progress < self::PROGRESS_PERCENT_DIST_UPGRADE_FINISH
            && $this->_getPackagesUpgradeProgressFromLog() !== false
        ) {
            $progress += $this->_getPackagesUpgradeProgressFromLog();
        }

        return $progress;
    }

    /**
     * Try fetch UPGRADE packages PART of progress from log.
     *
     * @return boolean|integer
     */
    private function _getPackagesUpgradeProgressFromLog() {
        $upgradeLog = $this->getOutputLog();
        // _newPackages don't keep state!
        $countPackages = count($this->_newPackages ?? []);
        if (($countPackages === 0)
            && preg_match_all('/^(\d+) upgraded, (\d+) newly installed/m', $upgradeLog, $packagesMatch, PREG_SET_ORDER)
        ) {
            $upgradePackagesMatch = array_shift($packagesMatch);
            $countPackages        = ($upgradePackagesMatch[1] ?? 0) + ($upgradePackagesMatch[2] ?? 0);
        }

        if ($countPackages === 0) {
            return false;
        }

        $countUploaded = (int)preg_match_all('/^Get:/m', $upgradeLog);
        $countUnpacked = (int)preg_match_all('/^Unpacking/m', $upgradeLog);
        $countSetted   = (int)preg_match_all('/^Setting up/m', $upgradeLog);

        // why 29 = (96 - 9) / 3
        $percentUploaded = $countUploaded > $countPackages ? 29 : (int)(($countUploaded / $countPackages) * 29);
        $percentUnpacked = $countUnpacked > $countPackages ? 29 : (int)(($countUnpacked / $countPackages) * 29);
        $percentSet      = $countSetted > $countPackages ? 29 : (int)(($countSetted / $countPackages) * 29);

        return $percentUploaded + $percentUnpacked + $percentSet;
    }

    /**
     * Set upgrade progress value.
     *
     * @param float   $percent  Progress value.
     * @param boolean $addValue Add this value to old value.
     *
     * @return void
     */
    private function _setProgress(float $percent = 0, $addValue = false): void {
        $value = $addValue ? $this->_progress + $percent : $percent;

        $this->_progress = $value;
        try {
            $res = fopen(self::PROGRESS_FILE, 'w');
            fwrite($res, (int)$value);
            fclose($res);
        } catch (\Exception $e) {
            $this->getLogger()->info($e->getMessage());
        }
    }
}
