<?php

namespace Wildix\System\Upgrade;

use Wildix\Notifications\Email\Interfaces\EmailManagerInterface;

/**
 * Class NotificationSender
 */
class NotificationSender {

    const NO_NEW_VERSION              = 'noNewVersion';
    const UPGRADE_FINISHED            = 'upgradeFinished';
    const UPGRADE_FINISHED_WITH_ERROR = 'upgradeFinishedWithError';
    const BACKUP_CREATED              = 'backupCreated';
    const BACKUP_NOT_CREATED          = 'backupNotCreated';
    const NOT_VALID_REPO              = 'notValidRepo';
    const FILESYSTEM_WITH_ERRORS      = 'fsWithErrors';
    const NOT_ALLOWED_MAJOR_UPGRADE   = 'notAllowedMajorUpgrade';

    /**
     * Sendmail class.
     *
     * @var EmailManagerInterface
     */
    private $_gateway;

    private $_verboseFile;

    private $_pbxDomain = '';

    public function __construct(
        EmailManagerInterface $gateway
    ) {
        $this->_gateway = $gateway;
    }

    /**
     * Set PBX domain name.
     *
     * @param string $pbxDomain PBX domain name.
     *
     * @return void
     */
    public function setPbxDomain(string $pbxDomain): void {
        $this->_pbxDomain = $pbxDomain;
    }

    /**
     * Set verbose file path.
     *
     * @param string $verboseFile Verbose file path.
     *
     * @return void
     */
    public function setVerboseFile(string $verboseFile): void {
        $this->_verboseFile = $verboseFile;
    }

    /**
     * Get verbose file path.
     *
     * @return string
     */
    private function _getVerboseFile(): string {
        return $this->_verboseFile;
    }

    /**
     * Send notification.
     *
     * @param string $reason Reason key.
     * @param string $email  Email address.
     *
     * @return void
     */
    public function send(string $reason, string $email): void {
        if ($reason && $email) {
            $this->_gateway->send($this->_prepare($reason, $email));
        }
    }

    /**
     * Prepare template.
     *
     * @param string $reason Reason key.
     * @param string $email  Email address.
     *
     * @return array
     */
    private function _prepare(string $reason, string $email) {
        $attachFile = null;
        switch ($reason) {
            case self::NO_NEW_VERSION:
                $body  = _t('upgrade.AutoUpgradeNoNewVersion');
                $body .= PHP_EOL . PHP_EOL . _t('upgrade.AutoUpgradeDate');
                break;
            case self::UPGRADE_FINISHED:
                $body  = _t('upgrade.AutoUpgradeFinished');
                $body .= PHP_EOL . PHP_EOL . _t('upgrade.AutoUpgradeDate');

                $attachFile = $this->_getVerboseFile();
                break;
            case self::UPGRADE_FINISHED_WITH_ERROR:
                $body  = _t('upgrade.AutoUpgradeFinishedWithError');
                $body .= PHP_EOL . PHP_EOL . _t('upgrade.AutoUpgradeDate');

                $attachFile = $this->_getVerboseFile();
                break;
            case self::BACKUP_CREATED:
                $body  = _t('upgrade.AutoUpgradeBackupCreated');
                $body .= PHP_EOL . PHP_EOL . _t('upgrade.AutoUpgradeStarted');
                $body .= PHP_EOL . PHP_EOL . _t('upgrade.AutoUpgradeDate');

                $attachFile = trim(file_get_contents(UpgradeService::BACKUP_FILE_REPORT));
                break;
            case self::BACKUP_NOT_CREATED:
                $body  = _t('upgrade.AutoUpgradeBackupNotCreated');
                $body .= PHP_EOL . PHP_EOL . _t('upgrade.AutoUpgradeDate');
                break;
            case self::NOT_VALID_REPO:
                $body  = _t('upgrade.AutoUpgradeNotValidRepo');
                $body .= PHP_EOL . PHP_EOL . _t('upgrade.AutoUpgradeDate');
                break;
            case self::NOT_ALLOWED_MAJOR_UPGRADE:
                $body  = _t('upgrade.MajorUpgradeError');
                $body .= PHP_EOL . PHP_EOL . _t('upgrade.AutoUpgradeDate');
                break;
            case self::FILESYSTEM_WITH_ERRORS:
                $body  = _t('upgrade.AutoUpgradeFsWithErrorsMail');
                $body .= PHP_EOL . PHP_EOL . _t('upgrade.AutoUpgradeDate');

                $logFile    = \Wildix\System\FileSystem::getInstance()->getLogFile();
                $attachFile = file_exists($logFile) ? $logFile : '';
                break;
            default:
                $body  = _t('upgrade.AutoUpgradeReport');
                $body .= PHP_EOL . PHP_EOL . _t('upgrade.AutoUpgradeDate');
                break;
        }

        return [
            'type' => 'upgrade',
            'subject' => _t('upgrade.AutoUpgradeReport') . ': ' . $this->_pbxDomain,
            'to' => $email,
            'body' => $body . ' ' . date('Y-m-d H:i:s'),
            'template' => '',
            'attach' => $attachFile
        ];
    }
}
