<?php
namespace Wildix\System\Upgrade;

use Wildix\Validator;

/**
 * Class UpgradeService
 */
abstract class UpgradeService {

    const MODE_CHECK   = 'check';
    const MODE_UPGRADE = 'upgrade';

    const PROCESS_UPGRADE = 'wms_upgrade';
    const PROCESS_CHEKER  = 'upgrade_checker';

    const UPGRADE_SCRIPT = '/usr/sbin/wms_upgrade';
    const PROGRESS_FILE  = '/tmp/upgrade_progress.log';
    const SOURCES_LIST   = '/etc/apt/sources.list';

    const LOCAL_REPOSITORY = 'LOCAL REPOSITORY';
    const PACKAGE_NAME     = 'wms-interface-sqlite3';

    const UPGRADE_COMPLETED     = 'PBX upgrade was completed';
    const UPGRADE_NOT_COMPLETED = 'PBX upgrade not completed';

    const BACKUP_FILE_REPORT = '/dev/shm/.backup';

    const ATTEMPTS_CLEAN_DISK_SPACE = 5;
    const AMOUNT_CLEAN_DISK_SPACE   = '1.3G';

    /**
     * Verbose logger.
     *
     * @var \Wildix\System\Upgrade\VerboseLogger
     */
    protected $_verboseLogger;

    /**
     * Upgrade settings class.
     *
     * @var UpgradeSettings
     */
    protected $_upgradeSettingsService;

    /**
     * Models PBX class.
     *
     * @var \Models\PBX
     */
    protected $_pbx;

    /**
     * Apt-get service class.
     *
     * @var \Wildix\System\Upgrade\AptGetService
     */
    protected $_aptGetService;

    /**
     * Portal Services class.
     *
     * @var \Wildix\Services\Portal
     */
    protected $_portal;

    /**
     * Filesystem class.
     *
     * @var \Wildix\System\FileSystem
     */
    protected $_fileSystem;

    protected $_mode = self::MODE_UPGRADE;

    protected $_serverConfig;
    protected $_newVersionInfo;
    protected $_newPackages;

    protected $_progress = 0;

    /**
     * Set upgrade or check mode, default upgrade.
     *
     * @param string $mode Mode name.
     *
     * @return void
     */
    public function setMode(string $mode): void {
        $this->_mode = $mode;
        $this->getLogger()->setVerboseFileName($mode === self::MODE_CHECK ? VerboseLogger::LOG_CHECK : VerboseLogger::LOG_UPGRADE);
    }

    /**
     * Get Models PBX class.
     *
     * @return \Models\PBX Pbx class.
     */
    protected function getPbx() {
        return $this->_pbx;
    }

    /**
     * Get apt-get service class.
     *
     * @return \Wildix\System\Upgrade\AptGetService
     */
    protected function getAptGetService(): \Wildix\System\Upgrade\AptGetService {
        return $this->_aptGetService;
    }

    /**
     * Get verbose logger.
     *
     * @return VerboseLogger
     */
    public function getLogger(): VerboseLogger {
        return $this->_verboseLogger;
    }

    /**
     * Get server config.
     *
     * @throws \Exception Error read config.
     * @return null|array
     */
    public function getServerConfig(): ?array {
        if ($this->_serverConfig === null) {
            $this->checkUpgradeIni();
        }

        return $this->_serverConfig;
    }

    /**
     * Get upgrade settings.
     *
     * @return array
     * @throws \Exception Error read config.
     */
    protected function getSettings(): array {
        return $this->_upgradeSettingsService->getSettings();
    }

    /**
     * Get current repository name.
     *
     * @return string
     *
     * @throws \Exception Error.
     */
    public function getCurrentRepo(): string {
        return $this->_upgradeSettingsService->getCurrentRepo();
    }

    /**
     * Get current log data.
     *
     * @return string
     */
    public function getOutputLog(): string {
        $log = $this->getLogger()->getVerboseLog();
        return ($log !== false) ? $log : '';
    }

    public function hasErrors() {
        return !empty($this->getLogger()->getErrors());
    }


    /**
     * Prepare PBX to upgrade.
     * Check configs, new version wms-interface, permissions.
     *
     * @return boolean
     */
    public function prepareToUpgrade(): bool {
        if (!$this->hasErrors()) {
            $this->getLogger()->clearVerboseFile();
            $this->getLogger()->info(_t('strings.BeginningProcedure'), false);
        }

        $checkActions  = [
            'checkMajorUpdateAllowed',
            'checkUpgradeIni',
            'checkNewVersion',
            'checkSourcesList',
            'checkAllowingUpgradePermissions',
        ];
        $updateActions = [
            'aptGetUpdate',
            'cleaningDiskSpaceIfNeed',
            'checkDiskSpaceQuota',
            'calculatePackagesForUpgrade',
        ];
        foreach (array_merge($checkActions, $updateActions) as $action) {
            if ($this->hasErrors()) {
                return false;
            }
            // skip check additional packages if found new wms-interface version
            if ($action === 'calculatePackagesForUpgrade' && $this->_newVersionInfo['canUpgrade'] === true) {
                continue;
            }

            $this->$action();
        }

        return !$this->hasErrors();
    }

    /**
     * Check PBX source list.
     *
     * @throws \Exception Error check.
     *
     * @return boolean
     */
    protected function checkSourcesList(): bool {
        $this->getLogger()->info(_t('upgrade.CheckingFilename', _t('upgrade.repositoryList')), false);

        $repo = $this->getCurrentRepo();
        if (\Services\Helpers::isIp($repo)) {
            exec('sudo /usr/bin/custom-apt "' . $repo . '"');
        } else {
            exec(sprintf('sudo /usr/bin/jail-apt "%s" "%s"', $this->_serverConfig['getpath'], $this->_serverConfig['gbranch']));

            if (!\Services\Helpers::isReadableFile(self::SOURCES_LIST)) {
                $this->getLogger()->error(_t('upgrade.FilenameDoesNotExistOrIsNotReadable', self::SOURCES_LIST));
                return false;
            }
        }

        $sourcesList = array();
        foreach (file(self::SOURCES_LIST) as $row) {
            if (strpos($row, '#') !== false) {
                $row = substr($row, 0, strpos($row, '#'));
            }
            $row = trim($row);
            if ($row !== '') {
                $sourcesList[] = $row;
            }
        }
        $sourcesCount = count($sourcesList);

        $messageWildixRepoOnly = _t('upgrade.OnlyWmsRepositoryShouldBeEnabled', '/etc/apt/sources.list') . ': ' . PHP_EOL . '
                            <b>' . $this->_serverConfig['serverinfo'] . '</b>';

        if ($sourcesCount !== 1) {
            if ($sourcesCount === 0) {
                $this->getLogger()->error(_t('upgrade.spaceErrorNoEnabledRepositoryFound'));
            } else {
                $this->getLogger()->error(_t('upgrade.spaceError'));
            }
            $this->getLogger()->error($messageWildixRepoOnly);

            return false;
        } else if (\Services\Helpers::isIp($repo)) {
            $this->getLogger()->info(_t('upgrade.CheckingIfUpgradeIsAllowedForTheGateway'));

            return true;
        } else if (!preg_match($this->_serverConfig['server'], $sourcesList[0])) {
            $this->getLogger()->error($messageWildixRepoOnly);

            return false;
        }
        $this->getLogger()->info(_t('upgrade.spaceOk'));

        return true;
    }

    /**
     * Check to allow upgrade on the portal.
     *
     * @return boolean
     */
    protected function checkAllowingUpgradePermissions(): bool {
        $this->getLogger()->info(_t('upgrade.CheckingIfUpgradeIsAllowedForTheGateway'), false);

        $serial = strtolower($this->getPbx()->getSerial());
        if (empty($serial)) {
            $this->getLogger()->error('Cannot get gateway ID');
            return false;
        }

        if ($this->_serverConfig['gserv'] !== '5.10.0.1' && $this->_serverConfig['gserv'] !== '172.16.0.8') {
            $host = env('SERVICE_UPGRADE_PORTAL_URI');
        } else {
            $host = 'https://' . $this->_serverConfig['gserv'];
        }

        $url        = $host . $this->_serverConfig['guri'];
        $deniedInfo = $this->_portal->checkUpgradeIsDenied($url, rawurlencode($serial));

        if ($deniedInfo['errorCode'] !== 0) {
            $this->getLogger()->error($deniedInfo['error']);
            return false;
        }

        $this->getLogger()->info(_t('upgrade.spaceOk'));

        return true;
    }

    /**
     * Clean disk space to the required minimum for upgrading.
     *
     * @return void
     */
    protected function cleaningDiskSpaceIfNeed(): void {
        $requiredSpace = $this->getRequiredSpaceForUpgrade();

        $i = 0;
        while ($this->_fileSystem->getFreeDiskSpace() <= $requiredSpace && $i < self::ATTEMPTS_CLEAN_DISK_SPACE) {
            exec('sudo /usr/sbin/cleaning_space.sh ' . self::AMOUNT_CLEAN_DISK_SPACE);
            $i++;
        }
    }

    private function getRequiredSpaceForUpgrade(): float {
        return $this->_getSpaceRequiredForDownload() * 3;
    }

    /**
     * Should be run after apt update.
     *
     * @return boolean Success on not.
     */
    protected function checkDiskSpaceQuota(): bool {
        $this->getLogger()->info(_t('upgrade.CheckingFilename', 'free disk quota'), false);

        $freeDiskSpace = $this->_fileSystem->getFreeDiskSpace();
        $requiredSpace = $this->getRequiredSpaceForUpgrade();
        if ($freeDiskSpace <= $requiredSpace) {
            $this->getLogger()->info(PHP_EOL, false);

            $this->getLogger()->info('Free disk space: ' . $this->_humanReadableBytes($freeDiskSpace));
            $this->getLogger()->info('Required for upgrade at least: ' . $this->_humanReadableBytes($requiredSpace));

            $this->getLogger()->error('PBX has no space for upgrade.');
            $this->getLogger()->info('Try to cleanup it manually.');

            return false;
        }

        $this->getLogger()->info(_t('upgrade.spaceOk'));

        return true;
    }

    private function _getSpaceRequiredForDownload(): float {
        // fire error due to an abort of operation => return false instead of output
        // $output = $this->getAptGetService()->getAptGetDistUpgrade('--assume-no');
        exec(
            'sudo /usr/bin/apt-get --assume-no dist-upgrade',
            $output,
            $return
        );

        $requiredSpaceForDownload = 0;
        if (!is_array($output)) {
            return $requiredSpaceForDownload;
        }
        foreach ($output as $line) {
            // Need to get 123.4 Mb of archives.
            // Need to get 123 MB of archives.
            // Need to get 0 B/4774 kB of archives.
            // Need to get 1 MB/4774 kB of archives.
            if (preg_match('/^Need to get (?:(\d+(?:\.\d+)?)\s(\w+)\/)?(\d+.\d+)\s(\w+) of archives\./', $line, $matches)) {
                switch (count($matches)) {
                    case 3:
                        $requiredSpaceForDownload = $this->_convertToBytes($matches[1], $matches[2]);
                        break;
                    case 5:
                        $requiredSpaceForDownload = $this->_convertToBytes($matches[3], $matches[4]);
                        break;
                    default:
                        $requiredSpaceForDownload = 0;
                        break;
                }
            }
        }

        return $requiredSpaceForDownload;
    }

    private function _convertToBytes(float $number, string $unit): float {
        $siPrefix = ['B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'];
        $exponent = array_flip($siPrefix)[$unit] ?? null;
        if ($exponent === null) {
            return (float)0;
        }

        return $number * (1024 ** $exponent);
    }

    private function _humanReadableBytes(float $bytes): string {
        $siPrefix = ['B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'];
        $base     = 1024;
        $class    = min((int)log($bytes, $base), count($siPrefix) - 1);

        return sprintf('%1.2f', $bytes / ($base ** $class)) . ' ' . $siPrefix[$class];
    }

    /**
     * Run apt-get update.
     *
     * @return boolean
     */
    protected function aptGetUpdate(): bool {
        return $this->_aptGetService->aptGetUpdate();
    }

    /**
     * Calculate and sorting packages for upgrade.
     *
     * @param boolean $logging Write packages info to log file.
     *
     * @return boolean
     */
    protected function calculatePackagesForUpgrade($logging = true): bool {
        $this->_newPackages = [];

        $output = $this->getAptGetService()->getAptGetDistUpgrade('-us');
        if ($output === false) {
            return false;
        }

        $packages = [];
        foreach ($output as $line) {
            if (preg_match('/^Inst\s(.+)\s\[(.+)\]\s\(([^,\s]+)\s(.+)\).*$/', $line, $matches)) {
                $packages[] = [
                    'packname'       => $matches[1],
                    'current_verson' => $matches[2],
                    'avail_version'  => $matches[3],
                    'from_server'    => $matches[4],
                ];
            }
        }

        if (!empty($packages) && $logging) {
            $log       = _t('upgrade.PackagesReadyForUpgrade') . ': ';
            $tmpLogArr = array();
            foreach ($packages as $item) {
                $tmpLogArr[] = '<b>' . $item['packname'] . '</b>';
            }
            $log .= implode(', ', $tmpLogArr);
            $this->getLogger()->info($log);
        }

        $this->_newPackages = $packages;

        return true;
    }

    /**
     * Search running process by name.
     *
     * @param string $name Process name.
     *
     * @return array
     */
    protected function searchRunningProcess(string $name): array {
        exec(sprintf('pgrep %s', $name), $out, $return);
        return [
            'isRunning' => ($return === 0),
            'pid' => array_map('intval', $out),
        ];
    }

    /**
     * Convert log file data to array.
     *
     * @param string $log       Log data.
     * @param string $separator Separator.
     *
     * @return array
     */
    protected function formatLog(string $log, string $separator = PHP_EOL): array {
        return explode($separator, $log);
    }

    /**
     * Run upgrade script in background.
     *
     * @param boolean $inBackgroundMode No background mode.
     *
     * @return boolean
     */
    public function runUpgradeScript(bool $inBackgroundMode = true): bool {
        if ($this->searchRunningProcess(self::PROCESS_UPGRADE)['isRunning']
            || $this->searchRunningProcess(self::PROCESS_CHEKER)['isRunning']
        ) {
            $this->getLogger()->info('Not running autoupgrade script');

            return false;
        }

        $this->getLogger()->clearVerboseFile();
        $this->getLogger()->info('Run upgrade script');

        $cmd = sprintf(
            'nohup sudo %s 1>/dev/null 2>/dev/null%s',
            self::UPGRADE_SCRIPT,
            (($inBackgroundMode === true) ? ' &' : '')
        );
        exec($cmd);

        return true;
    }

    /**
     * Check and fix upgrade ini file.
     *
     * @throws \Exception Error check.
     *
     * @return boolean
     */
    protected function checkUpgradeIni(): bool {
        $repo = $this->getCurrentRepo();

        if (\Services\Helpers::isIp($repo)) {
            $this->getLogger()->info(_t('upgrade.spaceOk'));
            return true;
        }

        $settings     = $this->getSettings();
        $tunnelConfig = $settings[UpgradeSettings::SECTION_TUNNEL];

        $this->getLogger()->info(_t('upgrade.spaceOk'));
        $this->getLogger()->info(_t('upgrade.CheckingServerInformation'), false);

        $releases = $this->_getReleasesList($tunnelConfig['serverpath']);
        if (empty($releases)) {
            return false;
        }

        if (!isset($releases['wmsreleases'])
            || !isset($releases['wmsupgrades'])
            || !isset($releases['wmsupgrades'][$this->_upgradeSettingsService->getCurrentRelease()])
        ) {
            $this->getLogger()->error(_t('upgrade.ReleaseXIsUnknownForUpgrade', $repo));
            return false;
        }

        $possibleRepos = explode(',', $releases['wmsupgrades'][$this->_upgradeSettingsService->getCurrentRelease()]);

        if (!in_array($repo, $possibleRepos) || !isset($releases['wmsreleases'][$repo])) {
            $this->getLogger()->error(_t('upgrade.ItsNotPossibleToUpgradeUsingX', $repo));
            return false;
        }

        $releaseSettings = explode('|', $releases['wmsreleases'][$repo]);
        foreach ($releaseSettings as $item) {
            $tmp = explode('=', $item);
            $settings[$tmp[0]]['release'] = $tmp[1];
        }

        $tunnelConfig = $settings[UpgradeSettings::SECTION_TUNNEL];
        if (!isset($releases[$tunnelConfig['release']])) {
            $this->getLogger()->error(_t('upgrade.UnkownPackageTypeIsSetInFile', 'env.ini'));
            return false;
        }
        $this->getLogger()->info(_t('upgrade.spaceOk'));

        $this->_serverConfig = array_merge($releases[$tunnelConfig['release']], ['upgradeRepository' => $repo]);

        return true;
    }

    public function checkMajorUpdateAllowed(): bool {
        $from = $this->_getReleaseVersion($this->_upgradeSettingsService->getCurrentRelease());
        $to   = $this->_getReleaseVersion($this->getCurrentRepo());

        if ($from === $to) {
            return true;
        }

        $this->getLogger()->info(_t('upgrade.spaceOk'));
        $this->getLogger()->info(_t('upgrade.checkMajorVersions', $from, $to));

        exec(
            'callweaver -x\'is_upgrade_allowed\'',
            $output,
            $return
        );

        $isAllowed = Validator::isTrue($output[0] ?? false);
        if (!$isAllowed) {
            $this->getLogger()->error(_t('upgrade.MajorUpgradeError'));
        }

        return $isAllowed;
    }

    private function _getReleaseVersion(string $version): string {
        preg_match('/^rel(\d)+/', $version, $match);

        return strtolower($match[0] ?? '');
    }

    /**
     * Get releases list from server.
     *
     * @param string $path Path of server URL.
     *
     * @return array
     */
    private function _getReleasesList(string $path): array {
        $this->_portal->setPath($path);
        $result = $this->_portal->getReleasesList();

        if (empty($result)) {
            $this->getLogger()->error(_t('upgrade.CannotGetVersionInformationFromServer', env('SERVICE_UPGRADE_PORTAL_URI') . '/' . $path));
            return [];
        }

        return $result;
    }

    /**
     * Compare remote and local version wms-interface.
     *
     * @throws \Exception Error.
     *
     * @return array
     */
    protected function checkNewVersion(): array {
        $repo = $this->getCurrentRepo();
        if (\Services\Helpers::isIp($repo)) {
            $this->_newVersionInfo = [
                'canUpgrade' => true,
                'currentVersion' => WMS_VERSION_FULL,
                'serverVersion' => self::LOCAL_REPOSITORY,
            ];
        } else {
            $releaseConfig = $this->getServerConfig();
            $serverVersion = isset($releaseConfig['version']) ? trim($releaseConfig['version']) : '';

            $this->_newVersionInfo = [
                'canUpgrade' => version_compare(strtolower($serverVersion), strtolower(WMS_VERSION_FULL), '>'),
                'currentVersion' => WMS_VERSION_FULL,
                'serverVersion' => $serverVersion,
            ];
        }

        return $this->_newVersionInfo;
    }

    /**
     * Calc available new wms-interface or additional packages.
     *
     * No calc! just read the property that set elsewhere !
     *
     * @return boolean
     */
    protected function isAvailableUpgrade() {
        return (!empty($this->_newVersionInfo) && $this->_newVersionInfo['canUpgrade'] === true) || !empty($this->_newPackages);
    }
}
