<?php

namespace Wildix\System\Upgrade;

use Logger\Log;
use Services\Helpers;
use Wildix\Exception;
use Wildix\Settings\Env\EnvConfig;
use Wildix\System\Systemd;
use Wildix\Validator;

/**
 * Class UpgradeSettings
 */
class UpgradeSettings {

    const CURRENT_RELEASE              = 'rel60';
    const CUSTOM_FIRMWARE_NGINX_CONFIG = '/rw2/etc/nginx/app/customFirmwareServer.conf';

    const SECTION_WMS_RELEASE     = 'wmsrelease';
    const SECTION_TUNNEL          = 'tunnel';
    const SECTION_CUSTOM_FIRMWARE = 'customFirmwareServer';
    const SECTION_AUTO_UPGRADE    = 'autoUpgrade';

    const FIELD_REPOSITORY = 'repository';
    const FIELD_ENABLED    = 'enabled';
    const FIELD_INTERVAL   = 'howly';

    const FIELD_UPGRADE_REPOSITORY    = 'UPGRADE_REPOSITORY';
    const FIELD_AUTO_UPGRADE_ENABLED  = 'AUTO_UPGRADE_ENABLED';
    const FIELD_AUTO_UPGRADE_EMAIL    = 'AUTO_UPGRADE_EMAIL';
    const FIELD_AUTO_UPGRADE_INTERVAL = 'AUTO_UPGRADE_INTERVAL';

    const FIELD_CUSTOM_FIRMWARE_SERVER_ENABLED   = 'CUSTOM_FIRMWARE_SERVER_ENABLED';
    const FIELD_CUSTOM_FIRMWARE_SERVER_URL       = 'CUSTOM_FIRMWARE_SERVER_URL';
    const FIELD_CUSTOM_FIRMWARE_SERVER_CHECK_TLS = 'CUSTOM_FIRMWARE_SERVER_CHECK_TLS';

    const FIELD_TUNNEL_SERVERPATH = 'TUNNEL_SERVERPATH';
    const FIELD_TUNNEL_RELEASE    = 'TUNNEL_RELEASE';

    const LOG_AUTO_UPGRADE_MARKER = 'needLogAutoUpgradeChanges';

    const INTERVALS = [
        'daily',
        'weekly',
        'monthly',
    ];

    private $_errors = [];

    private $_envConfig;
    private $_nginx;
    private $_log;

    public function __construct(
        EnvConfig $envConfig,
        Systemd $nginx,
        Log $logger
    ) {
        $this->_envConfig = $envConfig;
        $this->_nginx     = $nginx;
        $this->_log       = $logger;

        $this->getLogger()->setLogLevel(Log::DEBUG);
    }

    public function getCurrentRelease(): string {
        return self::CURRENT_RELEASE;
    }

    public function getSettings(): array {
        return [
            self::SECTION_WMS_RELEASE => [
                self::FIELD_REPOSITORY => $this->getEnvConfig()->get(self::FIELD_UPGRADE_REPOSITORY),
                self::FIELD_ENABLED => $this->getEnvConfig()->get(self::FIELD_AUTO_UPGRADE_ENABLED),
                'email' => $this->getEnvConfig()->get(self::FIELD_AUTO_UPGRADE_EMAIL),
                'howly' => $this->getEnvConfig()->get(self::FIELD_AUTO_UPGRADE_INTERVAL),
            ],
            self::SECTION_TUNNEL => [
                'serverpath' => $this->getEnvConfig()->get(self::FIELD_TUNNEL_SERVERPATH),
                'release' => $this->getEnvConfig()->get(self::FIELD_TUNNEL_RELEASE),
            ],
            self::SECTION_CUSTOM_FIRMWARE => [
                self::FIELD_ENABLED  => $this->getEnvConfig()->get(self::FIELD_CUSTOM_FIRMWARE_SERVER_ENABLED),
                'url'      => $this->getEnvConfig()->get(self::FIELD_CUSTOM_FIRMWARE_SERVER_URL),
                'checkTls' => $this->getEnvConfig()->get(self::FIELD_CUSTOM_FIRMWARE_SERVER_CHECK_TLS),
            ]
        ];
    }

    public function updateSettings(array $data): bool {
        if ($this->_isEmptySavedData($data) || !$this->_validateSettings($data)) {
            throw new \Wildix\InvalidDataException(\Wildix\Error::getMessage(\Wildix\Error::BadRequest), $this->_errors);
        }

        if ($this->_isNeedLoggingAutoUpgradeChanges($data)) {
            $this->getLogger()->info(
                __LINE__,
                sprintf(
                    '[Auto-Upgrade]. Settings for auto-upgrade changed to %s, repo entered %s, recurrency %s',
                    ($data[self::SECTION_WMS_RELEASE][self::FIELD_ENABLED] ? 'enabled' : 'disabled'),
                    $data[self::SECTION_WMS_RELEASE][self::FIELD_REPOSITORY],
                    $data[self::SECTION_WMS_RELEASE][self::FIELD_INTERVAL]
                )
            );
        }

        $this->_writeUpgradeConfigFile($data);

        if (!empty($data[self::SECTION_CUSTOM_FIRMWARE])) {
            $this->_writeCustomNginxConfig($data[self::SECTION_CUSTOM_FIRMWARE]);
        }

        return true;
    }

    /**
     * Check saved data is empty.
     *
     * @param array $data Saved data.
     *
     * @return boolean
     */
    private function _isEmptySavedData(array $data): bool {
        return (empty($data)
            || (empty($data[self::SECTION_WMS_RELEASE])
                && empty($data[self::SECTION_CUSTOM_FIRMWARE])
            )
        );
    }

    /**
     * Get custom firmware settings.
     *
     * @throws \Exception Error read file.
     *
     * @return array
     */
    public function getCustomFirmwareConfig(): array {
        return [
            self::FIELD_ENABLED  => Validator::isTrue($this->getEnvConfig()->get(self::FIELD_CUSTOM_FIRMWARE_SERVER_ENABLED)),
            'url'      => (string)$this->getEnvConfig()->get(self::FIELD_CUSTOM_FIRMWARE_SERVER_URL),
            'checkTls' => Validator::isTrue($this->getEnvConfig()->get(self::FIELD_CUSTOM_FIRMWARE_SERVER_CHECK_TLS))
        ];
    }

    public function getCurrentRepo(): string {
        return $this->getEnvConfig()->get(self::FIELD_UPGRADE_REPOSITORY);
    }

    /**
     * Write settings to upgrade config file.
     *
     * @param array $data Settings data.
     *
     * @throws \Exception Save error.
     *
     * @return void
     */
    private function _writeUpgradeConfigFile(array $data) {
        if (!empty($data[self::SECTION_CUSTOM_FIRMWARE])) {
            $this->getEnvConfig()->set(
                self::FIELD_CUSTOM_FIRMWARE_SERVER_ENABLED,
                (bool)$data[self::SECTION_CUSTOM_FIRMWARE][self::FIELD_ENABLED] ?? $this->getEnvConfig()->get(self::FIELD_CUSTOM_FIRMWARE_SERVER_ENABLED)
            );
            $this->getEnvConfig()->set(
                self::FIELD_CUSTOM_FIRMWARE_SERVER_CHECK_TLS,
                (bool)$data[self::SECTION_CUSTOM_FIRMWARE]['checkTls'] ?? $this->getEnvConfig()->get(self::FIELD_CUSTOM_FIRMWARE_SERVER_CHECK_TLS)
            );
            $this->getEnvConfig()->set(
                self::FIELD_CUSTOM_FIRMWARE_SERVER_URL,
                (string)$data[self::SECTION_CUSTOM_FIRMWARE]['url'] ?? $this->getEnvConfig()->get(self::FIELD_CUSTOM_FIRMWARE_SERVER_URL)
            );
        }

        if (!empty($data[self::SECTION_WMS_RELEASE])) {
            $this->getEnvConfig()->set(
                self::FIELD_UPGRADE_REPOSITORY,
                $data[self::SECTION_WMS_RELEASE][self::FIELD_REPOSITORY] ?? $this->getCurrentRepo()
            );

            $this->getEnvConfig()->set(
                self::FIELD_AUTO_UPGRADE_ENABLED,
                (bool)$data[self::SECTION_WMS_RELEASE][self::FIELD_ENABLED] ?? $this->getEnvConfig()->get(self::FIELD_AUTO_UPGRADE_ENABLED)
            );

            if ($data[self::SECTION_WMS_RELEASE][self::FIELD_ENABLED]) {
                $this->getEnvConfig()->set(
                    self::FIELD_AUTO_UPGRADE_EMAIL,
                    (string)$data[self::SECTION_WMS_RELEASE]['email'] ?? $this->getEnvConfig()->get(self::FIELD_AUTO_UPGRADE_EMAIL)
                );

                $this->getEnvConfig()->set(
                    self::FIELD_AUTO_UPGRADE_INTERVAL,
                    (string)$data[self::SECTION_WMS_RELEASE][self::FIELD_INTERVAL] ?? $this->getEnvConfig()->get(self::FIELD_AUTO_UPGRADE_INTERVAL)
                );
            }
        }

        if ($this->getEnvConfig()->save()) {
            $this->getLogger()->info(__LINE__, sprintf('Save config file %s', $this->getEnvConfig()->getPath()));
        } else {
            $this->getLogger()->error(__LINE__, sprintf('Impossible write config %s', $this->getEnvConfig()->getPath()));
            throw new Exception('Impossible write config', 500);
        }
    }

    /**
     * Validate upgrade settings.
     *
     * @param array $data Settings data.
     *
     * @return boolean
     */
    private function _validateSettings(array $data) {
        if (!empty($data[self::SECTION_WMS_RELEASE])) {
            $form = new Validator($data[self::SECTION_WMS_RELEASE], ['repository', 'enabled', 'email', 'howly']);
            $form->rules(
                [
                    'upgradeRepository' => [
                        [self::FIELD_REPOSITORY],
                    ],
                    'required' => [
                        [
                            self::FIELD_REPOSITORY,
                            self::FIELD_ENABLED,
                        ],
                    ],
                    'boolean' => [
                        [self::FIELD_ENABLED],
                    ],
                    'in' => [
                        [
                            'howly',
                            self::INTERVALS,
                        ],
                    ],
                ]
            );

            if ($data[self::SECTION_WMS_RELEASE][self::FIELD_ENABLED]) {
                $form->rules(
                    [
                        'required' => ['howly'],
                        'email' => [
                            ['email'],
                        ],
                    ]
                );
            }

            if (!$form->validate()) {
                $this->_errors[self::SECTION_WMS_RELEASE] = $form->errors();
            }
        }

        if (!empty($data[self::SECTION_CUSTOM_FIRMWARE])) {
            $form = new Validator($data[self::SECTION_CUSTOM_FIRMWARE], [self::FIELD_ENABLED, 'url', 'checkTls']);
            $form->rules(
                [
                    'boolean' => [
                        [self::FIELD_ENABLED],
                        ['checkTls'],
                    ],
                    'url' => [
                        ['url'],
                    ],
                    'required' => [
                        [self::FIELD_ENABLED],
                        ['checkTls'],
                    ],
                ]
            );

            if ($data[self::SECTION_CUSTOM_FIRMWARE][self::FIELD_ENABLED]) {
                $form->rules(
                    [
                        'required' => [
                            ['url'],
                        ],
                    ]
                );
            }

            if (!$form->validate()) {
                $this->_errors[self::SECTION_CUSTOM_FIRMWARE] = $form->errors();
            }
        }

        return empty($this->_errors);
    }

    /**
     * Get autoupgrade settings.
     *
     * @throws \Exception Not readable error.
     *
     * @return array
     */
    public function getAutoUpgradeConfig(): array {
        return [
            self::FIELD_ENABLED => Validator::isTrue($this->getEnvConfig()->get(self::FIELD_AUTO_UPGRADE_ENABLED)),
            'email' => $this->getEnvConfig()->get(self::FIELD_AUTO_UPGRADE_EMAIL),
            'howly' => $this->getEnvConfig()->get(self::FIELD_AUTO_UPGRADE_INTERVAL),
        ];
    }

    /**
     * Update autoupgrade settings.
     *
     * @param array $data Saved data.
     *
     * @return array
     *
     * @throws Exception            Save Error.
     * @throws \Wildix\InvalidDataException Validation error.
     */
    public function updateAutoUpgradeConfig(array $data): array {
        $data[self::FIELD_REPOSITORY] = $this->getCurrentRepo();

        $this->updateSettings([self::SECTION_WMS_RELEASE => $data]);
        return $this->getAutoUpgradeConfig();
    }

    /**
     * Write Nginx config for custom firmware server.
     *
     * @param array $settings Custom Firmware Settings array.
     *
     * @throws \Exception OperationNotComplete.
     * @return void;
     */
    private function _writeCustomNginxConfig(array $settings) {
        if (!Helpers::isWritableFile(self::CUSTOM_FIRMWARE_NGINX_CONFIG)) {
            $this->getLogger()->error(__LINE__, sprintf('Not writable file %s', self::CUSTOM_FIRMWARE_NGINX_CONFIG));
            throw new Exception('Impossible write config', 500);
        }

        $customFirmwareServer = env('SERVICE_FIRMWARE_URI');
        $checkTls             = 'on';

        if ($settings[self::FIELD_ENABLED]) {
            $customFirmwareServer = trim($settings['url']);

            if (!$settings['checkTls']) {
                $checkTls = 'off';
            }
        }

        file_put_contents(
            self::CUSTOM_FIRMWARE_NGINX_CONFIG,
            $this->_generateCustomNginxConfig($customFirmwareServer, $checkTls)
        );
        $this->getLogger()->info(__LINE__, sprintf('Save config file %s', self::CUSTOM_FIRMWARE_NGINX_CONFIG));

        $this->_nginx->reloadOrRestart();
        $this->getLogger()->info(__LINE__, 'Nginx reload after save custom firmware config');
    }

    /**
     * Generate Nginx config for custom firmware server.
     *
     * @param string $firmwareServer Firmware server URL.
     * @param string $checkTls       Check TLS.
     *
     * @return string
     */
    private function _generateCustomNginxConfig(string $firmwareServer, string $checkTls = 'on'): string {
        if (!$firmwareServer) {
            $firmwareServer = env('SERVICE_FIRMWARE_URI');
        }

        return 'set $fwm_upstream ' . $firmwareServer . ';
        location ^~ /app/android {
            resolver 127.0.0.1;

            proxy_pass $fwm_upstream;
            proxy_set_header Host $host;

            proxy_ssl_trusted_certificate /usr/lib/ssl/certs/ca-certificates.crt;
            proxy_ssl_verify ' . $checkTls . ';
        }

        location /remoteupgrade/ {
            resolver 127.0.0.1;

            proxy_pass $fwm_upstream;
            proxy_set_header Host $host;

            proxy_ssl_trusted_certificate /usr/lib/ssl/certs/ca-certificates.crt;
            proxy_ssl_verify ' . $checkTls . ';
            proxy_connect_timeout 3s;
        }
        ';
    }

    /**
     * Check is new auto upgrade config has difference from current
     *
     * @param array $newConfig Upgrade config data.
     *
     * @return boolean
     *
     * @throws \Exception Error get current config file data
     */
    private function _isNeedLoggingAutoUpgradeChanges(array $newConfig): bool {
        unset($newConfig[self::SECTION_WMS_RELEASE][self::FIELD_REPOSITORY]);

        return isset($newConfig[self::LOG_AUTO_UPGRADE_MARKER])
            && $newConfig[self::LOG_AUTO_UPGRADE_MARKER] === true
            && isset($newConfig[self::SECTION_WMS_RELEASE])
            && $this->getAutoUpgradeConfig() !== $newConfig[self::SECTION_WMS_RELEASE];
    }

    /**
     * Build data for updating settings from request params data
     *
     * @param array $data Request params data.
     *
     * @return array
     */
    public function buildDataFromRequestParams(array $data) : array {
        $data[self::LOG_AUTO_UPGRADE_MARKER] = true;
        return $data;
    }

    private function getEnvConfig(): EnvConfig {
        return $this->_envConfig;
    }

    private function getLogger(): Log {
        return $this->_log;
    }
}
