<?php

namespace Wildix\System\Upgrade;

/**
 * Class VerboseLogger
 */
class VerboseLogger {

    const MAX_COUNT_LOG_FILES = 30;

    const LOG_UPGRADE = 'wmsupgrade.out';
    const LOG_CHECK   = 'check_upgrade.out';

    /**
     * Logger class.
     *
     * @var \Logger\Log
     */
    private $_logger;

    private $_errors = [];

    private $_verboseFile = self::LOG_UPGRADE;

    public function __construct(
        \Logger\Log $logger
    ) {
        $this->_logger = $logger;
    }

    /**
     * Setter fo logLevel parameter.
     *
     * @param integer $logLevel Log level parameter.
     *
     * @return void
     */
    public function setLogLevel($logLevel): void {
        $this->_logger->setLogLevel($logLevel);
    }

    /**
     * Add error row.
     *
     * @param string $error Error row.
     *
     * @return void
     */
    private function _saveError(string $error): void {
        $this->_errors[] = $error;
    }

    /**
     * Get all errors.
     *
     * @return array
     */
    public function getErrors(): array {
        return $this->_errors;
    }

    /**
     * Set verbose file name.
     *
     * @param string $name Filename.
     *
     * @return void
     */
    public function setVerboseFileName(string $name): void {
        $this->_verboseFile = $name;
    }

    /**
     * Get current verbose filename.
     *
     * @return string File name.
     */
    public function getVerboseFilePath(): string {
        return BASEPATH . 'upgrade' . DIRECTORY_SEPARATOR . $this->_verboseFile;
    }

    /**
     * Add to log INFO content.
     *
     * @param string  $content         Info content.
     * @param boolean $appendLineBreak Add line break in the verbose file after this row.
     *
     * @return void
     */
    public function info(string $content, bool $appendLineBreak = true): void {
        $this->flushLog($content, false, $appendLineBreak);
    }

    /**
     * Add to log ERROR content.
     *
     * @param string  $content         Error content.
     * @param boolean $appendLineBreak Add line break in the verbose file after this row.
     *
     * @return void
     */
    public function error(string $content, bool $appendLineBreak = true): void {
        $this->flushLog($content, true, $appendLineBreak);
    }

    /**
     * Flush log into disc.
     *
     * @param string  $content         Logged data.
     * @param boolean $isError         Is error flag.
     * @param boolean $appendLineBreak Add line break in the verbose file after this row.
     *
     * @return boolean
     */
    private function flushLog(string $content, bool $isError = false, bool $appendLineBreak = true): bool {
        if ($isError) {
            $content = _t('upgrade.spaceError') . ' ' . $content;

            $this->_logger->error(__LINE__, $content);
            $this->_saveError($content);
        } else {
            $this->_logger->info(__LINE__, $content);
        }

        $res = fopen($this->getVerboseFilePath(), 'a');
        if ($res === false) {
            $fileError = _t('upgrade.spaceError') . ' ' . _t('upgrade.CannotOpenFileForWriting', $this->getVerboseFilePath());
            $this->_saveError($fileError);
            $this->_logger->error(__LINE__, $fileError);
            return false;
        }

        if ($appendLineBreak) {
            $content .= PHP_EOL;
        }

        if (fwrite($res, $content) === false) {
            $fileError = _t('upgrade.spaceError') . ' ' . _t('upgrade.CannotPutValidContentToFile', $this->getVerboseFilePath());
            $this->_saveError($fileError);
            $this->_logger->error(__LINE__, $fileError);
            return false;
        }

        if (fclose($res) === false) {
            $fileError = _t('upgrade.spaceError') . ' ' . _t('upgrade.CannotCloseHandleOfFile', $this->getVerboseFilePath());
            $this->_saveError($fileError);
            $this->_logger->error(__LINE__, $fileError);

            return false;
        }

        return true;
    }

    /**
     * Clear current verbose file.
     *
     * @return boolean
     */
    public function clearVerboseFile(): bool {
        $fileName = $this->getVerboseFilePath();
        if (file_exists($fileName) && !is_writeable($fileName)) {
            unlink($fileName);
        }

        $fp = fopen($fileName, 'w');
        if ($fp === false) {
            $fileError = _t('upgrade.spaceError') . ' ' . _t('upgrade.CannotOpenFileForWriting', $this->getVerboseFilePath());
            $this->_saveError($fileError);
            $this->_logger->error(__LINE__, $fileError);

            return false;
        }
        if (fclose($fp) === false) {
            $fileError = _t('upgrade.spaceError') . ' ' . _t('upgrade.CannotCloseHandleOfFile', $this->getVerboseFilePath());
            $this->_saveError($fileError);
            $this->_logger->error(__LINE__, $fileError);

            return false;
        }

        return true;
    }

    /**
     * Backup verbose file.
     *
     * @return boolean
     */
    public function backupVerboseFile(): bool {
        $fileName = $this->getVerboseFilePath();
        if (!file_exists($fileName) || file_get_contents($fileName) === PHP_EOL) {
            return true;
        }

        $copyOfVerboseFile = $fileName . '_log_' . date('Y.m.d-H:i');
        if (!copy($fileName, $copyOfVerboseFile)) {
            $fileError = _t('upgrade.spaceError') . ' ' . _t('strings.canNotRename', $fileName, $copyOfVerboseFile);
            $this->_saveError($fileError);
            $this->_logger->error(__LINE__, $fileError);

            return false;
        }

        $filesArray = array();
        foreach (glob($fileName . '_log_*') as $fName) {
            $filesArray[filemtime($fName)] = $fName;
        }

        $filesCount = count($filesArray);
        if ($filesCount >= self::MAX_COUNT_LOG_FILES) {
            ksort($filesArray);
            $i = $filesCount - self::MAX_COUNT_LOG_FILES;
            foreach ($filesArray as $key => $val) {
                if (!$i) {
                    break;
                }
                unlink($filesArray[$key]);
                $i--;
            }
        }

        return true;
    }

    /**
     * Get verbose file content.
     *
     * @return boolean|string
     */
    public function getVerboseLog() {
        if (\Services\Helpers::isReadableFile($this->getVerboseFilePath())) {
            return file_get_contents($this->getVerboseFilePath());
        }

        return false;
    }
}
