<?php

namespace Wildix\TaskQueue;

/**
 * Class QueueManager
 */
class QueueManager {
    const QUEUE_MANAGER_PATH = '/usr/sbin/queue_manager';
    const RUNNER_PATH        = BASEPATH . 'api/Console/Runner.php';

    /**
     * Queue manager execution time limit (seconds)
     *
     * Default: 10 minutes
     */
    const TIME_EXECUTION_LIMIT = 60 * 10;

    /**
     * Logger.
     *
     * @var \Logger\Log
     */
    private $_logger;

    /**
     * Queue storage.
     *
     * @var \Wildix\Databases\QueueDB
     */
    private $_storage;

    /**
     * Queue items array.
     *
     * @var array
     */
    private $_queue = [];

    /**
     * Queue manager start time (unix timestamp)
     *
     * @var integer
     */
    private $_startTime;

    /**
     * QueueManager constructor.
     *
     * @param \Wildix\Databases\QueueDB $storage Queue storage.
     * @param \Logger\Log               $logger  Logger.
     */
    public function __construct(\Wildix\Databases\QueueDB $storage, \Logger\Log $logger) {
        $this->_startTime = time();
        $this->_logger    = $logger;
        $this->_storage   = $storage;
        $this->getQueue();
        $this->processQueue();
    }

    /**
     * Run queue manager.
     *
     * @return void
     */
    public static function runQueueManager() : void {
        exec(self::QUEUE_MANAGER_PATH . ' > /dev/null 2>/dev/null &');
    }

    /**
     * Fill queue items array.
     *
     * @return void
     */
    private function getQueue() : void {
        $this->_queue = $this->_storage->getItems();
    }

    /**
     * Process queue items.
     *
     * @return void
     */
    private function processQueue() : void {
        foreach ($this->_queue as $queueItem) {
            if ($this->isExecutionTimeLimitReached()) {
                return;
            }
            $this->processQueueItem($queueItem);
        }
        if (!$this->isExecutionTimeLimitReached() && $this->checkNewTasks()) {
            $this->processQueue();
        }
    }

    /**
     * Check is new tasks available.
     *
     * @return boolean
     */
    private function checkNewTasks() : bool {
        $this->getQueue();
        return count($this->_queue) > 0;
    }

    /**
     * Precessing queue item
     *
     * @param array $queueItem Queue item.
     *
     * @return void
     */
    private function processQueueItem(array $queueItem) : void {
        try {
            $item   = new QueueItem($queueItem);
            $params = $item->getParams();

            $params[QueueItem::PARAM_CREATED]     = $item->getCreatedAt();
            $params[QueueItem::PARAM_RETRY_COUNT] = $item->getRetryCount();

            $workerParams = implode(
                ' ',
                array_map(
                    function ($key, $value) {
                        return escapeshellarg($key . '=' . $value);
                    },
                    array_keys($params),
                    $params
                )
            );

            $command = \sprintf(
                'php %s %s %s',
                self::RUNNER_PATH,
                escapeshellarg($item->getType()),
                $workerParams
            );

            exec(
                $command,
                $out,
                $res
            );

            if ($res === 0) {
                $this->_storage->delJobByField(['id' => $item->getId()]);
                $this->_logger->info(__LINE__, 'Queue manager: Processing complete. ' . $item->getType() . ' ' . json_encode($item->getParams()));
            } else {
                $this->_logger->error(__LINE__, 'Queue manager: [ ERROR ] Processing failed. ' . json_encode($out));
                $this->_storage->updateRetryCountJob($item->getRetryCount() + 1, $item->getId());
                $this->_storage->updateCreatedAtToCurrentTimeJob($item->getId());
            }
            unset($item);
            unset($params);
            unset($workerParams);
            unset($command);
        } catch (\Exception $e) {
            $this->_logger->error(__LINE__, 'Queue manager: [ ERROR ] Processing failed. ' . $e->getMessage());
        }
    }

    /**
     * Check is execution time no longer then limit
     *
     * @return boolean
     */
    private function isExecutionTimeLimitReached() : bool {
        if (time() >= $this->_startTime + self::TIME_EXECUTION_LIMIT) {
            $this->_logger->info(__LINE__, 'Queue manager: Execution time limit reached');
            return true;
        }
        return false;
    }
}
