<?php

namespace Wildix\Trunks\Groups;

use Wildix\Error;
use Wildix\Exception;
use Wildix\InvalidDataException;
use Wildix\Validator;

/**
 * Class Service
 */
class Service {

    private $_storage;

    public function __construct(Storage $storage) {
        $this->_storage = $storage;
    }

    /**
     * Trunk groups storage.
     *
     * @return Storage
     */
    private function _storage(): Storage {
        return $this->_storage;
    }

    /**
     * Get trunk group items
     *
     * @return array
     *
     * @throws \Exception Database error
     */
    public function getItems(): array {
        $items = $this->_storage()->getItems();
        return $this->_storage()->loadRelations($items);
    }

    public function getItem(int $id, bool $check = false): ?Group {
        $item = $this->_storage()->getItem($id);

        if (empty($item) && $check) {
            throw new Exception(Error::NotFound, 404);
        }

        return $item;
    }

    /**
     * Create trunk group with rules
     *
     * @param array $params Params array.
     *
     * @throws InvalidDataException Invalid data
     *
     * @return void
     */
    public function create(array $params): void {
        $this->validate($params);
        $this->_storage()->create($params);
    }

    /**
     * Update trunk group with rules
     *
     * @param integer $id   Trunk group id.
     * @param array   $data Trunk group data.
     *
     * @throws InvalidDataException Invalid data
     *
     * @return void
     */
    public function update(int $id, array $data): void {
        $this->getItem($id, true);
        $this->validate($data, $id);
        $this->_storage()->update($id, $data);
    }

    /**
     * Remove trunk group with rules by id
     *
     * @param integer $id Trunk group id.
     *
     * @return void
     *
     * @throws InvalidDataException Database error
     */
    public function remove(int $id): void {
        $item   = $this->getItem($id, true);
        $linked = $this->_storage()->getLinked($id);

        if (count($linked)) {
            throw new InvalidDataException(sprintf('You can’t remove %s. It’s linked to %s ', $item->getName(), implode(', ', $linked)), []);
        }

        $this->_storage()->delete($id);
    }

    /**
     * Build item
     *
     * @param Group $item Trunk group item.
     *
     * @return array
     *
     * @throws \Exception Database error
     */
    public function buildItem(Group $item) : array {
        return [
            Group::FIELD_ID => $item->getId(),
            Group::FIELD_NAME => $item->getName(),
            Group::FIELD_DESCRIPTION => $item->getDescription(),
            Group::FIELD_RULES => $item->getRules()
        ];
    }

    /**
     * Validate trunk group with rules
     *
     * @param array   $data Trunk group rules array.
     * @param integer $id   Trunk group id if updating. Default = null.
     *
     * @throws InvalidDataException Invalid data
     *
     * @return void
     */
    protected function validate(array $data, int $id = null): void {
        $errors    = [];
        $validator = new Validator($data);
        $validator->rule('regex', Group::FIELD_NAME, '/^[\w]+$/i')->reason('Invalid trunk group name');
        $validator->rule('regex', Group::FIELD_DESCRIPTION, '/^[a-zA-Zàèéêìíîòóôùú0-9!#\$\^&\*\(\)_\-\+=`\{\}\|\"\?\/\. ]+$/')->reason('Invalid trunk group description');
        $validator->rule('required', Group::FIELD_NAME);
        $items = $this->_getList();
        if ($id) {
            unset($items[$id]);
        }

        if (in_array((string)$data['name'], $items)) {
            $errors['name'] = [
                'code' => 'AlreadyPresent',
                'reason' => 'Already present'
            ];
        }

        if (isset($data['rules']) && is_array($data['rules'])) {
            $availableTrunksNames = $this->_storage()->getAvailableTrunkNames();

            foreach ($data['rules'] as $key => $item) {
                $validator->rule(
                    'min',
                    [
                        sprintf('rules.%s.%s', $key, Rule::FIELD_REMOVE),
                        sprintf('rules.%s.%s', $key, Rule::FIELD_MAX_CALLS),
                        sprintf('rules.%s.%s', $key, Rule::FIELD_TIMEOUT),
                    ],
                    0
                )->message(Error::MinLimit);

                $validator->rule(
                    'unsignedInteger',
                    [
                        sprintf('rules.%s.%s', $key, Rule::FIELD_REMOVE),
                        sprintf('rules.%s.%s', $key, Rule::FIELD_MAX_CALLS),
                        sprintf('rules.%s.%s', $key, Rule::FIELD_TIMEOUT),
                    ]
                )->message(Error::NotUnsignedInteger);

                $validator->rule('In', sprintf('rules.%s.%s', $key, Rule::FIELD_TRUNK_NAME), $availableTrunksNames)->reason('Unknown trunk');
                $validator->rule('required', sprintf('rules.%s.%s', $key, Rule::FIELD_TRUNK_NAME));
                $validator->rule(
                    'regex',
                    sprintf('rules.%s.%s', $key, Rule::FIELD_PREPEND),
                    '/^[\+]?[0-9\*\#]*$/'
                )->reason('Only +, *, #, 0-9 characters are allowed in the Prepend field.');
            }
        }

        if (!$validator->validate()) {
            $errors = array_merge($errors, $validator->errorsTree());
        }

        if (!empty($errors)) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $errors);
        }
    }

    private function _getList() {
        $res = [];

        foreach ($this->getItems() as $item) {
            $res[$item->getId()] = $item->getName();
        }

        return $res;
    }
}
