<?php

namespace Wildix\Trunks\Groups;

use Wildix\Databases\WMSDB;
use Wildix\InvalidDataException;
use Wildix\Trunks\PstnTrunks\Trunk as PstnTrunk;
use Wildix\Trunks\SipTrunks\Trunk as SipTrunk;

/**
 * Class Storage
 */
class Storage {

    private const GROUPS_MAIN_TABLE  = 'outgoing_groups_main';
    private const GROUPS_RULES_TABLE = 'outgoing_groups_rules';

    private const DIALPLAN_LINK = 1;
    private const LINK_GROUP    = 5;

    /**
     * WMS database
     *
     * @var WMSDB
     */
    private $_db;

    /**
     * Storage constructor.
     *
     * @param WMSDB $_db WMSDB.
     */
    public function __construct(WMSDB $_db) {
        $this->_db = $_db;
    }

    /**
     * WMS database getter
     *
     * @return WMSDB
     */
    private function _db(): WMSDB {
        return $this->_db;
    }

    /**
     * Get trunk groups
     *
     * @return array
     *
     * @throws \Exception Database error
     */
    public function getItems() {
        return $this->_db()->getRows('SELECT * FROM ' . self::GROUPS_MAIN_TABLE, null, Group::class);
    }

    public function getItem(int $id): ?Group {
        $item = $this->_db()->first('SELECT * FROM ' . self::GROUPS_MAIN_TABLE . ' WHERE id = ' . $id);

        if ($item) {
            $item = new Group($item);
        }

        return $item;
    }

    /**
     * Get trunk group rules
     *
     * @param integer $groupId Trunk group id.
     *
     * @return array
     *
     * @throws \Exception Database error
     */
    public function getRules(int $groupId) {
        return $this->_db()->getRows(
            sprintf(
                'SELECT * FROM `%s` LEFT JOIN `%s` ON (`%s`.`%s`=`%s`.`%s`) WHERE `%s`.`%s`=%d ORDER BY `%s`.`%s`;',
                self::GROUPS_RULES_TABLE,
                self::GROUPS_MAIN_TABLE,
                self::GROUPS_MAIN_TABLE,
                Group::ID,
                self::GROUPS_RULES_TABLE,
                Rule::GROUP_ID,
                self::GROUPS_MAIN_TABLE,
                Group::ID,
                (int)$groupId,
                self::GROUPS_RULES_TABLE,
                Rule::PRIORITY
            ),
            null,
            Rule::class
        );
    }

    /**
     * Delete trunk group with rules
     *
     * @param integer $id Trunk group id.
     *
     * @throws \Exception Database error
     *
     * @return void
     */
    public function delete(int $id) : void {
        $queries = [
            sprintf('DELETE FROM `%s` WHERE `%s` = %d;', self::GROUPS_MAIN_TABLE, Group::ID, (int)$id),
            sprintf('DELETE FROM `%s` WHERE `%s` = %d;', self::GROUPS_RULES_TABLE, Rule::GROUP_ID, (int)$id),
        ];
        $this->_db()->doTransaction($queries);
    }

    /**
     * Create trunk group item
     *
     * @param string $name        Trunk group name.
     * @param string $description Trunk group description.
     *
     * @return integer Created trunk group id.
     *
     * @throws \Exception Database error
     */
    private function _create(string $name, string $description) : int {
        $this->_db()->insert(
            self::GROUPS_MAIN_TABLE,
            [
                Group::NAME => $name,
                Group::DESCRIPTION => $description
            ]
        );

        return $this->_db()->getLastInsertId();
    }

    /**
     * Delete trunk group rules by trunk group id
     *
     * @param integer $id Trunk group id.
     *
     * @throws \Exception Database error
     *
     * @return void
     */
    private function _clearRules(int $id) : void {
        $query = sprintf(
            'DELETE FROM `%s` WHERE `%s` = %d',
            self::GROUPS_RULES_TABLE,
            Rule::GROUP_ID,
            $id
        );
        $this->_db()->exec($query);
    }

    /**
     * Add trunk group rule item
     *
     * @param integer $id       Trunk group id.
     * @param array   $rule     Trunk group rule content.
     * @param integer $priority Trunk group rule priority.
     *
     * @throws \Exception Database error
     *
     * @return void
     */
    private function _addRule(int $id, array $rule, int $priority) : void {
        $this->_db()->insert(
            self::GROUPS_RULES_TABLE,
            [
                Rule::GROUP_ID => (int)$id,
                Rule::REMOVE => $rule[Rule::FIELD_REMOVE],
                Rule::PREPEND => (string)$rule[Rule::FIELD_PREPEND],
                Rule::MAX_CALLS => $rule[Rule::FIELD_MAX_CALLS],
                Rule::TIMEOUT => $rule[Rule::FIELD_TIMEOUT],
                Rule::TRUNK_NAME => (string)$rule[Rule::FIELD_TRUNK_NAME],
                Rule::PRIORITY => (int)$priority,
            ]
        );
    }

    /**
     * Update trunk group fields
     *
     * @param integer $id          Trunk group id.
     * @param string  $name        Trunk group name.
     * @param string  $description Trunk group description.
     *
     * @throws \Exception Database error.
     *
     * @return void
     */
    private function _update(int $id, string $name, string $description) : void {
        $query = sprintf(
            'UPDATE `%s` SET `%s`=\'%s\', `%s`=\'%s\' WHERE `%s`=%d',
            self::GROUPS_MAIN_TABLE,
            Group::NAME,
            $this->_db()->realEscapeString($name),
            Group::DESCRIPTION,
            $this->_db()->realEscapeString($description),
            Group::ID,
            (int)$id
        );
        $this->_db()->exec($query);
    }

    /**
     * Clear old trunk group rules and add new rules
     *
     * @param integer $id    Trunk group id.
     * @param array   $rules Trunk group rules array.
     *
     * @throws InvalidDataException Invalid data
     *
     * @return void
     */
    private function _setRules(int $id, array $rules) : void {
        $this->_clearRules($id);
        $priority = 1;

        foreach ($rules as $rule) {
            $rule = is_array($rule) ? $rule : json_decode($rule, true);
            $this->_addRule($id, $rule, $priority);
            $priority++;
        }
    }

    /**
     * Create trunk group
     *
     * @param array $params Trunk group data.
     *
     * @throws InvalidDataException Invalid data
     *
     * @return void
     */
    public function create(array $params) : void {
        $this->_db()->beginTransaction();

        $id = $this->_create($params[Group::FIELD_NAME], $params[Group::FIELD_DESCRIPTION] ?? '');
        $this->_setRules($id, $params[Group::FIELD_RULES] ?? []);

        $this->_db()->endTransaction();
    }

    /**
     * Update trunk group with rules
     *
     * @param integer $id   Trunk group id.
     * @param array   $data Trunk group data.
     *
     * @throws InvalidDataException Invalid data
     *
     * @return void
     */
    public function update(int $id, array $data) : void {
        $this->_db()->beginTransaction();

        $this->_update($id, $data[Group::FIELD_NAME], $data[Group::FIELD_DESCRIPTION] ?? '');
        $this->_setRules($id, $data[Group::FIELD_RULES] ?? []);

        $this->_db()->endTransaction();
    }

    /**
     * Get groups count by name and id (optionally)
     *
     * @param string  $name Name.
     * @param integer $id   Id.
     *
     * @return integer
     *
     * @throws \Exception Database error
     */
    public function getCountBy(string $name, int $id) : int {
        $sql = sprintf(
            "SELECT COUNT(groupname) AS `present` FROM `%s` WHERE `groupname`='%s' AND `id` <> %d;",
            self::GROUPS_MAIN_TABLE,
            $this->_db()->realEscapeString($name),
            (int)$id
        );

        return $this->_db()->first($sql)['present'] ?? 0;
    }

    public function loadRelations(array $groups = []): array {
        $items = [];
        foreach ($groups as $group) {
            $items[$group->getId()] = $group;
        }

        $rules = $this->_getRules(array_keys($items));

        foreach ($rules as $rule) {
            if (isset($items[$rule->getParentId()])) {
                $items[$rule->getParentId()]->addRule(
                    [
                        Rule::FIELD_PRIORITY => $rule->getPriority(),
                        Rule::FIELD_TRUNK_NAME => $rule->getTrunkName(),
                        Rule::FIELD_PREPEND => $rule->getPrepend(),
                        Rule::FIELD_REMOVE => $rule->getRemove(),
                        Rule::FIELD_MAX_CALLS => $rule->getMaxCalls(),
                        Rule::FIELD_TIMEOUT => $rule->getTimeout(),
                    ]
                );
            }
        }

        return array_values($items);
    }

    private function _getRules(array $ids = []): array {
        $items = [];
        if ($ids) {
            $items = $this->_db()->getRows(
                sprintf(
                    'SELECT * FROM %s WHERE %s IN (%s)',
                    self::GROUPS_RULES_TABLE,
                    Rule::GROUP_ID,
                    implode(', ', $ids)
                ),
                null,
                Rule::class
            );
        }

        return $items;
    }

    public function getLinked($id): array {
        // phpcs:disable Generic.Files.LineLength.TooLong
        $res = $this->_db()->getRows(
            sprintf(
                'SELECT rn.rulename FROM %s as rl LEFT JOIN %s as rn ON rn.id=rl.object_from_id WHERE rl.object_from_type=%s AND rl.object_to_type=%s AND rl.object_to_id=%d',
                WMSDB::TABLE_RULES_LINKS,
                WMSDB::DIALPLAN_RULE_NAME,
                self::DIALPLAN_LINK,
                self::LINK_GROUP,
                $id
            )
        );
        // phpcs:enable Generic.Files.LineLength.TooLong
        return $res ? array_column($res, 'rulename') : [];
    }

    public function getAvailableTrunkNames() {
        $trunks = [];

        $sipTrunks = $this->_db()->getRows('SELECT * FROM ' . WMSDB::TABLE_TRUNKS, null, SipTrunk::class);

        foreach ($sipTrunks as $trunk) {
            $trunks[] = 'SIP/' . $trunk->getPeername();
        }

        $pstnTrunks = $this->_db()->getRows('SELECT * FROM ' . WMSDB::TABLE_PSTN_TRUNKS, null, PstnTrunk::class);

        foreach ($pstnTrunks as $trunk) {
            for ($port = 0, $trunkPorts = $trunk->getPorts(); $port < $trunkPorts; $port++) {
                $trunks[] = sprintf(
                    '%s/%s_%d',
                    strtoupper($trunk->getTechnology()),
                    $trunk->getUsername(),
                    $port
                );
            }
        }

        return $trunks;
    }
}
