<?php

namespace Wildix;

use Logger\Log;
use Wildix\Auth\Api\CustomerApplications;
use Wildix\Auth\Applications\Interfaces\SimpleTokenServiceInterface;
use Wildix\CallHistory\Calls;
use Wildix\Colleagues\Colleague;
use Wildix\Colleagues\Colleagues;
use Wildix\Devices\Devices;
use Wildix\Dialplan\CallGroups\CallGroupsService;
use Wildix\Dialplan\PagingGroups\PagingGroupsService;
use Wildix\Phonebooks\Phonebook;
use Wildix\Phonebooks\Phonebooks;
use Wildix\UserDataCleaner\Enums\UserDataFieldsEnum;
use Wildix\UserDataCleaner\Objects\UserData;
use Wildix\VoiceMail\VoiceMails;

class UserDataCleaner {

    private $_calls;
    private $_chatCleaner;
    private $_voiceMails;
    private $_phoneBooks;
    private $_callweaverManager;
    private $_logger;
    private $_devices;
    private $_colleagues;
    private $_callGroups;
    private $_pagingGroups;
    private $_simpleToken;
    private $_serverToServerService;

    public function __construct(
        Calls                       $calls,
        ChatHistoryCleaner          $chatCleaner,
        VoiceMails                  $voiceMails,
        Phonebooks                  $phoneBooks,
        CallweaverManager           $callweaverManager,
        Devices                     $devices,
        Colleagues                  $_colleagues,
        CallGroupsService           $_callGroups,
        PagingGroupsService         $_pagingGroups,
        SimpleTokenServiceInterface $_simpleTokenService,
        CustomerApplications        $_serverToServerService,
        Log                         $logger
    ) {
        $this->_calls                 = $calls;
        $this->_chatCleaner           = $chatCleaner;
        $this->_voiceMails            = $voiceMails;
        $this->_phoneBooks            = $phoneBooks;
        $this->_callweaverManager     = $callweaverManager;
        $this->_devices               = $devices;
        $this->_colleagues            = $_colleagues;
        $this->_callGroups            = $_callGroups;
        $this->_pagingGroups          = $_pagingGroups;
        $this->_simpleToken           = $_simpleTokenService;
        $this->_serverToServerService = $_serverToServerService;
        $this->_logger                = $logger;
    }

    public function clearUserData(Colleague $colleague, UserData $userData): bool {
        $this->_logger->info(__LINE__, sprintf('Start clear data by user extension: %s data: %s', $colleague->getExtension(), json_encode($userData->toArray())));

        foreach (get_object_vars($userData) as $userDataField => $userDataValue) {
            if ($userData->{$userDataField} !== UserDataFieldsEnum::USER_WITH_ALL_DATA
                && ($userData->{$userDataField} || $userData->{UserDataFieldsEnum::USER_WITH_ALL_DATA})
            ) {
                if (method_exists(__CLASS__, UserDataFieldsEnum::getValueById($userDataField))) {
                    $this->{UserDataFieldsEnum::getValueById($userDataField)}($colleague);
                }
            }
        }

        $this->clearMissedCallNotifications($colleague);

        return true;
    }

    private function clearCalls(Colleague $colleague): void {
        try {
            $result = $this->_calls->deleteCall(null, null, null, $colleague);
            $this->_logger->info(__LINE__, sprintf('Cleared calls %d', $result));
        } catch (\Throwable $e) {
            $this->_logger->error(__LINE__, $e->getMessage());
        }
    }

    private function clearMissedCallNotifications(Colleague $colleague): void {
        try {
            if ($this->_callweaverManager->isConnected()) {
                $this->_callweaverManager->Command(sprintf('sip notify clear-missed-call %s', $colleague->getExtension()));
            }
            $this->_logger->info(__LINE__, 'Cleared missed call notifications');
        } catch (\Throwable $e) {
            $this->_logger->error(__LINE__, $e->getMessage());
        }
    }

    private function clearChats(Colleague $colleague): void {
        try {
            $result = $this->_chatCleaner->clearDataByUser($colleague);

            $this->_logger->info(__LINE__, sprintf('Cleared chats: ' . json_encode($result)));
        } catch (\Throwable $e) {
            $this->_logger->error(__LINE__, $e->getMessage());
        }
    }

    private function clearVoiceMail(Colleague $colleague): void {
        try {
            $ext = $colleague->getExtension();

            $result = $this->_voiceMails->deleteVoiceMails(
                ['exten' => $ext . ',vm*' . $ext],
                null,
                null,
                null,
                null,
                null,
                null,
                $colleague
            );

            $this->_logger->info(__LINE__, sprintf('Cleared voice mail %d', $result['changed']));
        } catch (\Throwable $e) {
            $this->_logger->error(__LINE__, $e->getMessage());
        }
    }

    private function clearPhoneBooks(Colleague $colleague): void {
        try {
            $result = $this->_phoneBooks->deletePhonebooks(
                ['type' => Phonebook::TYPE_PERSONAL],
                null,
                null,
                $colleague
            );

            $this->_logger->info(__LINE__, sprintf('Delete personal phone books %d', $result));
        } catch (\Throwable $e) {
            $this->_logger->error(__LINE__, $e->getMessage());
        }
    }

    private function clearVoicemailGreetings(Colleague $colleague): void {
        try {
            $this->_voiceMails->deleteCustomGreetings($colleague->getExtension());
            $this->_logger->info(__LINE__, sprintf('Delete voicemail greetings. Extension: %s', $colleague->getExtension()));
        } catch (Exception $e) {
            $this->_logger->error(__LINE__, $e->getMessage());
        }
    }

    private function unassignDevices(Colleague $colleague): void {
        $this->_devices->unassignDevicesByExtension($colleague->getExtension());
        $this->_logger->info(__LINE__, sprintf('Devices unassigned. Extension: %s', $colleague->getExtension()));
    }

    private function clearPersonalData(Colleague $colleague): void {
        $colleague->resetPassword();
        $colleague->setEmail('');
        $colleague->setMobilePhone('');
        $this->_colleagues->removeUserAvatar($colleague);
        $this->_logger->info(__LINE__, sprintf('Email, mobile phone, avatar were cleared. Password reseted. Extension: %s', $colleague->getExtension()));
    }

    private function clearCallGroupMembers(Colleague $colleague): void {
        $this->_callGroups->clearUserByExtension($colleague->getExtension());
        $this->_logger->info(__LINE__, sprintf('Call group members were cleared. Extension: %s', $colleague->getExtension()));
    }

    private function clearCallGroupMissedCallManager(Colleague $colleague): void {
        $this->_callGroups->clearQueueManagerByExtension($colleague->getExtension());
        $this->_logger->info(__LINE__, sprintf('Call queue manager were cleared. Extension: %s', $colleague->getExtension()));
    }

    private function clearPagingGroupMember(Colleague $colleague): void {
        $this->_pagingGroups->clearUserByExtension($colleague->getExtension());
        $this->_logger->info(__LINE__, sprintf('Paging group members were cleared. Extension: %s', $colleague->getExtension()));
    }

    private function clearApplications(Colleague $colleague): void {
        $this->_simpleToken->removeByExtension($colleague->getExtension());
        $this->_serverToServerService->removeByExtension($colleague->getExtension());
        $this->_logger->info(__LINE__, sprintf('Applications were cleared. Extension: %s', $colleague->getExtension()));
    }
}
