<?php

namespace Wildix;

use DateTime;
use DateTimeZone;
use Slim\Psr7\UploadedFile;
use Wildix\Colleagues\Colleagues;

class Validator extends \Validator {

    const ON  = 'on';
    const OFF = 'off';

    private const ACCEPTABLE_TRUE  = ['yes', self::ON, 1, '1', true, 'true'];
    private const ACCEPTABLE_FALSE = ['no', self::OFF, 0, '0', false, 'false'];

    /**
     * Connection to LDAP.
     *
     * @var \WMS_Ldap
     */
    private $_ldap;

    protected $_instanceRuleMessage = array(
        'required' => Error::FieldRequired, //  Required field
        'equals' => Error::UnknownError, // Field must match another field (email/password confirmation)
        'different' => Error::UnknownError, // Field must be different than another field
        'accepted' => Error::UnknownError, // Checkbox or Radio must be accepted (yes, on, 1, true)
        'numeric' => Error::NotNumeric, // Must be numeric
        'integer' => Error::NotInteger, // Must be integer number
        'unsignedInteger' => Error::NotUnsignedInteger, // Must be integer number
        'boolean' => Error::IsNotBoolean, // Must be boolean
        'array' => Error::NotArray, //  Must be array
        'length' => Error::InvalidLength, // String must be certain length
        'lengthBetween' => Error::UnknownError, // String must be between given lengths
        'lengthMin' => Error::UnknownError, // String must be greater than given length
        'lengthMax' => Error::UnknownError, // String must be less than given length
        'min' => Error::MinLimit, // Minimum
        'max' => Error::MaxLimit, // Maximum
        'between' => Error::NotBetween, // Maximum
        'in' => Error::NotInRange, // Performs in_array check on given array values
        'notIn' => Error::UnknownError, // Negation of in rule (not in array of values)
        'ip' => Error::InvalidIpAddress, // Valid IP address
        'domainNameSuffix' => Error::InvalidDomainNameSuffix, // Valid domain name suffix
        'email' => Error::WrongEmailPattern, // Valid email address
        'url' => Error::InvalidUrl, // Valid URL
        'urlActive' => Error::UnknownError, // Valid URL with active DNS record
        'alpha' => Error::NotAlpha, // Alphabetic characters only
        'alphaNum' => Error::AlphaNum, // Alphabetic and numeric characters only
        'slug' => Error::UnknownError, // URL slug characters (a-z, 0-9, -, _)
        'regex' => Error::WrongRegEx, // Field matches given regex pattern
        'date' => Error::UnknownError, // Field is a valid date
        'dateFormat' => Error::UnknownError, // Field is a valid date in the given format
        'dateBefore' => Error::UnknownError, // Field is a valid date and is before the given date
        'dateAfter' => Error::UnknownError, // Field is a valid date and is after the given date
        'contains' => Error::UnknownError, // Field is a string and contains the given string
        'creditCard' => Error::UnknownError, // Field is a valid credit card number
        'instanceOf' => Error::UnknownError, // Field contains an instance of the given class
        'optional' => Error::UnknownError, // Value does not need to be included in data array. If it is however, it must pass validation.
        'DomainOrIp' => Error::HostNotFound,
        'DateTime' => Error::WrongDateTime,
        'PhoneNumber' => Error::WrongPhoneNumber,
        'PrincipalName' => Error::WrongPrincipalName,
        'CommonName' => Error::WrongCommonName,
        'extenNumber' => Error::WrongExtenNumber,
        'checkUserLicenseType' => Error::WrongLicenseType,
        'hasUser' => Error::NotFoundColleague,
        'password' => Error::InvalidPassword,
        'IsFile' => Error::IsFile,
        'SchemeAndDomain' => Error::InvalidSchemeAndDomain,
        'CountryCode' => Error::CountryCode,
        'AdKeyTab' => Error::AdKeyTab,
        'AdFqdn' => Error::AdFqdn,
        'sipPassword' => Error::InvalidPassword,
        'pbxDn' => Error::InvalidPbxDn,
        'groupName' => Error::InvalidData,
        'uniqueLdapProp' => Error::MustBeUnique,
        'existsLdapObj' => Error::InvalidData,
        'aclToG' => Error::InvalidData,
        'isEmptyOrDifferent' => Error::UnknownError,
        'macAddress' => Error::MAC_ADDRESS,
        'ipPort' => Error::IP_PORT,
        'domainOrIp' => Error::DomainOrIp,
        'image' => Error::InvalidData,
        'upgradeRepository' => Error::UpgradeRepository,
        'IncorrectFormat' => Error::IncorrectFormat,
        'CallHistoryTag' => Error::InvalidData,
        'featureCode' => Error::IncorrectFormat,
        'json' => Error::IncorrectFormat,
        'string' => Error::String,
        'StrictInteger' => Error::StrictInteger,
        'ExistsLdapPbxBySerial' => Error::UnknownPbxInLdap,
        'StrictBoolean' => Error::StrictBoolean,
    );

    public function __construct(array $data = array(), array $fields = array()) {
        parent::__construct($data, $fields);
    }

    public function setLdap(\WMS_Ldap $ldap): void {
        $this->_ldap = $ldap;
    }

    /**
     * Validate data and return an array of validation errors.
     *
     * @param array $data   Data for validation.
     * @param array $fields Fields for validation.
     *
     * @return array
     */
    public function validateData(array $data, array $fields) {
        $cloneWithData = $this->withData($data, $fields);
        $extraFields   = array_diff(array_keys($data), $fields);

        if (count($extraFields) > 0) {
            array_walk(
                $extraFields,
                function ($field) use ($cloneWithData) {
                    $cloneWithData->error($field, Error::UnknownError);
                }
            );
        }

        return $cloneWithData->validate() ? array() : $cloneWithData->errors();
    }

    protected function validateDomainOrIp($field, $value) {
        return (preg_match('/^([a-z\d](-*[a-z\d])*)(\.([a-z\d](-*[a-z\d])*))*$/i', $value)
        // valid chars check
            && preg_match('/^.{1,253}$/', $value)
        // overall length check
            && preg_match('/^[^\.]{1,63}(\.[^\.]{1,63})*$/', $value));
        // length of each label
    }

    protected function validateDomainNameSuffix($field, $value) {
        return preg_match('/^((\w+\-+)|(\w+\.))*\w{1,63}\.[a-zA-Z]{2,6}$/', $value);
    }

    protected function validateSchemeAndDomain($field, $value) {
        $parts = parse_url($value);
        if ($parts
            && $this->validateUrl('', $value)
            && isset($parts['host'])
            && $this->validateDomainOrIp('', $parts['host'])
            && !isset($parts['path'])
        ) {
            return true;
        }

        return false;
    }

    protected function validateMacAddress($field, $value) {
        return preg_match('/^[0-9a-f]{12}$/i', $value);
    }

    protected function validateIpPort($field, $value) {
        return preg_match('/^\d{1,5}$/', (string)$value) && $value > 0 && $value <= 65535;
    }

    protected function validateDateTime($field, $value) {
        return DateTime::createFromFormat('d/m/Y H:i:s', $value, new DateTimeZone('UTC')) !== false;
    }

    protected function validatePhoneNumber($field, $value) {
        return preg_match('/^[\+]?[0-9]{2,50}\z/', $value);
    }

    protected function validatePrincipalName($field, $value) {
        return preg_match('/^[^^\/\\\|\+\=\}\{\]\[\)\(\*&\?%\$#\!\,\":;<>]{2,50}\z/i', $value);
    }

    protected function validateCommonName($field, $value) {
        return preg_match('/^[^^\/\\\@\|\+\=\}\{\]\[\)\(\*&\?%\$#@\!\,\":;<>]{2,50}\z/i', $value);
    }

    protected function validateExtenNumber($field, $value) {
        return preg_match('/^[0-9]{2,50}\z/', $value);
    }

    protected function validateHasUser($field, $value) {
        $user = Colleagues::getInstance()->getUserByExtension($value);
        return (boolean)$user;
    }

    protected function validateCheckUserLicenseType($field, $value, $params) {
        $user = Colleagues::getInstance()->getUserByExtension($value);
        return $user && in_array($user->getLicenseType(), $params[0], true);
    }

    protected function validateAdKeyTab($field, $value) {
        return preg_match('/[\w\d\-_@.]+\.keytab/', $value);

    }

    protected function validateAdFqdn($field, $value) {
        return preg_match('/(^((?!-)[a-zA-Z0-9-]{0,62}[a-zA-Z0-9]\.)+[a-zA-Z]{2,63}$)/', $value);
    }

    protected function validateCountryCode($field, $value) {
        return preg_match('/^(1[0-9]{0,3}|[2-9][0-9]{0,2})$/', $value);
    }

    protected function validateLanguageCode($field, $value) {
        return \Language::getInstance()->hasCode($value);
    }

    protected function validatePbxDn($field, $value) {
        return preg_match('/^pbxid=[0-9a-z]{12},dc=wildix$/', $value);
    }

    protected function validateDtmf($field, $value) {
        return preg_match('/^[0-9#*]$/', $value);
    }

    protected function validateDialplan($field, $value) {
        // @ToDo: Now list of dial plan has built in function 'get_contextnames'. Should be object for manipulate with it.
        // @ToDo: Use object for checking dial plan is able.
        return (is_string($value) && strlen($value) > 0);
    }

    protected function validatePassword($field, $value) {
        $notAllowedWebPassCharsRegexp = "/[^a-zA-Z0-9~`!?@#\$%\^&\*\(\)\-_\+=\[\]<>\{\}'\";\\\|\.\/\,]+/";
        $strongWebPassRegexp          = '/^(?=.{8,})(((?=.*[A-Z])(?=.*[a-z])(?=.*\W))|((?=.*[A-Z])(?=.*\W)(?=.*[0-9]))|((?=.*\W)(?=.*[a-z])(?=.*[0-9]))).*$/';
        $mediumWebPassRegexp          = '/^(?=.{8,})(((?=.*[a-z])(?=.*\W))|((?=.*[0-9])(?=.*\W))|((?=.*[A-Z])(?=.*\W))).*$/';

        $isLengsSufficient      = (is_string($value) && strlen($value) >= 8);
        $isNotUseForbiddenChars = (preg_match($notAllowedWebPassCharsRegexp, $value) !== 1);
        $isPasswordStrongEnough = (preg_match($strongWebPassRegexp, $value) === 1 || preg_match($mediumWebPassRegexp, $value) === 1);

        return ($isLengsSufficient && $isNotUseForbiddenChars && $isPasswordStrongEnough);
    }

    protected function validateImage($field, $value) {
        if ($value instanceof UploadedFile) {
            $value = $value->getFilePath();
        }

        return is_file($value) ? is_array(@getimagesize($value)) : false;
    }

    protected function validateIsFile($field, $value) {
        return is_file($value->getFilePath()) && !$value->getError();
    }

    protected function validateCssFile($field, $value) {
        if ($value instanceof UploadedFile) {
            return $value->getClientMediaType() === 'text/css';
        } else {
            return is_file($value) ? mime_content_type($value) === 'text/css' : false;
        }
    }

    /**
     * Validate sip password
     * Sip password must consist of at least 10 symbols and contain at least one uppercase (A-Z) letter,
     * lowercase (a-z) letter, digit (0-9) and special character (%!@#*).
     *
     * @param string $field Field name.
     * @param string $value Value.
     *
     * @return boolean
     */
    protected function validateSipPassword(string $field, string $value): bool {
        $strongSipPassRegexp = '/^(?=.{10,})(?=.*[A-Z])(?=.*[a-z])(?=.*[0-9])(?=.*[%!@#*])[A-Za-z0-9%!@#*]{10,}$/';

        return preg_match($strongSipPassRegexp, $value) === 1;
    }

    protected function validateBoolean($field, $value) {
        if (in_array($value, self::ACCEPTABLE_TRUE, true)) {
            $this->_fields[$field] = true;
        } else if (in_array($value, self::ACCEPTABLE_FALSE, true)) {
            $this->_fields[$field] = false;
        }
        return in_array($value, array_merge(self::ACCEPTABLE_TRUE, self::ACCEPTABLE_FALSE), true) ? true : false;
    }

    /**
     * Validate that a field contains a boolean.
     *
     * @param  string $field
     * @param  mixed  $value
     * @return bool
     */
    protected function validateStrictBoolean($field, $value) {
        return is_bool($value);
    }

    public static function isTrue($s) {
        return in_array($s, self::ACCEPTABLE_TRUE, true);
    }

    protected function validateGroupName(string $field, $value): bool {
        return preg_match('/^[^^\\/\\@\|\\+\=\{\}\\[\\]\\(\\)\\*&\\?%\$#@\!\\,\'\":;<>\\\\]+$/i', $value)
            && preg_match('/[a-z]/i', $value);
    }

    /**
     * Checks that LDAP has no one object with a specified property.
     * This rule required LDAP connection.
     *
     * @param string $field   the name of the field to check.
     * @param mixed  $value   the property value.
     * @param array  $options should contains next items [<LDAP class>, <property name>, <data field name with DN>]
     * @param array  $data    data to validation.
     *
     * @return boolean
     */
    protected function validateUniqueLdapProp(string $field, $value, array $options, array $data): bool {
        if (!isset($this->_ldap)) {
            throw new \LogicException('Rule "uniqueLdapProp" needed LDAP connection.');
        }

        [$class, $property, $dnDataField] = $options;

        $entry = $this->_ldap->getEntry($class, sprintf('(%s=%s)', $property, ldap_escape($value)), ['dn']);

        return (!is_array($entry) || (isset($data[$dnDataField]) && $data[$dnDataField] === $entry['dn']));
    }

    /**
     * Checks that LDAP contains object by his DN.
     *
     * @param string $field the name of the field to check.
     * @param mixed  $value of object DN or an associative array that contains 'dn' field.
     *
     * @return boolean
     */
    protected function validateExistsLdapObj(string $field, $value): bool {
        if (!isset($this->_ldap)) {
            throw new \LogicException('Rule "existsLdapObj" needed LDAP connection.');
        }

        if (!is_string($value)) {
            $value = $value['dn'] ?? null;
        }

        if (!isset($value)) {
            return false;
        }

        $entry = $this->_ldap->getEntryByDn($value, ['dn']);

        return is_array($entry);
    }

    /**
     * Checks that ACL rules of user space are correct.
     *
     * @param string $field the name of the field with ACL rules.
     * @param mixed  $value an ACL rule or an array with ACL rules.
     *
     * @return boolean
     */
    protected function validateAclToG(string $field, $value): bool {
        if (!isset($this->_ldap)) {
            throw new \LogicException('Rule "aclToG" needed LDAP connection.');
        }

        $isValidRuleCall       = function (array $rule): bool {
            return in_array($rule[\Wildix\Colleagues\Groups::FIELD_RULES_RULE], \Wildix\Colleagues\Colleague::CAN_CANNOT_CALL, true)
                && in_array($rule[\Wildix\Colleagues\Groups::FIELD_RULES_ABILITY], \Wildix\Colleagues\Groups::RULE_ABILITY_OPTIONS_CALL, true)
                && ($rule[\Wildix\Colleagues\Groups::FIELD_RULES_GROUP] === 'all' || $rule[\Wildix\Colleagues\Groups::FIELD_RULES_GROUP] === '');
        };
        $isValidRuleUse        = function (array $rule): bool {
            return in_array($rule[\Wildix\Colleagues\Groups::FIELD_RULES_RULE], \Wildix\Colleagues\Colleague::CAN_CANNOT_USE, true)
                && in_array($rule[\Wildix\Colleagues\Groups::FIELD_RULES_ABILITY], \Wildix\Colleagues\Groups::RULE_ABILITY_OPTIONS_USE, true)
                && ($rule[\Wildix\Colleagues\Groups::FIELD_RULES_GROUP] === 'all' || $rule[\Wildix\Colleagues\Groups::FIELD_RULES_GROUP] === '');
        };
        $isValidRuleSet        = function (array $rule): bool {
            return in_array($rule[\Wildix\Colleagues\Groups::FIELD_RULES_RULE], \Wildix\Colleagues\Colleague::CAN_CANNOT_SET, true)
                && in_array($rule[\Wildix\Colleagues\Groups::FIELD_RULES_ABILITY], \Wildix\Colleagues\Groups::RULE_ABILITY_OPTIONS_SET, true)
                && ($rule[\Wildix\Colleagues\Groups::FIELD_RULES_GROUP] === 'all' || $rule[\Wildix\Colleagues\Groups::FIELD_RULES_GROUP] === '');
        };
        $isValidRuleCanCommon  = function (array $rule): bool {
            return in_array($rule[\Wildix\Colleagues\Groups::FIELD_RULES_RULE], \Wildix\Colleagues\Colleague::CAN_CANNOT_COMMON, true)
                && in_array($rule[\Wildix\Colleagues\Groups::FIELD_RULES_ABILITY], \Wildix\Colleagues\Groups::RULE_ABILITY_OPTIONS_YES, true)
                && $rule[\Wildix\Colleagues\Groups::FIELD_RULES_GROUP] === 'all';
        };
        $isValidRuleCanToGroup = function (array $rule): bool {
            return in_array($rule[\Wildix\Colleagues\Groups::FIELD_RULES_RULE], \Wildix\Colleagues\Colleague::CAN_CANNOT_TO_GROUP, true)
                && in_array($rule[\Wildix\Colleagues\Groups::FIELD_RULES_ABILITY], \Wildix\Colleagues\Groups::RULE_ABILITY_OPTIONS_YES, true)
                && ($this->_ldap->getEntryByDn($rule[\Wildix\Colleagues\Groups::FIELD_RULES_GROUP], [\Wildix\Colleagues\Group::DN]) !== false);
        };

        $rules = is_array($value) ? $value : [$value];

        foreach ($rules as $rule) {
            if (
                !isset(
                    $rule[\Wildix\Colleagues\Groups::FIELD_RULES_RULE],
                    $rule[\Wildix\Colleagues\Groups::FIELD_RULES_ABILITY],
                    $rule[\Wildix\Colleagues\Groups::FIELD_RULES_GROUP]
                )
                || (
                    !$isValidRuleCall($rule)
                    && !$isValidRuleUse($rule)
                    && !$isValidRuleSet($rule)
                    && !$isValidRuleCanCommon($rule)
                    && !$isValidRuleCanToGroup($rule)
                )
            ) {
                return false;
            }
        }

        return true;
    }

    /**
     * @param string $field
     * @param string $value
     * @param array  $params
     *
     * @return boolean
     */
    protected function validateIsEmptyOrDifferent(string $field, string $value, array $params): bool {
        $field2 = isset($params[0]) && isset($this->_fields[$params[0]]) ? $this->_fields[$params[0]] : '';

        return empty($value) || empty($field2) || trim($field2) !== trim($value);
    }

    protected function validateUpgradeRepository(string $field, string $value): bool {
        return $this->validateSlug($field, $value) || $this->validateIp($field, $value);
    }

    /**
     * Check that $value is a correct call history tag
     *
     * @param string $field
     * @param string $value
     * @return bool
     */

    protected function validateCallHistoryTag(string $field, string $value) {
        return $this->validateRegex($field, $value, ['/^([0-9a-z_\-]+,?)*$/i']);
    }

    protected function validateFeatureCode(string $field, string $value) {
        $value = trim($value);
        return !empty($value) && preg_match('/^[*#0-9]+$/', $value) && !in_array($value, ['*', '#']);
    }

    /**
     * Add custom error for field
     *
     * @param string $field  Error field.
     * @param string $code   Error code.
     * @param string $reason Error reason.
     *
     * @return void
     */
    public function addCustomError(string $field, string $code = '', string $reason = ''): void {
        $this->_errors[$field] = [
            'code' => $code,
            'reason' => $reason,
        ];
    }

    protected function validateTime(string $field, string $value): bool {
        return (boolean)preg_match('/^([0-9]|0[0-9]|1[0-9]|2[0-3]):[0-5][0-9]:[0-5][0-9]$/', $value);
    }

    protected function validateMax($field, $value, $params) {
        if (is_array($value)) {
            return $params[0] >= count($value);
        }
        return parent::validateMax($field, $value, $params);
    }

    protected function validateNetworkPassword(string $field, $value): bool {
        return (boolean)preg_match('/^(?=.{24,})((?=.*[A-Z])(?=.*[a-z])(?=.*[0-9])).*$/', $value);
    }

    protected function validateString(string $field, $value): bool {
        return is_string($value);
    }

    protected function validateJson(string $field, $value): bool {
        json_decode($value);
        return (json_last_error() === JSON_ERROR_NONE);
    }

    protected function validateStrictInteger(string $field, $value): bool {
        return is_integer($value);
    }

    protected function validateExistsLdapPbxBySerial(string $field, $value): bool {
        if (!isset($this->_ldap)) {
            $this->setLdap(\WMS_Ldap::getInstance());
        }

        return (bool)$this->_ldap->getDnByPbxSerial($value);
    }
}
