<?php
namespace Wildix\VoiceMail;

use Services\LuaService;
use Wildix\Colleagues\Colleague;
use Wildix\Exception;

class VoiceMails extends \Singleton {

    const DIR = '/var/spool/callweaver/voicemail';

    const CUSTOM_GREETING_FILES = [
        'busy.alaw',
        'greet.alaw',
        'unavail.alaw',
    ];

    private $_luaService;

    static function cmpVoicemails($a, $b) {
        if ($a['origtime'] == $b['origtime']) {
            return 0;
        }
        return ($a['origtime'] > $b['origtime']) ? -1 : 1;
    }

    function where($row, $filter = null) {
        if (is_array($filter)) {
            foreach ($filter as $key => $value) {
                if (!in_array($row[$key], $value)) {
                    return false;
                }
            }
        }

        return true;
    }

    private $_folder = array();
    private function getVoiceMailFolder(Colleague $colleague) {
        $extension = $colleague->getExtension();
        if (!$this->_folder[$extension]) {
            $sharedVM = $colleague->getSettings()->getSharedVoiceMail();
            if ($sharedVM && isset($sharedVM['status']) && $sharedVM['status'] == 1 && !empty($sharedVM['value'])) {
                $this->_folder[$extension] = $sharedVM['value'];
            } else {
                $this->_folder[$extension] = $extension;
            }
        }
        return $this->_folder[$extension];
    }

    /**
     *
     * @param  array                        $filter
     * @param  string                       $search
     * @param  array                        $searchFields
     * @param  integer                      $start
     * @param  integer                      $count
     * @param  string                       $sort
     * @param  string                       $dir          (ASC | DESC)
     * @param  Colleague $colleague
     * @return array <\Wildix\Phonebooks\Contact>
     */
    public function getVoiceMails($filter = null, $search = null, $searchFields = null, $start = null, $count = null, $sort = null, $dir = null, Colleague $colleague = null) {
        if ($filter) {
            foreach ($filter as $key => $value) {
                $filter[$key] = explode(',', $value);
                array_walk($filter[$key], 'trim');
            }
        }

        $files = array();

        if (empty($filter['folder']) || in_array('INBOX', $filter['folder'])) {
            $newPath = self::DIR . '/' . $this->getVoiceMailFolder($colleague) . '/INBOX/*.txt';
            foreach (glob($newPath) as $file) {
                $content = parse_ini_file($file, false, INI_SCANNER_RAW);

                if (!$content || !isset($content['msg_id']) || !$content['msg_id']) {
                    continue;
                }

                $content['folder'] = 'INBOX';
                $content['file']   = basename($file);
                $content['path']   = '';
                foreach (array('wav', 'mp3') as $fileExtension) {
                    if (file_exists(self::DIR . '/' . $this->getVoiceMailFolder($colleague) . '/INBOX/' . $content['msg_id'] . '.' . $fileExtension)) {
                        $content['path'] = $this->getVoiceMailFolder($colleague) . '/INBOX/' . $content['msg_id'] . '.' . $fileExtension;
                        break;
                    }
                }

                if ($this->where($content, $filter)) {
                    $files[] = $content;
                }
            }
        }

        if (empty($filter['folder']) || in_array('Old', $filter['folder'])) {
            $oldPath = self::DIR . '/' . $this->getVoiceMailFolder($colleague) . '/Old/*.txt';
            foreach (glob($oldPath) as $file) {
                $content = parse_ini_file($file, false, INI_SCANNER_RAW);

                if (!$content || !isset($content['msg_id']) || !$content['msg_id']) {
                    continue;
                }

                $content['folder'] = 'Old';
                $content['file']   = basename($file);
                $content['path']   = '';
                foreach (array('wav', 'mp3') as $fileExtension) {
                    if (file_exists(self::DIR . '/' . $this->getVoiceMailFolder($colleague) . '/INBOX/' . $content['msg_id'] . '.' . $fileExtension)) {
                        $content['path'] = $this->getVoiceMailFolder($colleague) . '/INBOX/' . $content['msg_id'] . '.' . $fileExtension;
                        break;
                    }
                }

                if ($this->where($content, $filter)) {
                    $files[] = $content;
                }
            }
        }

        usort($files, array("\Wildix\VoiceMail\VoiceMails", "cmpVoicemails"));

        return $files;
    }



    public function updateVoiceMails($data, $filter = null, $search = null, $searchFields = null, $start = null, $count = null, $sort = null, $dir = null, Colleague $colleague = null) {
        $result = array('');

        $files = $this->getVoiceMails($filter, $search, $searchFields, $start, $count, $sort, $dir, $colleague);

        $moveToFolder = false;
        if (!empty($data['folder'])) {
            $moveToFolder = ($data['folder'] == 'INBOX') ? 'INBOX' : 'Old';
            unset($data['folder']);
        }

        if ($moveToFolder !== false) {
            $msg_ids = array_column($files, 'msg_id');
            if ($moveToFolder == 'Old') {
                $result = $this->exec($this->getVoiceMailFolder($colleague), join(',', $msg_ids), 'read');
            }
        }
        return $result[0];
    }



    public function deleteVoiceMails($filter = null, $search = null, $searchFields = null, $start = null, $count = null, $sort = null, $dir = null, Colleague $colleague = null) {
        $result = array('');

        $files = $this->getVoiceMails($filter, $search, $searchFields, $start, $count, $sort, $dir, $colleague);

        $msg_ids = array_column($files, 'msg_id');
        $result  = $this->exec($this->getVoiceMailFolder($colleague), join(',', $msg_ids), 'delete');

        return isset($result[0]) ? $result[0] : $result;
    }

    private function getLuaService(): LuaService {
        if (!$this->_luaService) {
            $this->_luaService = new LuaService();
        }
        return $this->_luaService;
    }

    public function exec(string $extension, string $msgID, string $action) {
        return $this->getLuaService()->manageVoicemails($extension, $msgID, $action);
    }

    /**
     * Delete voice mails older than provided period.
     *
     * @param integer $time Period while we store the voice mails.
     *
     * @return void
     * @throws \Exception File cannot be removed.
     */
    public function deleteAllVoiceMailsAfter($time) {
        $period  = 'month';
        $newDate = new \DateTime();
        $newDate->modify('-' . $time . ' ' . $period);

        $cmd = 'find ' . self::DIR . ' -type d -name "INBOX"';
        exec($cmd, $result);
        foreach ($result as $folder) {
            $temp   = array();
            $filter = array();

            // get extension from folder path
            preg_match('/(\d*)\/INBOX/', $folder, $matches);
            $extension = $matches[1];
            $colleague = \Wildix\Colleagues\Colleagues::getInstance()->getUserByExtension($extension);
            // if colleague exists in system
            if ($colleague !== null) {
                // get all voice mails for current colleague
                $voiceMails = $this->getVoiceMails(array(), null, null, null, null, null, null, $colleague);
                foreach ($voiceMails as $voiceMail) {
                    if ($voiceMail['origtime'] < $newDate->format('U')) {
                        $temp[] = $voiceMail['msg_id'];

                        // delete record file if exists
                        if (!empty($voiceMail['path'])) {
                            $filename = self::DIR . '/' . $voiceMail['path'];
                            if (!unlink($filename)) {
                                throw new Exception('Cannot remove this file: ' . $filename);
                            }
                        }
                    }
                }
                // finally remove old voice mails for current colleague
                if (!empty($temp)) {
                    $filter['msg_id'] = implode(',', $temp);
                    $this->deleteVoiceMails($filter, null, null, null, null, null, null, $colleague);
                }
            } else {
                // if colleague does not exists in system(old files of removed users)
                $files = scandir($folder);
                foreach ($files as $file) {
                    preg_match('/msg.*.mp3$|msg.*.wav$/', $file, $matches);
                    if (!empty($matches[0])) {
                        $filename = $folder . '/' . $matches[0];
                        if (filemtime($filename) < $newDate->format('U')) {
                            if (!unlink($filename)) {
                                throw new Exception('Cannot remove this file: ' . $filename);
                            }
                        }
                    }
                }
            }
        }
    }

    public function deleteCustomGreetings(string $extension): void {
        foreach (self::CUSTOM_GREETING_FILES as $fileName) {
            $file = sprintf('%s/%s/%s', self::DIR, $extension, $fileName);

            if (file_exists($file)) {
                if (!unlink($file)) {
                    throw new Exception('Cannot remove this file: ' . $fileName);
                }
            }
        }
    }
}
