<?php

namespace Wildix\Webinar;

use \Cache\FileCache;
use \Logger\Log;
use Wildix\Services\WizyWebinar;

/**
 * Class WebinarService
 */
class WebinarService {

    /**
     * Contacts cache time (seconds), default 24 hour
     */
    const CONTACTS_CACHE_TTL = 24 * 3600;

    const CONTACTS_CACHE_KEY   = 'contacts';
    const CONTACTS_CACHE_GROUP = 'webinar';

    private $_isEnabled = false;

    /**
     * WizyWebinar class
     *
     * @var WizyWebinar $_httpClient
     */
    private $_httpClient;

    /**
     * File cache class
     *
     * @var FileCache $_cache
     */
    private $_cache;

    /**
     * Logger class
     *
     * @var Log $_logger
     */
    private $_logger;

    public function __construct(
        WizyWebinar $httpClient,
        bool $enabled,
        FileCache $cache,
        Log $logger
    ) {
        $this->_httpClient = $httpClient;
        $this->_isEnabled  = $enabled;
        $this->_cache      = $cache;
        $this->_logger     = $logger;

        $this->_logger->setLogLevel(Log::DEBUG);
    }

    /**
     * Get WizyWebinar enabled status
     *
     * @return boolean
     */
    public function isEnabledWebinar(): bool {
        return $this->_isEnabled;
    }

    /**
     * Check allowing WizyWebinar for current email
     *
     * @param string $email User email.
     *
     * @return boolean
     */
    public function isAllowedForEmail(string $email): bool {
        return $this->isEnabledWebinar()
            && $email
            && in_array(mb_strtolower($email), array_map('mb_strtolower', $this->getContactsEmail()), true);
    }

    /**
     * Get list of allowed email for WizyWebinar
     *
     * @param boolean $force Force read from service.
     *
     * @return array
     */
    public function getContactsEmail(bool $force = false): array {
        if (!$this->isEnabledWebinar()) {
            return [];
        }

        $this->_logger->info(__LINE__, 'Get WizyWebinar contacts');
        $contacts = $this->_cache->get(self::CONTACTS_CACHE_KEY, self::CONTACTS_CACHE_GROUP);
        if ($contacts === false || $force) {
            $contacts = $this->_getContactsFromService();
        }

        return $contacts;
    }

    private function _getContactsFromService(): array {
        $contacts = [];
        try {
            $this->_logger->info(__LINE__, 'Load WizyWebinar contacts from service');
            $contacts = $this->_client()->getWizyWebinarContacts();
            $contacts = array_column($contacts, 'email');
            $this->_cache->set(self::CONTACTS_CACHE_KEY, $contacts, self::CONTACTS_CACHE_TTL, self::CONTACTS_CACHE_GROUP);
            $this->_logger->info(__LINE__, 'Received WizyWebinar contacts from service: ' . json_encode($contacts));
        } catch (\Exception $e) {
            $this->_logger->error(__LINE__, 'Error received WizyWebinar contacts from service. ' . $e->getMessage());
        }

        return $contacts;
    }

    /**
     * Get access token for current email
     *
     * @param string $email User email.
     *
     * @return string
     */
    public function getAccessToken(string $email): string {
        $token = '';

        if ($email) {
            try {
                $token = $this->_client()->wizyWebinarAccessToken($email);
            } catch (\Exception $e) {
                $this->_logger->error(__LINE__, 'Error received WizyWebinar access token from service. ' . $e->getMessage());
            }
        }

        return $token;
    }

    private function _client(): WizyWebinar {
        return $this->_httpClient;
    }
}
