<?php
declare(strict_types=1);
namespace Wildix;

use Models\PBX;
use Services\Helpers;
use Slim\Psr7\UploadedFile;

/**
 * WhiteLabel Class.
 */
class WhiteLabel extends \Singleton {

    const CONF_FILE = '/rw2/etc/whitelabel.conf';

    const FILE_PATH = 'images/whitelabel/';
    const ROOT_PATH = '/var/www/' . self::FILE_PATH;

    /**
     * Default config
     *
     * @var array
     */
    const DEFAULT = [
        'favicon'           => 'favicon.ico',
        'logo'              => 'authorization/images/logo-wildix.svg',
        'smallLogo'         => 'images/logo_wildix_main.svg',
        'emailLogo'         => 'https://addons.wildix.com/img/wildix-logo-positive.png',
        'css'               => '',
        'brand'             => 'Wildix',
        'domain'            => 'wildixin.com',
        'isHelpLinkEnabled' => true,
        'isTopLinkEnabled'  => true,
        'googleMapsApiKey'  => '',
        'pbxName'           => 'Wildix PBX',
        'collaboration'     => 'Collaboration',
        'site'              => 'https://www.wildix.com/',
        'helpLink'          => 'https://wildix.atlassian.net/wiki/x/zhDOAQ',
        'legal'             => 'https://www.wildix.com/license-agreements/',
        'licenseLink'       => 'https://www.wildix.com/licensing/',
    ];

    const FILES = [
        'favicon' => 'favicon.ico',
        'logo' => 'logo.png',
        'smallLogo' => 'logo-small.png',
        'css' => 'style.css',
        'emailLogo' => 'email-logo.png',
    ];

    /**
     * Current config
     *
     * @var null
     */
    private $_config;

    private $_log;

    /**
     * Get settings
     *
     * @return array
     */
    public function getSettings(): array {
        return $this->_getConfig();
    }

    /**
     * Set setting
     *
     * @param array $data Data post and file.
     *
     * @return void
     * @throws Exception            Access Error.
     * @throws InvalidDataException Invalid Data Exception.
     */
    public function setSettings(array $data): void {
        $this->_createDirectory();
        $params = array_intersect_key($data, self::DEFAULT);
        $params = array_diff_key($params, self::FILES);
        $files  = array_intersect_key($data, self::FILES);
        $this->_validate(array_merge($params, $files));

        $result = array_merge($params, $this->_uploadFiles($files));
        if (isset($result['isHelpLinkEnabled'])) {
            $result['isHelpLinkEnabled'] = (int)Validator::isTrue($result['isHelpLinkEnabled']);
        }
        if (isset($result['isTopLinkEnabled'])) {
            $result['isTopLinkEnabled'] = (int)Validator::isTrue($result['isTopLinkEnabled']);
        }
        $current = $this->parseConfigFile();
        $this->writeConfigFile(array_merge($current, $result));
    }

    /**
     * Reset settings.
     *
     * @return void
     * @throws Exception Exception.
     */
    public function resetSettings(): void {
        $files = glob(self::ROOT_PATH . '*', GLOB_MARK);
        if (is_file(self::CONF_FILE)) {
            $files[] = self::CONF_FILE;
        }
        foreach ($files as $file) {
            if (is_file($file) && !unlink($file)) {
                throw new Exception(\Wildix\Error::AccessError, 403);
            }
        }
    }

    private function parseConfigFile(): array {
        $config = Helpers::iniFileGet(self::CONF_FILE);
        if (isset($config['small_logo'])) {
            $config['smallLogo'] = $config['small_logo'];
            unset($config['small_logo']);
        }
        return $config;
    }

    private function writeConfigFile($config): void {
        if (isset($config['smallLogo'])) {
            $config['small_logo'] = $config['smallLogo'];
            unset($config['smallLogo']);
        }
        Helpers::iniFilePut(self::CONF_FILE, $config);
    }

    /**
     * Retrieve msmtp daemon config path - /etc/companies.d/0/msmtprcphp
     *
     * @return array
     */
    private function _getConfig(): array {
        if (!is_array($this->_config)) {
            $this->_config = array_merge(self::DEFAULT, $this->parseConfigFile());
            $this->_config['isHelpLinkEnabled'] = Validator::isTrue($this->_config['isHelpLinkEnabled']);
            $this->_config['isTopLinkEnabled']  = Validator::isTrue($this->_config['isTopLinkEnabled']);
        }

        return $this->_config;
    }

    /**
     * Get congig value.
     *
     * @param string $key Key.
     *
     * @return string|boolean
     */
    private function _getValue(string $key) {
        $value  = '';
        $config = $this->_getConfig();
        if (isset($config[$key])) {
            $value = $config[$key];
        }
        return $value;
    }

    /**
     * Logger
     *
     * @return \Syslog
     */
    private function _getLog():\Syslog {
        if (!$this->_log) {
            $this->_log = new \Syslog(self::class, 4);
        }
        return $this->_log;
    }

    public function getFavicon(): string {
        return $this->_getValue('favicon');
    }

    public function getSmallLogo(): string {
        return $this->_getValue('smallLogo');
    }

    public function getBigLogo(): string {
        return $this->_getValue('logo');
    }

    public function getSite(): string {
        return $this->_getValue('site');
    }

    public function getBrand(): string {
        return $this->_getValue('brand');
    }

    public function getCss(): string {
        return $this->_getValue('css');
    }

    public function isHelpLinkEnabled(): bool {
        return $this->_getValue('isHelpLinkEnabled');
    }

    public function getHelpLink(): string {
        return $this->_getValue('helpLink');
    }

    public function getLicenseLink(): string {
        return $this->_getValue('licenseLink');
    }

    public function isTopLinkEnabled(): bool {
        return $this->_getValue('isTopLinkEnabled');
    }

    public function getGoogleMapsApiKey(): string {
        return $this->_getValue('googleMapsApiKey');
    }

    public function getCollaboration(): string {
        return $this->_getValue('collaboration');
    }

    public function getPbxName(): string {
        return $this->_getValue('pbxName');
    }

    public function getEmailLogo(): string {
        $logo = $this->_getValue('emailLogo');
        if ($logo !== self::DEFAULT['emailLogo']) {
            $logo = 'https://' . PBX::getInstance()->getDomain() . '/' . $logo;
        }
        return $logo;
    }

    public function getLegal(): string {
        return $this->_getValue('legal');
    }

    /**
     * Upload file.
     *
     * @param array $files Array files from $_FILES.
     *
     * @return array
     */
    private function _uploadFiles(array $files): array {
        $res = [];
        foreach ($files as $key => $file) {
            $name = self::ROOT_PATH . self::FILES[$key];

            if ($file instanceof UploadedFile) {
                $file = $file->getFilePath();
            }

            if (move_uploaded_file($file, $name)) {
                $res[$key] = self::FILE_PATH . self::FILES[$key];
            }
        }
        return $res;
    }

    /**
     * Validate
     *
     * @param array $data Data from request.
     *
     * @return void
     * @throws InvalidDataException Invalid Data Exception.
     */
    private function _validate(array $data): void {
        $validator = new Validator($data);
        $validator->rules(
            [
                'url' => [
                    ['site'],
                    ['helpLink'],
                    ['legal'],
                    ['licenseLink'],
                ],
                'domainOrIp' => [
                    ['domain'],
                ],
                'image' => [
                    ['favicon'],
                    ['logo'],
                    ['smallLogo'],
                    ['emailLogo'],
                ],
                'cssFile' => [['css']]
            ]
        );
        $validator->validate();
        if (!empty($validator->errors())) {
            throw new InvalidDataException(Error::getMessage(Error::VALIDATION_ERROR), $validator->errors());
        }
    }

    /**
     * Check and create directory.
     *
     * @return void
     * @throws Exception Access Error.
     */
    private function _createDirectory(): void {
        if (!is_dir(self::ROOT_PATH) && !mkdir(self::ROOT_PATH)) {
            $this->_getLog(__LINE__, 'create directory failed');
            throw new Exception(\Wildix\Error::AccessError, 403);
        }
    }
}
