<?php

/**
 * Custom principal backend
 */

class CustomDAVACLPrincipalBackend extends \Sabre\DAVACL\PrincipalBackend\AbstractBackend {

    /**
     * PDO table name for 'principals'
     *
     * @var string
     */
    protected $tableName;

    /**
     * PDO table name for 'group members'
     *
     * @var string
     */
    protected $groupMembersTableName;

    /**
     * A list of additional fields to support
     *
     * @var array
     */
    protected $fieldMap = array(

        /**
         * This property can be used to display the users' real name.
         */
        '{DAV:}displayname' => array(
            'dbField' => 'displayname',
        ),

        /**
         * This property is actually used by the CardDAV plugin, where it gets
         * mapped to {http://calendarserver.orgi/ns/}me-card.
         *
         * The reason we don't straight-up use that property, is because
         * me-card is defined as a property on the users' addressbook
         * collection.
         */
        '{http://sabredav.org/ns}vcard-url' => array(
            'dbField' => 'vcardurl',
        ),
        /**
         * This is the users' primary email-address.
         */
        '{http://sabredav.org/ns}email-address' => array(
            'dbField' => 'email',
        ),
    );

    /**
     * Sets up the backend.
     *
     * @param string $tableName
     * @param string $groupMembersTableName
     */
    public function __construct($tableName = 'principals', $groupMembersTableName = 'groupmembers') {

        $this->tableName = $tableName;
        $this->groupMembersTableName = $groupMembersTableName;
    }


    /**
     * Returns a list of principals based on a prefix.
     *
     * This prefix will often contain something like 'principals'. You are only
     * expected to return principals that are in this base path.
     *
     * You are expected to return at least a 'uri' for every user, you can
     * return any additional properties if you wish so. Common properties are:
     *   {DAV:}displayname
     *   {http://sabredav.org/ns}email-address - This is a custom SabreDAV
     *     field that's actualy injected in a number of other properties. If
     *     you have an email address, use this property.
     *
     * @param string $prefixPath
     * @return array
     */
    public function getPrincipalsByPrefix($prefixPath) {
    	//logger($prefixPath, 'getPrincipalsByPrefix');

    	$fields = array(
    			'uri',
    	);

    	foreach($this->fieldMap as $key=>$value) {
    		$fields[] = $value['dbField'];
    	}

    	$colleagues = class_colleagues::getInstance()->getColleagues();

    	$principals = array();

    	foreach($colleagues as $row){

    		$row['cn'] = str_replace(' ', '_', $row['cn']);

    		$uri = $prefixPath.'/'.$row['telephoneNumber'];
    		$row['displayname'] =$row['cn'];
    		if(isOSXAddressBook()){
    		    $row['vcardurl'] = 'addressbooks/'.$row['telephoneNumber'].'/Contacts/'.$row['telephoneNumber'];
    		}else{
    		    $row['vcardurl'] = 'addressbooks/'.$row['telephoneNumber'].'/Colleagues/'.$row['telephoneNumber'];
    		}

    		$row['email'] = $row['mail'];

    		// Checking if the principal is in the prefix
    		list($rowPrefix) = \Sabre\DAV\URLUtil::splitPath($uri);
    		if ($rowPrefix !== $prefixPath) continue;

    		$principal = array(
    				'uri' => $uri,
    		);
    		foreach($this->fieldMap as $key=>$value) {
    			if ($row[$value['dbField']]) {
    				$principal[$key] = $row[$value['dbField']];
    			}
    		}
    		$principals[] = $principal;
    	}
        return $principals;

    }

    /**
     * Returns a specific principal, specified by it's path.
     * The returned structure should be the exact same as from
     * getPrincipalsByPrefix.
     *
     * @param string $path
     * @return array
     */

    private $_principals = array();
    public function getPrincipalByPath($path) {
        //logger($path, 'getPrincipalByPath');

    	if(!isset($this->_principals[$path])){
    	    $fields = array(
    	            'id',
    	            'uri',
    	    );
    	    foreach($this->fieldMap as $key=>$value) {
    	        $fields[] = $value['dbField'];
    	    }

    	    $userName = str_replace('principals/', '', $path);
    	    $colleague = class_colleagues::getInstance()->getColleagueByName($userName, true);

    	    if (!$colleague) return;

    	    $colleague['cn'] = str_replace(' ', '_', $colleague['cn']);
    	    $colleague['displayname'] =$colleague['cn'];
    	    //$colleague['vcardurl'] = 'vcard/'.$colleague['cn'];
    	    if(isOSXAddressBook()){
    	        $colleague['vcardurl'] = 'addressbooks/'.$colleague['telephoneNumber'].'/Contacts/'.$colleague['telephoneNumber'];
    	    }else{
    	        $colleague['vcardurl'] = 'addressbooks/'.$colleague['telephoneNumber'].'/Colleagues/'.$colleague['telephoneNumber'];
    	    }

    	    $colleague['email'] = $colleague['mail'];

    	    $this->_principals[$path] = array(
    	            'id'  => $colleague['uid'],
    	            'uri' => 'principals/'.$colleague['telephoneNumber'],
    	    );

    	    foreach($this->fieldMap as $key=>$value) {
    	        if ($colleague[$value['dbField']]) {
    	            $this->_principals[$path][$key] = $colleague[$value['dbField']];
    	        }
    	    }
    	}
        return $this->_principals[$path];
    }

    /**
     * Updates one ore more webdav properties on a principal.
     *
     * The list of mutations is supplied as an array. Each key in the array is
     * a propertyname, such as {DAV:}displayname.
     *
     * Each value is the actual value to be updated. If a value is null, it
     * must be deleted.
     *
     * This method should be atomic. It must either completely succeed, or
     * completely fail. Success and failure can simply be returned as 'true' or
     * 'false'.
     *
     * It is also possible to return detailed failure information. In that case
     * an array such as this should be returned:
     *
     * array(
     *   200 => array(
     *      '{DAV:}prop1' => null,
     *   ),
     *   201 => array(
     *      '{DAV:}prop2' => null,
     *   ),
     *   403 => array(
     *      '{DAV:}prop3' => null,
     *   ),
     *   424 => array(
     *      '{DAV:}prop4' => null,
     *   ),
     * );
     *
     * In this previous example prop1 was successfully updated or deleted, and
     * prop2 was succesfully created.
     *
     * prop3 failed to update due to '403 Forbidden' and because of this prop4
     * also could not be updated with '424 Failed dependency'.
     *
     * This last example was actually incorrect. While 200 and 201 could appear
     * in 1 response, if there's any error (403) the other properties should
     * always fail with 423 (failed dependency).
     *
     * But anyway, if you don't want to scratch your head over this, just
     * return true or false.
     *
     * @param string $path
     * @param array $mutations
     * @return array|bool
     */
    public function updatePrincipal($path, $mutations) {
    	//logger($path, 'updatePrincipal');

        return true;
    }

    /**
     * This method is used to search for principals matching a set of
     * properties.
     *
     * This search is specifically used by RFC3744's principal-property-search
     * REPORT. You should at least allow searching on
     * http://sabredav.org/ns}email-address.
     *
     * The actual search should be a unicode-non-case-sensitive search. The
     * keys in searchProperties are the WebDAV property names, while the values
     * are the property values to search on.
     *
     * If multiple properties are being searched on, the search should be
     * AND'ed.
     *
     * This method should simply return an array with full principal uri's.
     *
     * If somebody attempted to search on a property the backend does not
     * support, you should simply return 0 results.
     *
     * You can also just return 0 results if you choose to not support
     * searching at all, but keep in mind that this may stop certain features
     * from working.
     *
     * @param string $prefixPath
     * @param array $searchProperties
     * @return array
     */
    public function searchPrincipals($prefixPath, array $searchProperties) {
    	//logger($prefixPath, 'searchPrincipals');

        $principals = array();
        return $principals;

    }

    /**
     * Returns the list of members for a group-principal
     *
     * @param string $principal
     * @return array
     */
    public function getGroupMemberSet($principal) {
    	//logger($principal, 'getGroupMemberSet');

        $result = array();
        return $result;
    }

    /**
     * Returns the list of groups a principal is a member of
     *
     * @param string $principal
     * @return array
     */
    public function getGroupMembership($principal) {
    	//logger($principal, 'getGroupMembership');

        $result = array();
        return $result;
    }

    /**
     * Updates the list of group members for a group principal.
     *
     * The principals should be passed as a list of uri's.
     *
     * @param string $principal
     * @param array $members
     * @return void
     */
    public function setGroupMemberSet($principal, array $members) {
    	//logger($principal, 'setGroupMemberSet');
    }

}
