<?php

use Wildix\Devices\Devices;

class Device {

    /**
     * Wildix gateways models.
     *
     * @var array List of models names.
     */
    const WILDIX_GATEWAYS = array(
        'MGW1', 'MGW', 'W01GSM',
        'W-AIR', 'W-AIR SB', 'FXSR2',
        'W-PA', 'FXSR3', 'W-PA2',
    );


	private static $_Instance = null;
	private $_PBXClass = null;
    private $_logger;

    private function __construct() {
        $this->_logger = \Logger\LoggerFactory::getInstance();
    }
	private function __clone() {}

	public static function getInstance(){
		if(self::$_Instance === null){
			self::$_Instance = new Device();
		}
		return self::$_Instance;
	}

	private function getPBXClass() {
		if($this->_PBXClass === null){
			$this->_PBXClass = PBX::getInstance();
		}
		return $this->_PBXClass;
	}

	/**
	 * @return \Wildix\Databases\USRLOC
	 */
	private function getUsrLoc() {
		return \Wildix\Databases\USRLOC::getInstance();
	}

    /**
     * @return \Wildix\Databases\DEVDB
     */
	private function getDevdb() {
		return \Wildix\Databases\DEVDB::getInstance();
	}

    /**
     * Get Devices instance.
     *
     * @return Devices
     */
	private function getDevicesService():Devices  {
        $diContainer = new \Wildix\DependencyInjection\Container();

        return $diContainer->get(Devices::class);
    }

    /**
     * Getting the username assigned to the device by registration in kamailio
     *
     * @param string $deviceMac Device mac address.
     *
     * @return string Username or empty string.
     * @throws Exception If query to USRLOC db was failed.
     */
    private function getSyncUserNameByRegistrations($deviceMac) {
        $query = 'SELECT id, username, contact FROM location WHERE user_agent LIKE "%'.$this->getUsrLoc()->realEscapeString($deviceMac).'%"';

        $syncUserName  = '';
        $registrations = $this->getUsrLoc()->getRows($query);

        foreach ($registrations as $registration) {
            $syncUserName = $registration['username'];
            if (strpos($syncUserName,'unknown') === false) {
                break;
            }
        }

        return $syncUserName;
    }

    /**
     * Getting info to send correct check-sync to device
     *
     * @param array $device Device object with following structure:
     * {
     *      mac : MAC,
     *      address : ip address,
     *      username : username assigned to the device,
     *      model : model,
     *      newfw ?: firmware version, only for devices with 'vendor:additional' MGW and WP
     *  }
     *
     * @return array Info.
     * @throws Exception If query to USRLOC db was failed.
     */
    private function getInfoForCheckSync(array $device) {
        $result = array();

	    if (empty($device)) {
	        return $result;
        }

        $userName    = $device['username'];
        $wildixModel = Wildix\Devices\Vendors::getDeviceInfoByMac($device['mac'])['modelLine'];

        if (in_array($wildixModel, self::WILDIX_GATEWAYS, true)){
            $userName = $this->getSyncUserNameByRegistrations($device['mac']);
        }

        if ($userName === '') { // check-sync will be sent only for registered devices
            $message = sprintf('User not assigned to device %s %s %s', $device['model'], $device['address'], $device['mac']);
            $this->getLog()->error(__LINE__, $message);
            return $result;
        }

        if (strpos($userName, 'login') !== false) {
            $userName = str_replace('login', 'unknown', $userName);
        }

        $result['username']   = $userName;
        $result['model']      = $wildixModel;
        $result['needReboot'] = in_array($wildixModel, Wildix\Devices\Vendors::WPA_FAMILY);

        // set flag about upgrading firmware
        if (!empty($device['newfw'])) {
            $devMac     = $this->getDevdb()->realEscapeString($device['mac']);
            $upgradeSQL = 'INSERT OR REPLACE INTO autoprov (id,param,value,mac) VALUES ((select id from autoprov where param="%s" and mac="%s"),"%s","%s","%s")';

            $result['setFlagQuery'] = sprintf($upgradeSQL, 'upgrade', $devMac, 'upgrade', 'remupgrade', $devMac);
        }

        return $result;
    }

    /**
     * Sync devices by SIP.
     *
     * @param array $devices List of devices.
     *
     * @return void
     * @throws Exception If query to USRLOC db was failed.
     */
    private function syncDevicesViaSipNotify(array $devices) {
        if (empty($devices)) {
            return;
        }

        $queries = array();

        $dataForSync = array(
            'withReboot' => array(
                'peers' => array(),
                'models' => array(),
                'needReboot' => true,
            ),
            'withoutReboot' => array(
                'peers' => array(),
                'models' => array(),
                'needReboot' => false,
            ),
        );

        foreach ($devices as $device) {
            $info = $this->getInfoForCheckSync($device);

            if (isset($info['username'])) {
                $type = $info['needReboot'] ? 'withReboot' : 'withoutReboot';

                $dataForSync[$type]['peers'][]  = $info['username'];
                $dataForSync[$type]['models'][] = $info['model'];
            }

            if (isset($info['setFlagQuery'])) {
                $queries[] = $info['setFlagQuery'];
            }
        }

        try {
            $this->getDevdb()->doTransaction($queries);
        } catch (\Exception $e) {
            $this->getLog()->error(__LINE__, $e->getMessage());
        }

        foreach ($dataForSync as $data) {
            if (empty($data['peers'])) {
                continue;
            }

            $peers = array_unique($data['peers']);

            $this->getLog()->info(__LINE__, 'Send check-sync to ' . implode (',', $data['models']) . ' (' . implode (',', $peers) . ')');
            $this->getPBXClass()->sip_notify_check_sync($peers, $data['needReboot']);
        }
    }

    /**
     * After upgrade and reboot pbx to send SIP notify check-sync to devices, that have new firmware.
     * Calls only in wildix-base package.
     *
     * @return void
     * @throws Exception If transaction in DEVDB failed and the changes were not rollback.
     */
    public function syncDevicesAfterUpgrade() {
        if ($this->getDevicesService()->isDisableSync(Devices::DEVICE_SYNC_MANUAL)) {
            return;
        }
        $devices = $this->getSyncDevicesAfterUpgrade();

        $this->syncDevicesViaSipNotify($devices);
    }

    /**
     * Getting list of devices that need to be updated after upgrade pbx.
     *
     * @return array List of devices.
     * @throws Exception If query to DEVDB was failed.
     */
    public function getSyncDevicesAfterUpgrade() {
        $devices         = $this->getDevdb()->query("select t1.value as device, t2.value as username from autoprov as t1 left join autoprov as t2 ON (t1.mac = t2.mac and t2.param='username') where t1.param='device' group by t1.mac");
        $localFirmwares  = $this->getDevicesService()->getLocalFirmwares();
        $remoteFirmwares = $this->getDevicesService()->getRemoteFirmwares();
        $data            = array();


        while ($result = $devices->fetchArray(SQLITE3_ASSOC)) {
            if (!isset($result['device'])) {
                continue;
            }

            $dev = json_decode($result['device'], true);

            if (!is_array($dev)) {
                continue;
            }

            $devAdditionalVendor      = $dev['vendor:additional'] ?? '';
            $devFirmwareVersion       = $dev['fw:version'] ?? '';
            $devModel                 = $dev['model'] ?? '';
            $devLocalFirmwareVersion  = $localFirmwares[$devModel] ?? '';
            $devRemoteFirmwareVersion = '';

            if ($devAdditionalVendor !== '' && isset($remoteFirmwares[$devAdditionalVendor])) {
                if ($devAdditionalVendor === 'MGW') {
                    $devRemoteFirmwareVersion = $remoteFirmwares[$devAdditionalVendor][$dev['technology']]['version'] ?? '';
                } else if ($devAdditionalVendor === 'WP') {
                    $devRemoteFirmwareVersion = $remoteFirmwares[$devAdditionalVendor][$devModel]['version'] ?? '';
                }
            }

            $existsNewRemoteFirmwareVersion = $devRemoteFirmwareVersion !== '' && $devFirmwareVersion !== $devRemoteFirmwareVersion;
            $existsNewLocalFirmwareVersion  = $devLocalFirmwareVersion !== '' && $devFirmwareVersion !== $devLocalFirmwareVersion;

            if ($existsNewRemoteFirmwareVersion || $existsNewLocalFirmwareVersion) {
                $data[] = array(
                    'mac' => $dev['mac'],
                    'address' => $dev['address'],
                    'username' => str_replace('unknown', 'login', $result['username']),
                    'model' => $devModel,
                    'newfw' => $existsNewRemoteFirmwareVersion ? $devRemoteFirmwareVersion : '',
                );
            }
        }

        return $data;
    }

    /**
     * Get logger instance
     *
     * @return \Logger\Log
     */
    private function getLog() {
        return $this->_logger;
    }

}
