<?php
class GoogleAPI{

    private $_logLevel = 4;

    const TOKEN_VALID = 1;
    const NO_ACCESS_TOKEN = 2;
    const TOKEN_EXPIRED = 3;

    private $_authUrl = 'https://accounts.google.com/o/oauth2/auth';
    private $_tokenUrl = 'https://accounts.google.com/o/oauth2/token';
    private $_userInfoUrl = 'https://www.googleapis.com/oauth2/v3/userinfo';

    private $_contact = [
        'id' => '',
        'updated' => '',
        'title' => '',
        'birthday' => '',
        'nickname' => '',
        'notes' => '',
        'name' => [
            'name' => '',
            'additionalName' => '',
            'familyName' => '',
            'namePrefix' => '',
            'nameSuffix' => '',
        ],
        'website' => [],
        'email' => [],
        'organization' => [],
        'phone' => [],
        'address' => [],
        'sources' => [],
    ];

    private $_contactGroup = [
        'system' => false,
        'id' => '',
        'title' => '',
        'updated' => '',
        'name' => '',
        'memberCount' => 0
    ];

    private static $_Instances = null;
    public static function getInstance(){
        if (self::$_Instances === null) {
            self::$_Instances = new self();
        }
        return self::$_Instances;
    }

    private $_log = null;
    private function getLog(){
        if($this->_log === null){
            $this->_log = new \Syslog('\GoogleApi', $this->_logLevel);
        }
        return $this->_log;
    }

    private $_googleIntegrationApp = null;
    private function getGoogleIntegrationApp($param){
    	if($this->_googleIntegrationApp === null){
    		$kiteService = new \Wildix\Services\Kite();
    		$this->_googleIntegrationApp = $kiteService->getGoogleIntegrationApp();
    	}

    	return $this->_googleIntegrationApp[$param];
    }

    public function getAuthUrl($scope, $callback, $access_type = 'offline', $rePrompted = 'force'){
        $params = array(
                'scope'=> $scope, // email profile
                'state'=> $callback,
                'redirect_uri'=> $this->getGoogleIntegrationApp('redirectUrl'),
                'response_type'=> 'code',
                'client_id'=> $this->getGoogleIntegrationApp('clientId'),
                'access_type'=> $access_type,
                'approval_prompt'=> $rePrompted
        );

        return $this->_authUrl.'?'.urldecode(http_build_query($params));
    }

    /**
     *
     * @var token
     * array(
     *     'received_token_time'=> time
     *     'access_token' => string
     *     'token_type' => string
     *     'expires_in' => int
     *     'refresh_token' => string
     * )
     */
    private $_token = null;
    public function getToken($code = null){
        if($this->_token === null && $code !== null){
            $params = array(
                    'client_id'     => $this->getGoogleIntegrationApp('clientId'),
                    'client_secret' => $this->getGoogleIntegrationApp('clientSecret'),
                    'redirect_uri'  => $this->getGoogleIntegrationApp('redirectUrl'),
                    'grant_type'    => 'authorization_code',
                    'code'          => $code
            );
            $this->_token = $this->_postRequest($this->_tokenUrl, $params);
            $this->_token['received_token_time'] = time();
        }

        return $this->_token;
    }

    public function refreshToken($token){

        $params = array(
                'client_id'     => $this->getGoogleIntegrationApp('clientId'),
                'client_secret' => $this->getGoogleIntegrationApp('clientSecret'),
                'refresh_token'  => $token['refresh_token'],
                'grant_type'    => 'refresh_token'
        );

        $newToken = $this->_postRequest($this->_tokenUrl, $params);
        $newToken['received_token_time'] = time();
        $newToken['refresh_token'] = $token['refresh_token'];
        return $newToken;
    }

    public function isValidToken($token){
        if(empty($token['access_token'])){
            return self::NO_ACCESS_TOKEN;
        }
        if(!empty($token['received_token_time']) && !empty($token['expires_in'])){
            if($token['received_token_time'] + $token['expires_in'] < time()){
                return self::TOKEN_EXPIRED;
            }
        }

        return self::TOKEN_VALID;
    }

    public function setToken($token){
        $this->getLog()->debug(__LINE__, 'setToken: '.print_r($token, true));

        $result = $this->isValidToken($token);
        if($result == self::TOKEN_VALID){
            $this->_token = $token;
        }else if($result = self::TOKEN_EXPIRED){
            $token = $this->refreshToken($token);
            $result = $this->isValidToken($token);
            if($result == self::TOKEN_VALID){
                $this->_token = $token;
            }
        }
        return $result;
    }


    private $_userInfo = null;
    public function getUserInfo($field = null){
        if($this->_userInfo === null){
            $this->_userInfo = $this->_getRequest($this->_userInfoUrl);
        }

        if(!empty($field)){
            return $this->_userInfo[$field];
        }

        return $this->_userInfo;
    }

    public function getPhonebooks($count = 100): array {
        // https://developers.google.com/people/api/rest/v1/contactGroups/list
        $contactGroups = $this->_getRequest('https://people.googleapis.com/v1/contactGroups', ['pageSize' => $count]);

        return $this->_prepareContactGroupsResponse($contactGroups);
    }

    private function _prepareContactGroupsResponse($response): array {
        $result = [];

        foreach ($response['contactGroups'] as $entry) {
            $result['entry'][] = $this->_parseContactGroup($entry);
        }

        if (!empty($response['totalItems'])) {
            $result['total'] = $response['totalItems'];
        }
        if (!empty($response['nextPageToken'])) {
            $result['nextPageToken'] = $response['nextPageToken'];
        }
        if (!empty($response['nextSyncToken'])) {
            $result['nextSyncToken'] = $response['nextSyncToken'];
        }

        return $result;
    }

    private function _parseContactGroup($contactGroup) {
        $group = $this->_contactGroup;

        $fieldMapping = [
            'resourceName' => 'id',
            'name' => 'title',
            'formattedName' => 'name',
            'memberCount' => 'memberCount',
        ];

        foreach ($fieldMapping as $key => $val) {
            if (!empty($contactGroup[$key])) {
                $group[$val] = $contactGroup[$key];
            }
        }

        if (!empty($contactGroup['groupType'])) {
            $group['system'] = $contactGroup['groupType'] === 'SYSTEM_CONTACT_GROUP';
        }
        if (!empty($contactGroup['metadata']['updateTime'])) {
            $group['updated'] = $contactGroup['metadata']['updateTime'];
        }

        return $group;
    }

    public function getContacts($phonebookId = null) {
        $contacts = [
            'entry' => [],
            'errors' => [],
        ];
        if (empty($phonebookId)) {
            return $contacts;
        }

        $peopleResources = $this->getContactResourcesByGroup($phonebookId);
        $this->getLog()->info(__LINE__, 'People resources for ' . $phonebookId . ' count ' . count($peopleResources));

        if (count($peopleResources)) {
            // https://developers.google.com/people/api/rest/v1/people/getBatchGet
            // You can include up to 200 resource names in one request.
            $chunks = array_chunk($peopleResources, 200);

            $tmp    = [];
            $errors = [];
            foreach ($chunks as $n => $chunk) {
                // Limits and Quotas on Google API Requests
                // 50,000 requests per project per day
                // 10 queries per second (QPS) per IP address
                // 100 requests per 100 seconds per user
                // https://developers.google.com/analytics/devguides/config/mgmt/v3/limits-quotas
                usleep(700000);

                $params = [
                    'personFields' => 'addresses,biographies,birthdays,clientData,emailAddresses,genders,memberships,metadata,names,nicknames,occupations,organizations,phoneNumbers,relations,urls',
                    'resourceNames' => $chunk
                ];
                $this->getLog()->info(__LINE__, sprintf('Data chunk %d count people %d', $n, count($chunk)));

                $response = $this->_getRequest('https://people.googleapis.com/v1/people:batchGet', $params);
                if (isset($response['responses']) && is_countable($response['responses'])) {
                    $data = $this->_prepareContactsResponse($response['responses']);

                    $tmp[]    = $data['entry'];
                    $errors[] = $data['errors'];
                } else {
                    $this->getLog()->error(__LINE__, 'ERROR RESPONSE ' . json_encode($response));
                    $error = !empty($response['error']['message']) ? $response['error']['message'] : json_encode($response);
                    $errors[] = [$error];
                }
            }
            $contacts['entry'] = array_merge(...$tmp);
            unset($tmp);

            if (!empty($errors)) {
                $contacts['errors'] = array_merge(...$errors);
                unset($errors);
            }
        }

        return $contacts;
    }

    public function getContactResourcesByGroup($groupId, $count = 100000) {
        $params   = [
            'maxMembers' => $count
        ];
        $response = $this->_getRequest('https://people.googleapis.com/v1/' . $groupId, $params);

        return $response['memberResourceNames'] ?? [];
    }

    private function _prepareContactsResponse($response): array {
        $result = [
            'entry' => [],
            'errors' => [],
        ];

        foreach ($response as $entry) {
            if ($entry['httpStatusCode'] !== 200) {
                $result['errors'][] = sprintf('Incorrect response code %s for %s', $entry['httpStatusCode'], $entry['requestedResourceName']);
                continue;
            }

            $result['entry'][] = $this->_parsePerson($entry['person']);
        }

        return $result;
    }

    private function _parsePerson($googlePerson) {
        $person = $this->_contact;

        $person['id'] = $googlePerson['resourceName'];

        if (!empty($googlePerson['metadata']['sources'])) {
            foreach ($googlePerson['metadata']['sources'] as $source) {
                $person['sources'][] = $source['type'];

                if ($source['type'] === 'CONTACT' && !empty($source['updateTime'])) {
                    $person['updated'] = $source['updateTime'];
                }
            }
        }

        if (!empty($googlePerson['names'][0])) {
            $person['title'] = $googlePerson['names'][0]['displayName'];

            if (!empty($googlePerson['names'][0]['givenName'])) {
                $person['name']['name'] = $googlePerson['names'][0]['givenName'];
            }
            if (!empty($googlePerson['names'][0]['middleName'])) {
                $person['name']['additionalName'] = $googlePerson['names'][0]['middleName'];
            }
            if (!empty($googlePerson['names'][0]['familyName'])) {
                $person['name']['familyName'] = $googlePerson['names'][0]['familyName'];
            }
            if (!empty($googlePerson['names'][0]['honorificPrefix'])) {
                $person['name']['namePrefix'] = $googlePerson['names'][0]['honorificPrefix'];
            }
            if (!empty($googlePerson['names'][0]['honorificSuffix'])) {
                $person['name']['nameSuffix'] = $googlePerson['names'][0]['honorificSuffix'];
            }
        }

        if (!empty($googlePerson['birthdays'])) {
            $person['birthday'] = $googlePerson['birthdays'][0]['text'];
        }

        if (!empty($googlePerson['nicknames'])) {
            $person['nickname'] = $googlePerson['nicknames'][0]['value'];
        }

        if (!empty($googlePerson['biographies'])) {
            $person['notes'] = $googlePerson['biographies'][0]['value'];
        }

        if (!empty($googlePerson['urls'])) {
            foreach ($googlePerson['urls'] as $url) {
                $type = !empty($url['type']) ? $url['type'] : 'other';

                if (empty($person['website'][$type])) {
                    $person['website'][$type] = $url['value'];
                }
            }
        }

        if (!empty($googlePerson['emailAddresses'])) {
            foreach ($googlePerson['emailAddresses'] as $email) {
                $type = !empty($email['type']) ? $email['type'] : 'other';
                if (empty($person['email'][$type])) {
                    $person['email'][$type] = $email['value'];
                }
            }
        }

        if (!empty($googlePerson['organizations'])) {
            foreach ($googlePerson['organizations'] as $organization) {
                $type = !empty($organization['type']) ? $organization['type'] : 'other';
                if (empty($person['organization'][$type])) {
                    $person['organization'][$type] = $organization['name'];
                }
            }
        }

        if (!empty($googlePerson['phoneNumbers'])) {
            foreach ($googlePerson['phoneNumbers'] as $phoneNumber) {
                $type = !empty($phoneNumber['type']) ? $phoneNumber['type'] : 'other';
                if (empty($person['phone'][$type])) {
                    $person['phone'][$type] = $phoneNumber['value'];
                }
            }
        }

        if (!empty($googlePerson['addresses'])) {
            foreach ($googlePerson['addresses'] as $address) {
                $type = !empty($address['type']) ? $address['type'] : 'other';
                if (empty($person['address'][$type])) {
                    $person['address'][$type] = [
                        'city'             => $address['city'] ?? '',
                        'country'          => $address['country'] ?? '',
                        'neighborhood'     => $address['extendedAddress'] ?? '',
                        'pobox'            => $address['poBox'] ?? '',
                        'postcode'         => $address['postalCode'] ?? '',
                        'region'           => $address['region'] ?? '',
                        'street'           => $address['streetAddress'] ?? '',
                        'formattedAddress' => $address['formattedValue'] ?? '',
                    ];
                }
            }
        }

        return $person;
    }

    public function getDomainUsers($pageToken = null, $try = false) {
        $params = [
            'customer' => 'my_customer',
            'viewType' => 'domain_public'
        ];
        if ($pageToken) {
            $params['pageToken'] = $pageToken;
        }
        if ($try) {
            $params['maxResults'] = 1;
        }

        $response = $this->_getRequest('https://www.googleapis.com/admin/directory/v1/users', $params);

        $result = $this->_prepareUsersResponse($response);

        if ($try) {
            // if request was OK, purge users and next page token, so that response will look like flag if token is OK
            $result['users'] = [];
            unset($result['nextPageToken']);
        }

        return $result;
    }

    private function _prepareUsersResponse($response): array {
        $result = ['users' => []];

        foreach ($response['users'] as $user) {
            $result['users'][] = $this->_parseUser($user);
        }

        $result['nextPageToken'] = $response['nextPageToken'] ?? '';

        return $result;
    }

    private function _parseUser($googleUser) {
        $user = [
            'name'         => '',
            'firstName'    => '',
            'lastName'     => '',
            'login'        => '',
            'phone'        => '',
            'phoneWorkFax' => '',
            'phoneWork'    => '',
            'phoneMobile'  => '',
            'email'        => '',
            'group'        => '',
            'password'     => '',
            'avatar'       => '',
            'department'   => '',
            'departmentJob' => '',
        ];

        $user['id']        = $googleUser['id'];
        $user['name']      = $googleUser['name']['fullName'] ?? '';
        $user['firstName'] = $googleUser['name']['givenName'] ?? '';
        $user['lastName']  = $googleUser['name']['familyName'] ?? '';

        $mail          = explode('@', $googleUser['primaryEmail']);
        $user['login'] = $mail[0];

        if (isset($googleUser['phones'])) {
            foreach ($googleUser['phones'] as $phone) {
                switch ($phone['type']) {
                    case 'main':
                    case 'home':
                        $user['phone'] = $phone['value'];
                        break;
                    case 'work':
                        $user['phoneWork'] = $phone['value'];
                        break;
                    case 'work_fax':
                        $user['phoneWorkFax'] = $phone['value'];
                        break;
                    case 'mobile':
                        $user['phoneMobile'] = $phone['value'];
                        break;
                }
            }
        }
        $user['email']    = $googleUser['primaryEmail'] ?? '';
        $user['group']    = $googleUser['isAdmin'] ? 'Admin' : 'Default';
        $user['password'] = $googleUser['password'] ?? '';
        $user['avatar']   = $googleUser['thumbnailPhotoUrl'] ?? '';

        $data = $this->getOrganizationDepartment($googleUser['organizations'] ?? []);

        $user['department']    = $data['department'];
        $user['departmentJob'] = $data['departmentJob'];

        return $user;
    }

    private function getOrganizationDepartment($organizations = []): array {
        $result = [
            'department' => '',
            'departmentJob' => '',
        ];

        foreach ($organizations as $organization) {
            $isPrimary = ($organization['primary'] ?? false) === true;
            if ($isPrimary && !empty($organization['department'])) {
                $result['department']    = $organization['department'];
                $result['departmentJob'] = $organization['department'];
                if (!empty($organization['title'])) {
                    $result['departmentJob'] .= sprintf(' %s %s', env('DEPARTMENTS_TREE_SEPARATOR'), $organization['title']);
                }
            }
        }

        return $result;
    }

    private function _getRequest($url, $params = null) {
        if ($params === null) {
            $params = [];
        }
        $token = $this->getToken(null);

        $params['access_token'] = $token['access_token'];

        $additionalParams = '';
        if (!empty($params['resourceNames'])) {
            $additionalParams = implode(
                '&',
                array_map(
                    function ($item) {
                        return '&resourceNames=' . $item;
                    },
                    $params['resourceNames']
                )
            );

            unset($params['resourceNames']);
        }

        $url .= '?' . http_build_query($params);

        if ($additionalParams) {
            $url .= $additionalParams;
        }

        $this->getLog()->debug(__LINE__, 'GET: ' . $url);

        $curl = curl_init();
        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_HTTPGET, true);
        curl_setopt($curl, CURLOPT_CONNECTTIMEOUT, 5);
        curl_setopt($curl, CURLOPT_TIMEOUT, 120);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, false);
        $result = curl_exec($curl);

        if ($result == false) {
            $result['error'] = [
                'code' => curl_errno($curl),
                'message' => curl_error($curl)
            ];
        } else {
            $result = json_decode($result, true);
        }
        curl_close($curl);
        $this->getLog()->debug(__LINE__, 'Response: '.print_r($result, true));

        return $result;
    }

    private function _postRequest($url, $postParams = null, $getParams=null){

        if($postParams === null){
            $postParams = array();
        }

        $this->getLog()->info(__LINE__, 'POST: '.$url);
        $this->getLog()->debug(__LINE__, print_r($postParams, true));

        $curl = curl_init();
        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_POST, true);
        curl_setopt($curl, CURLOPT_POSTFIELDS, urldecode(http_build_query($postParams)));
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_TIMEOUT, 5);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, false);
        $result = curl_exec($curl);
        if($result == false){
            $result['error'] = array(
                    'code' => curl_errno($curl),
                    'message' => curl_error($curl)
            );
        }else{
            $result = json_decode($result, true);
        }
        curl_close($curl);
        $this->getLog()->debug(__LINE__, 'Response: '.print_r($result, true));
        return $result;
    }
}
