<?php
/**
 * Ldap class
 *
 * @copyright Wildix
 * @package Base
 * @version 1.0
 */
class WMS_Ldap {
	private $configPath = '/rw2/etc/pbx/ldap_connect.conf';
	private $defaultParams = array (
		'binddn'        => 'cn=admin,dc=wildix',
		'bindpw'        => 'wildix',
		'basedn'        => 'dc=wildix',
		'host'          => 'localhost',
		'host_rw'       => 'localhost',             // host to write
        'port_rw'       => '443',                   // port to write
        'rootdn'        => 'cn=admin,dc=wildix',
        'rootpw'        => 'wildix'
	);
	private $params = array();
	private $connect = null;
	private $connect_rw = null;
	private static $object = false;

	const PERSONS = 'wPerson';
	const GROUPS = 'WGroup';
	const PBX = 'wPbx';
	const FAILOVER = 'wFailover';
	const ORGANIZATION_NAME = 'com0';

	private $recursionLimit = 10;

    /**
     * @var \Logger\Log $logger
     */
    private $logger;

	private function __construct() {
	    $this->setParams();
	}

    private function getLogger(): \Logger\Log {
        if (!$this->logger) {
            $this->logger = \Logger\LoggerFactory::getInstance();
        }

        return $this->logger;
    }

    private function getSafeAttributes($attributes) {
        if (!is_array($attributes)) {
            return $attributes;
        }

        $safeAttr = ['userPassword', 'wuserPasswordSip', 'userPIN'];
        foreach ($safeAttr as $attr) {
            if (!empty($attributes[$attr])) {
                $attributes[$attr] = '********';
            }
        }

        return $attributes;
    }

	public static function getInstance() {
		return self::$object == false ? self::$object = new WMS_Ldap() : self::$object;
	}

	public static function closeInstance() {
		if (self::$object != false) {
			self::$object->__destruct();
			self::$object = false;
		}
		return true;
	}

	public function getConnect() {
        $this->setParams();
		if($this->connect === null) {
            // connect read
            $this->connect = @ldap_connect($this->params['host']);
            ldap_set_option($this->connect, LDAP_OPT_PROTOCOL_VERSION, 3);
            if (!ldap_bind($this->connect, $this->params['rootdn'] , $this->params['rootpw'])) {
                return ldap_error($this->connect);
            }
        }

		return $this->connect;
	}

	private function getConnectRw() {
		if($this->params['host_rw'] == $this->params['host']){
			return $this->getConnect();
		}else{
            $this->setParams();
			if($this->connect_rw === null){
                $ldapUri          = 'ldaps://' . $this->params['host_rw'] . ':' . $this->params['port_rw'];
                $this->connect_rw = @ldap_connect($ldapUri);
                ldap_set_option($this->connect_rw, LDAP_OPT_PROTOCOL_VERSION, 3);
                if (!ldap_bind($this->connect_rw, $this->params['binddn'], $this->params['bindpw'])) {
                    return ldap_error($this->connect_rw);
                }
			}

			return $this->connect_rw;
		}
	}

    /**
     * Returns connection for configuration.
     *
     * @param string $host     Hostname.
     * @param string $dn       Dn.
     * @param string $password Password.
     *
     * @return resource|string
     */
    public function getConfigurationConnect(string $host = '', string $dn = '', string $password = '') {
        if ($host === '') {
            $host = $this->params['host'];
        }
        if ($dn === '') {
            $dn = $this->params['binddn'];
        }
        if ($password === '') {
            $password = $this->params['bindpw'];
        }

        $ldapConnect = @ldap_connect($host);
        ldap_set_option($ldapConnect, LDAP_OPT_PROTOCOL_VERSION, 3);

        if (!ldap_bind($ldapConnect, $dn, $password)) {
            return ldap_error($ldapConnect);
        }

        return $ldapConnect;
    }

    public function closeConnect() {
        self::closeInstance();
    }

	public function __destruct() {
        if ($this->connect !== null) {
            ldap_close($this->connect);
        }
        if ($this->connect_rw !== null) {
            ldap_close($this->connect_rw);
        }
	}

	/**
	 * Get LDAP connect parameters
	 *
	 * @return array
	 */
	public function getConnectParams() {
		return $this->params;
	}

	public function getEntry($objectClass, $searchString, $attributes) {
        try {
            $searchResult = ldap_search(
                $this->getConnect(),
                $this->params['basedn'],
                "(&(objectClass={$objectClass}){$searchString})",
                $attributes
            );
            if (!$searchResult) {
                return ldap_error($this->getConnect());
            }
            $entry = ldap_get_entries($this->getConnect(), $searchResult);
            if (!isset($entry['count']) || $entry['count'] != 1) {
                return false;
            }
            unset($entry['count']);
            $result = [];
            foreach ($entry as $row) {
                foreach ($attributes as $attr) {
                    $attrLdap = strtolower($attr);
                    if (!isset($row[$attrLdap]) || !isset($row[$attrLdap][0])) {
                        continue;
                    }
                    if (is_array($row[$attrLdap])) {
                        $result[$attr] = $row[$attrLdap][0];
                    } else {
                        $result[$attr] = $row[$attrLdap];
                    }
                }
            }
            ldap_free_result($searchResult);
            unset($searchResult, $entry, $attrLdap, $attr);

            return $result;
        } catch (\Throwable $throwable){
            return null;
        }
	}

    public function getEntryFromMaster($objectClass, $searchString, $attributes) {
        if (!$this->getConnectRw()) {
            return false;
        }
        $searchResult = ldap_search($this->getConnectRw(), $this->params['basedn'], "(&(objectClass={$objectClass}){$searchString})", $attributes);
        if (!$searchResult) {
            return ldap_error($this->getConnectRw());
        }
        $entry = @ldap_get_entries($this->getConnectRw(), $searchResult);
        if (!isset($entry['count']) || $entry['count'] != 1) {
            return false;
        }
        unset($entry['count']);
        $result = array();
        foreach ($entry as $row) {
            foreach ($attributes as $attr) {
                $attrLdap = strtolower($attr);
                if (!isset($row[$attrLdap]) || !isset($row[$attrLdap][0])) {
                    continue;
                }
                if (is_array($row[$attrLdap])) {
                    $result[$attr] = $row[$attrLdap][0];
                } else {
                    $result[$attr] = $row[$attrLdap];
                }
            }
        }
        @ldap_free_result($searchResult);
        unset($searchResult, $entry, $attrLdap, $attr);
        return $result;
    }

	/**
	 * Get one entry by dn
	 *
	 * @param string $dn
	 * @param array $attributes what attributes return
	 * @param array $attributesAsArray array with attributes which should be returned as array
	 * @return mixed boolean false or array
	 * @author Artem
	 * @author Igor Kuts <igor.kuts@wildix.com>
	 */
	public function getEntryByDn($dn, $attributes = array(), $attributesAsArray = array()) {
		array_walk(
		    $attributesAsArray,
		    function (&$v) {
		        $v = strtolower($v);
            }
        );
		$searchResult = ldap_search($this->getConnect(), $dn, '(&(objectClass=*))', $attributes);
		if (!$searchResult) {
			return ldap_error($this->getConnect());
		}
		$entry = ldap_get_entries($this->getConnect(), $searchResult);
		if (!is_array($entry) || $entry['count'] != 1 || !isset($entry[0])) {
			return false;
		}
		$result = array();
		foreach ($attributes as $attr) {
			$attrLdap = strtolower($attr);
			if (!isset($entry[0][$attrLdap]) || !isset($entry[0][$attrLdap][0])) {
				continue;
			}
			if (is_array($entry[0][$attrLdap]) && !in_array($attrLdap, $attributesAsArray)) {
				$result[$attr] = $entry[0][$attrLdap][0];
			} else {
				if (is_array($entry[0][$attrLdap])) {
					unset($entry[0][$attrLdap]['count']);
				}
				$result[$attr] = $entry[0][$attrLdap];
			}
		}
		ldap_free_result($searchResult);
		unset($searchResult, $entry, $attrLdap, $attr);
		return $result;
	}

	/**
	 * Perform search in LDAP and get found enries as array
	 *
	 * @param string $objectClass
	 * @param string $searchString
	 * @param array $attributes what attributes return
	 * @param array $attributesAsArray array with attributes which should be returned as array
	 * @return mixed boolean false or array
	 * @author Artem
	 * @author Igor Kuts <igor.kuts@wildix.com>
	 */
	public function getEntries($objectClass, $searchString, $attributes, $attributesAsArray = array(), $arrayKey = false) {
		array_walk(
		    $attributesAsArray,
		    function (&$v) {
		        $v = strtolower($v);
            }
        );
		$searchResult = @ldap_search($this->getConnect(), $this->params['basedn'], "(&(objectClass={$objectClass}){$searchString})", $attributes);
		if (!$searchResult) {
			return ldap_error($this->getConnect());
		}
		$entry = @ldap_get_entries($this->getConnect(), $searchResult);
		if (!isset($entry['count']) || $entry['count'] == 0) {
			return false;
		}
		unset($entry['count']);
		$result = array();
		foreach ($entry as $row) {
			$tmpArr = array();
			foreach ($attributes as $attr) {
				$attrLdap = strtolower($attr);
				if (!isset($row[$attrLdap]) || !isset($row[$attrLdap][0])) {
					continue;
				}
				if (is_array($row[$attrLdap]) && !in_array($attrLdap, $attributesAsArray)) {
					$tmpArr[$attr] = $row[$attrLdap][0];
				} else {
					if (is_array($row[$attrLdap])) {
						unset($row[$attrLdap]['count']);
					}
					$tmpArr[$attr] = $row[$attrLdap];
				}
			}
			if($arrayKey && $tmpArr[$arrayKey]){
				$result[$tmpArr[$arrayKey]] = $tmpArr;
			}else{
				$result[] = $tmpArr;
			}

		}
		@ldap_free_result($searchResult);
		unset($searchResult, $tmpArr, $arrayKey, $row, $attr, $attrLdap);
		return $result;
	}

	private function setParams() {
        $this->params = @parse_ini_file($this->configPath, false, INI_SCANNER_RAW);
        if (!is_array($this->params)) {
            $this->params = array();
        }
        $this->params = array_merge($this->defaultParams, $this->params);
    }

    private function parseEntry($attributes, $entry){
        $result = [];
	    foreach ($attributes as $attr) {
	        $result[$attr] = '';
            $ldapKey = strtolower($attr);

            if(!isset($entry[$ldapKey])){
                continue;
            }
            if (isset($entry[$ldapKey]['count'])) {
                if ($entry[$ldapKey]['count']==1) {
                    $result[$attr] = $entry[$ldapKey][0];
                } else {
                    $result[$attr] = $entry[$ldapKey];
                    unset($result[$attr]['count']);
                }
            } else {
                $result[$attr] = $entry[$ldapKey];
            }
	    }

        $result['groupName'] = $this->getEntryByDn($result['owner'], ['cn'])['cn'] ?? '';
        $result['pbx'] = $this->getEntryByDn($result['pbxowner'], ['pbxDomain'])['pbxDomain'] ?? '';

        return $result;
	}

    private function sort_ldap_entries(&$e, $fld, $dir='asc') {
        usort($e, function ($a, $b) use ($fld, $dir) {
            $a = strtolower($a[$fld]);
            $b = strtolower($b[$fld]);
            return $dir === 'asc' ? $a <=> $b : $b <=> $a;
        });
    }

	public function getEntriesSortLimit($objectClass, $searchString, $attributes, $startIndex=0, $limit = 0, $sortKey = false, $sortDir = false) {
        $oldMemory_Limit = ini_get('memory_limit');
        ini_set('memory_limit', '512M');
		$searchResult = @ldap_search($this->getConnect(), $this->params['basedn'], "(&(objectClass={$objectClass}){$searchString})", $attributes);
		if (!$searchResult) {
			return ldap_error($this->getConnect());
		}

        $entries = ldap_get_entries($this->getConnect(), $searchResult);
        @ldap_free_result($searchResult);

        $count = $entries['count'];
        unset($entries['count']);

        if (!isset($count) || $count == 0) {
            return false;
        }

        $result = [];
        foreach ($entries as $entry) {
            $result[] = $this->parseEntry($attributes, $entry);
        }

        if ($sortKey){
            $this->sort_ldap_entries($result, $sortKey, $sortDir);
		}

        if ($limit) {
            $result = array_slice($result, $startIndex, $limit);
        }
        $result['count'] = $count;

        ini_set('memory_limit', $oldMemory_Limit);
		return $result;
	}

    /**
     * Returns count of records.
     *
     * @param string $objectClass  Object class.
     * @param string $searchString Search string.
     *
     * @return int
     */
    public function getCount(string $objectClass, string $searchString): int {
        $searchResult = ldap_search($this->getConnect(), $this->params['basedn'], "(&(objectClass={$objectClass}){$searchString})", ['dn'], 1, 0);
        if (!$searchResult) {
            return 0;
        }
        $count = ldap_count_entries($this->getConnect(), $searchResult);

        ldap_free_result($searchResult);

        return $count;
    }

    public function getEntriesPaged($filter, $attributes, $pageSize) {
        $cookie  = '';
        $entries = [];

        do {
            $searchResult = ldap_search(
                $this->getConnect(),
                $this->params['basedn'],
                $filter,
                $attributes,
                0,
                0,
                2,
                LDAP_DEREF_NEVER,
                [
                    [
                        'oid'   => LDAP_CONTROL_PAGEDRESULTS,
                        'value' => [
                            'size'   => $pageSize,
                            'cookie' => $cookie,
                        ],
                    ],
                ]
            );

            ldap_parse_result($this->getConnect(), $searchResult, $errcode, $matcheddn, $errmsg, $referrals, $controls);
            if (!$searchResult) {
                return ldap_error($this->getConnect());
            }

            $entryID = ldap_first_entry($this->getConnect(), $searchResult);
            if (!$entryID) {
                return $entries;
            }
            do {
                $row = ldap_get_attributes($this->getConnect(), $entryID);

                $tmpArr = [];
                foreach ($attributes as $attr) {
                    if ($attr == 'dn') {
                        $tmpArr['dn'] = ldap_get_dn($this->getConnect(), $entryID);
                    } else {
                        if (isset($row[$attr]['count'])) {
                            if ($row[$attr]['count'] == 1) {
                                $tmpArr[$attr] = $row[$attr][0];
                            } else {
                                unset($row[$attr]['count']);
                                $tmpArr[$attr] = $row[$attr];
                            }
                        } else {
                            $tmpArr[$attr] = $row[$attr] ?? '';
                        }
                    }
                }

                // omit users without extension or cn.
                // include(not bypass) default admin (no extension)
                if (($tmpArr['cn'] ?? '') !== 'admin' && !isset($tmpArr[LDAP_EXTEN_NUMBER], $tmpArr['cn'])) {
                    continue;
                }

                $entries[$tmpArr['uid']] = $tmpArr;

            } while ($entryID = ldap_next_entry($this->getConnect(), $entryID));

            unset($row, $tmpArr, $entryID);
            ldap_free_result($searchResult);

            $cookie = '';
            if (isset($controls[LDAP_CONTROL_PAGEDRESULTS]['value']['cookie'])) {
                $cookie = $controls[LDAP_CONTROL_PAGEDRESULTS]['value']['cookie'];
            }

        } while (!empty($cookie));

        unset($cookie);
        ldap_close($this->connect);
        $this->connect = null;

        return $entries;
    }


    /**
     * Delete entry by dn
	 *
	 * @param string $dn
	 * @return boolean
	 * @author Igor Kuts <igor.kuts@wildix.com>
	 */
	public function deleteEntryByDn($dn) {
        $this->getLogger()->info(__LINE__, '[wms_ldap] deleteEntryByDn ' . $dn);

		$result = @ldap_delete($this->getConnectRw(), $dn);
		if (!$result) {
			return ldap_error($this->getConnectRw());
		}
		return $result;
	}

	private function add($dn, $attributes) {
        $this->getLogger()->info(__LINE__, '[wms_ldap] add ' . $dn . ' ' . json_encode($this->getSafeAttributes($attributes)));

		$result = @ldap_add($this->getConnectRw(), $dn, $attributes);
		if (!$result) {
			return ldap_error($this->getConnectRw());
		}
		return $result;
	}

	/**
	 * Add user to wPerson object of LDAP
	 *
	 * @param array $attributes
	 * @return boolean
	 * @author Artem
	 * @author Igor Kuts <igor.kuts@wildix.com>
	 */
	public function addUser($attributes) {
		// If user dn not presented in attributes - start generation dn process
		if (!array_key_exists('dn', $attributes)) {
			if ($this->recursionLimit <= 0) {
				$this->recursionLimit = 10;
				return false;
			}
			$uid = rand(1000000, 10000000);
			$resultTest = $this->getEntries(WMS_Ldap::PERSONS, '(uid=' . $uid . ')', array('cn'));
			if ($resultTest) {
				--$this->recursionLimit;
				return $this->addUser($attributes);
			} else {
				$this->recursionLimit = 10;
				$dn = 'uid=' . $uid . ',o=' . $attributes['o'] . ',' . $this->params['basedn'];
			}
		} else { // Get user dn from attributes
			$dn = $attributes['dn'];
			unset($attributes['dn']);
		}
		if (!isset($attributes['objectClass'])) {
			if (file_exists('/etc/ldap/ldap_rtrt')) {
				$attributes['objectClass'] = array('top', WMS_Ldap::PERSONS, 'person', 'organizationalPerson', 'inetOrgPerson');
			} else {
				$attributes['objectClass'] = WMS_Ldap::PERSONS;
			}
		}
		$result = array();
		foreach ($attributes as $key => $val) {
			if (is_array($val) || trim($val || '') != '') {
				$result[$key] = $val;
			}
		}

		unset($uid, $resultTest, $key, $val);

		return $this->add($dn, $result);
	}

	/**
	 * Add group to wGroup object of LDAP
	 *
	 * @param array $attributes
	 * @return boolean
	 * @author Artem
	 * @author Igor Kuts <igor.kuts@wildix.com>
	 */
	public function addGroup($attributes) {
        $this->getLogger()->info(__LINE__, '[wms_ldap] addGroup ' . json_encode($this->getSafeAttributes($attributes)));

		// If group dn not presented in attributes - start generation dn process
		if (!array_key_exists('dn', $attributes)) {
			if ($this->recursionLimit <= 0) {
				$this->recursionLimit = 10;
				return false;
			}
			$gid = rand(1000000, 10000000);
			$resultTest = $this->getEntries(WMS_Ldap::GROUPS, '(gid=' . $gid . ')', array('cn'));
			if ($resultTest) {
				--$this->recursionLimit;
				return $this->addGroup($attributes);
			}
			$this->recursionLimit = 10;
			$dn = 'gid=' . $gid . ',o=' . $attributes['o'] . ',' . $this->params['basedn'];
		} else { // Get user dn from attributes
			$dn = $attributes['dn'];
			unset($attributes['dn']);
		}
		if (!isset($attributes['objectClass'])) {
			$attributes['objectClass'] = WMS_Ldap::GROUPS;
		}
		$result = array();
		foreach ($attributes as $key => $val) {
			if (is_array($val) || trim($val) != '') {
				$result[$key] = $val;
			}
		}
		return $this->add($dn, $result);
	}

	public function addPbx($attributes) {
		if ($this->recursionLimit <= 0) {
			$this->recursionLimit = 10;
			return false;
		}
		$this->recursionLimit = 10;
		$dn = 'pbxid=' . $attributes['pbxid'] . ',' . $this->params['basedn'];
		if (!isset($attributes['objectClass'])) {
			$attributes['objectClass'] = WMS_Ldap::PBX;
		}
		$result = array();
		foreach ($attributes as $key => $val) {
			if (trim($val) != '') {
				$result[$key] = $val;
			}
		}
		return $this->add($dn, $result);
	}

	public function addFailover($attributes){
		// If uid not presented in attributes - start generation dn process
		if (!array_key_exists('uid', $attributes)) {
			if ($this->recursionLimit <= 0) {
				$this->recursionLimit = 10;
				return false;
			}
			$uid = rand(1000000, 10000000);
			$resultTest = $this->getEntries(WMS_Ldap::FAILOVER, '(uid=' . $uid . ')', array('dn'));
			if ($resultTest) {
				--$this->recursionLimit;
				return $this->addFailover($attributes);
			}
			$this->recursionLimit = 10;
		} else { // Get user uid from attributes
			$uid = $attributes['uid'];
			unset($attributes['uid']);
		}
		if (!isset($attributes['objectClass'])) {
			$attributes['objectClass'] = WMS_Ldap::FAILOVER;
		}
		$dn = 'uid=' . $uid . ',o=com0,' . $this->params['basedn'];
		$result = array();
		foreach ($attributes as $key => $val) {
			if (is_array($val) || trim($val) != '') {
				$result[$key] = $val;
			}
		}
		$result['uid'] = $uid;
		return $this->add($dn, $result);
	}

	/**
	 * Add external user (used for Settings->System->Directory)
	 *
	 * @param array $attributes
	 * @return mixed boolean or error message
	 * @author Igor Kuts <igor.kuts@wildix.com>
	 */
	public function addExternalUser($attributes) {
		// If user dn or objectClass not presented in attributes - terminate execution
		if (!array_key_exists('dn', $attributes) || !isset($attributes['objectClass'])) {
			return false;
		} else { // Get user dn from attributes
			$dn = $attributes['dn'];
			unset($attributes['dn']);
		}
		$result = array();
		foreach ($attributes as $key => $val) {
			if (is_array($val) || trim($val) != '') {
				$result[$key] = $val;
			}
		}
		return $this->add($dn, $result);
	}

    /**
     * Update LDAP entry.
     *
     * @param string        $dn         Dn.
     * @param array         $attributes Attributes.
     * @param string        $filter     Search filter
     * @param null|resource $connect    LDAP connect.
     *
     * @return boolean|string
     */
    public function update($dn, $attributes = array(), $filter = '(&(objectClass=*))', $connect = null) {
        $this->getLogger()->info(__LINE__, '[wms_ldap] update ' . $dn . ' ' . json_encode($this->getSafeAttributes($attributes)));

		$existKeys = $this->getAllAttrsByDn($dn, $filter, $connect);
		if (!is_array($existKeys)) {
			return false;
		}
		$attributes = array_change_key_case($attributes, CASE_LOWER);
		$attrKeys = array_keys($attributes);
		$notExistkeys = array_diff($attrKeys, $existKeys);

        if ($connect === null) {
            $connect = $this->getConnectRw();
        }

		foreach ($attributes as $key => $val) {
			if (in_array($key, $notExistkeys)) {
				if ($val == '' || empty($val)) {
					continue;
				} else {
					//Add attribute
					$result = @ldap_mod_add($connect, $dn, array($key => $val));
				}
			} else {
				if ($val == '' || empty($val)) {
					// Delete attribute
					$result = @ldap_mod_del($connect, $dn, array($key => array()));
				} else {
					//Update attribute
					$result = @ldap_mod_replace($connect, $dn, array($key => $val));
				}
			}
			if (!$result) {
				$err = ldap_error($connect);
				return $err;
			}
		}
		unset($existKeys, $attrKeys, $notExistkeys, $val, $val , $result);
		return true;
	}

	public function deleteAttrs($dn, $attributes = array()) {
		$ret = true;
        $this->getLogger()->info(__LINE__, '[wms_ldap] deleteAttrs ' . $dn . ' ' . json_encode($this->getSafeAttributes($attributes)));

		$result = @ldap_mod_del($this->getConnectRw(), $dn, $attributes);
		if (!$result) {
			$ret = ldap_error($this->getConnectRw());
		}

		return $ret;
	}

    /**
     * Returns all attributes by dn.
     *
     * @param string        $dn         Dn.
     * @param string        $filter     Search filter
     * @param null|resource $connect    LDAP connect.
     *
     * @return boolean|array
     */
	private function getAllAttrsByDn($dn, $filter = '(&(objectClass=*))', $connect = null) {
        if ($connect === null) {
            $connect = $this->getConnect();
        }

		$searchResult = ldap_search($connect, $dn, $filter);
		if (!$searchResult) {
			return false;
		}
		$entry = ldap_get_entries($connect, $searchResult);
		if (!is_array($entry) || $entry['count'] != 1 || !isset($entry[0])) {
			return false;
		}
		if (!isset($entry['count']) || $entry['count'] != 1 || !isset($entry[0])) {
			return false;
		}
		$keys = array_keys($entry[0]);
		$resultKeys = array();
		foreach ($keys as $key) {
			if (!is_int($key) && $key != 'objectclass' && $key != 'count' && $key != 'dn') {
				$resultKeys[] = $key;
			}
		}
		unset($searchResult, $entry, $keys, $key);
		return $resultKeys;
	}

	public function moveUser($user, $old_pbx = '', $new_pbx = '') {
		if (!empty($new_pbx)) { //add user to new pbx
            $this->getLogger()->info(__LINE__, '[wms_ldap] moveUser ' . $user . ' ' . $old_pbx . ' -> ' . $new_pbx);
			//change ownership for the user
			$colleague = $this->getColleagueByDn($user);
			$res       = $this->update($user, array('pbxowner' => $new_pbx));

			if ($res !== true) {
				return $res;
			}

			\Wildix\Ejabberd::getInstance()->moveUsers([$colleague->getExtension()], $this->getPBXByDn($old_pbx)->getDomain(), $this->getPBXByDn($new_pbx)->getDomain());
		}
		return true;
	}

    /**
     * Move users from one PBX to another.
     *
     * @param string $old_pbx_dn Old PBX dn.
     * @param string $new_pbx_dn New PBX dn.
     *
     * @return boolean|string
     */
    public function moveUsers($old_pbx_dn, $new_pbx_dn) {
        $this->getLogger()->info(__LINE__, '[wms_ldap] moveUsers ' . $old_pbx_dn . ' -> ' . $new_pbx_dn);

        //move users from one dn to other
        $users = $this->getEntriesPaged('(&(objectClass=' . WMS_Ldap::PERSONS . ')(pbxowner=' . $old_pbx_dn . '))', ['dn', 'cn', 'uid', 'telephoneNumber'], 1000);

        $result   = true;
        $oldPbx   = $this->getPBXByDn($old_pbx_dn);
        $newPbx   = $this->getPBXByDn($new_pbx_dn);
        $usersExt = [];

        $maxMoveUsersCount = 500;

        foreach($users as $user){
            $colleague = $this->getColleagueByDn($user['dn']);

            if ($colleague instanceof \Wildix\Colleagues\Colleague) {
                $usersExt[] = $colleague->getExtension();
            }

            $result = $this->update($user['dn'], array('pbxowner' => $new_pbx_dn));
            if ($result !== true) {
                break;
            }


            if (count($usersExt) === $maxMoveUsersCount) {
                \Wildix\Ejabberd::getInstance()->moveUsers($usersExt, $oldPbx->getDomain(), $newPbx->getDomain());
                $usersExt = [];
            }
        }

        if (count($usersExt) > 0) {
            \Wildix\Ejabberd::getInstance()->moveUsers($usersExt, $oldPbx->getDomain(), $newPbx->getDomain());
        }

        return $result;
    }

	public function escapeString($str = ''){

	    $metaChars = array('\\00', '\\', '(', ')', '*');
	    $quotedMetaChars = array ();
	    foreach ($metaChars as $key => $value) {
	        $quotedMetaChars[$key] = '\\'. dechex (ord ($value));
	    }
	    $str = str_replace (
	        $metaChars, $quotedMetaChars, $str
	    ); //replace them
	    return ($str);
	}

	public function getLDAPBackend(){
	    return 'mdb';
	}

    public function getDnByPbxSerial(string $serial): string {
        return $this->getEntry(self::PBX, sprintf('(%s=%s)', 'pbxSerial', ldap_escape($serial)), ['dn'])['dn'] ?? '';
    }

    /**
     * Returns Colleague by dn.
     *
     * @param string $userDn User dn.
     *
     * @return null|\Wildix\Colleagues\Colleague
     */
	private function getColleagueByDn(string $userDn) {
        if (preg_match('/^uid=([^,]+),.*$/m', $userDn, $matches) && !empty($matches[1])) {
            return \Wildix\Colleagues\Colleagues::getInstance()->getUserByID($matches[1]);
        }
        return null;
    }

    /**
     * Returns PBX by dn.
     *
     * @param string $pbxDn PBX dn.
     *
     * @return null|\Wildix\PBXes\PBX
     */
    private function getPBXByDn(string $pbxDn) {
        if (preg_match('/^pbxid=([^,]+),dc=wildix$/m', $pbxDn, $matches) && !empty($matches[1])) {
            return \Wildix\PBXes\PBXes::getInstance()->getPBXByID($matches[1]);
        }
        return null;
    }
}
