<?php
namespace CTI;
/**
 * Class extends SQLite3. It creates sqlite instance
 *
 * @copyright Wildix
 * @version 1.0
 * @package Base
 */
class DB extends \SQLite3 {

	private static $db = array();

	/**
	 * Method return connect to sqlite database
	 * @param string $dbPath Path to sqlite database
	 */
    public static function getInstance($dbPath) {
        if (!isset(self::$db[$dbPath])) {
            self::$db[$dbPath] = new \CTI\DB($dbPath);
        }
        return self::$db[$dbPath];
    }

    /**
     * Path to sqlite database
     * @var string
     */
    private $dbPath;

    /**
     * Method creates connect to sqlite database
     * @param string $dbPath Path to sqlite database
     */
    public function __construct($dbPath) {
        $this->dbPath = $dbPath;
        $this->connect();
    }

    /**
     * Method connects to sqlite database
     */
    public function connect() {
		$this->open($this->dbPath, SQLITE3_OPEN_READWRITE);
    }

    /**
     * Method closes connect to database
     */
    public function disconnect() {
        $this->close();
    }

    public function checkTable($tableName, $num) {
    	$query = "SELECT * FROM {$tableName} LIMIT 1;";
    	$result = $this->query($query);
    	$count = 0;
		if ($row = $result->fetchArray(SQLITE3_ASSOC)) {
            $count =  $result->numColumns();
            $result->finalize();
        }
        if ($count != $num) {
        	$this->exec('DROP TABLE ' . $tableName);
        }
    }

    /**
     * Auto disconnect
     */
    public function  __destruct() {
        $this->disconnect();
    }

    /**
     * Overwrite escape function
     * @param String $string
     * @return String
     */
    public function escape($string) {
        return strip_tags($this->escapeString($string));
    }

    /*
     * wrapper for query function of sqlite3
     * @param string $query
     */
    public function query($query): \SQLite3Result {
        $res = @parent::query($query);
        if(!$res && $this->lastErrorCode() == 14){ // The database disk image is malformed
            if($this->removeJournal()){
                $res = @parent::query($query);
            }
        }
        if(!$res && $this->lastErrorCode() == 11){ // The database disk image is malformed
            if($this->restoreDB()){
                $res = @parent::query($query);
            }
        }
        return $res;
    }

    /*
     * wrapper for exec function of sqlite3
     * @param string $query
     */
    public function exec($query){
        $res = @parent::exec($query);
        if(!$res && $this->lastErrorCode() == 14){ // The database disk image is malformed
            if($this->removeJournal()){
                $res = @parent::exec($query);
            }
        }
        if(!$res && $this->lastErrorCode() == 11){ // The database disk image is malformed
            if($this->restoreDB()){
                $res = @parent::exec($query);
            }
        }
        return $res;
    }

    /*
     * Restoring DB from the backup if exists. Also send to user notification about restoration
     */
    public function restoreDB(){
        $path = dirname($this->dbPath);
        $name = basename($this->dbPath);
        $backup = $path.DIRECTORY_SEPARATOR.$name.'.dmp.gz';
        $schemaRestore = '/usr/sbin/init_'.$name;
        if(file_exists($backup)){
            $this->disconnect();
            unlink($this->dbPath);
            exec('zcat "'.$backup.'" | sqlite3 "'.$this->dbPath.'"');
            $this->connect($this->dbPath);

            Logger::getInstance('cti_db.php')->info(__LINE__, 'Database '.$name.' was restored from daily backup over corrupted database image');

            return true;
        }elseif(file_exists($schemaRestore)){
            $this->disconnect();
            unlink($this->dbPath);
            exec('/bin/sh '.$schemaRestore);
            $this->connect($this->dbPath);

            Logger::getInstance('cti_db.php')->info(__LINE__, 'Schema of the database '.$name.' was recreated over corrupted database image');

            return true;
        }else{
        	Logger::getInstance('cti_db.php')->error(__LINE__, 'Database '.$name.' not restored');
        }

        return false;
    }

    private function removeJournal() {
        if (file_exists($this->dbPath.'-journal') ){
            //Used sudo rm due to sometimes not enough rights
            exec('sudo rm '.$this->dbPath.'-journal', $out, $ret);
            Logger::getInstance('cti_db.php')->info(__LINE__, 'Sqlite journal file exists:'.$this->dbPath.'-journal - '. ($ret==0?'deleted':'can not delete it'));
            return $ret==0;
        }
        return false;
    }
}
