<?php
namespace CTI;

use Api\Translation\Enums\LanguagesEnum;

/**
 * Class for creating xml, xsl and generating xsl
 *
 * @copyright Wildix
 * @version   1.0
 * @package   Base
 */
class Templater {

    /**
     * XML variable, only one in all extended classes
     *
     * @var object
     */
    private static $xml = false;
    /**
     * XML root tag pointer, only one in all extended classes
     *
     * @var object
     */
    private static $rootTag;
    /**
     * XML modules tag pointer, only one in all extended classes
     *
     * @var object
     */
    private static $modulesTag = null;
     /**
      * XML modules tag pointer, only one in all extended classes
      *
      * @var object
      */
    private static $includesTag = null;

    /**
     * Method generates html from xsl + xml. Method is final, can't be overrided
     *
     * @param  string $stylesheet Main stylesheet file path
     * @return string result HTML
     */
    final public function generateHTML($stylesheet = 'index.xsl') {
        $xslt = new \XSLTProcessor();
        $XSL  = new \DOMDocument();
        $XSL->load('data/templates/' . $stylesheet);
        $xslt->importStylesheet($XSL);
        if (self::$xml == false) {
            $this->newDocument('root');
        }

        return $xslt->transformToXML(self::$xml);
    }

    /**
     * Method creates new DOMDocument and modules tag. Method is final, can't be overrided
     *
     * @param string $tagName    Root tag.
     * @param array  $attributes Root attributes.
     */
    final public function newDocument($tagName, $attributes = array(), $lang = 'en') {
        self::$xml = new \DOMDocument('1.0', 'UTF-8');

        self::$xml->loadXML($this->buildRootTemplate($lang));
        self::$rootTag = self::$xml->getElementsByTagName($tagName)->item(0);

        foreach ($attributes as $name => $value) {
            self::$rootTag->setAttribute($name, $value);
        }

        self::$modulesTag  = $this->addTag('modules', null, self::$rootTag);
        self::$includesTag = $this->addTag('includes', null, self::$rootTag);
    }

    /**
     * Method adds tag to xml tree. Method is final, can't be overrided
     *
     * @param  string      $tagName    Tag name
     * @param  string      $text       Tag value
     * @param  DOMDocument $parentNode Pointer to parent node
     * @param  array       $attributes Tag attributes
     * @return DOMDocument Node object
     */
    final public function addTag($tagName, $text, $parentNode, $attributes = array()) {
        $element = self::$xml->createElement($tagName, $text);
        foreach ($attributes as $name => $value) {
            $element->setAttribute($name, $value);
        }
        $parentNode->appendChild($element);
        return $element;
    }

    /**
     * Method returns xml as text. Method is final, can't be overrided
     *
     * @return string XML data
     */
    final public function getXML() {
        return self::$xml->saveXML();
    }

    /**
     * Method returns root tag. Method is final, can't be overrided
     *
     * @return DOMDocument Root tag
     */
    final public function getRootElement() {
        return self::$rootTag;
    }

    /**
     * Method returns modules tag. Method is final, can't be overrided
     *
     * @return DOMDocument Modules tag
     */
    final public function getModulesElement() {
        return self::$modulesTag;
    }

    /**
     * Method returns modules tag. Method is final, can't be overrided
     *
     * @return DOMDocument Modules tag
     */
    final public function getIncludesElement() {
        return self::$includesTag;
    }

    /**
     * Method adds string-xml to DOMDocument. Method is final, can't be overrided
     *
     * @param  string      $tagName    Tag name.
     * @param  string      $xml        XML data.
     * @param  DOMDocument $parentNode Parent node.
     * @return boolean Result.
     */
    final public function addText($tagName, $xml, $parentNode) {
        $txtNode   = $this->addTag($tagName, null, $parentNode);
        $importXML = simplexml_load_string($xml);
        if ($importXML === false) {
            return false;
        }
        $importXML  = dom_import_simplexml($importXML);
        $importNode = self::$xml->importNode($importXML, true);
        $txtNode->appendChild($importNode);
        return true;
    }

    /**
     * Method gets xml object. Method is final, can't be overrided
     *
     * @return DOMDocument XML object
     */
    final public function getDocument() {
        return self::$xml;
    }

    final public function addCData($nodeName, $text, $node) {
        $nodeCData = self::$xml->createCDATASection($text);
        $tag       = $this->addTag($nodeName, '', $node);
        $tag->appendChild($nodeCData);
    }

    private function buildRootTemplate($lang): string {
        $data   = [];
        foreach (LanguagesEnum::getCollaborationDictionaryByCode($lang) as $key => $value) {
            $data[] = sprintf('<%s><![CDATA[%s]]></%s>', $key, $value, $key);
        }

        return sprintf('<?xml version="1.0" encoding="UTF-8"?><root><trans_texts>%s</trans_texts></root>', implode($data));
    }
}
