<?php
namespace CTI;
/**
 * Class for user data
 * Class uses singleton pattern (can be called without creating object)
 *
 * @copyright Wildix
 * @version 1.0
 * @package Base
 */
class User {

    const ERROR_CANT_GET_USER_DATA = 10000;


    private static $object;

    /**
     * User data
     * @var array
     */
    private $userData = array();

    /**
     * User features taken from lua file
     * @var array
     */
    private $features = null;

    /**
     * User acl
     * @var array
     */
    private $acl = null;

    private $featuresToTest = array('CTI_PRO', 'CTI_PO', 'CTI_CDR', 'CTI_CDR_VIEW', 'PERSONALRECORD', 'SMS', 'FAX', 'BROWSER_PHONE', 'VOICEMAIL', 'CONTACT_CENTER');

    const FEATURES = 1;
    const USER_DATA = 2;
    const ACL = 3;

    private function __construct() {
        $this->init();
    }

    public static function getInstance() {
        return self::$object == false ? self::$object = new \CTI\User() : self::$object;
    }

    /**
     * Init user data from ldap, get user features and acl
     * @return bool Init result
     */
    public function init() {
        $this->getUserData();
        $this->getUserAclAllowed();
        $this->getUserFeatures();
        return true;
    }

    private function getUserData() {
        if (empty($this->userData)) {

            if (\User::getInstance()->isLogged()) {
                $this->userData = array(
                        'dn' => \User::getInstance()->getDn(),
                        'ou' => \User::getInstance()->getDepartment(),
                        'userDialplan' => \User::getInstance()->getDialplan(),
                        'userFaxDialplan' => \User::getInstance()->getFaxDialplan(),
                        'preferredLanguage' => \User::getInstance()->getLanguage(),
                        'userKind' => \User::getInstance()->getRole(),
                        'mail' => \User::getInstance()->getEmail(),
                        'mobile' => \User::getInstance()->getMobilePhone(),
                        'owner' => \User::getInstance()->getGroupDn(),
                        'pbxowner' => \User::getInstance()->getPBXDn(),
                        'userPassword' => \User::getInstance()->getPassword(),
                        'uid' => \User::getInstance()->getID(),
                        'cn' => \User::getInstance()->getName(),
                        LDAP_EXTEN_NUMBER => \User::getInstance()->getExtension(),
                        'imageUrl' => \User::getInstance()->getPicture(),
                        'facsimileTelephoneNumber' => \User::getInstance()->getFaxNumber(),
                        'wuserPasswordSip' => \User::getInstance()->getSipPassword(),
                        'wOfficePhone' => \User::getInstance()->getOfficePhone(),
                        \Wildix\Colleagues\Colleague::LICENSE_TYPE => \User::getInstance()->getLicenseType(),
                );
            }

            if (!is_array($this->userData)) {
                throw new \CTI\Exception('CAN`T GET USER DATA', self::ERROR_CANT_GET_USER_DATA);
            }

            if (!empty($this->userData[LDAP_EXTEN_NUMBER])) {
                $this->userData['extenNumber'] = $this->userData[LDAP_EXTEN_NUMBER];
            }
            \CTI\Session::getInstance()->set('language', \Language::getInstance()->getLanguageNameByCode($this->userData['preferredLanguage']));
        }

        return $this->userData;
    }

    private function getUserFeatures() {
        if ($this->features == null) {
            $this->features = getUserFeatures($this->get(self::USER_DATA, 'uid'));
            if (!is_array($this->features)) {
                throw new \CTI\Exception('CAN`T GET FEATURES', \CTI\Exception::ERROR_WARNING);
            }
        }
        return $this->features;
    }

    private function getUserAclAllowed() {
        if ($this->acl == null) {
            $this->acl = array();

            $userDataOwner   = $this->get(self::USER_DATA, 'owner');
            $userLicenseType = $this->get(self::USER_DATA, \Wildix\Colleagues\Colleague::LICENSE_TYPE);

            foreach ($this->featuresToTest as $feature) {
                $this->acl[$feature] = is_acl_subject_allowed($userDataOwner, 'canuse', $feature, $userLicenseType);
            }

            $this->acl['P_INFO']          = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::PERSONAL_INFORMATION, $userLicenseType);
            $this->acl['QUICK_STATUS']    = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::ADVANCED_STATUS, $userLicenseType);
            $this->acl['DND']             = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::DO_NOT_DISTURB, $userLicenseType);
            $this->acl['FKEYS']           = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::FUNCTION_KEYS, $userLicenseType);
            $this->acl['FCF']             = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::FORCE_CALL_FORWARD, $userLicenseType);
            $this->acl['BCF']             = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::BUSY_CALL_FORWARD, $userLicenseType);
            $this->acl['UCF']             = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::UNAVAILABLE_CALL_FORWARD, $userLicenseType);
            $this->acl['CW']              = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::CALL_WAITING, $userLicenseType);
            $this->acl['VM']              = is_acl_subject_allowed($userDataOwner, 'canset', 'VM', $userLicenseType);
            $this->acl['TOUT']            = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::DIAL_TIMEOUT, $userLicenseType);
            $this->acl['MEXT']            = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::MOBILITY_EXTENSION_MANAGEMENT, $userLicenseType);
            $this->acl['RETXFER']         = is_acl_subject_allowed($userDataOwner, 'canset', 'RETXFER', $userLicenseType);
            $this->acl['MOBCONFIRMATION'] = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::MOBILITY_CONFIRMATION, $userLicenseType);
            $this->acl['PHONEBOOK']       = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::PHONEBOOKS, $userLicenseType);

            $this->acl['STIME']   = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::TIMETABLE, $userLicenseType);
            $this->acl['3SWITCH'] = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::THREE_STATE_SWITCH, $userLicenseType);
            $this->acl['SWITCH']  = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::SWITCH_, $userLicenseType);

            $this->acl['EMAILMCN'] = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::EMAIL_MISSED_CALL_NOTIFICATION, $userLicenseType);
            $this->acl['SMSMCN']   = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::SMS_MISSED_CALL_NOTIFICATION, $userLicenseType);
            $this->acl['RING']     = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::RINGTONE, $userLicenseType);

            $this->acl['EDIT_PP']           = is_acl_subject_allowed($userDataOwner, 'yesno', \Wildix\Colleagues\Colleague::MODIFY_PUBLIC_PHONEBOOKS, $userLicenseType, 'all');
            $this->acl['CREATE_CONFERENCE'] = is_acl_subject_allowed($userDataOwner, 'yesno', \Wildix\Colleagues\Colleague::CREATE_CONFERENCE, $userLicenseType, 'all');
            $this->acl['SEE_EXTENSIONS']    = is_acl_subject_allowed($userDataOwner, 'yesno', \Wildix\Colleagues\Colleague::SEE_EXTENSIONS, $userLicenseType, 'all');
            $this->acl['DELETE_CALLS']      = is_acl_subject_allowed($userDataOwner, 'yesno', \Wildix\Colleagues\Colleague::DELETE_CALLS, $userLicenseType, 'all');

            $this->acl['TWO_FACTOR_AUTH_OFF'] = is_acl_subject_allowed($userDataOwner, 'yesno', \Wildix\Colleagues\Colleague::TWO_FACTOR_AUTH_OFF, $userLicenseType, 'all');
            $this->acl['ENABLE_VIDEO_CALL'] = is_acl_subject_allowed($userDataOwner, 'yesno', \Wildix\Colleagues\Colleague::ENABLE_VIDEO_CALL, $userLicenseType, 'all');

            $this->acl['SEE_CALL_RECORDINGS'] = is_acl_subject_allowed($userDataOwner, 'yesno', \Wildix\Colleagues\Colleague::SEE_CALL_RECORDINGS, $userLicenseType, 'all');

            $groups = \Wildix\Colleagues\Groups::getInstance()->getGroups(\User::getInstance());
            if ($groups) {
                foreach ($groups as $key => $group) {
                    $dn = $group->getDn();
                    $this->acl['SEE_COLLEAGUES'][$dn]  = is_acl_subject_allowed($userDataOwner, 'yesno', \Wildix\Colleagues\Colleague::VIEW, $userLicenseType, $dn);
                    $this->acl['PICKUP'][$dn]          = is_acl_subject_allowed($userDataOwner, 'yesno', \Wildix\Colleagues\Colleague::CALL_PICKUP, $userLicenseType, $dn);
                    $this->acl['SUPERVISION'][$dn]     = is_acl_subject_allowed($userDataOwner, 'yesno', \Wildix\Colleagues\Colleague::INTRUSION, $userLicenseType, $dn);
                    $this->acl['SEE_CALLS'][$dn]       = is_acl_subject_allowed($userDataOwner, 'yesno', \Wildix\Colleagues\Colleague::SEE_CALLS_OF_USERS, $userLicenseType, $dn);
                    $this->acl['CHANGE_PRESENCE'][$dn] = is_acl_subject_allowed($userDataOwner, 'yesno', \Wildix\Colleagues\Colleague::CHANGE_PRESENCE, $userLicenseType, $dn);
                    $this->acl['VIEW_GEO'][$dn]        = is_acl_subject_allowed($userDataOwner, 'yesno', \Wildix\Colleagues\Colleague::VIEW_GEO, $userLicenseType, $dn);
                    $this->acl['SEE_VOICEMAIL'][$dn]   = is_acl_subject_allowed($userDataOwner, 'yesno', \Wildix\Colleagues\Colleague::SEE_VOICEMAIL, $userLicenseType, $dn);
                }
            }
            if (empty($this->acl['SEE_COLLEAGUES'])) {
                $this->acl['SEE_COLLEAGUES'] = 0;
            }

            $exten = $this->get(self::USER_DATA, 'extenNumber');
            if (empty($exten)) {
                $this->acl['CTI_PRO'] = 0;
                $this->acl['CTI_PO']  = 0;
            }
            $this->acl['FAXSERVERSET'] = is_acl_subject_allowed($userDataOwner, 'canset', \Wildix\Colleagues\Colleague::FAXSERVERSET, $userLicenseType);
        }
        return $this->acl;
    }

    /**
     * Method gets option or right.
     *
     * @param string $type Option type.
     * @param string $name Option name.
     *
     * @return string Option|Right
     */
    public function get($type, $name = false) {
        $result = null;
        switch ($type) {
            case self::USER_DATA:
                if (isset($this->userData[$name]) || $name === false) {
                    $result = $name === false ? $this->userData : $this->userData[$name];
                }
                break;
            case self::FEATURES:
                if (isset($this->features[$name]) || $name === false) {
                    $result = $name === false ? $this->features : $this->features[$name];
                }
                break;
            case self::ACL:
                if (isset($this->acl[$name]) || $name === false) {
                    $result = $name === false ? $this->acl : $this->acl[$name];
                }
                break;
            default:
                $result = null;
                break;
        }
        return $result;
    }

    /**
     * Method sets option or right.
     *
     * @param string $type Option type.
     * @param string $name Option name.
     * @param string $val  Option value.
     *
     * @return void
     */
    public function set($type, $name, $val) {
        switch ($type) {
            case self::USER_DATA:
                $this->userData[$name] = $val;
                break;
            case self::FEATURES:
                $this->features[$name] = $val;
                break;
            case self::ACL:
                $this->acl[$name] = $val;
                break;
            default:
                break;
        }
    }

    public function isOwnedThisPBX() {
        $pbxowner  = $this->get(self::USER_DATA, 'pbxowner');
        $mac       = \Models\PBX::getInstance()->getSerial();
        $curentPBX = \Wildix\PBXes\PBXes::getInstance()->getPBXBySerial($mac)->getDn();
        if ($pbxowner != $curentPBX) {
            if ($this->isPbxFailover($pbxowner, $curentPBX) == false) {
                return false;
            }
        }
        return true;
    }

    private $_failover = null;
    public function isPbxFailover($ownerDn = null, $pbxDn = null) {
        if ($this->_failover === null) {
            $this->_failover = false;
            if ($ownerDn == null) {
                $ownerDn = $this->get(self::USER_DATA, 'pbxowner');
                $mac     = \Models\PBX::getInstance()->getSerial();
                $pbxDn   = \Wildix\PBXes\PBXes::getInstance()->getPBXBySerial($mac)->getDn();
            }
            $failover = \WMS_Ldap::getInstance()->getEntries(\WMS_Ldap::FAILOVER, '(&(pbxowner=' . $ownerDn . ')(enabled=1)(pbxFailover= ' . $pbxDn . '))', array('pbxFailover'));
            if ($failover) {
                $this->_failover = true;
            }
        }
        return $this->_failover;
    }

    public function getPbxOwnerDomain() {
        $remotePBX = \WMS_Ldap::getInstance()->getEntryByDn($this->get(self::USER_DATA, 'pbxowner'), array('pbxDomain'));
        return $remotePBX['pbxDomain'];
    }
}
