<?php

/**
 * Class DhcpConfigConverter
 */
class DhcpConfigConverter {

    private $_readFromMainFile = false;

    private $_foundConfig = false;

    /**
     * Converter.
     *
     * @var \Wildix\System\Dhcp\Converter
     */
    private $_converter;

    private $_RFC_options = array(
        '1' => 'netmask',
        '2' => 'time-offset',
        '3' => 'router',
        '6' => 'dns-server',
        '7' => 'log-server',
        '9' => 'lpr-server',
        '13' => 'boot-file-size',
        '15' => 'domain-name',
        '16' => 'swap-server',
        '17' => 'root-path',
        '18' => 'extension-path',
        '19' => 'ip-forward-enable',
        '20' => 'non-local-source-routing',
        '21' => 'policy-filter',
        '22' => 'max-datagram-reassembly',
        '23' => 'default-ttl',
        '26' => 'mtu',
        '27' => 'all-subnets-local',
        '31' => 'router-discovery',
        '32' => 'router-solicitation',
        '33' => 'static-route',
        '34' => 'trailer-encapsulation',
        '35' => 'arp-timeout',
        '36' => 'ethernet-encap',
        '37' => 'tcp-ttl',
        '38' => 'tcp-keepalive',
        '40' => 'nis-domain',
        '41' => 'nis-server',
        '42' => 'ntp-server',
        '44' => 'netbios-ns', //WINS servers
        '45' => 'netbios-dd',
        '46' => 'netbios-nodetype',
        '47' => 'netbios-scope',
        '48' => 'x-windows-fs',
        '49' => 'x-windows-dm',
        '60' => 'vendor-class',
        '64' => 'nis-domain',
        '65' => 'nis-server',
        '66' => 'tftp-server',
        '67' => 'bootfile-name',
        '68' => 'mobile-ip-home',
        '69' => 'smtp-server',
        '70' => 'pop3-server',
        '71' => 'nntp-server',
        '74' => 'irc-server',
        '77' => 'user-class',
        '119' => 'domain-search',
        '120' => 'sip-server',
        '121' => 'classless-static-route',
    );

    private $_config = array(
        'general' => array(
            'interface' => 'eth0,eth1,eth2,eth3',
            'dhcp-range' => '10.135.0.10,10.135.0.99,255.255.255.0,3600m',
            'read-ethers' => '',
            'log-dhcp' => '',
            'dhcp-lease-max' => '150',
            'no-dhcp-interface' => 'eth0,eth2,eth3',
            'dhcp-leasefile' => '/rw2/etc/dnsmasq.leases',
        ),
        'options' => array(
            'domain-name' => '',
            'netbios-ns' => '',
            'dns-server' => '',
            'router' => '',
            'ntp-server' => '',
        )
    );

    /**
     * Get config
     *
     * @return array
     */
    public function getConfig(): array {
        return $this->_config;
    }

    /**
     * Is found old config to convert
     *
     * @return boolean
     */
    public function isFoundConfig() {
        return $this->_foundConfig;
    }

    /**
     * DhcpConfigConverter constructor.
     *
     * @param \Wildix\System\Dhcp\Converter $converter Converter class.
     *
     * @return void
     */
    public function __construct(\Wildix\System\Dhcp\Converter $converter) {
        $this->_converter = $converter;
        $this->_readConfigFromFile();
    }

    /**
     * Read DHCP config from old format
     *
     * @return boolean
     */
    private function _readConfigFromFile() {
        $file = \Wildix\System\Dhcp\ServerManager::CONFIG_DHCP . '_';
        if (!file_exists($file)) {
            $file = \Wildix\System\Dhcp\ServerManager::CONFIG_DHCP;

            $this->_readFromMainFile = true;
        }

        if (!file_exists($file)) {
            return false;
        }

        $data = file($file);
        if (!count($data)) {
            return false;
        }

        $this->_foundConfig = true;
        foreach ($data as $line) {
            if (strpos(trim($line), '#') === 0) {
                continue;
            }

            if (($line = trim($line)) == '') {
                continue;
            }

            list ($option, $value) = explode('=', $line);

            if ($option == 'dhcp-option' && strpos($value, 'net:') !== false) {
                continue;
            }

            if (in_array(
                $option,
                array(
                    'dhcp-host',
                    'dhcp-boot',
                    'alias',
                    'dhcp-script',
                    'dhcp-mac',
                )
            )
            ) {
                continue;
            }

            if ($option == 'dhcp-option') {
                if (preg_match('/option\:([^,]+)\,([^\s+]*)/', $value, $matches)) {
                    if (!is_array($this->_config['options'][$matches[1]])) {
                        $this->_config['options'][$matches[1]] = array();
                    }
                    if (strpos($matches[2], ',') !== false) {
                        foreach (explode(',', $matches[2]) as $val) {
                            $this->_config['options'][$matches[1]][] = $val;
                        }
                    } else {
                        $this->_config['options'][$matches[1]][] = $matches[2];
                    }
                } else {
                    list($opt,) = explode(',', $value);

                    $vals = substr($value, strlen($opt) + 1);
                    $opt  = trim($opt);

                    // convert
                    if (is_numeric($opt) && isset($this->_RFC_options[$opt])) {
                        $opt = $this->_RFC_options[$opt];
                    }

                    if (!is_array($this->_config['options'][$opt])) {
                        $this->_config['options'][$opt] = array();
                    }
                    if (strpos($vals, ',') !== false) {
                        foreach (explode(',', $vals) as $val) {
                            $this->_config['options'][$opt][] = $val;
                        }
                    } else {
                        $this->_config['options'][$opt][] = $vals;
                    }
                }
            } else {
                $this->_config['general'][$option] = $value;
            }
        }
    }

    /**
     * Convert config into new format.
     *
     * @return array
     */
    public function convert() {
        $newConfig = $this->_converter->convertFileToConfig($this->_config);

        $newConfig['deny'] = $newConfig['deny'] ? \Wildix\System\Dhcp\Server::ON : \Wildix\System\Dhcp\Server::OFF;

        $currentInterface = $this->_getCurrentInterfaceFromConfigFile();

        if (isset($currentInterface['interface'])) {
            $newConfig['currentInterface'] = $currentInterface['interface'];
            $tmp = \Wildix\System\Network::generateNetwork($newConfig['start'], $newConfig['networkMask']);

            $newConfig['ipAddress'] = $tmp['network'];

            $newConfig['status'] = $this->_readFromMainFile ? \Wildix\System\Dhcp\Server::ON : \Wildix\System\Dhcp\Server::OFF;
        } else {
            $newConfig['currentInterface'] = 'eth1';
            $newConfig['status']           = \Wildix\System\Dhcp\Server::OFF;
        }

        if ($newConfig['start'] && $newConfig['end']) {
            $newConfig['ntpServerPrimary']   = $this->_convertDomainNameToIp($newConfig['ntpServerPrimary'], $newConfig['interfaceIP']);
            $newConfig['ntpServerSecondary'] = $this->_convertDomainNameToIp($newConfig['ntpServerSecondary'], $newConfig['interfaceIP']);
        }

        return $newConfig;
    }

    /**
     * Get current interface from config file
     *
     * @return array
     */
    private function _getCurrentInterfaceFromConfigFile() {
        $output = [];

        $data = file(\Wildix\System\Dhcp\ServerManager::CONFIG_DNSMASQ);
        foreach ($data as $row) {
            if (strpos($row, 'DMASQINTERFACE=') === 0) {
                $output['interface'] = trim(str_replace('DMASQINTERFACE=', '', $row));
            }
            if (strpos($row, 'DMASQADDR=') === 0) {
                list(
                    $output['address'],
                    $output['netmask']
                    ) = explode('/', trim(str_replace('DMASQADDR=', '', $row)));
            }
        }

        return $output;
    }

    /**
     * Convert domain name to IP
     *
     * @param string $server      Domain name.
     * @param string $interfaceIP Interface IP.
     *
     * @return string
     */
    private function _convertDomainNameToIp($server, $interfaceIP) {
        if (!$server) {
            return '';
        }

        if (!\Wildix\System\Network::isIp($server)) {
            $ip = gethostbyname($server);
            // fallback: if we can not resolve the host will substitute it with PBX ip address
            if (!\Wildix\System\Network::isIp($ip)) {
                $ip = $interfaceIP;
            }
            $server = $ip;
        }

        return $server;
    }
}
