/* LdapRadiusSettings module */
(function universalModuleDefinition(root, factory){
    if (typeof exports === 'object'){
        // CommonJS
        module.exports = factory(require('jquery'));
    } else if (typeof define === 'function' && define.amd){
        // AMD
        define(['jquery'], factory);
    } else {
        // Browser
        root.LdapRadiusSettings = factory(jQuery);
    }
}(this, function ($) {
    'use strict';
    /**
     * @param {array} data Form data.
     * @return {undefined}
     */
    function LdapRadiusSettings(data) {
        this.constructor(data);
    }

    $.extend(LdapRadiusSettings.prototype, {
        constructor: function (data) {
            this._data = data;

            this.mainForm = $('#ldap_net_settings_form');
            this.mainForm.find('#ldap_radius_settings_save').click($.proxy(this.save, this));
            this.mainForm.find('#ldap_connection_enable').change($.proxy(this._onChangeEnable, this));

            this.mainForm.find('#ldap_connection_pass').passwordToggler({
                textShowPassword: WMS._t('strings.Show'),
                textHidePassword: WMS._t('strings.Hide')
            });

            this.mainForm.checkForm({
                rules: [ // rule, element_id, message
                    'required,ldap_connection_pass,' + WMS._t('rules.CannotBeEmpty')
                ]
            });

            $(window).bind('beforeunload', $.proxy(function () {
                if (this.isDirty()) {
                    return WMS._t('prefs.are_you_sure_want_to_exit_without_save');
                }
                return undefined;
            }, this));
        },

        load: function () {
            if (this._data) {
                this.setData(this._data);
            }
            this._onChangeEnable();
            this._setStrongPassword();
            $('#ldap_connection_pass').change();
            return this;
        },

        validate: function () {
            if ($('#ldap_connection_enable').is(':checked')) {
                return this.mainForm.checkForm('validate');
            }
            return true;
        },

        getData: function () {
            var data = {};
            var elements = [
                'ldap_connection_enable',
                'ldap_connection_pass'
            ];
            elements.forEach(function (name) {
                var value = $('#' + name).val();
                if (['ldap_connection_enable'].indexOf(name) > -1) {
                    value = $('#' + name).is(':checked') ? 'on' : 'off';
                }
                data[name.split('ldap_')[1]] = value;
            });
            return data;
        },

        setData: function (data) {
            $.each(data, $.proxy(function (name, value) {
                if (Object.isArray(value)) {
                    value = value.join('\n');
                }
                if (['connection_enable'].indexOf(name) > -1) {
                    if (this.mainForm.find('#ldap_' + name).prop('checked') != (value === 'on')) {
                        this.mainForm.find('#ldap_' + name).prop('checked', value === 'on');
                    }
                } else {
                    if (this.mainForm.find('#ldap_' + name).val() != value) {
                        this.mainForm.find('#ldap_' + name).val(value);
                    }
                }
            }, this));
        },

        isDirty: function () {
            var result = false;
            $.each(this.getData(), $.proxy(function (name, value) {
                if (Object.isArray(value)) {
                    value = value.join('\n');
                }
                var dataValue = this._data[name];
                if (Object.isArray(dataValue)) {
                    dataValue = dataValue.join('\n');
                }
                if (dataValue !== value) {
                    result = true;
                    return false;
                }
                return undefined;
            }, this));
            return result;
        },

        save: function () {
            if (!this.validate()) {
                return;
            }

            ap_showWaitMessage('waitDiv', 1);
            $.ajax({
                url: 'api/v1/PBX/settings/ldapRadius/',
                type: 'POST',
                dataType: 'json',
                data: {
                    data: this.getData(),
                    _METHOD: 'PUT'
                },
                success: $.proxy(function (response) {
                    ap_showWaitMessage('waitDiv', 1);
                    $.ajax({
                        url: 'api/v1/PBX/settings/ldapRadius/',
                        type: 'GET',
                        cache: false,
                        data: {
                            force: true
                        },
                        dataType: 'json',
                        success: $.proxy(function (response2) {
                            if (response2 && response2.result) {
                                if (response2.result != this._data) {
                                    this._data = response2.result;
                                    this.load();
                                } else {
                                    this._data = response2.result;
                                }
                            } else {
                                this._data = this.getData();
                            }
                        }, this),
                        complete: function () {
                            ap_showWaitMessage('waitDiv', 0);
                        }
                    });
                }, this),
                error: $.proxy(function (jqXHR) {
                    var response = JSON.parse(jqXHR.responseText);
                    if (jqXHR.status == 400 && response.code == 'BadRequest') {
                        if (response.errors) {
                            var inputs = {};
                            $.each(response.errors, function (index, value) {
                                var dataEl = this.mainForm.find('#ldap_' + index);
                                if (dataEl.length) {
                                    inputs[index] = {el: dataEl[0], error: value.reason};
                                    dataEl.addClass('wms-error-field').tooltip({
                                        tooltipClass: 'wms-error-bubble',
                                        items: 'input, select, textarea, button',
                                        position: {
                                            my: 'left top',
                                            at: 'right+10 top-7',
                                            collision: 'none'
                                        },
                                        classes: {
                                            'ui-tooltip': 'wms-error-bubble'
                                        },
                                        content: value.reason
                                    });
                                }
                            }.bind(this));

                            this.mainForm.data('wms-errors', inputs);
                            this.mainForm.find(':visible.wms-error-field').first().tooltip('open');
                        }
                    }
                }, this),
                complete: function () {
                    ap_showWaitMessage('waitDiv', 0);
                }
            });
        },

        _generateRandomString: function (length) {
            let string = '';
            for (let i = 0; i < length; i++) {
                let numI = this._getRandomNum();
                string = string + String.fromCharCode(numI);
            }
            return string;
        },

        _getRandomNum: function () {
            let rndNum = Math.random();
            rndNum = parseInt(rndNum * 1000);
            rndNum = (rndNum % 94) + 33;

            return rndNum === 38 ? this._getRandomNum() : rndNum;
        },

        _setStrongPassword: function () {
            const inputPasswordEl = $('#ldap_connection_pass');
            if (!inputPasswordEl.val()) {
                inputPasswordEl.val(this._generateRandomString(24));
            }
        },

        _onChangeEnable: function () {
            if(this.mainForm.find('#ldap_connection_enable').prop('checked')){
                this.mainForm.find('#ldap_connection_enable').val(1);
                this.mainForm.find('input[type=text], input[type=password]').each(function () {
                    $(this).prop('disabled', false);
                });
            } else {
                this.mainForm.find('#ldap_connection_enable').val(0);
                this.mainForm.find('input[type=text], input[type=password]').each(function () {
                    $(this).prop('disabled', true);
                });
            }
        }

    });

    return LdapRadiusSettings;
}));
